/* Generated by Cython 3.0.12 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "hummingbot/core/cpp/LimitOrder.h",
            "hummingbot/core/cpp/OrderBookEntry.h",
            "hummingbot/core/cpp/OrderExpirationEntry.h",
            "hummingbot/core/cpp/PyRef.h",
            "hummingbot/core/cpp/Utils.h"
        ],
        "include_dirs": [
            "hummingbot/core",
            "hummingbot/core/data_type"
        ],
        "language": "c++",
        "name": "hummingbot.connector.exchange.paper_trade.paper_trade_exchange",
        "sources": [
            "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx",
            "hummingbot/core/cpp/Utils.cpp",
            "hummingbot/core/cpp/LimitOrder.cpp",
            "hummingbot/core/cpp/OrderExpirationEntry.cpp"
        ]
    },
    "module_name": "hummingbot.connector.exchange.paper_trade.paper_trade_exchange"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"

    #if PY_MAJOR_VERSION >= 3
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj)
    #else
      #define __Pyx_PyFloat_FromString(obj)  PyFloat_FromString(obj, NULL)
    #endif
    

    #if PY_MAJOR_VERSION <= 2
    #define PyDict_GetItemWithError _PyDict_GetItemWithError
    #endif
    

    #if (PY_VERSION_HEX < 0x030700b1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM < 0x07030600)) && !defined(PyContextVar_Get)
    #define PyContextVar_Get(var, d, v)         ((d) ?             ((void)(var), Py_INCREF(d), (v)[0] = (d), 0) :             ((v)[0] = NULL, 0)         )
    #endif
    
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#if defined(CYTHON_LIMITED_API) && CYTHON_LIMITED_API
#define __PYX_EXTRA_ABI_MODULE_NAME "limited"
#else
#define __PYX_EXTRA_ABI_MODULE_NAME ""
#endif
#define CYTHON_ABI "3_0_12" __PYX_EXTRA_ABI_MODULE_NAME
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x03000CF0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(Py_GIL_DISABLED) || defined(Py_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 0
  #endif
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000 && PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(const U& other) const { return *ptr == other; }
    template<typename U> bool operator !=(const U& other) const { return *ptr != other; }
    template<typename U> bool operator==(const __Pyx_FakeReference<U>& other) const { return *ptr == *other.ptr; }
    template<typename U> bool operator!=(const __Pyx_FakeReference<U>& other) const { return *ptr != *other.ptr; }
  private:
    T *ptr;
};

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    static CYTHON_INLINE PyObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030B0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030B0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
    #ifndef CO_OPTIMIZED
    #define CO_OPTIMIZED 0x0001
    #endif
    #ifndef CO_NEWLOCALS
    #define CO_NEWLOCALS 0x0002
    #endif
    #ifndef CO_VARARGS
    #define CO_VARARGS 0x0004
    #endif
    #ifndef CO_VARKEYWORDS
    #define CO_VARKEYWORDS 0x0008
    #endif
    #ifndef CO_ASYNC_GENERATOR
    #define CO_ASYNC_GENERATOR 0x0200
    #endif
    #ifndef CO_GENERATOR
    #define CO_GENERATOR 0x0020
    #endif
    #ifndef CO_COROUTINE
    #define CO_COROUTINE 0x0080
    #endif
#elif PY_VERSION_HEX >= 0x030B0000
  static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    PyObject *empty_bytes = PyBytes_FromStringAndSize("", 0);
    if (!empty_bytes) return NULL;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, empty_bytes);
    Py_DECREF(empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_MAJOR_VERSION >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void *cfunc) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000 || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && PY_VERSION_HEX < 0x030d0000 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyObject_GenericSetAttr((PyObject*)tp, k, v)
#else
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyDict_SetItem(tp->tp_dict, k, v)
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535U : 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = (Py_UNICODE) ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static CYTHON_INLINE PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define __Pyx_Py3Int_Check(op)       PyLong_Check(op)
  #define __Pyx_Py3Int_CheckExact(op)  PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#else
  #define __Pyx_Py3Int_Check(op)       (PyLong_Check(op) || PyInt_Check(op))
  #define __Pyx_Py3Int_CheckExact(op)  (PyLong_CheckExact(op) || PyInt_CheckExact(op))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__;  (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
    #define __PYX_EXTERN_C extern "C++"
#endif

#define __PYX_HAVE__hummingbot__connector__exchange__paper_trade__paper_trade_exchange
#define __PYX_HAVE_API__hummingbot__connector__exchange__paper_trade__paper_trade_exchange
/* Early includes */
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <utility>

    #if __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600)
    // move should be defined for these versions of MSVC, but __cplusplus isn't set usefully
    #include <type_traits>

    namespace cython_std {
    template <typename T> typename std::remove_reference<T>::type&& move(T& t) noexcept { return std::move(t); }
    template <typename T> typename std::remove_reference<T>::type&& move(T&& t) noexcept { return std::move(t); }
    }

    #endif
    
#include <set>
#include <string.h>
#include <string>
#include <unordered_map>
#include "../cpp/LimitOrder.h"
#include "../cpp/OrderExpirationEntry.h"
#include <stdint.h>
#include <unordered_set>
#include "cpp/PyRef.h"
#include <vector>
#include "../cpp/OrderBookEntry.h"
#include <stdio.h>

    /* Using NumPy API declarations from "numpy/__init__.cython-30.pxd" */
    
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"
#include <stddef.h>
#include "pythread.h"
#include <iterator>
#include "cpp/Utils.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
#include <string.h>
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = (char) c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#include <string.h>
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif (defined(_Complex_I) && !defined(_MSC_VER)) || ((defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) && !defined(__STDC_NO_COMPLEX__) && !defined(_MSC_VER))
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx",
  "<stringsource>",
  "__init__.cython-30.pxd",
  "contextvars.pxd",
  "hummingbot/core/pubsub.pxd",
  "hummingbot/core/event/event_reporter.pxd",
  "hummingbot/core/event/event_logger.pxd",
  "hummingbot/core/data_type/order_book_query_result.pxd",
  "type.pxd",
  "bool.pxd",
  "complex.pxd",
  "hummingbot/core/data_type/order_book.pxd",
  "hummingbot/core/clock.pxd",
  "hummingbot/core/time_iterator.pxd",
  "hummingbot/core/network_iterator.pxd",
  "hummingbot/connector/connector_base.pxd",
  "hummingbot/core/data_type/composite_order_book.pxd",
  "hummingbot/core/data_type/limit_order.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* #### Code section: numeric_typedefs ### */

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":730
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":731
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":732
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":733
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":737
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":738
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":739
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":740
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":744
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":745
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":754
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":755
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":757
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":758
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":760
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":761
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":763
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":764
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":765
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* #### Code section: complex_type_declarations ### */
/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub;
struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener;
struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter;
struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger;
struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult;
struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult;
struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook;
struct __pyx_obj_10hummingbot_4core_5clock_Clock;
struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator;
struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator;
struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase;
struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase;
struct __pyx_obj_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook;
struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder;
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams;
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange;
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder;
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener;
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener;
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network;
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network;
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network;
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all;
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async;

/* "hummingbot/core/pubsub.pxd":10
 * from hummingbot.core.event.event_listener cimport EventListener
 * 
 * ctypedef unordered_set[PyRef] EventListenersCollection             # <<<<<<<<<<<<<<
 * ctypedef unordered_set[PyRef].iterator EventListenersIterator
 * ctypedef unordered_map[int64_t, EventListenersCollection] Events
 */
typedef std::unordered_set<PyRef>  __pyx_t_10hummingbot_4core_6pubsub_EventListenersCollection;

/* "hummingbot/core/pubsub.pxd":11
 * 
 * ctypedef unordered_set[PyRef] EventListenersCollection
 * ctypedef unordered_set[PyRef].iterator EventListenersIterator             # <<<<<<<<<<<<<<
 * ctypedef unordered_map[int64_t, EventListenersCollection] Events
 * ctypedef unordered_map[int64_t, EventListenersCollection].iterator EventsIterator
 */
typedef std::unordered_set<PyRef> ::iterator __pyx_t_10hummingbot_4core_6pubsub_EventListenersIterator;

/* "hummingbot/core/pubsub.pxd":12
 * ctypedef unordered_set[PyRef] EventListenersCollection
 * ctypedef unordered_set[PyRef].iterator EventListenersIterator
 * ctypedef unordered_map[int64_t, EventListenersCollection] Events             # <<<<<<<<<<<<<<
 * ctypedef unordered_map[int64_t, EventListenersCollection].iterator EventsIterator
 * ctypedef pair[int64_t, EventListenersCollection] EventsPair
 */
typedef std::unordered_map<int64_t,__pyx_t_10hummingbot_4core_6pubsub_EventListenersCollection>  __pyx_t_10hummingbot_4core_6pubsub_Events;

/* "hummingbot/core/pubsub.pxd":13
 * ctypedef unordered_set[PyRef].iterator EventListenersIterator
 * ctypedef unordered_map[int64_t, EventListenersCollection] Events
 * ctypedef unordered_map[int64_t, EventListenersCollection].iterator EventsIterator             # <<<<<<<<<<<<<<
 * ctypedef pair[int64_t, EventListenersCollection] EventsPair
 * 
 */
typedef std::unordered_map<int64_t,__pyx_t_10hummingbot_4core_6pubsub_EventListenersCollection> ::iterator __pyx_t_10hummingbot_4core_6pubsub_EventsIterator;

/* "hummingbot/core/pubsub.pxd":14
 * ctypedef unordered_map[int64_t, EventListenersCollection] Events
 * ctypedef unordered_map[int64_t, EventListenersCollection].iterator EventsIterator
 * ctypedef pair[int64_t, EventListenersCollection] EventsPair             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef std::pair<int64_t,__pyx_t_10hummingbot_4core_6pubsub_EventListenersCollection>  __pyx_t_10hummingbot_4core_6pubsub_EventsPair;
struct __pyx_opt_args_7cpython_11contextvars_get_value;
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default;

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value {
  int __pyx_n;
  PyObject *default_value;
};

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */
struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default {
  int __pyx_n;
  PyObject *default_value;
};

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":767
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":768
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":769
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":771
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_buy;
struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_sell;
struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_quantize_order_amount;

/* "hummingbot/connector/connector_base.pxd":21
 *         public object _client_config
 * 
 *     cdef str c_buy(self, str trading_pair, object amount, object order_type=*, object price=*, dict kwargs=*)             # <<<<<<<<<<<<<<
 *     cdef str c_sell(self, str trading_pair, object amount, object order_type=*, object price=*, dict kwargs=*)
 *     cdef c_cancel(self, str trading_pair, str client_order_id)
 */
struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_buy {
  int __pyx_n;
  PyObject *order_type;
  PyObject *price;
  PyObject *kwargs;
};

/* "hummingbot/connector/connector_base.pxd":22
 * 
 *     cdef str c_buy(self, str trading_pair, object amount, object order_type=*, object price=*, dict kwargs=*)
 *     cdef str c_sell(self, str trading_pair, object amount, object order_type=*, object price=*, dict kwargs=*)             # <<<<<<<<<<<<<<
 *     cdef c_cancel(self, str trading_pair, str client_order_id)
 *     cdef c_stop_tracking_order(self, str order_id)
 */
struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_sell {
  int __pyx_n;
  PyObject *order_type;
  PyObject *price;
  PyObject *kwargs;
};

/* "hummingbot/connector/connector_base.pxd":31
 *     cdef object c_get_order_size_quantum(self, str trading_pair, object order_size)
 *     cdef object c_quantize_order_price(self, str trading_pair, object price)
 *     cdef object c_quantize_order_amount(self, str trading_pair, object amount, object price=*)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_quantize_order_amount {
  int __pyx_n;
  PyObject *price;
};
struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_buy;
struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_sell;
struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_get_fee;

/* "hummingbot/connector/exchange_base.pxd":17
 *         object _mapping_initialization_lock
 * 
 *     cdef str c_buy(self, str trading_pair, object amount, object order_type= *, object price= *, dict kwargs= *)             # <<<<<<<<<<<<<<
 *     cdef str c_sell(self, str trading_pair, object amount, object order_type= *, object price= *, dict kwargs= *)
 *     cdef c_cancel(self, str trading_pair, str client_order_id)
 */
struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_buy {
  int __pyx_n;
  PyObject *order_type;
  PyObject *price;
  PyObject *kwargs;
};

/* "hummingbot/connector/exchange_base.pxd":18
 * 
 *     cdef str c_buy(self, str trading_pair, object amount, object order_type= *, object price= *, dict kwargs= *)
 *     cdef str c_sell(self, str trading_pair, object amount, object order_type= *, object price= *, dict kwargs= *)             # <<<<<<<<<<<<<<
 *     cdef c_cancel(self, str trading_pair, str client_order_id)
 *     cdef c_stop_tracking_order(self, str order_id)
 */
struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_sell {
  int __pyx_n;
  PyObject *order_type;
  PyObject *price;
  PyObject *kwargs;
};

/* "hummingbot/connector/exchange_base.pxd":33
 *     cdef ClientOrderBookQueryResult c_get_price_for_quote_volume(self, str trading_pair, bint is_buy, double volume)
 *     cdef ClientOrderBookQueryResult c_get_price_for_volume(self, str trading_pair, bint is_buy, object volume)
 *     cdef object c_get_fee(             # <<<<<<<<<<<<<<
 *         self,
 *         str base_currency,
 */
struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_get_fee {
  int __pyx_n;
  PyObject *is_maker;
};
struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_fee;
struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_cancel_order_from_orders_map;
struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_buy;
struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_sell;
struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_quantize_order_amount;

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pxd":12
 * 
 * 
 * ctypedef cpp_set[CPPLimitOrder] SingleTradingPairLimitOrders             # <<<<<<<<<<<<<<
 * ctypedef unordered_map[string, SingleTradingPairLimitOrders].iterator LimitOrdersIterator
 * ctypedef pair[string, SingleTradingPairLimitOrders] LimitOrdersPair
 */
typedef std::set<LimitOrder>  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders;

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pxd":13
 * 
 * ctypedef cpp_set[CPPLimitOrder] SingleTradingPairLimitOrders
 * ctypedef unordered_map[string, SingleTradingPairLimitOrders].iterator LimitOrdersIterator             # <<<<<<<<<<<<<<
 * ctypedef pair[string, SingleTradingPairLimitOrders] LimitOrdersPair
 * ctypedef unordered_map[string, SingleTradingPairLimitOrders] LimitOrders
 */
typedef std::unordered_map<std::string,__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders> ::iterator __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator;

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pxd":14
 * ctypedef cpp_set[CPPLimitOrder] SingleTradingPairLimitOrders
 * ctypedef unordered_map[string, SingleTradingPairLimitOrders].iterator LimitOrdersIterator
 * ctypedef pair[string, SingleTradingPairLimitOrders] LimitOrdersPair             # <<<<<<<<<<<<<<
 * ctypedef unordered_map[string, SingleTradingPairLimitOrders] LimitOrders
 * ctypedef cpp_set[CPPLimitOrder].iterator SingleTradingPairLimitOrdersIterator
 */
typedef std::pair<std::string,__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders>  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersPair;

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pxd":15
 * ctypedef unordered_map[string, SingleTradingPairLimitOrders].iterator LimitOrdersIterator
 * ctypedef pair[string, SingleTradingPairLimitOrders] LimitOrdersPair
 * ctypedef unordered_map[string, SingleTradingPairLimitOrders] LimitOrders             # <<<<<<<<<<<<<<
 * ctypedef cpp_set[CPPLimitOrder].iterator SingleTradingPairLimitOrdersIterator
 * ctypedef cpp_set[CPPLimitOrder].reverse_iterator SingleTradingPairLimitOrdersRIterator
 */
typedef std::unordered_map<std::string,__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders>  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders;

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pxd":16
 * ctypedef pair[string, SingleTradingPairLimitOrders] LimitOrdersPair
 * ctypedef unordered_map[string, SingleTradingPairLimitOrders] LimitOrders
 * ctypedef cpp_set[CPPLimitOrder].iterator SingleTradingPairLimitOrdersIterator             # <<<<<<<<<<<<<<
 * ctypedef cpp_set[CPPLimitOrder].reverse_iterator SingleTradingPairLimitOrdersRIterator
 * ctypedef cpp_set[CPPOrderExpirationEntry] LimitOrderExpirationSet
 */
typedef std::set<LimitOrder> ::iterator __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator;

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pxd":17
 * ctypedef unordered_map[string, SingleTradingPairLimitOrders] LimitOrders
 * ctypedef cpp_set[CPPLimitOrder].iterator SingleTradingPairLimitOrdersIterator
 * ctypedef cpp_set[CPPLimitOrder].reverse_iterator SingleTradingPairLimitOrdersRIterator             # <<<<<<<<<<<<<<
 * ctypedef cpp_set[CPPOrderExpirationEntry] LimitOrderExpirationSet
 * ctypedef cpp_set[CPPOrderExpirationEntry].iterator LimitOrderExpirationSetIterator
 */
typedef std::set<LimitOrder> ::reverse_iterator __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersRIterator;

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pxd":18
 * ctypedef cpp_set[CPPLimitOrder].iterator SingleTradingPairLimitOrdersIterator
 * ctypedef cpp_set[CPPLimitOrder].reverse_iterator SingleTradingPairLimitOrdersRIterator
 * ctypedef cpp_set[CPPOrderExpirationEntry] LimitOrderExpirationSet             # <<<<<<<<<<<<<<
 * ctypedef cpp_set[CPPOrderExpirationEntry].iterator LimitOrderExpirationSetIterator
 * 
 */
typedef std::set<OrderExpirationEntry>  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrderExpirationSet;

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pxd":19
 * ctypedef cpp_set[CPPLimitOrder].reverse_iterator SingleTradingPairLimitOrdersRIterator
 * ctypedef cpp_set[CPPOrderExpirationEntry] LimitOrderExpirationSet
 * ctypedef cpp_set[CPPOrderExpirationEntry].iterator LimitOrderExpirationSetIterator             # <<<<<<<<<<<<<<
 * 
 * cdef class QuantizationParams:
 */
typedef std::set<OrderExpirationEntry> ::iterator __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrderExpirationSetIterator;

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pxd":48
 *     cdef c_process_market_orders(self)
 *     cdef c_set_balance(self, str currency, object amount)
 *     cdef object c_get_fee(self,             # <<<<<<<<<<<<<<
 *                           str base_asset,
 *                           str quote_asset,
 */
struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_fee {
  int __pyx_n;
  PyObject *is_maker;
};

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pxd":79
 *     cdef c_process_crossed_limit_orders(self)
 *     cdef c_match_trade_to_limit_orders(self, object order_book_trade_event)
 *     cdef object c_cancel_order_from_orders_map(self,             # <<<<<<<<<<<<<<
 *                                                LimitOrders *orders_map,
 *                                                str trading_pair_str,
 */
struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_cancel_order_from_orders_map {
  int __pyx_n;
  int cancel_all;
  PyObject *client_order_id;
};

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":323
 *         self.c_process_crossed_limit_orders()
 * 
 *     cdef str c_buy(self,             # <<<<<<<<<<<<<<
 *                    str trading_pair_str,
 *                    object amount,
 */
struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_buy {
  int __pyx_n;
  PyObject *order_type;
  PyObject *price;
  PyObject *kwargs;
};

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":384
 *         return order_id
 * 
 *     cdef str c_sell(self,             # <<<<<<<<<<<<<<
 *                     str trading_pair_str,
 *                     object amount,
 */
struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_sell {
  int __pyx_n;
  PyObject *order_type;
  PyObject *price;
  PyObject *kwargs;
};

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1077
 *         return (price // price_quantum) * price_quantum
 * 
 *     cdef object c_quantize_order_amount(self,             # <<<<<<<<<<<<<<
 *                                         str trading_pair,
 *                                         object amount,
 */
struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_quantize_order_amount {
  int __pyx_n;
  PyObject *price;
};

/* "hummingbot/core/pubsub.pxd":17
 * 
 * 
 * cdef class PubSub:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Events _events
 */
struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub {
  PyObject_HEAD
  struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub *__pyx_vtab;
  __pyx_t_10hummingbot_4core_6pubsub_Events _events;
  PyObject *__weakref__;
};


/* "event_listener.pxd":6
 * 
 * 
 * cdef class EventListener:             # <<<<<<<<<<<<<<
 *     cdef:
 *         object __weakref__
 */
struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener {
  PyObject_HEAD
  struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener *__pyx_vtab;
  PyObject *__weakref__;
  int64_t _current_event_tag;
  struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *_current_event_caller;
};


/* "hummingbot/core/event/event_reporter.pxd":4
 * 
 * 
 * cdef class EventReporter(EventListener):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str event_source
 */
struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter {
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener __pyx_base;
  PyObject *event_source;
};


/* "hummingbot/core/event/event_logger.pxd":4
 * 
 * 
 * cdef class EventLogger(EventListener):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str _event_source
 */
struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger {
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener __pyx_base;
  PyObject *_event_source;
  PyObject *_logged_events;
  PyObject *_generic_logged_events;
  PyObject *_order_filled_logged_events;
  PyObject *_waiting;
  PyObject *_wait_returns;
};


/* "order_book_query_result.pxd":3
 * # distutils: language=c++
 * 
 * cdef class OrderBookQueryResult:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double query_price
 */
struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult {
  PyObject_HEAD
  double query_price;
  double query_volume;
  double result_price;
  double result_volume;
};


/* "order_book_query_result.pxd":11
 * 
 * 
 * cdef class ClientOrderBookQueryResult:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object query_price
 */
struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult {
  PyObject_HEAD
  PyObject *query_price;
  PyObject *query_volume;
  PyObject *result_price;
  PyObject *result_volume;
};


/* "hummingbot/core/data_type/order_book.pxd":12
 * 
 * 
 * cdef class OrderBook(PubSub):             # <<<<<<<<<<<<<<
 *     cdef set[OrderBookEntry] _bid_book
 *     cdef set[OrderBookEntry] _ask_book
 */
struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook {
  struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub __pyx_base;
  std::set<OrderBookEntry>  _bid_book;
  std::set<OrderBookEntry>  _ask_book;
  int64_t _snapshot_uid;
  int64_t _last_diff_uid;
  double _best_bid;
  double _best_ask;
  double _last_trade_price;
  double _last_applied_trade;
  double _last_trade_price_rest_updated;
  int _dex;
};


/* "hummingbot/core/clock.pxd":3
 * # distutils: language=c++
 * 
 * cdef class Clock:             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _clock_mode
 */
struct __pyx_obj_10hummingbot_4core_5clock_Clock {
  PyObject_HEAD
  PyObject *_clock_mode;
  double _tick_size;
  double _start_time;
  double _end_time;
  PyObject *_child_iterators;
  PyObject *_current_context;
  double _current_tick;
  int _started;
};


/* "hummingbot/core/time_iterator.pxd":7
 * 
 * 
 * cdef class TimeIterator(PubSub):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double _current_timestamp
 */
struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator {
  struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub __pyx_base;
  double _current_timestamp;
  struct __pyx_obj_10hummingbot_4core_5clock_Clock *_clock;
};


/* "hummingbot/core/network_iterator.pxd":6
 * 
 * 
 * cdef class NetworkIterator(TimeIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _network_status
 */
struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator {
  struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator __pyx_base;
  PyObject *_network_status;
  double _last_connected_timestamp;
  double _check_network_interval;
  double _check_network_timeout;
  double _network_error_wait_time;
  PyObject *_check_network_task;
};


/* "hummingbot/connector/connector_base.pxd":5
 * from hummingbot.core.network_iterator cimport NetworkIterator
 * 
 * cdef class ConnectorBase(NetworkIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         EventReporter _event_reporter
 */
struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase {
  struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator __pyx_base;
  struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter *_event_reporter;
  struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger *_event_logger;
  int _trading_required;
  PyObject *_account_available_balances;
  PyObject *_account_balances;
  int _real_time_balance_update;
  PyObject *_in_flight_orders_snapshot;
  double _in_flight_orders_snapshot_timestamp;
  PyObject *_current_trade_fills;
  PyObject *_exchange_order_ids;
  PyObject *_trade_fee_schema;
  PyObject *_trade_volume_metric_collector;
  PyObject *_client_config;
};


/* "hummingbot/connector/exchange_base.pxd":10
 * )
 * 
 * cdef class ExchangeBase(ConnectorBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _order_book_tracker
 */
struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase {
  struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase __pyx_base;
  PyObject *_order_book_tracker;
  PyObject *_budget_checker;
  PyObject *_trading_pair_symbol_map;
  PyObject *_mapping_initialization_lock;
};


/* "hummingbot/core/data_type/composite_order_book.pxd":4
 * from hummingbot.core.data_type.order_book cimport OrderBook
 * 
 * cdef class CompositeOrderBook(OrderBook):             # <<<<<<<<<<<<<<
 *     cdef:
 *         OrderBook _traded_order_book
 */
struct __pyx_obj_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook {
  struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook __pyx_base;
  struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *_traded_order_book;
};


/* "hummingbot/core/data_type/limit_order.pxd":6
 * 
 * 
 * cdef class LimitOrder:             # <<<<<<<<<<<<<<
 *     cdef:
 *         CPPLimitOrder _cpp_limit_order
 */
struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder {
  PyObject_HEAD
  struct __pyx_vtabstruct_10hummingbot_4core_9data_type_11limit_order_LimitOrder *__pyx_vtab;
  LimitOrder _cpp_limit_order;
};


/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pxd":21
 * ctypedef cpp_set[CPPOrderExpirationEntry].iterator LimitOrderExpirationSetIterator
 * 
 * cdef class QuantizationParams:             # <<<<<<<<<<<<<<
 *     cdef:
 *         str trading_pair
 */
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams {
  PyObject_HEAD
  PyObject *trading_pair;
  int price_precision;
  int price_decimals;
  int order_size_precision;
  int order_size_decimals;
};


/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pxd":30
 * 
 * 
 * cdef class PaperTradeExchange(ExchangeBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         LimitOrders _bid_limit_orders
 */
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase __pyx_base;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders _bid_limit_orders;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders _ask_limit_orders;
  int _paper_trade_market_initialized;
  PyObject *_trading_pairs;
  PyObject *_queued_orders;
  PyObject *_quantization_params;
  PyObject *_order_book_trade_listener;
  PyObject *_market_order_filled_listener;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrderExpirationSet _limit_order_expiration_set;
  PyObject *_target_market;
  PyObject *_exchange_name;
};


/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":73
 * 
 * 
 * cdef class QueuedOrder:             # <<<<<<<<<<<<<<
 *     cdef:
 *         double create_timestamp
 */
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder {
  PyObject_HEAD
  double create_timestamp;
  PyObject *_order_id;
  int _is_buy;
  PyObject *_trading_pair;
  PyObject *_amount;
};


/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":113
 * 
 * 
 * cdef class OrderBookTradeListener(EventListener):             # <<<<<<<<<<<<<<
 *     cdef:
 *         ExchangeBase _market
 */
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener {
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener __pyx_base;
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *_market;
};


/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":127
 *             self.logger().error("Error call trade listener.", exc_info=True)
 * 
 * cdef class OrderBookMarketOrderFillListener(EventListener):             # <<<<<<<<<<<<<<
 *     cdef:
 *         ExchangeBase _market
 */
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener {
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener __pyx_base;
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *_market;
};


/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":299
 *         ExchangeBase.c_start(self, clock, timestamp)
 * 
 *     async def start_network(self):             # <<<<<<<<<<<<<<
 *         await self.stop_network()
 *         self.order_book_tracker.start()
 */
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network {
  PyObject_HEAD
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self;
};


/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":303
 *         self.order_book_tracker.start()
 * 
 *     async def stop_network(self):             # <<<<<<<<<<<<<<
 *         self.order_book_tracker.stop()
 * 
 */
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network {
  PyObject_HEAD
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self;
};


/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":306
 *         self.order_book_tracker.stop()
 * 
 *     async def check_network(self) -> NetworkStatus:             # <<<<<<<<<<<<<<
 *         return NetworkStatus.CONNECTED
 * 
 */
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network {
  PyObject_HEAD
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self;
};


/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":945
 *         return self.available_balances.get(currency.upper(), s_decimal_0)
 * 
 *     async def cancel_all(self, timeout_seconds: float) -> List[CancellationResult]:             # <<<<<<<<<<<<<<
 *         cdef:
 *             LimitOrders *limit_orders_map_ptr
 */
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all {
  PyObject_HEAD
  PyObject *__pyx_v_cancellation_results;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *__pyx_v_limit_orders_map_ptr;
  PyObject *__pyx_v_results;
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self;
  PyObject *__pyx_v_timeout_seconds;
  PyObject *__pyx_v_trading_pair_str;
};


/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1134
 *         return OrderType.LIMIT
 * 
 *     async def trigger_event_async(self,             # <<<<<<<<<<<<<<
 *                                   event_tag,
 *                                   event):
 */
struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async {
  PyObject_HEAD
  PyObject *__pyx_v_event;
  PyObject *__pyx_v_event_tag;
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self;
};



/* "hummingbot/core/pubsub.pxd":17
 * 
 * 
 * cdef class PubSub:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Events _events
 */

struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub {
  PyObject *(*c_log_exception)(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *, int64_t, PyObject *);
  PyObject *(*c_add_listener)(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *, int64_t, struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *);
  PyObject *(*c_remove_listener)(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *, int64_t, struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *);
  PyObject *(*c_remove_dead_listeners)(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *, int64_t);
  PyObject *(*c_get_listeners)(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *, int64_t);
  PyObject *(*c_trigger_event)(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *, int64_t, PyObject *);
};
static struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub *__pyx_vtabptr_10hummingbot_4core_6pubsub_PubSub;


/* "event_listener.pxd":6
 * 
 * 
 * cdef class EventListener:             # <<<<<<<<<<<<<<
 *     cdef:
 *         object __weakref__
 */

struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener {
  PyObject *(*c_set_event_info)(struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *, int64_t, struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *);
  PyObject *(*c_call)(struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *, PyObject *);
};
static struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener *__pyx_vtabptr_10hummingbot_4core_5event_14event_listener_EventListener;


/* "hummingbot/core/event/event_reporter.pxd":4
 * 
 * 
 * cdef class EventReporter(EventListener):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str event_source
 */

struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_reporter_EventReporter {
  struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener __pyx_base;
};
static struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_reporter_EventReporter *__pyx_vtabptr_10hummingbot_4core_5event_14event_reporter_EventReporter;


/* "hummingbot/core/event/event_logger.pxd":4
 * 
 * 
 * cdef class EventLogger(EventListener):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str _event_source
 */

struct __pyx_vtabstruct_10hummingbot_4core_5event_12event_logger_EventLogger {
  struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener __pyx_base;
};
static struct __pyx_vtabstruct_10hummingbot_4core_5event_12event_logger_EventLogger *__pyx_vtabptr_10hummingbot_4core_5event_12event_logger_EventLogger;


/* "hummingbot/core/data_type/order_book.pxd":12
 * 
 * 
 * cdef class OrderBook(PubSub):             # <<<<<<<<<<<<<<
 *     cdef set[OrderBookEntry] _bid_book
 *     cdef set[OrderBookEntry] _ask_book
 */

struct __pyx_vtabstruct_10hummingbot_4core_9data_type_10order_book_OrderBook {
  struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub __pyx_base;
  PyObject *(*c_apply_diffs)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, std::vector<OrderBookEntry> , std::vector<OrderBookEntry> , int64_t);
  PyObject *(*c_apply_snapshot)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, std::vector<OrderBookEntry> , std::vector<OrderBookEntry> , int64_t);
  PyObject *(*c_apply_trade)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, PyObject *);
  PyObject *(*c_apply_numpy_diffs)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, PyArrayObject *, PyArrayObject *);
  PyObject *(*c_apply_numpy_snapshot)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, PyArrayObject *, PyArrayObject *);
  double (*c_get_price)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult *(*c_get_price_for_volume)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int, double);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult *(*c_get_price_for_quote_volume)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int, double);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult *(*c_get_volume_for_price)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int, double);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult *(*c_get_quote_volume_for_price)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int, double);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult *(*c_get_vwap_for_volume)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int, double);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult *(*c_get_quote_volume_for_base_amount)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int, double);
};
static struct __pyx_vtabstruct_10hummingbot_4core_9data_type_10order_book_OrderBook *__pyx_vtabptr_10hummingbot_4core_9data_type_10order_book_OrderBook;


/* "hummingbot/core/time_iterator.pxd":7
 * 
 * 
 * cdef class TimeIterator(PubSub):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double _current_timestamp
 */

struct __pyx_vtabstruct_10hummingbot_4core_13time_iterator_TimeIterator {
  struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub __pyx_base;
  PyObject *(*c_start)(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *, struct __pyx_obj_10hummingbot_4core_5clock_Clock *, double);
  PyObject *(*c_stop)(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *, struct __pyx_obj_10hummingbot_4core_5clock_Clock *);
  PyObject *(*c_tick)(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *, double);
};
static struct __pyx_vtabstruct_10hummingbot_4core_13time_iterator_TimeIterator *__pyx_vtabptr_10hummingbot_4core_13time_iterator_TimeIterator;


/* "hummingbot/core/network_iterator.pxd":6
 * 
 * 
 * cdef class NetworkIterator(TimeIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _network_status
 */

struct __pyx_vtabstruct_10hummingbot_4core_16network_iterator_NetworkIterator {
  struct __pyx_vtabstruct_10hummingbot_4core_13time_iterator_TimeIterator __pyx_base;
};
static struct __pyx_vtabstruct_10hummingbot_4core_16network_iterator_NetworkIterator *__pyx_vtabptr_10hummingbot_4core_16network_iterator_NetworkIterator;


/* "hummingbot/connector/connector_base.pxd":5
 * from hummingbot.core.network_iterator cimport NetworkIterator
 * 
 * cdef class ConnectorBase(NetworkIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         EventReporter _event_reporter
 */

struct __pyx_vtabstruct_10hummingbot_9connector_14connector_base_ConnectorBase {
  struct __pyx_vtabstruct_10hummingbot_4core_16network_iterator_NetworkIterator __pyx_base;
  PyObject *(*c_buy)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_buy *__pyx_optional_args);
  PyObject *(*c_sell)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_sell *__pyx_optional_args);
  PyObject *(*c_cancel)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *);
  PyObject *(*c_stop_tracking_order)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *);
  PyObject *(*c_get_balance)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *);
  PyObject *(*c_get_available_balance)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *);
  PyObject *(*c_get_price)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, int);
  PyObject *(*c_get_order_price_quantum)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *);
  PyObject *(*c_get_order_size_quantum)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *);
  PyObject *(*c_quantize_order_price)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *);
  PyObject *(*c_quantize_order_amount)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_quantize_order_amount *__pyx_optional_args);
};
static struct __pyx_vtabstruct_10hummingbot_9connector_14connector_base_ConnectorBase *__pyx_vtabptr_10hummingbot_9connector_14connector_base_ConnectorBase;


/* "hummingbot/connector/exchange_base.pxd":10
 * )
 * 
 * cdef class ExchangeBase(ConnectorBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _order_book_tracker
 */

struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase {
  struct __pyx_vtabstruct_10hummingbot_9connector_14connector_base_ConnectorBase __pyx_base;
  struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *(*c_get_order_book)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult *(*c_get_quote_volume_for_base_amount)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, int, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult *(*c_get_volume_for_price)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, int, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult *(*c_get_quote_volume_for_price)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, int, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult *(*c_get_vwap_for_volume)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, int, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult *(*c_get_price_for_quote_volume)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, int, double);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult *(*c_get_price_for_volume)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, int, PyObject *);
  PyObject *(*c_get_fee)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_get_fee *__pyx_optional_args);
};
static struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_vtabptr_10hummingbot_9connector_13exchange_base_ExchangeBase;


/* "hummingbot/core/data_type/composite_order_book.pxd":4
 * from hummingbot.core.data_type.order_book cimport OrderBook
 * 
 * cdef class CompositeOrderBook(OrderBook):             # <<<<<<<<<<<<<<
 *     cdef:
 *         OrderBook _traded_order_book
 */

struct __pyx_vtabstruct_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook {
  struct __pyx_vtabstruct_10hummingbot_4core_9data_type_10order_book_OrderBook __pyx_base;
};
static struct __pyx_vtabstruct_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook *__pyx_vtabptr_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook;


/* "hummingbot/core/data_type/limit_order.pxd":6
 * 
 * 
 * cdef class LimitOrder:             # <<<<<<<<<<<<<<
 *     cdef:
 *         CPPLimitOrder _cpp_limit_order
 */

struct __pyx_vtabstruct_10hummingbot_4core_9data_type_11limit_order_LimitOrder {
  PY_LONG_LONG (*c_age)(struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *);
  PY_LONG_LONG (*c_age_til)(struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *, PY_LONG_LONG);
};
static struct __pyx_vtabstruct_10hummingbot_4core_9data_type_11limit_order_LimitOrder *__pyx_vtabptr_10hummingbot_4core_9data_type_11limit_order_LimitOrder;


/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":143
 * 
 * 
 * cdef class PaperTradeExchange(ExchangeBase):             # <<<<<<<<<<<<<<
 *     TRADE_EXECUTION_DELAY = 5.0
 *     ORDER_FILLED_EVENT_TAG = MarketEvent.OrderFilled.value
 */

struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange {
  struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase __pyx_base;
  PyObject *(*c_execute_buy)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, PyObject *, PyObject *, PyObject *);
  PyObject *(*c_execute_sell)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, PyObject *, PyObject *, PyObject *);
  PyObject *(*c_process_market_orders)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *);
  PyObject *(*c_set_balance)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, PyObject *, PyObject *);
  PyObject *(*c_delete_limit_order)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator const );
  PyObject *(*c_process_limit_order)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, int, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator);
  PyObject *(*c_process_limit_bid_order)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator);
  PyObject *(*c_process_limit_ask_order)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator);
  PyObject *(*c_process_crossed_limit_orders_for_trading_pair)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, int, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *);
  PyObject *(*c_process_crossed_limit_orders)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *);
  PyObject *(*c_match_trade_to_limit_orders)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, PyObject *);
  PyObject *(*c_cancel_order_from_orders_map)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *, PyObject *, struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_cancel_order_from_orders_map *__pyx_optional_args);
};
static struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_vtabptr_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange;


/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":113
 * 
 * 
 * cdef class OrderBookTradeListener(EventListener):             # <<<<<<<<<<<<<<
 *     cdef:
 *         ExchangeBase _market
 */

struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener {
  struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener __pyx_base;
};
static struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *__pyx_vtabptr_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener;


/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":127
 *             self.logger().error("Error call trade listener.", exc_info=True)
 * 
 * cdef class OrderBookMarketOrderFillListener(EventListener):             # <<<<<<<<<<<<<<
 *     cdef:
 *         ExchangeBase _market
 */

struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener {
  struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener __pyx_base;
};
static struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *__pyx_vtabptr_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_VARARGS(args, i) PySequence_GetItem(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#else
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GetItem(args, i)
#endif
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_NewRef_VARARGS(arg) __Pyx_NewRef(arg)
    #define __Pyx_Arg_XDECREF_VARARGS(arg) Py_XDECREF(arg)
#else
    #define __Pyx_Arg_NewRef_VARARGS(arg) arg
    #define __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
    #define __Pyx_Arg_NewRef_FASTCALL(arg) arg  /* no-op, __Pyx_Arg_FASTCALL is direct and this needs
                                                   to have the same reference counting */
    #define __Pyx_Arg_XDECREF_FASTCALL(arg)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
    #define __Pyx_Arg_NewRef_FASTCALL(arg) __Pyx_Arg_NewRef_VARARGS(arg)
    #define __Pyx_Arg_XDECREF_FASTCALL(arg) __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* PyUnicode_Unicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* BuildPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char);

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kw, const char* function_name, int kw_allowed);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *typesModule=NULL, *methodType=NULL, *result=NULL;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    typesModule = PyImport_ImportModule("types");
    if (!typesModule) return NULL;
    methodType = PyObject_GetAttrString(typesModule, "MethodType");
    Py_DECREF(typesModule);
    if (!methodType) return NULL;
    result = PyObject_CallFunctionObjArgs(methodType, func, self, NULL);
    Py_DECREF(methodType);
    return result;
}
#elif PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc);
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* UnicodeConcatInPlace.proto */
# if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
    #if CYTHON_REFNANNY
        #define __Pyx_PyUnicode_ConcatInPlace(left, right) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right, __pyx_refnanny)
    #else
        #define __Pyx_PyUnicode_ConcatInPlace(left, right) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right)
    #endif
    static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    );
#else
#define __Pyx_PyUnicode_ConcatInPlace __Pyx_PyUnicode_Concat
#endif
#define __Pyx_PyUnicode_ConcatInPlaceSafe(left, right) ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_ConcatInPlace(left, right))

/* PyObjectFormat.proto */
#if CYTHON_USE_UNICODE_WRITER
static PyObject* __Pyx_PyObject_Format(PyObject* s, PyObject* f);
#else
#define __Pyx_PyObject_Format(s, f) PyObject_Format(s, f)
#endif

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* AssertionsEnabled.proto */
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (1)
#elif CYTHON_COMPILING_IN_LIMITED_API  ||  (CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030C0000)
  static int __pyx_assertions_enabled_flag;
  #define __pyx_assertions_enabled() (__pyx_assertions_enabled_flag)
  static int __Pyx_init_assertions_enabled(void) {
    PyObject *builtins, *debug, *debug_str;
    int flag;
    builtins = PyEval_GetBuiltins();
    if (!builtins) goto bad;
    debug_str = PyUnicode_FromStringAndSize("__debug__", 9);
    if (!debug_str) goto bad;
    debug = PyObject_GetItem(builtins, debug_str);
    Py_DECREF(debug_str);
    if (!debug) goto bad;
    flag = PyObject_IsTrue(debug);
    Py_DECREF(debug);
    if (flag == -1) goto bad;
    __pyx_assertions_enabled_flag = flag;
    return 0;
  bad:
    __pyx_assertions_enabled_flag = 1;
    return -1;
  }
#else
  #define __Pyx_init_assertions_enabled()  (0)
  #define __pyx_assertions_enabled()  (!Py_OptimizeFlag)
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
struct __pyx_CoroutineObject;
typedef PyObject *(*__pyx_coroutine_body_t)(struct __pyx_CoroutineObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct __pyx_CoroutineObject {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Coroutine.proto */
#define __Pyx_Coroutine_USED
#define __Pyx_Coroutine_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_CoroutineType)
#define __Pyx_Coroutine_Check(obj) __Pyx_Coroutine_CheckExact(obj)
#define __Pyx_CoroutineAwait_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_CoroutineAwaitType)
#define __Pyx_Coroutine_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_CoroutineType, body, code, closure, name, qualname, module_name)
static int __pyx_Coroutine_init(PyObject *module);
static PyObject *__Pyx__Coroutine_await(PyObject *coroutine);
typedef struct {
    PyObject_HEAD
    PyObject *coroutine;
} __pyx_CoroutineAwaitObject;
static PyObject *__Pyx_CoroutineAwait_Close(__pyx_CoroutineAwaitObject *self, PyObject *arg);
static PyObject *__Pyx_CoroutineAwait_Throw(__pyx_CoroutineAwaitObject *self, PyObject *args);

/* GetAwaitIter.proto */
static CYTHON_INLINE PyObject *__Pyx_Coroutine_GetAwaitableIter(PyObject *o);
static PyObject *__Pyx__Coroutine_GetAwaitableIter(PyObject *o);

/* CoroutineYieldFrom.proto */
static CYTHON_INLINE PyObject* __Pyx_Coroutine_Yield_From(__pyx_CoroutineObject *gen, PyObject *source);

/* pep479.proto */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen);

/* ReturnWithStopIteration.proto */
#define __Pyx_ReturnWithStopIteration(value)\
    if (value == Py_None) PyErr_SetNone(PyExc_StopIteration); else __Pyx__ReturnWithStopIteration(value)
static void __Pyx__ReturnWithStopIteration(PyObject* value);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* MoveIfSupported.proto */
#if CYTHON_USE_CPP_STD_MOVE
  #include <utility>
  #define __PYX_STD_MOVE_IF_SUPPORTED(x) std::move(x)
#else
  #define __PYX_STD_MOVE_IF_SUPPORTED(x) x
#endif

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* pybytes_as_double.proto */
static double __Pyx_SlowPyString_AsDouble(PyObject *obj);
static double __Pyx__PyBytes_AsDouble(PyObject *obj, const char* start, Py_ssize_t length);
static CYTHON_INLINE double __Pyx_PyBytes_AsDouble(PyObject *obj) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
    as_c_string = PyBytes_AS_STRING(obj);
    size = PyBytes_GET_SIZE(obj);
#else
    if (PyBytes_AsStringAndSize(obj, &as_c_string, &size) < 0) {
        return (double)-1;
    }
#endif
    return __Pyx__PyBytes_AsDouble(obj, as_c_string, size);
}
static CYTHON_INLINE double __Pyx_PyByteArray_AsDouble(PyObject *obj) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
    as_c_string = PyByteArray_AS_STRING(obj);
    size = PyByteArray_GET_SIZE(obj);
#else
    as_c_string = PyByteArray_AsString(obj);
    if (as_c_string == NULL) {
        return (double)-1;
    }
    size = PyByteArray_Size(obj);
#endif
    return __Pyx__PyBytes_AsDouble(obj, as_c_string, size);
}

/* pyunicode_as_double.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY && CYTHON_ASSUME_SAFE_MACROS
static const char* __Pyx__PyUnicode_AsDouble_Copy(const void* data, const int kind, char* buffer, Py_ssize_t start, Py_ssize_t end) {
    int last_was_punctuation;
    Py_ssize_t i;
    last_was_punctuation = 1;
    for (i=start; i <= end; i++) {
        Py_UCS4 chr = PyUnicode_READ(kind, data, i);
        int is_punctuation = (chr == '_') | (chr == '.');
        *buffer = (char)chr;
        buffer += (chr != '_');
        if (unlikely(chr > 127)) goto parse_failure;
        if (unlikely(last_was_punctuation & is_punctuation)) goto parse_failure;
        last_was_punctuation = is_punctuation;
    }
    if (unlikely(last_was_punctuation)) goto parse_failure;
    *buffer = '\0';
    return buffer;
parse_failure:
    return NULL;
}
static double __Pyx__PyUnicode_AsDouble_inf_nan(const void* data, int kind, Py_ssize_t start, Py_ssize_t length) {
    int matches = 1;
    Py_UCS4 chr;
    Py_UCS4 sign = PyUnicode_READ(kind, data, start);
    int is_signed = (sign == '-') | (sign == '+');
    start += is_signed;
    length -= is_signed;
    switch (PyUnicode_READ(kind, data, start)) {
        #ifdef Py_NAN
        case 'n':
        case 'N':
            if (unlikely(length != 3)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+1);
            matches &= (chr == 'a') | (chr == 'A');
            chr = PyUnicode_READ(kind, data, start+2);
            matches &= (chr == 'n') | (chr == 'N');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_NAN : Py_NAN;
        #endif
        case 'i':
        case 'I':
            if (unlikely(length < 3)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+1);
            matches &= (chr == 'n') | (chr == 'N');
            chr = PyUnicode_READ(kind, data, start+2);
            matches &= (chr == 'f') | (chr == 'F');
            if (likely(length == 3 && matches))
                return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
            if (unlikely(length != 8)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+3);
            matches &= (chr == 'i') | (chr == 'I');
            chr = PyUnicode_READ(kind, data, start+4);
            matches &= (chr == 'n') | (chr == 'N');
            chr = PyUnicode_READ(kind, data, start+5);
            matches &= (chr == 'i') | (chr == 'I');
            chr = PyUnicode_READ(kind, data, start+6);
            matches &= (chr == 't') | (chr == 'T');
            chr = PyUnicode_READ(kind, data, start+7);
            matches &= (chr == 'y') | (chr == 'Y');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
        case '.': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
            break;
        default:
            goto parse_failure;
    }
    return 0.0;
parse_failure:
    return -1.0;
}
static double __Pyx_PyUnicode_AsDouble_WithSpaces(PyObject *obj) {
    double value;
    const char *last;
    char *end;
    Py_ssize_t start, length = PyUnicode_GET_LENGTH(obj);
    const int kind = PyUnicode_KIND(obj);
    const void* data = PyUnicode_DATA(obj);
    start = 0;
    while (Py_UNICODE_ISSPACE(PyUnicode_READ(kind, data, start)))
        start++;
    while (start < length - 1 && Py_UNICODE_ISSPACE(PyUnicode_READ(kind, data, length - 1)))
        length--;
    length -= start;
    if (unlikely(length <= 0)) goto fallback;
    value = __Pyx__PyUnicode_AsDouble_inf_nan(data, kind, start, length);
    if (unlikely(value == -1.0)) goto fallback;
    if (value != 0.0) return value;
    if (length < 40) {
        char number[40];
        last = __Pyx__PyUnicode_AsDouble_Copy(data, kind, number, start, start + length);
        if (unlikely(!last)) goto fallback;
        value = PyOS_string_to_double(number, &end, NULL);
    } else {
        char *number = (char*) PyMem_Malloc((length + 1) * sizeof(char));
        if (unlikely(!number)) goto fallback;
        last = __Pyx__PyUnicode_AsDouble_Copy(data, kind, number, start, start + length);
        if (unlikely(!last)) {
            PyMem_Free(number);
            goto fallback;
        }
        value = PyOS_string_to_double(number, &end, NULL);
        PyMem_Free(number);
    }
    if (likely(end == last) || (value == (double)-1 && PyErr_Occurred())) {
        return value;
    }
fallback:
    return __Pyx_SlowPyString_AsDouble(obj);
}
#endif
static CYTHON_INLINE double __Pyx_PyUnicode_AsDouble(PyObject *obj) {
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY && CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(__Pyx_PyUnicode_READY(obj) == -1))
        return (double)-1;
    if (likely(PyUnicode_IS_ASCII(obj))) {
        const char *s;
        Py_ssize_t length;
        s = PyUnicode_AsUTF8AndSize(obj, &length);
        return __Pyx__PyBytes_AsDouble(obj, s, length);
    }
    return __Pyx_PyUnicode_AsDouble_WithSpaces(obj);
#else
    return __Pyx_SlowPyString_AsDouble(obj);
#endif
}

/* pynumber_float.proto */
static CYTHON_INLINE PyObject* __Pyx__PyNumber_Float(PyObject* obj);
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : __Pyx__PyNumber_Float(x))

/* IncludeCppStringH.proto */
#include <string>

/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* decode_cpp_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_cpp_string(
         std::string cppstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    return __Pyx_decode_c_bytes(
        cppstring.data(), cppstring.size(), start, stop, encoding, errors, decode_func);
}

/* PyObject_Unicode.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))
#else
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Unicode(obj))
#endif

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* CallNextTpDealloc.proto */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear);

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetupReduce.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce(PyObject* type_obj);
#endif

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_0_12
#define __PYX_HAVE_RT_ImportType_proto_3_0_12
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_12(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_12(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_0_12 {
   __Pyx_ImportType_CheckSize_Error_3_0_12 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_0_12 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_0_12 = 2
};
static PyTypeObject *__Pyx_ImportType_3_0_12(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_12 check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* MergeVTables.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type);
#endif

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple);
#endif

/* PatchInspect.proto */
static PyObject* __Pyx_patch_inspect(PyObject* module);

/* PatchAsyncIO.proto */
static PyObject* __Pyx_patch_asyncio(PyObject* module);

/* ClassMethod.proto */
#include "descrobject.h"
CYTHON_UNUSED static PyObject* __Pyx_Method_ClassMethod(PyObject *method);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* None.proto */
#include <new>

/* CppExceptionConversion.proto */
#ifndef __Pyx_CppExn2PyErr
#include <new>
#include <typeinfo>
#include <stdexcept>
#include <ios>
static void __Pyx_CppExn2PyErr() {
  try {
    if (PyErr_Occurred())
      ; // let the latest Python exn pass through and ignore the current one
    else
      throw;
  } catch (const std::bad_alloc& exn) {
    PyErr_SetString(PyExc_MemoryError, exn.what());
  } catch (const std::bad_cast& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::bad_typeid& exn) {
    PyErr_SetString(PyExc_TypeError, exn.what());
  } catch (const std::domain_error& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::invalid_argument& exn) {
    PyErr_SetString(PyExc_ValueError, exn.what());
  } catch (const std::ios_base::failure& exn) {
    PyErr_SetString(PyExc_IOError, exn.what());
  } catch (const std::out_of_range& exn) {
    PyErr_SetString(PyExc_IndexError, exn.what());
  } catch (const std::overflow_error& exn) {
    PyErr_SetString(PyExc_OverflowError, exn.what());
  } catch (const std::range_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::underflow_error& exn) {
    PyErr_SetString(PyExc_ArithmeticError, exn.what());
  } catch (const std::exception& exn) {
    PyErr_SetString(PyExc_RuntimeError, exn.what());
  }
  catch (...)
  {
    PyErr_SetString(PyExc_RuntimeError, "Unknown exception");
  }
}
#endif

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_As_PY_LONG_LONG(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* FunctionImport.proto */
static int __Pyx_ImportFunction_3_0_12(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

/* #### Code section: module_declarations ### */
static CYTHON_INLINE PyObject *__pyx_f_5numpy_7ndarray_4base_base(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyArray_Descr *__pyx_f_5numpy_7ndarray_5descr_descr(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_7ndarray_4ndim_ndim(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_5shape_shape(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_7strides_strides(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_7ndarray_4size_size(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE char *__pyx_f_5numpy_7ndarray_4data_data(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_c_call(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *__pyx_v_self, PyObject *__pyx_v_event_object); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_c_call(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *__pyx_v_self, PyObject *__pyx_v_event_object); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_start(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, struct __pyx_obj_10hummingbot_4core_5clock_Clock *__pyx_v_clock, double __pyx_v_timestamp); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_set_balance(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_currency, PyObject *__pyx_v_balance); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_balance(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_currency); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_tick(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, double __pyx_v_timestamp); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_buy(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair_str, PyObject *__pyx_v_amount, struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_buy *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_sell(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair_str, PyObject *__pyx_v_amount, struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_sell *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_execute_buy(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_order_id, PyObject *__pyx_v_trading_pair_str, PyObject *__pyx_v_amount); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_execute_sell(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_order_id, PyObject *__pyx_v_trading_pair_str, PyObject *__pyx_v_amount); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_process_market_orders(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_delete_limit_order(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *__pyx_v_limit_orders_map_ptr, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *__pyx_v_map_it_ptr, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator const __pyx_v_orders_it); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_process_limit_bid_order(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *__pyx_v_limit_orders_map_ptr, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *__pyx_v_map_it_ptr, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator __pyx_v_orders_it); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_process_limit_ask_order(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *__pyx_v_limit_orders_map_ptr, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *__pyx_v_map_it_ptr, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator __pyx_v_orders_it); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_process_limit_order(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, int __pyx_v_is_buy, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *__pyx_v_limit_orders_map_ptr, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *__pyx_v_map_it_ptr, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator __pyx_v_orders_it); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_process_crossed_limit_orders_for_trading_pair(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, int __pyx_v_is_buy, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *__pyx_v_limit_orders_map_ptr, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *__pyx_v_map_it_ptr); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_process_crossed_limit_orders(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_match_trade_to_limit_orders(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_order_book_trade_event); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_available_balance(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_currency); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_cancel_order_from_orders_map(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *__pyx_v_orders_map, PyObject *__pyx_v_trading_pair_str, struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_cancel_order_from_orders_map *__pyx_optional_args); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_cancel(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair_str, PyObject *__pyx_v_client_order_id); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_fee(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_base_asset, PyObject *__pyx_v_quote_asset, PyObject *__pyx_v_order_type, PyObject *__pyx_v_order_side, PyObject *__pyx_v_amount, PyObject *__pyx_v_price, struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_fee *__pyx_optional_args); /* proto*/
static struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_order_book(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_order_price_quantum(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_price); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_order_size_quantum(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_order_size); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_quantize_order_price(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_price); /* proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_quantize_order_amount(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_amount, struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_quantize_order_amount *__pyx_optional_args); /* proto*/

/* Module declarations from "libcpp.utility" */

/* Module declarations from "libcpp.set" */

/* Module declarations from "libc.string" */

/* Module declarations from "libcpp.string" */

/* Module declarations from "libcpp.unordered_map" */

/* Module declarations from "libcpp" */

/* Module declarations from "hummingbot.core.data_type.LimitOrder" */

/* Module declarations from "hummingbot.core.data_type.OrderExpirationEntry" */

/* Module declarations from "libc.stdint" */

/* Module declarations from "libcpp.unordered_set" */

/* Module declarations from "hummingbot.core.PyRef" */

/* Module declarations from "hummingbot.core.pubsub" */

/* Module declarations from "hummingbot.core.event.event_listener" */

/* Module declarations from "hummingbot.core.event.event_reporter" */

/* Module declarations from "hummingbot.core.event.event_logger" */

/* Module declarations from "libcpp.vector" */

/* Module declarations from "hummingbot.core.data_type.OrderBookEntry" */

/* Module declarations from "hummingbot.core.data_type.order_book_query_result" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython.version" */

/* Module declarations from "cpython.exc" */

/* Module declarations from "cpython.module" */

/* Module declarations from "cpython.mem" */

/* Module declarations from "cpython.tuple" */

/* Module declarations from "cpython.list" */

/* Module declarations from "cpython.sequence" */

/* Module declarations from "cpython.mapping" */

/* Module declarations from "cpython.iterator" */

/* Module declarations from "cpython.number" */

/* Module declarations from "cpython.int" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.bool" */

/* Module declarations from "cpython.long" */

/* Module declarations from "cpython.float" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.complex" */

/* Module declarations from "cpython.string" */

/* Module declarations from "libc.stddef" */

/* Module declarations from "cpython.unicode" */

/* Module declarations from "cpython.pyport" */

/* Module declarations from "cpython.dict" */

/* Module declarations from "cpython.instance" */

/* Module declarations from "cpython.function" */

/* Module declarations from "cpython.method" */

/* Module declarations from "cpython.weakref" */

/* Module declarations from "cpython.getargs" */

/* Module declarations from "cpython.pythread" */

/* Module declarations from "cpython.pystate" */

/* Module declarations from "cpython.cobject" */

/* Module declarations from "cpython.oldbuffer" */

/* Module declarations from "cpython.set" */

/* Module declarations from "cpython.buffer" */

/* Module declarations from "cpython.bytes" */

/* Module declarations from "cpython.pycapsule" */

/* Module declarations from "cpython.contextvars" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "numpy" */

/* Module declarations from "numpy" */

/* Module declarations from "hummingbot.core.data_type.order_book" */

/* Module declarations from "hummingbot.core.clock" */

/* Module declarations from "hummingbot.core.time_iterator" */

/* Module declarations from "hummingbot.core.network_iterator" */

/* Module declarations from "hummingbot.connector.connector_base" */

/* Module declarations from "hummingbot.connector.exchange_base" */

/* Module declarations from "hummingbot.core.data_type.composite_order_book" */

/* Module declarations from "hummingbot.core.data_type.limit_order" */
static struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *(*__pyx_f_10hummingbot_4core_9data_type_11limit_order_c_create_limit_order_from_cpp_limit_order)(LimitOrder const ); /*proto*/

/* Module declarations from "hummingbot.core.Utils" */

/* Module declarations from "hummingbot.connector.exchange.paper_trade.paper_trade_exchange" */
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_QuantizationParams__set_state(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *, PyObject *); /*proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_QueuedOrder__set_state(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *, PyObject *); /*proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_OrderBookTradeListener__set_state(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *, PyObject *); /*proto*/
static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_OrderBookMarketOrderFillListener__set_state(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *, PyObject *); /*proto*/
static std::string __pyx_convert_string_from_py_6libcpp_6string_std__in_string(PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string(std::string const &); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string(std::string const &); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyStr_string_to_py_6libcpp_6string_std__in_string(std::string const &); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string(std::string const &); /*proto*/
static CYTHON_INLINE PyObject *__pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string(std::string const &); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "hummingbot.connector.exchange.paper_trade.paper_trade_exchange"
extern int __pyx_module_is_main_hummingbot__connector__exchange__paper_trade__paper_trade_exchange;
int __pyx_module_is_main_hummingbot__connector__exchange__paper_trade__paper_trade_exchange = 0;

/* Implementation of "hummingbot.connector.exchange.paper_trade.paper_trade_exchange" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_super;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_AssertionError;
static PyObject *__pyx_builtin_all;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_ImportError;
/* #### Code section: string_decls ### */
static const char __pyx_k_0[] = "0";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_1e[] = "1e-";
static const char __pyx_k_7g[] = "%.7g";
static const char __pyx_k_8g[] = ".8g";
static const char __pyx_k__3[] = "', ";
static const char __pyx_k__4[] = ", ";
static const char __pyx_k__5[] = ")";
static const char __pyx_k__6[] = ", '";
static const char __pyx_k__9[] = "://";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_02x[] = "02x";
static const char __pyx_k_BUY[] = "BUY";
static const char __pyx_k_NIL[] = "NIL";
static const char __pyx_k__10[] = "/";
static const char __pyx_k__11[] = "";
static const char __pyx_k__15[] = ".";
static const char __pyx_k__20[] = " ";
static const char __pyx_k__22[] = ". ";
static const char __pyx_k__26[] = "'.";
static const char __pyx_k__36[] = "*";
static const char __pyx_k__88[] = "?";
static const char __pyx_k_all[] = "all";
static const char __pyx_k_buy[] = "buy";
static const char __pyx_k_cls[] = "cls";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_str[] = "str";
static const char __pyx_k_val[] = "val";
static const char __pyx_k_1e_2[] = "1e";
static const char __pyx_k_1e_7[] = "1e-7";
static const char __pyx_k_Dict[] = "Dict";
static const char __pyx_k_List[] = "List";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_SELL[] = "SELL";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_bool[] = "bool";
static const char __pyx_k_ceil[] = "ceil";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_math[] = "math";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_sell[] = "sell";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "_time";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_utf8[] = "utf8";
static const char __pyx_k_vals[] = "vals";
static const char __pyx_k_1e_10[] = "1e-10";
static const char __pyx_k_Clock[] = "Clock";
static const char __pyx_k_LIMIT[] = "LIMIT";
static const char __pyx_k_Tuple[] = "Tuple";
static const char __pyx_k_await[] = "__await__";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_deque[] = "deque";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_event[] = "event";
static const char __pyx_k_float[] = "float";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_log10[] = "log10";
static const char __pyx_k_price[] = "price";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_ready[] = "ready";
static const char __pyx_k_sleep[] = "sleep";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_upper[] = "upper";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_MARKET[] = "MARKET";
static const char __pyx_k_amount[] = "amount";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_cancel[] = "cancel";
static const char __pyx_k_choice[] = "choice";
static const char __pyx_k_dict_2[] = "_dict";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_is_buy[] = "is_buy";
static const char __pyx_k_kwargs[] = "kwargs";
static const char __pyx_k_logger[] = "logger";
static const char __pyx_k_market[] = "market";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_random[] = "random";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_return[] = "return";
static const char __pyx_k_typing[] = "typing";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_Decimal[] = "Decimal";
static const char __pyx_k_asyncio[] = "asyncio";
static const char __pyx_k_balance[] = "balance";
static const char __pyx_k_decimal[] = "decimal";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_get_fee[] = "get_fee";
static const char __pyx_k_inspect[] = "inspect";
static const char __pyx_k_popleft[] = "popleft";
static const char __pyx_k_results[] = "results";
static const char __pyx_k_warning[] = "warning";
static const char __pyx_k_Callable[] = "Callable";
static const char __pyx_k_Optional[] = "Optional";
static const char __pyx_k_currency[] = "currency";
static const char __pyx_k_exc_info[] = "exc_info";
static const char __pyx_k_exchange[] = "exchange";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_is_maker[] = "is_maker";
static const char __pyx_k_order_id[] = "order_id";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_quantity[] = "quantity";
static const char __pyx_k_quantize[] = "quantize";
static const char __pyx_k_rounding[] = "rounding";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_CONNECTED[] = "CONNECTED";
static const char __pyx_k_OrderBook[] = "OrderBook";
static const char __pyx_k_OrderType[] = "OrderType";
static const char __pyx_k_TradeType[] = "TradeType";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_available[] = " available.";
static const char __pyx_k_event_tag[] = "event_tag";
static const char __pyx_k_get_price[] = "get_price";
static const char __pyx_k_is_finite[] = "is_finite";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_needed_vs[] = " needed vs. ";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_LimitOrder[] = "LimitOrder";
static const char __pyx_k_Not_enough[] = "Not enough ";
static const char __pyx_k_PaperTrade[] = "_PaperTrade";
static const char __pyx_k_ROUND_DOWN[] = "ROUND_DOWN";
static const char __pyx_k_TradeEvent[] = "TradeEvent";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_base_asset[] = "base_asset";
static const char __pyx_k_cancel_all[] = "cancel_all";
static const char __pyx_k_order_book[] = "order_book";
static const char __pyx_k_order_side[] = "order_side";
static const char __pyx_k_order_type[] = "order_type";
static const char __pyx_k_ptm_logger[] = "ptm_logger";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_LIMIT_MAKER[] = "LIMIT_MAKER";
static const char __pyx_k_MarketEvent[] = "MarketEvent";
static const char __pyx_k_OrderFilled[] = "OrderFilled";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_QueuedOrder[] = "QueuedOrder(";
static const char __pyx_k_TradingPair[] = "TradingPair";
static const char __pyx_k_collections[] = "collections";
static const char __pyx_k_data_source[] = "data_source";
static const char __pyx_k_defaultdict[] = "defaultdict";
static const char __pyx_k_order_books[] = "order_books";
static const char __pyx_k_quote_asset[] = "quote_asset";
static const char __pyx_k_s_decimal_0[] = "s_decimal_0";
static const char __pyx_k_set_balance[] = "set_balance";
static const char __pyx_k_status_dict[] = "status_dict";
static const char __pyx_k_ExchangeBase[] = "ExchangeBase";
static const char __pyx_k_Insufficient[] = "Insufficient ";
static const char __pyx_k_OrderFailure[] = "OrderFailure";
static const char __pyx_k_Trading_pair[] = "Trading pair '";
static const char __pyx_k_available_vs[] = " available vs. ";
static const char __pyx_k_event_object[] = "event_object";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_limit_orders[] = "limit_orders";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_simulate_buy[] = "simulate_buy";
static const char __pyx_k_stop_network[] = "stop_network";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_trading_pair[] = "trading_pair";
static const char __pyx_k_use_setstate[] = "use_setstate";
static const char __pyx_k_BudgetChecker[] = "BudgetChecker";
static const char __pyx_k_NetworkStatus[] = "NetworkStatus";
static const char __pyx_k_Optional_bool[] = "Optional[bool]";
static const char __pyx_k_QueuedOrder_2[] = "QueuedOrder";
static const char __pyx_k_TYPE_CHECKING[] = "TYPE_CHECKING";
static const char __pyx_k_Tuple_str_str[] = "Tuple[str, str]";
static const char __pyx_k_asyncio_tasks[] = "asyncio.tasks";
static const char __pyx_k_base_currency[] = "base_currency";
static const char __pyx_k_check_network[] = "check_network";
static const char __pyx_k_class_getitem[] = "__class_getitem__";
static const char __pyx_k_exchange_name[] = "exchange_name";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_simulate_sell[] = "simulate_sell";
static const char __pyx_k_start_network[] = "start_network";
static const char __pyx_k_target_market[] = "target_market";
static const char __pyx_k_AssertionError[] = "AssertionError";
static const char __pyx_k_OrderBookEvent[] = "OrderBookEvent";
static const char __pyx_k_OrderCancelled[] = "OrderCancelled";
static const char __pyx_k_OrderCandidate[] = "OrderCandidate";
static const char __pyx_k_get_order_book[] = "get_order_book";
static const char __pyx_k_price_decimals[] = "price_decimals";
static const char __pyx_k_quote_currency[] = "quote_currency";
static const char __pyx_k_BuyOrderCreated[] = "BuyOrderCreated";
static const char __pyx_k_build_trade_fee[] = "build_trade_fee";
static const char __pyx_k_client_order_id[] = "client_order_id";
static const char __pyx_k_price_precision[] = "price_precision";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_random_order_id[] = "random_order_id";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_timeout_seconds[] = "timeout_seconds";
static const char __pyx_k_Dict_str_Decimal[] = "Dict[str, Decimal]";
static const char __pyx_k_OrderBookTracker[] = "OrderBookTracker";
static const char __pyx_k_OrderFilledEvent[] = "OrderFilledEvent";
static const char __pyx_k_SellOrderCreated[] = "SellOrderCreated";
static const char __pyx_k_create_timestamp[] = "create_timestamp";
static const char __pyx_k_get_all_balances[] = "get_all_balances";
static const char __pyx_k_on_hold_balances[] = "on_hold_balances";
static const char __pyx_k_order_collateral[] = "order_collateral";
static const char __pyx_k_trading_pair_str[] = "trading_pair_str";
static const char __pyx_k_BuyOrderCompleted[] = "BuyOrderCompleted";
static const char __pyx_k_client_config_map[] = "client_config_map";
static const char __pyx_k_exchange_trade_id[] = "exchange_trade_id";
static const char __pyx_k_potential_returns[] = "potential_returns";
static const char __pyx_k_CancellationResult[] = "CancellationResult";
static const char __pyx_k_CompositeOrderBook[] = "CompositeOrderBook";
static const char __pyx_k_PaperTradeExchange[] = "PaperTradeExchange";
static const char __pyx_k_QuantizationParams[] = "QuantizationParams('";
static const char __pyx_k_SellOrderCompleted[] = "SellOrderCompleted";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_available_balances[] = "available_balances";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_order_book_tracker[] = "order_book_tracker";
static const char __pyx_k_safe_ensure_future[] = "safe_ensure_future";
static const char __pyx_k_split_trading_pair[] = "split_trading_pair";
static const char __pyx_k_ClientConfigAdapter[] = "ClientConfigAdapter";
static const char __pyx_k_OrderBookTradeEvent[] = "OrderBookTradeEvent";
static const char __pyx_k_OrderCancelledEvent[] = "OrderCancelledEvent";
static const char __pyx_k_from_total_balances[] = "from_total_balances";
static const char __pyx_k_order_size_decimals[] = "order_size_decimals";
static const char __pyx_k_record_filled_order[] = "record_filled_order";
static const char __pyx_k_trigger_event_async[] = "trigger_event_async";
static const char __pyx_k_BuyOrderCreatedEvent[] = "BuyOrderCreatedEvent";
static const char __pyx_k_QuantizationParams_2[] = "QuantizationParams";
static const char __pyx_k_cancellation_results[] = "cancellation_results";
static const char __pyx_k_get_maker_order_type[] = "get_maker_order_type";
static const char __pyx_k_get_taker_order_type[] = "get_taker_order_type";
static const char __pyx_k_init___locals_lambda[] = "__init__.<locals>.<lambda>";
static const char __pyx_k_limit_orders_map_ptr[] = "limit_orders_map_ptr";
static const char __pyx_k_order_size_precision[] = "order_size_precision";
static const char __pyx_k_DummyMetricsCollector[] = "DummyMetricsCollector";
static const char __pyx_k_Error_canceling_order[] = "Error canceling order.";
static const char __pyx_k_SellOrderCreatedEvent[] = "SellOrderCreatedEvent";
static const char __pyx_k_TRADE_EXECUTION_DELAY[] = "TRADE_EXECUTION_DELAY";
static const char __pyx_k_get_available_balance[] = "get_available_balance";
static const char __pyx_k_hummingbot_core_clock[] = "hummingbot.core.clock";
static const char __pyx_k_BuyOrderCompletedEvent[] = "BuyOrderCompletedEvent";
static const char __pyx_k_ORDER_FILLED_EVENT_TAG[] = "ORDER_FILLED_EVENT_TAG";
static const char __pyx_k_OrderBookTradeListener[] = "OrderBookTradeListener";
static const char __pyx_k_PaperTradeExchange_buy[] = "PaperTradeExchange.buy";
static const char __pyx_k_required_for_the_order[] = " required for the order.";
static const char __pyx_k_set_order_book_tracker[] = "_set_order_book_tracker";
static const char __pyx_k_List_CancellationResult[] = "List[CancellationResult]";
static const char __pyx_k_MarketOrderFailureEvent[] = "MarketOrderFailureEvent";
static const char __pyx_k_PaperTradeExchange_sell[] = "PaperTradeExchange.sell";
static const char __pyx_k_SellOrderCompletedEvent[] = "SellOrderCompletedEvent";
static const char __pyx_k_get_order_price_quantum[] = "get_order_price_quantum";
static const char __pyx_k_init_paper_trade_market[] = "init_paper_trade_market";
static const char __pyx_k_order_books_initialized[] = "order_books_initialized";
static const char __pyx_k_No_order_book_exists_for[] = "No order book exists for '";
static const char __pyx_k_pyx_unpickle_QueuedOrder[] = "__pyx_unpickle_QueuedOrder";
static const char __pyx_k_Error_call_trade_listener[] = "Error call trade listener.";
static const char __pyx_k_PaperTradeExchange_cancel[] = "PaperTradeExchange.cancel";
static const char __pyx_k_ORDER_BOOK_TRADE_EVENT_TAG[] = "ORDER_BOOK_TRADE_EVENT_TAG";
static const char __pyx_k_PaperTradeExchange_get_fee[] = "PaperTradeExchange.get_fee";
static const char __pyx_k_order_book_create_function[] = "order_book_create_function";
static const char __pyx_k_QueuedOrder___reduce_cython[] = "QueuedOrder.__reduce_cython__";
static const char __pyx_k_match_trade_to_limit_orders[] = "match_trade_to_limit_orders";
static const char __pyx_k_populate_collateral_entries[] = "populate_collateral_entries";
static const char __pyx_k_Error_executing_queued_order[] = "Error executing queued order.";
static const char __pyx_k_Error_processing_limit_order[] = "Error processing limit order.";
static const char __pyx_k_PaperTradeExchange_get_price[] = "PaperTradeExchange.get_price";
static const char __pyx_k_hummingbot_core_event_events[] = "hummingbot.core.event.events";
static const char __pyx_k_BUY_ORDER_COMPLETED_EVENT_TAG[] = "BUY_ORDER_COMPLETED_EVENT_TAG";
static const char __pyx_k_PaperTradeExchange_cancel_all[] = "PaperTradeExchange.cancel_all";
static const char __pyx_k_QueuedOrder___setstate_cython[] = "QueuedOrder.__setstate_cython__";
static const char __pyx_k_MARKET_ORDER_FAILURE_EVENT_TAG[] = "MARKET_ORDER_FAILURE_EVENT_TAG";
static const char __pyx_k_PaperTradeExchange_set_balance[] = "PaperTradeExchange.set_balance";
static const char __pyx_k_SELL_ORDER_COMPLETED_EVENT_TAG[] = "SELL_ORDER_COMPLETED_EVENT_TAG";
static const char __pyx_k_pyx_unpickle_OrderBookMarketOr[] = "__pyx_unpickle_OrderBookMarketOrderFillListener";
static const char __pyx_k_pyx_unpickle_OrderBookTradeLis[] = "__pyx_unpickle_OrderBookTradeListener";
static const char __pyx_k_pyx_unpickle_QuantizationParam[] = "__pyx_unpickle_QuantizationParams";
static const char __pyx_k_MARKET_ORDER_CANCELED_EVENT_TAG[] = "MARKET_ORDER_CANCELED_EVENT_TAG";
static const char __pyx_k_MARKET_SELL_ORDER_CREATED_EVENT[] = "MARKET_SELL_ORDER_CREATED_EVENT_TAG";
static const char __pyx_k_OrderBookTradeListener___reduce[] = "OrderBookTradeListener.__reduce_cython__";
static const char __pyx_k_PaperTradeExchange_random_order[] = "PaperTradeExchange.random_order_id";
static const char __pyx_k_PaperTradeExchange_stop_network[] = "PaperTradeExchange.stop_network";
static const char __pyx_k_balance_available_for_buy_order[] = " balance available for buy order. ";
static const char __pyx_k_balance_available_for_sell_orde[] = " balance available for sell order. ";
static const char __pyx_k_balance_to_fill_limit_buy_order[] = " balance to fill limit buy order on ";
static const char __pyx_k_balance_to_fill_limit_sell_orde[] = " balance to fill limit sell order on ";
static const char __pyx_k_does_not_existing_in_current_da[] = "' does not existing in current data set.";
static const char __pyx_k_hummingbot_client_config_config[] = "hummingbot.client.config.config_helpers";
static const char __pyx_k_hummingbot_core_data_type_limit[] = "hummingbot.core.data_type.limit_order";
static const char __pyx_k_hummingbot_core_data_type_order[] = "hummingbot.core.data_type.order_book_tracker";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_Account_balance_does_not_have_as[] = "Account balance does not have asset ";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0xa74d68e, 0x820b24c, 0x7857e48) = (order_size_decimals, order_size_precision, price_decimals, price_precision, trading_pair))";
static const char __pyx_k_MARKET_BUY_ORDER_CREATED_EVENT_T[] = "MARKET_BUY_ORDER_CREATED_EVENT_TAG";
static const char __pyx_k_OrderBookMarketOrderFillListener[] = "OrderBookMarketOrderFillListener";
static const char __pyx_k_OrderBookTradeListener___setstat[] = "OrderBookTradeListener.__setstate_cython__";
static const char __pyx_k_PaperTradeExchange___reduce_cyth[] = "PaperTradeExchange.__reduce_cython__";
static const char __pyx_k_PaperTradeExchange___setstate_cy[] = "PaperTradeExchange.__setstate_cython__";
static const char __pyx_k_PaperTradeExchange_check_network[] = "PaperTradeExchange.check_network";
static const char __pyx_k_PaperTradeExchange_get_all_balan[] = "PaperTradeExchange.get_all_balances";
static const char __pyx_k_PaperTradeExchange_get_available[] = "PaperTradeExchange.get_available_balance";
static const char __pyx_k_PaperTradeExchange_get_maker_ord[] = "PaperTradeExchange.get_maker_order_type";
static const char __pyx_k_PaperTradeExchange_get_order_boo[] = "PaperTradeExchange.get_order_book";
static const char __pyx_k_PaperTradeExchange_get_order_pri[] = "PaperTradeExchange.get_order_price_quantum";
static const char __pyx_k_PaperTradeExchange_get_taker_ord[] = "PaperTradeExchange.get_taker_order_type";
static const char __pyx_k_PaperTradeExchange_init_paper_tr[] = "PaperTradeExchange.init_paper_trade_market";
static const char __pyx_k_PaperTradeExchange_match_trade_t[] = "PaperTradeExchange.match_trade_to_limit_orders";
static const char __pyx_k_PaperTradeExchange_split_trading[] = "PaperTradeExchange.split_trading_pair";
static const char __pyx_k_PaperTradeExchange_start_network[] = "PaperTradeExchange.start_network";
static const char __pyx_k_PaperTradeExchange_trigger_event[] = "PaperTradeExchange.trigger_event_async";
static const char __pyx_k_QuantizationParams___reduce_cyth[] = "QuantizationParams.__reduce_cython__";
static const char __pyx_k_QuantizationParams___setstate_cy[] = "QuantizationParams.__setstate_cython__";
static const char __pyx_k_convert_from_exchange_trading_pa[] = "convert_from_exchange_trading_pair";
static const char __pyx_k_convert_to_exchange_trading_pair[] = "convert_to_exchange_trading_pair";
static const char __pyx_k_hummingbot_connector_budget_chec[] = "hummingbot.connector.budget_checker";
static const char __pyx_k_hummingbot_connector_connector_m[] = "hummingbot.connector.connector_metrics_collector";
static const char __pyx_k_hummingbot_connector_exchange_ba[] = "hummingbot.connector.exchange_base";
static const char __pyx_k_hummingbot_connector_exchange_pa[] = "hummingbot.connector.exchange.paper_trade.paper_trade_exchange";
static const char __pyx_k_hummingbot_core_data_type_cancel[] = "hummingbot.core.data_type.cancellation_result";
static const char __pyx_k_hummingbot_core_data_type_common[] = "hummingbot.core.data_type.common";
static const char __pyx_k_hummingbot_core_data_type_compos[] = "hummingbot.core.data_type.composite_order_book";
static const char __pyx_k_hummingbot_core_network_iterator[] = "hummingbot.core.network_iterator";
static const char __pyx_k_hummingbot_core_utils_async_util[] = "hummingbot.core.utils.async_utils";
static const char __pyx_k_hummingbot_core_utils_estimate_f[] = "hummingbot.core.utils.estimate_fee";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_order_filled_events_from_order_b[] = "order_filled_events_from_order_book_rows";
static const char __pyx_k_self__ask_limit_orders_self__bid[] = "self._ask_limit_orders,self._bid_limit_orders,self._events,self._limit_order_expiration_set cannot be converted to a Python object for pickling";
static const char __pyx_k_hummingbot_core_data_type_order_2[] = "hummingbot.core.data_type.order_candidate";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_2[] = "Incompatible checksums (0x%x vs (0xefa59de, 0x3685ee0, 0x5aefc6c) = (_amount, _is_buy, _order_id, _trading_pair, create_timestamp))";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_3[] = "Incompatible checksums (0x%x vs (0x1bae46d, 0x16e8803, 0x37d9abe) = (_current_event_caller, _current_event_tag, _market))";
static const char __pyx_k_OrderBookMarketOrderFillListener_2[] = "OrderBookMarketOrderFillListener.__reduce_cython__";
static const char __pyx_k_OrderBookMarketOrderFillListener_3[] = "OrderBookMarketOrderFillListener.__setstate_cython__";
static const char __pyx_k_hummingbot_connector_exchange_pa_2[] = "hummingbot.connector.exchange.paper_trade.trading_pair";
static const char __pyx_k_hummingbot_connector_exchange_pa_3[] = "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx";
/* #### Code section: decls ### */
static int __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams___init__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *__pyx_v_self, PyObject *__pyx_v_trading_pair, int __pyx_v_price_precision, int __pyx_v_price_decimals, int __pyx_v_order_size_precision, int __pyx_v_order_size_decimals); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_2__repr__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_4__reduce_cython__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_6__setstate_cython__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder___init__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_self, PyObject *__pyx_v_create_timestamp, PyObject *__pyx_v_order_id, PyObject *__pyx_v_is_buy, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_amount); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_9timestamp___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_8order_id___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_6is_buy___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_12trading_pair___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_6amount___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_2__repr__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_4__reduce_cython__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_6__setstate_cython__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener___init__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *__pyx_v_self, PyObject *__pyx_v_market); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_2__reduce_cython__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_4__setstate_cython__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener___init__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *__pyx_v_self, PyObject *__pyx_v_market); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_2__reduce_cython__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_4__setstate_cython__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static int __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange___init__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_client_config_map, PyObject *__pyx_v_order_book_tracker, PyObject *__pyx_v_target_market, PyObject *__pyx_v_exchange_name); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_14budget_checker___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_2random_order_id(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_order_side, PyObject *__pyx_v_trading_pair); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_4init_paper_trade_market(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_6split_trading_pair(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12trading_pair___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_13trading_pairs___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_4name___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12display_name___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_11order_books___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_11status_dict___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_5ready___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_13queued_orders___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12limit_orders___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_16on_hold_balances___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_18available_balances___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_8start_network(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_11stop_network(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_14check_network(CYTHON_UNUSED struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_17cancel_all(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_timeout_seconds); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_20get_order_price_quantum(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_price); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_22get_available_balance(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_currency); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_24get_all_balances(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_26match_trade_to_limit_orders(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_event_object); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_28set_balance(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_currency, PyObject *__pyx_v_balance); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_30get_price(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_is_buy); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_32buy(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_amount, PyObject *__pyx_v_order_type, PyObject *__pyx_v_price, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_34sell(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_amount, PyObject *__pyx_v_order_type, PyObject *__pyx_v_price, PyObject *__pyx_v_kwargs); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_36cancel(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_client_order_id); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_38get_fee(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_base_currency, PyObject *__pyx_v_quote_currency, PyObject *__pyx_v_order_type, PyObject *__pyx_v_order_side, PyObject *__pyx_v_amount, PyObject *__pyx_v_price, PyObject *__pyx_v_is_maker); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_40get_order_book(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_42get_maker_order_type(CYTHON_UNUSED struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_44get_taker_order_type(CYTHON_UNUSED struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_46trigger_event_async(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_event_tag, PyObject *__pyx_v_event); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_49__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_51__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_QuantizationParams(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_2__pyx_unpickle_QueuedOrder(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_4__pyx_unpickle_OrderBookTradeListener(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_6__pyx_unpickle_OrderBookMarketOrderFillListener(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyUnicode_Type_upper = {0, 0, 0, 0, 0};
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_6pubsub_PubSub;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult;
  PyTypeObject *__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4bool_bool;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_7complex_complex;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5numpy_dtype;
  PyTypeObject *__pyx_ptype_5numpy_flatiter;
  PyTypeObject *__pyx_ptype_5numpy_broadcast;
  PyTypeObject *__pyx_ptype_5numpy_ndarray;
  PyTypeObject *__pyx_ptype_5numpy_generic;
  PyTypeObject *__pyx_ptype_5numpy_number;
  PyTypeObject *__pyx_ptype_5numpy_integer;
  PyTypeObject *__pyx_ptype_5numpy_signedinteger;
  PyTypeObject *__pyx_ptype_5numpy_unsignedinteger;
  PyTypeObject *__pyx_ptype_5numpy_inexact;
  PyTypeObject *__pyx_ptype_5numpy_floating;
  PyTypeObject *__pyx_ptype_5numpy_complexfloating;
  PyTypeObject *__pyx_ptype_5numpy_flexible;
  PyTypeObject *__pyx_ptype_5numpy_character;
  PyTypeObject *__pyx_ptype_5numpy_ufunc;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_5clock_Clock;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  PyObject *__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams;
  PyObject *__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange;
  PyObject *__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder;
  PyObject *__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener;
  PyObject *__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener;
  PyObject *__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network;
  PyObject *__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network;
  PyObject *__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network;
  PyObject *__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all;
  PyObject *__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async;
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams;
  PyTypeObject *__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange;
  PyTypeObject *__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder;
  PyTypeObject *__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener;
  PyTypeObject *__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener;
  PyTypeObject *__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network;
  PyTypeObject *__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network;
  PyTypeObject *__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network;
  PyTypeObject *__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all;
  PyTypeObject *__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async;
  PyObject *__pyx_kp_u_0;
  PyObject *__pyx_kp_u_02x;
  PyObject *__pyx_kp_u_1e;
  PyObject *__pyx_kp_u_1e_10;
  PyObject *__pyx_kp_u_1e_2;
  PyObject *__pyx_kp_u_1e_7;
  PyObject *__pyx_kp_u_7g;
  PyObject *__pyx_kp_u_8g;
  PyObject *__pyx_kp_u_Account_balance_does_not_have_as;
  PyObject *__pyx_n_s_AssertionError;
  PyObject *__pyx_n_s_BUY;
  PyObject *__pyx_n_u_BUY;
  PyObject *__pyx_n_s_BUY_ORDER_COMPLETED_EVENT_TAG;
  PyObject *__pyx_n_s_BudgetChecker;
  PyObject *__pyx_n_s_BuyOrderCompleted;
  PyObject *__pyx_n_s_BuyOrderCompletedEvent;
  PyObject *__pyx_n_s_BuyOrderCreated;
  PyObject *__pyx_n_s_BuyOrderCreatedEvent;
  PyObject *__pyx_n_s_CONNECTED;
  PyObject *__pyx_n_s_Callable;
  PyObject *__pyx_n_s_CancellationResult;
  PyObject *__pyx_n_s_ClientConfigAdapter;
  PyObject *__pyx_n_s_Clock;
  PyObject *__pyx_n_s_CompositeOrderBook;
  PyObject *__pyx_n_s_Decimal;
  PyObject *__pyx_n_s_Dict;
  PyObject *__pyx_kp_s_Dict_str_Decimal;
  PyObject *__pyx_n_s_DummyMetricsCollector;
  PyObject *__pyx_kp_u_Error_call_trade_listener;
  PyObject *__pyx_kp_u_Error_canceling_order;
  PyObject *__pyx_kp_u_Error_executing_queued_order;
  PyObject *__pyx_kp_u_Error_processing_limit_order;
  PyObject *__pyx_n_s_ExchangeBase;
  PyObject *__pyx_n_s_ImportError;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2;
  PyObject *__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3;
  PyObject *__pyx_kp_u_Insufficient;
  PyObject *__pyx_n_s_LIMIT;
  PyObject *__pyx_n_s_LIMIT_MAKER;
  PyObject *__pyx_n_s_LimitOrder;
  PyObject *__pyx_n_s_List;
  PyObject *__pyx_kp_s_List_CancellationResult;
  PyObject *__pyx_n_s_MARKET;
  PyObject *__pyx_n_s_MARKET_BUY_ORDER_CREATED_EVENT_T;
  PyObject *__pyx_n_s_MARKET_ORDER_CANCELED_EVENT_TAG;
  PyObject *__pyx_n_s_MARKET_ORDER_FAILURE_EVENT_TAG;
  PyObject *__pyx_n_s_MARKET_SELL_ORDER_CREATED_EVENT;
  PyObject *__pyx_n_s_MarketEvent;
  PyObject *__pyx_n_s_MarketOrderFailureEvent;
  PyObject *__pyx_n_b_NIL;
  PyObject *__pyx_n_s_NetworkStatus;
  PyObject *__pyx_kp_u_No_order_book_exists_for;
  PyObject *__pyx_kp_u_None;
  PyObject *__pyx_kp_u_Not_enough;
  PyObject *__pyx_n_s_ORDER_BOOK_TRADE_EVENT_TAG;
  PyObject *__pyx_n_s_ORDER_FILLED_EVENT_TAG;
  PyObject *__pyx_n_s_Optional;
  PyObject *__pyx_kp_s_Optional_bool;
  PyObject *__pyx_n_s_OrderBook;
  PyObject *__pyx_n_s_OrderBookEvent;
  PyObject *__pyx_n_s_OrderBookMarketOrderFillListener;
  PyObject *__pyx_n_s_OrderBookMarketOrderFillListener_2;
  PyObject *__pyx_n_s_OrderBookMarketOrderFillListener_3;
  PyObject *__pyx_n_s_OrderBookTracker;
  PyObject *__pyx_n_s_OrderBookTradeEvent;
  PyObject *__pyx_n_s_OrderBookTradeListener;
  PyObject *__pyx_n_s_OrderBookTradeListener___reduce;
  PyObject *__pyx_n_s_OrderBookTradeListener___setstat;
  PyObject *__pyx_n_s_OrderCancelled;
  PyObject *__pyx_n_s_OrderCancelledEvent;
  PyObject *__pyx_n_s_OrderCandidate;
  PyObject *__pyx_n_s_OrderFailure;
  PyObject *__pyx_n_s_OrderFilled;
  PyObject *__pyx_n_s_OrderFilledEvent;
  PyObject *__pyx_n_s_OrderType;
  PyObject *__pyx_n_u_PaperTrade;
  PyObject *__pyx_n_s_PaperTradeExchange;
  PyObject *__pyx_n_s_PaperTradeExchange___reduce_cyth;
  PyObject *__pyx_n_s_PaperTradeExchange___setstate_cy;
  PyObject *__pyx_n_s_PaperTradeExchange_buy;
  PyObject *__pyx_n_s_PaperTradeExchange_cancel;
  PyObject *__pyx_n_s_PaperTradeExchange_cancel_all;
  PyObject *__pyx_n_s_PaperTradeExchange_check_network;
  PyObject *__pyx_n_s_PaperTradeExchange_get_all_balan;
  PyObject *__pyx_n_s_PaperTradeExchange_get_available;
  PyObject *__pyx_n_s_PaperTradeExchange_get_fee;
  PyObject *__pyx_n_s_PaperTradeExchange_get_maker_ord;
  PyObject *__pyx_n_s_PaperTradeExchange_get_order_boo;
  PyObject *__pyx_n_s_PaperTradeExchange_get_order_pri;
  PyObject *__pyx_n_s_PaperTradeExchange_get_price;
  PyObject *__pyx_n_s_PaperTradeExchange_get_taker_ord;
  PyObject *__pyx_n_s_PaperTradeExchange_init_paper_tr;
  PyObject *__pyx_n_s_PaperTradeExchange_match_trade_t;
  PyObject *__pyx_n_s_PaperTradeExchange_random_order;
  PyObject *__pyx_n_s_PaperTradeExchange_sell;
  PyObject *__pyx_n_s_PaperTradeExchange_set_balance;
  PyObject *__pyx_n_s_PaperTradeExchange_split_trading;
  PyObject *__pyx_n_s_PaperTradeExchange_start_network;
  PyObject *__pyx_n_s_PaperTradeExchange_stop_network;
  PyObject *__pyx_n_s_PaperTradeExchange_trigger_event;
  PyObject *__pyx_n_s_PickleError;
  PyObject *__pyx_kp_u_QuantizationParams;
  PyObject *__pyx_n_s_QuantizationParams_2;
  PyObject *__pyx_n_s_QuantizationParams___reduce_cyth;
  PyObject *__pyx_n_s_QuantizationParams___setstate_cy;
  PyObject *__pyx_kp_u_QueuedOrder;
  PyObject *__pyx_n_s_QueuedOrder_2;
  PyObject *__pyx_n_s_QueuedOrder___reduce_cython;
  PyObject *__pyx_n_s_QueuedOrder___setstate_cython;
  PyObject *__pyx_n_s_ROUND_DOWN;
  PyObject *__pyx_n_s_SELL;
  PyObject *__pyx_n_s_SELL_ORDER_COMPLETED_EVENT_TAG;
  PyObject *__pyx_n_s_SellOrderCompleted;
  PyObject *__pyx_n_s_SellOrderCompletedEvent;
  PyObject *__pyx_n_s_SellOrderCreated;
  PyObject *__pyx_n_s_SellOrderCreatedEvent;
  PyObject *__pyx_n_s_TRADE_EXECUTION_DELAY;
  PyObject *__pyx_n_s_TYPE_CHECKING;
  PyObject *__pyx_n_s_TradeEvent;
  PyObject *__pyx_n_s_TradeType;
  PyObject *__pyx_n_s_TradingPair;
  PyObject *__pyx_kp_u_Trading_pair;
  PyObject *__pyx_n_s_Tuple;
  PyObject *__pyx_kp_s_Tuple_str_str;
  PyObject *__pyx_n_s_TypeError;
  PyObject *__pyx_n_s_ValueError;
  PyObject *__pyx_kp_u__10;
  PyObject *__pyx_kp_u__11;
  PyObject *__pyx_kp_u__15;
  PyObject *__pyx_kp_u__20;
  PyObject *__pyx_kp_u__22;
  PyObject *__pyx_kp_u__26;
  PyObject *__pyx_kp_u__3;
  PyObject *__pyx_n_s__36;
  PyObject *__pyx_kp_u__4;
  PyObject *__pyx_kp_u__5;
  PyObject *__pyx_kp_u__6;
  PyObject *__pyx_n_s__88;
  PyObject *__pyx_kp_u__9;
  PyObject *__pyx_n_s_all;
  PyObject *__pyx_n_s_amount;
  PyObject *__pyx_n_s_append;
  PyObject *__pyx_n_s_args;
  PyObject *__pyx_n_s_asyncio;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_n_s_asyncio_tasks;
  PyObject *__pyx_kp_u_available;
  PyObject *__pyx_n_s_available_balances;
  PyObject *__pyx_kp_u_available_vs;
  PyObject *__pyx_n_s_await;
  PyObject *__pyx_n_s_balance;
  PyObject *__pyx_kp_u_balance_available_for_buy_order;
  PyObject *__pyx_kp_u_balance_available_for_sell_orde;
  PyObject *__pyx_kp_u_balance_to_fill_limit_buy_order;
  PyObject *__pyx_kp_u_balance_to_fill_limit_sell_orde;
  PyObject *__pyx_n_s_base_asset;
  PyObject *__pyx_n_s_base_currency;
  PyObject *__pyx_n_s_bool;
  PyObject *__pyx_n_s_build_trade_fee;
  PyObject *__pyx_n_s_buy;
  PyObject *__pyx_n_u_buy;
  PyObject *__pyx_n_s_cancel;
  PyObject *__pyx_n_s_cancel_all;
  PyObject *__pyx_n_s_cancellation_results;
  PyObject *__pyx_n_s_ceil;
  PyObject *__pyx_n_s_check_network;
  PyObject *__pyx_n_s_choice;
  PyObject *__pyx_n_s_class_getitem;
  PyObject *__pyx_n_s_client_config_map;
  PyObject *__pyx_n_s_client_order_id;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_close;
  PyObject *__pyx_n_s_cls;
  PyObject *__pyx_n_s_collections;
  PyObject *__pyx_n_s_convert_from_exchange_trading_pa;
  PyObject *__pyx_n_s_convert_to_exchange_trading_pair;
  PyObject *__pyx_n_s_create_timestamp;
  PyObject *__pyx_n_s_currency;
  PyObject *__pyx_n_s_data_source;
  PyObject *__pyx_n_s_decimal;
  PyObject *__pyx_n_s_defaultdict;
  PyObject *__pyx_n_s_deque;
  PyObject *__pyx_n_s_dict;
  PyObject *__pyx_n_s_dict_2;
  PyObject *__pyx_kp_u_disable;
  PyObject *__pyx_kp_u_does_not_existing_in_current_da;
  PyObject *__pyx_kp_u_enable;
  PyObject *__pyx_n_s_encode;
  PyObject *__pyx_n_s_error;
  PyObject *__pyx_n_s_event;
  PyObject *__pyx_n_s_event_object;
  PyObject *__pyx_n_s_event_tag;
  PyObject *__pyx_n_s_exc_info;
  PyObject *__pyx_n_s_exchange;
  PyObject *__pyx_n_s_exchange_name;
  PyObject *__pyx_n_s_exchange_trade_id;
  PyObject *__pyx_n_s_float;
  PyObject *__pyx_n_s_from_total_balances;
  PyObject *__pyx_kp_u_gc;
  PyObject *__pyx_n_s_get;
  PyObject *__pyx_n_s_get_all_balances;
  PyObject *__pyx_n_s_get_available_balance;
  PyObject *__pyx_n_s_get_fee;
  PyObject *__pyx_n_s_get_maker_order_type;
  PyObject *__pyx_n_s_get_order_book;
  PyObject *__pyx_n_s_get_order_price_quantum;
  PyObject *__pyx_n_s_get_price;
  PyObject *__pyx_n_s_get_taker_order_type;
  PyObject *__pyx_n_s_getstate;
  PyObject *__pyx_n_s_hummingbot_client_config_config;
  PyObject *__pyx_n_s_hummingbot_connector_budget_chec;
  PyObject *__pyx_n_s_hummingbot_connector_connector_m;
  PyObject *__pyx_n_s_hummingbot_connector_exchange_ba;
  PyObject *__pyx_n_s_hummingbot_connector_exchange_pa;
  PyObject *__pyx_n_s_hummingbot_connector_exchange_pa_2;
  PyObject *__pyx_kp_s_hummingbot_connector_exchange_pa_3;
  PyObject *__pyx_n_s_hummingbot_core_clock;
  PyObject *__pyx_n_s_hummingbot_core_data_type_cancel;
  PyObject *__pyx_n_s_hummingbot_core_data_type_common;
  PyObject *__pyx_n_s_hummingbot_core_data_type_compos;
  PyObject *__pyx_n_s_hummingbot_core_data_type_limit;
  PyObject *__pyx_n_s_hummingbot_core_data_type_order;
  PyObject *__pyx_n_s_hummingbot_core_data_type_order_2;
  PyObject *__pyx_n_s_hummingbot_core_event_events;
  PyObject *__pyx_n_s_hummingbot_core_network_iterator;
  PyObject *__pyx_n_s_hummingbot_core_utils_async_util;
  PyObject *__pyx_n_s_hummingbot_core_utils_estimate_f;
  PyObject *__pyx_n_s_i;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_init;
  PyObject *__pyx_n_s_init___locals_lambda;
  PyObject *__pyx_n_s_init_paper_trade_market;
  PyObject *__pyx_n_s_initializing;
  PyObject *__pyx_n_s_inspect;
  PyObject *__pyx_n_s_is_buy;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_n_s_is_finite;
  PyObject *__pyx_n_s_is_maker;
  PyObject *__pyx_kp_u_isenabled;
  PyObject *__pyx_n_s_items;
  PyObject *__pyx_n_s_keys;
  PyObject *__pyx_n_s_kwargs;
  PyObject *__pyx_n_s_limit_orders;
  PyObject *__pyx_n_s_limit_orders_map_ptr;
  PyObject *__pyx_n_s_log10;
  PyObject *__pyx_n_s_logger;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_s_market;
  PyObject *__pyx_n_s_match_trade_to_limit_orders;
  PyObject *__pyx_n_s_math;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_s_name_2;
  PyObject *__pyx_kp_u_needed_vs;
  PyObject *__pyx_n_s_new;
  PyObject *__pyx_kp_u_numpy_core_multiarray_failed_to;
  PyObject *__pyx_kp_u_numpy_core_umath_failed_to_impor;
  PyObject *__pyx_n_s_on_hold_balances;
  PyObject *__pyx_n_s_order_book;
  PyObject *__pyx_n_s_order_book_create_function;
  PyObject *__pyx_n_s_order_book_tracker;
  PyObject *__pyx_n_s_order_books;
  PyObject *__pyx_n_u_order_books_initialized;
  PyObject *__pyx_n_s_order_collateral;
  PyObject *__pyx_n_s_order_filled_events_from_order_b;
  PyObject *__pyx_n_s_order_id;
  PyObject *__pyx_n_s_order_side;
  PyObject *__pyx_n_s_order_size_decimals;
  PyObject *__pyx_n_s_order_size_precision;
  PyObject *__pyx_n_s_order_type;
  PyObject *__pyx_n_s_pickle;
  PyObject *__pyx_n_s_popleft;
  PyObject *__pyx_n_s_populate_collateral_entries;
  PyObject *__pyx_n_s_potential_returns;
  PyObject *__pyx_n_s_price;
  PyObject *__pyx_n_s_price_decimals;
  PyObject *__pyx_n_s_price_precision;
  PyObject *__pyx_n_s_ptm_logger;
  PyObject *__pyx_n_s_pyx_PickleError;
  PyObject *__pyx_n_s_pyx_checksum;
  PyObject *__pyx_n_s_pyx_result;
  PyObject *__pyx_n_s_pyx_state;
  PyObject *__pyx_n_s_pyx_type;
  PyObject *__pyx_n_s_pyx_unpickle_OrderBookMarketOr;
  PyObject *__pyx_n_s_pyx_unpickle_OrderBookTradeLis;
  PyObject *__pyx_n_s_pyx_unpickle_QuantizationParam;
  PyObject *__pyx_n_s_pyx_unpickle_QueuedOrder;
  PyObject *__pyx_n_s_pyx_vtable;
  PyObject *__pyx_n_s_quantity;
  PyObject *__pyx_n_s_quantize;
  PyObject *__pyx_n_s_quote_asset;
  PyObject *__pyx_n_s_quote_currency;
  PyObject *__pyx_n_s_random;
  PyObject *__pyx_n_s_random_order_id;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_s_ready;
  PyObject *__pyx_n_s_record_filled_order;
  PyObject *__pyx_n_s_reduce;
  PyObject *__pyx_n_s_reduce_cython;
  PyObject *__pyx_n_s_reduce_ex;
  PyObject *__pyx_kp_u_required_for_the_order;
  PyObject *__pyx_n_s_results;
  PyObject *__pyx_n_s_return;
  PyObject *__pyx_n_s_rounding;
  PyObject *__pyx_n_s_s_decimal_0;
  PyObject *__pyx_n_s_safe_ensure_future;
  PyObject *__pyx_n_s_self;
  PyObject *__pyx_kp_s_self__ask_limit_orders_self__bid;
  PyObject *__pyx_n_s_sell;
  PyObject *__pyx_n_u_sell;
  PyObject *__pyx_n_s_send;
  PyObject *__pyx_n_s_set_balance;
  PyObject *__pyx_n_s_set_order_book_tracker;
  PyObject *__pyx_n_s_setstate;
  PyObject *__pyx_n_s_setstate_cython;
  PyObject *__pyx_n_s_simulate_buy;
  PyObject *__pyx_n_s_simulate_sell;
  PyObject *__pyx_n_s_sleep;
  PyObject *__pyx_n_s_spec;
  PyObject *__pyx_n_s_split_trading_pair;
  PyObject *__pyx_n_s_start;
  PyObject *__pyx_n_s_start_network;
  PyObject *__pyx_n_s_state;
  PyObject *__pyx_n_s_status_dict;
  PyObject *__pyx_n_s_stop;
  PyObject *__pyx_n_s_stop_network;
  PyObject *__pyx_n_s_str;
  PyObject *__pyx_kp_s_stringsource;
  PyObject *__pyx_n_s_super;
  PyObject *__pyx_n_s_target_market;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_throw;
  PyObject *__pyx_n_s_time;
  PyObject *__pyx_n_s_timeout_seconds;
  PyObject *__pyx_n_s_trading_pair;
  PyObject *__pyx_n_s_trading_pair_str;
  PyObject *__pyx_n_s_trigger_event_async;
  PyObject *__pyx_n_s_type;
  PyObject *__pyx_n_s_typing;
  PyObject *__pyx_n_s_update;
  PyObject *__pyx_n_s_upper;
  PyObject *__pyx_n_s_use_setstate;
  PyObject *__pyx_n_u_utf8;
  PyObject *__pyx_n_s_val;
  PyObject *__pyx_n_s_vals;
  PyObject *__pyx_n_s_value;
  PyObject *__pyx_n_s_values;
  PyObject *__pyx_n_s_warning;
  PyObject *__pyx_float_0_0;
  PyObject *__pyx_float_1e6;
  PyObject *__pyx_float_5_0;
  PyObject *__pyx_float_0_01;
  PyObject *__pyx_float_1eneg_7;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_256;
  PyObject *__pyx_int_24020995;
  PyObject *__pyx_int_29025389;
  PyObject *__pyx_int_57171680;
  PyObject *__pyx_int_58563262;
  PyObject *__pyx_int_95353964;
  PyObject *__pyx_int_126189128;
  PyObject *__pyx_int_136360524;
  PyObject *__pyx_int_175429262;
  PyObject *__pyx_int_251288030;
  PyObject *__pyx_k__16;
  PyObject *__pyx_k__17;
  PyObject *__pyx_k__18;
  PyObject *__pyx_k__19;
  PyObject *__pyx_k__27;
  PyObject *__pyx_k__28;
  PyObject *__pyx_k__29;
  PyObject *__pyx_k__30;
  PyObject *__pyx_k__31;
  PyObject *__pyx_k__55;
  PyObject *__pyx_k__56;
  PyObject *__pyx_k__60;
  PyObject *__pyx_tuple_;
  PyObject *__pyx_tuple__2;
  PyObject *__pyx_tuple__7;
  PyObject *__pyx_tuple__8;
  PyObject *__pyx_tuple__21;
  PyObject *__pyx_tuple__23;
  PyObject *__pyx_tuple__25;
  PyObject *__pyx_tuple__33;
  PyObject *__pyx_tuple__34;
  PyObject *__pyx_tuple__35;
  PyObject *__pyx_tuple__37;
  PyObject *__pyx_tuple__38;
  PyObject *__pyx_tuple__40;
  PyObject *__pyx_tuple__48;
  PyObject *__pyx_tuple__50;
  PyObject *__pyx_tuple__52;
  PyObject *__pyx_tuple__54;
  PyObject *__pyx_tuple__57;
  PyObject *__pyx_tuple__58;
  PyObject *__pyx_tuple__61;
  PyObject *__pyx_tuple__64;
  PyObject *__pyx_tuple__66;
  PyObject *__pyx_tuple__68;
  PyObject *__pyx_tuple__70;
  PyObject *__pyx_tuple__73;
  PyObject *__pyx_tuple__75;
  PyObject *__pyx_tuple__80;
  PyObject *__pyx_tuple__83;
  PyObject *__pyx_codeobj__12;
  PyObject *__pyx_codeobj__13;
  PyObject *__pyx_codeobj__14;
  PyObject *__pyx_codeobj__24;
  PyObject *__pyx_codeobj__32;
  PyObject *__pyx_codeobj__39;
  PyObject *__pyx_codeobj__41;
  PyObject *__pyx_codeobj__42;
  PyObject *__pyx_codeobj__43;
  PyObject *__pyx_codeobj__44;
  PyObject *__pyx_codeobj__45;
  PyObject *__pyx_codeobj__46;
  PyObject *__pyx_codeobj__47;
  PyObject *__pyx_codeobj__49;
  PyObject *__pyx_codeobj__51;
  PyObject *__pyx_codeobj__53;
  PyObject *__pyx_codeobj__59;
  PyObject *__pyx_codeobj__62;
  PyObject *__pyx_codeobj__63;
  PyObject *__pyx_codeobj__65;
  PyObject *__pyx_codeobj__67;
  PyObject *__pyx_codeobj__69;
  PyObject *__pyx_codeobj__71;
  PyObject *__pyx_codeobj__72;
  PyObject *__pyx_codeobj__74;
  PyObject *__pyx_codeobj__76;
  PyObject *__pyx_codeobj__77;
  PyObject *__pyx_codeobj__78;
  PyObject *__pyx_codeobj__79;
  PyObject *__pyx_codeobj__81;
  PyObject *__pyx_codeobj__82;
  PyObject *__pyx_codeobj__84;
  PyObject *__pyx_codeobj__85;
  PyObject *__pyx_codeobj__86;
  PyObject *__pyx_codeobj__87;
} __pyx_mstate;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#else
static __pyx_mstate __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstate *__pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_6pubsub_PubSub);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_dtype);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_flatiter);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_broadcast);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_ndarray);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_generic);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_number);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_integer);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_signedinteger);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_unsignedinteger);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_inexact);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_floating);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_complexfloating);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_flexible);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_character);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_ufunc);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_5clock_Clock);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async);
  Py_CLEAR(clear_module_state->__pyx_kp_u_0);
  Py_CLEAR(clear_module_state->__pyx_kp_u_02x);
  Py_CLEAR(clear_module_state->__pyx_kp_u_1e);
  Py_CLEAR(clear_module_state->__pyx_kp_u_1e_10);
  Py_CLEAR(clear_module_state->__pyx_kp_u_1e_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_1e_7);
  Py_CLEAR(clear_module_state->__pyx_kp_u_7g);
  Py_CLEAR(clear_module_state->__pyx_kp_u_8g);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Account_balance_does_not_have_as);
  Py_CLEAR(clear_module_state->__pyx_n_s_AssertionError);
  Py_CLEAR(clear_module_state->__pyx_n_s_BUY);
  Py_CLEAR(clear_module_state->__pyx_n_u_BUY);
  Py_CLEAR(clear_module_state->__pyx_n_s_BUY_ORDER_COMPLETED_EVENT_TAG);
  Py_CLEAR(clear_module_state->__pyx_n_s_BudgetChecker);
  Py_CLEAR(clear_module_state->__pyx_n_s_BuyOrderCompleted);
  Py_CLEAR(clear_module_state->__pyx_n_s_BuyOrderCompletedEvent);
  Py_CLEAR(clear_module_state->__pyx_n_s_BuyOrderCreated);
  Py_CLEAR(clear_module_state->__pyx_n_s_BuyOrderCreatedEvent);
  Py_CLEAR(clear_module_state->__pyx_n_s_CONNECTED);
  Py_CLEAR(clear_module_state->__pyx_n_s_Callable);
  Py_CLEAR(clear_module_state->__pyx_n_s_CancellationResult);
  Py_CLEAR(clear_module_state->__pyx_n_s_ClientConfigAdapter);
  Py_CLEAR(clear_module_state->__pyx_n_s_Clock);
  Py_CLEAR(clear_module_state->__pyx_n_s_CompositeOrderBook);
  Py_CLEAR(clear_module_state->__pyx_n_s_Decimal);
  Py_CLEAR(clear_module_state->__pyx_n_s_Dict);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Dict_str_Decimal);
  Py_CLEAR(clear_module_state->__pyx_n_s_DummyMetricsCollector);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Error_call_trade_listener);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Error_canceling_order);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Error_executing_queued_order);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Error_processing_limit_order);
  Py_CLEAR(clear_module_state->__pyx_n_s_ExchangeBase);
  Py_CLEAR(clear_module_state->__pyx_n_s_ImportError);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Insufficient);
  Py_CLEAR(clear_module_state->__pyx_n_s_LIMIT);
  Py_CLEAR(clear_module_state->__pyx_n_s_LIMIT_MAKER);
  Py_CLEAR(clear_module_state->__pyx_n_s_LimitOrder);
  Py_CLEAR(clear_module_state->__pyx_n_s_List);
  Py_CLEAR(clear_module_state->__pyx_kp_s_List_CancellationResult);
  Py_CLEAR(clear_module_state->__pyx_n_s_MARKET);
  Py_CLEAR(clear_module_state->__pyx_n_s_MARKET_BUY_ORDER_CREATED_EVENT_T);
  Py_CLEAR(clear_module_state->__pyx_n_s_MARKET_ORDER_CANCELED_EVENT_TAG);
  Py_CLEAR(clear_module_state->__pyx_n_s_MARKET_ORDER_FAILURE_EVENT_TAG);
  Py_CLEAR(clear_module_state->__pyx_n_s_MARKET_SELL_ORDER_CREATED_EVENT);
  Py_CLEAR(clear_module_state->__pyx_n_s_MarketEvent);
  Py_CLEAR(clear_module_state->__pyx_n_s_MarketOrderFailureEvent);
  Py_CLEAR(clear_module_state->__pyx_n_b_NIL);
  Py_CLEAR(clear_module_state->__pyx_n_s_NetworkStatus);
  Py_CLEAR(clear_module_state->__pyx_kp_u_No_order_book_exists_for);
  Py_CLEAR(clear_module_state->__pyx_kp_u_None);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Not_enough);
  Py_CLEAR(clear_module_state->__pyx_n_s_ORDER_BOOK_TRADE_EVENT_TAG);
  Py_CLEAR(clear_module_state->__pyx_n_s_ORDER_FILLED_EVENT_TAG);
  Py_CLEAR(clear_module_state->__pyx_n_s_Optional);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Optional_bool);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderBook);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderBookEvent);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderBookMarketOrderFillListener);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderBookMarketOrderFillListener_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderBookMarketOrderFillListener_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderBookTracker);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderBookTradeEvent);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderBookTradeListener);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderBookTradeListener___reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderBookTradeListener___setstat);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderCancelled);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderCancelledEvent);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderCandidate);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderFailure);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderFilled);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderFilledEvent);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderType);
  Py_CLEAR(clear_module_state->__pyx_n_u_PaperTrade);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange___reduce_cyth);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange___setstate_cy);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_buy);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_cancel);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_cancel_all);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_check_network);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_get_all_balan);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_get_available);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_get_fee);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_get_maker_ord);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_get_order_boo);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_get_order_pri);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_get_price);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_get_taker_ord);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_init_paper_tr);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_match_trade_t);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_random_order);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_sell);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_set_balance);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_split_trading);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_start_network);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_stop_network);
  Py_CLEAR(clear_module_state->__pyx_n_s_PaperTradeExchange_trigger_event);
  Py_CLEAR(clear_module_state->__pyx_n_s_PickleError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_QuantizationParams);
  Py_CLEAR(clear_module_state->__pyx_n_s_QuantizationParams_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_QuantizationParams___reduce_cyth);
  Py_CLEAR(clear_module_state->__pyx_n_s_QuantizationParams___setstate_cy);
  Py_CLEAR(clear_module_state->__pyx_kp_u_QueuedOrder);
  Py_CLEAR(clear_module_state->__pyx_n_s_QueuedOrder_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_QueuedOrder___reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_QueuedOrder___setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_ROUND_DOWN);
  Py_CLEAR(clear_module_state->__pyx_n_s_SELL);
  Py_CLEAR(clear_module_state->__pyx_n_s_SELL_ORDER_COMPLETED_EVENT_TAG);
  Py_CLEAR(clear_module_state->__pyx_n_s_SellOrderCompleted);
  Py_CLEAR(clear_module_state->__pyx_n_s_SellOrderCompletedEvent);
  Py_CLEAR(clear_module_state->__pyx_n_s_SellOrderCreated);
  Py_CLEAR(clear_module_state->__pyx_n_s_SellOrderCreatedEvent);
  Py_CLEAR(clear_module_state->__pyx_n_s_TRADE_EXECUTION_DELAY);
  Py_CLEAR(clear_module_state->__pyx_n_s_TYPE_CHECKING);
  Py_CLEAR(clear_module_state->__pyx_n_s_TradeEvent);
  Py_CLEAR(clear_module_state->__pyx_n_s_TradeType);
  Py_CLEAR(clear_module_state->__pyx_n_s_TradingPair);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Trading_pair);
  Py_CLEAR(clear_module_state->__pyx_n_s_Tuple);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Tuple_str_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_TypeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_ValueError);
  Py_CLEAR(clear_module_state->__pyx_kp_u__10);
  Py_CLEAR(clear_module_state->__pyx_kp_u__11);
  Py_CLEAR(clear_module_state->__pyx_kp_u__15);
  Py_CLEAR(clear_module_state->__pyx_kp_u__20);
  Py_CLEAR(clear_module_state->__pyx_kp_u__22);
  Py_CLEAR(clear_module_state->__pyx_kp_u__26);
  Py_CLEAR(clear_module_state->__pyx_kp_u__3);
  Py_CLEAR(clear_module_state->__pyx_n_s__36);
  Py_CLEAR(clear_module_state->__pyx_kp_u__4);
  Py_CLEAR(clear_module_state->__pyx_kp_u__5);
  Py_CLEAR(clear_module_state->__pyx_kp_u__6);
  Py_CLEAR(clear_module_state->__pyx_n_s__88);
  Py_CLEAR(clear_module_state->__pyx_kp_u__9);
  Py_CLEAR(clear_module_state->__pyx_n_s_all);
  Py_CLEAR(clear_module_state->__pyx_n_s_amount);
  Py_CLEAR(clear_module_state->__pyx_n_s_append);
  Py_CLEAR(clear_module_state->__pyx_n_s_args);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_tasks);
  Py_CLEAR(clear_module_state->__pyx_kp_u_available);
  Py_CLEAR(clear_module_state->__pyx_n_s_available_balances);
  Py_CLEAR(clear_module_state->__pyx_kp_u_available_vs);
  Py_CLEAR(clear_module_state->__pyx_n_s_await);
  Py_CLEAR(clear_module_state->__pyx_n_s_balance);
  Py_CLEAR(clear_module_state->__pyx_kp_u_balance_available_for_buy_order);
  Py_CLEAR(clear_module_state->__pyx_kp_u_balance_available_for_sell_orde);
  Py_CLEAR(clear_module_state->__pyx_kp_u_balance_to_fill_limit_buy_order);
  Py_CLEAR(clear_module_state->__pyx_kp_u_balance_to_fill_limit_sell_orde);
  Py_CLEAR(clear_module_state->__pyx_n_s_base_asset);
  Py_CLEAR(clear_module_state->__pyx_n_s_base_currency);
  Py_CLEAR(clear_module_state->__pyx_n_s_bool);
  Py_CLEAR(clear_module_state->__pyx_n_s_build_trade_fee);
  Py_CLEAR(clear_module_state->__pyx_n_s_buy);
  Py_CLEAR(clear_module_state->__pyx_n_u_buy);
  Py_CLEAR(clear_module_state->__pyx_n_s_cancel);
  Py_CLEAR(clear_module_state->__pyx_n_s_cancel_all);
  Py_CLEAR(clear_module_state->__pyx_n_s_cancellation_results);
  Py_CLEAR(clear_module_state->__pyx_n_s_ceil);
  Py_CLEAR(clear_module_state->__pyx_n_s_check_network);
  Py_CLEAR(clear_module_state->__pyx_n_s_choice);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_getitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_client_config_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_client_order_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_cls);
  Py_CLEAR(clear_module_state->__pyx_n_s_collections);
  Py_CLEAR(clear_module_state->__pyx_n_s_convert_from_exchange_trading_pa);
  Py_CLEAR(clear_module_state->__pyx_n_s_convert_to_exchange_trading_pair);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_timestamp);
  Py_CLEAR(clear_module_state->__pyx_n_s_currency);
  Py_CLEAR(clear_module_state->__pyx_n_s_data_source);
  Py_CLEAR(clear_module_state->__pyx_n_s_decimal);
  Py_CLEAR(clear_module_state->__pyx_n_s_defaultdict);
  Py_CLEAR(clear_module_state->__pyx_n_s_deque);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_disable);
  Py_CLEAR(clear_module_state->__pyx_kp_u_does_not_existing_in_current_da);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enable);
  Py_CLEAR(clear_module_state->__pyx_n_s_encode);
  Py_CLEAR(clear_module_state->__pyx_n_s_error);
  Py_CLEAR(clear_module_state->__pyx_n_s_event);
  Py_CLEAR(clear_module_state->__pyx_n_s_event_object);
  Py_CLEAR(clear_module_state->__pyx_n_s_event_tag);
  Py_CLEAR(clear_module_state->__pyx_n_s_exc_info);
  Py_CLEAR(clear_module_state->__pyx_n_s_exchange);
  Py_CLEAR(clear_module_state->__pyx_n_s_exchange_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_exchange_trade_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_float);
  Py_CLEAR(clear_module_state->__pyx_n_s_from_total_balances);
  Py_CLEAR(clear_module_state->__pyx_kp_u_gc);
  Py_CLEAR(clear_module_state->__pyx_n_s_get);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_all_balances);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_available_balance);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_fee);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_maker_order_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_order_book);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_order_price_quantum);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_price);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_taker_order_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_getstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_client_config_config);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_connector_budget_chec);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_connector_connector_m);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_connector_exchange_ba);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_connector_exchange_pa);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_connector_exchange_pa_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_hummingbot_connector_exchange_pa_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_core_clock);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_core_data_type_cancel);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_core_data_type_common);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_core_data_type_compos);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_core_data_type_limit);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_core_data_type_order);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_core_data_type_order_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_core_event_events);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_core_network_iterator);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_core_utils_async_util);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_core_utils_estimate_f);
  Py_CLEAR(clear_module_state->__pyx_n_s_i);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_init);
  Py_CLEAR(clear_module_state->__pyx_n_s_init___locals_lambda);
  Py_CLEAR(clear_module_state->__pyx_n_s_init_paper_trade_market);
  Py_CLEAR(clear_module_state->__pyx_n_s_initializing);
  Py_CLEAR(clear_module_state->__pyx_n_s_inspect);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_buy);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_finite);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_maker);
  Py_CLEAR(clear_module_state->__pyx_kp_u_isenabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_items);
  Py_CLEAR(clear_module_state->__pyx_n_s_keys);
  Py_CLEAR(clear_module_state->__pyx_n_s_kwargs);
  Py_CLEAR(clear_module_state->__pyx_n_s_limit_orders);
  Py_CLEAR(clear_module_state->__pyx_n_s_limit_orders_map_ptr);
  Py_CLEAR(clear_module_state->__pyx_n_s_log10);
  Py_CLEAR(clear_module_state->__pyx_n_s_logger);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_s_market);
  Py_CLEAR(clear_module_state->__pyx_n_s_match_trade_to_limit_orders);
  Py_CLEAR(clear_module_state->__pyx_n_s_math);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_name_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_needed_vs);
  Py_CLEAR(clear_module_state->__pyx_n_s_new);
  Py_CLEAR(clear_module_state->__pyx_kp_u_numpy_core_multiarray_failed_to);
  Py_CLEAR(clear_module_state->__pyx_kp_u_numpy_core_umath_failed_to_impor);
  Py_CLEAR(clear_module_state->__pyx_n_s_on_hold_balances);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_book);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_book_create_function);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_book_tracker);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_books);
  Py_CLEAR(clear_module_state->__pyx_n_u_order_books_initialized);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_collateral);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_filled_events_from_order_b);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_side);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_size_decimals);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_size_precision);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_pickle);
  Py_CLEAR(clear_module_state->__pyx_n_s_popleft);
  Py_CLEAR(clear_module_state->__pyx_n_s_populate_collateral_entries);
  Py_CLEAR(clear_module_state->__pyx_n_s_potential_returns);
  Py_CLEAR(clear_module_state->__pyx_n_s_price);
  Py_CLEAR(clear_module_state->__pyx_n_s_price_decimals);
  Py_CLEAR(clear_module_state->__pyx_n_s_price_precision);
  Py_CLEAR(clear_module_state->__pyx_n_s_ptm_logger);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_PickleError);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_checksum);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_result);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_OrderBookMarketOr);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_OrderBookTradeLis);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_QuantizationParam);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_unpickle_QueuedOrder);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_vtable);
  Py_CLEAR(clear_module_state->__pyx_n_s_quantity);
  Py_CLEAR(clear_module_state->__pyx_n_s_quantize);
  Py_CLEAR(clear_module_state->__pyx_n_s_quote_asset);
  Py_CLEAR(clear_module_state->__pyx_n_s_quote_currency);
  Py_CLEAR(clear_module_state->__pyx_n_s_random);
  Py_CLEAR(clear_module_state->__pyx_n_s_random_order_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_ready);
  Py_CLEAR(clear_module_state->__pyx_n_s_record_filled_order);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_ex);
  Py_CLEAR(clear_module_state->__pyx_kp_u_required_for_the_order);
  Py_CLEAR(clear_module_state->__pyx_n_s_results);
  Py_CLEAR(clear_module_state->__pyx_n_s_return);
  Py_CLEAR(clear_module_state->__pyx_n_s_rounding);
  Py_CLEAR(clear_module_state->__pyx_n_s_s_decimal_0);
  Py_CLEAR(clear_module_state->__pyx_n_s_safe_ensure_future);
  Py_CLEAR(clear_module_state->__pyx_n_s_self);
  Py_CLEAR(clear_module_state->__pyx_kp_s_self__ask_limit_orders_self__bid);
  Py_CLEAR(clear_module_state->__pyx_n_s_sell);
  Py_CLEAR(clear_module_state->__pyx_n_u_sell);
  Py_CLEAR(clear_module_state->__pyx_n_s_send);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_balance);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_order_book_tracker);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_simulate_buy);
  Py_CLEAR(clear_module_state->__pyx_n_s_simulate_sell);
  Py_CLEAR(clear_module_state->__pyx_n_s_sleep);
  Py_CLEAR(clear_module_state->__pyx_n_s_spec);
  Py_CLEAR(clear_module_state->__pyx_n_s_split_trading_pair);
  Py_CLEAR(clear_module_state->__pyx_n_s_start);
  Py_CLEAR(clear_module_state->__pyx_n_s_start_network);
  Py_CLEAR(clear_module_state->__pyx_n_s_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_status_dict);
  Py_CLEAR(clear_module_state->__pyx_n_s_stop);
  Py_CLEAR(clear_module_state->__pyx_n_s_stop_network);
  Py_CLEAR(clear_module_state->__pyx_n_s_str);
  Py_CLEAR(clear_module_state->__pyx_kp_s_stringsource);
  Py_CLEAR(clear_module_state->__pyx_n_s_super);
  Py_CLEAR(clear_module_state->__pyx_n_s_target_market);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_throw);
  Py_CLEAR(clear_module_state->__pyx_n_s_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_timeout_seconds);
  Py_CLEAR(clear_module_state->__pyx_n_s_trading_pair);
  Py_CLEAR(clear_module_state->__pyx_n_s_trading_pair_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_trigger_event_async);
  Py_CLEAR(clear_module_state->__pyx_n_s_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_typing);
  Py_CLEAR(clear_module_state->__pyx_n_s_update);
  Py_CLEAR(clear_module_state->__pyx_n_s_upper);
  Py_CLEAR(clear_module_state->__pyx_n_s_use_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_u_utf8);
  Py_CLEAR(clear_module_state->__pyx_n_s_val);
  Py_CLEAR(clear_module_state->__pyx_n_s_vals);
  Py_CLEAR(clear_module_state->__pyx_n_s_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_values);
  Py_CLEAR(clear_module_state->__pyx_n_s_warning);
  Py_CLEAR(clear_module_state->__pyx_float_0_0);
  Py_CLEAR(clear_module_state->__pyx_float_1e6);
  Py_CLEAR(clear_module_state->__pyx_float_5_0);
  Py_CLEAR(clear_module_state->__pyx_float_0_01);
  Py_CLEAR(clear_module_state->__pyx_float_1eneg_7);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_256);
  Py_CLEAR(clear_module_state->__pyx_int_24020995);
  Py_CLEAR(clear_module_state->__pyx_int_29025389);
  Py_CLEAR(clear_module_state->__pyx_int_57171680);
  Py_CLEAR(clear_module_state->__pyx_int_58563262);
  Py_CLEAR(clear_module_state->__pyx_int_95353964);
  Py_CLEAR(clear_module_state->__pyx_int_126189128);
  Py_CLEAR(clear_module_state->__pyx_int_136360524);
  Py_CLEAR(clear_module_state->__pyx_int_175429262);
  Py_CLEAR(clear_module_state->__pyx_int_251288030);
  Py_CLEAR(clear_module_state->__pyx_k__16);
  Py_CLEAR(clear_module_state->__pyx_k__17);
  Py_CLEAR(clear_module_state->__pyx_k__18);
  Py_CLEAR(clear_module_state->__pyx_k__19);
  Py_CLEAR(clear_module_state->__pyx_k__27);
  Py_CLEAR(clear_module_state->__pyx_k__28);
  Py_CLEAR(clear_module_state->__pyx_k__29);
  Py_CLEAR(clear_module_state->__pyx_k__30);
  Py_CLEAR(clear_module_state->__pyx_k__31);
  Py_CLEAR(clear_module_state->__pyx_k__55);
  Py_CLEAR(clear_module_state->__pyx_k__56);
  Py_CLEAR(clear_module_state->__pyx_k__60);
  Py_CLEAR(clear_module_state->__pyx_tuple_);
  Py_CLEAR(clear_module_state->__pyx_tuple__2);
  Py_CLEAR(clear_module_state->__pyx_tuple__7);
  Py_CLEAR(clear_module_state->__pyx_tuple__8);
  Py_CLEAR(clear_module_state->__pyx_tuple__21);
  Py_CLEAR(clear_module_state->__pyx_tuple__23);
  Py_CLEAR(clear_module_state->__pyx_tuple__25);
  Py_CLEAR(clear_module_state->__pyx_tuple__33);
  Py_CLEAR(clear_module_state->__pyx_tuple__34);
  Py_CLEAR(clear_module_state->__pyx_tuple__35);
  Py_CLEAR(clear_module_state->__pyx_tuple__37);
  Py_CLEAR(clear_module_state->__pyx_tuple__38);
  Py_CLEAR(clear_module_state->__pyx_tuple__40);
  Py_CLEAR(clear_module_state->__pyx_tuple__48);
  Py_CLEAR(clear_module_state->__pyx_tuple__50);
  Py_CLEAR(clear_module_state->__pyx_tuple__52);
  Py_CLEAR(clear_module_state->__pyx_tuple__54);
  Py_CLEAR(clear_module_state->__pyx_tuple__57);
  Py_CLEAR(clear_module_state->__pyx_tuple__58);
  Py_CLEAR(clear_module_state->__pyx_tuple__61);
  Py_CLEAR(clear_module_state->__pyx_tuple__64);
  Py_CLEAR(clear_module_state->__pyx_tuple__66);
  Py_CLEAR(clear_module_state->__pyx_tuple__68);
  Py_CLEAR(clear_module_state->__pyx_tuple__70);
  Py_CLEAR(clear_module_state->__pyx_tuple__73);
  Py_CLEAR(clear_module_state->__pyx_tuple__75);
  Py_CLEAR(clear_module_state->__pyx_tuple__80);
  Py_CLEAR(clear_module_state->__pyx_tuple__83);
  Py_CLEAR(clear_module_state->__pyx_codeobj__12);
  Py_CLEAR(clear_module_state->__pyx_codeobj__13);
  Py_CLEAR(clear_module_state->__pyx_codeobj__14);
  Py_CLEAR(clear_module_state->__pyx_codeobj__24);
  Py_CLEAR(clear_module_state->__pyx_codeobj__32);
  Py_CLEAR(clear_module_state->__pyx_codeobj__39);
  Py_CLEAR(clear_module_state->__pyx_codeobj__41);
  Py_CLEAR(clear_module_state->__pyx_codeobj__42);
  Py_CLEAR(clear_module_state->__pyx_codeobj__43);
  Py_CLEAR(clear_module_state->__pyx_codeobj__44);
  Py_CLEAR(clear_module_state->__pyx_codeobj__45);
  Py_CLEAR(clear_module_state->__pyx_codeobj__46);
  Py_CLEAR(clear_module_state->__pyx_codeobj__47);
  Py_CLEAR(clear_module_state->__pyx_codeobj__49);
  Py_CLEAR(clear_module_state->__pyx_codeobj__51);
  Py_CLEAR(clear_module_state->__pyx_codeobj__53);
  Py_CLEAR(clear_module_state->__pyx_codeobj__59);
  Py_CLEAR(clear_module_state->__pyx_codeobj__62);
  Py_CLEAR(clear_module_state->__pyx_codeobj__63);
  Py_CLEAR(clear_module_state->__pyx_codeobj__65);
  Py_CLEAR(clear_module_state->__pyx_codeobj__67);
  Py_CLEAR(clear_module_state->__pyx_codeobj__69);
  Py_CLEAR(clear_module_state->__pyx_codeobj__71);
  Py_CLEAR(clear_module_state->__pyx_codeobj__72);
  Py_CLEAR(clear_module_state->__pyx_codeobj__74);
  Py_CLEAR(clear_module_state->__pyx_codeobj__76);
  Py_CLEAR(clear_module_state->__pyx_codeobj__77);
  Py_CLEAR(clear_module_state->__pyx_codeobj__78);
  Py_CLEAR(clear_module_state->__pyx_codeobj__79);
  Py_CLEAR(clear_module_state->__pyx_codeobj__81);
  Py_CLEAR(clear_module_state->__pyx_codeobj__82);
  Py_CLEAR(clear_module_state->__pyx_codeobj__84);
  Py_CLEAR(clear_module_state->__pyx_codeobj__85);
  Py_CLEAR(clear_module_state->__pyx_codeobj__86);
  Py_CLEAR(clear_module_state->__pyx_codeobj__87);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_6pubsub_PubSub);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4bool_bool);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_7complex_complex);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_dtype);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_flatiter);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_broadcast);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_ndarray);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_generic);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_number);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_integer);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_signedinteger);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_unsignedinteger);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_inexact);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_floating);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_complexfloating);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_flexible);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_character);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_ufunc);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_5clock_Clock);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async);
  Py_VISIT(traverse_module_state->__pyx_kp_u_0);
  Py_VISIT(traverse_module_state->__pyx_kp_u_02x);
  Py_VISIT(traverse_module_state->__pyx_kp_u_1e);
  Py_VISIT(traverse_module_state->__pyx_kp_u_1e_10);
  Py_VISIT(traverse_module_state->__pyx_kp_u_1e_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_1e_7);
  Py_VISIT(traverse_module_state->__pyx_kp_u_7g);
  Py_VISIT(traverse_module_state->__pyx_kp_u_8g);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Account_balance_does_not_have_as);
  Py_VISIT(traverse_module_state->__pyx_n_s_AssertionError);
  Py_VISIT(traverse_module_state->__pyx_n_s_BUY);
  Py_VISIT(traverse_module_state->__pyx_n_u_BUY);
  Py_VISIT(traverse_module_state->__pyx_n_s_BUY_ORDER_COMPLETED_EVENT_TAG);
  Py_VISIT(traverse_module_state->__pyx_n_s_BudgetChecker);
  Py_VISIT(traverse_module_state->__pyx_n_s_BuyOrderCompleted);
  Py_VISIT(traverse_module_state->__pyx_n_s_BuyOrderCompletedEvent);
  Py_VISIT(traverse_module_state->__pyx_n_s_BuyOrderCreated);
  Py_VISIT(traverse_module_state->__pyx_n_s_BuyOrderCreatedEvent);
  Py_VISIT(traverse_module_state->__pyx_n_s_CONNECTED);
  Py_VISIT(traverse_module_state->__pyx_n_s_Callable);
  Py_VISIT(traverse_module_state->__pyx_n_s_CancellationResult);
  Py_VISIT(traverse_module_state->__pyx_n_s_ClientConfigAdapter);
  Py_VISIT(traverse_module_state->__pyx_n_s_Clock);
  Py_VISIT(traverse_module_state->__pyx_n_s_CompositeOrderBook);
  Py_VISIT(traverse_module_state->__pyx_n_s_Decimal);
  Py_VISIT(traverse_module_state->__pyx_n_s_Dict);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Dict_str_Decimal);
  Py_VISIT(traverse_module_state->__pyx_n_s_DummyMetricsCollector);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Error_call_trade_listener);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Error_canceling_order);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Error_executing_queued_order);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Error_processing_limit_order);
  Py_VISIT(traverse_module_state->__pyx_n_s_ExchangeBase);
  Py_VISIT(traverse_module_state->__pyx_n_s_ImportError);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Insufficient);
  Py_VISIT(traverse_module_state->__pyx_n_s_LIMIT);
  Py_VISIT(traverse_module_state->__pyx_n_s_LIMIT_MAKER);
  Py_VISIT(traverse_module_state->__pyx_n_s_LimitOrder);
  Py_VISIT(traverse_module_state->__pyx_n_s_List);
  Py_VISIT(traverse_module_state->__pyx_kp_s_List_CancellationResult);
  Py_VISIT(traverse_module_state->__pyx_n_s_MARKET);
  Py_VISIT(traverse_module_state->__pyx_n_s_MARKET_BUY_ORDER_CREATED_EVENT_T);
  Py_VISIT(traverse_module_state->__pyx_n_s_MARKET_ORDER_CANCELED_EVENT_TAG);
  Py_VISIT(traverse_module_state->__pyx_n_s_MARKET_ORDER_FAILURE_EVENT_TAG);
  Py_VISIT(traverse_module_state->__pyx_n_s_MARKET_SELL_ORDER_CREATED_EVENT);
  Py_VISIT(traverse_module_state->__pyx_n_s_MarketEvent);
  Py_VISIT(traverse_module_state->__pyx_n_s_MarketOrderFailureEvent);
  Py_VISIT(traverse_module_state->__pyx_n_b_NIL);
  Py_VISIT(traverse_module_state->__pyx_n_s_NetworkStatus);
  Py_VISIT(traverse_module_state->__pyx_kp_u_No_order_book_exists_for);
  Py_VISIT(traverse_module_state->__pyx_kp_u_None);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Not_enough);
  Py_VISIT(traverse_module_state->__pyx_n_s_ORDER_BOOK_TRADE_EVENT_TAG);
  Py_VISIT(traverse_module_state->__pyx_n_s_ORDER_FILLED_EVENT_TAG);
  Py_VISIT(traverse_module_state->__pyx_n_s_Optional);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Optional_bool);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderBook);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderBookEvent);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderBookMarketOrderFillListener);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderBookMarketOrderFillListener_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderBookMarketOrderFillListener_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderBookTracker);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderBookTradeEvent);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderBookTradeListener);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderBookTradeListener___reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderBookTradeListener___setstat);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderCancelled);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderCancelledEvent);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderCandidate);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderFailure);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderFilled);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderFilledEvent);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderType);
  Py_VISIT(traverse_module_state->__pyx_n_u_PaperTrade);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange___reduce_cyth);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange___setstate_cy);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_buy);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_cancel);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_cancel_all);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_check_network);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_get_all_balan);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_get_available);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_get_fee);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_get_maker_ord);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_get_order_boo);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_get_order_pri);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_get_price);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_get_taker_ord);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_init_paper_tr);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_match_trade_t);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_random_order);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_sell);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_set_balance);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_split_trading);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_start_network);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_stop_network);
  Py_VISIT(traverse_module_state->__pyx_n_s_PaperTradeExchange_trigger_event);
  Py_VISIT(traverse_module_state->__pyx_n_s_PickleError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_QuantizationParams);
  Py_VISIT(traverse_module_state->__pyx_n_s_QuantizationParams_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_QuantizationParams___reduce_cyth);
  Py_VISIT(traverse_module_state->__pyx_n_s_QuantizationParams___setstate_cy);
  Py_VISIT(traverse_module_state->__pyx_kp_u_QueuedOrder);
  Py_VISIT(traverse_module_state->__pyx_n_s_QueuedOrder_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_QueuedOrder___reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_QueuedOrder___setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_ROUND_DOWN);
  Py_VISIT(traverse_module_state->__pyx_n_s_SELL);
  Py_VISIT(traverse_module_state->__pyx_n_s_SELL_ORDER_COMPLETED_EVENT_TAG);
  Py_VISIT(traverse_module_state->__pyx_n_s_SellOrderCompleted);
  Py_VISIT(traverse_module_state->__pyx_n_s_SellOrderCompletedEvent);
  Py_VISIT(traverse_module_state->__pyx_n_s_SellOrderCreated);
  Py_VISIT(traverse_module_state->__pyx_n_s_SellOrderCreatedEvent);
  Py_VISIT(traverse_module_state->__pyx_n_s_TRADE_EXECUTION_DELAY);
  Py_VISIT(traverse_module_state->__pyx_n_s_TYPE_CHECKING);
  Py_VISIT(traverse_module_state->__pyx_n_s_TradeEvent);
  Py_VISIT(traverse_module_state->__pyx_n_s_TradeType);
  Py_VISIT(traverse_module_state->__pyx_n_s_TradingPair);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Trading_pair);
  Py_VISIT(traverse_module_state->__pyx_n_s_Tuple);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Tuple_str_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_TypeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_ValueError);
  Py_VISIT(traverse_module_state->__pyx_kp_u__10);
  Py_VISIT(traverse_module_state->__pyx_kp_u__11);
  Py_VISIT(traverse_module_state->__pyx_kp_u__15);
  Py_VISIT(traverse_module_state->__pyx_kp_u__20);
  Py_VISIT(traverse_module_state->__pyx_kp_u__22);
  Py_VISIT(traverse_module_state->__pyx_kp_u__26);
  Py_VISIT(traverse_module_state->__pyx_kp_u__3);
  Py_VISIT(traverse_module_state->__pyx_n_s__36);
  Py_VISIT(traverse_module_state->__pyx_kp_u__4);
  Py_VISIT(traverse_module_state->__pyx_kp_u__5);
  Py_VISIT(traverse_module_state->__pyx_kp_u__6);
  Py_VISIT(traverse_module_state->__pyx_n_s__88);
  Py_VISIT(traverse_module_state->__pyx_kp_u__9);
  Py_VISIT(traverse_module_state->__pyx_n_s_all);
  Py_VISIT(traverse_module_state->__pyx_n_s_amount);
  Py_VISIT(traverse_module_state->__pyx_n_s_append);
  Py_VISIT(traverse_module_state->__pyx_n_s_args);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_tasks);
  Py_VISIT(traverse_module_state->__pyx_kp_u_available);
  Py_VISIT(traverse_module_state->__pyx_n_s_available_balances);
  Py_VISIT(traverse_module_state->__pyx_kp_u_available_vs);
  Py_VISIT(traverse_module_state->__pyx_n_s_await);
  Py_VISIT(traverse_module_state->__pyx_n_s_balance);
  Py_VISIT(traverse_module_state->__pyx_kp_u_balance_available_for_buy_order);
  Py_VISIT(traverse_module_state->__pyx_kp_u_balance_available_for_sell_orde);
  Py_VISIT(traverse_module_state->__pyx_kp_u_balance_to_fill_limit_buy_order);
  Py_VISIT(traverse_module_state->__pyx_kp_u_balance_to_fill_limit_sell_orde);
  Py_VISIT(traverse_module_state->__pyx_n_s_base_asset);
  Py_VISIT(traverse_module_state->__pyx_n_s_base_currency);
  Py_VISIT(traverse_module_state->__pyx_n_s_bool);
  Py_VISIT(traverse_module_state->__pyx_n_s_build_trade_fee);
  Py_VISIT(traverse_module_state->__pyx_n_s_buy);
  Py_VISIT(traverse_module_state->__pyx_n_u_buy);
  Py_VISIT(traverse_module_state->__pyx_n_s_cancel);
  Py_VISIT(traverse_module_state->__pyx_n_s_cancel_all);
  Py_VISIT(traverse_module_state->__pyx_n_s_cancellation_results);
  Py_VISIT(traverse_module_state->__pyx_n_s_ceil);
  Py_VISIT(traverse_module_state->__pyx_n_s_check_network);
  Py_VISIT(traverse_module_state->__pyx_n_s_choice);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_getitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_client_config_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_client_order_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_cls);
  Py_VISIT(traverse_module_state->__pyx_n_s_collections);
  Py_VISIT(traverse_module_state->__pyx_n_s_convert_from_exchange_trading_pa);
  Py_VISIT(traverse_module_state->__pyx_n_s_convert_to_exchange_trading_pair);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_timestamp);
  Py_VISIT(traverse_module_state->__pyx_n_s_currency);
  Py_VISIT(traverse_module_state->__pyx_n_s_data_source);
  Py_VISIT(traverse_module_state->__pyx_n_s_decimal);
  Py_VISIT(traverse_module_state->__pyx_n_s_defaultdict);
  Py_VISIT(traverse_module_state->__pyx_n_s_deque);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_disable);
  Py_VISIT(traverse_module_state->__pyx_kp_u_does_not_existing_in_current_da);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enable);
  Py_VISIT(traverse_module_state->__pyx_n_s_encode);
  Py_VISIT(traverse_module_state->__pyx_n_s_error);
  Py_VISIT(traverse_module_state->__pyx_n_s_event);
  Py_VISIT(traverse_module_state->__pyx_n_s_event_object);
  Py_VISIT(traverse_module_state->__pyx_n_s_event_tag);
  Py_VISIT(traverse_module_state->__pyx_n_s_exc_info);
  Py_VISIT(traverse_module_state->__pyx_n_s_exchange);
  Py_VISIT(traverse_module_state->__pyx_n_s_exchange_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_exchange_trade_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_float);
  Py_VISIT(traverse_module_state->__pyx_n_s_from_total_balances);
  Py_VISIT(traverse_module_state->__pyx_kp_u_gc);
  Py_VISIT(traverse_module_state->__pyx_n_s_get);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_all_balances);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_available_balance);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_fee);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_maker_order_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_order_book);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_order_price_quantum);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_price);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_taker_order_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_getstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_client_config_config);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_connector_budget_chec);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_connector_connector_m);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_connector_exchange_ba);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_connector_exchange_pa);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_connector_exchange_pa_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_hummingbot_connector_exchange_pa_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_core_clock);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_core_data_type_cancel);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_core_data_type_common);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_core_data_type_compos);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_core_data_type_limit);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_core_data_type_order);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_core_data_type_order_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_core_event_events);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_core_network_iterator);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_core_utils_async_util);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_core_utils_estimate_f);
  Py_VISIT(traverse_module_state->__pyx_n_s_i);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_init);
  Py_VISIT(traverse_module_state->__pyx_n_s_init___locals_lambda);
  Py_VISIT(traverse_module_state->__pyx_n_s_init_paper_trade_market);
  Py_VISIT(traverse_module_state->__pyx_n_s_initializing);
  Py_VISIT(traverse_module_state->__pyx_n_s_inspect);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_buy);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_finite);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_maker);
  Py_VISIT(traverse_module_state->__pyx_kp_u_isenabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_items);
  Py_VISIT(traverse_module_state->__pyx_n_s_keys);
  Py_VISIT(traverse_module_state->__pyx_n_s_kwargs);
  Py_VISIT(traverse_module_state->__pyx_n_s_limit_orders);
  Py_VISIT(traverse_module_state->__pyx_n_s_limit_orders_map_ptr);
  Py_VISIT(traverse_module_state->__pyx_n_s_log10);
  Py_VISIT(traverse_module_state->__pyx_n_s_logger);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_s_market);
  Py_VISIT(traverse_module_state->__pyx_n_s_match_trade_to_limit_orders);
  Py_VISIT(traverse_module_state->__pyx_n_s_math);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_name_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_needed_vs);
  Py_VISIT(traverse_module_state->__pyx_n_s_new);
  Py_VISIT(traverse_module_state->__pyx_kp_u_numpy_core_multiarray_failed_to);
  Py_VISIT(traverse_module_state->__pyx_kp_u_numpy_core_umath_failed_to_impor);
  Py_VISIT(traverse_module_state->__pyx_n_s_on_hold_balances);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_book);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_book_create_function);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_book_tracker);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_books);
  Py_VISIT(traverse_module_state->__pyx_n_u_order_books_initialized);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_collateral);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_filled_events_from_order_b);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_side);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_size_decimals);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_size_precision);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_pickle);
  Py_VISIT(traverse_module_state->__pyx_n_s_popleft);
  Py_VISIT(traverse_module_state->__pyx_n_s_populate_collateral_entries);
  Py_VISIT(traverse_module_state->__pyx_n_s_potential_returns);
  Py_VISIT(traverse_module_state->__pyx_n_s_price);
  Py_VISIT(traverse_module_state->__pyx_n_s_price_decimals);
  Py_VISIT(traverse_module_state->__pyx_n_s_price_precision);
  Py_VISIT(traverse_module_state->__pyx_n_s_ptm_logger);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_PickleError);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_checksum);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_result);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_OrderBookMarketOr);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_OrderBookTradeLis);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_QuantizationParam);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_unpickle_QueuedOrder);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_vtable);
  Py_VISIT(traverse_module_state->__pyx_n_s_quantity);
  Py_VISIT(traverse_module_state->__pyx_n_s_quantize);
  Py_VISIT(traverse_module_state->__pyx_n_s_quote_asset);
  Py_VISIT(traverse_module_state->__pyx_n_s_quote_currency);
  Py_VISIT(traverse_module_state->__pyx_n_s_random);
  Py_VISIT(traverse_module_state->__pyx_n_s_random_order_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_ready);
  Py_VISIT(traverse_module_state->__pyx_n_s_record_filled_order);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_ex);
  Py_VISIT(traverse_module_state->__pyx_kp_u_required_for_the_order);
  Py_VISIT(traverse_module_state->__pyx_n_s_results);
  Py_VISIT(traverse_module_state->__pyx_n_s_return);
  Py_VISIT(traverse_module_state->__pyx_n_s_rounding);
  Py_VISIT(traverse_module_state->__pyx_n_s_s_decimal_0);
  Py_VISIT(traverse_module_state->__pyx_n_s_safe_ensure_future);
  Py_VISIT(traverse_module_state->__pyx_n_s_self);
  Py_VISIT(traverse_module_state->__pyx_kp_s_self__ask_limit_orders_self__bid);
  Py_VISIT(traverse_module_state->__pyx_n_s_sell);
  Py_VISIT(traverse_module_state->__pyx_n_u_sell);
  Py_VISIT(traverse_module_state->__pyx_n_s_send);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_balance);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_order_book_tracker);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_simulate_buy);
  Py_VISIT(traverse_module_state->__pyx_n_s_simulate_sell);
  Py_VISIT(traverse_module_state->__pyx_n_s_sleep);
  Py_VISIT(traverse_module_state->__pyx_n_s_spec);
  Py_VISIT(traverse_module_state->__pyx_n_s_split_trading_pair);
  Py_VISIT(traverse_module_state->__pyx_n_s_start);
  Py_VISIT(traverse_module_state->__pyx_n_s_start_network);
  Py_VISIT(traverse_module_state->__pyx_n_s_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_status_dict);
  Py_VISIT(traverse_module_state->__pyx_n_s_stop);
  Py_VISIT(traverse_module_state->__pyx_n_s_stop_network);
  Py_VISIT(traverse_module_state->__pyx_n_s_str);
  Py_VISIT(traverse_module_state->__pyx_kp_s_stringsource);
  Py_VISIT(traverse_module_state->__pyx_n_s_super);
  Py_VISIT(traverse_module_state->__pyx_n_s_target_market);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_throw);
  Py_VISIT(traverse_module_state->__pyx_n_s_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_timeout_seconds);
  Py_VISIT(traverse_module_state->__pyx_n_s_trading_pair);
  Py_VISIT(traverse_module_state->__pyx_n_s_trading_pair_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_trigger_event_async);
  Py_VISIT(traverse_module_state->__pyx_n_s_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_typing);
  Py_VISIT(traverse_module_state->__pyx_n_s_update);
  Py_VISIT(traverse_module_state->__pyx_n_s_upper);
  Py_VISIT(traverse_module_state->__pyx_n_s_use_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_u_utf8);
  Py_VISIT(traverse_module_state->__pyx_n_s_val);
  Py_VISIT(traverse_module_state->__pyx_n_s_vals);
  Py_VISIT(traverse_module_state->__pyx_n_s_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_values);
  Py_VISIT(traverse_module_state->__pyx_n_s_warning);
  Py_VISIT(traverse_module_state->__pyx_float_0_0);
  Py_VISIT(traverse_module_state->__pyx_float_1e6);
  Py_VISIT(traverse_module_state->__pyx_float_5_0);
  Py_VISIT(traverse_module_state->__pyx_float_0_01);
  Py_VISIT(traverse_module_state->__pyx_float_1eneg_7);
  Py_VISIT(traverse_module_state->__pyx_int_0);
  Py_VISIT(traverse_module_state->__pyx_int_256);
  Py_VISIT(traverse_module_state->__pyx_int_24020995);
  Py_VISIT(traverse_module_state->__pyx_int_29025389);
  Py_VISIT(traverse_module_state->__pyx_int_57171680);
  Py_VISIT(traverse_module_state->__pyx_int_58563262);
  Py_VISIT(traverse_module_state->__pyx_int_95353964);
  Py_VISIT(traverse_module_state->__pyx_int_126189128);
  Py_VISIT(traverse_module_state->__pyx_int_136360524);
  Py_VISIT(traverse_module_state->__pyx_int_175429262);
  Py_VISIT(traverse_module_state->__pyx_int_251288030);
  Py_VISIT(traverse_module_state->__pyx_k__16);
  Py_VISIT(traverse_module_state->__pyx_k__17);
  Py_VISIT(traverse_module_state->__pyx_k__18);
  Py_VISIT(traverse_module_state->__pyx_k__19);
  Py_VISIT(traverse_module_state->__pyx_k__27);
  Py_VISIT(traverse_module_state->__pyx_k__28);
  Py_VISIT(traverse_module_state->__pyx_k__29);
  Py_VISIT(traverse_module_state->__pyx_k__30);
  Py_VISIT(traverse_module_state->__pyx_k__31);
  Py_VISIT(traverse_module_state->__pyx_k__55);
  Py_VISIT(traverse_module_state->__pyx_k__56);
  Py_VISIT(traverse_module_state->__pyx_k__60);
  Py_VISIT(traverse_module_state->__pyx_tuple_);
  Py_VISIT(traverse_module_state->__pyx_tuple__2);
  Py_VISIT(traverse_module_state->__pyx_tuple__7);
  Py_VISIT(traverse_module_state->__pyx_tuple__8);
  Py_VISIT(traverse_module_state->__pyx_tuple__21);
  Py_VISIT(traverse_module_state->__pyx_tuple__23);
  Py_VISIT(traverse_module_state->__pyx_tuple__25);
  Py_VISIT(traverse_module_state->__pyx_tuple__33);
  Py_VISIT(traverse_module_state->__pyx_tuple__34);
  Py_VISIT(traverse_module_state->__pyx_tuple__35);
  Py_VISIT(traverse_module_state->__pyx_tuple__37);
  Py_VISIT(traverse_module_state->__pyx_tuple__38);
  Py_VISIT(traverse_module_state->__pyx_tuple__40);
  Py_VISIT(traverse_module_state->__pyx_tuple__48);
  Py_VISIT(traverse_module_state->__pyx_tuple__50);
  Py_VISIT(traverse_module_state->__pyx_tuple__52);
  Py_VISIT(traverse_module_state->__pyx_tuple__54);
  Py_VISIT(traverse_module_state->__pyx_tuple__57);
  Py_VISIT(traverse_module_state->__pyx_tuple__58);
  Py_VISIT(traverse_module_state->__pyx_tuple__61);
  Py_VISIT(traverse_module_state->__pyx_tuple__64);
  Py_VISIT(traverse_module_state->__pyx_tuple__66);
  Py_VISIT(traverse_module_state->__pyx_tuple__68);
  Py_VISIT(traverse_module_state->__pyx_tuple__70);
  Py_VISIT(traverse_module_state->__pyx_tuple__73);
  Py_VISIT(traverse_module_state->__pyx_tuple__75);
  Py_VISIT(traverse_module_state->__pyx_tuple__80);
  Py_VISIT(traverse_module_state->__pyx_tuple__83);
  Py_VISIT(traverse_module_state->__pyx_codeobj__12);
  Py_VISIT(traverse_module_state->__pyx_codeobj__13);
  Py_VISIT(traverse_module_state->__pyx_codeobj__14);
  Py_VISIT(traverse_module_state->__pyx_codeobj__24);
  Py_VISIT(traverse_module_state->__pyx_codeobj__32);
  Py_VISIT(traverse_module_state->__pyx_codeobj__39);
  Py_VISIT(traverse_module_state->__pyx_codeobj__41);
  Py_VISIT(traverse_module_state->__pyx_codeobj__42);
  Py_VISIT(traverse_module_state->__pyx_codeobj__43);
  Py_VISIT(traverse_module_state->__pyx_codeobj__44);
  Py_VISIT(traverse_module_state->__pyx_codeobj__45);
  Py_VISIT(traverse_module_state->__pyx_codeobj__46);
  Py_VISIT(traverse_module_state->__pyx_codeobj__47);
  Py_VISIT(traverse_module_state->__pyx_codeobj__49);
  Py_VISIT(traverse_module_state->__pyx_codeobj__51);
  Py_VISIT(traverse_module_state->__pyx_codeobj__53);
  Py_VISIT(traverse_module_state->__pyx_codeobj__59);
  Py_VISIT(traverse_module_state->__pyx_codeobj__62);
  Py_VISIT(traverse_module_state->__pyx_codeobj__63);
  Py_VISIT(traverse_module_state->__pyx_codeobj__65);
  Py_VISIT(traverse_module_state->__pyx_codeobj__67);
  Py_VISIT(traverse_module_state->__pyx_codeobj__69);
  Py_VISIT(traverse_module_state->__pyx_codeobj__71);
  Py_VISIT(traverse_module_state->__pyx_codeobj__72);
  Py_VISIT(traverse_module_state->__pyx_codeobj__74);
  Py_VISIT(traverse_module_state->__pyx_codeobj__76);
  Py_VISIT(traverse_module_state->__pyx_codeobj__77);
  Py_VISIT(traverse_module_state->__pyx_codeobj__78);
  Py_VISIT(traverse_module_state->__pyx_codeobj__79);
  Py_VISIT(traverse_module_state->__pyx_codeobj__81);
  Py_VISIT(traverse_module_state->__pyx_codeobj__82);
  Py_VISIT(traverse_module_state->__pyx_codeobj__84);
  Py_VISIT(traverse_module_state->__pyx_codeobj__85);
  Py_VISIT(traverse_module_state->__pyx_codeobj__86);
  Py_VISIT(traverse_module_state->__pyx_codeobj__87);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#ifdef __Pyx_Generator_USED
#define __pyx_GeneratorType __pyx_mstate_global->__pyx_GeneratorType
#endif
#ifdef __Pyx_IterableCoroutine_USED
#define __pyx_IterableCoroutineType __pyx_mstate_global->__pyx_IterableCoroutineType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineAwaitType __pyx_mstate_global->__pyx_CoroutineAwaitType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineType __pyx_mstate_global->__pyx_CoroutineType
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_6pubsub_PubSub __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_6pubsub_PubSub
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult
#define __pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4bool_bool __pyx_mstate_global->__pyx_ptype_7cpython_4bool_bool
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_7complex_complex __pyx_mstate_global->__pyx_ptype_7cpython_7complex_complex
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5numpy_dtype __pyx_mstate_global->__pyx_ptype_5numpy_dtype
#define __pyx_ptype_5numpy_flatiter __pyx_mstate_global->__pyx_ptype_5numpy_flatiter
#define __pyx_ptype_5numpy_broadcast __pyx_mstate_global->__pyx_ptype_5numpy_broadcast
#define __pyx_ptype_5numpy_ndarray __pyx_mstate_global->__pyx_ptype_5numpy_ndarray
#define __pyx_ptype_5numpy_generic __pyx_mstate_global->__pyx_ptype_5numpy_generic
#define __pyx_ptype_5numpy_number __pyx_mstate_global->__pyx_ptype_5numpy_number
#define __pyx_ptype_5numpy_integer __pyx_mstate_global->__pyx_ptype_5numpy_integer
#define __pyx_ptype_5numpy_signedinteger __pyx_mstate_global->__pyx_ptype_5numpy_signedinteger
#define __pyx_ptype_5numpy_unsignedinteger __pyx_mstate_global->__pyx_ptype_5numpy_unsignedinteger
#define __pyx_ptype_5numpy_inexact __pyx_mstate_global->__pyx_ptype_5numpy_inexact
#define __pyx_ptype_5numpy_floating __pyx_mstate_global->__pyx_ptype_5numpy_floating
#define __pyx_ptype_5numpy_complexfloating __pyx_mstate_global->__pyx_ptype_5numpy_complexfloating
#define __pyx_ptype_5numpy_flexible __pyx_mstate_global->__pyx_ptype_5numpy_flexible
#define __pyx_ptype_5numpy_character __pyx_mstate_global->__pyx_ptype_5numpy_character
#define __pyx_ptype_5numpy_ufunc __pyx_mstate_global->__pyx_ptype_5numpy_ufunc
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_5clock_Clock __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_5clock_Clock
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#define __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams __pyx_mstate_global->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams
#define __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange __pyx_mstate_global->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange
#define __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder __pyx_mstate_global->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder
#define __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener __pyx_mstate_global->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener
#define __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener __pyx_mstate_global->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener
#define __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network __pyx_mstate_global->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network
#define __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network __pyx_mstate_global->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network
#define __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network __pyx_mstate_global->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network
#define __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all __pyx_mstate_global->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all
#define __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async __pyx_mstate_global->__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async
#endif
#define __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams
#define __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange
#define __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder
#define __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener
#define __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener
#define __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network
#define __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network
#define __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network
#define __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all
#define __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async
#define __pyx_kp_u_0 __pyx_mstate_global->__pyx_kp_u_0
#define __pyx_kp_u_02x __pyx_mstate_global->__pyx_kp_u_02x
#define __pyx_kp_u_1e __pyx_mstate_global->__pyx_kp_u_1e
#define __pyx_kp_u_1e_10 __pyx_mstate_global->__pyx_kp_u_1e_10
#define __pyx_kp_u_1e_2 __pyx_mstate_global->__pyx_kp_u_1e_2
#define __pyx_kp_u_1e_7 __pyx_mstate_global->__pyx_kp_u_1e_7
#define __pyx_kp_u_7g __pyx_mstate_global->__pyx_kp_u_7g
#define __pyx_kp_u_8g __pyx_mstate_global->__pyx_kp_u_8g
#define __pyx_kp_u_Account_balance_does_not_have_as __pyx_mstate_global->__pyx_kp_u_Account_balance_does_not_have_as
#define __pyx_n_s_AssertionError __pyx_mstate_global->__pyx_n_s_AssertionError
#define __pyx_n_s_BUY __pyx_mstate_global->__pyx_n_s_BUY
#define __pyx_n_u_BUY __pyx_mstate_global->__pyx_n_u_BUY
#define __pyx_n_s_BUY_ORDER_COMPLETED_EVENT_TAG __pyx_mstate_global->__pyx_n_s_BUY_ORDER_COMPLETED_EVENT_TAG
#define __pyx_n_s_BudgetChecker __pyx_mstate_global->__pyx_n_s_BudgetChecker
#define __pyx_n_s_BuyOrderCompleted __pyx_mstate_global->__pyx_n_s_BuyOrderCompleted
#define __pyx_n_s_BuyOrderCompletedEvent __pyx_mstate_global->__pyx_n_s_BuyOrderCompletedEvent
#define __pyx_n_s_BuyOrderCreated __pyx_mstate_global->__pyx_n_s_BuyOrderCreated
#define __pyx_n_s_BuyOrderCreatedEvent __pyx_mstate_global->__pyx_n_s_BuyOrderCreatedEvent
#define __pyx_n_s_CONNECTED __pyx_mstate_global->__pyx_n_s_CONNECTED
#define __pyx_n_s_Callable __pyx_mstate_global->__pyx_n_s_Callable
#define __pyx_n_s_CancellationResult __pyx_mstate_global->__pyx_n_s_CancellationResult
#define __pyx_n_s_ClientConfigAdapter __pyx_mstate_global->__pyx_n_s_ClientConfigAdapter
#define __pyx_n_s_Clock __pyx_mstate_global->__pyx_n_s_Clock
#define __pyx_n_s_CompositeOrderBook __pyx_mstate_global->__pyx_n_s_CompositeOrderBook
#define __pyx_n_s_Decimal __pyx_mstate_global->__pyx_n_s_Decimal
#define __pyx_n_s_Dict __pyx_mstate_global->__pyx_n_s_Dict
#define __pyx_kp_s_Dict_str_Decimal __pyx_mstate_global->__pyx_kp_s_Dict_str_Decimal
#define __pyx_n_s_DummyMetricsCollector __pyx_mstate_global->__pyx_n_s_DummyMetricsCollector
#define __pyx_kp_u_Error_call_trade_listener __pyx_mstate_global->__pyx_kp_u_Error_call_trade_listener
#define __pyx_kp_u_Error_canceling_order __pyx_mstate_global->__pyx_kp_u_Error_canceling_order
#define __pyx_kp_u_Error_executing_queued_order __pyx_mstate_global->__pyx_kp_u_Error_executing_queued_order
#define __pyx_kp_u_Error_processing_limit_order __pyx_mstate_global->__pyx_kp_u_Error_processing_limit_order
#define __pyx_n_s_ExchangeBase __pyx_mstate_global->__pyx_n_s_ExchangeBase
#define __pyx_n_s_ImportError __pyx_mstate_global->__pyx_n_s_ImportError
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2
#define __pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3 __pyx_mstate_global->__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3
#define __pyx_kp_u_Insufficient __pyx_mstate_global->__pyx_kp_u_Insufficient
#define __pyx_n_s_LIMIT __pyx_mstate_global->__pyx_n_s_LIMIT
#define __pyx_n_s_LIMIT_MAKER __pyx_mstate_global->__pyx_n_s_LIMIT_MAKER
#define __pyx_n_s_LimitOrder __pyx_mstate_global->__pyx_n_s_LimitOrder
#define __pyx_n_s_List __pyx_mstate_global->__pyx_n_s_List
#define __pyx_kp_s_List_CancellationResult __pyx_mstate_global->__pyx_kp_s_List_CancellationResult
#define __pyx_n_s_MARKET __pyx_mstate_global->__pyx_n_s_MARKET
#define __pyx_n_s_MARKET_BUY_ORDER_CREATED_EVENT_T __pyx_mstate_global->__pyx_n_s_MARKET_BUY_ORDER_CREATED_EVENT_T
#define __pyx_n_s_MARKET_ORDER_CANCELED_EVENT_TAG __pyx_mstate_global->__pyx_n_s_MARKET_ORDER_CANCELED_EVENT_TAG
#define __pyx_n_s_MARKET_ORDER_FAILURE_EVENT_TAG __pyx_mstate_global->__pyx_n_s_MARKET_ORDER_FAILURE_EVENT_TAG
#define __pyx_n_s_MARKET_SELL_ORDER_CREATED_EVENT __pyx_mstate_global->__pyx_n_s_MARKET_SELL_ORDER_CREATED_EVENT
#define __pyx_n_s_MarketEvent __pyx_mstate_global->__pyx_n_s_MarketEvent
#define __pyx_n_s_MarketOrderFailureEvent __pyx_mstate_global->__pyx_n_s_MarketOrderFailureEvent
#define __pyx_n_b_NIL __pyx_mstate_global->__pyx_n_b_NIL
#define __pyx_n_s_NetworkStatus __pyx_mstate_global->__pyx_n_s_NetworkStatus
#define __pyx_kp_u_No_order_book_exists_for __pyx_mstate_global->__pyx_kp_u_No_order_book_exists_for
#define __pyx_kp_u_None __pyx_mstate_global->__pyx_kp_u_None
#define __pyx_kp_u_Not_enough __pyx_mstate_global->__pyx_kp_u_Not_enough
#define __pyx_n_s_ORDER_BOOK_TRADE_EVENT_TAG __pyx_mstate_global->__pyx_n_s_ORDER_BOOK_TRADE_EVENT_TAG
#define __pyx_n_s_ORDER_FILLED_EVENT_TAG __pyx_mstate_global->__pyx_n_s_ORDER_FILLED_EVENT_TAG
#define __pyx_n_s_Optional __pyx_mstate_global->__pyx_n_s_Optional
#define __pyx_kp_s_Optional_bool __pyx_mstate_global->__pyx_kp_s_Optional_bool
#define __pyx_n_s_OrderBook __pyx_mstate_global->__pyx_n_s_OrderBook
#define __pyx_n_s_OrderBookEvent __pyx_mstate_global->__pyx_n_s_OrderBookEvent
#define __pyx_n_s_OrderBookMarketOrderFillListener __pyx_mstate_global->__pyx_n_s_OrderBookMarketOrderFillListener
#define __pyx_n_s_OrderBookMarketOrderFillListener_2 __pyx_mstate_global->__pyx_n_s_OrderBookMarketOrderFillListener_2
#define __pyx_n_s_OrderBookMarketOrderFillListener_3 __pyx_mstate_global->__pyx_n_s_OrderBookMarketOrderFillListener_3
#define __pyx_n_s_OrderBookTracker __pyx_mstate_global->__pyx_n_s_OrderBookTracker
#define __pyx_n_s_OrderBookTradeEvent __pyx_mstate_global->__pyx_n_s_OrderBookTradeEvent
#define __pyx_n_s_OrderBookTradeListener __pyx_mstate_global->__pyx_n_s_OrderBookTradeListener
#define __pyx_n_s_OrderBookTradeListener___reduce __pyx_mstate_global->__pyx_n_s_OrderBookTradeListener___reduce
#define __pyx_n_s_OrderBookTradeListener___setstat __pyx_mstate_global->__pyx_n_s_OrderBookTradeListener___setstat
#define __pyx_n_s_OrderCancelled __pyx_mstate_global->__pyx_n_s_OrderCancelled
#define __pyx_n_s_OrderCancelledEvent __pyx_mstate_global->__pyx_n_s_OrderCancelledEvent
#define __pyx_n_s_OrderCandidate __pyx_mstate_global->__pyx_n_s_OrderCandidate
#define __pyx_n_s_OrderFailure __pyx_mstate_global->__pyx_n_s_OrderFailure
#define __pyx_n_s_OrderFilled __pyx_mstate_global->__pyx_n_s_OrderFilled
#define __pyx_n_s_OrderFilledEvent __pyx_mstate_global->__pyx_n_s_OrderFilledEvent
#define __pyx_n_s_OrderType __pyx_mstate_global->__pyx_n_s_OrderType
#define __pyx_n_u_PaperTrade __pyx_mstate_global->__pyx_n_u_PaperTrade
#define __pyx_n_s_PaperTradeExchange __pyx_mstate_global->__pyx_n_s_PaperTradeExchange
#define __pyx_n_s_PaperTradeExchange___reduce_cyth __pyx_mstate_global->__pyx_n_s_PaperTradeExchange___reduce_cyth
#define __pyx_n_s_PaperTradeExchange___setstate_cy __pyx_mstate_global->__pyx_n_s_PaperTradeExchange___setstate_cy
#define __pyx_n_s_PaperTradeExchange_buy __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_buy
#define __pyx_n_s_PaperTradeExchange_cancel __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_cancel
#define __pyx_n_s_PaperTradeExchange_cancel_all __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_cancel_all
#define __pyx_n_s_PaperTradeExchange_check_network __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_check_network
#define __pyx_n_s_PaperTradeExchange_get_all_balan __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_get_all_balan
#define __pyx_n_s_PaperTradeExchange_get_available __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_get_available
#define __pyx_n_s_PaperTradeExchange_get_fee __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_get_fee
#define __pyx_n_s_PaperTradeExchange_get_maker_ord __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_get_maker_ord
#define __pyx_n_s_PaperTradeExchange_get_order_boo __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_get_order_boo
#define __pyx_n_s_PaperTradeExchange_get_order_pri __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_get_order_pri
#define __pyx_n_s_PaperTradeExchange_get_price __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_get_price
#define __pyx_n_s_PaperTradeExchange_get_taker_ord __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_get_taker_ord
#define __pyx_n_s_PaperTradeExchange_init_paper_tr __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_init_paper_tr
#define __pyx_n_s_PaperTradeExchange_match_trade_t __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_match_trade_t
#define __pyx_n_s_PaperTradeExchange_random_order __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_random_order
#define __pyx_n_s_PaperTradeExchange_sell __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_sell
#define __pyx_n_s_PaperTradeExchange_set_balance __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_set_balance
#define __pyx_n_s_PaperTradeExchange_split_trading __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_split_trading
#define __pyx_n_s_PaperTradeExchange_start_network __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_start_network
#define __pyx_n_s_PaperTradeExchange_stop_network __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_stop_network
#define __pyx_n_s_PaperTradeExchange_trigger_event __pyx_mstate_global->__pyx_n_s_PaperTradeExchange_trigger_event
#define __pyx_n_s_PickleError __pyx_mstate_global->__pyx_n_s_PickleError
#define __pyx_kp_u_QuantizationParams __pyx_mstate_global->__pyx_kp_u_QuantizationParams
#define __pyx_n_s_QuantizationParams_2 __pyx_mstate_global->__pyx_n_s_QuantizationParams_2
#define __pyx_n_s_QuantizationParams___reduce_cyth __pyx_mstate_global->__pyx_n_s_QuantizationParams___reduce_cyth
#define __pyx_n_s_QuantizationParams___setstate_cy __pyx_mstate_global->__pyx_n_s_QuantizationParams___setstate_cy
#define __pyx_kp_u_QueuedOrder __pyx_mstate_global->__pyx_kp_u_QueuedOrder
#define __pyx_n_s_QueuedOrder_2 __pyx_mstate_global->__pyx_n_s_QueuedOrder_2
#define __pyx_n_s_QueuedOrder___reduce_cython __pyx_mstate_global->__pyx_n_s_QueuedOrder___reduce_cython
#define __pyx_n_s_QueuedOrder___setstate_cython __pyx_mstate_global->__pyx_n_s_QueuedOrder___setstate_cython
#define __pyx_n_s_ROUND_DOWN __pyx_mstate_global->__pyx_n_s_ROUND_DOWN
#define __pyx_n_s_SELL __pyx_mstate_global->__pyx_n_s_SELL
#define __pyx_n_s_SELL_ORDER_COMPLETED_EVENT_TAG __pyx_mstate_global->__pyx_n_s_SELL_ORDER_COMPLETED_EVENT_TAG
#define __pyx_n_s_SellOrderCompleted __pyx_mstate_global->__pyx_n_s_SellOrderCompleted
#define __pyx_n_s_SellOrderCompletedEvent __pyx_mstate_global->__pyx_n_s_SellOrderCompletedEvent
#define __pyx_n_s_SellOrderCreated __pyx_mstate_global->__pyx_n_s_SellOrderCreated
#define __pyx_n_s_SellOrderCreatedEvent __pyx_mstate_global->__pyx_n_s_SellOrderCreatedEvent
#define __pyx_n_s_TRADE_EXECUTION_DELAY __pyx_mstate_global->__pyx_n_s_TRADE_EXECUTION_DELAY
#define __pyx_n_s_TYPE_CHECKING __pyx_mstate_global->__pyx_n_s_TYPE_CHECKING
#define __pyx_n_s_TradeEvent __pyx_mstate_global->__pyx_n_s_TradeEvent
#define __pyx_n_s_TradeType __pyx_mstate_global->__pyx_n_s_TradeType
#define __pyx_n_s_TradingPair __pyx_mstate_global->__pyx_n_s_TradingPair
#define __pyx_kp_u_Trading_pair __pyx_mstate_global->__pyx_kp_u_Trading_pair
#define __pyx_n_s_Tuple __pyx_mstate_global->__pyx_n_s_Tuple
#define __pyx_kp_s_Tuple_str_str __pyx_mstate_global->__pyx_kp_s_Tuple_str_str
#define __pyx_n_s_TypeError __pyx_mstate_global->__pyx_n_s_TypeError
#define __pyx_n_s_ValueError __pyx_mstate_global->__pyx_n_s_ValueError
#define __pyx_kp_u__10 __pyx_mstate_global->__pyx_kp_u__10
#define __pyx_kp_u__11 __pyx_mstate_global->__pyx_kp_u__11
#define __pyx_kp_u__15 __pyx_mstate_global->__pyx_kp_u__15
#define __pyx_kp_u__20 __pyx_mstate_global->__pyx_kp_u__20
#define __pyx_kp_u__22 __pyx_mstate_global->__pyx_kp_u__22
#define __pyx_kp_u__26 __pyx_mstate_global->__pyx_kp_u__26
#define __pyx_kp_u__3 __pyx_mstate_global->__pyx_kp_u__3
#define __pyx_n_s__36 __pyx_mstate_global->__pyx_n_s__36
#define __pyx_kp_u__4 __pyx_mstate_global->__pyx_kp_u__4
#define __pyx_kp_u__5 __pyx_mstate_global->__pyx_kp_u__5
#define __pyx_kp_u__6 __pyx_mstate_global->__pyx_kp_u__6
#define __pyx_n_s__88 __pyx_mstate_global->__pyx_n_s__88
#define __pyx_kp_u__9 __pyx_mstate_global->__pyx_kp_u__9
#define __pyx_n_s_all __pyx_mstate_global->__pyx_n_s_all
#define __pyx_n_s_amount __pyx_mstate_global->__pyx_n_s_amount
#define __pyx_n_s_append __pyx_mstate_global->__pyx_n_s_append
#define __pyx_n_s_args __pyx_mstate_global->__pyx_n_s_args
#define __pyx_n_s_asyncio __pyx_mstate_global->__pyx_n_s_asyncio
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_n_s_asyncio_tasks __pyx_mstate_global->__pyx_n_s_asyncio_tasks
#define __pyx_kp_u_available __pyx_mstate_global->__pyx_kp_u_available
#define __pyx_n_s_available_balances __pyx_mstate_global->__pyx_n_s_available_balances
#define __pyx_kp_u_available_vs __pyx_mstate_global->__pyx_kp_u_available_vs
#define __pyx_n_s_await __pyx_mstate_global->__pyx_n_s_await
#define __pyx_n_s_balance __pyx_mstate_global->__pyx_n_s_balance
#define __pyx_kp_u_balance_available_for_buy_order __pyx_mstate_global->__pyx_kp_u_balance_available_for_buy_order
#define __pyx_kp_u_balance_available_for_sell_orde __pyx_mstate_global->__pyx_kp_u_balance_available_for_sell_orde
#define __pyx_kp_u_balance_to_fill_limit_buy_order __pyx_mstate_global->__pyx_kp_u_balance_to_fill_limit_buy_order
#define __pyx_kp_u_balance_to_fill_limit_sell_orde __pyx_mstate_global->__pyx_kp_u_balance_to_fill_limit_sell_orde
#define __pyx_n_s_base_asset __pyx_mstate_global->__pyx_n_s_base_asset
#define __pyx_n_s_base_currency __pyx_mstate_global->__pyx_n_s_base_currency
#define __pyx_n_s_bool __pyx_mstate_global->__pyx_n_s_bool
#define __pyx_n_s_build_trade_fee __pyx_mstate_global->__pyx_n_s_build_trade_fee
#define __pyx_n_s_buy __pyx_mstate_global->__pyx_n_s_buy
#define __pyx_n_u_buy __pyx_mstate_global->__pyx_n_u_buy
#define __pyx_n_s_cancel __pyx_mstate_global->__pyx_n_s_cancel
#define __pyx_n_s_cancel_all __pyx_mstate_global->__pyx_n_s_cancel_all
#define __pyx_n_s_cancellation_results __pyx_mstate_global->__pyx_n_s_cancellation_results
#define __pyx_n_s_ceil __pyx_mstate_global->__pyx_n_s_ceil
#define __pyx_n_s_check_network __pyx_mstate_global->__pyx_n_s_check_network
#define __pyx_n_s_choice __pyx_mstate_global->__pyx_n_s_choice
#define __pyx_n_s_class_getitem __pyx_mstate_global->__pyx_n_s_class_getitem
#define __pyx_n_s_client_config_map __pyx_mstate_global->__pyx_n_s_client_config_map
#define __pyx_n_s_client_order_id __pyx_mstate_global->__pyx_n_s_client_order_id
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_close __pyx_mstate_global->__pyx_n_s_close
#define __pyx_n_s_cls __pyx_mstate_global->__pyx_n_s_cls
#define __pyx_n_s_collections __pyx_mstate_global->__pyx_n_s_collections
#define __pyx_n_s_convert_from_exchange_trading_pa __pyx_mstate_global->__pyx_n_s_convert_from_exchange_trading_pa
#define __pyx_n_s_convert_to_exchange_trading_pair __pyx_mstate_global->__pyx_n_s_convert_to_exchange_trading_pair
#define __pyx_n_s_create_timestamp __pyx_mstate_global->__pyx_n_s_create_timestamp
#define __pyx_n_s_currency __pyx_mstate_global->__pyx_n_s_currency
#define __pyx_n_s_data_source __pyx_mstate_global->__pyx_n_s_data_source
#define __pyx_n_s_decimal __pyx_mstate_global->__pyx_n_s_decimal
#define __pyx_n_s_defaultdict __pyx_mstate_global->__pyx_n_s_defaultdict
#define __pyx_n_s_deque __pyx_mstate_global->__pyx_n_s_deque
#define __pyx_n_s_dict __pyx_mstate_global->__pyx_n_s_dict
#define __pyx_n_s_dict_2 __pyx_mstate_global->__pyx_n_s_dict_2
#define __pyx_kp_u_disable __pyx_mstate_global->__pyx_kp_u_disable
#define __pyx_kp_u_does_not_existing_in_current_da __pyx_mstate_global->__pyx_kp_u_does_not_existing_in_current_da
#define __pyx_kp_u_enable __pyx_mstate_global->__pyx_kp_u_enable
#define __pyx_n_s_encode __pyx_mstate_global->__pyx_n_s_encode
#define __pyx_n_s_error __pyx_mstate_global->__pyx_n_s_error
#define __pyx_n_s_event __pyx_mstate_global->__pyx_n_s_event
#define __pyx_n_s_event_object __pyx_mstate_global->__pyx_n_s_event_object
#define __pyx_n_s_event_tag __pyx_mstate_global->__pyx_n_s_event_tag
#define __pyx_n_s_exc_info __pyx_mstate_global->__pyx_n_s_exc_info
#define __pyx_n_s_exchange __pyx_mstate_global->__pyx_n_s_exchange
#define __pyx_n_s_exchange_name __pyx_mstate_global->__pyx_n_s_exchange_name
#define __pyx_n_s_exchange_trade_id __pyx_mstate_global->__pyx_n_s_exchange_trade_id
#define __pyx_n_s_float __pyx_mstate_global->__pyx_n_s_float
#define __pyx_n_s_from_total_balances __pyx_mstate_global->__pyx_n_s_from_total_balances
#define __pyx_kp_u_gc __pyx_mstate_global->__pyx_kp_u_gc
#define __pyx_n_s_get __pyx_mstate_global->__pyx_n_s_get
#define __pyx_n_s_get_all_balances __pyx_mstate_global->__pyx_n_s_get_all_balances
#define __pyx_n_s_get_available_balance __pyx_mstate_global->__pyx_n_s_get_available_balance
#define __pyx_n_s_get_fee __pyx_mstate_global->__pyx_n_s_get_fee
#define __pyx_n_s_get_maker_order_type __pyx_mstate_global->__pyx_n_s_get_maker_order_type
#define __pyx_n_s_get_order_book __pyx_mstate_global->__pyx_n_s_get_order_book
#define __pyx_n_s_get_order_price_quantum __pyx_mstate_global->__pyx_n_s_get_order_price_quantum
#define __pyx_n_s_get_price __pyx_mstate_global->__pyx_n_s_get_price
#define __pyx_n_s_get_taker_order_type __pyx_mstate_global->__pyx_n_s_get_taker_order_type
#define __pyx_n_s_getstate __pyx_mstate_global->__pyx_n_s_getstate
#define __pyx_n_s_hummingbot_client_config_config __pyx_mstate_global->__pyx_n_s_hummingbot_client_config_config
#define __pyx_n_s_hummingbot_connector_budget_chec __pyx_mstate_global->__pyx_n_s_hummingbot_connector_budget_chec
#define __pyx_n_s_hummingbot_connector_connector_m __pyx_mstate_global->__pyx_n_s_hummingbot_connector_connector_m
#define __pyx_n_s_hummingbot_connector_exchange_ba __pyx_mstate_global->__pyx_n_s_hummingbot_connector_exchange_ba
#define __pyx_n_s_hummingbot_connector_exchange_pa __pyx_mstate_global->__pyx_n_s_hummingbot_connector_exchange_pa
#define __pyx_n_s_hummingbot_connector_exchange_pa_2 __pyx_mstate_global->__pyx_n_s_hummingbot_connector_exchange_pa_2
#define __pyx_kp_s_hummingbot_connector_exchange_pa_3 __pyx_mstate_global->__pyx_kp_s_hummingbot_connector_exchange_pa_3
#define __pyx_n_s_hummingbot_core_clock __pyx_mstate_global->__pyx_n_s_hummingbot_core_clock
#define __pyx_n_s_hummingbot_core_data_type_cancel __pyx_mstate_global->__pyx_n_s_hummingbot_core_data_type_cancel
#define __pyx_n_s_hummingbot_core_data_type_common __pyx_mstate_global->__pyx_n_s_hummingbot_core_data_type_common
#define __pyx_n_s_hummingbot_core_data_type_compos __pyx_mstate_global->__pyx_n_s_hummingbot_core_data_type_compos
#define __pyx_n_s_hummingbot_core_data_type_limit __pyx_mstate_global->__pyx_n_s_hummingbot_core_data_type_limit
#define __pyx_n_s_hummingbot_core_data_type_order __pyx_mstate_global->__pyx_n_s_hummingbot_core_data_type_order
#define __pyx_n_s_hummingbot_core_data_type_order_2 __pyx_mstate_global->__pyx_n_s_hummingbot_core_data_type_order_2
#define __pyx_n_s_hummingbot_core_event_events __pyx_mstate_global->__pyx_n_s_hummingbot_core_event_events
#define __pyx_n_s_hummingbot_core_network_iterator __pyx_mstate_global->__pyx_n_s_hummingbot_core_network_iterator
#define __pyx_n_s_hummingbot_core_utils_async_util __pyx_mstate_global->__pyx_n_s_hummingbot_core_utils_async_util
#define __pyx_n_s_hummingbot_core_utils_estimate_f __pyx_mstate_global->__pyx_n_s_hummingbot_core_utils_estimate_f
#define __pyx_n_s_i __pyx_mstate_global->__pyx_n_s_i
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_init __pyx_mstate_global->__pyx_n_s_init
#define __pyx_n_s_init___locals_lambda __pyx_mstate_global->__pyx_n_s_init___locals_lambda
#define __pyx_n_s_init_paper_trade_market __pyx_mstate_global->__pyx_n_s_init_paper_trade_market
#define __pyx_n_s_initializing __pyx_mstate_global->__pyx_n_s_initializing
#define __pyx_n_s_inspect __pyx_mstate_global->__pyx_n_s_inspect
#define __pyx_n_s_is_buy __pyx_mstate_global->__pyx_n_s_is_buy
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_n_s_is_finite __pyx_mstate_global->__pyx_n_s_is_finite
#define __pyx_n_s_is_maker __pyx_mstate_global->__pyx_n_s_is_maker
#define __pyx_kp_u_isenabled __pyx_mstate_global->__pyx_kp_u_isenabled
#define __pyx_n_s_items __pyx_mstate_global->__pyx_n_s_items
#define __pyx_n_s_keys __pyx_mstate_global->__pyx_n_s_keys
#define __pyx_n_s_kwargs __pyx_mstate_global->__pyx_n_s_kwargs
#define __pyx_n_s_limit_orders __pyx_mstate_global->__pyx_n_s_limit_orders
#define __pyx_n_s_limit_orders_map_ptr __pyx_mstate_global->__pyx_n_s_limit_orders_map_ptr
#define __pyx_n_s_log10 __pyx_mstate_global->__pyx_n_s_log10
#define __pyx_n_s_logger __pyx_mstate_global->__pyx_n_s_logger
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_s_market __pyx_mstate_global->__pyx_n_s_market
#define __pyx_n_s_match_trade_to_limit_orders __pyx_mstate_global->__pyx_n_s_match_trade_to_limit_orders
#define __pyx_n_s_math __pyx_mstate_global->__pyx_n_s_math
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_s_name_2 __pyx_mstate_global->__pyx_n_s_name_2
#define __pyx_kp_u_needed_vs __pyx_mstate_global->__pyx_kp_u_needed_vs
#define __pyx_n_s_new __pyx_mstate_global->__pyx_n_s_new
#define __pyx_kp_u_numpy_core_multiarray_failed_to __pyx_mstate_global->__pyx_kp_u_numpy_core_multiarray_failed_to
#define __pyx_kp_u_numpy_core_umath_failed_to_impor __pyx_mstate_global->__pyx_kp_u_numpy_core_umath_failed_to_impor
#define __pyx_n_s_on_hold_balances __pyx_mstate_global->__pyx_n_s_on_hold_balances
#define __pyx_n_s_order_book __pyx_mstate_global->__pyx_n_s_order_book
#define __pyx_n_s_order_book_create_function __pyx_mstate_global->__pyx_n_s_order_book_create_function
#define __pyx_n_s_order_book_tracker __pyx_mstate_global->__pyx_n_s_order_book_tracker
#define __pyx_n_s_order_books __pyx_mstate_global->__pyx_n_s_order_books
#define __pyx_n_u_order_books_initialized __pyx_mstate_global->__pyx_n_u_order_books_initialized
#define __pyx_n_s_order_collateral __pyx_mstate_global->__pyx_n_s_order_collateral
#define __pyx_n_s_order_filled_events_from_order_b __pyx_mstate_global->__pyx_n_s_order_filled_events_from_order_b
#define __pyx_n_s_order_id __pyx_mstate_global->__pyx_n_s_order_id
#define __pyx_n_s_order_side __pyx_mstate_global->__pyx_n_s_order_side
#define __pyx_n_s_order_size_decimals __pyx_mstate_global->__pyx_n_s_order_size_decimals
#define __pyx_n_s_order_size_precision __pyx_mstate_global->__pyx_n_s_order_size_precision
#define __pyx_n_s_order_type __pyx_mstate_global->__pyx_n_s_order_type
#define __pyx_n_s_pickle __pyx_mstate_global->__pyx_n_s_pickle
#define __pyx_n_s_popleft __pyx_mstate_global->__pyx_n_s_popleft
#define __pyx_n_s_populate_collateral_entries __pyx_mstate_global->__pyx_n_s_populate_collateral_entries
#define __pyx_n_s_potential_returns __pyx_mstate_global->__pyx_n_s_potential_returns
#define __pyx_n_s_price __pyx_mstate_global->__pyx_n_s_price
#define __pyx_n_s_price_decimals __pyx_mstate_global->__pyx_n_s_price_decimals
#define __pyx_n_s_price_precision __pyx_mstate_global->__pyx_n_s_price_precision
#define __pyx_n_s_ptm_logger __pyx_mstate_global->__pyx_n_s_ptm_logger
#define __pyx_n_s_pyx_PickleError __pyx_mstate_global->__pyx_n_s_pyx_PickleError
#define __pyx_n_s_pyx_checksum __pyx_mstate_global->__pyx_n_s_pyx_checksum
#define __pyx_n_s_pyx_result __pyx_mstate_global->__pyx_n_s_pyx_result
#define __pyx_n_s_pyx_state __pyx_mstate_global->__pyx_n_s_pyx_state
#define __pyx_n_s_pyx_type __pyx_mstate_global->__pyx_n_s_pyx_type
#define __pyx_n_s_pyx_unpickle_OrderBookMarketOr __pyx_mstate_global->__pyx_n_s_pyx_unpickle_OrderBookMarketOr
#define __pyx_n_s_pyx_unpickle_OrderBookTradeLis __pyx_mstate_global->__pyx_n_s_pyx_unpickle_OrderBookTradeLis
#define __pyx_n_s_pyx_unpickle_QuantizationParam __pyx_mstate_global->__pyx_n_s_pyx_unpickle_QuantizationParam
#define __pyx_n_s_pyx_unpickle_QueuedOrder __pyx_mstate_global->__pyx_n_s_pyx_unpickle_QueuedOrder
#define __pyx_n_s_pyx_vtable __pyx_mstate_global->__pyx_n_s_pyx_vtable
#define __pyx_n_s_quantity __pyx_mstate_global->__pyx_n_s_quantity
#define __pyx_n_s_quantize __pyx_mstate_global->__pyx_n_s_quantize
#define __pyx_n_s_quote_asset __pyx_mstate_global->__pyx_n_s_quote_asset
#define __pyx_n_s_quote_currency __pyx_mstate_global->__pyx_n_s_quote_currency
#define __pyx_n_s_random __pyx_mstate_global->__pyx_n_s_random
#define __pyx_n_s_random_order_id __pyx_mstate_global->__pyx_n_s_random_order_id
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_s_ready __pyx_mstate_global->__pyx_n_s_ready
#define __pyx_n_s_record_filled_order __pyx_mstate_global->__pyx_n_s_record_filled_order
#define __pyx_n_s_reduce __pyx_mstate_global->__pyx_n_s_reduce
#define __pyx_n_s_reduce_cython __pyx_mstate_global->__pyx_n_s_reduce_cython
#define __pyx_n_s_reduce_ex __pyx_mstate_global->__pyx_n_s_reduce_ex
#define __pyx_kp_u_required_for_the_order __pyx_mstate_global->__pyx_kp_u_required_for_the_order
#define __pyx_n_s_results __pyx_mstate_global->__pyx_n_s_results
#define __pyx_n_s_return __pyx_mstate_global->__pyx_n_s_return
#define __pyx_n_s_rounding __pyx_mstate_global->__pyx_n_s_rounding
#define __pyx_n_s_s_decimal_0 __pyx_mstate_global->__pyx_n_s_s_decimal_0
#define __pyx_n_s_safe_ensure_future __pyx_mstate_global->__pyx_n_s_safe_ensure_future
#define __pyx_n_s_self __pyx_mstate_global->__pyx_n_s_self
#define __pyx_kp_s_self__ask_limit_orders_self__bid __pyx_mstate_global->__pyx_kp_s_self__ask_limit_orders_self__bid
#define __pyx_n_s_sell __pyx_mstate_global->__pyx_n_s_sell
#define __pyx_n_u_sell __pyx_mstate_global->__pyx_n_u_sell
#define __pyx_n_s_send __pyx_mstate_global->__pyx_n_s_send
#define __pyx_n_s_set_balance __pyx_mstate_global->__pyx_n_s_set_balance
#define __pyx_n_s_set_order_book_tracker __pyx_mstate_global->__pyx_n_s_set_order_book_tracker
#define __pyx_n_s_setstate __pyx_mstate_global->__pyx_n_s_setstate
#define __pyx_n_s_setstate_cython __pyx_mstate_global->__pyx_n_s_setstate_cython
#define __pyx_n_s_simulate_buy __pyx_mstate_global->__pyx_n_s_simulate_buy
#define __pyx_n_s_simulate_sell __pyx_mstate_global->__pyx_n_s_simulate_sell
#define __pyx_n_s_sleep __pyx_mstate_global->__pyx_n_s_sleep
#define __pyx_n_s_spec __pyx_mstate_global->__pyx_n_s_spec
#define __pyx_n_s_split_trading_pair __pyx_mstate_global->__pyx_n_s_split_trading_pair
#define __pyx_n_s_start __pyx_mstate_global->__pyx_n_s_start
#define __pyx_n_s_start_network __pyx_mstate_global->__pyx_n_s_start_network
#define __pyx_n_s_state __pyx_mstate_global->__pyx_n_s_state
#define __pyx_n_s_status_dict __pyx_mstate_global->__pyx_n_s_status_dict
#define __pyx_n_s_stop __pyx_mstate_global->__pyx_n_s_stop
#define __pyx_n_s_stop_network __pyx_mstate_global->__pyx_n_s_stop_network
#define __pyx_n_s_str __pyx_mstate_global->__pyx_n_s_str
#define __pyx_kp_s_stringsource __pyx_mstate_global->__pyx_kp_s_stringsource
#define __pyx_n_s_super __pyx_mstate_global->__pyx_n_s_super
#define __pyx_n_s_target_market __pyx_mstate_global->__pyx_n_s_target_market
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_throw __pyx_mstate_global->__pyx_n_s_throw
#define __pyx_n_s_time __pyx_mstate_global->__pyx_n_s_time
#define __pyx_n_s_timeout_seconds __pyx_mstate_global->__pyx_n_s_timeout_seconds
#define __pyx_n_s_trading_pair __pyx_mstate_global->__pyx_n_s_trading_pair
#define __pyx_n_s_trading_pair_str __pyx_mstate_global->__pyx_n_s_trading_pair_str
#define __pyx_n_s_trigger_event_async __pyx_mstate_global->__pyx_n_s_trigger_event_async
#define __pyx_n_s_type __pyx_mstate_global->__pyx_n_s_type
#define __pyx_n_s_typing __pyx_mstate_global->__pyx_n_s_typing
#define __pyx_n_s_update __pyx_mstate_global->__pyx_n_s_update
#define __pyx_n_s_upper __pyx_mstate_global->__pyx_n_s_upper
#define __pyx_n_s_use_setstate __pyx_mstate_global->__pyx_n_s_use_setstate
#define __pyx_n_u_utf8 __pyx_mstate_global->__pyx_n_u_utf8
#define __pyx_n_s_val __pyx_mstate_global->__pyx_n_s_val
#define __pyx_n_s_vals __pyx_mstate_global->__pyx_n_s_vals
#define __pyx_n_s_value __pyx_mstate_global->__pyx_n_s_value
#define __pyx_n_s_values __pyx_mstate_global->__pyx_n_s_values
#define __pyx_n_s_warning __pyx_mstate_global->__pyx_n_s_warning
#define __pyx_float_0_0 __pyx_mstate_global->__pyx_float_0_0
#define __pyx_float_1e6 __pyx_mstate_global->__pyx_float_1e6
#define __pyx_float_5_0 __pyx_mstate_global->__pyx_float_5_0
#define __pyx_float_0_01 __pyx_mstate_global->__pyx_float_0_01
#define __pyx_float_1eneg_7 __pyx_mstate_global->__pyx_float_1eneg_7
#define __pyx_int_0 __pyx_mstate_global->__pyx_int_0
#define __pyx_int_256 __pyx_mstate_global->__pyx_int_256
#define __pyx_int_24020995 __pyx_mstate_global->__pyx_int_24020995
#define __pyx_int_29025389 __pyx_mstate_global->__pyx_int_29025389
#define __pyx_int_57171680 __pyx_mstate_global->__pyx_int_57171680
#define __pyx_int_58563262 __pyx_mstate_global->__pyx_int_58563262
#define __pyx_int_95353964 __pyx_mstate_global->__pyx_int_95353964
#define __pyx_int_126189128 __pyx_mstate_global->__pyx_int_126189128
#define __pyx_int_136360524 __pyx_mstate_global->__pyx_int_136360524
#define __pyx_int_175429262 __pyx_mstate_global->__pyx_int_175429262
#define __pyx_int_251288030 __pyx_mstate_global->__pyx_int_251288030
#define __pyx_k__16 __pyx_mstate_global->__pyx_k__16
#define __pyx_k__17 __pyx_mstate_global->__pyx_k__17
#define __pyx_k__18 __pyx_mstate_global->__pyx_k__18
#define __pyx_k__19 __pyx_mstate_global->__pyx_k__19
#define __pyx_k__27 __pyx_mstate_global->__pyx_k__27
#define __pyx_k__28 __pyx_mstate_global->__pyx_k__28
#define __pyx_k__29 __pyx_mstate_global->__pyx_k__29
#define __pyx_k__30 __pyx_mstate_global->__pyx_k__30
#define __pyx_k__31 __pyx_mstate_global->__pyx_k__31
#define __pyx_k__55 __pyx_mstate_global->__pyx_k__55
#define __pyx_k__56 __pyx_mstate_global->__pyx_k__56
#define __pyx_k__60 __pyx_mstate_global->__pyx_k__60
#define __pyx_tuple_ __pyx_mstate_global->__pyx_tuple_
#define __pyx_tuple__2 __pyx_mstate_global->__pyx_tuple__2
#define __pyx_tuple__7 __pyx_mstate_global->__pyx_tuple__7
#define __pyx_tuple__8 __pyx_mstate_global->__pyx_tuple__8
#define __pyx_tuple__21 __pyx_mstate_global->__pyx_tuple__21
#define __pyx_tuple__23 __pyx_mstate_global->__pyx_tuple__23
#define __pyx_tuple__25 __pyx_mstate_global->__pyx_tuple__25
#define __pyx_tuple__33 __pyx_mstate_global->__pyx_tuple__33
#define __pyx_tuple__34 __pyx_mstate_global->__pyx_tuple__34
#define __pyx_tuple__35 __pyx_mstate_global->__pyx_tuple__35
#define __pyx_tuple__37 __pyx_mstate_global->__pyx_tuple__37
#define __pyx_tuple__38 __pyx_mstate_global->__pyx_tuple__38
#define __pyx_tuple__40 __pyx_mstate_global->__pyx_tuple__40
#define __pyx_tuple__48 __pyx_mstate_global->__pyx_tuple__48
#define __pyx_tuple__50 __pyx_mstate_global->__pyx_tuple__50
#define __pyx_tuple__52 __pyx_mstate_global->__pyx_tuple__52
#define __pyx_tuple__54 __pyx_mstate_global->__pyx_tuple__54
#define __pyx_tuple__57 __pyx_mstate_global->__pyx_tuple__57
#define __pyx_tuple__58 __pyx_mstate_global->__pyx_tuple__58
#define __pyx_tuple__61 __pyx_mstate_global->__pyx_tuple__61
#define __pyx_tuple__64 __pyx_mstate_global->__pyx_tuple__64
#define __pyx_tuple__66 __pyx_mstate_global->__pyx_tuple__66
#define __pyx_tuple__68 __pyx_mstate_global->__pyx_tuple__68
#define __pyx_tuple__70 __pyx_mstate_global->__pyx_tuple__70
#define __pyx_tuple__73 __pyx_mstate_global->__pyx_tuple__73
#define __pyx_tuple__75 __pyx_mstate_global->__pyx_tuple__75
#define __pyx_tuple__80 __pyx_mstate_global->__pyx_tuple__80
#define __pyx_tuple__83 __pyx_mstate_global->__pyx_tuple__83
#define __pyx_codeobj__12 __pyx_mstate_global->__pyx_codeobj__12
#define __pyx_codeobj__13 __pyx_mstate_global->__pyx_codeobj__13
#define __pyx_codeobj__14 __pyx_mstate_global->__pyx_codeobj__14
#define __pyx_codeobj__24 __pyx_mstate_global->__pyx_codeobj__24
#define __pyx_codeobj__32 __pyx_mstate_global->__pyx_codeobj__32
#define __pyx_codeobj__39 __pyx_mstate_global->__pyx_codeobj__39
#define __pyx_codeobj__41 __pyx_mstate_global->__pyx_codeobj__41
#define __pyx_codeobj__42 __pyx_mstate_global->__pyx_codeobj__42
#define __pyx_codeobj__43 __pyx_mstate_global->__pyx_codeobj__43
#define __pyx_codeobj__44 __pyx_mstate_global->__pyx_codeobj__44
#define __pyx_codeobj__45 __pyx_mstate_global->__pyx_codeobj__45
#define __pyx_codeobj__46 __pyx_mstate_global->__pyx_codeobj__46
#define __pyx_codeobj__47 __pyx_mstate_global->__pyx_codeobj__47
#define __pyx_codeobj__49 __pyx_mstate_global->__pyx_codeobj__49
#define __pyx_codeobj__51 __pyx_mstate_global->__pyx_codeobj__51
#define __pyx_codeobj__53 __pyx_mstate_global->__pyx_codeobj__53
#define __pyx_codeobj__59 __pyx_mstate_global->__pyx_codeobj__59
#define __pyx_codeobj__62 __pyx_mstate_global->__pyx_codeobj__62
#define __pyx_codeobj__63 __pyx_mstate_global->__pyx_codeobj__63
#define __pyx_codeobj__65 __pyx_mstate_global->__pyx_codeobj__65
#define __pyx_codeobj__67 __pyx_mstate_global->__pyx_codeobj__67
#define __pyx_codeobj__69 __pyx_mstate_global->__pyx_codeobj__69
#define __pyx_codeobj__71 __pyx_mstate_global->__pyx_codeobj__71
#define __pyx_codeobj__72 __pyx_mstate_global->__pyx_codeobj__72
#define __pyx_codeobj__74 __pyx_mstate_global->__pyx_codeobj__74
#define __pyx_codeobj__76 __pyx_mstate_global->__pyx_codeobj__76
#define __pyx_codeobj__77 __pyx_mstate_global->__pyx_codeobj__77
#define __pyx_codeobj__78 __pyx_mstate_global->__pyx_codeobj__78
#define __pyx_codeobj__79 __pyx_mstate_global->__pyx_codeobj__79
#define __pyx_codeobj__81 __pyx_mstate_global->__pyx_codeobj__81
#define __pyx_codeobj__82 __pyx_mstate_global->__pyx_codeobj__82
#define __pyx_codeobj__84 __pyx_mstate_global->__pyx_codeobj__84
#define __pyx_codeobj__85 __pyx_mstate_global->__pyx_codeobj__85
#define __pyx_codeobj__86 __pyx_mstate_global->__pyx_codeobj__86
#define __pyx_codeobj__87 __pyx_mstate_global->__pyx_codeobj__87
/* #### Code section: module_code ### */

/* "string.from_py":13
 * 
 * @cname("__pyx_convert_string_from_py_6libcpp_6string_std__in_string")
 * cdef string __pyx_convert_string_from_py_6libcpp_6string_std__in_string(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t length = 0
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)
 */

static std::string __pyx_convert_string_from_py_6libcpp_6string_std__in_string(PyObject *__pyx_v_o) {
  Py_ssize_t __pyx_v_length;
  char const *__pyx_v_data;
  std::string __pyx_r;
  char const *__pyx_t_1;
  std::string __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "string.from_py":14
 * @cname("__pyx_convert_string_from_py_6libcpp_6string_std__in_string")
 * cdef string __pyx_convert_string_from_py_6libcpp_6string_std__in_string(object o) except *:
 *     cdef Py_ssize_t length = 0             # <<<<<<<<<<<<<<
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)
 *     return string(data, length)
 */
  __pyx_v_length = 0;

  /* "string.from_py":15
 * cdef string __pyx_convert_string_from_py_6libcpp_6string_std__in_string(object o) except *:
 *     cdef Py_ssize_t length = 0
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)             # <<<<<<<<<<<<<<
 *     return string(data, length)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_AsStringAndSize(__pyx_v_o, (&__pyx_v_length)); if (unlikely(__pyx_t_1 == ((char const *)NULL))) __PYX_ERR(1, 15, __pyx_L1_error)
  __pyx_v_data = __pyx_t_1;

  /* "string.from_py":16
 *     cdef Py_ssize_t length = 0
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)
 *     return string(data, length)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  try {
    __pyx_t_2 = std::string(__pyx_v_data, __pyx_v_length);
  } catch(...) {
    __Pyx_CppExn2PyErr();
    __PYX_ERR(1, 16, __pyx_L1_error)
  }
  __pyx_r = __pyx_t_2;
  goto __pyx_L0;

  /* "string.from_py":13
 * 
 * @cname("__pyx_convert_string_from_py_6libcpp_6string_std__in_string")
 * cdef string __pyx_convert_string_from_py_6libcpp_6string_std__in_string(object o) except *:             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t length = 0
 *     cdef const char* data = __Pyx_PyObject_AsStringAndSize(o, &length)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("string.from_py.__pyx_convert_string_from_py_6libcpp_6string_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_pretend_to_initialize(&__pyx_r);
  __pyx_L0:;
  return __pyx_r;
}

/* "string.to_py":31
 * 
 * @cname("__pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyObject_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

static CYTHON_INLINE PyObject *__pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string", 1);

  /* "string.to_py":32
 * @cname("__pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     return __Pyx_PyObject_FromStringAndSize(s.data(), s.size())             # <<<<<<<<<<<<<<
 * cdef extern from *:
 *     cdef object __Pyx_PyUnicode_FromStringAndSize(const char*, size_t)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_FromStringAndSize(__pyx_v_s.data(), __pyx_v_s.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "string.to_py":31
 * 
 * @cname("__pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyObject_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyObject_string_to_py_6libcpp_6string_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.to_py":37
 * 
 * @cname("__pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyUnicode_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

static CYTHON_INLINE PyObject *__pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string", 1);

  /* "string.to_py":38
 * @cname("__pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     return __Pyx_PyUnicode_FromStringAndSize(s.data(), s.size())             # <<<<<<<<<<<<<<
 * cdef extern from *:
 *     cdef object __Pyx_PyStr_FromStringAndSize(const char*, size_t)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyUnicode_FromStringAndSize(__pyx_v_s.data(), __pyx_v_s.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "string.to_py":37
 * 
 * @cname("__pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyUnicode_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyUnicode_string_to_py_6libcpp_6string_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.to_py":43
 * 
 * @cname("__pyx_convert_PyStr_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyStr_string_to_py_6libcpp_6string_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyStr_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

static CYTHON_INLINE PyObject *__pyx_convert_PyStr_string_to_py_6libcpp_6string_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyStr_string_to_py_6libcpp_6string_std__in_string", 1);

  /* "string.to_py":44
 * @cname("__pyx_convert_PyStr_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyStr_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     return __Pyx_PyStr_FromStringAndSize(s.data(), s.size())             # <<<<<<<<<<<<<<
 * cdef extern from *:
 *     cdef object __Pyx_PyBytes_FromStringAndSize(const char*, size_t)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyStr_FromStringAndSize(__pyx_v_s.data(), __pyx_v_s.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "string.to_py":43
 * 
 * @cname("__pyx_convert_PyStr_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyStr_string_to_py_6libcpp_6string_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyStr_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyStr_string_to_py_6libcpp_6string_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.to_py":49
 * 
 * @cname("__pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyBytes_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

static CYTHON_INLINE PyObject *__pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string", 1);

  /* "string.to_py":50
 * @cname("__pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     return __Pyx_PyBytes_FromStringAndSize(s.data(), s.size())             # <<<<<<<<<<<<<<
 * cdef extern from *:
 *     cdef object __Pyx_PyByteArray_FromStringAndSize(const char*, size_t)
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_s.data(), __pyx_v_s.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "string.to_py":49
 * 
 * @cname("__pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyBytes_FromStringAndSize(s.data(), s.size())
 * cdef extern from *:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyBytes_string_to_py_6libcpp_6string_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "string.to_py":55
 * 
 * @cname("__pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyByteArray_FromStringAndSize(s.data(), s.size())
 * 
 */

static CYTHON_INLINE PyObject *__pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string(std::string const &__pyx_v_s) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string", 1);

  /* "string.to_py":56
 * @cname("__pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string(const string& s):
 *     return __Pyx_PyByteArray_FromStringAndSize(s.data(), s.size())             # <<<<<<<<<<<<<<
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyByteArray_FromStringAndSize(__pyx_v_s.data(), __pyx_v_s.size()); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "string.to_py":55
 * 
 * @cname("__pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string")
 * cdef inline object __pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string(const string& s):             # <<<<<<<<<<<<<<
 *     return __Pyx_PyByteArray_FromStringAndSize(s.data(), s.size())
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("string.to_py.__pyx_convert_PyByteArray_string_to_py_6libcpp_6string_std__in_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":245
 * 
 *         @property
 *         cdef inline PyObject* base(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_7ndarray_4base_base(PyArrayObject *__pyx_v_self) {
  PyObject *__pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":248
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 *             return PyArray_BASE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_BASE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":245
 * 
 *         @property
 *         cdef inline PyObject* base(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":251
 * 
 *         @property
 *         cdef inline dtype descr(self):             # <<<<<<<<<<<<<<
 *             """Returns an owned reference to the dtype of the array.
 *             """
 */

static CYTHON_INLINE PyArray_Descr *__pyx_f_5numpy_7ndarray_5descr_descr(PyArrayObject *__pyx_v_self) {
  PyArray_Descr *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyArray_Descr *__pyx_t_1;
  __Pyx_RefNannySetupContext("descr", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":254
 *             """Returns an owned reference to the dtype of the array.
 *             """
 *             return <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyArray_DESCR(__pyx_v_self);
  __Pyx_INCREF((PyObject *)((PyArray_Descr *)__pyx_t_1));
  __pyx_r = ((PyArray_Descr *)__pyx_t_1);
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":251
 * 
 *         @property
 *         cdef inline dtype descr(self):             # <<<<<<<<<<<<<<
 *             """Returns an owned reference to the dtype of the array.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":257
 * 
 *         @property
 *         cdef inline int ndim(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns the number of dimensions in the array.
 *             """
 */

static CYTHON_INLINE int __pyx_f_5numpy_7ndarray_4ndim_ndim(PyArrayObject *__pyx_v_self) {
  int __pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":260
 *             """Returns the number of dimensions in the array.
 *             """
 *             return PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_NDIM(__pyx_v_self);
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":257
 * 
 *         @property
 *         cdef inline int ndim(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns the number of dimensions in the array.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":263
 * 
 *         @property
 *         cdef inline npy_intp *shape(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the dimensions/shape of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_5shape_shape(PyArrayObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":268
 *             Can return NULL for 0-dimensional arrays.
 *             """
 *             return PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_DIMS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":263
 * 
 *         @property
 *         cdef inline npy_intp *shape(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the dimensions/shape of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":271
 * 
 *         @property
 *         cdef inline npy_intp *strides(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the strides of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_7strides_strides(PyArrayObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":275
 *             The number of elements matches the number of dimensions of the array (ndim).
 *             """
 *             return PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_STRIDES(__pyx_v_self);
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":271
 * 
 *         @property
 *         cdef inline npy_intp *strides(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the strides of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":278
 * 
 *         @property
 *         cdef inline npy_intp size(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns the total size (in number of elements) of the array.
 *             """
 */

static CYTHON_INLINE npy_intp __pyx_f_5numpy_7ndarray_4size_size(PyArrayObject *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":281
 *             """Returns the total size (in number of elements) of the array.
 *             """
 *             return PyArray_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_SIZE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":278
 * 
 *         @property
 *         cdef inline npy_intp size(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns the total size (in number of elements) of the array.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":284
 * 
 *         @property
 *         cdef inline char* data(self) nogil:             # <<<<<<<<<<<<<<
 *             """The pointer to the data buffer as a char*.
 *             This is provided for legacy reasons to avoid direct struct field access.
 */

static CYTHON_INLINE char *__pyx_f_5numpy_7ndarray_4data_data(PyArrayObject *__pyx_v_self) {
  char *__pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":290
 *             of `PyArray_DATA()` instead, which returns a 'void*'.
 *             """
 *             return PyArray_BYTES(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef unsigned char      npy_bool
 */
  __pyx_r = PyArray_BYTES(__pyx_v_self);
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":284
 * 
 *         @property
 *         cdef inline char* data(self) nogil:             # <<<<<<<<<<<<<<
 *             """The pointer to the data buffer as a char*.
 *             This is provided for legacy reasons to avoid direct struct field access.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":773
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":774
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":773
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":776
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":777
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":776
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":779
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":780
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":779
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":782
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":783
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":782
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":785
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":786
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":785
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":788
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":789
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = PyDataType_HASSUBARRAY(__pyx_v_d);
  if (__pyx_t_1) {

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":790
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":789
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":792
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":788
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":968
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":969
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":970
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  __pyx_t_1 = PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 970, __pyx_L1_error)

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":968
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("numpy.set_array_base", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":972
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":973
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":974
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = (__pyx_v_base == NULL);
  if (__pyx_t_1) {

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":975
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":974
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":976
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":972
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":980
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":981
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":982
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 982, __pyx_L3_error)

      /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":981
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":983
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 983, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":984
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 984, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 984, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":981
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":980
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":986
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":987
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":988
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 988, __pyx_L3_error)

      /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":987
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":989
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 989, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":990
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 990, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 990, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":987
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":986
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":992
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":993
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":994
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 994, __pyx_L3_error)

      /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":993
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":995
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 995, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":996
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 996, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 996, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":993
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":992
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":999
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1011
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":999
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1014
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1026
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1014
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1029
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {
  npy_datetime __pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1036
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1029
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1039
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {
  npy_timedelta __pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1043
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1039
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1046
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1050
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1046
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4real_real(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":20
 *         @property
 *         cdef inline double real(self) noexcept:
 *             return self.cval.real             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = __pyx_v_self->cval.real;
  goto __pyx_L0;

  /* "cpython/complex.pxd":19
 * 
 *         @property
 *         cdef inline double real(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.real
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

static CYTHON_INLINE double __pyx_f_7cpython_7complex_7complex_4imag_imag(PyComplexObject *__pyx_v_self) {
  double __pyx_r;

  /* "cpython/complex.pxd":24
 *         @property
 *         cdef inline double imag(self) noexcept:
 *             return self.cval.imag             # <<<<<<<<<<<<<<
 * 
 *     # PyTypeObject PyComplex_Type
 */
  __pyx_r = __pyx_v_self->cval.imag;
  goto __pyx_L0;

  /* "cpython/complex.pxd":23
 * 
 *         @property
 *         cdef inline double imag(self) noexcept:             # <<<<<<<<<<<<<<
 *             return self.cval.imag
 * 
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":117
 *     or None if no such value or default was found.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":118
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)             # <<<<<<<<<<<<<<
 *     if value is NULL:
 *         # context variable does not have a default
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, NULL, (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(3, 118, __pyx_L1_error)

  /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
  __pyx_t_2 = (__pyx_v_value == NULL);
  if (__pyx_t_2) {

    /* "cpython/contextvars.pxd":121
 *     if value is NULL:
 *         # context variable does not have a default
 *         pyvalue = default_value             # <<<<<<<<<<<<<<
 *     else:
 *         # value or default value of context variable
 */
    __Pyx_INCREF(__pyx_v_default_value);
    __pyx_v_pyvalue = __pyx_v_default_value;

    /* "cpython/contextvars.pxd":119
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, NULL, &value)
 *     if value is NULL:             # <<<<<<<<<<<<<<
 *         # context variable does not have a default
 *         pyvalue = default_value
 */
    goto __pyx_L3;
  }

  /* "cpython/contextvars.pxd":124
 *     else:
 *         # value or default value of context variable
 *         pyvalue = <object>value             # <<<<<<<<<<<<<<
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  /*else*/ {
    __pyx_t_3 = ((PyObject *)__pyx_v_value);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_pyvalue = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "cpython/contextvars.pxd":125
 *         # value or default value of context variable
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 * 
 */
    Py_XDECREF(__pyx_v_value);
  }
  __pyx_L3:;

  /* "cpython/contextvars.pxd":126
 *         pyvalue = <object>value
 *         Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":112
 * 
 * 
 * cdef inline object get_value(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the default value of the context variable,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("cpython.contextvars.get_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

static CYTHON_INLINE PyObject *__pyx_f_7cpython_11contextvars_get_value_no_default(PyObject *__pyx_v_var, struct __pyx_opt_args_7cpython_11contextvars_get_value_no_default *__pyx_optional_args) {
  PyObject *__pyx_v_default_value = ((PyObject *)Py_None);
  PyObject *__pyx_v_value;
  PyObject *__pyx_v_pyvalue = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_value_no_default", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_default_value = __pyx_optional_args->default_value;
    }
  }

  /* "cpython/contextvars.pxd":135
 *     Ignores the default value of the context variable, if any.
 *     """
 *     cdef PyObject *value = NULL             # <<<<<<<<<<<<<<
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 */
  __pyx_v_value = NULL;

  /* "cpython/contextvars.pxd":136
 *     """
 *     cdef PyObject *value = NULL
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)             # <<<<<<<<<<<<<<
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 */
  __pyx_t_1 = PyContextVar_Get(__pyx_v_var, ((PyObject *)__pyx_v_default_value), (&__pyx_v_value)); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(3, 136, __pyx_L1_error)

  /* "cpython/contextvars.pxd":138
 *     PyContextVar_Get(var, <PyObject*>default_value, &value)
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value             # <<<<<<<<<<<<<<
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue
 */
  __pyx_t_2 = ((PyObject *)__pyx_v_value);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_pyvalue = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "cpython/contextvars.pxd":139
 *     # value of context variable or 'default_value'
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'             # <<<<<<<<<<<<<<
 *     return pyvalue
 */
  Py_XDECREF(__pyx_v_value);

  /* "cpython/contextvars.pxd":140
 *     pyvalue = <object>value
 *     Py_XDECREF(value)  # PyContextVar_Get() returned an owned reference as 'PyObject*'
 *     return pyvalue             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_pyvalue);
  __pyx_r = __pyx_v_pyvalue;
  goto __pyx_L0;

  /* "cpython/contextvars.pxd":129
 * 
 * 
 * cdef inline object get_value_no_default(var, default_value=None):             # <<<<<<<<<<<<<<
 *     """Return a new reference to the value of the context variable,
 *     or the provided default value if no such value was found.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("cpython.contextvars.get_value_no_default", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pyvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":56
 * 
 * cdef class QuantizationParams:
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  str trading_pair,
 *                  int price_precision,
 */

/* Python wrapper */
static int __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_trading_pair = 0;
  int __pyx_v_price_precision;
  int __pyx_v_price_decimals;
  int __pyx_v_order_size_precision;
  int __pyx_v_order_size_decimals;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_trading_pair,&__pyx_n_s_price_precision,&__pyx_n_s_price_decimals,&__pyx_n_s_order_size_precision,&__pyx_n_s_order_size_decimals,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_trading_pair)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 56, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_price_precision)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 56, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 1); __PYX_ERR(0, 56, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_price_decimals)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 56, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 2); __PYX_ERR(0, 56, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_order_size_precision)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 56, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 3); __PYX_ERR(0, 56, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_order_size_decimals)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[4]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 56, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 4); __PYX_ERR(0, 56, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 56, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
      values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
      values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
    }
    __pyx_v_trading_pair = ((PyObject*)values[0]);
    __pyx_v_price_precision = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_price_precision == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 58, __pyx_L3_error)
    __pyx_v_price_decimals = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_price_decimals == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 59, __pyx_L3_error)
    __pyx_v_order_size_precision = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_order_size_precision == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 60, __pyx_L3_error)
    __pyx_v_order_size_decimals = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_order_size_decimals == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 61, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 56, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.QuantizationParams.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_trading_pair), (&PyUnicode_Type), 1, "trading_pair", 1))) __PYX_ERR(0, 57, __pyx_L1_error)
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams___init__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *)__pyx_v_self), __pyx_v_trading_pair, __pyx_v_price_precision, __pyx_v_price_decimals, __pyx_v_order_size_precision, __pyx_v_order_size_decimals);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams___init__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *__pyx_v_self, PyObject *__pyx_v_trading_pair, int __pyx_v_price_precision, int __pyx_v_price_decimals, int __pyx_v_order_size_precision, int __pyx_v_order_size_decimals) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":62
 *                  int order_size_precision,
 *                  int order_size_decimals):
 *         self.trading_pair = trading_pair             # <<<<<<<<<<<<<<
 *         self.price_precision = price_precision
 *         self.price_decimals = price_decimals
 */
  __Pyx_INCREF(__pyx_v_trading_pair);
  __Pyx_GIVEREF(__pyx_v_trading_pair);
  __Pyx_GOTREF(__pyx_v_self->trading_pair);
  __Pyx_DECREF(__pyx_v_self->trading_pair);
  __pyx_v_self->trading_pair = __pyx_v_trading_pair;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":63
 *                  int order_size_decimals):
 *         self.trading_pair = trading_pair
 *         self.price_precision = price_precision             # <<<<<<<<<<<<<<
 *         self.price_decimals = price_decimals
 *         self.order_size_precision = order_size_precision
 */
  __pyx_v_self->price_precision = __pyx_v_price_precision;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":64
 *         self.trading_pair = trading_pair
 *         self.price_precision = price_precision
 *         self.price_decimals = price_decimals             # <<<<<<<<<<<<<<
 *         self.order_size_precision = order_size_precision
 *         self.order_size_decimals = order_size_decimals
 */
  __pyx_v_self->price_decimals = __pyx_v_price_decimals;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":65
 *         self.price_precision = price_precision
 *         self.price_decimals = price_decimals
 *         self.order_size_precision = order_size_precision             # <<<<<<<<<<<<<<
 *         self.order_size_decimals = order_size_decimals
 * 
 */
  __pyx_v_self->order_size_precision = __pyx_v_order_size_precision;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":66
 *         self.price_decimals = price_decimals
 *         self.order_size_precision = order_size_precision
 *         self.order_size_decimals = order_size_decimals             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self) -> str:
 */
  __pyx_v_self->order_size_decimals = __pyx_v_order_size_decimals;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":56
 * 
 * cdef class QuantizationParams:
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  str trading_pair,
 *                  int price_precision,
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":68
 *         self.order_size_decimals = order_size_decimals
 * 
 *     def __repr__(self) -> str:             # <<<<<<<<<<<<<<
 *         return (f"QuantizationParams('{self.trading_pair}', {self.price_precision}, {self.price_decimals}, "
 *                 f"{self.order_size_precision}, {self.order_size_decimals})")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_2__repr__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_2__repr__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":69
 * 
 *     def __repr__(self) -> str:
 *         return (f"QuantizationParams('{self.trading_pair}', {self.price_precision}, {self.price_decimals}, "             # <<<<<<<<<<<<<<
 *                 f"{self.order_size_precision}, {self.order_size_decimals})")
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_QuantizationParams);
  __pyx_t_2 += 20;
  __Pyx_GIVEREF(__pyx_kp_u_QuantizationParams);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_QuantizationParams);
  __pyx_t_4 = __Pyx_PyUnicode_Unicode(__pyx_v_self->trading_pair); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__3);
  __pyx_t_2 += 3;
  __Pyx_GIVEREF(__pyx_kp_u__3);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u__3);
  __pyx_t_4 = __Pyx_PyUnicode_From_int(__pyx_v_self->price_precision, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__4);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u__4);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u__4);
  __pyx_t_4 = __Pyx_PyUnicode_From_int(__pyx_v_self->price_decimals, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__4);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u__4);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u__4);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":70
 *     def __repr__(self) -> str:
 *         return (f"QuantizationParams('{self.trading_pair}', {self.price_precision}, {self.price_decimals}, "
 *                 f"{self.order_size_precision}, {self.order_size_decimals})")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __Pyx_PyUnicode_From_int(__pyx_v_self->order_size_precision, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__4);
  __pyx_t_2 += 2;
  __Pyx_GIVEREF(__pyx_kp_u__4);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_kp_u__4);
  __pyx_t_4 = __Pyx_PyUnicode_From_int(__pyx_v_self->order_size_decimals, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__5);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__5);
  PyTuple_SET_ITEM(__pyx_t_1, 10, __pyx_kp_u__5);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":69
 * 
 *     def __repr__(self) -> str:
 *         return (f"QuantizationParams('{self.trading_pair}', {self.price_precision}, {self.price_decimals}, "             # <<<<<<<<<<<<<<
 *                 f"{self.order_size_precision}, {self.order_size_decimals})")
 * 
 */
  __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_1, 11, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":68
 *         self.order_size_decimals = order_size_decimals
 * 
 *     def __repr__(self) -> str:             # <<<<<<<<<<<<<<
 *         return (f"QuantizationParams('{self.trading_pair}', {self.price_precision}, {self.price_decimals}, "
 *                 f"{self.order_size_precision}, {self.order_size_decimals})")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.QuantizationParams.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_4__reduce_cython__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_4__reduce_cython__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.order_size_decimals, self.order_size_precision, self.price_decimals, self.price_precision, self.trading_pair)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->order_size_decimals); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->order_size_precision); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->price_decimals); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->price_precision); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyTuple_New(5); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->trading_pair);
  __Pyx_GIVEREF(__pyx_v_self->trading_pair);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_v_self->trading_pair)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.order_size_decimals, self.order_size_precision, self.price_decimals, self.price_precision, self.trading_pair)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_5 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v__dict = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "(tree fragment)":7
 *     state = (self.order_size_decimals, self.order_size_precision, self.price_decimals, self.price_precision, self.trading_pair)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_6 = (__pyx_v__dict != Py_None);
  if (__pyx_t_6) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.trading_pair is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.order_size_decimals, self.order_size_precision, self.price_decimals, self.price_precision, self.trading_pair)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.trading_pair is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_QuantizationParams, (type(self), 0xa74d68e, None), state
 */
  /*else*/ {
    __pyx_t_6 = (__pyx_v_self->trading_pair != ((PyObject*)Py_None));
    __pyx_v_use_setstate = __pyx_t_6;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.trading_pair is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_QuantizationParams, (type(self), 0xa74d68e, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self.trading_pair is not None
 *     if use_setstate:
 *         return __pyx_unpickle_QuantizationParams, (type(self), 0xa74d68e, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_QuantizationParams, (type(self), 0xa74d68e, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_QuantizationParam); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_175429262);
    __Pyx_GIVEREF(__pyx_int_175429262);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_175429262)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.trading_pair is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_QuantizationParams, (type(self), 0xa74d68e, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_QuantizationParams, (type(self), 0xa74d68e, None), state
 *     else:
 *         return __pyx_unpickle_QuantizationParams, (type(self), 0xa74d68e, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_QuantizationParams__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pyx_unpickle_QuantizationParam); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_175429262);
    __Pyx_GIVEREF(__pyx_int_175429262);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_175429262)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_5 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.QuantizationParams.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_QuantizationParams, (type(self), 0xa74d68e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_QuantizationParams__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.QuantizationParams.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_6__setstate_cython__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_6__setstate_cython__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_QuantizationParams, (type(self), 0xa74d68e, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_QuantizationParams__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_QuantizationParams__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_QuantizationParams, (type(self), 0xa74d68e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_QuantizationParams__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.QuantizationParams.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":81
 *         object _amount
 * 
 *     def __init__(self, create_timestamp: float, order_id: str, is_buy: bool, trading_pair: str, amount: Decimal):             # <<<<<<<<<<<<<<
 *         self.create_timestamp = create_timestamp
 *         self._order_id = order_id
 */

/* Python wrapper */
static int __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_create_timestamp = 0;
  PyObject *__pyx_v_order_id = 0;
  PyObject *__pyx_v_is_buy = 0;
  PyObject *__pyx_v_trading_pair = 0;
  PyObject *__pyx_v_amount = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[5] = {0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_create_timestamp,&__pyx_n_s_order_id,&__pyx_n_s_is_buy,&__pyx_n_s_trading_pair,&__pyx_n_s_amount,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  5: values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_create_timestamp)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 81, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_order_id)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 81, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 1); __PYX_ERR(0, 81, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_is_buy)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 81, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 2); __PYX_ERR(0, 81, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_trading_pair)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 81, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 3); __PYX_ERR(0, 81, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_amount)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[4]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 81, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, 4); __PYX_ERR(0, 81, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 81, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 5)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
      values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
      values[4] = __Pyx_Arg_VARARGS(__pyx_args, 4);
    }
    __pyx_v_create_timestamp = values[0];
    __pyx_v_order_id = values[1];
    __pyx_v_is_buy = values[2];
    __pyx_v_trading_pair = values[3];
    __pyx_v_amount = values[4];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 5, 5, __pyx_nargs); __PYX_ERR(0, 81, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.QueuedOrder.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder___init__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *)__pyx_v_self), __pyx_v_create_timestamp, __pyx_v_order_id, __pyx_v_is_buy, __pyx_v_trading_pair, __pyx_v_amount);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder___init__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_self, PyObject *__pyx_v_create_timestamp, PyObject *__pyx_v_order_id, PyObject *__pyx_v_is_buy, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_amount) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":82
 * 
 *     def __init__(self, create_timestamp: float, order_id: str, is_buy: bool, trading_pair: str, amount: Decimal):
 *         self.create_timestamp = create_timestamp             # <<<<<<<<<<<<<<
 *         self._order_id = order_id
 *         self._is_buy = is_buy
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_create_timestamp); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 82, __pyx_L1_error)
  __pyx_v_self->create_timestamp = __pyx_t_1;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":83
 *     def __init__(self, create_timestamp: float, order_id: str, is_buy: bool, trading_pair: str, amount: Decimal):
 *         self.create_timestamp = create_timestamp
 *         self._order_id = order_id             # <<<<<<<<<<<<<<
 *         self._is_buy = is_buy
 *         self._trading_pair = trading_pair
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_order_id))||((__pyx_v_order_id) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_order_id))) __PYX_ERR(0, 83, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_order_id;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_order_id);
  __Pyx_DECREF(__pyx_v_self->_order_id);
  __pyx_v_self->_order_id = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":84
 *         self.create_timestamp = create_timestamp
 *         self._order_id = order_id
 *         self._is_buy = is_buy             # <<<<<<<<<<<<<<
 *         self._trading_pair = trading_pair
 *         self._amount = amount
 */
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_v_is_buy); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 84, __pyx_L1_error)
  __pyx_v_self->_is_buy = __pyx_t_3;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":85
 *         self._order_id = order_id
 *         self._is_buy = is_buy
 *         self._trading_pair = trading_pair             # <<<<<<<<<<<<<<
 *         self._amount = amount
 * 
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_trading_pair))||((__pyx_v_trading_pair) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_trading_pair))) __PYX_ERR(0, 85, __pyx_L1_error)
  __pyx_t_2 = __pyx_v_trading_pair;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_trading_pair);
  __Pyx_DECREF(__pyx_v_self->_trading_pair);
  __pyx_v_self->_trading_pair = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":86
 *         self._is_buy = is_buy
 *         self._trading_pair = trading_pair
 *         self._amount = amount             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_INCREF(__pyx_v_amount);
  __Pyx_GIVEREF(__pyx_v_amount);
  __Pyx_GOTREF(__pyx_v_self->_amount);
  __Pyx_DECREF(__pyx_v_self->_amount);
  __pyx_v_self->_amount = __pyx_v_amount;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":81
 *         object _amount
 * 
 *     def __init__(self, create_timestamp: float, order_id: str, is_buy: bool, trading_pair: str, amount: Decimal):             # <<<<<<<<<<<<<<
 *         self.create_timestamp = create_timestamp
 *         self._order_id = order_id
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.QueuedOrder.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":88
 *         self._amount = amount
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def timestamp(self) -> double:
 *         return self.create_timestamp
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_9timestamp_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_9timestamp_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_9timestamp___get__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_9timestamp___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":90
 *     @property
 *     def timestamp(self) -> double:
 *         return self.create_timestamp             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->create_timestamp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":88
 *         self._amount = amount
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def timestamp(self) -> double:
 *         return self.create_timestamp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.QueuedOrder.timestamp.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":92
 *         return self.create_timestamp
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_id(self) -> str:
 *         return self._order_id
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_8order_id_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_8order_id_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_8order_id___get__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_8order_id___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":94
 *     @property
 *     def order_id(self) -> str:
 *         return self._order_id             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_order_id);
  __pyx_r = __pyx_v_self->_order_id;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":92
 *         return self.create_timestamp
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_id(self) -> str:
 *         return self._order_id
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":96
 *         return self._order_id
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def is_buy(self) -> bint:
 *         return self._is_buy
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_6is_buy_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_6is_buy_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_6is_buy___get__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_6is_buy___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":98
 *     @property
 *     def is_buy(self) -> bint:
 *         return self._is_buy             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_is_buy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":96
 *         return self._order_id
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def is_buy(self) -> bint:
 *         return self._is_buy
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.QueuedOrder.is_buy.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":100
 *         return self._is_buy
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def trading_pair(self) -> str:
 *         return self._trading_pair
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_12trading_pair_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_12trading_pair_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_12trading_pair___get__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_12trading_pair___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":102
 *     @property
 *     def trading_pair(self) -> str:
 *         return self._trading_pair             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_trading_pair);
  __pyx_r = __pyx_v_self->_trading_pair;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":100
 *         return self._is_buy
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def trading_pair(self) -> str:
 *         return self._trading_pair
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":104
 *         return self._trading_pair
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def amount(self) -> Decimal:
 *         return self._amount
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_6amount_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_6amount_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_6amount___get__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_6amount___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":106
 *     @property
 *     def amount(self) -> Decimal:
 *         return self._amount             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self) -> str:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_amount);
  __pyx_r = __pyx_v_self->_amount;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":104
 *         return self._trading_pair
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def amount(self) -> Decimal:
 *         return self._amount
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":108
 *         return self._amount
 * 
 *     def __repr__(self) -> str:             # <<<<<<<<<<<<<<
 *         return (f"QueuedOrder({self.create_timestamp}, '{self.order_id}', {self.is_buy}, '{self.trading_pair}', "
 *                 f"{self.amount})")
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_3__repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_3__repr__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_2__repr__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_2__repr__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_UCS4 __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":109
 * 
 *     def __repr__(self) -> str:
 *         return (f"QueuedOrder({self.create_timestamp}, '{self.order_id}', {self.is_buy}, '{self.trading_pair}', "             # <<<<<<<<<<<<<<
 *                 f"{self.amount})")
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = 0;
  __pyx_t_3 = 127;
  __Pyx_INCREF(__pyx_kp_u_QueuedOrder);
  __pyx_t_2 += 12;
  __Pyx_GIVEREF(__pyx_kp_u_QueuedOrder);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_QueuedOrder);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->create_timestamp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u__6);
  __pyx_t_2 += 3;
  __Pyx_GIVEREF(__pyx_kp_u__6);
  PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u__6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_id); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__3);
  __pyx_t_2 += 3;
  __Pyx_GIVEREF(__pyx_kp_u__3);
  PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u__3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_is_buy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u__6);
  __pyx_t_2 += 3;
  __Pyx_GIVEREF(__pyx_kp_u__6);
  PyTuple_SET_ITEM(__pyx_t_1, 6, __pyx_kp_u__6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_INCREF(__pyx_kp_u__3);
  __pyx_t_2 += 3;
  __Pyx_GIVEREF(__pyx_kp_u__3);
  PyTuple_SET_ITEM(__pyx_t_1, 8, __pyx_kp_u__3);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":110
 *     def __repr__(self) -> str:
 *         return (f"QueuedOrder({self.create_timestamp}, '{self.order_id}', {self.is_buy}, '{self.trading_pair}', "
 *                 f"{self.amount})")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_amount); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_empty_unicode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_3) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_3;
  __pyx_t_2 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_1, 9, __pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_INCREF(__pyx_kp_u__5);
  __pyx_t_2 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__5);
  PyTuple_SET_ITEM(__pyx_t_1, 10, __pyx_kp_u__5);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":109
 * 
 *     def __repr__(self) -> str:
 *         return (f"QueuedOrder({self.create_timestamp}, '{self.order_id}', {self.is_buy}, '{self.trading_pair}', "             # <<<<<<<<<<<<<<
 *                 f"{self.amount})")
 * 
 */
  __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_1, 11, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":108
 *         return self._amount
 * 
 *     def __repr__(self) -> str:             # <<<<<<<<<<<<<<
 *         return (f"QueuedOrder({self.create_timestamp}, '{self.order_id}', {self.is_buy}, '{self.trading_pair}', "
 *                 f"{self.amount})")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.QueuedOrder.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_5__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_5__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_4__reduce_cython__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_4__reduce_cython__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._amount, self._is_buy, self._order_id, self._trading_pair, self.create_timestamp)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_is_buy); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->create_timestamp); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->_amount);
  __Pyx_GIVEREF(__pyx_v_self->_amount);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->_amount)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_order_id);
  __Pyx_GIVEREF(__pyx_v_self->_order_id);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_self->_order_id)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_trading_pair);
  __Pyx_GIVEREF(__pyx_v_self->_trading_pair);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_v_self->_trading_pair)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._amount, self._is_buy, self._order_id, self._trading_pair, self.create_timestamp)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self._amount, self._is_buy, self._order_id, self._trading_pair, self.create_timestamp)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._amount is not None or self._order_id is not None or self._trading_pair is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._amount, self._is_buy, self._order_id, self._trading_pair, self.create_timestamp)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._amount is not None or self._order_id is not None or self._trading_pair is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_QueuedOrder, (type(self), 0xefa59de, None), state
 */
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->_amount != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->_order_id != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->_trading_pair != ((PyObject*)Py_None));
    __pyx_t_4 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._amount is not None or self._order_id is not None or self._trading_pair is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_QueuedOrder, (type(self), 0xefa59de, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._amount is not None or self._order_id is not None or self._trading_pair is not None
 *     if use_setstate:
 *         return __pyx_unpickle_QueuedOrder, (type(self), 0xefa59de, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_QueuedOrder, (type(self), 0xefa59de, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pyx_unpickle_QueuedOrder); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_251288030);
    __Pyx_GIVEREF(__pyx_int_251288030);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_251288030)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._amount is not None or self._order_id is not None or self._trading_pair is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_QueuedOrder, (type(self), 0xefa59de, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_QueuedOrder, (type(self), 0xefa59de, None), state
 *     else:
 *         return __pyx_unpickle_QueuedOrder, (type(self), 0xefa59de, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_QueuedOrder__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_QueuedOrder); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_251288030);
    __Pyx_GIVEREF(__pyx_int_251288030);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_251288030)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.QueuedOrder.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_QueuedOrder, (type(self), 0xefa59de, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_QueuedOrder__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_7__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_7__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.QueuedOrder.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_6__setstate_cython__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_6__setstate_cython__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_QueuedOrder, (type(self), 0xefa59de, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_QueuedOrder__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_QueuedOrder__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_QueuedOrder, (type(self), 0xefa59de, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_QueuedOrder__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.QueuedOrder.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":117
 *         ExchangeBase _market
 * 
 *     def __init__(self, market: ExchangeBase):             # <<<<<<<<<<<<<<
 *         super().__init__()
 *         self._market = market
 */

/* Python wrapper */
static int __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_market = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_market,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_market)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 117, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 117, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_market = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 117, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.OrderBookTradeListener.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener___init__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *)__pyx_v_self), __pyx_v_market);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener___init__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *__pyx_v_self, PyObject *__pyx_v_market) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":118
 * 
 *     def __init__(self, market: ExchangeBase):
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         self._market = market
 * 
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener))) __PYX_ERR(0, 118, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 118, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":119
 *     def __init__(self, market: ExchangeBase):
 *         super().__init__()
 *         self._market = market             # <<<<<<<<<<<<<<
 * 
 *     cdef c_call(self, object event_object):
 */
  if (!(likely(((__pyx_v_market) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_market, __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 119, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_market;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_market);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_market);
  __pyx_v_self->_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":117
 *         ExchangeBase _market
 * 
 *     def __init__(self, market: ExchangeBase):             # <<<<<<<<<<<<<<
 *         super().__init__()
 *         self._market = market
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.OrderBookTradeListener.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":121
 *         self._market = market
 * 
 *     cdef c_call(self, object event_object):             # <<<<<<<<<<<<<<
 *         try:
 *             self._market.match_trade_to_limit_orders(event_object)
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_c_call(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *__pyx_v_self, PyObject *__pyx_v_event_object) {
  CYTHON_UNUSED PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_call", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":122
 * 
 *     cdef c_call(self, object event_object):
 *         try:             # <<<<<<<<<<<<<<
 *             self._market.match_trade_to_limit_orders(event_object)
 *         except Exception as e:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":123
 *     cdef c_call(self, object event_object):
 *         try:
 *             self._market.match_trade_to_limit_orders(event_object)             # <<<<<<<<<<<<<<
 *         except Exception as e:
 *             self.logger().error("Error call trade listener.", exc_info=True)
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_market), __pyx_n_s_match_trade_to_limit_orders); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 123, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_event_object};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 123, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":122
 * 
 *     cdef c_call(self, object event_object):
 *         try:             # <<<<<<<<<<<<<<
 *             self._market.match_trade_to_limit_orders(event_object)
 *         except Exception as e:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":124
 *         try:
 *             self._market.match_trade_to_limit_orders(event_object)
 *         except Exception as e:             # <<<<<<<<<<<<<<
 *             self.logger().error("Error call trade listener.", exc_info=True)
 * 
 */
    __pyx_t_8 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_8) {
      __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.OrderBookTradeListener.c_call", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_5, &__pyx_t_6) < 0) __PYX_ERR(0, 124, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_5);
      __pyx_v_e = __pyx_t_5;
      /*try:*/ {

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":125
 *             self._market.match_trade_to_limit_orders(event_object)
 *         except Exception as e:
 *             self.logger().error("Error call trade listener.", exc_info=True)             # <<<<<<<<<<<<<<
 * 
 * cdef class OrderBookMarketOrderFillListener(EventListener):
 */
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 125, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = NULL;
        __pyx_t_7 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
            __pyx_t_7 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_11, NULL};
          __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 125, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_error); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 125, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_9 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 125, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_exc_info, Py_True) < 0) __PYX_ERR(0, 125, __pyx_L14_error)
        __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_tuple__7, __pyx_t_9); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 125, __pyx_L14_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":124
 *         try:
 *             self._market.match_trade_to_limit_orders(event_object)
 *         except Exception as e:             # <<<<<<<<<<<<<<
 *             self.logger().error("Error call trade listener.", exc_info=True)
 * 
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
          goto __pyx_L15;
        }
        __pyx_L14_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_8 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L5_except_error;
        }
        __pyx_L15:;
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":122
 * 
 *     cdef c_call(self, object event_object):
 *         try:             # <<<<<<<<<<<<<<
 *             self._market.match_trade_to_limit_orders(event_object)
 *         except Exception as e:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":121
 *         self._market = market
 * 
 *     cdef c_call(self, object event_object):             # <<<<<<<<<<<<<<
 *         try:
 *             self._market.match_trade_to_limit_orders(event_object)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.OrderBookTradeListener.c_call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_2__reduce_cython__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_2__reduce_cython__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._current_event_caller, self._current_event_tag, self._market)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base._current_event_tag); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base._current_event_caller);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base._current_event_caller);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self->__pyx_base._current_event_caller))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->_market);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->_market);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, ((PyObject *)__pyx_v_self->_market))) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._current_event_caller, self._current_event_tag, self._market)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self._current_event_caller, self._current_event_tag, self._market)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._current_event_caller is not None or self._market is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._current_event_caller, self._current_event_tag, self._market)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._current_event_caller is not None or self._market is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_OrderBookTradeListener, (type(self), 0x1bae46d, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base._current_event_caller) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->_market) != Py_None);
    __pyx_t_3 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._current_event_caller is not None or self._market is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_OrderBookTradeListener, (type(self), 0x1bae46d, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._current_event_caller is not None or self._market is not None
 *     if use_setstate:
 *         return __pyx_unpickle_OrderBookTradeListener, (type(self), 0x1bae46d, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_OrderBookTradeListener, (type(self), 0x1bae46d, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_OrderBookTradeLis); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_29025389);
    __Pyx_GIVEREF(__pyx_int_29025389);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_29025389)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._current_event_caller is not None or self._market is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_OrderBookTradeListener, (type(self), 0x1bae46d, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_OrderBookTradeListener, (type(self), 0x1bae46d, None), state
 *     else:
 *         return __pyx_unpickle_OrderBookTradeListener, (type(self), 0x1bae46d, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_OrderBookTradeListener__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_OrderBookTradeLis); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_29025389);
    __Pyx_GIVEREF(__pyx_int_29025389);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_29025389)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.OrderBookTradeListener.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_OrderBookTradeListener, (type(self), 0x1bae46d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_OrderBookTradeListener__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.OrderBookTradeListener.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_4__setstate_cython__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_4__setstate_cython__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_OrderBookTradeListener, (type(self), 0x1bae46d, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_OrderBookTradeListener__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_OrderBookTradeListener__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_OrderBookTradeListener, (type(self), 0x1bae46d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_OrderBookTradeListener__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.OrderBookTradeListener.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":131
 *         ExchangeBase _market
 * 
 *     def __init__(self, market: ExchangeBase):             # <<<<<<<<<<<<<<
 *         super().__init__()
 *         self._market = market
 */

/* Python wrapper */
static int __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_market = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_market,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_market)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 131, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 131, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
    }
    __pyx_v_market = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 131, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.OrderBookMarketOrderFillListener.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener___init__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *)__pyx_v_self), __pyx_v_market);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener___init__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *__pyx_v_self, PyObject *__pyx_v_market) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":132
 * 
 *     def __init__(self, market: ExchangeBase):
 *         super().__init__()             # <<<<<<<<<<<<<<
 *         self._market = market
 * 
 */
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener))) __PYX_ERR(0, 132, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 132, __pyx_L1_error);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_init); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 132, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":133
 *     def __init__(self, market: ExchangeBase):
 *         super().__init__()
 *         self._market = market             # <<<<<<<<<<<<<<
 * 
 *     cdef c_call(self, object event_object):
 */
  if (!(likely(((__pyx_v_market) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_market, __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 133, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_market;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_market);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_market);
  __pyx_v_self->_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":131
 *         ExchangeBase _market
 * 
 *     def __init__(self, market: ExchangeBase):             # <<<<<<<<<<<<<<
 *         super().__init__()
 *         self._market = market
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.OrderBookMarketOrderFillListener.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":135
 *         self._market = market
 * 
 *     cdef c_call(self, object event_object):             # <<<<<<<<<<<<<<
 * 
 *         if event_object.trading_pair not in self._market.order_books or event_object.order_type != OrderType.MARKET:
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_c_call(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *__pyx_v_self, PyObject *__pyx_v_event_object) {
  PyObject *__pyx_v_order_book = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_call", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":137
 *     cdef c_call(self, object event_object):
 * 
 *         if event_object.trading_pair not in self._market.order_books or event_object.order_type != OrderType.MARKET:             # <<<<<<<<<<<<<<
 *             return
 *         order_book = self._market.order_books[event_object.trading_pair]
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_event_object, __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_market), __pyx_n_s_order_books); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_t_2, __pyx_t_3, Py_NE)); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!__pyx_t_4) {
  } else {
    __pyx_t_1 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_event_object, __pyx_n_s_order_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MARKET); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_3, __pyx_t_5, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":138
 * 
 *         if event_object.trading_pair not in self._market.order_books or event_object.order_type != OrderType.MARKET:
 *             return             # <<<<<<<<<<<<<<
 *         order_book = self._market.order_books[event_object.trading_pair]
 *         order_book.record_filled_order(event_object)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":137
 *     cdef c_call(self, object event_object):
 * 
 *         if event_object.trading_pair not in self._market.order_books or event_object.order_type != OrderType.MARKET:             # <<<<<<<<<<<<<<
 *             return
 *         order_book = self._market.order_books[event_object.trading_pair]
 */
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":139
 *         if event_object.trading_pair not in self._market.order_books or event_object.order_type != OrderType.MARKET:
 *             return
 *         order_book = self._market.order_books[event_object.trading_pair]             # <<<<<<<<<<<<<<
 *         order_book.record_filled_order(event_object)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_market), __pyx_n_s_order_books); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_event_object, __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_order_book = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":140
 *             return
 *         order_book = self._market.order_books[event_object.trading_pair]
 *         order_book.record_filled_order(event_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_book, __pyx_n_s_record_filled_order); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_event_object};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":135
 *         self._market = market
 * 
 *     cdef c_call(self, object event_object):             # <<<<<<<<<<<<<<
 * 
 *         if event_object.trading_pair not in self._market.order_books or event_object.order_type != OrderType.MARKET:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.OrderBookMarketOrderFillListener.c_call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_order_book);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_2__reduce_cython__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_2__reduce_cython__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._current_event_caller, self._current_event_tag, self._market)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->__pyx_base._current_event_tag); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base._current_event_caller);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base._current_event_caller);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_v_self->__pyx_base._current_event_caller))) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->_market);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->_market);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, ((PyObject *)__pyx_v_self->_market))) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._current_event_caller, self._current_event_tag, self._market)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_2 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v__dict = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":7
 *     state = (self._current_event_caller, self._current_event_tag, self._market)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_3 = (__pyx_v__dict != Py_None);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v__dict)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._current_event_caller is not None or self._market is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._current_event_caller, self._current_event_tag, self._market)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._current_event_caller is not None or self._market is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_OrderBookMarketOrderFillListener, (type(self), 0x1bae46d, None), state
 */
  /*else*/ {
    __pyx_t_4 = (((PyObject *)__pyx_v_self->__pyx_base._current_event_caller) != Py_None);
    if (!__pyx_t_4) {
    } else {
      __pyx_t_3 = __pyx_t_4;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = (((PyObject *)__pyx_v_self->_market) != Py_None);
    __pyx_t_3 = __pyx_t_4;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._current_event_caller is not None or self._market is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_OrderBookMarketOrderFillListener, (type(self), 0x1bae46d, None), state
 *     else:
 */
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._current_event_caller is not None or self._market is not None
 *     if use_setstate:
 *         return __pyx_unpickle_OrderBookMarketOrderFillListener, (type(self), 0x1bae46d, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_OrderBookMarketOrderFillListener, (type(self), 0x1bae46d, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pyx_unpickle_OrderBookMarketOr); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_29025389);
    __Pyx_GIVEREF(__pyx_int_29025389);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_29025389)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_None)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._current_event_caller is not None or self._market is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_OrderBookMarketOrderFillListener, (type(self), 0x1bae46d, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_OrderBookMarketOrderFillListener, (type(self), 0x1bae46d, None), state
 *     else:
 *         return __pyx_unpickle_OrderBookMarketOrderFillListener, (type(self), 0x1bae46d, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_OrderBookMarketOr); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))))) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_29025389);
    __Pyx_GIVEREF(__pyx_int_29025389);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_29025389)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_v_state)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_2 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.OrderBookMarketOrderFillListener.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_OrderBookMarketOrderFillListener, (type(self), 0x1bae46d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 16, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 16, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.OrderBookMarketOrderFillListener.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_4__setstate_cython__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_4__setstate_cython__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_OrderBookMarketOrderFillListener, (type(self), 0x1bae46d, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_OrderBookMarketOrderFillListener__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_OrderBookMarketOrderFillListener, (type(self), 0x1bae46d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.OrderBookMarketOrderFillListener.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":154
 *     MARKET_BUY_ORDER_CREATED_EVENT_TAG = MarketEvent.BuyOrderCreated.value
 * 
 *     def __init__(             # <<<<<<<<<<<<<<
 *         self,
 *         client_config_map: "ClientConfigAdapter",
 */

/* Python wrapper */
static int __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_client_config_map = 0;
  PyObject *__pyx_v_order_book_tracker = 0;
  PyObject *__pyx_v_target_market = 0;
  PyObject *__pyx_v_exchange_name = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_client_config_map,&__pyx_n_s_order_book_tracker,&__pyx_n_s_target_market,&__pyx_n_s_exchange_name,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_VARARGS(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_client_config_map)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 154, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_order_book_tracker)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 154, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 1); __PYX_ERR(0, 154, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_target_market)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 154, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 2); __PYX_ERR(0, 154, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_VARARGS(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_exchange_name)) != 0)) {
          (void)__Pyx_Arg_NewRef_VARARGS(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 154, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, 3); __PYX_ERR(0, 154, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__init__") < 0)) __PYX_ERR(0, 154, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 4)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_VARARGS(__pyx_args, 0);
      values[1] = __Pyx_Arg_VARARGS(__pyx_args, 1);
      values[2] = __Pyx_Arg_VARARGS(__pyx_args, 2);
      values[3] = __Pyx_Arg_VARARGS(__pyx_args, 3);
    }
    __pyx_v_client_config_map = values[0];
    __pyx_v_order_book_tracker = values[1];
    __pyx_v_target_market = values[2];
    __pyx_v_exchange_name = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 4, 4, __pyx_nargs); __PYX_ERR(0, 154, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange___init__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self), __pyx_v_client_config_map, __pyx_v_order_book_tracker, __pyx_v_target_market, __pyx_v_exchange_name);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_VARARGS(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":161
 *         exchange_name: str,
 *     ):
 *         order_book_tracker.data_source.order_book_create_function = lambda: CompositeOrderBook()             # <<<<<<<<<<<<<<
 *         self._set_order_book_tracker(order_book_tracker)
 *         self._budget_checker = BudgetChecker(exchange=self)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_8__init___lambda(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_8__init___lambda = {"lambda", (PyCFunction)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_8__init___lambda, METH_NOARGS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_8__init___lambda(PyObject *__pyx_self, CYTHON_UNUSED PyObject *unused) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_lambda_funcdef_lambda(__pyx_self);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda(CYTHON_UNUSED PyObject *__pyx_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.__init__.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":154
 *     MARKET_BUY_ORDER_CREATED_EVENT_TAG = MarketEvent.BuyOrderCreated.value
 * 
 *     def __init__(             # <<<<<<<<<<<<<<
 *         self,
 *         client_config_map: "ClientConfigAdapter",
 */

static int __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange___init__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_client_config_map, PyObject *__pyx_v_order_book_tracker, PyObject *__pyx_v_target_market, PyObject *__pyx_v_exchange_name) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int64_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":161
 *         exchange_name: str,
 *     ):
 *         order_book_tracker.data_source.order_book_create_function = lambda: CompositeOrderBook()             # <<<<<<<<<<<<<<
 *         self._set_order_book_tracker(order_book_tracker)
 *         self._budget_checker = BudgetChecker(exchange=self)
 */
  __pyx_t_1 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_8__init___lambda, 0, __pyx_n_s_init___locals_lambda, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_book_tracker, __pyx_n_s_data_source); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_t_2, __pyx_n_s_order_book_create_function, __pyx_t_1) < 0) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":162
 *     ):
 *         order_book_tracker.data_source.order_book_create_function = lambda: CompositeOrderBook()
 *         self._set_order_book_tracker(order_book_tracker)             # <<<<<<<<<<<<<<
 *         self._budget_checker = BudgetChecker(exchange=self)
 *         super(ExchangeBase, self).__init__(client_config_map)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_set_order_book_tracker); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_order_book_tracker};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":163
 *         order_book_tracker.data_source.order_book_create_function = lambda: CompositeOrderBook()
 *         self._set_order_book_tracker(order_book_tracker)
 *         self._budget_checker = BudgetChecker(exchange=self)             # <<<<<<<<<<<<<<
 *         super(ExchangeBase, self).__init__(client_config_map)
 *         self._exchange_name = exchange_name
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BudgetChecker); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_exchange, ((PyObject *)__pyx_v_self)) < 0) __PYX_ERR(0, 163, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base._budget_checker);
  __Pyx_DECREF(__pyx_v_self->__pyx_base._budget_checker);
  __pyx_v_self->__pyx_base._budget_checker = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":164
 *         self._set_order_book_tracker(order_book_tracker)
 *         self._budget_checker = BudgetChecker(exchange=self)
 *         super(ExchangeBase, self).__init__(client_config_map)             # <<<<<<<<<<<<<<
 *         self._exchange_name = exchange_name
 *         self._account_balances = {}
 */
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF((PyObject *)__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase);
  __Pyx_GIVEREF((PyObject *)__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))) __PYX_ERR(0, 164, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 164, __pyx_L1_error);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_init); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_client_config_map};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":165
 *         self._budget_checker = BudgetChecker(exchange=self)
 *         super(ExchangeBase, self).__init__(client_config_map)
 *         self._exchange_name = exchange_name             # <<<<<<<<<<<<<<
 *         self._account_balances = {}
 *         self._account_available_balances = {}
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_exchange_name))||((__pyx_v_exchange_name) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_exchange_name))) __PYX_ERR(0, 165, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_exchange_name;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_exchange_name);
  __Pyx_DECREF(__pyx_v_self->_exchange_name);
  __pyx_v_self->_exchange_name = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":166
 *         super(ExchangeBase, self).__init__(client_config_map)
 *         self._exchange_name = exchange_name
 *         self._account_balances = {}             # <<<<<<<<<<<<<<
 *         self._account_available_balances = {}
 *         self._paper_trade_market_initialized = False
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.__pyx_base._account_balances);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.__pyx_base._account_balances);
  __pyx_v_self->__pyx_base.__pyx_base._account_balances = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":167
 *         self._exchange_name = exchange_name
 *         self._account_balances = {}
 *         self._account_available_balances = {}             # <<<<<<<<<<<<<<
 *         self._paper_trade_market_initialized = False
 *         self._trading_pairs = {}
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.__pyx_base._account_available_balances);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.__pyx_base._account_available_balances);
  __pyx_v_self->__pyx_base.__pyx_base._account_available_balances = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":168
 *         self._account_balances = {}
 *         self._account_available_balances = {}
 *         self._paper_trade_market_initialized = False             # <<<<<<<<<<<<<<
 *         self._trading_pairs = {}
 *         self._queued_orders = deque()
 */
  __pyx_v_self->_paper_trade_market_initialized = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":169
 *         self._account_available_balances = {}
 *         self._paper_trade_market_initialized = False
 *         self._trading_pairs = {}             # <<<<<<<<<<<<<<
 *         self._queued_orders = deque()
 *         self._quantization_params = {}
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_trading_pairs);
  __Pyx_DECREF(__pyx_v_self->_trading_pairs);
  __pyx_v_self->_trading_pairs = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":170
 *         self._paper_trade_market_initialized = False
 *         self._trading_pairs = {}
 *         self._queued_orders = deque()             # <<<<<<<<<<<<<<
 *         self._quantization_params = {}
 *         self._order_book_trade_listener = OrderBookTradeListener(self)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_deque); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 170, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_queued_orders);
  __Pyx_DECREF(__pyx_v_self->_queued_orders);
  __pyx_v_self->_queued_orders = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":171
 *         self._trading_pairs = {}
 *         self._queued_orders = deque()
 *         self._quantization_params = {}             # <<<<<<<<<<<<<<
 *         self._order_book_trade_listener = OrderBookTradeListener(self)
 *         self._target_market = target_market
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_quantization_params);
  __Pyx_DECREF(__pyx_v_self->_quantization_params);
  __pyx_v_self->_quantization_params = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":172
 *         self._queued_orders = deque()
 *         self._quantization_params = {}
 *         self._order_book_trade_listener = OrderBookTradeListener(self)             # <<<<<<<<<<<<<<
 *         self._target_market = target_market
 *         self._market_order_filled_listener = OrderBookMarketOrderFillListener(self)
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener), ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_order_book_trade_listener);
  __Pyx_DECREF(__pyx_v_self->_order_book_trade_listener);
  __pyx_v_self->_order_book_trade_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":173
 *         self._quantization_params = {}
 *         self._order_book_trade_listener = OrderBookTradeListener(self)
 *         self._target_market = target_market             # <<<<<<<<<<<<<<
 *         self._market_order_filled_listener = OrderBookMarketOrderFillListener(self)
 *         self.c_add_listener(self.ORDER_FILLED_EVENT_TAG, self._market_order_filled_listener)
 */
  __Pyx_INCREF(__pyx_v_target_market);
  __Pyx_GIVEREF(__pyx_v_target_market);
  __Pyx_GOTREF(__pyx_v_self->_target_market);
  __Pyx_DECREF(__pyx_v_self->_target_market);
  __pyx_v_self->_target_market = __pyx_v_target_market;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":174
 *         self._order_book_trade_listener = OrderBookTradeListener(self)
 *         self._target_market = target_market
 *         self._market_order_filled_listener = OrderBookMarketOrderFillListener(self)             # <<<<<<<<<<<<<<
 *         self.c_add_listener(self.ORDER_FILLED_EVENT_TAG, self._market_order_filled_listener)
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener), ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_market_order_filled_listener);
  __Pyx_DECREF(__pyx_v_self->_market_order_filled_listener);
  __pyx_v_self->_market_order_filled_listener = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":175
 *         self._target_market = target_market
 *         self._market_order_filled_listener = OrderBookMarketOrderFillListener(self)
 *         self.c_add_listener(self.ORDER_FILLED_EVENT_TAG, self._market_order_filled_listener)             # <<<<<<<<<<<<<<
 * 
 *         # Trade volume metrics should never be gather for paper trade connector
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_ORDER_FILLED_EVENT_TAG); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_As_int64_t(__pyx_t_3); if (unlikely((__pyx_t_5 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (!(likely(((__pyx_v_self->_market_order_filled_listener) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_self->_market_order_filled_listener, __pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener))))) __PYX_ERR(0, 175, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_self->_market_order_filled_listener;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.c_add_listener(((struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *)__pyx_v_self), __pyx_t_5, ((struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *)__pyx_t_3)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":178
 * 
 *         # Trade volume metrics should never be gather for paper trade connector
 *         self._trade_volume_metric_collector = DummyMetricsCollector()             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_DummyMetricsCollector); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->__pyx_base.__pyx_base._trade_volume_metric_collector);
  __Pyx_DECREF(__pyx_v_self->__pyx_base.__pyx_base._trade_volume_metric_collector);
  __pyx_v_self->__pyx_base.__pyx_base._trade_volume_metric_collector = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":154
 *     MARKET_BUY_ORDER_CREATED_EVENT_TAG = MarketEvent.BuyOrderCreated.value
 * 
 *     def __init__(             # <<<<<<<<<<<<<<
 *         self,
 *         client_config_map: "ClientConfigAdapter",
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":180
 *         self._trade_volume_metric_collector = DummyMetricsCollector()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def budget_checker(self) -> BudgetChecker:
 *         return self._budget_checker
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_14budget_checker_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_14budget_checker_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_14budget_checker___get__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_14budget_checker___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":182
 *     @property
 *     def budget_checker(self) -> BudgetChecker:
 *         return self._budget_checker             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._budget_checker);
  __pyx_r = __pyx_v_self->__pyx_base._budget_checker;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":180
 *         self._trade_volume_metric_collector = DummyMetricsCollector()
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def budget_checker(self) -> BudgetChecker:
 *         return self._budget_checker
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":184
 *         return self._budget_checker
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def random_order_id(cls, order_side: str, trading_pair: str) -> str:
 *         vals = [random.choice(range(0, 256)) for i in range(0, 13)]
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_3random_order_id(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_3random_order_id = {"random_order_id", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_3random_order_id, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_3random_order_id(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_order_side = 0;
  PyObject *__pyx_v_trading_pair = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("random_order_id (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_order_side,&__pyx_n_s_trading_pair,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_order_side)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 184, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_trading_pair)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 184, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("random_order_id", 1, 2, 2, 1); __PYX_ERR(0, 184, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "random_order_id") < 0)) __PYX_ERR(0, 184, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_order_side = values[0];
    __pyx_v_trading_pair = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("random_order_id", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 184, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.random_order_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_2random_order_id(((PyTypeObject*)__pyx_v_cls), __pyx_v_order_side, __pyx_v_trading_pair);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_2random_order_id(CYTHON_UNUSED PyTypeObject *__pyx_v_cls, PyObject *__pyx_v_order_side, PyObject *__pyx_v_trading_pair) {
  PyObject *__pyx_v_vals = NULL;
  CYTHON_UNUSED long __pyx_7genexpr__pyx_v_i;
  PyObject *__pyx_8genexpr1__pyx_v_val = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  long __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("random_order_id", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":186
 *     @classmethod
 *     def random_order_id(cls, order_side: str, trading_pair: str) -> str:
 *         vals = [random.choice(range(0, 256)) for i in range(0, 13)]             # <<<<<<<<<<<<<<
 *         return f"{order_side}://" + trading_pair + "/" + "".join([f"{val:02x}" for val in vals])
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 186, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    for (__pyx_t_2 = 0; __pyx_t_2 < 13; __pyx_t_2+=1) {
      __pyx_7genexpr__pyx_v_i = __pyx_t_2;
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_random); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_choice); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_range, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_4};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 186, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_3))) __PYX_ERR(0, 186, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
  } /* exit inner scope */
  __pyx_v_vals = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":187
 *     def random_order_id(cls, order_side: str, trading_pair: str) -> str:
 *         vals = [random.choice(range(0, 256)) for i in range(0, 13)]
 *         return f"{order_side}://" + trading_pair + "/" + "".join([f"{val:02x}" for val in vals])             # <<<<<<<<<<<<<<
 * 
 *     def init_paper_trade_market(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_v_order_side, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_1, __pyx_kp_u__9); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_t_3, __pyx_v_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_kp_u__10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 187, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __pyx_v_vals; __Pyx_INCREF(__pyx_t_5);
    __pyx_t_8 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 187, __pyx_L7_error)
        #endif
        if (__pyx_t_8 >= __pyx_temp) break;
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_8); __Pyx_INCREF(__pyx_t_4); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 187, __pyx_L7_error)
      #else
      __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 187, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
      __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_val, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Format(__pyx_8genexpr1__pyx_v_val, __pyx_kp_u_02x); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 187, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 187, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_val); __pyx_8genexpr1__pyx_v_val = 0;
    goto __pyx_L11_exit_scope;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_val); __pyx_8genexpr1__pyx_v_val = 0;
    goto __pyx_L1_error;
    __pyx_L11_exit_scope:;
  } /* exit inner scope */
  __pyx_t_5 = PyUnicode_Join(__pyx_kp_u__11, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Add(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":184
 *         return self._budget_checker
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def random_order_id(cls, order_side: str, trading_pair: str) -> str:
 *         vals = [random.choice(range(0, 256)) for i in range(0, 13)]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.random_order_id", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_vals);
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_val);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":189
 *         return f"{order_side}://" + trading_pair + "/" + "".join([f"{val:02x}" for val in vals])
 * 
 *     def init_paper_trade_market(self):             # <<<<<<<<<<<<<<
 *         for trading_pair_str, order_book in self.order_book_tracker.order_books.items():
 *             assert type(order_book) is CompositeOrderBook
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_5init_paper_trade_market(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_5init_paper_trade_market = {"init_paper_trade_market", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_5init_paper_trade_market, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_5init_paper_trade_market(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("init_paper_trade_market (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("init_paper_trade_market", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "init_paper_trade_market", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_4init_paper_trade_market(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_4init_paper_trade_market(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  PyObject *__pyx_v_trading_pair_str = NULL;
  PyObject *__pyx_v_order_book = NULL;
  PyObject *__pyx_v_base_asset = NULL;
  PyObject *__pyx_v_quote_asset = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *(*__pyx_t_12)(PyObject *);
  int64_t __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init_paper_trade_market", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":190
 * 
 *     def init_paper_trade_market(self):
 *         for trading_pair_str, order_book in self.order_book_tracker.order_books.items():             # <<<<<<<<<<<<<<
 *             assert type(order_book) is CompositeOrderBook
 *             base_asset, quote_asset = self.split_trading_pair(trading_pair_str)
 */
  __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_book_tracker); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_order_books); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(__pyx_t_6 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 190, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_t_6, 0, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, &__pyx_t_6, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_trading_pair_str, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_order_book, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":191
 *     def init_paper_trade_market(self):
 *         for trading_pair_str, order_book in self.order_book_tracker.order_books.items():
 *             assert type(order_book) is CompositeOrderBook             # <<<<<<<<<<<<<<
 *             base_asset, quote_asset = self.split_trading_pair(trading_pair_str)
 *             self._trading_pairs[self._target_market.convert_from_exchange_trading_pair(trading_pair_str)] = TradingPair(trading_pair_str, base_asset, quote_asset)
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(__pyx_assertions_enabled())) {
      __pyx_t_8 = (((PyObject *)Py_TYPE(__pyx_v_order_book)) == ((PyObject *)__pyx_ptype_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook));
      if (unlikely(!__pyx_t_8)) {
        __Pyx_Raise(__pyx_builtin_AssertionError, 0, 0, 0);
        __PYX_ERR(0, 191, __pyx_L1_error)
      }
    }
    #else
    if ((1)); else __PYX_ERR(0, 191, __pyx_L1_error)
    #endif

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":192
 *         for trading_pair_str, order_book in self.order_book_tracker.order_books.items():
 *             assert type(order_book) is CompositeOrderBook
 *             base_asset, quote_asset = self.split_trading_pair(trading_pair_str)             # <<<<<<<<<<<<<<
 *             self._trading_pairs[self._target_market.convert_from_exchange_trading_pair(trading_pair_str)] = TradingPair(trading_pair_str, base_asset, quote_asset)
 *             (<CompositeOrderBook>order_book).c_add_listener(
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_split_trading_pair); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = NULL;
    __pyx_t_10 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_10 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_v_trading_pair_str};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 192, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
      PyObject* sequence = __pyx_t_6;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 192, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 192, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 192, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_11 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 192, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_12 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_11);
      index = 0; __pyx_t_5 = __pyx_t_12(__pyx_t_11); if (unlikely(!__pyx_t_5)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_9 = __pyx_t_12(__pyx_t_11); if (unlikely(!__pyx_t_9)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_11), 2) < 0) __PYX_ERR(0, 192, __pyx_L1_error)
      __pyx_t_12 = NULL;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_12 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 192, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_base_asset, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_quote_asset, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":193
 *             assert type(order_book) is CompositeOrderBook
 *             base_asset, quote_asset = self.split_trading_pair(trading_pair_str)
 *             self._trading_pairs[self._target_market.convert_from_exchange_trading_pair(trading_pair_str)] = TradingPair(trading_pair_str, base_asset, quote_asset)             # <<<<<<<<<<<<<<
 *             (<CompositeOrderBook>order_book).c_add_listener(
 *                 self.ORDER_BOOK_TRADE_EVENT_TAG,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_TradingPair); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_5 = NULL;
    __pyx_t_10 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_10 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_5, __pyx_v_trading_pair_str, __pyx_v_base_asset, __pyx_v_quote_asset};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_10, 3+__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 193, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    if (unlikely(__pyx_v_self->_trading_pairs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 193, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_target_market, __pyx_n_s_convert_from_exchange_trading_pa); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_11 = NULL;
    __pyx_t_10 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_11)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_11);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_10 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_v_trading_pair_str};
      __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 193, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    if (unlikely((PyDict_SetItem(__pyx_v_self->_trading_pairs, __pyx_t_9, __pyx_t_6) < 0))) __PYX_ERR(0, 193, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":195
 *             self._trading_pairs[self._target_market.convert_from_exchange_trading_pair(trading_pair_str)] = TradingPair(trading_pair_str, base_asset, quote_asset)
 *             (<CompositeOrderBook>order_book).c_add_listener(
 *                 self.ORDER_BOOK_TRADE_EVENT_TAG,             # <<<<<<<<<<<<<<
 *                 self._order_book_trade_listener
 *             )
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_ORDER_BOOK_TRADE_EVENT_TAG); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_13 = __Pyx_PyInt_As_int64_t(__pyx_t_6); if (unlikely((__pyx_t_13 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 195, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":196
 *             (<CompositeOrderBook>order_book).c_add_listener(
 *                 self.ORDER_BOOK_TRADE_EVENT_TAG,
 *                 self._order_book_trade_listener             # <<<<<<<<<<<<<<
 *             )
 * 
 */
    if (!(likely(((__pyx_v_self->_order_book_trade_listener) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_self->_order_book_trade_listener, __pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener))))) __PYX_ERR(0, 196, __pyx_L1_error)
    __pyx_t_6 = __pyx_v_self->_order_book_trade_listener;
    __Pyx_INCREF(__pyx_t_6);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":194
 *             base_asset, quote_asset = self.split_trading_pair(trading_pair_str)
 *             self._trading_pairs[self._target_market.convert_from_exchange_trading_pair(trading_pair_str)] = TradingPair(trading_pair_str, base_asset, quote_asset)
 *             (<CompositeOrderBook>order_book).c_add_listener(             # <<<<<<<<<<<<<<
 *                 self.ORDER_BOOK_TRADE_EVENT_TAG,
 *                 self._order_book_trade_listener
 */
    __pyx_t_9 = ((struct __pyx_vtabstruct_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook *)((struct __pyx_obj_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook *)__pyx_v_order_book)->__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_add_listener(((struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *)((struct __pyx_obj_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook *)__pyx_v_order_book)), __pyx_t_13, ((struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *)__pyx_t_6)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":189
 *         return f"{order_side}://" + trading_pair + "/" + "".join([f"{val:02x}" for val in vals])
 * 
 *     def init_paper_trade_market(self):             # <<<<<<<<<<<<<<
 *         for trading_pair_str, order_book in self.order_book_tracker.order_books.items():
 *             assert type(order_book) is CompositeOrderBook
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.init_paper_trade_market", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_trading_pair_str);
  __Pyx_XDECREF(__pyx_v_order_book);
  __Pyx_XDECREF(__pyx_v_base_asset);
  __Pyx_XDECREF(__pyx_v_quote_asset);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":199
 *             )
 * 
 *     def split_trading_pair(self, trading_pair: str) -> Tuple[str, str]:             # <<<<<<<<<<<<<<
 *         return self._target_market.split_trading_pair(trading_pair)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_7split_trading_pair(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_7split_trading_pair = {"split_trading_pair", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_7split_trading_pair, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_7split_trading_pair(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_trading_pair = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("split_trading_pair (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_trading_pair,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_trading_pair)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 199, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "split_trading_pair") < 0)) __PYX_ERR(0, 199, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_trading_pair = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("split_trading_pair", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 199, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.split_trading_pair", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_6split_trading_pair(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self), __pyx_v_trading_pair);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_6split_trading_pair(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("split_trading_pair", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":200
 * 
 *     def split_trading_pair(self, trading_pair: str) -> Tuple[str, str]:
 *         return self._target_market.split_trading_pair(trading_pair)             # <<<<<<<<<<<<<<
 * 
 *     #  <editor-fold desc="Property">
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_target_market, __pyx_n_s_split_trading_pair); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_trading_pair};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 200, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":199
 *             )
 * 
 *     def split_trading_pair(self, trading_pair: str) -> Tuple[str, str]:             # <<<<<<<<<<<<<<
 *         return self._target_market.split_trading_pair(trading_pair)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.split_trading_pair", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":203
 * 
 *     #  <editor-fold desc="Property">
 *     @property             # <<<<<<<<<<<<<<
 *     def trading_pair(self) -> Dict[str, TradingPair]:
 *         return self._trading_pairs
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12trading_pair_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12trading_pair_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12trading_pair___get__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12trading_pair___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":205
 *     @property
 *     def trading_pair(self) -> Dict[str, TradingPair]:
 *         return self._trading_pairs             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_trading_pairs);
  __pyx_r = __pyx_v_self->_trading_pairs;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":203
 * 
 *     #  <editor-fold desc="Property">
 *     @property             # <<<<<<<<<<<<<<
 *     def trading_pair(self) -> Dict[str, TradingPair]:
 *         return self._trading_pairs
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":207
 *         return self._trading_pairs
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def trading_pairs(self) -> List[str]:
 *         return [trading_pair for trading_pair in self._trading_pairs]
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_13trading_pairs_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_13trading_pairs_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_13trading_pairs___get__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_13trading_pairs___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  PyObject *__pyx_8genexpr2__pyx_v_trading_pair = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":209
 *     @property
 *     def trading_pairs(self) -> List[str]:
 *         return [trading_pair for trading_pair in self._trading_pairs]             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 209, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = 0;
    if (unlikely(__pyx_v_self->_trading_pairs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 209, __pyx_L5_error)
    }
    __pyx_t_6 = __Pyx_dict_iterator(__pyx_v_self->_trading_pairs, 1, ((PyObject *)NULL), (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 209, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_6;
    __pyx_t_6 = 0;
    while (1) {
      __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, &__pyx_t_6, NULL, NULL, __pyx_t_5);
      if (unlikely(__pyx_t_7 == 0)) break;
      if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 209, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_trading_pair, __pyx_t_6);
      __pyx_t_6 = 0;
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_8genexpr2__pyx_v_trading_pair))) __PYX_ERR(0, 209, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_trading_pair); __pyx_8genexpr2__pyx_v_trading_pair = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_trading_pair); __pyx_8genexpr2__pyx_v_trading_pair = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":207
 *         return self._trading_pairs
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def trading_pairs(self) -> List[str]:
 *         return [trading_pair for trading_pair in self._trading_pairs]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.trading_pairs.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_trading_pair);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":211
 *         return [trading_pair for trading_pair in self._trading_pairs]
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def name(self) -> str:
 *         return self._exchange_name
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_4name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_4name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_4name___get__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_4name___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":213
 *     @property
 *     def name(self) -> str:
 *         return self._exchange_name             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_exchange_name);
  __pyx_r = __pyx_v_self->_exchange_name;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":211
 *         return [trading_pair for trading_pair in self._trading_pairs]
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def name(self) -> str:
 *         return self._exchange_name
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":215
 *         return self._exchange_name
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def display_name(self) -> str:
 *         return f"{self._exchange_name}_PaperTrade"
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12display_name_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12display_name_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12display_name___get__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12display_name___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":217
 *     @property
 *     def display_name(self) -> str:
 *         return f"{self._exchange_name}_PaperTrade"             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyUnicode_Unicode(__pyx_v_self->_exchange_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_1, __pyx_n_u_PaperTrade); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":215
 *         return self._exchange_name
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def display_name(self) -> str:
 *         return f"{self._exchange_name}_PaperTrade"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.display_name.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":219
 *         return f"{self._exchange_name}_PaperTrade"
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_books(self) -> Dict[str, CompositeOrderBook]:
 *         return self.order_book_tracker.order_books
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_11order_books_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_11order_books_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_11order_books___get__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_11order_books___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":221
 *     @property
 *     def order_books(self) -> Dict[str, CompositeOrderBook]:
 *         return self.order_book_tracker.order_books             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_book_tracker); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_order_books); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":219
 *         return f"{self._exchange_name}_PaperTrade"
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_books(self) -> Dict[str, CompositeOrderBook]:
 *         return self.order_book_tracker.order_books
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.order_books.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":223
 *         return self.order_book_tracker.order_books
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def status_dict(self) -> Dict[str, bool]:
 *         return {
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_11status_dict_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_11status_dict_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_11status_dict___get__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_11status_dict___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":225
 *     @property
 *     def status_dict(self) -> Dict[str, bool]:
 *         return {             # <<<<<<<<<<<<<<
 *             "order_books_initialized": self.order_book_tracker and len(self.order_book_tracker.order_books) > 0
 *         }
 */
  __Pyx_XDECREF(__pyx_r);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":226
 *     def status_dict(self) -> Dict[str, bool]:
 *         return {
 *             "order_books_initialized": self.order_book_tracker and len(self.order_book_tracker.order_books) > 0             # <<<<<<<<<<<<<<
 *         }
 * 
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_book_tracker); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 226, __pyx_L1_error)
  if (__pyx_t_4) {
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_2 = __pyx_t_3;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_book_tracker); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_order_books); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = PyObject_Length(__pyx_t_5); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_4 = (__pyx_t_6 > 0);
  __pyx_t_5 = __Pyx_PyBool_FromLong(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = __pyx_t_5;
  __pyx_t_5 = 0;
  __pyx_L3_bool_binop_done:;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_u_order_books_initialized, __pyx_t_2) < 0) __PYX_ERR(0, 226, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":223
 *         return self.order_book_tracker.order_books
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def status_dict(self) -> Dict[str, bool]:
 *         return {
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.status_dict.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":229
 *         }
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ready(self):
 *         if not self.order_book_tracker.ready:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_5ready_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_5ready_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_5ready___get__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_5ready___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":231
 *     @property
 *     def ready(self):
 *         if not self.order_book_tracker.ready:             # <<<<<<<<<<<<<<
 *             return False
 *         if all(self.status_dict.values()):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_book_tracker); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ready); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (!__pyx_t_3);
  if (__pyx_t_4) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":232
 *     def ready(self):
 *         if not self.order_book_tracker.ready:
 *             return False             # <<<<<<<<<<<<<<
 *         if all(self.status_dict.values()):
 *             if not self._paper_trade_market_initialized:
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":231
 *     @property
 *     def ready(self):
 *         if not self.order_book_tracker.ready:             # <<<<<<<<<<<<<<
 *             return False
 *         if all(self.status_dict.values()):
 */
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":233
 *         if not self.order_book_tracker.ready:
 *             return False
 *         if all(self.status_dict.values()):             # <<<<<<<<<<<<<<
 *             if not self._paper_trade_market_initialized:
 *                 self.init_paper_trade_market()
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_status_dict); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_values); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 233, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_builtin_all, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 233, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_4) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":234
 *             return False
 *         if all(self.status_dict.values()):
 *             if not self._paper_trade_market_initialized:             # <<<<<<<<<<<<<<
 *                 self.init_paper_trade_market()
 *                 self._paper_trade_market_initialized = True
 */
    __pyx_t_4 = (!__pyx_v_self->_paper_trade_market_initialized);
    if (__pyx_t_4) {

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":235
 *         if all(self.status_dict.values()):
 *             if not self._paper_trade_market_initialized:
 *                 self.init_paper_trade_market()             # <<<<<<<<<<<<<<
 *                 self._paper_trade_market_initialized = True
 *             return True
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_init_paper_trade_market); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 235, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 235, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":236
 *             if not self._paper_trade_market_initialized:
 *                 self.init_paper_trade_market()
 *                 self._paper_trade_market_initialized = True             # <<<<<<<<<<<<<<
 *             return True
 *         else:
 */
      __pyx_v_self->_paper_trade_market_initialized = 1;

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":234
 *             return False
 *         if all(self.status_dict.values()):
 *             if not self._paper_trade_market_initialized:             # <<<<<<<<<<<<<<
 *                 self.init_paper_trade_market()
 *                 self._paper_trade_market_initialized = True
 */
    }

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":237
 *                 self.init_paper_trade_market()
 *                 self._paper_trade_market_initialized = True
 *             return True             # <<<<<<<<<<<<<<
 *         else:
 *             return False
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_True);
    __pyx_r = Py_True;
    goto __pyx_L0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":233
 *         if not self.order_book_tracker.ready:
 *             return False
 *         if all(self.status_dict.values()):             # <<<<<<<<<<<<<<
 *             if not self._paper_trade_market_initialized:
 *                 self.init_paper_trade_market()
 */
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":239
 *             return True
 *         else:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":229
 *         }
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def ready(self):
 *         if not self.order_book_tracker.ready:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.ready.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":241
 *             return False
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def queued_orders(self) -> List[QueuedOrder]:
 *         return self._queued_orders
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_13queued_orders_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_13queued_orders_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_13queued_orders___get__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_13queued_orders___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":243
 *     @property
 *     def queued_orders(self) -> List[QueuedOrder]:
 *         return self._queued_orders             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_queued_orders);
  __pyx_r = __pyx_v_self->_queued_orders;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":241
 *             return False
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def queued_orders(self) -> List[QueuedOrder]:
 *         return self._queued_orders
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":245
 *         return self._queued_orders
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def limit_orders(self) -> List[LimitOrder]:
 *         cdef:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12limit_orders_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12limit_orders_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12limit_orders___get__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12limit_orders___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator __pyx_v_map_it;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders *__pyx_v_single_trading_pair_collection_ptr;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator __pyx_v_collection_it;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersRIterator __pyx_v_collection_rit;
  LimitOrder const *__pyx_v_cpp_limit_order_ptr;
  PyObject *__pyx_v_retval = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":253
 *             SingleTradingPairLimitOrdersRIterator collection_rit
 *             const CPPLimitOrder *cpp_limit_order_ptr
 *             list retval = []             # <<<<<<<<<<<<<<
 * 
 *         map_it = self._bid_limit_orders.begin()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_retval = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":255
 *             list retval = []
 * 
 *         map_it = self._bid_limit_orders.begin()             # <<<<<<<<<<<<<<
 *         while map_it != self._bid_limit_orders.end():
 *             single_trading_pair_collection_ptr = address(deref(map_it).second)
 */
  __pyx_v_map_it = __pyx_v_self->_bid_limit_orders.begin();

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":256
 * 
 *         map_it = self._bid_limit_orders.begin()
 *         while map_it != self._bid_limit_orders.end():             # <<<<<<<<<<<<<<
 *             single_trading_pair_collection_ptr = address(deref(map_it).second)
 *             collection_rit = single_trading_pair_collection_ptr.rbegin()
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_map_it != __pyx_v_self->_bid_limit_orders.end());
    if (!__pyx_t_2) break;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":257
 *         map_it = self._bid_limit_orders.begin()
 *         while map_it != self._bid_limit_orders.end():
 *             single_trading_pair_collection_ptr = address(deref(map_it).second)             # <<<<<<<<<<<<<<
 *             collection_rit = single_trading_pair_collection_ptr.rbegin()
 *             while collection_rit != single_trading_pair_collection_ptr.rend():
 */
    __pyx_v_single_trading_pair_collection_ptr = (&(*__pyx_v_map_it).second);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":258
 *         while map_it != self._bid_limit_orders.end():
 *             single_trading_pair_collection_ptr = address(deref(map_it).second)
 *             collection_rit = single_trading_pair_collection_ptr.rbegin()             # <<<<<<<<<<<<<<
 *             while collection_rit != single_trading_pair_collection_ptr.rend():
 *                 cpp_limit_order_ptr = address(deref(collection_rit))
 */
    __pyx_v_collection_rit = __pyx_v_single_trading_pair_collection_ptr->rbegin();

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":259
 *             single_trading_pair_collection_ptr = address(deref(map_it).second)
 *             collection_rit = single_trading_pair_collection_ptr.rbegin()
 *             while collection_rit != single_trading_pair_collection_ptr.rend():             # <<<<<<<<<<<<<<
 *                 cpp_limit_order_ptr = address(deref(collection_rit))
 *                 retval.append(c_create_limit_order_from_cpp_limit_order(deref(cpp_limit_order_ptr)))
 */
    while (1) {
      __pyx_t_2 = (__pyx_v_collection_rit != __pyx_v_single_trading_pair_collection_ptr->rend());
      if (!__pyx_t_2) break;

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":260
 *             collection_rit = single_trading_pair_collection_ptr.rbegin()
 *             while collection_rit != single_trading_pair_collection_ptr.rend():
 *                 cpp_limit_order_ptr = address(deref(collection_rit))             # <<<<<<<<<<<<<<
 *                 retval.append(c_create_limit_order_from_cpp_limit_order(deref(cpp_limit_order_ptr)))
 *                 inc(collection_rit)
 */
      __pyx_v_cpp_limit_order_ptr = (&(*__pyx_v_collection_rit));

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":261
 *             while collection_rit != single_trading_pair_collection_ptr.rend():
 *                 cpp_limit_order_ptr = address(deref(collection_rit))
 *                 retval.append(c_create_limit_order_from_cpp_limit_order(deref(cpp_limit_order_ptr)))             # <<<<<<<<<<<<<<
 *                 inc(collection_rit)
 *             inc(map_it)
 */
      __pyx_t_1 = ((PyObject *)__pyx_f_10hummingbot_4core_9data_type_11limit_order_c_create_limit_order_from_cpp_limit_order((*__pyx_v_cpp_limit_order_ptr))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_retval, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 261, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":262
 *                 cpp_limit_order_ptr = address(deref(collection_rit))
 *                 retval.append(c_create_limit_order_from_cpp_limit_order(deref(cpp_limit_order_ptr)))
 *                 inc(collection_rit)             # <<<<<<<<<<<<<<
 *             inc(map_it)
 * 
 */
      (void)((__pyx_v_collection_rit++));
    }

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":263
 *                 retval.append(c_create_limit_order_from_cpp_limit_order(deref(cpp_limit_order_ptr)))
 *                 inc(collection_rit)
 *             inc(map_it)             # <<<<<<<<<<<<<<
 * 
 *         map_it = self._ask_limit_orders.begin()
 */
    (void)((__pyx_v_map_it++));
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":265
 *             inc(map_it)
 * 
 *         map_it = self._ask_limit_orders.begin()             # <<<<<<<<<<<<<<
 *         while map_it != self._ask_limit_orders.end():
 *             single_trading_pair_collection_ptr = address(deref(map_it).second)
 */
  __pyx_v_map_it = __pyx_v_self->_ask_limit_orders.begin();

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":266
 * 
 *         map_it = self._ask_limit_orders.begin()
 *         while map_it != self._ask_limit_orders.end():             # <<<<<<<<<<<<<<
 *             single_trading_pair_collection_ptr = address(deref(map_it).second)
 *             collection_it = single_trading_pair_collection_ptr.begin()
 */
  while (1) {
    __pyx_t_2 = (__pyx_v_map_it != __pyx_v_self->_ask_limit_orders.end());
    if (!__pyx_t_2) break;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":267
 *         map_it = self._ask_limit_orders.begin()
 *         while map_it != self._ask_limit_orders.end():
 *             single_trading_pair_collection_ptr = address(deref(map_it).second)             # <<<<<<<<<<<<<<
 *             collection_it = single_trading_pair_collection_ptr.begin()
 *             while collection_it != single_trading_pair_collection_ptr.end():
 */
    __pyx_v_single_trading_pair_collection_ptr = (&(*__pyx_v_map_it).second);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":268
 *         while map_it != self._ask_limit_orders.end():
 *             single_trading_pair_collection_ptr = address(deref(map_it).second)
 *             collection_it = single_trading_pair_collection_ptr.begin()             # <<<<<<<<<<<<<<
 *             while collection_it != single_trading_pair_collection_ptr.end():
 *                 cpp_limit_order_ptr = address(deref(collection_it))
 */
    __pyx_v_collection_it = __pyx_v_single_trading_pair_collection_ptr->begin();

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":269
 *             single_trading_pair_collection_ptr = address(deref(map_it).second)
 *             collection_it = single_trading_pair_collection_ptr.begin()
 *             while collection_it != single_trading_pair_collection_ptr.end():             # <<<<<<<<<<<<<<
 *                 cpp_limit_order_ptr = address(deref(collection_it))
 *                 retval.append(c_create_limit_order_from_cpp_limit_order(deref(cpp_limit_order_ptr)))
 */
    while (1) {
      __pyx_t_2 = (__pyx_v_collection_it != __pyx_v_single_trading_pair_collection_ptr->end());
      if (!__pyx_t_2) break;

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":270
 *             collection_it = single_trading_pair_collection_ptr.begin()
 *             while collection_it != single_trading_pair_collection_ptr.end():
 *                 cpp_limit_order_ptr = address(deref(collection_it))             # <<<<<<<<<<<<<<
 *                 retval.append(c_create_limit_order_from_cpp_limit_order(deref(cpp_limit_order_ptr)))
 *                 inc(collection_it)
 */
      __pyx_v_cpp_limit_order_ptr = (&(*__pyx_v_collection_it));

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":271
 *             while collection_it != single_trading_pair_collection_ptr.end():
 *                 cpp_limit_order_ptr = address(deref(collection_it))
 *                 retval.append(c_create_limit_order_from_cpp_limit_order(deref(cpp_limit_order_ptr)))             # <<<<<<<<<<<<<<
 *                 inc(collection_it)
 *             inc(map_it)
 */
      __pyx_t_1 = ((PyObject *)__pyx_f_10hummingbot_4core_9data_type_11limit_order_c_create_limit_order_from_cpp_limit_order((*__pyx_v_cpp_limit_order_ptr))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyList_Append(__pyx_v_retval, __pyx_t_1); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(0, 271, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":272
 *                 cpp_limit_order_ptr = address(deref(collection_it))
 *                 retval.append(c_create_limit_order_from_cpp_limit_order(deref(cpp_limit_order_ptr)))
 *                 inc(collection_it)             # <<<<<<<<<<<<<<
 *             inc(map_it)
 * 
 */
      (void)((__pyx_v_collection_it++));
    }

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":273
 *                 retval.append(c_create_limit_order_from_cpp_limit_order(deref(cpp_limit_order_ptr)))
 *                 inc(collection_it)
 *             inc(map_it)             # <<<<<<<<<<<<<<
 * 
 *         return retval
 */
    (void)((__pyx_v_map_it++));
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":275
 *             inc(map_it)
 * 
 *         return retval             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_retval);
  __pyx_r = __pyx_v_retval;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":245
 *         return self._queued_orders
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def limit_orders(self) -> List[LimitOrder]:
 *         cdef:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.limit_orders.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_retval);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":277
 *         return retval
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def on_hold_balances(self) -> Dict[str, Decimal]:
 *         _on_hold_balances = defaultdict(Decimal)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_16on_hold_balances_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_16on_hold_balances_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_16on_hold_balances___get__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_16on_hold_balances___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  PyObject *__pyx_v__on_hold_balances = NULL;
  PyObject *__pyx_v_limit_order = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":279
 *     @property
 *     def on_hold_balances(self) -> Dict[str, Decimal]:
 *         _on_hold_balances = defaultdict(Decimal)             # <<<<<<<<<<<<<<
 *         for limit_order in self.limit_orders:
 *             if limit_order.is_buy:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_defaultdict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v__on_hold_balances = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":280
 *     def on_hold_balances(self) -> Dict[str, Decimal]:
 *         _on_hold_balances = defaultdict(Decimal)
 *         for limit_order in self.limit_orders:             # <<<<<<<<<<<<<<
 *             if limit_order.is_buy:
 *                 _on_hold_balances[limit_order.quote_currency] += limit_order.quantity * limit_order.price
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_limit_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 280, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 280, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 280, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 280, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 280, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 280, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_7(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 280, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_limit_order, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":281
 *         _on_hold_balances = defaultdict(Decimal)
 *         for limit_order in self.limit_orders:
 *             if limit_order.is_buy:             # <<<<<<<<<<<<<<
 *                 _on_hold_balances[limit_order.quote_currency] += limit_order.quantity * limit_order.price
 *             else:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order, __pyx_n_s_is_buy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 281, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_8) {

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":282
 *         for limit_order in self.limit_orders:
 *             if limit_order.is_buy:
 *                 _on_hold_balances[limit_order.quote_currency] += limit_order.quantity * limit_order.price             # <<<<<<<<<<<<<<
 *             else:
 *                 _on_hold_balances[limit_order.base_currency] += limit_order.quantity
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order, __pyx_n_s_quote_currency); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 282, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_v__on_hold_balances, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 282, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order, __pyx_n_s_quantity); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 282, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order, __pyx_n_s_price); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 282, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = PyNumber_Multiply(__pyx_t_4, __pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 282, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 282, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely((PyObject_SetItem(__pyx_v__on_hold_balances, __pyx_t_1, __pyx_t_9) < 0))) __PYX_ERR(0, 282, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":281
 *         _on_hold_balances = defaultdict(Decimal)
 *         for limit_order in self.limit_orders:
 *             if limit_order.is_buy:             # <<<<<<<<<<<<<<
 *                 _on_hold_balances[limit_order.quote_currency] += limit_order.quantity * limit_order.price
 *             else:
 */
      goto __pyx_L5;
    }

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":284
 *                 _on_hold_balances[limit_order.quote_currency] += limit_order.quantity * limit_order.price
 *             else:
 *                 _on_hold_balances[limit_order.base_currency] += limit_order.quantity             # <<<<<<<<<<<<<<
 *         return _on_hold_balances
 * 
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order, __pyx_n_s_base_currency); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = __Pyx_PyObject_GetItem(__pyx_v__on_hold_balances, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_limit_order, __pyx_n_s_quantity); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely((PyObject_SetItem(__pyx_v__on_hold_balances, __pyx_t_1, __pyx_t_3) < 0))) __PYX_ERR(0, 284, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_L5:;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":280
 *     def on_hold_balances(self) -> Dict[str, Decimal]:
 *         _on_hold_balances = defaultdict(Decimal)
 *         for limit_order in self.limit_orders:             # <<<<<<<<<<<<<<
 *             if limit_order.is_buy:
 *                 _on_hold_balances[limit_order.quote_currency] += limit_order.quantity * limit_order.price
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":285
 *             else:
 *                 _on_hold_balances[limit_order.base_currency] += limit_order.quantity
 *         return _on_hold_balances             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v__on_hold_balances);
  __pyx_r = __pyx_v__on_hold_balances;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":277
 *         return retval
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def on_hold_balances(self) -> Dict[str, Decimal]:
 *         _on_hold_balances = defaultdict(Decimal)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.on_hold_balances.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__on_hold_balances);
  __Pyx_XDECREF(__pyx_v_limit_order);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":287
 *         return _on_hold_balances
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def available_balances(self) -> Dict[str, Decimal]:
 *         _available_balances = self._account_balances.copy()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_18available_balances_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_18available_balances_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_18available_balances___get__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_18available_balances___get__(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  PyObject *__pyx_v__available_balances = NULL;
  PyObject *__pyx_v_trading_pair_str = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_balance = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":289
 *     @property
 *     def available_balances(self) -> Dict[str, Decimal]:
 *         _available_balances = self._account_balances.copy()             # <<<<<<<<<<<<<<
 *         for trading_pair_str, balance in _available_balances.items():
 *             _available_balances[trading_pair_str] -= self.on_hold_balances[trading_pair_str]
 */
  if (unlikely(__pyx_v_self->__pyx_base.__pyx_base._account_balances == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "copy");
    __PYX_ERR(0, 289, __pyx_L1_error)
  }
  __pyx_t_1 = PyDict_Copy(__pyx_v_self->__pyx_base.__pyx_base._account_balances); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 289, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__available_balances = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":290
 *     def available_balances(self) -> Dict[str, Decimal]:
 *         _available_balances = self._account_balances.copy()
 *         for trading_pair_str, balance in _available_balances.items():             # <<<<<<<<<<<<<<
 *             _available_balances[trading_pair_str] -= self.on_hold_balances[trading_pair_str]
 *         return _available_balances
 */
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v__available_balances == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 290, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_v__available_balances, 1, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, &__pyx_t_6, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 290, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF_SET(__pyx_v_trading_pair_str, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_balance, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":291
 *         _available_balances = self._account_balances.copy()
 *         for trading_pair_str, balance in _available_balances.items():
 *             _available_balances[trading_pair_str] -= self.on_hold_balances[trading_pair_str]             # <<<<<<<<<<<<<<
 *         return _available_balances
 * 
 */
    if (unlikely(__pyx_v__available_balances == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 291, __pyx_L1_error)
    }
    __Pyx_INCREF(__pyx_v__available_balances);
    __pyx_t_8 = __pyx_v__available_balances;
    __Pyx_INCREF(__pyx_v_trading_pair_str);
    __pyx_t_6 = __pyx_v_trading_pair_str;
    if (unlikely(__pyx_t_8 == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 291, __pyx_L1_error)
    }
    __pyx_t_5 = __Pyx_PyDict_GetItem(__pyx_t_8, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 291, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_on_hold_balances); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 291, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_t_9, __pyx_v_trading_pair_str); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 291, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyNumber_InPlaceSubtract(__pyx_t_5, __pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 291, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_t_8 == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 291, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__pyx_t_8, __pyx_t_6, __pyx_t_9) < 0))) __PYX_ERR(0, 291, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":292
 *         for trading_pair_str, balance in _available_balances.items():
 *             _available_balances[trading_pair_str] -= self.on_hold_balances[trading_pair_str]
 *         return _available_balances             # <<<<<<<<<<<<<<
 * 
 *     # </editor-fold>
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v__available_balances);
  __pyx_r = __pyx_v__available_balances;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":287
 *         return _on_hold_balances
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def available_balances(self) -> Dict[str, Decimal]:
 *         _available_balances = self._account_balances.copy()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.available_balances.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__available_balances);
  __Pyx_XDECREF(__pyx_v_trading_pair_str);
  __Pyx_XDECREF(__pyx_v_balance);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":296
 *     # </editor-fold>
 * 
 *     cdef c_start(self, Clock clock, double timestamp):             # <<<<<<<<<<<<<<
 *         ExchangeBase.c_start(self, clock, timestamp)
 * 
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_start(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, struct __pyx_obj_10hummingbot_4core_5clock_Clock *__pyx_v_clock, double __pyx_v_timestamp) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_start", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":297
 * 
 *     cdef c_start(self, Clock clock, double timestamp):
 *         ExchangeBase.c_start(self, clock, timestamp)             # <<<<<<<<<<<<<<
 * 
 *     async def start_network(self):
 */
  __pyx_t_1 = __pyx_vtabptr_10hummingbot_9connector_13exchange_base_ExchangeBase->__pyx_base.__pyx_base.__pyx_base.c_start(((struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *)__pyx_v_self), __pyx_v_clock, __pyx_v_timestamp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":296
 *     # </editor-fold>
 * 
 *     cdef c_start(self, Clock clock, double timestamp):             # <<<<<<<<<<<<<<
 *         ExchangeBase.c_start(self, clock, timestamp)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_10generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":299
 *         ExchangeBase.c_start(self, clock, timestamp)
 * 
 *     async def start_network(self):             # <<<<<<<<<<<<<<
 *         await self.stop_network()
 *         self.order_book_tracker.start()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_9start_network(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_9start_network = {"start_network", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_9start_network, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_9start_network(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("start_network (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("start_network", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "start_network", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_8start_network(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_8start_network(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("start_network", 0);
  __pyx_cur_scope = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network *)__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 299, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_10generator, __pyx_codeobj__12, (PyObject *) __pyx_cur_scope, __pyx_n_s_start_network, __pyx_n_s_PaperTradeExchange_start_network, __pyx_n_s_hummingbot_connector_exchange_pa); if (unlikely(!gen)) __PYX_ERR(0, 299, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.start_network", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_10generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network *__pyx_cur_scope = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("start_network", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L4_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 299, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":300
 * 
 *     async def start_network(self):
 *         await self.stop_network()             # <<<<<<<<<<<<<<
 *         self.order_book_tracker.start()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_stop_network); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 300, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, awaiting value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L4_resume_from_await:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 300, __pyx_L1_error)
  } else {
    PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (exc_type) {
      if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
      else __PYX_ERR(0, 300, __pyx_L1_error)
    }
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":301
 *     async def start_network(self):
 *         await self.stop_network()
 *         self.order_book_tracker.start()             # <<<<<<<<<<<<<<
 * 
 *     async def stop_network(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_order_book_tracker); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_start); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 301, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 301, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":299
 *         ExchangeBase.c_start(self, clock, timestamp)
 * 
 *     async def start_network(self):             # <<<<<<<<<<<<<<
 *         await self.stop_network()
 *         self.order_book_tracker.start()
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("start_network", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_13generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":303
 *         self.order_book_tracker.start()
 * 
 *     async def stop_network(self):             # <<<<<<<<<<<<<<
 *         self.order_book_tracker.stop()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12stop_network(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12stop_network = {"stop_network", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12stop_network, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12stop_network(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("stop_network (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("stop_network", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "stop_network", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_11stop_network(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_11stop_network(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("stop_network", 0);
  __pyx_cur_scope = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network *)__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 303, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_13generator1, __pyx_codeobj__13, (PyObject *) __pyx_cur_scope, __pyx_n_s_stop_network, __pyx_n_s_PaperTradeExchange_stop_network, __pyx_n_s_hummingbot_connector_exchange_pa); if (unlikely(!gen)) __PYX_ERR(0, 303, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.stop_network", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_13generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network *__pyx_cur_scope = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("stop_network", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 303, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":304
 * 
 *     async def stop_network(self):
 *         self.order_book_tracker.stop()             # <<<<<<<<<<<<<<
 * 
 *     async def check_network(self) -> NetworkStatus:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_order_book_tracker); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_stop); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":303
 *         self.order_book_tracker.start()
 * 
 *     async def stop_network(self):             # <<<<<<<<<<<<<<
 *         self.order_book_tracker.stop()
 * 
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("stop_network", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_16generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":306
 *         self.order_book_tracker.stop()
 * 
 *     async def check_network(self) -> NetworkStatus:             # <<<<<<<<<<<<<<
 *         return NetworkStatus.CONNECTED
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_15check_network(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_15check_network = {"check_network", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_15check_network, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_15check_network(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("check_network (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("check_network", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "check_network", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_14check_network(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_14check_network(CYTHON_UNUSED struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("check_network", 0);
  __pyx_cur_scope = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network *)__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 306, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_16generator2, __pyx_codeobj__14, (PyObject *) __pyx_cur_scope, __pyx_n_s_check_network, __pyx_n_s_PaperTradeExchange_check_network, __pyx_n_s_hummingbot_connector_exchange_pa); if (unlikely(!gen)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.check_network", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_16generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network *__pyx_cur_scope = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("check_network", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 306, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":307
 * 
 *     async def check_network(self) -> NetworkStatus:
 *         return NetworkStatus.CONNECTED             # <<<<<<<<<<<<<<
 * 
 *     cdef c_set_balance(self, str currency, object balance):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NetworkStatus); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_CONNECTED); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  goto __pyx_L0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":306
 *         self.order_book_tracker.stop()
 * 
 *     async def check_network(self) -> NetworkStatus:             # <<<<<<<<<<<<<<
 *         return NetworkStatus.CONNECTED
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("check_network", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":309
 *         return NetworkStatus.CONNECTED
 * 
 *     cdef c_set_balance(self, str currency, object balance):             # <<<<<<<<<<<<<<
 *         self._account_balances[currency.upper()] = Decimal(balance)
 * 
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_set_balance(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_currency, PyObject *__pyx_v_balance) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_set_balance", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":310
 * 
 *     cdef c_set_balance(self, str currency, object balance):
 *         self._account_balances[currency.upper()] = Decimal(balance)             # <<<<<<<<<<<<<<
 * 
 *     cdef object c_get_balance(self, str currency):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_balance};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 310, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if (unlikely(__pyx_v_self->__pyx_base.__pyx_base._account_balances == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 310, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyUnicode_Type_upper, __pyx_v_currency); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely((PyDict_SetItem(__pyx_v_self->__pyx_base.__pyx_base._account_balances, __pyx_t_2, __pyx_t_1) < 0))) __PYX_ERR(0, 310, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":309
 *         return NetworkStatus.CONNECTED
 * 
 *     cdef c_set_balance(self, str currency, object balance):             # <<<<<<<<<<<<<<
 *         self._account_balances[currency.upper()] = Decimal(balance)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_set_balance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":312
 *         self._account_balances[currency.upper()] = Decimal(balance)
 * 
 *     cdef object c_get_balance(self, str currency):             # <<<<<<<<<<<<<<
 *         if currency.upper() not in self._account_balances:
 *             self.logger().warning(f"Account balance does not have asset {currency.upper()}.")
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_balance(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_currency) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_get_balance", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":313
 * 
 *     cdef object c_get_balance(self, str currency):
 *         if currency.upper() not in self._account_balances:             # <<<<<<<<<<<<<<
 *             self.logger().warning(f"Account balance does not have asset {currency.upper()}.")
 *             return Decimal(0.0)
 */
  __pyx_t_1 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyUnicode_Type_upper, __pyx_v_currency); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(__pyx_v_self->__pyx_base.__pyx_base._account_balances == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 313, __pyx_L1_error)
  }
  __pyx_t_2 = (__Pyx_PyDict_ContainsTF(__pyx_t_1, __pyx_v_self->__pyx_base.__pyx_base._account_balances, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":314
 *     cdef object c_get_balance(self, str currency):
 *         if currency.upper() not in self._account_balances:
 *             self.logger().warning(f"Account balance does not have asset {currency.upper()}.")             # <<<<<<<<<<<<<<
 *             return Decimal(0.0)
 *         return self._account_balances[currency.upper()]
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 314, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warning); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = 0;
    __pyx_t_8 = 127;
    __Pyx_INCREF(__pyx_kp_u_Account_balance_does_not_have_as);
    __pyx_t_7 += 36;
    __Pyx_GIVEREF(__pyx_kp_u_Account_balance_does_not_have_as);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Account_balance_does_not_have_as);
    __pyx_t_5 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyUnicode_Type_upper, __pyx_v_currency); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = __Pyx_PyObject_FormatSimple(__pyx_t_5, __pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) : __pyx_t_8;
    __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_9);
    __pyx_t_9 = 0;
    __Pyx_INCREF(__pyx_kp_u__15);
    __pyx_t_7 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__15);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__15);
    __pyx_t_9 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 314, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_9};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 314, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":315
 *         if currency.upper() not in self._account_balances:
 *             self.logger().warning(f"Account balance does not have asset {currency.upper()}.")
 *             return Decimal(0.0)             # <<<<<<<<<<<<<<
 *         return self._account_balances[currency.upper()]
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_float_0_0};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 315, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":313
 * 
 *     cdef object c_get_balance(self, str currency):
 *         if currency.upper() not in self._account_balances:             # <<<<<<<<<<<<<<
 *             self.logger().warning(f"Account balance does not have asset {currency.upper()}.")
 *             return Decimal(0.0)
 */
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":316
 *             self.logger().warning(f"Account balance does not have asset {currency.upper()}.")
 *             return Decimal(0.0)
 *         return self._account_balances[currency.upper()]             # <<<<<<<<<<<<<<
 * 
 *     cdef c_tick(self, double timestamp):
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->__pyx_base.__pyx_base._account_balances == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 316, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyUnicode_Type_upper, __pyx_v_currency); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyDict_GetItem(__pyx_v_self->__pyx_base.__pyx_base._account_balances, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 316, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":312
 *         self._account_balances[currency.upper()] = Decimal(balance)
 * 
 *     cdef object c_get_balance(self, str currency):             # <<<<<<<<<<<<<<
 *         if currency.upper() not in self._account_balances:
 *             self.logger().warning(f"Account balance does not have asset {currency.upper()}.")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_get_balance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":318
 *         return self._account_balances[currency.upper()]
 * 
 *     cdef c_tick(self, double timestamp):             # <<<<<<<<<<<<<<
 *         ExchangeBase.c_tick(self, timestamp)
 *         self.c_process_market_orders()
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_tick(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, double __pyx_v_timestamp) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_tick", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":319
 * 
 *     cdef c_tick(self, double timestamp):
 *         ExchangeBase.c_tick(self, timestamp)             # <<<<<<<<<<<<<<
 *         self.c_process_market_orders()
 *         self.c_process_crossed_limit_orders()
 */
  __pyx_t_1 = __pyx_vtabptr_10hummingbot_9connector_13exchange_base_ExchangeBase->__pyx_base.__pyx_base.__pyx_base.c_tick(((struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *)__pyx_v_self), __pyx_v_timestamp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":320
 *     cdef c_tick(self, double timestamp):
 *         ExchangeBase.c_tick(self, timestamp)
 *         self.c_process_market_orders()             # <<<<<<<<<<<<<<
 *         self.c_process_crossed_limit_orders()
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_process_market_orders(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 320, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":321
 *         ExchangeBase.c_tick(self, timestamp)
 *         self.c_process_market_orders()
 *         self.c_process_crossed_limit_orders()             # <<<<<<<<<<<<<<
 * 
 *     cdef str c_buy(self,
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_process_crossed_limit_orders(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":318
 *         return self._account_balances[currency.upper()]
 * 
 *     cdef c_tick(self, double timestamp):             # <<<<<<<<<<<<<<
 *         ExchangeBase.c_tick(self, timestamp)
 *         self.c_process_market_orders()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_tick", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":323
 *         self.c_process_crossed_limit_orders()
 * 
 *     cdef str c_buy(self,             # <<<<<<<<<<<<<<
 *                    str trading_pair_str,
 *                    object amount,
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_buy(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair_str, PyObject *__pyx_v_amount, struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_buy *__pyx_optional_args) {
  PyObject *__pyx_v_order_type = __pyx_k__16;
  PyObject *__pyx_v_price = __pyx_k__17;
  PyObject *__pyx_v_order_id = 0;
  PyObject *__pyx_v_quote_asset = 0;
  std::string __pyx_v_cpp_order_id;
  std::string __pyx_v_cpp_trading_pair_str;
  std::string __pyx_v_cpp_base_asset;
  std::string __pyx_v_cpp_quote_asset;
  std::string __pyx_v_cpp_position;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator __pyx_v_map_it;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders *__pyx_v_limit_orders_collection_ptr;
  std::pair<std::unordered_map<std::string,__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders> ::iterator,bool>  __pyx_v_insert_result;
  PyObject *__pyx_v_quantized_price = NULL;
  PyObject *__pyx_v_quantized_amount = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_UCS4 __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  std::string __pyx_t_8;
  int __pyx_t_9;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders __pyx_t_10;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersPair __pyx_t_11;
  std::pair<std::unordered_map<std::string,__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders> ::iterator,int>  __pyx_t_12;
  std::unordered_map<std::string,__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders> ::iterator __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_buy", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_order_type = __pyx_optional_args->order_type;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_price = __pyx_optional_args->price;
      }
    }
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":329
 *                    object price=s_decimal_0,
 *                    dict kwargs={}):
 *         if trading_pair_str not in self._trading_pairs:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Trading pair '{trading_pair_str}' does not existing in current data set.")
 * 
 */
  if (unlikely(__pyx_v_self->_trading_pairs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 329, __pyx_L1_error)
  }
  __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_v_trading_pair_str, __pyx_v_self->_trading_pairs, Py_NE)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 329, __pyx_L1_error)
  if (unlikely(__pyx_t_1)) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":330
 *                    dict kwargs={}):
 *         if trading_pair_str not in self._trading_pairs:
 *             raise ValueError(f"Trading pair '{trading_pair_str}' does not existing in current data set.")             # <<<<<<<<<<<<<<
 * 
 *         cdef:
 */
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = 127;
    __Pyx_INCREF(__pyx_kp_u_Trading_pair);
    __pyx_t_3 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_Trading_pair);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_Trading_pair);
    __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_v_trading_pair_str); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_4;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u_does_not_existing_in_current_da);
    __pyx_t_3 += 40;
    __Pyx_GIVEREF(__pyx_kp_u_does_not_existing_in_current_da);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_does_not_existing_in_current_da);
    __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_2, 3, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 330, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 330, __pyx_L1_error)

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":329
 *                    object price=s_decimal_0,
 *                    dict kwargs={}):
 *         if trading_pair_str not in self._trading_pairs:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Trading pair '{trading_pair_str}' does not existing in current data set.")
 * 
 */
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":333
 * 
 *         cdef:
 *             str order_id = self.random_order_id("buy", trading_pair_str)             # <<<<<<<<<<<<<<
 *             str quote_asset = self._trading_pairs[trading_pair_str].quote_asset
 *             string cpp_order_id = order_id.encode("utf8")
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_random_order_id); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 333, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_n_u_buy, __pyx_v_trading_pair_str};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 333, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_2))) __PYX_ERR(0, 333, __pyx_L1_error)
  __pyx_v_order_id = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":334
 *         cdef:
 *             str order_id = self.random_order_id("buy", trading_pair_str)
 *             str quote_asset = self._trading_pairs[trading_pair_str].quote_asset             # <<<<<<<<<<<<<<
 *             string cpp_order_id = order_id.encode("utf8")
 *             string cpp_trading_pair_str = trading_pair_str.encode("utf8")
 */
  if (unlikely(__pyx_v_self->_trading_pairs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 334, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_self->_trading_pairs, __pyx_v_trading_pair_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_quote_asset); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_5))) __PYX_ERR(0, 334, __pyx_L1_error)
  __pyx_v_quote_asset = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":335
 *             str order_id = self.random_order_id("buy", trading_pair_str)
 *             str quote_asset = self._trading_pairs[trading_pair_str].quote_asset
 *             string cpp_order_id = order_id.encode("utf8")             # <<<<<<<<<<<<<<
 *             string cpp_trading_pair_str = trading_pair_str.encode("utf8")
 *             string cpp_base_asset = self._trading_pairs[trading_pair_str].base_asset.encode("utf8")
 */
  if (unlikely(__pyx_v_order_id == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 335, __pyx_L1_error)
  }
  __pyx_t_5 = PyUnicode_AsUTF8String(__pyx_v_order_id); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __pyx_convert_string_from_py_6libcpp_6string_std__in_string(__pyx_t_5); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 335, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_cpp_order_id = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_8);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":336
 *             str quote_asset = self._trading_pairs[trading_pair_str].quote_asset
 *             string cpp_order_id = order_id.encode("utf8")
 *             string cpp_trading_pair_str = trading_pair_str.encode("utf8")             # <<<<<<<<<<<<<<
 *             string cpp_base_asset = self._trading_pairs[trading_pair_str].base_asset.encode("utf8")
 *             string cpp_quote_asset = quote_asset.encode("utf8")
 */
  if (unlikely(__pyx_v_trading_pair_str == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 336, __pyx_L1_error)
  }
  __pyx_t_5 = PyUnicode_AsUTF8String(__pyx_v_trading_pair_str); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __pyx_convert_string_from_py_6libcpp_6string_std__in_string(__pyx_t_5); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_cpp_trading_pair_str = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_8);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":337
 *             string cpp_order_id = order_id.encode("utf8")
 *             string cpp_trading_pair_str = trading_pair_str.encode("utf8")
 *             string cpp_base_asset = self._trading_pairs[trading_pair_str].base_asset.encode("utf8")             # <<<<<<<<<<<<<<
 *             string cpp_quote_asset = quote_asset.encode("utf8")
 *             string cpp_position = "NIL".encode("utf8")
 */
  if (unlikely(__pyx_v_self->_trading_pairs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 337, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_self->_trading_pairs, __pyx_v_trading_pair_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_base_asset); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_n_u_utf8};
    __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 337, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_8 = __pyx_convert_string_from_py_6libcpp_6string_std__in_string(__pyx_t_5); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 337, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_cpp_base_asset = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_8);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":338
 *             string cpp_trading_pair_str = trading_pair_str.encode("utf8")
 *             string cpp_base_asset = self._trading_pairs[trading_pair_str].base_asset.encode("utf8")
 *             string cpp_quote_asset = quote_asset.encode("utf8")             # <<<<<<<<<<<<<<
 *             string cpp_position = "NIL".encode("utf8")
 *             LimitOrdersIterator map_it
 */
  if (unlikely(__pyx_v_quote_asset == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 338, __pyx_L1_error)
  }
  __pyx_t_5 = PyUnicode_AsUTF8String(__pyx_v_quote_asset); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __pyx_convert_string_from_py_6libcpp_6string_std__in_string(__pyx_t_5); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 338, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_cpp_quote_asset = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_8);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":339
 *             string cpp_base_asset = self._trading_pairs[trading_pair_str].base_asset.encode("utf8")
 *             string cpp_quote_asset = quote_asset.encode("utf8")
 *             string cpp_position = "NIL".encode("utf8")             # <<<<<<<<<<<<<<
 *             LimitOrdersIterator map_it
 *             SingleTradingPairLimitOrders *limit_orders_collection_ptr = NULL
 */
  __pyx_t_8 = __pyx_convert_string_from_py_6libcpp_6string_std__in_string(__pyx_n_b_NIL); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 339, __pyx_L1_error)
  __pyx_v_cpp_position = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_8);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":341
 *             string cpp_position = "NIL".encode("utf8")
 *             LimitOrdersIterator map_it
 *             SingleTradingPairLimitOrders *limit_orders_collection_ptr = NULL             # <<<<<<<<<<<<<<
 *             pair[LimitOrders.iterator, cppbool] insert_result
 * 
 */
  __pyx_v_limit_orders_collection_ptr = NULL;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":345
 * 
 *         quantized_price = (self.c_quantize_order_price(trading_pair_str, price)
 *                            if order_type is OrderType.LIMIT             # <<<<<<<<<<<<<<
 *                            else s_decimal_0)
 *         quantized_amount = self.c_quantize_order_amount(trading_pair_str, amount)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_LIMIT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = (__pyx_v_order_type == __pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_1) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":344
 *             pair[LimitOrders.iterator, cppbool] insert_result
 * 
 *         quantized_price = (self.c_quantize_order_price(trading_pair_str, price)             # <<<<<<<<<<<<<<
 *                            if order_type is OrderType.LIMIT
 *                            else s_decimal_0)
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), __pyx_v_trading_pair_str, __pyx_v_price); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 344, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __pyx_t_6;
    __pyx_t_6 = 0;
  } else {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":346
 *         quantized_price = (self.c_quantize_order_price(trading_pair_str, price)
 *                            if order_type is OrderType.LIMIT
 *                            else s_decimal_0)             # <<<<<<<<<<<<<<
 *         quantized_amount = self.c_quantize_order_amount(trading_pair_str, amount)
 *         if order_type is OrderType.MARKET:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_s_decimal_0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __pyx_t_6;
    __pyx_t_6 = 0;
  }
  __pyx_v_quantized_price = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":347
 *                            if order_type is OrderType.LIMIT
 *                            else s_decimal_0)
 *         quantized_amount = self.c_quantize_order_amount(trading_pair_str, amount)             # <<<<<<<<<<<<<<
 *         if order_type is OrderType.MARKET:
 *             self._queued_orders.append(QueuedOrder(self._current_timestamp, order_id, True, trading_pair_str,
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), __pyx_v_trading_pair_str, __pyx_v_amount, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_quantized_amount = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":348
 *                            else s_decimal_0)
 *         quantized_amount = self.c_quantize_order_amount(trading_pair_str, amount)
 *         if order_type is OrderType.MARKET:             # <<<<<<<<<<<<<<
 *             self._queued_orders.append(QueuedOrder(self._current_timestamp, order_id, True, trading_pair_str,
 *                                                    quantized_amount))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_MARKET); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 348, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = (__pyx_v_order_type == __pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_1) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":349
 *         quantized_amount = self.c_quantize_order_amount(trading_pair_str, amount)
 *         if order_type is OrderType.MARKET:
 *             self._queued_orders.append(QueuedOrder(self._current_timestamp, order_id, True, trading_pair_str,             # <<<<<<<<<<<<<<
 *                                                    quantized_amount))
 *         elif order_type is OrderType.LIMIT:
 */
    __pyx_t_6 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 349, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":350
 *         if order_type is OrderType.MARKET:
 *             self._queued_orders.append(QueuedOrder(self._current_timestamp, order_id, True, trading_pair_str,
 *                                                    quantized_amount))             # <<<<<<<<<<<<<<
 *         elif order_type is OrderType.LIMIT:
 * 
 */
    __pyx_t_5 = PyTuple_New(5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 349, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6)) __PYX_ERR(0, 349, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_order_id);
    __Pyx_GIVEREF(__pyx_v_order_id);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_order_id)) __PYX_ERR(0, 349, __pyx_L1_error);
    __Pyx_INCREF(Py_True);
    __Pyx_GIVEREF(Py_True);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, Py_True)) __PYX_ERR(0, 349, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_trading_pair_str);
    __Pyx_GIVEREF(__pyx_v_trading_pair_str);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_v_trading_pair_str)) __PYX_ERR(0, 349, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_quantized_amount);
    __Pyx_GIVEREF(__pyx_v_quantized_amount);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_v_quantized_amount)) __PYX_ERR(0, 349, __pyx_L1_error);
    __pyx_t_6 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":349
 *         quantized_amount = self.c_quantize_order_amount(trading_pair_str, amount)
 *         if order_type is OrderType.MARKET:
 *             self._queued_orders.append(QueuedOrder(self._current_timestamp, order_id, True, trading_pair_str,             # <<<<<<<<<<<<<<
 *                                                    quantized_amount))
 *         elif order_type is OrderType.LIMIT:
 */
    __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder), __pyx_t_5, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 349, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_9 = __Pyx_PyObject_Append(__pyx_v_self->_queued_orders, __pyx_t_6); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 349, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":348
 *                            else s_decimal_0)
 *         quantized_amount = self.c_quantize_order_amount(trading_pair_str, amount)
 *         if order_type is OrderType.MARKET:             # <<<<<<<<<<<<<<
 *             self._queued_orders.append(QueuedOrder(self._current_timestamp, order_id, True, trading_pair_str,
 *                                                    quantized_amount))
 */
    goto __pyx_L4;
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":351
 *             self._queued_orders.append(QueuedOrder(self._current_timestamp, order_id, True, trading_pair_str,
 *                                                    quantized_amount))
 *         elif order_type is OrderType.LIMIT:             # <<<<<<<<<<<<<<
 * 
 *             map_it = self._bid_limit_orders.find(cpp_trading_pair_str)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_LIMIT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_1 = (__pyx_v_order_type == __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_1) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":353
 *         elif order_type is OrderType.LIMIT:
 * 
 *             map_it = self._bid_limit_orders.find(cpp_trading_pair_str)             # <<<<<<<<<<<<<<
 * 
 *             if map_it == self._bid_limit_orders.end():
 */
    __pyx_v_map_it = __pyx_v_self->_bid_limit_orders.find(__pyx_v_cpp_trading_pair_str);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":355
 *             map_it = self._bid_limit_orders.find(cpp_trading_pair_str)
 * 
 *             if map_it == self._bid_limit_orders.end():             # <<<<<<<<<<<<<<
 *                 insert_result = self._bid_limit_orders.insert(LimitOrdersPair(cpp_trading_pair_str,
 *                                                                               SingleTradingPairLimitOrders()))
 */
    __pyx_t_1 = (__pyx_v_map_it == __pyx_v_self->_bid_limit_orders.end());
    if (__pyx_t_1) {

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":357
 *             if map_it == self._bid_limit_orders.end():
 *                 insert_result = self._bid_limit_orders.insert(LimitOrdersPair(cpp_trading_pair_str,
 *                                                                               SingleTradingPairLimitOrders()))             # <<<<<<<<<<<<<<
 *                 map_it = insert_result.first
 *             limit_orders_collection_ptr = address(deref(map_it).second)
 */
      try {
        __pyx_t_10 = __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders();
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 357, __pyx_L1_error)
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":356
 * 
 *             if map_it == self._bid_limit_orders.end():
 *                 insert_result = self._bid_limit_orders.insert(LimitOrdersPair(cpp_trading_pair_str,             # <<<<<<<<<<<<<<
 *                                                                               SingleTradingPairLimitOrders()))
 *                 map_it = insert_result.first
 */
      try {
        __pyx_t_11 = __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersPair(__pyx_v_cpp_trading_pair_str, __pyx_t_10);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 356, __pyx_L1_error)
      }
      try {
        __pyx_t_12 = __pyx_v_self->_bid_limit_orders.insert(__pyx_t_11);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 356, __pyx_L1_error)
      }
      __pyx_v_insert_result = __PYX_STD_MOVE_IF_SUPPORTED(((std::pair<std::unordered_map<std::string,__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders> ::iterator,bool> )__pyx_t_12));

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":358
 *                 insert_result = self._bid_limit_orders.insert(LimitOrdersPair(cpp_trading_pair_str,
 *                                                                               SingleTradingPairLimitOrders()))
 *                 map_it = insert_result.first             # <<<<<<<<<<<<<<
 *             limit_orders_collection_ptr = address(deref(map_it).second)
 *             limit_orders_collection_ptr.insert(CPPLimitOrder(
 */
      __pyx_t_13 = __pyx_v_insert_result.first;
      __pyx_v_map_it = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_13);

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":355
 *             map_it = self._bid_limit_orders.find(cpp_trading_pair_str)
 * 
 *             if map_it == self._bid_limit_orders.end():             # <<<<<<<<<<<<<<
 *                 insert_result = self._bid_limit_orders.insert(LimitOrdersPair(cpp_trading_pair_str,
 *                                                                               SingleTradingPairLimitOrders()))
 */
    }

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":359
 *                                                                               SingleTradingPairLimitOrders()))
 *                 map_it = insert_result.first
 *             limit_orders_collection_ptr = address(deref(map_it).second)             # <<<<<<<<<<<<<<
 *             limit_orders_collection_ptr.insert(CPPLimitOrder(
 *                 cpp_order_id,
 */
    __pyx_v_limit_orders_collection_ptr = (&(*__pyx_v_map_it).second);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":360
 *                 map_it = insert_result.first
 *             limit_orders_collection_ptr = address(deref(map_it).second)
 *             limit_orders_collection_ptr.insert(CPPLimitOrder(             # <<<<<<<<<<<<<<
 *                 cpp_order_id,
 *                 cpp_trading_pair_str,
 */
    try {
      __pyx_v_limit_orders_collection_ptr->insert(LimitOrder(__pyx_v_cpp_order_id, __pyx_v_cpp_trading_pair_str, 1, __pyx_v_cpp_base_asset, __pyx_v_cpp_quote_asset, ((PyObject *)__pyx_v_quantized_price), ((PyObject *)__pyx_v_quantized_amount), ((PyObject *)Py_None), ((PY_LONG_LONG)(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp * 1e6)), 0, __pyx_v_cpp_position));
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 360, __pyx_L1_error)
    }

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":351
 *             self._queued_orders.append(QueuedOrder(self._current_timestamp, order_id, True, trading_pair_str,
 *                                                    quantized_amount))
 *         elif order_type is OrderType.LIMIT:             # <<<<<<<<<<<<<<
 * 
 *             map_it = self._bid_limit_orders.find(cpp_trading_pair_str)
 */
  }
  __pyx_L4:;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":373
 *                 cpp_position,
 *             ))
 *         safe_ensure_future(self.trigger_event_async(             # <<<<<<<<<<<<<<
 *             self.MARKET_BUY_ORDER_CREATED_EVENT_TAG,
 *             BuyOrderCreatedEvent(self._current_timestamp,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_safe_ensure_future); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trigger_event_async); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 373, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":374
 *             ))
 *         safe_ensure_future(self.trigger_event_async(
 *             self.MARKET_BUY_ORDER_CREATED_EVENT_TAG,             # <<<<<<<<<<<<<<
 *             BuyOrderCreatedEvent(self._current_timestamp,
 *                                  order_type,
 */
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_MARKET_BUY_ORDER_CREATED_EVENT_T); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":375
 *         safe_ensure_future(self.trigger_event_async(
 *             self.MARKET_BUY_ORDER_CREATED_EVENT_TAG,
 *             BuyOrderCreatedEvent(self._current_timestamp,             # <<<<<<<<<<<<<<
 *                                  order_type,
 *                                  trading_pair_str,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_n_s_BuyOrderCreatedEvent); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_18 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 375, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":381
 *                                  quantized_price,
 *                                  order_id,
 *                                  self._current_timestamp)))             # <<<<<<<<<<<<<<
 *         return order_id
 * 
 */
  __pyx_t_19 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_20 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_17))) {
    __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_17);
    if (likely(__pyx_t_20)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
      __Pyx_INCREF(__pyx_t_20);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_17, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[8] = {__pyx_t_20, __pyx_t_18, __pyx_v_order_type, __pyx_v_trading_pair_str, __pyx_v_quantized_amount, __pyx_v_quantized_price, __pyx_v_order_id, __pyx_t_19};
    __pyx_t_16 = __Pyx_PyObject_FastCall(__pyx_t_17, __pyx_callargs+1-__pyx_t_7, 7+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 375, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  }
  __pyx_t_17 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_14))) {
    __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_14);
    if (likely(__pyx_t_17)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
      __Pyx_INCREF(__pyx_t_17);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_14, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_17, __pyx_t_15, __pyx_t_16};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_14, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  }
  __pyx_t_14 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_14)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_14);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_14, __pyx_t_2};
    __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 373, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":382
 *                                  order_id,
 *                                  self._current_timestamp)))
 *         return order_id             # <<<<<<<<<<<<<<
 * 
 *     cdef str c_sell(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_order_id);
  __pyx_r = __pyx_v_order_id;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":323
 *         self.c_process_crossed_limit_orders()
 * 
 *     cdef str c_buy(self,             # <<<<<<<<<<<<<<
 *                    str trading_pair_str,
 *                    object amount,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_buy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_order_id);
  __Pyx_XDECREF(__pyx_v_quote_asset);
  __Pyx_XDECREF(__pyx_v_quantized_price);
  __Pyx_XDECREF(__pyx_v_quantized_amount);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":384
 *         return order_id
 * 
 *     cdef str c_sell(self,             # <<<<<<<<<<<<<<
 *                     str trading_pair_str,
 *                     object amount,
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_sell(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair_str, PyObject *__pyx_v_amount, struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_sell *__pyx_optional_args) {
  PyObject *__pyx_v_order_type = __pyx_k__18;
  PyObject *__pyx_v_price = __pyx_k__19;
  PyObject *__pyx_v_order_id = 0;
  PyObject *__pyx_v_base_asset = 0;
  std::string __pyx_v_cpp_order_id;
  std::string __pyx_v_cpp_trading_pair_str;
  std::string __pyx_v_cpp_base_asset;
  std::string __pyx_v_cpp_quote_asset;
  std::string __pyx_v_cpp_position;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator __pyx_v_map_it;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders *__pyx_v_limit_orders_collection_ptr;
  std::pair<std::unordered_map<std::string,__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders> ::iterator,bool>  __pyx_v_insert_result;
  PyObject *__pyx_v_quantized_price = NULL;
  PyObject *__pyx_v_quantized_amount = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_UCS4 __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  std::string __pyx_t_8;
  int __pyx_t_9;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders __pyx_t_10;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersPair __pyx_t_11;
  std::pair<std::unordered_map<std::string,__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders> ::iterator,int>  __pyx_t_12;
  std::unordered_map<std::string,__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders> ::iterator __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_sell", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_order_type = __pyx_optional_args->order_type;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_price = __pyx_optional_args->price;
      }
    }
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":391
 *                     dict kwargs={}):
 * 
 *         if trading_pair_str not in self._trading_pairs:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Trading pair '{trading_pair_str}' does not existing in current data set.")
 *         cdef:
 */
  if (unlikely(__pyx_v_self->_trading_pairs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 391, __pyx_L1_error)
  }
  __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_v_trading_pair_str, __pyx_v_self->_trading_pairs, Py_NE)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 391, __pyx_L1_error)
  if (unlikely(__pyx_t_1)) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":392
 * 
 *         if trading_pair_str not in self._trading_pairs:
 *             raise ValueError(f"Trading pair '{trading_pair_str}' does not existing in current data set.")             # <<<<<<<<<<<<<<
 *         cdef:
 *             str order_id = self.random_order_id("sell", trading_pair_str)
 */
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = 127;
    __Pyx_INCREF(__pyx_kp_u_Trading_pair);
    __pyx_t_3 += 14;
    __Pyx_GIVEREF(__pyx_kp_u_Trading_pair);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_Trading_pair);
    __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_v_trading_pair_str); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_4;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u_does_not_existing_in_current_da);
    __pyx_t_3 += 40;
    __Pyx_GIVEREF(__pyx_kp_u_does_not_existing_in_current_da);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_does_not_existing_in_current_da);
    __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_2, 3, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 392, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 392, __pyx_L1_error)

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":391
 *                     dict kwargs={}):
 * 
 *         if trading_pair_str not in self._trading_pairs:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"Trading pair '{trading_pair_str}' does not existing in current data set.")
 *         cdef:
 */
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":394
 *             raise ValueError(f"Trading pair '{trading_pair_str}' does not existing in current data set.")
 *         cdef:
 *             str order_id = self.random_order_id("sell", trading_pair_str)             # <<<<<<<<<<<<<<
 *             str base_asset = self._trading_pairs[trading_pair_str].base_asset
 *             string cpp_order_id = order_id.encode("utf8")
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_random_order_id); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_n_u_sell, __pyx_v_trading_pair_str};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_2))) __PYX_ERR(0, 394, __pyx_L1_error)
  __pyx_v_order_id = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":395
 *         cdef:
 *             str order_id = self.random_order_id("sell", trading_pair_str)
 *             str base_asset = self._trading_pairs[trading_pair_str].base_asset             # <<<<<<<<<<<<<<
 *             string cpp_order_id = order_id.encode("utf8")
 *             string cpp_trading_pair_str = trading_pair_str.encode("utf8")
 */
  if (unlikely(__pyx_v_self->_trading_pairs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 395, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_self->_trading_pairs, __pyx_v_trading_pair_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_base_asset); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_5))) __PYX_ERR(0, 395, __pyx_L1_error)
  __pyx_v_base_asset = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":396
 *             str order_id = self.random_order_id("sell", trading_pair_str)
 *             str base_asset = self._trading_pairs[trading_pair_str].base_asset
 *             string cpp_order_id = order_id.encode("utf8")             # <<<<<<<<<<<<<<
 *             string cpp_trading_pair_str = trading_pair_str.encode("utf8")
 *             string cpp_base_asset = base_asset.encode("utf8")
 */
  if (unlikely(__pyx_v_order_id == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 396, __pyx_L1_error)
  }
  __pyx_t_5 = PyUnicode_AsUTF8String(__pyx_v_order_id); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __pyx_convert_string_from_py_6libcpp_6string_std__in_string(__pyx_t_5); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 396, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_cpp_order_id = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_8);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":397
 *             str base_asset = self._trading_pairs[trading_pair_str].base_asset
 *             string cpp_order_id = order_id.encode("utf8")
 *             string cpp_trading_pair_str = trading_pair_str.encode("utf8")             # <<<<<<<<<<<<<<
 *             string cpp_base_asset = base_asset.encode("utf8")
 *             string cpp_quote_asset = self._trading_pairs[trading_pair_str].quote_asset.encode("utf8")
 */
  if (unlikely(__pyx_v_trading_pair_str == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 397, __pyx_L1_error)
  }
  __pyx_t_5 = PyUnicode_AsUTF8String(__pyx_v_trading_pair_str); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __pyx_convert_string_from_py_6libcpp_6string_std__in_string(__pyx_t_5); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 397, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_cpp_trading_pair_str = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_8);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":398
 *             string cpp_order_id = order_id.encode("utf8")
 *             string cpp_trading_pair_str = trading_pair_str.encode("utf8")
 *             string cpp_base_asset = base_asset.encode("utf8")             # <<<<<<<<<<<<<<
 *             string cpp_quote_asset = self._trading_pairs[trading_pair_str].quote_asset.encode("utf8")
 *             string cpp_position = "NIL".encode("utf8")
 */
  if (unlikely(__pyx_v_base_asset == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 398, __pyx_L1_error)
  }
  __pyx_t_5 = PyUnicode_AsUTF8String(__pyx_v_base_asset); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = __pyx_convert_string_from_py_6libcpp_6string_std__in_string(__pyx_t_5); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 398, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_cpp_base_asset = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_8);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":399
 *             string cpp_trading_pair_str = trading_pair_str.encode("utf8")
 *             string cpp_base_asset = base_asset.encode("utf8")
 *             string cpp_quote_asset = self._trading_pairs[trading_pair_str].quote_asset.encode("utf8")             # <<<<<<<<<<<<<<
 *             string cpp_position = "NIL".encode("utf8")
 *             LimitOrdersIterator map_it
 */
  if (unlikely(__pyx_v_self->_trading_pairs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 399, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_self->_trading_pairs, __pyx_v_trading_pair_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_quote_asset); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_n_u_utf8};
    __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_8 = __pyx_convert_string_from_py_6libcpp_6string_std__in_string(__pyx_t_5); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_cpp_quote_asset = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_8);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":400
 *             string cpp_base_asset = base_asset.encode("utf8")
 *             string cpp_quote_asset = self._trading_pairs[trading_pair_str].quote_asset.encode("utf8")
 *             string cpp_position = "NIL".encode("utf8")             # <<<<<<<<<<<<<<
 *             LimitOrdersIterator map_it
 *             SingleTradingPairLimitOrders *limit_orders_collection_ptr = NULL
 */
  __pyx_t_8 = __pyx_convert_string_from_py_6libcpp_6string_std__in_string(__pyx_n_b_NIL); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 400, __pyx_L1_error)
  __pyx_v_cpp_position = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_8);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":402
 *             string cpp_position = "NIL".encode("utf8")
 *             LimitOrdersIterator map_it
 *             SingleTradingPairLimitOrders *limit_orders_collection_ptr = NULL             # <<<<<<<<<<<<<<
 *             pair[LimitOrders.iterator, cppbool] insert_result
 * 
 */
  __pyx_v_limit_orders_collection_ptr = NULL;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":406
 * 
 *         quantized_price = (self.c_quantize_order_price(trading_pair_str, price)
 *                            if order_type is OrderType.LIMIT             # <<<<<<<<<<<<<<
 *                            else s_decimal_0)
 *         quantized_amount = self.c_quantize_order_amount(trading_pair_str, amount)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_LIMIT); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = (__pyx_v_order_type == __pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_1) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":405
 *             pair[LimitOrders.iterator, cppbool] insert_result
 * 
 *         quantized_price = (self.c_quantize_order_price(trading_pair_str, price)             # <<<<<<<<<<<<<<
 *                            if order_type is OrderType.LIMIT
 *                            else s_decimal_0)
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), __pyx_v_trading_pair_str, __pyx_v_price); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __pyx_t_6;
    __pyx_t_6 = 0;
  } else {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":407
 *         quantized_price = (self.c_quantize_order_price(trading_pair_str, price)
 *                            if order_type is OrderType.LIMIT
 *                            else s_decimal_0)             # <<<<<<<<<<<<<<
 *         quantized_amount = self.c_quantize_order_amount(trading_pair_str, amount)
 *         if order_type is OrderType.MARKET:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_s_decimal_0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __pyx_t_6;
    __pyx_t_6 = 0;
  }
  __pyx_v_quantized_price = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":408
 *                            if order_type is OrderType.LIMIT
 *                            else s_decimal_0)
 *         quantized_amount = self.c_quantize_order_amount(trading_pair_str, amount)             # <<<<<<<<<<<<<<
 *         if order_type is OrderType.MARKET:
 *             self._queued_orders.append(QueuedOrder(self._current_timestamp, order_id, False, trading_pair_str,
 */
  __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), __pyx_v_trading_pair_str, __pyx_v_amount, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_quantized_amount = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":409
 *                            else s_decimal_0)
 *         quantized_amount = self.c_quantize_order_amount(trading_pair_str, amount)
 *         if order_type is OrderType.MARKET:             # <<<<<<<<<<<<<<
 *             self._queued_orders.append(QueuedOrder(self._current_timestamp, order_id, False, trading_pair_str,
 *                                                    quantized_amount))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_MARKET); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = (__pyx_v_order_type == __pyx_t_6);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_1) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":410
 *         quantized_amount = self.c_quantize_order_amount(trading_pair_str, amount)
 *         if order_type is OrderType.MARKET:
 *             self._queued_orders.append(QueuedOrder(self._current_timestamp, order_id, False, trading_pair_str,             # <<<<<<<<<<<<<<
 *                                                    quantized_amount))
 *         elif order_type is OrderType.LIMIT:
 */
    __pyx_t_6 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":411
 *         if order_type is OrderType.MARKET:
 *             self._queued_orders.append(QueuedOrder(self._current_timestamp, order_id, False, trading_pair_str,
 *                                                    quantized_amount))             # <<<<<<<<<<<<<<
 *         elif order_type is OrderType.LIMIT:
 *             map_it = self._ask_limit_orders.find(cpp_trading_pair_str)
 */
    __pyx_t_5 = PyTuple_New(5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6)) __PYX_ERR(0, 410, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_order_id);
    __Pyx_GIVEREF(__pyx_v_order_id);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_order_id)) __PYX_ERR(0, 410, __pyx_L1_error);
    __Pyx_INCREF(Py_False);
    __Pyx_GIVEREF(Py_False);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, Py_False)) __PYX_ERR(0, 410, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_trading_pair_str);
    __Pyx_GIVEREF(__pyx_v_trading_pair_str);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_v_trading_pair_str)) __PYX_ERR(0, 410, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_quantized_amount);
    __Pyx_GIVEREF(__pyx_v_quantized_amount);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 4, __pyx_v_quantized_amount)) __PYX_ERR(0, 410, __pyx_L1_error);
    __pyx_t_6 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":410
 *         quantized_amount = self.c_quantize_order_amount(trading_pair_str, amount)
 *         if order_type is OrderType.MARKET:
 *             self._queued_orders.append(QueuedOrder(self._current_timestamp, order_id, False, trading_pair_str,             # <<<<<<<<<<<<<<
 *                                                    quantized_amount))
 *         elif order_type is OrderType.LIMIT:
 */
    __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder), __pyx_t_5, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_9 = __Pyx_PyObject_Append(__pyx_v_self->_queued_orders, __pyx_t_6); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":409
 *                            else s_decimal_0)
 *         quantized_amount = self.c_quantize_order_amount(trading_pair_str, amount)
 *         if order_type is OrderType.MARKET:             # <<<<<<<<<<<<<<
 *             self._queued_orders.append(QueuedOrder(self._current_timestamp, order_id, False, trading_pair_str,
 *                                                    quantized_amount))
 */
    goto __pyx_L4;
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":412
 *             self._queued_orders.append(QueuedOrder(self._current_timestamp, order_id, False, trading_pair_str,
 *                                                    quantized_amount))
 *         elif order_type is OrderType.LIMIT:             # <<<<<<<<<<<<<<
 *             map_it = self._ask_limit_orders.find(cpp_trading_pair_str)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_LIMIT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_1 = (__pyx_v_order_type == __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_1) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":413
 *                                                    quantized_amount))
 *         elif order_type is OrderType.LIMIT:
 *             map_it = self._ask_limit_orders.find(cpp_trading_pair_str)             # <<<<<<<<<<<<<<
 * 
 *             if map_it == self._ask_limit_orders.end():
 */
    __pyx_v_map_it = __pyx_v_self->_ask_limit_orders.find(__pyx_v_cpp_trading_pair_str);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":415
 *             map_it = self._ask_limit_orders.find(cpp_trading_pair_str)
 * 
 *             if map_it == self._ask_limit_orders.end():             # <<<<<<<<<<<<<<
 *                 insert_result = self._ask_limit_orders.insert(LimitOrdersPair(cpp_trading_pair_str,
 *                                                                               SingleTradingPairLimitOrders()))
 */
    __pyx_t_1 = (__pyx_v_map_it == __pyx_v_self->_ask_limit_orders.end());
    if (__pyx_t_1) {

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":417
 *             if map_it == self._ask_limit_orders.end():
 *                 insert_result = self._ask_limit_orders.insert(LimitOrdersPair(cpp_trading_pair_str,
 *                                                                               SingleTradingPairLimitOrders()))             # <<<<<<<<<<<<<<
 *                 map_it = insert_result.first
 *             limit_orders_collection_ptr = address(deref(map_it).second)
 */
      try {
        __pyx_t_10 = __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders();
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 417, __pyx_L1_error)
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":416
 * 
 *             if map_it == self._ask_limit_orders.end():
 *                 insert_result = self._ask_limit_orders.insert(LimitOrdersPair(cpp_trading_pair_str,             # <<<<<<<<<<<<<<
 *                                                                               SingleTradingPairLimitOrders()))
 *                 map_it = insert_result.first
 */
      try {
        __pyx_t_11 = __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersPair(__pyx_v_cpp_trading_pair_str, __pyx_t_10);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 416, __pyx_L1_error)
      }
      try {
        __pyx_t_12 = __pyx_v_self->_ask_limit_orders.insert(__pyx_t_11);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 416, __pyx_L1_error)
      }
      __pyx_v_insert_result = __PYX_STD_MOVE_IF_SUPPORTED(((std::pair<std::unordered_map<std::string,__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders> ::iterator,bool> )__pyx_t_12));

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":418
 *                 insert_result = self._ask_limit_orders.insert(LimitOrdersPair(cpp_trading_pair_str,
 *                                                                               SingleTradingPairLimitOrders()))
 *                 map_it = insert_result.first             # <<<<<<<<<<<<<<
 *             limit_orders_collection_ptr = address(deref(map_it).second)
 *             limit_orders_collection_ptr.insert(CPPLimitOrder(
 */
      __pyx_t_13 = __pyx_v_insert_result.first;
      __pyx_v_map_it = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_13);

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":415
 *             map_it = self._ask_limit_orders.find(cpp_trading_pair_str)
 * 
 *             if map_it == self._ask_limit_orders.end():             # <<<<<<<<<<<<<<
 *                 insert_result = self._ask_limit_orders.insert(LimitOrdersPair(cpp_trading_pair_str,
 *                                                                               SingleTradingPairLimitOrders()))
 */
    }

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":419
 *                                                                               SingleTradingPairLimitOrders()))
 *                 map_it = insert_result.first
 *             limit_orders_collection_ptr = address(deref(map_it).second)             # <<<<<<<<<<<<<<
 *             limit_orders_collection_ptr.insert(CPPLimitOrder(
 *                 cpp_order_id,
 */
    __pyx_v_limit_orders_collection_ptr = (&(*__pyx_v_map_it).second);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":420
 *                 map_it = insert_result.first
 *             limit_orders_collection_ptr = address(deref(map_it).second)
 *             limit_orders_collection_ptr.insert(CPPLimitOrder(             # <<<<<<<<<<<<<<
 *                 cpp_order_id,
 *                 cpp_trading_pair_str,
 */
    try {
      __pyx_v_limit_orders_collection_ptr->insert(LimitOrder(__pyx_v_cpp_order_id, __pyx_v_cpp_trading_pair_str, 0, __pyx_v_cpp_base_asset, __pyx_v_cpp_quote_asset, ((PyObject *)__pyx_v_quantized_price), ((PyObject *)__pyx_v_quantized_amount), ((PyObject *)Py_None), ((PY_LONG_LONG)(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp * 1e6)), 0, __pyx_v_cpp_position));
    } catch(...) {
      __Pyx_CppExn2PyErr();
      __PYX_ERR(0, 420, __pyx_L1_error)
    }

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":412
 *             self._queued_orders.append(QueuedOrder(self._current_timestamp, order_id, False, trading_pair_str,
 *                                                    quantized_amount))
 *         elif order_type is OrderType.LIMIT:             # <<<<<<<<<<<<<<
 *             map_it = self._ask_limit_orders.find(cpp_trading_pair_str)
 * 
 */
  }
  __pyx_L4:;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":433
 *                 cpp_position,
 *             ))
 *         safe_ensure_future(self.trigger_event_async(             # <<<<<<<<<<<<<<
 *             self.MARKET_SELL_ORDER_CREATED_EVENT_TAG,
 *             SellOrderCreatedEvent(self._current_timestamp,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_safe_ensure_future); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trigger_event_async); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":434
 *             ))
 *         safe_ensure_future(self.trigger_event_async(
 *             self.MARKET_SELL_ORDER_CREATED_EVENT_TAG,             # <<<<<<<<<<<<<<
 *             SellOrderCreatedEvent(self._current_timestamp,
 *                                   order_type,
 */
  __pyx_t_15 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_MARKET_SELL_ORDER_CREATED_EVENT); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":435
 *         safe_ensure_future(self.trigger_event_async(
 *             self.MARKET_SELL_ORDER_CREATED_EVENT_TAG,
 *             SellOrderCreatedEvent(self._current_timestamp,             # <<<<<<<<<<<<<<
 *                                   order_type,
 *                                   trading_pair_str,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_17, __pyx_n_s_SellOrderCreatedEvent); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_t_18 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_18);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":441
 *                                   quantized_price,
 *                                   order_id,
 *                                   self._current_timestamp)))             # <<<<<<<<<<<<<<
 *         return order_id
 * 
 */
  __pyx_t_19 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_19);
  __pyx_t_20 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_17))) {
    __pyx_t_20 = PyMethod_GET_SELF(__pyx_t_17);
    if (likely(__pyx_t_20)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
      __Pyx_INCREF(__pyx_t_20);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_17, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[8] = {__pyx_t_20, __pyx_t_18, __pyx_v_order_type, __pyx_v_trading_pair_str, __pyx_v_quantized_amount, __pyx_v_quantized_price, __pyx_v_order_id, __pyx_t_19};
    __pyx_t_16 = __Pyx_PyObject_FastCall(__pyx_t_17, __pyx_callargs+1-__pyx_t_7, 7+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
    __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
    __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
    if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 435, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
  }
  __pyx_t_17 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_14))) {
    __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_14);
    if (likely(__pyx_t_17)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
      __Pyx_INCREF(__pyx_t_17);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_14, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_17, __pyx_t_15, __pyx_t_16};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_14, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  }
  __pyx_t_14 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_6))) {
    __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_6);
    if (likely(__pyx_t_14)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_14);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_6, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_14, __pyx_t_2};
    __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":442
 *                                   order_id,
 *                                   self._current_timestamp)))
 *         return order_id             # <<<<<<<<<<<<<<
 * 
 *     cdef c_execute_buy(self, str order_id, str trading_pair_str, object amount):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_order_id);
  __pyx_r = __pyx_v_order_id;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":384
 *         return order_id
 * 
 *     cdef str c_sell(self,             # <<<<<<<<<<<<<<
 *                     str trading_pair_str,
 *                     object amount,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_sell", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_order_id);
  __Pyx_XDECREF(__pyx_v_base_asset);
  __Pyx_XDECREF(__pyx_v_quantized_price);
  __Pyx_XDECREF(__pyx_v_quantized_amount);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":444
 *         return order_id
 * 
 *     cdef c_execute_buy(self, str order_id, str trading_pair_str, object amount):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str quote_asset = self._trading_pairs[trading_pair_str].quote_asset
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_execute_buy(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_order_id, PyObject *__pyx_v_trading_pair_str, PyObject *__pyx_v_amount) {
  PyObject *__pyx_v_quote_asset = 0;
  PyObject *__pyx_v_base_asset = 0;
  PyObject *__pyx_v_quote_balance = 0;
  PyObject *__pyx_v_base_balance = 0;
  PyObject *__pyx_v_order_book = NULL;
  PyObject *__pyx_v_buy_entries = NULL;
  PyObject *__pyx_v_avg_price = NULL;
  PyObject *__pyx_v_entry = NULL;
  PyObject *__pyx_v_order_candidate = NULL;
  PyObject *__pyx_v_adjusted_order_candidate = NULL;
  PyObject *__pyx_v_paid_amount = NULL;
  PyObject *__pyx_v_acquired_amount = NULL;
  PyObject *__pyx_v_fees = NULL;
  PyObject *__pyx_v_order_filled_events = NULL;
  PyObject *__pyx_v_order_filled_event = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  Py_UCS4 __pyx_t_11;
  int64_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_execute_buy", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":446
 *     cdef c_execute_buy(self, str order_id, str trading_pair_str, object amount):
 *         cdef:
 *             str quote_asset = self._trading_pairs[trading_pair_str].quote_asset             # <<<<<<<<<<<<<<
 *             str base_asset = self._trading_pairs[trading_pair_str].base_asset
 *             object quote_balance = self.c_get_balance(quote_asset)
 */
  if (unlikely(__pyx_v_self->_trading_pairs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 446, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_self->_trading_pairs, __pyx_v_trading_pair_str); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_quote_asset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_2))) __PYX_ERR(0, 446, __pyx_L1_error)
  __pyx_v_quote_asset = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":447
 *         cdef:
 *             str quote_asset = self._trading_pairs[trading_pair_str].quote_asset
 *             str base_asset = self._trading_pairs[trading_pair_str].base_asset             # <<<<<<<<<<<<<<
 *             object quote_balance = self.c_get_balance(quote_asset)
 *             object base_balance = self.c_get_balance(base_asset)
 */
  if (unlikely(__pyx_v_self->_trading_pairs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 447, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_self->_trading_pairs, __pyx_v_trading_pair_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 447, __pyx_L1_error)
  __pyx_v_base_asset = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":448
 *             str quote_asset = self._trading_pairs[trading_pair_str].quote_asset
 *             str base_asset = self._trading_pairs[trading_pair_str].base_asset
 *             object quote_balance = self.c_get_balance(quote_asset)             # <<<<<<<<<<<<<<
 *             object base_balance = self.c_get_balance(base_asset)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_get_balance(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), __pyx_v_quote_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_quote_balance = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":449
 *             str base_asset = self._trading_pairs[trading_pair_str].base_asset
 *             object quote_balance = self.c_get_balance(quote_asset)
 *             object base_balance = self.c_get_balance(base_asset)             # <<<<<<<<<<<<<<
 * 
 *         order_book = self.order_books[trading_pair_str]
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_get_balance(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), __pyx_v_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_base_balance = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":451
 *             object base_balance = self.c_get_balance(base_asset)
 * 
 *         order_book = self.order_books[trading_pair_str]             # <<<<<<<<<<<<<<
 * 
 *         buy_entries = order_book.simulate_buy(float(amount))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_books); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_v_trading_pair_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_order_book = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":453
 *         order_book = self.order_books[trading_pair_str]
 * 
 *         buy_entries = order_book.simulate_buy(float(amount))             # <<<<<<<<<<<<<<
 * 
 *         # Get the weighted average price of the trade
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_book, __pyx_n_s_simulate_buy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyNumber_Float(__pyx_v_amount); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_v_buy_entries = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":456
 * 
 *         # Get the weighted average price of the trade
 *         avg_price = Decimal(0)             # <<<<<<<<<<<<<<
 *         for entry in buy_entries:
 *             avg_price += Decimal(entry.price) * Decimal(entry.amount)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_int_0};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 456, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_v_avg_price = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":457
 *         # Get the weighted average price of the trade
 *         avg_price = Decimal(0)
 *         for entry in buy_entries:             # <<<<<<<<<<<<<<
 *             avg_price += Decimal(entry.price) * Decimal(entry.amount)
 *         avg_price = avg_price / amount
 */
  if (likely(PyList_CheckExact(__pyx_v_buy_entries)) || PyTuple_CheckExact(__pyx_v_buy_entries)) {
    __pyx_t_2 = __pyx_v_buy_entries; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_buy_entries); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 457, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 457, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 457, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 457, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 457, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 457, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 457, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 457, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_7(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 457, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_entry, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":458
 *         avg_price = Decimal(0)
 *         for entry in buy_entries:
 *             avg_price += Decimal(entry.price) * Decimal(entry.amount)             # <<<<<<<<<<<<<<
 *         avg_price = avg_price / amount
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_entry, __pyx_n_s_price); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 458, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_entry, __pyx_n_s_amount); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_8};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 458, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_4 = PyNumber_Multiply(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_avg_price, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 458, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_avg_price, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":457
 *         # Get the weighted average price of the trade
 *         avg_price = Decimal(0)
 *         for entry in buy_entries:             # <<<<<<<<<<<<<<
 *             avg_price += Decimal(entry.price) * Decimal(entry.amount)
 *         avg_price = avg_price / amount
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":459
 *         for entry in buy_entries:
 *             avg_price += Decimal(entry.price) * Decimal(entry.amount)
 *         avg_price = avg_price / amount             # <<<<<<<<<<<<<<
 * 
 *         order_candidate = OrderCandidate(
 */
  __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_v_avg_price, __pyx_v_amount); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_avg_price, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":461
 *         avg_price = avg_price / amount
 * 
 *         order_candidate = OrderCandidate(             # <<<<<<<<<<<<<<
 *             trading_pair=trading_pair_str,
 *             # Market orders are not maker orders
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_OrderCandidate); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":462
 * 
 *         order_candidate = OrderCandidate(
 *             trading_pair=trading_pair_str,             # <<<<<<<<<<<<<<
 *             # Market orders are not maker orders
 *             is_maker=False,
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_trading_pair, __pyx_v_trading_pair_str) < 0) __PYX_ERR(0, 462, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":464
 *             trading_pair=trading_pair_str,
 *             # Market orders are not maker orders
 *             is_maker=False,             # <<<<<<<<<<<<<<
 *             order_type=OrderType.MARKET,
 *             order_side=TradeType.BUY,
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_is_maker, Py_False) < 0) __PYX_ERR(0, 462, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":465
 *             # Market orders are not maker orders
 *             is_maker=False,
 *             order_type=OrderType.MARKET,             # <<<<<<<<<<<<<<
 *             order_side=TradeType.BUY,
 *             amount=amount,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_MARKET); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 465, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_order_type, __pyx_t_1) < 0) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":466
 *             is_maker=False,
 *             order_type=OrderType.MARKET,
 *             order_side=TradeType.BUY,             # <<<<<<<<<<<<<<
 *             amount=amount,
 *             price=avg_price,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_BUY); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_order_side, __pyx_t_4) < 0) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":467
 *             order_type=OrderType.MARKET,
 *             order_side=TradeType.BUY,
 *             amount=amount,             # <<<<<<<<<<<<<<
 *             price=avg_price,
 *             from_total_balances=True
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_amount, __pyx_v_amount) < 0) __PYX_ERR(0, 462, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":468
 *             order_side=TradeType.BUY,
 *             amount=amount,
 *             price=avg_price,             # <<<<<<<<<<<<<<
 *             from_total_balances=True
 *         )
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_price, __pyx_v_avg_price) < 0) __PYX_ERR(0, 462, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":469
 *             amount=amount,
 *             price=avg_price,
 *             from_total_balances=True             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_from_total_balances, Py_True) < 0) __PYX_ERR(0, 462, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":461
 *         avg_price = avg_price / amount
 * 
 *         order_candidate = OrderCandidate(             # <<<<<<<<<<<<<<
 *             trading_pair=trading_pair_str,
 *             # Market orders are not maker orders
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 461, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_order_candidate = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":472
 *         )
 * 
 *         adjusted_order_candidate = self._budget_checker.populate_collateral_entries(order_candidate)             # <<<<<<<<<<<<<<
 * 
 *         # Quote currency used, including fees.
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base._budget_checker, __pyx_n_s_populate_collateral_entries); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_order_candidate};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_adjusted_order_candidate = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":475
 * 
 *         # Quote currency used, including fees.
 *         paid_amount = adjusted_order_candidate.order_collateral.amount             # <<<<<<<<<<<<<<
 *         # Base currency acquired, including fees.
 *         acquired_amount = adjusted_order_candidate.potential_returns.amount
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_adjusted_order_candidate, __pyx_n_s_order_collateral); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_amount); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_paid_amount = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":477
 *         paid_amount = adjusted_order_candidate.order_collateral.amount
 *         # Base currency acquired, including fees.
 *         acquired_amount = adjusted_order_candidate.potential_returns.amount             # <<<<<<<<<<<<<<
 * 
 *         # It's not possible to fulfill the order, the possible acquired amount is less than requested
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_adjusted_order_candidate, __pyx_n_s_potential_returns); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_amount); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 477, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_acquired_amount = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":480
 * 
 *         # It's not possible to fulfill the order, the possible acquired amount is less than requested
 *         if paid_amount > quote_balance:             # <<<<<<<<<<<<<<
 *             self.logger().warning(f"Insufficient {quote_asset} balance available for buy order. "
 *                                   f"{quote_balance} {quote_asset} available vs. "
 */
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_paid_amount, __pyx_v_quote_balance, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 480, __pyx_L1_error)
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_10) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":481
 *         # It's not possible to fulfill the order, the possible acquired amount is less than requested
 *         if paid_amount > quote_balance:
 *             self.logger().warning(f"Insufficient {quote_asset} balance available for buy order. "             # <<<<<<<<<<<<<<
 *                                   f"{quote_balance} {quote_asset} available vs. "
 *                                   f"{paid_amount} {quote_asset} required for the order.")
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 481, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warning); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = 0;
    __pyx_t_11 = 127;
    __Pyx_INCREF(__pyx_kp_u_Insufficient);
    __pyx_t_6 += 13;
    __Pyx_GIVEREF(__pyx_kp_u_Insufficient);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Insufficient);
    __pyx_t_1 = __Pyx_PyUnicode_Unicode(__pyx_v_quote_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_11;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u_balance_available_for_buy_order);
    __pyx_t_6 += 34;
    __Pyx_GIVEREF(__pyx_kp_u_balance_available_for_buy_order);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u_balance_available_for_buy_order);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":482
 *         if paid_amount > quote_balance:
 *             self.logger().warning(f"Insufficient {quote_asset} balance available for buy order. "
 *                                   f"{quote_balance} {quote_asset} available vs. "             # <<<<<<<<<<<<<<
 *                                   f"{paid_amount} {quote_asset} required for the order.")
 *             self.c_trigger_event(
 */
    __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_v_quote_balance, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 482, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_11;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u__20);
    __pyx_t_6 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__20);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u__20);
    __pyx_t_1 = __Pyx_PyUnicode_Unicode(__pyx_v_quote_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 482, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_11;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u_available_vs);
    __pyx_t_6 += 15;
    __Pyx_GIVEREF(__pyx_kp_u_available_vs);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_available_vs);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":483
 *             self.logger().warning(f"Insufficient {quote_asset} balance available for buy order. "
 *                                   f"{quote_balance} {quote_asset} available vs. "
 *                                   f"{paid_amount} {quote_asset} required for the order.")             # <<<<<<<<<<<<<<
 *             self.c_trigger_event(
 *                 self.MARKET_ORDER_FAILURE_EVENT_TAG,
 */
    __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_v_paid_amount, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_11;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u__20);
    __pyx_t_6 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__20);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u__20);
    __pyx_t_1 = __Pyx_PyUnicode_Unicode(__pyx_v_quote_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_11;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u_required_for_the_order);
    __pyx_t_6 += 24;
    __Pyx_GIVEREF(__pyx_kp_u_required_for_the_order);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_required_for_the_order);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":481
 *         # It's not possible to fulfill the order, the possible acquired amount is less than requested
 *         if paid_amount > quote_balance:
 *             self.logger().warning(f"Insufficient {quote_asset} balance available for buy order. "             # <<<<<<<<<<<<<<
 *                                   f"{quote_balance} {quote_asset} available vs. "
 *                                   f"{paid_amount} {quote_asset} required for the order.")
 */
    __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_6, __pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_1};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 481, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":485
 *                                   f"{paid_amount} {quote_asset} required for the order.")
 *             self.c_trigger_event(
 *                 self.MARKET_ORDER_FAILURE_EVENT_TAG,             # <<<<<<<<<<<<<<
 *                 MarketOrderFailureEvent(self._current_timestamp, order_id, OrderType.MARKET)
 *             )
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_MARKET_ORDER_FAILURE_EVENT_TAG); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 485, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_12 = __Pyx_PyInt_As_int64_t(__pyx_t_4); if (unlikely((__pyx_t_12 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 485, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":486
 *             self.c_trigger_event(
 *                 self.MARKET_ORDER_FAILURE_EVENT_TAG,
 *                 MarketOrderFailureEvent(self._current_timestamp, order_id, OrderType.MARKET)             # <<<<<<<<<<<<<<
 *             )
 *             return
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MarketOrderFailureEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MARKET); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 486, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_t_1, __pyx_v_order_id, __pyx_t_8};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 486, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":484
 *                                   f"{quote_balance} {quote_asset} available vs. "
 *                                   f"{paid_amount} {quote_asset} required for the order.")
 *             self.c_trigger_event(             # <<<<<<<<<<<<<<
 *                 self.MARKET_ORDER_FAILURE_EVENT_TAG,
 *                 MarketOrderFailureEvent(self._current_timestamp, order_id, OrderType.MARKET)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.c_trigger_event(((struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *)__pyx_v_self), __pyx_t_12, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 484, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":488
 *                 MarketOrderFailureEvent(self._current_timestamp, order_id, OrderType.MARKET)
 *             )
 *             return             # <<<<<<<<<<<<<<
 * 
 *         # The order was successfully executed
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":480
 * 
 *         # It's not possible to fulfill the order, the possible acquired amount is less than requested
 *         if paid_amount > quote_balance:             # <<<<<<<<<<<<<<
 *             self.logger().warning(f"Insufficient {quote_asset} balance available for buy order. "
 *                                   f"{quote_balance} {quote_asset} available vs. "
 */
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":492
 *         # The order was successfully executed
 *         self.c_set_balance(quote_asset,
 *                            quote_balance - paid_amount)             # <<<<<<<<<<<<<<
 *         self.c_set_balance(base_asset,
 *                            base_balance + acquired_amount)
 */
  __pyx_t_2 = PyNumber_Subtract(__pyx_v_quote_balance, __pyx_v_paid_amount); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":491
 * 
 *         # The order was successfully executed
 *         self.c_set_balance(quote_asset,             # <<<<<<<<<<<<<<
 *                            quote_balance - paid_amount)
 *         self.c_set_balance(base_asset,
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_set_balance(__pyx_v_self, __pyx_v_quote_asset, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":494
 *                            quote_balance - paid_amount)
 *         self.c_set_balance(base_asset,
 *                            base_balance + acquired_amount)             # <<<<<<<<<<<<<<
 * 
 *         # add fee
 */
  __pyx_t_4 = PyNumber_Add(__pyx_v_base_balance, __pyx_v_acquired_amount); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":493
 *         self.c_set_balance(quote_asset,
 *                            quote_balance - paid_amount)
 *         self.c_set_balance(base_asset,             # <<<<<<<<<<<<<<
 *                            base_balance + acquired_amount)
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_set_balance(__pyx_v_self, __pyx_v_base_asset, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":497
 * 
 *         # add fee
 *         fees = build_trade_fee(             # <<<<<<<<<<<<<<
 *             exchange=self.name,
 *             is_maker=False,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_build_trade_fee); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":498
 *         # add fee
 *         fees = build_trade_fee(
 *             exchange=self.name,             # <<<<<<<<<<<<<<
 *             is_maker=False,
 *             base_currency="",
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_name); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_exchange, __pyx_t_8) < 0) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":499
 *         fees = build_trade_fee(
 *             exchange=self.name,
 *             is_maker=False,             # <<<<<<<<<<<<<<
 *             base_currency="",
 *             quote_currency="",
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_is_maker, Py_False) < 0) __PYX_ERR(0, 498, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_base_currency, __pyx_kp_u__11) < 0) __PYX_ERR(0, 498, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_quote_currency, __pyx_kp_u__11) < 0) __PYX_ERR(0, 498, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":502
 *             base_currency="",
 *             quote_currency="",
 *             order_type=OrderType.LIMIT,             # <<<<<<<<<<<<<<
 *             order_side=TradeType.BUY,
 *             amount=Decimal("0"),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_LIMIT); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_order_type, __pyx_t_1) < 0) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":503
 *             quote_currency="",
 *             order_type=OrderType.LIMIT,
 *             order_side=TradeType.BUY,             # <<<<<<<<<<<<<<
 *             amount=Decimal("0"),
 *             price=Decimal("0"),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_BUY); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 503, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_order_side, __pyx_t_8) < 0) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":504
 *             order_type=OrderType.LIMIT,
 *             order_side=TradeType.BUY,
 *             amount=Decimal("0"),             # <<<<<<<<<<<<<<
 *             price=Decimal("0"),
 *         )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_u_0};
    __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 504, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_amount, __pyx_t_8) < 0) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":505
 *             order_side=TradeType.BUY,
 *             amount=Decimal("0"),
 *             price=Decimal("0"),             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 505, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_u_0};
    __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 505, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_price, __pyx_t_8) < 0) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":497
 * 
 *         # add fee
 *         fees = build_trade_fee(             # <<<<<<<<<<<<<<
 *             exchange=self.name,
 *             is_maker=False,
 */
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_fees = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":508
 *         )
 * 
 *         order_filled_events = OrderFilledEvent.order_filled_events_from_order_book_rows(             # <<<<<<<<<<<<<<
 *             self._current_timestamp, order_id, trading_pair_str, TradeType.BUY, OrderType.MARKET,
 *             fees, buy_entries
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_OrderFilledEvent); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_order_filled_events_from_order_b); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 508, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":509
 * 
 *         order_filled_events = OrderFilledEvent.order_filled_events_from_order_book_rows(
 *             self._current_timestamp, order_id, trading_pair_str, TradeType.BUY, OrderType.MARKET,             # <<<<<<<<<<<<<<
 *             fees, buy_entries
 *         )
 */
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_BUY); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_MARKET); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 509, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":510
 *         order_filled_events = OrderFilledEvent.order_filled_events_from_order_book_rows(
 *             self._current_timestamp, order_id, trading_pair_str, TradeType.BUY, OrderType.MARKET,
 *             fees, buy_entries             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __pyx_t_1 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[8] = {__pyx_t_1, __pyx_t_4, __pyx_v_order_id, __pyx_v_trading_pair_str, __pyx_t_3, __pyx_t_9, __pyx_v_fees, __pyx_v_buy_entries};
    __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 7+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 508, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_order_filled_events = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":513
 *         )
 * 
 *         for order_filled_event in order_filled_events:             # <<<<<<<<<<<<<<
 *             self.c_trigger_event(self.ORDER_FILLED_EVENT_TAG, order_filled_event)
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_order_filled_events)) || PyTuple_CheckExact(__pyx_v_order_filled_events)) {
    __pyx_t_8 = __pyx_v_order_filled_events; __Pyx_INCREF(__pyx_t_8);
    __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_v_order_filled_events); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 513, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 513, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_8))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_8);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 513, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 513, __pyx_L1_error)
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 513, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_8);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 513, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 513, __pyx_L1_error)
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 513, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_8);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 513, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_order_filled_event, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":514
 * 
 *         for order_filled_event in order_filled_events:
 *             self.c_trigger_event(self.ORDER_FILLED_EVENT_TAG, order_filled_event)             # <<<<<<<<<<<<<<
 * 
 *         self.c_trigger_event(
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_ORDER_FILLED_EVENT_TAG); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_12 = __Pyx_PyInt_As_int64_t(__pyx_t_2); if (unlikely((__pyx_t_12 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 514, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.c_trigger_event(((struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *)__pyx_v_self), __pyx_t_12, __pyx_v_order_filled_event); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":513
 *         )
 * 
 *         for order_filled_event in order_filled_events:             # <<<<<<<<<<<<<<
 *             self.c_trigger_event(self.ORDER_FILLED_EVENT_TAG, order_filled_event)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":517
 * 
 *         self.c_trigger_event(
 *             self.BUY_ORDER_COMPLETED_EVENT_TAG,             # <<<<<<<<<<<<<<
 *             BuyOrderCompletedEvent(self._current_timestamp,
 *                                    order_id,
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_BUY_ORDER_COMPLETED_EVENT_TAG); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_12 = __Pyx_PyInt_As_int64_t(__pyx_t_8); if (unlikely((__pyx_t_12 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 517, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":518
 *         self.c_trigger_event(
 *             self.BUY_ORDER_COMPLETED_EVENT_TAG,
 *             BuyOrderCompletedEvent(self._current_timestamp,             # <<<<<<<<<<<<<<
 *                                    order_id,
 *                                    base_asset,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_BuyOrderCompletedEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":524
 *                                    acquired_amount,
 *                                    paid_amount,
 *                                    OrderType.MARKET))             # <<<<<<<<<<<<<<
 * 
 *     cdef c_execute_sell(self, str order_id, str trading_pair_str, object amount):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MARKET); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[8] = {__pyx_t_3, __pyx_t_9, __pyx_v_order_id, __pyx_v_base_asset, __pyx_v_quote_asset, __pyx_v_acquired_amount, __pyx_v_paid_amount, __pyx_t_4};
    __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 7+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 518, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":516
 *             self.c_trigger_event(self.ORDER_FILLED_EVENT_TAG, order_filled_event)
 * 
 *         self.c_trigger_event(             # <<<<<<<<<<<<<<
 *             self.BUY_ORDER_COMPLETED_EVENT_TAG,
 *             BuyOrderCompletedEvent(self._current_timestamp,
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.c_trigger_event(((struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *)__pyx_v_self), __pyx_t_12, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":444
 *         return order_id
 * 
 *     cdef c_execute_buy(self, str order_id, str trading_pair_str, object amount):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str quote_asset = self._trading_pairs[trading_pair_str].quote_asset
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_execute_buy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_quote_asset);
  __Pyx_XDECREF(__pyx_v_base_asset);
  __Pyx_XDECREF(__pyx_v_quote_balance);
  __Pyx_XDECREF(__pyx_v_base_balance);
  __Pyx_XDECREF(__pyx_v_order_book);
  __Pyx_XDECREF(__pyx_v_buy_entries);
  __Pyx_XDECREF(__pyx_v_avg_price);
  __Pyx_XDECREF(__pyx_v_entry);
  __Pyx_XDECREF(__pyx_v_order_candidate);
  __Pyx_XDECREF(__pyx_v_adjusted_order_candidate);
  __Pyx_XDECREF(__pyx_v_paid_amount);
  __Pyx_XDECREF(__pyx_v_acquired_amount);
  __Pyx_XDECREF(__pyx_v_fees);
  __Pyx_XDECREF(__pyx_v_order_filled_events);
  __Pyx_XDECREF(__pyx_v_order_filled_event);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":526
 *                                    OrderType.MARKET))
 * 
 *     cdef c_execute_sell(self, str order_id, str trading_pair_str, object amount):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str quote_asset = self._trading_pairs[trading_pair_str].quote_asset
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_execute_sell(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_order_id, PyObject *__pyx_v_trading_pair_str, PyObject *__pyx_v_amount) {
  PyObject *__pyx_v_quote_asset = 0;
  PyObject *__pyx_v_base_asset = 0;
  PyObject *__pyx_v_quote_balance = 0;
  PyObject *__pyx_v_base_balance = 0;
  PyObject *__pyx_v_order_book = NULL;
  PyObject *__pyx_v_sell_entries = NULL;
  PyObject *__pyx_v_avg_price = NULL;
  PyObject *__pyx_v_entry = NULL;
  PyObject *__pyx_v_order_candidate = NULL;
  PyObject *__pyx_v_adjusted_order_candidate = NULL;
  PyObject *__pyx_v_sold_amount = NULL;
  PyObject *__pyx_v_acquired_amount = NULL;
  PyObject *__pyx_v_fees = NULL;
  PyObject *__pyx_v_order_filled_events = NULL;
  PyObject *__pyx_v_order_filled_event = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  Py_UCS4 __pyx_t_11;
  int64_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_execute_sell", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":528
 *     cdef c_execute_sell(self, str order_id, str trading_pair_str, object amount):
 *         cdef:
 *             str quote_asset = self._trading_pairs[trading_pair_str].quote_asset             # <<<<<<<<<<<<<<
 *             str base_asset = self._trading_pairs[trading_pair_str].base_asset
 *             object quote_balance = self.c_get_balance(quote_asset)
 */
  if (unlikely(__pyx_v_self->_trading_pairs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 528, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyDict_GetItem(__pyx_v_self->_trading_pairs, __pyx_v_trading_pair_str); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_quote_asset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_2))) __PYX_ERR(0, 528, __pyx_L1_error)
  __pyx_v_quote_asset = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":529
 *         cdef:
 *             str quote_asset = self._trading_pairs[trading_pair_str].quote_asset
 *             str base_asset = self._trading_pairs[trading_pair_str].base_asset             # <<<<<<<<<<<<<<
 *             object quote_balance = self.c_get_balance(quote_asset)
 *             object base_balance = self.c_get_balance(base_asset)
 */
  if (unlikely(__pyx_v_self->_trading_pairs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(0, 529, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_self->_trading_pairs, __pyx_v_trading_pair_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 529, __pyx_L1_error)
  __pyx_v_base_asset = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":530
 *             str quote_asset = self._trading_pairs[trading_pair_str].quote_asset
 *             str base_asset = self._trading_pairs[trading_pair_str].base_asset
 *             object quote_balance = self.c_get_balance(quote_asset)             # <<<<<<<<<<<<<<
 *             object base_balance = self.c_get_balance(base_asset)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_get_balance(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), __pyx_v_quote_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_quote_balance = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":531
 *             str base_asset = self._trading_pairs[trading_pair_str].base_asset
 *             object quote_balance = self.c_get_balance(quote_asset)
 *             object base_balance = self.c_get_balance(base_asset)             # <<<<<<<<<<<<<<
 * 
 *         order_book = self.order_books[trading_pair_str]
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_get_balance(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), __pyx_v_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_base_balance = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":533
 *             object base_balance = self.c_get_balance(base_asset)
 * 
 *         order_book = self.order_books[trading_pair_str]             # <<<<<<<<<<<<<<
 * 
 *         sell_entries = order_book.simulate_sell(float(amount))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_books); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Dict_GetItem(__pyx_t_1, __pyx_v_trading_pair_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_order_book = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":535
 *         order_book = self.order_books[trading_pair_str]
 * 
 *         sell_entries = order_book.simulate_sell(float(amount))             # <<<<<<<<<<<<<<
 * 
 *         # Get the weighted average price of the trade
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_book, __pyx_n_s_simulate_sell); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyNumber_Float(__pyx_v_amount); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 535, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_v_sell_entries = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":538
 * 
 *         # Get the weighted average price of the trade
 *         avg_price = Decimal(0)             # <<<<<<<<<<<<<<
 *         for entry in sell_entries:
 *             avg_price += Decimal(entry.price) * Decimal(entry.amount)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_int_0};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 538, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_v_avg_price = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":539
 *         # Get the weighted average price of the trade
 *         avg_price = Decimal(0)
 *         for entry in sell_entries:             # <<<<<<<<<<<<<<
 *             avg_price += Decimal(entry.price) * Decimal(entry.amount)
 *         avg_price = avg_price / amount
 */
  if (likely(PyList_CheckExact(__pyx_v_sell_entries)) || PyTuple_CheckExact(__pyx_v_sell_entries)) {
    __pyx_t_2 = __pyx_v_sell_entries; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_sell_entries); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 539, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 539, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 539, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 539, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 539, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 539, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 539, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 539, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_7(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 539, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_entry, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":540
 *         avg_price = Decimal(0)
 *         for entry in sell_entries:
 *             avg_price += Decimal(entry.price) * Decimal(entry.amount)             # <<<<<<<<<<<<<<
 *         avg_price = avg_price / amount
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_entry, __pyx_n_s_price); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_entry, __pyx_n_s_amount); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_8};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 540, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_4 = PyNumber_Multiply(__pyx_t_1, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_avg_price, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF_SET(__pyx_v_avg_price, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":539
 *         # Get the weighted average price of the trade
 *         avg_price = Decimal(0)
 *         for entry in sell_entries:             # <<<<<<<<<<<<<<
 *             avg_price += Decimal(entry.price) * Decimal(entry.amount)
 *         avg_price = avg_price / amount
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":541
 *         for entry in sell_entries:
 *             avg_price += Decimal(entry.price) * Decimal(entry.amount)
 *         avg_price = avg_price / amount             # <<<<<<<<<<<<<<
 * 
 *         order_candidate = OrderCandidate(
 */
  __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_v_avg_price, __pyx_v_amount); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 541, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF_SET(__pyx_v_avg_price, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":543
 *         avg_price = avg_price / amount
 * 
 *         order_candidate = OrderCandidate(             # <<<<<<<<<<<<<<
 *             trading_pair=trading_pair_str,
 *             # Market orders are not maker orders
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_OrderCandidate); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":544
 * 
 *         order_candidate = OrderCandidate(
 *             trading_pair=trading_pair_str,             # <<<<<<<<<<<<<<
 *             # Market orders are not maker orders
 *             is_maker=False,
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_trading_pair, __pyx_v_trading_pair_str) < 0) __PYX_ERR(0, 544, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":546
 *             trading_pair=trading_pair_str,
 *             # Market orders are not maker orders
 *             is_maker=False,             # <<<<<<<<<<<<<<
 *             order_type=OrderType.MARKET,
 *             order_side=TradeType.SELL,
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_is_maker, Py_False) < 0) __PYX_ERR(0, 544, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":547
 *             # Market orders are not maker orders
 *             is_maker=False,
 *             order_type=OrderType.MARKET,             # <<<<<<<<<<<<<<
 *             order_side=TradeType.SELL,
 *             amount=amount,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_MARKET); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_order_type, __pyx_t_1) < 0) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":548
 *             is_maker=False,
 *             order_type=OrderType.MARKET,
 *             order_side=TradeType.SELL,             # <<<<<<<<<<<<<<
 *             amount=amount,
 *             price=avg_price,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_SELL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_order_side, __pyx_t_4) < 0) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":549
 *             order_type=OrderType.MARKET,
 *             order_side=TradeType.SELL,
 *             amount=amount,             # <<<<<<<<<<<<<<
 *             price=avg_price,
 *             from_total_balances=True
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_amount, __pyx_v_amount) < 0) __PYX_ERR(0, 544, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":550
 *             order_side=TradeType.SELL,
 *             amount=amount,
 *             price=avg_price,             # <<<<<<<<<<<<<<
 *             from_total_balances=True
 *         )
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_price, __pyx_v_avg_price) < 0) __PYX_ERR(0, 544, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":551
 *             amount=amount,
 *             price=avg_price,
 *             from_total_balances=True             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_from_total_balances, Py_True) < 0) __PYX_ERR(0, 544, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":543
 *         avg_price = avg_price / amount
 * 
 *         order_candidate = OrderCandidate(             # <<<<<<<<<<<<<<
 *             trading_pair=trading_pair_str,
 *             # Market orders are not maker orders
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_order_candidate = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":554
 *         )
 * 
 *         adjusted_order_candidate = self._budget_checker.populate_collateral_entries(order_candidate)             # <<<<<<<<<<<<<<
 * 
 *         # Base currency used, including fees.
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base._budget_checker, __pyx_n_s_populate_collateral_entries); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_order_candidate};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 554, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_adjusted_order_candidate = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":557
 * 
 *         # Base currency used, including fees.
 *         sold_amount = adjusted_order_candidate.order_collateral.amount             # <<<<<<<<<<<<<<
 *         # Quote currency acquired, including fees.
 *         acquired_amount = adjusted_order_candidate.potential_returns.amount
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_adjusted_order_candidate, __pyx_n_s_order_collateral); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_amount); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 557, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_sold_amount = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":559
 *         sold_amount = adjusted_order_candidate.order_collateral.amount
 *         # Quote currency acquired, including fees.
 *         acquired_amount = adjusted_order_candidate.potential_returns.amount             # <<<<<<<<<<<<<<
 * 
 *         # It's not possible to fulfill the order, the possible sold amount is less than requested
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_adjusted_order_candidate, __pyx_n_s_potential_returns); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_amount); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 559, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_acquired_amount = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":562
 * 
 *         # It's not possible to fulfill the order, the possible sold amount is less than requested
 *         if sold_amount > base_balance:             # <<<<<<<<<<<<<<
 *             self.logger().warning(f"Insufficient {base_asset} balance available for sell order. "
 *                                   f"{base_balance} {base_asset} available vs. "
 */
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_sold_amount, __pyx_v_base_balance, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 562, __pyx_L1_error)
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 562, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_10) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":563
 *         # It's not possible to fulfill the order, the possible sold amount is less than requested
 *         if sold_amount > base_balance:
 *             self.logger().warning(f"Insufficient {base_asset} balance available for sell order. "             # <<<<<<<<<<<<<<
 *                                   f"{base_balance} {base_asset} available vs. "
 *                                   f"{amount} {base_asset} required for the order.")
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 0+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 563, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warning); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = 0;
    __pyx_t_11 = 127;
    __Pyx_INCREF(__pyx_kp_u_Insufficient);
    __pyx_t_6 += 13;
    __Pyx_GIVEREF(__pyx_kp_u_Insufficient);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Insufficient);
    __pyx_t_1 = __Pyx_PyUnicode_Unicode(__pyx_v_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_11;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u_balance_available_for_sell_orde);
    __pyx_t_6 += 35;
    __Pyx_GIVEREF(__pyx_kp_u_balance_available_for_sell_orde);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u_balance_available_for_sell_orde);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":564
 *         if sold_amount > base_balance:
 *             self.logger().warning(f"Insufficient {base_asset} balance available for sell order. "
 *                                   f"{base_balance} {base_asset} available vs. "             # <<<<<<<<<<<<<<
 *                                   f"{amount} {base_asset} required for the order.")
 *             self.c_trigger_event(
 */
    __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_v_base_balance, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_11;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u__20);
    __pyx_t_6 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__20);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u__20);
    __pyx_t_1 = __Pyx_PyUnicode_Unicode(__pyx_v_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_11;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u_available_vs);
    __pyx_t_6 += 15;
    __Pyx_GIVEREF(__pyx_kp_u_available_vs);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_available_vs);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":565
 *             self.logger().warning(f"Insufficient {base_asset} balance available for sell order. "
 *                                   f"{base_balance} {base_asset} available vs. "
 *                                   f"{amount} {base_asset} required for the order.")             # <<<<<<<<<<<<<<
 *             self.c_trigger_event(
 *                 self.MARKET_ORDER_FAILURE_EVENT_TAG,
 */
    __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_v_amount, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_11;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u__20);
    __pyx_t_6 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__20);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u__20);
    __pyx_t_1 = __Pyx_PyUnicode_Unicode(__pyx_v_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 565, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_11;
    __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_INCREF(__pyx_kp_u_required_for_the_order);
    __pyx_t_6 += 24;
    __Pyx_GIVEREF(__pyx_kp_u_required_for_the_order);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u_required_for_the_order);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":563
 *         # It's not possible to fulfill the order, the possible sold amount is less than requested
 *         if sold_amount > base_balance:
 *             self.logger().warning(f"Insufficient {base_asset} balance available for sell order. "             # <<<<<<<<<<<<<<
 *                                   f"{base_balance} {base_asset} available vs. "
 *                                   f"{amount} {base_asset} required for the order.")
 */
    __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_3, 11, __pyx_t_6, __pyx_t_11); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 563, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_1};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 563, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":567
 *                                   f"{amount} {base_asset} required for the order.")
 *             self.c_trigger_event(
 *                 self.MARKET_ORDER_FAILURE_EVENT_TAG,             # <<<<<<<<<<<<<<
 *                 MarketOrderFailureEvent(self._current_timestamp, order_id, OrderType.MARKET)
 *             )
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_MARKET_ORDER_FAILURE_EVENT_TAG); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 567, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_12 = __Pyx_PyInt_As_int64_t(__pyx_t_4); if (unlikely((__pyx_t_12 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 567, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":568
 *             self.c_trigger_event(
 *                 self.MARKET_ORDER_FAILURE_EVENT_TAG,
 *                 MarketOrderFailureEvent(self._current_timestamp, order_id, OrderType.MARKET)             # <<<<<<<<<<<<<<
 *             )
 *             return
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MarketOrderFailureEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MARKET); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_t_1, __pyx_v_order_id, __pyx_t_8};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 3+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 568, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":566
 *                                   f"{base_balance} {base_asset} available vs. "
 *                                   f"{amount} {base_asset} required for the order.")
 *             self.c_trigger_event(             # <<<<<<<<<<<<<<
 *                 self.MARKET_ORDER_FAILURE_EVENT_TAG,
 *                 MarketOrderFailureEvent(self._current_timestamp, order_id, OrderType.MARKET)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.c_trigger_event(((struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *)__pyx_v_self), __pyx_t_12, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 566, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":570
 *                 MarketOrderFailureEvent(self._current_timestamp, order_id, OrderType.MARKET)
 *             )
 *             return             # <<<<<<<<<<<<<<
 * 
 *         # The order was successfully executed
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":562
 * 
 *         # It's not possible to fulfill the order, the possible sold amount is less than requested
 *         if sold_amount > base_balance:             # <<<<<<<<<<<<<<
 *             self.logger().warning(f"Insufficient {base_asset} balance available for sell order. "
 *                                   f"{base_balance} {base_asset} available vs. "
 */
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":574
 *         # The order was successfully executed
 *         self.c_set_balance(quote_asset,
 *                            quote_balance + acquired_amount)             # <<<<<<<<<<<<<<
 *         self.c_set_balance(base_asset,
 *                            base_balance - sold_amount)
 */
  __pyx_t_2 = PyNumber_Add(__pyx_v_quote_balance, __pyx_v_acquired_amount); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":573
 * 
 *         # The order was successfully executed
 *         self.c_set_balance(quote_asset,             # <<<<<<<<<<<<<<
 *                            quote_balance + acquired_amount)
 *         self.c_set_balance(base_asset,
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_set_balance(__pyx_v_self, __pyx_v_quote_asset, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 573, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":576
 *                            quote_balance + acquired_amount)
 *         self.c_set_balance(base_asset,
 *                            base_balance - sold_amount)             # <<<<<<<<<<<<<<
 * 
 *         # add fee
 */
  __pyx_t_4 = PyNumber_Subtract(__pyx_v_base_balance, __pyx_v_sold_amount); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 576, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":575
 *         self.c_set_balance(quote_asset,
 *                            quote_balance + acquired_amount)
 *         self.c_set_balance(base_asset,             # <<<<<<<<<<<<<<
 *                            base_balance - sold_amount)
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_set_balance(__pyx_v_self, __pyx_v_base_asset, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 575, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":579
 * 
 *         # add fee
 *         fees = build_trade_fee(             # <<<<<<<<<<<<<<
 *             exchange=self.name,
 *             is_maker=False,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_build_trade_fee); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":580
 *         # add fee
 *         fees = build_trade_fee(
 *             exchange=self.name,             # <<<<<<<<<<<<<<
 *             is_maker=False,
 *             base_currency="",
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_name); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_exchange, __pyx_t_8) < 0) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":581
 *         fees = build_trade_fee(
 *             exchange=self.name,
 *             is_maker=False,             # <<<<<<<<<<<<<<
 *             base_currency="",
 *             quote_currency="",
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_is_maker, Py_False) < 0) __PYX_ERR(0, 580, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_base_currency, __pyx_kp_u__11) < 0) __PYX_ERR(0, 580, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_quote_currency, __pyx_kp_u__11) < 0) __PYX_ERR(0, 580, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":584
 *             base_currency="",
 *             quote_currency="",
 *             order_type=OrderType.LIMIT,             # <<<<<<<<<<<<<<
 *             order_side=TradeType.BUY,
 *             amount=Decimal("0"),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_LIMIT); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_order_type, __pyx_t_1) < 0) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":585
 *             quote_currency="",
 *             order_type=OrderType.LIMIT,
 *             order_side=TradeType.BUY,             # <<<<<<<<<<<<<<
 *             amount=Decimal("0"),
 *             price=Decimal("0"),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_BUY); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 585, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_order_side, __pyx_t_8) < 0) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":586
 *             order_type=OrderType.LIMIT,
 *             order_side=TradeType.BUY,
 *             amount=Decimal("0"),             # <<<<<<<<<<<<<<
 *             price=Decimal("0"),
 *         )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_u_0};
    __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 586, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_amount, __pyx_t_8) < 0) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":587
 *             order_side=TradeType.BUY,
 *             amount=Decimal("0"),
 *             price=Decimal("0"),             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_u_0};
    __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 587, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_price, __pyx_t_8) < 0) __PYX_ERR(0, 580, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":579
 * 
 *         # add fee
 *         fees = build_trade_fee(             # <<<<<<<<<<<<<<
 *             exchange=self.name,
 *             is_maker=False,
 */
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_fees = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":590
 *         )
 * 
 *         order_filled_events = OrderFilledEvent.order_filled_events_from_order_book_rows(             # <<<<<<<<<<<<<<
 *             self._current_timestamp, order_id, trading_pair_str, TradeType.SELL,
 *             OrderType.MARKET, fees, sell_entries
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_OrderFilledEvent); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_order_filled_events_from_order_b); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":591
 * 
 *         order_filled_events = OrderFilledEvent.order_filled_events_from_order_book_rows(
 *             self._current_timestamp, order_id, trading_pair_str, TradeType.SELL,             # <<<<<<<<<<<<<<
 *             OrderType.MARKET, fees, sell_entries
 *         )
 */
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_SELL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":592
 *         order_filled_events = OrderFilledEvent.order_filled_events_from_order_book_rows(
 *             self._current_timestamp, order_id, trading_pair_str, TradeType.SELL,
 *             OrderType.MARKET, fees, sell_entries             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_MARKET); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[8] = {__pyx_t_1, __pyx_t_4, __pyx_v_order_id, __pyx_v_trading_pair_str, __pyx_t_3, __pyx_t_9, __pyx_v_fees, __pyx_v_sell_entries};
    __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 7+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_order_filled_events = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":595
 *         )
 * 
 *         for order_filled_event in order_filled_events:             # <<<<<<<<<<<<<<
 *             self.c_trigger_event(self.ORDER_FILLED_EVENT_TAG, order_filled_event)
 * 
 */
  if (likely(PyList_CheckExact(__pyx_v_order_filled_events)) || PyTuple_CheckExact(__pyx_v_order_filled_events)) {
    __pyx_t_8 = __pyx_v_order_filled_events; __Pyx_INCREF(__pyx_t_8);
    __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_6 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_v_order_filled_events); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 595, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 595, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_8))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_8);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 595, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 595, __pyx_L1_error)
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 595, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_8);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 595, __pyx_L1_error)
          #endif
          if (__pyx_t_6 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_6); __Pyx_INCREF(__pyx_t_2); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 595, __pyx_L1_error)
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 595, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_7(__pyx_t_8);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 595, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_order_filled_event, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":596
 * 
 *         for order_filled_event in order_filled_events:
 *             self.c_trigger_event(self.ORDER_FILLED_EVENT_TAG, order_filled_event)             # <<<<<<<<<<<<<<
 * 
 *         self.c_trigger_event(
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_ORDER_FILLED_EVENT_TAG); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 596, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_12 = __Pyx_PyInt_As_int64_t(__pyx_t_2); if (unlikely((__pyx_t_12 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 596, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.c_trigger_event(((struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *)__pyx_v_self), __pyx_t_12, __pyx_v_order_filled_event); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 596, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":595
 *         )
 * 
 *         for order_filled_event in order_filled_events:             # <<<<<<<<<<<<<<
 *             self.c_trigger_event(self.ORDER_FILLED_EVENT_TAG, order_filled_event)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":599
 * 
 *         self.c_trigger_event(
 *             self.SELL_ORDER_COMPLETED_EVENT_TAG,             # <<<<<<<<<<<<<<
 *             SellOrderCompletedEvent(self._current_timestamp,
 *                                     order_id,
 */
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_SELL_ORDER_COMPLETED_EVENT_TAG); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_12 = __Pyx_PyInt_As_int64_t(__pyx_t_8); if (unlikely((__pyx_t_12 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":600
 *         self.c_trigger_event(
 *             self.SELL_ORDER_COMPLETED_EVENT_TAG,
 *             SellOrderCompletedEvent(self._current_timestamp,             # <<<<<<<<<<<<<<
 *                                     order_id,
 *                                     base_asset,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_SellOrderCompletedEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 600, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":606
 *                                     sold_amount,
 *                                     acquired_amount,
 *                                     OrderType.MARKET))             # <<<<<<<<<<<<<<
 * 
 *     cdef c_process_market_orders(self):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MARKET); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[8] = {__pyx_t_3, __pyx_t_9, __pyx_v_order_id, __pyx_v_base_asset, __pyx_v_quote_asset, __pyx_v_sold_amount, __pyx_v_acquired_amount, __pyx_t_4};
    __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 7+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 600, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":598
 *             self.c_trigger_event(self.ORDER_FILLED_EVENT_TAG, order_filled_event)
 * 
 *         self.c_trigger_event(             # <<<<<<<<<<<<<<
 *             self.SELL_ORDER_COMPLETED_EVENT_TAG,
 *             SellOrderCompletedEvent(self._current_timestamp,
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.c_trigger_event(((struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *)__pyx_v_self), __pyx_t_12, __pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 598, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":526
 *                                    OrderType.MARKET))
 * 
 *     cdef c_execute_sell(self, str order_id, str trading_pair_str, object amount):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str quote_asset = self._trading_pairs[trading_pair_str].quote_asset
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_execute_sell", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_quote_asset);
  __Pyx_XDECREF(__pyx_v_base_asset);
  __Pyx_XDECREF(__pyx_v_quote_balance);
  __Pyx_XDECREF(__pyx_v_base_balance);
  __Pyx_XDECREF(__pyx_v_order_book);
  __Pyx_XDECREF(__pyx_v_sell_entries);
  __Pyx_XDECREF(__pyx_v_avg_price);
  __Pyx_XDECREF(__pyx_v_entry);
  __Pyx_XDECREF(__pyx_v_order_candidate);
  __Pyx_XDECREF(__pyx_v_adjusted_order_candidate);
  __Pyx_XDECREF(__pyx_v_sold_amount);
  __Pyx_XDECREF(__pyx_v_acquired_amount);
  __Pyx_XDECREF(__pyx_v_fees);
  __Pyx_XDECREF(__pyx_v_order_filled_events);
  __Pyx_XDECREF(__pyx_v_order_filled_event);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":608
 *                                     OrderType.MARKET))
 * 
 *     cdef c_process_market_orders(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             QueuedOrder front_order = None
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_process_market_orders(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v_front_order = 0;
  CYTHON_UNUSED PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  char const *__pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  PyObject *__pyx_t_21 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_process_market_orders", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":610
 *     cdef c_process_market_orders(self):
 *         cdef:
 *             QueuedOrder front_order = None             # <<<<<<<<<<<<<<
 *         while len(self._queued_orders) > 0:
 *             front_order = self._queued_orders[0]
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_front_order = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *)Py_None);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":611
 *         cdef:
 *             QueuedOrder front_order = None
 *         while len(self._queued_orders) > 0:             # <<<<<<<<<<<<<<
 *             front_order = self._queued_orders[0]
 *             if front_order.create_timestamp <= self._current_timestamp - self.TRADE_EXECUTION_DELAY:
 */
  while (1) {
    __pyx_t_1 = __pyx_v_self->_queued_orders;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 611, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = (__pyx_t_2 > 0);
    if (!__pyx_t_3) break;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":612
 *             QueuedOrder front_order = None
 *         while len(self._queued_orders) > 0:
 *             front_order = self._queued_orders[0]             # <<<<<<<<<<<<<<
 *             if front_order.create_timestamp <= self._current_timestamp - self.TRADE_EXECUTION_DELAY:
 *                 self._queued_orders.popleft()
 */
    __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_self->_queued_orders, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 612, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder))))) __PYX_ERR(0, 612, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_front_order, ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":613
 *         while len(self._queued_orders) > 0:
 *             front_order = self._queued_orders[0]
 *             if front_order.create_timestamp <= self._current_timestamp - self.TRADE_EXECUTION_DELAY:             # <<<<<<<<<<<<<<
 *                 self._queued_orders.popleft()
 *                 try:
 */
    __pyx_t_1 = PyFloat_FromDouble(__pyx_v_front_order->create_timestamp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_TRADE_EXECUTION_DELAY); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyNumber_Subtract(__pyx_t_4, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_t_6, Py_LE); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 613, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_3) {

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":614
 *             front_order = self._queued_orders[0]
 *             if front_order.create_timestamp <= self._current_timestamp - self.TRADE_EXECUTION_DELAY:
 *                 self._queued_orders.popleft()             # <<<<<<<<<<<<<<
 *                 try:
 *                     if front_order.is_buy:
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_queued_orders, __pyx_n_s_popleft); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 614, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = NULL;
      __pyx_t_7 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_7 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 614, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":615
 *             if front_order.create_timestamp <= self._current_timestamp - self.TRADE_EXECUTION_DELAY:
 *                 self._queued_orders.popleft()
 *                 try:             # <<<<<<<<<<<<<<
 *                     if front_order.is_buy:
 *                         self.c_execute_buy(front_order.order_id, front_order.trading_pair, front_order.amount)
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_8, &__pyx_t_9, &__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_8);
        __Pyx_XGOTREF(__pyx_t_9);
        __Pyx_XGOTREF(__pyx_t_10);
        /*try:*/ {

          /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":616
 *                 self._queued_orders.popleft()
 *                 try:
 *                     if front_order.is_buy:             # <<<<<<<<<<<<<<
 *                         self.c_execute_buy(front_order.order_id, front_order.trading_pair, front_order.amount)
 *                     else:
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_front_order), __pyx_n_s_is_buy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 616, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 616, __pyx_L6_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (__pyx_t_3) {

            /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":617
 *                 try:
 *                     if front_order.is_buy:
 *                         self.c_execute_buy(front_order.order_id, front_order.trading_pair, front_order.amount)             # <<<<<<<<<<<<<<
 *                     else:
 *                         self.c_execute_sell(front_order.order_id, front_order.trading_pair, front_order.amount)
 */
            __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_front_order), __pyx_n_s_order_id); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 617, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_5))) __PYX_ERR(0, 617, __pyx_L6_error)
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_front_order), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 617, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_6);
            if (!(likely(PyUnicode_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_6))) __PYX_ERR(0, 617, __pyx_L6_error)
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_front_order), __pyx_n_s_amount); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 617, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __pyx_t_4 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_execute_buy(__pyx_v_self, ((PyObject*)__pyx_t_5), ((PyObject*)__pyx_t_6), __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 617, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

            /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":616
 *                 self._queued_orders.popleft()
 *                 try:
 *                     if front_order.is_buy:             # <<<<<<<<<<<<<<
 *                         self.c_execute_buy(front_order.order_id, front_order.trading_pair, front_order.amount)
 *                     else:
 */
            goto __pyx_L14;
          }

          /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":619
 *                         self.c_execute_buy(front_order.order_id, front_order.trading_pair, front_order.amount)
 *                     else:
 *                         self.c_execute_sell(front_order.order_id, front_order.trading_pair, front_order.amount)             # <<<<<<<<<<<<<<
 *                 except Exception as e:
 *                     self.logger().error("Error executing queued order.", exc_info=True)
 */
          /*else*/ {
            __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_front_order), __pyx_n_s_order_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 619, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (!(likely(PyUnicode_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_4))) __PYX_ERR(0, 619, __pyx_L6_error)
            __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_front_order), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 619, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 619, __pyx_L6_error)
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_front_order), __pyx_n_s_amount); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 619, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_execute_sell(__pyx_v_self, ((PyObject*)__pyx_t_4), ((PyObject*)__pyx_t_1), __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 619, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          }
          __pyx_L14:;

          /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":615
 *             if front_order.create_timestamp <= self._current_timestamp - self.TRADE_EXECUTION_DELAY:
 *                 self._queued_orders.popleft()
 *                 try:             # <<<<<<<<<<<<<<
 *                     if front_order.is_buy:
 *                         self.c_execute_buy(front_order.order_id, front_order.trading_pair, front_order.amount)
 */
        }
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        goto __pyx_L13_try_end;
        __pyx_L6_error:;
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":620
 *                     else:
 *                         self.c_execute_sell(front_order.order_id, front_order.trading_pair, front_order.amount)
 *                 except Exception as e:             # <<<<<<<<<<<<<<
 *                     self.logger().error("Error executing queued order.", exc_info=True)
 *             else:
 */
        __pyx_t_11 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
        if (__pyx_t_11) {
          __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_process_market_orders", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_1) < 0) __PYX_ERR(0, 620, __pyx_L8_except_error)
          __Pyx_XGOTREF(__pyx_t_5);
          __Pyx_XGOTREF(__pyx_t_6);
          __Pyx_XGOTREF(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_6);
          __pyx_v_e = __pyx_t_6;
          /*try:*/ {

            /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":621
 *                         self.c_execute_sell(front_order.order_id, front_order.trading_pair, front_order.amount)
 *                 except Exception as e:
 *                     self.logger().error("Error executing queued order.", exc_info=True)             # <<<<<<<<<<<<<<
 *             else:
 *                 return
 */
            __pyx_t_12 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 621, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_13 = NULL;
            __pyx_t_7 = 0;
            #if CYTHON_UNPACK_METHODS
            if (likely(PyMethod_Check(__pyx_t_12))) {
              __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
              if (likely(__pyx_t_13)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
                __Pyx_INCREF(__pyx_t_13);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_12, function);
                __pyx_t_7 = 1;
              }
            }
            #endif
            {
              PyObject *__pyx_callargs[2] = {__pyx_t_13, NULL};
              __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_7, 0+__pyx_t_7);
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 621, __pyx_L20_error)
              __Pyx_GOTREF(__pyx_t_4);
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            }
            __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_error); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 621, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 621, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_4);
            if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_exc_info, Py_True) < 0) __PYX_ERR(0, 621, __pyx_L20_error)
            __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_tuple__21, __pyx_t_4); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 621, __pyx_L20_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }

          /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":620
 *                     else:
 *                         self.c_execute_sell(front_order.order_id, front_order.trading_pair, front_order.amount)
 *                 except Exception as e:             # <<<<<<<<<<<<<<
 *                     self.logger().error("Error executing queued order.", exc_info=True)
 *             else:
 */
          /*finally:*/ {
            /*normal exit:*/{
              __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
              goto __pyx_L21;
            }
            __pyx_L20_error:;
            /*exception exit:*/{
              __Pyx_PyThreadState_declare
              __Pyx_PyThreadState_assign
              __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
              __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
              __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
              if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_19, &__pyx_t_20, &__pyx_t_21);
              if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18) < 0)) __Pyx_ErrFetch(&__pyx_t_16, &__pyx_t_17, &__pyx_t_18);
              __Pyx_XGOTREF(__pyx_t_16);
              __Pyx_XGOTREF(__pyx_t_17);
              __Pyx_XGOTREF(__pyx_t_18);
              __Pyx_XGOTREF(__pyx_t_19);
              __Pyx_XGOTREF(__pyx_t_20);
              __Pyx_XGOTREF(__pyx_t_21);
              __pyx_t_11 = __pyx_lineno; __pyx_t_14 = __pyx_clineno; __pyx_t_15 = __pyx_filename;
              {
                __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
              }
              if (PY_MAJOR_VERSION >= 3) {
                __Pyx_XGIVEREF(__pyx_t_19);
                __Pyx_XGIVEREF(__pyx_t_20);
                __Pyx_XGIVEREF(__pyx_t_21);
                __Pyx_ExceptionReset(__pyx_t_19, __pyx_t_20, __pyx_t_21);
              }
              __Pyx_XGIVEREF(__pyx_t_16);
              __Pyx_XGIVEREF(__pyx_t_17);
              __Pyx_XGIVEREF(__pyx_t_18);
              __Pyx_ErrRestore(__pyx_t_16, __pyx_t_17, __pyx_t_18);
              __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0; __pyx_t_21 = 0;
              __pyx_lineno = __pyx_t_11; __pyx_clineno = __pyx_t_14; __pyx_filename = __pyx_t_15;
              goto __pyx_L8_except_error;
            }
            __pyx_L21:;
          }
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
          goto __pyx_L7_exception_handled;
        }
        goto __pyx_L8_except_error;

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":615
 *             if front_order.create_timestamp <= self._current_timestamp - self.TRADE_EXECUTION_DELAY:
 *                 self._queued_orders.popleft()
 *                 try:             # <<<<<<<<<<<<<<
 *                     if front_order.is_buy:
 *                         self.c_execute_buy(front_order.order_id, front_order.trading_pair, front_order.amount)
 */
        __pyx_L8_except_error:;
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
        goto __pyx_L1_error;
        __pyx_L7_exception_handled:;
        __Pyx_XGIVEREF(__pyx_t_8);
        __Pyx_XGIVEREF(__pyx_t_9);
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_ExceptionReset(__pyx_t_8, __pyx_t_9, __pyx_t_10);
        __pyx_L13_try_end:;
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":613
 *         while len(self._queued_orders) > 0:
 *             front_order = self._queued_orders[0]
 *             if front_order.create_timestamp <= self._current_timestamp - self.TRADE_EXECUTION_DELAY:             # <<<<<<<<<<<<<<
 *                 self._queued_orders.popleft()
 *                 try:
 */
      goto __pyx_L5;
    }

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":623
 *                     self.logger().error("Error executing queued order.", exc_info=True)
 *             else:
 *                 return             # <<<<<<<<<<<<<<
 * 
 *     cdef c_delete_limit_order(self,
 */
    /*else*/ {
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;
    }
    __pyx_L5:;
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":608
 *                                     OrderType.MARKET))
 * 
 *     cdef c_process_market_orders(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             QueuedOrder front_order = None
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_process_market_orders", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_front_order);
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":625
 *                 return
 * 
 *     cdef c_delete_limit_order(self,             # <<<<<<<<<<<<<<
 *                               LimitOrders *limit_orders_map_ptr,
 *                               LimitOrdersIterator *map_it_ptr,
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_delete_limit_order(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *__pyx_v_limit_orders_map_ptr, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *__pyx_v_map_it_ptr, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator const __pyx_v_orders_it) {
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders *__pyx_v_orders_collection_ptr;
  CYTHON_UNUSED PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("c_delete_limit_order", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":630
 *                               const SingleTradingPairLimitOrdersIterator orders_it):
 *         cdef:
 *             SingleTradingPairLimitOrders *orders_collection_ptr = address(deref(deref(map_it_ptr)).second)             # <<<<<<<<<<<<<<
 *         try:
 *             orders_collection_ptr.erase(orders_it)
 */
  __pyx_v_orders_collection_ptr = (&(*(*__pyx_v_map_it_ptr)).second);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":631
 *         cdef:
 *             SingleTradingPairLimitOrders *orders_collection_ptr = address(deref(deref(map_it_ptr)).second)
 *         try:             # <<<<<<<<<<<<<<
 *             orders_collection_ptr.erase(orders_it)
 *             if orders_collection_ptr.empty():
 */
  {
    (void)__pyx_t_1; (void)__pyx_t_2; (void)__pyx_t_3; /* mark used */
    /*try:*/ {

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":632
 *             SingleTradingPairLimitOrders *orders_collection_ptr = address(deref(deref(map_it_ptr)).second)
 *         try:
 *             orders_collection_ptr.erase(orders_it)             # <<<<<<<<<<<<<<
 *             if orders_collection_ptr.empty():
 *                 map_it_ptr[0] = limit_orders_map_ptr.erase(deref(map_it_ptr))
 */
      (void)(__pyx_v_orders_collection_ptr->erase(__pyx_v_orders_it));

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":633
 *         try:
 *             orders_collection_ptr.erase(orders_it)
 *             if orders_collection_ptr.empty():             # <<<<<<<<<<<<<<
 *                 map_it_ptr[0] = limit_orders_map_ptr.erase(deref(map_it_ptr))
 *             return True
 */
      __pyx_t_4 = __pyx_v_orders_collection_ptr->empty();
      if (__pyx_t_4) {

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":634
 *             orders_collection_ptr.erase(orders_it)
 *             if orders_collection_ptr.empty():
 *                 map_it_ptr[0] = limit_orders_map_ptr.erase(deref(map_it_ptr))             # <<<<<<<<<<<<<<
 *             return True
 *         except Exception as err:
 */
        (__pyx_v_map_it_ptr[0]) = __pyx_v_limit_orders_map_ptr->erase((*__pyx_v_map_it_ptr));

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":633
 *         try:
 *             orders_collection_ptr.erase(orders_it)
 *             if orders_collection_ptr.empty():             # <<<<<<<<<<<<<<
 *                 map_it_ptr[0] = limit_orders_map_ptr.erase(deref(map_it_ptr))
 *             return True
 */
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":635
 *             if orders_collection_ptr.empty():
 *                 map_it_ptr[0] = limit_orders_map_ptr.erase(deref(map_it_ptr))
 *             return True             # <<<<<<<<<<<<<<
 *         except Exception as err:
 *             self.logger().error("Error deleting limit order.", exc_info=True)
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_True);
      __pyx_r = Py_True;
      goto __pyx_L7_try_return;

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":631
 *         cdef:
 *             SingleTradingPairLimitOrders *orders_collection_ptr = address(deref(deref(map_it_ptr)).second)
 *         try:             # <<<<<<<<<<<<<<
 *             orders_collection_ptr.erase(orders_it)
 *             if orders_collection_ptr.empty():
 */
    }
    __pyx_L7_try_return:;
    goto __pyx_L0;
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":625
 *                 return
 * 
 *     cdef c_delete_limit_order(self,             # <<<<<<<<<<<<<<
 *                               LimitOrders *limit_orders_map_ptr,
 *                               LimitOrdersIterator *map_it_ptr,
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":640
 *             return False
 * 
 *     cdef c_process_limit_bid_order(self,             # <<<<<<<<<<<<<<
 *                                    LimitOrders *limit_orders_map_ptr,
 *                                    LimitOrdersIterator *map_it_ptr,
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_process_limit_bid_order(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *__pyx_v_limit_orders_map_ptr, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *__pyx_v_map_it_ptr, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator __pyx_v_orders_it) {
  LimitOrder const *__pyx_v_cpp_limit_order_ptr;
  PyObject *__pyx_v_trading_pair_str = 0;
  PyObject *__pyx_v_quote_asset = 0;
  PyObject *__pyx_v_base_asset = 0;
  PyObject *__pyx_v_order_id = 0;
  PyObject *__pyx_v_amount = 0;
  PyObject *__pyx_v_price = 0;
  PyObject *__pyx_v_quote_balance = 0;
  PyObject *__pyx_v_base_balance = 0;
  PyObject *__pyx_v_order_candidate = NULL;
  PyObject *__pyx_v_adjusted_order_candidate = NULL;
  PyObject *__pyx_v_paid_amount = NULL;
  PyObject *__pyx_v_acquired_amount = NULL;
  PyObject *__pyx_v_fees = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_UCS4 __pyx_t_9;
  int64_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_process_limit_bid_order", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":645
 *                                    SingleTradingPairLimitOrdersIterator orders_it):
 *         cdef:
 *             const CPPLimitOrder *cpp_limit_order_ptr = address(deref(orders_it))             # <<<<<<<<<<<<<<
 *             str trading_pair_str = cpp_limit_order_ptr.getTradingPair().decode("utf8")
 *             str quote_asset = cpp_limit_order_ptr.getQuoteCurrency().decode("utf8")
 */
  __pyx_v_cpp_limit_order_ptr = (&(*__pyx_v_orders_it));

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":646
 *         cdef:
 *             const CPPLimitOrder *cpp_limit_order_ptr = address(deref(orders_it))
 *             str trading_pair_str = cpp_limit_order_ptr.getTradingPair().decode("utf8")             # <<<<<<<<<<<<<<
 *             str quote_asset = cpp_limit_order_ptr.getQuoteCurrency().decode("utf8")
 *             str base_asset = cpp_limit_order_ptr.getBaseCurrency().decode("utf8")
 */
  __pyx_t_1 = __Pyx_decode_cpp_string(__pyx_v_cpp_limit_order_ptr->getTradingPair(), 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_trading_pair_str = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":647
 *             const CPPLimitOrder *cpp_limit_order_ptr = address(deref(orders_it))
 *             str trading_pair_str = cpp_limit_order_ptr.getTradingPair().decode("utf8")
 *             str quote_asset = cpp_limit_order_ptr.getQuoteCurrency().decode("utf8")             # <<<<<<<<<<<<<<
 *             str base_asset = cpp_limit_order_ptr.getBaseCurrency().decode("utf8")
 *             str order_id = cpp_limit_order_ptr.getClientOrderID().decode("utf8")
 */
  __pyx_t_1 = __Pyx_decode_cpp_string(__pyx_v_cpp_limit_order_ptr->getQuoteCurrency(), 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 647, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_quote_asset = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":648
 *             str trading_pair_str = cpp_limit_order_ptr.getTradingPair().decode("utf8")
 *             str quote_asset = cpp_limit_order_ptr.getQuoteCurrency().decode("utf8")
 *             str base_asset = cpp_limit_order_ptr.getBaseCurrency().decode("utf8")             # <<<<<<<<<<<<<<
 *             str order_id = cpp_limit_order_ptr.getClientOrderID().decode("utf8")
 *             object amount = <object> cpp_limit_order_ptr.getQuantity()
 */
  __pyx_t_1 = __Pyx_decode_cpp_string(__pyx_v_cpp_limit_order_ptr->getBaseCurrency(), 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 648, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_base_asset = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":649
 *             str quote_asset = cpp_limit_order_ptr.getQuoteCurrency().decode("utf8")
 *             str base_asset = cpp_limit_order_ptr.getBaseCurrency().decode("utf8")
 *             str order_id = cpp_limit_order_ptr.getClientOrderID().decode("utf8")             # <<<<<<<<<<<<<<
 *             object amount = <object> cpp_limit_order_ptr.getQuantity()
 *             object price = <object> cpp_limit_order_ptr.getPrice()
 */
  __pyx_t_1 = __Pyx_decode_cpp_string(__pyx_v_cpp_limit_order_ptr->getClientOrderID(), 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 649, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_order_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":650
 *             str base_asset = cpp_limit_order_ptr.getBaseCurrency().decode("utf8")
 *             str order_id = cpp_limit_order_ptr.getClientOrderID().decode("utf8")
 *             object amount = <object> cpp_limit_order_ptr.getQuantity()             # <<<<<<<<<<<<<<
 *             object price = <object> cpp_limit_order_ptr.getPrice()
 *             object quote_balance = self.c_get_balance(quote_asset)
 */
  __pyx_t_2 = __pyx_v_cpp_limit_order_ptr->getQuantity();
  __pyx_t_1 = ((PyObject *)__pyx_t_2);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_amount = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":651
 *             str order_id = cpp_limit_order_ptr.getClientOrderID().decode("utf8")
 *             object amount = <object> cpp_limit_order_ptr.getQuantity()
 *             object price = <object> cpp_limit_order_ptr.getPrice()             # <<<<<<<<<<<<<<
 *             object quote_balance = self.c_get_balance(quote_asset)
 *             object base_balance = self.c_get_balance(base_asset)
 */
  __pyx_t_2 = __pyx_v_cpp_limit_order_ptr->getPrice();
  __pyx_t_1 = ((PyObject *)__pyx_t_2);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_price = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":652
 *             object amount = <object> cpp_limit_order_ptr.getQuantity()
 *             object price = <object> cpp_limit_order_ptr.getPrice()
 *             object quote_balance = self.c_get_balance(quote_asset)             # <<<<<<<<<<<<<<
 *             object base_balance = self.c_get_balance(base_asset)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_get_balance(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), __pyx_v_quote_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 652, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_quote_balance = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":653
 *             object price = <object> cpp_limit_order_ptr.getPrice()
 *             object quote_balance = self.c_get_balance(quote_asset)
 *             object base_balance = self.c_get_balance(base_asset)             # <<<<<<<<<<<<<<
 * 
 *         order_candidate = OrderCandidate(
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_get_balance(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), __pyx_v_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_base_balance = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":655
 *             object base_balance = self.c_get_balance(base_asset)
 * 
 *         order_candidate = OrderCandidate(             # <<<<<<<<<<<<<<
 *             trading_pair=trading_pair_str,
 *             is_maker=True,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_OrderCandidate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":656
 * 
 *         order_candidate = OrderCandidate(
 *             trading_pair=trading_pair_str,             # <<<<<<<<<<<<<<
 *             is_maker=True,
 *             order_type=OrderType.LIMIT,
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_trading_pair, __pyx_v_trading_pair_str) < 0) __PYX_ERR(0, 656, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":657
 *         order_candidate = OrderCandidate(
 *             trading_pair=trading_pair_str,
 *             is_maker=True,             # <<<<<<<<<<<<<<
 *             order_type=OrderType.LIMIT,
 *             order_side=TradeType.BUY,
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_is_maker, Py_True) < 0) __PYX_ERR(0, 656, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":658
 *             trading_pair=trading_pair_str,
 *             is_maker=True,
 *             order_type=OrderType.LIMIT,             # <<<<<<<<<<<<<<
 *             order_side=TradeType.BUY,
 *             amount=amount,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_LIMIT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_order_type, __pyx_t_5) < 0) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":659
 *             is_maker=True,
 *             order_type=OrderType.LIMIT,
 *             order_side=TradeType.BUY,             # <<<<<<<<<<<<<<
 *             amount=amount,
 *             price=price,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_BUY); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 659, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_order_side, __pyx_t_4) < 0) __PYX_ERR(0, 656, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":660
 *             order_type=OrderType.LIMIT,
 *             order_side=TradeType.BUY,
 *             amount=amount,             # <<<<<<<<<<<<<<
 *             price=price,
 *             from_total_balances=True
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_amount, __pyx_v_amount) < 0) __PYX_ERR(0, 656, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":661
 *             order_side=TradeType.BUY,
 *             amount=amount,
 *             price=price,             # <<<<<<<<<<<<<<
 *             from_total_balances=True
 *         )
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_price, __pyx_v_price) < 0) __PYX_ERR(0, 656, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":662
 *             amount=amount,
 *             price=price,
 *             from_total_balances=True             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_from_total_balances, Py_True) < 0) __PYX_ERR(0, 656, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":655
 *             object base_balance = self.c_get_balance(base_asset)
 * 
 *         order_candidate = OrderCandidate(             # <<<<<<<<<<<<<<
 *             trading_pair=trading_pair_str,
 *             is_maker=True,
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_order_candidate = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":665
 *         )
 * 
 *         adjusted_order_candidate = self._budget_checker.populate_collateral_entries(order_candidate)             # <<<<<<<<<<<<<<
 * 
 *         # Quote currency used, including fees.
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base._budget_checker, __pyx_n_s_populate_collateral_entries); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 665, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_order_candidate};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_adjusted_order_candidate = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":668
 * 
 *         # Quote currency used, including fees.
 *         paid_amount = adjusted_order_candidate.order_collateral.amount             # <<<<<<<<<<<<<<
 *         # Base currency acquired, including fees.
 *         acquired_amount = adjusted_order_candidate.potential_returns.amount
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_adjusted_order_candidate, __pyx_n_s_order_collateral); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_amount); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 668, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_paid_amount = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":670
 *         paid_amount = adjusted_order_candidate.order_collateral.amount
 *         # Base currency acquired, including fees.
 *         acquired_amount = adjusted_order_candidate.potential_returns.amount             # <<<<<<<<<<<<<<
 * 
 *         # It's not possible to fulfill the order, the possible acquired amount is less than requested
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_adjusted_order_candidate, __pyx_n_s_potential_returns); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_amount); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 670, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_acquired_amount = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":673
 * 
 *         # It's not possible to fulfill the order, the possible acquired amount is less than requested
 *         if paid_amount > quote_balance:             # <<<<<<<<<<<<<<
 *             self.logger().warning(f"Not enough {quote_asset} balance to fill limit buy order on {trading_pair_str}. "
 *                                   f"{paid_amount:.8g} {quote_asset} needed vs. "
 */
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_paid_amount, __pyx_v_quote_balance, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 673, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_7) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":674
 *         # It's not possible to fulfill the order, the possible acquired amount is less than requested
 *         if paid_amount > quote_balance:
 *             self.logger().warning(f"Not enough {quote_asset} balance to fill limit buy order on {trading_pair_str}. "             # <<<<<<<<<<<<<<
 *                                   f"{paid_amount:.8g} {quote_asset} needed vs. "
 *                                   f"{quote_balance:.8g} {quote_asset} available.")
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warning); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(13); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = 0;
    __pyx_t_9 = 127;
    __Pyx_INCREF(__pyx_kp_u_Not_enough);
    __pyx_t_8 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_Not_enough);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Not_enough);
    __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_v_quote_asset); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_9;
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u_balance_to_fill_limit_buy_order);
    __pyx_t_8 += 36;
    __Pyx_GIVEREF(__pyx_kp_u_balance_to_fill_limit_buy_order);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u_balance_to_fill_limit_buy_order);
    __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_v_trading_pair_str); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_9;
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u__22);
    __pyx_t_8 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__22);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u__22);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":675
 *         if paid_amount > quote_balance:
 *             self.logger().warning(f"Not enough {quote_asset} balance to fill limit buy order on {trading_pair_str}. "
 *                                   f"{paid_amount:.8g} {quote_asset} needed vs. "             # <<<<<<<<<<<<<<
 *                                   f"{quote_balance:.8g} {quote_asset} available.")
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_Format(__pyx_v_paid_amount, __pyx_kp_u_8g); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_9;
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u__20);
    __pyx_t_8 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__20);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u__20);
    __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_v_quote_asset); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_9;
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u_needed_vs);
    __pyx_t_8 += 12;
    __Pyx_GIVEREF(__pyx_kp_u_needed_vs);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_needed_vs);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":676
 *             self.logger().warning(f"Not enough {quote_asset} balance to fill limit buy order on {trading_pair_str}. "
 *                                   f"{paid_amount:.8g} {quote_asset} needed vs. "
 *                                   f"{quote_balance:.8g} {quote_asset} available.")             # <<<<<<<<<<<<<<
 * 
 *             self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 */
    __pyx_t_5 = __Pyx_PyObject_Format(__pyx_v_quote_balance, __pyx_kp_u_8g); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_9;
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u__20);
    __pyx_t_8 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__20);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u__20);
    __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_v_quote_asset); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_9;
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 11, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u_available);
    __pyx_t_8 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_available);
    PyTuple_SET_ITEM(__pyx_t_3, 12, __pyx_kp_u_available);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":674
 *         # It's not possible to fulfill the order, the possible acquired amount is less than requested
 *         if paid_amount > quote_balance:
 *             self.logger().warning(f"Not enough {quote_asset} balance to fill limit buy order on {trading_pair_str}. "             # <<<<<<<<<<<<<<
 *                                   f"{paid_amount:.8g} {quote_asset} needed vs. "
 *                                   f"{quote_balance:.8g} {quote_asset} available.")
 */
    __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_3, 13, __pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 674, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":678
 *                                   f"{quote_balance:.8g} {quote_asset} available.")
 * 
 *             self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)             # <<<<<<<<<<<<<<
 *             self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,
 *                                  OrderCancelledEvent(self._current_timestamp,
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_delete_limit_order(__pyx_v_self, __pyx_v_limit_orders_map_ptr, __pyx_v_map_it_ptr, __pyx_v_orders_it); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 678, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":679
 * 
 *             self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 *             self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,             # <<<<<<<<<<<<<<
 *                                  OrderCancelledEvent(self._current_timestamp,
 *                                                      order_id)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_MARKET_ORDER_CANCELED_EVENT_TAG); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_10 = __Pyx_PyInt_As_int64_t(__pyx_t_4); if (unlikely((__pyx_t_10 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 679, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":680
 *             self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 *             self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,
 *                                  OrderCancelledEvent(self._current_timestamp,             # <<<<<<<<<<<<<<
 *                                                      order_id)
 *                                  )
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_OrderCancelledEvent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 680, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 680, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":681
 *             self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,
 *                                  OrderCancelledEvent(self._current_timestamp,
 *                                                      order_id)             # <<<<<<<<<<<<<<
 *                                  )
 *             return
 */
    __pyx_t_3 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_t_5, __pyx_v_order_id};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 680, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":679
 * 
 *             self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 *             self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,             # <<<<<<<<<<<<<<
 *                                  OrderCancelledEvent(self._current_timestamp,
 *                                                      order_id)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.c_trigger_event(((struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *)__pyx_v_self), __pyx_t_10, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":683
 *                                                      order_id)
 *                                  )
 *             return             # <<<<<<<<<<<<<<
 * 
 *         # The order was successfully executed
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":673
 * 
 *         # It's not possible to fulfill the order, the possible acquired amount is less than requested
 *         if paid_amount > quote_balance:             # <<<<<<<<<<<<<<
 *             self.logger().warning(f"Not enough {quote_asset} balance to fill limit buy order on {trading_pair_str}. "
 *                                   f"{paid_amount:.8g} {quote_asset} needed vs. "
 */
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":687
 *         # The order was successfully executed
 *         self.c_set_balance(quote_asset,
 *                            quote_balance - paid_amount)             # <<<<<<<<<<<<<<
 *         self.c_set_balance(base_asset,
 *                            base_balance + acquired_amount)
 */
  __pyx_t_1 = PyNumber_Subtract(__pyx_v_quote_balance, __pyx_v_paid_amount); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":686
 * 
 *         # The order was successfully executed
 *         self.c_set_balance(quote_asset,             # <<<<<<<<<<<<<<
 *                            quote_balance - paid_amount)
 *         self.c_set_balance(base_asset,
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_set_balance(__pyx_v_self, __pyx_v_quote_asset, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 686, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":689
 *                            quote_balance - paid_amount)
 *         self.c_set_balance(base_asset,
 *                            base_balance + acquired_amount)             # <<<<<<<<<<<<<<
 * 
 *         # add fee
 */
  __pyx_t_4 = PyNumber_Add(__pyx_v_base_balance, __pyx_v_acquired_amount); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 689, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":688
 *         self.c_set_balance(quote_asset,
 *                            quote_balance - paid_amount)
 *         self.c_set_balance(base_asset,             # <<<<<<<<<<<<<<
 *                            base_balance + acquired_amount)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_set_balance(__pyx_v_self, __pyx_v_base_asset, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":692
 * 
 *         # add fee
 *         fees = build_trade_fee(             # <<<<<<<<<<<<<<
 *             exchange=self.name,
 *             is_maker=True,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_build_trade_fee); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":693
 *         # add fee
 *         fees = build_trade_fee(
 *             exchange=self.name,             # <<<<<<<<<<<<<<
 *             is_maker=True,
 *             base_currency="",
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_exchange, __pyx_t_5) < 0) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":694
 *         fees = build_trade_fee(
 *             exchange=self.name,
 *             is_maker=True,             # <<<<<<<<<<<<<<
 *             base_currency="",
 *             quote_currency="",
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_is_maker, Py_True) < 0) __PYX_ERR(0, 693, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_base_currency, __pyx_kp_u__11) < 0) __PYX_ERR(0, 693, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_quote_currency, __pyx_kp_u__11) < 0) __PYX_ERR(0, 693, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":697
 *             base_currency="",
 *             quote_currency="",
 *             order_type=OrderType.LIMIT,             # <<<<<<<<<<<<<<
 *             order_side=TradeType.BUY,
 *             amount=Decimal("0"),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_LIMIT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 697, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_order_type, __pyx_t_3) < 0) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":698
 *             quote_currency="",
 *             order_type=OrderType.LIMIT,
 *             order_side=TradeType.BUY,             # <<<<<<<<<<<<<<
 *             amount=Decimal("0"),
 *             price=Decimal("0"),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_BUY); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_order_side, __pyx_t_5) < 0) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":699
 *             order_type=OrderType.LIMIT,
 *             order_side=TradeType.BUY,
 *             amount=Decimal("0"),             # <<<<<<<<<<<<<<
 *             price=Decimal("0"),
 *         )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 699, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_11 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_11)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_11);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_kp_u_0};
    __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 699, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_amount, __pyx_t_5) < 0) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":700
 *             order_side=TradeType.BUY,
 *             amount=Decimal("0"),
 *             price=Decimal("0"),             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_11 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_11)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_11);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_kp_u_0};
    __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 700, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_price, __pyx_t_5) < 0) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":692
 * 
 *         # add fee
 *         fees = build_trade_fee(             # <<<<<<<<<<<<<<
 *             exchange=self.name,
 *             is_maker=True,
 */
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_fees = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":705
 *         # Emit the trade and order completed events.
 *         self.c_trigger_event(
 *             self.ORDER_FILLED_EVENT_TAG,             # <<<<<<<<<<<<<<
 *             OrderFilledEvent(
 *                 self._current_timestamp,
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_ORDER_FILLED_EVENT_TAG); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_PyInt_As_int64_t(__pyx_t_5); if (unlikely((__pyx_t_10 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":706
 *         self.c_trigger_event(
 *             self.ORDER_FILLED_EVENT_TAG,
 *             OrderFilledEvent(             # <<<<<<<<<<<<<<
 *                 self._current_timestamp,
 *                 order_id,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_OrderFilledEvent); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":707
 *             self.ORDER_FILLED_EVENT_TAG,
 *             OrderFilledEvent(
 *                 self._current_timestamp,             # <<<<<<<<<<<<<<
 *                 order_id,
 *                 trading_pair_str,
 */
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 707, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":710
 *                 order_id,
 *                 trading_pair_str,
 *                 TradeType.BUY,             # <<<<<<<<<<<<<<
 *                 OrderType.LIMIT,
 *                 <object> cpp_limit_order_ptr.getPrice(),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_BUY); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":711
 *                 trading_pair_str,
 *                 TradeType.BUY,
 *                 OrderType.LIMIT,             # <<<<<<<<<<<<<<
 *                 <object> cpp_limit_order_ptr.getPrice(),
 *                 <object> cpp_limit_order_ptr.getQuantity(),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LIMIT); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":712
 *                 TradeType.BUY,
 *                 OrderType.LIMIT,
 *                 <object> cpp_limit_order_ptr.getPrice(),             # <<<<<<<<<<<<<<
 *                 <object> cpp_limit_order_ptr.getQuantity(),
 *                 fees,
 */
  __pyx_t_2 = __pyx_v_cpp_limit_order_ptr->getPrice();

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":713
 *                 OrderType.LIMIT,
 *                 <object> cpp_limit_order_ptr.getPrice(),
 *                 <object> cpp_limit_order_ptr.getQuantity(),             # <<<<<<<<<<<<<<
 *                 fees,
 *                 exchange_trade_id=str(int(self._time() * 1e6))
 */
  __pyx_t_12 = __pyx_v_cpp_limit_order_ptr->getQuantity();

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":706
 *         self.c_trigger_event(
 *             self.ORDER_FILLED_EVENT_TAG,
 *             OrderFilledEvent(             # <<<<<<<<<<<<<<
 *                 self._current_timestamp,
 *                 order_id,
 */
  __pyx_t_1 = PyTuple_New(8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4)) __PYX_ERR(0, 706, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_order_id);
  __Pyx_GIVEREF(__pyx_v_order_id);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_order_id)) __PYX_ERR(0, 706, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_trading_pair_str);
  __Pyx_GIVEREF(__pyx_v_trading_pair_str);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_trading_pair_str)) __PYX_ERR(0, 706, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_3)) __PYX_ERR(0, 706, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_11);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_11)) __PYX_ERR(0, 706, __pyx_L1_error);
  __Pyx_INCREF(((PyObject *)__pyx_t_2));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_2));
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 5, ((PyObject *)__pyx_t_2))) __PYX_ERR(0, 706, __pyx_L1_error);
  __Pyx_INCREF(((PyObject *)__pyx_t_12));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_12));
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 6, ((PyObject *)__pyx_t_12))) __PYX_ERR(0, 706, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_fees);
  __Pyx_GIVEREF(__pyx_v_fees);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_v_fees)) __PYX_ERR(0, 706, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  __pyx_t_11 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":715
 *                 <object> cpp_limit_order_ptr.getQuantity(),
 *                 fees,
 *                 exchange_trade_id=str(int(self._time() * 1e6))             # <<<<<<<<<<<<<<
 *             ))
 * 
 */
  __pyx_t_11 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_time); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_13 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_13)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_13);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_13, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 715, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_4 = PyNumber_Multiply(__pyx_t_3, __pyx_float_1e6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Unicode(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_11, __pyx_n_s_exchange_trade_id, __pyx_t_4) < 0) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":706
 *         self.c_trigger_event(
 *             self.ORDER_FILLED_EVENT_TAG,
 *             OrderFilledEvent(             # <<<<<<<<<<<<<<
 *                 self._current_timestamp,
 *                 order_id,
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_1, __pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 706, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":704
 * 
 *         # Emit the trade and order completed events.
 *         self.c_trigger_event(             # <<<<<<<<<<<<<<
 *             self.ORDER_FILLED_EVENT_TAG,
 *             OrderFilledEvent(
 */
  __pyx_t_11 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.c_trigger_event(((struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *)__pyx_v_self), __pyx_t_10, __pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":719
 * 
 *         self.c_trigger_event(
 *             self.BUY_ORDER_COMPLETED_EVENT_TAG,             # <<<<<<<<<<<<<<
 *             BuyOrderCompletedEvent(
 *                 self._current_timestamp,
 */
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_BUY_ORDER_COMPLETED_EVENT_TAG); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_10 = __Pyx_PyInt_As_int64_t(__pyx_t_11); if (unlikely((__pyx_t_10 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 719, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":720
 *         self.c_trigger_event(
 *             self.BUY_ORDER_COMPLETED_EVENT_TAG,
 *             BuyOrderCompletedEvent(             # <<<<<<<<<<<<<<
 *                 self._current_timestamp,
 *                 order_id,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_BuyOrderCompletedEvent); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 720, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":721
 *             self.BUY_ORDER_COMPLETED_EVENT_TAG,
 *             BuyOrderCompletedEvent(
 *                 self._current_timestamp,             # <<<<<<<<<<<<<<
 *                 order_id,
 *                 base_asset,
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 721, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":727
 *                 acquired_amount,
 *                 paid_amount,
 *                 OrderType.LIMIT             # <<<<<<<<<<<<<<
 *             ))
 *         self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_LIMIT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[8] = {__pyx_t_5, __pyx_t_1, __pyx_v_order_id, __pyx_v_base_asset, __pyx_v_quote_asset, __pyx_v_acquired_amount, __pyx_v_paid_amount, __pyx_t_3};
    __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 7+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 720, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":718
 *             ))
 * 
 *         self.c_trigger_event(             # <<<<<<<<<<<<<<
 *             self.BUY_ORDER_COMPLETED_EVENT_TAG,
 *             BuyOrderCompletedEvent(
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.c_trigger_event(((struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *)__pyx_v_self), __pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":729
 *                 OrderType.LIMIT
 *             ))
 *         self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_process_limit_ask_order(self,
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_delete_limit_order(__pyx_v_self, __pyx_v_limit_orders_map_ptr, __pyx_v_map_it_ptr, __pyx_v_orders_it); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 729, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":640
 *             return False
 * 
 *     cdef c_process_limit_bid_order(self,             # <<<<<<<<<<<<<<
 *                                    LimitOrders *limit_orders_map_ptr,
 *                                    LimitOrdersIterator *map_it_ptr,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_process_limit_bid_order", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_trading_pair_str);
  __Pyx_XDECREF(__pyx_v_quote_asset);
  __Pyx_XDECREF(__pyx_v_base_asset);
  __Pyx_XDECREF(__pyx_v_order_id);
  __Pyx_XDECREF(__pyx_v_amount);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XDECREF(__pyx_v_quote_balance);
  __Pyx_XDECREF(__pyx_v_base_balance);
  __Pyx_XDECREF(__pyx_v_order_candidate);
  __Pyx_XDECREF(__pyx_v_adjusted_order_candidate);
  __Pyx_XDECREF(__pyx_v_paid_amount);
  __Pyx_XDECREF(__pyx_v_acquired_amount);
  __Pyx_XDECREF(__pyx_v_fees);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":731
 *         self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 * 
 *     cdef c_process_limit_ask_order(self,             # <<<<<<<<<<<<<<
 *                                    LimitOrders *limit_orders_map_ptr,
 *                                    LimitOrdersIterator *map_it_ptr,
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_process_limit_ask_order(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *__pyx_v_limit_orders_map_ptr, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *__pyx_v_map_it_ptr, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator __pyx_v_orders_it) {
  LimitOrder const *__pyx_v_cpp_limit_order_ptr;
  PyObject *__pyx_v_trading_pair_str = 0;
  PyObject *__pyx_v_quote_asset = 0;
  PyObject *__pyx_v_base_asset = 0;
  PyObject *__pyx_v_order_id = 0;
  PyObject *__pyx_v_amount = 0;
  PyObject *__pyx_v_price = 0;
  PyObject *__pyx_v_quote_balance = 0;
  PyObject *__pyx_v_base_balance = 0;
  PyObject *__pyx_v_order_candidate = NULL;
  PyObject *__pyx_v_adjusted_order_candidate = NULL;
  PyObject *__pyx_v_sold_amount = NULL;
  PyObject *__pyx_v_acquired_amount = NULL;
  PyObject *__pyx_v_fees = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_UCS4 __pyx_t_9;
  int64_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_process_limit_ask_order", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":736
 *                                    SingleTradingPairLimitOrdersIterator orders_it):
 *         cdef:
 *             const CPPLimitOrder *cpp_limit_order_ptr = address(deref(orders_it))             # <<<<<<<<<<<<<<
 *             str trading_pair_str = cpp_limit_order_ptr.getTradingPair().decode("utf8")
 *             str quote_asset = cpp_limit_order_ptr.getQuoteCurrency().decode("utf8")
 */
  __pyx_v_cpp_limit_order_ptr = (&(*__pyx_v_orders_it));

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":737
 *         cdef:
 *             const CPPLimitOrder *cpp_limit_order_ptr = address(deref(orders_it))
 *             str trading_pair_str = cpp_limit_order_ptr.getTradingPair().decode("utf8")             # <<<<<<<<<<<<<<
 *             str quote_asset = cpp_limit_order_ptr.getQuoteCurrency().decode("utf8")
 *             str base_asset = cpp_limit_order_ptr.getBaseCurrency().decode("utf8")
 */
  __pyx_t_1 = __Pyx_decode_cpp_string(__pyx_v_cpp_limit_order_ptr->getTradingPair(), 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 737, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_trading_pair_str = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":738
 *             const CPPLimitOrder *cpp_limit_order_ptr = address(deref(orders_it))
 *             str trading_pair_str = cpp_limit_order_ptr.getTradingPair().decode("utf8")
 *             str quote_asset = cpp_limit_order_ptr.getQuoteCurrency().decode("utf8")             # <<<<<<<<<<<<<<
 *             str base_asset = cpp_limit_order_ptr.getBaseCurrency().decode("utf8")
 *             str order_id = cpp_limit_order_ptr.getClientOrderID().decode("utf8")
 */
  __pyx_t_1 = __Pyx_decode_cpp_string(__pyx_v_cpp_limit_order_ptr->getQuoteCurrency(), 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 738, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_quote_asset = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":739
 *             str trading_pair_str = cpp_limit_order_ptr.getTradingPair().decode("utf8")
 *             str quote_asset = cpp_limit_order_ptr.getQuoteCurrency().decode("utf8")
 *             str base_asset = cpp_limit_order_ptr.getBaseCurrency().decode("utf8")             # <<<<<<<<<<<<<<
 *             str order_id = cpp_limit_order_ptr.getClientOrderID().decode("utf8")
 *             object amount = <object> cpp_limit_order_ptr.getQuantity()
 */
  __pyx_t_1 = __Pyx_decode_cpp_string(__pyx_v_cpp_limit_order_ptr->getBaseCurrency(), 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 739, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_base_asset = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":740
 *             str quote_asset = cpp_limit_order_ptr.getQuoteCurrency().decode("utf8")
 *             str base_asset = cpp_limit_order_ptr.getBaseCurrency().decode("utf8")
 *             str order_id = cpp_limit_order_ptr.getClientOrderID().decode("utf8")             # <<<<<<<<<<<<<<
 *             object amount = <object> cpp_limit_order_ptr.getQuantity()
 *             object price = <object> cpp_limit_order_ptr.getPrice()
 */
  __pyx_t_1 = __Pyx_decode_cpp_string(__pyx_v_cpp_limit_order_ptr->getClientOrderID(), 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 740, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_order_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":741
 *             str base_asset = cpp_limit_order_ptr.getBaseCurrency().decode("utf8")
 *             str order_id = cpp_limit_order_ptr.getClientOrderID().decode("utf8")
 *             object amount = <object> cpp_limit_order_ptr.getQuantity()             # <<<<<<<<<<<<<<
 *             object price = <object> cpp_limit_order_ptr.getPrice()
 *             object quote_balance = self.c_get_balance(quote_asset)
 */
  __pyx_t_2 = __pyx_v_cpp_limit_order_ptr->getQuantity();
  __pyx_t_1 = ((PyObject *)__pyx_t_2);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_amount = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":742
 *             str order_id = cpp_limit_order_ptr.getClientOrderID().decode("utf8")
 *             object amount = <object> cpp_limit_order_ptr.getQuantity()
 *             object price = <object> cpp_limit_order_ptr.getPrice()             # <<<<<<<<<<<<<<
 *             object quote_balance = self.c_get_balance(quote_asset)
 *             object base_balance = self.c_get_balance(base_asset)
 */
  __pyx_t_2 = __pyx_v_cpp_limit_order_ptr->getPrice();
  __pyx_t_1 = ((PyObject *)__pyx_t_2);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_price = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":743
 *             object amount = <object> cpp_limit_order_ptr.getQuantity()
 *             object price = <object> cpp_limit_order_ptr.getPrice()
 *             object quote_balance = self.c_get_balance(quote_asset)             # <<<<<<<<<<<<<<
 *             object base_balance = self.c_get_balance(base_asset)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_get_balance(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), __pyx_v_quote_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 743, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_quote_balance = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":744
 *             object price = <object> cpp_limit_order_ptr.getPrice()
 *             object quote_balance = self.c_get_balance(quote_asset)
 *             object base_balance = self.c_get_balance(base_asset)             # <<<<<<<<<<<<<<
 * 
 *         order_candidate = OrderCandidate(
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_get_balance(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), __pyx_v_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 744, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_base_balance = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":746
 *             object base_balance = self.c_get_balance(base_asset)
 * 
 *         order_candidate = OrderCandidate(             # <<<<<<<<<<<<<<
 *             trading_pair=trading_pair_str,
 *             # Market orders are not maker orders
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_OrderCandidate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":747
 * 
 *         order_candidate = OrderCandidate(
 *             trading_pair=trading_pair_str,             # <<<<<<<<<<<<<<
 *             # Market orders are not maker orders
 *             is_maker=True,
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 747, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_trading_pair, __pyx_v_trading_pair_str) < 0) __PYX_ERR(0, 747, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":749
 *             trading_pair=trading_pair_str,
 *             # Market orders are not maker orders
 *             is_maker=True,             # <<<<<<<<<<<<<<
 *             order_type=OrderType.LIMIT,
 *             order_side=TradeType.SELL,
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_is_maker, Py_True) < 0) __PYX_ERR(0, 747, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":750
 *             # Market orders are not maker orders
 *             is_maker=True,
 *             order_type=OrderType.LIMIT,             # <<<<<<<<<<<<<<
 *             order_side=TradeType.SELL,
 *             amount=amount,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_LIMIT); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_order_type, __pyx_t_5) < 0) __PYX_ERR(0, 747, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":751
 *             is_maker=True,
 *             order_type=OrderType.LIMIT,
 *             order_side=TradeType.SELL,             # <<<<<<<<<<<<<<
 *             amount=amount,
 *             price=price,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 751, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_SELL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 751, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_order_side, __pyx_t_4) < 0) __PYX_ERR(0, 747, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":752
 *             order_type=OrderType.LIMIT,
 *             order_side=TradeType.SELL,
 *             amount=amount,             # <<<<<<<<<<<<<<
 *             price=price,
 *             from_total_balances=True
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_amount, __pyx_v_amount) < 0) __PYX_ERR(0, 747, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":753
 *             order_side=TradeType.SELL,
 *             amount=amount,
 *             price=price,             # <<<<<<<<<<<<<<
 *             from_total_balances=True
 *         )
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_price, __pyx_v_price) < 0) __PYX_ERR(0, 747, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":754
 *             amount=amount,
 *             price=price,
 *             from_total_balances=True             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_from_total_balances, Py_True) < 0) __PYX_ERR(0, 747, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":746
 *             object base_balance = self.c_get_balance(base_asset)
 * 
 *         order_candidate = OrderCandidate(             # <<<<<<<<<<<<<<
 *             trading_pair=trading_pair_str,
 *             # Market orders are not maker orders
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 746, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_order_candidate = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":757
 *         )
 * 
 *         adjusted_order_candidate = self._budget_checker.populate_collateral_entries(order_candidate)             # <<<<<<<<<<<<<<
 * 
 *         # Base currency used, including fees.
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base._budget_checker, __pyx_n_s_populate_collateral_entries); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 757, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_order_candidate};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_adjusted_order_candidate = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":760
 * 
 *         # Base currency used, including fees.
 *         sold_amount = adjusted_order_candidate.order_collateral.amount             # <<<<<<<<<<<<<<
 *         # Quote currency acquired, including fees.
 *         acquired_amount = adjusted_order_candidate.potential_returns.amount
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_adjusted_order_candidate, __pyx_n_s_order_collateral); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 760, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_amount); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 760, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_sold_amount = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":762
 *         sold_amount = adjusted_order_candidate.order_collateral.amount
 *         # Quote currency acquired, including fees.
 *         acquired_amount = adjusted_order_candidate.potential_returns.amount             # <<<<<<<<<<<<<<
 * 
 *         # It's not possible to fulfill the order, the possible sold amount is less than requested
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_adjusted_order_candidate, __pyx_n_s_potential_returns); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_amount); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 762, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_acquired_amount = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":765
 * 
 *         # It's not possible to fulfill the order, the possible sold amount is less than requested
 *         if sold_amount > base_balance:             # <<<<<<<<<<<<<<
 *             self.logger().warning(f"Not enough {base_asset} balance to fill limit sell order on {trading_pair_str}. "
 *                                   f"{sold_amount:.8g} {base_asset} needed vs. "
 */
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_sold_amount, __pyx_v_base_balance, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 765, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 765, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_7) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":766
 *         # It's not possible to fulfill the order, the possible sold amount is less than requested
 *         if sold_amount > base_balance:
 *             self.logger().warning(f"Not enough {base_asset} balance to fill limit sell order on {trading_pair_str}. "             # <<<<<<<<<<<<<<
 *                                   f"{sold_amount:.8g} {base_asset} needed vs. "
 *                                   f"{base_balance:.8g} {base_asset} available.")
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_warning); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(13); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = 0;
    __pyx_t_9 = 127;
    __Pyx_INCREF(__pyx_kp_u_Not_enough);
    __pyx_t_8 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_Not_enough);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_Not_enough);
    __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_v_base_asset); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_9;
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u_balance_to_fill_limit_sell_orde);
    __pyx_t_8 += 37;
    __Pyx_GIVEREF(__pyx_kp_u_balance_to_fill_limit_sell_orde);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u_balance_to_fill_limit_sell_orde);
    __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_v_trading_pair_str); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_9;
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u__22);
    __pyx_t_8 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__22);
    PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u__22);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":767
 *         if sold_amount > base_balance:
 *             self.logger().warning(f"Not enough {base_asset} balance to fill limit sell order on {trading_pair_str}. "
 *                                   f"{sold_amount:.8g} {base_asset} needed vs. "             # <<<<<<<<<<<<<<
 *                                   f"{base_balance:.8g} {base_asset} available.")
 *             self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 */
    __pyx_t_5 = __Pyx_PyObject_Format(__pyx_v_sold_amount, __pyx_kp_u_8g); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 767, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_9;
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u__20);
    __pyx_t_8 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__20);
    PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u__20);
    __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_v_base_asset); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 767, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_9;
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u_needed_vs);
    __pyx_t_8 += 12;
    __Pyx_GIVEREF(__pyx_kp_u_needed_vs);
    PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_kp_u_needed_vs);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":768
 *             self.logger().warning(f"Not enough {base_asset} balance to fill limit sell order on {trading_pair_str}. "
 *                                   f"{sold_amount:.8g} {base_asset} needed vs. "
 *                                   f"{base_balance:.8g} {base_asset} available.")             # <<<<<<<<<<<<<<
 *             self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 *             self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,
 */
    __pyx_t_5 = __Pyx_PyObject_Format(__pyx_v_base_balance, __pyx_kp_u_8g); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 768, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_9;
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 9, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u__20);
    __pyx_t_8 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__20);
    PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_kp_u__20);
    __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_v_base_asset); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 768, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_9) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_9;
    __pyx_t_8 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 11, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u_available);
    __pyx_t_8 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_available);
    PyTuple_SET_ITEM(__pyx_t_3, 12, __pyx_kp_u_available);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":766
 *         # It's not possible to fulfill the order, the possible sold amount is less than requested
 *         if sold_amount > base_balance:
 *             self.logger().warning(f"Not enough {base_asset} balance to fill limit sell order on {trading_pair_str}. "             # <<<<<<<<<<<<<<
 *                                   f"{sold_amount:.8g} {base_asset} needed vs. "
 *                                   f"{base_balance:.8g} {base_asset} available.")
 */
    __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_3, 13, __pyx_t_8, __pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":769
 *                                   f"{sold_amount:.8g} {base_asset} needed vs. "
 *                                   f"{base_balance:.8g} {base_asset} available.")
 *             self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)             # <<<<<<<<<<<<<<
 *             self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,
 *                                  OrderCancelledEvent(self._current_timestamp,
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_delete_limit_order(__pyx_v_self, __pyx_v_limit_orders_map_ptr, __pyx_v_map_it_ptr, __pyx_v_orders_it); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 769, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":770
 *                                   f"{base_balance:.8g} {base_asset} available.")
 *             self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 *             self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,             # <<<<<<<<<<<<<<
 *                                  OrderCancelledEvent(self._current_timestamp,
 *                                                      order_id)
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_MARKET_ORDER_CANCELED_EVENT_TAG); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 770, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_10 = __Pyx_PyInt_As_int64_t(__pyx_t_4); if (unlikely((__pyx_t_10 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 770, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":771
 *             self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 *             self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,
 *                                  OrderCancelledEvent(self._current_timestamp,             # <<<<<<<<<<<<<<
 *                                                      order_id)
 *                                  )
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_OrderCancelledEvent); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 771, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 771, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":772
 *             self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,
 *                                  OrderCancelledEvent(self._current_timestamp,
 *                                                      order_id)             # <<<<<<<<<<<<<<
 *                                  )
 *             return
 */
    __pyx_t_3 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_t_5, __pyx_v_order_id};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 771, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":770
 *                                   f"{base_balance:.8g} {base_asset} available.")
 *             self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 *             self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,             # <<<<<<<<<<<<<<
 *                                  OrderCancelledEvent(self._current_timestamp,
 *                                                      order_id)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.c_trigger_event(((struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *)__pyx_v_self), __pyx_t_10, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 770, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":774
 *                                                      order_id)
 *                                  )
 *             return             # <<<<<<<<<<<<<<
 * 
 *         # The order was successfully executed
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":765
 * 
 *         # It's not possible to fulfill the order, the possible sold amount is less than requested
 *         if sold_amount > base_balance:             # <<<<<<<<<<<<<<
 *             self.logger().warning(f"Not enough {base_asset} balance to fill limit sell order on {trading_pair_str}. "
 *                                   f"{sold_amount:.8g} {base_asset} needed vs. "
 */
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":778
 *         # The order was successfully executed
 *         self.c_set_balance(quote_asset,
 *                            quote_balance + acquired_amount)             # <<<<<<<<<<<<<<
 *         self.c_set_balance(base_asset,
 *                            base_balance - sold_amount)
 */
  __pyx_t_1 = PyNumber_Add(__pyx_v_quote_balance, __pyx_v_acquired_amount); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 778, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":777
 * 
 *         # The order was successfully executed
 *         self.c_set_balance(quote_asset,             # <<<<<<<<<<<<<<
 *                            quote_balance + acquired_amount)
 *         self.c_set_balance(base_asset,
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_set_balance(__pyx_v_self, __pyx_v_quote_asset, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":780
 *                            quote_balance + acquired_amount)
 *         self.c_set_balance(base_asset,
 *                            base_balance - sold_amount)             # <<<<<<<<<<<<<<
 * 
 *         # add fee
 */
  __pyx_t_4 = PyNumber_Subtract(__pyx_v_base_balance, __pyx_v_sold_amount); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":779
 *         self.c_set_balance(quote_asset,
 *                            quote_balance + acquired_amount)
 *         self.c_set_balance(base_asset,             # <<<<<<<<<<<<<<
 *                            base_balance - sold_amount)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_set_balance(__pyx_v_self, __pyx_v_base_asset, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":783
 * 
 *         # add fee
 *         fees = build_trade_fee(             # <<<<<<<<<<<<<<
 *             exchange=self.name,
 *             is_maker=True,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_build_trade_fee); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":784
 *         # add fee
 *         fees = build_trade_fee(
 *             exchange=self.name,             # <<<<<<<<<<<<<<
 *             is_maker=True,
 *             base_currency="",
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_name); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_exchange, __pyx_t_5) < 0) __PYX_ERR(0, 784, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":785
 *         fees = build_trade_fee(
 *             exchange=self.name,
 *             is_maker=True,             # <<<<<<<<<<<<<<
 *             base_currency="",
 *             quote_currency="",
 */
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_is_maker, Py_True) < 0) __PYX_ERR(0, 784, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_base_currency, __pyx_kp_u__11) < 0) __PYX_ERR(0, 784, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_quote_currency, __pyx_kp_u__11) < 0) __PYX_ERR(0, 784, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":788
 *             base_currency="",
 *             quote_currency="",
 *             order_type=OrderType.LIMIT,             # <<<<<<<<<<<<<<
 *             order_side=TradeType.SELL,
 *             amount=Decimal("0"),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_LIMIT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 788, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_order_type, __pyx_t_3) < 0) __PYX_ERR(0, 784, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":789
 *             quote_currency="",
 *             order_type=OrderType.LIMIT,
 *             order_side=TradeType.SELL,             # <<<<<<<<<<<<<<
 *             amount=Decimal("0"),
 *             price=Decimal("0"),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_SELL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 789, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_order_side, __pyx_t_5) < 0) __PYX_ERR(0, 784, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":790
 *             order_type=OrderType.LIMIT,
 *             order_side=TradeType.SELL,
 *             amount=Decimal("0"),             # <<<<<<<<<<<<<<
 *             price=Decimal("0"),
 *         )
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_11 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_11)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_11);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_kp_u_0};
    __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 790, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_amount, __pyx_t_5) < 0) __PYX_ERR(0, 784, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":791
 *             order_side=TradeType.SELL,
 *             amount=Decimal("0"),
 *             price=Decimal("0"),             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 791, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_11 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_11)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_11);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_kp_u_0};
    __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 791, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_price, __pyx_t_5) < 0) __PYX_ERR(0, 784, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":783
 * 
 *         # add fee
 *         fees = build_trade_fee(             # <<<<<<<<<<<<<<
 *             exchange=self.name,
 *             is_maker=True,
 */
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_v_fees = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":796
 *         # Emit the trade and order completed events.
 *         self.c_trigger_event(
 *             self.ORDER_FILLED_EVENT_TAG,             # <<<<<<<<<<<<<<
 *             OrderFilledEvent(
 *                 self._current_timestamp,
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_ORDER_FILLED_EVENT_TAG); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_10 = __Pyx_PyInt_As_int64_t(__pyx_t_5); if (unlikely((__pyx_t_10 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 796, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":797
 *         self.c_trigger_event(
 *             self.ORDER_FILLED_EVENT_TAG,
 *             OrderFilledEvent(             # <<<<<<<<<<<<<<
 *                 self._current_timestamp,
 *                 order_id,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_OrderFilledEvent); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":798
 *             self.ORDER_FILLED_EVENT_TAG,
 *             OrderFilledEvent(
 *                 self._current_timestamp,             # <<<<<<<<<<<<<<
 *                 order_id,
 *                 trading_pair_str,
 */
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 798, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":801
 *                 order_id,
 *                 trading_pair_str,
 *                 TradeType.SELL,             # <<<<<<<<<<<<<<
 *                 OrderType.LIMIT,
 *                 <object> cpp_limit_order_ptr.getPrice(),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_SELL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":802
 *                 trading_pair_str,
 *                 TradeType.SELL,
 *                 OrderType.LIMIT,             # <<<<<<<<<<<<<<
 *                 <object> cpp_limit_order_ptr.getPrice(),
 *                 <object> cpp_limit_order_ptr.getQuantity(),
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LIMIT); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 802, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":803
 *                 TradeType.SELL,
 *                 OrderType.LIMIT,
 *                 <object> cpp_limit_order_ptr.getPrice(),             # <<<<<<<<<<<<<<
 *                 <object> cpp_limit_order_ptr.getQuantity(),
 *                 fees,
 */
  __pyx_t_2 = __pyx_v_cpp_limit_order_ptr->getPrice();

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":804
 *                 OrderType.LIMIT,
 *                 <object> cpp_limit_order_ptr.getPrice(),
 *                 <object> cpp_limit_order_ptr.getQuantity(),             # <<<<<<<<<<<<<<
 *                 fees,
 *                 exchange_trade_id=str(int(self._time() * 1e6))
 */
  __pyx_t_12 = __pyx_v_cpp_limit_order_ptr->getQuantity();

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":797
 *         self.c_trigger_event(
 *             self.ORDER_FILLED_EVENT_TAG,
 *             OrderFilledEvent(             # <<<<<<<<<<<<<<
 *                 self._current_timestamp,
 *                 order_id,
 */
  __pyx_t_1 = PyTuple_New(8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_4)) __PYX_ERR(0, 797, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_order_id);
  __Pyx_GIVEREF(__pyx_v_order_id);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_order_id)) __PYX_ERR(0, 797, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_trading_pair_str);
  __Pyx_GIVEREF(__pyx_v_trading_pair_str);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_trading_pair_str)) __PYX_ERR(0, 797, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_3)) __PYX_ERR(0, 797, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_11);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_t_11)) __PYX_ERR(0, 797, __pyx_L1_error);
  __Pyx_INCREF(((PyObject *)__pyx_t_2));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_2));
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 5, ((PyObject *)__pyx_t_2))) __PYX_ERR(0, 797, __pyx_L1_error);
  __Pyx_INCREF(((PyObject *)__pyx_t_12));
  __Pyx_GIVEREF(((PyObject *)__pyx_t_12));
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 6, ((PyObject *)__pyx_t_12))) __PYX_ERR(0, 797, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_fees);
  __Pyx_GIVEREF(__pyx_v_fees);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 7, __pyx_v_fees)) __PYX_ERR(0, 797, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  __pyx_t_11 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":806
 *                 <object> cpp_limit_order_ptr.getQuantity(),
 *                 fees,
 *                 exchange_trade_id=str(int(self._time() * 1e6))             # <<<<<<<<<<<<<<
 *             ))
 * 
 */
  __pyx_t_11 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_time); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_13 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_13)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_13);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_13, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 806, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_4 = PyNumber_Multiply(__pyx_t_3, __pyx_float_1e6); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyNumber_Int(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Unicode(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_t_11, __pyx_n_s_exchange_trade_id, __pyx_t_4) < 0) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":797
 *         self.c_trigger_event(
 *             self.ORDER_FILLED_EVENT_TAG,
 *             OrderFilledEvent(             # <<<<<<<<<<<<<<
 *                 self._current_timestamp,
 *                 order_id,
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_1, __pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":795
 * 
 *         # Emit the trade and order completed events.
 *         self.c_trigger_event(             # <<<<<<<<<<<<<<
 *             self.ORDER_FILLED_EVENT_TAG,
 *             OrderFilledEvent(
 */
  __pyx_t_11 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.c_trigger_event(((struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *)__pyx_v_self), __pyx_t_10, __pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":810
 * 
 *         self.c_trigger_event(
 *             self.SELL_ORDER_COMPLETED_EVENT_TAG,             # <<<<<<<<<<<<<<
 *             SellOrderCompletedEvent(
 *                 self._current_timestamp,
 */
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_SELL_ORDER_COMPLETED_EVENT_TAG); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_10 = __Pyx_PyInt_As_int64_t(__pyx_t_11); if (unlikely((__pyx_t_10 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 810, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":811
 *         self.c_trigger_event(
 *             self.SELL_ORDER_COMPLETED_EVENT_TAG,
 *             SellOrderCompletedEvent(             # <<<<<<<<<<<<<<
 *                 self._current_timestamp,
 *                 order_id,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_SellOrderCompletedEvent); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":812
 *             self.SELL_ORDER_COMPLETED_EVENT_TAG,
 *             SellOrderCompletedEvent(
 *                 self._current_timestamp,             # <<<<<<<<<<<<<<
 *                 order_id,
 *                 base_asset,
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":818
 *                 sold_amount,
 *                 acquired_amount,
 *                 OrderType.LIMIT             # <<<<<<<<<<<<<<
 *             ))
 *         self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 818, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_LIMIT); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 818, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[8] = {__pyx_t_5, __pyx_t_1, __pyx_v_order_id, __pyx_v_base_asset, __pyx_v_quote_asset, __pyx_v_sold_amount, __pyx_v_acquired_amount, __pyx_t_3};
    __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 7+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 811, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":809
 *             ))
 * 
 *         self.c_trigger_event(             # <<<<<<<<<<<<<<
 *             self.SELL_ORDER_COMPLETED_EVENT_TAG,
 *             SellOrderCompletedEvent(
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.c_trigger_event(((struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *)__pyx_v_self), __pyx_t_10, __pyx_t_11); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 809, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":820
 *                 OrderType.LIMIT
 *             ))
 *         self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_process_limit_order(self,
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_delete_limit_order(__pyx_v_self, __pyx_v_limit_orders_map_ptr, __pyx_v_map_it_ptr, __pyx_v_orders_it); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":731
 *         self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 * 
 *     cdef c_process_limit_ask_order(self,             # <<<<<<<<<<<<<<
 *                                    LimitOrders *limit_orders_map_ptr,
 *                                    LimitOrdersIterator *map_it_ptr,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_process_limit_ask_order", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_trading_pair_str);
  __Pyx_XDECREF(__pyx_v_quote_asset);
  __Pyx_XDECREF(__pyx_v_base_asset);
  __Pyx_XDECREF(__pyx_v_order_id);
  __Pyx_XDECREF(__pyx_v_amount);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XDECREF(__pyx_v_quote_balance);
  __Pyx_XDECREF(__pyx_v_base_balance);
  __Pyx_XDECREF(__pyx_v_order_candidate);
  __Pyx_XDECREF(__pyx_v_adjusted_order_candidate);
  __Pyx_XDECREF(__pyx_v_sold_amount);
  __Pyx_XDECREF(__pyx_v_acquired_amount);
  __Pyx_XDECREF(__pyx_v_fees);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":822
 *         self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 * 
 *     cdef c_process_limit_order(self,             # <<<<<<<<<<<<<<
 *                                bint is_buy,
 *                                LimitOrders *limit_orders_map_ptr,
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_process_limit_order(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, int __pyx_v_is_buy, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *__pyx_v_limit_orders_map_ptr, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *__pyx_v_map_it_ptr, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator __pyx_v_orders_it) {
  CYTHON_UNUSED PyObject *__pyx_v_e = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  unsigned int __pyx_t_11;
  int __pyx_t_12;
  char const *__pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_process_limit_order", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":827
 *                                LimitOrdersIterator *map_it_ptr,
 *                                SingleTradingPairLimitOrdersIterator orders_it):
 *         try:             # <<<<<<<<<<<<<<
 *             if is_buy:
 *                 self.c_process_limit_bid_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":828
 *                                SingleTradingPairLimitOrdersIterator orders_it):
 *         try:
 *             if is_buy:             # <<<<<<<<<<<<<<
 *                 self.c_process_limit_bid_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 *             else:
 */
      if (__pyx_v_is_buy) {

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":829
 *         try:
 *             if is_buy:
 *                 self.c_process_limit_bid_order(limit_orders_map_ptr, map_it_ptr, orders_it)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.c_process_limit_ask_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 */
        __pyx_t_4 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_process_limit_bid_order(__pyx_v_self, __pyx_v_limit_orders_map_ptr, __pyx_v_map_it_ptr, __pyx_v_orders_it); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 829, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":828
 *                                SingleTradingPairLimitOrdersIterator orders_it):
 *         try:
 *             if is_buy:             # <<<<<<<<<<<<<<
 *                 self.c_process_limit_bid_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 *             else:
 */
        goto __pyx_L9;
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":831
 *                 self.c_process_limit_bid_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 *             else:
 *                 self.c_process_limit_ask_order(limit_orders_map_ptr, map_it_ptr, orders_it)             # <<<<<<<<<<<<<<
 *         except Exception as e:
 *             self.logger().error(f"Error processing limit order.", exc_info=True)
 */
      /*else*/ {
        __pyx_t_4 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_process_limit_ask_order(__pyx_v_self, __pyx_v_limit_orders_map_ptr, __pyx_v_map_it_ptr, __pyx_v_orders_it); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 831, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __pyx_L9:;

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":827
 *                                LimitOrdersIterator *map_it_ptr,
 *                                SingleTradingPairLimitOrdersIterator orders_it):
 *         try:             # <<<<<<<<<<<<<<
 *             if is_buy:
 *                 self.c_process_limit_bid_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":832
 *             else:
 *                 self.c_process_limit_ask_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 *         except Exception as e:             # <<<<<<<<<<<<<<
 *             self.logger().error(f"Error processing limit order.", exc_info=True)
 * 
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_5) {
      __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_process_limit_order", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_4, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(0, 832, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_t_6);
      __pyx_v_e = __pyx_t_6;
      /*try:*/ {

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":833
 *                 self.c_process_limit_ask_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 *         except Exception as e:
 *             self.logger().error(f"Error processing limit order.", exc_info=True)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_process_crossed_limit_orders_for_trading_pair(self,
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 833, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = NULL;
        __pyx_t_11 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
            __pyx_t_11 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_10, NULL};
          __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 0+__pyx_t_11);
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 833, __pyx_L15_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        }
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_error); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 833, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 833, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_exc_info, Py_True) < 0) __PYX_ERR(0, 833, __pyx_L15_error)
        __pyx_t_10 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_tuple__23, __pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 833, __pyx_L15_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":832
 *             else:
 *                 self.c_process_limit_ask_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 *         except Exception as e:             # <<<<<<<<<<<<<<
 *             self.logger().error(f"Error processing limit order.", exc_info=True)
 * 
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
          goto __pyx_L16;
        }
        __pyx_L15_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_17, &__pyx_t_18, &__pyx_t_19);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16) < 0)) __Pyx_ErrFetch(&__pyx_t_14, &__pyx_t_15, &__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_14);
          __Pyx_XGOTREF(__pyx_t_15);
          __Pyx_XGOTREF(__pyx_t_16);
          __Pyx_XGOTREF(__pyx_t_17);
          __Pyx_XGOTREF(__pyx_t_18);
          __Pyx_XGOTREF(__pyx_t_19);
          __pyx_t_5 = __pyx_lineno; __pyx_t_12 = __pyx_clineno; __pyx_t_13 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_e); __pyx_v_e = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_17);
            __Pyx_XGIVEREF(__pyx_t_18);
            __Pyx_XGIVEREF(__pyx_t_19);
            __Pyx_ExceptionReset(__pyx_t_17, __pyx_t_18, __pyx_t_19);
          }
          __Pyx_XGIVEREF(__pyx_t_14);
          __Pyx_XGIVEREF(__pyx_t_15);
          __Pyx_XGIVEREF(__pyx_t_16);
          __Pyx_ErrRestore(__pyx_t_14, __pyx_t_15, __pyx_t_16);
          __pyx_t_14 = 0; __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0;
          __pyx_lineno = __pyx_t_5; __pyx_clineno = __pyx_t_12; __pyx_filename = __pyx_t_13;
          goto __pyx_L5_except_error;
        }
        __pyx_L16:;
      }
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":827
 *                                LimitOrdersIterator *map_it_ptr,
 *                                SingleTradingPairLimitOrdersIterator orders_it):
 *         try:             # <<<<<<<<<<<<<<
 *             if is_buy:
 *                 self.c_process_limit_bid_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":822
 *         self.c_delete_limit_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 * 
 *     cdef c_process_limit_order(self,             # <<<<<<<<<<<<<<
 *                                bint is_buy,
 *                                LimitOrders *limit_orders_map_ptr,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_process_limit_order", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_e);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":835
 *             self.logger().error(f"Error processing limit order.", exc_info=True)
 * 
 *     cdef c_process_crossed_limit_orders_for_trading_pair(self,             # <<<<<<<<<<<<<<
 *                                                          bint is_buy,
 *                                                          LimitOrders *limit_orders_map_ptr,
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_process_crossed_limit_orders_for_trading_pair(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, int __pyx_v_is_buy, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *__pyx_v_limit_orders_map_ptr, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *__pyx_v_map_it_ptr) {
  PyObject *__pyx_v_trading_pair = 0;
  PyObject *__pyx_v_opposite_order_book_price = 0;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders *__pyx_v_orders_collection_ptr;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator __pyx_v_orders_it;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersRIterator __pyx_v_orders_rit;
  std::vector<__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator>  __pyx_v_process_order_its;
  LimitOrder const *__pyx_v_cpp_limit_order_ptr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3;
  std::vector<__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator> ::iterator __pyx_t_4;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_process_crossed_limit_orders_for_trading_pair", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":848
 *         """
 *         cdef:
 *             str trading_pair = deref(deref(map_it_ptr)).first.decode("utf8")             # <<<<<<<<<<<<<<
 *             object opposite_order_book_price = self.c_get_price(trading_pair, is_buy)
 *             SingleTradingPairLimitOrders *orders_collection_ptr = address(deref(deref(map_it_ptr)).second)
 */
  __pyx_t_1 = __Pyx_decode_cpp_string((*(*__pyx_v_map_it_ptr)).first, 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 848, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_trading_pair = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":849
 *         cdef:
 *             str trading_pair = deref(deref(map_it_ptr)).first.decode("utf8")
 *             object opposite_order_book_price = self.c_get_price(trading_pair, is_buy)             # <<<<<<<<<<<<<<
 *             SingleTradingPairLimitOrders *orders_collection_ptr = address(deref(deref(map_it_ptr)).second)
 *             SingleTradingPairLimitOrdersIterator orders_it = orders_collection_ptr.begin()
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_get_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), __pyx_v_trading_pair, __pyx_v_is_buy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_opposite_order_book_price = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":850
 *             str trading_pair = deref(deref(map_it_ptr)).first.decode("utf8")
 *             object opposite_order_book_price = self.c_get_price(trading_pair, is_buy)
 *             SingleTradingPairLimitOrders *orders_collection_ptr = address(deref(deref(map_it_ptr)).second)             # <<<<<<<<<<<<<<
 *             SingleTradingPairLimitOrdersIterator orders_it = orders_collection_ptr.begin()
 *             SingleTradingPairLimitOrdersRIterator orders_rit = orders_collection_ptr.rbegin()
 */
  __pyx_v_orders_collection_ptr = (&(*(*__pyx_v_map_it_ptr)).second);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":851
 *             object opposite_order_book_price = self.c_get_price(trading_pair, is_buy)
 *             SingleTradingPairLimitOrders *orders_collection_ptr = address(deref(deref(map_it_ptr)).second)
 *             SingleTradingPairLimitOrdersIterator orders_it = orders_collection_ptr.begin()             # <<<<<<<<<<<<<<
 *             SingleTradingPairLimitOrdersRIterator orders_rit = orders_collection_ptr.rbegin()
 *             vector[SingleTradingPairLimitOrdersIterator] process_order_its
 */
  __pyx_v_orders_it = __pyx_v_orders_collection_ptr->begin();

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":852
 *             SingleTradingPairLimitOrders *orders_collection_ptr = address(deref(deref(map_it_ptr)).second)
 *             SingleTradingPairLimitOrdersIterator orders_it = orders_collection_ptr.begin()
 *             SingleTradingPairLimitOrdersRIterator orders_rit = orders_collection_ptr.rbegin()             # <<<<<<<<<<<<<<
 *             vector[SingleTradingPairLimitOrdersIterator] process_order_its
 *             const CPPLimitOrder *cpp_limit_order_ptr = NULL
 */
  __pyx_v_orders_rit = __pyx_v_orders_collection_ptr->rbegin();

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":854
 *             SingleTradingPairLimitOrdersRIterator orders_rit = orders_collection_ptr.rbegin()
 *             vector[SingleTradingPairLimitOrdersIterator] process_order_its
 *             const CPPLimitOrder *cpp_limit_order_ptr = NULL             # <<<<<<<<<<<<<<
 * 
 *         if is_buy:
 */
  __pyx_v_cpp_limit_order_ptr = NULL;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":856
 *             const CPPLimitOrder *cpp_limit_order_ptr = NULL
 * 
 *         if is_buy:             # <<<<<<<<<<<<<<
 *             while orders_rit != orders_collection_ptr.rend():
 *                 cpp_limit_order_ptr = address(deref(orders_rit))
 */
  if (__pyx_v_is_buy) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":857
 * 
 *         if is_buy:
 *             while orders_rit != orders_collection_ptr.rend():             # <<<<<<<<<<<<<<
 *                 cpp_limit_order_ptr = address(deref(orders_rit))
 *                 if opposite_order_book_price > <object>cpp_limit_order_ptr.getPrice():
 */
    while (1) {
      __pyx_t_2 = (__pyx_v_orders_rit != __pyx_v_orders_collection_ptr->rend());
      if (!__pyx_t_2) break;

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":858
 *         if is_buy:
 *             while orders_rit != orders_collection_ptr.rend():
 *                 cpp_limit_order_ptr = address(deref(orders_rit))             # <<<<<<<<<<<<<<
 *                 if opposite_order_book_price > <object>cpp_limit_order_ptr.getPrice():
 *                     break
 */
      __pyx_v_cpp_limit_order_ptr = (&(*__pyx_v_orders_rit));

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":859
 *             while orders_rit != orders_collection_ptr.rend():
 *                 cpp_limit_order_ptr = address(deref(orders_rit))
 *                 if opposite_order_book_price > <object>cpp_limit_order_ptr.getPrice():             # <<<<<<<<<<<<<<
 *                     break
 *                 process_order_its.push_back(getIteratorFromReverseIterator(
 */
      __pyx_t_3 = __pyx_v_cpp_limit_order_ptr->getPrice();
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_opposite_order_book_price, ((PyObject *)__pyx_t_3), Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 859, __pyx_L1_error)
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 859, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_2) {

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":860
 *                 cpp_limit_order_ptr = address(deref(orders_rit))
 *                 if opposite_order_book_price > <object>cpp_limit_order_ptr.getPrice():
 *                     break             # <<<<<<<<<<<<<<
 *                 process_order_its.push_back(getIteratorFromReverseIterator(
 *                     <reverse_iterator[SingleTradingPairLimitOrdersIterator]>orders_rit))
 */
        goto __pyx_L5_break;

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":859
 *             while orders_rit != orders_collection_ptr.rend():
 *                 cpp_limit_order_ptr = address(deref(orders_rit))
 *                 if opposite_order_book_price > <object>cpp_limit_order_ptr.getPrice():             # <<<<<<<<<<<<<<
 *                     break
 *                 process_order_its.push_back(getIteratorFromReverseIterator(
 */
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":861
 *                 if opposite_order_book_price > <object>cpp_limit_order_ptr.getPrice():
 *                     break
 *                 process_order_its.push_back(getIteratorFromReverseIterator(             # <<<<<<<<<<<<<<
 *                     <reverse_iterator[SingleTradingPairLimitOrdersIterator]>orders_rit))
 *                 inc(orders_rit)
 */
      try {
        __pyx_v_process_order_its.push_back(getIteratorFromReverseIterator<__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator>(((std::reverse_iterator<__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator> )__pyx_v_orders_rit)));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 861, __pyx_L1_error)
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":863
 *                 process_order_its.push_back(getIteratorFromReverseIterator(
 *                     <reverse_iterator[SingleTradingPairLimitOrdersIterator]>orders_rit))
 *                 inc(orders_rit)             # <<<<<<<<<<<<<<
 *         else:
 *             while orders_it != orders_collection_ptr.end():
 */
      (void)((__pyx_v_orders_rit++));
    }
    __pyx_L5_break:;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":856
 *             const CPPLimitOrder *cpp_limit_order_ptr = NULL
 * 
 *         if is_buy:             # <<<<<<<<<<<<<<
 *             while orders_rit != orders_collection_ptr.rend():
 *                 cpp_limit_order_ptr = address(deref(orders_rit))
 */
    goto __pyx_L3;
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":865
 *                 inc(orders_rit)
 *         else:
 *             while orders_it != orders_collection_ptr.end():             # <<<<<<<<<<<<<<
 *                 cpp_limit_order_ptr = address(deref(orders_it))
 *                 if opposite_order_book_price < <object>cpp_limit_order_ptr.getPrice():
 */
  /*else*/ {
    while (1) {
      __pyx_t_2 = (__pyx_v_orders_it != __pyx_v_orders_collection_ptr->end());
      if (!__pyx_t_2) break;

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":866
 *         else:
 *             while orders_it != orders_collection_ptr.end():
 *                 cpp_limit_order_ptr = address(deref(orders_it))             # <<<<<<<<<<<<<<
 *                 if opposite_order_book_price < <object>cpp_limit_order_ptr.getPrice():
 *                     break
 */
      __pyx_v_cpp_limit_order_ptr = (&(*__pyx_v_orders_it));

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":867
 *             while orders_it != orders_collection_ptr.end():
 *                 cpp_limit_order_ptr = address(deref(orders_it))
 *                 if opposite_order_book_price < <object>cpp_limit_order_ptr.getPrice():             # <<<<<<<<<<<<<<
 *                     break
 *                 process_order_its.push_back(orders_it)
 */
      __pyx_t_3 = __pyx_v_cpp_limit_order_ptr->getPrice();
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_opposite_order_book_price, ((PyObject *)__pyx_t_3), Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 867, __pyx_L1_error)
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 867, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_2) {

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":868
 *                 cpp_limit_order_ptr = address(deref(orders_it))
 *                 if opposite_order_book_price < <object>cpp_limit_order_ptr.getPrice():
 *                     break             # <<<<<<<<<<<<<<
 *                 process_order_its.push_back(orders_it)
 *                 inc(orders_it)
 */
        goto __pyx_L8_break;

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":867
 *             while orders_it != orders_collection_ptr.end():
 *                 cpp_limit_order_ptr = address(deref(orders_it))
 *                 if opposite_order_book_price < <object>cpp_limit_order_ptr.getPrice():             # <<<<<<<<<<<<<<
 *                     break
 *                 process_order_its.push_back(orders_it)
 */
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":869
 *                 if opposite_order_book_price < <object>cpp_limit_order_ptr.getPrice():
 *                     break
 *                 process_order_its.push_back(orders_it)             # <<<<<<<<<<<<<<
 *                 inc(orders_it)
 * 
 */
      try {
        __pyx_v_process_order_its.push_back(__pyx_v_orders_it);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 869, __pyx_L1_error)
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":870
 *                     break
 *                 process_order_its.push_back(orders_it)
 *                 inc(orders_it)             # <<<<<<<<<<<<<<
 * 
 *         for orders_it in process_order_its:
 */
      (void)((__pyx_v_orders_it++));
    }
    __pyx_L8_break:;
  }
  __pyx_L3:;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":872
 *                 inc(orders_it)
 * 
 *         for orders_it in process_order_its:             # <<<<<<<<<<<<<<
 *             self.c_process_limit_order(is_buy, limit_orders_map_ptr, map_it_ptr, orders_it)
 * 
 */
  __pyx_t_4 = __pyx_v_process_order_its.begin();
  for (;;) {
    if (!(__pyx_t_4 != __pyx_v_process_order_its.end())) break;
    __pyx_t_5 = *__pyx_t_4;
    ++__pyx_t_4;
    __pyx_v_orders_it = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_5);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":873
 * 
 *         for orders_it in process_order_its:
 *             self.c_process_limit_order(is_buy, limit_orders_map_ptr, map_it_ptr, orders_it)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_process_crossed_limit_orders(self):
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_process_limit_order(__pyx_v_self, __pyx_v_is_buy, __pyx_v_limit_orders_map_ptr, __pyx_v_map_it_ptr, __pyx_v_orders_it); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 873, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":872
 *                 inc(orders_it)
 * 
 *         for orders_it in process_order_its:             # <<<<<<<<<<<<<<
 *             self.c_process_limit_order(is_buy, limit_orders_map_ptr, map_it_ptr, orders_it)
 * 
 */
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":835
 *             self.logger().error(f"Error processing limit order.", exc_info=True)
 * 
 *     cdef c_process_crossed_limit_orders_for_trading_pair(self,             # <<<<<<<<<<<<<<
 *                                                          bint is_buy,
 *                                                          LimitOrders *limit_orders_map_ptr,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_process_crossed_limit_orders_for_trading_pair", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_trading_pair);
  __Pyx_XDECREF(__pyx_v_opposite_order_book_price);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":875
 *             self.c_process_limit_order(is_buy, limit_orders_map_ptr, map_it_ptr, orders_it)
 * 
 *     cdef c_process_crossed_limit_orders(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             LimitOrders *limit_orders_ptr = address(self._bid_limit_orders)
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_process_crossed_limit_orders(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *__pyx_v_limit_orders_ptr;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator __pyx_v_map_it;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_process_crossed_limit_orders", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":877
 *     cdef c_process_crossed_limit_orders(self):
 *         cdef:
 *             LimitOrders *limit_orders_ptr = address(self._bid_limit_orders)             # <<<<<<<<<<<<<<
 *             LimitOrdersIterator map_it = limit_orders_ptr.begin()
 * 
 */
  __pyx_v_limit_orders_ptr = (&__pyx_v_self->_bid_limit_orders);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":878
 *         cdef:
 *             LimitOrders *limit_orders_ptr = address(self._bid_limit_orders)
 *             LimitOrdersIterator map_it = limit_orders_ptr.begin()             # <<<<<<<<<<<<<<
 * 
 *         while map_it != limit_orders_ptr.end():
 */
  __pyx_v_map_it = __pyx_v_limit_orders_ptr->begin();

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":880
 *             LimitOrdersIterator map_it = limit_orders_ptr.begin()
 * 
 *         while map_it != limit_orders_ptr.end():             # <<<<<<<<<<<<<<
 *             self.c_process_crossed_limit_orders_for_trading_pair(True, limit_orders_ptr, address(map_it))
 *             if map_it != limit_orders_ptr.end():
 */
  while (1) {
    __pyx_t_1 = (__pyx_v_map_it != __pyx_v_limit_orders_ptr->end());
    if (!__pyx_t_1) break;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":881
 * 
 *         while map_it != limit_orders_ptr.end():
 *             self.c_process_crossed_limit_orders_for_trading_pair(True, limit_orders_ptr, address(map_it))             # <<<<<<<<<<<<<<
 *             if map_it != limit_orders_ptr.end():
 *                 inc(map_it)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_process_crossed_limit_orders_for_trading_pair(__pyx_v_self, 1, __pyx_v_limit_orders_ptr, (&__pyx_v_map_it)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 881, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":882
 *         while map_it != limit_orders_ptr.end():
 *             self.c_process_crossed_limit_orders_for_trading_pair(True, limit_orders_ptr, address(map_it))
 *             if map_it != limit_orders_ptr.end():             # <<<<<<<<<<<<<<
 *                 inc(map_it)
 * 
 */
    __pyx_t_1 = (__pyx_v_map_it != __pyx_v_limit_orders_ptr->end());
    if (__pyx_t_1) {

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":883
 *             self.c_process_crossed_limit_orders_for_trading_pair(True, limit_orders_ptr, address(map_it))
 *             if map_it != limit_orders_ptr.end():
 *                 inc(map_it)             # <<<<<<<<<<<<<<
 * 
 *         limit_orders_ptr = address(self._ask_limit_orders)
 */
      (void)((__pyx_v_map_it++));

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":882
 *         while map_it != limit_orders_ptr.end():
 *             self.c_process_crossed_limit_orders_for_trading_pair(True, limit_orders_ptr, address(map_it))
 *             if map_it != limit_orders_ptr.end():             # <<<<<<<<<<<<<<
 *                 inc(map_it)
 * 
 */
    }
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":885
 *                 inc(map_it)
 * 
 *         limit_orders_ptr = address(self._ask_limit_orders)             # <<<<<<<<<<<<<<
 *         map_it = limit_orders_ptr.begin()
 * 
 */
  __pyx_v_limit_orders_ptr = (&__pyx_v_self->_ask_limit_orders);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":886
 * 
 *         limit_orders_ptr = address(self._ask_limit_orders)
 *         map_it = limit_orders_ptr.begin()             # <<<<<<<<<<<<<<
 * 
 *         while map_it != limit_orders_ptr.end():
 */
  __pyx_v_map_it = __pyx_v_limit_orders_ptr->begin();

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":888
 *         map_it = limit_orders_ptr.begin()
 * 
 *         while map_it != limit_orders_ptr.end():             # <<<<<<<<<<<<<<
 *             self.c_process_crossed_limit_orders_for_trading_pair(False, limit_orders_ptr, address(map_it))
 *             if map_it != limit_orders_ptr.end():
 */
  while (1) {
    __pyx_t_1 = (__pyx_v_map_it != __pyx_v_limit_orders_ptr->end());
    if (!__pyx_t_1) break;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":889
 * 
 *         while map_it != limit_orders_ptr.end():
 *             self.c_process_crossed_limit_orders_for_trading_pair(False, limit_orders_ptr, address(map_it))             # <<<<<<<<<<<<<<
 *             if map_it != limit_orders_ptr.end():
 *                 inc(map_it)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_process_crossed_limit_orders_for_trading_pair(__pyx_v_self, 0, __pyx_v_limit_orders_ptr, (&__pyx_v_map_it)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 889, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":890
 *         while map_it != limit_orders_ptr.end():
 *             self.c_process_crossed_limit_orders_for_trading_pair(False, limit_orders_ptr, address(map_it))
 *             if map_it != limit_orders_ptr.end():             # <<<<<<<<<<<<<<
 *                 inc(map_it)
 * 
 */
    __pyx_t_1 = (__pyx_v_map_it != __pyx_v_limit_orders_ptr->end());
    if (__pyx_t_1) {

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":891
 *             self.c_process_crossed_limit_orders_for_trading_pair(False, limit_orders_ptr, address(map_it))
 *             if map_it != limit_orders_ptr.end():
 *                 inc(map_it)             # <<<<<<<<<<<<<<
 * 
 *     # <editor-fold desc="Event listener functions">
 */
      (void)((__pyx_v_map_it++));

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":890
 *         while map_it != limit_orders_ptr.end():
 *             self.c_process_crossed_limit_orders_for_trading_pair(False, limit_orders_ptr, address(map_it))
 *             if map_it != limit_orders_ptr.end():             # <<<<<<<<<<<<<<
 *                 inc(map_it)
 * 
 */
    }
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":875
 *             self.c_process_limit_order(is_buy, limit_orders_map_ptr, map_it_ptr, orders_it)
 * 
 *     cdef c_process_crossed_limit_orders(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             LimitOrders *limit_orders_ptr = address(self._bid_limit_orders)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_process_crossed_limit_orders", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":894
 * 
 *     # <editor-fold desc="Event listener functions">
 *     cdef c_match_trade_to_limit_orders(self, object order_book_trade_event):             # <<<<<<<<<<<<<<
 *         """
 *         Trigger limit orders when incoming market orders have crossed the limit order's price.
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_match_trade_to_limit_orders(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_order_book_trade_event) {
  std::string __pyx_v_cpp_trading_pair;
  int __pyx_v_is_maker_buy;
  PyObject *__pyx_v_trade_price = 0;
  CYTHON_UNUSED PyObject *__pyx_v_trade_quantity = 0;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *__pyx_v_limit_orders_map_ptr;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator __pyx_v_map_it;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders *__pyx_v_orders_collection_ptr;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator __pyx_v_orders_it;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersRIterator __pyx_v_orders_rit;
  std::vector<__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator>  __pyx_v_process_order_its;
  LimitOrder const *__pyx_v_cpp_limit_order_ptr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  std::string __pyx_t_5;
  int __pyx_t_6;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *__pyx_t_7;
  PyObject *__pyx_t_8;
  std::vector<__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator> ::iterator __pyx_t_9;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_match_trade_to_limit_orders", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":901
 *         """
 *         cdef:
 *             string cpp_trading_pair = order_book_trade_event.trading_pair.encode("utf8")             # <<<<<<<<<<<<<<
 *             bint is_maker_buy = order_book_trade_event.type is TradeType.SELL
 *             object trade_price = order_book_trade_event.price
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_book_trade_event, __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_n_u_utf8};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 901, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_5 = __pyx_convert_string_from_py_6libcpp_6string_std__in_string(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_cpp_trading_pair = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_5);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":902
 *         cdef:
 *             string cpp_trading_pair = order_book_trade_event.trading_pair.encode("utf8")
 *             bint is_maker_buy = order_book_trade_event.type is TradeType.SELL             # <<<<<<<<<<<<<<
 *             object trade_price = order_book_trade_event.price
 *             object trade_quantity = order_book_trade_event.amount
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_book_trade_event, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_SELL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_6 = (__pyx_t_1 == __pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_is_maker_buy = __pyx_t_6;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":903
 *             string cpp_trading_pair = order_book_trade_event.trading_pair.encode("utf8")
 *             bint is_maker_buy = order_book_trade_event.type is TradeType.SELL
 *             object trade_price = order_book_trade_event.price             # <<<<<<<<<<<<<<
 *             object trade_quantity = order_book_trade_event.amount
 *             LimitOrders *limit_orders_map_ptr = (address(self._bid_limit_orders)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_book_trade_event, __pyx_n_s_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_trade_price = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":904
 *             bint is_maker_buy = order_book_trade_event.type is TradeType.SELL
 *             object trade_price = order_book_trade_event.price
 *             object trade_quantity = order_book_trade_event.amount             # <<<<<<<<<<<<<<
 *             LimitOrders *limit_orders_map_ptr = (address(self._bid_limit_orders)
 *                                                  if is_maker_buy
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_book_trade_event, __pyx_n_s_amount); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_trade_quantity = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":906
 *             object trade_quantity = order_book_trade_event.amount
 *             LimitOrders *limit_orders_map_ptr = (address(self._bid_limit_orders)
 *                                                  if is_maker_buy             # <<<<<<<<<<<<<<
 *                                                  else address(self._ask_limit_orders))
 *             LimitOrdersIterator map_it = limit_orders_map_ptr.find(cpp_trading_pair)
 */
  if (__pyx_v_is_maker_buy) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":905
 *             object trade_price = order_book_trade_event.price
 *             object trade_quantity = order_book_trade_event.amount
 *             LimitOrders *limit_orders_map_ptr = (address(self._bid_limit_orders)             # <<<<<<<<<<<<<<
 *                                                  if is_maker_buy
 *                                                  else address(self._ask_limit_orders))
 */
    __pyx_t_7 = (&__pyx_v_self->_bid_limit_orders);
  } else {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":907
 *             LimitOrders *limit_orders_map_ptr = (address(self._bid_limit_orders)
 *                                                  if is_maker_buy
 *                                                  else address(self._ask_limit_orders))             # <<<<<<<<<<<<<<
 *             LimitOrdersIterator map_it = limit_orders_map_ptr.find(cpp_trading_pair)
 *             SingleTradingPairLimitOrders *orders_collection_ptr = NULL
 */
    __pyx_t_7 = (&__pyx_v_self->_ask_limit_orders);
  }
  __pyx_v_limit_orders_map_ptr = __pyx_t_7;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":908
 *                                                  if is_maker_buy
 *                                                  else address(self._ask_limit_orders))
 *             LimitOrdersIterator map_it = limit_orders_map_ptr.find(cpp_trading_pair)             # <<<<<<<<<<<<<<
 *             SingleTradingPairLimitOrders *orders_collection_ptr = NULL
 *             SingleTradingPairLimitOrdersIterator orders_it
 */
  __pyx_v_map_it = __pyx_v_limit_orders_map_ptr->find(__pyx_v_cpp_trading_pair);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":909
 *                                                  else address(self._ask_limit_orders))
 *             LimitOrdersIterator map_it = limit_orders_map_ptr.find(cpp_trading_pair)
 *             SingleTradingPairLimitOrders *orders_collection_ptr = NULL             # <<<<<<<<<<<<<<
 *             SingleTradingPairLimitOrdersIterator orders_it
 *             SingleTradingPairLimitOrdersRIterator orders_rit
 */
  __pyx_v_orders_collection_ptr = NULL;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":913
 *             SingleTradingPairLimitOrdersRIterator orders_rit
 *             vector[SingleTradingPairLimitOrdersIterator] process_order_its
 *             const CPPLimitOrder *cpp_limit_order_ptr = NULL             # <<<<<<<<<<<<<<
 * 
 *         if map_it == limit_orders_map_ptr.end():
 */
  __pyx_v_cpp_limit_order_ptr = NULL;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":915
 *             const CPPLimitOrder *cpp_limit_order_ptr = NULL
 * 
 *         if map_it == limit_orders_map_ptr.end():             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_6 = (__pyx_v_map_it == __pyx_v_limit_orders_map_ptr->end());
  if (__pyx_t_6) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":916
 * 
 *         if map_it == limit_orders_map_ptr.end():
 *             return             # <<<<<<<<<<<<<<
 * 
 *         orders_collection_ptr = address(deref(map_it).second)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":915
 *             const CPPLimitOrder *cpp_limit_order_ptr = NULL
 * 
 *         if map_it == limit_orders_map_ptr.end():             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":918
 *             return
 * 
 *         orders_collection_ptr = address(deref(map_it).second)             # <<<<<<<<<<<<<<
 *         if is_maker_buy:
 *             orders_rit = orders_collection_ptr.rbegin()
 */
  __pyx_v_orders_collection_ptr = (&(*__pyx_v_map_it).second);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":919
 * 
 *         orders_collection_ptr = address(deref(map_it).second)
 *         if is_maker_buy:             # <<<<<<<<<<<<<<
 *             orders_rit = orders_collection_ptr.rbegin()
 *             while orders_rit != orders_collection_ptr.rend():
 */
  if (__pyx_v_is_maker_buy) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":920
 *         orders_collection_ptr = address(deref(map_it).second)
 *         if is_maker_buy:
 *             orders_rit = orders_collection_ptr.rbegin()             # <<<<<<<<<<<<<<
 *             while orders_rit != orders_collection_ptr.rend():
 *                 cpp_limit_order_ptr = address(deref(orders_rit))
 */
    __pyx_v_orders_rit = __pyx_v_orders_collection_ptr->rbegin();

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":921
 *         if is_maker_buy:
 *             orders_rit = orders_collection_ptr.rbegin()
 *             while orders_rit != orders_collection_ptr.rend():             # <<<<<<<<<<<<<<
 *                 cpp_limit_order_ptr = address(deref(orders_rit))
 *                 if <object>cpp_limit_order_ptr.getPrice() <= trade_price:
 */
    while (1) {
      __pyx_t_6 = (__pyx_v_orders_rit != __pyx_v_orders_collection_ptr->rend());
      if (!__pyx_t_6) break;

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":922
 *             orders_rit = orders_collection_ptr.rbegin()
 *             while orders_rit != orders_collection_ptr.rend():
 *                 cpp_limit_order_ptr = address(deref(orders_rit))             # <<<<<<<<<<<<<<
 *                 if <object>cpp_limit_order_ptr.getPrice() <= trade_price:
 *                     break
 */
      __pyx_v_cpp_limit_order_ptr = (&(*__pyx_v_orders_rit));

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":923
 *             while orders_rit != orders_collection_ptr.rend():
 *                 cpp_limit_order_ptr = address(deref(orders_rit))
 *                 if <object>cpp_limit_order_ptr.getPrice() <= trade_price:             # <<<<<<<<<<<<<<
 *                     break
 *                 process_order_its.push_back(getIteratorFromReverseIterator(
 */
      __pyx_t_8 = __pyx_v_cpp_limit_order_ptr->getPrice();
      __pyx_t_2 = PyObject_RichCompare(((PyObject *)__pyx_t_8), __pyx_v_trade_price, Py_LE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 923, __pyx_L1_error)
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 923, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_6) {

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":924
 *                 cpp_limit_order_ptr = address(deref(orders_rit))
 *                 if <object>cpp_limit_order_ptr.getPrice() <= trade_price:
 *                     break             # <<<<<<<<<<<<<<
 *                 process_order_its.push_back(getIteratorFromReverseIterator(
 *                     <reverse_iterator[SingleTradingPairLimitOrdersIterator]>orders_rit))
 */
        goto __pyx_L6_break;

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":923
 *             while orders_rit != orders_collection_ptr.rend():
 *                 cpp_limit_order_ptr = address(deref(orders_rit))
 *                 if <object>cpp_limit_order_ptr.getPrice() <= trade_price:             # <<<<<<<<<<<<<<
 *                     break
 *                 process_order_its.push_back(getIteratorFromReverseIterator(
 */
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":925
 *                 if <object>cpp_limit_order_ptr.getPrice() <= trade_price:
 *                     break
 *                 process_order_its.push_back(getIteratorFromReverseIterator(             # <<<<<<<<<<<<<<
 *                     <reverse_iterator[SingleTradingPairLimitOrdersIterator]>orders_rit))
 *                 inc(orders_rit)
 */
      try {
        __pyx_v_process_order_its.push_back(getIteratorFromReverseIterator<__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator>(((std::reverse_iterator<__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator> )__pyx_v_orders_rit)));
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 925, __pyx_L1_error)
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":927
 *                 process_order_its.push_back(getIteratorFromReverseIterator(
 *                     <reverse_iterator[SingleTradingPairLimitOrdersIterator]>orders_rit))
 *                 inc(orders_rit)             # <<<<<<<<<<<<<<
 *         else:
 *             orders_it = orders_collection_ptr.begin()
 */
      (void)((__pyx_v_orders_rit++));
    }
    __pyx_L6_break:;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":919
 * 
 *         orders_collection_ptr = address(deref(map_it).second)
 *         if is_maker_buy:             # <<<<<<<<<<<<<<
 *             orders_rit = orders_collection_ptr.rbegin()
 *             while orders_rit != orders_collection_ptr.rend():
 */
    goto __pyx_L4;
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":929
 *                 inc(orders_rit)
 *         else:
 *             orders_it = orders_collection_ptr.begin()             # <<<<<<<<<<<<<<
 *             while orders_it != orders_collection_ptr.end():
 *                 cpp_limit_order_ptr = address(deref(orders_it))
 */
  /*else*/ {
    __pyx_v_orders_it = __pyx_v_orders_collection_ptr->begin();

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":930
 *         else:
 *             orders_it = orders_collection_ptr.begin()
 *             while orders_it != orders_collection_ptr.end():             # <<<<<<<<<<<<<<
 *                 cpp_limit_order_ptr = address(deref(orders_it))
 *                 if <object>cpp_limit_order_ptr.getPrice() >= trade_price:
 */
    while (1) {
      __pyx_t_6 = (__pyx_v_orders_it != __pyx_v_orders_collection_ptr->end());
      if (!__pyx_t_6) break;

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":931
 *             orders_it = orders_collection_ptr.begin()
 *             while orders_it != orders_collection_ptr.end():
 *                 cpp_limit_order_ptr = address(deref(orders_it))             # <<<<<<<<<<<<<<
 *                 if <object>cpp_limit_order_ptr.getPrice() >= trade_price:
 *                     break
 */
      __pyx_v_cpp_limit_order_ptr = (&(*__pyx_v_orders_it));

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":932
 *             while orders_it != orders_collection_ptr.end():
 *                 cpp_limit_order_ptr = address(deref(orders_it))
 *                 if <object>cpp_limit_order_ptr.getPrice() >= trade_price:             # <<<<<<<<<<<<<<
 *                     break
 *                 process_order_its.push_back(orders_it)
 */
      __pyx_t_8 = __pyx_v_cpp_limit_order_ptr->getPrice();
      __pyx_t_2 = PyObject_RichCompare(((PyObject *)__pyx_t_8), __pyx_v_trade_price, Py_GE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 932, __pyx_L1_error)
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 932, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_6) {

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":933
 *                 cpp_limit_order_ptr = address(deref(orders_it))
 *                 if <object>cpp_limit_order_ptr.getPrice() >= trade_price:
 *                     break             # <<<<<<<<<<<<<<
 *                 process_order_its.push_back(orders_it)
 *                 inc(orders_it)
 */
        goto __pyx_L9_break;

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":932
 *             while orders_it != orders_collection_ptr.end():
 *                 cpp_limit_order_ptr = address(deref(orders_it))
 *                 if <object>cpp_limit_order_ptr.getPrice() >= trade_price:             # <<<<<<<<<<<<<<
 *                     break
 *                 process_order_its.push_back(orders_it)
 */
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":934
 *                 if <object>cpp_limit_order_ptr.getPrice() >= trade_price:
 *                     break
 *                 process_order_its.push_back(orders_it)             # <<<<<<<<<<<<<<
 *                 inc(orders_it)
 * 
 */
      try {
        __pyx_v_process_order_its.push_back(__pyx_v_orders_it);
      } catch(...) {
        __Pyx_CppExn2PyErr();
        __PYX_ERR(0, 934, __pyx_L1_error)
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":935
 *                     break
 *                 process_order_its.push_back(orders_it)
 *                 inc(orders_it)             # <<<<<<<<<<<<<<
 * 
 *         for orders_it in process_order_its:
 */
      (void)((__pyx_v_orders_it++));
    }
    __pyx_L9_break:;
  }
  __pyx_L4:;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":937
 *                 inc(orders_it)
 * 
 *         for orders_it in process_order_its:             # <<<<<<<<<<<<<<
 *             self.c_process_limit_order(is_maker_buy, limit_orders_map_ptr, address(map_it), orders_it)
 * 
 */
  __pyx_t_9 = __pyx_v_process_order_its.begin();
  for (;;) {
    if (!(__pyx_t_9 != __pyx_v_process_order_its.end())) break;
    __pyx_t_10 = *__pyx_t_9;
    ++__pyx_t_9;
    __pyx_v_orders_it = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_10);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":938
 * 
 *         for orders_it in process_order_its:
 *             self.c_process_limit_order(is_maker_buy, limit_orders_map_ptr, address(map_it), orders_it)             # <<<<<<<<<<<<<<
 * 
 *     # </editor-fold>
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_process_limit_order(__pyx_v_self, __pyx_v_is_maker_buy, __pyx_v_limit_orders_map_ptr, (&__pyx_v_map_it), __pyx_v_orders_it); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 938, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":937
 *                 inc(orders_it)
 * 
 *         for orders_it in process_order_its:             # <<<<<<<<<<<<<<
 *             self.c_process_limit_order(is_maker_buy, limit_orders_map_ptr, address(map_it), orders_it)
 * 
 */
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":894
 * 
 *     # <editor-fold desc="Event listener functions">
 *     cdef c_match_trade_to_limit_orders(self, object order_book_trade_event):             # <<<<<<<<<<<<<<
 *         """
 *         Trigger limit orders when incoming market orders have crossed the limit order's price.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_match_trade_to_limit_orders", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_trade_price);
  __Pyx_XDECREF(__pyx_v_trade_quantity);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":942
 *     # </editor-fold>
 * 
 *     cdef object c_get_available_balance(self, str currency):             # <<<<<<<<<<<<<<
 *         return self.available_balances.get(currency.upper(), s_decimal_0)
 * 
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_available_balance(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_currency) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_get_available_balance", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":943
 * 
 *     cdef object c_get_available_balance(self, str currency):
 *         return self.available_balances.get(currency.upper(), s_decimal_0)             # <<<<<<<<<<<<<<
 * 
 *     async def cancel_all(self, timeout_seconds: float) -> List[CancellationResult]:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_available_balances); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyUnicode_Type_upper, __pyx_v_currency); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_s_decimal_0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 943, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":942
 *     # </editor-fold>
 * 
 *     cdef object c_get_available_balance(self, str currency):             # <<<<<<<<<<<<<<
 *         return self.available_balances.get(currency.upper(), s_decimal_0)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_get_available_balance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_19generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":945
 *         return self.available_balances.get(currency.upper(), s_decimal_0)
 * 
 *     async def cancel_all(self, timeout_seconds: float) -> List[CancellationResult]:             # <<<<<<<<<<<<<<
 *         cdef:
 *             LimitOrders *limit_orders_map_ptr
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_18cancel_all(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_18cancel_all = {"cancel_all", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_18cancel_all, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_18cancel_all(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v_timeout_seconds = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel_all (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_timeout_seconds,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timeout_seconds)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 945, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "cancel_all") < 0)) __PYX_ERR(0, 945, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_timeout_seconds = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cancel_all", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 945, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.cancel_all", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_17cancel_all(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self), __pyx_v_timeout_seconds);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_17cancel_all(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_timeout_seconds) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancel_all", 0);
  __pyx_cur_scope = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all *)__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 945, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_timeout_seconds = __pyx_v_timeout_seconds;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_timeout_seconds);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_timeout_seconds);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_19generator3, __pyx_codeobj__24, (PyObject *) __pyx_cur_scope, __pyx_n_s_cancel_all, __pyx_n_s_PaperTradeExchange_cancel_all, __pyx_n_s_hummingbot_connector_exchange_pa); if (unlikely(!gen)) __PYX_ERR(0, 945, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.cancel_all", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_19generator3(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all *__pyx_cur_scope = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_cancel_order_from_orders_map __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel_all", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 945, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":948
 *         cdef:
 *             LimitOrders *limit_orders_map_ptr
 *             list cancellation_results = []             # <<<<<<<<<<<<<<
 *         limit_orders_map_ptr = address(self._bid_limit_orders)
 *         for trading_pair_str in self._trading_pairs.keys():
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_cur_scope->__pyx_v_cancellation_results = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":949
 *             LimitOrders *limit_orders_map_ptr
 *             list cancellation_results = []
 *         limit_orders_map_ptr = address(self._bid_limit_orders)             # <<<<<<<<<<<<<<
 *         for trading_pair_str in self._trading_pairs.keys():
 *             results = self.c_cancel_order_from_orders_map(limit_orders_map_ptr, trading_pair_str, cancel_all=True)
 */
  __pyx_cur_scope->__pyx_v_limit_orders_map_ptr = (&__pyx_cur_scope->__pyx_v_self->_bid_limit_orders);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":950
 *             list cancellation_results = []
 *         limit_orders_map_ptr = address(self._bid_limit_orders)
 *         for trading_pair_str in self._trading_pairs.keys():             # <<<<<<<<<<<<<<
 *             results = self.c_cancel_order_from_orders_map(limit_orders_map_ptr, trading_pair_str, cancel_all=True)
 *             cancellation_results.extend(results)
 */
  __pyx_t_2 = 0;
  if (unlikely(__pyx_cur_scope->__pyx_v_self->_trading_pairs == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
    __PYX_ERR(0, 950, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_cur_scope->__pyx_v_self->_trading_pairs, 1, __pyx_n_s_keys, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 950, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_6 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_5, NULL, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_6 == 0)) break;
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 950, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_trading_pair_str);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_trading_pair_str, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":951
 *         limit_orders_map_ptr = address(self._bid_limit_orders)
 *         for trading_pair_str in self._trading_pairs.keys():
 *             results = self.c_cancel_order_from_orders_map(limit_orders_map_ptr, trading_pair_str, cancel_all=True)             # <<<<<<<<<<<<<<
 *             cancellation_results.extend(results)
 * 
 */
    if (!(likely(PyUnicode_CheckExact(__pyx_cur_scope->__pyx_v_trading_pair_str))||((__pyx_cur_scope->__pyx_v_trading_pair_str) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_cur_scope->__pyx_v_trading_pair_str))) __PYX_ERR(0, 951, __pyx_L1_error)
    __pyx_t_7.__pyx_n = 1;
    __pyx_t_7.cancel_all = 1;
    __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_cancel_order_from_orders_map(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_limit_orders_map_ptr, ((PyObject*)__pyx_cur_scope->__pyx_v_trading_pair_str), &__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 951, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_results);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_results, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":952
 *         for trading_pair_str in self._trading_pairs.keys():
 *             results = self.c_cancel_order_from_orders_map(limit_orders_map_ptr, trading_pair_str, cancel_all=True)
 *             cancellation_results.extend(results)             # <<<<<<<<<<<<<<
 * 
 *         limit_orders_map_ptr = address(self._ask_limit_orders)
 */
    __pyx_t_8 = __Pyx_PyList_Extend(__pyx_cur_scope->__pyx_v_cancellation_results, __pyx_cur_scope->__pyx_v_results); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 952, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":954
 *             cancellation_results.extend(results)
 * 
 *         limit_orders_map_ptr = address(self._ask_limit_orders)             # <<<<<<<<<<<<<<
 *         for trading_pair_str in self._trading_pairs.keys():
 *             results = self.c_cancel_order_from_orders_map(limit_orders_map_ptr, trading_pair_str, cancel_all=True)
 */
  __pyx_cur_scope->__pyx_v_limit_orders_map_ptr = (&__pyx_cur_scope->__pyx_v_self->_ask_limit_orders);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":955
 * 
 *         limit_orders_map_ptr = address(self._ask_limit_orders)
 *         for trading_pair_str in self._trading_pairs.keys():             # <<<<<<<<<<<<<<
 *             results = self.c_cancel_order_from_orders_map(limit_orders_map_ptr, trading_pair_str, cancel_all=True)
 *             cancellation_results.extend(results)
 */
  __pyx_t_3 = 0;
  if (unlikely(__pyx_cur_scope->__pyx_v_self->_trading_pairs == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
    __PYX_ERR(0, 955, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_dict_iterator(__pyx_cur_scope->__pyx_v_self->_trading_pairs, 1, __pyx_n_s_keys, (&__pyx_t_2), (&__pyx_t_4)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_5;
  __pyx_t_5 = 0;
  while (1) {
    __pyx_t_6 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_2, &__pyx_t_3, &__pyx_t_5, NULL, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_6 == 0)) break;
    if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 955, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_trading_pair_str);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_trading_pair_str, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":956
 *         limit_orders_map_ptr = address(self._ask_limit_orders)
 *         for trading_pair_str in self._trading_pairs.keys():
 *             results = self.c_cancel_order_from_orders_map(limit_orders_map_ptr, trading_pair_str, cancel_all=True)             # <<<<<<<<<<<<<<
 *             cancellation_results.extend(results)
 *         return cancellation_results
 */
    if (!(likely(PyUnicode_CheckExact(__pyx_cur_scope->__pyx_v_trading_pair_str))||((__pyx_cur_scope->__pyx_v_trading_pair_str) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_cur_scope->__pyx_v_trading_pair_str))) __PYX_ERR(0, 956, __pyx_L1_error)
    __pyx_t_7.__pyx_n = 1;
    __pyx_t_7.cancel_all = 1;
    __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_cancel_order_from_orders_map(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_limit_orders_map_ptr, ((PyObject*)__pyx_cur_scope->__pyx_v_trading_pair_str), &__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 956, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_results);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_results, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":957
 *         for trading_pair_str in self._trading_pairs.keys():
 *             results = self.c_cancel_order_from_orders_map(limit_orders_map_ptr, trading_pair_str, cancel_all=True)
 *             cancellation_results.extend(results)             # <<<<<<<<<<<<<<
 *         return cancellation_results
 * 
 */
    __pyx_t_8 = __Pyx_PyList_Extend(__pyx_cur_scope->__pyx_v_cancellation_results, __pyx_cur_scope->__pyx_v_results); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 957, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":958
 *             results = self.c_cancel_order_from_orders_map(limit_orders_map_ptr, trading_pair_str, cancel_all=True)
 *             cancellation_results.extend(results)
 *         return cancellation_results             # <<<<<<<<<<<<<<
 * 
 *     cdef object c_cancel_order_from_orders_map(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_r = NULL; __Pyx_ReturnWithStopIteration(__pyx_cur_scope->__pyx_v_cancellation_results);
  goto __pyx_L0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":945
 *         return self.available_balances.get(currency.upper(), s_decimal_0)
 * 
 *     async def cancel_all(self, timeout_seconds: float) -> List[CancellationResult]:             # <<<<<<<<<<<<<<
 *         cdef:
 *             LimitOrders *limit_orders_map_ptr
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("cancel_all", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":960
 *         return cancellation_results
 * 
 *     cdef object c_cancel_order_from_orders_map(self,             # <<<<<<<<<<<<<<
 *                                                LimitOrders *orders_map,
 *                                                str trading_pair_str,
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_cancel_order_from_orders_map(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *__pyx_v_orders_map, PyObject *__pyx_v_trading_pair_str, struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_cancel_order_from_orders_map *__pyx_optional_args) {

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":963
 *                                                LimitOrders *orders_map,
 *                                                str trading_pair_str,
 *                                                bint cancel_all=False,             # <<<<<<<<<<<<<<
 *                                                str client_order_id=None):
 *         cdef:
 */
  int __pyx_v_cancel_all = ((int)0);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":964
 *                                                str trading_pair_str,
 *                                                bint cancel_all=False,
 *                                                str client_order_id=None):             # <<<<<<<<<<<<<<
 *         cdef:
 *             string cpp_trading_pair = trading_pair_str.encode("utf8")
 */
  PyObject *__pyx_v_client_order_id = ((PyObject*)Py_None);
  std::string __pyx_v_cpp_trading_pair;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator __pyx_v_map_it;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrders *__pyx_v_limit_orders_collection_ptr;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator __pyx_v_orders_it;
  std::vector<__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator>  __pyx_v_process_order_its;
  LimitOrder const *__pyx_v_limit_order_ptr;
  PyObject *__pyx_v_limit_order_cid = 0;
  PyObject *__pyx_v_cancellation_results = 0;
  PyObject *__pyx_v_delete_success = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_err = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  std::string __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  std::vector<__pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator> ::iterator __pyx_t_8;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  unsigned int __pyx_t_12;
  int __pyx_t_13;
  int64_t __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_t_16;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_t_19;
  char const *__pyx_t_20;
  PyObject *__pyx_t_21 = NULL;
  PyObject *__pyx_t_22 = NULL;
  PyObject *__pyx_t_23 = NULL;
  PyObject *__pyx_t_24 = NULL;
  PyObject *__pyx_t_25 = NULL;
  PyObject *__pyx_t_26 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_cancel_order_from_orders_map", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_cancel_all = __pyx_optional_args->cancel_all;
      if (__pyx_optional_args->__pyx_n > 1) {
        __pyx_v_client_order_id = __pyx_optional_args->client_order_id;
      }
    }
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":966
 *                                                str client_order_id=None):
 *         cdef:
 *             string cpp_trading_pair = trading_pair_str.encode("utf8")             # <<<<<<<<<<<<<<
 *             LimitOrdersIterator map_it = orders_map.find(cpp_trading_pair)
 *             SingleTradingPairLimitOrders *limit_orders_collection_ptr = NULL
 */
  if (unlikely(__pyx_v_trading_pair_str == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 966, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_AsUTF8String(__pyx_v_trading_pair_str); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_convert_string_from_py_6libcpp_6string_std__in_string(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_cpp_trading_pair = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_2);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":967
 *         cdef:
 *             string cpp_trading_pair = trading_pair_str.encode("utf8")
 *             LimitOrdersIterator map_it = orders_map.find(cpp_trading_pair)             # <<<<<<<<<<<<<<
 *             SingleTradingPairLimitOrders *limit_orders_collection_ptr = NULL
 *             SingleTradingPairLimitOrdersIterator orders_it
 */
  __pyx_v_map_it = __pyx_v_orders_map->find(__pyx_v_cpp_trading_pair);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":968
 *             string cpp_trading_pair = trading_pair_str.encode("utf8")
 *             LimitOrdersIterator map_it = orders_map.find(cpp_trading_pair)
 *             SingleTradingPairLimitOrders *limit_orders_collection_ptr = NULL             # <<<<<<<<<<<<<<
 *             SingleTradingPairLimitOrdersIterator orders_it
 *             vector[SingleTradingPairLimitOrdersIterator] process_order_its
 */
  __pyx_v_limit_orders_collection_ptr = NULL;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":971
 *             SingleTradingPairLimitOrdersIterator orders_it
 *             vector[SingleTradingPairLimitOrdersIterator] process_order_its
 *             const CPPLimitOrder *limit_order_ptr = NULL             # <<<<<<<<<<<<<<
 *             str limit_order_cid
 *             list cancellation_results = []
 */
  __pyx_v_limit_order_ptr = NULL;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":973
 *             const CPPLimitOrder *limit_order_ptr = NULL
 *             str limit_order_cid
 *             list cancellation_results = []             # <<<<<<<<<<<<<<
 *         try:
 *             if map_it == orders_map.end():
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_cancellation_results = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":974
 *             str limit_order_cid
 *             list cancellation_results = []
 *         try:             # <<<<<<<<<<<<<<
 *             if map_it == orders_map.end():
 *                 return []
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":975
 *             list cancellation_results = []
 *         try:
 *             if map_it == orders_map.end():             # <<<<<<<<<<<<<<
 *                 return []
 * 
 */
      __pyx_t_6 = (__pyx_v_map_it == __pyx_v_orders_map->end());
      if (__pyx_t_6) {

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":976
 *         try:
 *             if map_it == orders_map.end():
 *                 return []             # <<<<<<<<<<<<<<
 * 
 *             limit_orders_collection_ptr = address(deref(map_it).second)
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 976, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L7_try_return;

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":975
 *             list cancellation_results = []
 *         try:
 *             if map_it == orders_map.end():             # <<<<<<<<<<<<<<
 *                 return []
 * 
 */
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":978
 *                 return []
 * 
 *             limit_orders_collection_ptr = address(deref(map_it).second)             # <<<<<<<<<<<<<<
 *             orders_it = limit_orders_collection_ptr.begin()
 *             while orders_it != limit_orders_collection_ptr.end():
 */
      __pyx_v_limit_orders_collection_ptr = (&(*__pyx_v_map_it).second);

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":979
 * 
 *             limit_orders_collection_ptr = address(deref(map_it).second)
 *             orders_it = limit_orders_collection_ptr.begin()             # <<<<<<<<<<<<<<
 *             while orders_it != limit_orders_collection_ptr.end():
 *                 limit_order_ptr = address(deref(orders_it))
 */
      __pyx_v_orders_it = __pyx_v_limit_orders_collection_ptr->begin();

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":980
 *             limit_orders_collection_ptr = address(deref(map_it).second)
 *             orders_it = limit_orders_collection_ptr.begin()
 *             while orders_it != limit_orders_collection_ptr.end():             # <<<<<<<<<<<<<<
 *                 limit_order_ptr = address(deref(orders_it))
 *                 limit_order_cid = limit_order_ptr.getClientOrderID().decode("utf8")
 */
      while (1) {
        __pyx_t_6 = (__pyx_v_orders_it != __pyx_v_limit_orders_collection_ptr->end());
        if (!__pyx_t_6) break;

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":981
 *             orders_it = limit_orders_collection_ptr.begin()
 *             while orders_it != limit_orders_collection_ptr.end():
 *                 limit_order_ptr = address(deref(orders_it))             # <<<<<<<<<<<<<<
 *                 limit_order_cid = limit_order_ptr.getClientOrderID().decode("utf8")
 *                 if (not cancel_all and limit_order_cid == client_order_id) or cancel_all:
 */
        __pyx_v_limit_order_ptr = (&(*__pyx_v_orders_it));

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":982
 *             while orders_it != limit_orders_collection_ptr.end():
 *                 limit_order_ptr = address(deref(orders_it))
 *                 limit_order_cid = limit_order_ptr.getClientOrderID().decode("utf8")             # <<<<<<<<<<<<<<
 *                 if (not cancel_all and limit_order_cid == client_order_id) or cancel_all:
 *                     process_order_its.push_back(orders_it)
 */
        __pyx_t_1 = __Pyx_decode_cpp_string(__pyx_v_limit_order_ptr->getClientOrderID(), 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 982, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_v_limit_order_cid, ((PyObject*)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":983
 *                 limit_order_ptr = address(deref(orders_it))
 *                 limit_order_cid = limit_order_ptr.getClientOrderID().decode("utf8")
 *                 if (not cancel_all and limit_order_cid == client_order_id) or cancel_all:             # <<<<<<<<<<<<<<
 *                     process_order_its.push_back(orders_it)
 *                 inc(orders_it)
 */
        __pyx_t_7 = (!__pyx_v_cancel_all);
        if (!__pyx_t_7) {
          goto __pyx_L14_next_or;
        } else {
        }
        __pyx_t_7 = (__Pyx_PyUnicode_Equals(__pyx_v_limit_order_cid, __pyx_v_client_order_id, Py_EQ)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 983, __pyx_L3_error)
        if (!__pyx_t_7) {
        } else {
          __pyx_t_6 = __pyx_t_7;
          goto __pyx_L13_bool_binop_done;
        }
        __pyx_L14_next_or:;
        __pyx_t_6 = __pyx_v_cancel_all;
        __pyx_L13_bool_binop_done:;
        if (__pyx_t_6) {

          /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":984
 *                 limit_order_cid = limit_order_ptr.getClientOrderID().decode("utf8")
 *                 if (not cancel_all and limit_order_cid == client_order_id) or cancel_all:
 *                     process_order_its.push_back(orders_it)             # <<<<<<<<<<<<<<
 *                 inc(orders_it)
 * 
 */
          try {
            __pyx_v_process_order_its.push_back(__pyx_v_orders_it);
          } catch(...) {
            __Pyx_CppExn2PyErr();
            __PYX_ERR(0, 984, __pyx_L3_error)
          }

          /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":983
 *                 limit_order_ptr = address(deref(orders_it))
 *                 limit_order_cid = limit_order_ptr.getClientOrderID().decode("utf8")
 *                 if (not cancel_all and limit_order_cid == client_order_id) or cancel_all:             # <<<<<<<<<<<<<<
 *                     process_order_its.push_back(orders_it)
 *                 inc(orders_it)
 */
        }

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":985
 *                 if (not cancel_all and limit_order_cid == client_order_id) or cancel_all:
 *                     process_order_its.push_back(orders_it)
 *                 inc(orders_it)             # <<<<<<<<<<<<<<
 * 
 *             for orders_it in process_order_its:
 */
        (void)((__pyx_v_orders_it++));
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":987
 *                 inc(orders_it)
 * 
 *             for orders_it in process_order_its:             # <<<<<<<<<<<<<<
 *                 limit_order_ptr = address(deref(orders_it))
 *                 limit_order_cid = limit_order_ptr.getClientOrderID().decode("utf8")
 */
      __pyx_t_8 = __pyx_v_process_order_its.begin();
      for (;;) {
        if (!(__pyx_t_8 != __pyx_v_process_order_its.end())) break;
        __pyx_t_9 = *__pyx_t_8;
        ++__pyx_t_8;
        __pyx_v_orders_it = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_9);

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":988
 * 
 *             for orders_it in process_order_its:
 *                 limit_order_ptr = address(deref(orders_it))             # <<<<<<<<<<<<<<
 *                 limit_order_cid = limit_order_ptr.getClientOrderID().decode("utf8")
 *                 delete_success = self.c_delete_limit_order(orders_map, address(map_it), orders_it)
 */
        __pyx_v_limit_order_ptr = (&(*__pyx_v_orders_it));

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":989
 *             for orders_it in process_order_its:
 *                 limit_order_ptr = address(deref(orders_it))
 *                 limit_order_cid = limit_order_ptr.getClientOrderID().decode("utf8")             # <<<<<<<<<<<<<<
 *                 delete_success = self.c_delete_limit_order(orders_map, address(map_it), orders_it)
 *                 cancellation_results.append(CancellationResult(limit_order_cid,
 */
        __pyx_t_1 = __Pyx_decode_cpp_string(__pyx_v_limit_order_ptr->getClientOrderID(), 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 989, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_v_limit_order_cid, ((PyObject*)__pyx_t_1));
        __pyx_t_1 = 0;

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":990
 *                 limit_order_ptr = address(deref(orders_it))
 *                 limit_order_cid = limit_order_ptr.getClientOrderID().decode("utf8")
 *                 delete_success = self.c_delete_limit_order(orders_map, address(map_it), orders_it)             # <<<<<<<<<<<<<<
 *                 cancellation_results.append(CancellationResult(limit_order_cid,
 *                                                                delete_success))
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_delete_limit_order(__pyx_v_self, __pyx_v_orders_map, (&__pyx_v_map_it), __pyx_v_orders_it); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 990, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF_SET(__pyx_v_delete_success, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":991
 *                 limit_order_cid = limit_order_ptr.getClientOrderID().decode("utf8")
 *                 delete_success = self.c_delete_limit_order(orders_map, address(map_it), orders_it)
 *                 cancellation_results.append(CancellationResult(limit_order_cid,             # <<<<<<<<<<<<<<
 *                                                                delete_success))
 *                 self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,
 */
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_CancellationResult); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 991, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_10);

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":992
 *                 delete_success = self.c_delete_limit_order(orders_map, address(map_it), orders_it)
 *                 cancellation_results.append(CancellationResult(limit_order_cid,
 *                                                                delete_success))             # <<<<<<<<<<<<<<
 *                 self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,
 *                                      OrderCancelledEvent(self._current_timestamp,
 */
        __pyx_t_11 = NULL;
        __pyx_t_12 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_11)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_11);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
            __pyx_t_12 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_11, __pyx_v_limit_order_cid, __pyx_v_delete_success};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_12, 2+__pyx_t_12);
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 991, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":991
 *                 limit_order_cid = limit_order_ptr.getClientOrderID().decode("utf8")
 *                 delete_success = self.c_delete_limit_order(orders_map, address(map_it), orders_it)
 *                 cancellation_results.append(CancellationResult(limit_order_cid,             # <<<<<<<<<<<<<<
 *                                                                delete_success))
 *                 self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,
 */
        __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_cancellation_results, __pyx_t_1); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 991, __pyx_L3_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":993
 *                 cancellation_results.append(CancellationResult(limit_order_cid,
 *                                                                delete_success))
 *                 self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,             # <<<<<<<<<<<<<<
 *                                      OrderCancelledEvent(self._current_timestamp,
 *                                                          limit_order_cid)
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_MARKET_ORDER_CANCELED_EVENT_TAG); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 993, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_14 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_14 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 993, __pyx_L3_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":994
 *                                                                delete_success))
 *                 self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,
 *                                      OrderCancelledEvent(self._current_timestamp,             # <<<<<<<<<<<<<<
 *                                                          limit_order_cid)
 *                                      )
 */
        __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_OrderCancelledEvent); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 994, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 994, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_11);

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":995
 *                 self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,
 *                                      OrderCancelledEvent(self._current_timestamp,
 *                                                          limit_order_cid)             # <<<<<<<<<<<<<<
 *                                      )
 *             return cancellation_results
 */
        __pyx_t_15 = NULL;
        __pyx_t_12 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_10);
          if (likely(__pyx_t_15)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
            __Pyx_INCREF(__pyx_t_15);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_10, function);
            __pyx_t_12 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_15, __pyx_t_11, __pyx_v_limit_order_cid};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_12, 2+__pyx_t_12);
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 994, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":993
 *                 cancellation_results.append(CancellationResult(limit_order_cid,
 *                                                                delete_success))
 *                 self.c_trigger_event(self.MARKET_ORDER_CANCELED_EVENT_TAG,             # <<<<<<<<<<<<<<
 *                                      OrderCancelledEvent(self._current_timestamp,
 *                                                          limit_order_cid)
 */
        __pyx_t_10 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.c_trigger_event(((struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *)__pyx_v_self), __pyx_t_14, __pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 993, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":987
 *                 inc(orders_it)
 * 
 *             for orders_it in process_order_its:             # <<<<<<<<<<<<<<
 *                 limit_order_ptr = address(deref(orders_it))
 *                 limit_order_cid = limit_order_ptr.getClientOrderID().decode("utf8")
 */
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":997
 *                                                          limit_order_cid)
 *                                      )
 *             return cancellation_results             # <<<<<<<<<<<<<<
 *         except Exception as err:
 *             self.logger().error(f"Error canceling order.", exc_info=True)
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_cancellation_results);
      __pyx_r = __pyx_v_cancellation_results;
      goto __pyx_L7_try_return;

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":974
 *             str limit_order_cid
 *             list cancellation_results = []
 *         try:             # <<<<<<<<<<<<<<
 *             if map_it == orders_map.end():
 *                 return []
 */
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":998
 *                                      )
 *             return cancellation_results
 *         except Exception as err:             # <<<<<<<<<<<<<<
 *             self.logger().error(f"Error canceling order.", exc_info=True)
 * 
 */
    __pyx_t_16 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_16) {
      __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_cancel_order_from_orders_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_10, &__pyx_t_1, &__pyx_t_11) < 0) __PYX_ERR(0, 998, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_1);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_v_err = __pyx_t_1;
      /*try:*/ {

        /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":999
 *             return cancellation_results
 *         except Exception as err:
 *             self.logger().error(f"Error canceling order.", exc_info=True)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_cancel(self, str trading_pair_str, str client_order_id):
 */
        __pyx_t_17 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 999, __pyx_L24_error)
        __Pyx_GOTREF(__pyx_t_17);
        __pyx_t_18 = NULL;
        __pyx_t_12 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_17))) {
          __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_17);
          if (likely(__pyx_t_18)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_17);
            __Pyx_INCREF(__pyx_t_18);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_17, function);
            __pyx_t_12 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_18, NULL};
          __pyx_t_15 = __Pyx_PyObject_FastCall(__pyx_t_17, __pyx_callargs+1-__pyx_t_12, 0+__pyx_t_12);
          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
          if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 999, __pyx_L24_error)
          __Pyx_GOTREF(__pyx_t_15);
          __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        }
        __pyx_t_17 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_error); if (unlikely(!__pyx_t_17)) __PYX_ERR(0, 999, __pyx_L24_error)
        __Pyx_GOTREF(__pyx_t_17);
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __pyx_t_15 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 999, __pyx_L24_error)
        __Pyx_GOTREF(__pyx_t_15);
        if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_exc_info, Py_True) < 0) __PYX_ERR(0, 999, __pyx_L24_error)
        __pyx_t_18 = __Pyx_PyObject_Call(__pyx_t_17, __pyx_tuple__25, __pyx_t_15); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 999, __pyx_L24_error)
        __Pyx_GOTREF(__pyx_t_18);
        __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
      }

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":998
 *                                      )
 *             return cancellation_results
 *         except Exception as err:             # <<<<<<<<<<<<<<
 *             self.logger().error(f"Error canceling order.", exc_info=True)
 * 
 */
      /*finally:*/ {
        /*normal exit:*/{
          __Pyx_DECREF(__pyx_v_err); __pyx_v_err = 0;
          goto __pyx_L25;
        }
        __pyx_L24_error:;
        /*exception exit:*/{
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0; __pyx_t_26 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
          if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_24, &__pyx_t_25, &__pyx_t_26);
          if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23) < 0)) __Pyx_ErrFetch(&__pyx_t_21, &__pyx_t_22, &__pyx_t_23);
          __Pyx_XGOTREF(__pyx_t_21);
          __Pyx_XGOTREF(__pyx_t_22);
          __Pyx_XGOTREF(__pyx_t_23);
          __Pyx_XGOTREF(__pyx_t_24);
          __Pyx_XGOTREF(__pyx_t_25);
          __Pyx_XGOTREF(__pyx_t_26);
          __pyx_t_16 = __pyx_lineno; __pyx_t_19 = __pyx_clineno; __pyx_t_20 = __pyx_filename;
          {
            __Pyx_DECREF(__pyx_v_err); __pyx_v_err = 0;
          }
          if (PY_MAJOR_VERSION >= 3) {
            __Pyx_XGIVEREF(__pyx_t_24);
            __Pyx_XGIVEREF(__pyx_t_25);
            __Pyx_XGIVEREF(__pyx_t_26);
            __Pyx_ExceptionReset(__pyx_t_24, __pyx_t_25, __pyx_t_26);
          }
          __Pyx_XGIVEREF(__pyx_t_21);
          __Pyx_XGIVEREF(__pyx_t_22);
          __Pyx_XGIVEREF(__pyx_t_23);
          __Pyx_ErrRestore(__pyx_t_21, __pyx_t_22, __pyx_t_23);
          __pyx_t_21 = 0; __pyx_t_22 = 0; __pyx_t_23 = 0; __pyx_t_24 = 0; __pyx_t_25 = 0; __pyx_t_26 = 0;
          __pyx_lineno = __pyx_t_16; __pyx_clineno = __pyx_t_19; __pyx_filename = __pyx_t_20;
          goto __pyx_L5_except_error;
        }
        __pyx_L25:;
      }
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":974
 *             str limit_order_cid
 *             list cancellation_results = []
 *         try:             # <<<<<<<<<<<<<<
 *             if map_it == orders_map.end():
 *                 return []
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L7_try_return:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L0;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":960
 *         return cancellation_results
 * 
 *     cdef object c_cancel_order_from_orders_map(self,             # <<<<<<<<<<<<<<
 *                                                LimitOrders *orders_map,
 *                                                str trading_pair_str,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_cancel_order_from_orders_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_limit_order_cid);
  __Pyx_XDECREF(__pyx_v_cancellation_results);
  __Pyx_XDECREF(__pyx_v_delete_success);
  __Pyx_XDECREF(__pyx_v_err);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1001
 *             self.logger().error(f"Error canceling order.", exc_info=True)
 * 
 *     cdef c_cancel(self, str trading_pair_str, str client_order_id):             # <<<<<<<<<<<<<<
 *         cdef:
 *             string cpp_trading_pair = trading_pair_str.encode("utf8")
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_cancel(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair_str, PyObject *__pyx_v_client_order_id) {
  CYTHON_UNUSED std::string __pyx_v_cpp_trading_pair;
  CYTHON_UNUSED std::string __pyx_v_cpp_client_order_id;
  PyObject *__pyx_v_trade_type = 0;
  int __pyx_v_is_maker_buy;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *__pyx_v_limit_orders_map_ptr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  std::string __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *__pyx_t_5;
  struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_cancel_order_from_orders_map __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_cancel", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1003
 *     cdef c_cancel(self, str trading_pair_str, str client_order_id):
 *         cdef:
 *             string cpp_trading_pair = trading_pair_str.encode("utf8")             # <<<<<<<<<<<<<<
 *             string cpp_client_order_id = client_order_id.encode("utf8")
 *             str trade_type = client_order_id.split("://")[0]
 */
  if (unlikely(__pyx_v_trading_pair_str == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 1003, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_AsUTF8String(__pyx_v_trading_pair_str); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_convert_string_from_py_6libcpp_6string_std__in_string(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1003, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_cpp_trading_pair = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_2);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1004
 *         cdef:
 *             string cpp_trading_pair = trading_pair_str.encode("utf8")
 *             string cpp_client_order_id = client_order_id.encode("utf8")             # <<<<<<<<<<<<<<
 *             str trade_type = client_order_id.split("://")[0]
 *             bint is_maker_buy = trade_type.upper() == "BUY"
 */
  if (unlikely(__pyx_v_client_order_id == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "encode");
    __PYX_ERR(0, 1004, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_AsUTF8String(__pyx_v_client_order_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_convert_string_from_py_6libcpp_6string_std__in_string(__pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1004, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_cpp_client_order_id = __PYX_STD_MOVE_IF_SUPPORTED(__pyx_t_2);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1005
 *             string cpp_trading_pair = trading_pair_str.encode("utf8")
 *             string cpp_client_order_id = client_order_id.encode("utf8")
 *             str trade_type = client_order_id.split("://")[0]             # <<<<<<<<<<<<<<
 *             bint is_maker_buy = trade_type.upper() == "BUY"
 *             LimitOrders *limit_orders_map_ptr = (address(self._bid_limit_orders)
 */
  if (unlikely(__pyx_v_client_order_id == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "split");
    __PYX_ERR(0, 1005, __pyx_L1_error)
  }
  __pyx_t_1 = PyUnicode_Split(__pyx_v_client_order_id, __Pyx_NoneAsNull(__pyx_kp_u__9), -1L); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_GetItemInt_List(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 1, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1005, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_3))) __PYX_ERR(0, 1005, __pyx_L1_error)
  __pyx_v_trade_type = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1006
 *             string cpp_client_order_id = client_order_id.encode("utf8")
 *             str trade_type = client_order_id.split("://")[0]
 *             bint is_maker_buy = trade_type.upper() == "BUY"             # <<<<<<<<<<<<<<
 *             LimitOrders *limit_orders_map_ptr = (address(self._bid_limit_orders)
 *                                                  if is_maker_buy
 */
  __pyx_t_3 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyUnicode_Type_upper, __pyx_v_trade_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_n_u_BUY, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_is_maker_buy = __pyx_t_4;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1008
 *             bint is_maker_buy = trade_type.upper() == "BUY"
 *             LimitOrders *limit_orders_map_ptr = (address(self._bid_limit_orders)
 *                                                  if is_maker_buy             # <<<<<<<<<<<<<<
 *                                                  else address(self._ask_limit_orders))
 *         self.c_cancel_order_from_orders_map(limit_orders_map_ptr, trading_pair_str, False, client_order_id)
 */
  if (__pyx_v_is_maker_buy) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1007
 *             str trade_type = client_order_id.split("://")[0]
 *             bint is_maker_buy = trade_type.upper() == "BUY"
 *             LimitOrders *limit_orders_map_ptr = (address(self._bid_limit_orders)             # <<<<<<<<<<<<<<
 *                                                  if is_maker_buy
 *                                                  else address(self._ask_limit_orders))
 */
    __pyx_t_5 = (&__pyx_v_self->_bid_limit_orders);
  } else {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1009
 *             LimitOrders *limit_orders_map_ptr = (address(self._bid_limit_orders)
 *                                                  if is_maker_buy
 *                                                  else address(self._ask_limit_orders))             # <<<<<<<<<<<<<<
 *         self.c_cancel_order_from_orders_map(limit_orders_map_ptr, trading_pair_str, False, client_order_id)
 * 
 */
    __pyx_t_5 = (&__pyx_v_self->_ask_limit_orders);
  }
  __pyx_v_limit_orders_map_ptr = __pyx_t_5;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1010
 *                                                  if is_maker_buy
 *                                                  else address(self._ask_limit_orders))
 *         self.c_cancel_order_from_orders_map(limit_orders_map_ptr, trading_pair_str, False, client_order_id)             # <<<<<<<<<<<<<<
 * 
 *     cdef object c_get_fee(self,
 */
  __pyx_t_6.__pyx_n = 2;
  __pyx_t_6.cancel_all = 0;
  __pyx_t_6.client_order_id = __pyx_v_client_order_id;
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_cancel_order_from_orders_map(__pyx_v_self, __pyx_v_limit_orders_map_ptr, __pyx_v_trading_pair_str, &__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1001
 *             self.logger().error(f"Error canceling order.", exc_info=True)
 * 
 *     cdef c_cancel(self, str trading_pair_str, str client_order_id):             # <<<<<<<<<<<<<<
 *         cdef:
 *             string cpp_trading_pair = trading_pair_str.encode("utf8")
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_trade_type);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1012
 *         self.c_cancel_order_from_orders_map(limit_orders_map_ptr, trading_pair_str, False, client_order_id)
 * 
 *     cdef object c_get_fee(self,             # <<<<<<<<<<<<<<
 *                           str base_asset,
 *                           str quote_asset,
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_fee(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_base_asset, PyObject *__pyx_v_quote_asset, PyObject *__pyx_v_order_type, PyObject *__pyx_v_order_side, PyObject *__pyx_v_amount, PyObject *__pyx_v_price, struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_fee *__pyx_optional_args) {

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1019
 *                           object amount,
 *                           object price,
 *                           object is_maker = None):             # <<<<<<<<<<<<<<
 *         return build_trade_fee(
 *             self.name,
 */
  PyObject *__pyx_v_is_maker = ((PyObject *)Py_None);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_get_fee", 1);
  if (__pyx_optional_args) {
    if (__pyx_optional_args->__pyx_n > 0) {
      __pyx_v_is_maker = __pyx_optional_args->is_maker;
    }
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1020
 *                           object price,
 *                           object is_maker = None):
 *         return build_trade_fee(             # <<<<<<<<<<<<<<
 *             self.name,
 *             is_maker=is_maker if is_maker is not None else order_type in [OrderType.LIMIT, OrderType.LIMIT_MAKER],
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_build_trade_fee); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1021
 *                           object is_maker = None):
 *         return build_trade_fee(
 *             self.name,             # <<<<<<<<<<<<<<
 *             is_maker=is_maker if is_maker is not None else order_type in [OrderType.LIMIT, OrderType.LIMIT_MAKER],
 *             base_currency=base_asset,
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1021, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1020
 *                           object price,
 *                           object is_maker = None):
 *         return build_trade_fee(             # <<<<<<<<<<<<<<
 *             self.name,
 *             is_maker=is_maker if is_maker is not None else order_type in [OrderType.LIMIT, OrderType.LIMIT_MAKER],
 */
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(0, 1020, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1022
 *         return build_trade_fee(
 *             self.name,
 *             is_maker=is_maker if is_maker is not None else order_type in [OrderType.LIMIT, OrderType.LIMIT_MAKER],             # <<<<<<<<<<<<<<
 *             base_currency=base_asset,
 *             quote_currency=quote_asset,
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1022, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = (__pyx_v_is_maker != Py_None);
  if (__pyx_t_5) {
    __Pyx_INCREF(__pyx_v_is_maker);
    __pyx_t_4 = __pyx_v_is_maker;
  } else {
    __Pyx_INCREF(__pyx_v_order_type);
    __pyx_t_6 = __pyx_v_order_type;
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_LIMIT); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyObject_RichCompare(__pyx_t_6, __pyx_t_9, Py_EQ); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (!__pyx_t_10) {
    } else {
      __pyx_t_7 = __pyx_t_10;
      goto __pyx_L3_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_8, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_LIMIT_MAKER); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = PyObject_RichCompare(__pyx_t_6, __pyx_t_9, Py_EQ); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_7 = __pyx_t_10;
    __pyx_L3_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyBool_FromLong(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_4 = __pyx_t_6;
    __pyx_t_6 = 0;
  }
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_is_maker, __pyx_t_4) < 0) __PYX_ERR(0, 1022, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1023
 *             self.name,
 *             is_maker=is_maker if is_maker is not None else order_type in [OrderType.LIMIT, OrderType.LIMIT_MAKER],
 *             base_currency=base_asset,             # <<<<<<<<<<<<<<
 *             quote_currency=quote_asset,
 *             order_type=order_type,
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_base_currency, __pyx_v_base_asset) < 0) __PYX_ERR(0, 1022, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1024
 *             is_maker=is_maker if is_maker is not None else order_type in [OrderType.LIMIT, OrderType.LIMIT_MAKER],
 *             base_currency=base_asset,
 *             quote_currency=quote_asset,             # <<<<<<<<<<<<<<
 *             order_type=order_type,
 *             order_side=order_side,
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_quote_currency, __pyx_v_quote_asset) < 0) __PYX_ERR(0, 1022, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1025
 *             base_currency=base_asset,
 *             quote_currency=quote_asset,
 *             order_type=order_type,             # <<<<<<<<<<<<<<
 *             order_side=order_side,
 *             amount=amount,
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_order_type, __pyx_v_order_type) < 0) __PYX_ERR(0, 1022, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1026
 *             quote_currency=quote_asset,
 *             order_type=order_type,
 *             order_side=order_side,             # <<<<<<<<<<<<<<
 *             amount=amount,
 *             price=price,
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_order_side, __pyx_v_order_side) < 0) __PYX_ERR(0, 1022, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1027
 *             order_type=order_type,
 *             order_side=order_side,
 *             amount=amount,             # <<<<<<<<<<<<<<
 *             price=price,
 *         )
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_amount, __pyx_v_amount) < 0) __PYX_ERR(0, 1022, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1028
 *             order_side=order_side,
 *             amount=amount,
 *             price=price,             # <<<<<<<<<<<<<<
 *         )
 * 
 */
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_price, __pyx_v_price) < 0) __PYX_ERR(0, 1022, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1020
 *                           object price,
 *                           object is_maker = None):
 *         return build_trade_fee(             # <<<<<<<<<<<<<<
 *             self.name,
 *             is_maker=is_maker if is_maker is not None else order_type in [OrderType.LIMIT, OrderType.LIMIT_MAKER],
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1020, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1012
 *         self.c_cancel_order_from_orders_map(limit_orders_map_ptr, trading_pair_str, False, client_order_id)
 * 
 *     cdef object c_get_fee(self,             # <<<<<<<<<<<<<<
 *                           str base_asset,
 *                           str quote_asset,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_get_fee", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1031
 *         )
 * 
 *     cdef OrderBook c_get_order_book(self, str trading_pair):             # <<<<<<<<<<<<<<
 *         if trading_pair not in self._trading_pairs:
 *             raise ValueError(f"No order book exists for '{trading_pair}'.")
 */

static struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_order_book(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair) {
  struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_UCS4 __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_get_order_book", 0);
  __Pyx_INCREF(__pyx_v_trading_pair);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1032
 * 
 *     cdef OrderBook c_get_order_book(self, str trading_pair):
 *         if trading_pair not in self._trading_pairs:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"No order book exists for '{trading_pair}'.")
 *         trading_pair = self._target_market.convert_to_exchange_trading_pair(trading_pair)
 */
  if (unlikely(__pyx_v_self->_trading_pairs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 1032, __pyx_L1_error)
  }
  __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_v_trading_pair, __pyx_v_self->_trading_pairs, Py_NE)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1032, __pyx_L1_error)
  if (unlikely(__pyx_t_1)) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1033
 *     cdef OrderBook c_get_order_book(self, str trading_pair):
 *         if trading_pair not in self._trading_pairs:
 *             raise ValueError(f"No order book exists for '{trading_pair}'.")             # <<<<<<<<<<<<<<
 *         trading_pair = self._target_market.convert_to_exchange_trading_pair(trading_pair)
 *         return self._order_book_tracker.order_books[trading_pair]
 */
    __pyx_t_2 = PyTuple_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1033, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = 127;
    __Pyx_INCREF(__pyx_kp_u_No_order_book_exists_for);
    __pyx_t_3 += 26;
    __Pyx_GIVEREF(__pyx_kp_u_No_order_book_exists_for);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_No_order_book_exists_for);
    __pyx_t_5 = __Pyx_PyUnicode_Unicode(__pyx_v_trading_pair); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1033, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_4) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_4;
    __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u__26);
    __pyx_t_3 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__26);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u__26);
    __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_2, 3, __pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1033, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1033, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1033, __pyx_L1_error)

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1032
 * 
 *     cdef OrderBook c_get_order_book(self, str trading_pair):
 *         if trading_pair not in self._trading_pairs:             # <<<<<<<<<<<<<<
 *             raise ValueError(f"No order book exists for '{trading_pair}'.")
 *         trading_pair = self._target_market.convert_to_exchange_trading_pair(trading_pair)
 */
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1034
 *         if trading_pair not in self._trading_pairs:
 *             raise ValueError(f"No order book exists for '{trading_pair}'.")
 *         trading_pair = self._target_market.convert_to_exchange_trading_pair(trading_pair)             # <<<<<<<<<<<<<<
 *         return self._order_book_tracker.order_books[trading_pair]
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_target_market, __pyx_n_s_convert_to_exchange_trading_pair); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = NULL;
  __pyx_t_7 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_7 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_trading_pair};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1034, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_2))) __PYX_ERR(0, 1034, __pyx_L1_error)
  __Pyx_DECREF_SET(__pyx_v_trading_pair, ((PyObject*)__pyx_t_2));
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1035
 *             raise ValueError(f"No order book exists for '{trading_pair}'.")
 *         trading_pair = self._target_market.convert_to_exchange_trading_pair(trading_pair)
 *         return self._order_book_tracker.order_books[trading_pair]             # <<<<<<<<<<<<<<
 * 
 *     cdef object c_get_order_price_quantum(self, str trading_pair, object price):
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->__pyx_base._order_book_tracker, __pyx_n_s_order_books); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_2, __pyx_v_trading_pair); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1035, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!(likely(((__pyx_t_5) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_5, __pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook))))) __PYX_ERR(0, 1035, __pyx_L1_error)
  __pyx_r = ((struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *)__pyx_t_5);
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1031
 *         )
 * 
 *     cdef OrderBook c_get_order_book(self, str trading_pair):             # <<<<<<<<<<<<<<
 *         if trading_pair not in self._trading_pairs:
 *             raise ValueError(f"No order book exists for '{trading_pair}'.")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_get_order_book", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_trading_pair);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1037
 *         return self._order_book_tracker.order_books[trading_pair]
 * 
 *     cdef object c_get_order_price_quantum(self, str trading_pair, object price):             # <<<<<<<<<<<<<<
 *         cdef:
 *             QuantizationParams q_params
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_order_price_quantum(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_price) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *__pyx_v_q_params = 0;
  PyObject *__pyx_v_decimals_quantum = NULL;
  PyObject *__pyx_v_precision_quantum = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_get_order_price_quantum", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1040
 *         cdef:
 *             QuantizationParams q_params
 *         if trading_pair in self._quantization_params:             # <<<<<<<<<<<<<<
 *             q_params = self._quantization_params[trading_pair]
 *             decimals_quantum = Decimal(f"1e-{q_params.price_decimals}")
 */
  if (unlikely(__pyx_v_self->_quantization_params == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 1040, __pyx_L1_error)
  }
  __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_v_trading_pair, __pyx_v_self->_quantization_params, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1040, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1041
 *             QuantizationParams q_params
 *         if trading_pair in self._quantization_params:
 *             q_params = self._quantization_params[trading_pair]             # <<<<<<<<<<<<<<
 *             decimals_quantum = Decimal(f"1e-{q_params.price_decimals}")
 *             if price.is_finite() and price > s_decimal_0:
 */
    if (unlikely(__pyx_v_self->_quantization_params == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 1041, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_self->_quantization_params, __pyx_v_trading_pair); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams))))) __PYX_ERR(0, 1041, __pyx_L1_error)
    __pyx_v_q_params = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1042
 *         if trading_pair in self._quantization_params:
 *             q_params = self._quantization_params[trading_pair]
 *             decimals_quantum = Decimal(f"1e-{q_params.price_decimals}")             # <<<<<<<<<<<<<<
 *             if price.is_finite() and price > s_decimal_0:
 *                 precision_quantum = Decimal(f"1e{math.ceil(math.log10(price)) - q_params.price_precision}")
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1042, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyUnicode_From_int(__pyx_v_q_params->price_decimals, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1042, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyUnicode_Concat(__pyx_kp_u_1e, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1042, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1042, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_v_decimals_quantum = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1043
 *             q_params = self._quantization_params[trading_pair]
 *             decimals_quantum = Decimal(f"1e-{q_params.price_decimals}")
 *             if price.is_finite() and price > s_decimal_0:             # <<<<<<<<<<<<<<
 *                 precision_quantum = Decimal(f"1e{math.ceil(math.log10(price)) - q_params.price_precision}")
 *             else:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_price, __pyx_n_s_is_finite); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1043, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1043, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_7) {
    } else {
      __pyx_t_1 = __pyx_t_7;
      goto __pyx_L5_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_s_decimal_0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_price, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1043, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1043, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = __pyx_t_7;
    __pyx_L5_bool_binop_done:;
    if (__pyx_t_1) {

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1044
 *             decimals_quantum = Decimal(f"1e-{q_params.price_decimals}")
 *             if price.is_finite() and price > s_decimal_0:
 *                 precision_quantum = Decimal(f"1e{math.ceil(math.log10(price)) - q_params.price_precision}")             # <<<<<<<<<<<<<<
 *             else:
 *                 precision_quantum = Decimal(0)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1044, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_math); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1044, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ceil); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1044, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_math); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1044, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_log10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1044, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_v_price};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1044, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __pyx_t_10 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_4};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1044, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_q_params->price_precision); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1044, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = PyNumber_Subtract(__pyx_t_5, __pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1044, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1044, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyUnicode_Concat(__pyx_kp_u_1e_2, __pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1044, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_4};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1044, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __pyx_v_precision_quantum = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1043
 *             q_params = self._quantization_params[trading_pair]
 *             decimals_quantum = Decimal(f"1e-{q_params.price_decimals}")
 *             if price.is_finite() and price > s_decimal_0:             # <<<<<<<<<<<<<<
 *                 precision_quantum = Decimal(f"1e{math.ceil(math.log10(price)) - q_params.price_precision}")
 *             else:
 */
      goto __pyx_L4;
    }

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1046
 *                 precision_quantum = Decimal(f"1e{math.ceil(math.log10(price)) - q_params.price_precision}")
 *             else:
 *                 precision_quantum = Decimal(0)             # <<<<<<<<<<<<<<
 *             return max(precision_quantum, decimals_quantum)
 *         else:
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1046, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_int_0};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1046, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __pyx_v_precision_quantum = __pyx_t_3;
      __pyx_t_3 = 0;
    }
    __pyx_L4:;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1047
 *             else:
 *                 precision_quantum = Decimal(0)
 *             return max(precision_quantum, decimals_quantum)             # <<<<<<<<<<<<<<
 *         else:
 *             return Decimal(f"1e-10")
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_decimals_quantum);
    __pyx_t_3 = __pyx_v_decimals_quantum;
    __Pyx_INCREF(__pyx_v_precision_quantum);
    __pyx_t_2 = __pyx_v_precision_quantum;
    __pyx_t_8 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1047, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1047, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_1) {
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = __pyx_t_3;
    } else {
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_4 = __pyx_t_2;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1040
 *         cdef:
 *             QuantizationParams q_params
 *         if trading_pair in self._quantization_params:             # <<<<<<<<<<<<<<
 *             q_params = self._quantization_params[trading_pair]
 *             decimals_quantum = Decimal(f"1e-{q_params.price_decimals}")
 */
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1049
 *             return max(precision_quantum, decimals_quantum)
 *         else:
 *             return Decimal(f"1e-10")             # <<<<<<<<<<<<<<
 * 
 *     def get_order_price_quantum(self, trading_pair: str, price: Decimal) -> Decimal:
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1049, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_kp_u_1e_10};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1037
 *         return self._order_book_tracker.order_books[trading_pair]
 * 
 *     cdef object c_get_order_price_quantum(self, str trading_pair, object price):             # <<<<<<<<<<<<<<
 *         cdef:
 *             QuantizationParams q_params
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_get_order_price_quantum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_q_params);
  __Pyx_XDECREF(__pyx_v_decimals_quantum);
  __Pyx_XDECREF(__pyx_v_precision_quantum);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1051
 *             return Decimal(f"1e-10")
 * 
 *     def get_order_price_quantum(self, trading_pair: str, price: Decimal) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_get_order_price_quantum(trading_pair, price)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_21get_order_price_quantum(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_21get_order_price_quantum = {"get_order_price_quantum", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_21get_order_price_quantum, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_21get_order_price_quantum(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_trading_pair = 0;
  PyObject *__pyx_v_price = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_order_price_quantum (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_trading_pair,&__pyx_n_s_price,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_trading_pair)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1051, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_price)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1051, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("get_order_price_quantum", 1, 2, 2, 1); __PYX_ERR(0, 1051, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_order_price_quantum") < 0)) __PYX_ERR(0, 1051, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_trading_pair = values[0];
    __pyx_v_price = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_order_price_quantum", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1051, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.get_order_price_quantum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_20get_order_price_quantum(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self), __pyx_v_trading_pair, __pyx_v_price);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_20get_order_price_quantum(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_price) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_order_price_quantum", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1052
 * 
 *     def get_order_price_quantum(self, trading_pair: str, price: Decimal) -> Decimal:
 *         return self.c_get_order_price_quantum(trading_pair, price)             # <<<<<<<<<<<<<<
 * 
 *     cdef object c_get_order_size_quantum(self,
 */
  __Pyx_XDECREF(__pyx_r);
  if (!(likely(PyUnicode_CheckExact(__pyx_v_trading_pair))||((__pyx_v_trading_pair) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_trading_pair))) __PYX_ERR(0, 1052, __pyx_L1_error)
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_get_order_price_quantum(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), ((PyObject*)__pyx_v_trading_pair), __pyx_v_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1052, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1051
 *             return Decimal(f"1e-10")
 * 
 *     def get_order_price_quantum(self, trading_pair: str, price: Decimal) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_get_order_price_quantum(trading_pair, price)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.get_order_price_quantum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1054
 *         return self.c_get_order_price_quantum(trading_pair, price)
 * 
 *     cdef object c_get_order_size_quantum(self,             # <<<<<<<<<<<<<<
 *                                          str trading_pair,
 *                                          object order_size):
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_order_size_quantum(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_order_size) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *__pyx_v_q_params = 0;
  PyObject *__pyx_v_decimals_quantum = NULL;
  PyObject *__pyx_v_precision_quantum = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_get_order_size_quantum", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1059
 *         cdef:
 *             QuantizationParams q_params
 *         if trading_pair in self._quantization_params:             # <<<<<<<<<<<<<<
 *             q_params = self._quantization_params[trading_pair]
 *             decimals_quantum = Decimal(f"1e-{q_params.order_size_decimals}")
 */
  if (unlikely(__pyx_v_self->_quantization_params == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 1059, __pyx_L1_error)
  }
  __pyx_t_1 = (__Pyx_PyDict_ContainsTF(__pyx_v_trading_pair, __pyx_v_self->_quantization_params, Py_EQ)); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1059, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1060
 *             QuantizationParams q_params
 *         if trading_pair in self._quantization_params:
 *             q_params = self._quantization_params[trading_pair]             # <<<<<<<<<<<<<<
 *             decimals_quantum = Decimal(f"1e-{q_params.order_size_decimals}")
 *             if order_size.is_finite() and order_size > s_decimal_0:
 */
    if (unlikely(__pyx_v_self->_quantization_params == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 1060, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_PyDict_GetItem(__pyx_v_self->_quantization_params, __pyx_v_trading_pair); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1060, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams))))) __PYX_ERR(0, 1060, __pyx_L1_error)
    __pyx_v_q_params = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1061
 *         if trading_pair in self._quantization_params:
 *             q_params = self._quantization_params[trading_pair]
 *             decimals_quantum = Decimal(f"1e-{q_params.order_size_decimals}")             # <<<<<<<<<<<<<<
 *             if order_size.is_finite() and order_size > s_decimal_0:
 *                 precision_quantum = Decimal(f"1e{math.ceil(math.log10(order_size)) - q_params.order_size_precision}")
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyUnicode_From_int(__pyx_v_q_params->order_size_decimals, 0, ' ', 'd'); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyUnicode_Concat(__pyx_kp_u_1e, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_5};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1061, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_v_decimals_quantum = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1062
 *             q_params = self._quantization_params[trading_pair]
 *             decimals_quantum = Decimal(f"1e-{q_params.order_size_decimals}")
 *             if order_size.is_finite() and order_size > s_decimal_0:             # <<<<<<<<<<<<<<
 *                 precision_quantum = Decimal(f"1e{math.ceil(math.log10(order_size)) - q_params.order_size_precision}")
 *             else:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_size, __pyx_n_s_is_finite); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1062, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1062, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1062, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_7) {
    } else {
      __pyx_t_1 = __pyx_t_7;
      goto __pyx_L5_bool_binop_done;
    }
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_s_decimal_0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1062, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_order_size, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1062, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1062, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = __pyx_t_7;
    __pyx_L5_bool_binop_done:;
    if (__pyx_t_1) {

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1063
 *             decimals_quantum = Decimal(f"1e-{q_params.order_size_decimals}")
 *             if order_size.is_finite() and order_size > s_decimal_0:
 *                 precision_quantum = Decimal(f"1e{math.ceil(math.log10(order_size)) - q_params.order_size_precision}")             # <<<<<<<<<<<<<<
 *             else:
 *                 precision_quantum = Decimal(0)
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1063, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_math); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1063, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_ceil); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1063, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_math); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1063, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_log10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1063, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_v_order_size};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1063, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __pyx_t_10 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_4};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1063, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_q_params->order_size_precision); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1063, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_4 = PyNumber_Subtract(__pyx_t_5, __pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1063, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1063, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyUnicode_Concat(__pyx_kp_u_1e_2, __pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1063, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_4};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1063, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __pyx_v_precision_quantum = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1062
 *             q_params = self._quantization_params[trading_pair]
 *             decimals_quantum = Decimal(f"1e-{q_params.order_size_decimals}")
 *             if order_size.is_finite() and order_size > s_decimal_0:             # <<<<<<<<<<<<<<
 *                 precision_quantum = Decimal(f"1e{math.ceil(math.log10(order_size)) - q_params.order_size_precision}")
 *             else:
 */
      goto __pyx_L4;
    }

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1065
 *                 precision_quantum = Decimal(f"1e{math.ceil(math.log10(order_size)) - q_params.order_size_precision}")
 *             else:
 *                 precision_quantum = Decimal(0)             # <<<<<<<<<<<<<<
 *             return max(precision_quantum, decimals_quantum)
 *         else:
 */
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_int_0};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1065, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __pyx_v_precision_quantum = __pyx_t_3;
      __pyx_t_3 = 0;
    }
    __pyx_L4:;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1066
 *             else:
 *                 precision_quantum = Decimal(0)
 *             return max(precision_quantum, decimals_quantum)             # <<<<<<<<<<<<<<
 *         else:
 *             return Decimal(f"1e-7")
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_decimals_quantum);
    __pyx_t_3 = __pyx_v_decimals_quantum;
    __Pyx_INCREF(__pyx_v_precision_quantum);
    __pyx_t_2 = __pyx_v_precision_quantum;
    __pyx_t_8 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1066, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1066, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_1) {
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = __pyx_t_3;
    } else {
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_4 = __pyx_t_2;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_r = __pyx_t_4;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1059
 *         cdef:
 *             QuantizationParams q_params
 *         if trading_pair in self._quantization_params:             # <<<<<<<<<<<<<<
 *             q_params = self._quantization_params[trading_pair]
 *             decimals_quantum = Decimal(f"1e-{q_params.order_size_decimals}")
 */
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1068
 *             return max(precision_quantum, decimals_quantum)
 *         else:
 *             return Decimal(f"1e-7")             # <<<<<<<<<<<<<<
 * 
 *     cdef object c_quantize_order_price(self,
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1068, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_kp_u_1e_7};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1068, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1054
 *         return self.c_get_order_price_quantum(trading_pair, price)
 * 
 *     cdef object c_get_order_size_quantum(self,             # <<<<<<<<<<<<<<
 *                                          str trading_pair,
 *                                          object order_size):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_get_order_size_quantum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_q_params);
  __Pyx_XDECREF(__pyx_v_decimals_quantum);
  __Pyx_XDECREF(__pyx_v_precision_quantum);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1070
 *             return Decimal(f"1e-7")
 * 
 *     cdef object c_quantize_order_price(self,             # <<<<<<<<<<<<<<
 *                                        str trading_pair,
 *                                        object price):
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_quantize_order_price(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_price) {
  PyObject *__pyx_v_price_quantum = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_quantize_order_price", 0);
  __Pyx_INCREF(__pyx_v_price);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1073
 *                                        str trading_pair,
 *                                        object price):
 *         price = Decimal('%.7g' % price)  # hard code to round to 8 significant digits             # <<<<<<<<<<<<<<
 *         price_quantum = self.c_get_order_price_quantum(trading_pair, price)
 *         return (price // price_quantum) * price_quantum
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyUnicode_FormatSafe(__pyx_kp_u_7g, __pyx_v_price); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF_SET(__pyx_v_price, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1074
 *                                        object price):
 *         price = Decimal('%.7g' % price)  # hard code to round to 8 significant digits
 *         price_quantum = self.c_get_order_price_quantum(trading_pair, price)             # <<<<<<<<<<<<<<
 *         return (price // price_quantum) * price_quantum
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_get_order_price_quantum(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), __pyx_v_trading_pair, __pyx_v_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1074, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_price_quantum = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1075
 *         price = Decimal('%.7g' % price)  # hard code to round to 8 significant digits
 *         price_quantum = self.c_get_order_price_quantum(trading_pair, price)
 *         return (price // price_quantum) * price_quantum             # <<<<<<<<<<<<<<
 * 
 *     cdef object c_quantize_order_amount(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyNumber_FloorDivide(__pyx_v_price, __pyx_v_price_quantum); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Multiply(__pyx_t_1, __pyx_v_price_quantum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1075, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1070
 *             return Decimal(f"1e-7")
 * 
 *     cdef object c_quantize_order_price(self,             # <<<<<<<<<<<<<<
 *                                        str trading_pair,
 *                                        object price):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_quantize_order_price", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_price_quantum);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1077
 *         return (price // price_quantum) * price_quantum
 * 
 *     cdef object c_quantize_order_amount(self,             # <<<<<<<<<<<<<<
 *                                         str trading_pair,
 *                                         object amount,
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_quantize_order_amount(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_amount, struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_quantize_order_amount *__pyx_optional_args) {
  PyObject *__pyx_v_order_size_quantum = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_quantize_order_amount", 0);
  if (__pyx_optional_args) {
  }
  __Pyx_INCREF(__pyx_v_amount);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1081
 *                                         object amount,
 *                                         object price=s_decimal_0):
 *         amount = amount.quantize(Decimal('1e-7'), rounding=ROUND_DOWN)             # <<<<<<<<<<<<<<
 *         if amount <= 1e-7:
 *             amount = Decimal("0")
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_amount, __pyx_n_s_quantize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_u_1e_7};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1081, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(0, 1081, __pyx_L1_error);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_ROUND_DOWN); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_rounding, __pyx_t_4) < 0) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1081, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_amount, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1082
 *                                         object price=s_decimal_0):
 *         amount = amount.quantize(Decimal('1e-7'), rounding=ROUND_DOWN)
 *         if amount <= 1e-7:             # <<<<<<<<<<<<<<
 *             amount = Decimal("0")
 *         order_size_quantum = self.c_get_order_size_quantum(trading_pair, amount)
 */
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_amount, __pyx_float_1eneg_7, Py_LE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1082, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 1082, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_6) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1083
 *         amount = amount.quantize(Decimal('1e-7'), rounding=ROUND_DOWN)
 *         if amount <= 1e-7:
 *             amount = Decimal("0")             # <<<<<<<<<<<<<<
 *         order_size_quantum = self.c_get_order_size_quantum(trading_pair, amount)
 *         return (amount // order_size_quantum) * order_size_quantum
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1083, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_u_0};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1083, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF_SET(__pyx_v_amount, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1082
 *                                         object price=s_decimal_0):
 *         amount = amount.quantize(Decimal('1e-7'), rounding=ROUND_DOWN)
 *         if amount <= 1e-7:             # <<<<<<<<<<<<<<
 *             amount = Decimal("0")
 *         order_size_quantum = self.c_get_order_size_quantum(trading_pair, amount)
 */
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1084
 *         if amount <= 1e-7:
 *             amount = Decimal("0")
 *         order_size_quantum = self.c_get_order_size_quantum(trading_pair, amount)             # <<<<<<<<<<<<<<
 *         return (amount // order_size_quantum) * order_size_quantum
 * 
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_get_order_size_quantum(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), __pyx_v_trading_pair, __pyx_v_amount); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_order_size_quantum = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1085
 *             amount = Decimal("0")
 *         order_size_quantum = self.c_get_order_size_quantum(trading_pair, amount)
 *         return (amount // order_size_quantum) * order_size_quantum             # <<<<<<<<<<<<<<
 * 
 *     def get_available_balance(self, currency: str) -> Decimal:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyNumber_FloorDivide(__pyx_v_amount, __pyx_v_order_size_quantum); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1085, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = PyNumber_Multiply(__pyx_t_4, __pyx_v_order_size_quantum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1085, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1077
 *         return (price // price_quantum) * price_quantum
 * 
 *     cdef object c_quantize_order_amount(self,             # <<<<<<<<<<<<<<
 *                                         str trading_pair,
 *                                         object amount,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.c_quantize_order_amount", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_order_size_quantum);
  __Pyx_XDECREF(__pyx_v_amount);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1087
 *         return (amount // order_size_quantum) * order_size_quantum
 * 
 *     def get_available_balance(self, currency: str) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_get_available_balance(currency)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_23get_available_balance(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_23get_available_balance = {"get_available_balance", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_23get_available_balance, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_23get_available_balance(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_currency = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_available_balance (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_currency,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_currency)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1087, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_available_balance") < 0)) __PYX_ERR(0, 1087, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_currency = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_available_balance", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1087, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.get_available_balance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_22get_available_balance(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self), __pyx_v_currency);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_22get_available_balance(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_currency) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_available_balance", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1088
 * 
 *     def get_available_balance(self, currency: str) -> Decimal:
 *         return self.c_get_available_balance(currency)             # <<<<<<<<<<<<<<
 * 
 *     def get_all_balances(self) -> Dict[str, Decimal]:
 */
  __Pyx_XDECREF(__pyx_r);
  if (!(likely(PyUnicode_CheckExact(__pyx_v_currency))||((__pyx_v_currency) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_currency))) __PYX_ERR(0, 1088, __pyx_L1_error)
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_get_available_balance(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), ((PyObject*)__pyx_v_currency)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1088, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1087
 *         return (amount // order_size_quantum) * order_size_quantum
 * 
 *     def get_available_balance(self, currency: str) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_get_available_balance(currency)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.get_available_balance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1090
 *         return self.c_get_available_balance(currency)
 * 
 *     def get_all_balances(self) -> Dict[str, Decimal]:             # <<<<<<<<<<<<<<
 *         return self._account_balances.copy()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_25get_all_balances(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_25get_all_balances = {"get_all_balances", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_25get_all_balances, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_25get_all_balances(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_all_balances (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_all_balances", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_all_balances", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_24get_all_balances(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_24get_all_balances(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_all_balances", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1091
 * 
 *     def get_all_balances(self) -> Dict[str, Decimal]:
 *         return self._account_balances.copy()             # <<<<<<<<<<<<<<
 * 
 *     # <editor-fold desc="Python wrapper for cdef functions">
 */
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(__pyx_v_self->__pyx_base.__pyx_base._account_balances == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "copy");
    __PYX_ERR(0, 1091, __pyx_L1_error)
  }
  __pyx_t_1 = PyDict_Copy(__pyx_v_self->__pyx_base.__pyx_base._account_balances); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1091, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1090
 *         return self.c_get_available_balance(currency)
 * 
 *     def get_all_balances(self) -> Dict[str, Decimal]:             # <<<<<<<<<<<<<<
 *         return self._account_balances.copy()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.get_all_balances", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1094
 * 
 *     # <editor-fold desc="Python wrapper for cdef functions">
 *     def match_trade_to_limit_orders(self, event_object: OrderBookTradeEvent):             # <<<<<<<<<<<<<<
 *         self.c_match_trade_to_limit_orders(event_object)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_27match_trade_to_limit_orders(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_27match_trade_to_limit_orders = {"match_trade_to_limit_orders", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_27match_trade_to_limit_orders, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_27match_trade_to_limit_orders(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_event_object = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("match_trade_to_limit_orders (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_event_object,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_event_object)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1094, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "match_trade_to_limit_orders") < 0)) __PYX_ERR(0, 1094, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_event_object = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("match_trade_to_limit_orders", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1094, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.match_trade_to_limit_orders", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_26match_trade_to_limit_orders(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self), __pyx_v_event_object);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_26match_trade_to_limit_orders(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_event_object) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("match_trade_to_limit_orders", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1095
 *     # <editor-fold desc="Python wrapper for cdef functions">
 *     def match_trade_to_limit_orders(self, event_object: OrderBookTradeEvent):
 *         self.c_match_trade_to_limit_orders(event_object)             # <<<<<<<<<<<<<<
 * 
 *     def set_balance(self, currency: str, balance: Decimal):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_match_trade_to_limit_orders(__pyx_v_self, __pyx_v_event_object); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1095, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1094
 * 
 *     # <editor-fold desc="Python wrapper for cdef functions">
 *     def match_trade_to_limit_orders(self, event_object: OrderBookTradeEvent):             # <<<<<<<<<<<<<<
 *         self.c_match_trade_to_limit_orders(event_object)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.match_trade_to_limit_orders", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1097
 *         self.c_match_trade_to_limit_orders(event_object)
 * 
 *     def set_balance(self, currency: str, balance: Decimal):             # <<<<<<<<<<<<<<
 *         self.c_set_balance(currency, balance)
 *     # </editor-fold>
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_29set_balance(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_29set_balance = {"set_balance", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_29set_balance, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_29set_balance(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_currency = 0;
  PyObject *__pyx_v_balance = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_balance (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_currency,&__pyx_n_s_balance,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_currency)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1097, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_balance)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1097, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("set_balance", 1, 2, 2, 1); __PYX_ERR(0, 1097, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "set_balance") < 0)) __PYX_ERR(0, 1097, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_currency = values[0];
    __pyx_v_balance = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_balance", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1097, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.set_balance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_28set_balance(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self), __pyx_v_currency, __pyx_v_balance);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_28set_balance(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_currency, PyObject *__pyx_v_balance) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_balance", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1098
 * 
 *     def set_balance(self, currency: str, balance: Decimal):
 *         self.c_set_balance(currency, balance)             # <<<<<<<<<<<<<<
 *     # </editor-fold>
 * 
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_currency))||((__pyx_v_currency) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_currency))) __PYX_ERR(0, 1098, __pyx_L1_error)
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_set_balance(__pyx_v_self, ((PyObject*)__pyx_v_currency), __pyx_v_balance); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1098, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1097
 *         self.c_match_trade_to_limit_orders(event_object)
 * 
 *     def set_balance(self, currency: str, balance: Decimal):             # <<<<<<<<<<<<<<
 *         self.c_set_balance(currency, balance)
 *     # </editor-fold>
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.set_balance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1101
 *     # </editor-fold>
 * 
 *     def get_price(self, trading_pair: str, is_buy: bool) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_get_price(trading_pair, is_buy)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_31get_price(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_31get_price = {"get_price", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_31get_price, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_31get_price(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_trading_pair = 0;
  PyObject *__pyx_v_is_buy = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_price (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_trading_pair,&__pyx_n_s_is_buy,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_trading_pair)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1101, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_is_buy)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1101, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("get_price", 1, 2, 2, 1); __PYX_ERR(0, 1101, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_price") < 0)) __PYX_ERR(0, 1101, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_trading_pair = values[0];
    __pyx_v_is_buy = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_price", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1101, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.get_price", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_30get_price(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self), __pyx_v_trading_pair, __pyx_v_is_buy);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_30get_price(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_is_buy) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_price", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1102
 * 
 *     def get_price(self, trading_pair: str, is_buy: bool) -> Decimal:
 *         return self.c_get_price(trading_pair, is_buy)             # <<<<<<<<<<<<<<
 * 
 *     def buy(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,
 */
  __Pyx_XDECREF(__pyx_r);
  if (!(likely(PyUnicode_CheckExact(__pyx_v_trading_pair))||((__pyx_v_trading_pair) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_trading_pair))) __PYX_ERR(0, 1102, __pyx_L1_error)
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_is_buy); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1102, __pyx_L1_error)
  __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_get_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), ((PyObject*)__pyx_v_trading_pair), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1101
 *     # </editor-fold>
 * 
 *     def get_price(self, trading_pair: str, is_buy: bool) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_get_price(trading_pair, is_buy)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.get_price", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1104
 *         return self.c_get_price(trading_pair, is_buy)
 * 
 *     def buy(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,             # <<<<<<<<<<<<<<
 *             price: Decimal = s_decimal_0, **kwargs) -> str:
 *         return self.c_buy(trading_pair, amount, order_type, price, kwargs)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_33buy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_33buy = {"buy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_33buy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_33buy(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_trading_pair = 0;
  PyObject *__pyx_v_amount = 0;
  PyObject *__pyx_v_order_type = 0;
  PyObject *__pyx_v_price = 0;
  PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("buy (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_trading_pair,&__pyx_n_s_amount,&__pyx_n_s_order_type,&__pyx_n_s_price,0};
    values[2] = __Pyx_Arg_NewRef_FASTCALL(__pyx_k__27);
    values[3] = __Pyx_Arg_NewRef_FASTCALL(__pyx_k__28);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_trading_pair)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1104, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_amount)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1104, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("buy", 0, 2, 4, 1); __PYX_ERR(0, 1104, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_order_type);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1104, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_price);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1104, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, __pyx_v_kwargs, values + 0, kwd_pos_args, "buy") < 0)) __PYX_ERR(0, 1104, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_trading_pair = values[0];
    __pyx_v_amount = values[1];
    __pyx_v_order_type = values[2];
    __pyx_v_price = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("buy", 0, 2, 4, __pyx_nargs); __PYX_ERR(0, 1104, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.buy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_32buy(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self), __pyx_v_trading_pair, __pyx_v_amount, __pyx_v_order_type, __pyx_v_price, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_kwargs);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_32buy(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_amount, PyObject *__pyx_v_order_type, PyObject *__pyx_v_price, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_buy __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("buy", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1106
 *     def buy(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,
 *             price: Decimal = s_decimal_0, **kwargs) -> str:
 *         return self.c_buy(trading_pair, amount, order_type, price, kwargs)             # <<<<<<<<<<<<<<
 * 
 *     def sell(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,
 */
  __Pyx_XDECREF(__pyx_r);
  if (!(likely(PyUnicode_CheckExact(__pyx_v_trading_pair))||((__pyx_v_trading_pair) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_trading_pair))) __PYX_ERR(0, 1106, __pyx_L1_error)
  __pyx_t_2.__pyx_n = 3;
  __pyx_t_2.order_type = __pyx_v_order_type;
  __pyx_t_2.price = __pyx_v_price;
  __pyx_t_2.kwargs = __pyx_v_kwargs;
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_buy(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), ((PyObject*)__pyx_v_trading_pair), __pyx_v_amount, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1104
 *         return self.c_get_price(trading_pair, is_buy)
 * 
 *     def buy(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,             # <<<<<<<<<<<<<<
 *             price: Decimal = s_decimal_0, **kwargs) -> str:
 *         return self.c_buy(trading_pair, amount, order_type, price, kwargs)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.buy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1108
 *         return self.c_buy(trading_pair, amount, order_type, price, kwargs)
 * 
 *     def sell(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,             # <<<<<<<<<<<<<<
 *              price: Decimal = s_decimal_0, **kwargs) -> str:
 *         return self.c_sell(trading_pair, amount, order_type, price, kwargs)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_35sell(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_35sell = {"sell", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_35sell, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_35sell(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_trading_pair = 0;
  PyObject *__pyx_v_amount = 0;
  PyObject *__pyx_v_order_type = 0;
  PyObject *__pyx_v_price = 0;
  PyObject *__pyx_v_kwargs = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[4] = {0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("sell (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  __pyx_v_kwargs = PyDict_New(); if (unlikely(!__pyx_v_kwargs)) return NULL;
  __Pyx_GOTREF(__pyx_v_kwargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_trading_pair,&__pyx_n_s_amount,&__pyx_n_s_order_type,&__pyx_n_s_price,0};
    values[2] = __Pyx_Arg_NewRef_FASTCALL(__pyx_k__29);
    values[3] = __Pyx_Arg_NewRef_FASTCALL(__pyx_k__30);
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_trading_pair)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1108, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_amount)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1108, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("sell", 0, 2, 4, 1); __PYX_ERR(0, 1108, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_order_type);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1108, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_price);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1108, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, __pyx_v_kwargs, values + 0, kwd_pos_args, "sell") < 0)) __PYX_ERR(0, 1108, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_trading_pair = values[0];
    __pyx_v_amount = values[1];
    __pyx_v_order_type = values[2];
    __pyx_v_price = values[3];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("sell", 0, 2, 4, __pyx_nargs); __PYX_ERR(0, 1108, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_DECREF(__pyx_v_kwargs); __pyx_v_kwargs = 0;
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.sell", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_34sell(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self), __pyx_v_trading_pair, __pyx_v_amount, __pyx_v_order_type, __pyx_v_price, __pyx_v_kwargs);

  /* function exit code */
  __Pyx_DECREF(__pyx_v_kwargs);
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_34sell(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_amount, PyObject *__pyx_v_order_type, PyObject *__pyx_v_price, PyObject *__pyx_v_kwargs) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_sell __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("sell", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1110
 *     def sell(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,
 *              price: Decimal = s_decimal_0, **kwargs) -> str:
 *         return self.c_sell(trading_pair, amount, order_type, price, kwargs)             # <<<<<<<<<<<<<<
 * 
 *     def cancel(self, trading_pair: str, client_order_id: str):
 */
  __Pyx_XDECREF(__pyx_r);
  if (!(likely(PyUnicode_CheckExact(__pyx_v_trading_pair))||((__pyx_v_trading_pair) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_trading_pair))) __PYX_ERR(0, 1110, __pyx_L1_error)
  __pyx_t_2.__pyx_n = 3;
  __pyx_t_2.order_type = __pyx_v_order_type;
  __pyx_t_2.price = __pyx_v_price;
  __pyx_t_2.kwargs = __pyx_v_kwargs;
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_sell(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), ((PyObject*)__pyx_v_trading_pair), __pyx_v_amount, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1108
 *         return self.c_buy(trading_pair, amount, order_type, price, kwargs)
 * 
 *     def sell(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,             # <<<<<<<<<<<<<<
 *              price: Decimal = s_decimal_0, **kwargs) -> str:
 *         return self.c_sell(trading_pair, amount, order_type, price, kwargs)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.sell", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1112
 *         return self.c_sell(trading_pair, amount, order_type, price, kwargs)
 * 
 *     def cancel(self, trading_pair: str, client_order_id: str):             # <<<<<<<<<<<<<<
 *         return self.c_cancel(trading_pair, client_order_id)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_37cancel(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_37cancel = {"cancel", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_37cancel, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_37cancel(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_trading_pair = 0;
  PyObject *__pyx_v_client_order_id = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_trading_pair,&__pyx_n_s_client_order_id,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_trading_pair)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1112, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_client_order_id)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1112, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("cancel", 1, 2, 2, 1); __PYX_ERR(0, 1112, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "cancel") < 0)) __PYX_ERR(0, 1112, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_trading_pair = values[0];
    __pyx_v_client_order_id = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cancel", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1112, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_36cancel(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self), __pyx_v_trading_pair, __pyx_v_client_order_id);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_36cancel(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair, PyObject *__pyx_v_client_order_id) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancel", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1113
 * 
 *     def cancel(self, trading_pair: str, client_order_id: str):
 *         return self.c_cancel(trading_pair, client_order_id)             # <<<<<<<<<<<<<<
 * 
 *     def get_fee(self,
 */
  __Pyx_XDECREF(__pyx_r);
  if (!(likely(PyUnicode_CheckExact(__pyx_v_trading_pair))||((__pyx_v_trading_pair) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_trading_pair))) __PYX_ERR(0, 1113, __pyx_L1_error)
  if (!(likely(PyUnicode_CheckExact(__pyx_v_client_order_id))||((__pyx_v_client_order_id) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_client_order_id))) __PYX_ERR(0, 1113, __pyx_L1_error)
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_cancel(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_self), ((PyObject*)__pyx_v_trading_pair), ((PyObject*)__pyx_v_client_order_id)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1112
 *         return self.c_sell(trading_pair, amount, order_type, price, kwargs)
 * 
 *     def cancel(self, trading_pair: str, client_order_id: str):             # <<<<<<<<<<<<<<
 *         return self.c_cancel(trading_pair, client_order_id)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.cancel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1115
 *         return self.c_cancel(trading_pair, client_order_id)
 * 
 *     def get_fee(self,             # <<<<<<<<<<<<<<
 *                 base_currency: str,
 *                 quote_currency: str,
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_39get_fee(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_39get_fee = {"get_fee", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_39get_fee, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_39get_fee(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_base_currency = 0;
  PyObject *__pyx_v_quote_currency = 0;
  PyObject *__pyx_v_order_type = 0;
  PyObject *__pyx_v_order_side = 0;
  PyObject *__pyx_v_amount = 0;
  PyObject *__pyx_v_price = 0;
  PyObject *__pyx_v_is_maker = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[7] = {0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_fee (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_base_currency,&__pyx_n_s_quote_currency,&__pyx_n_s_order_type,&__pyx_n_s_order_side,&__pyx_n_s_amount,&__pyx_n_s_price,&__pyx_n_s_is_maker,0};
    values[5] = __Pyx_Arg_NewRef_FASTCALL(__pyx_k__31);

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1122
 *                 amount: Decimal,
 *                 price: Decimal = s_decimal_0,
 *                 is_maker: Optional[bool] = None):             # <<<<<<<<<<<<<<
 *         return self.c_get_fee(base_currency, quote_currency, order_type, order_side, amount, price, is_maker)
 * 
 */
    values[6] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  7: values[6] = __Pyx_Arg_FASTCALL(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = __Pyx_Arg_FASTCALL(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_base_currency)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1115, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_quote_currency)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1115, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("get_fee", 0, 5, 7, 1); __PYX_ERR(0, 1115, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_order_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1115, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("get_fee", 0, 5, 7, 2); __PYX_ERR(0, 1115, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (likely((values[3] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_order_side)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[3]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1115, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("get_fee", 0, 5, 7, 3); __PYX_ERR(0, 1115, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (likely((values[4] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_amount)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[4]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1115, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("get_fee", 0, 5, 7, 4); __PYX_ERR(0, 1115, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_price);
          if (value) { values[5] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1115, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_is_maker);
          if (value) { values[6] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1115, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_fee") < 0)) __PYX_ERR(0, 1115, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  7: values[6] = __Pyx_Arg_FASTCALL(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = __Pyx_Arg_FASTCALL(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
        values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_base_currency = values[0];
    __pyx_v_quote_currency = values[1];
    __pyx_v_order_type = values[2];
    __pyx_v_order_side = values[3];
    __pyx_v_amount = values[4];
    __pyx_v_price = values[5];
    __pyx_v_is_maker = values[6];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_fee", 0, 5, 7, __pyx_nargs); __PYX_ERR(0, 1115, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.get_fee", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_38get_fee(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self), __pyx_v_base_currency, __pyx_v_quote_currency, __pyx_v_order_type, __pyx_v_order_side, __pyx_v_amount, __pyx_v_price, __pyx_v_is_maker);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1115
 *         return self.c_cancel(trading_pair, client_order_id)
 * 
 *     def get_fee(self,             # <<<<<<<<<<<<<<
 *                 base_currency: str,
 *                 quote_currency: str,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_38get_fee(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_base_currency, PyObject *__pyx_v_quote_currency, PyObject *__pyx_v_order_type, PyObject *__pyx_v_order_side, PyObject *__pyx_v_amount, PyObject *__pyx_v_price, PyObject *__pyx_v_is_maker) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_get_fee __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_fee", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1123
 *                 price: Decimal = s_decimal_0,
 *                 is_maker: Optional[bool] = None):
 *         return self.c_get_fee(base_currency, quote_currency, order_type, order_side, amount, price, is_maker)             # <<<<<<<<<<<<<<
 * 
 *     def get_order_book(self, trading_pair: str) -> OrderBook:
 */
  __Pyx_XDECREF(__pyx_r);
  if (!(likely(PyUnicode_CheckExact(__pyx_v_base_currency))||((__pyx_v_base_currency) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_base_currency))) __PYX_ERR(0, 1123, __pyx_L1_error)
  if (!(likely(PyUnicode_CheckExact(__pyx_v_quote_currency))||((__pyx_v_quote_currency) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_quote_currency))) __PYX_ERR(0, 1123, __pyx_L1_error)
  __pyx_t_2.__pyx_n = 1;
  __pyx_t_2.is_maker = __pyx_v_is_maker;
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_get_fee(((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_self), ((PyObject*)__pyx_v_base_currency), ((PyObject*)__pyx_v_quote_currency), __pyx_v_order_type, __pyx_v_order_side, __pyx_v_amount, __pyx_v_price, &__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1115
 *         return self.c_cancel(trading_pair, client_order_id)
 * 
 *     def get_fee(self,             # <<<<<<<<<<<<<<
 *                 base_currency: str,
 *                 quote_currency: str,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.get_fee", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1125
 *         return self.c_get_fee(base_currency, quote_currency, order_type, order_side, amount, price, is_maker)
 * 
 *     def get_order_book(self, trading_pair: str) -> OrderBook:             # <<<<<<<<<<<<<<
 *         return self.c_get_order_book(trading_pair)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_41get_order_book(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_41get_order_book = {"get_order_book", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_41get_order_book, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_41get_order_book(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_trading_pair = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_order_book (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_trading_pair,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_trading_pair)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1125, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_order_book") < 0)) __PYX_ERR(0, 1125, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_trading_pair = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_order_book", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1125, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.get_order_book", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_40get_order_book(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self), __pyx_v_trading_pair);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_40get_order_book(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_trading_pair) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_order_book", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1126
 * 
 *     def get_order_book(self, trading_pair: str) -> OrderBook:
 *         return self.c_get_order_book(trading_pair)             # <<<<<<<<<<<<<<
 * 
 *     def get_maker_order_type(self):
 */
  __Pyx_XDECREF(__pyx_r);
  if (!(likely(PyUnicode_CheckExact(__pyx_v_trading_pair))||((__pyx_v_trading_pair) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_trading_pair))) __PYX_ERR(0, 1126, __pyx_L1_error)
  __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_get_order_book(((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_self), ((PyObject*)__pyx_v_trading_pair))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1125
 *         return self.c_get_fee(base_currency, quote_currency, order_type, order_side, amount, price, is_maker)
 * 
 *     def get_order_book(self, trading_pair: str) -> OrderBook:             # <<<<<<<<<<<<<<
 *         return self.c_get_order_book(trading_pair)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.get_order_book", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1128
 *         return self.c_get_order_book(trading_pair)
 * 
 *     def get_maker_order_type(self):             # <<<<<<<<<<<<<<
 *         return OrderType.LIMIT
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_43get_maker_order_type(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_43get_maker_order_type = {"get_maker_order_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_43get_maker_order_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_43get_maker_order_type(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_maker_order_type (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_maker_order_type", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_maker_order_type", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_42get_maker_order_type(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_42get_maker_order_type(CYTHON_UNUSED struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_maker_order_type", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1129
 * 
 *     def get_maker_order_type(self):
 *         return OrderType.LIMIT             # <<<<<<<<<<<<<<
 * 
 *     def get_taker_order_type(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LIMIT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1128
 *         return self.c_get_order_book(trading_pair)
 * 
 *     def get_maker_order_type(self):             # <<<<<<<<<<<<<<
 *         return OrderType.LIMIT
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.get_maker_order_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1131
 *         return OrderType.LIMIT
 * 
 *     def get_taker_order_type(self):             # <<<<<<<<<<<<<<
 *         return OrderType.LIMIT
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_45get_taker_order_type(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_45get_taker_order_type = {"get_taker_order_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_45get_taker_order_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_45get_taker_order_type(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_taker_order_type (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_taker_order_type", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_taker_order_type", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_44get_taker_order_type(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_44get_taker_order_type(CYTHON_UNUSED struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_taker_order_type", 1);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1132
 * 
 *     def get_taker_order_type(self):
 *         return OrderType.LIMIT             # <<<<<<<<<<<<<<
 * 
 *     async def trigger_event_async(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_LIMIT); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1131
 *         return OrderType.LIMIT
 * 
 *     def get_taker_order_type(self):             # <<<<<<<<<<<<<<
 *         return OrderType.LIMIT
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.get_taker_order_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_48generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1134
 *         return OrderType.LIMIT
 * 
 *     async def trigger_event_async(self,             # <<<<<<<<<<<<<<
 *                                   event_tag,
 *                                   event):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_47trigger_event_async(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_47trigger_event_async = {"trigger_event_async", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_47trigger_event_async, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_47trigger_event_async(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_event_tag = 0;
  PyObject *__pyx_v_event = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("trigger_event_async (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_event_tag,&__pyx_n_s_event,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_event_tag)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1134, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_event)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1134, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("trigger_event_async", 1, 2, 2, 1); __PYX_ERR(0, 1134, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "trigger_event_async") < 0)) __PYX_ERR(0, 1134, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_event_tag = values[0];
    __pyx_v_event = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("trigger_event_async", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1134, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.trigger_event_async", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_46trigger_event_async(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self), __pyx_v_event_tag, __pyx_v_event);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_46trigger_event_async(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, PyObject *__pyx_v_event_tag, PyObject *__pyx_v_event) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("trigger_event_async", 0);
  __pyx_cur_scope = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async *)__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1134, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_event_tag = __pyx_v_event_tag;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_event_tag);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_event_tag);
  __pyx_cur_scope->__pyx_v_event = __pyx_v_event;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_event);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_event);
  {
    __pyx_CoroutineObject *gen = __Pyx_Coroutine_New((__pyx_coroutine_body_t) __pyx_gb_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_48generator4, __pyx_codeobj__32, (PyObject *) __pyx_cur_scope, __pyx_n_s_trigger_event_async, __pyx_n_s_PaperTradeExchange_trigger_event, __pyx_n_s_hummingbot_connector_exchange_pa); if (unlikely(!gen)) __PYX_ERR(0, 1134, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.trigger_event_async", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_48generator4(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async *__pyx_cur_scope = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int64_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("trigger_event_async", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L4_resume_from_await;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1134, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1137
 *                                   event_tag,
 *                                   event):
 *         await asyncio.sleep(0.01)             # <<<<<<<<<<<<<<
 *         self.c_trigger_event(event_tag, event)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_asyncio); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_sleep); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_float_0_01};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_r = __Pyx_Coroutine_Yield_From(__pyx_generator, __pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_XGOTREF(__pyx_r);
  if (likely(__pyx_r)) {
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
    /* return from generator, awaiting value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L4_resume_from_await:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1137, __pyx_L1_error)
  } else {
    PyObject* exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (exc_type) {
      if (likely(exc_type == PyExc_StopIteration || (exc_type != PyExc_GeneratorExit && __Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))) PyErr_Clear();
      else __PYX_ERR(0, 1137, __pyx_L1_error)
    }
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1138
 *                                   event):
 *         await asyncio.sleep(0.01)
 *         self.c_trigger_event(event_tag, event)             # <<<<<<<<<<<<<<
 */
  __pyx_t_5 = __Pyx_PyInt_As_int64_t(__pyx_cur_scope->__pyx_v_event_tag); if (unlikely((__pyx_t_5 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 1138, __pyx_L1_error)
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_cur_scope->__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.c_trigger_event(((struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *)__pyx_cur_scope->__pyx_v_self), __pyx_t_5, __pyx_cur_scope->__pyx_v_event); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1134
 *         return OrderType.LIMIT
 * 
 *     async def trigger_event_async(self,             # <<<<<<<<<<<<<<
 *                                   event_tag,
 *                                   event):
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("trigger_event_async", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ask_limit_orders,self._bid_limit_orders,self._events,self._limit_order_expiration_set cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_50__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_50__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_50__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_50__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_49__reduce_cython__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_49__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self._ask_limit_orders,self._bid_limit_orders,self._events,self._limit_order_expiration_set cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self._ask_limit_orders,self._bid_limit_orders,self._events,self._limit_order_expiration_set cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self__ask_limit_orders_self__bid, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ask_limit_orders,self._bid_limit_orders,self._events,self._limit_order_expiration_set cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._ask_limit_orders,self._bid_limit_orders,self._events,self._limit_order_expiration_set cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ask_limit_orders,self._bid_limit_orders,self._events,self._limit_order_expiration_set cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_52__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_52__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_52__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_52__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_51__setstate_cython__(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_51__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "self._ask_limit_orders,self._bid_limit_orders,self._events,self._limit_order_expiration_set cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self._ask_limit_orders,self._bid_limit_orders,self._events,self._limit_order_expiration_set cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self__ask_limit_orders_self__bid, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._ask_limit_orders,self._bid_limit_orders,self._events,self._limit_order_expiration_set cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ask_limit_orders,self._bid_limit_orders,self._events,self._limit_order_expiration_set cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_QuantizationParams(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_1__pyx_unpickle_QuantizationParams(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_1__pyx_unpickle_QuantizationParams = {"__pyx_unpickle_QuantizationParams", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_1__pyx_unpickle_QuantizationParams, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_1__pyx_unpickle_QuantizationParams(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_QuantizationParams (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_QuantizationParams", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_QuantizationParams", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_QuantizationParams") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_QuantizationParams", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.__pyx_unpickle_QuantizationParams", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_QuantizationParams(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_QuantizationParams(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_QuantizationParams", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xa74d68e, 0x820b24c, 0x7857e48):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xa74d68e, 0x820b24c, 0x7857e48) = (order_size_decimals, order_size_precision, price_decimals, price_precision, trading_pair))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__33, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xa74d68e, 0x820b24c, 0x7857e48):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xa74d68e, 0x820b24c, 0x7857e48) = (order_size_decimals, order_size_precision, price_decimals, price_precision, trading_pair))" % __pyx_checksum
 *     __pyx_result = QuantizationParams.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xa74d68e, 0x820b24c, 0x7857e48):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xa74d68e, 0x820b24c, 0x7857e48) = (order_size_decimals, order_size_precision, price_decimals, price_precision, trading_pair))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = QuantizationParams.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xa74d68e, 0x820b24c, 0x7857e48):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xa74d68e, 0x820b24c, 0x7857e48) = (order_size_decimals, order_size_precision, price_decimals, price_precision, trading_pair))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xa74d68e, 0x820b24c, 0x7857e48) = (order_size_decimals, order_size_precision, price_decimals, price_precision, trading_pair))" % __pyx_checksum
 *     __pyx_result = QuantizationParams.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_QuantizationParams__set_state(<QuantizationParams> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xa74d68e, 0x820b24c, 0x7857e48) = (order_size_decimals, order_size_precision, price_decimals, price_precision, trading_pair))" % __pyx_checksum
 *     __pyx_result = QuantizationParams.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_QuantizationParams__set_state(<QuantizationParams> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = QuantizationParams.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_QuantizationParams__set_state(<QuantizationParams> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_QuantizationParams__set_state(QuantizationParams __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_QuantizationParams__set_state(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xa74d68e, 0x820b24c, 0x7857e48) = (order_size_decimals, order_size_precision, price_decimals, price_precision, trading_pair))" % __pyx_checksum
 *     __pyx_result = QuantizationParams.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_QuantizationParams__set_state(<QuantizationParams> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_QuantizationParams__set_state(<QuantizationParams> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_QuantizationParams__set_state(QuantizationParams __pyx_result, tuple __pyx_state):
 *     __pyx_result.order_size_decimals = __pyx_state[0]; __pyx_result.order_size_precision = __pyx_state[1]; __pyx_result.price_decimals = __pyx_state[2]; __pyx_result.price_precision = __pyx_state[3]; __pyx_result.trading_pair = __pyx_state[4]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_QuantizationParams(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.__pyx_unpickle_QuantizationParams", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_QuantizationParams__set_state(<QuantizationParams> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_QuantizationParams__set_state(QuantizationParams __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.order_size_decimals = __pyx_state[0]; __pyx_result.order_size_precision = __pyx_state[1]; __pyx_result.price_decimals = __pyx_state[2]; __pyx_result.price_precision = __pyx_state[3]; __pyx_result.trading_pair = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_QuantizationParams__set_state(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_QuantizationParams__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_QuantizationParams__set_state(QuantizationParams __pyx_result, tuple __pyx_state):
 *     __pyx_result.order_size_decimals = __pyx_state[0]; __pyx_result.order_size_precision = __pyx_state[1]; __pyx_result.price_decimals = __pyx_state[2]; __pyx_result.price_precision = __pyx_state[3]; __pyx_result.trading_pair = __pyx_state[4]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->order_size_decimals = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->order_size_precision = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->price_decimals = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->price_precision = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->trading_pair);
  __Pyx_DECREF(__pyx_v___pyx_result->trading_pair);
  __pyx_v___pyx_result->trading_pair = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_QuantizationParams__set_state(QuantizationParams __pyx_result, tuple __pyx_state):
 *     __pyx_result.order_size_decimals = __pyx_state[0]; __pyx_result.order_size_precision = __pyx_state[1]; __pyx_result.price_decimals = __pyx_state[2]; __pyx_result.price_precision = __pyx_state[3]; __pyx_result.trading_pair = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 5);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result.order_size_decimals = __pyx_state[0]; __pyx_result.order_size_precision = __pyx_state[1]; __pyx_result.price_decimals = __pyx_state[2]; __pyx_result.price_precision = __pyx_state[3]; __pyx_result.trading_pair = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[5])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_QuantizationParams__set_state(QuantizationParams __pyx_result, tuple __pyx_state):
 *     __pyx_result.order_size_decimals = __pyx_state[0]; __pyx_result.order_size_precision = __pyx_state[1]; __pyx_result.price_decimals = __pyx_state[2]; __pyx_result.price_precision = __pyx_state[3]; __pyx_result.trading_pair = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_QuantizationParams__set_state(<QuantizationParams> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_QuantizationParams__set_state(QuantizationParams __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.order_size_decimals = __pyx_state[0]; __pyx_result.order_size_precision = __pyx_state[1]; __pyx_result.price_decimals = __pyx_state[2]; __pyx_result.price_precision = __pyx_state[3]; __pyx_result.trading_pair = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.__pyx_unpickle_QuantizationParams__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_QueuedOrder(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_3__pyx_unpickle_QueuedOrder(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_3__pyx_unpickle_QueuedOrder = {"__pyx_unpickle_QueuedOrder", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_3__pyx_unpickle_QueuedOrder, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_3__pyx_unpickle_QueuedOrder(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_QueuedOrder (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_QueuedOrder", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_QueuedOrder", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_QueuedOrder") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_QueuedOrder", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.__pyx_unpickle_QueuedOrder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_2__pyx_unpickle_QueuedOrder(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_2__pyx_unpickle_QueuedOrder(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_QueuedOrder", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xefa59de, 0x3685ee0, 0x5aefc6c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xefa59de, 0x3685ee0, 0x5aefc6c) = (_amount, _is_buy, _order_id, _trading_pair, create_timestamp))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__34, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xefa59de, 0x3685ee0, 0x5aefc6c):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xefa59de, 0x3685ee0, 0x5aefc6c) = (_amount, _is_buy, _order_id, _trading_pair, create_timestamp))" % __pyx_checksum
 *     __pyx_result = QueuedOrder.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xefa59de, 0x3685ee0, 0x5aefc6c):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xefa59de, 0x3685ee0, 0x5aefc6c) = (_amount, _is_buy, _order_id, _trading_pair, create_timestamp))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = QueuedOrder.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xefa59de, 0x3685ee0, 0x5aefc6c):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xefa59de, 0x3685ee0, 0x5aefc6c) = (_amount, _is_buy, _order_id, _trading_pair, create_timestamp))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xefa59de, 0x3685ee0, 0x5aefc6c) = (_amount, _is_buy, _order_id, _trading_pair, create_timestamp))" % __pyx_checksum
 *     __pyx_result = QueuedOrder.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_QueuedOrder__set_state(<QueuedOrder> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xefa59de, 0x3685ee0, 0x5aefc6c) = (_amount, _is_buy, _order_id, _trading_pair, create_timestamp))" % __pyx_checksum
 *     __pyx_result = QueuedOrder.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_QueuedOrder__set_state(<QueuedOrder> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = QueuedOrder.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_QueuedOrder__set_state(<QueuedOrder> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_QueuedOrder__set_state(QueuedOrder __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_QueuedOrder__set_state(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xefa59de, 0x3685ee0, 0x5aefc6c) = (_amount, _is_buy, _order_id, _trading_pair, create_timestamp))" % __pyx_checksum
 *     __pyx_result = QueuedOrder.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_QueuedOrder__set_state(<QueuedOrder> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_QueuedOrder__set_state(<QueuedOrder> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_QueuedOrder__set_state(QueuedOrder __pyx_result, tuple __pyx_state):
 *     __pyx_result._amount = __pyx_state[0]; __pyx_result._is_buy = __pyx_state[1]; __pyx_result._order_id = __pyx_state[2]; __pyx_result._trading_pair = __pyx_state[3]; __pyx_result.create_timestamp = __pyx_state[4]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_QueuedOrder(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.__pyx_unpickle_QueuedOrder", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_QueuedOrder__set_state(<QueuedOrder> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_QueuedOrder__set_state(QueuedOrder __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._amount = __pyx_state[0]; __pyx_result._is_buy = __pyx_state[1]; __pyx_result._order_id = __pyx_state[2]; __pyx_result._trading_pair = __pyx_state[3]; __pyx_result.create_timestamp = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_QueuedOrder__set_state(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  double __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_QueuedOrder__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_QueuedOrder__set_state(QueuedOrder __pyx_result, tuple __pyx_state):
 *     __pyx_result._amount = __pyx_state[0]; __pyx_result._is_buy = __pyx_state[1]; __pyx_result._order_id = __pyx_state[2]; __pyx_result._trading_pair = __pyx_state[3]; __pyx_result.create_timestamp = __pyx_state[4]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_amount);
  __Pyx_DECREF(__pyx_v___pyx_result->_amount);
  __pyx_v___pyx_result->_amount = __pyx_t_1;
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->_is_buy = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_order_id);
  __Pyx_DECREF(__pyx_v___pyx_result->_order_id);
  __pyx_v___pyx_result->_order_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->_trading_pair);
  __Pyx_DECREF(__pyx_v___pyx_result->_trading_pair);
  __pyx_v___pyx_result->_trading_pair = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 4, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_3 == (double)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->create_timestamp = __pyx_t_3;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_QueuedOrder__set_state(QueuedOrder __pyx_result, tuple __pyx_state):
 *     __pyx_result._amount = __pyx_state[0]; __pyx_result._is_buy = __pyx_state[1]; __pyx_result._order_id = __pyx_state[2]; __pyx_result._trading_pair = __pyx_state[3]; __pyx_result.create_timestamp = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 5);
  if (__pyx_t_5) {
  } else {
    __pyx_t_2 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result._amount = __pyx_state[0]; __pyx_result._is_buy = __pyx_state[1]; __pyx_result._order_id = __pyx_state[2]; __pyx_result._trading_pair = __pyx_state[3]; __pyx_result.create_timestamp = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[5])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 5, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_QueuedOrder__set_state(QueuedOrder __pyx_result, tuple __pyx_state):
 *     __pyx_result._amount = __pyx_state[0]; __pyx_result._is_buy = __pyx_state[1]; __pyx_result._order_id = __pyx_state[2]; __pyx_result._trading_pair = __pyx_state[3]; __pyx_result.create_timestamp = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[5])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_QueuedOrder__set_state(<QueuedOrder> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_QueuedOrder__set_state(QueuedOrder __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._amount = __pyx_state[0]; __pyx_result._is_buy = __pyx_state[1]; __pyx_result._order_id = __pyx_state[2]; __pyx_result._trading_pair = __pyx_state[3]; __pyx_result.create_timestamp = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.__pyx_unpickle_QueuedOrder__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_OrderBookTradeListener(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_5__pyx_unpickle_OrderBookTradeListener(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_5__pyx_unpickle_OrderBookTradeListener = {"__pyx_unpickle_OrderBookTradeListener", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_5__pyx_unpickle_OrderBookTradeListener, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_5__pyx_unpickle_OrderBookTradeListener(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_OrderBookTradeListener (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_OrderBookTradeListener", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_OrderBookTradeListener", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_OrderBookTradeListener") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_OrderBookTradeListener", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.__pyx_unpickle_OrderBookTradeListener", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_4__pyx_unpickle_OrderBookTradeListener(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_4__pyx_unpickle_OrderBookTradeListener(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_OrderBookTradeListener", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1bae46d, 0x16e8803, 0x37d9abe):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bae46d, 0x16e8803, 0x37d9abe) = (_current_event_caller, _current_event_tag, _market))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__35, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1bae46d, 0x16e8803, 0x37d9abe):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bae46d, 0x16e8803, 0x37d9abe) = (_current_event_caller, _current_event_tag, _market))" % __pyx_checksum
 *     __pyx_result = OrderBookTradeListener.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x1bae46d, 0x16e8803, 0x37d9abe):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bae46d, 0x16e8803, 0x37d9abe) = (_current_event_caller, _current_event_tag, _market))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = OrderBookTradeListener.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1bae46d, 0x16e8803, 0x37d9abe):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bae46d, 0x16e8803, 0x37d9abe) = (_current_event_caller, _current_event_tag, _market))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bae46d, 0x16e8803, 0x37d9abe) = (_current_event_caller, _current_event_tag, _market))" % __pyx_checksum
 *     __pyx_result = OrderBookTradeListener.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_OrderBookTradeListener__set_state(<OrderBookTradeListener> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bae46d, 0x16e8803, 0x37d9abe) = (_current_event_caller, _current_event_tag, _market))" % __pyx_checksum
 *     __pyx_result = OrderBookTradeListener.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_OrderBookTradeListener__set_state(<OrderBookTradeListener> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = OrderBookTradeListener.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_OrderBookTradeListener__set_state(<OrderBookTradeListener> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_OrderBookTradeListener__set_state(OrderBookTradeListener __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_OrderBookTradeListener__set_state(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bae46d, 0x16e8803, 0x37d9abe) = (_current_event_caller, _current_event_tag, _market))" % __pyx_checksum
 *     __pyx_result = OrderBookTradeListener.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_OrderBookTradeListener__set_state(<OrderBookTradeListener> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_OrderBookTradeListener__set_state(<OrderBookTradeListener> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_OrderBookTradeListener__set_state(OrderBookTradeListener __pyx_result, tuple __pyx_state):
 *     __pyx_result._current_event_caller = __pyx_state[0]; __pyx_result._current_event_tag = __pyx_state[1]; __pyx_result._market = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_OrderBookTradeListener(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.__pyx_unpickle_OrderBookTradeListener", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_OrderBookTradeListener__set_state(<OrderBookTradeListener> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_OrderBookTradeListener__set_state(OrderBookTradeListener __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._current_event_caller = __pyx_state[0]; __pyx_result._current_event_tag = __pyx_state[1]; __pyx_result._market = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_OrderBookTradeListener__set_state(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int64_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_OrderBookTradeListener__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_OrderBookTradeListener__set_state(OrderBookTradeListener __pyx_result, tuple __pyx_state):
 *     __pyx_result._current_event_caller = __pyx_state[0]; __pyx_result._current_event_tag = __pyx_state[1]; __pyx_result._market = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10hummingbot_4core_6pubsub_PubSub))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base._current_event_caller);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base._current_event_caller);
  __pyx_v___pyx_result->__pyx_base._current_event_caller = ((struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base._current_event_tag = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->_market);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->_market);
  __pyx_v___pyx_result->_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_OrderBookTradeListener__set_state(OrderBookTradeListener __pyx_result, tuple __pyx_state):
 *     __pyx_result._current_event_caller = __pyx_state[0]; __pyx_result._current_event_tag = __pyx_state[1]; __pyx_result._market = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 3);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result._current_event_caller = __pyx_state[0]; __pyx_result._current_event_tag = __pyx_state[1]; __pyx_result._market = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_OrderBookTradeListener__set_state(OrderBookTradeListener __pyx_result, tuple __pyx_state):
 *     __pyx_result._current_event_caller = __pyx_state[0]; __pyx_result._current_event_tag = __pyx_state[1]; __pyx_result._market = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_OrderBookTradeListener__set_state(<OrderBookTradeListener> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_OrderBookTradeListener__set_state(OrderBookTradeListener __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._current_event_caller = __pyx_state[0]; __pyx_result._current_event_tag = __pyx_state[1]; __pyx_result._market = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.__pyx_unpickle_OrderBookTradeListener__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_OrderBookMarketOrderFillListener(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_7__pyx_unpickle_OrderBookMarketOrderFillListener(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_7__pyx_unpickle_OrderBookMarketOrderFillListener = {"__pyx_unpickle_OrderBookMarketOrderFillListener", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_7__pyx_unpickle_OrderBookMarketOrderFillListener, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_7__pyx_unpickle_OrderBookMarketOrderFillListener(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_OrderBookMarketOrderFillListener (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_type)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_checksum)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_OrderBookMarketOrderFillListener", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[2]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_OrderBookMarketOrderFillListener", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__pyx_unpickle_OrderBookMarketOrderFillListener") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
      values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_OrderBookMarketOrderFillListener", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.__pyx_unpickle_OrderBookMarketOrderFillListener", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_6__pyx_unpickle_OrderBookMarketOrderFillListener(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_6__pyx_unpickle_OrderBookMarketOrderFillListener(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_OrderBookMarketOrderFillListener", 1);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1bae46d, 0x16e8803, 0x37d9abe):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bae46d, 0x16e8803, 0x37d9abe) = (_current_event_caller, _current_event_tag, _market))" % __pyx_checksum
 */
  __pyx_t_1 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_tuple__35, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1bae46d, 0x16e8803, 0x37d9abe):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bae46d, 0x16e8803, 0x37d9abe) = (_current_event_caller, _current_event_tag, _market))" % __pyx_checksum
 *     __pyx_result = OrderBookMarketOrderFillListener.__new__(__pyx_type)
 */
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_PickleError)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0x1bae46d, 0x16e8803, 0x37d9abe):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bae46d, 0x16e8803, 0x37d9abe) = (_current_event_caller, _current_event_tag, _market))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = OrderBookMarketOrderFillListener.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_3 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0x1bae46d, 0x16e8803, 0x37d9abe):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bae46d, 0x16e8803, 0x37d9abe) = (_current_event_caller, _current_event_tag, _market))" % __pyx_checksum
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bae46d, 0x16e8803, 0x37d9abe) = (_current_event_caller, _current_event_tag, _market))" % __pyx_checksum
 *     __pyx_result = OrderBookMarketOrderFillListener.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(<OrderBookMarketOrderFillListener> __pyx_result, __pyx_state)
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener), __pyx_n_s_new); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bae46d, 0x16e8803, 0x37d9abe) = (_current_event_caller, _current_event_tag, _market))" % __pyx_checksum
 *     __pyx_result = OrderBookMarketOrderFillListener.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(<OrderBookMarketOrderFillListener> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = OrderBookMarketOrderFillListener.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(<OrderBookMarketOrderFillListener> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(OrderBookMarketOrderFillListener __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_OrderBookMarketOrderFillListener__set_state(((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0x1bae46d, 0x16e8803, 0x37d9abe) = (_current_event_caller, _current_event_tag, _market))" % __pyx_checksum
 *     __pyx_result = OrderBookMarketOrderFillListener.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(<OrderBookMarketOrderFillListener> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(<OrderBookMarketOrderFillListener> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(OrderBookMarketOrderFillListener __pyx_result, tuple __pyx_state):
 *     __pyx_result._current_event_caller = __pyx_state[0]; __pyx_result._current_event_tag = __pyx_state[1]; __pyx_result._market = __pyx_state[2]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_OrderBookMarketOrderFillListener(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.__pyx_unpickle_OrderBookMarketOrderFillListener", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(<OrderBookMarketOrderFillListener> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(OrderBookMarketOrderFillListener __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._current_event_caller = __pyx_state[0]; __pyx_result._current_event_tag = __pyx_state[1]; __pyx_result._market = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_unpickle_OrderBookMarketOrderFillListener__set_state(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int64_t __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_OrderBookMarketOrderFillListener__set_state", 1);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(OrderBookMarketOrderFillListener __pyx_result, tuple __pyx_state):
 *     __pyx_result._current_event_caller = __pyx_state[0]; __pyx_result._current_event_tag = __pyx_state[1]; __pyx_result._market = __pyx_state[2]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10hummingbot_4core_6pubsub_PubSub))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base._current_event_caller);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base._current_event_caller);
  __pyx_v___pyx_result->__pyx_base._current_event_caller = ((struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *)__pyx_t_1);
  __pyx_t_1 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_2 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v___pyx_result->__pyx_base._current_event_tag = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->_market);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->_market);
  __pyx_v___pyx_result->_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(OrderBookMarketOrderFillListener __pyx_result, tuple __pyx_state):
 *     __pyx_result._current_event_caller = __pyx_state[0]; __pyx_result._current_event_tag = __pyx_state[1]; __pyx_result._market = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 > 3);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_5 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_3 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "(tree fragment)":14
 *     __pyx_result._current_event_caller = __pyx_state[0]; __pyx_result._current_event_tag = __pyx_state[1]; __pyx_result._market = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[3])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 3, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(OrderBookMarketOrderFillListener __pyx_result, tuple __pyx_state):
 *     __pyx_result._current_event_caller = __pyx_state[0]; __pyx_result._current_event_tag = __pyx_state[1]; __pyx_result._market = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[3])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(<OrderBookMarketOrderFillListener> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(OrderBookMarketOrderFillListener __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._current_event_caller = __pyx_state[0]; __pyx_result._current_event_tag = __pyx_state[1]; __pyx_result._market = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("hummingbot.connector.exchange.paper_trade.paper_trade_exchange.__pyx_unpickle_OrderBookMarketOrderFillListener__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *)o);
  p->trading_pair = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams(PyObject *o) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  Py_CLEAR(p->trading_pair);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyObject *__pyx_specialmethod___pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_3__repr__(self);
}

static PyMethodDef __pyx_methods_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams},
  {Py_tp_repr, (void *)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_3__repr__},
  {Py_tp_methods, (void *)__pyx_methods_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams},
  {Py_tp_init, (void *)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams_spec = {
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.QuantizationParams",
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.""QuantizationParams", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange;

static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub*)__pyx_vtabptr_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange;
  new((void*)&(p->_bid_limit_orders)) __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders();
  new((void*)&(p->_ask_limit_orders)) __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders();
  new((void*)&(p->_limit_order_expiration_set)) __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrderExpirationSet();
  p->_trading_pairs = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_queued_orders = Py_None; Py_INCREF(Py_None);
  p->_quantization_params = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_order_book_trade_listener = Py_None; Py_INCREF(Py_None);
  p->_market_order_filled_listener = Py_None; Py_INCREF(Py_None);
  p->_target_market = Py_None; Py_INCREF(Py_None);
  p->_exchange_name = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange(PyObject *o) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  __Pyx_call_destructor(p->_bid_limit_orders);
  __Pyx_call_destructor(p->_ask_limit_orders);
  __Pyx_call_destructor(p->_limit_order_expiration_set);
  Py_CLEAR(p->_trading_pairs);
  Py_CLEAR(p->_queued_orders);
  Py_CLEAR(p->_quantization_params);
  Py_CLEAR(p->_order_book_trade_listener);
  Py_CLEAR(p->_market_order_filled_listener);
  Py_CLEAR(p->_target_market);
  Py_CLEAR(p->_exchange_name);
  #if PY_MAJOR_VERSION < 3
  if (!(__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase) || PyType_IS_GC(__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase)) PyObject_GC_Track(o);
  #else
  if (PyType_IS_GC(__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase)) PyObject_GC_Track(o);
  #endif
  if (likely(__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase)) __Pyx_PyType_GetSlot(__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase, tp_dealloc, destructor)(o); else __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);
}

static int __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)o;
  e = ((likely(__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase)) ? ((__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase->tp_traverse) ? __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange)); if (e) return e;
  if (p->_trading_pairs) {
    e = (*v)(p->_trading_pairs, a); if (e) return e;
  }
  if (p->_queued_orders) {
    e = (*v)(p->_queued_orders, a); if (e) return e;
  }
  if (p->_quantization_params) {
    e = (*v)(p->_quantization_params, a); if (e) return e;
  }
  if (p->_order_book_trade_listener) {
    e = (*v)(p->_order_book_trade_listener, a); if (e) return e;
  }
  if (p->_market_order_filled_listener) {
    e = (*v)(p->_market_order_filled_listener, a); if (e) return e;
  }
  if (p->_target_market) {
    e = (*v)(p->_target_market, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *)o;
  if (likely(__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase)) { if (__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase->tp_clear) __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);
  tmp = ((PyObject*)p->_trading_pairs);
  p->_trading_pairs = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_queued_orders);
  p->_queued_orders = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_quantization_params);
  p->_quantization_params = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_order_book_trade_listener);
  p->_order_book_trade_listener = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_market_order_filled_listener);
  p->_market_order_filled_listener = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_target_market);
  p->_target_market = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_budget_checker(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_14budget_checker_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_trading_pair(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12trading_pair_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_trading_pairs(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_13trading_pairs_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_4name_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_display_name(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12display_name_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_order_books(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_11order_books_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_status_dict(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_11status_dict_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_ready(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_5ready_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_queued_orders(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_13queued_orders_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_limit_orders(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12limit_orders_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_on_hold_balances(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_16on_hold_balances_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_available_balances(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_18available_balances_1__get__(o);
}

static PyMethodDef __pyx_methods_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange[] = {
  {"random_order_id", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_3random_order_id, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"init_paper_trade_market", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_5init_paper_trade_market, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"split_trading_pair", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_7split_trading_pair, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"start_network", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_9start_network, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"stop_network", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12stop_network, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"check_network", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_15check_network, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"cancel_all", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_18cancel_all, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_order_price_quantum", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_21get_order_price_quantum, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_available_balance", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_23get_available_balance, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_all_balances", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_25get_all_balances, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"match_trade_to_limit_orders", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_27match_trade_to_limit_orders, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_balance", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_29set_balance, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_price", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_31get_price, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"buy", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_33buy, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"sell", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_35sell, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"cancel", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_37cancel, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_fee", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_39get_fee, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_order_book", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_41get_order_book, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_maker_order_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_43get_maker_order_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_taker_order_type", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_45get_taker_order_type, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"trigger_event_async", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_47trigger_event_async, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_50__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_52__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange[] = {
  {(char *)"budget_checker", __pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_budget_checker, 0, (char *)0, 0},
  {(char *)"trading_pair", __pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_trading_pair, 0, (char *)0, 0},
  {(char *)"trading_pairs", __pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_trading_pairs, 0, (char *)0, 0},
  {(char *)"name", __pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_name, 0, (char *)0, 0},
  {(char *)"display_name", __pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_display_name, 0, (char *)0, 0},
  {(char *)"order_books", __pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_order_books, 0, (char *)0, 0},
  {(char *)"status_dict", __pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_status_dict, 0, (char *)0, 0},
  {(char *)"ready", __pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_ready, 0, (char *)0, 0},
  {(char *)"queued_orders", __pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_queued_orders, 0, (char *)0, 0},
  {(char *)"limit_orders", __pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_limit_orders, 0, (char *)0, 0},
  {(char *)"on_hold_balances", __pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_on_hold_balances, 0, (char *)0, 0},
  {(char *)"available_balances", __pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_available_balances, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange},
  {Py_tp_clear, (void *)__pyx_tp_clear_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange},
  {Py_tp_methods, (void *)__pyx_methods_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange},
  {Py_tp_getset, (void *)__pyx_getsets_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange},
  {Py_tp_init, (void *)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange_spec = {
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.PaperTradeExchange",
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.""PaperTradeExchange", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, /*tp_traverse*/
  __pyx_tp_clear_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *)o);
  p->_order_id = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_trading_pair = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_amount = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder(PyObject *o) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_order_id);
  Py_CLEAR(p->_trading_pair);
  Py_CLEAR(p->_amount);
  #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *)o;
  if (p->_amount) {
    e = (*v)(p->_amount, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder *)o;
  tmp = ((PyObject*)p->_amount);
  p->_amount = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_timestamp(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_9timestamp_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_order_id(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_8order_id_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_is_buy(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_6is_buy_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_trading_pair(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_12trading_pair_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_amount(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_6amount_1__get__(o);
}

static PyObject *__pyx_specialmethod___pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_3__repr__(PyObject *self, CYTHON_UNUSED PyObject *arg) {
  return __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_3__repr__(self);
}

static PyMethodDef __pyx_methods_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder[] = {
  {"__repr__", (PyCFunction)__pyx_specialmethod___pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_3__repr__, METH_NOARGS|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_5__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_7__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder[] = {
  {(char *)"timestamp", __pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_timestamp, 0, (char *)0, 0},
  {(char *)"order_id", __pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_order_id, 0, (char *)0, 0},
  {(char *)"is_buy", __pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_is_buy, 0, (char *)0, 0},
  {(char *)"trading_pair", __pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_trading_pair, 0, (char *)0, 0},
  {(char *)"amount", __pyx_getprop_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_amount, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder},
  {Py_tp_repr, (void *)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_3__repr__},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder},
  {Py_tp_clear, (void *)__pyx_tp_clear_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder},
  {Py_tp_methods, (void *)__pyx_methods_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder},
  {Py_tp_getset, (void *)__pyx_getsets_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder},
  {Py_tp_init, (void *)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder_spec = {
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.QueuedOrder",
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.""QueuedOrder", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_3__repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder, /*tp_traverse*/
  __pyx_tp_clear_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener;

static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener*)__pyx_vtabptr_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener;
  p->_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener(PyObject *o) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_market);
  #if PY_MAJOR_VERSION < 3
  if (!(__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener) || PyType_IS_GC(__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener)) PyObject_GC_Track(o);
  #else
  if (PyType_IS_GC(__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener)) PyObject_GC_Track(o);
  #endif
  if (likely(__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener)) __Pyx_PyType_GetSlot(__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener, tp_dealloc, destructor)(o); else __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener);
}

static int __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *)o;
  e = ((likely(__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener)) ? ((__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener->tp_traverse) ? __pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener)); if (e) return e;
  if (p->_market) {
    e = (*v)(((PyObject *)p->_market), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener *)o;
  if (likely(__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener)) { if (__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener->tp_clear) __pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener);
  tmp = ((PyObject*)p->_market);
  p->_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener},
  {Py_tp_clear, (void *)__pyx_tp_clear_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener},
  {Py_tp_methods, (void *)__pyx_methods_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener},
  {Py_tp_init, (void *)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener_spec = {
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.OrderBookTradeListener",
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.""OrderBookTradeListener", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener, /*tp_traverse*/
  __pyx_tp_clear_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener;

static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener*)__pyx_vtabptr_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener;
  p->_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener(PyObject *o) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_market);
  #if PY_MAJOR_VERSION < 3
  if (!(__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener) || PyType_IS_GC(__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener)) PyObject_GC_Track(o);
  #else
  if (PyType_IS_GC(__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener)) PyObject_GC_Track(o);
  #endif
  if (likely(__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener)) __Pyx_PyType_GetSlot(__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener, tp_dealloc, destructor)(o); else __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener);
}

static int __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *)o;
  e = ((likely(__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener)) ? ((__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener->tp_traverse) ? __pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener)); if (e) return e;
  if (p->_market) {
    e = (*v)(((PyObject *)p->_market), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener *)o;
  if (likely(__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener)) { if (__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener->tp_clear) __pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener);
  tmp = ((PyObject*)p->_market);
  p->_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener[] = {
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener},
  {Py_tp_clear, (void *)__pyx_tp_clear_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener},
  {Py_tp_methods, (void *)__pyx_methods_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener},
  {Py_tp_init, (void *)__pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener_spec = {
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.OrderBookMarketOrderFillListener",
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.""OrderBookMarketOrderFillListener", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener, /*tp_traverse*/
  __pyx_tp_clear_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network *__pyx_freelist_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network[8];
static int __pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network = 0;
#endif

static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network)))) {
    o = (PyObject*)__pyx_freelist_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network[--__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network];
    memset(o, 0, sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network(PyObject *o) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network)))) {
    __pyx_freelist_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network[__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network++] = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network_spec = {
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.__pyx_scope_struct__start_network",
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.""__pyx_scope_struct__start_network", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network *__pyx_freelist_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network[8];
static int __pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network = 0;
#endif

static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network)))) {
    o = (PyObject*)__pyx_freelist_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network[--__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network];
    memset(o, 0, sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network(PyObject *o) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network)))) {
    __pyx_freelist_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network[__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network++] = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network_spec = {
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.__pyx_scope_struct_1_stop_network",
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.""__pyx_scope_struct_1_stop_network", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network *__pyx_freelist_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network[8];
static int __pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network = 0;
#endif

static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network)))) {
    o = (PyObject*)__pyx_freelist_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network[--__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network];
    memset(o, 0, sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network(PyObject *o) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network)))) {
    __pyx_freelist_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network[__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network++] = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network *)o;
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network_spec = {
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.__pyx_scope_struct_2_check_network",
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.""__pyx_scope_struct_2_check_network", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all *__pyx_freelist_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all[8];
static int __pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all = 0;
#endif

static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all)))) {
    o = (PyObject*)__pyx_freelist_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all[--__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all];
    memset(o, 0, sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all(PyObject *o) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_cancellation_results);
  Py_CLEAR(p->__pyx_v_results);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_timeout_seconds);
  Py_CLEAR(p->__pyx_v_trading_pair_str);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all)))) {
    __pyx_freelist_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all[__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all++] = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all *)o;
  if (p->__pyx_v_cancellation_results) {
    e = (*v)(p->__pyx_v_cancellation_results, a); if (e) return e;
  }
  if (p->__pyx_v_results) {
    e = (*v)(p->__pyx_v_results, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_timeout_seconds) {
    e = (*v)(p->__pyx_v_timeout_seconds, a); if (e) return e;
  }
  if (p->__pyx_v_trading_pair_str) {
    e = (*v)(p->__pyx_v_trading_pair_str, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all_spec = {
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.__pyx_scope_struct_3_cancel_all",
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.""__pyx_scope_struct_3_cancel_all", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async *__pyx_freelist_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async[8];
static int __pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async = 0;
#endif

static PyObject *__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async)))) {
    o = (PyObject*)__pyx_freelist_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async[--__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async];
    memset(o, 0, sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async(PyObject *o) {
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_event);
  Py_CLEAR(p->__pyx_v_event_tag);
  Py_CLEAR(p->__pyx_v_self);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async)))) {
    __pyx_freelist_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async[__pyx_freecount_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async++] = ((struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async *p = (struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async *)o;
  if (p->__pyx_v_event) {
    e = (*v)(p->__pyx_v_event, a); if (e) return e;
  }
  if (p->__pyx_v_event_tag) {
    e = (*v)(p->__pyx_v_event_tag, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject *)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async_spec = {
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.__pyx_scope_struct_4_trigger_event_async",
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.connector.exchange.paper_trade.paper_trade_exchange.""__pyx_scope_struct_4_trigger_event_async", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static int __Pyx_CreateStringTabAndInitStrings(void) {
  __Pyx_StringTabEntry __pyx_string_tab[] = {
    {&__pyx_kp_u_0, __pyx_k_0, sizeof(__pyx_k_0), 0, 1, 0, 0},
    {&__pyx_kp_u_02x, __pyx_k_02x, sizeof(__pyx_k_02x), 0, 1, 0, 0},
    {&__pyx_kp_u_1e, __pyx_k_1e, sizeof(__pyx_k_1e), 0, 1, 0, 0},
    {&__pyx_kp_u_1e_10, __pyx_k_1e_10, sizeof(__pyx_k_1e_10), 0, 1, 0, 0},
    {&__pyx_kp_u_1e_2, __pyx_k_1e_2, sizeof(__pyx_k_1e_2), 0, 1, 0, 0},
    {&__pyx_kp_u_1e_7, __pyx_k_1e_7, sizeof(__pyx_k_1e_7), 0, 1, 0, 0},
    {&__pyx_kp_u_7g, __pyx_k_7g, sizeof(__pyx_k_7g), 0, 1, 0, 0},
    {&__pyx_kp_u_8g, __pyx_k_8g, sizeof(__pyx_k_8g), 0, 1, 0, 0},
    {&__pyx_kp_u_Account_balance_does_not_have_as, __pyx_k_Account_balance_does_not_have_as, sizeof(__pyx_k_Account_balance_does_not_have_as), 0, 1, 0, 0},
    {&__pyx_n_s_AssertionError, __pyx_k_AssertionError, sizeof(__pyx_k_AssertionError), 0, 0, 1, 1},
    {&__pyx_n_s_BUY, __pyx_k_BUY, sizeof(__pyx_k_BUY), 0, 0, 1, 1},
    {&__pyx_n_u_BUY, __pyx_k_BUY, sizeof(__pyx_k_BUY), 0, 1, 0, 1},
    {&__pyx_n_s_BUY_ORDER_COMPLETED_EVENT_TAG, __pyx_k_BUY_ORDER_COMPLETED_EVENT_TAG, sizeof(__pyx_k_BUY_ORDER_COMPLETED_EVENT_TAG), 0, 0, 1, 1},
    {&__pyx_n_s_BudgetChecker, __pyx_k_BudgetChecker, sizeof(__pyx_k_BudgetChecker), 0, 0, 1, 1},
    {&__pyx_n_s_BuyOrderCompleted, __pyx_k_BuyOrderCompleted, sizeof(__pyx_k_BuyOrderCompleted), 0, 0, 1, 1},
    {&__pyx_n_s_BuyOrderCompletedEvent, __pyx_k_BuyOrderCompletedEvent, sizeof(__pyx_k_BuyOrderCompletedEvent), 0, 0, 1, 1},
    {&__pyx_n_s_BuyOrderCreated, __pyx_k_BuyOrderCreated, sizeof(__pyx_k_BuyOrderCreated), 0, 0, 1, 1},
    {&__pyx_n_s_BuyOrderCreatedEvent, __pyx_k_BuyOrderCreatedEvent, sizeof(__pyx_k_BuyOrderCreatedEvent), 0, 0, 1, 1},
    {&__pyx_n_s_CONNECTED, __pyx_k_CONNECTED, sizeof(__pyx_k_CONNECTED), 0, 0, 1, 1},
    {&__pyx_n_s_Callable, __pyx_k_Callable, sizeof(__pyx_k_Callable), 0, 0, 1, 1},
    {&__pyx_n_s_CancellationResult, __pyx_k_CancellationResult, sizeof(__pyx_k_CancellationResult), 0, 0, 1, 1},
    {&__pyx_n_s_ClientConfigAdapter, __pyx_k_ClientConfigAdapter, sizeof(__pyx_k_ClientConfigAdapter), 0, 0, 1, 1},
    {&__pyx_n_s_Clock, __pyx_k_Clock, sizeof(__pyx_k_Clock), 0, 0, 1, 1},
    {&__pyx_n_s_CompositeOrderBook, __pyx_k_CompositeOrderBook, sizeof(__pyx_k_CompositeOrderBook), 0, 0, 1, 1},
    {&__pyx_n_s_Decimal, __pyx_k_Decimal, sizeof(__pyx_k_Decimal), 0, 0, 1, 1},
    {&__pyx_n_s_Dict, __pyx_k_Dict, sizeof(__pyx_k_Dict), 0, 0, 1, 1},
    {&__pyx_kp_s_Dict_str_Decimal, __pyx_k_Dict_str_Decimal, sizeof(__pyx_k_Dict_str_Decimal), 0, 0, 1, 0},
    {&__pyx_n_s_DummyMetricsCollector, __pyx_k_DummyMetricsCollector, sizeof(__pyx_k_DummyMetricsCollector), 0, 0, 1, 1},
    {&__pyx_kp_u_Error_call_trade_listener, __pyx_k_Error_call_trade_listener, sizeof(__pyx_k_Error_call_trade_listener), 0, 1, 0, 0},
    {&__pyx_kp_u_Error_canceling_order, __pyx_k_Error_canceling_order, sizeof(__pyx_k_Error_canceling_order), 0, 1, 0, 0},
    {&__pyx_kp_u_Error_executing_queued_order, __pyx_k_Error_executing_queued_order, sizeof(__pyx_k_Error_executing_queued_order), 0, 1, 0, 0},
    {&__pyx_kp_u_Error_processing_limit_order, __pyx_k_Error_processing_limit_order, sizeof(__pyx_k_Error_processing_limit_order), 0, 1, 0, 0},
    {&__pyx_n_s_ExchangeBase, __pyx_k_ExchangeBase, sizeof(__pyx_k_ExchangeBase), 0, 0, 1, 1},
    {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0, __pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_2, __pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 0, 1, 0},
    {&__pyx_kp_s_Incompatible_checksums_0x_x_vs_0_3, __pyx_k_Incompatible_checksums_0x_x_vs_0_3, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_3), 0, 0, 1, 0},
    {&__pyx_kp_u_Insufficient, __pyx_k_Insufficient, sizeof(__pyx_k_Insufficient), 0, 1, 0, 0},
    {&__pyx_n_s_LIMIT, __pyx_k_LIMIT, sizeof(__pyx_k_LIMIT), 0, 0, 1, 1},
    {&__pyx_n_s_LIMIT_MAKER, __pyx_k_LIMIT_MAKER, sizeof(__pyx_k_LIMIT_MAKER), 0, 0, 1, 1},
    {&__pyx_n_s_LimitOrder, __pyx_k_LimitOrder, sizeof(__pyx_k_LimitOrder), 0, 0, 1, 1},
    {&__pyx_n_s_List, __pyx_k_List, sizeof(__pyx_k_List), 0, 0, 1, 1},
    {&__pyx_kp_s_List_CancellationResult, __pyx_k_List_CancellationResult, sizeof(__pyx_k_List_CancellationResult), 0, 0, 1, 0},
    {&__pyx_n_s_MARKET, __pyx_k_MARKET, sizeof(__pyx_k_MARKET), 0, 0, 1, 1},
    {&__pyx_n_s_MARKET_BUY_ORDER_CREATED_EVENT_T, __pyx_k_MARKET_BUY_ORDER_CREATED_EVENT_T, sizeof(__pyx_k_MARKET_BUY_ORDER_CREATED_EVENT_T), 0, 0, 1, 1},
    {&__pyx_n_s_MARKET_ORDER_CANCELED_EVENT_TAG, __pyx_k_MARKET_ORDER_CANCELED_EVENT_TAG, sizeof(__pyx_k_MARKET_ORDER_CANCELED_EVENT_TAG), 0, 0, 1, 1},
    {&__pyx_n_s_MARKET_ORDER_FAILURE_EVENT_TAG, __pyx_k_MARKET_ORDER_FAILURE_EVENT_TAG, sizeof(__pyx_k_MARKET_ORDER_FAILURE_EVENT_TAG), 0, 0, 1, 1},
    {&__pyx_n_s_MARKET_SELL_ORDER_CREATED_EVENT, __pyx_k_MARKET_SELL_ORDER_CREATED_EVENT, sizeof(__pyx_k_MARKET_SELL_ORDER_CREATED_EVENT), 0, 0, 1, 1},
    {&__pyx_n_s_MarketEvent, __pyx_k_MarketEvent, sizeof(__pyx_k_MarketEvent), 0, 0, 1, 1},
    {&__pyx_n_s_MarketOrderFailureEvent, __pyx_k_MarketOrderFailureEvent, sizeof(__pyx_k_MarketOrderFailureEvent), 0, 0, 1, 1},
    {&__pyx_n_b_NIL, __pyx_k_NIL, sizeof(__pyx_k_NIL), 0, 0, 0, 1},
    {&__pyx_n_s_NetworkStatus, __pyx_k_NetworkStatus, sizeof(__pyx_k_NetworkStatus), 0, 0, 1, 1},
    {&__pyx_kp_u_No_order_book_exists_for, __pyx_k_No_order_book_exists_for, sizeof(__pyx_k_No_order_book_exists_for), 0, 1, 0, 0},
    {&__pyx_kp_u_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 1, 0, 0},
    {&__pyx_kp_u_Not_enough, __pyx_k_Not_enough, sizeof(__pyx_k_Not_enough), 0, 1, 0, 0},
    {&__pyx_n_s_ORDER_BOOK_TRADE_EVENT_TAG, __pyx_k_ORDER_BOOK_TRADE_EVENT_TAG, sizeof(__pyx_k_ORDER_BOOK_TRADE_EVENT_TAG), 0, 0, 1, 1},
    {&__pyx_n_s_ORDER_FILLED_EVENT_TAG, __pyx_k_ORDER_FILLED_EVENT_TAG, sizeof(__pyx_k_ORDER_FILLED_EVENT_TAG), 0, 0, 1, 1},
    {&__pyx_n_s_Optional, __pyx_k_Optional, sizeof(__pyx_k_Optional), 0, 0, 1, 1},
    {&__pyx_kp_s_Optional_bool, __pyx_k_Optional_bool, sizeof(__pyx_k_Optional_bool), 0, 0, 1, 0},
    {&__pyx_n_s_OrderBook, __pyx_k_OrderBook, sizeof(__pyx_k_OrderBook), 0, 0, 1, 1},
    {&__pyx_n_s_OrderBookEvent, __pyx_k_OrderBookEvent, sizeof(__pyx_k_OrderBookEvent), 0, 0, 1, 1},
    {&__pyx_n_s_OrderBookMarketOrderFillListener, __pyx_k_OrderBookMarketOrderFillListener, sizeof(__pyx_k_OrderBookMarketOrderFillListener), 0, 0, 1, 1},
    {&__pyx_n_s_OrderBookMarketOrderFillListener_2, __pyx_k_OrderBookMarketOrderFillListener_2, sizeof(__pyx_k_OrderBookMarketOrderFillListener_2), 0, 0, 1, 1},
    {&__pyx_n_s_OrderBookMarketOrderFillListener_3, __pyx_k_OrderBookMarketOrderFillListener_3, sizeof(__pyx_k_OrderBookMarketOrderFillListener_3), 0, 0, 1, 1},
    {&__pyx_n_s_OrderBookTracker, __pyx_k_OrderBookTracker, sizeof(__pyx_k_OrderBookTracker), 0, 0, 1, 1},
    {&__pyx_n_s_OrderBookTradeEvent, __pyx_k_OrderBookTradeEvent, sizeof(__pyx_k_OrderBookTradeEvent), 0, 0, 1, 1},
    {&__pyx_n_s_OrderBookTradeListener, __pyx_k_OrderBookTradeListener, sizeof(__pyx_k_OrderBookTradeListener), 0, 0, 1, 1},
    {&__pyx_n_s_OrderBookTradeListener___reduce, __pyx_k_OrderBookTradeListener___reduce, sizeof(__pyx_k_OrderBookTradeListener___reduce), 0, 0, 1, 1},
    {&__pyx_n_s_OrderBookTradeListener___setstat, __pyx_k_OrderBookTradeListener___setstat, sizeof(__pyx_k_OrderBookTradeListener___setstat), 0, 0, 1, 1},
    {&__pyx_n_s_OrderCancelled, __pyx_k_OrderCancelled, sizeof(__pyx_k_OrderCancelled), 0, 0, 1, 1},
    {&__pyx_n_s_OrderCancelledEvent, __pyx_k_OrderCancelledEvent, sizeof(__pyx_k_OrderCancelledEvent), 0, 0, 1, 1},
    {&__pyx_n_s_OrderCandidate, __pyx_k_OrderCandidate, sizeof(__pyx_k_OrderCandidate), 0, 0, 1, 1},
    {&__pyx_n_s_OrderFailure, __pyx_k_OrderFailure, sizeof(__pyx_k_OrderFailure), 0, 0, 1, 1},
    {&__pyx_n_s_OrderFilled, __pyx_k_OrderFilled, sizeof(__pyx_k_OrderFilled), 0, 0, 1, 1},
    {&__pyx_n_s_OrderFilledEvent, __pyx_k_OrderFilledEvent, sizeof(__pyx_k_OrderFilledEvent), 0, 0, 1, 1},
    {&__pyx_n_s_OrderType, __pyx_k_OrderType, sizeof(__pyx_k_OrderType), 0, 0, 1, 1},
    {&__pyx_n_u_PaperTrade, __pyx_k_PaperTrade, sizeof(__pyx_k_PaperTrade), 0, 1, 0, 1},
    {&__pyx_n_s_PaperTradeExchange, __pyx_k_PaperTradeExchange, sizeof(__pyx_k_PaperTradeExchange), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange___reduce_cyth, __pyx_k_PaperTradeExchange___reduce_cyth, sizeof(__pyx_k_PaperTradeExchange___reduce_cyth), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange___setstate_cy, __pyx_k_PaperTradeExchange___setstate_cy, sizeof(__pyx_k_PaperTradeExchange___setstate_cy), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_buy, __pyx_k_PaperTradeExchange_buy, sizeof(__pyx_k_PaperTradeExchange_buy), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_cancel, __pyx_k_PaperTradeExchange_cancel, sizeof(__pyx_k_PaperTradeExchange_cancel), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_cancel_all, __pyx_k_PaperTradeExchange_cancel_all, sizeof(__pyx_k_PaperTradeExchange_cancel_all), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_check_network, __pyx_k_PaperTradeExchange_check_network, sizeof(__pyx_k_PaperTradeExchange_check_network), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_get_all_balan, __pyx_k_PaperTradeExchange_get_all_balan, sizeof(__pyx_k_PaperTradeExchange_get_all_balan), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_get_available, __pyx_k_PaperTradeExchange_get_available, sizeof(__pyx_k_PaperTradeExchange_get_available), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_get_fee, __pyx_k_PaperTradeExchange_get_fee, sizeof(__pyx_k_PaperTradeExchange_get_fee), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_get_maker_ord, __pyx_k_PaperTradeExchange_get_maker_ord, sizeof(__pyx_k_PaperTradeExchange_get_maker_ord), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_get_order_boo, __pyx_k_PaperTradeExchange_get_order_boo, sizeof(__pyx_k_PaperTradeExchange_get_order_boo), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_get_order_pri, __pyx_k_PaperTradeExchange_get_order_pri, sizeof(__pyx_k_PaperTradeExchange_get_order_pri), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_get_price, __pyx_k_PaperTradeExchange_get_price, sizeof(__pyx_k_PaperTradeExchange_get_price), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_get_taker_ord, __pyx_k_PaperTradeExchange_get_taker_ord, sizeof(__pyx_k_PaperTradeExchange_get_taker_ord), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_init_paper_tr, __pyx_k_PaperTradeExchange_init_paper_tr, sizeof(__pyx_k_PaperTradeExchange_init_paper_tr), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_match_trade_t, __pyx_k_PaperTradeExchange_match_trade_t, sizeof(__pyx_k_PaperTradeExchange_match_trade_t), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_random_order, __pyx_k_PaperTradeExchange_random_order, sizeof(__pyx_k_PaperTradeExchange_random_order), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_sell, __pyx_k_PaperTradeExchange_sell, sizeof(__pyx_k_PaperTradeExchange_sell), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_set_balance, __pyx_k_PaperTradeExchange_set_balance, sizeof(__pyx_k_PaperTradeExchange_set_balance), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_split_trading, __pyx_k_PaperTradeExchange_split_trading, sizeof(__pyx_k_PaperTradeExchange_split_trading), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_start_network, __pyx_k_PaperTradeExchange_start_network, sizeof(__pyx_k_PaperTradeExchange_start_network), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_stop_network, __pyx_k_PaperTradeExchange_stop_network, sizeof(__pyx_k_PaperTradeExchange_stop_network), 0, 0, 1, 1},
    {&__pyx_n_s_PaperTradeExchange_trigger_event, __pyx_k_PaperTradeExchange_trigger_event, sizeof(__pyx_k_PaperTradeExchange_trigger_event), 0, 0, 1, 1},
    {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
    {&__pyx_kp_u_QuantizationParams, __pyx_k_QuantizationParams, sizeof(__pyx_k_QuantizationParams), 0, 1, 0, 0},
    {&__pyx_n_s_QuantizationParams_2, __pyx_k_QuantizationParams_2, sizeof(__pyx_k_QuantizationParams_2), 0, 0, 1, 1},
    {&__pyx_n_s_QuantizationParams___reduce_cyth, __pyx_k_QuantizationParams___reduce_cyth, sizeof(__pyx_k_QuantizationParams___reduce_cyth), 0, 0, 1, 1},
    {&__pyx_n_s_QuantizationParams___setstate_cy, __pyx_k_QuantizationParams___setstate_cy, sizeof(__pyx_k_QuantizationParams___setstate_cy), 0, 0, 1, 1},
    {&__pyx_kp_u_QueuedOrder, __pyx_k_QueuedOrder, sizeof(__pyx_k_QueuedOrder), 0, 1, 0, 0},
    {&__pyx_n_s_QueuedOrder_2, __pyx_k_QueuedOrder_2, sizeof(__pyx_k_QueuedOrder_2), 0, 0, 1, 1},
    {&__pyx_n_s_QueuedOrder___reduce_cython, __pyx_k_QueuedOrder___reduce_cython, sizeof(__pyx_k_QueuedOrder___reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_QueuedOrder___setstate_cython, __pyx_k_QueuedOrder___setstate_cython, sizeof(__pyx_k_QueuedOrder___setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_ROUND_DOWN, __pyx_k_ROUND_DOWN, sizeof(__pyx_k_ROUND_DOWN), 0, 0, 1, 1},
    {&__pyx_n_s_SELL, __pyx_k_SELL, sizeof(__pyx_k_SELL), 0, 0, 1, 1},
    {&__pyx_n_s_SELL_ORDER_COMPLETED_EVENT_TAG, __pyx_k_SELL_ORDER_COMPLETED_EVENT_TAG, sizeof(__pyx_k_SELL_ORDER_COMPLETED_EVENT_TAG), 0, 0, 1, 1},
    {&__pyx_n_s_SellOrderCompleted, __pyx_k_SellOrderCompleted, sizeof(__pyx_k_SellOrderCompleted), 0, 0, 1, 1},
    {&__pyx_n_s_SellOrderCompletedEvent, __pyx_k_SellOrderCompletedEvent, sizeof(__pyx_k_SellOrderCompletedEvent), 0, 0, 1, 1},
    {&__pyx_n_s_SellOrderCreated, __pyx_k_SellOrderCreated, sizeof(__pyx_k_SellOrderCreated), 0, 0, 1, 1},
    {&__pyx_n_s_SellOrderCreatedEvent, __pyx_k_SellOrderCreatedEvent, sizeof(__pyx_k_SellOrderCreatedEvent), 0, 0, 1, 1},
    {&__pyx_n_s_TRADE_EXECUTION_DELAY, __pyx_k_TRADE_EXECUTION_DELAY, sizeof(__pyx_k_TRADE_EXECUTION_DELAY), 0, 0, 1, 1},
    {&__pyx_n_s_TYPE_CHECKING, __pyx_k_TYPE_CHECKING, sizeof(__pyx_k_TYPE_CHECKING), 0, 0, 1, 1},
    {&__pyx_n_s_TradeEvent, __pyx_k_TradeEvent, sizeof(__pyx_k_TradeEvent), 0, 0, 1, 1},
    {&__pyx_n_s_TradeType, __pyx_k_TradeType, sizeof(__pyx_k_TradeType), 0, 0, 1, 1},
    {&__pyx_n_s_TradingPair, __pyx_k_TradingPair, sizeof(__pyx_k_TradingPair), 0, 0, 1, 1},
    {&__pyx_kp_u_Trading_pair, __pyx_k_Trading_pair, sizeof(__pyx_k_Trading_pair), 0, 1, 0, 0},
    {&__pyx_n_s_Tuple, __pyx_k_Tuple, sizeof(__pyx_k_Tuple), 0, 0, 1, 1},
    {&__pyx_kp_s_Tuple_str_str, __pyx_k_Tuple_str_str, sizeof(__pyx_k_Tuple_str_str), 0, 0, 1, 0},
    {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
    {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
    {&__pyx_kp_u__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 1, 0, 0},
    {&__pyx_kp_u__11, __pyx_k__11, sizeof(__pyx_k__11), 0, 1, 0, 0},
    {&__pyx_kp_u__15, __pyx_k__15, sizeof(__pyx_k__15), 0, 1, 0, 0},
    {&__pyx_kp_u__20, __pyx_k__20, sizeof(__pyx_k__20), 0, 1, 0, 0},
    {&__pyx_kp_u__22, __pyx_k__22, sizeof(__pyx_k__22), 0, 1, 0, 0},
    {&__pyx_kp_u__26, __pyx_k__26, sizeof(__pyx_k__26), 0, 1, 0, 0},
    {&__pyx_kp_u__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
    {&__pyx_n_s__36, __pyx_k__36, sizeof(__pyx_k__36), 0, 0, 1, 1},
    {&__pyx_kp_u__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 1, 0, 0},
    {&__pyx_kp_u__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 1, 0, 0},
    {&__pyx_kp_u__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 1, 0, 0},
    {&__pyx_n_s__88, __pyx_k__88, sizeof(__pyx_k__88), 0, 0, 1, 1},
    {&__pyx_kp_u__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 1, 0, 0},
    {&__pyx_n_s_all, __pyx_k_all, sizeof(__pyx_k_all), 0, 0, 1, 1},
    {&__pyx_n_s_amount, __pyx_k_amount, sizeof(__pyx_k_amount), 0, 0, 1, 1},
    {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
    {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio, __pyx_k_asyncio, sizeof(__pyx_k_asyncio), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_tasks, __pyx_k_asyncio_tasks, sizeof(__pyx_k_asyncio_tasks), 0, 0, 1, 1},
    {&__pyx_kp_u_available, __pyx_k_available, sizeof(__pyx_k_available), 0, 1, 0, 0},
    {&__pyx_n_s_available_balances, __pyx_k_available_balances, sizeof(__pyx_k_available_balances), 0, 0, 1, 1},
    {&__pyx_kp_u_available_vs, __pyx_k_available_vs, sizeof(__pyx_k_available_vs), 0, 1, 0, 0},
    {&__pyx_n_s_await, __pyx_k_await, sizeof(__pyx_k_await), 0, 0, 1, 1},
    {&__pyx_n_s_balance, __pyx_k_balance, sizeof(__pyx_k_balance), 0, 0, 1, 1},
    {&__pyx_kp_u_balance_available_for_buy_order, __pyx_k_balance_available_for_buy_order, sizeof(__pyx_k_balance_available_for_buy_order), 0, 1, 0, 0},
    {&__pyx_kp_u_balance_available_for_sell_orde, __pyx_k_balance_available_for_sell_orde, sizeof(__pyx_k_balance_available_for_sell_orde), 0, 1, 0, 0},
    {&__pyx_kp_u_balance_to_fill_limit_buy_order, __pyx_k_balance_to_fill_limit_buy_order, sizeof(__pyx_k_balance_to_fill_limit_buy_order), 0, 1, 0, 0},
    {&__pyx_kp_u_balance_to_fill_limit_sell_orde, __pyx_k_balance_to_fill_limit_sell_orde, sizeof(__pyx_k_balance_to_fill_limit_sell_orde), 0, 1, 0, 0},
    {&__pyx_n_s_base_asset, __pyx_k_base_asset, sizeof(__pyx_k_base_asset), 0, 0, 1, 1},
    {&__pyx_n_s_base_currency, __pyx_k_base_currency, sizeof(__pyx_k_base_currency), 0, 0, 1, 1},
    {&__pyx_n_s_bool, __pyx_k_bool, sizeof(__pyx_k_bool), 0, 0, 1, 1},
    {&__pyx_n_s_build_trade_fee, __pyx_k_build_trade_fee, sizeof(__pyx_k_build_trade_fee), 0, 0, 1, 1},
    {&__pyx_n_s_buy, __pyx_k_buy, sizeof(__pyx_k_buy), 0, 0, 1, 1},
    {&__pyx_n_u_buy, __pyx_k_buy, sizeof(__pyx_k_buy), 0, 1, 0, 1},
    {&__pyx_n_s_cancel, __pyx_k_cancel, sizeof(__pyx_k_cancel), 0, 0, 1, 1},
    {&__pyx_n_s_cancel_all, __pyx_k_cancel_all, sizeof(__pyx_k_cancel_all), 0, 0, 1, 1},
    {&__pyx_n_s_cancellation_results, __pyx_k_cancellation_results, sizeof(__pyx_k_cancellation_results), 0, 0, 1, 1},
    {&__pyx_n_s_ceil, __pyx_k_ceil, sizeof(__pyx_k_ceil), 0, 0, 1, 1},
    {&__pyx_n_s_check_network, __pyx_k_check_network, sizeof(__pyx_k_check_network), 0, 0, 1, 1},
    {&__pyx_n_s_choice, __pyx_k_choice, sizeof(__pyx_k_choice), 0, 0, 1, 1},
    {&__pyx_n_s_class_getitem, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
    {&__pyx_n_s_client_config_map, __pyx_k_client_config_map, sizeof(__pyx_k_client_config_map), 0, 0, 1, 1},
    {&__pyx_n_s_client_order_id, __pyx_k_client_order_id, sizeof(__pyx_k_client_order_id), 0, 0, 1, 1},
    {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
    {&__pyx_n_s_cls, __pyx_k_cls, sizeof(__pyx_k_cls), 0, 0, 1, 1},
    {&__pyx_n_s_collections, __pyx_k_collections, sizeof(__pyx_k_collections), 0, 0, 1, 1},
    {&__pyx_n_s_convert_from_exchange_trading_pa, __pyx_k_convert_from_exchange_trading_pa, sizeof(__pyx_k_convert_from_exchange_trading_pa), 0, 0, 1, 1},
    {&__pyx_n_s_convert_to_exchange_trading_pair, __pyx_k_convert_to_exchange_trading_pair, sizeof(__pyx_k_convert_to_exchange_trading_pair), 0, 0, 1, 1},
    {&__pyx_n_s_create_timestamp, __pyx_k_create_timestamp, sizeof(__pyx_k_create_timestamp), 0, 0, 1, 1},
    {&__pyx_n_s_currency, __pyx_k_currency, sizeof(__pyx_k_currency), 0, 0, 1, 1},
    {&__pyx_n_s_data_source, __pyx_k_data_source, sizeof(__pyx_k_data_source), 0, 0, 1, 1},
    {&__pyx_n_s_decimal, __pyx_k_decimal, sizeof(__pyx_k_decimal), 0, 0, 1, 1},
    {&__pyx_n_s_defaultdict, __pyx_k_defaultdict, sizeof(__pyx_k_defaultdict), 0, 0, 1, 1},
    {&__pyx_n_s_deque, __pyx_k_deque, sizeof(__pyx_k_deque), 0, 0, 1, 1},
    {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
    {&__pyx_n_s_dict_2, __pyx_k_dict_2, sizeof(__pyx_k_dict_2), 0, 0, 1, 1},
    {&__pyx_kp_u_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
    {&__pyx_kp_u_does_not_existing_in_current_da, __pyx_k_does_not_existing_in_current_da, sizeof(__pyx_k_does_not_existing_in_current_da), 0, 1, 0, 0},
    {&__pyx_kp_u_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
    {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
    {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
    {&__pyx_n_s_event, __pyx_k_event, sizeof(__pyx_k_event), 0, 0, 1, 1},
    {&__pyx_n_s_event_object, __pyx_k_event_object, sizeof(__pyx_k_event_object), 0, 0, 1, 1},
    {&__pyx_n_s_event_tag, __pyx_k_event_tag, sizeof(__pyx_k_event_tag), 0, 0, 1, 1},
    {&__pyx_n_s_exc_info, __pyx_k_exc_info, sizeof(__pyx_k_exc_info), 0, 0, 1, 1},
    {&__pyx_n_s_exchange, __pyx_k_exchange, sizeof(__pyx_k_exchange), 0, 0, 1, 1},
    {&__pyx_n_s_exchange_name, __pyx_k_exchange_name, sizeof(__pyx_k_exchange_name), 0, 0, 1, 1},
    {&__pyx_n_s_exchange_trade_id, __pyx_k_exchange_trade_id, sizeof(__pyx_k_exchange_trade_id), 0, 0, 1, 1},
    {&__pyx_n_s_float, __pyx_k_float, sizeof(__pyx_k_float), 0, 0, 1, 1},
    {&__pyx_n_s_from_total_balances, __pyx_k_from_total_balances, sizeof(__pyx_k_from_total_balances), 0, 0, 1, 1},
    {&__pyx_kp_u_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
    {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
    {&__pyx_n_s_get_all_balances, __pyx_k_get_all_balances, sizeof(__pyx_k_get_all_balances), 0, 0, 1, 1},
    {&__pyx_n_s_get_available_balance, __pyx_k_get_available_balance, sizeof(__pyx_k_get_available_balance), 0, 0, 1, 1},
    {&__pyx_n_s_get_fee, __pyx_k_get_fee, sizeof(__pyx_k_get_fee), 0, 0, 1, 1},
    {&__pyx_n_s_get_maker_order_type, __pyx_k_get_maker_order_type, sizeof(__pyx_k_get_maker_order_type), 0, 0, 1, 1},
    {&__pyx_n_s_get_order_book, __pyx_k_get_order_book, sizeof(__pyx_k_get_order_book), 0, 0, 1, 1},
    {&__pyx_n_s_get_order_price_quantum, __pyx_k_get_order_price_quantum, sizeof(__pyx_k_get_order_price_quantum), 0, 0, 1, 1},
    {&__pyx_n_s_get_price, __pyx_k_get_price, sizeof(__pyx_k_get_price), 0, 0, 1, 1},
    {&__pyx_n_s_get_taker_order_type, __pyx_k_get_taker_order_type, sizeof(__pyx_k_get_taker_order_type), 0, 0, 1, 1},
    {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_client_config_config, __pyx_k_hummingbot_client_config_config, sizeof(__pyx_k_hummingbot_client_config_config), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_connector_budget_chec, __pyx_k_hummingbot_connector_budget_chec, sizeof(__pyx_k_hummingbot_connector_budget_chec), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_connector_connector_m, __pyx_k_hummingbot_connector_connector_m, sizeof(__pyx_k_hummingbot_connector_connector_m), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_connector_exchange_ba, __pyx_k_hummingbot_connector_exchange_ba, sizeof(__pyx_k_hummingbot_connector_exchange_ba), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_connector_exchange_pa, __pyx_k_hummingbot_connector_exchange_pa, sizeof(__pyx_k_hummingbot_connector_exchange_pa), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_connector_exchange_pa_2, __pyx_k_hummingbot_connector_exchange_pa_2, sizeof(__pyx_k_hummingbot_connector_exchange_pa_2), 0, 0, 1, 1},
    {&__pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_k_hummingbot_connector_exchange_pa_3, sizeof(__pyx_k_hummingbot_connector_exchange_pa_3), 0, 0, 1, 0},
    {&__pyx_n_s_hummingbot_core_clock, __pyx_k_hummingbot_core_clock, sizeof(__pyx_k_hummingbot_core_clock), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_core_data_type_cancel, __pyx_k_hummingbot_core_data_type_cancel, sizeof(__pyx_k_hummingbot_core_data_type_cancel), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_core_data_type_common, __pyx_k_hummingbot_core_data_type_common, sizeof(__pyx_k_hummingbot_core_data_type_common), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_core_data_type_compos, __pyx_k_hummingbot_core_data_type_compos, sizeof(__pyx_k_hummingbot_core_data_type_compos), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_core_data_type_limit, __pyx_k_hummingbot_core_data_type_limit, sizeof(__pyx_k_hummingbot_core_data_type_limit), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_core_data_type_order, __pyx_k_hummingbot_core_data_type_order, sizeof(__pyx_k_hummingbot_core_data_type_order), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_core_data_type_order_2, __pyx_k_hummingbot_core_data_type_order_2, sizeof(__pyx_k_hummingbot_core_data_type_order_2), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_core_event_events, __pyx_k_hummingbot_core_event_events, sizeof(__pyx_k_hummingbot_core_event_events), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_core_network_iterator, __pyx_k_hummingbot_core_network_iterator, sizeof(__pyx_k_hummingbot_core_network_iterator), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_core_utils_async_util, __pyx_k_hummingbot_core_utils_async_util, sizeof(__pyx_k_hummingbot_core_utils_async_util), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_core_utils_estimate_f, __pyx_k_hummingbot_core_utils_estimate_f, sizeof(__pyx_k_hummingbot_core_utils_estimate_f), 0, 0, 1, 1},
    {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
    {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
    {&__pyx_n_s_init, __pyx_k_init, sizeof(__pyx_k_init), 0, 0, 1, 1},
    {&__pyx_n_s_init___locals_lambda, __pyx_k_init___locals_lambda, sizeof(__pyx_k_init___locals_lambda), 0, 0, 1, 1},
    {&__pyx_n_s_init_paper_trade_market, __pyx_k_init_paper_trade_market, sizeof(__pyx_k_init_paper_trade_market), 0, 0, 1, 1},
    {&__pyx_n_s_initializing, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
    {&__pyx_n_s_inspect, __pyx_k_inspect, sizeof(__pyx_k_inspect), 0, 0, 1, 1},
    {&__pyx_n_s_is_buy, __pyx_k_is_buy, sizeof(__pyx_k_is_buy), 0, 0, 1, 1},
    {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
    {&__pyx_n_s_is_finite, __pyx_k_is_finite, sizeof(__pyx_k_is_finite), 0, 0, 1, 1},
    {&__pyx_n_s_is_maker, __pyx_k_is_maker, sizeof(__pyx_k_is_maker), 0, 0, 1, 1},
    {&__pyx_kp_u_isenabled, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
    {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
    {&__pyx_n_s_keys, __pyx_k_keys, sizeof(__pyx_k_keys), 0, 0, 1, 1},
    {&__pyx_n_s_kwargs, __pyx_k_kwargs, sizeof(__pyx_k_kwargs), 0, 0, 1, 1},
    {&__pyx_n_s_limit_orders, __pyx_k_limit_orders, sizeof(__pyx_k_limit_orders), 0, 0, 1, 1},
    {&__pyx_n_s_limit_orders_map_ptr, __pyx_k_limit_orders_map_ptr, sizeof(__pyx_k_limit_orders_map_ptr), 0, 0, 1, 1},
    {&__pyx_n_s_log10, __pyx_k_log10, sizeof(__pyx_k_log10), 0, 0, 1, 1},
    {&__pyx_n_s_logger, __pyx_k_logger, sizeof(__pyx_k_logger), 0, 0, 1, 1},
    {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
    {&__pyx_n_s_market, __pyx_k_market, sizeof(__pyx_k_market), 0, 0, 1, 1},
    {&__pyx_n_s_match_trade_to_limit_orders, __pyx_k_match_trade_to_limit_orders, sizeof(__pyx_k_match_trade_to_limit_orders), 0, 0, 1, 1},
    {&__pyx_n_s_math, __pyx_k_math, sizeof(__pyx_k_math), 0, 0, 1, 1},
    {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
    {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
    {&__pyx_kp_u_needed_vs, __pyx_k_needed_vs, sizeof(__pyx_k_needed_vs), 0, 1, 0, 0},
    {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
    {&__pyx_kp_u_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 1, 0, 0},
    {&__pyx_kp_u_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 1, 0, 0},
    {&__pyx_n_s_on_hold_balances, __pyx_k_on_hold_balances, sizeof(__pyx_k_on_hold_balances), 0, 0, 1, 1},
    {&__pyx_n_s_order_book, __pyx_k_order_book, sizeof(__pyx_k_order_book), 0, 0, 1, 1},
    {&__pyx_n_s_order_book_create_function, __pyx_k_order_book_create_function, sizeof(__pyx_k_order_book_create_function), 0, 0, 1, 1},
    {&__pyx_n_s_order_book_tracker, __pyx_k_order_book_tracker, sizeof(__pyx_k_order_book_tracker), 0, 0, 1, 1},
    {&__pyx_n_s_order_books, __pyx_k_order_books, sizeof(__pyx_k_order_books), 0, 0, 1, 1},
    {&__pyx_n_u_order_books_initialized, __pyx_k_order_books_initialized, sizeof(__pyx_k_order_books_initialized), 0, 1, 0, 1},
    {&__pyx_n_s_order_collateral, __pyx_k_order_collateral, sizeof(__pyx_k_order_collateral), 0, 0, 1, 1},
    {&__pyx_n_s_order_filled_events_from_order_b, __pyx_k_order_filled_events_from_order_b, sizeof(__pyx_k_order_filled_events_from_order_b), 0, 0, 1, 1},
    {&__pyx_n_s_order_id, __pyx_k_order_id, sizeof(__pyx_k_order_id), 0, 0, 1, 1},
    {&__pyx_n_s_order_side, __pyx_k_order_side, sizeof(__pyx_k_order_side), 0, 0, 1, 1},
    {&__pyx_n_s_order_size_decimals, __pyx_k_order_size_decimals, sizeof(__pyx_k_order_size_decimals), 0, 0, 1, 1},
    {&__pyx_n_s_order_size_precision, __pyx_k_order_size_precision, sizeof(__pyx_k_order_size_precision), 0, 0, 1, 1},
    {&__pyx_n_s_order_type, __pyx_k_order_type, sizeof(__pyx_k_order_type), 0, 0, 1, 1},
    {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
    {&__pyx_n_s_popleft, __pyx_k_popleft, sizeof(__pyx_k_popleft), 0, 0, 1, 1},
    {&__pyx_n_s_populate_collateral_entries, __pyx_k_populate_collateral_entries, sizeof(__pyx_k_populate_collateral_entries), 0, 0, 1, 1},
    {&__pyx_n_s_potential_returns, __pyx_k_potential_returns, sizeof(__pyx_k_potential_returns), 0, 0, 1, 1},
    {&__pyx_n_s_price, __pyx_k_price, sizeof(__pyx_k_price), 0, 0, 1, 1},
    {&__pyx_n_s_price_decimals, __pyx_k_price_decimals, sizeof(__pyx_k_price_decimals), 0, 0, 1, 1},
    {&__pyx_n_s_price_precision, __pyx_k_price_precision, sizeof(__pyx_k_price_precision), 0, 0, 1, 1},
    {&__pyx_n_s_ptm_logger, __pyx_k_ptm_logger, sizeof(__pyx_k_ptm_logger), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_OrderBookMarketOr, __pyx_k_pyx_unpickle_OrderBookMarketOr, sizeof(__pyx_k_pyx_unpickle_OrderBookMarketOr), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_OrderBookTradeLis, __pyx_k_pyx_unpickle_OrderBookTradeLis, sizeof(__pyx_k_pyx_unpickle_OrderBookTradeLis), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_QuantizationParam, __pyx_k_pyx_unpickle_QuantizationParam, sizeof(__pyx_k_pyx_unpickle_QuantizationParam), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_unpickle_QueuedOrder, __pyx_k_pyx_unpickle_QueuedOrder, sizeof(__pyx_k_pyx_unpickle_QueuedOrder), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
    {&__pyx_n_s_quantity, __pyx_k_quantity, sizeof(__pyx_k_quantity), 0, 0, 1, 1},
    {&__pyx_n_s_quantize, __pyx_k_quantize, sizeof(__pyx_k_quantize), 0, 0, 1, 1},
    {&__pyx_n_s_quote_asset, __pyx_k_quote_asset, sizeof(__pyx_k_quote_asset), 0, 0, 1, 1},
    {&__pyx_n_s_quote_currency, __pyx_k_quote_currency, sizeof(__pyx_k_quote_currency), 0, 0, 1, 1},
    {&__pyx_n_s_random, __pyx_k_random, sizeof(__pyx_k_random), 0, 0, 1, 1},
    {&__pyx_n_s_random_order_id, __pyx_k_random_order_id, sizeof(__pyx_k_random_order_id), 0, 0, 1, 1},
    {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
    {&__pyx_n_s_ready, __pyx_k_ready, sizeof(__pyx_k_ready), 0, 0, 1, 1},
    {&__pyx_n_s_record_filled_order, __pyx_k_record_filled_order, sizeof(__pyx_k_record_filled_order), 0, 0, 1, 1},
    {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
    {&__pyx_kp_u_required_for_the_order, __pyx_k_required_for_the_order, sizeof(__pyx_k_required_for_the_order), 0, 1, 0, 0},
    {&__pyx_n_s_results, __pyx_k_results, sizeof(__pyx_k_results), 0, 0, 1, 1},
    {&__pyx_n_s_return, __pyx_k_return, sizeof(__pyx_k_return), 0, 0, 1, 1},
    {&__pyx_n_s_rounding, __pyx_k_rounding, sizeof(__pyx_k_rounding), 0, 0, 1, 1},
    {&__pyx_n_s_s_decimal_0, __pyx_k_s_decimal_0, sizeof(__pyx_k_s_decimal_0), 0, 0, 1, 1},
    {&__pyx_n_s_safe_ensure_future, __pyx_k_safe_ensure_future, sizeof(__pyx_k_safe_ensure_future), 0, 0, 1, 1},
    {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
    {&__pyx_kp_s_self__ask_limit_orders_self__bid, __pyx_k_self__ask_limit_orders_self__bid, sizeof(__pyx_k_self__ask_limit_orders_self__bid), 0, 0, 1, 0},
    {&__pyx_n_s_sell, __pyx_k_sell, sizeof(__pyx_k_sell), 0, 0, 1, 1},
    {&__pyx_n_u_sell, __pyx_k_sell, sizeof(__pyx_k_sell), 0, 1, 0, 1},
    {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
    {&__pyx_n_s_set_balance, __pyx_k_set_balance, sizeof(__pyx_k_set_balance), 0, 0, 1, 1},
    {&__pyx_n_s_set_order_book_tracker, __pyx_k_set_order_book_tracker, sizeof(__pyx_k_set_order_book_tracker), 0, 0, 1, 1},
    {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_simulate_buy, __pyx_k_simulate_buy, sizeof(__pyx_k_simulate_buy), 0, 0, 1, 1},
    {&__pyx_n_s_simulate_sell, __pyx_k_simulate_sell, sizeof(__pyx_k_simulate_sell), 0, 0, 1, 1},
    {&__pyx_n_s_sleep, __pyx_k_sleep, sizeof(__pyx_k_sleep), 0, 0, 1, 1},
    {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
    {&__pyx_n_s_split_trading_pair, __pyx_k_split_trading_pair, sizeof(__pyx_k_split_trading_pair), 0, 0, 1, 1},
    {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
    {&__pyx_n_s_start_network, __pyx_k_start_network, sizeof(__pyx_k_start_network), 0, 0, 1, 1},
    {&__pyx_n_s_state, __pyx_k_state, sizeof(__pyx_k_state), 0, 0, 1, 1},
    {&__pyx_n_s_status_dict, __pyx_k_status_dict, sizeof(__pyx_k_status_dict), 0, 0, 1, 1},
    {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
    {&__pyx_n_s_stop_network, __pyx_k_stop_network, sizeof(__pyx_k_stop_network), 0, 0, 1, 1},
    {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
    {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
    {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
    {&__pyx_n_s_target_market, __pyx_k_target_market, sizeof(__pyx_k_target_market), 0, 0, 1, 1},
    {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
    {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
    {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
    {&__pyx_n_s_timeout_seconds, __pyx_k_timeout_seconds, sizeof(__pyx_k_timeout_seconds), 0, 0, 1, 1},
    {&__pyx_n_s_trading_pair, __pyx_k_trading_pair, sizeof(__pyx_k_trading_pair), 0, 0, 1, 1},
    {&__pyx_n_s_trading_pair_str, __pyx_k_trading_pair_str, sizeof(__pyx_k_trading_pair_str), 0, 0, 1, 1},
    {&__pyx_n_s_trigger_event_async, __pyx_k_trigger_event_async, sizeof(__pyx_k_trigger_event_async), 0, 0, 1, 1},
    {&__pyx_n_s_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
    {&__pyx_n_s_typing, __pyx_k_typing, sizeof(__pyx_k_typing), 0, 0, 1, 1},
    {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
    {&__pyx_n_s_upper, __pyx_k_upper, sizeof(__pyx_k_upper), 0, 0, 1, 1},
    {&__pyx_n_s_use_setstate, __pyx_k_use_setstate, sizeof(__pyx_k_use_setstate), 0, 0, 1, 1},
    {&__pyx_n_u_utf8, __pyx_k_utf8, sizeof(__pyx_k_utf8), 0, 1, 0, 1},
    {&__pyx_n_s_val, __pyx_k_val, sizeof(__pyx_k_val), 0, 0, 1, 1},
    {&__pyx_n_s_vals, __pyx_k_vals, sizeof(__pyx_k_vals), 0, 0, 1, 1},
    {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
    {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
    {&__pyx_n_s_warning, __pyx_k_warning, sizeof(__pyx_k_warning), 0, 0, 1, 1},
    {0, 0, 0, 0, 0, 0, 0}
  };
  return __Pyx_InitStrings(__pyx_string_tab);
}
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 118, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 186, __pyx_L1_error)
  __pyx_builtin_AssertionError = __Pyx_GetBuiltinName(__pyx_n_s_AssertionError); if (!__pyx_builtin_AssertionError) __PYX_ERR(0, 191, __pyx_L1_error)
  __pyx_builtin_all = __Pyx_GetBuiltinName(__pyx_n_s_all); if (!__pyx_builtin_all) __PYX_ERR(0, 233, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 330, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(2, 984, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":984
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_u_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple_)) __PYX_ERR(2, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":990
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(2, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":125
 *             self._market.match_trade_to_limit_orders(event_object)
 *         except Exception as e:
 *             self.logger().error("Error call trade listener.", exc_info=True)             # <<<<<<<<<<<<<<
 * 
 * cdef class OrderBookMarketOrderFillListener(EventListener):
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_u_Error_call_trade_listener); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":186
 *     @classmethod
 *     def random_order_id(cls, order_side: str, trading_pair: str) -> str:
 *         vals = [random.choice(range(0, 256)) for i in range(0, 13)]             # <<<<<<<<<<<<<<
 *         return f"{order_side}://" + trading_pair + "/" + "".join([f"{val:02x}" for val in vals])
 * 
 */
  __pyx_tuple__8 = PyTuple_Pack(2, __pyx_int_0, __pyx_int_256); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":621
 *                         self.c_execute_sell(front_order.order_id, front_order.trading_pair, front_order.amount)
 *                 except Exception as e:
 *                     self.logger().error("Error executing queued order.", exc_info=True)             # <<<<<<<<<<<<<<
 *             else:
 *                 return
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_u_Error_executing_queued_order); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 621, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":833
 *                 self.c_process_limit_ask_order(limit_orders_map_ptr, map_it_ptr, orders_it)
 *         except Exception as e:
 *             self.logger().error(f"Error processing limit order.", exc_info=True)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_process_crossed_limit_orders_for_trading_pair(self,
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_u_Error_processing_limit_order); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 833, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":999
 *             return cancellation_results
 *         except Exception as err:
 *             self.logger().error(f"Error canceling order.", exc_info=True)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_cancel(self, str trading_pair_str, str client_order_id):
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_u_Error_canceling_order); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 999, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xa74d68e, 0x820b24c, 0x7857e48):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xa74d68e, 0x820b24c, 0x7857e48) = (order_size_decimals, order_size_precision, price_decimals, price_precision, trading_pair))" % __pyx_checksum
 */
  __pyx_tuple__33 = PyTuple_Pack(3, __pyx_int_175429262, __pyx_int_136360524, __pyx_int_126189128); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_tuple__34 = PyTuple_Pack(3, __pyx_int_251288030, __pyx_int_57171680, __pyx_int_95353964); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);
  __pyx_tuple__35 = PyTuple_Pack(3, __pyx_int_29025389, __pyx_int_24020995, __pyx_int_58563262); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":52
 * 
 * ptm_logger = None
 * s_decimal_0 = Decimal(0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__37 = PyTuple_Pack(1, __pyx_int_0); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_tuple__38 = PyTuple_Pack(4, __pyx_n_s_self, __pyx_n_s_state, __pyx_n_s_dict_2, __pyx_n_s_use_setstate); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);
  __pyx_codeobj__39 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__39)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_QuantizationParams, (type(self), 0xa74d68e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_QuantizationParams__set_state(self, __pyx_state)
 */
  __pyx_tuple__40 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);
  __pyx_codeobj__41 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__40, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__41)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_QueuedOrder, (type(self), 0xefa59de, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_QueuedOrder__set_state(self, __pyx_state)
 */
  __pyx_codeobj__43 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__40, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__43)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_OrderBookTradeListener, (type(self), 0x1bae46d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_OrderBookTradeListener__set_state(self, __pyx_state)
 */
  __pyx_codeobj__45 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__40, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__45)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 4, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__38, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_OrderBookMarketOrderFillListener, (type(self), 0x1bae46d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(self, __pyx_state)
 */
  __pyx_codeobj__47 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__40, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__47)) __PYX_ERR(1, 16, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":184
 *         return self._budget_checker
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def random_order_id(cls, order_side: str, trading_pair: str) -> str:
 *         vals = [random.choice(range(0, 256)) for i in range(0, 13)]
 */
  __pyx_tuple__48 = PyTuple_Pack(6, __pyx_n_s_cls, __pyx_n_s_order_side, __pyx_n_s_trading_pair, __pyx_n_s_vals, __pyx_n_s_i, __pyx_n_s_val); if (unlikely(!__pyx_tuple__48)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__48);
  __Pyx_GIVEREF(__pyx_tuple__48);
  __pyx_codeobj__49 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__48, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_random_order_id, 184, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__49)) __PYX_ERR(0, 184, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":189
 *         return f"{order_side}://" + trading_pair + "/" + "".join([f"{val:02x}" for val in vals])
 * 
 *     def init_paper_trade_market(self):             # <<<<<<<<<<<<<<
 *         for trading_pair_str, order_book in self.order_book_tracker.order_books.items():
 *             assert type(order_book) is CompositeOrderBook
 */
  __pyx_tuple__50 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_trading_pair_str, __pyx_n_s_order_book, __pyx_n_s_base_asset, __pyx_n_s_quote_asset); if (unlikely(!__pyx_tuple__50)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__50);
  __Pyx_GIVEREF(__pyx_tuple__50);
  __pyx_codeobj__51 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__50, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_init_paper_trade_market, 189, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__51)) __PYX_ERR(0, 189, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":199
 *             )
 * 
 *     def split_trading_pair(self, trading_pair: str) -> Tuple[str, str]:             # <<<<<<<<<<<<<<
 *         return self._target_market.split_trading_pair(trading_pair)
 * 
 */
  __pyx_tuple__52 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_trading_pair); if (unlikely(!__pyx_tuple__52)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__52);
  __Pyx_GIVEREF(__pyx_tuple__52);
  __pyx_codeobj__53 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__52, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_split_trading_pair, 199, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__53)) __PYX_ERR(0, 199, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":299
 *         ExchangeBase.c_start(self, clock, timestamp)
 * 
 *     async def start_network(self):             # <<<<<<<<<<<<<<
 *         await self.stop_network()
 *         self.order_book_tracker.start()
 */
  __pyx_tuple__54 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__54)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__54);
  __Pyx_GIVEREF(__pyx_tuple__54);
  __pyx_codeobj__12 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__54, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_start_network, 299, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__12)) __PYX_ERR(0, 299, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":303
 *         self.order_book_tracker.start()
 * 
 *     async def stop_network(self):             # <<<<<<<<<<<<<<
 *         self.order_book_tracker.stop()
 * 
 */
  __pyx_codeobj__13 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__54, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_stop_network, 303, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__13)) __PYX_ERR(0, 303, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":306
 *         self.order_book_tracker.stop()
 * 
 *     async def check_network(self) -> NetworkStatus:             # <<<<<<<<<<<<<<
 *         return NetworkStatus.CONNECTED
 * 
 */
  __pyx_codeobj__14 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__54, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_check_network, 306, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__14)) __PYX_ERR(0, 306, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":945
 *         return self.available_balances.get(currency.upper(), s_decimal_0)
 * 
 *     async def cancel_all(self, timeout_seconds: float) -> List[CancellationResult]:             # <<<<<<<<<<<<<<
 *         cdef:
 *             LimitOrders *limit_orders_map_ptr
 */
  __pyx_tuple__57 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_timeout_seconds, __pyx_n_s_limit_orders_map_ptr, __pyx_n_s_cancellation_results, __pyx_n_s_trading_pair_str, __pyx_n_s_results); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(0, 945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__57, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_cancel_all, 945, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(0, 945, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1051
 *             return Decimal(f"1e-10")
 * 
 *     def get_order_price_quantum(self, trading_pair: str, price: Decimal) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_get_order_price_quantum(trading_pair, price)
 * 
 */
  __pyx_tuple__58 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_trading_pair, __pyx_n_s_price); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(0, 1051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);
  __pyx_codeobj__59 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__58, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_get_order_price_quantum, 1051, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__59)) __PYX_ERR(0, 1051, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1087
 *         return (amount // order_size_quantum) * order_size_quantum
 * 
 *     def get_available_balance(self, currency: str) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_get_available_balance(currency)
 * 
 */
  __pyx_tuple__61 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_currency); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__61, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_get_available_balance, 1087, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(0, 1087, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1090
 *         return self.c_get_available_balance(currency)
 * 
 *     def get_all_balances(self) -> Dict[str, Decimal]:             # <<<<<<<<<<<<<<
 *         return self._account_balances.copy()
 * 
 */
  __pyx_codeobj__63 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__54, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_get_all_balances, 1090, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__63)) __PYX_ERR(0, 1090, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1094
 * 
 *     # <editor-fold desc="Python wrapper for cdef functions">
 *     def match_trade_to_limit_orders(self, event_object: OrderBookTradeEvent):             # <<<<<<<<<<<<<<
 *         self.c_match_trade_to_limit_orders(event_object)
 * 
 */
  __pyx_tuple__64 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_event_object); if (unlikely(!__pyx_tuple__64)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__64);
  __Pyx_GIVEREF(__pyx_tuple__64);
  __pyx_codeobj__65 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__64, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_match_trade_to_limit_orders, 1094, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__65)) __PYX_ERR(0, 1094, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1097
 *         self.c_match_trade_to_limit_orders(event_object)
 * 
 *     def set_balance(self, currency: str, balance: Decimal):             # <<<<<<<<<<<<<<
 *         self.c_set_balance(currency, balance)
 *     # </editor-fold>
 */
  __pyx_tuple__66 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_currency, __pyx_n_s_balance); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(0, 1097, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);
  __pyx_codeobj__67 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_set_balance, 1097, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__67)) __PYX_ERR(0, 1097, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1101
 *     # </editor-fold>
 * 
 *     def get_price(self, trading_pair: str, is_buy: bool) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_get_price(trading_pair, is_buy)
 * 
 */
  __pyx_tuple__68 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_trading_pair, __pyx_n_s_is_buy); if (unlikely(!__pyx_tuple__68)) __PYX_ERR(0, 1101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__68);
  __Pyx_GIVEREF(__pyx_tuple__68);
  __pyx_codeobj__69 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__68, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_get_price, 1101, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__69)) __PYX_ERR(0, 1101, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1104
 *         return self.c_get_price(trading_pair, is_buy)
 * 
 *     def buy(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,             # <<<<<<<<<<<<<<
 *             price: Decimal = s_decimal_0, **kwargs) -> str:
 *         return self.c_buy(trading_pair, amount, order_type, price, kwargs)
 */
  __pyx_tuple__70 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_trading_pair, __pyx_n_s_amount, __pyx_n_s_order_type, __pyx_n_s_price, __pyx_n_s_kwargs); if (unlikely(!__pyx_tuple__70)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__70);
  __Pyx_GIVEREF(__pyx_tuple__70);
  __pyx_codeobj__71 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__70, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_buy, 1104, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__71)) __PYX_ERR(0, 1104, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1108
 *         return self.c_buy(trading_pair, amount, order_type, price, kwargs)
 * 
 *     def sell(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,             # <<<<<<<<<<<<<<
 *              price: Decimal = s_decimal_0, **kwargs) -> str:
 *         return self.c_sell(trading_pair, amount, order_type, price, kwargs)
 */
  __pyx_codeobj__72 = (PyObject*)__Pyx_PyCode_New(5, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_VARKEYWORDS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__70, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_sell, 1108, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__72)) __PYX_ERR(0, 1108, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1112
 *         return self.c_sell(trading_pair, amount, order_type, price, kwargs)
 * 
 *     def cancel(self, trading_pair: str, client_order_id: str):             # <<<<<<<<<<<<<<
 *         return self.c_cancel(trading_pair, client_order_id)
 * 
 */
  __pyx_tuple__73 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_trading_pair, __pyx_n_s_client_order_id); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(0, 1112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);
  __pyx_codeobj__74 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__73, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_cancel, 1112, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__74)) __PYX_ERR(0, 1112, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1115
 *         return self.c_cancel(trading_pair, client_order_id)
 * 
 *     def get_fee(self,             # <<<<<<<<<<<<<<
 *                 base_currency: str,
 *                 quote_currency: str,
 */
  __pyx_tuple__75 = PyTuple_Pack(8, __pyx_n_s_self, __pyx_n_s_base_currency, __pyx_n_s_quote_currency, __pyx_n_s_order_type, __pyx_n_s_order_side, __pyx_n_s_amount, __pyx_n_s_price, __pyx_n_s_is_maker); if (unlikely(!__pyx_tuple__75)) __PYX_ERR(0, 1115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__75);
  __Pyx_GIVEREF(__pyx_tuple__75);
  __pyx_codeobj__76 = (PyObject*)__Pyx_PyCode_New(8, 0, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__75, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_get_fee, 1115, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__76)) __PYX_ERR(0, 1115, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1125
 *         return self.c_get_fee(base_currency, quote_currency, order_type, order_side, amount, price, is_maker)
 * 
 *     def get_order_book(self, trading_pair: str) -> OrderBook:             # <<<<<<<<<<<<<<
 *         return self.c_get_order_book(trading_pair)
 * 
 */
  __pyx_codeobj__77 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__52, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_get_order_book, 1125, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__77)) __PYX_ERR(0, 1125, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1128
 *         return self.c_get_order_book(trading_pair)
 * 
 *     def get_maker_order_type(self):             # <<<<<<<<<<<<<<
 *         return OrderType.LIMIT
 * 
 */
  __pyx_codeobj__78 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__54, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_get_maker_order_type, 1128, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__78)) __PYX_ERR(0, 1128, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1131
 *         return OrderType.LIMIT
 * 
 *     def get_taker_order_type(self):             # <<<<<<<<<<<<<<
 *         return OrderType.LIMIT
 * 
 */
  __pyx_codeobj__79 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__54, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_get_taker_order_type, 1131, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__79)) __PYX_ERR(0, 1131, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1134
 *         return OrderType.LIMIT
 * 
 *     async def trigger_event_async(self,             # <<<<<<<<<<<<<<
 *                                   event_tag,
 *                                   event):
 */
  __pyx_tuple__80 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_event_tag, __pyx_n_s_event); if (unlikely(!__pyx_tuple__80)) __PYX_ERR(0, 1134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__80);
  __Pyx_GIVEREF(__pyx_tuple__80);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS|CO_COROUTINE, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__80, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_connector_exchange_pa_3, __pyx_n_s_trigger_event_async, 1134, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 1134, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ask_limit_orders,self._bid_limit_orders,self._events,self._limit_order_expiration_set cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__81 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__54, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__81)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._ask_limit_orders,self._bid_limit_orders,self._events,self._limit_order_expiration_set cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ask_limit_orders,self._bid_limit_orders,self._events,self._limit_order_expiration_set cannot be converted to a Python object for pickling"
 */
  __pyx_codeobj__82 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__40, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__82)) __PYX_ERR(1, 3, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __pyx_unpickle_QuantizationParams(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__83 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__83)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__83);
  __Pyx_GIVEREF(__pyx_tuple__83);
  __pyx_codeobj__84 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__83, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_QuantizationParam, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__84)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__85 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__83, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_QueuedOrder, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__85)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__86 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__83, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_OrderBookTradeLis, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__86)) __PYX_ERR(1, 1, __pyx_L1_error)
  __pyx_codeobj__87 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__83, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_OrderBookMarketOr, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__87)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  __pyx_umethod_PyUnicode_Type_upper.type = (PyObject*)&PyUnicode_Type;
  __pyx_umethod_PyUnicode_Type_upper.method_name = &__pyx_n_s_upper;
  if (__Pyx_CreateStringTabAndInitStrings() < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_0_0 = PyFloat_FromDouble(0.0); if (unlikely(!__pyx_float_0_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1e6 = PyFloat_FromDouble(1e6); if (unlikely(!__pyx_float_1e6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_5_0 = PyFloat_FromDouble(5.0); if (unlikely(!__pyx_float_5_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_0_01 = PyFloat_FromDouble(0.01); if (unlikely(!__pyx_float_0_01)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_float_1eneg_7 = PyFloat_FromDouble(1e-7); if (unlikely(!__pyx_float_1eneg_7)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_256 = PyInt_FromLong(256); if (unlikely(!__pyx_int_256)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_24020995 = PyInt_FromLong(24020995L); if (unlikely(!__pyx_int_24020995)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_29025389 = PyInt_FromLong(29025389L); if (unlikely(!__pyx_int_29025389)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_57171680 = PyInt_FromLong(57171680L); if (unlikely(!__pyx_int_57171680)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_58563262 = PyInt_FromLong(58563262L); if (unlikely(!__pyx_int_58563262)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_95353964 = PyInt_FromLong(95353964L); if (unlikely(!__pyx_int_95353964)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_126189128 = PyInt_FromLong(126189128L); if (unlikely(!__pyx_int_126189128)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_136360524 = PyInt_FromLong(136360524L); if (unlikely(!__pyx_int_136360524)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_175429262 = PyInt_FromLong(175429262L); if (unlikely(!__pyx_int_175429262)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_251288030 = PyInt_FromLong(251288030L); if (unlikely(!__pyx_int_251288030)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  /* AssertionsEnabled.init */
  if (likely(__Pyx_init_assertions_enabled() == 0)); else

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams_spec, NULL); if (unlikely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams)) __PYX_ERR(0, 55, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams_spec, __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams) < 0) __PYX_ERR(0, 55, __pyx_L1_error)
  #else
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams = &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams) < 0) __PYX_ERR(0, 55, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams->tp_dictoffset && __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_QuantizationParams_2, (PyObject *) __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams) < 0) __PYX_ERR(0, 55, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams) < 0) __PYX_ERR(0, 55, __pyx_L1_error)
  #endif
  __pyx_t_1 = PyImport_ImportModule("hummingbot.connector.exchange_base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.connector.exchange_base", "ExchangeBase", sizeof(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_9connector_13exchange_base_ExchangeBase = (struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase); if (unlikely(!__pyx_vtabptr_10hummingbot_9connector_13exchange_base_ExchangeBase)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange = &__pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.__pyx_base = *__pyx_vtabptr_10hummingbot_9connector_13exchange_base_ExchangeBase;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.__pyx_base.__pyx_base.__pyx_base.__pyx_base.c_start = (PyObject *(*)(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *, struct __pyx_obj_10hummingbot_4core_5clock_Clock *, double))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_start;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.__pyx_base.__pyx_base.__pyx_base.__pyx_base.c_tick = (PyObject *(*)(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *, double))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_tick;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.__pyx_base.__pyx_base.c_buy = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_buy *__pyx_optional_args))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_buy;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.__pyx_base.__pyx_base.c_sell = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_sell *__pyx_optional_args))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_sell;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.__pyx_base.__pyx_base.c_cancel = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_cancel;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.__pyx_base.__pyx_base.c_get_balance = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_balance;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.__pyx_base.__pyx_base.c_get_available_balance = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_available_balance;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.__pyx_base.__pyx_base.c_get_order_price_quantum = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_order_price_quantum;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.__pyx_base.__pyx_base.c_get_order_size_quantum = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_order_size_quantum;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.__pyx_base.__pyx_base.c_quantize_order_price = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_quantize_order_price;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.__pyx_base.__pyx_base.c_quantize_order_amount = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_quantize_order_amount *__pyx_optional_args))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_quantize_order_amount;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.__pyx_base.c_get_order_book = (struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *(*)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_order_book;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.__pyx_base.c_get_fee = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_get_fee *__pyx_optional_args))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_get_fee;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.c_execute_buy = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, PyObject *, PyObject *, PyObject *))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_execute_buy;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.c_execute_sell = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, PyObject *, PyObject *, PyObject *))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_execute_sell;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.c_process_market_orders = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_process_market_orders;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.c_set_balance = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, PyObject *, PyObject *))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_set_balance;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.c_delete_limit_order = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator const ))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_delete_limit_order;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.c_process_limit_order = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, int, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_process_limit_order;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.c_process_limit_bid_order = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_process_limit_bid_order;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.c_process_limit_ask_order = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_SingleTradingPairLimitOrdersIterator))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_process_limit_ask_order;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.c_process_crossed_limit_orders_for_trading_pair = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, int, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrdersIterator *))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_process_crossed_limit_orders_for_trading_pair;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.c_process_crossed_limit_orders = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_process_crossed_limit_orders;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.c_match_trade_to_limit_orders = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, PyObject *))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_match_trade_to_limit_orders;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange.c_cancel_order_from_orders_map = (PyObject *(*)(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange *, __pyx_t_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_LimitOrders *, PyObject *, struct __pyx_opt_args_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_cancel_order_from_orders_map *__pyx_optional_args))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_c_cancel_order_from_orders_map;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange)) __PYX_ERR(0, 143, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange_spec, __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  #else
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange = &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange->tp_base = __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange->tp_dictoffset && __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_vtabptr_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_PaperTradeExchange, (PyObject *) __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  if (__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange->tp_weaklistoffset == 0) __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange->tp_weaklistoffset = offsetof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_base.__weakref__);
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange) < 0) __PYX_ERR(0, 143, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder_spec, NULL); if (unlikely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder)) __PYX_ERR(0, 73, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder_spec, __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder) < 0) __PYX_ERR(0, 73, __pyx_L1_error)
  #else
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder = &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder) < 0) __PYX_ERR(0, 73, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder->tp_dictoffset && __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_QueuedOrder_2, (PyObject *) __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder) < 0) __PYX_ERR(0, 73, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder) < 0) __PYX_ERR(0, 73, __pyx_L1_error)
  #endif
  __pyx_t_2 = PyImport_ImportModule("hummingbot.core.event.event_listener"); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener = __Pyx_ImportType_3_0_12(__pyx_t_2, "hummingbot.core.event.event_listener", "EventListener", sizeof(struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_5event_14event_listener_EventListener = (struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_5event_14event_listener_EventListener)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener = &__pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener.__pyx_base = *__pyx_vtabptr_10hummingbot_4core_5event_14event_listener_EventListener;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener.__pyx_base.c_call = (PyObject *(*)(struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *, PyObject *))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_c_call;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_3 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener_spec, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener)) __PYX_ERR(0, 113, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener_spec, __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener) < 0) __PYX_ERR(0, 113, __pyx_L1_error)
  #else
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener = &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener->tp_base = __pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener) < 0) __PYX_ERR(0, 113, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener->tp_dictoffset && __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener, __pyx_vtabptr_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener) < 0) __PYX_ERR(0, 113, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener) < 0) __PYX_ERR(0, 113, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_OrderBookTradeListener, (PyObject *) __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener) < 0) __PYX_ERR(0, 113, __pyx_L1_error)
  if (__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener->tp_weaklistoffset == 0) __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener->tp_weaklistoffset = offsetof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener, __pyx_base.__weakref__);
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener) < 0) __PYX_ERR(0, 113, __pyx_L1_error)
  #endif
  __pyx_vtabptr_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener = &__pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener.__pyx_base = *__pyx_vtabptr_10hummingbot_4core_5event_14event_listener_EventListener;
  __pyx_vtable_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener.__pyx_base.c_call = (PyObject *(*)(struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *, PyObject *))__pyx_f_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_c_call;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_3 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener_spec, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener)) __PYX_ERR(0, 127, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener_spec, __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener) < 0) __PYX_ERR(0, 127, __pyx_L1_error)
  #else
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener = &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener->tp_base = __pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener) < 0) __PYX_ERR(0, 127, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener->tp_dictoffset && __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener, __pyx_vtabptr_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener) < 0) __PYX_ERR(0, 127, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener) < 0) __PYX_ERR(0, 127, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_OrderBookMarketOrderFillListener, (PyObject *) __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener) < 0) __PYX_ERR(0, 127, __pyx_L1_error)
  if (__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener->tp_weaklistoffset == 0) __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener->tp_weaklistoffset = offsetof(struct __pyx_obj_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener, __pyx_base.__weakref__);
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener) < 0) __PYX_ERR(0, 127, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network_spec, NULL); if (unlikely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network)) __PYX_ERR(0, 299, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network_spec, __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network) < 0) __PYX_ERR(0, 299, __pyx_L1_error)
  #else
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network = &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network) < 0) __PYX_ERR(0, 299, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network->tp_dictoffset && __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct__start_network->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network_spec, NULL); if (unlikely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network)) __PYX_ERR(0, 303, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network_spec, __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network) < 0) __PYX_ERR(0, 303, __pyx_L1_error)
  #else
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network = &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network) < 0) __PYX_ERR(0, 303, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network->tp_dictoffset && __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_1_stop_network->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network_spec, NULL); if (unlikely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network)) __PYX_ERR(0, 306, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network_spec, __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network) < 0) __PYX_ERR(0, 306, __pyx_L1_error)
  #else
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network = &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network) < 0) __PYX_ERR(0, 306, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network->tp_dictoffset && __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_2_check_network->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all_spec, NULL); if (unlikely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all)) __PYX_ERR(0, 945, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all_spec, __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all) < 0) __PYX_ERR(0, 945, __pyx_L1_error)
  #else
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all = &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all) < 0) __PYX_ERR(0, 945, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all->tp_dictoffset && __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_3_cancel_all->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async_spec, NULL); if (unlikely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async)) __PYX_ERR(0, 1134, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async_spec, __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async) < 0) __PYX_ERR(0, 1134, __pyx_L1_error)
  #else
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async = &__pyx_type_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async) < 0) __PYX_ERR(0, 1134, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async->tp_dictoffset && __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange___pyx_scope_struct_4_trigger_event_async->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.pubsub"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_6pubsub_PubSub = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.pubsub", "PubSub", sizeof(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_6pubsub_PubSub) __PYX_ERR(4, 17, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_6pubsub_PubSub = (struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_4core_6pubsub_PubSub); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_6pubsub_PubSub)) __PYX_ERR(4, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.event.event_reporter"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.event.event_reporter", "EventReporter", sizeof(struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter) __PYX_ERR(5, 4, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_5event_14event_reporter_EventReporter = (struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_reporter_EventReporter*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_5event_14event_reporter_EventReporter)) __PYX_ERR(5, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.event.event_logger"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.event.event_logger", "EventLogger", sizeof(struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger) __PYX_ERR(6, 4, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_5event_12event_logger_EventLogger = (struct __pyx_vtabstruct_10hummingbot_4core_5event_12event_logger_EventLogger*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_5event_12event_logger_EventLogger)) __PYX_ERR(6, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.data_type.order_book_query_result"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.data_type.order_book_query_result", "OrderBookQueryResult", sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult) __PYX_ERR(7, 3, __pyx_L1_error)
  __pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.data_type.order_book_query_result", "ClientOrderBookQueryResult", sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult) __PYX_ERR(7, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_0_12(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(8, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4bool_bool = __Pyx_ImportType_3_0_12(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "bool", sizeof(PyBoolObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyBoolObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_7cpython_4bool_bool) __PYX_ERR(9, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_7complex_complex = __Pyx_ImportType_3_0_12(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "complex", sizeof(PyComplexObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyComplexObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_7cpython_7complex_complex) __PYX_ERR(10, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyArray_Descr),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(2, 202, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyArrayIterObject),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(2, 225, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyArrayMultiIterObject),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(2, 229, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyArrayObject),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(2, 238, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_generic) __PYX_ERR(2, 809, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "number", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_number) __PYX_ERR(2, 811, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_integer) __PYX_ERR(2, 813, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(2, 815, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(2, 817, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(2, 819, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_floating) __PYX_ERR(2, 821, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(2, 823, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(2, 825, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "character", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_character) __PYX_ERR(2, 827, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyUFuncObject),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(2, 866, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.data_type.order_book"); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.data_type.order_book", "OrderBook", sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook) __PYX_ERR(11, 12, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_9data_type_10order_book_OrderBook = (struct __pyx_vtabstruct_10hummingbot_4core_9data_type_10order_book_OrderBook*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_9data_type_10order_book_OrderBook)) __PYX_ERR(11, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.clock"); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_5clock_Clock = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.clock", "Clock", sizeof(struct __pyx_obj_10hummingbot_4core_5clock_Clock), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_5clock_Clock),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_5clock_Clock) __PYX_ERR(12, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.time_iterator"); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.time_iterator", "TimeIterator", sizeof(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator) __PYX_ERR(13, 7, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_13time_iterator_TimeIterator = (struct __pyx_vtabstruct_10hummingbot_4core_13time_iterator_TimeIterator*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_13time_iterator_TimeIterator)) __PYX_ERR(13, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.network_iterator"); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.network_iterator", "NetworkIterator", sizeof(struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator) __PYX_ERR(14, 6, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_16network_iterator_NetworkIterator = (struct __pyx_vtabstruct_10hummingbot_4core_16network_iterator_NetworkIterator*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_16network_iterator_NetworkIterator)) __PYX_ERR(14, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.connector.connector_base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.connector.connector_base", "ConnectorBase", sizeof(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase) __PYX_ERR(15, 5, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_9connector_14connector_base_ConnectorBase = (struct __pyx_vtabstruct_10hummingbot_9connector_14connector_base_ConnectorBase*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase); if (unlikely(!__pyx_vtabptr_10hummingbot_9connector_14connector_base_ConnectorBase)) __PYX_ERR(15, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.data_type.composite_order_book"); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.data_type.composite_order_book", "CompositeOrderBook", sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook) __PYX_ERR(16, 4, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook = (struct __pyx_vtabstruct_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_9data_type_20composite_order_book_CompositeOrderBook)) __PYX_ERR(16, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.data_type.limit_order"); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.data_type.limit_order", "LimitOrder", sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder) __PYX_ERR(17, 6, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_9data_type_11limit_order_LimitOrder = (struct __pyx_vtabstruct_10hummingbot_4core_9data_type_11limit_order_LimitOrder*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_9data_type_11limit_order_LimitOrder)) __PYX_ERR(17, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.data_type.limit_order"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportFunction_3_0_12(__pyx_t_1, "c_create_limit_order_from_cpp_limit_order", (void (**)(void))&__pyx_f_10hummingbot_4core_9data_type_11limit_order_c_create_limit_order_from_cpp_limit_order, "struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *(LimitOrder const )") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_paper_trade_exchange(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_paper_trade_exchange},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "paper_trade_exchange",
      0, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initpaper_trade_exchange(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initpaper_trade_exchange(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_paper_trade_exchange(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_paper_trade_exchange(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_paper_trade_exchange(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'paper_trade_exchange' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("paper_trade_exchange", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #elif CYTHON_USE_MODULE_STATE
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "paper_trade_exchange" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = __Pyx_PyImport_AddModuleRef((const char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_paper_trade_exchange(void)", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_hummingbot__connector__exchange__paper_trade__paper_trade_exchange) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "hummingbot.connector.exchange.paper_trade.paper_trade_exchange")) {
      if (unlikely((PyDict_SetItemString(modules, "hummingbot.connector.exchange.paper_trade.paper_trade_exchange", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely((__Pyx_modinit_type_init_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  if (unlikely((__Pyx_modinit_function_import_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":3
 * # distutils: sources=['hummingbot/core/cpp/Utils.cpp', 'hummingbot/core/cpp/LimitOrder.cpp', 'hummingbot/core/cpp/OrderExpirationEntry.cpp']
 * 
 * import asyncio             # <<<<<<<<<<<<<<
 * import math
 * import random
 */
  __pyx_t_2 = __Pyx_patch_asyncio(__Pyx_ImportDottedModule(__pyx_n_s_asyncio, NULL)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_asyncio, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":4
 * 
 * import asyncio
 * import math             # <<<<<<<<<<<<<<
 * import random
 * from collections import defaultdict, deque
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_math, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_math, __pyx_t_2) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":5
 * import asyncio
 * import math
 * import random             # <<<<<<<<<<<<<<
 * from collections import defaultdict, deque
 * from decimal import Decimal, ROUND_DOWN
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_random, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_random, __pyx_t_2) < 0) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":6
 * import math
 * import random
 * from collections import defaultdict, deque             # <<<<<<<<<<<<<<
 * from decimal import Decimal, ROUND_DOWN
 * from typing import Callable, Dict, List, Optional, Tuple, TYPE_CHECKING
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_defaultdict);
  __Pyx_GIVEREF(__pyx_n_s_defaultdict);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_defaultdict)) __PYX_ERR(0, 6, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_deque);
  __Pyx_GIVEREF(__pyx_n_s_deque);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_deque)) __PYX_ERR(0, 6, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_collections, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_defaultdict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_defaultdict, __pyx_t_2) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_deque); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_deque, __pyx_t_2) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":7
 * import random
 * from collections import defaultdict, deque
 * from decimal import Decimal, ROUND_DOWN             # <<<<<<<<<<<<<<
 * from typing import Callable, Dict, List, Optional, Tuple, TYPE_CHECKING
 * 
 */
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_Decimal);
  __Pyx_GIVEREF(__pyx_n_s_Decimal);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_Decimal)) __PYX_ERR(0, 7, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_ROUND_DOWN);
  __Pyx_GIVEREF(__pyx_n_s_ROUND_DOWN);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_ROUND_DOWN)) __PYX_ERR(0, 7, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_decimal, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Decimal, __pyx_t_3) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_ROUND_DOWN); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ROUND_DOWN, __pyx_t_3) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":8
 * from collections import defaultdict, deque
 * from decimal import Decimal, ROUND_DOWN
 * from typing import Callable, Dict, List, Optional, Tuple, TYPE_CHECKING             # <<<<<<<<<<<<<<
 * 
 * from cpython cimport PyObject
 */
  __pyx_t_2 = PyList_New(6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_Callable);
  __Pyx_GIVEREF(__pyx_n_s_Callable);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Callable)) __PYX_ERR(0, 8, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_Dict);
  __Pyx_GIVEREF(__pyx_n_s_Dict);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_Dict)) __PYX_ERR(0, 8, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_List);
  __Pyx_GIVEREF(__pyx_n_s_List);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_List)) __PYX_ERR(0, 8, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_Optional);
  __Pyx_GIVEREF(__pyx_n_s_Optional);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 3, __pyx_n_s_Optional)) __PYX_ERR(0, 8, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_Tuple);
  __Pyx_GIVEREF(__pyx_n_s_Tuple);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 4, __pyx_n_s_Tuple)) __PYX_ERR(0, 8, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_TYPE_CHECKING);
  __Pyx_GIVEREF(__pyx_n_s_TYPE_CHECKING);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 5, __pyx_n_s_TYPE_CHECKING)) __PYX_ERR(0, 8, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_typing, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Callable); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Callable, __pyx_t_2) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Dict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Dict, __pyx_t_2) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_List); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_List, __pyx_t_2) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Optional); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Optional, __pyx_t_2) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Tuple); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Tuple, __pyx_t_2) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_TYPE_CHECKING); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TYPE_CHECKING, __pyx_t_2) < 0) __PYX_ERR(0, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":15
 * from libcpp.vector cimport vector
 * 
 * from hummingbot.connector.budget_checker import BudgetChecker             # <<<<<<<<<<<<<<
 * from hummingbot.connector.connector_metrics_collector import DummyMetricsCollector
 * from hummingbot.connector.exchange.paper_trade.trading_pair import TradingPair
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_BudgetChecker);
  __Pyx_GIVEREF(__pyx_n_s_BudgetChecker);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_BudgetChecker)) __PYX_ERR(0, 15, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_hummingbot_connector_budget_chec, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_BudgetChecker); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BudgetChecker, __pyx_t_3) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":16
 * 
 * from hummingbot.connector.budget_checker import BudgetChecker
 * from hummingbot.connector.connector_metrics_collector import DummyMetricsCollector             # <<<<<<<<<<<<<<
 * from hummingbot.connector.exchange.paper_trade.trading_pair import TradingPair
 * from hummingbot.connector.exchange_base import ExchangeBase
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_DummyMetricsCollector);
  __Pyx_GIVEREF(__pyx_n_s_DummyMetricsCollector);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_DummyMetricsCollector)) __PYX_ERR(0, 16, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_hummingbot_connector_connector_m, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_DummyMetricsCollector); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DummyMetricsCollector, __pyx_t_2) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":17
 * from hummingbot.connector.budget_checker import BudgetChecker
 * from hummingbot.connector.connector_metrics_collector import DummyMetricsCollector
 * from hummingbot.connector.exchange.paper_trade.trading_pair import TradingPair             # <<<<<<<<<<<<<<
 * from hummingbot.connector.exchange_base import ExchangeBase
 * from hummingbot.core.clock cimport Clock
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_TradingPair);
  __Pyx_GIVEREF(__pyx_n_s_TradingPair);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_TradingPair)) __PYX_ERR(0, 17, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_hummingbot_connector_exchange_pa_2, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_TradingPair); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TradingPair, __pyx_t_3) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":18
 * from hummingbot.connector.connector_metrics_collector import DummyMetricsCollector
 * from hummingbot.connector.exchange.paper_trade.trading_pair import TradingPair
 * from hummingbot.connector.exchange_base import ExchangeBase             # <<<<<<<<<<<<<<
 * from hummingbot.core.clock cimport Clock
 * from hummingbot.core.clock import Clock
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_ExchangeBase);
  __Pyx_GIVEREF(__pyx_n_s_ExchangeBase);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_ExchangeBase)) __PYX_ERR(0, 18, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_hummingbot_connector_exchange_ba, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":20
 * from hummingbot.connector.exchange_base import ExchangeBase
 * from hummingbot.core.clock cimport Clock
 * from hummingbot.core.clock import Clock             # <<<<<<<<<<<<<<
 * from hummingbot.core.data_type.cancellation_result import CancellationResult
 * from hummingbot.core.data_type.common import OrderType, TradeType
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_Clock);
  __Pyx_GIVEREF(__pyx_n_s_Clock);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_Clock)) __PYX_ERR(0, 20, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_hummingbot_core_clock, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":21
 * from hummingbot.core.clock cimport Clock
 * from hummingbot.core.clock import Clock
 * from hummingbot.core.data_type.cancellation_result import CancellationResult             # <<<<<<<<<<<<<<
 * from hummingbot.core.data_type.common import OrderType, TradeType
 * from hummingbot.core.data_type.composite_order_book import CompositeOrderBook
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_CancellationResult);
  __Pyx_GIVEREF(__pyx_n_s_CancellationResult);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_CancellationResult)) __PYX_ERR(0, 21, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_hummingbot_core_data_type_cancel, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_CancellationResult); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CancellationResult, __pyx_t_2) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":22
 * from hummingbot.core.clock import Clock
 * from hummingbot.core.data_type.cancellation_result import CancellationResult
 * from hummingbot.core.data_type.common import OrderType, TradeType             # <<<<<<<<<<<<<<
 * from hummingbot.core.data_type.composite_order_book import CompositeOrderBook
 * from hummingbot.core.data_type.composite_order_book cimport CompositeOrderBook
 */
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_OrderType);
  __Pyx_GIVEREF(__pyx_n_s_OrderType);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_OrderType)) __PYX_ERR(0, 22, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_TradeType);
  __Pyx_GIVEREF(__pyx_n_s_TradeType);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_TradeType)) __PYX_ERR(0, 22, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_hummingbot_core_data_type_common, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OrderType, __pyx_t_3) < 0) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TradeType, __pyx_t_3) < 0) __PYX_ERR(0, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":23
 * from hummingbot.core.data_type.cancellation_result import CancellationResult
 * from hummingbot.core.data_type.common import OrderType, TradeType
 * from hummingbot.core.data_type.composite_order_book import CompositeOrderBook             # <<<<<<<<<<<<<<
 * from hummingbot.core.data_type.composite_order_book cimport CompositeOrderBook
 * from hummingbot.core.data_type.limit_order import LimitOrder
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_CompositeOrderBook);
  __Pyx_GIVEREF(__pyx_n_s_CompositeOrderBook);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_CompositeOrderBook)) __PYX_ERR(0, 23, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_hummingbot_core_data_type_compos, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":25
 * from hummingbot.core.data_type.composite_order_book import CompositeOrderBook
 * from hummingbot.core.data_type.composite_order_book cimport CompositeOrderBook
 * from hummingbot.core.data_type.limit_order import LimitOrder             # <<<<<<<<<<<<<<
 * from hummingbot.core.data_type.limit_order cimport c_create_limit_order_from_cpp_limit_order
 * from hummingbot.core.data_type.order_book cimport OrderBook
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_LimitOrder);
  __Pyx_GIVEREF(__pyx_n_s_LimitOrder);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_LimitOrder)) __PYX_ERR(0, 25, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_hummingbot_core_data_type_limit, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_LimitOrder); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_LimitOrder, __pyx_t_3) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":28
 * from hummingbot.core.data_type.limit_order cimport c_create_limit_order_from_cpp_limit_order
 * from hummingbot.core.data_type.order_book cimport OrderBook
 * from hummingbot.core.data_type.order_book_tracker import OrderBookTracker             # <<<<<<<<<<<<<<
 * from hummingbot.core.data_type.order_candidate import OrderCandidate
 * from hummingbot.core.event.event_listener cimport EventListener
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_OrderBookTracker);
  __Pyx_GIVEREF(__pyx_n_s_OrderBookTracker);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_OrderBookTracker)) __PYX_ERR(0, 28, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_hummingbot_core_data_type_order, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_OrderBookTracker); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OrderBookTracker, __pyx_t_2) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":29
 * from hummingbot.core.data_type.order_book cimport OrderBook
 * from hummingbot.core.data_type.order_book_tracker import OrderBookTracker
 * from hummingbot.core.data_type.order_candidate import OrderCandidate             # <<<<<<<<<<<<<<
 * from hummingbot.core.event.event_listener cimport EventListener
 * from hummingbot.core.event.events import (
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_OrderCandidate);
  __Pyx_GIVEREF(__pyx_n_s_OrderCandidate);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_OrderCandidate)) __PYX_ERR(0, 29, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_hummingbot_core_data_type_order_2, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_OrderCandidate); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OrderCandidate, __pyx_t_3) < 0) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":32
 * from hummingbot.core.event.event_listener cimport EventListener
 * from hummingbot.core.event.events import (
 *     BuyOrderCompletedEvent,             # <<<<<<<<<<<<<<
 *     BuyOrderCreatedEvent,
 *     MarketEvent,
 */
  __pyx_t_2 = PyList_New(10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_BuyOrderCompletedEvent);
  __Pyx_GIVEREF(__pyx_n_s_BuyOrderCompletedEvent);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_BuyOrderCompletedEvent)) __PYX_ERR(0, 32, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_BuyOrderCreatedEvent);
  __Pyx_GIVEREF(__pyx_n_s_BuyOrderCreatedEvent);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_BuyOrderCreatedEvent)) __PYX_ERR(0, 32, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_MarketEvent);
  __Pyx_GIVEREF(__pyx_n_s_MarketEvent);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_MarketEvent)) __PYX_ERR(0, 32, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_MarketOrderFailureEvent);
  __Pyx_GIVEREF(__pyx_n_s_MarketOrderFailureEvent);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 3, __pyx_n_s_MarketOrderFailureEvent)) __PYX_ERR(0, 32, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_OrderBookEvent);
  __Pyx_GIVEREF(__pyx_n_s_OrderBookEvent);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 4, __pyx_n_s_OrderBookEvent)) __PYX_ERR(0, 32, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_OrderBookTradeEvent);
  __Pyx_GIVEREF(__pyx_n_s_OrderBookTradeEvent);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 5, __pyx_n_s_OrderBookTradeEvent)) __PYX_ERR(0, 32, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_OrderCancelledEvent);
  __Pyx_GIVEREF(__pyx_n_s_OrderCancelledEvent);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 6, __pyx_n_s_OrderCancelledEvent)) __PYX_ERR(0, 32, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_OrderFilledEvent);
  __Pyx_GIVEREF(__pyx_n_s_OrderFilledEvent);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 7, __pyx_n_s_OrderFilledEvent)) __PYX_ERR(0, 32, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_SellOrderCompletedEvent);
  __Pyx_GIVEREF(__pyx_n_s_SellOrderCompletedEvent);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 8, __pyx_n_s_SellOrderCompletedEvent)) __PYX_ERR(0, 32, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_SellOrderCreatedEvent);
  __Pyx_GIVEREF(__pyx_n_s_SellOrderCreatedEvent);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 9, __pyx_n_s_SellOrderCreatedEvent)) __PYX_ERR(0, 32, __pyx_L1_error);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":31
 * from hummingbot.core.data_type.order_candidate import OrderCandidate
 * from hummingbot.core.event.event_listener cimport EventListener
 * from hummingbot.core.event.events import (             # <<<<<<<<<<<<<<
 *     BuyOrderCompletedEvent,
 *     BuyOrderCreatedEvent,
 */
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_hummingbot_core_event_events, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_BuyOrderCompletedEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BuyOrderCompletedEvent, __pyx_t_2) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_BuyOrderCreatedEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BuyOrderCreatedEvent, __pyx_t_2) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_MarketEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MarketEvent, __pyx_t_2) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_MarketOrderFailureEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MarketOrderFailureEvent, __pyx_t_2) < 0) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_OrderBookEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OrderBookEvent, __pyx_t_2) < 0) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_OrderBookTradeEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OrderBookTradeEvent, __pyx_t_2) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_OrderCancelledEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OrderCancelledEvent, __pyx_t_2) < 0) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_OrderFilledEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OrderFilledEvent, __pyx_t_2) < 0) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_SellOrderCompletedEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SellOrderCompletedEvent, __pyx_t_2) < 0) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_SellOrderCreatedEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SellOrderCreatedEvent, __pyx_t_2) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":43
 *     SellOrderCreatedEvent,
 * )
 * from hummingbot.core.network_iterator import NetworkStatus             # <<<<<<<<<<<<<<
 * from hummingbot.core.Utils cimport getIteratorFromReverseIterator, reverse_iterator
 * from hummingbot.core.utils.async_utils import safe_ensure_future
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_NetworkStatus);
  __Pyx_GIVEREF(__pyx_n_s_NetworkStatus);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_NetworkStatus)) __PYX_ERR(0, 43, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_hummingbot_core_network_iterator, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_NetworkStatus); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NetworkStatus, __pyx_t_3) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":45
 * from hummingbot.core.network_iterator import NetworkStatus
 * from hummingbot.core.Utils cimport getIteratorFromReverseIterator, reverse_iterator
 * from hummingbot.core.utils.async_utils import safe_ensure_future             # <<<<<<<<<<<<<<
 * from hummingbot.core.utils.estimate_fee import build_trade_fee
 * 
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_safe_ensure_future);
  __Pyx_GIVEREF(__pyx_n_s_safe_ensure_future);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_safe_ensure_future)) __PYX_ERR(0, 45, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_hummingbot_core_utils_async_util, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_safe_ensure_future); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_safe_ensure_future, __pyx_t_2) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":46
 * from hummingbot.core.Utils cimport getIteratorFromReverseIterator, reverse_iterator
 * from hummingbot.core.utils.async_utils import safe_ensure_future
 * from hummingbot.core.utils.estimate_fee import build_trade_fee             # <<<<<<<<<<<<<<
 * 
 * if TYPE_CHECKING:
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_build_trade_fee);
  __Pyx_GIVEREF(__pyx_n_s_build_trade_fee);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_build_trade_fee)) __PYX_ERR(0, 46, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_hummingbot_core_utils_estimate_f, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_build_trade_fee); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_build_trade_fee, __pyx_t_3) < 0) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":48
 * from hummingbot.core.utils.estimate_fee import build_trade_fee
 * 
 * if TYPE_CHECKING:             # <<<<<<<<<<<<<<
 *     from hummingbot.client.config.config_helpers import ClientConfigAdapter
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_TYPE_CHECKING); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":49
 * 
 * if TYPE_CHECKING:
 *     from hummingbot.client.config.config_helpers import ClientConfigAdapter             # <<<<<<<<<<<<<<
 * 
 * ptm_logger = None
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_ClientConfigAdapter);
    __Pyx_GIVEREF(__pyx_n_s_ClientConfigAdapter);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_ClientConfigAdapter)) __PYX_ERR(0, 49, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_hummingbot_client_config_config, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_ClientConfigAdapter); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_ClientConfigAdapter, __pyx_t_2) < 0) __PYX_ERR(0, 49, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":48
 * from hummingbot.core.utils.estimate_fee import build_trade_fee
 * 
 * if TYPE_CHECKING:             # <<<<<<<<<<<<<<
 *     from hummingbot.client.config.config_helpers import ClientConfigAdapter
 * 
 */
  }

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":51
 *     from hummingbot.client.config.config_helpers import ClientConfigAdapter
 * 
 * ptm_logger = None             # <<<<<<<<<<<<<<
 * s_decimal_0 = Decimal(0)
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ptm_logger, Py_None) < 0) __PYX_ERR(0, 51, __pyx_L1_error)

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":52
 * 
 * ptm_logger = None
 * s_decimal_0 = Decimal(0)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__37, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_s_decimal_0, __pyx_t_2) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_QuantizationParams___reduce_cyth, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__39)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_QuantizationParams, (type(self), 0xa74d68e, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_QuantizationParams__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18QuantizationParams_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_QuantizationParams___setstate_cy, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__41)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QuantizationParams);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_5__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_QueuedOrder___reduce_cython, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_QueuedOrder, (type(self), 0xefa59de, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_QueuedOrder__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_11QueuedOrder_7__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_QueuedOrder___setstate_cython, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__43)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_QueuedOrder);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_OrderBookTradeListener___reduce, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_OrderBookTradeListener, (type(self), 0x1bae46d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_OrderBookTradeListener__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_22OrderBookTradeListener_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_OrderBookTradeListener___setstat, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__45)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookTradeListener);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_OrderBookMarketOrderFillListener_2, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__46)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener);

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_OrderBookMarketOrderFillListener, (type(self), 0x1bae46d, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_OrderBookMarketOrderFillListener__set_state(self, __pyx_state)
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_32OrderBookMarketOrderFillListener_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_OrderBookMarketOrderFillListener_3, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__47)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_OrderBookMarketOrderFillListener);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":144
 * 
 * cdef class PaperTradeExchange(ExchangeBase):
 *     TRADE_EXECUTION_DELAY = 5.0             # <<<<<<<<<<<<<<
 *     ORDER_FILLED_EVENT_TAG = MarketEvent.OrderFilled.value
 *     SELL_ORDER_COMPLETED_EVENT_TAG = MarketEvent.SellOrderCompleted.value
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_TRADE_EXECUTION_DELAY, __pyx_float_5_0) < 0) __PYX_ERR(0, 144, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":145
 * cdef class PaperTradeExchange(ExchangeBase):
 *     TRADE_EXECUTION_DELAY = 5.0
 *     ORDER_FILLED_EVENT_TAG = MarketEvent.OrderFilled.value             # <<<<<<<<<<<<<<
 *     SELL_ORDER_COMPLETED_EVENT_TAG = MarketEvent.SellOrderCompleted.value
 *     BUY_ORDER_COMPLETED_EVENT_TAG = MarketEvent.BuyOrderCompleted.value
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MarketEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_OrderFilled); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_ORDER_FILLED_EVENT_TAG, __pyx_t_2) < 0) __PYX_ERR(0, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":146
 *     TRADE_EXECUTION_DELAY = 5.0
 *     ORDER_FILLED_EVENT_TAG = MarketEvent.OrderFilled.value
 *     SELL_ORDER_COMPLETED_EVENT_TAG = MarketEvent.SellOrderCompleted.value             # <<<<<<<<<<<<<<
 *     BUY_ORDER_COMPLETED_EVENT_TAG = MarketEvent.BuyOrderCompleted.value
 *     MARKET_ORDER_CANCELED_EVENT_TAG = MarketEvent.OrderCancelled.value
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MarketEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_SellOrderCompleted); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_SELL_ORDER_COMPLETED_EVENT_TAG, __pyx_t_2) < 0) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":147
 *     ORDER_FILLED_EVENT_TAG = MarketEvent.OrderFilled.value
 *     SELL_ORDER_COMPLETED_EVENT_TAG = MarketEvent.SellOrderCompleted.value
 *     BUY_ORDER_COMPLETED_EVENT_TAG = MarketEvent.BuyOrderCompleted.value             # <<<<<<<<<<<<<<
 *     MARKET_ORDER_CANCELED_EVENT_TAG = MarketEvent.OrderCancelled.value
 *     MARKET_ORDER_FAILURE_EVENT_TAG = MarketEvent.OrderFailure.value
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MarketEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BuyOrderCompleted); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_BUY_ORDER_COMPLETED_EVENT_TAG, __pyx_t_2) < 0) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":148
 *     SELL_ORDER_COMPLETED_EVENT_TAG = MarketEvent.SellOrderCompleted.value
 *     BUY_ORDER_COMPLETED_EVENT_TAG = MarketEvent.BuyOrderCompleted.value
 *     MARKET_ORDER_CANCELED_EVENT_TAG = MarketEvent.OrderCancelled.value             # <<<<<<<<<<<<<<
 *     MARKET_ORDER_FAILURE_EVENT_TAG = MarketEvent.OrderFailure.value
 *     ORDER_BOOK_TRADE_EVENT_TAG = OrderBookEvent.TradeEvent.value
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MarketEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_OrderCancelled); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_MARKET_ORDER_CANCELED_EVENT_TAG, __pyx_t_2) < 0) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":149
 *     BUY_ORDER_COMPLETED_EVENT_TAG = MarketEvent.BuyOrderCompleted.value
 *     MARKET_ORDER_CANCELED_EVENT_TAG = MarketEvent.OrderCancelled.value
 *     MARKET_ORDER_FAILURE_EVENT_TAG = MarketEvent.OrderFailure.value             # <<<<<<<<<<<<<<
 *     ORDER_BOOK_TRADE_EVENT_TAG = OrderBookEvent.TradeEvent.value
 *     MARKET_SELL_ORDER_CREATED_EVENT_TAG = MarketEvent.SellOrderCreated.value
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MarketEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_OrderFailure); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_MARKET_ORDER_FAILURE_EVENT_TAG, __pyx_t_2) < 0) __PYX_ERR(0, 149, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":150
 *     MARKET_ORDER_CANCELED_EVENT_TAG = MarketEvent.OrderCancelled.value
 *     MARKET_ORDER_FAILURE_EVENT_TAG = MarketEvent.OrderFailure.value
 *     ORDER_BOOK_TRADE_EVENT_TAG = OrderBookEvent.TradeEvent.value             # <<<<<<<<<<<<<<
 *     MARKET_SELL_ORDER_CREATED_EVENT_TAG = MarketEvent.SellOrderCreated.value
 *     MARKET_BUY_ORDER_CREATED_EVENT_TAG = MarketEvent.BuyOrderCreated.value
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_OrderBookEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_TradeEvent); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_ORDER_BOOK_TRADE_EVENT_TAG, __pyx_t_2) < 0) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":151
 *     MARKET_ORDER_FAILURE_EVENT_TAG = MarketEvent.OrderFailure.value
 *     ORDER_BOOK_TRADE_EVENT_TAG = OrderBookEvent.TradeEvent.value
 *     MARKET_SELL_ORDER_CREATED_EVENT_TAG = MarketEvent.SellOrderCreated.value             # <<<<<<<<<<<<<<
 *     MARKET_BUY_ORDER_CREATED_EVENT_TAG = MarketEvent.BuyOrderCreated.value
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MarketEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_SellOrderCreated); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_MARKET_SELL_ORDER_CREATED_EVENT, __pyx_t_2) < 0) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":152
 *     ORDER_BOOK_TRADE_EVENT_TAG = OrderBookEvent.TradeEvent.value
 *     MARKET_SELL_ORDER_CREATED_EVENT_TAG = MarketEvent.SellOrderCreated.value
 *     MARKET_BUY_ORDER_CREATED_EVENT_TAG = MarketEvent.BuyOrderCreated.value             # <<<<<<<<<<<<<<
 * 
 *     def __init__(
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_MarketEvent); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_BuyOrderCreated); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_MARKET_BUY_ORDER_CREATED_EVENT_T, __pyx_t_2) < 0) __PYX_ERR(0, 152, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":184
 *         return self._budget_checker
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def random_order_id(cls, order_side: str, trading_pair: str) -> str:
 *         vals = [random.choice(range(0, 256)) for i in range(0, 13)]
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_order_side, __pyx_n_s_str) < 0) __PYX_ERR(0, 184, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_trading_pair, __pyx_n_s_str) < 0) __PYX_ERR(0, 184, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_str) < 0) __PYX_ERR(0, 184, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_3random_order_id, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaperTradeExchange_random_order, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__49)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_random_order_id, __pyx_t_3) < 0) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);
  __Pyx_GetNameInClass(__pyx_t_3, (PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_random_order_id); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Method_ClassMethod(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_random_order_id, __pyx_t_2) < 0) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":189
 *         return f"{order_side}://" + trading_pair + "/" + "".join([f"{val:02x}" for val in vals])
 * 
 *     def init_paper_trade_market(self):             # <<<<<<<<<<<<<<
 *         for trading_pair_str, order_book in self.order_book_tracker.order_books.items():
 *             assert type(order_book) is CompositeOrderBook
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_5init_paper_trade_market, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaperTradeExchange_init_paper_tr, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__51)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_init_paper_trade_market, __pyx_t_2) < 0) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":199
 *             )
 * 
 *     def split_trading_pair(self, trading_pair: str) -> Tuple[str, str]:             # <<<<<<<<<<<<<<
 *         return self._target_market.split_trading_pair(trading_pair)
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_trading_pair, __pyx_n_s_str) < 0) __PYX_ERR(0, 199, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_kp_s_Tuple_str_str) < 0) __PYX_ERR(0, 199, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_7split_trading_pair, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaperTradeExchange_split_trading, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__53)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_split_trading_pair, __pyx_t_3) < 0) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":299
 *         ExchangeBase.c_start(self, clock, timestamp)
 * 
 *     async def start_network(self):             # <<<<<<<<<<<<<<
 *         await self.stop_network()
 *         self.order_book_tracker.start()
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_9start_network, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_PaperTradeExchange_start_network, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__12)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_start_network, __pyx_t_3) < 0) __PYX_ERR(0, 299, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":303
 *         self.order_book_tracker.start()
 * 
 *     async def stop_network(self):             # <<<<<<<<<<<<<<
 *         self.order_book_tracker.stop()
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_12stop_network, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_PaperTradeExchange_stop_network, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__13)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_stop_network, __pyx_t_3) < 0) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":306
 *         self.order_book_tracker.stop()
 * 
 *     async def check_network(self) -> NetworkStatus:             # <<<<<<<<<<<<<<
 *         return NetworkStatus.CONNECTED
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_NetworkStatus) < 0) __PYX_ERR(0, 306, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_15check_network, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_PaperTradeExchange_check_network, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__14)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_check_network, __pyx_t_2) < 0) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":326
 *                    str trading_pair_str,
 *                    object amount,
 *                    object order_type=OrderType.MARKET,             # <<<<<<<<<<<<<<
 *                    object price=s_decimal_0,
 *                    dict kwargs={}):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MARKET); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 326, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_k__16 = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":327
 *                    object amount,
 *                    object order_type=OrderType.MARKET,
 *                    object price=s_decimal_0,             # <<<<<<<<<<<<<<
 *                    dict kwargs={}):
 *         if trading_pair_str not in self._trading_pairs:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_s_decimal_0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_k__17 = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":328
 *                    object order_type=OrderType.MARKET,
 *                    object price=s_decimal_0,
 *                    dict kwargs={}):             # <<<<<<<<<<<<<<
 *         if trading_pair_str not in self._trading_pairs:
 *             raise ValueError(f"Trading pair '{trading_pair_str}' does not existing in current data set.")
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_k__55 = ((PyObject*)__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":387
 *                     str trading_pair_str,
 *                     object amount,
 *                     object order_type=OrderType.MARKET,             # <<<<<<<<<<<<<<
 *                     object price=s_decimal_0,
 *                     dict kwargs={}):
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MARKET); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 387, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_k__18 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":388
 *                     object amount,
 *                     object order_type=OrderType.MARKET,
 *                     object price=s_decimal_0,             # <<<<<<<<<<<<<<
 *                     dict kwargs={}):
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_s_decimal_0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 388, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__19 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":389
 *                     object order_type=OrderType.MARKET,
 *                     object price=s_decimal_0,
 *                     dict kwargs={}):             # <<<<<<<<<<<<<<
 * 
 *         if trading_pair_str not in self._trading_pairs:
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 389, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__56 = ((PyObject*)__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":945
 *         return self.available_balances.get(currency.upper(), s_decimal_0)
 * 
 *     async def cancel_all(self, timeout_seconds: float) -> List[CancellationResult]:             # <<<<<<<<<<<<<<
 *         cdef:
 *             LimitOrders *limit_orders_map_ptr
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_timeout_seconds, __pyx_n_s_float) < 0) __PYX_ERR(0, 945, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_kp_s_List_CancellationResult) < 0) __PYX_ERR(0, 945, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_18cancel_all, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_PaperTradeExchange_cancel_all, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__24)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_cancel_all, __pyx_t_3) < 0) __PYX_ERR(0, 945, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1051
 *             return Decimal(f"1e-10")
 * 
 *     def get_order_price_quantum(self, trading_pair: str, price: Decimal) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_get_order_price_quantum(trading_pair, price)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_trading_pair, __pyx_n_s_str) < 0) __PYX_ERR(0, 1051, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_price, __pyx_n_s_Decimal) < 0) __PYX_ERR(0, 1051, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_Decimal) < 0) __PYX_ERR(0, 1051, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_21get_order_price_quantum, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaperTradeExchange_get_order_pri, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__59)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1051, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_get_order_price_quantum, __pyx_t_2) < 0) __PYX_ERR(0, 1051, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1080
 *                                         str trading_pair,
 *                                         object amount,
 *                                         object price=s_decimal_0):             # <<<<<<<<<<<<<<
 *         amount = amount.quantize(Decimal('1e-7'), rounding=ROUND_DOWN)
 *         if amount <= 1e-7:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_s_decimal_0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1080, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__60 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1087
 *         return (amount // order_size_quantum) * order_size_quantum
 * 
 *     def get_available_balance(self, currency: str) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_get_available_balance(currency)
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_currency, __pyx_n_s_str) < 0) __PYX_ERR(0, 1087, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_Decimal) < 0) __PYX_ERR(0, 1087, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_23get_available_balance, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaperTradeExchange_get_available, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__62)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_get_available_balance, __pyx_t_3) < 0) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1090
 *         return self.c_get_available_balance(currency)
 * 
 *     def get_all_balances(self) -> Dict[str, Decimal]:             # <<<<<<<<<<<<<<
 *         return self._account_balances.copy()
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_kp_s_Dict_str_Decimal) < 0) __PYX_ERR(0, 1090, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_25get_all_balances, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaperTradeExchange_get_all_balan, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__63)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_get_all_balances, __pyx_t_2) < 0) __PYX_ERR(0, 1090, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1094
 * 
 *     # <editor-fold desc="Python wrapper for cdef functions">
 *     def match_trade_to_limit_orders(self, event_object: OrderBookTradeEvent):             # <<<<<<<<<<<<<<
 *         self.c_match_trade_to_limit_orders(event_object)
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_event_object, __pyx_n_s_OrderBookTradeEvent) < 0) __PYX_ERR(0, 1094, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_27match_trade_to_limit_orders, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaperTradeExchange_match_trade_t, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__65)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_match_trade_to_limit_orders, __pyx_t_3) < 0) __PYX_ERR(0, 1094, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1097
 *         self.c_match_trade_to_limit_orders(event_object)
 * 
 *     def set_balance(self, currency: str, balance: Decimal):             # <<<<<<<<<<<<<<
 *         self.c_set_balance(currency, balance)
 *     # </editor-fold>
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1097, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_currency, __pyx_n_s_str) < 0) __PYX_ERR(0, 1097, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_balance, __pyx_n_s_Decimal) < 0) __PYX_ERR(0, 1097, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_29set_balance, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaperTradeExchange_set_balance, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__67)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1097, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_set_balance, __pyx_t_2) < 0) __PYX_ERR(0, 1097, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1101
 *     # </editor-fold>
 * 
 *     def get_price(self, trading_pair: str, is_buy: bool) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_get_price(trading_pair, is_buy)
 * 
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_trading_pair, __pyx_n_s_str) < 0) __PYX_ERR(0, 1101, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_is_buy, __pyx_n_s_bool) < 0) __PYX_ERR(0, 1101, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_Decimal) < 0) __PYX_ERR(0, 1101, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_31get_price, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaperTradeExchange_get_price, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__69)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_get_price, __pyx_t_3) < 0) __PYX_ERR(0, 1101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1104
 *         return self.c_get_price(trading_pair, is_buy)
 * 
 *     def buy(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,             # <<<<<<<<<<<<<<
 *             price: Decimal = s_decimal_0, **kwargs) -> str:
 *         return self.c_buy(trading_pair, amount, order_type, price, kwargs)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MARKET); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_k__27 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1105
 * 
 *     def buy(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,
 *             price: Decimal = s_decimal_0, **kwargs) -> str:             # <<<<<<<<<<<<<<
 *         return self.c_buy(trading_pair, amount, order_type, price, kwargs)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_s_decimal_0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__28 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1104
 *         return self.c_get_price(trading_pair, is_buy)
 * 
 *     def buy(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,             # <<<<<<<<<<<<<<
 *             price: Decimal = s_decimal_0, **kwargs) -> str:
 *         return self.c_buy(trading_pair, amount, order_type, price, kwargs)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MARKET); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1105
 * 
 *     def buy(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,
 *             price: Decimal = s_decimal_0, **kwargs) -> str:             # <<<<<<<<<<<<<<
 *         return self.c_buy(trading_pair, amount, order_type, price, kwargs)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_s_decimal_0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1104
 *         return self.c_get_price(trading_pair, is_buy)
 * 
 *     def buy(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,             # <<<<<<<<<<<<<<
 *             price: Decimal = s_decimal_0, **kwargs) -> str:
 *         return self.c_buy(trading_pair, amount, order_type, price, kwargs)
 */
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(0, 1104, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(0, 1104, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_trading_pair, __pyx_n_s_str) < 0) __PYX_ERR(0, 1104, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_amount, __pyx_n_s_Decimal) < 0) __PYX_ERR(0, 1104, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_price, __pyx_n_s_Decimal) < 0) __PYX_ERR(0, 1104, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_str) < 0) __PYX_ERR(0, 1104, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_33buy, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaperTradeExchange_buy, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__71)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_buy, __pyx_t_3) < 0) __PYX_ERR(0, 1104, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1108
 *         return self.c_buy(trading_pair, amount, order_type, price, kwargs)
 * 
 *     def sell(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,             # <<<<<<<<<<<<<<
 *              price: Decimal = s_decimal_0, **kwargs) -> str:
 *         return self.c_sell(trading_pair, amount, order_type, price, kwargs)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MARKET); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_k__29 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1109
 * 
 *     def sell(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,
 *              price: Decimal = s_decimal_0, **kwargs) -> str:             # <<<<<<<<<<<<<<
 *         return self.c_sell(trading_pair, amount, order_type, price, kwargs)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_s_decimal_0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__30 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1108
 *         return self.c_buy(trading_pair, amount, order_type, price, kwargs)
 * 
 *     def sell(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,             # <<<<<<<<<<<<<<
 *              price: Decimal = s_decimal_0, **kwargs) -> str:
 *         return self.c_sell(trading_pair, amount, order_type, price, kwargs)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_MARKET); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1109
 * 
 *     def sell(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,
 *              price: Decimal = s_decimal_0, **kwargs) -> str:             # <<<<<<<<<<<<<<
 *         return self.c_sell(trading_pair, amount, order_type, price, kwargs)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_s_decimal_0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1108
 *         return self.c_buy(trading_pair, amount, order_type, price, kwargs)
 * 
 *     def sell(self, trading_pair: str, amount: Decimal, order_type=OrderType.MARKET,             # <<<<<<<<<<<<<<
 *              price: Decimal = s_decimal_0, **kwargs) -> str:
 *         return self.c_sell(trading_pair, amount, order_type, price, kwargs)
 */
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(0, 1108, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2)) __PYX_ERR(0, 1108, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyDict_NewPresized(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_trading_pair, __pyx_n_s_str) < 0) __PYX_ERR(0, 1108, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_amount, __pyx_n_s_Decimal) < 0) __PYX_ERR(0, 1108, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_price, __pyx_n_s_Decimal) < 0) __PYX_ERR(0, 1108, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_return, __pyx_n_s_str) < 0) __PYX_ERR(0, 1108, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_35sell, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaperTradeExchange_sell, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__72)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_3, __pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_sell, __pyx_t_3) < 0) __PYX_ERR(0, 1108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1112
 *         return self.c_sell(trading_pair, amount, order_type, price, kwargs)
 * 
 *     def cancel(self, trading_pair: str, client_order_id: str):             # <<<<<<<<<<<<<<
 *         return self.c_cancel(trading_pair, client_order_id)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_trading_pair, __pyx_n_s_str) < 0) __PYX_ERR(0, 1112, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_client_order_id, __pyx_n_s_str) < 0) __PYX_ERR(0, 1112, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_37cancel, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaperTradeExchange_cancel, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__74)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_cancel, __pyx_t_2) < 0) __PYX_ERR(0, 1112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1121
 *                 order_side: TradeType,
 *                 amount: Decimal,
 *                 price: Decimal = s_decimal_0,             # <<<<<<<<<<<<<<
 *                 is_maker: Optional[bool] = None):
 *         return self.c_get_fee(base_currency, quote_currency, order_type, order_side, amount, price, is_maker)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_s_decimal_0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_k__31 = __pyx_t_2;
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1115
 *         return self.c_cancel(trading_pair, client_order_id)
 * 
 *     def get_fee(self,             # <<<<<<<<<<<<<<
 *                 base_currency: str,
 *                 quote_currency: str,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_s_decimal_0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1122
 *                 amount: Decimal,
 *                 price: Decimal = s_decimal_0,
 *                 is_maker: Optional[bool] = None):             # <<<<<<<<<<<<<<
 *         return self.c_get_fee(base_currency, quote_currency, order_type, order_side, amount, price, is_maker)
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_2)) __PYX_ERR(0, 1115, __pyx_L1_error);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, Py_None)) __PYX_ERR(0, 1115, __pyx_L1_error);
  __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1115
 *         return self.c_cancel(trading_pair, client_order_id)
 * 
 *     def get_fee(self,             # <<<<<<<<<<<<<<
 *                 base_currency: str,
 *                 quote_currency: str,
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_base_currency, __pyx_n_s_str) < 0) __PYX_ERR(0, 1115, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_quote_currency, __pyx_n_s_str) < 0) __PYX_ERR(0, 1115, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_order_type, __pyx_n_s_OrderType) < 0) __PYX_ERR(0, 1115, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_order_side, __pyx_n_s_TradeType) < 0) __PYX_ERR(0, 1115, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_amount, __pyx_n_s_Decimal) < 0) __PYX_ERR(0, 1115, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_price, __pyx_n_s_Decimal) < 0) __PYX_ERR(0, 1115, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_is_maker, __pyx_kp_s_Optional_bool) < 0) __PYX_ERR(0, 1115, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_39get_fee, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaperTradeExchange_get_fee, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__76)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_get_fee, __pyx_t_5) < 0) __PYX_ERR(0, 1115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1125
 *         return self.c_get_fee(base_currency, quote_currency, order_type, order_side, amount, price, is_maker)
 * 
 *     def get_order_book(self, trading_pair: str) -> OrderBook:             # <<<<<<<<<<<<<<
 *         return self.c_get_order_book(trading_pair)
 * 
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_trading_pair, __pyx_n_s_str) < 0) __PYX_ERR(0, 1125, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_OrderBook) < 0) __PYX_ERR(0, 1125, __pyx_L1_error)
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_41get_order_book, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaperTradeExchange_get_order_boo, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__77)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_2, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_get_order_book, __pyx_t_2) < 0) __PYX_ERR(0, 1125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1128
 *         return self.c_get_order_book(trading_pair)
 * 
 *     def get_maker_order_type(self):             # <<<<<<<<<<<<<<
 *         return OrderType.LIMIT
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_43get_maker_order_type, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaperTradeExchange_get_maker_ord, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__78)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_get_maker_order_type, __pyx_t_2) < 0) __PYX_ERR(0, 1128, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1131
 *         return OrderType.LIMIT
 * 
 *     def get_taker_order_type(self):             # <<<<<<<<<<<<<<
 *         return OrderType.LIMIT
 * 
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_45get_taker_order_type, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaperTradeExchange_get_taker_ord, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__79)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_get_taker_order_type, __pyx_t_2) < 0) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1134
 *         return OrderType.LIMIT
 * 
 *     async def trigger_event_async(self,             # <<<<<<<<<<<<<<
 *                                   event_tag,
 *                                   event):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_47trigger_event_async, __Pyx_CYFUNCTION_CCLASS | __Pyx_CYFUNCTION_COROUTINE, __pyx_n_s_PaperTradeExchange_trigger_event, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__32)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange, __pyx_n_s_trigger_event_async, __pyx_t_2) < 0) __PYX_ERR(0, 1134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_PaperTradeExchange);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ask_limit_orders,self._bid_limit_orders,self._events,self._limit_order_expiration_set cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_50__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaperTradeExchange___reduce_cyth, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__81)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._ask_limit_orders,self._bid_limit_orders,self._events,self._limit_order_expiration_set cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._ask_limit_orders,self._bid_limit_orders,self._events,self._limit_order_expiration_set cannot be converted to a Python object for pickling"
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_18PaperTradeExchange_52__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_PaperTradeExchange___setstate_cy, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__82)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_QuantizationParams(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_1__pyx_unpickle_QuantizationParams, 0, __pyx_n_s_pyx_unpickle_QuantizationParam, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__84)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_QuantizationParam, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_QuantizationParams__set_state(<QuantizationParams> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_QuantizationParams__set_state(QuantizationParams __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.order_size_decimals = __pyx_state[0]; __pyx_result.order_size_precision = __pyx_state[1]; __pyx_result.price_decimals = __pyx_state[2]; __pyx_result.price_precision = __pyx_state[3]; __pyx_result.trading_pair = __pyx_state[4]
 *     if len(__pyx_state) > 5 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_3__pyx_unpickle_QueuedOrder, 0, __pyx_n_s_pyx_unpickle_QueuedOrder, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__85)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_QueuedOrder, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_OrderBookTradeListener(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_5__pyx_unpickle_OrderBookTradeListener, 0, __pyx_n_s_pyx_unpickle_OrderBookTradeLis, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__86)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_OrderBookTradeLis, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_OrderBookTradeListener__set_state(<OrderBookTradeListener> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_OrderBookTradeListener__set_state(OrderBookTradeListener __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._current_event_caller = __pyx_state[0]; __pyx_result._current_event_tag = __pyx_state[1]; __pyx_result._market = __pyx_state[2]
 *     if len(__pyx_state) > 3 and hasattr(__pyx_result, '__dict__'):
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_9connector_8exchange_11paper_trade_20paper_trade_exchange_7__pyx_unpickle_OrderBookMarketOrderFillListener, 0, __pyx_n_s_pyx_unpickle_OrderBookMarketOr, NULL, __pyx_n_s_hummingbot_connector_exchange_pa, __pyx_d, ((PyObject *)__pyx_codeobj__87)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_OrderBookMarketOr, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/connector/exchange/paper_trade/paper_trade_exchange.pyx":1
 * # distutils: sources=['hummingbot/core/cpp/Utils.cpp', 'hummingbot/core/cpp/LimitOrder.cpp', 'hummingbot/core/cpp/OrderExpirationEntry.cpp']             # <<<<<<<<<<<<<<
 * 
 * import asyncio
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init hummingbot.connector.exchange.paper_trade.paper_trade_exchange", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init hummingbot.connector.exchange.paper_trade.paper_trade_exchange");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C00A6
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE && PY_VERSION_HEX >= 0x030C00A6
    if (unlikely(tstate->current_exception))
#elif CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
      #if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
      #elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs = PyTuple_GET_SIZE(kwnames);
    PyObject *dict;
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        Py_XDECREF(key); key = NULL;
        Py_XDECREF(value); value = NULL;
        if (kwds_is_tuple) {
            Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(kwds);
#else
            size = PyTuple_Size(kwds);
            if (size < 0) goto bad;
#endif
            if (pos >= size) break;
#if CYTHON_AVOID_BORROWED_REFS
            key = __Pyx_PySequence_ITEM(kwds, pos);
            if (!key) goto bad;
#elif CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kwds, pos);
#else
            key = PyTuple_GetItem(kwds, pos);
            if (!key) goto bad;
#endif
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
#endif
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(value);
            Py_DECREF(key);
#endif
            key = NULL;
            value = NULL;
            continue;
        }
#if !CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(key);
#endif
        Py_INCREF(value);
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    Py_XDECREF(key);
    Py_XDECREF(value);
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    Py_XDECREF(key);
    Py_XDECREF(value);
    return -1;
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    type_name = __Pyx_PyType_GetName(type);
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")", name, type_name, obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* PyUnicode_Unicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj) {
    if (unlikely(obj == Py_None))
        obj = __pyx_kp_u_None;
    return __Pyx_NewRef(obj);
}

/* CIntToDigits */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
#if CYTHON_PEP393_ENABLED
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
#else
    Py_UNICODE *udata;
    uval = PyUnicode_FromUnicode(NULL, ulength);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_AS_UNICODE(uval);
#endif
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp;
                    PyObject *repeat = PyInt_FromSsize_t(uoffset - prepend_sign);
                    if (unlikely(!repeat)) goto done_or_error;
                    tmp = PyNumber_Multiply(padding, repeat);
                    Py_DECREF(repeat);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyNumber_Add(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyNumber_Add(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(int)*3+2];
    char *dpos, *end = digits + sizeof(int)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    int remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (int) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (int) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (int) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    CYTHON_MAYBE_UNUSED_VAR(max_char);
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    kind_shift = (result_ukind == 4) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    CYTHON_UNUSED_VAR(value_count);
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kw,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kw, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
        Py_ssize_t kwsize;
#if CYTHON_ASSUME_SAFE_MACROS
        kwsize = PyTuple_GET_SIZE(kw);
#else
        kwsize = PyTuple_Size(kw);
        if (kwsize < 0) return 0;
#endif
        if (unlikely(kwsize == 0))
            return 1;
        if (!kw_allowed) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, 0);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            goto invalid_keyword;
        }
#if PY_VERSION_HEX < 0x03090000
        for (pos = 0; pos < kwsize; pos++) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, pos);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
        }
#endif
        return 1;
    }
    while (PyDict_Next(kw, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if (!kw_allowed && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* GetAttr3 */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    int res = PyObject_GetOptionalAttr(o, n, &r);
    return (res != 0) ? r : __Pyx_NewRef(d);
#else
  #if CYTHON_USE_TYPE_SLOTS
    if (likely(PyString_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
  #endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
#endif
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && PY_VERSION_HEX < 0x030d0000
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object"))) {
        return NULL;
    }
    #endif
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) < 0) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
        #if PY_VERSION_HEX < 0x03090000
        vectorcallfunc f = _PyVectorcall_Function(func);
        #else
        vectorcallfunc f = PyVectorcall_Function(func);
        #endif
        if (f) {
            return f(func, args, (size_t)nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, (size_t)nargs, NULL);
        }
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_n_s_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* FixUpExtensionType */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef((char*) __PYX_ABI_MODULE_NAME);
}

/* FetchCommonType */
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#elif PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyObject *py_name = NULL;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
    if (!py_name) return NULL;
    PyErr_Format(PyExc_TypeError, "%.200S() takes no keyword arguments",
                 py_name);
    Py_DECREF(py_name);
#else
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
#endif
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_MACROS
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(!argc) < 0) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* UnicodeConcatInPlace */
# if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
static int
__Pyx_unicode_modifiable(PyObject *unicode)
{
    if (Py_REFCNT(unicode) != 1)
        return 0;
    if (!PyUnicode_CheckExact(unicode))
        return 0;
    if (PyUnicode_CHECK_INTERNED(unicode))
        return 0;
    return 1;
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    ) {
    PyObject *left = *p_left;
    Py_ssize_t left_len, right_len, new_len;
    if (unlikely(__Pyx_PyUnicode_READY(left) == -1))
        return NULL;
    if (unlikely(__Pyx_PyUnicode_READY(right) == -1))
        return NULL;
    left_len = PyUnicode_GET_LENGTH(left);
    if (left_len == 0) {
        Py_INCREF(right);
        return right;
    }
    right_len = PyUnicode_GET_LENGTH(right);
    if (right_len == 0) {
        Py_INCREF(left);
        return left;
    }
    if (unlikely(left_len > PY_SSIZE_T_MAX - right_len)) {
        PyErr_SetString(PyExc_OverflowError,
                        "strings are too large to concat");
        return NULL;
    }
    new_len = left_len + right_len;
    if (__Pyx_unicode_modifiable(left)
            && PyUnicode_CheckExact(right)
            && PyUnicode_KIND(right) <= PyUnicode_KIND(left)
            && !(PyUnicode_IS_ASCII(left) && !PyUnicode_IS_ASCII(right))) {
        int ret;
        __Pyx_GIVEREF(*p_left);
        ret = PyUnicode_Resize(p_left, new_len);
        __Pyx_GOTREF(*p_left);
        if (unlikely(ret != 0))
            return NULL;
        #if PY_VERSION_HEX >= 0x030d0000
        if (unlikely(PyUnicode_CopyCharacters(*p_left, left_len, right, 0, right_len) < 0)) return NULL;
        #else
        _PyUnicode_FastCopyCharacters(*p_left, left_len, right, 0, right_len);
        #endif
        __Pyx_INCREF(*p_left);
        __Pyx_GIVEREF(*p_left);
        return *p_left;
    } else {
        return __Pyx_PyUnicode_Concat(left, right);
    }
  }
#endif

/* PyObjectFormat */
#if CYTHON_USE_UNICODE_WRITER
static PyObject* __Pyx_PyObject_Format(PyObject* obj, PyObject* format_spec) {
    int ret;
    _PyUnicodeWriter writer;
    if (likely(PyFloat_CheckExact(obj))) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x03040000
        _PyUnicodeWriter_Init(&writer, 0);
#else
        _PyUnicodeWriter_Init(&writer);
#endif
        ret = _PyFloat_FormatAdvancedWriter(
            &writer,
            obj,
            format_spec, 0, PyUnicode_GET_LENGTH(format_spec));
    } else if (likely(PyLong_CheckExact(obj))) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x03040000
        _PyUnicodeWriter_Init(&writer, 0);
#else
        _PyUnicodeWriter_Init(&writer);
#endif
        ret = _PyLong_FormatAdvancedWriter(
            &writer,
            obj,
            format_spec, 0, PyUnicode_GET_LENGTH(format_spec));
    } else {
        return PyObject_Format(obj, format_spec);
    }
    if (unlikely(ret == -1)) {
        _PyUnicodeWriter_Dealloc(&writer);
        return NULL;
    }
    return _PyUnicodeWriter_Finish(&writer);
}
#endif

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
#if CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
#include <string.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallMethod1 */
#if !(CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2)
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_GetMethod;
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* CoroutineBase */
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6 && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    CYTHON_UNUSED_VAR(__pyx_tstate);
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (likely(__Pyx_IS_TYPE(ev, (PyTypeObject*)PyExc_StopIteration))) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_CLEAR(exc_state->exc_value);
#else
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
#endif
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(__pyx_CoroutineObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(PyObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(PyObject *gen, PyObject *value, int closing) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    CYTHON_MAYBE_UNUSED_VAR(closing);
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_value) {
        #if CYTHON_COMPILING_IN_PYPY
        #else
        PyObject *exc_tb;
        #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
        exc_tb = PyException_GetTraceback(exc_state->exc_value);
        #elif PY_VERSION_HEX >= 0x030B00a4
        exc_tb = ((PyBaseExceptionObject*) exc_state->exc_value)->traceback;
        #else
        exc_tb = exc_state->exc_traceback;
        #endif
        if (exc_tb) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
            #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
            Py_DECREF(exc_tb);
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body(self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
#if CYTHON_COMPILING_IN_PYPY
    CYTHON_UNUSED_VAR(exc_state);
#else
    PyObject *exc_tb;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (!exc_state->exc_value) return;
    exc_tb = PyException_GetTraceback(exc_state->exc_value);
    #else
    exc_tb = exc_state->exc_traceback;
    #endif
    if (likely(exc_tb)) {
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
        #if PY_VERSION_HEX >= 0x030B00a4
        Py_DECREF(exc_tb);
        #endif
    }
#endif
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(PyObject* gen, PyObject *retval) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
#if PY_VERSION_HEX < 0x030d00A1
            _PyGen_SetStopIterationValue(result);
#else
            if (!PyTuple_Check(result) && !PyExceptionInstance_Check(result)) {
                PyErr_SetObject(PyExc_StopIteration, result);
            } else {
                PyObject *exc = __Pyx_PyObject_CallOneArg(PyExc_StopIteration, result);
                if (likely(exc != NULL)) {
                    PyErr_SetObject(PyExc_StopIteration, exc);
                    Py_DECREF(exc);
                }
            }
#endif
        }
        Py_DECREF(result);
        result = NULL;
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (unlikely(PyErr_Occurred())) {
                PyErr_WriteUnraisable(yf);
            }
        } else {
            retval = __Pyx_PyObject_CallNoArg(meth);
            Py_DECREF(meth);
            if (unlikely(!retval))
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (unlikely(PyErr_Occurred())) {
                    gen->is_running = 0;
                    return NULL;
                }
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = __Pyx_PyObject_Call(meth, args, NULL);
            } else {
                PyObject *cargs[4] = {NULL, typ, val, tb};
                ret = __Pyx_PyObject_FastCall(meth, cargs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (unlikely(!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb)))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_VISIT(exc_state->exc_value);
#else
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
#endif
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (unlikely(PyObject_CallFinalizerFromDealloc(self)))
#else
        Py_TYPE(gen)->tp_del(self);
        if (unlikely(Py_REFCNT(self) > 0))
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    __Pyx_PyHeapTypeObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (likely(--self->ob_refcnt == 0)) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_name;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_name, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_qualname;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_qualname, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, void *context)
{
    PyObject *frame = self->gi_frame;
    CYTHON_UNUSED_VAR(context);
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    #if PY_VERSION_HEX >= 0x030B00a4
    gen->gi_exc_state.exc_value = NULL;
    #else
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
    #endif
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (unlikely(!module)) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Coroutine */
static void __Pyx_CoroutineAwait_dealloc(PyObject *self) {
    PyObject_GC_UnTrack(self);
    Py_CLEAR(((__pyx_CoroutineAwaitObject*)self)->coroutine);
    __Pyx_PyHeapTypeObject_GC_Del(self);
}
static int __Pyx_CoroutineAwait_traverse(__pyx_CoroutineAwaitObject *self, visitproc visit, void *arg) {
    Py_VISIT(self->coroutine);
    return 0;
}
static int __Pyx_CoroutineAwait_clear(__pyx_CoroutineAwaitObject *self) {
    Py_CLEAR(self->coroutine);
    return 0;
}
static PyObject *__Pyx_CoroutineAwait_Next(__pyx_CoroutineAwaitObject *self) {
    return __Pyx_Generator_Next(self->coroutine);
}
static PyObject *__Pyx_CoroutineAwait_Send(__pyx_CoroutineAwaitObject *self, PyObject *value) {
    return __Pyx_Coroutine_Send(self->coroutine, value);
}
static PyObject *__Pyx_CoroutineAwait_Throw(__pyx_CoroutineAwaitObject *self, PyObject *args) {
    return __Pyx_Coroutine_Throw(self->coroutine, args);
}
static PyObject *__Pyx_CoroutineAwait_Close(__pyx_CoroutineAwaitObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx_Coroutine_Close(self->coroutine);
}
static PyObject *__Pyx_CoroutineAwait_self(PyObject *self) {
    Py_INCREF(self);
    return self;
}
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_CoroutineAwait_no_new(PyTypeObject *type, PyObject *args, PyObject *kwargs) {
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(args);
    CYTHON_UNUSED_VAR(kwargs);
    PyErr_SetString(PyExc_TypeError, "cannot instantiate type, use 'await coroutine' instead");
    return NULL;
}
#endif
static PyObject *__Pyx_CoroutineAwait_reduce_ex(__pyx_CoroutineAwaitObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    PyErr_Format(PyExc_TypeError, "cannot pickle '%.200s' object",
                         Py_TYPE(self)->tp_name);
    return NULL;
}
static PyMethodDef __pyx_CoroutineAwait_methods[] = {
    {"send", (PyCFunction) __Pyx_CoroutineAwait_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into coroutine,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_CoroutineAwait_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in coroutine,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_CoroutineAwait_Close, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside coroutine.")},
    {"__reduce_ex__", (PyCFunction) __Pyx_CoroutineAwait_reduce_ex, METH_O, 0},
    {"__reduce__", (PyCFunction) __Pyx_CoroutineAwait_reduce_ex, METH_NOARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CoroutineAwaitType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CoroutineAwait_dealloc},
    {Py_tp_traverse, (void *)__Pyx_CoroutineAwait_traverse},
    {Py_tp_clear, (void *)__Pyx_CoroutineAwait_clear},
#if !CYTHON_COMPILING_IN_PYPY
    {Py_tp_new, (void *)__Pyx_CoroutineAwait_no_new},
#endif
    {Py_tp_methods, (void *)__pyx_CoroutineAwait_methods},
    {Py_tp_iter, (void *)__Pyx_CoroutineAwait_self},
    {Py_tp_iternext, (void *)__Pyx_CoroutineAwait_Next},
    {0, 0},
};
static PyType_Spec __pyx_CoroutineAwaitType_spec = {
    __PYX_TYPE_MODULE_PREFIX "coroutine_wrapper",
    sizeof(__pyx_CoroutineAwaitObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    __pyx_CoroutineAwaitType_slots
};
#else
static PyTypeObject __pyx_CoroutineAwaitType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "coroutine_wrapper",
    sizeof(__pyx_CoroutineAwaitObject),
    0,
    (destructor) __Pyx_CoroutineAwait_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    PyDoc_STR("A wrapper object implementing __await__ for coroutines."),
    (traverseproc) __Pyx_CoroutineAwait_traverse,
    (inquiry) __Pyx_CoroutineAwait_clear,
    0,
    0,
    __Pyx_CoroutineAwait_self,
    (iternextfunc) __Pyx_CoroutineAwait_Next,
    __pyx_CoroutineAwait_methods,
    0                         ,
    0                      ,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if !CYTHON_COMPILING_IN_PYPY
    __Pyx_CoroutineAwait_no_new,
#else
    0,
#endif
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
#if PY_VERSION_HEX < 0x030500B1 || defined(__Pyx_IterableCoroutine_USED) || CYTHON_USE_ASYNC_SLOTS
static CYTHON_INLINE PyObject *__Pyx__Coroutine_await(PyObject *coroutine) {
    __pyx_CoroutineAwaitObject *await = PyObject_GC_New(__pyx_CoroutineAwaitObject, __pyx_CoroutineAwaitType);
    if (unlikely(!await)) return NULL;
    Py_INCREF(coroutine);
    await->coroutine = coroutine;
    PyObject_GC_Track(await);
    return (PyObject*)await;
}
#endif
#if PY_VERSION_HEX < 0x030500B1
static PyObject *__Pyx_Coroutine_await_method(PyObject *coroutine, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx__Coroutine_await(coroutine);
}
#endif
#if defined(__Pyx_IterableCoroutine_USED) || CYTHON_USE_ASYNC_SLOTS
static PyObject *__Pyx_Coroutine_await(PyObject *coroutine) {
    if (unlikely(!coroutine || !__Pyx_Coroutine_Check(coroutine))) {
        PyErr_SetString(PyExc_TypeError, "invalid input, expected coroutine");
        return NULL;
    }
    return __Pyx__Coroutine_await(coroutine);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3 && PY_VERSION_HEX < 0x030500B1
static PyObject *__Pyx_Coroutine_compare(PyObject *obj, PyObject *other, int op) {
    PyObject* result;
    switch (op) {
        case Py_EQ: result = (other == obj) ? Py_True : Py_False; break;
        case Py_NE: result = (other != obj) ? Py_True : Py_False; break;
        default:
            result = Py_NotImplemented;
    }
    Py_INCREF(result);
    return result;
}
#endif
static PyMethodDef __pyx_Coroutine_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into coroutine,\nreturn next iterated value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in coroutine,\nreturn next iterated value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside coroutine.")},
#if PY_VERSION_HEX < 0x030500B1
    {"__await__", (PyCFunction) __Pyx_Coroutine_await_method, METH_NOARGS,
     (char*) PyDoc_STR("__await__() -> return an iterator to be used in await expression.")},
#endif
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Coroutine_memberlist[] = {
    {(char *) "cr_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "cr_await", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being awaited, or None")},
    {(char*) "cr_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Coroutine_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the coroutine"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the coroutine"), 0},
    {(char *) "cr_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the coroutine"), 0},
    {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CoroutineType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_am_await, (void *)&__Pyx_Coroutine_await},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_methods, (void *)__pyx_Coroutine_methods},
    {Py_tp_members, (void *)__pyx_Coroutine_memberlist},
    {Py_tp_getset, (void *)__pyx_Coroutine_getsets},
    {Py_tp_getattro, (void *) __Pyx_PyObject_GenericGetAttrNoDict},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
    {0, 0},
};
static PyType_Spec __pyx_CoroutineType_spec = {
    __PYX_TYPE_MODULE_PREFIX "coroutine",
    sizeof(__pyx_CoroutineObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    __pyx_CoroutineType_slots
};
#else
#if CYTHON_USE_ASYNC_SLOTS
static __Pyx_PyAsyncMethodsStruct __pyx_Coroutine_as_async = {
    __Pyx_Coroutine_await,
    0,
    0,
#if PY_VERSION_HEX >= 0x030A00A3
    0,
#endif
};
#endif
static PyTypeObject __pyx_CoroutineType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "coroutine",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
#if CYTHON_USE_ASYNC_SLOTS
    &__pyx_Coroutine_as_async,
#else
    0,
#endif
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
#if CYTHON_USE_ASYNC_SLOTS && CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3 && PY_VERSION_HEX < 0x030500B1
    __Pyx_Coroutine_compare,
#else
    0,
#endif
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    0,
    __pyx_Coroutine_methods,
    __pyx_Coroutine_memberlist,
    __pyx_Coroutine_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_Coroutine_init(PyObject *module) {
    CYTHON_MAYBE_UNUSED_VAR(module);
#if CYTHON_USE_TYPE_SPECS
    __pyx_CoroutineType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CoroutineType_spec, NULL);
#else
    __pyx_CoroutineType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_CoroutineType = __Pyx_FetchCommonType(&__pyx_CoroutineType_type);
#endif
    if (unlikely(!__pyx_CoroutineType))
        return -1;
#ifdef __Pyx_IterableCoroutine_USED
    if (unlikely(__pyx_IterableCoroutine_init(module) == -1))
        return -1;
#endif
#if CYTHON_USE_TYPE_SPECS
    __pyx_CoroutineAwaitType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CoroutineAwaitType_spec, NULL);
#else
    __pyx_CoroutineAwaitType = __Pyx_FetchCommonType(&__pyx_CoroutineAwaitType_type);
#endif
    if (unlikely(!__pyx_CoroutineAwaitType))
        return -1;
    return 0;
}

/* GetAwaitIter */
static CYTHON_INLINE PyObject *__Pyx_Coroutine_GetAwaitableIter(PyObject *o) {
#ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(o)) {
        return __Pyx_NewRef(o);
    }
#endif
    return __Pyx__Coroutine_GetAwaitableIter(o);
}
static void __Pyx_Coroutine_AwaitableIterError(PyObject *source) {
#if PY_VERSION_HEX >= 0x030600B3 && PY_VERSION_HEX < 0x030d0000 || defined(_PyErr_FormatFromCause)
    __Pyx_TypeName source_type_name = __Pyx_PyType_GetName(Py_TYPE(source));
    _PyErr_FormatFromCause(PyExc_TypeError,
        "'async for' received an invalid object from __anext__: " __Pyx_FMT_TYPENAME, source_type_name);
    __Pyx_DECREF_TypeName(source_type_name);
#elif PY_MAJOR_VERSION >= 3
    PyObject *exc, *val, *val2, *tb;
    __Pyx_TypeName source_type_name = __Pyx_PyType_GetName(Py_TYPE(source));
    assert(PyErr_Occurred());
    PyErr_Fetch(&exc, &val, &tb);
    PyErr_NormalizeException(&exc, &val, &tb);
    if (tb != NULL) {
        PyException_SetTraceback(val, tb);
        Py_DECREF(tb);
    }
    Py_DECREF(exc);
    assert(!PyErr_Occurred());
    PyErr_Format(PyExc_TypeError,
        "'async for' received an invalid object from __anext__: " __Pyx_FMT_TYPENAME, source_type_name);
    __Pyx_DECREF_TypeName(source_type_name);
    PyErr_Fetch(&exc, &val2, &tb);
    PyErr_NormalizeException(&exc, &val2, &tb);
    Py_INCREF(val);
    PyException_SetCause(val2, val);
    PyException_SetContext(val2, val);
    PyErr_Restore(exc, val2, tb);
#else
    CYTHON_UNUSED_VAR(source);
#endif
}
static PyObject *__Pyx__Coroutine_GetAwaitableIter(PyObject *obj) {
    PyObject *res;
#if CYTHON_USE_ASYNC_SLOTS
    __Pyx_PyAsyncMethodsStruct* am = __Pyx_PyType_AsAsync(obj);
    if (likely(am && am->am_await)) {
        res = (*am->am_await)(obj);
    } else
#endif
#if PY_VERSION_HEX >= 0x030500B2 || defined(PyCoro_CheckExact)
    if (PyCoro_CheckExact(obj)) {
        return __Pyx_NewRef(obj);
    } else
#endif
#if CYTHON_COMPILING_IN_CPYTHON && defined(CO_ITERABLE_COROUTINE)
#if PY_VERSION_HEX >= 0x030C00A6
    if (PyGen_CheckExact(obj) && (PyGen_GetCode((PyGenObject*)obj)->co_flags & CO_ITERABLE_COROUTINE)) {
#else
    if (PyGen_CheckExact(obj) && ((PyGenObject*)obj)->gi_code && ((PyCodeObject *)((PyGenObject*)obj)->gi_code)->co_flags & CO_ITERABLE_COROUTINE) {
#endif
        return __Pyx_NewRef(obj);
    } else
#endif
    {
        PyObject *method = NULL;
        int is_method = __Pyx_PyObject_GetMethod(obj, __pyx_n_s_await, &method);
        if (likely(is_method)) {
            res = __Pyx_PyObject_CallOneArg(method, obj);
        } else if (likely(method)) {
            res = __Pyx_PyObject_CallNoArg(method);
        } else
            goto slot_error;
        Py_DECREF(method);
    }
    if (unlikely(!res)) {
        __Pyx_Coroutine_AwaitableIterError(obj);
        goto bad;
    }
    if (unlikely(!PyIter_Check(res))) {
        __Pyx_TypeName res_type_name = __Pyx_PyType_GetName(Py_TYPE(res));
        PyErr_Format(PyExc_TypeError,
            "__await__() returned non-iterator of type '" __Pyx_FMT_TYPENAME "'", res_type_name);
        __Pyx_DECREF_TypeName(res_type_name);
        Py_CLEAR(res);
    } else {
        int is_coroutine = 0;
        #ifdef __Pyx_Coroutine_USED
        is_coroutine |= __Pyx_Coroutine_Check(res);
        #endif
        #if PY_VERSION_HEX >= 0x030500B2 || defined(PyCoro_CheckExact)
        is_coroutine |= PyCoro_CheckExact(res);
        #endif
        if (unlikely(is_coroutine)) {
            /* __await__ must return an *iterator*, not
               a coroutine or another awaitable (see PEP 492) */
            PyErr_SetString(PyExc_TypeError,
                            "__await__() returned a coroutine");
            Py_CLEAR(res);
        }
    }
    return res;
slot_error:
    {
        __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
        PyErr_Format(PyExc_TypeError,
            "object " __Pyx_FMT_TYPENAME " can't be used in 'await' expression", obj_type_name);
        __Pyx_DECREF_TypeName(obj_type_name);
    }
bad:
    return NULL;
}

/* CoroutineYieldFrom */
  static PyObject* __Pyx__Coroutine_Yield_From_Generic(__pyx_CoroutineObject *gen, PyObject *source) {
    PyObject *retval;
    PyObject *source_gen = __Pyx__Coroutine_GetAwaitableIter(source);
    if (unlikely(!source_gen)) {
        return NULL;
    }
    if (__Pyx_Coroutine_Check(source_gen)) {
        retval = __Pyx_Generator_Next(source_gen);
    } else {
        retval = __Pyx_PyObject_GetIterNextFunc(source_gen)(source_gen);
    }
    if (retval) {
        gen->yieldfrom = source_gen;
        return retval;
    }
    Py_DECREF(source_gen);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_Coroutine_Yield_From(__pyx_CoroutineObject *gen, PyObject *source) {
    PyObject *retval;
    if (__Pyx_Coroutine_Check(source)) {
        if (unlikely(((__pyx_CoroutineObject*)source)->yieldfrom)) {
            PyErr_SetString(
                PyExc_RuntimeError,
                "coroutine is being awaited already");
            return NULL;
        }
        retval = __Pyx_Generator_Next(source);
#ifdef __Pyx_AsyncGen_USED
    } else if (__pyx_PyAsyncGenASend_CheckExact(source)) {
        retval = __Pyx_async_gen_asend_iternext(source);
#endif
    } else {
        return __Pyx__Coroutine_Yield_From_Generic(gen, source);
    }
    if (retval) {
        Py_INCREF(source);
        gen->yieldfrom = source;
    }
    return retval;
}

/* pep479 */
  static void __Pyx_Generator_Replace_StopIteration(int in_async_gen) {
    PyObject *exc, *val, *tb, *cur_exc;
    __Pyx_PyThreadState_declare
    #ifdef __Pyx_StopAsyncIteration_USED
    int is_async_stopiteration = 0;
    #endif
    CYTHON_MAYBE_UNUSED_VAR(in_async_gen);
    cur_exc = PyErr_Occurred();
    if (likely(!__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopIteration))) {
        #ifdef __Pyx_StopAsyncIteration_USED
        if (in_async_gen && unlikely(__Pyx_PyErr_GivenExceptionMatches(cur_exc, __Pyx_PyExc_StopAsyncIteration))) {
            is_async_stopiteration = 1;
        } else
        #endif
            return;
    }
    __Pyx_PyThreadState_assign
    __Pyx_GetException(&exc, &val, &tb);
    Py_XDECREF(exc);
    Py_XDECREF(val);
    Py_XDECREF(tb);
    PyErr_SetString(PyExc_RuntimeError,
        #ifdef __Pyx_StopAsyncIteration_USED
        is_async_stopiteration ? "async generator raised StopAsyncIteration" :
        in_async_gen ? "async generator raised StopIteration" :
        #endif
        "generator raised StopIteration");
}

/* ReturnWithStopIteration */
  static void __Pyx__ReturnWithStopIteration(PyObject* value) {
    PyObject *exc, *args;
#if CYTHON_COMPILING_IN_CPYTHON
    __Pyx_PyThreadState_declare
    if (PY_VERSION_HEX >= 0x030C00A6
            || unlikely(PyTuple_Check(value) || PyExceptionInstance_Check(value))) {
        args = PyTuple_New(1);
        if (unlikely(!args)) return;
        Py_INCREF(value);
        PyTuple_SET_ITEM(args, 0, value);
        exc = PyType_Type.tp_call(PyExc_StopIteration, args, NULL);
        Py_DECREF(args);
        if (!exc) return;
    } else {
        Py_INCREF(value);
        exc = value;
    }
    #if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    #if CYTHON_USE_EXC_INFO_STACK
    if (!__pyx_tstate->exc_info->exc_value)
    #else
    if (!__pyx_tstate->exc_type)
    #endif
    {
        Py_INCREF(PyExc_StopIteration);
        __Pyx_ErrRestore(PyExc_StopIteration, exc, NULL);
        return;
    }
    #endif
#else
    args = PyTuple_Pack(1, value);
    if (unlikely(!args)) return;
    exc = PyObject_Call(PyExc_StopIteration, args, NULL);
    Py_DECREF(args);
    if (unlikely(!exc)) return;
#endif
    PyErr_SetObject(PyExc_StopIteration, exc);
    Py_DECREF(exc);
}

/* UnpackUnboundCMethod */
  static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
     METH_VARARGS | METH_KEYWORDS,
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #else
    if (likely(!__Pyx_CyOrPyCFunction_Check(method)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            target->method = unbound_method;
        }
    }
    return 0;
}

/* CallUnboundCMethod0 */
  static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* append */
  static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* pybytes_as_double */
  static double __Pyx_SlowPyString_AsDouble(PyObject *obj) {
    PyObject *float_value;
#if PY_MAJOR_VERSION >= 3
    float_value = PyFloat_FromString(obj);
#else
    float_value = PyFloat_FromString(obj, 0);
#endif
    if (likely(float_value)) {
#if CYTHON_ASSUME_SAFE_MACROS
        double value = PyFloat_AS_DOUBLE(float_value);
#else
        double value = PyFloat_AsDouble(float_value);
#endif
        Py_DECREF(float_value);
        return value;
    }
    return (double)-1;
}
static const char* __Pyx__PyBytes_AsDouble_Copy(const char* start, char* buffer, Py_ssize_t length) {
    int last_was_punctuation = 1;
    Py_ssize_t i;
    for (i=0; i < length; i++) {
        char chr = start[i];
        int is_punctuation = (chr == '_') | (chr == '.') | (chr == 'e') | (chr == 'E');
        *buffer = chr;
        buffer += (chr != '_');
        if (unlikely(last_was_punctuation & is_punctuation)) goto parse_failure;
        last_was_punctuation = is_punctuation;
    }
    if (unlikely(last_was_punctuation)) goto parse_failure;
    *buffer = '\0';
    return buffer;
parse_failure:
    return NULL;
}
static double __Pyx__PyBytes_AsDouble_inf_nan(const char* start, Py_ssize_t length) {
    int matches = 1;
    char sign = start[0];
    int is_signed = (sign == '+') | (sign == '-');
    start += is_signed;
    length -= is_signed;
    switch (start[0]) {
        #ifdef Py_NAN
        case 'n':
        case 'N':
            if (unlikely(length != 3)) goto parse_failure;
            matches &= (start[1] == 'a' || start[1] == 'A');
            matches &= (start[2] == 'n' || start[2] == 'N');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_NAN : Py_NAN;
        #endif
        case 'i':
        case 'I':
            if (unlikely(length < 3)) goto parse_failure;
            matches &= (start[1] == 'n' || start[1] == 'N');
            matches &= (start[2] == 'f' || start[2] == 'F');
            if (likely(length == 3 && matches))
                return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
            if (unlikely(length != 8)) goto parse_failure;
            matches &= (start[3] == 'i' || start[3] == 'I');
            matches &= (start[4] == 'n' || start[4] == 'N');
            matches &= (start[5] == 'i' || start[5] == 'I');
            matches &= (start[6] == 't' || start[6] == 'T');
            matches &= (start[7] == 'y' || start[7] == 'Y');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
        case '.': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
            break;
        default:
            goto parse_failure;
    }
    return 0.0;
parse_failure:
    return -1.0;
}
static CYTHON_INLINE int __Pyx__PyBytes_AsDouble_IsSpace(char ch) {
    return (ch == 0x20) | !((ch < 0x9) | (ch > 0xd));
}
CYTHON_UNUSED static double __Pyx__PyBytes_AsDouble(PyObject *obj, const char* start, Py_ssize_t length) {
    double value;
    Py_ssize_t i, digits;
    const char *last = start + length;
    char *end;
    while (__Pyx__PyBytes_AsDouble_IsSpace(*start))
        start++;
    while (start < last - 1 && __Pyx__PyBytes_AsDouble_IsSpace(last[-1]))
        last--;
    length = last - start;
    if (unlikely(length <= 0)) goto fallback;
    value = __Pyx__PyBytes_AsDouble_inf_nan(start, length);
    if (unlikely(value == -1.0)) goto fallback;
    if (value != 0.0) return value;
    digits = 0;
    for (i=0; i < length; digits += start[i++] != '_');
    if (likely(digits == length)) {
        value = PyOS_string_to_double(start, &end, NULL);
    } else if (digits < 40) {
        char number[40];
        last = __Pyx__PyBytes_AsDouble_Copy(start, number, length);
        if (unlikely(!last)) goto fallback;
        value = PyOS_string_to_double(number, &end, NULL);
    } else {
        char *number = (char*) PyMem_Malloc((digits + 1) * sizeof(char));
        if (unlikely(!number)) goto fallback;
        last = __Pyx__PyBytes_AsDouble_Copy(start, number, length);
        if (unlikely(!last)) {
            PyMem_Free(number);
            goto fallback;
        }
        value = PyOS_string_to_double(number, &end, NULL);
        PyMem_Free(number);
    }
    if (likely(end == last) || (value == (double)-1 && PyErr_Occurred())) {
        return value;
    }
fallback:
    return __Pyx_SlowPyString_AsDouble(obj);
}

/* pynumber_float */
  static CYTHON_INLINE PyObject* __Pyx__PyNumber_Float(PyObject* obj) {
    double val;
    if (PyLong_CheckExact(obj)) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (likely(__Pyx_PyLong_IsCompact(obj))) {
            val = (double) __Pyx_PyLong_CompactValue(obj);
            goto no_error;
        }
#endif
        val = PyLong_AsDouble(obj);
    } else if (PyUnicode_CheckExact(obj)) {
        val = __Pyx_PyUnicode_AsDouble(obj);
    } else if (PyBytes_CheckExact(obj)) {
        val = __Pyx_PyBytes_AsDouble(obj);
    } else if (PyByteArray_CheckExact(obj)) {
        val = __Pyx_PyByteArray_AsDouble(obj);
    } else {
        return PyNumber_Float(obj);
    }
    if (unlikely(val == -1 && PyErr_Occurred())) {
        return NULL;
    }
#if CYTHON_USE_PYLONG_INTERNALS
no_error:
#endif
    return PyFloat_FromDouble(val);
}

/* decode_c_bytes */
  static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    if (unlikely((start < 0) | (stop < 0))) {
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (stop > length)
        stop = length;
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* Import */
  static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.') != NULL) {
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportFrom */
  static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_kp_u__15);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* GetAttr */
  static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* HasAttr */
  static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (!r) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* CallNextTpDealloc */
  static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc) {
    PyTypeObject* type = Py_TYPE(obj);
    destructor tp_dealloc = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_dealloc, destructor) != current_tp_dealloc)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_dealloc = __Pyx_PyType_GetSlot(type, tp_dealloc, destructor)) == current_tp_dealloc)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type)
        tp_dealloc(obj);
}

/* CallNextTpTraverse */
  static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    traverseproc tp_traverse = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc) != current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_traverse = __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc)) == current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_traverse)
        return tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
  static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    inquiry tp_clear = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_clear, inquiry) != current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_clear = __Pyx_PyType_GetSlot(type, tp_clear, inquiry)) == current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_clear)
        tp_clear(obj);
}

/* ValidateBasesTuple */
  #if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_MACROS
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (n < 0) return -1;
#endif
    for (i = 1; i < n; i++)
    {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        PyTypeObject *b;
#if PY_MAJOR_VERSION < 3
        if (PyClass_Check(b0))
        {
            PyErr_Format(PyExc_TypeError, "base class '%.200s' is an old-style class",
                         PyString_AS_STRING(((PyClassObject*)b0)->cl_name));
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !(CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY)
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
  static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !(CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API) || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if PY_VERSION_HEX >= 0x030700a1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400)
        gc = PyImport_GetModule(__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* PyObject_GenericGetAttr */
  #if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetupReduce */
  #if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}
#endif

/* TypeImport */
  #ifndef __PYX_HAVE_RT_ImportType_3_0_12
#define __PYX_HAVE_RT_ImportType_3_0_12
static PyTypeObject *__Pyx_ImportType_3_0_12(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_12 check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_0_12 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_0_12 && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
  static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_n_s_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_n_s_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* SetVTable */
  static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_n_s_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_n_s_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* MergeVTables */
  #if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i;
    void** base_vtables;
    __Pyx_TypeName tp_base_name;
    __Pyx_TypeName base_name;
    void* unknown = (void*)-1;
    PyObject* bases = type->tp_bases;
    int base_depth = 0;
    {
        PyTypeObject* base = type->tp_base;
        while (base) {
            base_depth += 1;
            base = base->tp_base;
        }
    }
    base_vtables = (void**) malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
    for (i = 1; i < PyTuple_GET_SIZE(bases); i++) {
        void* base_vtable = __Pyx_GetVtable(((PyTypeObject*)PyTuple_GET_ITEM(bases, i)));
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = type->tp_base;
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = base->tp_base;
            }
        }
    }
    PyErr_Clear();
    free(base_vtables);
    return 0;
bad:
    tp_base_name = __Pyx_PyType_GetName(type->tp_base);
    base_name = __Pyx_PyType_GetName((PyTypeObject*)PyTuple_GET_ITEM(bases, i));
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
    free(base_vtables);
    return -1;
}
#endif

/* ImportDottedModule */
  #if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
    if (likely(PyTuple_GET_SIZE(parts_tuple) == count)) {
        partial_name = name;
    } else {
        slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
#if PY_MAJOR_VERSION < 3
        PyExc_ImportError,
        "No module named '%s'", PyString_AS_STRING(partial_name));
#else
#if PY_VERSION_HEX >= 0x030600B1
        PyExc_ModuleNotFoundError,
#else
        PyExc_ImportError,
#endif
        "No module named '%U'", partial_name);
#endif
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple) {
    Py_ssize_t i, nparts;
    nparts = PyTuple_GET_SIZE(parts_tuple);
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = PySequence_ITEM(parts_tuple, i);
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (unlikely(!module)) {
        return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
    }
    return module;
}
#endif
static PyObject *__Pyx__ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if PY_MAJOR_VERSION < 3
    PyObject *module, *from_list, *star = __pyx_n_s__36;
    CYTHON_UNUSED_VAR(parts_tuple);
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
    module = __Pyx_Import(name, from_list, 0);
    Py_DECREF(from_list);
    return module;
#else
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    return __Pyx_ImportDottedModule_WalkParts(module, name, parts_tuple);
#endif
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030400B1
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_n_s_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_n_s_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* PatchInspect */
  static PyObject* __Pyx_patch_inspect(PyObject* module) {
#if defined(__Pyx_Generator_USED) && (!defined(CYTHON_PATCH_INSPECT) || CYTHON_PATCH_INSPECT)
    static int inspect_patched = 0;
    if (unlikely((!inspect_patched) && module)) {
        module = __Pyx_Coroutine_patch_module(
            module, ""
"old_types = getattr(_module.isgenerator, '_cython_generator_types', None)\n"
"if old_types is None or not isinstance(old_types, set):\n"
"    old_types = set()\n"
"    def cy_wrap(orig_func, type=type, cython_generator_types=old_types):\n"
"        def cy_isgenerator(obj): return type(obj) in cython_generator_types or orig_func(obj)\n"
"        cy_isgenerator._cython_generator_types = cython_generator_types\n"
"        return cy_isgenerator\n"
"    _module.isgenerator = cy_wrap(_module.isgenerator)\n"
"old_types.add(_cython_generator_type)\n"
        );
        inspect_patched = 1;
    }
#else
    if ((0)) return __Pyx_Coroutine_patch_module(module, NULL);
#endif
    return module;
}

/* PatchAsyncIO */
  static PyObject* __Pyx_patch_asyncio(PyObject* module) {
#if PY_VERSION_HEX < 0x030500B2 &&\
        (defined(__Pyx_Coroutine_USED) || defined(__Pyx_Generator_USED)) &&\
        (!defined(CYTHON_PATCH_ASYNCIO) || CYTHON_PATCH_ASYNCIO)
    PyObject *patch_module = NULL;
    static int asyncio_patched = 0;
    if (unlikely((!asyncio_patched) && module)) {
        PyObject *package;
        package = __Pyx_Import(__pyx_n_s_asyncio_coroutines, NULL, 0);
        if (package) {
            patch_module = __Pyx_Coroutine_patch_module(
                PyObject_GetAttrString(package, "coroutines"), ""
"try:\n"
"    coro_types = _module._COROUTINE_TYPES\n"
"except AttributeError: pass\n"
"else:\n"
"    if _cython_coroutine_type is not None and _cython_coroutine_type not in coro_types:\n"
"        coro_types = tuple(coro_types) + (_cython_coroutine_type,)\n"
"    if _cython_generator_type is not None and _cython_generator_type not in coro_types:\n"
"        coro_types = tuple(coro_types) + (_cython_generator_type,)\n"
"_module._COROUTINE_TYPES = coro_types\n"
            );
        } else {
            PyErr_Clear();
            package = __Pyx_Import(__pyx_n_s_asyncio_tasks, NULL, 0);
            if (unlikely(!package)) goto asyncio_done;
            patch_module = __Pyx_Coroutine_patch_module(
                PyObject_GetAttrString(package, "tasks"), ""
"if hasattr(_module, 'iscoroutine'):\n"
"    old_types = getattr(_module.iscoroutine, '_cython_coroutine_types', None)\n"
"    if old_types is None or not isinstance(old_types, set):\n"
"        old_types = set()\n"
"        def cy_wrap(orig_func, type=type, cython_coroutine_types=old_types):\n"
"            def cy_iscoroutine(obj): return type(obj) in cython_coroutine_types or orig_func(obj)\n"
"            cy_iscoroutine._cython_coroutine_types = cython_coroutine_types\n"
"            return cy_iscoroutine\n"
"        _module.iscoroutine = cy_wrap(_module.iscoroutine)\n"
"    if _cython_coroutine_type is not None:\n"
"        old_types.add(_cython_coroutine_type)\n"
"    if _cython_generator_type is not None:\n"
"        old_types.add(_cython_generator_type)\n"
            );
        }
        Py_DECREF(package);
        if (unlikely(!patch_module)) goto ignore;
asyncio_done:
        PyErr_Clear();
        asyncio_patched = 1;
#ifdef __Pyx_Generator_USED
        {
            PyObject *inspect_module;
            if (patch_module) {
                inspect_module = PyObject_GetAttr(patch_module, __pyx_n_s_inspect);
                Py_DECREF(patch_module);
            } else {
                inspect_module = __Pyx_Import(__pyx_n_s_inspect, NULL, 0);
            }
            if (unlikely(!inspect_module)) goto ignore;
            inspect_module = __Pyx_patch_inspect(inspect_module);
            if (unlikely(!inspect_module)) {
                Py_DECREF(module);
                module = NULL;
            }
            Py_XDECREF(inspect_module);
        }
#else
        if ((0)) return __Pyx_patch_inspect(module);
#endif
    }
    return module;
ignore:
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch asyncio package with custom generator type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    if ((0)) return __Pyx_patch_inspect(__Pyx_Coroutine_patch_module(module, NULL));
#endif
    return module;
}

/* ClassMethod */
  static PyObject* __Pyx_Method_ClassMethod(PyObject *method) {
#if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM <= 0x05080000
    if (PyObject_TypeCheck(method, &PyWrapperDescr_Type)) {
        return PyClassMethod_New(method);
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (PyMethodDescr_Check(method))
#else
    #if PY_MAJOR_VERSION == 2
    static PyTypeObject *methoddescr_type = NULL;
    if (unlikely(methoddescr_type == NULL)) {
       PyObject *meth = PyObject_GetAttrString((PyObject*)&PyList_Type, "append");
       if (unlikely(!meth)) return NULL;
       methoddescr_type = Py_TYPE(meth);
       Py_DECREF(meth);
    }
    #else
    PyTypeObject *methoddescr_type = &PyMethodDescr_Type;
    #endif
    if (__Pyx_TypeCheck(method, methoddescr_type))
#endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject *)method;
        #if PY_VERSION_HEX < 0x03020000
        PyTypeObject *d_type = descr->d_type;
        #else
        PyTypeObject *d_type = descr->d_common.d_type;
        #endif
        return PyDescr_NewClassMethod(d_type, descr->d_method);
    }
#endif
    else if (PyMethod_Check(method)) {
        return PyClassMethod_New(PyMethod_GET_FUNCTION(method));
    }
    else {
        return PyClassMethod_New(method);
    }
}

/* GetNameInClass */
  static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    PyObject *dict;
    assert(PyType_Check(nmspace));
#if CYTHON_USE_TYPE_SLOTS
    dict = ((PyTypeObject*)nmspace)->tp_dict;
    Py_XINCREF(dict);
#else
    dict = PyObject_GetAttr(nmspace, __pyx_n_s_dict);
#endif
    if (likely(dict)) {
        result = PyObject_GetItem(dict, name);
        Py_DECREF(dict);
        if (result) {
            return result;
        }
    }
    PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}

/* CLineInTraceback */
  #ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
  #if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
  #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result;
        result = PyObject_Call(replace, __pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    #if __PYX_LIMITED_VERSION_HEX < 0x030780000
    {
        PyObject *compiled = NULL, *result = NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "code", code))) return NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "type", (PyObject*)(&PyType_Type)))) return NULL;
        compiled = Py_CompileString(
            "out = type(code)(\n"
            "  code.co_argcount, code.co_kwonlyargcount, code.co_nlocals, code.co_stacksize,\n"
            "  code.co_flags, code.co_code, code.co_consts, code.co_names,\n"
            "  code.co_varnames, code.co_filename, co_name, co_firstlineno,\n"
            "  code.co_lnotab)\n", "<dummy>", Py_file_input);
        if (!compiled) return NULL;
        result = PyEval_EvalCode(compiled, scratch_dict, scratch_dict);
        Py_DECREF(compiled);
        if (!result) PyErr_Print();
        Py_DECREF(result);
        result = PyDict_GetItemString(scratch_dict, "out");
        if (result) Py_INCREF(result);
        return result;
    }
    #else
    return NULL;
    #endif
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
    if (unlikely(!code_object)) goto bad;
    py_py_line = PyLong_FromLong(py_line);
    if (unlikely(!py_py_line)) goto bad;
    py_funcname = PyUnicode_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    dict = PyDict_New();
    if (unlikely(!dict)) goto bad;
    {
        PyObject *old_code_object = code_object;
        code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
        Py_DECREF(old_code_object);
    }
    if (unlikely(!code_object)) goto bad;
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
  #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Declarations */
  #if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
  #if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
  #if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
  static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int64_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int64_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        int64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int64_t) -1;
        val = __Pyx_PyInt_As_int64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 2 * PyLong_SHIFT)) {
                            return (int64_t) (((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 3 * PyLong_SHIFT)) {
                            return (int64_t) (((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 4 * PyLong_SHIFT)) {
                            return (int64_t) (((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int64_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int64_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int64_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) ((((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int64_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int64_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int64_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int64_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int64_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int64_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int64_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int64_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int64_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int64_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int64_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int64_t) 1) << (sizeof(int64_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int64_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int64_t");
    return (int64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int64_t");
    return (int64_t) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
  static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_As_PY_LONG_LONG(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const PY_LONG_LONG neg_one = (PY_LONG_LONG) -1, const_zero = (PY_LONG_LONG) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(PY_LONG_LONG) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (PY_LONG_LONG) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (PY_LONG_LONG) -1;
        val = __Pyx_PyInt_As_PY_LONG_LONG(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(PY_LONG_LONG) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(PY_LONG_LONG) >= 2 * PyLong_SHIFT)) {
                            return (PY_LONG_LONG) (((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(PY_LONG_LONG) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(PY_LONG_LONG) >= 3 * PyLong_SHIFT)) {
                            return (PY_LONG_LONG) (((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(PY_LONG_LONG) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(PY_LONG_LONG) >= 4 * PyLong_SHIFT)) {
                            return (PY_LONG_LONG) (((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (PY_LONG_LONG) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(PY_LONG_LONG) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(PY_LONG_LONG) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT)) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(PY_LONG_LONG) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT)) {
                            return (PY_LONG_LONG) ((((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT)) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(PY_LONG_LONG) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT)) {
                            return (PY_LONG_LONG) ((((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT)) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(PY_LONG_LONG) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT)) {
                            return (PY_LONG_LONG) ((((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(PY_LONG_LONG) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(PY_LONG_LONG) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        PY_LONG_LONG val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (PY_LONG_LONG) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (PY_LONG_LONG) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (PY_LONG_LONG) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (PY_LONG_LONG) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(PY_LONG_LONG) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((PY_LONG_LONG) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(PY_LONG_LONG) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((PY_LONG_LONG) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((PY_LONG_LONG) 1) << (sizeof(PY_LONG_LONG) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (PY_LONG_LONG) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to PY_LONG_LONG");
    return (PY_LONG_LONG) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to PY_LONG_LONG");
    return (PY_LONG_LONG) -1;
}

/* CIntToPy */
  static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* FormatTypeName */
  #if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name_2);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XDECREF(name);
        name = __Pyx_NewRef(__pyx_n_s__88);
    }
    return name;
}
#endif

/* FastTypeChecks */
  #if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
  static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030B00A4
    return Py_Version & ~0xFFUL;
#else
    const char* rt_version = Py_GetVersion();
    unsigned long version = 0;
    unsigned long factor = 0x01000000UL;
    unsigned int digit = 0;
    int i = 0;
    while (factor) {
        while ('0' <= rt_version[i] && rt_version[i] <= '9') {
            digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
            ++i;
        }
        version += factor * digit;
        if (rt_version[i] != '.')
            break;
        digit = 0;
        factor >>= 8;
        ++i;
    }
    return version;
#endif
}
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* FunctionImport */
  #ifndef __PYX_HAVE_RT_ImportFunction_3_0_12
#define __PYX_HAVE_RT_ImportFunction_3_0_12
static int __Pyx_ImportFunction_3_0_12(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
  #if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
