/* Generated by Cython 3.0.12 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "hummingbot/core/cpp/LimitOrder.h",
            "hummingbot/core/cpp/OrderBookEntry.h",
            "hummingbot/core/cpp/PyRef.h"
        ],
        "include_dirs": [
            "hummingbot/core",
            "hummingbot/core/data_type"
        ],
        "language": "c++",
        "name": "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making",
        "sources": [
            "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx"
        ]
    },
    "module_name": "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#if defined(CYTHON_LIMITED_API) && 0
  #ifndef Py_LIMITED_API
    #if CYTHON_LIMITED_API+0 > 0x03030000
      #define Py_LIMITED_API CYTHON_LIMITED_API
    #else
      #define Py_LIMITED_API 0x03030000
    #endif
  #endif
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02070000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.7+ or Python 3.3+.
#else
#if defined(CYTHON_LIMITED_API) && CYTHON_LIMITED_API
#define __PYX_EXTRA_ABI_MODULE_NAME "limited"
#else
#define __PYX_EXTRA_ABI_MODULE_NAME ""
#endif
#define CYTHON_ABI "3_0_12" __PYX_EXTRA_ABI_MODULE_NAME
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."
#define CYTHON_HEX_VERSION 0x03000CF0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS (PY_MAJOR_VERSION >= 3)
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1 && PYPY_VERSION_NUM >= 0x07030C00)
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 1
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(Py_GIL_DISABLED) || defined(Py_NOGIL)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #ifndef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 0
  #endif
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_NOGIL 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0 || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL (PY_MAJOR_VERSION < 3 || PY_VERSION_HEX >= 0x03060000 && PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL (PY_VERSION_HEX >= 0x030700A1)
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #if PY_VERSION_HEX < 0x030400a1
    #undef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #elif !defined(CYTHON_USE_TP_FINALIZE)
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #if PY_VERSION_HEX < 0x030600B1
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5)
  #endif
  #if PY_VERSION_HEX < 0x030700A3
    #undef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 0
  #elif !defined(CYTHON_USE_EXC_INFO_STACK)
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if !defined(CYTHON_VECTORCALL)
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_MAJOR_VERSION < 3
    #include "longintrepr.h"
  #endif
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

#ifndef __cplusplus
  #error "Cython files generated with the C++ option must be compiled with a C++ compiler."
#endif
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #else
    #define CYTHON_INLINE inline
  #endif
#endif
template<typename T>
void __Pyx_call_destructor(T& x) {
    x.~T();
}
template<typename T>
class __Pyx_FakeReference {
  public:
    __Pyx_FakeReference() : ptr(NULL) { }
    __Pyx_FakeReference(const T& ref) : ptr(const_cast<T*>(&ref)) { }
    T *operator->() { return ptr; }
    T *operator&() { return ptr; }
    operator T&() { return *ptr; }
    template<typename U> bool operator ==(const U& other) const { return *ptr == other; }
    template<typename U> bool operator !=(const U& other) const { return *ptr != other; }
    template<typename U> bool operator==(const __Pyx_FakeReference<U>& other) const { return *ptr == *other.ptr; }
    template<typename U> bool operator!=(const __Pyx_FakeReference<U>& other) const { return *ptr != *other.ptr; }
  private:
    T *ptr;
};

#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_DefaultClassType PyClass_Type
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    static CYTHON_INLINE PyObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030B0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030B0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
    #ifndef CO_OPTIMIZED
    #define CO_OPTIMIZED 0x0001
    #endif
    #ifndef CO_NEWLOCALS
    #define CO_NEWLOCALS 0x0002
    #endif
    #ifndef CO_VARARGS
    #define CO_VARARGS 0x0004
    #endif
    #ifndef CO_VARKEYWORDS
    #define CO_VARKEYWORDS 0x0008
    #endif
    #ifndef CO_ASYNC_GENERATOR
    #define CO_ASYNC_GENERATOR 0x0200
    #endif
    #ifndef CO_GENERATOR
    #define CO_GENERATOR 0x0020
    #endif
    #ifndef CO_COROUTINE
    #define CO_COROUTINE 0x0080
    #endif
#elif PY_VERSION_HEX >= 0x030B0000
  static CYTHON_INLINE PyCodeObject* __Pyx_PyCode_New(int a, int p, int k, int l, int s, int f,
                                                    PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                                    PyObject *fv, PyObject *cell, PyObject* fn,
                                                    PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    PyObject *empty_bytes = PyBytes_FromStringAndSize("", 0);
    if (!empty_bytes) return NULL;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, empty_bytes);
    Py_DECREF(empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
#endif
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef CO_COROUTINE
  #define CO_COROUTINE 0x80
#endif
#ifndef CO_ASYNC_GENERATOR
  #define CO_ASYNC_GENERATOR 0x200
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_MAJOR_VERSION >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void *cfunc) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x030900B1
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE void *__Pyx_PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE(obj), name, func_ctype)
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if PY_MAJOR_VERSION < 3
    #if CYTHON_COMPILING_IN_PYPY
        #if PYPY_VERSION_NUM < 0x07030600
            #if defined(__cplusplus) && __cplusplus >= 201402L
                [[deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")]]
            #elif defined(__GNUC__) || defined(__clang__)
                __attribute__ ((__deprecated__("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6")))
            #elif defined(_MSC_VER)
                __declspec(deprecated("`with nogil:` inside a nogil function will not release the GIL in PyPy2 < 7.3.6"))
            #endif
            static CYTHON_INLINE int PyGILState_Check(void) {
                return 0;
            }
        #else  // PYPY_VERSION_NUM < 0x07030600
        #endif  // PYPY_VERSION_NUM < 0x07030600
    #else
        static CYTHON_INLINE int PyGILState_Check(void) {
            PyThreadState * tstate = _PyThreadState_Current;
            return tstate && (tstate == PyGILState_GetThisThreadState());
        }
    #endif
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000 || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX > 0x030600B4 && PY_VERSION_HEX < 0x030d0000 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif PY_MAJOR_VERSION >= 3 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000)
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  (Py_TYPE(obj)->tp_iternext)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
  #define __Pyx_PyObject_GetIterNextFunc(obj)  PyIter_Next
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyObject_GenericSetAttr((PyObject*)tp, k, v)
#else
  #define __Pyx_SetItemOnTypeDict(tp, k, v) PyDict_SetItem(tp->tp_dict, k, v)
#endif
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GetLength(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#elif PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535U : 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((int)sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = (Py_UNICODE) ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains) || (PY_MAJOR_VERSION == 2 && PYPY_VERSION_NUM < 0x07030500)
    #undef PyUnicode_Contains
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static CYTHON_INLINE PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define __Pyx_Py3Int_Check(op)       PyLong_Check(op)
  #define __Pyx_Py3Int_CheckExact(op)  PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#else
  #define __Pyx_Py3Int_Check(op)       (PyLong_Check(op) || PyInt_Check(op))
  #define __Pyx_Py3Int_CheckExact(op)  (PyLong_CheckExact(op) || PyInt_CheckExact(op))
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsHash_t
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   __Pyx_PyIndex_AsSsize_t
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #if !defined(_USE_MATH_DEFINES)
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__;  (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
    #define __PYX_EXTERN_C extern "C++"
#endif

#define __PYX_HAVE__hummingbot__strategy__avellaneda_market_making__avellaneda_market_making
#define __PYX_HAVE_API__hummingbot__strategy__avellaneda_market_making__avellaneda_market_making
/* Early includes */
#include <stdint.h>
#include "ios"
#include "new"
#include "stdexcept"
#include "typeinfo"
#include <utility>

    #if __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600)
    // move should be defined for these versions of MSVC, but __cplusplus isn't set usefully
    #include <type_traits>

    namespace cython_std {
    template <typename T> typename std::remove_reference<T>::type&& move(T& t) noexcept { return std::move(t); }
    template <typename T> typename std::remove_reference<T>::type&& move(T&& t) noexcept { return std::move(t); }
    }

    #endif
    
#include <set>
#include "../cpp/OrderBookEntry.h"
#include <vector>
#include <unordered_map>
#include <unordered_set>
#include "cpp/PyRef.h"
#include <string.h>
#include <stdio.h>

    /* Using NumPy API declarations from "numpy/__init__.cython-30.pxd" */
    
#include "numpy/arrayobject.h"
#include "numpy/ndarrayobject.h"
#include "numpy/ndarraytypes.h"
#include "numpy/arrayscalars.h"
#include "numpy/ufuncobject.h"
#include <string>
#include "../cpp/LimitOrder.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
#include <string.h>
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = (char) c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#include <string.h>
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif (defined(_Complex_I) && !defined(_MSC_VER)) || ((defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) && !defined(__STDC_NO_COMPLEX__) && !defined(_MSC_VER))
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif

/* #### Code section: filename_table ### */

static const char *__pyx_f[] = {
  "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx",
  "<stringsource>",
  "__init__.cython-30.pxd",
  "hummingbot/core/event/event_listener.pxd",
  "hummingbot/core/pubsub.pxd",
  "hummingbot/core/data_type/order_book_query_result.pxd",
  "type.pxd",
  "hummingbot/core/data_type/order_book.pxd",
  "hummingbot/strategy/__utils__/trailing_indicators/trading_intensity.pxd",
  "hummingbot/core/clock.pxd",
  "hummingbot/core/time_iterator.pxd",
  "hummingbot/core/data_type/limit_order.pxd",
  "hummingbot/strategy/order_tracker.pxd",
  "hummingbot/core/event/event_reporter.pxd",
  "hummingbot/core/event/event_logger.pxd",
  "hummingbot/core/network_iterator.pxd",
  "hummingbot/connector/connector_base.pxd",
  "hummingbot/connector/exchange_base.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* #### Code section: numeric_typedefs ### */

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":730
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":731
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":732
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":733
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":737
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":738
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":739
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":740
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":744
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":745
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":754
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":755
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":757
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":758
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":760
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":761
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":763
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":764
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":765
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;
/* #### Code section: complex_type_declarations ### */
/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);

/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener;
struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub;
struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult;
struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult;
struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook;
struct __pyx_obj_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator;
struct __pyx_obj_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradesForwarder;
struct __pyx_obj_10hummingbot_4core_5clock_Clock;
struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator;
struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder;
struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker;
struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase;
struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter;
struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger;
struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator;
struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase;
struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase;
struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy;
struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start;
struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr;
struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal;
struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr;
struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr;

/* "hummingbot/core/pubsub.pxd":10
 * from hummingbot.core.event.event_listener cimport EventListener
 * 
 * ctypedef unordered_set[PyRef] EventListenersCollection             # <<<<<<<<<<<<<<
 * ctypedef unordered_set[PyRef].iterator EventListenersIterator
 * ctypedef unordered_map[int64_t, EventListenersCollection] Events
 */
typedef std::unordered_set<PyRef>  __pyx_t_10hummingbot_4core_6pubsub_EventListenersCollection;

/* "hummingbot/core/pubsub.pxd":11
 * 
 * ctypedef unordered_set[PyRef] EventListenersCollection
 * ctypedef unordered_set[PyRef].iterator EventListenersIterator             # <<<<<<<<<<<<<<
 * ctypedef unordered_map[int64_t, EventListenersCollection] Events
 * ctypedef unordered_map[int64_t, EventListenersCollection].iterator EventsIterator
 */
typedef std::unordered_set<PyRef> ::iterator __pyx_t_10hummingbot_4core_6pubsub_EventListenersIterator;

/* "hummingbot/core/pubsub.pxd":12
 * ctypedef unordered_set[PyRef] EventListenersCollection
 * ctypedef unordered_set[PyRef].iterator EventListenersIterator
 * ctypedef unordered_map[int64_t, EventListenersCollection] Events             # <<<<<<<<<<<<<<
 * ctypedef unordered_map[int64_t, EventListenersCollection].iterator EventsIterator
 * ctypedef pair[int64_t, EventListenersCollection] EventsPair
 */
typedef std::unordered_map<int64_t,__pyx_t_10hummingbot_4core_6pubsub_EventListenersCollection>  __pyx_t_10hummingbot_4core_6pubsub_Events;

/* "hummingbot/core/pubsub.pxd":13
 * ctypedef unordered_set[PyRef].iterator EventListenersIterator
 * ctypedef unordered_map[int64_t, EventListenersCollection] Events
 * ctypedef unordered_map[int64_t, EventListenersCollection].iterator EventsIterator             # <<<<<<<<<<<<<<
 * ctypedef pair[int64_t, EventListenersCollection] EventsPair
 * 
 */
typedef std::unordered_map<int64_t,__pyx_t_10hummingbot_4core_6pubsub_EventListenersCollection> ::iterator __pyx_t_10hummingbot_4core_6pubsub_EventsIterator;

/* "hummingbot/core/pubsub.pxd":14
 * ctypedef unordered_map[int64_t, EventListenersCollection] Events
 * ctypedef unordered_map[int64_t, EventListenersCollection].iterator EventsIterator
 * ctypedef pair[int64_t, EventListenersCollection] EventsPair             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef std::pair<int64_t,__pyx_t_10hummingbot_4core_6pubsub_EventListenersCollection>  __pyx_t_10hummingbot_4core_6pubsub_EventsPair;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":767
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":768
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":769
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":771
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_opt_args_10hummingbot_8strategy_13strategy_base_12StrategyBase_c_buy_with_specific_market;
struct __pyx_opt_args_10hummingbot_8strategy_13strategy_base_12StrategyBase_c_sell_with_specific_market;

/* "hummingbot/strategy/strategy_base.pxd":57
 *     cdef c_did_complete_sell_order_tracker(self, object order_completed_event)
 * 
 *     cdef str c_buy_with_specific_market(self, object market_trading_pair_tuple, object amount, object order_type = *,             # <<<<<<<<<<<<<<
 *                                         object price = *, double expiration_seconds = *, position_action = *)
 *     cdef str c_sell_with_specific_market(self, object market_trading_pair_tuple, object amount, object order_type = *,
 */
struct __pyx_opt_args_10hummingbot_8strategy_13strategy_base_12StrategyBase_c_buy_with_specific_market {
  int __pyx_n;
  PyObject *order_type;
  PyObject *price;
  double expiration_seconds;
  PyObject *position_action;
};

/* "hummingbot/strategy/strategy_base.pxd":59
 *     cdef str c_buy_with_specific_market(self, object market_trading_pair_tuple, object amount, object order_type = *,
 *                                         object price = *, double expiration_seconds = *, position_action = *)
 *     cdef str c_sell_with_specific_market(self, object market_trading_pair_tuple, object amount, object order_type = *,             # <<<<<<<<<<<<<<
 *                                          object price = *, double expiration_seconds = *, position_action = *, )
 *     cdef c_cancel_order(self, object market_pair, str order_id)
 */
struct __pyx_opt_args_10hummingbot_8strategy_13strategy_base_12StrategyBase_c_sell_with_specific_market {
  int __pyx_n;
  PyObject *order_type;
  PyObject *price;
  double expiration_seconds;
  PyObject *position_action;
};
struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_buy;
struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_sell;
struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_quantize_order_amount;

/* "hummingbot/connector/connector_base.pxd":21
 *         public object _client_config
 * 
 *     cdef str c_buy(self, str trading_pair, object amount, object order_type=*, object price=*, dict kwargs=*)             # <<<<<<<<<<<<<<
 *     cdef str c_sell(self, str trading_pair, object amount, object order_type=*, object price=*, dict kwargs=*)
 *     cdef c_cancel(self, str trading_pair, str client_order_id)
 */
struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_buy {
  int __pyx_n;
  PyObject *order_type;
  PyObject *price;
  PyObject *kwargs;
};

/* "hummingbot/connector/connector_base.pxd":22
 * 
 *     cdef str c_buy(self, str trading_pair, object amount, object order_type=*, object price=*, dict kwargs=*)
 *     cdef str c_sell(self, str trading_pair, object amount, object order_type=*, object price=*, dict kwargs=*)             # <<<<<<<<<<<<<<
 *     cdef c_cancel(self, str trading_pair, str client_order_id)
 *     cdef c_stop_tracking_order(self, str order_id)
 */
struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_sell {
  int __pyx_n;
  PyObject *order_type;
  PyObject *price;
  PyObject *kwargs;
};

/* "hummingbot/connector/connector_base.pxd":31
 *     cdef object c_get_order_size_quantum(self, str trading_pair, object order_size)
 *     cdef object c_quantize_order_price(self, str trading_pair, object price)
 *     cdef object c_quantize_order_amount(self, str trading_pair, object amount, object price=*)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_quantize_order_amount {
  int __pyx_n;
  PyObject *price;
};
struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_buy;
struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_sell;
struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_get_fee;

/* "hummingbot/connector/exchange_base.pxd":17
 *         object _mapping_initialization_lock
 * 
 *     cdef str c_buy(self, str trading_pair, object amount, object order_type= *, object price= *, dict kwargs= *)             # <<<<<<<<<<<<<<
 *     cdef str c_sell(self, str trading_pair, object amount, object order_type= *, object price= *, dict kwargs= *)
 *     cdef c_cancel(self, str trading_pair, str client_order_id)
 */
struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_buy {
  int __pyx_n;
  PyObject *order_type;
  PyObject *price;
  PyObject *kwargs;
};

/* "hummingbot/connector/exchange_base.pxd":18
 * 
 *     cdef str c_buy(self, str trading_pair, object amount, object order_type= *, object price= *, dict kwargs= *)
 *     cdef str c_sell(self, str trading_pair, object amount, object order_type= *, object price= *, dict kwargs= *)             # <<<<<<<<<<<<<<
 *     cdef c_cancel(self, str trading_pair, str client_order_id)
 *     cdef c_stop_tracking_order(self, str order_id)
 */
struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_sell {
  int __pyx_n;
  PyObject *order_type;
  PyObject *price;
  PyObject *kwargs;
};

/* "hummingbot/connector/exchange_base.pxd":33
 *     cdef ClientOrderBookQueryResult c_get_price_for_quote_volume(self, str trading_pair, bint is_buy, double volume)
 *     cdef ClientOrderBookQueryResult c_get_price_for_volume(self, str trading_pair, bint is_buy, object volume)
 *     cdef object c_get_fee(             # <<<<<<<<<<<<<<
 *         self,
 *         str base_currency,
 */
struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_get_fee {
  int __pyx_n;
  PyObject *is_maker;
};

/* "hummingbot/core/event/event_listener.pxd":6
 * 
 * 
 * cdef class EventListener:             # <<<<<<<<<<<<<<
 *     cdef:
 *         object __weakref__
 */
struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener {
  PyObject_HEAD
  struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener *__pyx_vtab;
  PyObject *__weakref__;
  int64_t _current_event_tag;
  struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *_current_event_caller;
};


/* "hummingbot/core/pubsub.pxd":17
 * 
 * 
 * cdef class PubSub:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Events _events
 */
struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub {
  PyObject_HEAD
  struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub *__pyx_vtab;
  __pyx_t_10hummingbot_4core_6pubsub_Events _events;
  PyObject *__weakref__;
};


/* "order_book_query_result.pxd":3
 * # distutils: language=c++
 * 
 * cdef class OrderBookQueryResult:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public double query_price
 */
struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult {
  PyObject_HEAD
  double query_price;
  double query_volume;
  double result_price;
  double result_volume;
};


/* "order_book_query_result.pxd":11
 * 
 * 
 * cdef class ClientOrderBookQueryResult:             # <<<<<<<<<<<<<<
 *     cdef:
 *         public object query_price
 */
struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult {
  PyObject_HEAD
  PyObject *query_price;
  PyObject *query_volume;
  PyObject *result_price;
  PyObject *result_volume;
};


/* "hummingbot/core/data_type/order_book.pxd":12
 * 
 * 
 * cdef class OrderBook(PubSub):             # <<<<<<<<<<<<<<
 *     cdef set[OrderBookEntry] _bid_book
 *     cdef set[OrderBookEntry] _ask_book
 */
struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook {
  struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub __pyx_base;
  std::set<OrderBookEntry>  _bid_book;
  std::set<OrderBookEntry>  _ask_book;
  int64_t _snapshot_uid;
  int64_t _last_diff_uid;
  double _best_bid;
  double _best_ask;
  double _last_trade_price;
  double _last_applied_trade;
  double _last_trade_price_rest_updated;
  int _dex;
};


/* "hummingbot/strategy/__utils__/trailing_indicators/trading_intensity.pxd":11
 * from hummingbot.core.event.event_listener cimport EventListener
 * 
 * cdef class TradingIntensityIndicator:             # <<<<<<<<<<<<<<
 *     cdef:
 *         double _alpha
 */
struct __pyx_obj_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator {
  PyObject_HEAD
  struct __pyx_vtabstruct_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator *__pyx_vtab;
  double _alpha;
  double _kappa;
  PyObject *_trade_samples;
  PyObject *_current_trade_sample;
  PyObject *_trades_forwarder;
  struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *_order_book;
  PyObject *_price_delegate;
  PyObject *_last_quotes;
  int _sampling_length;
  int _samples_length;
};


/* "hummingbot/strategy/__utils__/trailing_indicators/trading_intensity.pxd":28
 *     cdef c_estimate_intensity(self)
 * 
 * cdef class TradesForwarder(EventListener):             # <<<<<<<<<<<<<<
 *     cdef:
 *         TradingIntensityIndicator _indicator
 */
struct __pyx_obj_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradesForwarder {
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener __pyx_base;
  struct __pyx_obj_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator *_indicator;
};


/* "hummingbot/core/clock.pxd":3
 * # distutils: language=c++
 * 
 * cdef class Clock:             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _clock_mode
 */
struct __pyx_obj_10hummingbot_4core_5clock_Clock {
  PyObject_HEAD
  PyObject *_clock_mode;
  double _tick_size;
  double _start_time;
  double _end_time;
  PyObject *_child_iterators;
  PyObject *_current_context;
  double _current_tick;
  int _started;
};


/* "hummingbot/core/time_iterator.pxd":7
 * 
 * 
 * cdef class TimeIterator(PubSub):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double _current_timestamp
 */
struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator {
  struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub __pyx_base;
  double _current_timestamp;
  struct __pyx_obj_10hummingbot_4core_5clock_Clock *_clock;
};


/* "hummingbot/core/data_type/limit_order.pxd":6
 * 
 * 
 * cdef class LimitOrder:             # <<<<<<<<<<<<<<
 *     cdef:
 *         CPPLimitOrder _cpp_limit_order
 */
struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder {
  PyObject_HEAD
  struct __pyx_vtabstruct_10hummingbot_4core_9data_type_11limit_order_LimitOrder *__pyx_vtab;
  LimitOrder _cpp_limit_order;
};


/* "order_tracker.pxd":7
 * 
 * 
 * cdef class OrderTracker(TimeIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         dict _tracked_limit_orders
 */
struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker {
  struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator __pyx_base;
  PyObject *_tracked_limit_orders;
  PyObject *_tracked_market_orders;
  PyObject *_order_id_to_market_pair;
  PyObject *_shadow_tracked_limit_orders;
  PyObject *_shadow_order_id_to_market_pair;
  PyObject *_shadow_gc_requests;
  PyObject *_in_flight_cancels;
  PyObject *_in_flight_pending_created;
};


/* "hummingbot/strategy/strategy_base.pxd":8
 * from .order_tracker cimport OrderTracker
 * 
 * cdef class StrategyBase(TimeIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         set _sb_markets
 */
struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase {
  struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator __pyx_base;
  PyObject *_sb_markets;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_create_buy_order_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_create_sell_order_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_fill_order_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_fail_order_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_cancel_order_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_expire_order_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_complete_buy_order_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_complete_sell_order_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_complete_funding_payment_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_position_mode_change_success_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_position_mode_change_failure_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_range_position_liquidity_added_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_range_position_liquidity_removed_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_range_position_update_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_range_position_update_failure_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_range_position_fee_collected_listener;
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *_sb_range_position_closed_listener;
  int _sb_delegate_lock;
  struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *_sb_order_tracker;
};


/* "hummingbot/core/event/event_reporter.pxd":4
 * 
 * 
 * cdef class EventReporter(EventListener):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str event_source
 */
struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter {
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener __pyx_base;
  PyObject *event_source;
};


/* "hummingbot/core/event/event_logger.pxd":4
 * 
 * 
 * cdef class EventLogger(EventListener):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str _event_source
 */
struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger {
  struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener __pyx_base;
  PyObject *_event_source;
  PyObject *_logged_events;
  PyObject *_generic_logged_events;
  PyObject *_order_filled_logged_events;
  PyObject *_waiting;
  PyObject *_wait_returns;
};


/* "hummingbot/core/network_iterator.pxd":6
 * 
 * 
 * cdef class NetworkIterator(TimeIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _network_status
 */
struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator {
  struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator __pyx_base;
  PyObject *_network_status;
  double _last_connected_timestamp;
  double _check_network_interval;
  double _check_network_timeout;
  double _network_error_wait_time;
  PyObject *_check_network_task;
};


/* "hummingbot/connector/connector_base.pxd":5
 * from hummingbot.core.network_iterator cimport NetworkIterator
 * 
 * cdef class ConnectorBase(NetworkIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         EventReporter _event_reporter
 */
struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase {
  struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator __pyx_base;
  struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter *_event_reporter;
  struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger *_event_logger;
  int _trading_required;
  PyObject *_account_available_balances;
  PyObject *_account_balances;
  int _real_time_balance_update;
  PyObject *_in_flight_orders_snapshot;
  double _in_flight_orders_snapshot_timestamp;
  PyObject *_current_trade_fills;
  PyObject *_exchange_order_ids;
  PyObject *_trade_fee_schema;
  PyObject *_trade_volume_metric_collector;
  PyObject *_client_config;
};


/* "hummingbot/connector/exchange_base.pxd":10
 * )
 * 
 * cdef class ExchangeBase(ConnectorBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _order_book_tracker
 */
struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase {
  struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase __pyx_base;
  PyObject *_order_book_tracker;
  PyObject *_budget_checker;
  PyObject *_trading_pair_symbol_map;
  PyObject *_mapping_initialization_lock;
};


/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pxd":8
 * 
 * 
 * cdef class AvellanedaMarketMakingStrategy(StrategyBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _config_map
 */
struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy {
  struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase __pyx_base;
  PyObject *_config_map;
  PyObject *_market_info;
  PyObject *_price_delegate;
  PyObject *_minimum_spread;
  int _hanging_orders_enabled;
  PyObject *_hanging_orders_cancel_pct;
  PyObject *_hanging_orders_tracker;
  int _add_transaction_costs_to_orders;
  int _hb_app_notification;
  int _is_debug;
  double _cancel_timestamp;
  double _create_timestamp;
  PyObject *_limit_order_type;
  int _all_markets_ready;
  int _filled_buys_balance;
  int _filled_sells_balance;
  double _last_timestamp;
  double _status_report_interval;
  int64_t _logging_options;
  PyObject *_last_own_trade_price;
  int _volatility_sampling_period;
  double _last_sampling_timestamp;
  int _parameters_based_on_spread;
  int _volatility_buffer_size;
  int _trading_intensity_buffer_size;
  int _ticks_to_be_ready;
  PyObject *_alpha;
  PyObject *_kappa;
  PyObject *_gamma;
  PyObject *_eta;
  PyObject *_execution_mode;
  PyObject *_execution_timeframe;
  PyObject *_execution_state;
  PyObject *_start_time;
  PyObject *_end_time;
  double _min_spread;
  PyObject *_q_adjustment_factor;
  PyObject *_reservation_price;
  PyObject *_optimal_spread;
  PyObject *_optimal_bid;
  PyObject *_optimal_ask;
  PyObject *_debug_csv_path;
  PyObject *_avg_vol;
  struct __pyx_obj_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator *_trading_intensity;
  int _should_wait_order_cancel_confirmation;
};


/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":569
 *         return self.c_cancel_order(self._market_info, order_id)
 * 
 *     cdef c_start(self, Clock clock, double timestamp):             # <<<<<<<<<<<<<<
 *         StrategyBase.c_start(self, clock, timestamp)
 *         self.update_from_config_map()
 */
struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start {
  PyObject_HEAD
  PyObject *__pyx_v_order_id;
};


/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":580
 *             restored_order_ids = self.c_track_restored_orders(self.market_info)
 *             for order_id in restored_order_ids:
 *                 order = next(o for o in self.market_info.market.limit_orders if o.client_order_id == order_id)             # <<<<<<<<<<<<<<
 *                 if order:
 *                     self._hanging_orders_tracker.add_as_hanging_order(order)
 */
struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *__pyx_outer_scope;
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_o;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1271
 *         return self.c_to_create_orders(proposal)
 * 
 *     cdef c_execute_orders_proposal(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double expiration_seconds = NaN
 */
struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal {
  PyObject_HEAD
  PyObject *__pyx_v_ask_order_id;
  PyObject *__pyx_v_bid_order_id;
};


/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1298
 *                 orders_created = True
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))             # <<<<<<<<<<<<<<
 *                     if order:
 *                         self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(
 */
struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr {
  PyObject_HEAD
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal *__pyx_outer_scope;
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_o;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1321
 *                 orders_created = True
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))             # <<<<<<<<<<<<<<
 *                     if order:
 *                         self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order
 */
struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr {
  PyObject_HEAD
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal *__pyx_outer_scope;
  PyObject *__pyx_genexpr_arg_0;
  PyObject *__pyx_v_o;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};



/* "hummingbot/core/event/event_listener.pxd":6
 * 
 * 
 * cdef class EventListener:             # <<<<<<<<<<<<<<
 *     cdef:
 *         object __weakref__
 */

struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener {
  PyObject *(*c_set_event_info)(struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *, int64_t, struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *);
  PyObject *(*c_call)(struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *, PyObject *);
};
static struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener *__pyx_vtabptr_10hummingbot_4core_5event_14event_listener_EventListener;


/* "hummingbot/core/pubsub.pxd":17
 * 
 * 
 * cdef class PubSub:             # <<<<<<<<<<<<<<
 *     cdef:
 *         Events _events
 */

struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub {
  PyObject *(*c_log_exception)(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *, int64_t, PyObject *);
  PyObject *(*c_add_listener)(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *, int64_t, struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *);
  PyObject *(*c_remove_listener)(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *, int64_t, struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener *);
  PyObject *(*c_remove_dead_listeners)(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *, int64_t);
  PyObject *(*c_get_listeners)(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *, int64_t);
  PyObject *(*c_trigger_event)(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub *, int64_t, PyObject *);
};
static struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub *__pyx_vtabptr_10hummingbot_4core_6pubsub_PubSub;


/* "hummingbot/core/data_type/order_book.pxd":12
 * 
 * 
 * cdef class OrderBook(PubSub):             # <<<<<<<<<<<<<<
 *     cdef set[OrderBookEntry] _bid_book
 *     cdef set[OrderBookEntry] _ask_book
 */

struct __pyx_vtabstruct_10hummingbot_4core_9data_type_10order_book_OrderBook {
  struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub __pyx_base;
  PyObject *(*c_apply_diffs)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, std::vector<OrderBookEntry> , std::vector<OrderBookEntry> , int64_t);
  PyObject *(*c_apply_snapshot)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, std::vector<OrderBookEntry> , std::vector<OrderBookEntry> , int64_t);
  PyObject *(*c_apply_trade)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, PyObject *);
  PyObject *(*c_apply_numpy_diffs)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, PyArrayObject *, PyArrayObject *);
  PyObject *(*c_apply_numpy_snapshot)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, PyArrayObject *, PyArrayObject *);
  double (*c_get_price)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult *(*c_get_price_for_volume)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int, double);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult *(*c_get_price_for_quote_volume)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int, double);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult *(*c_get_volume_for_price)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int, double);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult *(*c_get_quote_volume_for_price)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int, double);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult *(*c_get_vwap_for_volume)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int, double);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult *(*c_get_quote_volume_for_base_amount)(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *, int, double);
};
static struct __pyx_vtabstruct_10hummingbot_4core_9data_type_10order_book_OrderBook *__pyx_vtabptr_10hummingbot_4core_9data_type_10order_book_OrderBook;


/* "hummingbot/strategy/__utils__/trailing_indicators/trading_intensity.pxd":11
 * from hummingbot.core.event.event_listener cimport EventListener
 * 
 * cdef class TradingIntensityIndicator:             # <<<<<<<<<<<<<<
 *     cdef:
 *         double _alpha
 */

struct __pyx_vtabstruct_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator {
  PyObject *(*c_calculate)(struct __pyx_obj_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator *, PyObject *);
  PyObject *(*c_register_trade)(struct __pyx_obj_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator *, PyObject *);
  PyObject *(*c_estimate_intensity)(struct __pyx_obj_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator *);
};
static struct __pyx_vtabstruct_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator *__pyx_vtabptr_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator;


/* "hummingbot/strategy/__utils__/trailing_indicators/trading_intensity.pxd":28
 *     cdef c_estimate_intensity(self)
 * 
 * cdef class TradesForwarder(EventListener):             # <<<<<<<<<<<<<<
 *     cdef:
 *         TradingIntensityIndicator _indicator
 */

struct __pyx_vtabstruct_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradesForwarder {
  struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener __pyx_base;
};
static struct __pyx_vtabstruct_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradesForwarder *__pyx_vtabptr_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradesForwarder;


/* "hummingbot/core/time_iterator.pxd":7
 * 
 * 
 * cdef class TimeIterator(PubSub):             # <<<<<<<<<<<<<<
 *     cdef:
 *         double _current_timestamp
 */

struct __pyx_vtabstruct_10hummingbot_4core_13time_iterator_TimeIterator {
  struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub __pyx_base;
  PyObject *(*c_start)(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *, struct __pyx_obj_10hummingbot_4core_5clock_Clock *, double);
  PyObject *(*c_stop)(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *, struct __pyx_obj_10hummingbot_4core_5clock_Clock *);
  PyObject *(*c_tick)(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *, double);
};
static struct __pyx_vtabstruct_10hummingbot_4core_13time_iterator_TimeIterator *__pyx_vtabptr_10hummingbot_4core_13time_iterator_TimeIterator;


/* "hummingbot/core/data_type/limit_order.pxd":6
 * 
 * 
 * cdef class LimitOrder:             # <<<<<<<<<<<<<<
 *     cdef:
 *         CPPLimitOrder _cpp_limit_order
 */

struct __pyx_vtabstruct_10hummingbot_4core_9data_type_11limit_order_LimitOrder {
  PY_LONG_LONG (*c_age)(struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *);
  PY_LONG_LONG (*c_age_til)(struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *, PY_LONG_LONG);
};
static struct __pyx_vtabstruct_10hummingbot_4core_9data_type_11limit_order_LimitOrder *__pyx_vtabptr_10hummingbot_4core_9data_type_11limit_order_LimitOrder;


/* "order_tracker.pxd":7
 * 
 * 
 * cdef class OrderTracker(TimeIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         dict _tracked_limit_orders
 */

struct __pyx_vtabstruct_10hummingbot_8strategy_13order_tracker_OrderTracker {
  struct __pyx_vtabstruct_10hummingbot_4core_13time_iterator_TimeIterator __pyx_base;
  PyObject *(*c_get_limit_orders)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *);
  PyObject *(*c_get_market_orders)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *);
  PyObject *(*c_get_shadow_limit_orders)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *);
  int (*c_has_in_flight_cancel)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *);
  int (*c_check_and_track_cancel)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *);
  PyObject *(*c_get_market_pair_from_order_id)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *);
  PyObject *(*c_get_shadow_market_pair_from_order_id)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *(*c_get_limit_order)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *, PyObject *);
  PyObject *(*c_get_market_order)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *(*c_get_shadow_limit_order)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *);
  PyObject *(*c_start_tracking_limit_order)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *, PyObject *, int, PyObject *, PyObject *);
  PyObject *(*c_stop_tracking_limit_order)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *, PyObject *);
  PyObject *(*c_start_tracking_market_order)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *, PyObject *, int, PyObject *);
  PyObject *(*c_stop_tracking_market_order)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *, PyObject *);
  PyObject *(*c_check_and_cleanup_shadow_records)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *);
  PyObject *(*c_add_create_order_pending)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *);
  PyObject *(*c_remove_create_order_pending)(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *, PyObject *);
};
static struct __pyx_vtabstruct_10hummingbot_8strategy_13order_tracker_OrderTracker *__pyx_vtabptr_10hummingbot_8strategy_13order_tracker_OrderTracker;


/* "hummingbot/strategy/strategy_base.pxd":8
 * from .order_tracker cimport OrderTracker
 * 
 * cdef class StrategyBase(TimeIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         set _sb_markets
 */

struct __pyx_vtabstruct_10hummingbot_8strategy_13strategy_base_StrategyBase {
  struct __pyx_vtabstruct_10hummingbot_4core_13time_iterator_TimeIterator __pyx_base;
  PyObject *(*c_add_markets)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_remove_markets)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_create_buy_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_create_sell_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_fill_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_fail_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_cancel_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_expire_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_complete_buy_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_complete_sell_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_complete_funding_payment)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_change_position_mode_succeed)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_change_position_mode_fail)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_add_liquidity)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_remove_liquidity)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_update_lp_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_fail_lp_update)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_collect_fee)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_close_position)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_fail_order_tracker)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_cancel_order_tracker)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_expire_order_tracker)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_complete_buy_order_tracker)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_did_complete_sell_order_tracker)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_buy_with_specific_market)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_8strategy_13strategy_base_12StrategyBase_c_buy_with_specific_market *__pyx_optional_args);
  PyObject *(*c_sell_with_specific_market)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_8strategy_13strategy_base_12StrategyBase_c_sell_with_specific_market *__pyx_optional_args);
  PyObject *(*c_cancel_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *, PyObject *);
  PyObject *(*c_start_tracking_limit_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *, PyObject *, int, PyObject *, PyObject *);
  PyObject *(*c_stop_tracking_limit_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *, PyObject *);
  PyObject *(*c_start_tracking_market_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *, PyObject *, int, PyObject *);
  PyObject *(*c_stop_tracking_market_order)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *, PyObject *);
  PyObject *(*c_track_restored_orders)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *);
  PyObject *(*c_sum_flat_fees)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *, PyObject *);
};
static struct __pyx_vtabstruct_10hummingbot_8strategy_13strategy_base_StrategyBase *__pyx_vtabptr_10hummingbot_8strategy_13strategy_base_StrategyBase;


/* "hummingbot/core/event/event_reporter.pxd":4
 * 
 * 
 * cdef class EventReporter(EventListener):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str event_source
 */

struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_reporter_EventReporter {
  struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener __pyx_base;
};
static struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_reporter_EventReporter *__pyx_vtabptr_10hummingbot_4core_5event_14event_reporter_EventReporter;


/* "hummingbot/core/event/event_logger.pxd":4
 * 
 * 
 * cdef class EventLogger(EventListener):             # <<<<<<<<<<<<<<
 *     cdef:
 *         str _event_source
 */

struct __pyx_vtabstruct_10hummingbot_4core_5event_12event_logger_EventLogger {
  struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener __pyx_base;
};
static struct __pyx_vtabstruct_10hummingbot_4core_5event_12event_logger_EventLogger *__pyx_vtabptr_10hummingbot_4core_5event_12event_logger_EventLogger;


/* "hummingbot/core/network_iterator.pxd":6
 * 
 * 
 * cdef class NetworkIterator(TimeIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _network_status
 */

struct __pyx_vtabstruct_10hummingbot_4core_16network_iterator_NetworkIterator {
  struct __pyx_vtabstruct_10hummingbot_4core_13time_iterator_TimeIterator __pyx_base;
};
static struct __pyx_vtabstruct_10hummingbot_4core_16network_iterator_NetworkIterator *__pyx_vtabptr_10hummingbot_4core_16network_iterator_NetworkIterator;


/* "hummingbot/connector/connector_base.pxd":5
 * from hummingbot.core.network_iterator cimport NetworkIterator
 * 
 * cdef class ConnectorBase(NetworkIterator):             # <<<<<<<<<<<<<<
 *     cdef:
 *         EventReporter _event_reporter
 */

struct __pyx_vtabstruct_10hummingbot_9connector_14connector_base_ConnectorBase {
  struct __pyx_vtabstruct_10hummingbot_4core_16network_iterator_NetworkIterator __pyx_base;
  PyObject *(*c_buy)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_buy *__pyx_optional_args);
  PyObject *(*c_sell)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_sell *__pyx_optional_args);
  PyObject *(*c_cancel)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *);
  PyObject *(*c_stop_tracking_order)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *);
  PyObject *(*c_get_balance)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *);
  PyObject *(*c_get_available_balance)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *);
  PyObject *(*c_get_price)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, int);
  PyObject *(*c_get_order_price_quantum)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *);
  PyObject *(*c_get_order_size_quantum)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *);
  PyObject *(*c_quantize_order_price)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *);
  PyObject *(*c_quantize_order_amount)(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_9connector_14connector_base_13ConnectorBase_c_quantize_order_amount *__pyx_optional_args);
};
static struct __pyx_vtabstruct_10hummingbot_9connector_14connector_base_ConnectorBase *__pyx_vtabptr_10hummingbot_9connector_14connector_base_ConnectorBase;


/* "hummingbot/connector/exchange_base.pxd":10
 * )
 * 
 * cdef class ExchangeBase(ConnectorBase):             # <<<<<<<<<<<<<<
 *     cdef:
 *         object _order_book_tracker
 */

struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase {
  struct __pyx_vtabstruct_10hummingbot_9connector_14connector_base_ConnectorBase __pyx_base;
  struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook *(*c_get_order_book)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult *(*c_get_quote_volume_for_base_amount)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, int, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult *(*c_get_volume_for_price)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, int, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult *(*c_get_quote_volume_for_price)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, int, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult *(*c_get_vwap_for_volume)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, int, PyObject *);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult *(*c_get_price_for_quote_volume)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, int, double);
  struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult *(*c_get_price_for_volume)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, int, PyObject *);
  PyObject *(*c_get_fee)(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, PyObject *, struct __pyx_opt_args_10hummingbot_9connector_13exchange_base_12ExchangeBase_c_get_fee *__pyx_optional_args);
};
static struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_vtabptr_10hummingbot_9connector_13exchange_base_ExchangeBase;


/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":60
 * 
 * 
 * cdef class AvellanedaMarketMakingStrategy(StrategyBase):             # <<<<<<<<<<<<<<
 *     OPTION_LOG_CREATE_ORDER = 1 << 3
 *     OPTION_LOG_MAKER_ORDER_FILLED = 1 << 4
 */

struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy {
  struct __pyx_vtabstruct_10hummingbot_8strategy_13strategy_base_StrategyBase __pyx_base;
  PyObject *(*c_get_mid_price)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *);
  PyObject *(*_create_proposal_based_on_order_levels)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *);
  PyObject *(*_create_proposal_based_on_order_override)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *);
  PyObject *(*_create_basic_proposal)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *);
  PyObject *(*c_create_base_proposal)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *);
  PyObject *(*c_get_adjusted_available_balance)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *);
  PyObject *(*c_apply_order_price_modifiers)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *);
  PyObject *(*c_apply_order_amount_eta_transformation)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *);
  PyObject *(*c_apply_budget_constraint)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *);
  PyObject *(*c_apply_order_optimization)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *);
  PyObject *(*c_apply_add_transaction_costs)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *);
  int (*c_is_within_tolerance)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *, PyObject *);
  PyObject *(*c_cancel_active_orders)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *);
  PyObject *(*c_cancel_active_orders_on_max_age_limit)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *);
  int (*c_to_create_orders)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *);
  PyObject *(*c_execute_orders_proposal)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *);
  PyObject *(*c_set_timers)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *);
  double (*c_get_spread)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *);
  PyObject *(*c_collect_market_variables)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, double);
  int (*c_is_algorithm_ready)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *);
  int (*c_is_algorithm_changed)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *);
  PyObject *(*c_measure_order_book_liquidity)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *);
  PyObject *(*c_calculate_reservation_price_and_optimal_spread)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *);
  PyObject *(*c_calculate_target_inventory)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *);
  PyObject *(*c_calculate_inventory)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *);
  PyObject *(*c_did_complete_order)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *);
};
static struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_vtabptr_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__))
  #define __Pyx_RefNannyFinishContextNogil() __Pyx_RefNannyFinishContext()
#endif
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_VARARGS(args, i) PySequence_GetItem(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GET_ITEM(args, i)
#else
    #define __Pyx_Arg_VARARGS(args, i) PyTuple_GetItem(args, i)
#endif
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_Arg_NewRef_VARARGS(arg) __Pyx_NewRef(arg)
    #define __Pyx_Arg_XDECREF_VARARGS(arg) Py_XDECREF(arg)
#else
    #define __Pyx_Arg_NewRef_VARARGS(arg) arg
    #define __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_Arg_FASTCALL(args, i) args[i]
    #define __Pyx_NumKwargs_FASTCALL(kwds) PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
    #define __Pyx_Arg_NewRef_FASTCALL(arg) arg  /* no-op, __Pyx_Arg_FASTCALL is direct and this needs
                                                   to have the same reference counting */
    #define __Pyx_Arg_XDECREF_FASTCALL(arg)
#else
    #define __Pyx_Arg_FASTCALL __Pyx_Arg_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
    #define __Pyx_Arg_NewRef_FASTCALL(arg) __Pyx_Arg_NewRef_VARARGS(arg)
    #define __Pyx_Arg_XDECREF_FASTCALL(arg) __Pyx_Arg_XDECREF_VARARGS(arg)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_VARARGS(args, start), stop - start)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(&__Pyx_Arg_FASTCALL(args, start), stop - start)
#else
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* KeywordStringCheck.proto */
static int __Pyx_CheckKeywordStrings(PyObject *kw, const char* function_name, int kw_allowed);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,
    const char* function_name);

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* set_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set);
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set);

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* pybytes_as_double.proto */
static double __Pyx_SlowPyString_AsDouble(PyObject *obj);
static double __Pyx__PyBytes_AsDouble(PyObject *obj, const char* start, Py_ssize_t length);
static CYTHON_INLINE double __Pyx_PyBytes_AsDouble(PyObject *obj) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
    as_c_string = PyBytes_AS_STRING(obj);
    size = PyBytes_GET_SIZE(obj);
#else
    if (PyBytes_AsStringAndSize(obj, &as_c_string, &size) < 0) {
        return (double)-1;
    }
#endif
    return __Pyx__PyBytes_AsDouble(obj, as_c_string, size);
}
static CYTHON_INLINE double __Pyx_PyByteArray_AsDouble(PyObject *obj) {
    char* as_c_string;
    Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
    as_c_string = PyByteArray_AS_STRING(obj);
    size = PyByteArray_GET_SIZE(obj);
#else
    as_c_string = PyByteArray_AsString(obj);
    if (as_c_string == NULL) {
        return (double)-1;
    }
    size = PyByteArray_Size(obj);
#endif
    return __Pyx__PyBytes_AsDouble(obj, as_c_string, size);
}

/* pyunicode_as_double.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY && CYTHON_ASSUME_SAFE_MACROS
static const char* __Pyx__PyUnicode_AsDouble_Copy(const void* data, const int kind, char* buffer, Py_ssize_t start, Py_ssize_t end) {
    int last_was_punctuation;
    Py_ssize_t i;
    last_was_punctuation = 1;
    for (i=start; i <= end; i++) {
        Py_UCS4 chr = PyUnicode_READ(kind, data, i);
        int is_punctuation = (chr == '_') | (chr == '.');
        *buffer = (char)chr;
        buffer += (chr != '_');
        if (unlikely(chr > 127)) goto parse_failure;
        if (unlikely(last_was_punctuation & is_punctuation)) goto parse_failure;
        last_was_punctuation = is_punctuation;
    }
    if (unlikely(last_was_punctuation)) goto parse_failure;
    *buffer = '\0';
    return buffer;
parse_failure:
    return NULL;
}
static double __Pyx__PyUnicode_AsDouble_inf_nan(const void* data, int kind, Py_ssize_t start, Py_ssize_t length) {
    int matches = 1;
    Py_UCS4 chr;
    Py_UCS4 sign = PyUnicode_READ(kind, data, start);
    int is_signed = (sign == '-') | (sign == '+');
    start += is_signed;
    length -= is_signed;
    switch (PyUnicode_READ(kind, data, start)) {
        #ifdef Py_NAN
        case 'n':
        case 'N':
            if (unlikely(length != 3)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+1);
            matches &= (chr == 'a') | (chr == 'A');
            chr = PyUnicode_READ(kind, data, start+2);
            matches &= (chr == 'n') | (chr == 'N');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_NAN : Py_NAN;
        #endif
        case 'i':
        case 'I':
            if (unlikely(length < 3)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+1);
            matches &= (chr == 'n') | (chr == 'N');
            chr = PyUnicode_READ(kind, data, start+2);
            matches &= (chr == 'f') | (chr == 'F');
            if (likely(length == 3 && matches))
                return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
            if (unlikely(length != 8)) goto parse_failure;
            chr = PyUnicode_READ(kind, data, start+3);
            matches &= (chr == 'i') | (chr == 'I');
            chr = PyUnicode_READ(kind, data, start+4);
            matches &= (chr == 'n') | (chr == 'N');
            chr = PyUnicode_READ(kind, data, start+5);
            matches &= (chr == 'i') | (chr == 'I');
            chr = PyUnicode_READ(kind, data, start+6);
            matches &= (chr == 't') | (chr == 'T');
            chr = PyUnicode_READ(kind, data, start+7);
            matches &= (chr == 'y') | (chr == 'Y');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
        case '.': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
            break;
        default:
            goto parse_failure;
    }
    return 0.0;
parse_failure:
    return -1.0;
}
static double __Pyx_PyUnicode_AsDouble_WithSpaces(PyObject *obj) {
    double value;
    const char *last;
    char *end;
    Py_ssize_t start, length = PyUnicode_GET_LENGTH(obj);
    const int kind = PyUnicode_KIND(obj);
    const void* data = PyUnicode_DATA(obj);
    start = 0;
    while (Py_UNICODE_ISSPACE(PyUnicode_READ(kind, data, start)))
        start++;
    while (start < length - 1 && Py_UNICODE_ISSPACE(PyUnicode_READ(kind, data, length - 1)))
        length--;
    length -= start;
    if (unlikely(length <= 0)) goto fallback;
    value = __Pyx__PyUnicode_AsDouble_inf_nan(data, kind, start, length);
    if (unlikely(value == -1.0)) goto fallback;
    if (value != 0.0) return value;
    if (length < 40) {
        char number[40];
        last = __Pyx__PyUnicode_AsDouble_Copy(data, kind, number, start, start + length);
        if (unlikely(!last)) goto fallback;
        value = PyOS_string_to_double(number, &end, NULL);
    } else {
        char *number = (char*) PyMem_Malloc((length + 1) * sizeof(char));
        if (unlikely(!number)) goto fallback;
        last = __Pyx__PyUnicode_AsDouble_Copy(data, kind, number, start, start + length);
        if (unlikely(!last)) {
            PyMem_Free(number);
            goto fallback;
        }
        value = PyOS_string_to_double(number, &end, NULL);
        PyMem_Free(number);
    }
    if (likely(end == last) || (value == (double)-1 && PyErr_Occurred())) {
        return value;
    }
fallback:
    return __Pyx_SlowPyString_AsDouble(obj);
}
#endif
static CYTHON_INLINE double __Pyx_PyUnicode_AsDouble(PyObject *obj) {
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY && CYTHON_ASSUME_SAFE_MACROS
    if (unlikely(__Pyx_PyUnicode_READY(obj) == -1))
        return (double)-1;
    if (likely(PyUnicode_IS_ASCII(obj))) {
        const char *s;
        Py_ssize_t length;
        s = PyUnicode_AsUTF8AndSize(obj, &length);
        return __Pyx__PyBytes_AsDouble(obj, s, length);
    }
    return __Pyx_PyUnicode_AsDouble_WithSpaces(obj);
#else
    return __Pyx_SlowPyString_AsDouble(obj);
#endif
}

/* pyobject_as_double.proto */
static double __Pyx__PyObject_AsDouble(PyObject* obj);
#if CYTHON_COMPILING_IN_PYPY
#define __Pyx_PyObject_AsDouble(obj)\
(likely(PyFloat_CheckExact(obj)) ? PyFloat_AS_DOUBLE(obj) :\
 likely(PyInt_CheckExact(obj)) ?\
 PyFloat_AsDouble(obj) : __Pyx__PyObject_AsDouble(obj))
#else
#define __Pyx_PyObject_AsDouble(obj)\
((likely(PyFloat_CheckExact(obj))) ?  PyFloat_AS_DOUBLE(obj) :\
 likely(PyLong_CheckExact(obj)) ?\
 PyLong_AsDouble(obj) : __Pyx__PyObject_AsDouble(obj))
#endif

/* PyObjectFormatSimple.proto */
#if CYTHON_COMPILING_IN_PYPY
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#elif PY_MAJOR_VERSION < 3
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyString_CheckExact(s)) ? PyUnicode_FromEncodedObject(s, NULL, "strict") :\
        PyObject_Format(s, f))
#elif CYTHON_USE_TYPE_SLOTS
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        likely(PyLong_CheckExact(s)) ? PyLong_Type.tp_repr(s) :\
        likely(PyFloat_CheckExact(s)) ? PyFloat_Type.tp_repr(s) :\
        PyObject_Format(s, f))
#else
    #define __Pyx_PyObject_FormatSimple(s, f) (\
        likely(PyUnicode_CheckExact(s)) ? (Py_INCREF(s), s) :\
        PyObject_Format(s, f))
#endif

/* JoinPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectFormat.proto */
#if CYTHON_USE_UNICODE_WRITER
static PyObject* __Pyx_PyObject_Format(PyObject* s, PyObject* f);
#else
#define __Pyx_PyObject_Format(s, f) PyObject_Format(s, f)
#endif

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* FixUpExtensionType.proto */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);
#endif

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* FetchCommonType.proto */
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);
#else
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases);
#endif

/* PyMethodNew.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *typesModule=NULL, *methodType=NULL, *result=NULL;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    typesModule = PyImport_ImportModule("types");
    if (!typesModule) return NULL;
    methodType = PyObject_GetAttrString(typesModule, "MethodType");
    Py_DECREF(typesModule);
    if (!methodType) return NULL;
    result = PyObject_CallFunctionObjArgs(methodType, func, self, NULL);
    Py_DECREF(methodType);
    return result;
}
#elif PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#else
    #define __Pyx_PyMethod_New PyMethod_New
#endif

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    void *defaults;
    int defaults_pyobjects;
    size_t defaults_size;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc);
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyIntCompare.proto */
static CYTHON_INLINE int __Pyx_PyInt_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* py_abs.proto */
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject *__Pyx_PyLong_AbsNeg(PyObject *num);
#define __Pyx_PyNumber_Absolute(x)\
    ((likely(PyLong_CheckExact(x))) ?\
         (likely(__Pyx_PyLong_IsNonNeg(x)) ? (Py_INCREF(x), (x)) : __Pyx_PyLong_AbsNeg(x)) :\
         PyNumber_Absolute(x))
#else
#define __Pyx_PyNumber_Absolute(x)  PyNumber_Absolute(x)
#endif

/* RaiseUnboundLocalError.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* pynumber_float.proto */
static CYTHON_INLINE PyObject* __Pyx__PyNumber_Float(PyObject* obj);
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : __Pyx__PyNumber_Float(x))

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* PyObject_Unicode.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Str(obj))
#else
#define __Pyx_PyObject_Unicode(obj)\
    (likely(PyUnicode_CheckExact(obj)) ? __Pyx_NewRef(obj) : PyObject_Unicode(obj))
#endif

/* UnicodeConcatInPlace.proto */
# if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
    #if CYTHON_REFNANNY
        #define __Pyx_PyUnicode_ConcatInPlace(left, right) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right, __pyx_refnanny)
    #else
        #define __Pyx_PyUnicode_ConcatInPlace(left, right) __Pyx_PyUnicode_ConcatInPlaceImpl(&left, right)
    #endif
    static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    );
#else
#define __Pyx_PyUnicode_ConcatInPlace __Pyx_PyUnicode_Concat
#endif
#define __Pyx_PyUnicode_ConcatInPlaceSafe(left, right) ((unlikely((left) == Py_None) || unlikely((right) == Py_None)) ?\
    PyNumber_InPlaceAdd(left, right) : __Pyx_PyUnicode_ConcatInPlace(left, right))

/* RaiseClosureNameError.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* pep479.proto */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen);

/* IterNext.proto */
#define __Pyx_PyIter_Next(obj) __Pyx_PyIter_Next2(obj, NULL)
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject *, PyObject *);

/* ModInt[long].proto */
static CYTHON_INLINE long __Pyx_mod_long(long, long);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* BuildPyUnicode.proto */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char);

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char);

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyIntCompare.proto */
static CYTHON_INLINE int __Pyx_PyInt_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_MultiplyCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_MultiplyCObj(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceMultiply(op1, op2) : PyNumber_Multiply(op1, op2))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddCObj(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_TrueDivideObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_TrueDivideObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceTrueDivide(op1, op2) : PyNumber_TrueDivide(op1, op2))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_MultiplyObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_MultiplyObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceMultiply(op1, op2) : PyNumber_Multiply(op1, op2))
#endif

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* PyUnicode_Unicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;

/* CallUnboundCMethod0.proto */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CallUnboundCMethod0(cfunc, self)\
    (likely((cfunc)->func) ?\
        (likely((cfunc)->flag == METH_NOARGS) ?  (*((cfunc)->func))(self, NULL) :\
         (PY_VERSION_HEX >= 0x030600B1 && likely((cfunc)->flag == METH_FASTCALL) ?\
            (PY_VERSION_HEX >= 0x030700A0 ?\
                (*(__Pyx_PyCFunctionFast)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0) :\
                (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL)) :\
          (PY_VERSION_HEX >= 0x030700A0 && (cfunc)->flag == (METH_FASTCALL | METH_KEYWORDS) ?\
            (*(__Pyx_PyCFunctionFastWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, &__pyx_empty_tuple, 0, NULL) :\
            (likely((cfunc)->flag == (METH_VARARGS | METH_KEYWORDS)) ?  ((*(PyCFunctionWithKeywords)(void*)(PyCFunction)(cfunc)->func)(self, __pyx_empty_tuple, NULL)) :\
               ((cfunc)->flag == METH_VARARGS ?  (*((cfunc)->func))(self, __pyx_empty_tuple) :\
               __Pyx__CallUnboundCMethod0(cfunc, self)))))) :\
        __Pyx__CallUnboundCMethod0(cfunc, self))
#else
#define __Pyx_CallUnboundCMethod0(cfunc, self)  __Pyx__CallUnboundCMethod0(cfunc, self)
#endif

/* CIntToPyUnicode.proto */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char);

/* CallNextTpDealloc.proto */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc);

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_0_12
#define __PYX_HAVE_RT_ImportType_proto_3_0_12
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_12(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_0_12(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_0_12 {
   __Pyx_ImportType_CheckSize_Error_3_0_12 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_0_12 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_0_12 = 2
};
static PyTypeObject *__Pyx_ImportType_3_0_12(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_12 check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* MergeVTables.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type);
#endif

/* SetupReduce.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce(PyObject* type_obj);
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple);
#endif

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* ClassMethod.proto */
#include "descrobject.h"
CYTHON_UNUSED static PyObject* __Pyx_Method_ClassMethod(PyObject *method);

/* GetNameInClass.proto */
#define __Pyx_GetNameInClass(var, nmspace, name)  (var) = __Pyx__GetNameInClass(nmspace, name)
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
#if !CYTHON_COMPILING_IN_LIMITED_API
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);
#endif

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* None.proto */
#include <new>

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
static __Pyx_TypeName __Pyx_PyType_GetName(PyTypeObject* tp);
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#else
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
struct __pyx_CoroutineObject;
typedef PyObject *(*__pyx_coroutine_body_t)(struct __pyx_CoroutineObject *, PyThreadState *, PyObject *);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_ExcInfoStruct  _PyErr_StackItem
#else
typedef struct {
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
} __Pyx_ExcInfoStruct;
#endif
typedef struct __pyx_CoroutineObject {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    __Pyx_ExcInfoStruct gi_exc_state;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    PyObject *gi_code;
    PyObject *gi_frame;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name);
static CYTHON_INLINE void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *self);
static int __Pyx_Coroutine_clear(PyObject *self);
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#if CYTHON_USE_EXC_INFO_STACK
#define __Pyx_Coroutine_SwapException(self)
#define __Pyx_Coroutine_ResetAndClearException(self)  __Pyx_Coroutine_ExceptionClear(&(self)->gi_exc_state)
#else
#define __Pyx_Coroutine_SwapException(self) {\
    __Pyx_ExceptionSwap(&(self)->gi_exc_state.exc_type, &(self)->gi_exc_state.exc_value, &(self)->gi_exc_state.exc_traceback);\
    __Pyx_Coroutine_ResetFrameBackpointer(&(self)->gi_exc_state);\
    }
#define __Pyx_Coroutine_ResetAndClearException(self) {\
    __Pyx_ExceptionReset((self)->gi_exc_state.exc_type, (self)->gi_exc_state.exc_value, (self)->gi_exc_state.exc_traceback);\
    (self)->gi_exc_state.exc_type = (self)->gi_exc_state.exc_value = (self)->gi_exc_state.exc_traceback = NULL;\
    }
#endif
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__pyx_tstate, pvalue)
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue)\
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, pvalue)
#endif
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *tstate, PyObject **pvalue);
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state);

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
#define __Pyx_Generator_CheckExact(obj) __Pyx_IS_TYPE(obj, __pyx_GeneratorType)
#define __Pyx_Generator_New(body, code, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, code, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(PyObject *module);

/* CheckBinaryVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

/* #### Code section: module_declarations ### */
static CYTHON_INLINE PyObject *__pyx_f_5numpy_7ndarray_4base_base(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE PyArray_Descr *__pyx_f_5numpy_7ndarray_5descr_descr(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE int __pyx_f_5numpy_7ndarray_4ndim_ndim(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_5shape_shape(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_7strides_strides(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE npy_intp __pyx_f_5numpy_7ndarray_4size_size(PyArrayObject *__pyx_v_self); /* proto*/
static CYTHON_INLINE char *__pyx_f_5numpy_7ndarray_4data_data(PyArrayObject *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_get_mid_price(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_start(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, struct __pyx_obj_10hummingbot_4core_5clock_Clock *__pyx_v_clock, double __pyx_v_timestamp); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_stop(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, struct __pyx_obj_10hummingbot_4core_5clock_Clock *__pyx_v_clock); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_tick(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, double __pyx_v_timestamp); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_collect_market_variables(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, double __pyx_v_timestamp); /* proto*/
static double __pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_get_spread(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_measure_order_book_liquidity(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_calculate_reservation_price_and_optimal_spread(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_calculate_target_inventory(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_calculate_inventory(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto*/
static int __pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_is_algorithm_ready(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto*/
static int __pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_is_algorithm_changed(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy__create_proposal_based_on_order_override(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy__create_proposal_based_on_order_levels(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy__create_basic_proposal(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_create_base_proposal(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_get_adjusted_available_balance(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_orders); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_apply_order_price_modifiers(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_apply_budget_constraint(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_apply_order_optimization(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_apply_order_amount_eta_transformation(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_apply_add_transaction_costs(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_did_fill_order(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_order_filled_event); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_did_complete_buy_order(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_order_completed_event); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_did_complete_sell_order(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_order_completed_event); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_did_complete_order(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_order_completed_event); /* proto*/
static int __pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_is_within_tolerance(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_current_prices, PyObject *__pyx_v_proposal_prices); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_cancel_active_orders_on_max_age_limit(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_cancel_active_orders(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static int __pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_to_create_orders(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_execute_orders_proposal(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto*/
static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_set_timers(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto*/

/* Module declarations from "libc.stdint" */

/* Module declarations from "libcpp.utility" */

/* Module declarations from "libcpp.set" */

/* Module declarations from "hummingbot.core.data_type.OrderBookEntry" */

/* Module declarations from "libcpp.vector" */

/* Module declarations from "libcpp.unordered_map" */

/* Module declarations from "libcpp.unordered_set" */

/* Module declarations from "hummingbot.core.PyRef" */

/* Module declarations from "hummingbot.core.event.event_listener" */

/* Module declarations from "hummingbot.core.pubsub" */

/* Module declarations from "hummingbot.core.data_type.order_book_query_result" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdio" */

/* Module declarations from "__builtin__" */

/* Module declarations from "cpython.type" */

/* Module declarations from "cpython" */

/* Module declarations from "cpython.object" */

/* Module declarations from "cpython.ref" */

/* Module declarations from "numpy" */

/* Module declarations from "numpy" */

/* Module declarations from "hummingbot.core.data_type.order_book" */

/* Module declarations from "hummingbot.strategy.__utils__.trailing_indicators.trading_intensity" */

/* Module declarations from "hummingbot.core.clock" */

/* Module declarations from "hummingbot.core.time_iterator" */

/* Module declarations from "libcpp" */

/* Module declarations from "libcpp.string" */

/* Module declarations from "hummingbot.core.data_type.LimitOrder" */

/* Module declarations from "hummingbot.core.data_type.limit_order" */

/* Module declarations from "hummingbot.strategy.order_tracker" */

/* Module declarations from "hummingbot.strategy.strategy_base" */

/* Module declarations from "hummingbot.core.event.event_reporter" */

/* Module declarations from "hummingbot.core.event.event_logger" */

/* Module declarations from "hummingbot.core.network_iterator" */

/* Module declarations from "hummingbot.connector.connector_base" */

/* Module declarations from "hummingbot.connector.exchange_base" */

/* Module declarations from "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making" */
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making"
extern int __pyx_module_is_main_hummingbot__strategy__avellaneda_market_making__avellaneda_market_making;
int __pyx_module_is_main_hummingbot__strategy__avellaneda_market_making__avellaneda_market_making = 0;

/* Implementation of "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_all;
static PyObject *__pyx_builtin_AttributeError;
static PyObject *__pyx_builtin_round;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_max;
static PyObject *__pyx_builtin_sorted;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_super;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_ImportError;
/* #### Code section: string_decls ### */
static const char __pyx_k_0[] = "0";
static const char __pyx_k_1[] = ".1%";
static const char __pyx_k_2[] = ".2%";
static const char __pyx_k_a[] = "a";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_o[] = "o";
static const char __pyx_k_p[] = "p";
static const char __pyx_k_q[] = "q=";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_x[] = "x";
static const char __pyx_k_10[] = "10";
static const char __pyx_k_3f[] = ".3f";
static const char __pyx_k_4f[] = ".4f";
static const char __pyx_k_5E[] = ".5E";
static const char __pyx_k__4[] = "";
static const char __pyx_k__5[] = ")";
static const char __pyx_k__6[] = "\n";
static const char __pyx_k__7[] = "    ";
static const char __pyx_k__8[] = "{:<";
static const char __pyx_k__9[] = "}";
static const char __pyx_k_df[] = "df";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_ln[] = "ln";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_pd[] = "pd";
static const char __pyx_k_100[] = "100";
static const char __pyx_k_10f[] = ".10f";
static const char __pyx_k_1_2[] = "1";
static const char __pyx_k_Age[] = "Age";
static const char __pyx_k_BUY[] = "BUY";
static const char __pyx_k_NaN[] = "NaN";
static const char __pyx_k__10[] = "%";
static const char __pyx_k__11[] = "(";
static const char __pyx_k__12[] = " ";
static const char __pyx_k__13[] = " (";
static const char __pyx_k__14[] = " @ ";
static const char __pyx_k__15[] = ", ";
static const char __pyx_k__17[] = "*";
static const char __pyx_k__18[] = ".";
static const char __pyx_k__96[] = "?";
static const char __pyx_k_age[] = "age";
static const char __pyx_k_all[] = "all";
static const char __pyx_k_buy[] = "buy";
static const char __pyx_k_cls[] = "cls";
static const char __pyx_k_eta[] = "eta";
static const char __pyx_k_exp[] = "exp";
static const char __pyx_k_idx[] = "idx";
static const char __pyx_k_int[] = "int";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_len[] = "len";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_msg[] = "msg";
static const char __pyx_k_nan[] = "nan";
static const char __pyx_k_str[] = "str";
static const char __pyx_k_vol[] = " | vol=";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_Dict[] = "Dict";
static const char __pyx_k_INFO[] = "INFO";
static const char __pyx_k_List[] = "List";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_SELL[] = "SELL";
static const char __pyx_k_Type[] = "Type";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_bool[] = "bool";
static const char __pyx_k_buys[] = "buys";
static const char __pyx_k_ceil[] = "ceil";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_hang[] = "hang";
static const char __pyx_k_info[] = "info";
static const char __pyx_k_line[] = "line";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_math[] = "math";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_path[] = "path";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_sell[] = "sell";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_sort[] = "sort";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_time[] = "time";
static const char __pyx_k_unit[] = "unit";
static const char __pyx_k_Clock[] = "Clock";
static const char __pyx_k_H_M_S[] = "%H:%M:%S";
static const char __pyx_k_LIMIT[] = "LIMIT";
static const char __pyx_k_Level[] = "Level";
static const char __pyx_k_Maker[] = ") Maker ";
static const char __pyx_k_Price[] = "Price";
static const char __pyx_k_Tuple[] = "Tuple";
static const char __pyx_k_Union[] = "Union";
static const char __pyx_k_alpha[] = "alpha=";
static const char __pyx_k_apply[] = "apply";
static const char __pyx_k_clock[] = "clock";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_float[] = "float";
static const char __pyx_k_floor[] = "floor";
static const char __pyx_k_gamma[] = "gamma";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_isnan[] = "isnan";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_kappa[] = " | kappa=";
static const char __pyx_k_level[] = "level";
static const char __pyx_k_lines[] = "lines";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_order[] = " order ";
static const char __pyx_k_price[] = "price";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_ready[] = "ready";
static const char __pyx_k_regex[] = "regex";
static const char __pyx_k_round[] = "round";
static const char __pyx_k_sells[] = "sells";
static const char __pyx_k_split[] = "split";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_title[] = "title";
static const char __pyx_k_upper[] = "upper";
static const char __pyx_k_vol_2[] = "vol";
static const char __pyx_k_Assets[] = "  Assets:";
static const char __pyx_k_Market[] = "Market";
static const char __pyx_k_Orders[] = "  Orders:";
static const char __pyx_k_Spread[] = "Spread";
static const char __pyx_k_amount[] = "amount";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_exists[] = "exists";
static const char __pyx_k_filled[] = " filled.";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_header[] = "header";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_is_buy[] = "is_buy";
static const char __pyx_k_logger[] = "logger";
static const char __pyx_k_market[] = "market";
static const char __pyx_k_orders[] = "orders";
static const char __pyx_k_pandas[] = "pandas";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_return[] = "return";
static const char __pyx_k_sorted[] = "sorted";
static const char __pyx_k_spread[] = "spread";
static const char __pyx_k_to_csv[] = "to_csv";
static const char __pyx_k_typing[] = "typing";
static const char __pyx_k_unlink[] = "unlink";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_Current[] = "Current %";
static const char __pyx_k_Decimal[] = "Decimal";
static const char __pyx_k_Maker_2[] = "Maker ";
static const char __pyx_k_Markets[] = "  Markets:";
static const char __pyx_k_alpha_2[] = "alpha";
static const char __pyx_k_columns[] = "columns";
static const char __pyx_k_decimal[] = "decimal";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_kappa_2[] = "kappa";
static const char __pyx_k_logging[] = "logging";
static const char __pyx_k_lvl_buy[] = "lvl_buy";
static const char __pyx_k_new_ask[] = "new_ask";
static const char __pyx_k_new_bid[] = "new_bid";
static const char __pyx_k_order_2[] = "order";
static const char __pyx_k_percent[] = "percent";
static const char __pyx_k_replace[] = "replace";
static const char __pyx_k_reverse[] = "reverse";
static const char __pyx_k_seconds[] = "seconds";
static const char __pyx_k_warning[] = "warning";
static const char __pyx_k_Best_Ask[] = "Best Ask";
static const char __pyx_k_Best_Bid[] = "Best Bid";
static const char __pyx_k_Creating[] = ") Creating ";
static const char __pyx_k_Exchange[] = "Exchange";
static const char __pyx_k_MidPrice[] = "MidPrice";
static const char __pyx_k_Proposal[] = "Proposal";
static const char __pyx_k_WARNINGS[] = "*** WARNINGS ***";
static const char __pyx_k_best_ask[] = "best_ask";
static const char __pyx_k_best_bid[] = "best_bid";
static const char __pyx_k_datetime[] = "datetime";
static const char __pyx_k_df_lines[] = "df_lines";
static const char __pyx_k_end_time[] = "end_time";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_is_debug[] = "is_debug";
static const char __pyx_k_lvl_sell[] = "lvl_sell";
static const char __pyx_k_no_sells[] = "no_sells";
static const char __pyx_k_order_id[] = "order_id";
static const char __pyx_k_proposal[] = "proposal";
static const char __pyx_k_quantity[] = "quantity";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_strftime[] = "strftime";
static const char __pyx_k_CONNECTED[] = "CONNECTED";
static const char __pyx_k_DataFrame[] = "DataFrame";
static const char __pyx_k_OrderType[] = "OrderType";
static const char __pyx_k_PriceSize[] = "PriceSize";
static const char __pyx_k_PriceType[] = "PriceType";
static const char __pyx_k_Timestamp[] = "Timestamp";
static const char __pyx_k_TradeType[] = "TradeType";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_ask_price[] = "ask_price";
static const char __pyx_k_assets_df[] = "assets_df";
static const char __pyx_k_bid_price[] = "bid_price";
static const char __pyx_k_calculate[] = "calculate";
static const char __pyx_k_df_header[] = "df_header";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_getLogger[] = "getLogger";
static const char __pyx_k_get_price[] = "get_price";
static const char __pyx_k_is_filled[] = " is filled.";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_mid_price[] = "mid_price=";
static const char __pyx_k_normalize[] = "normalize";
static const char __pyx_k_order_age[] = "order_age";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_ref_price[] = "ref_price";
static const char __pyx_k_time_left[] = "time_left";
static const char __pyx_k_timedelta[] = "timedelta";
static const char __pyx_k_timestamp[] = "timestamp";
static const char __pyx_k_to_string[] = "to_string";
static const char __pyx_k_Amount_Adj[] = "Amount (Adj)";
static const char __pyx_k_LimitOrder[] = "LimitOrder";
static const char __pyx_k_add_sample[] = "add_sample";
static const char __pyx_k_all_orders[] = "all_orders";
static const char __pyx_k_base_asset[] = "base_asset";
static const char __pyx_k_base_ratio[] = "base_ratio";
static const char __pyx_k_base_value[] = "base_value";
static const char __pyx_k_config_map[] = "config_map";
static const char __pyx_k_formatters[] = "formatters";
static const char __pyx_k_get_spread[] = "get_spread";
static const char __pyx_k_level_step[] = "level_step";
static const char __pyx_k_markets_df[] = "markets_df";
static const char __pyx_k_min_spread[] = "min_spread";
static const char __pyx_k_order_book[] = "order_book";
static const char __pyx_k_pmm_logger[] = "pmm_logger";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_sell_order[] = "sell_order";
static const char __pyx_k_set_timers[] = "set_timers";
static const char __pyx_k_start_time[] = "start_time";
static const char __pyx_k_target_inv[] = "target_inv";
static const char __pyx_k_trade_type[] = "trade_type";
static const char __pyx_k_volatility[] = "    volatility= ";
static const char __pyx_k_Amount_Orig[] = "Amount (Orig)";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_amount_orig[] = "amount_orig";
static const char __pyx_k_current_inv[] = "current_inv";
static const char __pyx_k_get_balance[] = "get_balance";
static const char __pyx_k_init_params[] = "init_params";
static const char __pyx_k_market_info[] = "market_info";
static const char __pyx_k_mid_price_2[] = "mid_price";
static const char __pyx_k_non_hanging[] = "non_hanging";
static const char __pyx_k_optimal_ask[] = "% | optimal_ask=";
static const char __pyx_k_optimal_bid[] = "optimal_bid=";
static const char __pyx_k_quote_asset[] = "quote_asset";
static const char __pyx_k_quote_ratio[] = "quote_ratio";
static const char __pyx_k_risk_factor[] = "risk_factor";
static const char __pyx_k_ExchangeBase[] = "ExchangeBase";
static const char __pyx_k_List_Decimal[] = "List[Decimal]";
static const char __pyx_k_StrategyBase[] = "StrategyBase";
static const char __pyx_k_base_balance[] = "base_balance";
static const char __pyx_k_cancel_order[] = "cancel_order";
static const char __pyx_k_closing_time[] = "closing_time";
static const char __pyx_k_display_name[] = "display_name";
static const char __pyx_k_end_datetime[] = "end_datetime";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_limit_orders[] = "limit_orders";
static const char __pyx_k_market_books[] = "market_books";
static const char __pyx_k_markets_data[] = "markets_data";
static const char __pyx_k_model_config[] = "model_config";
static const char __pyx_k_order_amount[] = "order_amount";
static const char __pyx_k_order_levels[] = "order_levels";
static const char __pyx_k_pd_DataFrame[] = "pd.DataFrame";
static const char __pyx_k_process_tick[] = "process_tick";
static const char __pyx_k_result_price[] = "result_price";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_trading_pair[] = "trading_pair";
static const char __pyx_k_volatility_2[] = "volatility";
static const char __pyx_k_Current_Value[] = "Current Value (";
static const char __pyx_k_NetworkStatus[] = "NetworkStatus";
static const char __pyx_k_Total_Balance[] = "Total Balance";
static const char __pyx_k_active_orders[] = "active_orders";
static const char __pyx_k_base_currency[] = "base_currency";
static const char __pyx_k_class_getitem[] = "__class_getitem__";
static const char __pyx_k_current_value[] = "current_value";
static const char __pyx_k_end_timestamp[] = "end_timestamp";
static const char __pyx_k_format_status[] = "format_status";
static const char __pyx_k_get_mid_price[] = "get_mid_price";
static const char __pyx_k_map_df_to_str[] = "map_df_to_str";
static const char __pyx_k_max_order_age[] = "max_order_age";
static const char __pyx_k_notify_hb_app[] = "notify_hb_app";
static const char __pyx_k_optimal_ask_2[] = "optimal_ask";
static const char __pyx_k_optimal_bid_2[] = "optimal_bid";
static const char __pyx_k_quote_balance[] = "quote_balance";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_risk_factor_2[] = "    risk_factor(\316\263)= ";
static const char __pyx_k_s_decimal_one[] = "s_decimal_one";
static const char __pyx_k_warning_lines[] = "warning_lines";
static const char __pyx_k_AttributeError[] = "AttributeError";
static const char __pyx_k_OPTION_LOG_ALL[] = "OPTION_LOG_ALL";
static const char __pyx_k_Optimal_Spread[] = "Optimal Spread";
static const char __pyx_k_active_markets[] = "active_markets";
static const char __pyx_k_current_prices[] = "current_prices";
static const char __pyx_k_debug_csv_path[] = "debug_csv_path";
static const char __pyx_k_execution_mode[] = "execution_mode";
static const char __pyx_k_get_last_price[] = "get_last_price";
static const char __pyx_k_get_volatility[] = "get_volatility";
static const char __pyx_k_log_with_clock[] = "log_with_clock";
static const char __pyx_k_network_status[] = "network_status";
static const char __pyx_k_optimal_spread[] = " | optimal_spread=";
static const char __pyx_k_order_override[] = "order_override";
static const char __pyx_k_price_delegate[] = "price_delegate";
static const char __pyx_k_quote_currency[] = "quote_currency";
static const char __pyx_k_s_decimal_zero[] = "s_decimal_zero";
static const char __pyx_k_start_datetime[] = "start_datetime";
static const char __pyx_k_total_in_quote[] = "total_in_quote";
static const char __pyx_k_volatility_pct[] = "volatility_pct";
static const char __pyx_k_List_LimitOrder[] = "List[LimitOrder]";
static const char __pyx_k_balance_warning[] = "balance_warning";
static const char __pyx_k_client_order_id[] = "client_order_id";
static const char __pyx_k_execution_state[] = "execution_state";
static const char __pyx_k_level_distances[] = "level_distances";
static const char __pyx_k_logging_options[] = "logging_options";
static const char __pyx_k_markets_columns[] = "markets_columns";
static const char __pyx_k_network_warning[] = "network_warning";
static const char __pyx_k_proposal_prices[] = "proposal_prices";
static const char __pyx_k_register_events[] = "register_events";
static const char __pyx_k_sampling_length[] = "sampling_length";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_start_timestamp[] = "start_timestamp";
static const char __pyx_k_active_orders_df[] = "active_orders_df";
static const char __pyx_k_first_col_length[] = "first_col_length";
static const char __pyx_k_get_limit_orders[] = "get_limit_orders";
static const char __pyx_k_is_hanging_order[] = "is_hanging_order";
static const char __pyx_k_optimal_spread_2[] = "optimal_spread";
static const char __pyx_k_to_create_orders[] = "to_create_orders";
static const char __pyx_k_Available_Balance[] = "Available Balance";
static const char __pyx_k_FileNotFoundError[] = "FileNotFoundError";
static const char __pyx_k_Reservation_Price[] = "Reservation Price";
static const char __pyx_k_all_markets_ready[] = "all_markets_ready";
static const char __pyx_k_ask_level_spreads[] = "ask_level_spreads";
static const char __pyx_k_bid_level_spreads[] = "bid_level_spreads";
static const char __pyx_k_get_level_spreads[] = "_get_level_spreads";
static const char __pyx_k_get_price_by_type[] = "get_price_by_type";
static const char __pyx_k_in_flight_cancels[] = "in_flight_cancels";
static const char __pyx_k_mid_price_std_dev[] = "mid_price std_dev";
static const char __pyx_k_order_levels_mode[] = "order_levels_mode";
static const char __pyx_k_pure_mm_assets_df[] = "pure_mm_assets_df";
static const char __pyx_k_reservation_price[] = " | reservation_price=";
static const char __pyx_k_s_decimal_neg_one[] = "s_decimal_neg_one";
static const char __pyx_k_unregister_events[] = "unregister_events";
static const char __pyx_k_Maker_buy_order_of[] = ") Maker buy order of ";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_filled_order_delay[] = "filled_order_delay";
static const char __pyx_k_is_algorithm_ready[] = "is_algorithm_ready";
static const char __pyx_k_mid_price_variance[] = "mid_price_variance";
static const char __pyx_k_optimal_ask_to_mid[] = "optimal_ask_to_mid_%";
static const char __pyx_k_optimal_bid_to_mid[] = "optimal_bid_to_mid_%";
static const char __pyx_k_order_refresh_time[] = "order_refresh_time";
static const char __pyx_k_time_left_fraction[] = "time_left_fraction";
static const char __pyx_k_ClientConfigAdapter[] = "ClientConfigAdapter";
static const char __pyx_k_CreatedPairOfOrders[] = "CreatedPairOfOrders";
static const char __pyx_k_FromDateToDateModel[] = "FromDateToDateModel";
static const char __pyx_k_Maker_sell_order_of[] = ") Maker sell order of ";
static const char __pyx_k_Strategy_parameters[] = "  Strategy parameters:";
static const char __pyx_k_calculate_inventory[] = "calculate_inventory";
static const char __pyx_k_execution_timeframe[] = "execution_timeframe";
static const char __pyx_k_hanging_orders_mode[] = "hanging_orders_mode";
static const char __pyx_k_hb_app_notification[] = "hb_app_notification";
static const char __pyx_k_is_within_tolerance[] = "is_within_tolerance";
static const char __pyx_k_reservation_price_2[] = "reservation_price";
static const char __pyx_k_to_show_current_pct[] = "to_show_current_pct";
static const char __pyx_k_HangingOrdersTracker[] = "HangingOrdersTracker";
static const char __pyx_k_MultiOrderLevelModel[] = "MultiOrderLevelModel";
static const char __pyx_k_add_as_hanging_order[] = "add_as_hanging_order";
static const char __pyx_k_cancel_active_orders[] = "cancel_active_orders";
static const char __pyx_k_create_base_proposal[] = "create_base_proposal";
static const char __pyx_k_dump_debug_variables[] = "dump_debug_variables";
static const char __pyx_k_get_maker_order_type[] = "get_maker_order_type";
static const char __pyx_k_get_price_for_volume[] = "get_price_for_volume";
static const char __pyx_k_inventory_target_pct[] = "inventory_target_pct";
static const char __pyx_k_is_algorithm_changed[] = "is_algorithm_changed";
static const char __pyx_k_add_transaction_costs[] = "add_transaction_costs";
static const char __pyx_k_create_basic_proposal[] = "create_basic_proposal";
static const char __pyx_k_get_available_balance[] = "get_available_balance";
static const char __pyx_k_hummingbot_core_utils[] = "hummingbot.core.utils";
static const char __pyx_k_inventory_target_base[] = "inventory_target_base";
static const char __pyx_k_ticks_to_fill_buffers[] = " ticks to fill buffers";
static const char __pyx_k_DailyBetweenTimesModel[] = "DailyBetweenTimesModel";
static const char __pyx_k_MarketTradingPairTuple[] = "MarketTradingPairTuple";
static const char __pyx_k_No_active_maker_orders[] = "  No active maker orders.";
static const char __pyx_k_all_non_hanging_orders[] = "all_non_hanging_orders";
static const char __pyx_k_available_base_balance[] = "available_base_balance";
static const char __pyx_k_hanging_orders_enabled[] = "hanging_orders_enabled";
static const char __pyx_k_hanging_orders_tracker[] = "hanging_orders_tracker";
static const char __pyx_k_status_report_interval[] = "status_report_interval";
static const char __pyx_k_update_from_config_map[] = "update_from_config_map";
static const char __pyx_k_volatility_buffer_size[] = "volatility_buffer_size";
static const char __pyx_k_OPTION_LOG_CREATE_ORDER[] = "OPTION_LOG_CREATE_ORDER";
static const char __pyx_k_RunAlwaysExecutionState[] = "RunAlwaysExecutionState";
static const char __pyx_k_TrackHangingOrdersModel[] = "TrackHangingOrdersModel";
static const char __pyx_k_apply_budget_constraint[] = "apply_budget_constraint";
static const char __pyx_k_available_quote_balance[] = "available_quote_balance";
static const char __pyx_k_execute_orders_proposal[] = "execute_orders_proposal";
static const char __pyx_k_is_sampling_buffer_full[] = "is_sampling_buffer_full";
static const char __pyx_k_order_book_depth_factor[] = "    order_book_depth_factor(\316\272)= ";
static const char __pyx_k_order_refresh_tolerance[] = "order_refresh_tolerance";
static const char __pyx_k_ticks_to_be_ready_after[] = "ticks_to_be_ready_after";
static const char __pyx_k_OPTION_LOG_STATUS_REPORT[] = "OPTION_LOG_STATUS_REPORT";
static const char __pyx_k_apply_order_optimization[] = "apply_order_optimization";
static const char __pyx_k_ask_orders_at_Size_Price[] = " ask orders at (Size, Price): ";
static const char __pyx_k_bid_orders_at_Size_Price[] = " bid orders at (Size, Price): ";
static const char __pyx_k_candidate_hanging_orders[] = "candidate_hanging_orders";
static const char __pyx_k_collect_market_variables[] = "collect_market_variables";
static const char __pyx_k_execution_timeframe_mode[] = "execution_timeframe_mode";
static const char __pyx_k_market_status_data_frame[] = "market_status_data_frame";
static const char __pyx_k_ticks_to_be_ready_before[] = "ticks_to_be_ready_before";
static const char __pyx_k_TradingIntensityIndicator[] = "TradingIntensityIndicator";
static const char __pyx_k_active_non_hanging_orders[] = "active_non_hanging_orders";
static const char __pyx_k_get_config_map_indicators[] = "get_config_map_indicators";
static const char __pyx_k_hanging_orders_cancel_pct[] = "hanging_orders_cancel_pct";
static const char __pyx_k_hummingbot_strategy_utils[] = "hummingbot.strategy.utils";
static const char __pyx_k_inventory_target_base_pct[] = "inventory_target_base_pct";
static const char __pyx_k_order_amount_shape_factor[] = "order_amount_shape_factor";
static const char __pyx_k_InstantVolatilityIndicator[] = "InstantVolatilityIndicator";
static const char __pyx_k_calculate_target_inventory[] = "calculate_target_inventory";
static const char __pyx_k_has_been_completely_filled[] = ") has been completely filled.";
static const char __pyx_k_is_potential_hanging_order[] = "is_potential_hanging_order";
static const char __pyx_k_is_sampling_buffer_changed[] = "is_sampling_buffer_changed";
static const char __pyx_k_market_trading_pair_tuples[] = "market_trading_pair_tuples";
static const char __pyx_k_order_optimization_enabled[] = "order_optimization_enabled";
static const char __pyx_k_List_MarketTradingPairTuple[] = "List[MarketTradingPairTuple]";
static const char __pyx_k_OrderBookAssetPriceDelegate[] = "OrderBookAssetPriceDelegate";
static const char __pyx_k_apply_add_transaction_costs[] = "apply_add_transaction_costs";
static const char __pyx_k_apply_order_price_modifiers[] = "apply_order_price_modifiers";
static const char __pyx_k_order_book_intensity_factor[] = "    order_book_intensity_factor(\316\221)= ";
static const char __pyx_k_order_refresh_tolerance_pct[] = "order_refresh_tolerance_pct";
static const char __pyx_k_market_info_to_active_orders[] = "market_info_to_active_orders";
static const char __pyx_k_market_pair_to_active_orders[] = "market_pair_to_active_orders";
static const char __pyx_k_measure_order_book_liquidity[] = "measure_order_book_liquidity";
static const char __pyx_k_notify_hb_app_with_timestamp[] = "notify_hb_app_with_timestamp";
static const char __pyx_k_OPTION_LOG_MAKER_ORDER_FILLED[] = "OPTION_LOG_MAKER_ORDER_FILLED";
static const char __pyx_k_get_config_map_execution_mode[] = "get_config_map_execution_mode";
static const char __pyx_k_get_config_map_hanging_orders[] = "get_config_map_hanging_orders";
static const char __pyx_k_is_order_id_in_hanging_orders[] = "is_order_id_in_hanging_orders";
static const char __pyx_k_trading_intensity_buffer_size[] = "trading_intensity_buffer_size";
static const char __pyx_k_AvellanedaMarketMakingStrategy[] = "AvellanedaMarketMakingStrategy";
static const char __pyx_k_active_orders_df_locals_lambda[] = "active_orders_df.<locals>.<lambda>";
static const char __pyx_k_get_adjusted_available_balance[] = "get_adjusted_available_balance";
static const char __pyx_k_hummingbot_strategy_data_types[] = "hummingbot.strategy.data_types";
static const char __pyx_k_AvellanedaMarketMakingConfigMap[] = "AvellanedaMarketMakingConfigMap";
static const char __pyx_k_Market_connectors_are_not_ready[] = "Market connectors are not ready.";
static const char __pyx_k_Markets_are_not_ready_No_market[] = "Markets are not ready. No market making trades are permitted.";
static const char __pyx_k_calculate_reservation_price_and[] = "calculate_reservation_price_and_optimal_spread";
static const char __pyx_k_current_created_pairs_of_orders[] = "current_created_pairs_of_orders";
static const char __pyx_k_hummingbot_client_config_config[] = "hummingbot.client.config.config_helpers";
static const char __pyx_k_hummingbot_core_data_type_limit[] = "hummingbot.core.data_type.limit_order";
static const char __pyx_k_hummingbot_strategy_conditional[] = "hummingbot.strategy.conditional_execution_state";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_time_until_end_of_trading_cycle[] = "    time until end of trading cycle = ";
static const char __pyx_k_AvellanedaMarketMakingStrategy_2[] = "AvellanedaMarketMakingStrategy._get_level_spreads";
static const char __pyx_k_AvellanedaMarketMakingStrategy_3[] = "AvellanedaMarketMakingStrategy.__reduce_cython__";
static const char __pyx_k_AvellanedaMarketMakingStrategy_4[] = "AvellanedaMarketMakingStrategy.__setstate_cython__";
static const char __pyx_k_AvellanedaMarketMakingStrategy_a[] = "AvellanedaMarketMakingStrategy.all_markets_ready";
static const char __pyx_k_AvellanedaMarketMakingStrategy_c[] = "AvellanedaMarketMakingStrategy.c_start.<locals>.genexpr";
static const char __pyx_k_AvellanedaMarketMakingStrategy_d[] = "AvellanedaMarketMakingStrategy.dump_debug_variables";
static const char __pyx_k_AvellanedaMarketMakingStrategy_e[] = "AvellanedaMarketMakingStrategy.execute_orders_proposal";
static const char __pyx_k_AvellanedaMarketMakingStrategy_f[] = "AvellanedaMarketMakingStrategy.format_status";
static const char __pyx_k_AvellanedaMarketMakingStrategy_g[] = "AvellanedaMarketMakingStrategy.get_price";
static const char __pyx_k_AvellanedaMarketMakingStrategy_i[] = "AvellanedaMarketMakingStrategy.init_params";
static const char __pyx_k_AvellanedaMarketMakingStrategy_l[] = "AvellanedaMarketMakingStrategy.logger";
static const char __pyx_k_AvellanedaMarketMakingStrategy_m[] = "AvellanedaMarketMakingStrategy.market_status_data_frame";
static const char __pyx_k_AvellanedaMarketMakingStrategy_n[] = "AvellanedaMarketMakingStrategy.notify_hb_app";
static const char __pyx_k_AvellanedaMarketMakingStrategy_p[] = "AvellanedaMarketMakingStrategy.pure_mm_assets_df";
static const char __pyx_k_AvellanedaMarketMakingStrategy_s[] = "AvellanedaMarketMakingStrategy.start";
static const char __pyx_k_AvellanedaMarketMakingStrategy_t[] = "AvellanedaMarketMakingStrategy.to_create_orders";
static const char __pyx_k_AvellanedaMarketMakingStrategy_u[] = "AvellanedaMarketMakingStrategy.update_from_config_map";
static const char __pyx_k_Calculating_volatility_estimatin[] = "Calculating volatility, estimating order book liquidity ... ";
static const char __pyx_k_RunInTimeConditionalExecutionSta[] = "RunInTimeConditionalExecutionState";
static const char __pyx_k_Tuple_List_Proposal_List_Proposa[] = "Tuple[List[Proposal], List[Proposal]]";
static const char __pyx_k_Union_AvellanedaMarketMakingConf[] = "Union[AvellanedaMarketMakingConfigMap, ClientConfigAdapter]";
static const char __pyx_k_WARNING_Some_markets_are_not_con[] = "WARNING: Some markets are not connected or are down at the moment. Market making may be dangerous when markets or networks are unstable.";
static const char __pyx_k_add_current_pairs_of_proposal_or[] = "add_current_pairs_of_proposal_orders_executed_by_strategy";
static const char __pyx_k_adjusted_available_balance_for_o[] = "adjusted_available_balance_for_orders_budget_constrain";
static const char __pyx_k_apply_order_amount_eta_transform[] = "apply_order_amount_eta_transformation";
static const char __pyx_k_candidate_hanging_orders_from_pa[] = "candidate_hanging_orders_from_pairs";
static const char __pyx_k_create_proposal_based_on_order_l[] = "create_proposal_based_on_order_levels";
static const char __pyx_k_create_proposal_based_on_order_o[] = "create_proposal_based_on_order_override";
static const char __pyx_k_hummingbot_connector_exchange_ba[] = "hummingbot.connector.exchange_base";
static const char __pyx_k_hummingbot_core_data_type_common[] = "hummingbot.core.data_type.common";
static const char __pyx_k_hummingbot_core_network_iterator[] = "hummingbot.core.network_iterator";
static const char __pyx_k_hummingbot_strategy___utils___tr[] = "hummingbot.strategy.__utils__.trailing_indicators.instant_volatility";
static const char __pyx_k_hummingbot_strategy_avellaneda_m[] = "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making";
static const char __pyx_k_hummingbot_strategy_hanging_orde[] = "hummingbot.strategy.hanging_orders_tracker";
static const char __pyx_k_hummingbot_strategy_market_tradi[] = "hummingbot.strategy.market_trading_pair_tuple";
static const char __pyx_k_hummingbot_strategy_order_book_a[] = "hummingbot.strategy.order_book_asset_price_delegate";
static const char __pyx_k_hummingbot_strategy_strategy_bas[] = "hummingbot.strategy.strategy_base";
static const char __pyx_k_is_order_id_in_completed_hanging[] = "is_order_id_in_completed_hanging_orders";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_self__events_cannot_be_converted[] = "self._events cannot be converted to a Python object for pickling";
static const char __pyx_k_should_wait_order_cancel_confirm[] = "should_wait_order_cancel_confirmation";
static const char __pyx_k_update_strategy_orders_with_equi[] = "update_strategy_orders_with_equivalent_orders";
static const char __pyx_k_time_until_end_of_trading_cycle_2[] = "    time until end of trading cycle = N/A";
static const char __pyx_k_AvellanedaMarketMakingStrategy_a_2[] = "AvellanedaMarketMakingStrategy.active_orders_df";
static const char __pyx_k_AvellanedaMarketMakingStrategy_a_3[] = "AvellanedaMarketMakingStrategy.apply_order_price_modifiers";
static const char __pyx_k_AvellanedaMarketMakingStrategy_a_4[] = "AvellanedaMarketMakingStrategy.apply_budget_constraint";
static const char __pyx_k_AvellanedaMarketMakingStrategy_a_5[] = "AvellanedaMarketMakingStrategy.adjusted_available_balance_for_orders_budget_constrain";
static const char __pyx_k_AvellanedaMarketMakingStrategy_a_6[] = "AvellanedaMarketMakingStrategy.apply_order_optimization";
static const char __pyx_k_AvellanedaMarketMakingStrategy_a_7[] = "AvellanedaMarketMakingStrategy.apply_order_amount_eta_transformation";
static const char __pyx_k_AvellanedaMarketMakingStrategy_a_8[] = "AvellanedaMarketMakingStrategy.apply_add_transaction_costs";
static const char __pyx_k_AvellanedaMarketMakingStrategy_c_2[] = "AvellanedaMarketMakingStrategy.c_apply_order_optimization.<locals>.<lambda>";
static const char __pyx_k_AvellanedaMarketMakingStrategy_c_3[] = "AvellanedaMarketMakingStrategy.c_execute_orders_proposal.<locals>.genexpr";
static const char __pyx_k_AvellanedaMarketMakingStrategy_c_4[] = "AvellanedaMarketMakingStrategy.cancel_order";
static const char __pyx_k_AvellanedaMarketMakingStrategy_c_5[] = "AvellanedaMarketMakingStrategy.collect_market_variables";
static const char __pyx_k_AvellanedaMarketMakingStrategy_c_6[] = "AvellanedaMarketMakingStrategy.calculate_reservation_price_and_optimal_spread";
static const char __pyx_k_AvellanedaMarketMakingStrategy_c_7[] = "AvellanedaMarketMakingStrategy.calculate_target_inventory";
static const char __pyx_k_AvellanedaMarketMakingStrategy_c_8[] = "AvellanedaMarketMakingStrategy.calculate_inventory";
static const char __pyx_k_AvellanedaMarketMakingStrategy_c_9[] = "AvellanedaMarketMakingStrategy.create_proposal_based_on_order_override";
static const char __pyx_k_AvellanedaMarketMakingStrategy_g_2[] = "AvellanedaMarketMakingStrategy.get_last_price";
static const char __pyx_k_AvellanedaMarketMakingStrategy_g_3[] = "AvellanedaMarketMakingStrategy.get_mid_price";
static const char __pyx_k_AvellanedaMarketMakingStrategy_g_4[] = "AvellanedaMarketMakingStrategy.get_config_map_execution_mode";
static const char __pyx_k_AvellanedaMarketMakingStrategy_g_5[] = "AvellanedaMarketMakingStrategy.get_config_map_hanging_orders";
static const char __pyx_k_AvellanedaMarketMakingStrategy_g_6[] = "AvellanedaMarketMakingStrategy.get_config_map_indicators";
static const char __pyx_k_AvellanedaMarketMakingStrategy_g_7[] = "AvellanedaMarketMakingStrategy.get_spread";
static const char __pyx_k_AvellanedaMarketMakingStrategy_g_8[] = "AvellanedaMarketMakingStrategy.get_volatility";
static const char __pyx_k_AvellanedaMarketMakingStrategy_g_9[] = "AvellanedaMarketMakingStrategy.get_adjusted_available_balance";
static const char __pyx_k_AvellanedaMarketMakingStrategy_i_2[] = "AvellanedaMarketMakingStrategy.is_algorithm_ready";
static const char __pyx_k_AvellanedaMarketMakingStrategy_i_3[] = "AvellanedaMarketMakingStrategy.is_algorithm_changed";
static const char __pyx_k_AvellanedaMarketMakingStrategy_i_4[] = "AvellanedaMarketMakingStrategy.is_within_tolerance";
static const char __pyx_k_AvellanedaMarketMakingStrategy_m_2[] = "AvellanedaMarketMakingStrategy.measure_order_book_liquidity";
static const char __pyx_k_AvellanedaMarketMakingStrategy_p_2[] = "AvellanedaMarketMakingStrategy.process_tick";
static const char __pyx_k_AvellanedaMarketMakingStrategy_s_2[] = "AvellanedaMarketMakingStrategy.set_timers";
static const char __pyx_k_Calculating_volatility_estimatin_2[] = "Calculating volatility, estimating order book liquidity ... no trades tick";
static const char __pyx_k_hummingbot_strategy___utils___tr_2[] = "hummingbot.strategy.__utils__.trailing_indicators.trading_intensity";
static const char __pyx_k_hummingbot_strategy_avellaneda_m_2[] = "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making_config_map_pydantic";
static const char __pyx_k_hummingbot_strategy_avellaneda_m_3[] = "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx";
static const char __pyx_k_AvellanedaMarketMakingStrategy_c_10[] = "AvellanedaMarketMakingStrategy.create_proposal_based_on_order_levels";
static const char __pyx_k_AvellanedaMarketMakingStrategy_c_11[] = "AvellanedaMarketMakingStrategy.create_basic_proposal";
static const char __pyx_k_AvellanedaMarketMakingStrategy_c_12[] = "AvellanedaMarketMakingStrategy.create_base_proposal";
static const char __pyx_k_AvellanedaMarketMakingStrategy_c_13[] = "AvellanedaMarketMakingStrategy.cancel_active_orders";
/* #### Code section: decls ### */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_logger(CYTHON_UNUSED PyTypeObject *__pyx_v_cls); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_2init_params(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_config_map, PyObject *__pyx_v_market_info, PyObject *__pyx_v_logging_options, PyObject *__pyx_v_status_report_interval, PyObject *__pyx_v_hb_app_notification, PyObject *__pyx_v_debug_csv_path, PyObject *__pyx_v_is_debug); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_4all_markets_ready(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10min_spread___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7avg_vol___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7avg_vol_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_indicator); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17trading_intensity___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17trading_intensity_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_indicator); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11market_info___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_27order_refresh_tolerance_pct___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_23order_refresh_tolerance___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12order_amount___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25inventory_target_base_pct___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_21inventory_target_base___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_21inventory_target_base_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26order_optimization_enabled___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_18order_refresh_time___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_18filled_order_delay___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_14order_override___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12order_levels___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15level_distances___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_13max_order_age___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_31add_transaction_costs_to_orders___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10base_asset___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11quote_asset___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12trading_pair___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5gamma___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5alpha___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5alpha_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5kappa___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5kappa_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_3eta___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17reservation_price___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17reservation_price_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_14optimal_spread___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_ask___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_ask_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_bid___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_bid_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_19execution_timeframe___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10start_time___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10start_time_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_8end_time___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_8end_time_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_6get_price(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_8get_last_price(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10get_mid_price(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_28market_info_to_active_orders___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_13active_orders___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25active_non_hanging_orders___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11active_buys___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12active_sells___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15logging_options___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15logging_options_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, int64_t __pyx_v_logging_options); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_22hanging_orders_tracker___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12update_from_config_map(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_14get_config_map_execution_mode(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_16get_config_map_hanging_orders(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_18get_config_map_indicators(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_20pure_mm_assets_df(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_to_show_current_pct); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_22active_orders_df(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_24market_status_data_frame(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_market_trading_pair_tuples); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26format_status(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_28execute_orders_proposal(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_30cancel_order(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_order_id); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7c_start_genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_32start(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_clock, PyObject *__pyx_v_timestamp); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_34process_tick(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_timestamp); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_36collect_market_variables(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_timestamp); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_38get_spread(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_40get_volatility(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_42measure_order_book_liquidity(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_44calculate_reservation_price_and_optimal_spread(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_46calculate_target_inventory(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_48calculate_inventory(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_50is_algorithm_ready(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_52is_algorithm_changed(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_54_get_level_spreads(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_56create_proposal_based_on_order_override(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_58create_proposal_based_on_order_levels(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_60create_basic_proposal(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_62create_base_proposal(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_64get_adjusted_available_balance(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_orders); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_66apply_order_price_modifiers(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_68apply_budget_constraint(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_70adjusted_available_balance_for_orders_budget_constrain(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_72apply_budget_constraint(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda2(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda3(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_74apply_order_optimization(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_76apply_order_amount_eta_transformation(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_78apply_add_transaction_costs(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_80is_within_tolerance(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_current_prices, PyObject *__pyx_v_proposal_prices); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_82cancel_active_orders(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_84to_create_orders(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25c_execute_orders_proposal_genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25c_execute_orders_proposal_3genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_86execute_orders_proposal(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_88set_timers(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_90notify_hb_app(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_msg); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_92dump_debug_variables(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_94__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_96__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static __Pyx_CachedCFunction __pyx_umethod_PyUnicode_Type_upper = {0, 0, 0, 0, 0};
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_6pubsub_PubSub;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult;
  PyTypeObject *__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_7cpython_4type_type;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_5numpy_dtype;
  PyTypeObject *__pyx_ptype_5numpy_flatiter;
  PyTypeObject *__pyx_ptype_5numpy_broadcast;
  PyTypeObject *__pyx_ptype_5numpy_ndarray;
  PyTypeObject *__pyx_ptype_5numpy_generic;
  PyTypeObject *__pyx_ptype_5numpy_number;
  PyTypeObject *__pyx_ptype_5numpy_integer;
  PyTypeObject *__pyx_ptype_5numpy_signedinteger;
  PyTypeObject *__pyx_ptype_5numpy_unsignedinteger;
  PyTypeObject *__pyx_ptype_5numpy_inexact;
  PyTypeObject *__pyx_ptype_5numpy_floating;
  PyTypeObject *__pyx_ptype_5numpy_complexfloating;
  PyTypeObject *__pyx_ptype_5numpy_flexible;
  PyTypeObject *__pyx_ptype_5numpy_character;
  PyTypeObject *__pyx_ptype_5numpy_ufunc;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator;
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradesForwarder;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_5clock_Clock;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator;
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_13order_tracker_OrderTracker;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase;
  #if CYTHON_USE_MODULE_STATE
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase;
  #if CYTHON_USE_MODULE_STATE
  PyObject *__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy;
  PyObject *__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start;
  PyObject *__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr;
  PyObject *__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal;
  PyObject *__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr;
  PyObject *__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr;
  #endif
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy;
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start;
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr;
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal;
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr;
  PyTypeObject *__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr;
  PyObject *__pyx_kp_u_0;
  PyObject *__pyx_kp_u_1;
  PyObject *__pyx_kp_u_10;
  PyObject *__pyx_kp_u_100;
  PyObject *__pyx_kp_u_10f;
  PyObject *__pyx_kp_u_1_2;
  PyObject *__pyx_kp_u_2;
  PyObject *__pyx_kp_u_3f;
  PyObject *__pyx_kp_u_4f;
  PyObject *__pyx_kp_u_5E;
  PyObject *__pyx_n_u_Age;
  PyObject *__pyx_kp_u_Amount_Adj;
  PyObject *__pyx_kp_u_Amount_Orig;
  PyObject *__pyx_kp_u_Assets;
  PyObject *__pyx_n_s_AttributeError;
  PyObject *__pyx_kp_u_Available_Balance;
  PyObject *__pyx_n_s_AvellanedaMarketMakingConfigMap;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_2;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_3;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_4;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_a;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_a_2;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_a_3;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_a_4;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_a_5;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_a_6;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_a_7;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_a_8;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_c;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_c_10;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_c_11;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_c_12;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_c_13;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_c_2;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_c_3;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_c_4;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_c_5;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_c_6;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_c_7;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_c_8;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_c_9;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_d;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_e;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_f;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_g;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_g_2;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_g_3;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_g_4;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_g_5;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_g_6;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_g_7;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_g_8;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_g_9;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_i;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_i_2;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_i_3;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_i_4;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_l;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_m;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_m_2;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_n;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_p;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_p_2;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_s;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_s_2;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_t;
  PyObject *__pyx_n_s_AvellanedaMarketMakingStrategy_u;
  PyObject *__pyx_n_s_BUY;
  PyObject *__pyx_kp_u_Best_Ask;
  PyObject *__pyx_kp_u_Best_Bid;
  PyObject *__pyx_n_s_CONNECTED;
  PyObject *__pyx_kp_u_Calculating_volatility_estimatin;
  PyObject *__pyx_kp_u_Calculating_volatility_estimatin_2;
  PyObject *__pyx_n_s_ClientConfigAdapter;
  PyObject *__pyx_n_s_Clock;
  PyObject *__pyx_n_s_CreatedPairOfOrders;
  PyObject *__pyx_kp_u_Creating;
  PyObject *__pyx_kp_u_Current;
  PyObject *__pyx_kp_u_Current_Value;
  PyObject *__pyx_n_s_DailyBetweenTimesModel;
  PyObject *__pyx_n_s_DataFrame;
  PyObject *__pyx_n_s_Decimal;
  PyObject *__pyx_n_s_Dict;
  PyObject *__pyx_n_u_Exchange;
  PyObject *__pyx_n_s_ExchangeBase;
  PyObject *__pyx_n_s_FileNotFoundError;
  PyObject *__pyx_n_s_FromDateToDateModel;
  PyObject *__pyx_kp_u_H_M_S;
  PyObject *__pyx_n_s_HangingOrdersTracker;
  PyObject *__pyx_n_s_INFO;
  PyObject *__pyx_n_s_ImportError;
  PyObject *__pyx_n_s_InstantVolatilityIndicator;
  PyObject *__pyx_n_s_LIMIT;
  PyObject *__pyx_n_u_Level;
  PyObject *__pyx_n_s_LimitOrder;
  PyObject *__pyx_n_s_List;
  PyObject *__pyx_kp_s_List_Decimal;
  PyObject *__pyx_kp_s_List_LimitOrder;
  PyObject *__pyx_kp_s_List_MarketTradingPairTuple;
  PyObject *__pyx_kp_u_Maker;
  PyObject *__pyx_kp_u_Maker_2;
  PyObject *__pyx_kp_u_Maker_buy_order_of;
  PyObject *__pyx_kp_u_Maker_sell_order_of;
  PyObject *__pyx_n_u_Market;
  PyObject *__pyx_n_s_MarketTradingPairTuple;
  PyObject *__pyx_kp_u_Market_connectors_are_not_ready;
  PyObject *__pyx_kp_u_Markets;
  PyObject *__pyx_kp_u_Markets_are_not_ready_No_market;
  PyObject *__pyx_n_s_MidPrice;
  PyObject *__pyx_n_u_MidPrice;
  PyObject *__pyx_n_s_MultiOrderLevelModel;
  PyObject *__pyx_n_s_NaN;
  PyObject *__pyx_n_s_NetworkStatus;
  PyObject *__pyx_kp_u_No_active_maker_orders;
  PyObject *__pyx_kp_u_None;
  PyObject *__pyx_n_s_OPTION_LOG_ALL;
  PyObject *__pyx_n_s_OPTION_LOG_CREATE_ORDER;
  PyObject *__pyx_n_s_OPTION_LOG_MAKER_ORDER_FILLED;
  PyObject *__pyx_n_s_OPTION_LOG_STATUS_REPORT;
  PyObject *__pyx_kp_u_Optimal_Spread;
  PyObject *__pyx_n_s_OrderBookAssetPriceDelegate;
  PyObject *__pyx_n_s_OrderType;
  PyObject *__pyx_kp_u_Orders;
  PyObject *__pyx_n_u_Price;
  PyObject *__pyx_n_s_PriceSize;
  PyObject *__pyx_n_s_PriceType;
  PyObject *__pyx_n_s_Proposal;
  PyObject *__pyx_kp_u_Reservation_Price;
  PyObject *__pyx_n_s_RunAlwaysExecutionState;
  PyObject *__pyx_n_s_RunInTimeConditionalExecutionSta;
  PyObject *__pyx_n_s_SELL;
  PyObject *__pyx_n_u_Spread;
  PyObject *__pyx_n_s_StrategyBase;
  PyObject *__pyx_kp_u_Strategy_parameters;
  PyObject *__pyx_n_s_Timestamp;
  PyObject *__pyx_kp_u_Total_Balance;
  PyObject *__pyx_n_s_TrackHangingOrdersModel;
  PyObject *__pyx_n_s_TradeType;
  PyObject *__pyx_n_s_TradingIntensityIndicator;
  PyObject *__pyx_n_s_Tuple;
  PyObject *__pyx_kp_s_Tuple_List_Proposal_List_Proposa;
  PyObject *__pyx_n_u_Type;
  PyObject *__pyx_n_s_TypeError;
  PyObject *__pyx_n_s_Union;
  PyObject *__pyx_kp_s_Union_AvellanedaMarketMakingConf;
  PyObject *__pyx_kp_u_WARNINGS;
  PyObject *__pyx_kp_u_WARNING_Some_markets_are_not_con;
  PyObject *__pyx_kp_u__10;
  PyObject *__pyx_kp_u__11;
  PyObject *__pyx_kp_u__12;
  PyObject *__pyx_kp_u__13;
  PyObject *__pyx_kp_u__14;
  PyObject *__pyx_kp_u__15;
  PyObject *__pyx_n_s__17;
  PyObject *__pyx_kp_u__18;
  PyObject *__pyx_kp_u__4;
  PyObject *__pyx_kp_u__5;
  PyObject *__pyx_kp_u__6;
  PyObject *__pyx_kp_u__7;
  PyObject *__pyx_kp_u__8;
  PyObject *__pyx_kp_u__9;
  PyObject *__pyx_n_s__96;
  PyObject *__pyx_n_u_a;
  PyObject *__pyx_n_s_active_markets;
  PyObject *__pyx_n_s_active_non_hanging_orders;
  PyObject *__pyx_n_s_active_orders;
  PyObject *__pyx_n_s_active_orders_df;
  PyObject *__pyx_n_s_active_orders_df_locals_lambda;
  PyObject *__pyx_n_s_add_as_hanging_order;
  PyObject *__pyx_n_s_add_current_pairs_of_proposal_or;
  PyObject *__pyx_n_s_add_sample;
  PyObject *__pyx_n_s_add_transaction_costs;
  PyObject *__pyx_n_s_adjusted_available_balance_for_o;
  PyObject *__pyx_n_s_age;
  PyObject *__pyx_n_s_all;
  PyObject *__pyx_n_s_all_markets_ready;
  PyObject *__pyx_n_s_all_non_hanging_orders;
  PyObject *__pyx_n_s_all_orders;
  PyObject *__pyx_kp_u_alpha;
  PyObject *__pyx_n_u_alpha_2;
  PyObject *__pyx_n_s_amount;
  PyObject *__pyx_n_s_amount_orig;
  PyObject *__pyx_n_s_apply;
  PyObject *__pyx_n_s_apply_add_transaction_costs;
  PyObject *__pyx_n_s_apply_budget_constraint;
  PyObject *__pyx_n_s_apply_order_amount_eta_transform;
  PyObject *__pyx_n_s_apply_order_optimization;
  PyObject *__pyx_n_s_apply_order_price_modifiers;
  PyObject *__pyx_n_s_args;
  PyObject *__pyx_n_s_ask_level_spreads;
  PyObject *__pyx_kp_u_ask_orders_at_Size_Price;
  PyObject *__pyx_n_s_ask_price;
  PyObject *__pyx_n_s_assets_df;
  PyObject *__pyx_n_s_asyncio_coroutines;
  PyObject *__pyx_n_s_available_base_balance;
  PyObject *__pyx_n_s_available_quote_balance;
  PyObject *__pyx_n_s_balance_warning;
  PyObject *__pyx_n_s_base_asset;
  PyObject *__pyx_n_s_base_balance;
  PyObject *__pyx_n_s_base_currency;
  PyObject *__pyx_n_s_base_ratio;
  PyObject *__pyx_n_s_base_value;
  PyObject *__pyx_n_s_best_ask;
  PyObject *__pyx_n_u_best_ask;
  PyObject *__pyx_n_s_best_bid;
  PyObject *__pyx_n_u_best_bid;
  PyObject *__pyx_n_s_bid_level_spreads;
  PyObject *__pyx_kp_u_bid_orders_at_Size_Price;
  PyObject *__pyx_n_s_bid_price;
  PyObject *__pyx_n_s_bool;
  PyObject *__pyx_n_u_buy;
  PyObject *__pyx_n_s_buys;
  PyObject *__pyx_n_s_calculate;
  PyObject *__pyx_n_s_calculate_inventory;
  PyObject *__pyx_n_s_calculate_reservation_price_and;
  PyObject *__pyx_n_s_calculate_target_inventory;
  PyObject *__pyx_n_s_cancel_active_orders;
  PyObject *__pyx_n_s_cancel_order;
  PyObject *__pyx_n_s_candidate_hanging_orders;
  PyObject *__pyx_n_s_candidate_hanging_orders_from_pa;
  PyObject *__pyx_n_s_ceil;
  PyObject *__pyx_n_s_class_getitem;
  PyObject *__pyx_n_s_client_order_id;
  PyObject *__pyx_n_s_cline_in_traceback;
  PyObject *__pyx_n_s_clock;
  PyObject *__pyx_n_s_close;
  PyObject *__pyx_n_s_closing_time;
  PyObject *__pyx_n_s_cls;
  PyObject *__pyx_n_s_collect_market_variables;
  PyObject *__pyx_n_s_columns;
  PyObject *__pyx_n_s_config_map;
  PyObject *__pyx_n_s_create_base_proposal;
  PyObject *__pyx_n_s_create_basic_proposal;
  PyObject *__pyx_n_s_create_proposal_based_on_order_l;
  PyObject *__pyx_n_s_create_proposal_based_on_order_o;
  PyObject *__pyx_n_s_current_created_pairs_of_orders;
  PyObject *__pyx_n_u_current_inv;
  PyObject *__pyx_n_s_current_prices;
  PyObject *__pyx_n_s_current_value;
  PyObject *__pyx_n_s_data;
  PyObject *__pyx_n_s_datetime;
  PyObject *__pyx_n_s_debug_csv_path;
  PyObject *__pyx_n_s_decimal;
  PyObject *__pyx_n_s_df;
  PyObject *__pyx_n_s_df_header;
  PyObject *__pyx_n_s_df_lines;
  PyObject *__pyx_n_s_dict;
  PyObject *__pyx_kp_u_disable;
  PyObject *__pyx_n_s_display_name;
  PyObject *__pyx_n_s_dump_debug_variables;
  PyObject *__pyx_kp_u_enable;
  PyObject *__pyx_n_s_end_datetime;
  PyObject *__pyx_n_s_end_time;
  PyObject *__pyx_n_s_end_timestamp;
  PyObject *__pyx_n_s_enumerate;
  PyObject *__pyx_n_s_eta;
  PyObject *__pyx_n_u_eta;
  PyObject *__pyx_n_s_execute_orders_proposal;
  PyObject *__pyx_n_s_execution_mode;
  PyObject *__pyx_n_s_execution_state;
  PyObject *__pyx_n_s_execution_timeframe;
  PyObject *__pyx_n_s_execution_timeframe_mode;
  PyObject *__pyx_n_s_exists;
  PyObject *__pyx_n_s_exp;
  PyObject *__pyx_kp_u_filled;
  PyObject *__pyx_n_s_filled_order_delay;
  PyObject *__pyx_n_s_first_col_length;
  PyObject *__pyx_n_s_float;
  PyObject *__pyx_n_s_floor;
  PyObject *__pyx_n_s_format;
  PyObject *__pyx_n_s_format_status;
  PyObject *__pyx_n_s_formatters;
  PyObject *__pyx_n_s_gamma;
  PyObject *__pyx_n_u_gamma;
  PyObject *__pyx_kp_u_gc;
  PyObject *__pyx_n_s_genexpr;
  PyObject *__pyx_n_s_getLogger;
  PyObject *__pyx_n_s_get_adjusted_available_balance;
  PyObject *__pyx_n_s_get_available_balance;
  PyObject *__pyx_n_s_get_balance;
  PyObject *__pyx_n_s_get_config_map_execution_mode;
  PyObject *__pyx_n_s_get_config_map_hanging_orders;
  PyObject *__pyx_n_s_get_config_map_indicators;
  PyObject *__pyx_n_s_get_last_price;
  PyObject *__pyx_n_s_get_level_spreads;
  PyObject *__pyx_n_s_get_limit_orders;
  PyObject *__pyx_n_s_get_maker_order_type;
  PyObject *__pyx_n_s_get_mid_price;
  PyObject *__pyx_n_s_get_price;
  PyObject *__pyx_n_s_get_price_by_type;
  PyObject *__pyx_n_s_get_price_for_volume;
  PyObject *__pyx_n_s_get_spread;
  PyObject *__pyx_n_s_get_volatility;
  PyObject *__pyx_n_s_getstate;
  PyObject *__pyx_n_u_hang;
  PyObject *__pyx_n_s_hanging_orders_cancel_pct;
  PyObject *__pyx_n_s_hanging_orders_enabled;
  PyObject *__pyx_n_s_hanging_orders_mode;
  PyObject *__pyx_n_s_hanging_orders_tracker;
  PyObject *__pyx_kp_u_has_been_completely_filled;
  PyObject *__pyx_n_s_hb_app_notification;
  PyObject *__pyx_n_s_header;
  PyObject *__pyx_n_s_hummingbot_client_config_config;
  PyObject *__pyx_n_s_hummingbot_connector_exchange_ba;
  PyObject *__pyx_n_s_hummingbot_core_data_type_common;
  PyObject *__pyx_n_s_hummingbot_core_data_type_limit;
  PyObject *__pyx_n_s_hummingbot_core_network_iterator;
  PyObject *__pyx_n_s_hummingbot_core_utils;
  PyObject *__pyx_n_s_hummingbot_strategy___utils___tr;
  PyObject *__pyx_n_s_hummingbot_strategy___utils___tr_2;
  PyObject *__pyx_n_s_hummingbot_strategy_avellaneda_m;
  PyObject *__pyx_n_s_hummingbot_strategy_avellaneda_m_2;
  PyObject *__pyx_kp_s_hummingbot_strategy_avellaneda_m_3;
  PyObject *__pyx_n_s_hummingbot_strategy_conditional;
  PyObject *__pyx_n_s_hummingbot_strategy_data_types;
  PyObject *__pyx_n_s_hummingbot_strategy_hanging_orde;
  PyObject *__pyx_n_s_hummingbot_strategy_market_tradi;
  PyObject *__pyx_n_s_hummingbot_strategy_order_book_a;
  PyObject *__pyx_n_s_hummingbot_strategy_strategy_bas;
  PyObject *__pyx_n_s_hummingbot_strategy_utils;
  PyObject *__pyx_n_s_i;
  PyObject *__pyx_n_s_idx;
  PyObject *__pyx_n_s_import;
  PyObject *__pyx_n_s_in_flight_cancels;
  PyObject *__pyx_n_s_index;
  PyObject *__pyx_n_s_info;
  PyObject *__pyx_n_s_init_params;
  PyObject *__pyx_n_s_initializing;
  PyObject *__pyx_n_s_int;
  PyObject *__pyx_n_s_inventory_target_base;
  PyObject *__pyx_n_s_inventory_target_base_pct;
  PyObject *__pyx_n_u_inventory_target_pct;
  PyObject *__pyx_n_s_is_algorithm_changed;
  PyObject *__pyx_n_s_is_algorithm_ready;
  PyObject *__pyx_n_s_is_buy;
  PyObject *__pyx_n_s_is_coroutine;
  PyObject *__pyx_n_s_is_debug;
  PyObject *__pyx_kp_u_is_filled;
  PyObject *__pyx_n_s_is_hanging_order;
  PyObject *__pyx_n_s_is_order_id_in_completed_hanging;
  PyObject *__pyx_n_s_is_order_id_in_hanging_orders;
  PyObject *__pyx_n_s_is_potential_hanging_order;
  PyObject *__pyx_n_s_is_sampling_buffer_changed;
  PyObject *__pyx_n_s_is_sampling_buffer_full;
  PyObject *__pyx_n_s_is_within_tolerance;
  PyObject *__pyx_kp_u_isenabled;
  PyObject *__pyx_n_s_isnan;
  PyObject *__pyx_n_s_items;
  PyObject *__pyx_kp_u_kappa;
  PyObject *__pyx_n_u_kappa_2;
  PyObject *__pyx_n_s_key;
  PyObject *__pyx_n_s_len;
  PyObject *__pyx_n_s_level;
  PyObject *__pyx_n_s_level_distances;
  PyObject *__pyx_n_s_level_step;
  PyObject *__pyx_n_s_limit_orders;
  PyObject *__pyx_n_s_line;
  PyObject *__pyx_n_s_lines;
  PyObject *__pyx_n_s_ln;
  PyObject *__pyx_n_s_log_with_clock;
  PyObject *__pyx_n_s_logger;
  PyObject *__pyx_n_s_logging;
  PyObject *__pyx_n_s_logging_options;
  PyObject *__pyx_n_s_lvl_buy;
  PyObject *__pyx_n_s_lvl_sell;
  PyObject *__pyx_n_s_main;
  PyObject *__pyx_n_s_map_df_to_str;
  PyObject *__pyx_n_s_market;
  PyObject *__pyx_n_s_market_books;
  PyObject *__pyx_n_s_market_info;
  PyObject *__pyx_n_s_market_info_to_active_orders;
  PyObject *__pyx_n_s_market_pair_to_active_orders;
  PyObject *__pyx_n_s_market_status_data_frame;
  PyObject *__pyx_n_s_market_trading_pair_tuples;
  PyObject *__pyx_n_s_markets_columns;
  PyObject *__pyx_n_s_markets_data;
  PyObject *__pyx_n_s_markets_df;
  PyObject *__pyx_n_s_math;
  PyObject *__pyx_n_s_max;
  PyObject *__pyx_n_s_max_order_age;
  PyObject *__pyx_n_s_measure_order_book_liquidity;
  PyObject *__pyx_kp_u_mid_price;
  PyObject *__pyx_n_s_mid_price_2;
  PyObject *__pyx_n_u_mid_price_2;
  PyObject *__pyx_kp_u_mid_price_std_dev;
  PyObject *__pyx_n_s_mid_price_variance;
  PyObject *__pyx_n_u_mid_price_variance;
  PyObject *__pyx_n_s_min_spread;
  PyObject *__pyx_n_s_mode;
  PyObject *__pyx_n_s_model_config;
  PyObject *__pyx_n_s_msg;
  PyObject *__pyx_n_s_name;
  PyObject *__pyx_n_s_nan;
  PyObject *__pyx_n_u_nan;
  PyObject *__pyx_n_s_network_status;
  PyObject *__pyx_n_s_network_warning;
  PyObject *__pyx_n_s_new_ask;
  PyObject *__pyx_n_s_new_bid;
  PyObject *__pyx_n_s_no_sells;
  PyObject *__pyx_n_s_non_hanging;
  PyObject *__pyx_n_s_normalize;
  PyObject *__pyx_n_s_notify_hb_app;
  PyObject *__pyx_n_s_notify_hb_app_with_timestamp;
  PyObject *__pyx_n_s_np;
  PyObject *__pyx_n_s_numpy;
  PyObject *__pyx_kp_u_numpy_core_multiarray_failed_to;
  PyObject *__pyx_kp_u_numpy_core_umath_failed_to_impor;
  PyObject *__pyx_n_s_o;
  PyObject *__pyx_kp_u_optimal_ask;
  PyObject *__pyx_n_u_optimal_ask_2;
  PyObject *__pyx_kp_u_optimal_ask_to_mid;
  PyObject *__pyx_kp_u_optimal_bid;
  PyObject *__pyx_n_u_optimal_bid_2;
  PyObject *__pyx_kp_u_optimal_bid_to_mid;
  PyObject *__pyx_kp_u_optimal_spread;
  PyObject *__pyx_n_u_optimal_spread_2;
  PyObject *__pyx_kp_u_order;
  PyObject *__pyx_n_s_order_2;
  PyObject *__pyx_n_s_order_age;
  PyObject *__pyx_n_s_order_amount;
  PyObject *__pyx_n_s_order_amount_shape_factor;
  PyObject *__pyx_n_s_order_book;
  PyObject *__pyx_kp_u_order_book_depth_factor;
  PyObject *__pyx_kp_u_order_book_intensity_factor;
  PyObject *__pyx_n_s_order_id;
  PyObject *__pyx_n_s_order_levels;
  PyObject *__pyx_n_s_order_levels_mode;
  PyObject *__pyx_n_s_order_optimization_enabled;
  PyObject *__pyx_n_s_order_override;
  PyObject *__pyx_n_s_order_refresh_time;
  PyObject *__pyx_n_s_order_refresh_tolerance;
  PyObject *__pyx_n_s_order_refresh_tolerance_pct;
  PyObject *__pyx_n_s_orders;
  PyObject *__pyx_n_s_os;
  PyObject *__pyx_n_s_p;
  PyObject *__pyx_n_s_pandas;
  PyObject *__pyx_n_s_path;
  PyObject *__pyx_n_s_pd;
  PyObject *__pyx_kp_s_pd_DataFrame;
  PyObject *__pyx_n_s_percent;
  PyObject *__pyx_n_s_pmm_logger;
  PyObject *__pyx_n_s_price;
  PyObject *__pyx_n_s_price_delegate;
  PyObject *__pyx_n_s_process_tick;
  PyObject *__pyx_n_s_proposal;
  PyObject *__pyx_n_s_proposal_prices;
  PyObject *__pyx_n_s_pure_mm_assets_df;
  PyObject *__pyx_n_s_pyx_state;
  PyObject *__pyx_n_s_pyx_vtable;
  PyObject *__pyx_kp_u_q;
  PyObject *__pyx_n_s_quantity;
  PyObject *__pyx_n_s_quote_asset;
  PyObject *__pyx_n_s_quote_balance;
  PyObject *__pyx_n_s_quote_currency;
  PyObject *__pyx_n_s_quote_ratio;
  PyObject *__pyx_n_s_range;
  PyObject *__pyx_n_s_ready;
  PyObject *__pyx_n_s_reduce;
  PyObject *__pyx_n_s_reduce_cython;
  PyObject *__pyx_n_s_reduce_ex;
  PyObject *__pyx_n_s_ref_price;
  PyObject *__pyx_n_s_regex;
  PyObject *__pyx_n_s_register_events;
  PyObject *__pyx_n_s_replace;
  PyObject *__pyx_kp_u_reservation_price;
  PyObject *__pyx_n_u_reservation_price_2;
  PyObject *__pyx_n_s_result_price;
  PyObject *__pyx_n_s_return;
  PyObject *__pyx_n_s_reverse;
  PyObject *__pyx_n_s_risk_factor;
  PyObject *__pyx_n_u_risk_factor;
  PyObject *__pyx_kp_u_risk_factor_2;
  PyObject *__pyx_n_s_round;
  PyObject *__pyx_n_u_s;
  PyObject *__pyx_n_s_s_decimal_neg_one;
  PyObject *__pyx_n_s_s_decimal_one;
  PyObject *__pyx_n_s_s_decimal_zero;
  PyObject *__pyx_n_s_sampling_length;
  PyObject *__pyx_n_s_seconds;
  PyObject *__pyx_n_s_self;
  PyObject *__pyx_kp_s_self__events_cannot_be_converted;
  PyObject *__pyx_n_u_sell;
  PyObject *__pyx_n_s_sell_order;
  PyObject *__pyx_n_s_sells;
  PyObject *__pyx_n_s_send;
  PyObject *__pyx_n_s_set_timers;
  PyObject *__pyx_n_s_setstate;
  PyObject *__pyx_n_s_setstate_cython;
  PyObject *__pyx_n_s_should_wait_order_cancel_confirm;
  PyObject *__pyx_n_s_size;
  PyObject *__pyx_n_s_sort;
  PyObject *__pyx_n_s_sorted;
  PyObject *__pyx_n_s_spec;
  PyObject *__pyx_n_s_split;
  PyObject *__pyx_n_s_spread;
  PyObject *__pyx_n_s_start;
  PyObject *__pyx_n_s_start_datetime;
  PyObject *__pyx_n_s_start_time;
  PyObject *__pyx_n_s_start_timestamp;
  PyObject *__pyx_n_s_status_report_interval;
  PyObject *__pyx_n_s_str;
  PyObject *__pyx_n_s_strftime;
  PyObject *__pyx_kp_s_stringsource;
  PyObject *__pyx_n_s_super;
  PyObject *__pyx_n_u_target_inv;
  PyObject *__pyx_n_s_test;
  PyObject *__pyx_n_s_throw;
  PyObject *__pyx_n_s_ticks_to_be_ready_after;
  PyObject *__pyx_n_s_ticks_to_be_ready_before;
  PyObject *__pyx_kp_u_ticks_to_fill_buffers;
  PyObject *__pyx_n_s_time;
  PyObject *__pyx_n_s_time_left;
  PyObject *__pyx_n_s_time_left_fraction;
  PyObject *__pyx_n_u_time_left_fraction;
  PyObject *__pyx_kp_u_time_until_end_of_trading_cycle;
  PyObject *__pyx_kp_u_time_until_end_of_trading_cycle_2;
  PyObject *__pyx_n_s_timedelta;
  PyObject *__pyx_n_s_timestamp;
  PyObject *__pyx_n_s_title;
  PyObject *__pyx_n_u_title;
  PyObject *__pyx_n_s_to_create_orders;
  PyObject *__pyx_n_s_to_csv;
  PyObject *__pyx_n_s_to_show_current_pct;
  PyObject *__pyx_n_s_to_string;
  PyObject *__pyx_n_s_total_in_quote;
  PyObject *__pyx_n_s_trade_type;
  PyObject *__pyx_n_s_trading_intensity_buffer_size;
  PyObject *__pyx_n_s_trading_pair;
  PyObject *__pyx_n_s_typing;
  PyObject *__pyx_n_s_unit;
  PyObject *__pyx_n_s_unlink;
  PyObject *__pyx_n_s_unregister_events;
  PyObject *__pyx_n_s_update_from_config_map;
  PyObject *__pyx_n_s_update_strategy_orders_with_equi;
  PyObject *__pyx_n_s_upper;
  PyObject *__pyx_n_s_values;
  PyObject *__pyx_kp_u_vol;
  PyObject *__pyx_n_s_vol_2;
  PyObject *__pyx_kp_u_volatility;
  PyObject *__pyx_n_u_volatility_2;
  PyObject *__pyx_n_s_volatility_buffer_size;
  PyObject *__pyx_n_s_volatility_pct;
  PyObject *__pyx_n_s_warning;
  PyObject *__pyx_n_s_warning_lines;
  PyObject *__pyx_n_s_x;
  PyObject *__pyx_n_s_zip;
  PyObject *__pyx_float_100_0;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_1;
  PyObject *__pyx_int_2;
  PyObject *__pyx_int_4;
  PyObject *__pyx_int_5;
  PyObject *__pyx_int_8;
  PyObject *__pyx_int_16;
  PyObject *__pyx_int_32;
  PyObject *__pyx_int_100;
  PyObject *__pyx_int_900;
  PyObject *__pyx_int_0x7fffffffffffffff;
  PyObject *__pyx_int_neg_1;
  PyObject *__pyx_k__3;
  PyObject *__pyx_tuple_;
  PyObject *__pyx_tuple__2;
  PyObject *__pyx_tuple__16;
  PyObject *__pyx_tuple__19;
  PyObject *__pyx_tuple__20;
  PyObject *__pyx_tuple__21;
  PyObject *__pyx_tuple__22;
  PyObject *__pyx_tuple__24;
  PyObject *__pyx_tuple__26;
  PyObject *__pyx_tuple__28;
  PyObject *__pyx_tuple__33;
  PyObject *__pyx_tuple__35;
  PyObject *__pyx_tuple__37;
  PyObject *__pyx_tuple__39;
  PyObject *__pyx_tuple__41;
  PyObject *__pyx_tuple__43;
  PyObject *__pyx_tuple__45;
  PyObject *__pyx_tuple__47;
  PyObject *__pyx_tuple__49;
  PyObject *__pyx_tuple__51;
  PyObject *__pyx_tuple__53;
  PyObject *__pyx_tuple__55;
  PyObject *__pyx_tuple__58;
  PyObject *__pyx_tuple__66;
  PyObject *__pyx_tuple__72;
  PyObject *__pyx_tuple__76;
  PyObject *__pyx_tuple__82;
  PyObject *__pyx_tuple__85;
  PyObject *__pyx_tuple__89;
  PyObject *__pyx_tuple__91;
  PyObject *__pyx_tuple__94;
  PyObject *__pyx_codeobj__23;
  PyObject *__pyx_codeobj__25;
  PyObject *__pyx_codeobj__27;
  PyObject *__pyx_codeobj__29;
  PyObject *__pyx_codeobj__30;
  PyObject *__pyx_codeobj__31;
  PyObject *__pyx_codeobj__32;
  PyObject *__pyx_codeobj__34;
  PyObject *__pyx_codeobj__36;
  PyObject *__pyx_codeobj__38;
  PyObject *__pyx_codeobj__40;
  PyObject *__pyx_codeobj__42;
  PyObject *__pyx_codeobj__44;
  PyObject *__pyx_codeobj__46;
  PyObject *__pyx_codeobj__48;
  PyObject *__pyx_codeobj__50;
  PyObject *__pyx_codeobj__52;
  PyObject *__pyx_codeobj__54;
  PyObject *__pyx_codeobj__56;
  PyObject *__pyx_codeobj__57;
  PyObject *__pyx_codeobj__59;
  PyObject *__pyx_codeobj__60;
  PyObject *__pyx_codeobj__61;
  PyObject *__pyx_codeobj__62;
  PyObject *__pyx_codeobj__63;
  PyObject *__pyx_codeobj__64;
  PyObject *__pyx_codeobj__65;
  PyObject *__pyx_codeobj__67;
  PyObject *__pyx_codeobj__68;
  PyObject *__pyx_codeobj__69;
  PyObject *__pyx_codeobj__70;
  PyObject *__pyx_codeobj__71;
  PyObject *__pyx_codeobj__73;
  PyObject *__pyx_codeobj__74;
  PyObject *__pyx_codeobj__75;
  PyObject *__pyx_codeobj__77;
  PyObject *__pyx_codeobj__78;
  PyObject *__pyx_codeobj__79;
  PyObject *__pyx_codeobj__80;
  PyObject *__pyx_codeobj__81;
  PyObject *__pyx_codeobj__83;
  PyObject *__pyx_codeobj__84;
  PyObject *__pyx_codeobj__86;
  PyObject *__pyx_codeobj__87;
  PyObject *__pyx_codeobj__88;
  PyObject *__pyx_codeobj__90;
  PyObject *__pyx_codeobj__92;
  PyObject *__pyx_codeobj__93;
  PyObject *__pyx_codeobj__95;
} __pyx_mstate;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
  extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate(o) ((__pyx_mstate *)__Pyx_PyModule_GetState(o))

#define __pyx_mstate_global (__pyx_mstate(PyState_FindModule(&__pyx_moduledef)))

#define __pyx_m (PyState_FindModule(&__pyx_moduledef))
#else
static __pyx_mstate __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstate *__pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_clear(PyObject *m) {
  __pyx_mstate *clear_module_state = __pyx_mstate(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_6pubsub_PubSub);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult);
  Py_CLEAR(clear_module_state->__pyx_ptype_7cpython_4type_type);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_dtype);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_flatiter);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_broadcast);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_ndarray);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_generic);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_number);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_integer);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_signedinteger);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_unsignedinteger);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_inexact);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_floating);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_complexfloating);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_flexible);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_character);
  Py_CLEAR(clear_module_state->__pyx_ptype_5numpy_ufunc);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradesForwarder);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_5clock_Clock);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_13order_tracker_OrderTracker);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr);
  Py_CLEAR(clear_module_state->__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr);
  Py_CLEAR(clear_module_state->__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr);
  Py_CLEAR(clear_module_state->__pyx_kp_u_0);
  Py_CLEAR(clear_module_state->__pyx_kp_u_1);
  Py_CLEAR(clear_module_state->__pyx_kp_u_10);
  Py_CLEAR(clear_module_state->__pyx_kp_u_100);
  Py_CLEAR(clear_module_state->__pyx_kp_u_10f);
  Py_CLEAR(clear_module_state->__pyx_kp_u_1_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_3f);
  Py_CLEAR(clear_module_state->__pyx_kp_u_4f);
  Py_CLEAR(clear_module_state->__pyx_kp_u_5E);
  Py_CLEAR(clear_module_state->__pyx_n_u_Age);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Amount_Adj);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Amount_Orig);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Assets);
  Py_CLEAR(clear_module_state->__pyx_n_s_AttributeError);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Available_Balance);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingConfigMap);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_4);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_a);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_a_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_a_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_a_4);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_a_5);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_a_6);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_a_7);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_a_8);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_10);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_11);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_12);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_13);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_4);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_5);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_6);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_7);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_8);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_9);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_d);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_e);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_f);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_g);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_g_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_g_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_g_4);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_g_5);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_g_6);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_g_7);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_g_8);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_g_9);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_i);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_i_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_i_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_i_4);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_l);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_m);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_m_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_n);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_p);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_p_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_s_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_t);
  Py_CLEAR(clear_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_u);
  Py_CLEAR(clear_module_state->__pyx_n_s_BUY);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Best_Ask);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Best_Bid);
  Py_CLEAR(clear_module_state->__pyx_n_s_CONNECTED);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Calculating_volatility_estimatin);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Calculating_volatility_estimatin_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_ClientConfigAdapter);
  Py_CLEAR(clear_module_state->__pyx_n_s_Clock);
  Py_CLEAR(clear_module_state->__pyx_n_s_CreatedPairOfOrders);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Creating);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Current);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Current_Value);
  Py_CLEAR(clear_module_state->__pyx_n_s_DailyBetweenTimesModel);
  Py_CLEAR(clear_module_state->__pyx_n_s_DataFrame);
  Py_CLEAR(clear_module_state->__pyx_n_s_Decimal);
  Py_CLEAR(clear_module_state->__pyx_n_s_Dict);
  Py_CLEAR(clear_module_state->__pyx_n_u_Exchange);
  Py_CLEAR(clear_module_state->__pyx_n_s_ExchangeBase);
  Py_CLEAR(clear_module_state->__pyx_n_s_FileNotFoundError);
  Py_CLEAR(clear_module_state->__pyx_n_s_FromDateToDateModel);
  Py_CLEAR(clear_module_state->__pyx_kp_u_H_M_S);
  Py_CLEAR(clear_module_state->__pyx_n_s_HangingOrdersTracker);
  Py_CLEAR(clear_module_state->__pyx_n_s_INFO);
  Py_CLEAR(clear_module_state->__pyx_n_s_ImportError);
  Py_CLEAR(clear_module_state->__pyx_n_s_InstantVolatilityIndicator);
  Py_CLEAR(clear_module_state->__pyx_n_s_LIMIT);
  Py_CLEAR(clear_module_state->__pyx_n_u_Level);
  Py_CLEAR(clear_module_state->__pyx_n_s_LimitOrder);
  Py_CLEAR(clear_module_state->__pyx_n_s_List);
  Py_CLEAR(clear_module_state->__pyx_kp_s_List_Decimal);
  Py_CLEAR(clear_module_state->__pyx_kp_s_List_LimitOrder);
  Py_CLEAR(clear_module_state->__pyx_kp_s_List_MarketTradingPairTuple);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Maker);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Maker_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Maker_buy_order_of);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Maker_sell_order_of);
  Py_CLEAR(clear_module_state->__pyx_n_u_Market);
  Py_CLEAR(clear_module_state->__pyx_n_s_MarketTradingPairTuple);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Market_connectors_are_not_ready);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Markets);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Markets_are_not_ready_No_market);
  Py_CLEAR(clear_module_state->__pyx_n_s_MidPrice);
  Py_CLEAR(clear_module_state->__pyx_n_u_MidPrice);
  Py_CLEAR(clear_module_state->__pyx_n_s_MultiOrderLevelModel);
  Py_CLEAR(clear_module_state->__pyx_n_s_NaN);
  Py_CLEAR(clear_module_state->__pyx_n_s_NetworkStatus);
  Py_CLEAR(clear_module_state->__pyx_kp_u_No_active_maker_orders);
  Py_CLEAR(clear_module_state->__pyx_kp_u_None);
  Py_CLEAR(clear_module_state->__pyx_n_s_OPTION_LOG_ALL);
  Py_CLEAR(clear_module_state->__pyx_n_s_OPTION_LOG_CREATE_ORDER);
  Py_CLEAR(clear_module_state->__pyx_n_s_OPTION_LOG_MAKER_ORDER_FILLED);
  Py_CLEAR(clear_module_state->__pyx_n_s_OPTION_LOG_STATUS_REPORT);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Optimal_Spread);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderBookAssetPriceDelegate);
  Py_CLEAR(clear_module_state->__pyx_n_s_OrderType);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Orders);
  Py_CLEAR(clear_module_state->__pyx_n_u_Price);
  Py_CLEAR(clear_module_state->__pyx_n_s_PriceSize);
  Py_CLEAR(clear_module_state->__pyx_n_s_PriceType);
  Py_CLEAR(clear_module_state->__pyx_n_s_Proposal);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Reservation_Price);
  Py_CLEAR(clear_module_state->__pyx_n_s_RunAlwaysExecutionState);
  Py_CLEAR(clear_module_state->__pyx_n_s_RunInTimeConditionalExecutionSta);
  Py_CLEAR(clear_module_state->__pyx_n_s_SELL);
  Py_CLEAR(clear_module_state->__pyx_n_u_Spread);
  Py_CLEAR(clear_module_state->__pyx_n_s_StrategyBase);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Strategy_parameters);
  Py_CLEAR(clear_module_state->__pyx_n_s_Timestamp);
  Py_CLEAR(clear_module_state->__pyx_kp_u_Total_Balance);
  Py_CLEAR(clear_module_state->__pyx_n_s_TrackHangingOrdersModel);
  Py_CLEAR(clear_module_state->__pyx_n_s_TradeType);
  Py_CLEAR(clear_module_state->__pyx_n_s_TradingIntensityIndicator);
  Py_CLEAR(clear_module_state->__pyx_n_s_Tuple);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Tuple_List_Proposal_List_Proposa);
  Py_CLEAR(clear_module_state->__pyx_n_u_Type);
  Py_CLEAR(clear_module_state->__pyx_n_s_TypeError);
  Py_CLEAR(clear_module_state->__pyx_n_s_Union);
  Py_CLEAR(clear_module_state->__pyx_kp_s_Union_AvellanedaMarketMakingConf);
  Py_CLEAR(clear_module_state->__pyx_kp_u_WARNINGS);
  Py_CLEAR(clear_module_state->__pyx_kp_u_WARNING_Some_markets_are_not_con);
  Py_CLEAR(clear_module_state->__pyx_kp_u__10);
  Py_CLEAR(clear_module_state->__pyx_kp_u__11);
  Py_CLEAR(clear_module_state->__pyx_kp_u__12);
  Py_CLEAR(clear_module_state->__pyx_kp_u__13);
  Py_CLEAR(clear_module_state->__pyx_kp_u__14);
  Py_CLEAR(clear_module_state->__pyx_kp_u__15);
  Py_CLEAR(clear_module_state->__pyx_n_s__17);
  Py_CLEAR(clear_module_state->__pyx_kp_u__18);
  Py_CLEAR(clear_module_state->__pyx_kp_u__4);
  Py_CLEAR(clear_module_state->__pyx_kp_u__5);
  Py_CLEAR(clear_module_state->__pyx_kp_u__6);
  Py_CLEAR(clear_module_state->__pyx_kp_u__7);
  Py_CLEAR(clear_module_state->__pyx_kp_u__8);
  Py_CLEAR(clear_module_state->__pyx_kp_u__9);
  Py_CLEAR(clear_module_state->__pyx_n_s__96);
  Py_CLEAR(clear_module_state->__pyx_n_u_a);
  Py_CLEAR(clear_module_state->__pyx_n_s_active_markets);
  Py_CLEAR(clear_module_state->__pyx_n_s_active_non_hanging_orders);
  Py_CLEAR(clear_module_state->__pyx_n_s_active_orders);
  Py_CLEAR(clear_module_state->__pyx_n_s_active_orders_df);
  Py_CLEAR(clear_module_state->__pyx_n_s_active_orders_df_locals_lambda);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_as_hanging_order);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_current_pairs_of_proposal_or);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_sample);
  Py_CLEAR(clear_module_state->__pyx_n_s_add_transaction_costs);
  Py_CLEAR(clear_module_state->__pyx_n_s_adjusted_available_balance_for_o);
  Py_CLEAR(clear_module_state->__pyx_n_s_age);
  Py_CLEAR(clear_module_state->__pyx_n_s_all);
  Py_CLEAR(clear_module_state->__pyx_n_s_all_markets_ready);
  Py_CLEAR(clear_module_state->__pyx_n_s_all_non_hanging_orders);
  Py_CLEAR(clear_module_state->__pyx_n_s_all_orders);
  Py_CLEAR(clear_module_state->__pyx_kp_u_alpha);
  Py_CLEAR(clear_module_state->__pyx_n_u_alpha_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_amount);
  Py_CLEAR(clear_module_state->__pyx_n_s_amount_orig);
  Py_CLEAR(clear_module_state->__pyx_n_s_apply);
  Py_CLEAR(clear_module_state->__pyx_n_s_apply_add_transaction_costs);
  Py_CLEAR(clear_module_state->__pyx_n_s_apply_budget_constraint);
  Py_CLEAR(clear_module_state->__pyx_n_s_apply_order_amount_eta_transform);
  Py_CLEAR(clear_module_state->__pyx_n_s_apply_order_optimization);
  Py_CLEAR(clear_module_state->__pyx_n_s_apply_order_price_modifiers);
  Py_CLEAR(clear_module_state->__pyx_n_s_args);
  Py_CLEAR(clear_module_state->__pyx_n_s_ask_level_spreads);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ask_orders_at_Size_Price);
  Py_CLEAR(clear_module_state->__pyx_n_s_ask_price);
  Py_CLEAR(clear_module_state->__pyx_n_s_assets_df);
  Py_CLEAR(clear_module_state->__pyx_n_s_asyncio_coroutines);
  Py_CLEAR(clear_module_state->__pyx_n_s_available_base_balance);
  Py_CLEAR(clear_module_state->__pyx_n_s_available_quote_balance);
  Py_CLEAR(clear_module_state->__pyx_n_s_balance_warning);
  Py_CLEAR(clear_module_state->__pyx_n_s_base_asset);
  Py_CLEAR(clear_module_state->__pyx_n_s_base_balance);
  Py_CLEAR(clear_module_state->__pyx_n_s_base_currency);
  Py_CLEAR(clear_module_state->__pyx_n_s_base_ratio);
  Py_CLEAR(clear_module_state->__pyx_n_s_base_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_best_ask);
  Py_CLEAR(clear_module_state->__pyx_n_u_best_ask);
  Py_CLEAR(clear_module_state->__pyx_n_s_best_bid);
  Py_CLEAR(clear_module_state->__pyx_n_u_best_bid);
  Py_CLEAR(clear_module_state->__pyx_n_s_bid_level_spreads);
  Py_CLEAR(clear_module_state->__pyx_kp_u_bid_orders_at_Size_Price);
  Py_CLEAR(clear_module_state->__pyx_n_s_bid_price);
  Py_CLEAR(clear_module_state->__pyx_n_s_bool);
  Py_CLEAR(clear_module_state->__pyx_n_u_buy);
  Py_CLEAR(clear_module_state->__pyx_n_s_buys);
  Py_CLEAR(clear_module_state->__pyx_n_s_calculate);
  Py_CLEAR(clear_module_state->__pyx_n_s_calculate_inventory);
  Py_CLEAR(clear_module_state->__pyx_n_s_calculate_reservation_price_and);
  Py_CLEAR(clear_module_state->__pyx_n_s_calculate_target_inventory);
  Py_CLEAR(clear_module_state->__pyx_n_s_cancel_active_orders);
  Py_CLEAR(clear_module_state->__pyx_n_s_cancel_order);
  Py_CLEAR(clear_module_state->__pyx_n_s_candidate_hanging_orders);
  Py_CLEAR(clear_module_state->__pyx_n_s_candidate_hanging_orders_from_pa);
  Py_CLEAR(clear_module_state->__pyx_n_s_ceil);
  Py_CLEAR(clear_module_state->__pyx_n_s_class_getitem);
  Py_CLEAR(clear_module_state->__pyx_n_s_client_order_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_cline_in_traceback);
  Py_CLEAR(clear_module_state->__pyx_n_s_clock);
  Py_CLEAR(clear_module_state->__pyx_n_s_close);
  Py_CLEAR(clear_module_state->__pyx_n_s_closing_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_cls);
  Py_CLEAR(clear_module_state->__pyx_n_s_collect_market_variables);
  Py_CLEAR(clear_module_state->__pyx_n_s_columns);
  Py_CLEAR(clear_module_state->__pyx_n_s_config_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_base_proposal);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_basic_proposal);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_proposal_based_on_order_l);
  Py_CLEAR(clear_module_state->__pyx_n_s_create_proposal_based_on_order_o);
  Py_CLEAR(clear_module_state->__pyx_n_s_current_created_pairs_of_orders);
  Py_CLEAR(clear_module_state->__pyx_n_u_current_inv);
  Py_CLEAR(clear_module_state->__pyx_n_s_current_prices);
  Py_CLEAR(clear_module_state->__pyx_n_s_current_value);
  Py_CLEAR(clear_module_state->__pyx_n_s_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_datetime);
  Py_CLEAR(clear_module_state->__pyx_n_s_debug_csv_path);
  Py_CLEAR(clear_module_state->__pyx_n_s_decimal);
  Py_CLEAR(clear_module_state->__pyx_n_s_df);
  Py_CLEAR(clear_module_state->__pyx_n_s_df_header);
  Py_CLEAR(clear_module_state->__pyx_n_s_df_lines);
  Py_CLEAR(clear_module_state->__pyx_n_s_dict);
  Py_CLEAR(clear_module_state->__pyx_kp_u_disable);
  Py_CLEAR(clear_module_state->__pyx_n_s_display_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_dump_debug_variables);
  Py_CLEAR(clear_module_state->__pyx_kp_u_enable);
  Py_CLEAR(clear_module_state->__pyx_n_s_end_datetime);
  Py_CLEAR(clear_module_state->__pyx_n_s_end_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_end_timestamp);
  Py_CLEAR(clear_module_state->__pyx_n_s_enumerate);
  Py_CLEAR(clear_module_state->__pyx_n_s_eta);
  Py_CLEAR(clear_module_state->__pyx_n_u_eta);
  Py_CLEAR(clear_module_state->__pyx_n_s_execute_orders_proposal);
  Py_CLEAR(clear_module_state->__pyx_n_s_execution_mode);
  Py_CLEAR(clear_module_state->__pyx_n_s_execution_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_execution_timeframe);
  Py_CLEAR(clear_module_state->__pyx_n_s_execution_timeframe_mode);
  Py_CLEAR(clear_module_state->__pyx_n_s_exists);
  Py_CLEAR(clear_module_state->__pyx_n_s_exp);
  Py_CLEAR(clear_module_state->__pyx_kp_u_filled);
  Py_CLEAR(clear_module_state->__pyx_n_s_filled_order_delay);
  Py_CLEAR(clear_module_state->__pyx_n_s_first_col_length);
  Py_CLEAR(clear_module_state->__pyx_n_s_float);
  Py_CLEAR(clear_module_state->__pyx_n_s_floor);
  Py_CLEAR(clear_module_state->__pyx_n_s_format);
  Py_CLEAR(clear_module_state->__pyx_n_s_format_status);
  Py_CLEAR(clear_module_state->__pyx_n_s_formatters);
  Py_CLEAR(clear_module_state->__pyx_n_s_gamma);
  Py_CLEAR(clear_module_state->__pyx_n_u_gamma);
  Py_CLEAR(clear_module_state->__pyx_kp_u_gc);
  Py_CLEAR(clear_module_state->__pyx_n_s_genexpr);
  Py_CLEAR(clear_module_state->__pyx_n_s_getLogger);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_adjusted_available_balance);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_available_balance);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_balance);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_config_map_execution_mode);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_config_map_hanging_orders);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_config_map_indicators);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_last_price);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_level_spreads);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_limit_orders);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_maker_order_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_mid_price);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_price);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_price_by_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_price_for_volume);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_spread);
  Py_CLEAR(clear_module_state->__pyx_n_s_get_volatility);
  Py_CLEAR(clear_module_state->__pyx_n_s_getstate);
  Py_CLEAR(clear_module_state->__pyx_n_u_hang);
  Py_CLEAR(clear_module_state->__pyx_n_s_hanging_orders_cancel_pct);
  Py_CLEAR(clear_module_state->__pyx_n_s_hanging_orders_enabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_hanging_orders_mode);
  Py_CLEAR(clear_module_state->__pyx_n_s_hanging_orders_tracker);
  Py_CLEAR(clear_module_state->__pyx_kp_u_has_been_completely_filled);
  Py_CLEAR(clear_module_state->__pyx_n_s_hb_app_notification);
  Py_CLEAR(clear_module_state->__pyx_n_s_header);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_client_config_config);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_connector_exchange_ba);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_core_data_type_common);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_core_data_type_limit);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_core_network_iterator);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_core_utils);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_strategy___utils___tr);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_strategy___utils___tr_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_strategy_avellaneda_m);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_strategy_avellaneda_m_2);
  Py_CLEAR(clear_module_state->__pyx_kp_s_hummingbot_strategy_avellaneda_m_3);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_strategy_conditional);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_strategy_data_types);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_strategy_hanging_orde);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_strategy_market_tradi);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_strategy_order_book_a);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_strategy_strategy_bas);
  Py_CLEAR(clear_module_state->__pyx_n_s_hummingbot_strategy_utils);
  Py_CLEAR(clear_module_state->__pyx_n_s_i);
  Py_CLEAR(clear_module_state->__pyx_n_s_idx);
  Py_CLEAR(clear_module_state->__pyx_n_s_import);
  Py_CLEAR(clear_module_state->__pyx_n_s_in_flight_cancels);
  Py_CLEAR(clear_module_state->__pyx_n_s_index);
  Py_CLEAR(clear_module_state->__pyx_n_s_info);
  Py_CLEAR(clear_module_state->__pyx_n_s_init_params);
  Py_CLEAR(clear_module_state->__pyx_n_s_initializing);
  Py_CLEAR(clear_module_state->__pyx_n_s_int);
  Py_CLEAR(clear_module_state->__pyx_n_s_inventory_target_base);
  Py_CLEAR(clear_module_state->__pyx_n_s_inventory_target_base_pct);
  Py_CLEAR(clear_module_state->__pyx_n_u_inventory_target_pct);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_algorithm_changed);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_algorithm_ready);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_buy);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_coroutine);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_debug);
  Py_CLEAR(clear_module_state->__pyx_kp_u_is_filled);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_hanging_order);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_order_id_in_completed_hanging);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_order_id_in_hanging_orders);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_potential_hanging_order);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_sampling_buffer_changed);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_sampling_buffer_full);
  Py_CLEAR(clear_module_state->__pyx_n_s_is_within_tolerance);
  Py_CLEAR(clear_module_state->__pyx_kp_u_isenabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_isnan);
  Py_CLEAR(clear_module_state->__pyx_n_s_items);
  Py_CLEAR(clear_module_state->__pyx_kp_u_kappa);
  Py_CLEAR(clear_module_state->__pyx_n_u_kappa_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_key);
  Py_CLEAR(clear_module_state->__pyx_n_s_len);
  Py_CLEAR(clear_module_state->__pyx_n_s_level);
  Py_CLEAR(clear_module_state->__pyx_n_s_level_distances);
  Py_CLEAR(clear_module_state->__pyx_n_s_level_step);
  Py_CLEAR(clear_module_state->__pyx_n_s_limit_orders);
  Py_CLEAR(clear_module_state->__pyx_n_s_line);
  Py_CLEAR(clear_module_state->__pyx_n_s_lines);
  Py_CLEAR(clear_module_state->__pyx_n_s_ln);
  Py_CLEAR(clear_module_state->__pyx_n_s_log_with_clock);
  Py_CLEAR(clear_module_state->__pyx_n_s_logger);
  Py_CLEAR(clear_module_state->__pyx_n_s_logging);
  Py_CLEAR(clear_module_state->__pyx_n_s_logging_options);
  Py_CLEAR(clear_module_state->__pyx_n_s_lvl_buy);
  Py_CLEAR(clear_module_state->__pyx_n_s_lvl_sell);
  Py_CLEAR(clear_module_state->__pyx_n_s_main);
  Py_CLEAR(clear_module_state->__pyx_n_s_map_df_to_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_market);
  Py_CLEAR(clear_module_state->__pyx_n_s_market_books);
  Py_CLEAR(clear_module_state->__pyx_n_s_market_info);
  Py_CLEAR(clear_module_state->__pyx_n_s_market_info_to_active_orders);
  Py_CLEAR(clear_module_state->__pyx_n_s_market_pair_to_active_orders);
  Py_CLEAR(clear_module_state->__pyx_n_s_market_status_data_frame);
  Py_CLEAR(clear_module_state->__pyx_n_s_market_trading_pair_tuples);
  Py_CLEAR(clear_module_state->__pyx_n_s_markets_columns);
  Py_CLEAR(clear_module_state->__pyx_n_s_markets_data);
  Py_CLEAR(clear_module_state->__pyx_n_s_markets_df);
  Py_CLEAR(clear_module_state->__pyx_n_s_math);
  Py_CLEAR(clear_module_state->__pyx_n_s_max);
  Py_CLEAR(clear_module_state->__pyx_n_s_max_order_age);
  Py_CLEAR(clear_module_state->__pyx_n_s_measure_order_book_liquidity);
  Py_CLEAR(clear_module_state->__pyx_kp_u_mid_price);
  Py_CLEAR(clear_module_state->__pyx_n_s_mid_price_2);
  Py_CLEAR(clear_module_state->__pyx_n_u_mid_price_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_mid_price_std_dev);
  Py_CLEAR(clear_module_state->__pyx_n_s_mid_price_variance);
  Py_CLEAR(clear_module_state->__pyx_n_u_mid_price_variance);
  Py_CLEAR(clear_module_state->__pyx_n_s_min_spread);
  Py_CLEAR(clear_module_state->__pyx_n_s_mode);
  Py_CLEAR(clear_module_state->__pyx_n_s_model_config);
  Py_CLEAR(clear_module_state->__pyx_n_s_msg);
  Py_CLEAR(clear_module_state->__pyx_n_s_name);
  Py_CLEAR(clear_module_state->__pyx_n_s_nan);
  Py_CLEAR(clear_module_state->__pyx_n_u_nan);
  Py_CLEAR(clear_module_state->__pyx_n_s_network_status);
  Py_CLEAR(clear_module_state->__pyx_n_s_network_warning);
  Py_CLEAR(clear_module_state->__pyx_n_s_new_ask);
  Py_CLEAR(clear_module_state->__pyx_n_s_new_bid);
  Py_CLEAR(clear_module_state->__pyx_n_s_no_sells);
  Py_CLEAR(clear_module_state->__pyx_n_s_non_hanging);
  Py_CLEAR(clear_module_state->__pyx_n_s_normalize);
  Py_CLEAR(clear_module_state->__pyx_n_s_notify_hb_app);
  Py_CLEAR(clear_module_state->__pyx_n_s_notify_hb_app_with_timestamp);
  Py_CLEAR(clear_module_state->__pyx_n_s_np);
  Py_CLEAR(clear_module_state->__pyx_n_s_numpy);
  Py_CLEAR(clear_module_state->__pyx_kp_u_numpy_core_multiarray_failed_to);
  Py_CLEAR(clear_module_state->__pyx_kp_u_numpy_core_umath_failed_to_impor);
  Py_CLEAR(clear_module_state->__pyx_n_s_o);
  Py_CLEAR(clear_module_state->__pyx_kp_u_optimal_ask);
  Py_CLEAR(clear_module_state->__pyx_n_u_optimal_ask_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_optimal_ask_to_mid);
  Py_CLEAR(clear_module_state->__pyx_kp_u_optimal_bid);
  Py_CLEAR(clear_module_state->__pyx_n_u_optimal_bid_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_optimal_bid_to_mid);
  Py_CLEAR(clear_module_state->__pyx_kp_u_optimal_spread);
  Py_CLEAR(clear_module_state->__pyx_n_u_optimal_spread_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_order);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_age);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_amount);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_amount_shape_factor);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_book);
  Py_CLEAR(clear_module_state->__pyx_kp_u_order_book_depth_factor);
  Py_CLEAR(clear_module_state->__pyx_kp_u_order_book_intensity_factor);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_id);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_levels);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_levels_mode);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_optimization_enabled);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_override);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_refresh_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_refresh_tolerance);
  Py_CLEAR(clear_module_state->__pyx_n_s_order_refresh_tolerance_pct);
  Py_CLEAR(clear_module_state->__pyx_n_s_orders);
  Py_CLEAR(clear_module_state->__pyx_n_s_os);
  Py_CLEAR(clear_module_state->__pyx_n_s_p);
  Py_CLEAR(clear_module_state->__pyx_n_s_pandas);
  Py_CLEAR(clear_module_state->__pyx_n_s_path);
  Py_CLEAR(clear_module_state->__pyx_n_s_pd);
  Py_CLEAR(clear_module_state->__pyx_kp_s_pd_DataFrame);
  Py_CLEAR(clear_module_state->__pyx_n_s_percent);
  Py_CLEAR(clear_module_state->__pyx_n_s_pmm_logger);
  Py_CLEAR(clear_module_state->__pyx_n_s_price);
  Py_CLEAR(clear_module_state->__pyx_n_s_price_delegate);
  Py_CLEAR(clear_module_state->__pyx_n_s_process_tick);
  Py_CLEAR(clear_module_state->__pyx_n_s_proposal);
  Py_CLEAR(clear_module_state->__pyx_n_s_proposal_prices);
  Py_CLEAR(clear_module_state->__pyx_n_s_pure_mm_assets_df);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_state);
  Py_CLEAR(clear_module_state->__pyx_n_s_pyx_vtable);
  Py_CLEAR(clear_module_state->__pyx_kp_u_q);
  Py_CLEAR(clear_module_state->__pyx_n_s_quantity);
  Py_CLEAR(clear_module_state->__pyx_n_s_quote_asset);
  Py_CLEAR(clear_module_state->__pyx_n_s_quote_balance);
  Py_CLEAR(clear_module_state->__pyx_n_s_quote_currency);
  Py_CLEAR(clear_module_state->__pyx_n_s_quote_ratio);
  Py_CLEAR(clear_module_state->__pyx_n_s_range);
  Py_CLEAR(clear_module_state->__pyx_n_s_ready);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_reduce_ex);
  Py_CLEAR(clear_module_state->__pyx_n_s_ref_price);
  Py_CLEAR(clear_module_state->__pyx_n_s_regex);
  Py_CLEAR(clear_module_state->__pyx_n_s_register_events);
  Py_CLEAR(clear_module_state->__pyx_n_s_replace);
  Py_CLEAR(clear_module_state->__pyx_kp_u_reservation_price);
  Py_CLEAR(clear_module_state->__pyx_n_u_reservation_price_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_result_price);
  Py_CLEAR(clear_module_state->__pyx_n_s_return);
  Py_CLEAR(clear_module_state->__pyx_n_s_reverse);
  Py_CLEAR(clear_module_state->__pyx_n_s_risk_factor);
  Py_CLEAR(clear_module_state->__pyx_n_u_risk_factor);
  Py_CLEAR(clear_module_state->__pyx_kp_u_risk_factor_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_round);
  Py_CLEAR(clear_module_state->__pyx_n_u_s);
  Py_CLEAR(clear_module_state->__pyx_n_s_s_decimal_neg_one);
  Py_CLEAR(clear_module_state->__pyx_n_s_s_decimal_one);
  Py_CLEAR(clear_module_state->__pyx_n_s_s_decimal_zero);
  Py_CLEAR(clear_module_state->__pyx_n_s_sampling_length);
  Py_CLEAR(clear_module_state->__pyx_n_s_seconds);
  Py_CLEAR(clear_module_state->__pyx_n_s_self);
  Py_CLEAR(clear_module_state->__pyx_kp_s_self__events_cannot_be_converted);
  Py_CLEAR(clear_module_state->__pyx_n_u_sell);
  Py_CLEAR(clear_module_state->__pyx_n_s_sell_order);
  Py_CLEAR(clear_module_state->__pyx_n_s_sells);
  Py_CLEAR(clear_module_state->__pyx_n_s_send);
  Py_CLEAR(clear_module_state->__pyx_n_s_set_timers);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate);
  Py_CLEAR(clear_module_state->__pyx_n_s_setstate_cython);
  Py_CLEAR(clear_module_state->__pyx_n_s_should_wait_order_cancel_confirm);
  Py_CLEAR(clear_module_state->__pyx_n_s_size);
  Py_CLEAR(clear_module_state->__pyx_n_s_sort);
  Py_CLEAR(clear_module_state->__pyx_n_s_sorted);
  Py_CLEAR(clear_module_state->__pyx_n_s_spec);
  Py_CLEAR(clear_module_state->__pyx_n_s_split);
  Py_CLEAR(clear_module_state->__pyx_n_s_spread);
  Py_CLEAR(clear_module_state->__pyx_n_s_start);
  Py_CLEAR(clear_module_state->__pyx_n_s_start_datetime);
  Py_CLEAR(clear_module_state->__pyx_n_s_start_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_start_timestamp);
  Py_CLEAR(clear_module_state->__pyx_n_s_status_report_interval);
  Py_CLEAR(clear_module_state->__pyx_n_s_str);
  Py_CLEAR(clear_module_state->__pyx_n_s_strftime);
  Py_CLEAR(clear_module_state->__pyx_kp_s_stringsource);
  Py_CLEAR(clear_module_state->__pyx_n_s_super);
  Py_CLEAR(clear_module_state->__pyx_n_u_target_inv);
  Py_CLEAR(clear_module_state->__pyx_n_s_test);
  Py_CLEAR(clear_module_state->__pyx_n_s_throw);
  Py_CLEAR(clear_module_state->__pyx_n_s_ticks_to_be_ready_after);
  Py_CLEAR(clear_module_state->__pyx_n_s_ticks_to_be_ready_before);
  Py_CLEAR(clear_module_state->__pyx_kp_u_ticks_to_fill_buffers);
  Py_CLEAR(clear_module_state->__pyx_n_s_time);
  Py_CLEAR(clear_module_state->__pyx_n_s_time_left);
  Py_CLEAR(clear_module_state->__pyx_n_s_time_left_fraction);
  Py_CLEAR(clear_module_state->__pyx_n_u_time_left_fraction);
  Py_CLEAR(clear_module_state->__pyx_kp_u_time_until_end_of_trading_cycle);
  Py_CLEAR(clear_module_state->__pyx_kp_u_time_until_end_of_trading_cycle_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_timedelta);
  Py_CLEAR(clear_module_state->__pyx_n_s_timestamp);
  Py_CLEAR(clear_module_state->__pyx_n_s_title);
  Py_CLEAR(clear_module_state->__pyx_n_u_title);
  Py_CLEAR(clear_module_state->__pyx_n_s_to_create_orders);
  Py_CLEAR(clear_module_state->__pyx_n_s_to_csv);
  Py_CLEAR(clear_module_state->__pyx_n_s_to_show_current_pct);
  Py_CLEAR(clear_module_state->__pyx_n_s_to_string);
  Py_CLEAR(clear_module_state->__pyx_n_s_total_in_quote);
  Py_CLEAR(clear_module_state->__pyx_n_s_trade_type);
  Py_CLEAR(clear_module_state->__pyx_n_s_trading_intensity_buffer_size);
  Py_CLEAR(clear_module_state->__pyx_n_s_trading_pair);
  Py_CLEAR(clear_module_state->__pyx_n_s_typing);
  Py_CLEAR(clear_module_state->__pyx_n_s_unit);
  Py_CLEAR(clear_module_state->__pyx_n_s_unlink);
  Py_CLEAR(clear_module_state->__pyx_n_s_unregister_events);
  Py_CLEAR(clear_module_state->__pyx_n_s_update_from_config_map);
  Py_CLEAR(clear_module_state->__pyx_n_s_update_strategy_orders_with_equi);
  Py_CLEAR(clear_module_state->__pyx_n_s_upper);
  Py_CLEAR(clear_module_state->__pyx_n_s_values);
  Py_CLEAR(clear_module_state->__pyx_kp_u_vol);
  Py_CLEAR(clear_module_state->__pyx_n_s_vol_2);
  Py_CLEAR(clear_module_state->__pyx_kp_u_volatility);
  Py_CLEAR(clear_module_state->__pyx_n_u_volatility_2);
  Py_CLEAR(clear_module_state->__pyx_n_s_volatility_buffer_size);
  Py_CLEAR(clear_module_state->__pyx_n_s_volatility_pct);
  Py_CLEAR(clear_module_state->__pyx_n_s_warning);
  Py_CLEAR(clear_module_state->__pyx_n_s_warning_lines);
  Py_CLEAR(clear_module_state->__pyx_n_s_x);
  Py_CLEAR(clear_module_state->__pyx_n_s_zip);
  Py_CLEAR(clear_module_state->__pyx_float_100_0);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_1);
  Py_CLEAR(clear_module_state->__pyx_int_2);
  Py_CLEAR(clear_module_state->__pyx_int_4);
  Py_CLEAR(clear_module_state->__pyx_int_5);
  Py_CLEAR(clear_module_state->__pyx_int_8);
  Py_CLEAR(clear_module_state->__pyx_int_16);
  Py_CLEAR(clear_module_state->__pyx_int_32);
  Py_CLEAR(clear_module_state->__pyx_int_100);
  Py_CLEAR(clear_module_state->__pyx_int_900);
  Py_CLEAR(clear_module_state->__pyx_int_0x7fffffffffffffff);
  Py_CLEAR(clear_module_state->__pyx_int_neg_1);
  Py_CLEAR(clear_module_state->__pyx_k__3);
  Py_CLEAR(clear_module_state->__pyx_tuple_);
  Py_CLEAR(clear_module_state->__pyx_tuple__2);
  Py_CLEAR(clear_module_state->__pyx_tuple__16);
  Py_CLEAR(clear_module_state->__pyx_tuple__19);
  Py_CLEAR(clear_module_state->__pyx_tuple__20);
  Py_CLEAR(clear_module_state->__pyx_tuple__21);
  Py_CLEAR(clear_module_state->__pyx_tuple__22);
  Py_CLEAR(clear_module_state->__pyx_tuple__24);
  Py_CLEAR(clear_module_state->__pyx_tuple__26);
  Py_CLEAR(clear_module_state->__pyx_tuple__28);
  Py_CLEAR(clear_module_state->__pyx_tuple__33);
  Py_CLEAR(clear_module_state->__pyx_tuple__35);
  Py_CLEAR(clear_module_state->__pyx_tuple__37);
  Py_CLEAR(clear_module_state->__pyx_tuple__39);
  Py_CLEAR(clear_module_state->__pyx_tuple__41);
  Py_CLEAR(clear_module_state->__pyx_tuple__43);
  Py_CLEAR(clear_module_state->__pyx_tuple__45);
  Py_CLEAR(clear_module_state->__pyx_tuple__47);
  Py_CLEAR(clear_module_state->__pyx_tuple__49);
  Py_CLEAR(clear_module_state->__pyx_tuple__51);
  Py_CLEAR(clear_module_state->__pyx_tuple__53);
  Py_CLEAR(clear_module_state->__pyx_tuple__55);
  Py_CLEAR(clear_module_state->__pyx_tuple__58);
  Py_CLEAR(clear_module_state->__pyx_tuple__66);
  Py_CLEAR(clear_module_state->__pyx_tuple__72);
  Py_CLEAR(clear_module_state->__pyx_tuple__76);
  Py_CLEAR(clear_module_state->__pyx_tuple__82);
  Py_CLEAR(clear_module_state->__pyx_tuple__85);
  Py_CLEAR(clear_module_state->__pyx_tuple__89);
  Py_CLEAR(clear_module_state->__pyx_tuple__91);
  Py_CLEAR(clear_module_state->__pyx_tuple__94);
  Py_CLEAR(clear_module_state->__pyx_codeobj__23);
  Py_CLEAR(clear_module_state->__pyx_codeobj__25);
  Py_CLEAR(clear_module_state->__pyx_codeobj__27);
  Py_CLEAR(clear_module_state->__pyx_codeobj__29);
  Py_CLEAR(clear_module_state->__pyx_codeobj__30);
  Py_CLEAR(clear_module_state->__pyx_codeobj__31);
  Py_CLEAR(clear_module_state->__pyx_codeobj__32);
  Py_CLEAR(clear_module_state->__pyx_codeobj__34);
  Py_CLEAR(clear_module_state->__pyx_codeobj__36);
  Py_CLEAR(clear_module_state->__pyx_codeobj__38);
  Py_CLEAR(clear_module_state->__pyx_codeobj__40);
  Py_CLEAR(clear_module_state->__pyx_codeobj__42);
  Py_CLEAR(clear_module_state->__pyx_codeobj__44);
  Py_CLEAR(clear_module_state->__pyx_codeobj__46);
  Py_CLEAR(clear_module_state->__pyx_codeobj__48);
  Py_CLEAR(clear_module_state->__pyx_codeobj__50);
  Py_CLEAR(clear_module_state->__pyx_codeobj__52);
  Py_CLEAR(clear_module_state->__pyx_codeobj__54);
  Py_CLEAR(clear_module_state->__pyx_codeobj__56);
  Py_CLEAR(clear_module_state->__pyx_codeobj__57);
  Py_CLEAR(clear_module_state->__pyx_codeobj__59);
  Py_CLEAR(clear_module_state->__pyx_codeobj__60);
  Py_CLEAR(clear_module_state->__pyx_codeobj__61);
  Py_CLEAR(clear_module_state->__pyx_codeobj__62);
  Py_CLEAR(clear_module_state->__pyx_codeobj__63);
  Py_CLEAR(clear_module_state->__pyx_codeobj__64);
  Py_CLEAR(clear_module_state->__pyx_codeobj__65);
  Py_CLEAR(clear_module_state->__pyx_codeobj__67);
  Py_CLEAR(clear_module_state->__pyx_codeobj__68);
  Py_CLEAR(clear_module_state->__pyx_codeobj__69);
  Py_CLEAR(clear_module_state->__pyx_codeobj__70);
  Py_CLEAR(clear_module_state->__pyx_codeobj__71);
  Py_CLEAR(clear_module_state->__pyx_codeobj__73);
  Py_CLEAR(clear_module_state->__pyx_codeobj__74);
  Py_CLEAR(clear_module_state->__pyx_codeobj__75);
  Py_CLEAR(clear_module_state->__pyx_codeobj__77);
  Py_CLEAR(clear_module_state->__pyx_codeobj__78);
  Py_CLEAR(clear_module_state->__pyx_codeobj__79);
  Py_CLEAR(clear_module_state->__pyx_codeobj__80);
  Py_CLEAR(clear_module_state->__pyx_codeobj__81);
  Py_CLEAR(clear_module_state->__pyx_codeobj__83);
  Py_CLEAR(clear_module_state->__pyx_codeobj__84);
  Py_CLEAR(clear_module_state->__pyx_codeobj__86);
  Py_CLEAR(clear_module_state->__pyx_codeobj__87);
  Py_CLEAR(clear_module_state->__pyx_codeobj__88);
  Py_CLEAR(clear_module_state->__pyx_codeobj__90);
  Py_CLEAR(clear_module_state->__pyx_codeobj__92);
  Py_CLEAR(clear_module_state->__pyx_codeobj__93);
  Py_CLEAR(clear_module_state->__pyx_codeobj__95);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstate *traverse_module_state = __pyx_mstate(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  Py_VISIT(traverse_module_state->__pyx_empty_tuple);
  Py_VISIT(traverse_module_state->__pyx_empty_bytes);
  Py_VISIT(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_6pubsub_PubSub);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult);
  Py_VISIT(traverse_module_state->__pyx_ptype_7cpython_4type_type);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_dtype);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_flatiter);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_broadcast);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_ndarray);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_generic);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_number);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_integer);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_signedinteger);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_unsignedinteger);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_inexact);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_floating);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_complexfloating);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_flexible);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_character);
  Py_VISIT(traverse_module_state->__pyx_ptype_5numpy_ufunc);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradesForwarder);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_5clock_Clock);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_13order_tracker_OrderTracker);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr);
  Py_VISIT(traverse_module_state->__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr);
  Py_VISIT(traverse_module_state->__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr);
  Py_VISIT(traverse_module_state->__pyx_kp_u_0);
  Py_VISIT(traverse_module_state->__pyx_kp_u_1);
  Py_VISIT(traverse_module_state->__pyx_kp_u_10);
  Py_VISIT(traverse_module_state->__pyx_kp_u_100);
  Py_VISIT(traverse_module_state->__pyx_kp_u_10f);
  Py_VISIT(traverse_module_state->__pyx_kp_u_1_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_3f);
  Py_VISIT(traverse_module_state->__pyx_kp_u_4f);
  Py_VISIT(traverse_module_state->__pyx_kp_u_5E);
  Py_VISIT(traverse_module_state->__pyx_n_u_Age);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Amount_Adj);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Amount_Orig);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Assets);
  Py_VISIT(traverse_module_state->__pyx_n_s_AttributeError);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Available_Balance);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingConfigMap);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_4);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_a);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_a_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_a_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_a_4);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_a_5);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_a_6);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_a_7);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_a_8);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_10);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_11);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_12);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_13);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_4);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_5);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_6);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_7);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_8);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_c_9);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_d);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_e);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_f);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_g);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_g_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_g_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_g_4);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_g_5);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_g_6);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_g_7);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_g_8);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_g_9);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_i);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_i_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_i_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_i_4);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_l);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_m);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_m_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_n);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_p);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_p_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_s_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_t);
  Py_VISIT(traverse_module_state->__pyx_n_s_AvellanedaMarketMakingStrategy_u);
  Py_VISIT(traverse_module_state->__pyx_n_s_BUY);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Best_Ask);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Best_Bid);
  Py_VISIT(traverse_module_state->__pyx_n_s_CONNECTED);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Calculating_volatility_estimatin);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Calculating_volatility_estimatin_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_ClientConfigAdapter);
  Py_VISIT(traverse_module_state->__pyx_n_s_Clock);
  Py_VISIT(traverse_module_state->__pyx_n_s_CreatedPairOfOrders);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Creating);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Current);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Current_Value);
  Py_VISIT(traverse_module_state->__pyx_n_s_DailyBetweenTimesModel);
  Py_VISIT(traverse_module_state->__pyx_n_s_DataFrame);
  Py_VISIT(traverse_module_state->__pyx_n_s_Decimal);
  Py_VISIT(traverse_module_state->__pyx_n_s_Dict);
  Py_VISIT(traverse_module_state->__pyx_n_u_Exchange);
  Py_VISIT(traverse_module_state->__pyx_n_s_ExchangeBase);
  Py_VISIT(traverse_module_state->__pyx_n_s_FileNotFoundError);
  Py_VISIT(traverse_module_state->__pyx_n_s_FromDateToDateModel);
  Py_VISIT(traverse_module_state->__pyx_kp_u_H_M_S);
  Py_VISIT(traverse_module_state->__pyx_n_s_HangingOrdersTracker);
  Py_VISIT(traverse_module_state->__pyx_n_s_INFO);
  Py_VISIT(traverse_module_state->__pyx_n_s_ImportError);
  Py_VISIT(traverse_module_state->__pyx_n_s_InstantVolatilityIndicator);
  Py_VISIT(traverse_module_state->__pyx_n_s_LIMIT);
  Py_VISIT(traverse_module_state->__pyx_n_u_Level);
  Py_VISIT(traverse_module_state->__pyx_n_s_LimitOrder);
  Py_VISIT(traverse_module_state->__pyx_n_s_List);
  Py_VISIT(traverse_module_state->__pyx_kp_s_List_Decimal);
  Py_VISIT(traverse_module_state->__pyx_kp_s_List_LimitOrder);
  Py_VISIT(traverse_module_state->__pyx_kp_s_List_MarketTradingPairTuple);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Maker);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Maker_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Maker_buy_order_of);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Maker_sell_order_of);
  Py_VISIT(traverse_module_state->__pyx_n_u_Market);
  Py_VISIT(traverse_module_state->__pyx_n_s_MarketTradingPairTuple);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Market_connectors_are_not_ready);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Markets);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Markets_are_not_ready_No_market);
  Py_VISIT(traverse_module_state->__pyx_n_s_MidPrice);
  Py_VISIT(traverse_module_state->__pyx_n_u_MidPrice);
  Py_VISIT(traverse_module_state->__pyx_n_s_MultiOrderLevelModel);
  Py_VISIT(traverse_module_state->__pyx_n_s_NaN);
  Py_VISIT(traverse_module_state->__pyx_n_s_NetworkStatus);
  Py_VISIT(traverse_module_state->__pyx_kp_u_No_active_maker_orders);
  Py_VISIT(traverse_module_state->__pyx_kp_u_None);
  Py_VISIT(traverse_module_state->__pyx_n_s_OPTION_LOG_ALL);
  Py_VISIT(traverse_module_state->__pyx_n_s_OPTION_LOG_CREATE_ORDER);
  Py_VISIT(traverse_module_state->__pyx_n_s_OPTION_LOG_MAKER_ORDER_FILLED);
  Py_VISIT(traverse_module_state->__pyx_n_s_OPTION_LOG_STATUS_REPORT);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Optimal_Spread);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderBookAssetPriceDelegate);
  Py_VISIT(traverse_module_state->__pyx_n_s_OrderType);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Orders);
  Py_VISIT(traverse_module_state->__pyx_n_u_Price);
  Py_VISIT(traverse_module_state->__pyx_n_s_PriceSize);
  Py_VISIT(traverse_module_state->__pyx_n_s_PriceType);
  Py_VISIT(traverse_module_state->__pyx_n_s_Proposal);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Reservation_Price);
  Py_VISIT(traverse_module_state->__pyx_n_s_RunAlwaysExecutionState);
  Py_VISIT(traverse_module_state->__pyx_n_s_RunInTimeConditionalExecutionSta);
  Py_VISIT(traverse_module_state->__pyx_n_s_SELL);
  Py_VISIT(traverse_module_state->__pyx_n_u_Spread);
  Py_VISIT(traverse_module_state->__pyx_n_s_StrategyBase);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Strategy_parameters);
  Py_VISIT(traverse_module_state->__pyx_n_s_Timestamp);
  Py_VISIT(traverse_module_state->__pyx_kp_u_Total_Balance);
  Py_VISIT(traverse_module_state->__pyx_n_s_TrackHangingOrdersModel);
  Py_VISIT(traverse_module_state->__pyx_n_s_TradeType);
  Py_VISIT(traverse_module_state->__pyx_n_s_TradingIntensityIndicator);
  Py_VISIT(traverse_module_state->__pyx_n_s_Tuple);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Tuple_List_Proposal_List_Proposa);
  Py_VISIT(traverse_module_state->__pyx_n_u_Type);
  Py_VISIT(traverse_module_state->__pyx_n_s_TypeError);
  Py_VISIT(traverse_module_state->__pyx_n_s_Union);
  Py_VISIT(traverse_module_state->__pyx_kp_s_Union_AvellanedaMarketMakingConf);
  Py_VISIT(traverse_module_state->__pyx_kp_u_WARNINGS);
  Py_VISIT(traverse_module_state->__pyx_kp_u_WARNING_Some_markets_are_not_con);
  Py_VISIT(traverse_module_state->__pyx_kp_u__10);
  Py_VISIT(traverse_module_state->__pyx_kp_u__11);
  Py_VISIT(traverse_module_state->__pyx_kp_u__12);
  Py_VISIT(traverse_module_state->__pyx_kp_u__13);
  Py_VISIT(traverse_module_state->__pyx_kp_u__14);
  Py_VISIT(traverse_module_state->__pyx_kp_u__15);
  Py_VISIT(traverse_module_state->__pyx_n_s__17);
  Py_VISIT(traverse_module_state->__pyx_kp_u__18);
  Py_VISIT(traverse_module_state->__pyx_kp_u__4);
  Py_VISIT(traverse_module_state->__pyx_kp_u__5);
  Py_VISIT(traverse_module_state->__pyx_kp_u__6);
  Py_VISIT(traverse_module_state->__pyx_kp_u__7);
  Py_VISIT(traverse_module_state->__pyx_kp_u__8);
  Py_VISIT(traverse_module_state->__pyx_kp_u__9);
  Py_VISIT(traverse_module_state->__pyx_n_s__96);
  Py_VISIT(traverse_module_state->__pyx_n_u_a);
  Py_VISIT(traverse_module_state->__pyx_n_s_active_markets);
  Py_VISIT(traverse_module_state->__pyx_n_s_active_non_hanging_orders);
  Py_VISIT(traverse_module_state->__pyx_n_s_active_orders);
  Py_VISIT(traverse_module_state->__pyx_n_s_active_orders_df);
  Py_VISIT(traverse_module_state->__pyx_n_s_active_orders_df_locals_lambda);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_as_hanging_order);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_current_pairs_of_proposal_or);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_sample);
  Py_VISIT(traverse_module_state->__pyx_n_s_add_transaction_costs);
  Py_VISIT(traverse_module_state->__pyx_n_s_adjusted_available_balance_for_o);
  Py_VISIT(traverse_module_state->__pyx_n_s_age);
  Py_VISIT(traverse_module_state->__pyx_n_s_all);
  Py_VISIT(traverse_module_state->__pyx_n_s_all_markets_ready);
  Py_VISIT(traverse_module_state->__pyx_n_s_all_non_hanging_orders);
  Py_VISIT(traverse_module_state->__pyx_n_s_all_orders);
  Py_VISIT(traverse_module_state->__pyx_kp_u_alpha);
  Py_VISIT(traverse_module_state->__pyx_n_u_alpha_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_amount);
  Py_VISIT(traverse_module_state->__pyx_n_s_amount_orig);
  Py_VISIT(traverse_module_state->__pyx_n_s_apply);
  Py_VISIT(traverse_module_state->__pyx_n_s_apply_add_transaction_costs);
  Py_VISIT(traverse_module_state->__pyx_n_s_apply_budget_constraint);
  Py_VISIT(traverse_module_state->__pyx_n_s_apply_order_amount_eta_transform);
  Py_VISIT(traverse_module_state->__pyx_n_s_apply_order_optimization);
  Py_VISIT(traverse_module_state->__pyx_n_s_apply_order_price_modifiers);
  Py_VISIT(traverse_module_state->__pyx_n_s_args);
  Py_VISIT(traverse_module_state->__pyx_n_s_ask_level_spreads);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ask_orders_at_Size_Price);
  Py_VISIT(traverse_module_state->__pyx_n_s_ask_price);
  Py_VISIT(traverse_module_state->__pyx_n_s_assets_df);
  Py_VISIT(traverse_module_state->__pyx_n_s_asyncio_coroutines);
  Py_VISIT(traverse_module_state->__pyx_n_s_available_base_balance);
  Py_VISIT(traverse_module_state->__pyx_n_s_available_quote_balance);
  Py_VISIT(traverse_module_state->__pyx_n_s_balance_warning);
  Py_VISIT(traverse_module_state->__pyx_n_s_base_asset);
  Py_VISIT(traverse_module_state->__pyx_n_s_base_balance);
  Py_VISIT(traverse_module_state->__pyx_n_s_base_currency);
  Py_VISIT(traverse_module_state->__pyx_n_s_base_ratio);
  Py_VISIT(traverse_module_state->__pyx_n_s_base_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_best_ask);
  Py_VISIT(traverse_module_state->__pyx_n_u_best_ask);
  Py_VISIT(traverse_module_state->__pyx_n_s_best_bid);
  Py_VISIT(traverse_module_state->__pyx_n_u_best_bid);
  Py_VISIT(traverse_module_state->__pyx_n_s_bid_level_spreads);
  Py_VISIT(traverse_module_state->__pyx_kp_u_bid_orders_at_Size_Price);
  Py_VISIT(traverse_module_state->__pyx_n_s_bid_price);
  Py_VISIT(traverse_module_state->__pyx_n_s_bool);
  Py_VISIT(traverse_module_state->__pyx_n_u_buy);
  Py_VISIT(traverse_module_state->__pyx_n_s_buys);
  Py_VISIT(traverse_module_state->__pyx_n_s_calculate);
  Py_VISIT(traverse_module_state->__pyx_n_s_calculate_inventory);
  Py_VISIT(traverse_module_state->__pyx_n_s_calculate_reservation_price_and);
  Py_VISIT(traverse_module_state->__pyx_n_s_calculate_target_inventory);
  Py_VISIT(traverse_module_state->__pyx_n_s_cancel_active_orders);
  Py_VISIT(traverse_module_state->__pyx_n_s_cancel_order);
  Py_VISIT(traverse_module_state->__pyx_n_s_candidate_hanging_orders);
  Py_VISIT(traverse_module_state->__pyx_n_s_candidate_hanging_orders_from_pa);
  Py_VISIT(traverse_module_state->__pyx_n_s_ceil);
  Py_VISIT(traverse_module_state->__pyx_n_s_class_getitem);
  Py_VISIT(traverse_module_state->__pyx_n_s_client_order_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_cline_in_traceback);
  Py_VISIT(traverse_module_state->__pyx_n_s_clock);
  Py_VISIT(traverse_module_state->__pyx_n_s_close);
  Py_VISIT(traverse_module_state->__pyx_n_s_closing_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_cls);
  Py_VISIT(traverse_module_state->__pyx_n_s_collect_market_variables);
  Py_VISIT(traverse_module_state->__pyx_n_s_columns);
  Py_VISIT(traverse_module_state->__pyx_n_s_config_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_base_proposal);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_basic_proposal);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_proposal_based_on_order_l);
  Py_VISIT(traverse_module_state->__pyx_n_s_create_proposal_based_on_order_o);
  Py_VISIT(traverse_module_state->__pyx_n_s_current_created_pairs_of_orders);
  Py_VISIT(traverse_module_state->__pyx_n_u_current_inv);
  Py_VISIT(traverse_module_state->__pyx_n_s_current_prices);
  Py_VISIT(traverse_module_state->__pyx_n_s_current_value);
  Py_VISIT(traverse_module_state->__pyx_n_s_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_datetime);
  Py_VISIT(traverse_module_state->__pyx_n_s_debug_csv_path);
  Py_VISIT(traverse_module_state->__pyx_n_s_decimal);
  Py_VISIT(traverse_module_state->__pyx_n_s_df);
  Py_VISIT(traverse_module_state->__pyx_n_s_df_header);
  Py_VISIT(traverse_module_state->__pyx_n_s_df_lines);
  Py_VISIT(traverse_module_state->__pyx_n_s_dict);
  Py_VISIT(traverse_module_state->__pyx_kp_u_disable);
  Py_VISIT(traverse_module_state->__pyx_n_s_display_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_dump_debug_variables);
  Py_VISIT(traverse_module_state->__pyx_kp_u_enable);
  Py_VISIT(traverse_module_state->__pyx_n_s_end_datetime);
  Py_VISIT(traverse_module_state->__pyx_n_s_end_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_end_timestamp);
  Py_VISIT(traverse_module_state->__pyx_n_s_enumerate);
  Py_VISIT(traverse_module_state->__pyx_n_s_eta);
  Py_VISIT(traverse_module_state->__pyx_n_u_eta);
  Py_VISIT(traverse_module_state->__pyx_n_s_execute_orders_proposal);
  Py_VISIT(traverse_module_state->__pyx_n_s_execution_mode);
  Py_VISIT(traverse_module_state->__pyx_n_s_execution_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_execution_timeframe);
  Py_VISIT(traverse_module_state->__pyx_n_s_execution_timeframe_mode);
  Py_VISIT(traverse_module_state->__pyx_n_s_exists);
  Py_VISIT(traverse_module_state->__pyx_n_s_exp);
  Py_VISIT(traverse_module_state->__pyx_kp_u_filled);
  Py_VISIT(traverse_module_state->__pyx_n_s_filled_order_delay);
  Py_VISIT(traverse_module_state->__pyx_n_s_first_col_length);
  Py_VISIT(traverse_module_state->__pyx_n_s_float);
  Py_VISIT(traverse_module_state->__pyx_n_s_floor);
  Py_VISIT(traverse_module_state->__pyx_n_s_format);
  Py_VISIT(traverse_module_state->__pyx_n_s_format_status);
  Py_VISIT(traverse_module_state->__pyx_n_s_formatters);
  Py_VISIT(traverse_module_state->__pyx_n_s_gamma);
  Py_VISIT(traverse_module_state->__pyx_n_u_gamma);
  Py_VISIT(traverse_module_state->__pyx_kp_u_gc);
  Py_VISIT(traverse_module_state->__pyx_n_s_genexpr);
  Py_VISIT(traverse_module_state->__pyx_n_s_getLogger);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_adjusted_available_balance);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_available_balance);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_balance);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_config_map_execution_mode);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_config_map_hanging_orders);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_config_map_indicators);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_last_price);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_level_spreads);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_limit_orders);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_maker_order_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_mid_price);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_price);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_price_by_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_price_for_volume);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_spread);
  Py_VISIT(traverse_module_state->__pyx_n_s_get_volatility);
  Py_VISIT(traverse_module_state->__pyx_n_s_getstate);
  Py_VISIT(traverse_module_state->__pyx_n_u_hang);
  Py_VISIT(traverse_module_state->__pyx_n_s_hanging_orders_cancel_pct);
  Py_VISIT(traverse_module_state->__pyx_n_s_hanging_orders_enabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_hanging_orders_mode);
  Py_VISIT(traverse_module_state->__pyx_n_s_hanging_orders_tracker);
  Py_VISIT(traverse_module_state->__pyx_kp_u_has_been_completely_filled);
  Py_VISIT(traverse_module_state->__pyx_n_s_hb_app_notification);
  Py_VISIT(traverse_module_state->__pyx_n_s_header);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_client_config_config);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_connector_exchange_ba);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_core_data_type_common);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_core_data_type_limit);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_core_network_iterator);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_core_utils);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_strategy___utils___tr);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_strategy___utils___tr_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_strategy_avellaneda_m);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_strategy_avellaneda_m_2);
  Py_VISIT(traverse_module_state->__pyx_kp_s_hummingbot_strategy_avellaneda_m_3);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_strategy_conditional);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_strategy_data_types);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_strategy_hanging_orde);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_strategy_market_tradi);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_strategy_order_book_a);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_strategy_strategy_bas);
  Py_VISIT(traverse_module_state->__pyx_n_s_hummingbot_strategy_utils);
  Py_VISIT(traverse_module_state->__pyx_n_s_i);
  Py_VISIT(traverse_module_state->__pyx_n_s_idx);
  Py_VISIT(traverse_module_state->__pyx_n_s_import);
  Py_VISIT(traverse_module_state->__pyx_n_s_in_flight_cancels);
  Py_VISIT(traverse_module_state->__pyx_n_s_index);
  Py_VISIT(traverse_module_state->__pyx_n_s_info);
  Py_VISIT(traverse_module_state->__pyx_n_s_init_params);
  Py_VISIT(traverse_module_state->__pyx_n_s_initializing);
  Py_VISIT(traverse_module_state->__pyx_n_s_int);
  Py_VISIT(traverse_module_state->__pyx_n_s_inventory_target_base);
  Py_VISIT(traverse_module_state->__pyx_n_s_inventory_target_base_pct);
  Py_VISIT(traverse_module_state->__pyx_n_u_inventory_target_pct);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_algorithm_changed);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_algorithm_ready);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_buy);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_coroutine);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_debug);
  Py_VISIT(traverse_module_state->__pyx_kp_u_is_filled);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_hanging_order);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_order_id_in_completed_hanging);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_order_id_in_hanging_orders);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_potential_hanging_order);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_sampling_buffer_changed);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_sampling_buffer_full);
  Py_VISIT(traverse_module_state->__pyx_n_s_is_within_tolerance);
  Py_VISIT(traverse_module_state->__pyx_kp_u_isenabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_isnan);
  Py_VISIT(traverse_module_state->__pyx_n_s_items);
  Py_VISIT(traverse_module_state->__pyx_kp_u_kappa);
  Py_VISIT(traverse_module_state->__pyx_n_u_kappa_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_key);
  Py_VISIT(traverse_module_state->__pyx_n_s_len);
  Py_VISIT(traverse_module_state->__pyx_n_s_level);
  Py_VISIT(traverse_module_state->__pyx_n_s_level_distances);
  Py_VISIT(traverse_module_state->__pyx_n_s_level_step);
  Py_VISIT(traverse_module_state->__pyx_n_s_limit_orders);
  Py_VISIT(traverse_module_state->__pyx_n_s_line);
  Py_VISIT(traverse_module_state->__pyx_n_s_lines);
  Py_VISIT(traverse_module_state->__pyx_n_s_ln);
  Py_VISIT(traverse_module_state->__pyx_n_s_log_with_clock);
  Py_VISIT(traverse_module_state->__pyx_n_s_logger);
  Py_VISIT(traverse_module_state->__pyx_n_s_logging);
  Py_VISIT(traverse_module_state->__pyx_n_s_logging_options);
  Py_VISIT(traverse_module_state->__pyx_n_s_lvl_buy);
  Py_VISIT(traverse_module_state->__pyx_n_s_lvl_sell);
  Py_VISIT(traverse_module_state->__pyx_n_s_main);
  Py_VISIT(traverse_module_state->__pyx_n_s_map_df_to_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_market);
  Py_VISIT(traverse_module_state->__pyx_n_s_market_books);
  Py_VISIT(traverse_module_state->__pyx_n_s_market_info);
  Py_VISIT(traverse_module_state->__pyx_n_s_market_info_to_active_orders);
  Py_VISIT(traverse_module_state->__pyx_n_s_market_pair_to_active_orders);
  Py_VISIT(traverse_module_state->__pyx_n_s_market_status_data_frame);
  Py_VISIT(traverse_module_state->__pyx_n_s_market_trading_pair_tuples);
  Py_VISIT(traverse_module_state->__pyx_n_s_markets_columns);
  Py_VISIT(traverse_module_state->__pyx_n_s_markets_data);
  Py_VISIT(traverse_module_state->__pyx_n_s_markets_df);
  Py_VISIT(traverse_module_state->__pyx_n_s_math);
  Py_VISIT(traverse_module_state->__pyx_n_s_max);
  Py_VISIT(traverse_module_state->__pyx_n_s_max_order_age);
  Py_VISIT(traverse_module_state->__pyx_n_s_measure_order_book_liquidity);
  Py_VISIT(traverse_module_state->__pyx_kp_u_mid_price);
  Py_VISIT(traverse_module_state->__pyx_n_s_mid_price_2);
  Py_VISIT(traverse_module_state->__pyx_n_u_mid_price_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_mid_price_std_dev);
  Py_VISIT(traverse_module_state->__pyx_n_s_mid_price_variance);
  Py_VISIT(traverse_module_state->__pyx_n_u_mid_price_variance);
  Py_VISIT(traverse_module_state->__pyx_n_s_min_spread);
  Py_VISIT(traverse_module_state->__pyx_n_s_mode);
  Py_VISIT(traverse_module_state->__pyx_n_s_model_config);
  Py_VISIT(traverse_module_state->__pyx_n_s_msg);
  Py_VISIT(traverse_module_state->__pyx_n_s_name);
  Py_VISIT(traverse_module_state->__pyx_n_s_nan);
  Py_VISIT(traverse_module_state->__pyx_n_u_nan);
  Py_VISIT(traverse_module_state->__pyx_n_s_network_status);
  Py_VISIT(traverse_module_state->__pyx_n_s_network_warning);
  Py_VISIT(traverse_module_state->__pyx_n_s_new_ask);
  Py_VISIT(traverse_module_state->__pyx_n_s_new_bid);
  Py_VISIT(traverse_module_state->__pyx_n_s_no_sells);
  Py_VISIT(traverse_module_state->__pyx_n_s_non_hanging);
  Py_VISIT(traverse_module_state->__pyx_n_s_normalize);
  Py_VISIT(traverse_module_state->__pyx_n_s_notify_hb_app);
  Py_VISIT(traverse_module_state->__pyx_n_s_notify_hb_app_with_timestamp);
  Py_VISIT(traverse_module_state->__pyx_n_s_np);
  Py_VISIT(traverse_module_state->__pyx_n_s_numpy);
  Py_VISIT(traverse_module_state->__pyx_kp_u_numpy_core_multiarray_failed_to);
  Py_VISIT(traverse_module_state->__pyx_kp_u_numpy_core_umath_failed_to_impor);
  Py_VISIT(traverse_module_state->__pyx_n_s_o);
  Py_VISIT(traverse_module_state->__pyx_kp_u_optimal_ask);
  Py_VISIT(traverse_module_state->__pyx_n_u_optimal_ask_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_optimal_ask_to_mid);
  Py_VISIT(traverse_module_state->__pyx_kp_u_optimal_bid);
  Py_VISIT(traverse_module_state->__pyx_n_u_optimal_bid_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_optimal_bid_to_mid);
  Py_VISIT(traverse_module_state->__pyx_kp_u_optimal_spread);
  Py_VISIT(traverse_module_state->__pyx_n_u_optimal_spread_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_order);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_age);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_amount);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_amount_shape_factor);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_book);
  Py_VISIT(traverse_module_state->__pyx_kp_u_order_book_depth_factor);
  Py_VISIT(traverse_module_state->__pyx_kp_u_order_book_intensity_factor);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_id);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_levels);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_levels_mode);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_optimization_enabled);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_override);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_refresh_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_refresh_tolerance);
  Py_VISIT(traverse_module_state->__pyx_n_s_order_refresh_tolerance_pct);
  Py_VISIT(traverse_module_state->__pyx_n_s_orders);
  Py_VISIT(traverse_module_state->__pyx_n_s_os);
  Py_VISIT(traverse_module_state->__pyx_n_s_p);
  Py_VISIT(traverse_module_state->__pyx_n_s_pandas);
  Py_VISIT(traverse_module_state->__pyx_n_s_path);
  Py_VISIT(traverse_module_state->__pyx_n_s_pd);
  Py_VISIT(traverse_module_state->__pyx_kp_s_pd_DataFrame);
  Py_VISIT(traverse_module_state->__pyx_n_s_percent);
  Py_VISIT(traverse_module_state->__pyx_n_s_pmm_logger);
  Py_VISIT(traverse_module_state->__pyx_n_s_price);
  Py_VISIT(traverse_module_state->__pyx_n_s_price_delegate);
  Py_VISIT(traverse_module_state->__pyx_n_s_process_tick);
  Py_VISIT(traverse_module_state->__pyx_n_s_proposal);
  Py_VISIT(traverse_module_state->__pyx_n_s_proposal_prices);
  Py_VISIT(traverse_module_state->__pyx_n_s_pure_mm_assets_df);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_state);
  Py_VISIT(traverse_module_state->__pyx_n_s_pyx_vtable);
  Py_VISIT(traverse_module_state->__pyx_kp_u_q);
  Py_VISIT(traverse_module_state->__pyx_n_s_quantity);
  Py_VISIT(traverse_module_state->__pyx_n_s_quote_asset);
  Py_VISIT(traverse_module_state->__pyx_n_s_quote_balance);
  Py_VISIT(traverse_module_state->__pyx_n_s_quote_currency);
  Py_VISIT(traverse_module_state->__pyx_n_s_quote_ratio);
  Py_VISIT(traverse_module_state->__pyx_n_s_range);
  Py_VISIT(traverse_module_state->__pyx_n_s_ready);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_reduce_ex);
  Py_VISIT(traverse_module_state->__pyx_n_s_ref_price);
  Py_VISIT(traverse_module_state->__pyx_n_s_regex);
  Py_VISIT(traverse_module_state->__pyx_n_s_register_events);
  Py_VISIT(traverse_module_state->__pyx_n_s_replace);
  Py_VISIT(traverse_module_state->__pyx_kp_u_reservation_price);
  Py_VISIT(traverse_module_state->__pyx_n_u_reservation_price_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_result_price);
  Py_VISIT(traverse_module_state->__pyx_n_s_return);
  Py_VISIT(traverse_module_state->__pyx_n_s_reverse);
  Py_VISIT(traverse_module_state->__pyx_n_s_risk_factor);
  Py_VISIT(traverse_module_state->__pyx_n_u_risk_factor);
  Py_VISIT(traverse_module_state->__pyx_kp_u_risk_factor_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_round);
  Py_VISIT(traverse_module_state->__pyx_n_u_s);
  Py_VISIT(traverse_module_state->__pyx_n_s_s_decimal_neg_one);
  Py_VISIT(traverse_module_state->__pyx_n_s_s_decimal_one);
  Py_VISIT(traverse_module_state->__pyx_n_s_s_decimal_zero);
  Py_VISIT(traverse_module_state->__pyx_n_s_sampling_length);
  Py_VISIT(traverse_module_state->__pyx_n_s_seconds);
  Py_VISIT(traverse_module_state->__pyx_n_s_self);
  Py_VISIT(traverse_module_state->__pyx_kp_s_self__events_cannot_be_converted);
  Py_VISIT(traverse_module_state->__pyx_n_u_sell);
  Py_VISIT(traverse_module_state->__pyx_n_s_sell_order);
  Py_VISIT(traverse_module_state->__pyx_n_s_sells);
  Py_VISIT(traverse_module_state->__pyx_n_s_send);
  Py_VISIT(traverse_module_state->__pyx_n_s_set_timers);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate);
  Py_VISIT(traverse_module_state->__pyx_n_s_setstate_cython);
  Py_VISIT(traverse_module_state->__pyx_n_s_should_wait_order_cancel_confirm);
  Py_VISIT(traverse_module_state->__pyx_n_s_size);
  Py_VISIT(traverse_module_state->__pyx_n_s_sort);
  Py_VISIT(traverse_module_state->__pyx_n_s_sorted);
  Py_VISIT(traverse_module_state->__pyx_n_s_spec);
  Py_VISIT(traverse_module_state->__pyx_n_s_split);
  Py_VISIT(traverse_module_state->__pyx_n_s_spread);
  Py_VISIT(traverse_module_state->__pyx_n_s_start);
  Py_VISIT(traverse_module_state->__pyx_n_s_start_datetime);
  Py_VISIT(traverse_module_state->__pyx_n_s_start_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_start_timestamp);
  Py_VISIT(traverse_module_state->__pyx_n_s_status_report_interval);
  Py_VISIT(traverse_module_state->__pyx_n_s_str);
  Py_VISIT(traverse_module_state->__pyx_n_s_strftime);
  Py_VISIT(traverse_module_state->__pyx_kp_s_stringsource);
  Py_VISIT(traverse_module_state->__pyx_n_s_super);
  Py_VISIT(traverse_module_state->__pyx_n_u_target_inv);
  Py_VISIT(traverse_module_state->__pyx_n_s_test);
  Py_VISIT(traverse_module_state->__pyx_n_s_throw);
  Py_VISIT(traverse_module_state->__pyx_n_s_ticks_to_be_ready_after);
  Py_VISIT(traverse_module_state->__pyx_n_s_ticks_to_be_ready_before);
  Py_VISIT(traverse_module_state->__pyx_kp_u_ticks_to_fill_buffers);
  Py_VISIT(traverse_module_state->__pyx_n_s_time);
  Py_VISIT(traverse_module_state->__pyx_n_s_time_left);
  Py_VISIT(traverse_module_state->__pyx_n_s_time_left_fraction);
  Py_VISIT(traverse_module_state->__pyx_n_u_time_left_fraction);
  Py_VISIT(traverse_module_state->__pyx_kp_u_time_until_end_of_trading_cycle);
  Py_VISIT(traverse_module_state->__pyx_kp_u_time_until_end_of_trading_cycle_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_timedelta);
  Py_VISIT(traverse_module_state->__pyx_n_s_timestamp);
  Py_VISIT(traverse_module_state->__pyx_n_s_title);
  Py_VISIT(traverse_module_state->__pyx_n_u_title);
  Py_VISIT(traverse_module_state->__pyx_n_s_to_create_orders);
  Py_VISIT(traverse_module_state->__pyx_n_s_to_csv);
  Py_VISIT(traverse_module_state->__pyx_n_s_to_show_current_pct);
  Py_VISIT(traverse_module_state->__pyx_n_s_to_string);
  Py_VISIT(traverse_module_state->__pyx_n_s_total_in_quote);
  Py_VISIT(traverse_module_state->__pyx_n_s_trade_type);
  Py_VISIT(traverse_module_state->__pyx_n_s_trading_intensity_buffer_size);
  Py_VISIT(traverse_module_state->__pyx_n_s_trading_pair);
  Py_VISIT(traverse_module_state->__pyx_n_s_typing);
  Py_VISIT(traverse_module_state->__pyx_n_s_unit);
  Py_VISIT(traverse_module_state->__pyx_n_s_unlink);
  Py_VISIT(traverse_module_state->__pyx_n_s_unregister_events);
  Py_VISIT(traverse_module_state->__pyx_n_s_update_from_config_map);
  Py_VISIT(traverse_module_state->__pyx_n_s_update_strategy_orders_with_equi);
  Py_VISIT(traverse_module_state->__pyx_n_s_upper);
  Py_VISIT(traverse_module_state->__pyx_n_s_values);
  Py_VISIT(traverse_module_state->__pyx_kp_u_vol);
  Py_VISIT(traverse_module_state->__pyx_n_s_vol_2);
  Py_VISIT(traverse_module_state->__pyx_kp_u_volatility);
  Py_VISIT(traverse_module_state->__pyx_n_u_volatility_2);
  Py_VISIT(traverse_module_state->__pyx_n_s_volatility_buffer_size);
  Py_VISIT(traverse_module_state->__pyx_n_s_volatility_pct);
  Py_VISIT(traverse_module_state->__pyx_n_s_warning);
  Py_VISIT(traverse_module_state->__pyx_n_s_warning_lines);
  Py_VISIT(traverse_module_state->__pyx_n_s_x);
  Py_VISIT(traverse_module_state->__pyx_n_s_zip);
  Py_VISIT(traverse_module_state->__pyx_float_100_0);
  Py_VISIT(traverse_module_state->__pyx_int_0);
  Py_VISIT(traverse_module_state->__pyx_int_1);
  Py_VISIT(traverse_module_state->__pyx_int_2);
  Py_VISIT(traverse_module_state->__pyx_int_4);
  Py_VISIT(traverse_module_state->__pyx_int_5);
  Py_VISIT(traverse_module_state->__pyx_int_8);
  Py_VISIT(traverse_module_state->__pyx_int_16);
  Py_VISIT(traverse_module_state->__pyx_int_32);
  Py_VISIT(traverse_module_state->__pyx_int_100);
  Py_VISIT(traverse_module_state->__pyx_int_900);
  Py_VISIT(traverse_module_state->__pyx_int_0x7fffffffffffffff);
  Py_VISIT(traverse_module_state->__pyx_int_neg_1);
  Py_VISIT(traverse_module_state->__pyx_k__3);
  Py_VISIT(traverse_module_state->__pyx_tuple_);
  Py_VISIT(traverse_module_state->__pyx_tuple__2);
  Py_VISIT(traverse_module_state->__pyx_tuple__16);
  Py_VISIT(traverse_module_state->__pyx_tuple__19);
  Py_VISIT(traverse_module_state->__pyx_tuple__20);
  Py_VISIT(traverse_module_state->__pyx_tuple__21);
  Py_VISIT(traverse_module_state->__pyx_tuple__22);
  Py_VISIT(traverse_module_state->__pyx_tuple__24);
  Py_VISIT(traverse_module_state->__pyx_tuple__26);
  Py_VISIT(traverse_module_state->__pyx_tuple__28);
  Py_VISIT(traverse_module_state->__pyx_tuple__33);
  Py_VISIT(traverse_module_state->__pyx_tuple__35);
  Py_VISIT(traverse_module_state->__pyx_tuple__37);
  Py_VISIT(traverse_module_state->__pyx_tuple__39);
  Py_VISIT(traverse_module_state->__pyx_tuple__41);
  Py_VISIT(traverse_module_state->__pyx_tuple__43);
  Py_VISIT(traverse_module_state->__pyx_tuple__45);
  Py_VISIT(traverse_module_state->__pyx_tuple__47);
  Py_VISIT(traverse_module_state->__pyx_tuple__49);
  Py_VISIT(traverse_module_state->__pyx_tuple__51);
  Py_VISIT(traverse_module_state->__pyx_tuple__53);
  Py_VISIT(traverse_module_state->__pyx_tuple__55);
  Py_VISIT(traverse_module_state->__pyx_tuple__58);
  Py_VISIT(traverse_module_state->__pyx_tuple__66);
  Py_VISIT(traverse_module_state->__pyx_tuple__72);
  Py_VISIT(traverse_module_state->__pyx_tuple__76);
  Py_VISIT(traverse_module_state->__pyx_tuple__82);
  Py_VISIT(traverse_module_state->__pyx_tuple__85);
  Py_VISIT(traverse_module_state->__pyx_tuple__89);
  Py_VISIT(traverse_module_state->__pyx_tuple__91);
  Py_VISIT(traverse_module_state->__pyx_tuple__94);
  Py_VISIT(traverse_module_state->__pyx_codeobj__23);
  Py_VISIT(traverse_module_state->__pyx_codeobj__25);
  Py_VISIT(traverse_module_state->__pyx_codeobj__27);
  Py_VISIT(traverse_module_state->__pyx_codeobj__29);
  Py_VISIT(traverse_module_state->__pyx_codeobj__30);
  Py_VISIT(traverse_module_state->__pyx_codeobj__31);
  Py_VISIT(traverse_module_state->__pyx_codeobj__32);
  Py_VISIT(traverse_module_state->__pyx_codeobj__34);
  Py_VISIT(traverse_module_state->__pyx_codeobj__36);
  Py_VISIT(traverse_module_state->__pyx_codeobj__38);
  Py_VISIT(traverse_module_state->__pyx_codeobj__40);
  Py_VISIT(traverse_module_state->__pyx_codeobj__42);
  Py_VISIT(traverse_module_state->__pyx_codeobj__44);
  Py_VISIT(traverse_module_state->__pyx_codeobj__46);
  Py_VISIT(traverse_module_state->__pyx_codeobj__48);
  Py_VISIT(traverse_module_state->__pyx_codeobj__50);
  Py_VISIT(traverse_module_state->__pyx_codeobj__52);
  Py_VISIT(traverse_module_state->__pyx_codeobj__54);
  Py_VISIT(traverse_module_state->__pyx_codeobj__56);
  Py_VISIT(traverse_module_state->__pyx_codeobj__57);
  Py_VISIT(traverse_module_state->__pyx_codeobj__59);
  Py_VISIT(traverse_module_state->__pyx_codeobj__60);
  Py_VISIT(traverse_module_state->__pyx_codeobj__61);
  Py_VISIT(traverse_module_state->__pyx_codeobj__62);
  Py_VISIT(traverse_module_state->__pyx_codeobj__63);
  Py_VISIT(traverse_module_state->__pyx_codeobj__64);
  Py_VISIT(traverse_module_state->__pyx_codeobj__65);
  Py_VISIT(traverse_module_state->__pyx_codeobj__67);
  Py_VISIT(traverse_module_state->__pyx_codeobj__68);
  Py_VISIT(traverse_module_state->__pyx_codeobj__69);
  Py_VISIT(traverse_module_state->__pyx_codeobj__70);
  Py_VISIT(traverse_module_state->__pyx_codeobj__71);
  Py_VISIT(traverse_module_state->__pyx_codeobj__73);
  Py_VISIT(traverse_module_state->__pyx_codeobj__74);
  Py_VISIT(traverse_module_state->__pyx_codeobj__75);
  Py_VISIT(traverse_module_state->__pyx_codeobj__77);
  Py_VISIT(traverse_module_state->__pyx_codeobj__78);
  Py_VISIT(traverse_module_state->__pyx_codeobj__79);
  Py_VISIT(traverse_module_state->__pyx_codeobj__80);
  Py_VISIT(traverse_module_state->__pyx_codeobj__81);
  Py_VISIT(traverse_module_state->__pyx_codeobj__83);
  Py_VISIT(traverse_module_state->__pyx_codeobj__84);
  Py_VISIT(traverse_module_state->__pyx_codeobj__86);
  Py_VISIT(traverse_module_state->__pyx_codeobj__87);
  Py_VISIT(traverse_module_state->__pyx_codeobj__88);
  Py_VISIT(traverse_module_state->__pyx_codeobj__90);
  Py_VISIT(traverse_module_state->__pyx_codeobj__92);
  Py_VISIT(traverse_module_state->__pyx_codeobj__93);
  Py_VISIT(traverse_module_state->__pyx_codeobj__95);
  return 0;
}
#endif
/* #### Code section: module_state_defines ### */
#define __pyx_d __pyx_mstate_global->__pyx_d
#define __pyx_b __pyx_mstate_global->__pyx_b
#define __pyx_cython_runtime __pyx_mstate_global->__pyx_cython_runtime
#define __pyx_empty_tuple __pyx_mstate_global->__pyx_empty_tuple
#define __pyx_empty_bytes __pyx_mstate_global->__pyx_empty_bytes
#define __pyx_empty_unicode __pyx_mstate_global->__pyx_empty_unicode
#ifdef __Pyx_CyFunction_USED
#define __pyx_CyFunctionType __pyx_mstate_global->__pyx_CyFunctionType
#endif
#ifdef __Pyx_FusedFunction_USED
#define __pyx_FusedFunctionType __pyx_mstate_global->__pyx_FusedFunctionType
#endif
#ifdef __Pyx_Generator_USED
#define __pyx_GeneratorType __pyx_mstate_global->__pyx_GeneratorType
#endif
#ifdef __Pyx_IterableCoroutine_USED
#define __pyx_IterableCoroutineType __pyx_mstate_global->__pyx_IterableCoroutineType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineAwaitType __pyx_mstate_global->__pyx_CoroutineAwaitType
#endif
#ifdef __Pyx_Coroutine_USED
#define __pyx_CoroutineType __pyx_mstate_global->__pyx_CoroutineType
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_6pubsub_PubSub __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_6pubsub_PubSub
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult
#define __pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_7cpython_4type_type __pyx_mstate_global->__pyx_ptype_7cpython_4type_type
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_5numpy_dtype __pyx_mstate_global->__pyx_ptype_5numpy_dtype
#define __pyx_ptype_5numpy_flatiter __pyx_mstate_global->__pyx_ptype_5numpy_flatiter
#define __pyx_ptype_5numpy_broadcast __pyx_mstate_global->__pyx_ptype_5numpy_broadcast
#define __pyx_ptype_5numpy_ndarray __pyx_mstate_global->__pyx_ptype_5numpy_ndarray
#define __pyx_ptype_5numpy_generic __pyx_mstate_global->__pyx_ptype_5numpy_generic
#define __pyx_ptype_5numpy_number __pyx_mstate_global->__pyx_ptype_5numpy_number
#define __pyx_ptype_5numpy_integer __pyx_mstate_global->__pyx_ptype_5numpy_integer
#define __pyx_ptype_5numpy_signedinteger __pyx_mstate_global->__pyx_ptype_5numpy_signedinteger
#define __pyx_ptype_5numpy_unsignedinteger __pyx_mstate_global->__pyx_ptype_5numpy_unsignedinteger
#define __pyx_ptype_5numpy_inexact __pyx_mstate_global->__pyx_ptype_5numpy_inexact
#define __pyx_ptype_5numpy_floating __pyx_mstate_global->__pyx_ptype_5numpy_floating
#define __pyx_ptype_5numpy_complexfloating __pyx_mstate_global->__pyx_ptype_5numpy_complexfloating
#define __pyx_ptype_5numpy_flexible __pyx_mstate_global->__pyx_ptype_5numpy_flexible
#define __pyx_ptype_5numpy_character __pyx_mstate_global->__pyx_ptype_5numpy_character
#define __pyx_ptype_5numpy_ufunc __pyx_mstate_global->__pyx_ptype_5numpy_ufunc
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator
#define __pyx_ptype_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradesForwarder __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradesForwarder
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_5clock_Clock __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_5clock_Clock
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_8strategy_13order_tracker_OrderTracker __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_13order_tracker_OrderTracker
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator __pyx_mstate_global->__pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase
#if CYTHON_USE_MODULE_STATE
#endif
#define __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase __pyx_mstate_global->__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase
#if CYTHON_USE_MODULE_STATE
#define __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy __pyx_mstate_global->__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy
#define __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start __pyx_mstate_global->__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start
#define __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr __pyx_mstate_global->__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr
#define __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal __pyx_mstate_global->__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal
#define __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr __pyx_mstate_global->__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr
#define __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr __pyx_mstate_global->__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr
#endif
#define __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy
#define __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start
#define __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr
#define __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal
#define __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr
#define __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr __pyx_mstate_global->__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr
#define __pyx_kp_u_0 __pyx_mstate_global->__pyx_kp_u_0
#define __pyx_kp_u_1 __pyx_mstate_global->__pyx_kp_u_1
#define __pyx_kp_u_10 __pyx_mstate_global->__pyx_kp_u_10
#define __pyx_kp_u_100 __pyx_mstate_global->__pyx_kp_u_100
#define __pyx_kp_u_10f __pyx_mstate_global->__pyx_kp_u_10f
#define __pyx_kp_u_1_2 __pyx_mstate_global->__pyx_kp_u_1_2
#define __pyx_kp_u_2 __pyx_mstate_global->__pyx_kp_u_2
#define __pyx_kp_u_3f __pyx_mstate_global->__pyx_kp_u_3f
#define __pyx_kp_u_4f __pyx_mstate_global->__pyx_kp_u_4f
#define __pyx_kp_u_5E __pyx_mstate_global->__pyx_kp_u_5E
#define __pyx_n_u_Age __pyx_mstate_global->__pyx_n_u_Age
#define __pyx_kp_u_Amount_Adj __pyx_mstate_global->__pyx_kp_u_Amount_Adj
#define __pyx_kp_u_Amount_Orig __pyx_mstate_global->__pyx_kp_u_Amount_Orig
#define __pyx_kp_u_Assets __pyx_mstate_global->__pyx_kp_u_Assets
#define __pyx_n_s_AttributeError __pyx_mstate_global->__pyx_n_s_AttributeError
#define __pyx_kp_u_Available_Balance __pyx_mstate_global->__pyx_kp_u_Available_Balance
#define __pyx_n_s_AvellanedaMarketMakingConfigMap __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingConfigMap
#define __pyx_n_s_AvellanedaMarketMakingStrategy __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy
#define __pyx_n_s_AvellanedaMarketMakingStrategy_2 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_2
#define __pyx_n_s_AvellanedaMarketMakingStrategy_3 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_3
#define __pyx_n_s_AvellanedaMarketMakingStrategy_4 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_4
#define __pyx_n_s_AvellanedaMarketMakingStrategy_a __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_a
#define __pyx_n_s_AvellanedaMarketMakingStrategy_a_2 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_a_2
#define __pyx_n_s_AvellanedaMarketMakingStrategy_a_3 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_a_3
#define __pyx_n_s_AvellanedaMarketMakingStrategy_a_4 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_a_4
#define __pyx_n_s_AvellanedaMarketMakingStrategy_a_5 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_a_5
#define __pyx_n_s_AvellanedaMarketMakingStrategy_a_6 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_a_6
#define __pyx_n_s_AvellanedaMarketMakingStrategy_a_7 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_a_7
#define __pyx_n_s_AvellanedaMarketMakingStrategy_a_8 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_a_8
#define __pyx_n_s_AvellanedaMarketMakingStrategy_c __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_c
#define __pyx_n_s_AvellanedaMarketMakingStrategy_c_10 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_c_10
#define __pyx_n_s_AvellanedaMarketMakingStrategy_c_11 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_c_11
#define __pyx_n_s_AvellanedaMarketMakingStrategy_c_12 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_c_12
#define __pyx_n_s_AvellanedaMarketMakingStrategy_c_13 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_c_13
#define __pyx_n_s_AvellanedaMarketMakingStrategy_c_2 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_c_2
#define __pyx_n_s_AvellanedaMarketMakingStrategy_c_3 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_c_3
#define __pyx_n_s_AvellanedaMarketMakingStrategy_c_4 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_c_4
#define __pyx_n_s_AvellanedaMarketMakingStrategy_c_5 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_c_5
#define __pyx_n_s_AvellanedaMarketMakingStrategy_c_6 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_c_6
#define __pyx_n_s_AvellanedaMarketMakingStrategy_c_7 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_c_7
#define __pyx_n_s_AvellanedaMarketMakingStrategy_c_8 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_c_8
#define __pyx_n_s_AvellanedaMarketMakingStrategy_c_9 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_c_9
#define __pyx_n_s_AvellanedaMarketMakingStrategy_d __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_d
#define __pyx_n_s_AvellanedaMarketMakingStrategy_e __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_e
#define __pyx_n_s_AvellanedaMarketMakingStrategy_f __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_f
#define __pyx_n_s_AvellanedaMarketMakingStrategy_g __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_g
#define __pyx_n_s_AvellanedaMarketMakingStrategy_g_2 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_g_2
#define __pyx_n_s_AvellanedaMarketMakingStrategy_g_3 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_g_3
#define __pyx_n_s_AvellanedaMarketMakingStrategy_g_4 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_g_4
#define __pyx_n_s_AvellanedaMarketMakingStrategy_g_5 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_g_5
#define __pyx_n_s_AvellanedaMarketMakingStrategy_g_6 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_g_6
#define __pyx_n_s_AvellanedaMarketMakingStrategy_g_7 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_g_7
#define __pyx_n_s_AvellanedaMarketMakingStrategy_g_8 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_g_8
#define __pyx_n_s_AvellanedaMarketMakingStrategy_g_9 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_g_9
#define __pyx_n_s_AvellanedaMarketMakingStrategy_i __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_i
#define __pyx_n_s_AvellanedaMarketMakingStrategy_i_2 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_i_2
#define __pyx_n_s_AvellanedaMarketMakingStrategy_i_3 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_i_3
#define __pyx_n_s_AvellanedaMarketMakingStrategy_i_4 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_i_4
#define __pyx_n_s_AvellanedaMarketMakingStrategy_l __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_l
#define __pyx_n_s_AvellanedaMarketMakingStrategy_m __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_m
#define __pyx_n_s_AvellanedaMarketMakingStrategy_m_2 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_m_2
#define __pyx_n_s_AvellanedaMarketMakingStrategy_n __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_n
#define __pyx_n_s_AvellanedaMarketMakingStrategy_p __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_p
#define __pyx_n_s_AvellanedaMarketMakingStrategy_p_2 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_p_2
#define __pyx_n_s_AvellanedaMarketMakingStrategy_s __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_s
#define __pyx_n_s_AvellanedaMarketMakingStrategy_s_2 __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_s_2
#define __pyx_n_s_AvellanedaMarketMakingStrategy_t __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_t
#define __pyx_n_s_AvellanedaMarketMakingStrategy_u __pyx_mstate_global->__pyx_n_s_AvellanedaMarketMakingStrategy_u
#define __pyx_n_s_BUY __pyx_mstate_global->__pyx_n_s_BUY
#define __pyx_kp_u_Best_Ask __pyx_mstate_global->__pyx_kp_u_Best_Ask
#define __pyx_kp_u_Best_Bid __pyx_mstate_global->__pyx_kp_u_Best_Bid
#define __pyx_n_s_CONNECTED __pyx_mstate_global->__pyx_n_s_CONNECTED
#define __pyx_kp_u_Calculating_volatility_estimatin __pyx_mstate_global->__pyx_kp_u_Calculating_volatility_estimatin
#define __pyx_kp_u_Calculating_volatility_estimatin_2 __pyx_mstate_global->__pyx_kp_u_Calculating_volatility_estimatin_2
#define __pyx_n_s_ClientConfigAdapter __pyx_mstate_global->__pyx_n_s_ClientConfigAdapter
#define __pyx_n_s_Clock __pyx_mstate_global->__pyx_n_s_Clock
#define __pyx_n_s_CreatedPairOfOrders __pyx_mstate_global->__pyx_n_s_CreatedPairOfOrders
#define __pyx_kp_u_Creating __pyx_mstate_global->__pyx_kp_u_Creating
#define __pyx_kp_u_Current __pyx_mstate_global->__pyx_kp_u_Current
#define __pyx_kp_u_Current_Value __pyx_mstate_global->__pyx_kp_u_Current_Value
#define __pyx_n_s_DailyBetweenTimesModel __pyx_mstate_global->__pyx_n_s_DailyBetweenTimesModel
#define __pyx_n_s_DataFrame __pyx_mstate_global->__pyx_n_s_DataFrame
#define __pyx_n_s_Decimal __pyx_mstate_global->__pyx_n_s_Decimal
#define __pyx_n_s_Dict __pyx_mstate_global->__pyx_n_s_Dict
#define __pyx_n_u_Exchange __pyx_mstate_global->__pyx_n_u_Exchange
#define __pyx_n_s_ExchangeBase __pyx_mstate_global->__pyx_n_s_ExchangeBase
#define __pyx_n_s_FileNotFoundError __pyx_mstate_global->__pyx_n_s_FileNotFoundError
#define __pyx_n_s_FromDateToDateModel __pyx_mstate_global->__pyx_n_s_FromDateToDateModel
#define __pyx_kp_u_H_M_S __pyx_mstate_global->__pyx_kp_u_H_M_S
#define __pyx_n_s_HangingOrdersTracker __pyx_mstate_global->__pyx_n_s_HangingOrdersTracker
#define __pyx_n_s_INFO __pyx_mstate_global->__pyx_n_s_INFO
#define __pyx_n_s_ImportError __pyx_mstate_global->__pyx_n_s_ImportError
#define __pyx_n_s_InstantVolatilityIndicator __pyx_mstate_global->__pyx_n_s_InstantVolatilityIndicator
#define __pyx_n_s_LIMIT __pyx_mstate_global->__pyx_n_s_LIMIT
#define __pyx_n_u_Level __pyx_mstate_global->__pyx_n_u_Level
#define __pyx_n_s_LimitOrder __pyx_mstate_global->__pyx_n_s_LimitOrder
#define __pyx_n_s_List __pyx_mstate_global->__pyx_n_s_List
#define __pyx_kp_s_List_Decimal __pyx_mstate_global->__pyx_kp_s_List_Decimal
#define __pyx_kp_s_List_LimitOrder __pyx_mstate_global->__pyx_kp_s_List_LimitOrder
#define __pyx_kp_s_List_MarketTradingPairTuple __pyx_mstate_global->__pyx_kp_s_List_MarketTradingPairTuple
#define __pyx_kp_u_Maker __pyx_mstate_global->__pyx_kp_u_Maker
#define __pyx_kp_u_Maker_2 __pyx_mstate_global->__pyx_kp_u_Maker_2
#define __pyx_kp_u_Maker_buy_order_of __pyx_mstate_global->__pyx_kp_u_Maker_buy_order_of
#define __pyx_kp_u_Maker_sell_order_of __pyx_mstate_global->__pyx_kp_u_Maker_sell_order_of
#define __pyx_n_u_Market __pyx_mstate_global->__pyx_n_u_Market
#define __pyx_n_s_MarketTradingPairTuple __pyx_mstate_global->__pyx_n_s_MarketTradingPairTuple
#define __pyx_kp_u_Market_connectors_are_not_ready __pyx_mstate_global->__pyx_kp_u_Market_connectors_are_not_ready
#define __pyx_kp_u_Markets __pyx_mstate_global->__pyx_kp_u_Markets
#define __pyx_kp_u_Markets_are_not_ready_No_market __pyx_mstate_global->__pyx_kp_u_Markets_are_not_ready_No_market
#define __pyx_n_s_MidPrice __pyx_mstate_global->__pyx_n_s_MidPrice
#define __pyx_n_u_MidPrice __pyx_mstate_global->__pyx_n_u_MidPrice
#define __pyx_n_s_MultiOrderLevelModel __pyx_mstate_global->__pyx_n_s_MultiOrderLevelModel
#define __pyx_n_s_NaN __pyx_mstate_global->__pyx_n_s_NaN
#define __pyx_n_s_NetworkStatus __pyx_mstate_global->__pyx_n_s_NetworkStatus
#define __pyx_kp_u_No_active_maker_orders __pyx_mstate_global->__pyx_kp_u_No_active_maker_orders
#define __pyx_kp_u_None __pyx_mstate_global->__pyx_kp_u_None
#define __pyx_n_s_OPTION_LOG_ALL __pyx_mstate_global->__pyx_n_s_OPTION_LOG_ALL
#define __pyx_n_s_OPTION_LOG_CREATE_ORDER __pyx_mstate_global->__pyx_n_s_OPTION_LOG_CREATE_ORDER
#define __pyx_n_s_OPTION_LOG_MAKER_ORDER_FILLED __pyx_mstate_global->__pyx_n_s_OPTION_LOG_MAKER_ORDER_FILLED
#define __pyx_n_s_OPTION_LOG_STATUS_REPORT __pyx_mstate_global->__pyx_n_s_OPTION_LOG_STATUS_REPORT
#define __pyx_kp_u_Optimal_Spread __pyx_mstate_global->__pyx_kp_u_Optimal_Spread
#define __pyx_n_s_OrderBookAssetPriceDelegate __pyx_mstate_global->__pyx_n_s_OrderBookAssetPriceDelegate
#define __pyx_n_s_OrderType __pyx_mstate_global->__pyx_n_s_OrderType
#define __pyx_kp_u_Orders __pyx_mstate_global->__pyx_kp_u_Orders
#define __pyx_n_u_Price __pyx_mstate_global->__pyx_n_u_Price
#define __pyx_n_s_PriceSize __pyx_mstate_global->__pyx_n_s_PriceSize
#define __pyx_n_s_PriceType __pyx_mstate_global->__pyx_n_s_PriceType
#define __pyx_n_s_Proposal __pyx_mstate_global->__pyx_n_s_Proposal
#define __pyx_kp_u_Reservation_Price __pyx_mstate_global->__pyx_kp_u_Reservation_Price
#define __pyx_n_s_RunAlwaysExecutionState __pyx_mstate_global->__pyx_n_s_RunAlwaysExecutionState
#define __pyx_n_s_RunInTimeConditionalExecutionSta __pyx_mstate_global->__pyx_n_s_RunInTimeConditionalExecutionSta
#define __pyx_n_s_SELL __pyx_mstate_global->__pyx_n_s_SELL
#define __pyx_n_u_Spread __pyx_mstate_global->__pyx_n_u_Spread
#define __pyx_n_s_StrategyBase __pyx_mstate_global->__pyx_n_s_StrategyBase
#define __pyx_kp_u_Strategy_parameters __pyx_mstate_global->__pyx_kp_u_Strategy_parameters
#define __pyx_n_s_Timestamp __pyx_mstate_global->__pyx_n_s_Timestamp
#define __pyx_kp_u_Total_Balance __pyx_mstate_global->__pyx_kp_u_Total_Balance
#define __pyx_n_s_TrackHangingOrdersModel __pyx_mstate_global->__pyx_n_s_TrackHangingOrdersModel
#define __pyx_n_s_TradeType __pyx_mstate_global->__pyx_n_s_TradeType
#define __pyx_n_s_TradingIntensityIndicator __pyx_mstate_global->__pyx_n_s_TradingIntensityIndicator
#define __pyx_n_s_Tuple __pyx_mstate_global->__pyx_n_s_Tuple
#define __pyx_kp_s_Tuple_List_Proposal_List_Proposa __pyx_mstate_global->__pyx_kp_s_Tuple_List_Proposal_List_Proposa
#define __pyx_n_u_Type __pyx_mstate_global->__pyx_n_u_Type
#define __pyx_n_s_TypeError __pyx_mstate_global->__pyx_n_s_TypeError
#define __pyx_n_s_Union __pyx_mstate_global->__pyx_n_s_Union
#define __pyx_kp_s_Union_AvellanedaMarketMakingConf __pyx_mstate_global->__pyx_kp_s_Union_AvellanedaMarketMakingConf
#define __pyx_kp_u_WARNINGS __pyx_mstate_global->__pyx_kp_u_WARNINGS
#define __pyx_kp_u_WARNING_Some_markets_are_not_con __pyx_mstate_global->__pyx_kp_u_WARNING_Some_markets_are_not_con
#define __pyx_kp_u__10 __pyx_mstate_global->__pyx_kp_u__10
#define __pyx_kp_u__11 __pyx_mstate_global->__pyx_kp_u__11
#define __pyx_kp_u__12 __pyx_mstate_global->__pyx_kp_u__12
#define __pyx_kp_u__13 __pyx_mstate_global->__pyx_kp_u__13
#define __pyx_kp_u__14 __pyx_mstate_global->__pyx_kp_u__14
#define __pyx_kp_u__15 __pyx_mstate_global->__pyx_kp_u__15
#define __pyx_n_s__17 __pyx_mstate_global->__pyx_n_s__17
#define __pyx_kp_u__18 __pyx_mstate_global->__pyx_kp_u__18
#define __pyx_kp_u__4 __pyx_mstate_global->__pyx_kp_u__4
#define __pyx_kp_u__5 __pyx_mstate_global->__pyx_kp_u__5
#define __pyx_kp_u__6 __pyx_mstate_global->__pyx_kp_u__6
#define __pyx_kp_u__7 __pyx_mstate_global->__pyx_kp_u__7
#define __pyx_kp_u__8 __pyx_mstate_global->__pyx_kp_u__8
#define __pyx_kp_u__9 __pyx_mstate_global->__pyx_kp_u__9
#define __pyx_n_s__96 __pyx_mstate_global->__pyx_n_s__96
#define __pyx_n_u_a __pyx_mstate_global->__pyx_n_u_a
#define __pyx_n_s_active_markets __pyx_mstate_global->__pyx_n_s_active_markets
#define __pyx_n_s_active_non_hanging_orders __pyx_mstate_global->__pyx_n_s_active_non_hanging_orders
#define __pyx_n_s_active_orders __pyx_mstate_global->__pyx_n_s_active_orders
#define __pyx_n_s_active_orders_df __pyx_mstate_global->__pyx_n_s_active_orders_df
#define __pyx_n_s_active_orders_df_locals_lambda __pyx_mstate_global->__pyx_n_s_active_orders_df_locals_lambda
#define __pyx_n_s_add_as_hanging_order __pyx_mstate_global->__pyx_n_s_add_as_hanging_order
#define __pyx_n_s_add_current_pairs_of_proposal_or __pyx_mstate_global->__pyx_n_s_add_current_pairs_of_proposal_or
#define __pyx_n_s_add_sample __pyx_mstate_global->__pyx_n_s_add_sample
#define __pyx_n_s_add_transaction_costs __pyx_mstate_global->__pyx_n_s_add_transaction_costs
#define __pyx_n_s_adjusted_available_balance_for_o __pyx_mstate_global->__pyx_n_s_adjusted_available_balance_for_o
#define __pyx_n_s_age __pyx_mstate_global->__pyx_n_s_age
#define __pyx_n_s_all __pyx_mstate_global->__pyx_n_s_all
#define __pyx_n_s_all_markets_ready __pyx_mstate_global->__pyx_n_s_all_markets_ready
#define __pyx_n_s_all_non_hanging_orders __pyx_mstate_global->__pyx_n_s_all_non_hanging_orders
#define __pyx_n_s_all_orders __pyx_mstate_global->__pyx_n_s_all_orders
#define __pyx_kp_u_alpha __pyx_mstate_global->__pyx_kp_u_alpha
#define __pyx_n_u_alpha_2 __pyx_mstate_global->__pyx_n_u_alpha_2
#define __pyx_n_s_amount __pyx_mstate_global->__pyx_n_s_amount
#define __pyx_n_s_amount_orig __pyx_mstate_global->__pyx_n_s_amount_orig
#define __pyx_n_s_apply __pyx_mstate_global->__pyx_n_s_apply
#define __pyx_n_s_apply_add_transaction_costs __pyx_mstate_global->__pyx_n_s_apply_add_transaction_costs
#define __pyx_n_s_apply_budget_constraint __pyx_mstate_global->__pyx_n_s_apply_budget_constraint
#define __pyx_n_s_apply_order_amount_eta_transform __pyx_mstate_global->__pyx_n_s_apply_order_amount_eta_transform
#define __pyx_n_s_apply_order_optimization __pyx_mstate_global->__pyx_n_s_apply_order_optimization
#define __pyx_n_s_apply_order_price_modifiers __pyx_mstate_global->__pyx_n_s_apply_order_price_modifiers
#define __pyx_n_s_args __pyx_mstate_global->__pyx_n_s_args
#define __pyx_n_s_ask_level_spreads __pyx_mstate_global->__pyx_n_s_ask_level_spreads
#define __pyx_kp_u_ask_orders_at_Size_Price __pyx_mstate_global->__pyx_kp_u_ask_orders_at_Size_Price
#define __pyx_n_s_ask_price __pyx_mstate_global->__pyx_n_s_ask_price
#define __pyx_n_s_assets_df __pyx_mstate_global->__pyx_n_s_assets_df
#define __pyx_n_s_asyncio_coroutines __pyx_mstate_global->__pyx_n_s_asyncio_coroutines
#define __pyx_n_s_available_base_balance __pyx_mstate_global->__pyx_n_s_available_base_balance
#define __pyx_n_s_available_quote_balance __pyx_mstate_global->__pyx_n_s_available_quote_balance
#define __pyx_n_s_balance_warning __pyx_mstate_global->__pyx_n_s_balance_warning
#define __pyx_n_s_base_asset __pyx_mstate_global->__pyx_n_s_base_asset
#define __pyx_n_s_base_balance __pyx_mstate_global->__pyx_n_s_base_balance
#define __pyx_n_s_base_currency __pyx_mstate_global->__pyx_n_s_base_currency
#define __pyx_n_s_base_ratio __pyx_mstate_global->__pyx_n_s_base_ratio
#define __pyx_n_s_base_value __pyx_mstate_global->__pyx_n_s_base_value
#define __pyx_n_s_best_ask __pyx_mstate_global->__pyx_n_s_best_ask
#define __pyx_n_u_best_ask __pyx_mstate_global->__pyx_n_u_best_ask
#define __pyx_n_s_best_bid __pyx_mstate_global->__pyx_n_s_best_bid
#define __pyx_n_u_best_bid __pyx_mstate_global->__pyx_n_u_best_bid
#define __pyx_n_s_bid_level_spreads __pyx_mstate_global->__pyx_n_s_bid_level_spreads
#define __pyx_kp_u_bid_orders_at_Size_Price __pyx_mstate_global->__pyx_kp_u_bid_orders_at_Size_Price
#define __pyx_n_s_bid_price __pyx_mstate_global->__pyx_n_s_bid_price
#define __pyx_n_s_bool __pyx_mstate_global->__pyx_n_s_bool
#define __pyx_n_u_buy __pyx_mstate_global->__pyx_n_u_buy
#define __pyx_n_s_buys __pyx_mstate_global->__pyx_n_s_buys
#define __pyx_n_s_calculate __pyx_mstate_global->__pyx_n_s_calculate
#define __pyx_n_s_calculate_inventory __pyx_mstate_global->__pyx_n_s_calculate_inventory
#define __pyx_n_s_calculate_reservation_price_and __pyx_mstate_global->__pyx_n_s_calculate_reservation_price_and
#define __pyx_n_s_calculate_target_inventory __pyx_mstate_global->__pyx_n_s_calculate_target_inventory
#define __pyx_n_s_cancel_active_orders __pyx_mstate_global->__pyx_n_s_cancel_active_orders
#define __pyx_n_s_cancel_order __pyx_mstate_global->__pyx_n_s_cancel_order
#define __pyx_n_s_candidate_hanging_orders __pyx_mstate_global->__pyx_n_s_candidate_hanging_orders
#define __pyx_n_s_candidate_hanging_orders_from_pa __pyx_mstate_global->__pyx_n_s_candidate_hanging_orders_from_pa
#define __pyx_n_s_ceil __pyx_mstate_global->__pyx_n_s_ceil
#define __pyx_n_s_class_getitem __pyx_mstate_global->__pyx_n_s_class_getitem
#define __pyx_n_s_client_order_id __pyx_mstate_global->__pyx_n_s_client_order_id
#define __pyx_n_s_cline_in_traceback __pyx_mstate_global->__pyx_n_s_cline_in_traceback
#define __pyx_n_s_clock __pyx_mstate_global->__pyx_n_s_clock
#define __pyx_n_s_close __pyx_mstate_global->__pyx_n_s_close
#define __pyx_n_s_closing_time __pyx_mstate_global->__pyx_n_s_closing_time
#define __pyx_n_s_cls __pyx_mstate_global->__pyx_n_s_cls
#define __pyx_n_s_collect_market_variables __pyx_mstate_global->__pyx_n_s_collect_market_variables
#define __pyx_n_s_columns __pyx_mstate_global->__pyx_n_s_columns
#define __pyx_n_s_config_map __pyx_mstate_global->__pyx_n_s_config_map
#define __pyx_n_s_create_base_proposal __pyx_mstate_global->__pyx_n_s_create_base_proposal
#define __pyx_n_s_create_basic_proposal __pyx_mstate_global->__pyx_n_s_create_basic_proposal
#define __pyx_n_s_create_proposal_based_on_order_l __pyx_mstate_global->__pyx_n_s_create_proposal_based_on_order_l
#define __pyx_n_s_create_proposal_based_on_order_o __pyx_mstate_global->__pyx_n_s_create_proposal_based_on_order_o
#define __pyx_n_s_current_created_pairs_of_orders __pyx_mstate_global->__pyx_n_s_current_created_pairs_of_orders
#define __pyx_n_u_current_inv __pyx_mstate_global->__pyx_n_u_current_inv
#define __pyx_n_s_current_prices __pyx_mstate_global->__pyx_n_s_current_prices
#define __pyx_n_s_current_value __pyx_mstate_global->__pyx_n_s_current_value
#define __pyx_n_s_data __pyx_mstate_global->__pyx_n_s_data
#define __pyx_n_s_datetime __pyx_mstate_global->__pyx_n_s_datetime
#define __pyx_n_s_debug_csv_path __pyx_mstate_global->__pyx_n_s_debug_csv_path
#define __pyx_n_s_decimal __pyx_mstate_global->__pyx_n_s_decimal
#define __pyx_n_s_df __pyx_mstate_global->__pyx_n_s_df
#define __pyx_n_s_df_header __pyx_mstate_global->__pyx_n_s_df_header
#define __pyx_n_s_df_lines __pyx_mstate_global->__pyx_n_s_df_lines
#define __pyx_n_s_dict __pyx_mstate_global->__pyx_n_s_dict
#define __pyx_kp_u_disable __pyx_mstate_global->__pyx_kp_u_disable
#define __pyx_n_s_display_name __pyx_mstate_global->__pyx_n_s_display_name
#define __pyx_n_s_dump_debug_variables __pyx_mstate_global->__pyx_n_s_dump_debug_variables
#define __pyx_kp_u_enable __pyx_mstate_global->__pyx_kp_u_enable
#define __pyx_n_s_end_datetime __pyx_mstate_global->__pyx_n_s_end_datetime
#define __pyx_n_s_end_time __pyx_mstate_global->__pyx_n_s_end_time
#define __pyx_n_s_end_timestamp __pyx_mstate_global->__pyx_n_s_end_timestamp
#define __pyx_n_s_enumerate __pyx_mstate_global->__pyx_n_s_enumerate
#define __pyx_n_s_eta __pyx_mstate_global->__pyx_n_s_eta
#define __pyx_n_u_eta __pyx_mstate_global->__pyx_n_u_eta
#define __pyx_n_s_execute_orders_proposal __pyx_mstate_global->__pyx_n_s_execute_orders_proposal
#define __pyx_n_s_execution_mode __pyx_mstate_global->__pyx_n_s_execution_mode
#define __pyx_n_s_execution_state __pyx_mstate_global->__pyx_n_s_execution_state
#define __pyx_n_s_execution_timeframe __pyx_mstate_global->__pyx_n_s_execution_timeframe
#define __pyx_n_s_execution_timeframe_mode __pyx_mstate_global->__pyx_n_s_execution_timeframe_mode
#define __pyx_n_s_exists __pyx_mstate_global->__pyx_n_s_exists
#define __pyx_n_s_exp __pyx_mstate_global->__pyx_n_s_exp
#define __pyx_kp_u_filled __pyx_mstate_global->__pyx_kp_u_filled
#define __pyx_n_s_filled_order_delay __pyx_mstate_global->__pyx_n_s_filled_order_delay
#define __pyx_n_s_first_col_length __pyx_mstate_global->__pyx_n_s_first_col_length
#define __pyx_n_s_float __pyx_mstate_global->__pyx_n_s_float
#define __pyx_n_s_floor __pyx_mstate_global->__pyx_n_s_floor
#define __pyx_n_s_format __pyx_mstate_global->__pyx_n_s_format
#define __pyx_n_s_format_status __pyx_mstate_global->__pyx_n_s_format_status
#define __pyx_n_s_formatters __pyx_mstate_global->__pyx_n_s_formatters
#define __pyx_n_s_gamma __pyx_mstate_global->__pyx_n_s_gamma
#define __pyx_n_u_gamma __pyx_mstate_global->__pyx_n_u_gamma
#define __pyx_kp_u_gc __pyx_mstate_global->__pyx_kp_u_gc
#define __pyx_n_s_genexpr __pyx_mstate_global->__pyx_n_s_genexpr
#define __pyx_n_s_getLogger __pyx_mstate_global->__pyx_n_s_getLogger
#define __pyx_n_s_get_adjusted_available_balance __pyx_mstate_global->__pyx_n_s_get_adjusted_available_balance
#define __pyx_n_s_get_available_balance __pyx_mstate_global->__pyx_n_s_get_available_balance
#define __pyx_n_s_get_balance __pyx_mstate_global->__pyx_n_s_get_balance
#define __pyx_n_s_get_config_map_execution_mode __pyx_mstate_global->__pyx_n_s_get_config_map_execution_mode
#define __pyx_n_s_get_config_map_hanging_orders __pyx_mstate_global->__pyx_n_s_get_config_map_hanging_orders
#define __pyx_n_s_get_config_map_indicators __pyx_mstate_global->__pyx_n_s_get_config_map_indicators
#define __pyx_n_s_get_last_price __pyx_mstate_global->__pyx_n_s_get_last_price
#define __pyx_n_s_get_level_spreads __pyx_mstate_global->__pyx_n_s_get_level_spreads
#define __pyx_n_s_get_limit_orders __pyx_mstate_global->__pyx_n_s_get_limit_orders
#define __pyx_n_s_get_maker_order_type __pyx_mstate_global->__pyx_n_s_get_maker_order_type
#define __pyx_n_s_get_mid_price __pyx_mstate_global->__pyx_n_s_get_mid_price
#define __pyx_n_s_get_price __pyx_mstate_global->__pyx_n_s_get_price
#define __pyx_n_s_get_price_by_type __pyx_mstate_global->__pyx_n_s_get_price_by_type
#define __pyx_n_s_get_price_for_volume __pyx_mstate_global->__pyx_n_s_get_price_for_volume
#define __pyx_n_s_get_spread __pyx_mstate_global->__pyx_n_s_get_spread
#define __pyx_n_s_get_volatility __pyx_mstate_global->__pyx_n_s_get_volatility
#define __pyx_n_s_getstate __pyx_mstate_global->__pyx_n_s_getstate
#define __pyx_n_u_hang __pyx_mstate_global->__pyx_n_u_hang
#define __pyx_n_s_hanging_orders_cancel_pct __pyx_mstate_global->__pyx_n_s_hanging_orders_cancel_pct
#define __pyx_n_s_hanging_orders_enabled __pyx_mstate_global->__pyx_n_s_hanging_orders_enabled
#define __pyx_n_s_hanging_orders_mode __pyx_mstate_global->__pyx_n_s_hanging_orders_mode
#define __pyx_n_s_hanging_orders_tracker __pyx_mstate_global->__pyx_n_s_hanging_orders_tracker
#define __pyx_kp_u_has_been_completely_filled __pyx_mstate_global->__pyx_kp_u_has_been_completely_filled
#define __pyx_n_s_hb_app_notification __pyx_mstate_global->__pyx_n_s_hb_app_notification
#define __pyx_n_s_header __pyx_mstate_global->__pyx_n_s_header
#define __pyx_n_s_hummingbot_client_config_config __pyx_mstate_global->__pyx_n_s_hummingbot_client_config_config
#define __pyx_n_s_hummingbot_connector_exchange_ba __pyx_mstate_global->__pyx_n_s_hummingbot_connector_exchange_ba
#define __pyx_n_s_hummingbot_core_data_type_common __pyx_mstate_global->__pyx_n_s_hummingbot_core_data_type_common
#define __pyx_n_s_hummingbot_core_data_type_limit __pyx_mstate_global->__pyx_n_s_hummingbot_core_data_type_limit
#define __pyx_n_s_hummingbot_core_network_iterator __pyx_mstate_global->__pyx_n_s_hummingbot_core_network_iterator
#define __pyx_n_s_hummingbot_core_utils __pyx_mstate_global->__pyx_n_s_hummingbot_core_utils
#define __pyx_n_s_hummingbot_strategy___utils___tr __pyx_mstate_global->__pyx_n_s_hummingbot_strategy___utils___tr
#define __pyx_n_s_hummingbot_strategy___utils___tr_2 __pyx_mstate_global->__pyx_n_s_hummingbot_strategy___utils___tr_2
#define __pyx_n_s_hummingbot_strategy_avellaneda_m __pyx_mstate_global->__pyx_n_s_hummingbot_strategy_avellaneda_m
#define __pyx_n_s_hummingbot_strategy_avellaneda_m_2 __pyx_mstate_global->__pyx_n_s_hummingbot_strategy_avellaneda_m_2
#define __pyx_kp_s_hummingbot_strategy_avellaneda_m_3 __pyx_mstate_global->__pyx_kp_s_hummingbot_strategy_avellaneda_m_3
#define __pyx_n_s_hummingbot_strategy_conditional __pyx_mstate_global->__pyx_n_s_hummingbot_strategy_conditional
#define __pyx_n_s_hummingbot_strategy_data_types __pyx_mstate_global->__pyx_n_s_hummingbot_strategy_data_types
#define __pyx_n_s_hummingbot_strategy_hanging_orde __pyx_mstate_global->__pyx_n_s_hummingbot_strategy_hanging_orde
#define __pyx_n_s_hummingbot_strategy_market_tradi __pyx_mstate_global->__pyx_n_s_hummingbot_strategy_market_tradi
#define __pyx_n_s_hummingbot_strategy_order_book_a __pyx_mstate_global->__pyx_n_s_hummingbot_strategy_order_book_a
#define __pyx_n_s_hummingbot_strategy_strategy_bas __pyx_mstate_global->__pyx_n_s_hummingbot_strategy_strategy_bas
#define __pyx_n_s_hummingbot_strategy_utils __pyx_mstate_global->__pyx_n_s_hummingbot_strategy_utils
#define __pyx_n_s_i __pyx_mstate_global->__pyx_n_s_i
#define __pyx_n_s_idx __pyx_mstate_global->__pyx_n_s_idx
#define __pyx_n_s_import __pyx_mstate_global->__pyx_n_s_import
#define __pyx_n_s_in_flight_cancels __pyx_mstate_global->__pyx_n_s_in_flight_cancels
#define __pyx_n_s_index __pyx_mstate_global->__pyx_n_s_index
#define __pyx_n_s_info __pyx_mstate_global->__pyx_n_s_info
#define __pyx_n_s_init_params __pyx_mstate_global->__pyx_n_s_init_params
#define __pyx_n_s_initializing __pyx_mstate_global->__pyx_n_s_initializing
#define __pyx_n_s_int __pyx_mstate_global->__pyx_n_s_int
#define __pyx_n_s_inventory_target_base __pyx_mstate_global->__pyx_n_s_inventory_target_base
#define __pyx_n_s_inventory_target_base_pct __pyx_mstate_global->__pyx_n_s_inventory_target_base_pct
#define __pyx_n_u_inventory_target_pct __pyx_mstate_global->__pyx_n_u_inventory_target_pct
#define __pyx_n_s_is_algorithm_changed __pyx_mstate_global->__pyx_n_s_is_algorithm_changed
#define __pyx_n_s_is_algorithm_ready __pyx_mstate_global->__pyx_n_s_is_algorithm_ready
#define __pyx_n_s_is_buy __pyx_mstate_global->__pyx_n_s_is_buy
#define __pyx_n_s_is_coroutine __pyx_mstate_global->__pyx_n_s_is_coroutine
#define __pyx_n_s_is_debug __pyx_mstate_global->__pyx_n_s_is_debug
#define __pyx_kp_u_is_filled __pyx_mstate_global->__pyx_kp_u_is_filled
#define __pyx_n_s_is_hanging_order __pyx_mstate_global->__pyx_n_s_is_hanging_order
#define __pyx_n_s_is_order_id_in_completed_hanging __pyx_mstate_global->__pyx_n_s_is_order_id_in_completed_hanging
#define __pyx_n_s_is_order_id_in_hanging_orders __pyx_mstate_global->__pyx_n_s_is_order_id_in_hanging_orders
#define __pyx_n_s_is_potential_hanging_order __pyx_mstate_global->__pyx_n_s_is_potential_hanging_order
#define __pyx_n_s_is_sampling_buffer_changed __pyx_mstate_global->__pyx_n_s_is_sampling_buffer_changed
#define __pyx_n_s_is_sampling_buffer_full __pyx_mstate_global->__pyx_n_s_is_sampling_buffer_full
#define __pyx_n_s_is_within_tolerance __pyx_mstate_global->__pyx_n_s_is_within_tolerance
#define __pyx_kp_u_isenabled __pyx_mstate_global->__pyx_kp_u_isenabled
#define __pyx_n_s_isnan __pyx_mstate_global->__pyx_n_s_isnan
#define __pyx_n_s_items __pyx_mstate_global->__pyx_n_s_items
#define __pyx_kp_u_kappa __pyx_mstate_global->__pyx_kp_u_kappa
#define __pyx_n_u_kappa_2 __pyx_mstate_global->__pyx_n_u_kappa_2
#define __pyx_n_s_key __pyx_mstate_global->__pyx_n_s_key
#define __pyx_n_s_len __pyx_mstate_global->__pyx_n_s_len
#define __pyx_n_s_level __pyx_mstate_global->__pyx_n_s_level
#define __pyx_n_s_level_distances __pyx_mstate_global->__pyx_n_s_level_distances
#define __pyx_n_s_level_step __pyx_mstate_global->__pyx_n_s_level_step
#define __pyx_n_s_limit_orders __pyx_mstate_global->__pyx_n_s_limit_orders
#define __pyx_n_s_line __pyx_mstate_global->__pyx_n_s_line
#define __pyx_n_s_lines __pyx_mstate_global->__pyx_n_s_lines
#define __pyx_n_s_ln __pyx_mstate_global->__pyx_n_s_ln
#define __pyx_n_s_log_with_clock __pyx_mstate_global->__pyx_n_s_log_with_clock
#define __pyx_n_s_logger __pyx_mstate_global->__pyx_n_s_logger
#define __pyx_n_s_logging __pyx_mstate_global->__pyx_n_s_logging
#define __pyx_n_s_logging_options __pyx_mstate_global->__pyx_n_s_logging_options
#define __pyx_n_s_lvl_buy __pyx_mstate_global->__pyx_n_s_lvl_buy
#define __pyx_n_s_lvl_sell __pyx_mstate_global->__pyx_n_s_lvl_sell
#define __pyx_n_s_main __pyx_mstate_global->__pyx_n_s_main
#define __pyx_n_s_map_df_to_str __pyx_mstate_global->__pyx_n_s_map_df_to_str
#define __pyx_n_s_market __pyx_mstate_global->__pyx_n_s_market
#define __pyx_n_s_market_books __pyx_mstate_global->__pyx_n_s_market_books
#define __pyx_n_s_market_info __pyx_mstate_global->__pyx_n_s_market_info
#define __pyx_n_s_market_info_to_active_orders __pyx_mstate_global->__pyx_n_s_market_info_to_active_orders
#define __pyx_n_s_market_pair_to_active_orders __pyx_mstate_global->__pyx_n_s_market_pair_to_active_orders
#define __pyx_n_s_market_status_data_frame __pyx_mstate_global->__pyx_n_s_market_status_data_frame
#define __pyx_n_s_market_trading_pair_tuples __pyx_mstate_global->__pyx_n_s_market_trading_pair_tuples
#define __pyx_n_s_markets_columns __pyx_mstate_global->__pyx_n_s_markets_columns
#define __pyx_n_s_markets_data __pyx_mstate_global->__pyx_n_s_markets_data
#define __pyx_n_s_markets_df __pyx_mstate_global->__pyx_n_s_markets_df
#define __pyx_n_s_math __pyx_mstate_global->__pyx_n_s_math
#define __pyx_n_s_max __pyx_mstate_global->__pyx_n_s_max
#define __pyx_n_s_max_order_age __pyx_mstate_global->__pyx_n_s_max_order_age
#define __pyx_n_s_measure_order_book_liquidity __pyx_mstate_global->__pyx_n_s_measure_order_book_liquidity
#define __pyx_kp_u_mid_price __pyx_mstate_global->__pyx_kp_u_mid_price
#define __pyx_n_s_mid_price_2 __pyx_mstate_global->__pyx_n_s_mid_price_2
#define __pyx_n_u_mid_price_2 __pyx_mstate_global->__pyx_n_u_mid_price_2
#define __pyx_kp_u_mid_price_std_dev __pyx_mstate_global->__pyx_kp_u_mid_price_std_dev
#define __pyx_n_s_mid_price_variance __pyx_mstate_global->__pyx_n_s_mid_price_variance
#define __pyx_n_u_mid_price_variance __pyx_mstate_global->__pyx_n_u_mid_price_variance
#define __pyx_n_s_min_spread __pyx_mstate_global->__pyx_n_s_min_spread
#define __pyx_n_s_mode __pyx_mstate_global->__pyx_n_s_mode
#define __pyx_n_s_model_config __pyx_mstate_global->__pyx_n_s_model_config
#define __pyx_n_s_msg __pyx_mstate_global->__pyx_n_s_msg
#define __pyx_n_s_name __pyx_mstate_global->__pyx_n_s_name
#define __pyx_n_s_nan __pyx_mstate_global->__pyx_n_s_nan
#define __pyx_n_u_nan __pyx_mstate_global->__pyx_n_u_nan
#define __pyx_n_s_network_status __pyx_mstate_global->__pyx_n_s_network_status
#define __pyx_n_s_network_warning __pyx_mstate_global->__pyx_n_s_network_warning
#define __pyx_n_s_new_ask __pyx_mstate_global->__pyx_n_s_new_ask
#define __pyx_n_s_new_bid __pyx_mstate_global->__pyx_n_s_new_bid
#define __pyx_n_s_no_sells __pyx_mstate_global->__pyx_n_s_no_sells
#define __pyx_n_s_non_hanging __pyx_mstate_global->__pyx_n_s_non_hanging
#define __pyx_n_s_normalize __pyx_mstate_global->__pyx_n_s_normalize
#define __pyx_n_s_notify_hb_app __pyx_mstate_global->__pyx_n_s_notify_hb_app
#define __pyx_n_s_notify_hb_app_with_timestamp __pyx_mstate_global->__pyx_n_s_notify_hb_app_with_timestamp
#define __pyx_n_s_np __pyx_mstate_global->__pyx_n_s_np
#define __pyx_n_s_numpy __pyx_mstate_global->__pyx_n_s_numpy
#define __pyx_kp_u_numpy_core_multiarray_failed_to __pyx_mstate_global->__pyx_kp_u_numpy_core_multiarray_failed_to
#define __pyx_kp_u_numpy_core_umath_failed_to_impor __pyx_mstate_global->__pyx_kp_u_numpy_core_umath_failed_to_impor
#define __pyx_n_s_o __pyx_mstate_global->__pyx_n_s_o
#define __pyx_kp_u_optimal_ask __pyx_mstate_global->__pyx_kp_u_optimal_ask
#define __pyx_n_u_optimal_ask_2 __pyx_mstate_global->__pyx_n_u_optimal_ask_2
#define __pyx_kp_u_optimal_ask_to_mid __pyx_mstate_global->__pyx_kp_u_optimal_ask_to_mid
#define __pyx_kp_u_optimal_bid __pyx_mstate_global->__pyx_kp_u_optimal_bid
#define __pyx_n_u_optimal_bid_2 __pyx_mstate_global->__pyx_n_u_optimal_bid_2
#define __pyx_kp_u_optimal_bid_to_mid __pyx_mstate_global->__pyx_kp_u_optimal_bid_to_mid
#define __pyx_kp_u_optimal_spread __pyx_mstate_global->__pyx_kp_u_optimal_spread
#define __pyx_n_u_optimal_spread_2 __pyx_mstate_global->__pyx_n_u_optimal_spread_2
#define __pyx_kp_u_order __pyx_mstate_global->__pyx_kp_u_order
#define __pyx_n_s_order_2 __pyx_mstate_global->__pyx_n_s_order_2
#define __pyx_n_s_order_age __pyx_mstate_global->__pyx_n_s_order_age
#define __pyx_n_s_order_amount __pyx_mstate_global->__pyx_n_s_order_amount
#define __pyx_n_s_order_amount_shape_factor __pyx_mstate_global->__pyx_n_s_order_amount_shape_factor
#define __pyx_n_s_order_book __pyx_mstate_global->__pyx_n_s_order_book
#define __pyx_kp_u_order_book_depth_factor __pyx_mstate_global->__pyx_kp_u_order_book_depth_factor
#define __pyx_kp_u_order_book_intensity_factor __pyx_mstate_global->__pyx_kp_u_order_book_intensity_factor
#define __pyx_n_s_order_id __pyx_mstate_global->__pyx_n_s_order_id
#define __pyx_n_s_order_levels __pyx_mstate_global->__pyx_n_s_order_levels
#define __pyx_n_s_order_levels_mode __pyx_mstate_global->__pyx_n_s_order_levels_mode
#define __pyx_n_s_order_optimization_enabled __pyx_mstate_global->__pyx_n_s_order_optimization_enabled
#define __pyx_n_s_order_override __pyx_mstate_global->__pyx_n_s_order_override
#define __pyx_n_s_order_refresh_time __pyx_mstate_global->__pyx_n_s_order_refresh_time
#define __pyx_n_s_order_refresh_tolerance __pyx_mstate_global->__pyx_n_s_order_refresh_tolerance
#define __pyx_n_s_order_refresh_tolerance_pct __pyx_mstate_global->__pyx_n_s_order_refresh_tolerance_pct
#define __pyx_n_s_orders __pyx_mstate_global->__pyx_n_s_orders
#define __pyx_n_s_os __pyx_mstate_global->__pyx_n_s_os
#define __pyx_n_s_p __pyx_mstate_global->__pyx_n_s_p
#define __pyx_n_s_pandas __pyx_mstate_global->__pyx_n_s_pandas
#define __pyx_n_s_path __pyx_mstate_global->__pyx_n_s_path
#define __pyx_n_s_pd __pyx_mstate_global->__pyx_n_s_pd
#define __pyx_kp_s_pd_DataFrame __pyx_mstate_global->__pyx_kp_s_pd_DataFrame
#define __pyx_n_s_percent __pyx_mstate_global->__pyx_n_s_percent
#define __pyx_n_s_pmm_logger __pyx_mstate_global->__pyx_n_s_pmm_logger
#define __pyx_n_s_price __pyx_mstate_global->__pyx_n_s_price
#define __pyx_n_s_price_delegate __pyx_mstate_global->__pyx_n_s_price_delegate
#define __pyx_n_s_process_tick __pyx_mstate_global->__pyx_n_s_process_tick
#define __pyx_n_s_proposal __pyx_mstate_global->__pyx_n_s_proposal
#define __pyx_n_s_proposal_prices __pyx_mstate_global->__pyx_n_s_proposal_prices
#define __pyx_n_s_pure_mm_assets_df __pyx_mstate_global->__pyx_n_s_pure_mm_assets_df
#define __pyx_n_s_pyx_state __pyx_mstate_global->__pyx_n_s_pyx_state
#define __pyx_n_s_pyx_vtable __pyx_mstate_global->__pyx_n_s_pyx_vtable
#define __pyx_kp_u_q __pyx_mstate_global->__pyx_kp_u_q
#define __pyx_n_s_quantity __pyx_mstate_global->__pyx_n_s_quantity
#define __pyx_n_s_quote_asset __pyx_mstate_global->__pyx_n_s_quote_asset
#define __pyx_n_s_quote_balance __pyx_mstate_global->__pyx_n_s_quote_balance
#define __pyx_n_s_quote_currency __pyx_mstate_global->__pyx_n_s_quote_currency
#define __pyx_n_s_quote_ratio __pyx_mstate_global->__pyx_n_s_quote_ratio
#define __pyx_n_s_range __pyx_mstate_global->__pyx_n_s_range
#define __pyx_n_s_ready __pyx_mstate_global->__pyx_n_s_ready
#define __pyx_n_s_reduce __pyx_mstate_global->__pyx_n_s_reduce
#define __pyx_n_s_reduce_cython __pyx_mstate_global->__pyx_n_s_reduce_cython
#define __pyx_n_s_reduce_ex __pyx_mstate_global->__pyx_n_s_reduce_ex
#define __pyx_n_s_ref_price __pyx_mstate_global->__pyx_n_s_ref_price
#define __pyx_n_s_regex __pyx_mstate_global->__pyx_n_s_regex
#define __pyx_n_s_register_events __pyx_mstate_global->__pyx_n_s_register_events
#define __pyx_n_s_replace __pyx_mstate_global->__pyx_n_s_replace
#define __pyx_kp_u_reservation_price __pyx_mstate_global->__pyx_kp_u_reservation_price
#define __pyx_n_u_reservation_price_2 __pyx_mstate_global->__pyx_n_u_reservation_price_2
#define __pyx_n_s_result_price __pyx_mstate_global->__pyx_n_s_result_price
#define __pyx_n_s_return __pyx_mstate_global->__pyx_n_s_return
#define __pyx_n_s_reverse __pyx_mstate_global->__pyx_n_s_reverse
#define __pyx_n_s_risk_factor __pyx_mstate_global->__pyx_n_s_risk_factor
#define __pyx_n_u_risk_factor __pyx_mstate_global->__pyx_n_u_risk_factor
#define __pyx_kp_u_risk_factor_2 __pyx_mstate_global->__pyx_kp_u_risk_factor_2
#define __pyx_n_s_round __pyx_mstate_global->__pyx_n_s_round
#define __pyx_n_u_s __pyx_mstate_global->__pyx_n_u_s
#define __pyx_n_s_s_decimal_neg_one __pyx_mstate_global->__pyx_n_s_s_decimal_neg_one
#define __pyx_n_s_s_decimal_one __pyx_mstate_global->__pyx_n_s_s_decimal_one
#define __pyx_n_s_s_decimal_zero __pyx_mstate_global->__pyx_n_s_s_decimal_zero
#define __pyx_n_s_sampling_length __pyx_mstate_global->__pyx_n_s_sampling_length
#define __pyx_n_s_seconds __pyx_mstate_global->__pyx_n_s_seconds
#define __pyx_n_s_self __pyx_mstate_global->__pyx_n_s_self
#define __pyx_kp_s_self__events_cannot_be_converted __pyx_mstate_global->__pyx_kp_s_self__events_cannot_be_converted
#define __pyx_n_u_sell __pyx_mstate_global->__pyx_n_u_sell
#define __pyx_n_s_sell_order __pyx_mstate_global->__pyx_n_s_sell_order
#define __pyx_n_s_sells __pyx_mstate_global->__pyx_n_s_sells
#define __pyx_n_s_send __pyx_mstate_global->__pyx_n_s_send
#define __pyx_n_s_set_timers __pyx_mstate_global->__pyx_n_s_set_timers
#define __pyx_n_s_setstate __pyx_mstate_global->__pyx_n_s_setstate
#define __pyx_n_s_setstate_cython __pyx_mstate_global->__pyx_n_s_setstate_cython
#define __pyx_n_s_should_wait_order_cancel_confirm __pyx_mstate_global->__pyx_n_s_should_wait_order_cancel_confirm
#define __pyx_n_s_size __pyx_mstate_global->__pyx_n_s_size
#define __pyx_n_s_sort __pyx_mstate_global->__pyx_n_s_sort
#define __pyx_n_s_sorted __pyx_mstate_global->__pyx_n_s_sorted
#define __pyx_n_s_spec __pyx_mstate_global->__pyx_n_s_spec
#define __pyx_n_s_split __pyx_mstate_global->__pyx_n_s_split
#define __pyx_n_s_spread __pyx_mstate_global->__pyx_n_s_spread
#define __pyx_n_s_start __pyx_mstate_global->__pyx_n_s_start
#define __pyx_n_s_start_datetime __pyx_mstate_global->__pyx_n_s_start_datetime
#define __pyx_n_s_start_time __pyx_mstate_global->__pyx_n_s_start_time
#define __pyx_n_s_start_timestamp __pyx_mstate_global->__pyx_n_s_start_timestamp
#define __pyx_n_s_status_report_interval __pyx_mstate_global->__pyx_n_s_status_report_interval
#define __pyx_n_s_str __pyx_mstate_global->__pyx_n_s_str
#define __pyx_n_s_strftime __pyx_mstate_global->__pyx_n_s_strftime
#define __pyx_kp_s_stringsource __pyx_mstate_global->__pyx_kp_s_stringsource
#define __pyx_n_s_super __pyx_mstate_global->__pyx_n_s_super
#define __pyx_n_u_target_inv __pyx_mstate_global->__pyx_n_u_target_inv
#define __pyx_n_s_test __pyx_mstate_global->__pyx_n_s_test
#define __pyx_n_s_throw __pyx_mstate_global->__pyx_n_s_throw
#define __pyx_n_s_ticks_to_be_ready_after __pyx_mstate_global->__pyx_n_s_ticks_to_be_ready_after
#define __pyx_n_s_ticks_to_be_ready_before __pyx_mstate_global->__pyx_n_s_ticks_to_be_ready_before
#define __pyx_kp_u_ticks_to_fill_buffers __pyx_mstate_global->__pyx_kp_u_ticks_to_fill_buffers
#define __pyx_n_s_time __pyx_mstate_global->__pyx_n_s_time
#define __pyx_n_s_time_left __pyx_mstate_global->__pyx_n_s_time_left
#define __pyx_n_s_time_left_fraction __pyx_mstate_global->__pyx_n_s_time_left_fraction
#define __pyx_n_u_time_left_fraction __pyx_mstate_global->__pyx_n_u_time_left_fraction
#define __pyx_kp_u_time_until_end_of_trading_cycle __pyx_mstate_global->__pyx_kp_u_time_until_end_of_trading_cycle
#define __pyx_kp_u_time_until_end_of_trading_cycle_2 __pyx_mstate_global->__pyx_kp_u_time_until_end_of_trading_cycle_2
#define __pyx_n_s_timedelta __pyx_mstate_global->__pyx_n_s_timedelta
#define __pyx_n_s_timestamp __pyx_mstate_global->__pyx_n_s_timestamp
#define __pyx_n_s_title __pyx_mstate_global->__pyx_n_s_title
#define __pyx_n_u_title __pyx_mstate_global->__pyx_n_u_title
#define __pyx_n_s_to_create_orders __pyx_mstate_global->__pyx_n_s_to_create_orders
#define __pyx_n_s_to_csv __pyx_mstate_global->__pyx_n_s_to_csv
#define __pyx_n_s_to_show_current_pct __pyx_mstate_global->__pyx_n_s_to_show_current_pct
#define __pyx_n_s_to_string __pyx_mstate_global->__pyx_n_s_to_string
#define __pyx_n_s_total_in_quote __pyx_mstate_global->__pyx_n_s_total_in_quote
#define __pyx_n_s_trade_type __pyx_mstate_global->__pyx_n_s_trade_type
#define __pyx_n_s_trading_intensity_buffer_size __pyx_mstate_global->__pyx_n_s_trading_intensity_buffer_size
#define __pyx_n_s_trading_pair __pyx_mstate_global->__pyx_n_s_trading_pair
#define __pyx_n_s_typing __pyx_mstate_global->__pyx_n_s_typing
#define __pyx_n_s_unit __pyx_mstate_global->__pyx_n_s_unit
#define __pyx_n_s_unlink __pyx_mstate_global->__pyx_n_s_unlink
#define __pyx_n_s_unregister_events __pyx_mstate_global->__pyx_n_s_unregister_events
#define __pyx_n_s_update_from_config_map __pyx_mstate_global->__pyx_n_s_update_from_config_map
#define __pyx_n_s_update_strategy_orders_with_equi __pyx_mstate_global->__pyx_n_s_update_strategy_orders_with_equi
#define __pyx_n_s_upper __pyx_mstate_global->__pyx_n_s_upper
#define __pyx_n_s_values __pyx_mstate_global->__pyx_n_s_values
#define __pyx_kp_u_vol __pyx_mstate_global->__pyx_kp_u_vol
#define __pyx_n_s_vol_2 __pyx_mstate_global->__pyx_n_s_vol_2
#define __pyx_kp_u_volatility __pyx_mstate_global->__pyx_kp_u_volatility
#define __pyx_n_u_volatility_2 __pyx_mstate_global->__pyx_n_u_volatility_2
#define __pyx_n_s_volatility_buffer_size __pyx_mstate_global->__pyx_n_s_volatility_buffer_size
#define __pyx_n_s_volatility_pct __pyx_mstate_global->__pyx_n_s_volatility_pct
#define __pyx_n_s_warning __pyx_mstate_global->__pyx_n_s_warning
#define __pyx_n_s_warning_lines __pyx_mstate_global->__pyx_n_s_warning_lines
#define __pyx_n_s_x __pyx_mstate_global->__pyx_n_s_x
#define __pyx_n_s_zip __pyx_mstate_global->__pyx_n_s_zip
#define __pyx_float_100_0 __pyx_mstate_global->__pyx_float_100_0
#define __pyx_int_0 __pyx_mstate_global->__pyx_int_0
#define __pyx_int_1 __pyx_mstate_global->__pyx_int_1
#define __pyx_int_2 __pyx_mstate_global->__pyx_int_2
#define __pyx_int_4 __pyx_mstate_global->__pyx_int_4
#define __pyx_int_5 __pyx_mstate_global->__pyx_int_5
#define __pyx_int_8 __pyx_mstate_global->__pyx_int_8
#define __pyx_int_16 __pyx_mstate_global->__pyx_int_16
#define __pyx_int_32 __pyx_mstate_global->__pyx_int_32
#define __pyx_int_100 __pyx_mstate_global->__pyx_int_100
#define __pyx_int_900 __pyx_mstate_global->__pyx_int_900
#define __pyx_int_0x7fffffffffffffff __pyx_mstate_global->__pyx_int_0x7fffffffffffffff
#define __pyx_int_neg_1 __pyx_mstate_global->__pyx_int_neg_1
#define __pyx_k__3 __pyx_mstate_global->__pyx_k__3
#define __pyx_tuple_ __pyx_mstate_global->__pyx_tuple_
#define __pyx_tuple__2 __pyx_mstate_global->__pyx_tuple__2
#define __pyx_tuple__16 __pyx_mstate_global->__pyx_tuple__16
#define __pyx_tuple__19 __pyx_mstate_global->__pyx_tuple__19
#define __pyx_tuple__20 __pyx_mstate_global->__pyx_tuple__20
#define __pyx_tuple__21 __pyx_mstate_global->__pyx_tuple__21
#define __pyx_tuple__22 __pyx_mstate_global->__pyx_tuple__22
#define __pyx_tuple__24 __pyx_mstate_global->__pyx_tuple__24
#define __pyx_tuple__26 __pyx_mstate_global->__pyx_tuple__26
#define __pyx_tuple__28 __pyx_mstate_global->__pyx_tuple__28
#define __pyx_tuple__33 __pyx_mstate_global->__pyx_tuple__33
#define __pyx_tuple__35 __pyx_mstate_global->__pyx_tuple__35
#define __pyx_tuple__37 __pyx_mstate_global->__pyx_tuple__37
#define __pyx_tuple__39 __pyx_mstate_global->__pyx_tuple__39
#define __pyx_tuple__41 __pyx_mstate_global->__pyx_tuple__41
#define __pyx_tuple__43 __pyx_mstate_global->__pyx_tuple__43
#define __pyx_tuple__45 __pyx_mstate_global->__pyx_tuple__45
#define __pyx_tuple__47 __pyx_mstate_global->__pyx_tuple__47
#define __pyx_tuple__49 __pyx_mstate_global->__pyx_tuple__49
#define __pyx_tuple__51 __pyx_mstate_global->__pyx_tuple__51
#define __pyx_tuple__53 __pyx_mstate_global->__pyx_tuple__53
#define __pyx_tuple__55 __pyx_mstate_global->__pyx_tuple__55
#define __pyx_tuple__58 __pyx_mstate_global->__pyx_tuple__58
#define __pyx_tuple__66 __pyx_mstate_global->__pyx_tuple__66
#define __pyx_tuple__72 __pyx_mstate_global->__pyx_tuple__72
#define __pyx_tuple__76 __pyx_mstate_global->__pyx_tuple__76
#define __pyx_tuple__82 __pyx_mstate_global->__pyx_tuple__82
#define __pyx_tuple__85 __pyx_mstate_global->__pyx_tuple__85
#define __pyx_tuple__89 __pyx_mstate_global->__pyx_tuple__89
#define __pyx_tuple__91 __pyx_mstate_global->__pyx_tuple__91
#define __pyx_tuple__94 __pyx_mstate_global->__pyx_tuple__94
#define __pyx_codeobj__23 __pyx_mstate_global->__pyx_codeobj__23
#define __pyx_codeobj__25 __pyx_mstate_global->__pyx_codeobj__25
#define __pyx_codeobj__27 __pyx_mstate_global->__pyx_codeobj__27
#define __pyx_codeobj__29 __pyx_mstate_global->__pyx_codeobj__29
#define __pyx_codeobj__30 __pyx_mstate_global->__pyx_codeobj__30
#define __pyx_codeobj__31 __pyx_mstate_global->__pyx_codeobj__31
#define __pyx_codeobj__32 __pyx_mstate_global->__pyx_codeobj__32
#define __pyx_codeobj__34 __pyx_mstate_global->__pyx_codeobj__34
#define __pyx_codeobj__36 __pyx_mstate_global->__pyx_codeobj__36
#define __pyx_codeobj__38 __pyx_mstate_global->__pyx_codeobj__38
#define __pyx_codeobj__40 __pyx_mstate_global->__pyx_codeobj__40
#define __pyx_codeobj__42 __pyx_mstate_global->__pyx_codeobj__42
#define __pyx_codeobj__44 __pyx_mstate_global->__pyx_codeobj__44
#define __pyx_codeobj__46 __pyx_mstate_global->__pyx_codeobj__46
#define __pyx_codeobj__48 __pyx_mstate_global->__pyx_codeobj__48
#define __pyx_codeobj__50 __pyx_mstate_global->__pyx_codeobj__50
#define __pyx_codeobj__52 __pyx_mstate_global->__pyx_codeobj__52
#define __pyx_codeobj__54 __pyx_mstate_global->__pyx_codeobj__54
#define __pyx_codeobj__56 __pyx_mstate_global->__pyx_codeobj__56
#define __pyx_codeobj__57 __pyx_mstate_global->__pyx_codeobj__57
#define __pyx_codeobj__59 __pyx_mstate_global->__pyx_codeobj__59
#define __pyx_codeobj__60 __pyx_mstate_global->__pyx_codeobj__60
#define __pyx_codeobj__61 __pyx_mstate_global->__pyx_codeobj__61
#define __pyx_codeobj__62 __pyx_mstate_global->__pyx_codeobj__62
#define __pyx_codeobj__63 __pyx_mstate_global->__pyx_codeobj__63
#define __pyx_codeobj__64 __pyx_mstate_global->__pyx_codeobj__64
#define __pyx_codeobj__65 __pyx_mstate_global->__pyx_codeobj__65
#define __pyx_codeobj__67 __pyx_mstate_global->__pyx_codeobj__67
#define __pyx_codeobj__68 __pyx_mstate_global->__pyx_codeobj__68
#define __pyx_codeobj__69 __pyx_mstate_global->__pyx_codeobj__69
#define __pyx_codeobj__70 __pyx_mstate_global->__pyx_codeobj__70
#define __pyx_codeobj__71 __pyx_mstate_global->__pyx_codeobj__71
#define __pyx_codeobj__73 __pyx_mstate_global->__pyx_codeobj__73
#define __pyx_codeobj__74 __pyx_mstate_global->__pyx_codeobj__74
#define __pyx_codeobj__75 __pyx_mstate_global->__pyx_codeobj__75
#define __pyx_codeobj__77 __pyx_mstate_global->__pyx_codeobj__77
#define __pyx_codeobj__78 __pyx_mstate_global->__pyx_codeobj__78
#define __pyx_codeobj__79 __pyx_mstate_global->__pyx_codeobj__79
#define __pyx_codeobj__80 __pyx_mstate_global->__pyx_codeobj__80
#define __pyx_codeobj__81 __pyx_mstate_global->__pyx_codeobj__81
#define __pyx_codeobj__83 __pyx_mstate_global->__pyx_codeobj__83
#define __pyx_codeobj__84 __pyx_mstate_global->__pyx_codeobj__84
#define __pyx_codeobj__86 __pyx_mstate_global->__pyx_codeobj__86
#define __pyx_codeobj__87 __pyx_mstate_global->__pyx_codeobj__87
#define __pyx_codeobj__88 __pyx_mstate_global->__pyx_codeobj__88
#define __pyx_codeobj__90 __pyx_mstate_global->__pyx_codeobj__90
#define __pyx_codeobj__92 __pyx_mstate_global->__pyx_codeobj__92
#define __pyx_codeobj__93 __pyx_mstate_global->__pyx_codeobj__93
#define __pyx_codeobj__95 __pyx_mstate_global->__pyx_codeobj__95
/* #### Code section: module_code ### */

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":245
 * 
 *         @property
 *         cdef inline PyObject* base(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_7ndarray_4base_base(PyArrayObject *__pyx_v_self) {
  PyObject *__pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":248
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 *             return PyArray_BASE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_BASE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":245
 * 
 *         @property
 *         cdef inline PyObject* base(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns a borrowed reference to the object owning the data/memory.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":251
 * 
 *         @property
 *         cdef inline dtype descr(self):             # <<<<<<<<<<<<<<
 *             """Returns an owned reference to the dtype of the array.
 *             """
 */

static CYTHON_INLINE PyArray_Descr *__pyx_f_5numpy_7ndarray_5descr_descr(PyArrayObject *__pyx_v_self) {
  PyArray_Descr *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyArray_Descr *__pyx_t_1;
  __Pyx_RefNannySetupContext("descr", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":254
 *             """Returns an owned reference to the dtype of the array.
 *             """
 *             return <dtype>PyArray_DESCR(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_1 = PyArray_DESCR(__pyx_v_self);
  __Pyx_INCREF((PyObject *)((PyArray_Descr *)__pyx_t_1));
  __pyx_r = ((PyArray_Descr *)__pyx_t_1);
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":251
 * 
 *         @property
 *         cdef inline dtype descr(self):             # <<<<<<<<<<<<<<
 *             """Returns an owned reference to the dtype of the array.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":257
 * 
 *         @property
 *         cdef inline int ndim(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns the number of dimensions in the array.
 *             """
 */

static CYTHON_INLINE int __pyx_f_5numpy_7ndarray_4ndim_ndim(PyArrayObject *__pyx_v_self) {
  int __pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":260
 *             """Returns the number of dimensions in the array.
 *             """
 *             return PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_NDIM(__pyx_v_self);
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":257
 * 
 *         @property
 *         cdef inline int ndim(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns the number of dimensions in the array.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":263
 * 
 *         @property
 *         cdef inline npy_intp *shape(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the dimensions/shape of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_5shape_shape(PyArrayObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":268
 *             Can return NULL for 0-dimensional arrays.
 *             """
 *             return PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_DIMS(__pyx_v_self);
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":263
 * 
 *         @property
 *         cdef inline npy_intp *shape(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the dimensions/shape of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":271
 * 
 *         @property
 *         cdef inline npy_intp *strides(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the strides of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

static CYTHON_INLINE npy_intp *__pyx_f_5numpy_7ndarray_7strides_strides(PyArrayObject *__pyx_v_self) {
  npy_intp *__pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":275
 *             The number of elements matches the number of dimensions of the array (ndim).
 *             """
 *             return PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_STRIDES(__pyx_v_self);
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":271
 * 
 *         @property
 *         cdef inline npy_intp *strides(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns a pointer to the strides of the array.
 *             The number of elements matches the number of dimensions of the array (ndim).
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":278
 * 
 *         @property
 *         cdef inline npy_intp size(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns the total size (in number of elements) of the array.
 *             """
 */

static CYTHON_INLINE npy_intp __pyx_f_5numpy_7ndarray_4size_size(PyArrayObject *__pyx_v_self) {
  npy_intp __pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":281
 *             """Returns the total size (in number of elements) of the array.
 *             """
 *             return PyArray_SIZE(self)             # <<<<<<<<<<<<<<
 * 
 *         @property
 */
  __pyx_r = PyArray_SIZE(__pyx_v_self);
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":278
 * 
 *         @property
 *         cdef inline npy_intp size(self) nogil:             # <<<<<<<<<<<<<<
 *             """Returns the total size (in number of elements) of the array.
 *             """
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":284
 * 
 *         @property
 *         cdef inline char* data(self) nogil:             # <<<<<<<<<<<<<<
 *             """The pointer to the data buffer as a char*.
 *             This is provided for legacy reasons to avoid direct struct field access.
 */

static CYTHON_INLINE char *__pyx_f_5numpy_7ndarray_4data_data(PyArrayObject *__pyx_v_self) {
  char *__pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":290
 *             of `PyArray_DATA()` instead, which returns a 'void*'.
 *             """
 *             return PyArray_BYTES(self)             # <<<<<<<<<<<<<<
 * 
 *     ctypedef unsigned char      npy_bool
 */
  __pyx_r = PyArray_BYTES(__pyx_v_self);
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":284
 * 
 *         @property
 *         cdef inline char* data(self) nogil:             # <<<<<<<<<<<<<<
 *             """The pointer to the data buffer as a char*.
 *             This is provided for legacy reasons to avoid direct struct field access.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":773
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":774
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":773
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":776
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":777
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":776
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":779
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":780
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":779
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":782
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":783
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":782
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":785
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":786
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":785
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":788
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyDataType_SHAPE(PyArray_Descr *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("PyDataType_SHAPE", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":789
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  __pyx_t_1 = PyDataType_HASSUBARRAY(__pyx_v_d);
  if (__pyx_t_1) {

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":790
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape             # <<<<<<<<<<<<<<
 *     else:
 *         return ()
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject*)__pyx_v_d->subarray->shape));
    __pyx_r = ((PyObject*)__pyx_v_d->subarray->shape);
    goto __pyx_L0;

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":789
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):
 *     if PyDataType_HASSUBARRAY(d):             # <<<<<<<<<<<<<<
 *         return <tuple>d.subarray.shape
 *     else:
 */
  }

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":792
 *         return <tuple>d.subarray.shape
 *     else:
 *         return ()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_empty_tuple);
    __pyx_r = __pyx_empty_tuple;
    goto __pyx_L0;
  }

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":788
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline tuple PyDataType_SHAPE(dtype d):             # <<<<<<<<<<<<<<
 *     if PyDataType_HASSUBARRAY(d):
 *         return <tuple>d.subarray.shape
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":968
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":969
 * 
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!             # <<<<<<<<<<<<<<
 *     PyArray_SetBaseObject(arr, base)
 * 
 */
  Py_INCREF(__pyx_v_base);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":970
 * cdef inline void set_array_base(ndarray arr, object base):
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  __pyx_t_1 = PyArray_SetBaseObject(__pyx_v_arr, __pyx_v_base); if (unlikely(__pyx_t_1 == ((int)-1))) __PYX_ERR(2, 970, __pyx_L1_error)

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":968
 *     int _import_umath() except -1
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *     Py_INCREF(base) # important to do this before stealing the reference below!
 *     PyArray_SetBaseObject(arr, base)
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("numpy.set_array_base", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":972
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_v_base;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":973
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)             # <<<<<<<<<<<<<<
 *     if base is NULL:
 *         return None
 */
  __pyx_v_base = PyArray_BASE(__pyx_v_arr);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":974
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  __pyx_t_1 = (__pyx_v_base == NULL);
  if (__pyx_t_1) {

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":975
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     return <object>base
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":974
 * cdef inline object get_array_base(ndarray arr):
 *     base = PyArray_BASE(arr)
 *     if base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     return <object>base
 */
  }

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":976
 *     if base is NULL:
 *         return None
 *     return <object>base             # <<<<<<<<<<<<<<
 * 
 * # Versions of the import_* functions which are more suitable for
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_base));
  __pyx_r = ((PyObject *)__pyx_v_base);
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":972
 *     PyArray_SetBaseObject(arr, base)
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     base = PyArray_BASE(arr)
 *     if base is NULL:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":980
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_array", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":981
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":982
 * cdef inline int import_array() except -1:
 *     try:
 *         __pyx_import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 982, __pyx_L3_error)

      /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":981
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":983
 *     try:
 *         __pyx_import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 983, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":984
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 984, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 984, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":981
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         __pyx_import_array()
 *     except Exception:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":980
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         __pyx_import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":986
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_umath", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":987
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":988
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 988, __pyx_L3_error)

      /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":987
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":989
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 989, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":990
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 990, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 990, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":987
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":986
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":992
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("import_ufunc", 1);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":993
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":994
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(2, 994, __pyx_L3_error)

      /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":993
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":995
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(2, 995, __pyx_L5_except_error)
      __Pyx_XGOTREF(__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_6);
      __Pyx_XGOTREF(__pyx_t_7);

      /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":996
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(2, 996, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(2, 996, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;

    /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":993
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L8_try_end:;
  }

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":992
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":999
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_timedelta64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1011
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyTimedeltaArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyTimedeltaArrType_Type));
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":999
 * 
 * 
 * cdef inline bint is_timedelta64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.timedelta64)`
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1014
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

static CYTHON_INLINE int __pyx_f_5numpy_is_datetime64_object(PyObject *__pyx_v_obj) {
  int __pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1026
 *     bool
 *     """
 *     return PyObject_TypeCheck(obj, &PyDatetimeArrType_Type)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = PyObject_TypeCheck(__pyx_v_obj, (&PyDatetimeArrType_Type));
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1014
 * 
 * 
 * cdef inline bint is_datetime64_object(object obj):             # <<<<<<<<<<<<<<
 *     """
 *     Cython equivalent of `isinstance(obj, np.datetime64)`
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1029
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

static CYTHON_INLINE npy_datetime __pyx_f_5numpy_get_datetime64_value(PyObject *__pyx_v_obj) {
  npy_datetime __pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1036
 *     also needed.  That can be found using `get_datetime64_unit`.
 *     """
 *     return (<PyDatetimeScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyDatetimeScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1029
 * 
 * 
 * cdef inline npy_datetime get_datetime64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy datetime64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1039
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

static CYTHON_INLINE npy_timedelta __pyx_f_5numpy_get_timedelta64_value(PyObject *__pyx_v_obj) {
  npy_timedelta __pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1043
 *     returns the int64 value underlying scalar numpy timedelta64 object
 *     """
 *     return (<PyTimedeltaScalarObject*>obj).obval             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((PyTimedeltaScalarObject *)__pyx_v_obj)->obval;
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1039
 * 
 * 
 * cdef inline npy_timedelta get_timedelta64_value(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the int64 value underlying scalar numpy timedelta64 object
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1046
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

static CYTHON_INLINE NPY_DATETIMEUNIT __pyx_f_5numpy_get_datetime64_unit(PyObject *__pyx_v_obj) {
  NPY_DATETIMEUNIT __pyx_r;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1050
 *     returns the unit part of the dtype for a numpy datetime64 object.
 *     """
 *     return <NPY_DATETIMEUNIT>(<PyDatetimeScalarObject*>obj).obmeta.base             # <<<<<<<<<<<<<<
 */
  __pyx_r = ((NPY_DATETIMEUNIT)((PyDatetimeScalarObject *)__pyx_v_obj)->obmeta.base);
  goto __pyx_L0;

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":1046
 * 
 * 
 * cdef inline NPY_DATETIMEUNIT get_datetime64_unit(object obj) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     returns the unit part of the dtype for a numpy datetime64 object.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":66
 *     OPTION_LOG_ALL = 0x7fffffffffffffff
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def logger(cls):
 *         global pmm_logger
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_1logger(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_1logger = {"logger", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_1logger, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_1logger(PyObject *__pyx_v_cls, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("logger (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("logger", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "logger", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_logger(((PyTypeObject*)__pyx_v_cls));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_logger(CYTHON_UNUSED PyTypeObject *__pyx_v_cls) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("logger", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":69
 *     def logger(cls):
 *         global pmm_logger
 *         if pmm_logger is None:             # <<<<<<<<<<<<<<
 *             pmm_logger = logging.getLogger(__name__)
 *         return pmm_logger
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pmm_logger); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":70
 *         global pmm_logger
 *         if pmm_logger is None:
 *             pmm_logger = logging.getLogger(__name__)             # <<<<<<<<<<<<<<
 *         return pmm_logger
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_logging); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_getLogger); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 70, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_pmm_logger, __pyx_t_1) < 0) __PYX_ERR(0, 70, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":69
 *     def logger(cls):
 *         global pmm_logger
 *         if pmm_logger is None:             # <<<<<<<<<<<<<<
 *             pmm_logger = logging.getLogger(__name__)
 *         return pmm_logger
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":71
 *         if pmm_logger is None:
 *             pmm_logger = logging.getLogger(__name__)
 *         return pmm_logger             # <<<<<<<<<<<<<<
 * 
 *     def init_params(self,
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pmm_logger); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":66
 *     OPTION_LOG_ALL = 0x7fffffffffffffff
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def logger(cls):
 *         global pmm_logger
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.logger", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":73
 *         return pmm_logger
 * 
 *     def init_params(self,             # <<<<<<<<<<<<<<
 *                     config_map: Union[AvellanedaMarketMakingConfigMap, ClientConfigAdapter],
 *                     market_info: MarketTradingPairTuple,
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_3init_params(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_3init_params = {"init_params", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_3init_params, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_3init_params(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_config_map = 0;
  PyObject *__pyx_v_market_info = 0;
  PyObject *__pyx_v_logging_options = 0;
  PyObject *__pyx_v_status_report_interval = 0;
  PyObject *__pyx_v_hb_app_notification = 0;
  PyObject *__pyx_v_debug_csv_path = 0;
  PyObject *__pyx_v_is_debug = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[7] = {0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("init_params (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_config_map,&__pyx_n_s_market_info,&__pyx_n_s_logging_options,&__pyx_n_s_status_report_interval,&__pyx_n_s_hb_app_notification,&__pyx_n_s_debug_csv_path,&__pyx_n_s_is_debug,0};
    values[2] = __Pyx_Arg_NewRef_FASTCALL(__pyx_k__3);
    values[3] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_int_900));

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":78
 *                     logging_options: int = OPTION_LOG_ALL,
 *                     status_report_interval: float = 900,
 *                     hb_app_notification: bool = False,             # <<<<<<<<<<<<<<
 *                     debug_csv_path: str = '',
 *                     is_debug: bool = False,
 */
    values[4] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_False));
    values[5] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)__pyx_kp_u__4));

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":80
 *                     hb_app_notification: bool = False,
 *                     debug_csv_path: str = '',
 *                     is_debug: bool = False,             # <<<<<<<<<<<<<<
 *                     ):
 *         self._sb_order_tracker = OrderTracker()
 */
    values[6] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_False));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  7: values[6] = __Pyx_Arg_FASTCALL(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = __Pyx_Arg_FASTCALL(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_config_map)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 73, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_market_info)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 73, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("init_params", 0, 2, 7, 1); __PYX_ERR(0, 73, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_logging_options);
          if (value) { values[2] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 73, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_status_report_interval);
          if (value) { values[3] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 73, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_hb_app_notification);
          if (value) { values[4] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 73, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  5:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_debug_csv_path);
          if (value) { values[5] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 73, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  6:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_is_debug);
          if (value) { values[6] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 73, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "init_params") < 0)) __PYX_ERR(0, 73, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  7: values[6] = __Pyx_Arg_FASTCALL(__pyx_args, 6);
        CYTHON_FALLTHROUGH;
        case  6: values[5] = __Pyx_Arg_FASTCALL(__pyx_args, 5);
        CYTHON_FALLTHROUGH;
        case  5: values[4] = __Pyx_Arg_FASTCALL(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = __Pyx_Arg_FASTCALL(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = __Pyx_Arg_FASTCALL(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_config_map = values[0];
    __pyx_v_market_info = values[1];
    __pyx_v_logging_options = values[2];
    __pyx_v_status_report_interval = values[3];
    __pyx_v_hb_app_notification = values[4];
    __pyx_v_debug_csv_path = values[5];
    __pyx_v_is_debug = values[6];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("init_params", 0, 2, 7, __pyx_nargs); __PYX_ERR(0, 73, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.init_params", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_2init_params(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_config_map, __pyx_v_market_info, __pyx_v_logging_options, __pyx_v_status_report_interval, __pyx_v_hb_app_notification, __pyx_v_debug_csv_path, __pyx_v_is_debug);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":73
 *         return pmm_logger
 * 
 *     def init_params(self,             # <<<<<<<<<<<<<<
 *                     config_map: Union[AvellanedaMarketMakingConfigMap, ClientConfigAdapter],
 *                     market_info: MarketTradingPairTuple,
 */

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_2init_params(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_config_map, PyObject *__pyx_v_market_info, PyObject *__pyx_v_logging_options, PyObject *__pyx_v_status_report_interval, PyObject *__pyx_v_hb_app_notification, PyObject *__pyx_v_debug_csv_path, PyObject *__pyx_v_is_debug) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int64_t __pyx_t_8;
  double __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("init_params", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":82
 *                     is_debug: bool = False,
 *                     ):
 *         self._sb_order_tracker = OrderTracker()             # <<<<<<<<<<<<<<
 *         self._config_map = config_map
 *         self._market_info = market_info
 */
  __pyx_t_1 = __Pyx_PyObject_CallNoArg(((PyObject *)__pyx_ptype_10hummingbot_8strategy_13order_tracker_OrderTracker)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->__pyx_base._sb_order_tracker);
  __Pyx_DECREF((PyObject *)__pyx_v_self->__pyx_base._sb_order_tracker);
  __pyx_v_self->__pyx_base._sb_order_tracker = ((struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":83
 *                     ):
 *         self._sb_order_tracker = OrderTracker()
 *         self._config_map = config_map             # <<<<<<<<<<<<<<
 *         self._market_info = market_info
 *         self._price_delegate = OrderBookAssetPriceDelegate(market_info.market, market_info.trading_pair)
 */
  __Pyx_INCREF(__pyx_v_config_map);
  __Pyx_GIVEREF(__pyx_v_config_map);
  __Pyx_GOTREF(__pyx_v_self->_config_map);
  __Pyx_DECREF(__pyx_v_self->_config_map);
  __pyx_v_self->_config_map = __pyx_v_config_map;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":84
 *         self._sb_order_tracker = OrderTracker()
 *         self._config_map = config_map
 *         self._market_info = market_info             # <<<<<<<<<<<<<<
 *         self._price_delegate = OrderBookAssetPriceDelegate(market_info.market, market_info.trading_pair)
 *         self._hb_app_notification = hb_app_notification
 */
  __Pyx_INCREF(__pyx_v_market_info);
  __Pyx_GIVEREF(__pyx_v_market_info);
  __Pyx_GOTREF(__pyx_v_self->_market_info);
  __Pyx_DECREF(__pyx_v_self->_market_info);
  __pyx_v_self->_market_info = __pyx_v_market_info;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":85
 *         self._config_map = config_map
 *         self._market_info = market_info
 *         self._price_delegate = OrderBookAssetPriceDelegate(market_info.market, market_info.trading_pair)             # <<<<<<<<<<<<<<
 *         self._hb_app_notification = hb_app_notification
 *         self._hanging_orders_enabled = False
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_OrderBookAssetPriceDelegate); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_market_info, __pyx_n_s_market); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_market_info, __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 85, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_price_delegate);
  __Pyx_DECREF(__pyx_v_self->_price_delegate);
  __pyx_v_self->_price_delegate = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":86
 *         self._market_info = market_info
 *         self._price_delegate = OrderBookAssetPriceDelegate(market_info.market, market_info.trading_pair)
 *         self._hb_app_notification = hb_app_notification             # <<<<<<<<<<<<<<
 *         self._hanging_orders_enabled = False
 *         self._hanging_orders_cancel_pct = Decimal("10")
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_hb_app_notification); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 86, __pyx_L1_error)
  __pyx_v_self->_hb_app_notification = __pyx_t_7;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":87
 *         self._price_delegate = OrderBookAssetPriceDelegate(market_info.market, market_info.trading_pair)
 *         self._hb_app_notification = hb_app_notification
 *         self._hanging_orders_enabled = False             # <<<<<<<<<<<<<<
 *         self._hanging_orders_cancel_pct = Decimal("10")
 *         self._hanging_orders_tracker = HangingOrdersTracker(self,
 */
  __pyx_v_self->_hanging_orders_enabled = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":88
 *         self._hb_app_notification = hb_app_notification
 *         self._hanging_orders_enabled = False
 *         self._hanging_orders_cancel_pct = Decimal("10")             # <<<<<<<<<<<<<<
 *         self._hanging_orders_tracker = HangingOrdersTracker(self,
 *                                                             self._hanging_orders_cancel_pct / Decimal('100'))
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_u_10};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 88, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_hanging_orders_cancel_pct);
  __Pyx_DECREF(__pyx_v_self->_hanging_orders_cancel_pct);
  __pyx_v_self->_hanging_orders_cancel_pct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":89
 *         self._hanging_orders_enabled = False
 *         self._hanging_orders_cancel_pct = Decimal("10")
 *         self._hanging_orders_tracker = HangingOrdersTracker(self,             # <<<<<<<<<<<<<<
 *                                                             self._hanging_orders_cancel_pct / Decimal('100'))
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_HangingOrdersTracker); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":90
 *         self._hanging_orders_cancel_pct = Decimal("10")
 *         self._hanging_orders_tracker = HangingOrdersTracker(self,
 *                                                             self._hanging_orders_cancel_pct / Decimal('100'))             # <<<<<<<<<<<<<<
 * 
 *         self._cancel_timestamp = 0
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_kp_u_100};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 90, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = __Pyx_PyNumber_Divide(__pyx_v_self->_hanging_orders_cancel_pct, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_v_self), __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":89
 *         self._hanging_orders_enabled = False
 *         self._hanging_orders_cancel_pct = Decimal("10")
 *         self._hanging_orders_tracker = HangingOrdersTracker(self,             # <<<<<<<<<<<<<<
 *                                                             self._hanging_orders_cancel_pct / Decimal('100'))
 * 
 */
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_hanging_orders_tracker);
  __Pyx_DECREF(__pyx_v_self->_hanging_orders_tracker);
  __pyx_v_self->_hanging_orders_tracker = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":92
 *                                                             self._hanging_orders_cancel_pct / Decimal('100'))
 * 
 *         self._cancel_timestamp = 0             # <<<<<<<<<<<<<<
 *         self._create_timestamp = 0
 *         self._limit_order_type = self._market_info.market.get_maker_order_type()
 */
  __pyx_v_self->_cancel_timestamp = 0.0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":93
 * 
 *         self._cancel_timestamp = 0
 *         self._create_timestamp = 0             # <<<<<<<<<<<<<<
 *         self._limit_order_type = self._market_info.market.get_maker_order_type()
 *         self._all_markets_ready = False
 */
  __pyx_v_self->_create_timestamp = 0.0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":94
 *         self._cancel_timestamp = 0
 *         self._create_timestamp = 0
 *         self._limit_order_type = self._market_info.market.get_maker_order_type()             # <<<<<<<<<<<<<<
 *         self._all_markets_ready = False
 *         self._filled_buys_balance = 0
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_market); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_get_maker_order_type); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 94, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_limit_order_type);
  __Pyx_DECREF(__pyx_v_self->_limit_order_type);
  __pyx_v_self->_limit_order_type = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":95
 *         self._create_timestamp = 0
 *         self._limit_order_type = self._market_info.market.get_maker_order_type()
 *         self._all_markets_ready = False             # <<<<<<<<<<<<<<
 *         self._filled_buys_balance = 0
 *         self._filled_sells_balance = 0
 */
  __pyx_v_self->_all_markets_ready = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":96
 *         self._limit_order_type = self._market_info.market.get_maker_order_type()
 *         self._all_markets_ready = False
 *         self._filled_buys_balance = 0             # <<<<<<<<<<<<<<
 *         self._filled_sells_balance = 0
 *         self._logging_options = logging_options
 */
  __pyx_v_self->_filled_buys_balance = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":97
 *         self._all_markets_ready = False
 *         self._filled_buys_balance = 0
 *         self._filled_sells_balance = 0             # <<<<<<<<<<<<<<
 *         self._logging_options = logging_options
 *         self._last_timestamp = 0
 */
  __pyx_v_self->_filled_sells_balance = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":98
 *         self._filled_buys_balance = 0
 *         self._filled_sells_balance = 0
 *         self._logging_options = logging_options             # <<<<<<<<<<<<<<
 *         self._last_timestamp = 0
 *         self._status_report_interval = status_report_interval
 */
  __pyx_t_8 = __Pyx_PyInt_As_int64_t(__pyx_v_logging_options); if (unlikely((__pyx_t_8 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 98, __pyx_L1_error)
  __pyx_v_self->_logging_options = __pyx_t_8;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":99
 *         self._filled_sells_balance = 0
 *         self._logging_options = logging_options
 *         self._last_timestamp = 0             # <<<<<<<<<<<<<<
 *         self._status_report_interval = status_report_interval
 *         self._last_own_trade_price = Decimal('nan')
 */
  __pyx_v_self->_last_timestamp = 0.0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":100
 *         self._logging_options = logging_options
 *         self._last_timestamp = 0
 *         self._status_report_interval = status_report_interval             # <<<<<<<<<<<<<<
 *         self._last_own_trade_price = Decimal('nan')
 * 
 */
  __pyx_t_9 = __pyx_PyFloat_AsDouble(__pyx_v_status_report_interval); if (unlikely((__pyx_t_9 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 100, __pyx_L1_error)
  __pyx_v_self->_status_report_interval = __pyx_t_9;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":101
 *         self._last_timestamp = 0
 *         self._status_report_interval = status_report_interval
 *         self._last_own_trade_price = Decimal('nan')             # <<<<<<<<<<<<<<
 * 
 *         self.c_add_markets([market_info.market])
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_n_u_nan};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_last_own_trade_price);
  __Pyx_DECREF(__pyx_v_self->_last_own_trade_price);
  __pyx_v_self->_last_own_trade_price = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":103
 *         self._last_own_trade_price = Decimal('nan')
 * 
 *         self.c_add_markets([market_info.market])             # <<<<<<<<<<<<<<
 *         self._volatility_buffer_size = 0
 *         self._trading_intensity_buffer_size = 0
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_market_info, __pyx_n_s_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(0, 103, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_add_markets(((struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *)__pyx_v_self), ((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":104
 * 
 *         self.c_add_markets([market_info.market])
 *         self._volatility_buffer_size = 0             # <<<<<<<<<<<<<<
 *         self._trading_intensity_buffer_size = 0
 *         self._ticks_to_be_ready = -1
 */
  __pyx_v_self->_volatility_buffer_size = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":105
 *         self.c_add_markets([market_info.market])
 *         self._volatility_buffer_size = 0
 *         self._trading_intensity_buffer_size = 0             # <<<<<<<<<<<<<<
 *         self._ticks_to_be_ready = -1
 *         self._avg_vol = None
 */
  __pyx_v_self->_trading_intensity_buffer_size = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":106
 *         self._volatility_buffer_size = 0
 *         self._trading_intensity_buffer_size = 0
 *         self._ticks_to_be_ready = -1             # <<<<<<<<<<<<<<
 *         self._avg_vol = None
 *         self._trading_intensity = None
 */
  __pyx_v_self->_ticks_to_be_ready = -1;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":107
 *         self._trading_intensity_buffer_size = 0
 *         self._ticks_to_be_ready = -1
 *         self._avg_vol = None             # <<<<<<<<<<<<<<
 *         self._trading_intensity = None
 *         self._last_sampling_timestamp = 0
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_avg_vol);
  __Pyx_DECREF(__pyx_v_self->_avg_vol);
  __pyx_v_self->_avg_vol = Py_None;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":108
 *         self._ticks_to_be_ready = -1
 *         self._avg_vol = None
 *         self._trading_intensity = None             # <<<<<<<<<<<<<<
 *         self._last_sampling_timestamp = 0
 *         self._alpha = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_trading_intensity);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_trading_intensity);
  __pyx_v_self->_trading_intensity = ((struct __pyx_obj_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator *)Py_None);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":109
 *         self._avg_vol = None
 *         self._trading_intensity = None
 *         self._last_sampling_timestamp = 0             # <<<<<<<<<<<<<<
 *         self._alpha = None
 *         self._kappa = None
 */
  __pyx_v_self->_last_sampling_timestamp = 0.0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":110
 *         self._trading_intensity = None
 *         self._last_sampling_timestamp = 0
 *         self._alpha = None             # <<<<<<<<<<<<<<
 *         self._kappa = None
 *         self._execution_mode = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_alpha);
  __Pyx_DECREF(__pyx_v_self->_alpha);
  __pyx_v_self->_alpha = Py_None;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":111
 *         self._last_sampling_timestamp = 0
 *         self._alpha = None
 *         self._kappa = None             # <<<<<<<<<<<<<<
 *         self._execution_mode = None
 *         self._execution_timeframe = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_kappa);
  __Pyx_DECREF(__pyx_v_self->_kappa);
  __pyx_v_self->_kappa = Py_None;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":112
 *         self._alpha = None
 *         self._kappa = None
 *         self._execution_mode = None             # <<<<<<<<<<<<<<
 *         self._execution_timeframe = None
 *         self._execution_state = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_execution_mode);
  __Pyx_DECREF(__pyx_v_self->_execution_mode);
  __pyx_v_self->_execution_mode = ((PyObject*)Py_None);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":113
 *         self._kappa = None
 *         self._execution_mode = None
 *         self._execution_timeframe = None             # <<<<<<<<<<<<<<
 *         self._execution_state = None
 *         self._start_time = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_execution_timeframe);
  __Pyx_DECREF(__pyx_v_self->_execution_timeframe);
  __pyx_v_self->_execution_timeframe = ((PyObject*)Py_None);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":114
 *         self._execution_mode = None
 *         self._execution_timeframe = None
 *         self._execution_state = None             # <<<<<<<<<<<<<<
 *         self._start_time = None
 *         self._end_time = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_execution_state);
  __Pyx_DECREF(__pyx_v_self->_execution_state);
  __pyx_v_self->_execution_state = Py_None;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":115
 *         self._execution_timeframe = None
 *         self._execution_state = None
 *         self._start_time = None             # <<<<<<<<<<<<<<
 *         self._end_time = None
 *         self._reservation_price = s_decimal_zero
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_start_time);
  __Pyx_DECREF(__pyx_v_self->_start_time);
  __pyx_v_self->_start_time = Py_None;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":116
 *         self._execution_state = None
 *         self._start_time = None
 *         self._end_time = None             # <<<<<<<<<<<<<<
 *         self._reservation_price = s_decimal_zero
 *         self._optimal_spread = s_decimal_zero
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_end_time);
  __Pyx_DECREF(__pyx_v_self->_end_time);
  __pyx_v_self->_end_time = Py_None;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":117
 *         self._start_time = None
 *         self._end_time = None
 *         self._reservation_price = s_decimal_zero             # <<<<<<<<<<<<<<
 *         self._optimal_spread = s_decimal_zero
 *         self._optimal_ask = s_decimal_zero
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_s_decimal_zero); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_reservation_price);
  __Pyx_DECREF(__pyx_v_self->_reservation_price);
  __pyx_v_self->_reservation_price = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":118
 *         self._end_time = None
 *         self._reservation_price = s_decimal_zero
 *         self._optimal_spread = s_decimal_zero             # <<<<<<<<<<<<<<
 *         self._optimal_ask = s_decimal_zero
 *         self._optimal_bid = s_decimal_zero
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_s_decimal_zero); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_optimal_spread);
  __Pyx_DECREF(__pyx_v_self->_optimal_spread);
  __pyx_v_self->_optimal_spread = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":119
 *         self._reservation_price = s_decimal_zero
 *         self._optimal_spread = s_decimal_zero
 *         self._optimal_ask = s_decimal_zero             # <<<<<<<<<<<<<<
 *         self._optimal_bid = s_decimal_zero
 *         self._debug_csv_path = debug_csv_path
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_s_decimal_zero); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_optimal_ask);
  __Pyx_DECREF(__pyx_v_self->_optimal_ask);
  __pyx_v_self->_optimal_ask = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":120
 *         self._optimal_spread = s_decimal_zero
 *         self._optimal_ask = s_decimal_zero
 *         self._optimal_bid = s_decimal_zero             # <<<<<<<<<<<<<<
 *         self._debug_csv_path = debug_csv_path
 *         self._is_debug = is_debug
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_s_decimal_zero); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_optimal_bid);
  __Pyx_DECREF(__pyx_v_self->_optimal_bid);
  __pyx_v_self->_optimal_bid = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":121
 *         self._optimal_ask = s_decimal_zero
 *         self._optimal_bid = s_decimal_zero
 *         self._debug_csv_path = debug_csv_path             # <<<<<<<<<<<<<<
 *         self._is_debug = is_debug
 *         try:
 */
  if (!(likely(PyUnicode_CheckExact(__pyx_v_debug_csv_path))||((__pyx_v_debug_csv_path) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_debug_csv_path))) __PYX_ERR(0, 121, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_debug_csv_path;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_debug_csv_path);
  __Pyx_DECREF(__pyx_v_self->_debug_csv_path);
  __pyx_v_self->_debug_csv_path = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":122
 *         self._optimal_bid = s_decimal_zero
 *         self._debug_csv_path = debug_csv_path
 *         self._is_debug = is_debug             # <<<<<<<<<<<<<<
 *         try:
 *             if self._is_debug:
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_is_debug); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 122, __pyx_L1_error)
  __pyx_v_self->_is_debug = __pyx_t_7;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":123
 *         self._debug_csv_path = debug_csv_path
 *         self._is_debug = is_debug
 *         try:             # <<<<<<<<<<<<<<
 *             if self._is_debug:
 *                 os.unlink(self._debug_csv_path)
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
    __Pyx_XGOTREF(__pyx_t_10);
    __Pyx_XGOTREF(__pyx_t_11);
    __Pyx_XGOTREF(__pyx_t_12);
    /*try:*/ {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":124
 *         self._is_debug = is_debug
 *         try:
 *             if self._is_debug:             # <<<<<<<<<<<<<<
 *                 os.unlink(self._debug_csv_path)
 *         except FileNotFoundError:
 */
      if (__pyx_v_self->_is_debug) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":125
 *         try:
 *             if self._is_debug:
 *                 os.unlink(self._debug_csv_path)             # <<<<<<<<<<<<<<
 *         except FileNotFoundError:
 *             pass
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_os); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 125, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_unlink); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 125, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        __pyx_t_6 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_2))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_2, function);
            __pyx_t_6 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_self->_debug_csv_path};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 125, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":124
 *         self._is_debug = is_debug
 *         try:
 *             if self._is_debug:             # <<<<<<<<<<<<<<
 *                 os.unlink(self._debug_csv_path)
 *         except FileNotFoundError:
 */
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":123
 *         self._debug_csv_path = debug_csv_path
 *         self._is_debug = is_debug
 *         try:             # <<<<<<<<<<<<<<
 *             if self._is_debug:
 *                 os.unlink(self._debug_csv_path)
 */
    }
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":126
 *             if self._is_debug:
 *                 os.unlink(self._debug_csv_path)
 *         except FileNotFoundError:             # <<<<<<<<<<<<<<
 *             pass
 * 
 */
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_FileNotFoundError); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 126, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_13 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_t_1 = 0; __pyx_t_2 = 0; __pyx_t_3 = 0;
    if (__pyx_t_13) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":123
 *         self._debug_csv_path = debug_csv_path
 *         self._is_debug = is_debug
 *         try:             # <<<<<<<<<<<<<<
 *             if self._is_debug:
 *                 os.unlink(self._debug_csv_path)
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_XGIVEREF(__pyx_t_12);
    __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_10);
    __Pyx_XGIVEREF(__pyx_t_11);
    __Pyx_XGIVEREF(__pyx_t_12);
    __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
    __pyx_L8_try_end:;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":129
 *             pass
 * 
 *         self.get_config_map_execution_mode()             # <<<<<<<<<<<<<<
 *         self.get_config_map_hanging_orders()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_config_map_execution_mode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":130
 * 
 *         self.get_config_map_execution_mode()
 *         self.get_config_map_hanging_orders()             # <<<<<<<<<<<<<<
 * 
 *     def all_markets_ready(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_config_map_hanging_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 130, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":73
 *         return pmm_logger
 * 
 *     def init_params(self,             # <<<<<<<<<<<<<<
 *                     config_map: Union[AvellanedaMarketMakingConfigMap, ClientConfigAdapter],
 *                     market_info: MarketTradingPairTuple,
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.init_params", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":132
 *         self.get_config_map_hanging_orders()
 * 
 *     def all_markets_ready(self):             # <<<<<<<<<<<<<<
 *         return all([market.ready for market in self._sb_markets])
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5all_markets_ready(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5all_markets_ready = {"all_markets_ready", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5all_markets_ready, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5all_markets_ready(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("all_markets_ready (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("all_markets_ready", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "all_markets_ready", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_4all_markets_ready(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_4all_markets_ready(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_7genexpr__pyx_v_market = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("all_markets_ready", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":133
 * 
 *     def all_markets_ready(self):
 *         return all([market.ready for market in self._sb_markets])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_set_iterator(__pyx_v_self->__pyx_base._sb_markets, 1, (&__pyx_t_4), (&__pyx_t_5)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 133, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2);
    __pyx_t_2 = __pyx_t_6;
    __pyx_t_6 = 0;
    while (1) {
      __pyx_t_7 = __Pyx_set_iter_next(__pyx_t_2, __pyx_t_4, &__pyx_t_3, &__pyx_t_6, __pyx_t_5);
      if (unlikely(__pyx_t_7 == 0)) break;
      if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 133, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_7genexpr__pyx_v_market, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_7genexpr__pyx_v_market, __pyx_n_s_ready); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 133, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_6))) __PYX_ERR(0, 133, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_market); __pyx_7genexpr__pyx_v_market = 0;
    goto __pyx_L8_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_7genexpr__pyx_v_market); __pyx_7genexpr__pyx_v_market = 0;
    goto __pyx_L1_error;
    __pyx_L8_exit_scope:;
  } /* exit inner scope */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_all, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":132
 *         self.get_config_map_hanging_orders()
 * 
 *     def all_markets_ready(self):             # <<<<<<<<<<<<<<
 *         return all([market.ready for market in self._sb_markets])
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.all_markets_ready", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_7genexpr__pyx_v_market);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":135
 *         return all([market.ready for market in self._sb_markets])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def min_spread(self):
 *         return self._config_map.min_spread
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10min_spread_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10min_spread_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10min_spread___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10min_spread___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":137
 *     @property
 *     def min_spread(self):
 *         return self._config_map.min_spread             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_min_spread); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":135
 *         return all([market.ready for market in self._sb_markets])
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def min_spread(self):
 *         return self._config_map.min_spread
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.min_spread.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":139
 *         return self._config_map.min_spread
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def avg_vol(self):
 *         return self._avg_vol
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7avg_vol_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7avg_vol_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7avg_vol___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7avg_vol___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":141
 *     @property
 *     def avg_vol(self):
 *         return self._avg_vol             # <<<<<<<<<<<<<<
 * 
 *     @avg_vol.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_avg_vol);
  __pyx_r = __pyx_v_self->_avg_vol;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":139
 *         return self._config_map.min_spread
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def avg_vol(self):
 *         return self._avg_vol
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":143
 *         return self._avg_vol
 * 
 *     @avg_vol.setter             # <<<<<<<<<<<<<<
 *     def avg_vol(self, indicator: InstantVolatilityIndicator):
 *         self._avg_vol = indicator
 */

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7avg_vol_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_indicator); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7avg_vol_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_indicator) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7avg_vol_2__set__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_indicator));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7avg_vol_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_indicator) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":145
 *     @avg_vol.setter
 *     def avg_vol(self, indicator: InstantVolatilityIndicator):
 *         self._avg_vol = indicator             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_INCREF(__pyx_v_indicator);
  __Pyx_GIVEREF(__pyx_v_indicator);
  __Pyx_GOTREF(__pyx_v_self->_avg_vol);
  __Pyx_DECREF(__pyx_v_self->_avg_vol);
  __pyx_v_self->_avg_vol = __pyx_v_indicator;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":143
 *         return self._avg_vol
 * 
 *     @avg_vol.setter             # <<<<<<<<<<<<<<
 *     def avg_vol(self, indicator: InstantVolatilityIndicator):
 *         self._avg_vol = indicator
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":147
 *         self._avg_vol = indicator
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def trading_intensity(self):
 *         return self._trading_intensity
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17trading_intensity_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17trading_intensity_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17trading_intensity___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17trading_intensity___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":149
 *     @property
 *     def trading_intensity(self):
 *         return self._trading_intensity             # <<<<<<<<<<<<<<
 * 
 *     @trading_intensity.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_self->_trading_intensity);
  __pyx_r = ((PyObject *)__pyx_v_self->_trading_intensity);
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":147
 *         self._avg_vol = indicator
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def trading_intensity(self):
 *         return self._trading_intensity
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":151
 *         return self._trading_intensity
 * 
 *     @trading_intensity.setter             # <<<<<<<<<<<<<<
 *     def trading_intensity(self, indicator: TradingIntensityIndicator):
 *         self._trading_intensity = indicator
 */

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17trading_intensity_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_indicator); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17trading_intensity_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_indicator) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17trading_intensity_2__set__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_indicator));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17trading_intensity_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_indicator) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":153
 *     @trading_intensity.setter
 *     def trading_intensity(self, indicator: TradingIntensityIndicator):
 *         self._trading_intensity = indicator             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  if (!(likely(((__pyx_v_indicator) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_indicator, __pyx_ptype_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator))))) __PYX_ERR(0, 153, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_indicator;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_trading_intensity);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_trading_intensity);
  __pyx_v_self->_trading_intensity = ((struct __pyx_obj_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":151
 *         return self._trading_intensity
 * 
 *     @trading_intensity.setter             # <<<<<<<<<<<<<<
 *     def trading_intensity(self, indicator: TradingIntensityIndicator):
 *         self._trading_intensity = indicator
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.trading_intensity.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":155
 *         self._trading_intensity = indicator
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def market_info(self) -> MarketTradingPairTuple:
 *         return self._market_info
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11market_info_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11market_info_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11market_info___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11market_info___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":157
 *     @property
 *     def market_info(self) -> MarketTradingPairTuple:
 *         return self._market_info             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_market_info);
  __pyx_r = __pyx_v_self->_market_info;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":155
 *         self._trading_intensity = indicator
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def market_info(self) -> MarketTradingPairTuple:
 *         return self._market_info
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":159
 *         return self._market_info
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_refresh_tolerance_pct(self) -> Decimal:
 *         return self._config_map.order_refresh_tolerance_pct
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_27order_refresh_tolerance_pct_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_27order_refresh_tolerance_pct_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_27order_refresh_tolerance_pct___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_27order_refresh_tolerance_pct___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":161
 *     @property
 *     def order_refresh_tolerance_pct(self) -> Decimal:
 *         return self._config_map.order_refresh_tolerance_pct             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_order_refresh_tolerance_pct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":159
 *         return self._market_info
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_refresh_tolerance_pct(self) -> Decimal:
 *         return self._config_map.order_refresh_tolerance_pct
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.order_refresh_tolerance_pct.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":163
 *         return self._config_map.order_refresh_tolerance_pct
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_refresh_tolerance(self) -> Decimal:
 *         return self._config_map.order_refresh_tolerance_pct / Decimal('100')
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_23order_refresh_tolerance_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_23order_refresh_tolerance_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_23order_refresh_tolerance___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_23order_refresh_tolerance___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":165
 *     @property
 *     def order_refresh_tolerance(self) -> Decimal:
 *         return self._config_map.order_refresh_tolerance_pct / Decimal('100')             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_order_refresh_tolerance_pct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_u_100};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 165, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":163
 *         return self._config_map.order_refresh_tolerance_pct
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_refresh_tolerance(self) -> Decimal:
 *         return self._config_map.order_refresh_tolerance_pct / Decimal('100')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.order_refresh_tolerance.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":167
 *         return self._config_map.order_refresh_tolerance_pct / Decimal('100')
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_amount(self) -> Decimal:
 *         return self._config_map.order_amount
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12order_amount_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12order_amount_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12order_amount___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12order_amount___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":169
 *     @property
 *     def order_amount(self) -> Decimal:
 *         return self._config_map.order_amount             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_order_amount); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":167
 *         return self._config_map.order_refresh_tolerance_pct / Decimal('100')
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_amount(self) -> Decimal:
 *         return self._config_map.order_amount
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.order_amount.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":171
 *         return self._config_map.order_amount
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def inventory_target_base_pct(self) -> Decimal:
 *         return self._config_map.inventory_target_base_pct
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25inventory_target_base_pct_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25inventory_target_base_pct_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25inventory_target_base_pct___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25inventory_target_base_pct___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":173
 *     @property
 *     def inventory_target_base_pct(self) -> Decimal:
 *         return self._config_map.inventory_target_base_pct             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_inventory_target_base_pct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":171
 *         return self._config_map.order_amount
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def inventory_target_base_pct(self) -> Decimal:
 *         return self._config_map.inventory_target_base_pct
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.inventory_target_base_pct.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":175
 *         return self._config_map.inventory_target_base_pct
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def inventory_target_base(self) -> Decimal:
 *         return self.inventory_target_base_pct / Decimal('100')
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_21inventory_target_base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_21inventory_target_base_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_21inventory_target_base___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_21inventory_target_base___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":177
 *     @property
 *     def inventory_target_base(self) -> Decimal:
 *         return self.inventory_target_base_pct / Decimal('100')             # <<<<<<<<<<<<<<
 * 
 *     @inventory_target_base.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_inventory_target_base_pct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_u_100};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":175
 *         return self._config_map.inventory_target_base_pct
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def inventory_target_base(self) -> Decimal:
 *         return self.inventory_target_base_pct / Decimal('100')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.inventory_target_base.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":179
 *         return self.inventory_target_base_pct / Decimal('100')
 * 
 *     @inventory_target_base.setter             # <<<<<<<<<<<<<<
 *     def inventory_target_base(self, value: Decimal):
 *         self.inventory_target_base_pct = value * Decimal('100')
 */

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_21inventory_target_base_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_21inventory_target_base_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_21inventory_target_base_2__set__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_21inventory_target_base_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":181
 *     @inventory_target_base.setter
 *     def inventory_target_base(self, value: Decimal):
 *         self.inventory_target_base_pct = value * Decimal('100')             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_u_100};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = PyNumber_Multiply(__pyx_v_value, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_inventory_target_base_pct, __pyx_t_2) < 0) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":179
 *         return self.inventory_target_base_pct / Decimal('100')
 * 
 *     @inventory_target_base.setter             # <<<<<<<<<<<<<<
 *     def inventory_target_base(self, value: Decimal):
 *         self.inventory_target_base_pct = value * Decimal('100')
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.inventory_target_base.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":183
 *         self.inventory_target_base_pct = value * Decimal('100')
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_optimization_enabled(self) -> bool:
 *         return self._config_map.order_optimization_enabled
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26order_optimization_enabled_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26order_optimization_enabled_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26order_optimization_enabled___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26order_optimization_enabled___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":185
 *     @property
 *     def order_optimization_enabled(self) -> bool:
 *         return self._config_map.order_optimization_enabled             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_order_optimization_enabled); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":183
 *         self.inventory_target_base_pct = value * Decimal('100')
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_optimization_enabled(self) -> bool:
 *         return self._config_map.order_optimization_enabled
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.order_optimization_enabled.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":187
 *         return self._config_map.order_optimization_enabled
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_refresh_time(self) -> float:
 *         return self._config_map.order_refresh_time
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_18order_refresh_time_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_18order_refresh_time_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_18order_refresh_time___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_18order_refresh_time___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":189
 *     @property
 *     def order_refresh_time(self) -> float:
 *         return self._config_map.order_refresh_time             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_order_refresh_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":187
 *         return self._config_map.order_optimization_enabled
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_refresh_time(self) -> float:
 *         return self._config_map.order_refresh_time
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.order_refresh_time.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":191
 *         return self._config_map.order_refresh_time
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def filled_order_delay(self) -> float:
 *         return self._config_map.filled_order_delay
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_18filled_order_delay_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_18filled_order_delay_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_18filled_order_delay___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_18filled_order_delay___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":193
 *     @property
 *     def filled_order_delay(self) -> float:
 *         return self._config_map.filled_order_delay             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_filled_order_delay); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 193, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":191
 *         return self._config_map.order_refresh_time
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def filled_order_delay(self) -> float:
 *         return self._config_map.filled_order_delay
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.filled_order_delay.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":195
 *         return self._config_map.filled_order_delay
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_override(self) -> Dict[str, any]:
 *         return self._config_map.order_override
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_14order_override_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_14order_override_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_14order_override___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_14order_override___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":197
 *     @property
 *     def order_override(self) -> Dict[str, any]:
 *         return self._config_map.order_override             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_order_override); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":195
 *         return self._config_map.filled_order_delay
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_override(self) -> Dict[str, any]:
 *         return self._config_map.order_override
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.order_override.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":199
 *         return self._config_map.order_override
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_levels(self) -> int:
 *         if self._config_map.order_levels_mode.title == MultiOrderLevelModel.model_config["title"]:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12order_levels_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12order_levels_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12order_levels___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12order_levels___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":201
 *     @property
 *     def order_levels(self) -> int:
 *         if self._config_map.order_levels_mode.title == MultiOrderLevelModel.model_config["title"]:             # <<<<<<<<<<<<<<
 *             return self._config_map.order_levels_mode.order_levels
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_order_levels_mode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_title); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_MultiOrderLevelModel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_model_config); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_title); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":202
 *     def order_levels(self) -> int:
 *         if self._config_map.order_levels_mode.title == MultiOrderLevelModel.model_config["title"]:
 *             return self._config_map.order_levels_mode.order_levels             # <<<<<<<<<<<<<<
 *         else:
 *             return 0
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_order_levels_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_order_levels); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":201
 *     @property
 *     def order_levels(self) -> int:
 *         if self._config_map.order_levels_mode.title == MultiOrderLevelModel.model_config["title"]:             # <<<<<<<<<<<<<<
 *             return self._config_map.order_levels_mode.order_levels
 *         else:
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":204
 *             return self._config_map.order_levels_mode.order_levels
 *         else:
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_int_0);
    __pyx_r = __pyx_int_0;
    goto __pyx_L0;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":199
 *         return self._config_map.order_override
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def order_levels(self) -> int:
 *         if self._config_map.order_levels_mode.title == MultiOrderLevelModel.model_config["title"]:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.order_levels.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":206
 *             return 0
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def level_distances(self) -> int:
 *         if self._config_map.order_levels_mode.title == MultiOrderLevelModel.model_config["title"]:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15level_distances_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15level_distances_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15level_distances___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15level_distances___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":208
 *     @property
 *     def level_distances(self) -> int:
 *         if self._config_map.order_levels_mode.title == MultiOrderLevelModel.model_config["title"]:             # <<<<<<<<<<<<<<
 *             return self._config_map.order_levels_mode.level_distances
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_order_levels_mode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_title); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_MultiOrderLevelModel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_model_config); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_title); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":209
 *     def level_distances(self) -> int:
 *         if self._config_map.order_levels_mode.title == MultiOrderLevelModel.model_config["title"]:
 *             return self._config_map.order_levels_mode.level_distances             # <<<<<<<<<<<<<<
 *         else:
 *             return 0
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_order_levels_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_level_distances); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":208
 *     @property
 *     def level_distances(self) -> int:
 *         if self._config_map.order_levels_mode.title == MultiOrderLevelModel.model_config["title"]:             # <<<<<<<<<<<<<<
 *             return self._config_map.order_levels_mode.level_distances
 *         else:
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":211
 *             return self._config_map.order_levels_mode.level_distances
 *         else:
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_int_0);
    __pyx_r = __pyx_int_0;
    goto __pyx_L0;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":206
 *             return 0
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def level_distances(self) -> int:
 *         if self._config_map.order_levels_mode.title == MultiOrderLevelModel.model_config["title"]:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.level_distances.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":213
 *             return 0
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_order_age(self):
 *         return self._config_map.max_order_age
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_13max_order_age_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_13max_order_age_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_13max_order_age___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_13max_order_age___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":215
 *     @property
 *     def max_order_age(self):
 *         return self._config_map.max_order_age             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_max_order_age); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":213
 *             return 0
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def max_order_age(self):
 *         return self._config_map.max_order_age
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.max_order_age.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":217
 *         return self._config_map.max_order_age
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def add_transaction_costs_to_orders(self) -> bool:
 *         return self._config_map.add_transaction_costs
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_31add_transaction_costs_to_orders_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_31add_transaction_costs_to_orders_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_31add_transaction_costs_to_orders___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_31add_transaction_costs_to_orders___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":219
 *     @property
 *     def add_transaction_costs_to_orders(self) -> bool:
 *         return self._config_map.add_transaction_costs             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_add_transaction_costs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":217
 *         return self._config_map.max_order_age
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def add_transaction_costs_to_orders(self) -> bool:
 *         return self._config_map.add_transaction_costs
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.add_transaction_costs_to_orders.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":221
 *         return self._config_map.add_transaction_costs
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def base_asset(self):
 *         return self._market_info.base_asset
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10base_asset_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10base_asset_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10base_asset___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10base_asset___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":223
 *     @property
 *     def base_asset(self):
 *         return self._market_info.base_asset             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":221
 *         return self._config_map.add_transaction_costs
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def base_asset(self):
 *         return self._market_info.base_asset
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.base_asset.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":225
 *         return self._market_info.base_asset
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def quote_asset(self):
 *         return self._market_info.quote_asset
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11quote_asset_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11quote_asset_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11quote_asset___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11quote_asset___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":227
 *     @property
 *     def quote_asset(self):
 *         return self._market_info.quote_asset             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_quote_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":225
 *         return self._market_info.base_asset
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def quote_asset(self):
 *         return self._market_info.quote_asset
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.quote_asset.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":229
 *         return self._market_info.quote_asset
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def trading_pair(self):
 *         return self._market_info.trading_pair
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12trading_pair_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12trading_pair_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12trading_pair___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12trading_pair___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":231
 *     @property
 *     def trading_pair(self):
 *         return self._market_info.trading_pair             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":229
 *         return self._market_info.quote_asset
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def trading_pair(self):
 *         return self._market_info.trading_pair
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.trading_pair.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":233
 *         return self._market_info.trading_pair
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def gamma(self):
 *         return self._config_map.risk_factor
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5gamma_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5gamma_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5gamma___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5gamma___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":235
 *     @property
 *     def gamma(self):
 *         return self._config_map.risk_factor             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_risk_factor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 235, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":233
 *         return self._market_info.trading_pair
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def gamma(self):
 *         return self._config_map.risk_factor
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.gamma.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":237
 *         return self._config_map.risk_factor
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def alpha(self):
 *         return self._alpha
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5alpha_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5alpha_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5alpha___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5alpha___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":239
 *     @property
 *     def alpha(self):
 *         return self._alpha             # <<<<<<<<<<<<<<
 * 
 *     @alpha.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_alpha);
  __pyx_r = __pyx_v_self->_alpha;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":237
 *         return self._config_map.risk_factor
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def alpha(self):
 *         return self._alpha
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":241
 *         return self._alpha
 * 
 *     @alpha.setter             # <<<<<<<<<<<<<<
 *     def alpha(self, value):
 *         self._alpha = value
 */

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5alpha_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5alpha_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5alpha_2__set__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5alpha_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":243
 *     @alpha.setter
 *     def alpha(self, value):
 *         self._alpha = value             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_alpha);
  __Pyx_DECREF(__pyx_v_self->_alpha);
  __pyx_v_self->_alpha = __pyx_v_value;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":241
 *         return self._alpha
 * 
 *     @alpha.setter             # <<<<<<<<<<<<<<
 *     def alpha(self, value):
 *         self._alpha = value
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":245
 *         self._alpha = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def kappa(self):
 *         return self._kappa
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5kappa_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5kappa_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5kappa___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5kappa___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":247
 *     @property
 *     def kappa(self):
 *         return self._kappa             # <<<<<<<<<<<<<<
 * 
 *     @kappa.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_kappa);
  __pyx_r = __pyx_v_self->_kappa;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":245
 *         self._alpha = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def kappa(self):
 *         return self._kappa
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":249
 *         return self._kappa
 * 
 *     @kappa.setter             # <<<<<<<<<<<<<<
 *     def kappa(self, value):
 *         self._kappa = value
 */

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5kappa_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5kappa_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5kappa_2__set__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5kappa_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":251
 *     @kappa.setter
 *     def kappa(self, value):
 *         self._kappa = value             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_kappa);
  __Pyx_DECREF(__pyx_v_self->_kappa);
  __pyx_v_self->_kappa = __pyx_v_value;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":249
 *         return self._kappa
 * 
 *     @kappa.setter             # <<<<<<<<<<<<<<
 *     def kappa(self, value):
 *         self._kappa = value
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":253
 *         self._kappa = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def eta(self):
 *         return self._config_map.order_amount_shape_factor
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_3eta_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_3eta_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_3eta___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_3eta___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":255
 *     @property
 *     def eta(self):
 *         return self._config_map.order_amount_shape_factor             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_order_amount_shape_factor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":253
 *         self._kappa = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def eta(self):
 *         return self._config_map.order_amount_shape_factor
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.eta.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":257
 *         return self._config_map.order_amount_shape_factor
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def reservation_price(self):
 *         return self._reservation_price
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17reservation_price_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17reservation_price_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17reservation_price___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17reservation_price___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":259
 *     @property
 *     def reservation_price(self):
 *         return self._reservation_price             # <<<<<<<<<<<<<<
 * 
 *     @reservation_price.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_reservation_price);
  __pyx_r = __pyx_v_self->_reservation_price;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":257
 *         return self._config_map.order_amount_shape_factor
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def reservation_price(self):
 *         return self._reservation_price
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":261
 *         return self._reservation_price
 * 
 *     @reservation_price.setter             # <<<<<<<<<<<<<<
 *     def reservation_price(self, value):
 *         self._reservation_price = value
 */

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17reservation_price_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17reservation_price_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17reservation_price_2__set__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17reservation_price_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":263
 *     @reservation_price.setter
 *     def reservation_price(self, value):
 *         self._reservation_price = value             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_reservation_price);
  __Pyx_DECREF(__pyx_v_self->_reservation_price);
  __pyx_v_self->_reservation_price = __pyx_v_value;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":261
 *         return self._reservation_price
 * 
 *     @reservation_price.setter             # <<<<<<<<<<<<<<
 *     def reservation_price(self, value):
 *         self._reservation_price = value
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":265
 *         self._reservation_price = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def optimal_spread(self):
 *         return self._optimal_spread
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_14optimal_spread_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_14optimal_spread_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_14optimal_spread___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_14optimal_spread___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":267
 *     @property
 *     def optimal_spread(self):
 *         return self._optimal_spread             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_optimal_spread);
  __pyx_r = __pyx_v_self->_optimal_spread;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":265
 *         self._reservation_price = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def optimal_spread(self):
 *         return self._optimal_spread
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":269
 *         return self._optimal_spread
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def optimal_ask(self):
 *         return self._optimal_ask
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_ask_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_ask_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_ask___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_ask___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":271
 *     @property
 *     def optimal_ask(self):
 *         return self._optimal_ask             # <<<<<<<<<<<<<<
 * 
 *     @optimal_ask.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_optimal_ask);
  __pyx_r = __pyx_v_self->_optimal_ask;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":269
 *         return self._optimal_spread
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def optimal_ask(self):
 *         return self._optimal_ask
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":273
 *         return self._optimal_ask
 * 
 *     @optimal_ask.setter             # <<<<<<<<<<<<<<
 *     def optimal_ask(self, value):
 *         self._optimal_ask = value
 */

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_ask_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_ask_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_ask_2__set__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_ask_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":275
 *     @optimal_ask.setter
 *     def optimal_ask(self, value):
 *         self._optimal_ask = value             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_optimal_ask);
  __Pyx_DECREF(__pyx_v_self->_optimal_ask);
  __pyx_v_self->_optimal_ask = __pyx_v_value;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":273
 *         return self._optimal_ask
 * 
 *     @optimal_ask.setter             # <<<<<<<<<<<<<<
 *     def optimal_ask(self, value):
 *         self._optimal_ask = value
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":277
 *         self._optimal_ask = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def optimal_bid(self):
 *         return self._optimal_bid
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_bid_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_bid_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_bid___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_bid___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":279
 *     @property
 *     def optimal_bid(self):
 *         return self._optimal_bid             # <<<<<<<<<<<<<<
 * 
 *     @optimal_bid.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_optimal_bid);
  __pyx_r = __pyx_v_self->_optimal_bid;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":277
 *         self._optimal_ask = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def optimal_bid(self):
 *         return self._optimal_bid
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":281
 *         return self._optimal_bid
 * 
 *     @optimal_bid.setter             # <<<<<<<<<<<<<<
 *     def optimal_bid(self, value):
 *         self._optimal_bid = value
 */

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_bid_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_bid_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_bid_2__set__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_bid_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":283
 *     @optimal_bid.setter
 *     def optimal_bid(self, value):
 *         self._optimal_bid = value             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_optimal_bid);
  __Pyx_DECREF(__pyx_v_self->_optimal_bid);
  __pyx_v_self->_optimal_bid = __pyx_v_value;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":281
 *         return self._optimal_bid
 * 
 *     @optimal_bid.setter             # <<<<<<<<<<<<<<
 *     def optimal_bid(self, value):
 *         self._optimal_bid = value
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":285
 *         self._optimal_bid = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def execution_timeframe(self):
 *         return self._execution_timeframe
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_19execution_timeframe_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_19execution_timeframe_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_19execution_timeframe___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_19execution_timeframe___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":287
 *     @property
 *     def execution_timeframe(self):
 *         return self._execution_timeframe             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_execution_timeframe);
  __pyx_r = __pyx_v_self->_execution_timeframe;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":285
 *         self._optimal_bid = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def execution_timeframe(self):
 *         return self._execution_timeframe
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":289
 *         return self._execution_timeframe
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def start_time(self) -> time:
 *         return self._start_time
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10start_time_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10start_time_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10start_time___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10start_time___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":291
 *     @property
 *     def start_time(self) -> time:
 *         return self._start_time             # <<<<<<<<<<<<<<
 * 
 *     @start_time.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_start_time);
  __pyx_r = __pyx_v_self->_start_time;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":289
 *         return self._execution_timeframe
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def start_time(self) -> time:
 *         return self._start_time
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":293
 *         return self._start_time
 * 
 *     @start_time.setter             # <<<<<<<<<<<<<<
 *     def start_time(self, value):
 *         self._start_time = value
 */

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10start_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10start_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10start_time_2__set__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10start_time_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":295
 *     @start_time.setter
 *     def start_time(self, value):
 *         self._start_time = value             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_start_time);
  __Pyx_DECREF(__pyx_v_self->_start_time);
  __pyx_v_self->_start_time = __pyx_v_value;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":293
 *         return self._start_time
 * 
 *     @start_time.setter             # <<<<<<<<<<<<<<
 *     def start_time(self, value):
 *         self._start_time = value
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":297
 *         self._start_time = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def end_time(self) -> time:
 *         return self._end_time
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_8end_time_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_8end_time_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_8end_time___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_8end_time___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":299
 *     @property
 *     def end_time(self) -> time:
 *         return self._end_time             # <<<<<<<<<<<<<<
 * 
 *     @end_time.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_end_time);
  __pyx_r = __pyx_v_self->_end_time;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":297
 *         self._start_time = value
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def end_time(self) -> time:
 *         return self._end_time
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":301
 *         return self._end_time
 * 
 *     @end_time.setter             # <<<<<<<<<<<<<<
 *     def end_time(self, value):
 *         self._end_time = value
 */

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_8end_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_8end_time_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_v_value) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_8end_time_2__set__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_8end_time_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":303
 *     @end_time.setter
 *     def end_time(self, value):
 *         self._end_time = value             # <<<<<<<<<<<<<<
 * 
 *     def get_price(self) -> float:
 */
  __Pyx_INCREF(__pyx_v_value);
  __Pyx_GIVEREF(__pyx_v_value);
  __Pyx_GOTREF(__pyx_v_self->_end_time);
  __Pyx_DECREF(__pyx_v_self->_end_time);
  __pyx_v_self->_end_time = __pyx_v_value;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":301
 *         return self._end_time
 * 
 *     @end_time.setter             # <<<<<<<<<<<<<<
 *     def end_time(self, value):
 *         self._end_time = value
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":305
 *         self._end_time = value
 * 
 *     def get_price(self) -> float:             # <<<<<<<<<<<<<<
 *         return self.get_mid_price()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7get_price(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7get_price = {"get_price", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7get_price, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7get_price(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_price (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_price", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_price", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_6get_price(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_6get_price(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_price", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":306
 * 
 *     def get_price(self) -> float:
 *         return self.get_mid_price()             # <<<<<<<<<<<<<<
 * 
 *     def get_last_price(self) -> float:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_mid_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 306, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 306, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":305
 *         self._end_time = value
 * 
 *     def get_price(self) -> float:             # <<<<<<<<<<<<<<
 *         return self.get_mid_price()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.get_price", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":308
 *         return self.get_mid_price()
 * 
 *     def get_last_price(self) -> float:             # <<<<<<<<<<<<<<
 *         return self._market_info.get_last_price()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_9get_last_price(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_9get_last_price = {"get_last_price", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_9get_last_price, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_9get_last_price(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_last_price (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_last_price", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_last_price", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_8get_last_price(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_8get_last_price(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_last_price", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":309
 * 
 *     def get_last_price(self) -> float:
 *         return self._market_info.get_last_price()             # <<<<<<<<<<<<<<
 * 
 *     def get_mid_price(self) -> float:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_get_last_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 309, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":308
 *         return self.get_mid_price()
 * 
 *     def get_last_price(self) -> float:             # <<<<<<<<<<<<<<
 *         return self._market_info.get_last_price()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.get_last_price", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":311
 *         return self._market_info.get_last_price()
 * 
 *     def get_mid_price(self) -> float:             # <<<<<<<<<<<<<<
 *         return self.c_get_mid_price()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11get_mid_price(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11get_mid_price = {"get_mid_price", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11get_mid_price, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11get_mid_price(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_mid_price (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_mid_price", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_mid_price", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10get_mid_price(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10get_mid_price(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_mid_price", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":312
 * 
 *     def get_mid_price(self) -> float:
 *         return self.c_get_mid_price()             # <<<<<<<<<<<<<<
 * 
 *     cdef object c_get_mid_price(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_get_mid_price(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":311
 *         return self._market_info.get_last_price()
 * 
 *     def get_mid_price(self) -> float:             # <<<<<<<<<<<<<<
 *         return self.c_get_mid_price()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.get_mid_price", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":314
 *         return self.c_get_mid_price()
 * 
 *     cdef object c_get_mid_price(self):             # <<<<<<<<<<<<<<
 *         return self._price_delegate.get_price_by_type(PriceType.MidPrice)
 * 
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_get_mid_price(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_get_mid_price", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":315
 * 
 *     cdef object c_get_mid_price(self):
 *         return self._price_delegate.get_price_by_type(PriceType.MidPrice)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_price_delegate, __pyx_n_s_get_price_by_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_PriceType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_MidPrice); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 315, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 315, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":314
 *         return self.c_get_mid_price()
 * 
 *     cdef object c_get_mid_price(self):             # <<<<<<<<<<<<<<
 *         return self._price_delegate.get_price_by_type(PriceType.MidPrice)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_get_mid_price", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":317
 *         return self._price_delegate.get_price_by_type(PriceType.MidPrice)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def market_info_to_active_orders(self) -> Dict[MarketTradingPairTuple, List[LimitOrder]]:
 *         return self._sb_order_tracker.market_pair_to_active_orders
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_28market_info_to_active_orders_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_28market_info_to_active_orders_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_28market_info_to_active_orders___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_28market_info_to_active_orders___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":319
 *     @property
 *     def market_info_to_active_orders(self) -> Dict[MarketTradingPairTuple, List[LimitOrder]]:
 *         return self._sb_order_tracker.market_pair_to_active_orders             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base._sb_order_tracker), __pyx_n_s_market_pair_to_active_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 319, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":317
 *         return self._price_delegate.get_price_by_type(PriceType.MidPrice)
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def market_info_to_active_orders(self) -> Dict[MarketTradingPairTuple, List[LimitOrder]]:
 *         return self._sb_order_tracker.market_pair_to_active_orders
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.market_info_to_active_orders.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":321
 *         return self._sb_order_tracker.market_pair_to_active_orders
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def active_orders(self) -> List[LimitOrder]:
 *         if self._market_info not in self.market_info_to_active_orders:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_13active_orders_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_13active_orders_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_13active_orders___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_13active_orders___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":323
 *     @property
 *     def active_orders(self) -> List[LimitOrder]:
 *         if self._market_info not in self.market_info_to_active_orders:             # <<<<<<<<<<<<<<
 *             return []
 *         return self.market_info_to_active_orders[self._market_info]
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_market_info_to_active_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_v_self->_market_info, __pyx_t_1, Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 323, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":324
 *     def active_orders(self) -> List[LimitOrder]:
 *         if self._market_info not in self.market_info_to_active_orders:
 *             return []             # <<<<<<<<<<<<<<
 *         return self.market_info_to_active_orders[self._market_info]
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 324, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    goto __pyx_L0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":323
 *     @property
 *     def active_orders(self) -> List[LimitOrder]:
 *         if self._market_info not in self.market_info_to_active_orders:             # <<<<<<<<<<<<<<
 *             return []
 *         return self.market_info_to_active_orders[self._market_info]
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":325
 *         if self._market_info not in self.market_info_to_active_orders:
 *             return []
 *         return self.market_info_to_active_orders[self._market_info]             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_market_info_to_active_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_self->_market_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 325, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":321
 *         return self._sb_order_tracker.market_pair_to_active_orders
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def active_orders(self) -> List[LimitOrder]:
 *         if self._market_info not in self.market_info_to_active_orders:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.active_orders.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":327
 *         return self.market_info_to_active_orders[self._market_info]
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def active_non_hanging_orders(self) -> List[LimitOrder]:
 *         orders = [o for o in self.active_orders if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)]
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25active_non_hanging_orders_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25active_non_hanging_orders_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25active_non_hanging_orders___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25active_non_hanging_orders___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_v_orders = NULL;
  PyObject *__pyx_8genexpr1__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  unsigned int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":329
 *     @property
 *     def active_non_hanging_orders(self) -> List[LimitOrder]:
 *         orders = [o for o in self.active_orders if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)]             # <<<<<<<<<<<<<<
 *         return orders
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 329, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_active_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 329, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 329, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 329, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 329, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 329, __pyx_L5_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 329, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 329, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 329, __pyx_L5_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 329, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 329, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_o, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_n_s_is_order_id_in_hanging_orders); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 329, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr1__pyx_v_o, __pyx_n_s_client_order_id); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 329, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = NULL;
      __pyx_t_9 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_9 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_7};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_9, 1+__pyx_t_9);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 329, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 329, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_11 = (!__pyx_t_10);
      if (__pyx_t_11) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_8genexpr1__pyx_v_o))) __PYX_ERR(0, 329, __pyx_L5_error)
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_o); __pyx_8genexpr1__pyx_v_o = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_o); __pyx_8genexpr1__pyx_v_o = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __pyx_v_orders = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":330
 *     def active_non_hanging_orders(self) -> List[LimitOrder]:
 *         orders = [o for o in self.active_orders if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)]
 *         return orders             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_orders);
  __pyx_r = __pyx_v_orders;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":327
 *         return self.market_info_to_active_orders[self._market_info]
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def active_non_hanging_orders(self) -> List[LimitOrder]:
 *         orders = [o for o in self.active_orders if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.active_non_hanging_orders.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_orders);
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_o);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":332
 *         return orders
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def active_buys(self) -> List[LimitOrder]:
 *         return [o for o in self.active_orders if o.is_buy]
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11active_buys_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11active_buys_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11active_buys___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11active_buys___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_8genexpr2__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":334
 *     @property
 *     def active_buys(self) -> List[LimitOrder]:
 *         return [o for o in self.active_orders if o.is_buy]             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 334, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_active_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 334, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 334, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 334, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 334, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 334, __pyx_L5_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 334, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 334, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 334, __pyx_L5_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 334, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 334, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr2__pyx_v_o, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr2__pyx_v_o, __pyx_n_s_is_buy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 334, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 334, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_6) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_8genexpr2__pyx_v_o))) __PYX_ERR(0, 334, __pyx_L5_error)
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_o); __pyx_8genexpr2__pyx_v_o = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_o); __pyx_8genexpr2__pyx_v_o = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":332
 *         return orders
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def active_buys(self) -> List[LimitOrder]:
 *         return [o for o in self.active_orders if o.is_buy]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.active_buys.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr2__pyx_v_o);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":336
 *         return [o for o in self.active_orders if o.is_buy]
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def active_sells(self) -> List[LimitOrder]:
 *         return [o for o in self.active_orders if not o.is_buy]
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12active_sells_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12active_sells_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12active_sells___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12active_sells___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_8genexpr3__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":338
 *     @property
 *     def active_sells(self) -> List[LimitOrder]:
 *         return [o for o in self.active_orders if not o.is_buy]             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 338, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_active_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 338, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 338, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 338, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 338, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 338, __pyx_L5_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 338, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 338, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 338, __pyx_L5_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 338, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 338, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr3__pyx_v_o, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr3__pyx_v_o, __pyx_n_s_is_buy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 338, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 338, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_7 = (!__pyx_t_6);
      if (__pyx_t_7) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_8genexpr3__pyx_v_o))) __PYX_ERR(0, 338, __pyx_L5_error)
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_o); __pyx_8genexpr3__pyx_v_o = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_o); __pyx_8genexpr3__pyx_v_o = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":336
 *         return [o for o in self.active_orders if o.is_buy]
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def active_sells(self) -> List[LimitOrder]:
 *         return [o for o in self.active_orders if not o.is_buy]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.active_sells.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr3__pyx_v_o);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":340
 *         return [o for o in self.active_orders if not o.is_buy]
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def logging_options(self) -> int:
 *         return self._logging_options
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15logging_options_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15logging_options_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15logging_options___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15logging_options___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":342
 *     @property
 *     def logging_options(self) -> int:
 *         return self._logging_options             # <<<<<<<<<<<<<<
 * 
 *     @logging_options.setter
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->_logging_options); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 342, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":340
 *         return [o for o in self.active_orders if not o.is_buy]
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def logging_options(self) -> int:
 *         return self._logging_options
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.logging_options.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":344
 *         return self._logging_options
 * 
 *     @logging_options.setter             # <<<<<<<<<<<<<<
 *     def logging_options(self, int64_t logging_options):
 *         self._logging_options = logging_options
 */

/* Python wrapper */
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15logging_options_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_logging_options); /*proto*/
static int __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15logging_options_3__set__(PyObject *__pyx_v_self, PyObject *__pyx_arg_logging_options) {
  int64_t __pyx_v_logging_options;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__set__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  assert(__pyx_arg_logging_options); {
    __pyx_v_logging_options = __Pyx_PyInt_As_int64_t(__pyx_arg_logging_options); if (unlikely((__pyx_v_logging_options == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 345, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.logging_options.__set__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15logging_options_2__set__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), ((int64_t)__pyx_v_logging_options));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15logging_options_2__set__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, int64_t __pyx_v_logging_options) {
  int __pyx_r;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":346
 *     @logging_options.setter
 *     def logging_options(self, int64_t logging_options):
 *         self._logging_options = logging_options             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_v_self->_logging_options = __pyx_v_logging_options;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":344
 *         return self._logging_options
 * 
 *     @logging_options.setter             # <<<<<<<<<<<<<<
 *     def logging_options(self, int64_t logging_options):
 *         self._logging_options = logging_options
 */

  /* function exit code */
  __pyx_r = 0;
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":348
 *         self._logging_options = logging_options
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def hanging_orders_tracker(self):
 *         return self._hanging_orders_tracker
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_22hanging_orders_tracker_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_22hanging_orders_tracker_1__get__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_22hanging_orders_tracker___get__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_22hanging_orders_tracker___get__(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":350
 *     @property
 *     def hanging_orders_tracker(self):
 *         return self._hanging_orders_tracker             # <<<<<<<<<<<<<<
 * 
 *     def update_from_config_map(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_hanging_orders_tracker);
  __pyx_r = __pyx_v_self->_hanging_orders_tracker;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":348
 *         self._logging_options = logging_options
 * 
 *     @property             # <<<<<<<<<<<<<<
 *     def hanging_orders_tracker(self):
 *         return self._hanging_orders_tracker
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":352
 *         return self._hanging_orders_tracker
 * 
 *     def update_from_config_map(self):             # <<<<<<<<<<<<<<
 *         self.get_config_map_execution_mode()
 *         self.get_config_map_hanging_orders()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_13update_from_config_map(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_13update_from_config_map = {"update_from_config_map", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_13update_from_config_map, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_13update_from_config_map(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("update_from_config_map (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("update_from_config_map", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "update_from_config_map", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12update_from_config_map(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12update_from_config_map(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("update_from_config_map", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":353
 * 
 *     def update_from_config_map(self):
 *         self.get_config_map_execution_mode()             # <<<<<<<<<<<<<<
 *         self.get_config_map_hanging_orders()
 *         self.get_config_map_indicators()
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_config_map_execution_mode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 353, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":354
 *     def update_from_config_map(self):
 *         self.get_config_map_execution_mode()
 *         self.get_config_map_hanging_orders()             # <<<<<<<<<<<<<<
 *         self.get_config_map_indicators()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_config_map_hanging_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":355
 *         self.get_config_map_execution_mode()
 *         self.get_config_map_hanging_orders()
 *         self.get_config_map_indicators()             # <<<<<<<<<<<<<<
 * 
 *     def get_config_map_execution_mode(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_config_map_indicators); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 355, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":352
 *         return self._hanging_orders_tracker
 * 
 *     def update_from_config_map(self):             # <<<<<<<<<<<<<<
 *         self.get_config_map_execution_mode()
 *         self.get_config_map_hanging_orders()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.update_from_config_map", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":357
 *         self.get_config_map_indicators()
 * 
 *     def get_config_map_execution_mode(self):             # <<<<<<<<<<<<<<
 *         try:
 *             execution_mode = self._config_map.execution_timeframe_mode.title
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15get_config_map_execution_mode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15get_config_map_execution_mode = {"get_config_map_execution_mode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15get_config_map_execution_mode, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15get_config_map_execution_mode(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_config_map_execution_mode (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_config_map_execution_mode", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_config_map_execution_mode", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_14get_config_map_execution_mode(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_14get_config_map_execution_mode(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_v_execution_mode = NULL;
  PyObject *__pyx_v_execution_timeframe = NULL;
  PyObject *__pyx_v_start_time = NULL;
  PyObject *__pyx_v_end_time = NULL;
  PyObject *__pyx_v_execution_state = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_config_map_execution_mode", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":358
 * 
 *     def get_config_map_execution_mode(self):
 *         try:             # <<<<<<<<<<<<<<
 *             execution_mode = self._config_map.execution_timeframe_mode.title
 *             execution_timeframe = self._config_map.execution_timeframe_mode.model_config["title"]
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":359
 *     def get_config_map_execution_mode(self):
 *         try:
 *             execution_mode = self._config_map.execution_timeframe_mode.title             # <<<<<<<<<<<<<<
 *             execution_timeframe = self._config_map.execution_timeframe_mode.model_config["title"]
 *             if execution_mode == FromDateToDateModel.model_config["title"]:
 */
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_execution_timeframe_mode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 359, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_title); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 359, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_execution_mode = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":360
 *         try:
 *             execution_mode = self._config_map.execution_timeframe_mode.title
 *             execution_timeframe = self._config_map.execution_timeframe_mode.model_config["title"]             # <<<<<<<<<<<<<<
 *             if execution_mode == FromDateToDateModel.model_config["title"]:
 *                 start_time = self._config_map.execution_timeframe_mode.start_datetime
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_execution_timeframe_mode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 360, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_model_config); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 360, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_4, __pyx_n_u_title); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 360, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_execution_timeframe = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":361
 *             execution_mode = self._config_map.execution_timeframe_mode.title
 *             execution_timeframe = self._config_map.execution_timeframe_mode.model_config["title"]
 *             if execution_mode == FromDateToDateModel.model_config["title"]:             # <<<<<<<<<<<<<<
 *                 start_time = self._config_map.execution_timeframe_mode.start_datetime
 *                 end_time = self._config_map.execution_timeframe_mode.end_datetime
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_FromDateToDateModel); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 361, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_model_config); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 361, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Dict_GetItem(__pyx_t_4, __pyx_n_u_title); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 361, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_execution_mode, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 361, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 361, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":362
 *             execution_timeframe = self._config_map.execution_timeframe_mode.model_config["title"]
 *             if execution_mode == FromDateToDateModel.model_config["title"]:
 *                 start_time = self._config_map.execution_timeframe_mode.start_datetime             # <<<<<<<<<<<<<<
 *                 end_time = self._config_map.execution_timeframe_mode.end_datetime
 *                 execution_state = RunInTimeConditionalExecutionState(start_timestamp=start_time, end_timestamp=end_time)
 */
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_execution_timeframe_mode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 362, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_start_datetime); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 362, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_v_start_time = __pyx_t_5;
        __pyx_t_5 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":363
 *             if execution_mode == FromDateToDateModel.model_config["title"]:
 *                 start_time = self._config_map.execution_timeframe_mode.start_datetime
 *                 end_time = self._config_map.execution_timeframe_mode.end_datetime             # <<<<<<<<<<<<<<
 *                 execution_state = RunInTimeConditionalExecutionState(start_timestamp=start_time, end_timestamp=end_time)
 *             elif execution_mode == DailyBetweenTimesModel.model_config["title"]:
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_execution_timeframe_mode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 363, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_end_datetime); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 363, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_v_end_time = __pyx_t_4;
        __pyx_t_4 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":364
 *                 start_time = self._config_map.execution_timeframe_mode.start_datetime
 *                 end_time = self._config_map.execution_timeframe_mode.end_datetime
 *                 execution_state = RunInTimeConditionalExecutionState(start_timestamp=start_time, end_timestamp=end_time)             # <<<<<<<<<<<<<<
 *             elif execution_mode == DailyBetweenTimesModel.model_config["title"]:
 *                 start_time = self._config_map.execution_timeframe_mode.start_time
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_RunInTimeConditionalExecutionSta); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 364, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 364, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_start_timestamp, __pyx_v_start_time) < 0) __PYX_ERR(0, 364, __pyx_L3_error)
        if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_end_timestamp, __pyx_v_end_time) < 0) __PYX_ERR(0, 364, __pyx_L3_error)
        __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 364, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_v_execution_state = __pyx_t_7;
        __pyx_t_7 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":361
 *             execution_mode = self._config_map.execution_timeframe_mode.title
 *             execution_timeframe = self._config_map.execution_timeframe_mode.model_config["title"]
 *             if execution_mode == FromDateToDateModel.model_config["title"]:             # <<<<<<<<<<<<<<
 *                 start_time = self._config_map.execution_timeframe_mode.start_datetime
 *                 end_time = self._config_map.execution_timeframe_mode.end_datetime
 */
        goto __pyx_L9;
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":365
 *                 end_time = self._config_map.execution_timeframe_mode.end_datetime
 *                 execution_state = RunInTimeConditionalExecutionState(start_timestamp=start_time, end_timestamp=end_time)
 *             elif execution_mode == DailyBetweenTimesModel.model_config["title"]:             # <<<<<<<<<<<<<<
 *                 start_time = self._config_map.execution_timeframe_mode.start_time
 *                 end_time = self._config_map.execution_timeframe_mode.end_time
 */
      __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_DailyBetweenTimesModel); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 365, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_model_config); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 365, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_Dict_GetItem(__pyx_t_5, __pyx_n_u_title); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 365, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyObject_RichCompare(__pyx_v_execution_mode, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 365, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 365, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_6) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":366
 *                 execution_state = RunInTimeConditionalExecutionState(start_timestamp=start_time, end_timestamp=end_time)
 *             elif execution_mode == DailyBetweenTimesModel.model_config["title"]:
 *                 start_time = self._config_map.execution_timeframe_mode.start_time             # <<<<<<<<<<<<<<
 *                 end_time = self._config_map.execution_timeframe_mode.end_time
 *                 execution_state = RunInTimeConditionalExecutionState(start_timestamp=start_time, end_timestamp=end_time)
 */
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_execution_timeframe_mode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 366, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_start_time); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 366, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_v_start_time = __pyx_t_7;
        __pyx_t_7 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":367
 *             elif execution_mode == DailyBetweenTimesModel.model_config["title"]:
 *                 start_time = self._config_map.execution_timeframe_mode.start_time
 *                 end_time = self._config_map.execution_timeframe_mode.end_time             # <<<<<<<<<<<<<<
 *                 execution_state = RunInTimeConditionalExecutionState(start_timestamp=start_time, end_timestamp=end_time)
 *             else:
 */
        __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_execution_timeframe_mode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 367, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_end_time); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 367, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_end_time = __pyx_t_5;
        __pyx_t_5 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":368
 *                 start_time = self._config_map.execution_timeframe_mode.start_time
 *                 end_time = self._config_map.execution_timeframe_mode.end_time
 *                 execution_state = RunInTimeConditionalExecutionState(start_timestamp=start_time, end_timestamp=end_time)             # <<<<<<<<<<<<<<
 *             else:
 *                 start_time = None
 */
        __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_RunInTimeConditionalExecutionSta); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 368, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_7 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 368, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_7);
        if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_start_timestamp, __pyx_v_start_time) < 0) __PYX_ERR(0, 368, __pyx_L3_error)
        if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_end_timestamp, __pyx_v_end_time) < 0) __PYX_ERR(0, 368, __pyx_L3_error)
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 368, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __pyx_v_execution_state = __pyx_t_4;
        __pyx_t_4 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":365
 *                 end_time = self._config_map.execution_timeframe_mode.end_datetime
 *                 execution_state = RunInTimeConditionalExecutionState(start_timestamp=start_time, end_timestamp=end_time)
 *             elif execution_mode == DailyBetweenTimesModel.model_config["title"]:             # <<<<<<<<<<<<<<
 *                 start_time = self._config_map.execution_timeframe_mode.start_time
 *                 end_time = self._config_map.execution_timeframe_mode.end_time
 */
        goto __pyx_L9;
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":370
 *                 execution_state = RunInTimeConditionalExecutionState(start_timestamp=start_time, end_timestamp=end_time)
 *             else:
 *                 start_time = None             # <<<<<<<<<<<<<<
 *                 end_time = None
 *                 execution_state = RunAlwaysExecutionState()
 */
      /*else*/ {
        __Pyx_INCREF(Py_None);
        __pyx_v_start_time = Py_None;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":371
 *             else:
 *                 start_time = None
 *                 end_time = None             # <<<<<<<<<<<<<<
 *                 execution_state = RunAlwaysExecutionState()
 * 
 */
        __Pyx_INCREF(Py_None);
        __pyx_v_end_time = Py_None;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":372
 *                 start_time = None
 *                 end_time = None
 *                 execution_state = RunAlwaysExecutionState()             # <<<<<<<<<<<<<<
 * 
 *             # Something has changed?
 */
        __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_RunAlwaysExecutionState); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 372, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_5 = NULL;
        __pyx_t_8 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_7))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_7, function);
            __pyx_t_8 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 372, __pyx_L3_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __pyx_v_execution_state = __pyx_t_4;
        __pyx_t_4 = 0;
      }
      __pyx_L9:;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":375
 * 
 *             # Something has changed?
 *             if self._execution_state is None or self._execution_state != execution_state:             # <<<<<<<<<<<<<<
 *                 self._execution_state = execution_state
 *                 self._execution_mode = execution_mode
 */
      __pyx_t_9 = (__pyx_v_self->_execution_state == Py_None);
      if (!__pyx_t_9) {
      } else {
        __pyx_t_6 = __pyx_t_9;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_self->_execution_state, __pyx_v_execution_state, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 375, __pyx_L3_error)
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 375, __pyx_L3_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __pyx_t_9;
      __pyx_L11_bool_binop_done:;
      if (__pyx_t_6) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":376
 *             # Something has changed?
 *             if self._execution_state is None or self._execution_state != execution_state:
 *                 self._execution_state = execution_state             # <<<<<<<<<<<<<<
 *                 self._execution_mode = execution_mode
 *                 self._execution_timeframe = execution_timeframe
 */
        __Pyx_INCREF(__pyx_v_execution_state);
        __Pyx_GIVEREF(__pyx_v_execution_state);
        __Pyx_GOTREF(__pyx_v_self->_execution_state);
        __Pyx_DECREF(__pyx_v_self->_execution_state);
        __pyx_v_self->_execution_state = __pyx_v_execution_state;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":377
 *             if self._execution_state is None or self._execution_state != execution_state:
 *                 self._execution_state = execution_state
 *                 self._execution_mode = execution_mode             # <<<<<<<<<<<<<<
 *                 self._execution_timeframe = execution_timeframe
 *                 self._start_time = start_time
 */
        if (!(likely(PyUnicode_CheckExact(__pyx_v_execution_mode))||((__pyx_v_execution_mode) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_execution_mode))) __PYX_ERR(0, 377, __pyx_L3_error)
        __pyx_t_4 = __pyx_v_execution_mode;
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_v_self->_execution_mode);
        __Pyx_DECREF(__pyx_v_self->_execution_mode);
        __pyx_v_self->_execution_mode = ((PyObject*)__pyx_t_4);
        __pyx_t_4 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":378
 *                 self._execution_state = execution_state
 *                 self._execution_mode = execution_mode
 *                 self._execution_timeframe = execution_timeframe             # <<<<<<<<<<<<<<
 *                 self._start_time = start_time
 *                 self._end_time = end_time
 */
        if (!(likely(PyUnicode_CheckExact(__pyx_v_execution_timeframe))||((__pyx_v_execution_timeframe) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_execution_timeframe))) __PYX_ERR(0, 378, __pyx_L3_error)
        __pyx_t_4 = __pyx_v_execution_timeframe;
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        __Pyx_GOTREF(__pyx_v_self->_execution_timeframe);
        __Pyx_DECREF(__pyx_v_self->_execution_timeframe);
        __pyx_v_self->_execution_timeframe = ((PyObject*)__pyx_t_4);
        __pyx_t_4 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":379
 *                 self._execution_mode = execution_mode
 *                 self._execution_timeframe = execution_timeframe
 *                 self._start_time = start_time             # <<<<<<<<<<<<<<
 *                 self._end_time = end_time
 *         except AttributeError:
 */
        __Pyx_INCREF(__pyx_v_start_time);
        __Pyx_GIVEREF(__pyx_v_start_time);
        __Pyx_GOTREF(__pyx_v_self->_start_time);
        __Pyx_DECREF(__pyx_v_self->_start_time);
        __pyx_v_self->_start_time = __pyx_v_start_time;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":380
 *                 self._execution_timeframe = execution_timeframe
 *                 self._start_time = start_time
 *                 self._end_time = end_time             # <<<<<<<<<<<<<<
 *         except AttributeError:
 *             # A parameter is missing in the execution timeframe mode configuration
 */
        __Pyx_INCREF(__pyx_v_end_time);
        __Pyx_GIVEREF(__pyx_v_end_time);
        __Pyx_GOTREF(__pyx_v_self->_end_time);
        __Pyx_DECREF(__pyx_v_self->_end_time);
        __pyx_v_self->_end_time = __pyx_v_end_time;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":375
 * 
 *             # Something has changed?
 *             if self._execution_state is None or self._execution_state != execution_state:             # <<<<<<<<<<<<<<
 *                 self._execution_state = execution_state
 *                 self._execution_mode = execution_mode
 */
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":358
 * 
 *     def get_config_map_execution_mode(self):
 *         try:             # <<<<<<<<<<<<<<
 *             execution_mode = self._config_map.execution_timeframe_mode.title
 *             execution_timeframe = self._config_map.execution_timeframe_mode.model_config["title"]
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":381
 *                 self._start_time = start_time
 *                 self._end_time = end_time
 *         except AttributeError:             # <<<<<<<<<<<<<<
 *             # A parameter is missing in the execution timeframe mode configuration
 *             pass
 */
    __pyx_t_10 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_AttributeError);
    if (__pyx_t_10) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":358
 * 
 *     def get_config_map_execution_mode(self):
 *         try:             # <<<<<<<<<<<<<<
 *             execution_mode = self._config_map.execution_timeframe_mode.title
 *             execution_timeframe = self._config_map.execution_timeframe_mode.model_config["title"]
 */
    __pyx_L5_except_error:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":357
 *         self.get_config_map_indicators()
 * 
 *     def get_config_map_execution_mode(self):             # <<<<<<<<<<<<<<
 *         try:
 *             execution_mode = self._config_map.execution_timeframe_mode.title
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.get_config_map_execution_mode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_execution_mode);
  __Pyx_XDECREF(__pyx_v_execution_timeframe);
  __Pyx_XDECREF(__pyx_v_start_time);
  __Pyx_XDECREF(__pyx_v_end_time);
  __Pyx_XDECREF(__pyx_v_execution_state);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":385
 *             pass
 * 
 *     def get_config_map_hanging_orders(self):             # <<<<<<<<<<<<<<
 *         if self._config_map.hanging_orders_mode.title == TrackHangingOrdersModel.model_config['title']:
 *             hanging_orders_enabled = True
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17get_config_map_hanging_orders(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17get_config_map_hanging_orders = {"get_config_map_hanging_orders", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17get_config_map_hanging_orders, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17get_config_map_hanging_orders(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_config_map_hanging_orders (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_config_map_hanging_orders", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_config_map_hanging_orders", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_16get_config_map_hanging_orders(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_16get_config_map_hanging_orders(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  int __pyx_v_hanging_orders_enabled;
  PyObject *__pyx_v_hanging_orders_cancel_pct = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  unsigned int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_config_map_hanging_orders", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":386
 * 
 *     def get_config_map_hanging_orders(self):
 *         if self._config_map.hanging_orders_mode.title == TrackHangingOrdersModel.model_config['title']:             # <<<<<<<<<<<<<<
 *             hanging_orders_enabled = True
 *             hanging_orders_cancel_pct = self._config_map.hanging_orders_mode.hanging_orders_cancel_pct
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_hanging_orders_mode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_title); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_TrackHangingOrdersModel); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_model_config); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Dict_GetItem(__pyx_t_3, __pyx_n_u_title); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 386, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_4) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":387
 *     def get_config_map_hanging_orders(self):
 *         if self._config_map.hanging_orders_mode.title == TrackHangingOrdersModel.model_config['title']:
 *             hanging_orders_enabled = True             # <<<<<<<<<<<<<<
 *             hanging_orders_cancel_pct = self._config_map.hanging_orders_mode.hanging_orders_cancel_pct
 *         else:
 */
    __pyx_v_hanging_orders_enabled = 1;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":388
 *         if self._config_map.hanging_orders_mode.title == TrackHangingOrdersModel.model_config['title']:
 *             hanging_orders_enabled = True
 *             hanging_orders_cancel_pct = self._config_map.hanging_orders_mode.hanging_orders_cancel_pct             # <<<<<<<<<<<<<<
 *         else:
 *             hanging_orders_enabled = False
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_hanging_orders_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_hanging_orders_cancel_pct); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 388, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_hanging_orders_cancel_pct = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":386
 * 
 *     def get_config_map_hanging_orders(self):
 *         if self._config_map.hanging_orders_mode.title == TrackHangingOrdersModel.model_config['title']:             # <<<<<<<<<<<<<<
 *             hanging_orders_enabled = True
 *             hanging_orders_cancel_pct = self._config_map.hanging_orders_mode.hanging_orders_cancel_pct
 */
    goto __pyx_L3;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":390
 *             hanging_orders_cancel_pct = self._config_map.hanging_orders_mode.hanging_orders_cancel_pct
 *         else:
 *             hanging_orders_enabled = False             # <<<<<<<<<<<<<<
 *             hanging_orders_cancel_pct = Decimal("0")
 * 
 */
  /*else*/ {
    __pyx_v_hanging_orders_enabled = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":391
 *         else:
 *             hanging_orders_enabled = False
 *             hanging_orders_cancel_pct = Decimal("0")             # <<<<<<<<<<<<<<
 * 
 *         if self._hanging_orders_enabled != hanging_orders_enabled:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 391, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_kp_u_0};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 391, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_v_hanging_orders_cancel_pct = __pyx_t_1;
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":393
 *             hanging_orders_cancel_pct = Decimal("0")
 * 
 *         if self._hanging_orders_enabled != hanging_orders_enabled:             # <<<<<<<<<<<<<<
 *             # Hanging order tracker instance doesn't exist - create from scratch
 *             self._hanging_orders_enabled = hanging_orders_enabled
 */
  __pyx_t_4 = (__pyx_v_self->_hanging_orders_enabled != __pyx_v_hanging_orders_enabled);
  if (__pyx_t_4) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":395
 *         if self._hanging_orders_enabled != hanging_orders_enabled:
 *             # Hanging order tracker instance doesn't exist - create from scratch
 *             self._hanging_orders_enabled = hanging_orders_enabled             # <<<<<<<<<<<<<<
 *             self._hanging_orders_cancel_pct = hanging_orders_cancel_pct
 *             self._hanging_orders_tracker = HangingOrdersTracker(self,
 */
    __pyx_v_self->_hanging_orders_enabled = __pyx_v_hanging_orders_enabled;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":396
 *             # Hanging order tracker instance doesn't exist - create from scratch
 *             self._hanging_orders_enabled = hanging_orders_enabled
 *             self._hanging_orders_cancel_pct = hanging_orders_cancel_pct             # <<<<<<<<<<<<<<
 *             self._hanging_orders_tracker = HangingOrdersTracker(self,
 *                                                                 hanging_orders_cancel_pct / Decimal('100'))
 */
    __Pyx_INCREF(__pyx_v_hanging_orders_cancel_pct);
    __Pyx_GIVEREF(__pyx_v_hanging_orders_cancel_pct);
    __Pyx_GOTREF(__pyx_v_self->_hanging_orders_cancel_pct);
    __Pyx_DECREF(__pyx_v_self->_hanging_orders_cancel_pct);
    __pyx_v_self->_hanging_orders_cancel_pct = __pyx_v_hanging_orders_cancel_pct;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":397
 *             self._hanging_orders_enabled = hanging_orders_enabled
 *             self._hanging_orders_cancel_pct = hanging_orders_cancel_pct
 *             self._hanging_orders_tracker = HangingOrdersTracker(self,             # <<<<<<<<<<<<<<
 *                                                                 hanging_orders_cancel_pct / Decimal('100'))
 *             self._hanging_orders_tracker.register_events(self.active_markets)
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_HangingOrdersTracker); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":398
 *             self._hanging_orders_cancel_pct = hanging_orders_cancel_pct
 *             self._hanging_orders_tracker = HangingOrdersTracker(self,
 *                                                                 hanging_orders_cancel_pct / Decimal('100'))             # <<<<<<<<<<<<<<
 *             self._hanging_orders_tracker.register_events(self.active_markets)
 *         elif self._hanging_orders_cancel_pct != hanging_orders_cancel_pct:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_kp_u_100};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 398, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __pyx_t_6 = __Pyx_PyNumber_Divide(__pyx_v_hanging_orders_cancel_pct, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 398, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_2, ((PyObject *)__pyx_v_self), __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 397, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":397
 *             self._hanging_orders_enabled = hanging_orders_enabled
 *             self._hanging_orders_cancel_pct = hanging_orders_cancel_pct
 *             self._hanging_orders_tracker = HangingOrdersTracker(self,             # <<<<<<<<<<<<<<
 *                                                                 hanging_orders_cancel_pct / Decimal('100'))
 *             self._hanging_orders_tracker.register_events(self.active_markets)
 */
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->_hanging_orders_tracker);
    __Pyx_DECREF(__pyx_v_self->_hanging_orders_tracker);
    __pyx_v_self->_hanging_orders_tracker = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":399
 *             self._hanging_orders_tracker = HangingOrdersTracker(self,
 *                                                                 hanging_orders_cancel_pct / Decimal('100'))
 *             self._hanging_orders_tracker.register_events(self.active_markets)             # <<<<<<<<<<<<<<
 *         elif self._hanging_orders_cancel_pct != hanging_orders_cancel_pct:
 *             # Hanging order tracker instance existst - only update variable
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_n_s_register_events); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_active_markets); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 399, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 399, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":393
 *             hanging_orders_cancel_pct = Decimal("0")
 * 
 *         if self._hanging_orders_enabled != hanging_orders_enabled:             # <<<<<<<<<<<<<<
 *             # Hanging order tracker instance doesn't exist - create from scratch
 *             self._hanging_orders_enabled = hanging_orders_enabled
 */
    goto __pyx_L4;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":400
 *                                                                 hanging_orders_cancel_pct / Decimal('100'))
 *             self._hanging_orders_tracker.register_events(self.active_markets)
 *         elif self._hanging_orders_cancel_pct != hanging_orders_cancel_pct:             # <<<<<<<<<<<<<<
 *             # Hanging order tracker instance existst - only update variable
 *             self._hanging_orders_cancel_pct = hanging_orders_cancel_pct
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_self->_hanging_orders_cancel_pct, __pyx_v_hanging_orders_cancel_pct, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 400, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":402
 *         elif self._hanging_orders_cancel_pct != hanging_orders_cancel_pct:
 *             # Hanging order tracker instance existst - only update variable
 *             self._hanging_orders_cancel_pct = hanging_orders_cancel_pct             # <<<<<<<<<<<<<<
 *             self._hanging_orders_tracker.hanging_orders_cancel_pct = hanging_orders_cancel_pct / Decimal('100')
 * 
 */
    __Pyx_INCREF(__pyx_v_hanging_orders_cancel_pct);
    __Pyx_GIVEREF(__pyx_v_hanging_orders_cancel_pct);
    __Pyx_GOTREF(__pyx_v_self->_hanging_orders_cancel_pct);
    __Pyx_DECREF(__pyx_v_self->_hanging_orders_cancel_pct);
    __pyx_v_self->_hanging_orders_cancel_pct = __pyx_v_hanging_orders_cancel_pct;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":403
 *             # Hanging order tracker instance existst - only update variable
 *             self._hanging_orders_cancel_pct = hanging_orders_cancel_pct
 *             self._hanging_orders_tracker.hanging_orders_cancel_pct = hanging_orders_cancel_pct / Decimal('100')             # <<<<<<<<<<<<<<
 * 
 *     def get_config_map_indicators(self):
 */
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_kp_u_100};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 403, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_3 = __Pyx_PyNumber_Divide(__pyx_v_hanging_orders_cancel_pct, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 403, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_n_s_hanging_orders_cancel_pct, __pyx_t_3) < 0) __PYX_ERR(0, 403, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":400
 *                                                                 hanging_orders_cancel_pct / Decimal('100'))
 *             self._hanging_orders_tracker.register_events(self.active_markets)
 *         elif self._hanging_orders_cancel_pct != hanging_orders_cancel_pct:             # <<<<<<<<<<<<<<
 *             # Hanging order tracker instance existst - only update variable
 *             self._hanging_orders_cancel_pct = hanging_orders_cancel_pct
 */
  }
  __pyx_L4:;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":385
 *             pass
 * 
 *     def get_config_map_hanging_orders(self):             # <<<<<<<<<<<<<<
 *         if self._config_map.hanging_orders_mode.title == TrackHangingOrdersModel.model_config['title']:
 *             hanging_orders_enabled = True
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.get_config_map_hanging_orders", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_hanging_orders_cancel_pct);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":405
 *             self._hanging_orders_tracker.hanging_orders_cancel_pct = hanging_orders_cancel_pct / Decimal('100')
 * 
 *     def get_config_map_indicators(self):             # <<<<<<<<<<<<<<
 *         volatility_buffer_size = self._config_map.volatility_buffer_size
 *         trading_intensity_buffer_size = self._config_map.trading_intensity_buffer_size
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_19get_config_map_indicators(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_19get_config_map_indicators = {"get_config_map_indicators", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_19get_config_map_indicators, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_19get_config_map_indicators(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_config_map_indicators (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_config_map_indicators", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_config_map_indicators", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_18get_config_map_indicators(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_18get_config_map_indicators(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_v_volatility_buffer_size = NULL;
  PyObject *__pyx_v_trading_intensity_buffer_size = NULL;
  PyObject *__pyx_v_ticks_to_be_ready_after = NULL;
  PyObject *__pyx_v_ticks_to_be_ready_before = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_config_map_indicators", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":406
 * 
 *     def get_config_map_indicators(self):
 *         volatility_buffer_size = self._config_map.volatility_buffer_size             # <<<<<<<<<<<<<<
 *         trading_intensity_buffer_size = self._config_map.trading_intensity_buffer_size
 *         ticks_to_be_ready_after = max(volatility_buffer_size, trading_intensity_buffer_size)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_volatility_buffer_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_volatility_buffer_size = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":407
 *     def get_config_map_indicators(self):
 *         volatility_buffer_size = self._config_map.volatility_buffer_size
 *         trading_intensity_buffer_size = self._config_map.trading_intensity_buffer_size             # <<<<<<<<<<<<<<
 *         ticks_to_be_ready_after = max(volatility_buffer_size, trading_intensity_buffer_size)
 *         ticks_to_be_ready_before = max(self._volatility_buffer_size, self._trading_intensity_buffer_size)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_trading_intensity_buffer_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_trading_intensity_buffer_size = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":408
 *         volatility_buffer_size = self._config_map.volatility_buffer_size
 *         trading_intensity_buffer_size = self._config_map.trading_intensity_buffer_size
 *         ticks_to_be_ready_after = max(volatility_buffer_size, trading_intensity_buffer_size)             # <<<<<<<<<<<<<<
 *         ticks_to_be_ready_before = max(self._volatility_buffer_size, self._trading_intensity_buffer_size)
 * 
 */
  __Pyx_INCREF(__pyx_v_trading_intensity_buffer_size);
  __pyx_t_1 = __pyx_v_trading_intensity_buffer_size;
  __Pyx_INCREF(__pyx_v_volatility_buffer_size);
  __pyx_t_2 = __pyx_v_volatility_buffer_size;
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 408, __pyx_L1_error)
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_3 = __pyx_t_1;
  } else {
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = __pyx_t_2;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_ticks_to_be_ready_after = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":409
 *         trading_intensity_buffer_size = self._config_map.trading_intensity_buffer_size
 *         ticks_to_be_ready_after = max(volatility_buffer_size, trading_intensity_buffer_size)
 *         ticks_to_be_ready_before = max(self._volatility_buffer_size, self._trading_intensity_buffer_size)             # <<<<<<<<<<<<<<
 * 
 *         if self._volatility_buffer_size == 0 or self._volatility_buffer_size != volatility_buffer_size:
 */
  __pyx_t_6 = __pyx_v_self->_trading_intensity_buffer_size;
  __pyx_t_7 = __pyx_v_self->_volatility_buffer_size;
  __pyx_t_5 = (__pyx_t_6 > __pyx_t_7);
  if (__pyx_t_5) {
    __pyx_t_8 = __pyx_t_6;
  } else {
    __pyx_t_8 = __pyx_t_7;
  }
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 409, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_ticks_to_be_ready_before = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":411
 *         ticks_to_be_ready_before = max(self._volatility_buffer_size, self._trading_intensity_buffer_size)
 * 
 *         if self._volatility_buffer_size == 0 or self._volatility_buffer_size != volatility_buffer_size:             # <<<<<<<<<<<<<<
 *             self._volatility_buffer_size = volatility_buffer_size
 * 
 */
  __pyx_t_9 = (__pyx_v_self->_volatility_buffer_size == 0);
  if (!__pyx_t_9) {
  } else {
    __pyx_t_5 = __pyx_t_9;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_volatility_buffer_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_RichCompare(__pyx_t_1, __pyx_v_volatility_buffer_size, Py_NE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_5) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":412
 * 
 *         if self._volatility_buffer_size == 0 or self._volatility_buffer_size != volatility_buffer_size:
 *             self._volatility_buffer_size = volatility_buffer_size             # <<<<<<<<<<<<<<
 * 
 *             if self._avg_vol is None:
 */
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_v_volatility_buffer_size); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 412, __pyx_L1_error)
    __pyx_v_self->_volatility_buffer_size = __pyx_t_8;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":414
 *             self._volatility_buffer_size = volatility_buffer_size
 * 
 *             if self._avg_vol is None:             # <<<<<<<<<<<<<<
 *                 self._avg_vol = InstantVolatilityIndicator(sampling_length=volatility_buffer_size)
 *             else:
 */
    __pyx_t_5 = (__pyx_v_self->_avg_vol == Py_None);
    if (__pyx_t_5) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":415
 * 
 *             if self._avg_vol is None:
 *                 self._avg_vol = InstantVolatilityIndicator(sampling_length=volatility_buffer_size)             # <<<<<<<<<<<<<<
 *             else:
 *                 self._avg_vol.sampling_length = volatility_buffer_size
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_InstantVolatilityIndicator); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 415, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 415, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_sampling_length, __pyx_v_volatility_buffer_size) < 0) __PYX_ERR(0, 415, __pyx_L1_error)
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 415, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GIVEREF(__pyx_t_2);
      __Pyx_GOTREF(__pyx_v_self->_avg_vol);
      __Pyx_DECREF(__pyx_v_self->_avg_vol);
      __pyx_v_self->_avg_vol = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":414
 *             self._volatility_buffer_size = volatility_buffer_size
 * 
 *             if self._avg_vol is None:             # <<<<<<<<<<<<<<
 *                 self._avg_vol = InstantVolatilityIndicator(sampling_length=volatility_buffer_size)
 *             else:
 */
      goto __pyx_L6;
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":417
 *                 self._avg_vol = InstantVolatilityIndicator(sampling_length=volatility_buffer_size)
 *             else:
 *                 self._avg_vol.sampling_length = volatility_buffer_size             # <<<<<<<<<<<<<<
 * 
 *         if (
 */
    /*else*/ {
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->_avg_vol, __pyx_n_s_sampling_length, __pyx_v_volatility_buffer_size) < 0) __PYX_ERR(0, 417, __pyx_L1_error)
    }
    __pyx_L6:;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":411
 *         ticks_to_be_ready_before = max(self._volatility_buffer_size, self._trading_intensity_buffer_size)
 * 
 *         if self._volatility_buffer_size == 0 or self._volatility_buffer_size != volatility_buffer_size:             # <<<<<<<<<<<<<<
 *             self._volatility_buffer_size = volatility_buffer_size
 * 
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":420
 * 
 *         if (
 *             self._trading_intensity_buffer_size == 0             # <<<<<<<<<<<<<<
 *             or self._trading_intensity_buffer_size != trading_intensity_buffer_size
 *         ):
 */
  __pyx_t_9 = (__pyx_v_self->_trading_intensity_buffer_size == 0);
  if (!__pyx_t_9) {
  } else {
    __pyx_t_5 = __pyx_t_9;
    goto __pyx_L8_bool_binop_done;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":421
 *         if (
 *             self._trading_intensity_buffer_size == 0
 *             or self._trading_intensity_buffer_size != trading_intensity_buffer_size             # <<<<<<<<<<<<<<
 *         ):
 *             self._trading_intensity_buffer_size = trading_intensity_buffer_size
 */
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->_trading_intensity_buffer_size); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_v_trading_intensity_buffer_size, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __pyx_t_9;
  __pyx_L8_bool_binop_done:;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":419
 *                 self._avg_vol.sampling_length = volatility_buffer_size
 * 
 *         if (             # <<<<<<<<<<<<<<
 *             self._trading_intensity_buffer_size == 0
 *             or self._trading_intensity_buffer_size != trading_intensity_buffer_size
 */
  if (__pyx_t_5) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":423
 *             or self._trading_intensity_buffer_size != trading_intensity_buffer_size
 *         ):
 *             self._trading_intensity_buffer_size = trading_intensity_buffer_size             # <<<<<<<<<<<<<<
 *             if self._trading_intensity is not None:
 *                 self._trading_intensity.sampling_length = trading_intensity_buffer_size
 */
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_v_trading_intensity_buffer_size); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 423, __pyx_L1_error)
    __pyx_v_self->_trading_intensity_buffer_size = __pyx_t_8;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":424
 *         ):
 *             self._trading_intensity_buffer_size = trading_intensity_buffer_size
 *             if self._trading_intensity is not None:             # <<<<<<<<<<<<<<
 *                 self._trading_intensity.sampling_length = trading_intensity_buffer_size
 * 
 */
    __pyx_t_5 = (((PyObject *)__pyx_v_self->_trading_intensity) != Py_None);
    if (__pyx_t_5) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":425
 *             self._trading_intensity_buffer_size = trading_intensity_buffer_size
 *             if self._trading_intensity is not None:
 *                 self._trading_intensity.sampling_length = trading_intensity_buffer_size             # <<<<<<<<<<<<<<
 * 
 *         if self._trading_intensity is None and self.market_info.market.ready:
 */
      if (__Pyx_PyObject_SetAttrStr(((PyObject *)__pyx_v_self->_trading_intensity), __pyx_n_s_sampling_length, __pyx_v_trading_intensity_buffer_size) < 0) __PYX_ERR(0, 425, __pyx_L1_error)

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":424
 *         ):
 *             self._trading_intensity_buffer_size = trading_intensity_buffer_size
 *             if self._trading_intensity is not None:             # <<<<<<<<<<<<<<
 *                 self._trading_intensity.sampling_length = trading_intensity_buffer_size
 * 
 */
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":419
 *                 self._avg_vol.sampling_length = volatility_buffer_size
 * 
 *         if (             # <<<<<<<<<<<<<<
 *             self._trading_intensity_buffer_size == 0
 *             or self._trading_intensity_buffer_size != trading_intensity_buffer_size
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":427
 *                 self._trading_intensity.sampling_length = trading_intensity_buffer_size
 * 
 *         if self._trading_intensity is None and self.market_info.market.ready:             # <<<<<<<<<<<<<<
 *             self._trading_intensity = TradingIntensityIndicator(
 *                 order_book=self.market_info.order_book,
 */
  __pyx_t_9 = (((PyObject *)__pyx_v_self->_trading_intensity) == Py_None);
  if (__pyx_t_9) {
  } else {
    __pyx_t_5 = __pyx_t_9;
    goto __pyx_L12_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_market_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_market); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ready); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __pyx_t_9;
  __pyx_L12_bool_binop_done:;
  if (__pyx_t_5) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":429
 *         if self._trading_intensity is None and self.market_info.market.ready:
 *             self._trading_intensity = TradingIntensityIndicator(
 *                 order_book=self.market_info.order_book,             # <<<<<<<<<<<<<<
 *                 price_delegate=self._price_delegate,
 *                 sampling_length=self._trading_intensity_buffer_size,
 */
    __pyx_t_1 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_market_info); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_order_book); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_order_book, __pyx_t_3) < 0) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":430
 *             self._trading_intensity = TradingIntensityIndicator(
 *                 order_book=self.market_info.order_book,
 *                 price_delegate=self._price_delegate,             # <<<<<<<<<<<<<<
 *                 sampling_length=self._trading_intensity_buffer_size,
 *             )
 */
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_price_delegate, __pyx_v_self->_price_delegate) < 0) __PYX_ERR(0, 429, __pyx_L1_error)

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":431
 *                 order_book=self.market_info.order_book,
 *                 price_delegate=self._price_delegate,
 *                 sampling_length=self._trading_intensity_buffer_size,             # <<<<<<<<<<<<<<
 *             )
 * 
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->_trading_intensity_buffer_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_sampling_length, __pyx_t_3) < 0) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":428
 * 
 *         if self._trading_intensity is None and self.market_info.market.ready:
 *             self._trading_intensity = TradingIntensityIndicator(             # <<<<<<<<<<<<<<
 *                 order_book=self.market_info.order_book,
 *                 price_delegate=self._price_delegate,
 */
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator), __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GIVEREF(__pyx_t_3);
    __Pyx_GOTREF((PyObject *)__pyx_v_self->_trading_intensity);
    __Pyx_DECREF((PyObject *)__pyx_v_self->_trading_intensity);
    __pyx_v_self->_trading_intensity = ((struct __pyx_obj_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":427
 *                 self._trading_intensity.sampling_length = trading_intensity_buffer_size
 * 
 *         if self._trading_intensity is None and self.market_info.market.ready:             # <<<<<<<<<<<<<<
 *             self._trading_intensity = TradingIntensityIndicator(
 *                 order_book=self.market_info.order_book,
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":434
 *             )
 * 
 *         self._ticks_to_be_ready += (ticks_to_be_ready_after - ticks_to_be_ready_before)             # <<<<<<<<<<<<<<
 *         if self._ticks_to_be_ready < 0:
 *             self._ticks_to_be_ready = 0
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->_ticks_to_be_ready); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyNumber_Subtract(__pyx_v_ticks_to_be_ready_after, __pyx_v_ticks_to_be_ready_before); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 434, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_self->_ticks_to_be_ready = __pyx_t_8;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":435
 * 
 *         self._ticks_to_be_ready += (ticks_to_be_ready_after - ticks_to_be_ready_before)
 *         if self._ticks_to_be_ready < 0:             # <<<<<<<<<<<<<<
 *             self._ticks_to_be_ready = 0
 * 
 */
  __pyx_t_5 = (__pyx_v_self->_ticks_to_be_ready < 0);
  if (__pyx_t_5) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":436
 *         self._ticks_to_be_ready += (ticks_to_be_ready_after - ticks_to_be_ready_before)
 *         if self._ticks_to_be_ready < 0:
 *             self._ticks_to_be_ready = 0             # <<<<<<<<<<<<<<
 * 
 *     def pure_mm_assets_df(self, to_show_current_pct: bool) -> pd.DataFrame:
 */
    __pyx_v_self->_ticks_to_be_ready = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":435
 * 
 *         self._ticks_to_be_ready += (ticks_to_be_ready_after - ticks_to_be_ready_before)
 *         if self._ticks_to_be_ready < 0:             # <<<<<<<<<<<<<<
 *             self._ticks_to_be_ready = 0
 * 
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":405
 *             self._hanging_orders_tracker.hanging_orders_cancel_pct = hanging_orders_cancel_pct / Decimal('100')
 * 
 *     def get_config_map_indicators(self):             # <<<<<<<<<<<<<<
 *         volatility_buffer_size = self._config_map.volatility_buffer_size
 *         trading_intensity_buffer_size = self._config_map.trading_intensity_buffer_size
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.get_config_map_indicators", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_volatility_buffer_size);
  __Pyx_XDECREF(__pyx_v_trading_intensity_buffer_size);
  __Pyx_XDECREF(__pyx_v_ticks_to_be_ready_after);
  __Pyx_XDECREF(__pyx_v_ticks_to_be_ready_before);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":438
 *             self._ticks_to_be_ready = 0
 * 
 *     def pure_mm_assets_df(self, to_show_current_pct: bool) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self._price_delegate.get_price_by_type(PriceType.MidPrice)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_21pure_mm_assets_df(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_21pure_mm_assets_df = {"pure_mm_assets_df", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_21pure_mm_assets_df, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_21pure_mm_assets_df(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_to_show_current_pct = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pure_mm_assets_df (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_to_show_current_pct,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_to_show_current_pct)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 438, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "pure_mm_assets_df") < 0)) __PYX_ERR(0, 438, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_to_show_current_pct = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pure_mm_assets_df", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 438, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.pure_mm_assets_df", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_20pure_mm_assets_df(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_to_show_current_pct);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_20pure_mm_assets_df(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_to_show_current_pct) {
  PyObject *__pyx_v_market = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_trading_pair = NULL;
  PyObject *__pyx_v_base_asset = NULL;
  PyObject *__pyx_v_quote_asset = NULL;
  PyObject *__pyx_v_price = NULL;
  double __pyx_v_base_balance;
  double __pyx_v_quote_balance;
  double __pyx_v_available_base_balance;
  double __pyx_v_available_quote_balance;
  double __pyx_v_base_value;
  double __pyx_v_total_in_quote;
  double __pyx_v_base_ratio;
  double __pyx_v_quote_ratio;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_df = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  unsigned int __pyx_t_8;
  double __pyx_t_9;
  int __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  Py_UCS4 __pyx_t_12;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pure_mm_assets_df", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":439
 * 
 *     def pure_mm_assets_df(self, to_show_current_pct: bool) -> pd.DataFrame:
 *         market, trading_pair, base_asset, quote_asset = self._market_info             # <<<<<<<<<<<<<<
 *         price = self._price_delegate.get_price_by_type(PriceType.MidPrice)
 *         base_balance = float(market.get_balance(base_asset))
 */
  __pyx_t_1 = __pyx_v_self->_market_info;
  __Pyx_INCREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 4)) {
      if (size > 4) __Pyx_RaiseTooManyValuesError(4);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 439, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 3); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 3); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[4] = {&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5};
      for (i=0; i < 4; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 439, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[4] = {&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5};
    __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 439, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6);
    for (index=0; index < 4; index++) {
      PyObject* item = __pyx_t_7(__pyx_t_6); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 4) < 0) __PYX_ERR(0, 439, __pyx_L1_error)
    __pyx_t_7 = NULL;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 439, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_market = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_trading_pair = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_base_asset = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_quote_asset = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":440
 *     def pure_mm_assets_df(self, to_show_current_pct: bool) -> pd.DataFrame:
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self._price_delegate.get_price_by_type(PriceType.MidPrice)             # <<<<<<<<<<<<<<
 *         base_balance = float(market.get_balance(base_asset))
 *         quote_balance = float(market.get_balance(quote_asset))
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_price_delegate, __pyx_n_s_get_price_by_type); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_PriceType); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_MidPrice); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_v_price = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":441
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self._price_delegate.get_price_by_type(PriceType.MidPrice)
 *         base_balance = float(market.get_balance(base_asset))             # <<<<<<<<<<<<<<
 *         quote_balance = float(market.get_balance(quote_asset))
 *         available_base_balance = float(market.get_available_balance(base_asset))
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_market, __pyx_n_s_get_balance); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_base_asset};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_9 = __Pyx_PyObject_AsDouble(__pyx_t_1); if (unlikely(__pyx_t_9 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_base_balance = __pyx_t_9;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":442
 *         price = self._price_delegate.get_price_by_type(PriceType.MidPrice)
 *         base_balance = float(market.get_balance(base_asset))
 *         quote_balance = float(market.get_balance(quote_asset))             # <<<<<<<<<<<<<<
 *         available_base_balance = float(market.get_available_balance(base_asset))
 *         available_quote_balance = float(market.get_available_balance(quote_asset))
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_market, __pyx_n_s_get_balance); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_quote_asset};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 442, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_9 = __Pyx_PyObject_AsDouble(__pyx_t_1); if (unlikely(__pyx_t_9 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_quote_balance = __pyx_t_9;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":443
 *         base_balance = float(market.get_balance(base_asset))
 *         quote_balance = float(market.get_balance(quote_asset))
 *         available_base_balance = float(market.get_available_balance(base_asset))             # <<<<<<<<<<<<<<
 *         available_quote_balance = float(market.get_available_balance(quote_asset))
 *         base_value = base_balance * float(price)
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_market, __pyx_n_s_get_available_balance); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 443, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_base_asset};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 443, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_9 = __Pyx_PyObject_AsDouble(__pyx_t_1); if (unlikely(__pyx_t_9 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 443, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_available_base_balance = __pyx_t_9;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":444
 *         quote_balance = float(market.get_balance(quote_asset))
 *         available_base_balance = float(market.get_available_balance(base_asset))
 *         available_quote_balance = float(market.get_available_balance(quote_asset))             # <<<<<<<<<<<<<<
 *         base_value = base_balance * float(price)
 *         total_in_quote = base_value + quote_balance
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_market, __pyx_n_s_get_available_balance); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_quote_asset};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 444, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_9 = __Pyx_PyObject_AsDouble(__pyx_t_1); if (unlikely(__pyx_t_9 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_available_quote_balance = __pyx_t_9;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":445
 *         available_base_balance = float(market.get_available_balance(base_asset))
 *         available_quote_balance = float(market.get_available_balance(quote_asset))
 *         base_value = base_balance * float(price)             # <<<<<<<<<<<<<<
 *         total_in_quote = base_value + quote_balance
 *         base_ratio = base_value / total_in_quote if total_in_quote > 0 else 0
 */
  __pyx_t_9 = __Pyx_PyObject_AsDouble(__pyx_v_price); if (unlikely(__pyx_t_9 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 445, __pyx_L1_error)
  __pyx_v_base_value = (__pyx_v_base_balance * __pyx_t_9);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":446
 *         available_quote_balance = float(market.get_available_balance(quote_asset))
 *         base_value = base_balance * float(price)
 *         total_in_quote = base_value + quote_balance             # <<<<<<<<<<<<<<
 *         base_ratio = base_value / total_in_quote if total_in_quote > 0 else 0
 *         quote_ratio = quote_balance / total_in_quote if total_in_quote > 0 else 0
 */
  __pyx_v_total_in_quote = (__pyx_v_base_value + __pyx_v_quote_balance);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":447
 *         base_value = base_balance * float(price)
 *         total_in_quote = base_value + quote_balance
 *         base_ratio = base_value / total_in_quote if total_in_quote > 0 else 0             # <<<<<<<<<<<<<<
 *         quote_ratio = quote_balance / total_in_quote if total_in_quote > 0 else 0
 *         data = [
 */
  __pyx_t_10 = (__pyx_v_total_in_quote > 0.0);
  if (__pyx_t_10) {
    if (unlikely(__pyx_v_total_in_quote == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 447, __pyx_L1_error)
    }
    __pyx_t_9 = (__pyx_v_base_value / __pyx_v_total_in_quote);
  } else {
    __pyx_t_9 = 0.0;
  }
  __pyx_v_base_ratio = __pyx_t_9;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":448
 *         total_in_quote = base_value + quote_balance
 *         base_ratio = base_value / total_in_quote if total_in_quote > 0 else 0
 *         quote_ratio = quote_balance / total_in_quote if total_in_quote > 0 else 0             # <<<<<<<<<<<<<<
 *         data = [
 *             ["", base_asset, quote_asset],
 */
  __pyx_t_10 = (__pyx_v_total_in_quote > 0.0);
  if (__pyx_t_10) {
    if (unlikely(__pyx_v_total_in_quote == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "float division");
      __PYX_ERR(0, 448, __pyx_L1_error)
    }
    __pyx_t_9 = (__pyx_v_quote_balance / __pyx_v_total_in_quote);
  } else {
    __pyx_t_9 = 0.0;
  }
  __pyx_v_quote_ratio = __pyx_t_9;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":450
 *         quote_ratio = quote_balance / total_in_quote if total_in_quote > 0 else 0
 *         data = [
 *             ["", base_asset, quote_asset],             # <<<<<<<<<<<<<<
 *             ["Total Balance", round(base_balance, 4), round(quote_balance, 4)],
 *             ["Available Balance", round(available_base_balance, 4), round(available_quote_balance, 4)],
 */
  __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_kp_u__4);
  __Pyx_GIVEREF(__pyx_kp_u__4);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u__4)) __PYX_ERR(0, 450, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_base_asset);
  __Pyx_GIVEREF(__pyx_v_base_asset);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, __pyx_v_base_asset)) __PYX_ERR(0, 450, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_quote_asset);
  __Pyx_GIVEREF(__pyx_v_quote_asset);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 2, __pyx_v_quote_asset)) __PYX_ERR(0, 450, __pyx_L1_error);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":451
 *         data = [
 *             ["", base_asset, quote_asset],
 *             ["Total Balance", round(base_balance, 4), round(quote_balance, 4)],             # <<<<<<<<<<<<<<
 *             ["Available Balance", round(available_base_balance, 4), round(available_quote_balance, 4)],
 *             [f"Current Value ({quote_asset})", round(base_value, 4), round(quote_balance, 4)]
 */
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_base_balance); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_5)) __PYX_ERR(0, 451, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_4)) __PYX_ERR(0, 451, __pyx_L1_error);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_3, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_quote_balance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_int_4)) __PYX_ERR(0, 451, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyList_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_kp_u_Total_Balance);
  __Pyx_GIVEREF(__pyx_kp_u_Total_Balance);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u_Total_Balance)) __PYX_ERR(0, 451, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_t_5)) __PYX_ERR(0, 451, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 2, __pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error);
  __pyx_t_5 = 0;
  __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":452
 *             ["", base_asset, quote_asset],
 *             ["Total Balance", round(base_balance, 4), round(quote_balance, 4)],
 *             ["Available Balance", round(available_base_balance, 4), round(available_quote_balance, 4)],             # <<<<<<<<<<<<<<
 *             [f"Current Value ({quote_asset})", round(base_value, 4), round(quote_balance, 4)]
 *         ]
 */
  __pyx_t_3 = PyFloat_FromDouble(__pyx_v_available_base_balance); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_3)) __PYX_ERR(0, 452, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_int_4)) __PYX_ERR(0, 452, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_available_quote_balance); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5)) __PYX_ERR(0, 452, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_4)) __PYX_ERR(0, 452, __pyx_L1_error);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyList_New(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 452, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_kp_u_Available_Balance);
  __Pyx_GIVEREF(__pyx_kp_u_Available_Balance);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_Available_Balance)) __PYX_ERR(0, 452, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 1, __pyx_t_3)) __PYX_ERR(0, 452, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 2, __pyx_t_5)) __PYX_ERR(0, 452, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":453
 *             ["Total Balance", round(base_balance, 4), round(quote_balance, 4)],
 *             ["Available Balance", round(available_base_balance, 4), round(available_quote_balance, 4)],
 *             [f"Current Value ({quote_asset})", round(base_value, 4), round(quote_balance, 4)]             # <<<<<<<<<<<<<<
 *         ]
 *         if to_show_current_pct:
 */
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_11 = 0;
  __pyx_t_12 = 127;
  __Pyx_INCREF(__pyx_kp_u_Current_Value);
  __pyx_t_11 += 15;
  __Pyx_GIVEREF(__pyx_kp_u_Current_Value);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_kp_u_Current_Value);
  __pyx_t_3 = __Pyx_PyObject_FormatSimple(__pyx_v_quote_asset, __pyx_empty_unicode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_12 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_12) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_12;
  __pyx_t_11 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_3);
  __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_kp_u__5);
  __pyx_t_11 += 1;
  __Pyx_GIVEREF(__pyx_kp_u__5);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_kp_u__5);
  __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_5, 3, __pyx_t_11, __pyx_t_12); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = PyFloat_FromDouble(__pyx_v_base_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = PyTuple_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5)) __PYX_ERR(0, 453, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_int_4)) __PYX_ERR(0, 453, __pyx_L1_error);
  __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyFloat_FromDouble(__pyx_v_quote_balance); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_13 = PyTuple_New(2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_6)) __PYX_ERR(0, 453, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_4);
  __Pyx_GIVEREF(__pyx_int_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_int_4)) __PYX_ERR(0, 453, __pyx_L1_error);
  __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_13, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
  __pyx_t_13 = PyList_New(3); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_13, 0, __pyx_t_3)) __PYX_ERR(0, 453, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_5);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_13, 1, __pyx_t_5)) __PYX_ERR(0, 453, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_6);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_13, 2, __pyx_t_6)) __PYX_ERR(0, 453, __pyx_L1_error);
  __pyx_t_3 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":449
 *         base_ratio = base_value / total_in_quote if total_in_quote > 0 else 0
 *         quote_ratio = quote_balance / total_in_quote if total_in_quote > 0 else 0
 *         data = [             # <<<<<<<<<<<<<<
 *             ["", base_asset, quote_asset],
 *             ["Total Balance", round(base_balance, 4), round(quote_balance, 4)],
 */
  __pyx_t_6 = PyList_New(4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 0, __pyx_t_1)) __PYX_ERR(0, 449, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 1, __pyx_t_4)) __PYX_ERR(0, 449, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 2, __pyx_t_2)) __PYX_ERR(0, 449, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_13);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 3, __pyx_t_13)) __PYX_ERR(0, 449, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_4 = 0;
  __pyx_t_2 = 0;
  __pyx_t_13 = 0;
  __pyx_v_data = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":455
 *             [f"Current Value ({quote_asset})", round(base_value, 4), round(quote_balance, 4)]
 *         ]
 *         if to_show_current_pct:             # <<<<<<<<<<<<<<
 *             data.append(["Current %", f"{base_ratio:.1%}", f"{quote_ratio:.1%}"])
 *         df = pd.DataFrame(data=data)
 */
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_to_show_current_pct); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 455, __pyx_L1_error)
  if (__pyx_t_10) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":456
 *         ]
 *         if to_show_current_pct:
 *             data.append(["Current %", f"{base_ratio:.1%}", f"{quote_ratio:.1%}"])             # <<<<<<<<<<<<<<
 *         df = pd.DataFrame(data=data)
 *         return df
 */
    __pyx_t_6 = PyFloat_FromDouble(__pyx_v_base_ratio); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 456, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_13 = __Pyx_PyObject_Format(__pyx_t_6, __pyx_kp_u_1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 456, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyFloat_FromDouble(__pyx_v_quote_ratio); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 456, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = __Pyx_PyObject_Format(__pyx_t_6, __pyx_kp_u_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 456, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyList_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 456, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_kp_u_Current);
    __Pyx_GIVEREF(__pyx_kp_u_Current);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 0, __pyx_kp_u_Current)) __PYX_ERR(0, 456, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_13);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 1, __pyx_t_13)) __PYX_ERR(0, 456, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_6, 2, __pyx_t_2)) __PYX_ERR(0, 456, __pyx_L1_error);
    __pyx_t_13 = 0;
    __pyx_t_2 = 0;
    __pyx_t_14 = __Pyx_PyList_Append(__pyx_v_data, __pyx_t_6); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 456, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":455
 *             [f"Current Value ({quote_asset})", round(base_value, 4), round(quote_balance, 4)]
 *         ]
 *         if to_show_current_pct:             # <<<<<<<<<<<<<<
 *             data.append(["Current %", f"{base_ratio:.1%}", f"{quote_ratio:.1%}"])
 *         df = pd.DataFrame(data=data)
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":457
 *         if to_show_current_pct:
 *             data.append(["Current %", f"{base_ratio:.1%}", f"{quote_ratio:.1%}"])
 *         df = pd.DataFrame(data=data)             # <<<<<<<<<<<<<<
 *         return df
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_pd); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_DataFrame); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_data, __pyx_v_data) < 0) __PYX_ERR(0, 457, __pyx_L1_error)
  __pyx_t_13 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_6); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_v_df = __pyx_t_13;
  __pyx_t_13 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":458
 *             data.append(["Current %", f"{base_ratio:.1%}", f"{quote_ratio:.1%}"])
 *         df = pd.DataFrame(data=data)
 *         return df             # <<<<<<<<<<<<<<
 * 
 *     def active_orders_df(self) -> pd.DataFrame:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_df);
  __pyx_r = __pyx_v_df;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":438
 *             self._ticks_to_be_ready = 0
 * 
 *     def pure_mm_assets_df(self, to_show_current_pct: bool) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self._price_delegate.get_price_by_type(PriceType.MidPrice)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.pure_mm_assets_df", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_trading_pair);
  __Pyx_XDECREF(__pyx_v_base_asset);
  __Pyx_XDECREF(__pyx_v_quote_asset);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_df);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":460
 *         return df
 * 
 *     def active_orders_df(self) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self.get_price()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_23active_orders_df(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_23active_orders_df = {"active_orders_df", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_23active_orders_df, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_23active_orders_df(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("active_orders_df (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("active_orders_df", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "active_orders_df", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_22active_orders_df(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":466
 *         no_sells = len([o for o in active_orders if not o.is_buy and o.client_order_id and
 *                         not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)])
 *         active_orders.sort(key=lambda x: x.price, reverse=True)             # <<<<<<<<<<<<<<
 *         columns = ["Level", "Type", "Price", "Spread", "Amount (Orig)", "Amount (Adj)", "Age"]
 *         data = []
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_16active_orders_df_lambda(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_16active_orders_df_lambda = {"lambda", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_16active_orders_df_lambda, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_16active_orders_df_lambda(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_x = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_x,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_x)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 466, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "lambda") < 0)) __PYX_ERR(0, 466, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_x = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lambda", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 466, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.active_orders_df.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_lambda_funcdef_lambda(__pyx_self, __pyx_v_x);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_x) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_x, __pyx_n_s_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.active_orders_df.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":460
 *         return df
 * 
 *     def active_orders_df(self) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self.get_price()
 */

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_22active_orders_df(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  CYTHON_UNUSED PyObject *__pyx_v_market = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_trading_pair = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_base_asset = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_quote_asset = NULL;
  PyObject *__pyx_v_price = NULL;
  PyObject *__pyx_v_active_orders = NULL;
  PyObject *__pyx_v_no_sells = NULL;
  PyObject *__pyx_v_columns = NULL;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_lvl_buy = NULL;
  PyObject *__pyx_v_lvl_sell = NULL;
  Py_ssize_t __pyx_v_idx;
  PyObject *__pyx_v_order = NULL;
  PyObject *__pyx_v_is_hanging_order = NULL;
  PyObject *__pyx_v_level = NULL;
  PyObject *__pyx_v_spread = NULL;
  PyObject *__pyx_v_age = NULL;
  PyObject *__pyx_v_amount_orig = NULL;
  PyObject *__pyx_8genexpr4__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  unsigned int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *(*__pyx_t_10)(PyObject *);
  int __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  Py_ssize_t __pyx_t_15;
  int __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("active_orders_df", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":461
 * 
 *     def active_orders_df(self) -> pd.DataFrame:
 *         market, trading_pair, base_asset, quote_asset = self._market_info             # <<<<<<<<<<<<<<
 *         price = self.get_price()
 *         active_orders = self.active_orders
 */
  __pyx_t_1 = __pyx_v_self->_market_info;
  __Pyx_INCREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 4)) {
      if (size > 4) __Pyx_RaiseTooManyValuesError(4);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 461, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 3); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 3); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[4] = {&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5};
      for (i=0; i < 4; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 461, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[4] = {&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5};
    __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 461, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6);
    for (index=0; index < 4; index++) {
      PyObject* item = __pyx_t_7(__pyx_t_6); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 4) < 0) __PYX_ERR(0, 461, __pyx_L1_error)
    __pyx_t_7 = NULL;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 461, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_market = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_trading_pair = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_base_asset = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_quote_asset = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":462
 *     def active_orders_df(self) -> pd.DataFrame:
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self.get_price()             # <<<<<<<<<<<<<<
 *         active_orders = self.active_orders
 *         no_sells = len([o for o in active_orders if not o.is_buy and o.client_order_id and
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_price); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 462, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_v_price = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":463
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self.get_price()
 *         active_orders = self.active_orders             # <<<<<<<<<<<<<<
 *         no_sells = len([o for o in active_orders if not o.is_buy and o.client_order_id and
 *                         not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)])
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_active_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 463, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_active_orders = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":464
 *         price = self.get_price()
 *         active_orders = self.active_orders
 *         no_sells = len([o for o in active_orders if not o.is_buy and o.client_order_id and             # <<<<<<<<<<<<<<
 *                         not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)])
 *         active_orders.sort(key=lambda x: x.price, reverse=True)
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 464, __pyx_L7_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (likely(PyList_CheckExact(__pyx_v_active_orders)) || PyTuple_CheckExact(__pyx_v_active_orders)) {
      __pyx_t_5 = __pyx_v_active_orders; __Pyx_INCREF(__pyx_t_5);
      __pyx_t_9 = 0;
      __pyx_t_10 = NULL;
    } else {
      __pyx_t_9 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_v_active_orders); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 464, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 464, __pyx_L7_error)
    }
    for (;;) {
      if (likely(!__pyx_t_10)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 464, __pyx_L7_error)
            #endif
            if (__pyx_t_9 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_9); __Pyx_INCREF(__pyx_t_4); __pyx_t_9++; if (unlikely((0 < 0))) __PYX_ERR(0, 464, __pyx_L7_error)
          #else
          __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 464, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 464, __pyx_L7_error)
            #endif
            if (__pyx_t_9 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_9); __Pyx_INCREF(__pyx_t_4); __pyx_t_9++; if (unlikely((0 < 0))) __PYX_ERR(0, 464, __pyx_L7_error)
          #else
          __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_9); __pyx_t_9++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 464, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_10(__pyx_t_5);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 464, __pyx_L7_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr4__pyx_v_o, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr4__pyx_v_o, __pyx_n_s_is_buy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 464, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 464, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_13 = (!__pyx_t_12);
      if (__pyx_t_13) {
      } else {
        __pyx_t_11 = __pyx_t_13;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr4__pyx_v_o, __pyx_n_s_client_order_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 464, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 464, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_13) {
      } else {
        __pyx_t_11 = __pyx_t_13;
        goto __pyx_L11_bool_binop_done;
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":465
 *         active_orders = self.active_orders
 *         no_sells = len([o for o in active_orders if not o.is_buy and o.client_order_id and
 *                         not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)])             # <<<<<<<<<<<<<<
 *         active_orders.sort(key=lambda x: x.price, reverse=True)
 *         columns = ["Level", "Type", "Price", "Spread", "Amount (Orig)", "Amount (Adj)", "Age"]
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_n_s_is_order_id_in_hanging_orders); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 465, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr4__pyx_v_o, __pyx_n_s_client_order_id); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 465, __pyx_L7_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_6)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_6);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_2};
        __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 465, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 465, __pyx_L7_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_12 = (!__pyx_t_13);
      __pyx_t_11 = __pyx_t_12;
      __pyx_L11_bool_binop_done:;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":464
 *         price = self.get_price()
 *         active_orders = self.active_orders
 *         no_sells = len([o for o in active_orders if not o.is_buy and o.client_order_id and             # <<<<<<<<<<<<<<
 *                         not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)])
 *         active_orders.sort(key=lambda x: x.price, reverse=True)
 */
      if (__pyx_t_11) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_8genexpr4__pyx_v_o))) __PYX_ERR(0, 464, __pyx_L7_error)
      }
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_o); __pyx_8genexpr4__pyx_v_o = 0;
    goto __pyx_L15_exit_scope;
    __pyx_L7_error:;
    __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_o); __pyx_8genexpr4__pyx_v_o = 0;
    goto __pyx_L1_error;
    __pyx_L15_exit_scope:;
  } /* exit inner scope */
  __pyx_t_9 = __Pyx_PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 464, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_no_sells = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":466
 *         no_sells = len([o for o in active_orders if not o.is_buy and o.client_order_id and
 *                         not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)])
 *         active_orders.sort(key=lambda x: x.price, reverse=True)             # <<<<<<<<<<<<<<
 *         columns = ["Level", "Type", "Price", "Spread", "Amount (Orig)", "Amount (Adj)", "Age"]
 *         data = []
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_active_orders, __pyx_n_s_sort); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_16active_orders_df_lambda, 0, __pyx_n_s_active_orders_df_locals_lambda, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_key, __pyx_t_4) < 0) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_reverse, Py_True) < 0) __PYX_ERR(0, 466, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":467
 *                         not self._hanging_orders_tracker.is_order_id_in_hanging_orders(o.client_order_id)])
 *         active_orders.sort(key=lambda x: x.price, reverse=True)
 *         columns = ["Level", "Type", "Price", "Spread", "Amount (Orig)", "Amount (Adj)", "Age"]             # <<<<<<<<<<<<<<
 *         data = []
 *         lvl_buy, lvl_sell = 0, 0
 */
  __pyx_t_4 = PyList_New(7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 467, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_n_u_Level);
  __Pyx_GIVEREF(__pyx_n_u_Level);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_n_u_Level)) __PYX_ERR(0, 467, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_Type);
  __Pyx_GIVEREF(__pyx_n_u_Type);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_n_u_Type)) __PYX_ERR(0, 467, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_Price);
  __Pyx_GIVEREF(__pyx_n_u_Price);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 2, __pyx_n_u_Price)) __PYX_ERR(0, 467, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_Spread);
  __Pyx_GIVEREF(__pyx_n_u_Spread);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 3, __pyx_n_u_Spread)) __PYX_ERR(0, 467, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_Amount_Orig);
  __Pyx_GIVEREF(__pyx_kp_u_Amount_Orig);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 4, __pyx_kp_u_Amount_Orig)) __PYX_ERR(0, 467, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_Amount_Adj);
  __Pyx_GIVEREF(__pyx_kp_u_Amount_Adj);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 5, __pyx_kp_u_Amount_Adj)) __PYX_ERR(0, 467, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_Age);
  __Pyx_GIVEREF(__pyx_n_u_Age);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 6, __pyx_n_u_Age)) __PYX_ERR(0, 467, __pyx_L1_error);
  __pyx_v_columns = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":468
 *         active_orders.sort(key=lambda x: x.price, reverse=True)
 *         columns = ["Level", "Type", "Price", "Spread", "Amount (Orig)", "Amount (Adj)", "Age"]
 *         data = []             # <<<<<<<<<<<<<<
 *         lvl_buy, lvl_sell = 0, 0
 *         for idx in range(0, len(active_orders)):
 */
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 468, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_data = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":469
 *         columns = ["Level", "Type", "Price", "Spread", "Amount (Orig)", "Amount (Adj)", "Age"]
 *         data = []
 *         lvl_buy, lvl_sell = 0, 0             # <<<<<<<<<<<<<<
 *         for idx in range(0, len(active_orders)):
 *             order = active_orders[idx]
 */
  __pyx_t_4 = __pyx_int_0;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = __pyx_int_0;
  __Pyx_INCREF(__pyx_t_5);
  __pyx_v_lvl_buy = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_lvl_sell = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":470
 *         data = []
 *         lvl_buy, lvl_sell = 0, 0
 *         for idx in range(0, len(active_orders)):             # <<<<<<<<<<<<<<
 *             order = active_orders[idx]
 *             is_hanging_order = self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)
 */
  __pyx_t_9 = PyObject_Length(__pyx_v_active_orders); if (unlikely(__pyx_t_9 == ((Py_ssize_t)-1))) __PYX_ERR(0, 470, __pyx_L1_error)
  __pyx_t_14 = __pyx_t_9;
  for (__pyx_t_15 = 0; __pyx_t_15 < __pyx_t_14; __pyx_t_15+=1) {
    __pyx_v_idx = __pyx_t_15;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":471
 *         lvl_buy, lvl_sell = 0, 0
 *         for idx in range(0, len(active_orders)):
 *             order = active_orders[idx]             # <<<<<<<<<<<<<<
 *             is_hanging_order = self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)
 *             if not is_hanging_order:
 */
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_active_orders, __pyx_v_idx, Py_ssize_t, 1, PyInt_FromSsize_t, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 471, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_order, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":472
 *         for idx in range(0, len(active_orders)):
 *             order = active_orders[idx]
 *             is_hanging_order = self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)             # <<<<<<<<<<<<<<
 *             if not is_hanging_order:
 *                 if order.is_buy:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_n_s_is_order_id_in_hanging_orders); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_n_s_client_order_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_1};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 472, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_is_hanging_order, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":473
 *             order = active_orders[idx]
 *             is_hanging_order = self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)
 *             if not is_hanging_order:             # <<<<<<<<<<<<<<
 *                 if order.is_buy:
 *                     level = lvl_buy + 1
 */
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_is_hanging_order); if (unlikely((__pyx_t_11 < 0))) __PYX_ERR(0, 473, __pyx_L1_error)
    __pyx_t_12 = (!__pyx_t_11);
    if (__pyx_t_12) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":474
 *             is_hanging_order = self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)
 *             if not is_hanging_order:
 *                 if order.is_buy:             # <<<<<<<<<<<<<<
 *                     level = lvl_buy + 1
 *                     lvl_buy += 1
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_n_s_is_buy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 474, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 474, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_12) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":475
 *             if not is_hanging_order:
 *                 if order.is_buy:
 *                     level = lvl_buy + 1             # <<<<<<<<<<<<<<
 *                     lvl_buy += 1
 *                 else:
 */
        __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_v_lvl_buy, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 475, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_XDECREF_SET(__pyx_v_level, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":476
 *                 if order.is_buy:
 *                     level = lvl_buy + 1
 *                     lvl_buy += 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     level = no_sells - lvl_sell
 */
        __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_v_lvl_buy, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 476, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF_SET(__pyx_v_lvl_buy, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":474
 *             is_hanging_order = self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)
 *             if not is_hanging_order:
 *                 if order.is_buy:             # <<<<<<<<<<<<<<
 *                     level = lvl_buy + 1
 *                     lvl_buy += 1
 */
        goto __pyx_L19;
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":478
 *                     lvl_buy += 1
 *                 else:
 *                     level = no_sells - lvl_sell             # <<<<<<<<<<<<<<
 *                     lvl_sell += 1
 *             spread = 0 if price == 0 else abs(order.price - price) / price
 */
      /*else*/ {
        __pyx_t_5 = PyNumber_Subtract(__pyx_v_no_sells, __pyx_v_lvl_sell); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 478, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_XDECREF_SET(__pyx_v_level, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":479
 *                 else:
 *                     level = no_sells - lvl_sell
 *                     lvl_sell += 1             # <<<<<<<<<<<<<<
 *             spread = 0 if price == 0 else abs(order.price - price) / price
 *             age = pd.Timestamp(order_age(order, self._current_timestamp), unit='s').strftime('%H:%M:%S')
 */
        __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_v_lvl_sell, __pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 479, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF_SET(__pyx_v_lvl_sell, __pyx_t_5);
        __pyx_t_5 = 0;
      }
      __pyx_L19:;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":473
 *             order = active_orders[idx]
 *             is_hanging_order = self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)
 *             if not is_hanging_order:             # <<<<<<<<<<<<<<
 *                 if order.is_buy:
 *                     level = lvl_buy + 1
 */
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":480
 *                     level = no_sells - lvl_sell
 *                     lvl_sell += 1
 *             spread = 0 if price == 0 else abs(order.price - price) / price             # <<<<<<<<<<<<<<
 *             age = pd.Timestamp(order_age(order, self._current_timestamp), unit='s').strftime('%H:%M:%S')
 * 
 */
    __pyx_t_12 = (__Pyx_PyInt_BoolEqObjC(__pyx_v_price, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 480, __pyx_L1_error)
    if (__pyx_t_12) {
      __Pyx_INCREF(__pyx_int_0);
      __pyx_t_5 = __pyx_int_0;
    } else {
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_n_s_price); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 480, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = PyNumber_Subtract(__pyx_t_4, __pyx_v_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 480, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyNumber_Absolute(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 480, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_4, __pyx_v_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 480, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_5 = __pyx_t_1;
      __pyx_t_1 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_spread, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":481
 *                     lvl_sell += 1
 *             spread = 0 if price == 0 else abs(order.price - price) / price
 *             age = pd.Timestamp(order_age(order, self._current_timestamp), unit='s').strftime('%H:%M:%S')             # <<<<<<<<<<<<<<
 * 
 *             amount_orig = self._config_map.order_amount
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_pd); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_Timestamp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_order_age); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_order, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_8, 2+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 481, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1)) __PYX_ERR(0, 481, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_unit, __pyx_n_u_s) < 0) __PYX_ERR(0, 481, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_strftime); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 481, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_kp_u_H_M_S};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 481, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_age, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":483
 *             age = pd.Timestamp(order_age(order, self._current_timestamp), unit='s').strftime('%H:%M:%S')
 * 
 *             amount_orig = self._config_map.order_amount             # <<<<<<<<<<<<<<
 *             if is_hanging_order:
 *                 amount_orig = order.quantity
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_order_amount); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 483, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v_amount_orig, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":484
 * 
 *             amount_orig = self._config_map.order_amount
 *             if is_hanging_order:             # <<<<<<<<<<<<<<
 *                 amount_orig = order.quantity
 *                 level = "hang"
 */
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_v_is_hanging_order); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 484, __pyx_L1_error)
    if (__pyx_t_12) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":485
 *             amount_orig = self._config_map.order_amount
 *             if is_hanging_order:
 *                 amount_orig = order.quantity             # <<<<<<<<<<<<<<
 *                 level = "hang"
 *             data.append([
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_n_s_quantity); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 485, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF_SET(__pyx_v_amount_orig, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":486
 *             if is_hanging_order:
 *                 amount_orig = order.quantity
 *                 level = "hang"             # <<<<<<<<<<<<<<
 *             data.append([
 *                 level,
 */
      __Pyx_INCREF(__pyx_n_u_hang);
      __Pyx_XDECREF_SET(__pyx_v_level, __pyx_n_u_hang);

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":484
 * 
 *             amount_orig = self._config_map.order_amount
 *             if is_hanging_order:             # <<<<<<<<<<<<<<
 *                 amount_orig = order.quantity
 *                 level = "hang"
 */
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":488
 *                 level = "hang"
 *             data.append([
 *                 level,             # <<<<<<<<<<<<<<
 *                 "buy" if order.is_buy else "sell",
 *                 float(order.price),
 */
    if (unlikely(!__pyx_v_level)) { __Pyx_RaiseUnboundLocalError("level"); __PYX_ERR(0, 488, __pyx_L1_error) }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":489
 *             data.append([
 *                 level,
 *                 "buy" if order.is_buy else "sell",             # <<<<<<<<<<<<<<
 *                 float(order.price),
 *                 f"{spread:.2%}",
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_n_s_is_buy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 489, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_12) {
      __Pyx_INCREF(__pyx_n_u_buy);
      __pyx_t_5 = __pyx_n_u_buy;
    } else {
      __Pyx_INCREF(__pyx_n_u_sell);
      __pyx_t_5 = __pyx_n_u_sell;
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":490
 *                 level,
 *                 "buy" if order.is_buy else "sell",
 *                 float(order.price),             # <<<<<<<<<<<<<<
 *                 f"{spread:.2%}",
 *                 float(amount_orig),
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_n_s_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyNumber_Float(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":491
 *                 "buy" if order.is_buy else "sell",
 *                 float(order.price),
 *                 f"{spread:.2%}",             # <<<<<<<<<<<<<<
 *                 float(amount_orig),
 *                 float(order.quantity),
 */
    __pyx_t_1 = __Pyx_PyObject_Format(__pyx_v_spread, __pyx_kp_u_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 491, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":492
 *                 float(order.price),
 *                 f"{spread:.2%}",
 *                 float(amount_orig),             # <<<<<<<<<<<<<<
 *                 float(order.quantity),
 *                 age
 */
    __pyx_t_3 = __Pyx_PyNumber_Float(__pyx_v_amount_orig); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 492, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":493
 *                 f"{spread:.2%}",
 *                 float(amount_orig),
 *                 float(order.quantity),             # <<<<<<<<<<<<<<
 *                 age
 *             ])
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_n_s_quantity); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 493, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyNumber_Float(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 493, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":487
 *                 amount_orig = order.quantity
 *                 level = "hang"
 *             data.append([             # <<<<<<<<<<<<<<
 *                 level,
 *                 "buy" if order.is_buy else "sell",
 */
    __pyx_t_4 = PyList_New(7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_level);
    __Pyx_GIVEREF(__pyx_v_level);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_level)) __PYX_ERR(0, 487, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_5);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_t_5)) __PYX_ERR(0, 487, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 2, __pyx_t_2)) __PYX_ERR(0, 487, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 3, __pyx_t_1)) __PYX_ERR(0, 487, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 4, __pyx_t_3)) __PYX_ERR(0, 487, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 5, __pyx_t_6)) __PYX_ERR(0, 487, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_age);
    __Pyx_GIVEREF(__pyx_v_age);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 6, __pyx_v_age)) __PYX_ERR(0, 487, __pyx_L1_error);
    __pyx_t_5 = 0;
    __pyx_t_2 = 0;
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_t_6 = 0;
    __pyx_t_16 = __Pyx_PyList_Append(__pyx_v_data, __pyx_t_4); if (unlikely(__pyx_t_16 == ((int)-1))) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":497
 *             ])
 * 
 *         return pd.DataFrame(data=data, columns=columns)             # <<<<<<<<<<<<<<
 * 
 *     def market_status_data_frame(self, market_trading_pair_tuples: List[MarketTradingPairTuple]) -> pd.DataFrame:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pd); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_DataFrame); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_data, __pyx_v_data) < 0) __PYX_ERR(0, 497, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_columns, __pyx_v_columns) < 0) __PYX_ERR(0, 497, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_empty_tuple, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 497, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":460
 *         return df
 * 
 *     def active_orders_df(self) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self.get_price()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.active_orders_df", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_trading_pair);
  __Pyx_XDECREF(__pyx_v_base_asset);
  __Pyx_XDECREF(__pyx_v_quote_asset);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XDECREF(__pyx_v_active_orders);
  __Pyx_XDECREF(__pyx_v_no_sells);
  __Pyx_XDECREF(__pyx_v_columns);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_lvl_buy);
  __Pyx_XDECREF(__pyx_v_lvl_sell);
  __Pyx_XDECREF(__pyx_v_order);
  __Pyx_XDECREF(__pyx_v_is_hanging_order);
  __Pyx_XDECREF(__pyx_v_level);
  __Pyx_XDECREF(__pyx_v_spread);
  __Pyx_XDECREF(__pyx_v_age);
  __Pyx_XDECREF(__pyx_v_amount_orig);
  __Pyx_XDECREF(__pyx_8genexpr4__pyx_v_o);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":499
 *         return pd.DataFrame(data=data, columns=columns)
 * 
 *     def market_status_data_frame(self, market_trading_pair_tuples: List[MarketTradingPairTuple]) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         markets_data = []
 *         markets_columns = ["Exchange", "Market", "Best Bid", "Best Ask", f"MidPrice"]
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25market_status_data_frame(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25market_status_data_frame = {"market_status_data_frame", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25market_status_data_frame, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25market_status_data_frame(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v_market_trading_pair_tuples = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("market_status_data_frame (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_market_trading_pair_tuples,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_market_trading_pair_tuples)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 499, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "market_status_data_frame") < 0)) __PYX_ERR(0, 499, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_market_trading_pair_tuples = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("market_status_data_frame", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 499, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.market_status_data_frame", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_24market_status_data_frame(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_market_trading_pair_tuples);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_24market_status_data_frame(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_market_trading_pair_tuples) {
  PyObject *__pyx_v_markets_data = NULL;
  PyObject *__pyx_v_markets_columns = NULL;
  PyObject *__pyx_v_market_books = NULL;
  PyObject *__pyx_v_market = NULL;
  PyObject *__pyx_v_trading_pair = NULL;
  PyObject *__pyx_v_bid_price = NULL;
  PyObject *__pyx_v_ask_price = NULL;
  PyObject *__pyx_v_ref_price = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  unsigned int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("market_status_data_frame", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":500
 * 
 *     def market_status_data_frame(self, market_trading_pair_tuples: List[MarketTradingPairTuple]) -> pd.DataFrame:
 *         markets_data = []             # <<<<<<<<<<<<<<
 *         markets_columns = ["Exchange", "Market", "Best Bid", "Best Ask", f"MidPrice"]
 *         markets_columns.append('Reservation Price')
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 500, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_markets_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":501
 *     def market_status_data_frame(self, market_trading_pair_tuples: List[MarketTradingPairTuple]) -> pd.DataFrame:
 *         markets_data = []
 *         markets_columns = ["Exchange", "Market", "Best Bid", "Best Ask", f"MidPrice"]             # <<<<<<<<<<<<<<
 *         markets_columns.append('Reservation Price')
 *         markets_columns.append('Optimal Spread')
 */
  __pyx_t_1 = PyList_New(5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 501, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_u_Exchange);
  __Pyx_GIVEREF(__pyx_n_u_Exchange);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_u_Exchange)) __PYX_ERR(0, 501, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_Market);
  __Pyx_GIVEREF(__pyx_n_u_Market);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_u_Market)) __PYX_ERR(0, 501, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_Best_Bid);
  __Pyx_GIVEREF(__pyx_kp_u_Best_Bid);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_Best_Bid)) __PYX_ERR(0, 501, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_Best_Ask);
  __Pyx_GIVEREF(__pyx_kp_u_Best_Ask);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 3, __pyx_kp_u_Best_Ask)) __PYX_ERR(0, 501, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_u_MidPrice);
  __Pyx_GIVEREF(__pyx_n_u_MidPrice);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_u_MidPrice)) __PYX_ERR(0, 501, __pyx_L1_error);
  __pyx_v_markets_columns = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":502
 *         markets_data = []
 *         markets_columns = ["Exchange", "Market", "Best Bid", "Best Ask", f"MidPrice"]
 *         markets_columns.append('Reservation Price')             # <<<<<<<<<<<<<<
 *         markets_columns.append('Optimal Spread')
 *         market_books = [(self._market_info.market, self._market_info.trading_pair)]
 */
  __pyx_t_2 = __Pyx_PyList_Append(__pyx_v_markets_columns, __pyx_kp_u_Reservation_Price); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 502, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":503
 *         markets_columns = ["Exchange", "Market", "Best Bid", "Best Ask", f"MidPrice"]
 *         markets_columns.append('Reservation Price')
 *         markets_columns.append('Optimal Spread')             # <<<<<<<<<<<<<<
 *         market_books = [(self._market_info.market, self._market_info.trading_pair)]
 *         for market, trading_pair in market_books:
 */
  __pyx_t_2 = __Pyx_PyList_Append(__pyx_v_markets_columns, __pyx_kp_u_Optimal_Spread); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 503, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":504
 *         markets_columns.append('Reservation Price')
 *         markets_columns.append('Optimal Spread')
 *         market_books = [(self._market_info.market, self._market_info.trading_pair)]             # <<<<<<<<<<<<<<
 *         for market, trading_pair in market_books:
 *             bid_price = market.get_price(trading_pair, False)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1)) __PYX_ERR(0, 504, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_3)) __PYX_ERR(0, 504, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_t_4)) __PYX_ERR(0, 504, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_v_market_books = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":505
 *         markets_columns.append('Optimal Spread')
 *         market_books = [(self._market_info.market, self._market_info.trading_pair)]
 *         for market, trading_pair in market_books:             # <<<<<<<<<<<<<<
 *             bid_price = market.get_price(trading_pair, False)
 *             ask_price = market.get_price(trading_pair, True)
 */
  __pyx_t_3 = __pyx_v_market_books; __Pyx_INCREF(__pyx_t_3);
  __pyx_t_5 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 505, __pyx_L1_error)
      #endif
      if (__pyx_t_5 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_4); __pyx_t_5++; if (unlikely((0 < 0))) __PYX_ERR(0, 505, __pyx_L1_error)
    #else
    __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 505, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 505, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 505, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 505, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 505, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7);
      index = 0; __pyx_t_1 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_1)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L5_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 505, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L6_unpacking_done;
      __pyx_L5_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 505, __pyx_L1_error)
      __pyx_L6_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_market, __pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_trading_pair, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":506
 *         market_books = [(self._market_info.market, self._market_info.trading_pair)]
 *         for market, trading_pair in market_books:
 *             bid_price = market.get_price(trading_pair, False)             # <<<<<<<<<<<<<<
 *             ask_price = market.get_price(trading_pair, True)
 *             ref_price = self.get_price()
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_market, __pyx_n_s_get_price); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 506, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_v_trading_pair, Py_False};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_9, 2+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 506, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_bid_price, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":507
 *         for market, trading_pair in market_books:
 *             bid_price = market.get_price(trading_pair, False)
 *             ask_price = market.get_price(trading_pair, True)             # <<<<<<<<<<<<<<
 *             ref_price = self.get_price()
 *             markets_data.append([
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_market, __pyx_n_s_get_price); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 507, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_v_trading_pair, Py_True};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_9, 2+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 507, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_ask_price, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":508
 *             bid_price = market.get_price(trading_pair, False)
 *             ask_price = market.get_price(trading_pair, True)
 *             ref_price = self.get_price()             # <<<<<<<<<<<<<<
 *             markets_data.append([
 *                 market.display_name,
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_price); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 508, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = NULL;
    __pyx_t_9 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_9 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_9, 0+__pyx_t_9);
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 508, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_XDECREF_SET(__pyx_v_ref_price, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":510
 *             ref_price = self.get_price()
 *             markets_data.append([
 *                 market.display_name,             # <<<<<<<<<<<<<<
 *                 trading_pair,
 *                 float(bid_price),
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_market, __pyx_n_s_display_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":512
 *                 market.display_name,
 *                 trading_pair,
 *                 float(bid_price),             # <<<<<<<<<<<<<<
 *                 float(ask_price),
 *                 float(ref_price),
 */
    __pyx_t_6 = __Pyx_PyNumber_Float(__pyx_v_bid_price); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":513
 *                 trading_pair,
 *                 float(bid_price),
 *                 float(ask_price),             # <<<<<<<<<<<<<<
 *                 float(ref_price),
 *                 round(self._reservation_price, 5),
 */
    __pyx_t_1 = __Pyx_PyNumber_Float(__pyx_v_ask_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 513, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":514
 *                 float(bid_price),
 *                 float(ask_price),
 *                 float(ref_price),             # <<<<<<<<<<<<<<
 *                 round(self._reservation_price, 5),
 *                 round(self._optimal_spread, 5),
 */
    __pyx_t_7 = __Pyx_PyNumber_Float(__pyx_v_ref_price); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 514, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":515
 *                 float(ask_price),
 *                 float(ref_price),
 *                 round(self._reservation_price, 5),             # <<<<<<<<<<<<<<
 *                 round(self._optimal_spread, 5),
 *             ])
 */
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_v_self->_reservation_price);
    __Pyx_GIVEREF(__pyx_v_self->_reservation_price);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_self->_reservation_price)) __PYX_ERR(0, 515, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_5);
    __Pyx_GIVEREF(__pyx_int_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_int_5)) __PYX_ERR(0, 515, __pyx_L1_error);
    __pyx_t_11 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_10, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 515, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":516
 *                 float(ref_price),
 *                 round(self._reservation_price, 5),
 *                 round(self._optimal_spread, 5),             # <<<<<<<<<<<<<<
 *             ])
 *         return pd.DataFrame(data=markets_data, columns=markets_columns).replace(np.nan, '', regex=True)
 */
    __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_INCREF(__pyx_v_self->_optimal_spread);
    __Pyx_GIVEREF(__pyx_v_self->_optimal_spread);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_v_self->_optimal_spread)) __PYX_ERR(0, 516, __pyx_L1_error);
    __Pyx_INCREF(__pyx_int_5);
    __Pyx_GIVEREF(__pyx_int_5);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_int_5)) __PYX_ERR(0, 516, __pyx_L1_error);
    __pyx_t_12 = __Pyx_PyObject_Call(__pyx_builtin_round, __pyx_t_10, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 516, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":509
 *             ask_price = market.get_price(trading_pair, True)
 *             ref_price = self.get_price()
 *             markets_data.append([             # <<<<<<<<<<<<<<
 *                 market.display_name,
 *                 trading_pair,
 */
    __pyx_t_10 = PyList_New(7); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 509, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 0, __pyx_t_4)) __PYX_ERR(0, 509, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_trading_pair);
    __Pyx_GIVEREF(__pyx_v_trading_pair);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 1, __pyx_v_trading_pair)) __PYX_ERR(0, 509, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 2, __pyx_t_6)) __PYX_ERR(0, 509, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 3, __pyx_t_1)) __PYX_ERR(0, 509, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_7);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 4, __pyx_t_7)) __PYX_ERR(0, 509, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_11);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 5, __pyx_t_11)) __PYX_ERR(0, 509, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_12);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_10, 6, __pyx_t_12)) __PYX_ERR(0, 509, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_6 = 0;
    __pyx_t_1 = 0;
    __pyx_t_7 = 0;
    __pyx_t_11 = 0;
    __pyx_t_12 = 0;
    __pyx_t_2 = __Pyx_PyList_Append(__pyx_v_markets_data, __pyx_t_10); if (unlikely(__pyx_t_2 == ((int)-1))) __PYX_ERR(0, 509, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":505
 *         markets_columns.append('Optimal Spread')
 *         market_books = [(self._market_info.market, self._market_info.trading_pair)]
 *         for market, trading_pair in market_books:             # <<<<<<<<<<<<<<
 *             bid_price = market.get_price(trading_pair, False)
 *             ask_price = market.get_price(trading_pair, True)
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":518
 *                 round(self._optimal_spread, 5),
 *             ])
 *         return pd.DataFrame(data=markets_data, columns=markets_columns).replace(np.nan, '', regex=True)             # <<<<<<<<<<<<<<
 * 
 *     def format_status(self) -> str:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_pd); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_DataFrame); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_data, __pyx_v_markets_data) < 0) __PYX_ERR(0, 518, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_columns, __pyx_v_markets_columns) < 0) __PYX_ERR(0, 518, __pyx_L1_error)
  __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_replace); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_np); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_nan); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_10);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_10)) __PYX_ERR(0, 518, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u__4);
  __Pyx_GIVEREF(__pyx_kp_u__4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_kp_u__4)) __PYX_ERR(0, 518, __pyx_L1_error);
  __pyx_t_10 = 0;
  __pyx_t_10 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  if (PyDict_SetItem(__pyx_t_10, __pyx_n_s_regex, Py_True) < 0) __PYX_ERR(0, 518, __pyx_L1_error)
  __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_12, __pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 518, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_r = __pyx_t_11;
  __pyx_t_11 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":499
 *         return pd.DataFrame(data=data, columns=columns)
 * 
 *     def market_status_data_frame(self, market_trading_pair_tuples: List[MarketTradingPairTuple]) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         markets_data = []
 *         markets_columns = ["Exchange", "Market", "Best Bid", "Best Ask", f"MidPrice"]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.market_status_data_frame", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_markets_data);
  __Pyx_XDECREF(__pyx_v_markets_columns);
  __Pyx_XDECREF(__pyx_v_market_books);
  __Pyx_XDECREF(__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_trading_pair);
  __Pyx_XDECREF(__pyx_v_bid_price);
  __Pyx_XDECREF(__pyx_v_ask_price);
  __Pyx_XDECREF(__pyx_v_ref_price);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":520
 *         return pd.DataFrame(data=markets_data, columns=markets_columns).replace(np.nan, '', regex=True)
 * 
 *     def format_status(self) -> str:             # <<<<<<<<<<<<<<
 *         if not self._all_markets_ready:
 *             return "Market connectors are not ready."
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_27format_status(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_27format_status = {"format_status", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_27format_status, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_27format_status(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("format_status (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("format_status", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "format_status", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26format_status(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26format_status(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_v_lines = 0;
  PyObject *__pyx_v_warning_lines = 0;
  PyObject *__pyx_v_markets_df = NULL;
  PyObject *__pyx_v_assets_df = NULL;
  PyObject *__pyx_v_first_col_length = NULL;
  PyObject *__pyx_v_df_lines = NULL;
  PyObject *__pyx_v_df = NULL;
  PyObject *__pyx_v_volatility_pct = NULL;
  PyObject *__pyx_8genexpr5__pyx_v_line = NULL;
  PyObject *__pyx_8genexpr6__pyx_v_line = NULL;
  PyObject *__pyx_8genexpr7__pyx_v_line = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_t_12;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  Py_UCS4 __pyx_t_16;
  int __pyx_t_17;
  double __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("format_status", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":521
 * 
 *     def format_status(self) -> str:
 *         if not self._all_markets_ready:             # <<<<<<<<<<<<<<
 *             return "Market connectors are not ready."
 *         cdef:
 */
  __pyx_t_1 = (!__pyx_v_self->_all_markets_ready);
  if (__pyx_t_1) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":522
 *     def format_status(self) -> str:
 *         if not self._all_markets_ready:
 *             return "Market connectors are not ready."             # <<<<<<<<<<<<<<
 *         cdef:
 *             list lines = []
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_kp_u_Market_connectors_are_not_ready);
    __pyx_r = __pyx_kp_u_Market_connectors_are_not_ready;
    goto __pyx_L0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":521
 * 
 *     def format_status(self) -> str:
 *         if not self._all_markets_ready:             # <<<<<<<<<<<<<<
 *             return "Market connectors are not ready."
 *         cdef:
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":524
 *             return "Market connectors are not ready."
 *         cdef:
 *             list lines = []             # <<<<<<<<<<<<<<
 *             list warning_lines = []
 *         warning_lines.extend(self.network_warning([self._market_info]))
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 524, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_lines = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":525
 *         cdef:
 *             list lines = []
 *             list warning_lines = []             # <<<<<<<<<<<<<<
 *         warning_lines.extend(self.network_warning([self._market_info]))
 * 
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_warning_lines = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":526
 *             list lines = []
 *             list warning_lines = []
 *         warning_lines.extend(self.network_warning([self._market_info]))             # <<<<<<<<<<<<<<
 * 
 *         markets_df = self.market_status_data_frame([self._market_info])
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_network_warning); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_self->_market_info);
  __Pyx_GIVEREF(__pyx_v_self->_market_info);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_self->_market_info)) __PYX_ERR(0, 526, __pyx_L1_error);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_7 = __Pyx_PyList_Extend(__pyx_v_warning_lines, __pyx_t_2); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 526, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":528
 *         warning_lines.extend(self.network_warning([self._market_info]))
 * 
 *         markets_df = self.market_status_data_frame([self._market_info])             # <<<<<<<<<<<<<<
 *         lines.extend(["", "  Markets:"] + ["    " + line for line in markets_df.to_string(index=False).split("\n")])
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_market_status_data_frame); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyList_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 528, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_self->_market_info);
  __Pyx_GIVEREF(__pyx_v_self->_market_info);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_v_self->_market_info)) __PYX_ERR(0, 528, __pyx_L1_error);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 528, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_markets_df = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":529
 * 
 *         markets_df = self.market_status_data_frame([self._market_info])
 *         lines.extend(["", "  Markets:"] + ["    " + line for line in markets_df.to_string(index=False).split("\n")])             # <<<<<<<<<<<<<<
 * 
 *         assets_df = map_df_to_str(self.pure_mm_assets_df(True))
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_kp_u__4);
  __Pyx_GIVEREF(__pyx_kp_u__4);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u__4)) __PYX_ERR(0, 529, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_Markets);
  __Pyx_GIVEREF(__pyx_kp_u_Markets);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 1, __pyx_kp_u_Markets)) __PYX_ERR(0, 529, __pyx_L1_error);
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 529, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_markets_df, __pyx_n_s_to_string); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 529, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 529, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_8);
    if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_index, Py_False) < 0) __PYX_ERR(0, 529, __pyx_L6_error)
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_empty_tuple, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 529, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_split); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 529, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_kp_u__6};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 529, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
      __pyx_t_8 = __pyx_t_4; __Pyx_INCREF(__pyx_t_8);
      __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_10 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 529, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_11 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_8); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 529, __pyx_L6_error)
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    for (;;) {
      if (likely(!__pyx_t_11)) {
        if (likely(PyList_CheckExact(__pyx_t_8))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_8);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 529, __pyx_L6_error)
            #endif
            if (__pyx_t_10 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_10); __Pyx_INCREF(__pyx_t_4); __pyx_t_10++; if (unlikely((0 < 0))) __PYX_ERR(0, 529, __pyx_L6_error)
          #else
          __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 529, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_8);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 529, __pyx_L6_error)
            #endif
            if (__pyx_t_10 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_10); __Pyx_INCREF(__pyx_t_4); __pyx_t_10++; if (unlikely((0 < 0))) __PYX_ERR(0, 529, __pyx_L6_error)
          #else
          __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 529, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          #endif
        }
      } else {
        __pyx_t_4 = __pyx_t_11(__pyx_t_8);
        if (unlikely(!__pyx_t_4)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 529, __pyx_L6_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_4);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr5__pyx_v_line, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_4 = PyNumber_Add(__pyx_kp_u__7, __pyx_8genexpr5__pyx_v_line); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 529, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 529, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_line); __pyx_8genexpr5__pyx_v_line = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L6_error:;
    __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_line); __pyx_8genexpr5__pyx_v_line = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __pyx_t_8 = PyNumber_Add(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyList_Extend(__pyx_v_lines, __pyx_t_8); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":531
 *         lines.extend(["", "  Markets:"] + ["    " + line for line in markets_df.to_string(index=False).split("\n")])
 * 
 *         assets_df = map_df_to_str(self.pure_mm_assets_df(True))             # <<<<<<<<<<<<<<
 *         first_col_length = max(*assets_df[0].apply(len))
 *         df_lines = assets_df.to_string(index=False, header=False,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_map_df_to_str); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pure_mm_assets_df); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 531, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_9)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_9, Py_True};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_2};
    __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_assets_df = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":532
 * 
 *         assets_df = map_df_to_str(self.pure_mm_assets_df(True))
 *         first_col_length = max(*assets_df[0].apply(len))             # <<<<<<<<<<<<<<
 *         df_lines = assets_df.to_string(index=False, header=False,
 *                                        formatters={0: ("{:<" + str(first_col_length) + "}").format}).split("\n")
 */
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_assets_df, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_apply); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_GetBuiltinName(__pyx_n_s_len); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 532, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = __Pyx_PySequence_Tuple(__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_max, __pyx_t_2, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 532, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_first_col_length = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":533
 *         assets_df = map_df_to_str(self.pure_mm_assets_df(True))
 *         first_col_length = max(*assets_df[0].apply(len))
 *         df_lines = assets_df.to_string(index=False, header=False,             # <<<<<<<<<<<<<<
 *                                        formatters={0: ("{:<" + str(first_col_length) + "}").format}).split("\n")
 *         lines.extend(["", "  Assets:"] + ["    " + line for line in df_lines])
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_assets_df, __pyx_n_s_to_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_index, Py_False) < 0) __PYX_ERR(0, 533, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_header, Py_False) < 0) __PYX_ERR(0, 533, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":534
 *         first_col_length = max(*assets_df[0].apply(len))
 *         df_lines = assets_df.to_string(index=False, header=False,
 *                                        formatters={0: ("{:<" + str(first_col_length) + "}").format}).split("\n")             # <<<<<<<<<<<<<<
 *         lines.extend(["", "  Assets:"] + ["    " + line for line in df_lines])
 * 
 */
  __pyx_t_4 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_9 = __Pyx_PyObject_Unicode(__pyx_v_first_col_length); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_5 = __Pyx_PyUnicode_Concat(__pyx_kp_u__8, __pyx_t_9); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyUnicode_ConcatInPlace(__pyx_t_5, __pyx_kp_u__9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_4, __pyx_int_0, __pyx_t_5) < 0) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_formatters, __pyx_t_4) < 0) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":533
 *         assets_df = map_df_to_str(self.pure_mm_assets_df(True))
 *         first_col_length = max(*assets_df[0].apply(len))
 *         df_lines = assets_df.to_string(index=False, header=False,             # <<<<<<<<<<<<<<
 *                                        formatters={0: ("{:<" + str(first_col_length) + "}").format}).split("\n")
 *         lines.extend(["", "  Assets:"] + ["    " + line for line in df_lines])
 */
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":534
 *         first_col_length = max(*assets_df[0].apply(len))
 *         df_lines = assets_df.to_string(index=False, header=False,
 *                                        formatters={0: ("{:<" + str(first_col_length) + "}").format}).split("\n")             # <<<<<<<<<<<<<<
 *         lines.extend(["", "  Assets:"] + ["    " + line for line in df_lines])
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_split); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_u__6};
    __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 534, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_df_lines = __pyx_t_8;
  __pyx_t_8 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":535
 *         df_lines = assets_df.to_string(index=False, header=False,
 *                                        formatters={0: ("{:<" + str(first_col_length) + "}").format}).split("\n")
 *         lines.extend(["", "  Assets:"] + ["    " + line for line in df_lines])             # <<<<<<<<<<<<<<
 * 
 *         # See if there are any open orders.
 */
  __pyx_t_8 = PyList_New(2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_INCREF(__pyx_kp_u__4);
  __Pyx_GIVEREF(__pyx_kp_u__4);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 0, __pyx_kp_u__4)) __PYX_ERR(0, 535, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u_Assets);
  __Pyx_GIVEREF(__pyx_kp_u_Assets);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_8, 1, __pyx_kp_u_Assets)) __PYX_ERR(0, 535, __pyx_L1_error);
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 535, __pyx_L13_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (likely(PyList_CheckExact(__pyx_v_df_lines)) || PyTuple_CheckExact(__pyx_v_df_lines)) {
      __pyx_t_4 = __pyx_v_df_lines; __Pyx_INCREF(__pyx_t_4);
      __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
    } else {
      __pyx_t_10 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_df_lines); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 535, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_11 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 535, __pyx_L13_error)
    }
    for (;;) {
      if (likely(!__pyx_t_11)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 535, __pyx_L13_error)
            #endif
            if (__pyx_t_10 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely((0 < 0))) __PYX_ERR(0, 535, __pyx_L13_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 535, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 535, __pyx_L13_error)
            #endif
            if (__pyx_t_10 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_10); __Pyx_INCREF(__pyx_t_2); __pyx_t_10++; if (unlikely((0 < 0))) __PYX_ERR(0, 535, __pyx_L13_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 535, __pyx_L13_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_11(__pyx_t_4);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 535, __pyx_L13_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_8genexpr6__pyx_v_line, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Add(__pyx_kp_u__7, __pyx_8genexpr6__pyx_v_line); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 535, __pyx_L13_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 535, __pyx_L13_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_line); __pyx_8genexpr6__pyx_v_line = 0;
    goto __pyx_L17_exit_scope;
    __pyx_L13_error:;
    __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_line); __pyx_8genexpr6__pyx_v_line = 0;
    goto __pyx_L1_error;
    __pyx_L17_exit_scope:;
  } /* exit inner scope */
  __pyx_t_4 = PyNumber_Add(__pyx_t_8, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_7 = __Pyx_PyList_Extend(__pyx_v_lines, __pyx_t_4); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 535, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":538
 * 
 *         # See if there are any open orders.
 *         if len(self.active_orders) > 0:             # <<<<<<<<<<<<<<
 *             df = self.active_orders_df()
 *             lines.extend(["", "  Orders:"] + ["    " + line for line in df.to_string(index=False).split("\n")])
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_active_orders); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_10 = PyObject_Length(__pyx_t_4); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_1 = (__pyx_t_10 > 0);
  if (__pyx_t_1) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":539
 *         # See if there are any open orders.
 *         if len(self.active_orders) > 0:
 *             df = self.active_orders_df()             # <<<<<<<<<<<<<<
 *             lines.extend(["", "  Orders:"] + ["    " + line for line in df.to_string(index=False).split("\n")])
 *         else:
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_active_orders_df); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 539, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, NULL};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 539, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_v_df = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":540
 *         if len(self.active_orders) > 0:
 *             df = self.active_orders_df()
 *             lines.extend(["", "  Orders:"] + ["    " + line for line in df.to_string(index=False).split("\n")])             # <<<<<<<<<<<<<<
 *         else:
 *             lines.extend(["", "  No active maker orders."])
 */
    __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_kp_u__4);
    __Pyx_GIVEREF(__pyx_kp_u__4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u__4)) __PYX_ERR(0, 540, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_u_Orders);
    __Pyx_GIVEREF(__pyx_kp_u_Orders);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_kp_u_Orders)) __PYX_ERR(0, 540, __pyx_L1_error);
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 540, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_df, __pyx_n_s_to_string); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 540, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 540, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_index, Py_False) < 0) __PYX_ERR(0, 540, __pyx_L21_error)
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 540, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_split); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 540, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      __pyx_t_6 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_6 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_kp_u__6};
        __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 540, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      if (likely(PyList_CheckExact(__pyx_t_8)) || PyTuple_CheckExact(__pyx_t_8)) {
        __pyx_t_5 = __pyx_t_8; __Pyx_INCREF(__pyx_t_5);
        __pyx_t_10 = 0;
        __pyx_t_11 = NULL;
      } else {
        __pyx_t_10 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 540, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_11 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 540, __pyx_L21_error)
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      for (;;) {
        if (likely(!__pyx_t_11)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 540, __pyx_L21_error)
              #endif
              if (__pyx_t_10 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_8 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_8); __pyx_t_10++; if (unlikely((0 < 0))) __PYX_ERR(0, 540, __pyx_L21_error)
            #else
            __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 540, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_8);
            #endif
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 540, __pyx_L21_error)
              #endif
              if (__pyx_t_10 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_8); __pyx_t_10++; if (unlikely((0 < 0))) __PYX_ERR(0, 540, __pyx_L21_error)
            #else
            __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 540, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_8);
            #endif
          }
        } else {
          __pyx_t_8 = __pyx_t_11(__pyx_t_5);
          if (unlikely(!__pyx_t_8)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 540, __pyx_L21_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_8);
        }
        __Pyx_XDECREF_SET(__pyx_8genexpr7__pyx_v_line, __pyx_t_8);
        __pyx_t_8 = 0;
        __pyx_t_8 = PyNumber_Add(__pyx_kp_u__7, __pyx_8genexpr7__pyx_v_line); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 540, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_8))) __PYX_ERR(0, 540, __pyx_L21_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_line); __pyx_8genexpr7__pyx_v_line = 0;
      goto __pyx_L25_exit_scope;
      __pyx_L21_error:;
      __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_line); __pyx_8genexpr7__pyx_v_line = 0;
      goto __pyx_L1_error;
      __pyx_L25_exit_scope:;
    } /* exit inner scope */
    __pyx_t_5 = PyNumber_Add(__pyx_t_4, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_7 = __Pyx_PyList_Extend(__pyx_v_lines, __pyx_t_5); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 540, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":538
 * 
 *         # See if there are any open orders.
 *         if len(self.active_orders) > 0:             # <<<<<<<<<<<<<<
 *             df = self.active_orders_df()
 *             lines.extend(["", "  Orders:"] + ["    " + line for line in df.to_string(index=False).split("\n")])
 */
    goto __pyx_L18;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":542
 *             lines.extend(["", "  Orders:"] + ["    " + line for line in df.to_string(index=False).split("\n")])
 *         else:
 *             lines.extend(["", "  No active maker orders."])             # <<<<<<<<<<<<<<
 * 
 *         volatility_pct = self._avg_vol.current_value / float(self.get_price()) * 100.0
 */
  /*else*/ {
    __pyx_t_7 = __Pyx_ListComp_Append(__pyx_v_lines, __pyx_kp_u__4); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 542, __pyx_L1_error)
    __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_lines, __pyx_kp_u_No_active_maker_orders); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 542, __pyx_L1_error)
    (void)((__pyx_t_7 | __pyx_t_12));
  }
  __pyx_L18:;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":544
 *             lines.extend(["", "  No active maker orders."])
 * 
 *         volatility_pct = self._avg_vol.current_value / float(self.get_price()) * 100.0             # <<<<<<<<<<<<<<
 *         if all((self.gamma, self._alpha, self._kappa, not isnan(volatility_pct))):
 *             lines.extend(["", f"  Strategy parameters:",
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_avg_vol, __pyx_n_s_current_value); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_price); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_8 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_8, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 544, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_4 = __Pyx_PyNumber_Float(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyNumber_Divide(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = PyNumber_Multiply(__pyx_t_3, __pyx_float_100_0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 544, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_volatility_pct = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":545
 * 
 *         volatility_pct = self._avg_vol.current_value / float(self.get_price()) * 100.0
 *         if all((self.gamma, self._alpha, self._kappa, not isnan(volatility_pct))):             # <<<<<<<<<<<<<<
 *             lines.extend(["", f"  Strategy parameters:",
 *                           f"    risk_factor(\u03B3)= {self.gamma:.5E}",
 */
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_gamma); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_isnan); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_8 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_v_volatility_pct};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 545, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyBool_FromLong((!__pyx_t_1)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4)) __PYX_ERR(0, 545, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_alpha);
  __Pyx_GIVEREF(__pyx_v_self->_alpha);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_v_self->_alpha)) __PYX_ERR(0, 545, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_kappa);
  __Pyx_GIVEREF(__pyx_v_self->_kappa);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_self->_kappa)) __PYX_ERR(0, 545, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_3)) __PYX_ERR(0, 545, __pyx_L1_error);
  __pyx_t_4 = 0;
  __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_all, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 545, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_1) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":546
 *         volatility_pct = self._avg_vol.current_value / float(self.get_price()) * 100.0
 *         if all((self.gamma, self._alpha, self._kappa, not isnan(volatility_pct))):
 *             lines.extend(["", f"  Strategy parameters:",             # <<<<<<<<<<<<<<
 *                           f"    risk_factor(\u03B3)= {self.gamma:.5E}",
 *                           f"    order_book_intensity_factor(\u0391)= {self._alpha:.5E}",
 */
    __pyx_t_12 = __Pyx_ListComp_Append(__pyx_v_lines, __pyx_kp_u__4); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 546, __pyx_L1_error)
    __pyx_t_7 = __Pyx_ListComp_Append(__pyx_v_lines, __pyx_kp_u_Strategy_parameters); if (unlikely(__pyx_t_7 == ((int)-1))) __PYX_ERR(0, 546, __pyx_L1_error)

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":547
 *         if all((self.gamma, self._alpha, self._kappa, not isnan(volatility_pct))):
 *             lines.extend(["", f"  Strategy parameters:",
 *                           f"    risk_factor(\u03B3)= {self.gamma:.5E}",             # <<<<<<<<<<<<<<
 *                           f"    order_book_intensity_factor(\u0391)= {self._alpha:.5E}",
 *                           f"    order_book_depth_factor(\u03BA)= {self._kappa:.5E}",
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_gamma); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 547, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_Format(__pyx_t_3, __pyx_kp_u_5E); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 547, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_kp_u_risk_factor_2, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 547, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":546
 *         volatility_pct = self._avg_vol.current_value / float(self.get_price()) * 100.0
 *         if all((self.gamma, self._alpha, self._kappa, not isnan(volatility_pct))):
 *             lines.extend(["", f"  Strategy parameters:",             # <<<<<<<<<<<<<<
 *                           f"    risk_factor(\u03B3)= {self.gamma:.5E}",
 *                           f"    order_book_intensity_factor(\u0391)= {self._alpha:.5E}",
 */
    __pyx_t_13 = __Pyx_ListComp_Append(__pyx_v_lines, __pyx_t_3); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 546, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":548
 *             lines.extend(["", f"  Strategy parameters:",
 *                           f"    risk_factor(\u03B3)= {self.gamma:.5E}",
 *                           f"    order_book_intensity_factor(\u0391)= {self._alpha:.5E}",             # <<<<<<<<<<<<<<
 *                           f"    order_book_depth_factor(\u03BA)= {self._kappa:.5E}",
 *                           f"    volatility= {volatility_pct:.3f}%"])
 */
    __pyx_t_3 = __Pyx_PyObject_Format(__pyx_v_self->_alpha, __pyx_kp_u_5E); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 548, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyUnicode_Concat(__pyx_kp_u_order_book_intensity_factor, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 548, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":546
 *         volatility_pct = self._avg_vol.current_value / float(self.get_price()) * 100.0
 *         if all((self.gamma, self._alpha, self._kappa, not isnan(volatility_pct))):
 *             lines.extend(["", f"  Strategy parameters:",             # <<<<<<<<<<<<<<
 *                           f"    risk_factor(\u03B3)= {self.gamma:.5E}",
 *                           f"    order_book_intensity_factor(\u0391)= {self._alpha:.5E}",
 */
    __pyx_t_14 = __Pyx_ListComp_Append(__pyx_v_lines, __pyx_t_5); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 546, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":549
 *                           f"    risk_factor(\u03B3)= {self.gamma:.5E}",
 *                           f"    order_book_intensity_factor(\u0391)= {self._alpha:.5E}",
 *                           f"    order_book_depth_factor(\u03BA)= {self._kappa:.5E}",             # <<<<<<<<<<<<<<
 *                           f"    volatility= {volatility_pct:.3f}%"])
 *             if self._execution_state.time_left is not None:
 */
    __pyx_t_5 = __Pyx_PyObject_Format(__pyx_v_self->_kappa, __pyx_kp_u_5E); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyUnicode_Concat(__pyx_kp_u_order_book_depth_factor, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 549, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":546
 *         volatility_pct = self._avg_vol.current_value / float(self.get_price()) * 100.0
 *         if all((self.gamma, self._alpha, self._kappa, not isnan(volatility_pct))):
 *             lines.extend(["", f"  Strategy parameters:",             # <<<<<<<<<<<<<<
 *                           f"    risk_factor(\u03B3)= {self.gamma:.5E}",
 *                           f"    order_book_intensity_factor(\u0391)= {self._alpha:.5E}",
 */
    __pyx_t_15 = __Pyx_ListComp_Append(__pyx_v_lines, __pyx_t_3); if (unlikely(__pyx_t_15 == ((int)-1))) __PYX_ERR(0, 546, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":550
 *                           f"    order_book_intensity_factor(\u0391)= {self._alpha:.5E}",
 *                           f"    order_book_depth_factor(\u03BA)= {self._kappa:.5E}",
 *                           f"    volatility= {volatility_pct:.3f}%"])             # <<<<<<<<<<<<<<
 *             if self._execution_state.time_left is not None:
 *                 lines.extend([f"    time until end of trading cycle = {str(datetime.timedelta(seconds=float(self._execution_state.time_left)//1e3))}"])
 */
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 550, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = 0;
    __pyx_t_16 = 127;
    __Pyx_INCREF(__pyx_kp_u_volatility);
    __pyx_t_10 += 16;
    __Pyx_GIVEREF(__pyx_kp_u_volatility);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_volatility);
    __pyx_t_5 = __Pyx_PyObject_Format(__pyx_v_volatility_pct, __pyx_kp_u_3f); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 550, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_16 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) > __pyx_t_16) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_5) : __pyx_t_16;
    __pyx_t_10 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_INCREF(__pyx_kp_u__10);
    __pyx_t_10 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__10);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u__10);
    __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_3, 3, __pyx_t_10, __pyx_t_16); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 550, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":546
 *         volatility_pct = self._avg_vol.current_value / float(self.get_price()) * 100.0
 *         if all((self.gamma, self._alpha, self._kappa, not isnan(volatility_pct))):
 *             lines.extend(["", f"  Strategy parameters:",             # <<<<<<<<<<<<<<
 *                           f"    risk_factor(\u03B3)= {self.gamma:.5E}",
 *                           f"    order_book_intensity_factor(\u0391)= {self._alpha:.5E}",
 */
    __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_lines, __pyx_t_5); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 546, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    (void)((__pyx_t_12 | (__pyx_t_7 | (__pyx_t_13 | (__pyx_t_14 | (__pyx_t_15 | __pyx_t_17))))));

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":551
 *                           f"    order_book_depth_factor(\u03BA)= {self._kappa:.5E}",
 *                           f"    volatility= {volatility_pct:.3f}%"])
 *             if self._execution_state.time_left is not None:             # <<<<<<<<<<<<<<
 *                 lines.extend([f"    time until end of trading cycle = {str(datetime.timedelta(seconds=float(self._execution_state.time_left)//1e3))}"])
 *             else:
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_execution_state, __pyx_n_s_time_left); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 551, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = (__pyx_t_5 != Py_None);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_1) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":552
 *                           f"    volatility= {volatility_pct:.3f}%"])
 *             if self._execution_state.time_left is not None:
 *                 lines.extend([f"    time until end of trading cycle = {str(datetime.timedelta(seconds=float(self._execution_state.time_left)//1e3))}"])             # <<<<<<<<<<<<<<
 *             else:
 *                 lines.extend([f"    time until end of trading cycle = N/A"])
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_datetime); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 552, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_timedelta); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 552, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 552, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_execution_state, __pyx_n_s_time_left); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 552, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_18 = __Pyx_PyObject_AsDouble(__pyx_t_4); if (unlikely(__pyx_t_18 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 552, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = PyFloat_FromDouble(floor(__pyx_t_18 / 1e3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 552, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_seconds, __pyx_t_4) < 0) __PYX_ERR(0, 552, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_empty_tuple, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 552, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Unicode(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 552, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_4 = __Pyx_PyUnicode_Concat(__pyx_kp_u_time_until_end_of_trading_cycle, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 552, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_lines, __pyx_t_4); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 552, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":551
 *                           f"    order_book_depth_factor(\u03BA)= {self._kappa:.5E}",
 *                           f"    volatility= {volatility_pct:.3f}%"])
 *             if self._execution_state.time_left is not None:             # <<<<<<<<<<<<<<
 *                 lines.extend([f"    time until end of trading cycle = {str(datetime.timedelta(seconds=float(self._execution_state.time_left)//1e3))}"])
 *             else:
 */
      goto __pyx_L27;
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":554
 *                 lines.extend([f"    time until end of trading cycle = {str(datetime.timedelta(seconds=float(self._execution_state.time_left)//1e3))}"])
 *             else:
 *                 lines.extend([f"    time until end of trading cycle = N/A"])             # <<<<<<<<<<<<<<
 * 
 *         warning_lines.extend(self.balance_warning([self._market_info]))
 */
    /*else*/ {
      __pyx_t_17 = __Pyx_PyList_Append(__pyx_v_lines, __pyx_kp_u_time_until_end_of_trading_cycle_2); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 554, __pyx_L1_error)
    }
    __pyx_L27:;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":545
 * 
 *         volatility_pct = self._avg_vol.current_value / float(self.get_price()) * 100.0
 *         if all((self.gamma, self._alpha, self._kappa, not isnan(volatility_pct))):             # <<<<<<<<<<<<<<
 *             lines.extend(["", f"  Strategy parameters:",
 *                           f"    risk_factor(\u03B3)= {self.gamma:.5E}",
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":556
 *                 lines.extend([f"    time until end of trading cycle = N/A"])
 * 
 *         warning_lines.extend(self.balance_warning([self._market_info]))             # <<<<<<<<<<<<<<
 * 
 *         if len(warning_lines) > 0:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_balance_warning); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->_market_info);
  __Pyx_GIVEREF(__pyx_v_self->_market_info);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->_market_info)) __PYX_ERR(0, 556, __pyx_L1_error);
  __pyx_t_8 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_8)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_3};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 556, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_17 = __Pyx_PyList_Extend(__pyx_v_warning_lines, __pyx_t_4); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":558
 *         warning_lines.extend(self.balance_warning([self._market_info]))
 * 
 *         if len(warning_lines) > 0:             # <<<<<<<<<<<<<<
 *             lines.extend(["", "*** WARNINGS ***"] + warning_lines)
 * 
 */
  __pyx_t_10 = __Pyx_PyList_GET_SIZE(__pyx_v_warning_lines); if (unlikely(__pyx_t_10 == ((Py_ssize_t)-1))) __PYX_ERR(0, 558, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_10 > 0);
  if (__pyx_t_1) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":559
 * 
 *         if len(warning_lines) > 0:
 *             lines.extend(["", "*** WARNINGS ***"] + warning_lines)             # <<<<<<<<<<<<<<
 * 
 *         return "\n".join(lines)
 */
    __pyx_t_4 = PyList_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_kp_u__4);
    __Pyx_GIVEREF(__pyx_kp_u__4);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u__4)) __PYX_ERR(0, 559, __pyx_L1_error);
    __Pyx_INCREF(__pyx_kp_u_WARNINGS);
    __Pyx_GIVEREF(__pyx_kp_u_WARNINGS);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_4, 1, __pyx_kp_u_WARNINGS)) __PYX_ERR(0, 559, __pyx_L1_error);
    __pyx_t_5 = PyNumber_Add(__pyx_t_4, __pyx_v_warning_lines); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_17 = __Pyx_PyList_Extend(__pyx_v_lines, __pyx_t_5); if (unlikely(__pyx_t_17 == ((int)-1))) __PYX_ERR(0, 559, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":558
 *         warning_lines.extend(self.balance_warning([self._market_info]))
 * 
 *         if len(warning_lines) > 0:             # <<<<<<<<<<<<<<
 *             lines.extend(["", "*** WARNINGS ***"] + warning_lines)
 * 
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":561
 *             lines.extend(["", "*** WARNINGS ***"] + warning_lines)
 * 
 *         return "\n".join(lines)             # <<<<<<<<<<<<<<
 * 
 *     def execute_orders_proposal(self, proposal: Proposal):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_5 = PyUnicode_Join(__pyx_kp_u__6, __pyx_v_lines); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":520
 *         return pd.DataFrame(data=markets_data, columns=markets_columns).replace(np.nan, '', regex=True)
 * 
 *     def format_status(self) -> str:             # <<<<<<<<<<<<<<
 *         if not self._all_markets_ready:
 *             return "Market connectors are not ready."
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.format_status", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_lines);
  __Pyx_XDECREF(__pyx_v_warning_lines);
  __Pyx_XDECREF(__pyx_v_markets_df);
  __Pyx_XDECREF(__pyx_v_assets_df);
  __Pyx_XDECREF(__pyx_v_first_col_length);
  __Pyx_XDECREF(__pyx_v_df_lines);
  __Pyx_XDECREF(__pyx_v_df);
  __Pyx_XDECREF(__pyx_v_volatility_pct);
  __Pyx_XDECREF(__pyx_8genexpr5__pyx_v_line);
  __Pyx_XDECREF(__pyx_8genexpr6__pyx_v_line);
  __Pyx_XDECREF(__pyx_8genexpr7__pyx_v_line);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":563
 *         return "\n".join(lines)
 * 
 *     def execute_orders_proposal(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_execute_orders_proposal(proposal)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_29execute_orders_proposal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_29execute_orders_proposal = {"execute_orders_proposal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_29execute_orders_proposal, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_29execute_orders_proposal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_proposal = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("execute_orders_proposal (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_proposal,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_proposal)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 563, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "execute_orders_proposal") < 0)) __PYX_ERR(0, 563, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_proposal = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("execute_orders_proposal", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 563, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.execute_orders_proposal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_28execute_orders_proposal(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_proposal);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_28execute_orders_proposal(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("execute_orders_proposal", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":564
 * 
 *     def execute_orders_proposal(self, proposal: Proposal):
 *         return self.c_execute_orders_proposal(proposal)             # <<<<<<<<<<<<<<
 * 
 *     def cancel_order(self, order_id: str):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_execute_orders_proposal(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":563
 *         return "\n".join(lines)
 * 
 *     def execute_orders_proposal(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_execute_orders_proposal(proposal)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.execute_orders_proposal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":566
 *         return self.c_execute_orders_proposal(proposal)
 * 
 *     def cancel_order(self, order_id: str):             # <<<<<<<<<<<<<<
 *         return self.c_cancel_order(self._market_info, order_id)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_31cancel_order(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_31cancel_order = {"cancel_order", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_31cancel_order, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_31cancel_order(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_order_id = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel_order (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_order_id,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_order_id)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 566, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "cancel_order") < 0)) __PYX_ERR(0, 566, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_order_id = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cancel_order", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 566, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.cancel_order", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_30cancel_order(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_order_id);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_30cancel_order(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_order_id) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancel_order", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":567
 * 
 *     def cancel_order(self, order_id: str):
 *         return self.c_cancel_order(self._market_info, order_id)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_start(self, Clock clock, double timestamp):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_v_self->_market_info;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_v_order_id))||((__pyx_v_order_id) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_v_order_id))) __PYX_ERR(0, 567, __pyx_L1_error)
  __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_cancel_order(((struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *)__pyx_v_self), __pyx_t_1, ((PyObject*)__pyx_v_order_id)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 567, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":566
 *         return self.c_execute_orders_proposal(proposal)
 * 
 *     def cancel_order(self, order_id: str):             # <<<<<<<<<<<<<<
 *         return self.c_cancel_order(self._market_info, order_id)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.cancel_order", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7c_start_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":580
 *             restored_order_ids = self.c_track_restored_orders(self.market_info)
 *             for order_id in restored_order_ids:
 *                 order = next(o for o in self.market_info.market.limit_orders if o.client_order_id == order_id)             # <<<<<<<<<<<<<<
 *                 if order:
 *                     self._hanging_orders_tracker.add_as_hanging_order(order)
 */

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7c_start_genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr *)__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 580, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *) __pyx_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7c_start_2generator, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_AvellanedaMarketMakingStrategy_c, __pyx_n_s_hummingbot_strategy_avellaneda_m); if (unlikely(!gen)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_start.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7c_start_2generator(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 580, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 580, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_genexpr_arg_0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 580, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 580, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 580, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 580, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 580, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 580, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 580, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 580, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_o);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_o, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_o, __pyx_n_s_client_order_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_order_id)) { __Pyx_RaiseClosureNameError("order_id"); __PYX_ERR(0, 580, __pyx_L1_error) }
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_4, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_order_id, Py_EQ); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 580, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_6) {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_o);
      __pyx_r = __pyx_cur_scope->__pyx_v_o;
      __Pyx_XGIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L7_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 580, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":569
 *         return self.c_cancel_order(self._market_info, order_id)
 * 
 *     cdef c_start(self, Clock clock, double timestamp):             # <<<<<<<<<<<<<<
 *         StrategyBase.c_start(self, clock, timestamp)
 *         self.update_from_config_map()
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_start(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, struct __pyx_obj_10hummingbot_4core_5clock_Clock *__pyx_v_clock, double __pyx_v_timestamp) {
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *__pyx_cur_scope;
  PyObject *__pyx_v_restored_order_ids = NULL;
  PyObject *__pyx_v_order = NULL;
  PyObject *__pyx_gb_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7c_start_2generator = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_start", 0);
  __pyx_cur_scope = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *)__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 569, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":570
 * 
 *     cdef c_start(self, Clock clock, double timestamp):
 *         StrategyBase.c_start(self, clock, timestamp)             # <<<<<<<<<<<<<<
 *         self.update_from_config_map()
 *         self._last_timestamp = timestamp
 */
  __pyx_t_1 = __pyx_vtabptr_10hummingbot_8strategy_13strategy_base_StrategyBase->__pyx_base.c_start(((struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *)__pyx_v_self), __pyx_v_clock, __pyx_v_timestamp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":571
 *     cdef c_start(self, Clock clock, double timestamp):
 *         StrategyBase.c_start(self, clock, timestamp)
 *         self.update_from_config_map()             # <<<<<<<<<<<<<<
 *         self._last_timestamp = timestamp
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_update_from_config_map); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 571, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":572
 *         StrategyBase.c_start(self, clock, timestamp)
 *         self.update_from_config_map()
 *         self._last_timestamp = timestamp             # <<<<<<<<<<<<<<
 * 
 *         self._hanging_orders_tracker.register_events(self.active_markets)
 */
  __pyx_v_self->_last_timestamp = __pyx_v_timestamp;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":574
 *         self._last_timestamp = timestamp
 * 
 *         self._hanging_orders_tracker.register_events(self.active_markets)             # <<<<<<<<<<<<<<
 * 
 *         if self._hanging_orders_enabled:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_n_s_register_events); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_active_markets); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 574, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":576
 *         self._hanging_orders_tracker.register_events(self.active_markets)
 * 
 *         if self._hanging_orders_enabled:             # <<<<<<<<<<<<<<
 *             # start tracking any restored limit order
 *             restored_order_ids = self.c_track_restored_orders(self.market_info)
 */
  if (__pyx_v_self->_hanging_orders_enabled) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":578
 *         if self._hanging_orders_enabled:
 *             # start tracking any restored limit order
 *             restored_order_ids = self.c_track_restored_orders(self.market_info)             # <<<<<<<<<<<<<<
 *             for order_id in restored_order_ids:
 *                 order = next(o for o in self.market_info.market.limit_orders if o.client_order_id == order_id)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_market_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_track_restored_orders(((struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *)__pyx_v_self), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 578, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_restored_order_ids = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":579
 *             # start tracking any restored limit order
 *             restored_order_ids = self.c_track_restored_orders(self.market_info)
 *             for order_id in restored_order_ids:             # <<<<<<<<<<<<<<
 *                 order = next(o for o in self.market_info.market.limit_orders if o.client_order_id == order_id)
 *                 if order:
 */
    if (likely(PyList_CheckExact(__pyx_v_restored_order_ids)) || PyTuple_CheckExact(__pyx_v_restored_order_ids)) {
      __pyx_t_2 = __pyx_v_restored_order_ids; __Pyx_INCREF(__pyx_t_2);
      __pyx_t_6 = 0;
      __pyx_t_7 = NULL;
    } else {
      __pyx_t_6 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_restored_order_ids); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 579, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 579, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_7)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 579, __pyx_L1_error)
            #endif
            if (__pyx_t_6 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 579, __pyx_L1_error)
          #else
          __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 579, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 579, __pyx_L1_error)
            #endif
            if (__pyx_t_6 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_6); __Pyx_INCREF(__pyx_t_1); __pyx_t_6++; if (unlikely((0 < 0))) __PYX_ERR(0, 579, __pyx_L1_error)
          #else
          __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_6); __pyx_t_6++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 579, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          #endif
        }
      } else {
        __pyx_t_1 = __pyx_t_7(__pyx_t_2);
        if (unlikely(!__pyx_t_1)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 579, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_1);
      }
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_order_id);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_order_id, __pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      __pyx_t_1 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":580
 *             restored_order_ids = self.c_track_restored_orders(self.market_info)
 *             for order_id in restored_order_ids:
 *                 order = next(o for o in self.market_info.market.limit_orders if o.client_order_id == order_id)             # <<<<<<<<<<<<<<
 *                 if order:
 *                     self._hanging_orders_tracker.add_as_hanging_order(order)
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_market_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 580, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_market); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 580, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_limit_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 580, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7c_start_genexpr(((PyObject*)__pyx_cur_scope), __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 580, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyIter_Next(__pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 580, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_order, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":581
 *             for order_id in restored_order_ids:
 *                 order = next(o for o in self.market_info.market.limit_orders if o.client_order_id == order_id)
 *                 if order:             # <<<<<<<<<<<<<<
 *                     self._hanging_orders_tracker.add_as_hanging_order(order)
 * 
 */
      __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_v_order); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 581, __pyx_L1_error)
      if (__pyx_t_8) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":582
 *                 order = next(o for o in self.market_info.market.limit_orders if o.client_order_id == order_id)
 *                 if order:
 *                     self._hanging_orders_tracker.add_as_hanging_order(order)             # <<<<<<<<<<<<<<
 * 
 *         self._execution_state.time_left = self._execution_state.closing_time
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_n_s_add_as_hanging_order); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 582, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = NULL;
        __pyx_t_4 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
            __pyx_t_4 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_v_order};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 582, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":581
 *             for order_id in restored_order_ids:
 *                 order = next(o for o in self.market_info.market.limit_orders if o.client_order_id == order_id)
 *                 if order:             # <<<<<<<<<<<<<<
 *                     self._hanging_orders_tracker.add_as_hanging_order(order)
 * 
 */
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":579
 *             # start tracking any restored limit order
 *             restored_order_ids = self.c_track_restored_orders(self.market_info)
 *             for order_id in restored_order_ids:             # <<<<<<<<<<<<<<
 *                 order = next(o for o in self.market_info.market.limit_orders if o.client_order_id == order_id)
 *                 if order:
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":576
 *         self._hanging_orders_tracker.register_events(self.active_markets)
 * 
 *         if self._hanging_orders_enabled:             # <<<<<<<<<<<<<<
 *             # start tracking any restored limit order
 *             restored_order_ids = self.c_track_restored_orders(self.market_info)
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":584
 *                     self._hanging_orders_tracker.add_as_hanging_order(order)
 * 
 *         self._execution_state.time_left = self._execution_state.closing_time             # <<<<<<<<<<<<<<
 * 
 *     def start(self, clock: Clock, timestamp: float):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_execution_state, __pyx_n_s_closing_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->_execution_state, __pyx_n_s_time_left, __pyx_t_2) < 0) __PYX_ERR(0, 584, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":569
 *         return self.c_cancel_order(self._market_info, order_id)
 * 
 *     cdef c_start(self, Clock clock, double timestamp):             # <<<<<<<<<<<<<<
 *         StrategyBase.c_start(self, clock, timestamp)
 *         self.update_from_config_map()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_restored_order_ids);
  __Pyx_XDECREF(__pyx_v_order);
  __Pyx_XDECREF(__pyx_gb_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7c_start_2generator);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":586
 *         self._execution_state.time_left = self._execution_state.closing_time
 * 
 *     def start(self, clock: Clock, timestamp: float):             # <<<<<<<<<<<<<<
 *         self.c_start(clock, timestamp)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_33start(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_33start = {"start", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_33start, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_33start(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_clock = 0;
  PyObject *__pyx_v_timestamp = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("start (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_clock,&__pyx_n_s_timestamp,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_clock)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 586, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timestamp)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 586, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("start", 1, 2, 2, 1); __PYX_ERR(0, 586, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "start") < 0)) __PYX_ERR(0, 586, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_clock = values[0];
    __pyx_v_timestamp = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("start", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 586, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_32start(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_clock, __pyx_v_timestamp);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_32start(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_clock, PyObject *__pyx_v_timestamp) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("start", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":587
 * 
 *     def start(self, clock: Clock, timestamp: float):
 *         self.c_start(clock, timestamp)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_stop(self, Clock clock):
 */
  if (!(likely(((__pyx_v_clock) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_clock, __pyx_ptype_10hummingbot_4core_5clock_Clock))))) __PYX_ERR(0, 587, __pyx_L1_error)
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_timestamp); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 587, __pyx_L1_error)
  __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.__pyx_base.c_start(((struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *)__pyx_v_self), ((struct __pyx_obj_10hummingbot_4core_5clock_Clock *)__pyx_v_clock), __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":586
 *         self._execution_state.time_left = self._execution_state.closing_time
 * 
 *     def start(self, clock: Clock, timestamp: float):             # <<<<<<<<<<<<<<
 *         self.c_start(clock, timestamp)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.start", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":589
 *         self.c_start(clock, timestamp)
 * 
 *     cdef c_stop(self, Clock clock):             # <<<<<<<<<<<<<<
 *         self._hanging_orders_tracker.unregister_events(self.active_markets)
 *         StrategyBase.c_stop(self, clock)
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_stop(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, struct __pyx_obj_10hummingbot_4core_5clock_Clock *__pyx_v_clock) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_stop", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":590
 * 
 *     cdef c_stop(self, Clock clock):
 *         self._hanging_orders_tracker.unregister_events(self.active_markets)             # <<<<<<<<<<<<<<
 *         StrategyBase.c_stop(self, clock)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_n_s_unregister_events); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_active_markets); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 590, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 590, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":591
 *     cdef c_stop(self, Clock clock):
 *         self._hanging_orders_tracker.unregister_events(self.active_markets)
 *         StrategyBase.c_stop(self, clock)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_tick(self, double timestamp):
 */
  __pyx_t_1 = __pyx_vtabptr_10hummingbot_8strategy_13strategy_base_StrategyBase->__pyx_base.c_stop(((struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *)__pyx_v_self), __pyx_v_clock); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":589
 *         self.c_start(clock, timestamp)
 * 
 *     cdef c_stop(self, Clock clock):             # <<<<<<<<<<<<<<
 *         self._hanging_orders_tracker.unregister_events(self.active_markets)
 *         StrategyBase.c_stop(self, clock)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_stop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":593
 *         StrategyBase.c_stop(self, clock)
 * 
 *     cdef c_tick(self, double timestamp):             # <<<<<<<<<<<<<<
 *         StrategyBase.c_tick(self, timestamp)
 *         cdef:
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_tick(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, double __pyx_v_timestamp) {
  int64_t __pyx_v_current_tick;
  int64_t __pyx_v_last_tick;
  int __pyx_v_should_report_warnings;
  PyObject *__pyx_8genexpr9__pyx_v_mkt = NULL;
  PyObject *__pyx_9genexpr10__pyx_v_mkt = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  unsigned int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  Py_UCS4 __pyx_t_13;
  char const *__pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  PyObject *__pyx_t_18 = NULL;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_tick", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":594
 * 
 *     cdef c_tick(self, double timestamp):
 *         StrategyBase.c_tick(self, timestamp)             # <<<<<<<<<<<<<<
 *         cdef:
 *             int64_t current_tick = <int64_t>(timestamp // self._status_report_interval)
 */
  __pyx_t_1 = __pyx_vtabptr_10hummingbot_8strategy_13strategy_base_StrategyBase->__pyx_base.c_tick(((struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *)__pyx_v_self), __pyx_v_timestamp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 594, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":596
 *         StrategyBase.c_tick(self, timestamp)
 *         cdef:
 *             int64_t current_tick = <int64_t>(timestamp // self._status_report_interval)             # <<<<<<<<<<<<<<
 *             int64_t last_tick = <int64_t>(self._last_timestamp // self._status_report_interval)
 *             bint should_report_warnings = ((current_tick > last_tick) and
 */
  if (unlikely(__pyx_v_self->_status_report_interval == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 596, __pyx_L1_error)
  }
  __pyx_v_current_tick = ((int64_t)floor(__pyx_v_timestamp / __pyx_v_self->_status_report_interval));

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":597
 *         cdef:
 *             int64_t current_tick = <int64_t>(timestamp // self._status_report_interval)
 *             int64_t last_tick = <int64_t>(self._last_timestamp // self._status_report_interval)             # <<<<<<<<<<<<<<
 *             bint should_report_warnings = ((current_tick > last_tick) and
 *                                            (self._logging_options & self.OPTION_LOG_STATUS_REPORT))
 */
  if (unlikely(__pyx_v_self->_status_report_interval == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(0, 597, __pyx_L1_error)
  }
  __pyx_v_last_tick = ((int64_t)floor(__pyx_v_self->_last_timestamp / __pyx_v_self->_status_report_interval));

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":598
 *             int64_t current_tick = <int64_t>(timestamp // self._status_report_interval)
 *             int64_t last_tick = <int64_t>(self._last_timestamp // self._status_report_interval)
 *             bint should_report_warnings = ((current_tick > last_tick) and             # <<<<<<<<<<<<<<
 *                                            (self._logging_options & self.OPTION_LOG_STATUS_REPORT))
 *             object proposal
 */
  __pyx_t_3 = (__pyx_v_current_tick > __pyx_v_last_tick);
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L3_bool_binop_done;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":599
 *             int64_t last_tick = <int64_t>(self._last_timestamp // self._status_report_interval)
 *             bint should_report_warnings = ((current_tick > last_tick) and
 *                                            (self._logging_options & self.OPTION_LOG_STATUS_REPORT))             # <<<<<<<<<<<<<<
 *             object proposal
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->_logging_options); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_OPTION_LOG_STATUS_REPORT); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyNumber_And(__pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 599, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = __pyx_t_3;
  __pyx_L3_bool_binop_done:;
  __pyx_v_should_report_warnings = __pyx_t_2;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":602
 *             object proposal
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if not self._all_markets_ready:
 *                 self._all_markets_ready = all([mkt.ready for mkt in self._sb_markets])
 */
  /*try:*/ {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":603
 * 
 *         try:
 *             if not self._all_markets_ready:             # <<<<<<<<<<<<<<
 *                 self._all_markets_ready = all([mkt.ready for mkt in self._sb_markets])
 *                 if not self._all_markets_ready:
 */
    __pyx_t_2 = (!__pyx_v_self->_all_markets_ready);
    if (__pyx_t_2) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":604
 *         try:
 *             if not self._all_markets_ready:
 *                 self._all_markets_ready = all([mkt.ready for mkt in self._sb_markets])             # <<<<<<<<<<<<<<
 *                 if not self._all_markets_ready:
 *                     # Markets not ready yet. Don't do anything.
 */
      { /* enter inner scope */
        __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 604, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = 0;
        __pyx_t_1 = __Pyx_set_iterator(__pyx_v_self->__pyx_base._sb_markets, 1, (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 604, __pyx_L11_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_XDECREF(__pyx_t_4);
        __pyx_t_4 = __pyx_t_1;
        __pyx_t_1 = 0;
        while (1) {
          __pyx_t_9 = __Pyx_set_iter_next(__pyx_t_4, __pyx_t_7, &__pyx_t_6, &__pyx_t_1, __pyx_t_8);
          if (unlikely(__pyx_t_9 == 0)) break;
          if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 604, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_XDECREF_SET(__pyx_8genexpr9__pyx_v_mkt, __pyx_t_1);
          __pyx_t_1 = 0;
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_8genexpr9__pyx_v_mkt, __pyx_n_s_ready); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 604, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_1);
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_t_1))) __PYX_ERR(0, 604, __pyx_L11_error)
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_mkt); __pyx_8genexpr9__pyx_v_mkt = 0;
        goto __pyx_L14_exit_scope;
        __pyx_L11_error:;
        __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_mkt); __pyx_8genexpr9__pyx_v_mkt = 0;
        goto __pyx_L6_error;
        __pyx_L14_exit_scope:;
      } /* exit inner scope */
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_all, __pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 604, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 604, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_v_self->_all_markets_ready = __pyx_t_2;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":605
 *             if not self._all_markets_ready:
 *                 self._all_markets_ready = all([mkt.ready for mkt in self._sb_markets])
 *                 if not self._all_markets_ready:             # <<<<<<<<<<<<<<
 *                     # Markets not ready yet. Don't do anything.
 *                     if should_report_warnings:
 */
      __pyx_t_2 = (!__pyx_v_self->_all_markets_ready);
      if (__pyx_t_2) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":607
 *                 if not self._all_markets_ready:
 *                     # Markets not ready yet. Don't do anything.
 *                     if should_report_warnings:             # <<<<<<<<<<<<<<
 *                         self.logger().warning(f"Markets are not ready. No market making trades are permitted.")
 *                     return
 */
        if (__pyx_v_should_report_warnings) {

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":608
 *                     # Markets not ready yet. Don't do anything.
 *                     if should_report_warnings:
 *                         self.logger().warning(f"Markets are not ready. No market making trades are permitted.")             # <<<<<<<<<<<<<<
 *                     return
 * 
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 608, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_10 = NULL;
          __pyx_t_11 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_10)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_10);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_11 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_10, NULL};
            __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_11, 0+__pyx_t_11);
            __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_warning); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 608, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = NULL;
          __pyx_t_11 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_11 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_kp_u_Markets_are_not_ready_No_market};
            __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 608, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":607
 *                 if not self._all_markets_ready:
 *                     # Markets not ready yet. Don't do anything.
 *                     if should_report_warnings:             # <<<<<<<<<<<<<<
 *                         self.logger().warning(f"Markets are not ready. No market making trades are permitted.")
 *                     return
 */
        }

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":609
 *                     if should_report_warnings:
 *                         self.logger().warning(f"Markets are not ready. No market making trades are permitted.")
 *                     return             # <<<<<<<<<<<<<<
 * 
 *             if should_report_warnings:
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        goto __pyx_L5_return;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":605
 *             if not self._all_markets_ready:
 *                 self._all_markets_ready = all([mkt.ready for mkt in self._sb_markets])
 *                 if not self._all_markets_ready:             # <<<<<<<<<<<<<<
 *                     # Markets not ready yet. Don't do anything.
 *                     if should_report_warnings:
 */
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":603
 * 
 *         try:
 *             if not self._all_markets_ready:             # <<<<<<<<<<<<<<
 *                 self._all_markets_ready = all([mkt.ready for mkt in self._sb_markets])
 *                 if not self._all_markets_ready:
 */
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":611
 *                     return
 * 
 *             if should_report_warnings:             # <<<<<<<<<<<<<<
 *                 if not all([mkt.network_status is NetworkStatus.CONNECTED for mkt in self._sb_markets]):
 *                     self.logger().warning(f"WARNING: Some markets are not connected or are down at the moment. Market "
 */
    if (__pyx_v_should_report_warnings) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":612
 * 
 *             if should_report_warnings:
 *                 if not all([mkt.network_status is NetworkStatus.CONNECTED for mkt in self._sb_markets]):             # <<<<<<<<<<<<<<
 *                     self.logger().warning(f"WARNING: Some markets are not connected or are down at the moment. Market "
 *                                           f"making may be dangerous when markets or networks are unstable.")
 */
      { /* enter inner scope */
        __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 612, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_7 = 0;
        __pyx_t_5 = __Pyx_set_iterator(__pyx_v_self->__pyx_base._sb_markets, 1, (&__pyx_t_6), (&__pyx_t_8)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 612, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_XDECREF(__pyx_t_1);
        __pyx_t_1 = __pyx_t_5;
        __pyx_t_5 = 0;
        while (1) {
          __pyx_t_9 = __Pyx_set_iter_next(__pyx_t_1, __pyx_t_6, &__pyx_t_7, &__pyx_t_5, __pyx_t_8);
          if (unlikely(__pyx_t_9 == 0)) break;
          if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 612, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_XDECREF_SET(__pyx_9genexpr10__pyx_v_mkt, __pyx_t_5);
          __pyx_t_5 = 0;
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr10__pyx_v_mkt, __pyx_n_s_network_status); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 612, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_NetworkStatus); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 612, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_CONNECTED); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 612, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_2 = (__pyx_t_5 == __pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_12 = __Pyx_PyBool_FromLong(__pyx_t_2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 612, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_12);
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_4, (PyObject*)__pyx_t_12))) __PYX_ERR(0, 612, __pyx_L21_error)
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_mkt); __pyx_9genexpr10__pyx_v_mkt = 0;
        goto __pyx_L24_exit_scope;
        __pyx_L21_error:;
        __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_mkt); __pyx_9genexpr10__pyx_v_mkt = 0;
        goto __pyx_L6_error;
        __pyx_L24_exit_scope:;
      } /* exit inner scope */
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_all, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 612, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 612, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_3 = (!__pyx_t_2);
      if (__pyx_t_3) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":613
 *             if should_report_warnings:
 *                 if not all([mkt.network_status is NetworkStatus.CONNECTED for mkt in self._sb_markets]):
 *                     self.logger().warning(f"WARNING: Some markets are not connected or are down at the moment. Market "             # <<<<<<<<<<<<<<
 *                                           f"making may be dangerous when markets or networks are unstable.")
 * 
 */
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 613, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_5 = NULL;
        __pyx_t_11 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_5)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_5);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
            __pyx_t_11 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_11, 0+__pyx_t_11);
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 613, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_warning); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 613, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        __pyx_t_11 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_12))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_12);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_12, function);
            __pyx_t_11 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_u_WARNING_Some_markets_are_not_con};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 613, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":612
 * 
 *             if should_report_warnings:
 *                 if not all([mkt.network_status is NetworkStatus.CONNECTED for mkt in self._sb_markets]):             # <<<<<<<<<<<<<<
 *                     self.logger().warning(f"WARNING: Some markets are not connected or are down at the moment. Market "
 *                                           f"making may be dangerous when markets or networks are unstable.")
 */
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":611
 *                     return
 * 
 *             if should_report_warnings:             # <<<<<<<<<<<<<<
 *                 if not all([mkt.network_status is NetworkStatus.CONNECTED for mkt in self._sb_markets]):
 *                     self.logger().warning(f"WARNING: Some markets are not connected or are down at the moment. Market "
 */
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":617
 * 
 *             # Updates settings from config map if changed
 *             self.update_from_config_map()             # <<<<<<<<<<<<<<
 * 
 *             self.c_collect_market_variables(timestamp)
 */
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_update_from_config_map); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 617, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_4 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_12);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_12, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_11, 0+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 617, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":619
 *             self.update_from_config_map()
 * 
 *             self.c_collect_market_variables(timestamp)             # <<<<<<<<<<<<<<
 * 
 *             if self.c_is_algorithm_ready():
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_collect_market_variables(__pyx_v_self, __pyx_v_timestamp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 619, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":621
 *             self.c_collect_market_variables(timestamp)
 * 
 *             if self.c_is_algorithm_ready():             # <<<<<<<<<<<<<<
 *                 if self._create_timestamp <= self._current_timestamp:
 *                     # Measure order book liquidity
 */
    __pyx_t_3 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_is_algorithm_ready(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 621, __pyx_L6_error)
    if (__pyx_t_3) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":622
 * 
 *             if self.c_is_algorithm_ready():
 *                 if self._create_timestamp <= self._current_timestamp:             # <<<<<<<<<<<<<<
 *                     # Measure order book liquidity
 *                     self.c_measure_order_book_liquidity()
 */
      __pyx_t_3 = (__pyx_v_self->_create_timestamp <= __pyx_v_self->__pyx_base.__pyx_base._current_timestamp);
      if (__pyx_t_3) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":624
 *                 if self._create_timestamp <= self._current_timestamp:
 *                     # Measure order book liquidity
 *                     self.c_measure_order_book_liquidity()             # <<<<<<<<<<<<<<
 * 
 *                 self._hanging_orders_tracker.process_tick()
 */
        __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_measure_order_book_liquidity(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 624, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":622
 * 
 *             if self.c_is_algorithm_ready():
 *                 if self._create_timestamp <= self._current_timestamp:             # <<<<<<<<<<<<<<
 *                     # Measure order book liquidity
 *                     self.c_measure_order_book_liquidity()
 */
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":626
 *                     self.c_measure_order_book_liquidity()
 * 
 *                 self._hanging_orders_tracker.process_tick()             # <<<<<<<<<<<<<<
 * 
 *                 # Needs to be executed at all times to not to have active order leftovers after a trading session ends
 */
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_n_s_process_tick); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 626, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_4 = NULL;
      __pyx_t_11 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
          __pyx_t_11 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_11, 0+__pyx_t_11);
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 626, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":629
 * 
 *                 # Needs to be executed at all times to not to have active order leftovers after a trading session ends
 *                 self.c_cancel_active_orders_on_max_age_limit()             # <<<<<<<<<<<<<<
 * 
 *                 # process_tick() is only called if within a trading timeframe
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_cancel_active_orders_on_max_age_limit(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 629, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":632
 * 
 *                 # process_tick() is only called if within a trading timeframe
 *                 self._execution_state.process_tick(timestamp, self)             # <<<<<<<<<<<<<<
 * 
 *             else:
 */
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_execution_state, __pyx_n_s_process_tick); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 632, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_4 = PyFloat_FromDouble(__pyx_v_timestamp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 632, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_5 = NULL;
      __pyx_t_11 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
          __pyx_t_11 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_t_4, ((PyObject *)__pyx_v_self)};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_11, 2+__pyx_t_11);
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 632, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":621
 *             self.c_collect_market_variables(timestamp)
 * 
 *             if self.c_is_algorithm_ready():             # <<<<<<<<<<<<<<
 *                 if self._create_timestamp <= self._current_timestamp:
 *                     # Measure order book liquidity
 */
      goto __pyx_L25;
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":636
 *             else:
 *                 # Only if snapshots are different - for trading intensity - a market order happened
 *                 if self.c_is_algorithm_changed():             # <<<<<<<<<<<<<<
 *                     self._ticks_to_be_ready -= 1
 *                     if self._ticks_to_be_ready % 5 == 0:
 */
    /*else*/ {
      __pyx_t_3 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_is_algorithm_changed(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 636, __pyx_L6_error)
      if (__pyx_t_3) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":637
 *                 # Only if snapshots are different - for trading intensity - a market order happened
 *                 if self.c_is_algorithm_changed():
 *                     self._ticks_to_be_ready -= 1             # <<<<<<<<<<<<<<
 *                     if self._ticks_to_be_ready % 5 == 0:
 *                         self.logger().info(f"Calculating volatility, estimating order book liquidity ... {self._ticks_to_be_ready} ticks to fill buffers")
 */
        __pyx_v_self->_ticks_to_be_ready = (__pyx_v_self->_ticks_to_be_ready - 1);

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":638
 *                 if self.c_is_algorithm_changed():
 *                     self._ticks_to_be_ready -= 1
 *                     if self._ticks_to_be_ready % 5 == 0:             # <<<<<<<<<<<<<<
 *                         self.logger().info(f"Calculating volatility, estimating order book liquidity ... {self._ticks_to_be_ready} ticks to fill buffers")
 *                 else:
 */
        __pyx_t_3 = (__Pyx_mod_long(__pyx_v_self->_ticks_to_be_ready, 5) == 0);
        if (__pyx_t_3) {

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":639
 *                     self._ticks_to_be_ready -= 1
 *                     if self._ticks_to_be_ready % 5 == 0:
 *                         self.logger().info(f"Calculating volatility, estimating order book liquidity ... {self._ticks_to_be_ready} ticks to fill buffers")             # <<<<<<<<<<<<<<
 *                 else:
 *                     self.logger().info(f"Calculating volatility, estimating order book liquidity ... no trades tick")
 */
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 639, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __pyx_t_5 = NULL;
          __pyx_t_11 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_5)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_5);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_11 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
            __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_11, 0+__pyx_t_11);
            __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 639, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_12, __pyx_n_s_info); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 639, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_12 = PyTuple_New(3); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 639, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_6 = 0;
          __pyx_t_13 = 127;
          __Pyx_INCREF(__pyx_kp_u_Calculating_volatility_estimatin);
          __pyx_t_6 += 60;
          __Pyx_GIVEREF(__pyx_kp_u_Calculating_volatility_estimatin);
          PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_kp_u_Calculating_volatility_estimatin);
          __pyx_t_5 = __Pyx_PyUnicode_From_int(__pyx_v_self->_ticks_to_be_ready, 0, ' ', 'd'); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 639, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_5);
          __pyx_t_5 = 0;
          __Pyx_INCREF(__pyx_kp_u_ticks_to_fill_buffers);
          __pyx_t_6 += 22;
          __Pyx_GIVEREF(__pyx_kp_u_ticks_to_fill_buffers);
          PyTuple_SET_ITEM(__pyx_t_12, 2, __pyx_kp_u_ticks_to_fill_buffers);
          __pyx_t_5 = __Pyx_PyUnicode_Join(__pyx_t_12, 3, __pyx_t_6, __pyx_t_13); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 639, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_12 = NULL;
          __pyx_t_11 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_4))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_4);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_4, function);
              __pyx_t_11 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_t_5};
            __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 639, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_1);
            __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          }
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":638
 *                 if self.c_is_algorithm_changed():
 *                     self._ticks_to_be_ready -= 1
 *                     if self._ticks_to_be_ready % 5 == 0:             # <<<<<<<<<<<<<<
 *                         self.logger().info(f"Calculating volatility, estimating order book liquidity ... {self._ticks_to_be_ready} ticks to fill buffers")
 *                 else:
 */
        }

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":636
 *             else:
 *                 # Only if snapshots are different - for trading intensity - a market order happened
 *                 if self.c_is_algorithm_changed():             # <<<<<<<<<<<<<<
 *                     self._ticks_to_be_ready -= 1
 *                     if self._ticks_to_be_ready % 5 == 0:
 */
        goto __pyx_L27;
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":641
 *                         self.logger().info(f"Calculating volatility, estimating order book liquidity ... {self._ticks_to_be_ready} ticks to fill buffers")
 *                 else:
 *                     self.logger().info(f"Calculating volatility, estimating order book liquidity ... no trades tick")             # <<<<<<<<<<<<<<
 *         finally:
 *             self._last_timestamp = timestamp
 */
      /*else*/ {
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 641, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_12 = NULL;
        __pyx_t_11 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_12)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_12);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_11 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_12, NULL};
          __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_11, 0+__pyx_t_11);
          __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 641, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_info); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 641, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        __pyx_t_11 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_5))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_5, function);
            __pyx_t_11 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_kp_u_Calculating_volatility_estimatin_2};
          __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 641, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __pyx_L27:;
    }
    __pyx_L25:;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":643
 *                     self.logger().info(f"Calculating volatility, estimating order book liquidity ... no trades tick")
 *         finally:
 *             self._last_timestamp = timestamp             # <<<<<<<<<<<<<<
 * 
 *     def process_tick(self, timestamp: float):
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_v_self->_last_timestamp = __pyx_v_timestamp;
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_18, &__pyx_t_19, &__pyx_t_20);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17) < 0)) __Pyx_ErrFetch(&__pyx_t_15, &__pyx_t_16, &__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_15);
      __Pyx_XGOTREF(__pyx_t_16);
      __Pyx_XGOTREF(__pyx_t_17);
      __Pyx_XGOTREF(__pyx_t_18);
      __Pyx_XGOTREF(__pyx_t_19);
      __Pyx_XGOTREF(__pyx_t_20);
      __pyx_t_8 = __pyx_lineno; __pyx_t_9 = __pyx_clineno; __pyx_t_14 = __pyx_filename;
      {
        __pyx_v_self->_last_timestamp = __pyx_v_timestamp;
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_18);
        __Pyx_XGIVEREF(__pyx_t_19);
        __Pyx_XGIVEREF(__pyx_t_20);
        __Pyx_ExceptionReset(__pyx_t_18, __pyx_t_19, __pyx_t_20);
      }
      __Pyx_XGIVEREF(__pyx_t_15);
      __Pyx_XGIVEREF(__pyx_t_16);
      __Pyx_XGIVEREF(__pyx_t_17);
      __Pyx_ErrRestore(__pyx_t_15, __pyx_t_16, __pyx_t_17);
      __pyx_t_15 = 0; __pyx_t_16 = 0; __pyx_t_17 = 0; __pyx_t_18 = 0; __pyx_t_19 = 0; __pyx_t_20 = 0;
      __pyx_lineno = __pyx_t_8; __pyx_clineno = __pyx_t_9; __pyx_filename = __pyx_t_14;
      goto __pyx_L1_error;
    }
    __pyx_L5_return: {
      __pyx_t_20 = __pyx_r;
      __pyx_r = 0;
      __pyx_v_self->_last_timestamp = __pyx_v_timestamp;
      __pyx_r = __pyx_t_20;
      __pyx_t_20 = 0;
      goto __pyx_L0;
    }
    __pyx_L7:;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":593
 *         StrategyBase.c_stop(self, clock)
 * 
 *     cdef c_tick(self, double timestamp):             # <<<<<<<<<<<<<<
 *         StrategyBase.c_tick(self, timestamp)
 *         cdef:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_tick", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr9__pyx_v_mkt);
  __Pyx_XDECREF(__pyx_9genexpr10__pyx_v_mkt);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":645
 *             self._last_timestamp = timestamp
 * 
 *     def process_tick(self, timestamp: float):             # <<<<<<<<<<<<<<
 *         proposal = None
 *         # Trading is allowed
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_35process_tick(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_35process_tick = {"process_tick", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_35process_tick, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_35process_tick(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v_timestamp = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("process_tick (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_timestamp,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timestamp)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 645, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "process_tick") < 0)) __PYX_ERR(0, 645, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_timestamp = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("process_tick", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 645, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.process_tick", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_34process_tick(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_timestamp);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_34process_tick(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_timestamp) {
  PyObject *__pyx_v_proposal = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("process_tick", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":646
 * 
 *     def process_tick(self, timestamp: float):
 *         proposal = None             # <<<<<<<<<<<<<<
 *         # Trading is allowed
 *         if self._create_timestamp <= self._current_timestamp:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_proposal = Py_None;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":648
 *         proposal = None
 *         # Trading is allowed
 *         if self._create_timestamp <= self._current_timestamp:             # <<<<<<<<<<<<<<
 *             # 1. Calculate reservation price and optimal spread from gamma, alpha, kappa and volatility
 *             self.c_calculate_reservation_price_and_optimal_spread()
 */
  __pyx_t_1 = (__pyx_v_self->_create_timestamp <= __pyx_v_self->__pyx_base.__pyx_base._current_timestamp);
  if (__pyx_t_1) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":650
 *         if self._create_timestamp <= self._current_timestamp:
 *             # 1. Calculate reservation price and optimal spread from gamma, alpha, kappa and volatility
 *             self.c_calculate_reservation_price_and_optimal_spread()             # <<<<<<<<<<<<<<
 *             # 2. Check if calculated prices make sense
 *             if self._optimal_bid > 0 and self._optimal_ask > 0:
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_calculate_reservation_price_and_optimal_spread(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 650, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":652
 *             self.c_calculate_reservation_price_and_optimal_spread()
 *             # 2. Check if calculated prices make sense
 *             if self._optimal_bid > 0 and self._optimal_ask > 0:             # <<<<<<<<<<<<<<
 *                 # 3. Create base order proposals
 *                 proposal = self.c_create_base_proposal()
 */
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_self->_optimal_bid, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 652, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 652, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_3) {
    } else {
      __pyx_t_1 = __pyx_t_3;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = PyObject_RichCompare(__pyx_v_self->_optimal_ask, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 652, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 652, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_1 = __pyx_t_3;
    __pyx_L5_bool_binop_done:;
    if (__pyx_t_1) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":654
 *             if self._optimal_bid > 0 and self._optimal_ask > 0:
 *                 # 3. Create base order proposals
 *                 proposal = self.c_create_base_proposal()             # <<<<<<<<<<<<<<
 *                 # 4. Apply functions that modify orders amount
 *                 self.c_apply_order_amount_eta_transformation(proposal)
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_create_base_proposal(__pyx_v_self); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 654, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_proposal, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":656
 *                 proposal = self.c_create_base_proposal()
 *                 # 4. Apply functions that modify orders amount
 *                 self.c_apply_order_amount_eta_transformation(proposal)             # <<<<<<<<<<<<<<
 *                 # 5. Apply functions that modify orders price
 *                 self.c_apply_order_price_modifiers(proposal)
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_order_amount_eta_transformation(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 656, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":658
 *                 self.c_apply_order_amount_eta_transformation(proposal)
 *                 # 5. Apply functions that modify orders price
 *                 self.c_apply_order_price_modifiers(proposal)             # <<<<<<<<<<<<<<
 *                 # 6. Apply budget constraint, i.e. can't buy/sell more than what you have.
 *                 self.c_apply_budget_constraint(proposal)
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_order_price_modifiers(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 658, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":660
 *                 self.c_apply_order_price_modifiers(proposal)
 *                 # 6. Apply budget constraint, i.e. can't buy/sell more than what you have.
 *                 self.c_apply_budget_constraint(proposal)             # <<<<<<<<<<<<<<
 * 
 *                 self.c_cancel_active_orders(proposal)
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_budget_constraint(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 660, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":662
 *                 self.c_apply_budget_constraint(proposal)
 * 
 *                 self.c_cancel_active_orders(proposal)             # <<<<<<<<<<<<<<
 * 
 *         if self.c_to_create_orders(proposal):
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_cancel_active_orders(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 662, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":652
 *             self.c_calculate_reservation_price_and_optimal_spread()
 *             # 2. Check if calculated prices make sense
 *             if self._optimal_bid > 0 and self._optimal_ask > 0:             # <<<<<<<<<<<<<<
 *                 # 3. Create base order proposals
 *                 proposal = self.c_create_base_proposal()
 */
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":648
 *         proposal = None
 *         # Trading is allowed
 *         if self._create_timestamp <= self._current_timestamp:             # <<<<<<<<<<<<<<
 *             # 1. Calculate reservation price and optimal spread from gamma, alpha, kappa and volatility
 *             self.c_calculate_reservation_price_and_optimal_spread()
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":664
 *                 self.c_cancel_active_orders(proposal)
 * 
 *         if self.c_to_create_orders(proposal):             # <<<<<<<<<<<<<<
 *             self.c_execute_orders_proposal(proposal)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_to_create_orders(__pyx_v_self, __pyx_v_proposal); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 664, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":665
 * 
 *         if self.c_to_create_orders(proposal):
 *             self.c_execute_orders_proposal(proposal)             # <<<<<<<<<<<<<<
 * 
 *         if self._is_debug:
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_execute_orders_proposal(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 665, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":664
 *                 self.c_cancel_active_orders(proposal)
 * 
 *         if self.c_to_create_orders(proposal):             # <<<<<<<<<<<<<<
 *             self.c_execute_orders_proposal(proposal)
 * 
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":667
 *             self.c_execute_orders_proposal(proposal)
 * 
 *         if self._is_debug:             # <<<<<<<<<<<<<<
 *             self.dump_debug_variables()
 * 
 */
  if (__pyx_v_self->_is_debug) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":668
 * 
 *         if self._is_debug:
 *             self.dump_debug_variables()             # <<<<<<<<<<<<<<
 * 
 *     cdef c_collect_market_variables(self, double timestamp):
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dump_debug_variables); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 668, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 668, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":667
 *             self.c_execute_orders_proposal(proposal)
 * 
 *         if self._is_debug:             # <<<<<<<<<<<<<<
 *             self.dump_debug_variables()
 * 
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":645
 *             self._last_timestamp = timestamp
 * 
 *     def process_tick(self, timestamp: float):             # <<<<<<<<<<<<<<
 *         proposal = None
 *         # Trading is allowed
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.process_tick", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_proposal);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":670
 *             self.dump_debug_variables()
 * 
 *     cdef c_collect_market_variables(self, double timestamp):             # <<<<<<<<<<<<<<
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         self._last_sampling_timestamp = timestamp
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_collect_market_variables(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, double __pyx_v_timestamp) {
  PyObject *__pyx_v_market = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_trading_pair = NULL;
  PyObject *__pyx_v_base_asset = NULL;
  PyObject *__pyx_v_quote_asset = NULL;
  PyObject *__pyx_v_price = NULL;
  PyObject *__pyx_v_base_balance = NULL;
  PyObject *__pyx_v_quote_balance = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_inventory_in_base = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *(*__pyx_t_7)(PyObject *);
  unsigned int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_collect_market_variables", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":671
 * 
 *     cdef c_collect_market_variables(self, double timestamp):
 *         market, trading_pair, base_asset, quote_asset = self._market_info             # <<<<<<<<<<<<<<
 *         self._last_sampling_timestamp = timestamp
 * 
 */
  __pyx_t_1 = __pyx_v_self->_market_info;
  __Pyx_INCREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 4)) {
      if (size > 4) __Pyx_RaiseTooManyValuesError(4);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 671, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 3); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_4 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 3); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[4] = {&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5};
      for (i=0; i < 4; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 671, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[4] = {&__pyx_t_2,&__pyx_t_3,&__pyx_t_4,&__pyx_t_5};
    __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 671, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6);
    for (index=0; index < 4; index++) {
      PyObject* item = __pyx_t_7(__pyx_t_6); if (unlikely(!item)) goto __pyx_L3_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_7(__pyx_t_6), 4) < 0) __PYX_ERR(0, 671, __pyx_L1_error)
    __pyx_t_7 = NULL;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_7 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 671, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_market = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_trading_pair = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_v_base_asset = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_quote_asset = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":672
 *     cdef c_collect_market_variables(self, double timestamp):
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         self._last_sampling_timestamp = timestamp             # <<<<<<<<<<<<<<
 * 
 *         price = self.get_price()
 */
  __pyx_v_self->_last_sampling_timestamp = __pyx_v_timestamp;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":674
 *         self._last_sampling_timestamp = timestamp
 * 
 *         price = self.get_price()             # <<<<<<<<<<<<<<
 *         self._avg_vol.add_sample(price)
 *         self._trading_intensity.calculate(timestamp)
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_price); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 674, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 0+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 674, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_v_price = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":675
 * 
 *         price = self.get_price()
 *         self._avg_vol.add_sample(price)             # <<<<<<<<<<<<<<
 *         self._trading_intensity.calculate(timestamp)
 *         # Calculate adjustment factor to have 0.01% of inventory resolution
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_avg_vol, __pyx_n_s_add_sample); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 675, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_price};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 675, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":676
 *         price = self.get_price()
 *         self._avg_vol.add_sample(price)
 *         self._trading_intensity.calculate(timestamp)             # <<<<<<<<<<<<<<
 *         # Calculate adjustment factor to have 0.01% of inventory resolution
 *         base_balance = market.get_balance(base_asset)
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_trading_intensity), __pyx_n_s_calculate); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = PyFloat_FromDouble(__pyx_v_timestamp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 676, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":678
 *         self._trading_intensity.calculate(timestamp)
 *         # Calculate adjustment factor to have 0.01% of inventory resolution
 *         base_balance = market.get_balance(base_asset)             # <<<<<<<<<<<<<<
 *         quote_balance = market.get_balance(quote_asset)
 *         inventory_in_base = quote_balance / price + base_balance
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_market, __pyx_n_s_get_balance); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_base_asset};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 678, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_v_base_balance = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":679
 *         # Calculate adjustment factor to have 0.01% of inventory resolution
 *         base_balance = market.get_balance(base_asset)
 *         quote_balance = market.get_balance(quote_asset)             # <<<<<<<<<<<<<<
 *         inventory_in_base = quote_balance / price + base_balance
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_market, __pyx_n_s_get_balance); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 679, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = NULL;
  __pyx_t_8 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
      __pyx_t_8 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_quote_asset};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_v_quote_balance = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":680
 *         base_balance = market.get_balance(base_asset)
 *         quote_balance = market.get_balance(quote_asset)
 *         inventory_in_base = quote_balance / price + base_balance             # <<<<<<<<<<<<<<
 * 
 *     def collect_market_variables(self, timestamp: float):
 */
  __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_v_quote_balance, __pyx_v_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyNumber_Add(__pyx_t_1, __pyx_v_base_balance); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 680, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_inventory_in_base = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":670
 *             self.dump_debug_variables()
 * 
 *     cdef c_collect_market_variables(self, double timestamp):             # <<<<<<<<<<<<<<
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         self._last_sampling_timestamp = timestamp
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_collect_market_variables", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_trading_pair);
  __Pyx_XDECREF(__pyx_v_base_asset);
  __Pyx_XDECREF(__pyx_v_quote_asset);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XDECREF(__pyx_v_base_balance);
  __Pyx_XDECREF(__pyx_v_quote_balance);
  __Pyx_XDECREF(__pyx_v_inventory_in_base);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":682
 *         inventory_in_base = quote_balance / price + base_balance
 * 
 *     def collect_market_variables(self, timestamp: float):             # <<<<<<<<<<<<<<
 *         self.c_collect_market_variables(timestamp)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_37collect_market_variables(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_37collect_market_variables = {"collect_market_variables", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_37collect_market_variables, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_37collect_market_variables(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_timestamp = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("collect_market_variables (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_timestamp,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_timestamp)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 682, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "collect_market_variables") < 0)) __PYX_ERR(0, 682, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_timestamp = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("collect_market_variables", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 682, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.collect_market_variables", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_36collect_market_variables(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_timestamp);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_36collect_market_variables(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_timestamp) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("collect_market_variables", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":683
 * 
 *     def collect_market_variables(self, timestamp: float):
 *         self.c_collect_market_variables(timestamp)             # <<<<<<<<<<<<<<
 * 
 *     cdef double c_get_spread(self):
 */
  __pyx_t_1 = __pyx_PyFloat_AsDouble(__pyx_v_timestamp); if (unlikely((__pyx_t_1 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 683, __pyx_L1_error)
  __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_collect_market_variables(__pyx_v_self, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 683, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":682
 *         inventory_in_base = quote_balance / price + base_balance
 * 
 *     def collect_market_variables(self, timestamp: float):             # <<<<<<<<<<<<<<
 *         self.c_collect_market_variables(timestamp)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.collect_market_variables", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":685
 *         self.c_collect_market_variables(timestamp)
 * 
 *     cdef double c_get_spread(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

static double __pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_get_spread(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_trading_pair = 0;
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_get_spread", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":687
 *     cdef double c_get_spread(self):
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *             str trading_pair = self._market_info.trading_pair
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 687, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":688
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 *             str trading_pair = self._market_info.trading_pair             # <<<<<<<<<<<<<<
 * 
 *         return market.c_get_price(trading_pair, True) - market.c_get_price(trading_pair, False)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 688, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 688, __pyx_L1_error)
  __pyx_v_trading_pair = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":690
 *             str trading_pair = self._market_info.trading_pair
 * 
 *         return market.c_get_price(trading_pair, True) - market.c_get_price(trading_pair, False)             # <<<<<<<<<<<<<<
 * 
 *     def get_spread(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_get_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), __pyx_v_trading_pair, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_get_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), __pyx_v_trading_pair, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Subtract(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 690, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":685
 *         self.c_collect_market_variables(timestamp)
 * 
 *     cdef double c_get_spread(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_get_spread", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_trading_pair);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":692
 *         return market.c_get_price(trading_pair, True) - market.c_get_price(trading_pair, False)
 * 
 *     def get_spread(self):             # <<<<<<<<<<<<<<
 *         return self.c_get_spread()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_39get_spread(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_39get_spread = {"get_spread", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_39get_spread, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_39get_spread(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_spread (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_spread", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_spread", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_38get_spread(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_38get_spread(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_spread", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":693
 * 
 *     def get_spread(self):
 *         return self.c_get_spread()             # <<<<<<<<<<<<<<
 * 
 *     def get_volatility(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_get_spread(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 693, __pyx_L1_error)
  __pyx_t_2 = PyFloat_FromDouble(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 693, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":692
 *         return market.c_get_price(trading_pair, True) - market.c_get_price(trading_pair, False)
 * 
 *     def get_spread(self):             # <<<<<<<<<<<<<<
 *         return self.c_get_spread()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.get_spread", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":695
 *         return self.c_get_spread()
 * 
 *     def get_volatility(self):             # <<<<<<<<<<<<<<
 *         vol = Decimal(str(self._avg_vol.current_value))
 *         return vol
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_41get_volatility(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_41get_volatility = {"get_volatility", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_41get_volatility, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_41get_volatility(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_volatility (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("get_volatility", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "get_volatility", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_40get_volatility(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_40get_volatility(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_v_vol = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_volatility", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":696
 * 
 *     def get_volatility(self):
 *         vol = Decimal(str(self._avg_vol.current_value))             # <<<<<<<<<<<<<<
 *         return vol
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_avg_vol, __pyx_n_s_current_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_Unicode(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_vol = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":697
 *     def get_volatility(self):
 *         vol = Decimal(str(self._avg_vol.current_value))
 *         return vol             # <<<<<<<<<<<<<<
 * 
 *     cdef c_measure_order_book_liquidity(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_vol);
  __pyx_r = __pyx_v_vol;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":695
 *         return self.c_get_spread()
 * 
 *     def get_volatility(self):             # <<<<<<<<<<<<<<
 *         vol = Decimal(str(self._avg_vol.current_value))
 *         return vol
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.get_volatility", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_vol);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":699
 *         return vol
 * 
 *     cdef c_measure_order_book_liquidity(self):             # <<<<<<<<<<<<<<
 * 
 *         self._alpha, self._kappa = self._trading_intensity.current_value
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_measure_order_book_liquidity(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *(*__pyx_t_5)(PyObject *);
  unsigned int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_UCS4 __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_measure_order_book_liquidity", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":701
 *     cdef c_measure_order_book_liquidity(self):
 * 
 *         self._alpha, self._kappa = self._trading_intensity.current_value             # <<<<<<<<<<<<<<
 * 
 *         self._alpha = Decimal(self._alpha)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_trading_intensity), __pyx_n_s_current_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 701, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 701, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_4 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 701, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4);
    index = 0; __pyx_t_2 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_5(__pyx_t_4); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_5(__pyx_t_4), 2) < 0) __PYX_ERR(0, 701, __pyx_L1_error)
    __pyx_t_5 = NULL;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 701, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->_alpha);
  __Pyx_DECREF(__pyx_v_self->_alpha);
  __pyx_v_self->_alpha = __pyx_t_2;
  __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_kappa);
  __Pyx_DECREF(__pyx_v_self->_kappa);
  __pyx_v_self->_kappa = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":703
 *         self._alpha, self._kappa = self._trading_intensity.current_value
 * 
 *         self._alpha = Decimal(self._alpha)             # <<<<<<<<<<<<<<
 *         self._kappa = Decimal(self._kappa)
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_self->_alpha};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 703, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_alpha);
  __Pyx_DECREF(__pyx_v_self->_alpha);
  __pyx_v_self->_alpha = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":704
 * 
 *         self._alpha = Decimal(self._alpha)
 *         self._kappa = Decimal(self._kappa)             # <<<<<<<<<<<<<<
 * 
 *         if self._is_debug:
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 704, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_self->_kappa};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 704, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_kappa);
  __Pyx_DECREF(__pyx_v_self->_kappa);
  __pyx_v_self->_kappa = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":706
 *         self._kappa = Decimal(self._kappa)
 * 
 *         if self._is_debug:             # <<<<<<<<<<<<<<
 *             self.logger().info(f"alpha={self._alpha:.4f} | "
 *                                f"kappa={self._kappa:.4f}")
 */
  if (__pyx_v_self->_is_debug) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":707
 * 
 *         if self._is_debug:
 *             self.logger().info(f"alpha={self._alpha:.4f} | "             # <<<<<<<<<<<<<<
 *                                f"kappa={self._kappa:.4f}")
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 0+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_info); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyTuple_New(4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_7 = 0;
    __pyx_t_8 = 127;
    __Pyx_INCREF(__pyx_kp_u_alpha);
    __pyx_t_7 += 6;
    __Pyx_GIVEREF(__pyx_kp_u_alpha);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_alpha);
    __pyx_t_4 = __Pyx_PyObject_Format(__pyx_v_self->_alpha, __pyx_kp_u_4f); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_8;
    __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_4);
    __pyx_t_4 = 0;
    __Pyx_INCREF(__pyx_kp_u_kappa);
    __pyx_t_7 += 9;
    __Pyx_GIVEREF(__pyx_kp_u_kappa);
    PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u_kappa);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":708
 *         if self._is_debug:
 *             self.logger().info(f"alpha={self._alpha:.4f} | "
 *                                f"kappa={self._kappa:.4f}")             # <<<<<<<<<<<<<<
 * 
 *     def measure_order_book_liquidity(self):
 */
    __pyx_t_4 = __Pyx_PyObject_Format(__pyx_v_self->_kappa, __pyx_kp_u_4f); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 708, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_8) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_8;
    __pyx_t_7 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":707
 * 
 *         if self._is_debug:
 *             self.logger().info(f"alpha={self._alpha:.4f} | "             # <<<<<<<<<<<<<<
 *                                f"kappa={self._kappa:.4f}")
 * 
 */
    __pyx_t_4 = __Pyx_PyUnicode_Join(__pyx_t_3, 4, __pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 707, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 707, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":706
 *         self._kappa = Decimal(self._kappa)
 * 
 *         if self._is_debug:             # <<<<<<<<<<<<<<
 *             self.logger().info(f"alpha={self._alpha:.4f} | "
 *                                f"kappa={self._kappa:.4f}")
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":699
 *         return vol
 * 
 *     cdef c_measure_order_book_liquidity(self):             # <<<<<<<<<<<<<<
 * 
 *         self._alpha, self._kappa = self._trading_intensity.current_value
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_measure_order_book_liquidity", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":710
 *                                f"kappa={self._kappa:.4f}")
 * 
 *     def measure_order_book_liquidity(self):             # <<<<<<<<<<<<<<
 *         return self.c_measure_order_book_liquidity()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_43measure_order_book_liquidity(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_43measure_order_book_liquidity = {"measure_order_book_liquidity", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_43measure_order_book_liquidity, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_43measure_order_book_liquidity(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("measure_order_book_liquidity (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("measure_order_book_liquidity", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "measure_order_book_liquidity", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_42measure_order_book_liquidity(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_42measure_order_book_liquidity(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("measure_order_book_liquidity", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":711
 * 
 *     def measure_order_book_liquidity(self):
 *         return self.c_measure_order_book_liquidity()             # <<<<<<<<<<<<<<
 * 
 *     cdef c_calculate_reservation_price_and_optimal_spread(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_measure_order_book_liquidity(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 711, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":710
 *                                f"kappa={self._kappa:.4f}")
 * 
 *     def measure_order_book_liquidity(self):             # <<<<<<<<<<<<<<
 *         return self.c_measure_order_book_liquidity()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.measure_order_book_liquidity", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":713
 *         return self.c_measure_order_book_liquidity()
 * 
 *     cdef c_calculate_reservation_price_and_optimal_spread(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_calculate_reservation_price_and_optimal_spread(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_price = NULL;
  PyObject *__pyx_v_inventory = NULL;
  PyObject *__pyx_v_q_target = NULL;
  PyObject *__pyx_v_q = NULL;
  PyObject *__pyx_v_vol = NULL;
  PyObject *__pyx_v_time_left_fraction = NULL;
  PyObject *__pyx_v_min_spread = NULL;
  PyObject *__pyx_v_max_limit_bid = NULL;
  PyObject *__pyx_v_min_limit_ask = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  Py_UCS4 __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_calculate_reservation_price_and_optimal_spread", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":715
 *     cdef c_calculate_reservation_price_and_optimal_spread(self):
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 * 
 *         # Current mid price
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 715, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 715, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":718
 * 
 *         # Current mid price
 *         price = self.get_price()             # <<<<<<<<<<<<<<
 * 
 *         # The amount of stocks owned - q - has to be in relative units, not absolute, because changing the portfolio size shouldn't change the reservation price
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 718, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_price = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":722
 *         # The amount of stocks owned - q - has to be in relative units, not absolute, because changing the portfolio size shouldn't change the reservation price
 *         # The reservation price should concern itself only with the strategy performance, i.e. amount of stocks relative to the target
 *         inventory = Decimal(str(self.c_calculate_inventory()))             # <<<<<<<<<<<<<<
 *         if inventory == 0:
 *             return
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_calculate_inventory(__pyx_v_self); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_Unicode(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 722, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 722, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_inventory = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":723
 *         # The reservation price should concern itself only with the strategy performance, i.e. amount of stocks relative to the target
 *         inventory = Decimal(str(self.c_calculate_inventory()))
 *         if inventory == 0:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_6 = (__Pyx_PyInt_BoolEqObjC(__pyx_v_inventory, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 723, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":724
 *         inventory = Decimal(str(self.c_calculate_inventory()))
 *         if inventory == 0:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         q_target = Decimal(str(self.c_calculate_target_inventory()))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":723
 *         # The reservation price should concern itself only with the strategy performance, i.e. amount of stocks relative to the target
 *         inventory = Decimal(str(self.c_calculate_inventory()))
 *         if inventory == 0:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":726
 *             return
 * 
 *         q_target = Decimal(str(self.c_calculate_target_inventory()))             # <<<<<<<<<<<<<<
 *         q = (market.get_balance(self.base_asset) - q_target) / (inventory)
 *         # Volatility has to be in absolute values (prices) because in calculation of reservation price it's not multiplied by the current price, therefore
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_calculate_target_inventory(__pyx_v_self); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_3 = __Pyx_PyObject_Unicode(__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 726, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_5 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 726, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_q_target = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":727
 * 
 *         q_target = Decimal(str(self.c_calculate_target_inventory()))
 *         q = (market.get_balance(self.base_asset) - q_target) / (inventory)             # <<<<<<<<<<<<<<
 *         # Volatility has to be in absolute values (prices) because in calculation of reservation price it's not multiplied by the current price, therefore
 *         # it can't be a percentage. The result of the multiplication has to be an absolute price value because it's being subtracted from the current price
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_market), __pyx_n_s_get_balance); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base_asset); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 727, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = PyNumber_Subtract(__pyx_t_1, __pyx_v_q_target); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_2, __pyx_v_inventory); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_q = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":730
 *         # Volatility has to be in absolute values (prices) because in calculation of reservation price it's not multiplied by the current price, therefore
 *         # it can't be a percentage. The result of the multiplication has to be an absolute price value because it's being subtracted from the current price
 *         vol = self.get_volatility()             # <<<<<<<<<<<<<<
 * 
 *         # order book liquidity - kappa and alpha have to represent absolute values because the second member of the optimal spread equation has to be an absolute price
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_volatility); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 730, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 730, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_vol = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":734
 *         # order book liquidity - kappa and alpha have to represent absolute values because the second member of the optimal spread equation has to be an absolute price
 *         # and from the reservation price calculation we know that gamma's unit is not absolute price
 *         if all((self.gamma, self._kappa)) and self._alpha != 0 and self._kappa > 0 and vol != 0:             # <<<<<<<<<<<<<<
 *             if self._execution_state.time_left is not None and self._execution_state.closing_time is not None:
 *                 # Avellaneda-Stoikov for a fixed timespan
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_gamma); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1)) __PYX_ERR(0, 734, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_kappa);
  __Pyx_GIVEREF(__pyx_v_self->_kappa);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_self->_kappa)) __PYX_ERR(0, 734, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_all, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 734, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 734, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_7 = (__Pyx_PyInt_BoolNeObjC(__pyx_v_self->_alpha, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 734, __pyx_L1_error)
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_self->_kappa, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 734, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 734, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_7 = (__Pyx_PyInt_BoolNeObjC(__pyx_v_vol, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 734, __pyx_L1_error)
  __pyx_t_6 = __pyx_t_7;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_6) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":735
 *         # and from the reservation price calculation we know that gamma's unit is not absolute price
 *         if all((self.gamma, self._kappa)) and self._alpha != 0 and self._kappa > 0 and vol != 0:
 *             if self._execution_state.time_left is not None and self._execution_state.closing_time is not None:             # <<<<<<<<<<<<<<
 *                 # Avellaneda-Stoikov for a fixed timespan
 *                 time_left_fraction = Decimal(str(self._execution_state.time_left / self._execution_state.closing_time))
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_execution_state, __pyx_n_s_time_left); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 735, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = (__pyx_t_1 != Py_None);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L10_bool_binop_done;
    }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_execution_state, __pyx_n_s_closing_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 735, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = (__pyx_t_1 != Py_None);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __pyx_t_7;
    __pyx_L10_bool_binop_done:;
    if (__pyx_t_6) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":737
 *             if self._execution_state.time_left is not None and self._execution_state.closing_time is not None:
 *                 # Avellaneda-Stoikov for a fixed timespan
 *                 time_left_fraction = Decimal(str(self._execution_state.time_left / self._execution_state.closing_time))             # <<<<<<<<<<<<<<
 *             else:
 *                 # Avellaneda-Stoikov for an infinite timespan
 */
      __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 737, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_execution_state, __pyx_n_s_time_left); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 737, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_execution_state, __pyx_n_s_closing_time); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 737, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 737, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Unicode(__pyx_t_8); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 737, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_5};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 737, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __pyx_v_time_left_fraction = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":735
 *         # and from the reservation price calculation we know that gamma's unit is not absolute price
 *         if all((self.gamma, self._kappa)) and self._alpha != 0 and self._kappa > 0 and vol != 0:
 *             if self._execution_state.time_left is not None and self._execution_state.closing_time is not None:             # <<<<<<<<<<<<<<
 *                 # Avellaneda-Stoikov for a fixed timespan
 *                 time_left_fraction = Decimal(str(self._execution_state.time_left / self._execution_state.closing_time))
 */
      goto __pyx_L9;
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":746
 *                 # The proposed solution is to use the same equation as for the constrained timespan but with
 *                 # a fixed time left
 *                 time_left_fraction = 1             # <<<<<<<<<<<<<<
 * 
 *             # Here seems to be another mistake in the paper
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_int_1);
      __pyx_v_time_left_fraction = __pyx_int_1;
    }
    __pyx_L9:;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":755
 *             # This leads to normalization of the risk_factor and will guaranetee consistent behavior on all price ranges of the asset, and across assets
 * 
 *             self._reservation_price = price - (q * self.gamma * vol * time_left_fraction)             # <<<<<<<<<<<<<<
 * 
 *             self._optimal_spread = self.gamma * vol * time_left_fraction
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_gamma); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 755, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyNumber_Multiply(__pyx_v_q, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 755, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Multiply(__pyx_t_2, __pyx_v_vol); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 755, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_Multiply(__pyx_t_1, __pyx_v_time_left_fraction); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 755, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Subtract(__pyx_v_price, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 755, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->_reservation_price);
    __Pyx_DECREF(__pyx_v_self->_reservation_price);
    __pyx_v_self->_reservation_price = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":757
 *             self._reservation_price = price - (q * self.gamma * vol * time_left_fraction)
 * 
 *             self._optimal_spread = self.gamma * vol * time_left_fraction             # <<<<<<<<<<<<<<
 *             self._optimal_spread += 2 * Decimal(1 + self.gamma / self._kappa).ln() / self.gamma
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_gamma); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyNumber_Multiply(__pyx_t_1, __pyx_v_vol); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_Multiply(__pyx_t_2, __pyx_v_time_left_fraction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 757, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->_optimal_spread);
    __Pyx_DECREF(__pyx_v_self->_optimal_spread);
    __pyx_v_self->_optimal_spread = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":758
 * 
 *             self._optimal_spread = self.gamma * vol * time_left_fraction
 *             self._optimal_spread += 2 * Decimal(1 + self.gamma / self._kappa).ln() / self.gamma             # <<<<<<<<<<<<<<
 * 
 *             min_spread = price / 100 * Decimal(str(self._config_map.min_spread))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 758, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_gamma); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 758, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_3 = __Pyx_PyNumber_Divide(__pyx_t_8, __pyx_v_self->_kappa); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 758, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyInt_AddCObj(__pyx_int_1, __pyx_t_3, 1, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 758, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_8};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 758, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_ln); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 758, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 758, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_5 = __Pyx_PyInt_MultiplyCObj(__pyx_int_2, __pyx_t_1, 2, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 758, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_gamma); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 758, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 758, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_self->_optimal_spread, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 758, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->_optimal_spread);
    __Pyx_DECREF(__pyx_v_self->_optimal_spread);
    __pyx_v_self->_optimal_spread = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":760
 *             self._optimal_spread += 2 * Decimal(1 + self.gamma / self._kappa).ln() / self.gamma
 * 
 *             min_spread = price / 100 * Decimal(str(self._config_map.min_spread))             # <<<<<<<<<<<<<<
 * 
 *             max_limit_bid = price - min_spread / 2
 */
    __pyx_t_1 = __Pyx_PyInt_TrueDivideObjC(__pyx_v_price, __pyx_int_100, 0x64, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 760, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 760, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_min_spread); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 760, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_3 = __Pyx_PyObject_Unicode(__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 760, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_3};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 760, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_5 = PyNumber_Multiply(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 760, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_min_spread = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":762
 *             min_spread = price / 100 * Decimal(str(self._config_map.min_spread))
 * 
 *             max_limit_bid = price - min_spread / 2             # <<<<<<<<<<<<<<
 *             min_limit_ask = price + min_spread / 2
 * 
 */
    __pyx_t_5 = __Pyx_PyInt_TrueDivideObjC(__pyx_v_min_spread, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 762, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = PyNumber_Subtract(__pyx_v_price, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 762, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_max_limit_bid = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":763
 * 
 *             max_limit_bid = price - min_spread / 2
 *             min_limit_ask = price + min_spread / 2             # <<<<<<<<<<<<<<
 * 
 *             self._optimal_ask = max(self._reservation_price + self._optimal_spread / 2, min_limit_ask)
 */
    __pyx_t_2 = __Pyx_PyInt_TrueDivideObjC(__pyx_v_min_spread, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 763, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyNumber_Add(__pyx_v_price, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 763, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_min_limit_ask = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":765
 *             min_limit_ask = price + min_spread / 2
 * 
 *             self._optimal_ask = max(self._reservation_price + self._optimal_spread / 2, min_limit_ask)             # <<<<<<<<<<<<<<
 *             self._optimal_bid = min(self._reservation_price - self._optimal_spread / 2, max_limit_bid)
 * 
 */
    __Pyx_INCREF(__pyx_v_min_limit_ask);
    __pyx_t_5 = __pyx_v_min_limit_ask;
    __pyx_t_2 = __Pyx_PyInt_TrueDivideObjC(__pyx_v_self->_optimal_spread, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 765, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyNumber_Add(__pyx_v_self->_reservation_price, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 765, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_5, __pyx_t_1, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 765, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 765, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_6) {
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_2 = __pyx_t_5;
    } else {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_2 = __pyx_t_1;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __pyx_t_2;
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GIVEREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_v_self->_optimal_ask);
    __Pyx_DECREF(__pyx_v_self->_optimal_ask);
    __pyx_v_self->_optimal_ask = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":766
 * 
 *             self._optimal_ask = max(self._reservation_price + self._optimal_spread / 2, min_limit_ask)
 *             self._optimal_bid = min(self._reservation_price - self._optimal_spread / 2, max_limit_bid)             # <<<<<<<<<<<<<<
 * 
 *             # This is not what the algorithm will use as proposed bid and ask. This is just the raw output.
 */
    __Pyx_INCREF(__pyx_v_max_limit_bid);
    __pyx_t_5 = __pyx_v_max_limit_bid;
    __pyx_t_2 = __Pyx_PyInt_TrueDivideObjC(__pyx_v_self->_optimal_spread, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyNumber_Subtract(__pyx_v_self->_reservation_price, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_5, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 766, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 766, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_6) {
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_2 = __pyx_t_5;
    } else {
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_2 = __pyx_t_1;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __pyx_t_2;
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GIVEREF(__pyx_t_5);
    __Pyx_GOTREF(__pyx_v_self->_optimal_bid);
    __Pyx_DECREF(__pyx_v_self->_optimal_bid);
    __pyx_v_self->_optimal_bid = __pyx_t_5;
    __pyx_t_5 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":770
 *             # This is not what the algorithm will use as proposed bid and ask. This is just the raw output.
 *             # Optimal bid and optimal ask prices will be used
 *             if self._is_debug:             # <<<<<<<<<<<<<<
 *                 self.logger().info(f"q={q:.4f} | "
 *                                    f"vol={vol:.10f}")
 */
    if (__pyx_v_self->_is_debug) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":771
 *             # Optimal bid and optimal ask prices will be used
 *             if self._is_debug:
 *                 self.logger().info(f"q={q:.4f} | "             # <<<<<<<<<<<<<<
 *                                    f"vol={vol:.10f}")
 *                 self.logger().info(f"mid_price={price:.10f} | "
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 771, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 771, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 771, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 771, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_9 = 0;
      __pyx_t_10 = 127;
      __Pyx_INCREF(__pyx_kp_u_q);
      __pyx_t_9 += 2;
      __Pyx_GIVEREF(__pyx_kp_u_q);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u_q);
      __pyx_t_3 = __Pyx_PyObject_Format(__pyx_v_q, __pyx_kp_u_4f); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 771, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_INCREF(__pyx_kp_u_vol);
      __pyx_t_9 += 7;
      __Pyx_GIVEREF(__pyx_kp_u_vol);
      PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_vol);

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":772
 *             if self._is_debug:
 *                 self.logger().info(f"q={q:.4f} | "
 *                                    f"vol={vol:.10f}")             # <<<<<<<<<<<<<<
 *                 self.logger().info(f"mid_price={price:.10f} | "
 *                                    f"reservation_price={self._reservation_price:.10f} | "
 */
      __pyx_t_3 = __Pyx_PyObject_Format(__pyx_v_vol, __pyx_kp_u_10f); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 772, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_3) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":771
 *             # Optimal bid and optimal ask prices will be used
 *             if self._is_debug:
 *                 self.logger().info(f"q={q:.4f} | "             # <<<<<<<<<<<<<<
 *                                    f"vol={vol:.10f}")
 *                 self.logger().info(f"mid_price={price:.10f} | "
 */
      __pyx_t_3 = __Pyx_PyUnicode_Join(__pyx_t_2, 4, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 771, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_3};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 771, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":773
 *                 self.logger().info(f"q={q:.4f} | "
 *                                    f"vol={vol:.10f}")
 *                 self.logger().info(f"mid_price={price:.10f} | "             # <<<<<<<<<<<<<<
 *                                    f"reservation_price={self._reservation_price:.10f} | "
 *                                    f"optimal_spread={self._optimal_spread:.10f}")
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_2)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_2);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 773, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_info); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyTuple_New(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = 0;
      __pyx_t_10 = 127;
      __Pyx_INCREF(__pyx_kp_u_mid_price);
      __pyx_t_9 += 10;
      __Pyx_GIVEREF(__pyx_kp_u_mid_price);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u_mid_price);
      __pyx_t_2 = __Pyx_PyObject_Format(__pyx_v_price, __pyx_kp_u_10f); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_INCREF(__pyx_kp_u_reservation_price);
      __pyx_t_9 += 21;
      __Pyx_GIVEREF(__pyx_kp_u_reservation_price);
      PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_reservation_price);

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":774
 *                                    f"vol={vol:.10f}")
 *                 self.logger().info(f"mid_price={price:.10f} | "
 *                                    f"reservation_price={self._reservation_price:.10f} | "             # <<<<<<<<<<<<<<
 *                                    f"optimal_spread={self._optimal_spread:.10f}")
 *                 self.logger().info(f"optimal_bid={(price-(self._reservation_price - self._optimal_spread / 2)) / price * 100:.4f}% | "
 */
      __pyx_t_2 = __Pyx_PyObject_Format(__pyx_v_self->_reservation_price, __pyx_kp_u_10f); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 774, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_INCREF(__pyx_kp_u_optimal_spread);
      __pyx_t_9 += 18;
      __Pyx_GIVEREF(__pyx_kp_u_optimal_spread);
      PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_optimal_spread);

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":775
 *                 self.logger().info(f"mid_price={price:.10f} | "
 *                                    f"reservation_price={self._reservation_price:.10f} | "
 *                                    f"optimal_spread={self._optimal_spread:.10f}")             # <<<<<<<<<<<<<<
 *                 self.logger().info(f"optimal_bid={(price-(self._reservation_price - self._optimal_spread / 2)) / price * 100:.4f}% | "
 *                                    f"optimal_ask={((self._reservation_price + self._optimal_spread / 2) - price) / price * 100:.4f}%")
 */
      __pyx_t_2 = __Pyx_PyObject_Format(__pyx_v_self->_optimal_spread, __pyx_kp_u_10f); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 775, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":773
 *                 self.logger().info(f"q={q:.4f} | "
 *                                    f"vol={vol:.10f}")
 *                 self.logger().info(f"mid_price={price:.10f} | "             # <<<<<<<<<<<<<<
 *                                    f"reservation_price={self._reservation_price:.10f} | "
 *                                    f"optimal_spread={self._optimal_spread:.10f}")
 */
      __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_1, 6, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 773, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_t_2};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 773, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":776
 *                                    f"reservation_price={self._reservation_price:.10f} | "
 *                                    f"optimal_spread={self._optimal_spread:.10f}")
 *                 self.logger().info(f"optimal_bid={(price-(self._reservation_price - self._optimal_spread / 2)) / price * 100:.4f}% | "             # <<<<<<<<<<<<<<
 *                                    f"optimal_ask={((self._reservation_price + self._optimal_spread / 2) - price) / price * 100:.4f}%")
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 776, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_1 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
        __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 776, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_info); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 776, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyTuple_New(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 776, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = 0;
      __pyx_t_10 = 127;
      __Pyx_INCREF(__pyx_kp_u_optimal_bid);
      __pyx_t_9 += 12;
      __Pyx_GIVEREF(__pyx_kp_u_optimal_bid);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u_optimal_bid);
      __pyx_t_1 = __Pyx_PyInt_TrueDivideObjC(__pyx_v_self->_optimal_spread, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 776, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_8 = PyNumber_Subtract(__pyx_v_self->_reservation_price, __pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 776, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = PyNumber_Subtract(__pyx_v_price, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 776, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyNumber_Divide(__pyx_t_1, __pyx_v_price); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 776, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyInt_MultiplyObjC(__pyx_t_8, __pyx_int_100, 0x64, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 776, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_Format(__pyx_t_1, __pyx_kp_u_4f); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 776, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_INCREF(__pyx_kp_u_optimal_ask);
      __pyx_t_9 += 16;
      __Pyx_GIVEREF(__pyx_kp_u_optimal_ask);
      PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u_optimal_ask);

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":777
 *                                    f"optimal_spread={self._optimal_spread:.10f}")
 *                 self.logger().info(f"optimal_bid={(price-(self._reservation_price - self._optimal_spread / 2)) / price * 100:.4f}% | "
 *                                    f"optimal_ask={((self._reservation_price + self._optimal_spread / 2) - price) / price * 100:.4f}%")             # <<<<<<<<<<<<<<
 * 
 *     def calculate_reservation_price_and_optimal_spread(self):
 */
      __pyx_t_8 = __Pyx_PyInt_TrueDivideObjC(__pyx_v_self->_optimal_spread, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 777, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = PyNumber_Add(__pyx_v_self->_reservation_price, __pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 777, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = PyNumber_Subtract(__pyx_t_1, __pyx_v_price); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 777, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_8, __pyx_v_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 777, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyInt_MultiplyObjC(__pyx_t_1, __pyx_int_100, 0x64, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 777, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_PyObject_Format(__pyx_t_8, __pyx_kp_u_4f); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 777, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_10 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_10) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_10;
      __pyx_t_9 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_1);
      __pyx_t_1 = 0;
      __Pyx_INCREF(__pyx_kp_u__10);
      __pyx_t_9 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__10);
      PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u__10);

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":776
 *                                    f"reservation_price={self._reservation_price:.10f} | "
 *                                    f"optimal_spread={self._optimal_spread:.10f}")
 *                 self.logger().info(f"optimal_bid={(price-(self._reservation_price - self._optimal_spread / 2)) / price * 100:.4f}% | "             # <<<<<<<<<<<<<<
 *                                    f"optimal_ask={((self._reservation_price + self._optimal_spread / 2) - price) / price * 100:.4f}%")
 * 
 */
      __pyx_t_1 = __Pyx_PyUnicode_Join(__pyx_t_3, 5, __pyx_t_9, __pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 776, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_3)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_1};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 776, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":770
 *             # This is not what the algorithm will use as proposed bid and ask. This is just the raw output.
 *             # Optimal bid and optimal ask prices will be used
 *             if self._is_debug:             # <<<<<<<<<<<<<<
 *                 self.logger().info(f"q={q:.4f} | "
 *                                    f"vol={vol:.10f}")
 */
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":734
 *         # order book liquidity - kappa and alpha have to represent absolute values because the second member of the optimal spread equation has to be an absolute price
 *         # and from the reservation price calculation we know that gamma's unit is not absolute price
 *         if all((self.gamma, self._kappa)) and self._alpha != 0 and self._kappa > 0 and vol != 0:             # <<<<<<<<<<<<<<
 *             if self._execution_state.time_left is not None and self._execution_state.closing_time is not None:
 *                 # Avellaneda-Stoikov for a fixed timespan
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":713
 *         return self.c_measure_order_book_liquidity()
 * 
 *     cdef c_calculate_reservation_price_and_optimal_spread(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_calculate_reservation_price_and_optimal_spread", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XDECREF(__pyx_v_inventory);
  __Pyx_XDECREF(__pyx_v_q_target);
  __Pyx_XDECREF(__pyx_v_q);
  __Pyx_XDECREF(__pyx_v_vol);
  __Pyx_XDECREF(__pyx_v_time_left_fraction);
  __Pyx_XDECREF(__pyx_v_min_spread);
  __Pyx_XDECREF(__pyx_v_max_limit_bid);
  __Pyx_XDECREF(__pyx_v_min_limit_ask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":779
 *                                    f"optimal_ask={((self._reservation_price + self._optimal_spread / 2) - price) / price * 100:.4f}%")
 * 
 *     def calculate_reservation_price_and_optimal_spread(self):             # <<<<<<<<<<<<<<
 *         return self.c_calculate_reservation_price_and_optimal_spread()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_45calculate_reservation_price_and_optimal_spread(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_45calculate_reservation_price_and_optimal_spread = {"calculate_reservation_price_and_optimal_spread", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_45calculate_reservation_price_and_optimal_spread, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_45calculate_reservation_price_and_optimal_spread(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calculate_reservation_price_and_optimal_spread (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("calculate_reservation_price_and_optimal_spread", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "calculate_reservation_price_and_optimal_spread", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_44calculate_reservation_price_and_optimal_spread(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_44calculate_reservation_price_and_optimal_spread(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calculate_reservation_price_and_optimal_spread", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":780
 * 
 *     def calculate_reservation_price_and_optimal_spread(self):
 *         return self.c_calculate_reservation_price_and_optimal_spread()             # <<<<<<<<<<<<<<
 * 
 *     cdef object c_calculate_target_inventory(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_calculate_reservation_price_and_optimal_spread(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":779
 *                                    f"optimal_ask={((self._reservation_price + self._optimal_spread / 2) - price) / price * 100:.4f}%")
 * 
 *     def calculate_reservation_price_and_optimal_spread(self):             # <<<<<<<<<<<<<<
 *         return self.c_calculate_reservation_price_and_optimal_spread()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.calculate_reservation_price_and_optimal_spread", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":782
 *         return self.c_calculate_reservation_price_and_optimal_spread()
 * 
 *     cdef object c_calculate_target_inventory(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_calculate_target_inventory(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_trading_pair = 0;
  PyObject *__pyx_v_base_asset = 0;
  PyObject *__pyx_v_quote_asset = 0;
  PyObject *__pyx_v_base_value = 0;
  PyObject *__pyx_v_inventory_value = 0;
  PyObject *__pyx_v_target_inventory_value = 0;
  PyObject *__pyx_v_price = NULL;
  PyObject *__pyx_v_base_asset_amount = NULL;
  PyObject *__pyx_v_quote_asset_amount = NULL;
  PyObject *__pyx_v_target_inventory_amount = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_calculate_target_inventory", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":784
 *     cdef object c_calculate_target_inventory(self):
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *             str trading_pair = self._market_info.trading_pair
 *             str base_asset = self._market_info.base_asset
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 784, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 784, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":785
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 *             str trading_pair = self._market_info.trading_pair             # <<<<<<<<<<<<<<
 *             str base_asset = self._market_info.base_asset
 *             str quote_asset = self._market_info.quote_asset
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 785, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 785, __pyx_L1_error)
  __pyx_v_trading_pair = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":786
 *             ExchangeBase market = self._market_info.market
 *             str trading_pair = self._market_info.trading_pair
 *             str base_asset = self._market_info.base_asset             # <<<<<<<<<<<<<<
 *             str quote_asset = self._market_info.quote_asset
 *             object mid_price
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 786, __pyx_L1_error)
  __pyx_v_base_asset = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":787
 *             str trading_pair = self._market_info.trading_pair
 *             str base_asset = self._market_info.base_asset
 *             str quote_asset = self._market_info.quote_asset             # <<<<<<<<<<<<<<
 *             object mid_price
 *             object base_value
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_quote_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 787, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 787, __pyx_L1_error)
  __pyx_v_quote_asset = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":793
 *             object target_inventory_value
 * 
 *         price = self.get_price()             # <<<<<<<<<<<<<<
 *         base_asset_amount = market.get_balance(base_asset)
 *         quote_asset_amount = market.get_balance(quote_asset)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 793, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 793, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_price = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":794
 * 
 *         price = self.get_price()
 *         base_asset_amount = market.get_balance(base_asset)             # <<<<<<<<<<<<<<
 *         quote_asset_amount = market.get_balance(quote_asset)
 *         # Base asset value in quote asset prices
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_market), __pyx_n_s_get_balance); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 794, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_base_asset};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_base_asset_amount = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":795
 *         price = self.get_price()
 *         base_asset_amount = market.get_balance(base_asset)
 *         quote_asset_amount = market.get_balance(quote_asset)             # <<<<<<<<<<<<<<
 *         # Base asset value in quote asset prices
 *         base_value = base_asset_amount * price
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_market), __pyx_n_s_get_balance); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_quote_asset};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 795, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_quote_asset_amount = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":797
 *         quote_asset_amount = market.get_balance(quote_asset)
 *         # Base asset value in quote asset prices
 *         base_value = base_asset_amount * price             # <<<<<<<<<<<<<<
 *         # Total inventory value in quote asset prices
 *         inventory_value = base_value + quote_asset_amount
 */
  __pyx_t_1 = PyNumber_Multiply(__pyx_v_base_asset_amount, __pyx_v_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 797, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_base_value = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":799
 *         base_value = base_asset_amount * price
 *         # Total inventory value in quote asset prices
 *         inventory_value = base_value + quote_asset_amount             # <<<<<<<<<<<<<<
 *         # Target base asset value in quote asset prices
 *         target_inventory_value = inventory_value * self.inventory_target_base
 */
  __pyx_t_1 = PyNumber_Add(__pyx_v_base_value, __pyx_v_quote_asset_amount); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_inventory_value = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":801
 *         inventory_value = base_value + quote_asset_amount
 *         # Target base asset value in quote asset prices
 *         target_inventory_value = inventory_value * self.inventory_target_base             # <<<<<<<<<<<<<<
 *         # Target base asset amount
 *         target_inventory_amount = target_inventory_value / price
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_inventory_target_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Multiply(__pyx_v_inventory_value, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 801, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_target_inventory_value = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":803
 *         target_inventory_value = inventory_value * self.inventory_target_base
 *         # Target base asset amount
 *         target_inventory_amount = target_inventory_value / price             # <<<<<<<<<<<<<<
 *         return market.c_quantize_order_amount(trading_pair, Decimal(str(target_inventory_amount)))
 * 
 */
  __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_v_target_inventory_value, __pyx_v_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_target_inventory_amount = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":804
 *         # Target base asset amount
 *         target_inventory_amount = target_inventory_value / price
 *         return market.c_quantize_order_amount(trading_pair, Decimal(str(target_inventory_amount)))             # <<<<<<<<<<<<<<
 * 
 *     def calculate_target_inventory(self) -> Decimal:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_Unicode(__pyx_v_target_inventory_amount); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_3};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 804, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), __pyx_v_trading_pair, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":782
 *         return self.c_calculate_reservation_price_and_optimal_spread()
 * 
 *     cdef object c_calculate_target_inventory(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_calculate_target_inventory", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_trading_pair);
  __Pyx_XDECREF(__pyx_v_base_asset);
  __Pyx_XDECREF(__pyx_v_quote_asset);
  __Pyx_XDECREF(__pyx_v_base_value);
  __Pyx_XDECREF(__pyx_v_inventory_value);
  __Pyx_XDECREF(__pyx_v_target_inventory_value);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XDECREF(__pyx_v_base_asset_amount);
  __Pyx_XDECREF(__pyx_v_quote_asset_amount);
  __Pyx_XDECREF(__pyx_v_target_inventory_amount);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":806
 *         return market.c_quantize_order_amount(trading_pair, Decimal(str(target_inventory_amount)))
 * 
 *     def calculate_target_inventory(self) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_calculate_target_inventory()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_47calculate_target_inventory(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_47calculate_target_inventory = {"calculate_target_inventory", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_47calculate_target_inventory, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_47calculate_target_inventory(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calculate_target_inventory (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("calculate_target_inventory", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "calculate_target_inventory", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_46calculate_target_inventory(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_46calculate_target_inventory(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calculate_target_inventory", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":807
 * 
 *     def calculate_target_inventory(self) -> Decimal:
 *         return self.c_calculate_target_inventory()             # <<<<<<<<<<<<<<
 * 
 *     cdef c_calculate_inventory(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_calculate_target_inventory(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 807, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":806
 *         return market.c_quantize_order_amount(trading_pair, Decimal(str(target_inventory_amount)))
 * 
 *     def calculate_target_inventory(self) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_calculate_target_inventory()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.calculate_target_inventory", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":809
 *         return self.c_calculate_target_inventory()
 * 
 *     cdef c_calculate_inventory(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_calculate_inventory(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_base_asset = 0;
  PyObject *__pyx_v_quote_asset = 0;
  PyObject *__pyx_v_base_value = 0;
  PyObject *__pyx_v_inventory_value_quote = 0;
  PyObject *__pyx_v_inventory_value_base = 0;
  PyObject *__pyx_v_price = NULL;
  PyObject *__pyx_v_base_asset_amount = NULL;
  PyObject *__pyx_v_quote_asset_amount = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_calculate_inventory", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":811
 *     cdef c_calculate_inventory(self):
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *             str base_asset = self._market_info.base_asset
 *             str quote_asset = self._market_info.quote_asset
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 811, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 811, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":812
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 *             str base_asset = self._market_info.base_asset             # <<<<<<<<<<<<<<
 *             str quote_asset = self._market_info.quote_asset
 *             object mid_price
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 812, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 812, __pyx_L1_error)
  __pyx_v_base_asset = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":813
 *             ExchangeBase market = self._market_info.market
 *             str base_asset = self._market_info.base_asset
 *             str quote_asset = self._market_info.quote_asset             # <<<<<<<<<<<<<<
 *             object mid_price
 *             object base_value
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_quote_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 813, __pyx_L1_error)
  __pyx_v_quote_asset = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":820
 *             object inventory_value_base
 * 
 *         price = self.get_price()             # <<<<<<<<<<<<<<
 *         base_asset_amount = market.get_balance(base_asset)
 *         quote_asset_amount = market.get_balance(quote_asset)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 820, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 820, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_price = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":821
 * 
 *         price = self.get_price()
 *         base_asset_amount = market.get_balance(base_asset)             # <<<<<<<<<<<<<<
 *         quote_asset_amount = market.get_balance(quote_asset)
 *         # Base asset value in quote asset prices
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_market), __pyx_n_s_get_balance); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 821, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_base_asset};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 821, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_base_asset_amount = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":822
 *         price = self.get_price()
 *         base_asset_amount = market.get_balance(base_asset)
 *         quote_asset_amount = market.get_balance(quote_asset)             # <<<<<<<<<<<<<<
 *         # Base asset value in quote asset prices
 *         base_value = base_asset_amount * price
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_market), __pyx_n_s_get_balance); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_quote_asset};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 822, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_quote_asset_amount = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":824
 *         quote_asset_amount = market.get_balance(quote_asset)
 *         # Base asset value in quote asset prices
 *         base_value = base_asset_amount * price             # <<<<<<<<<<<<<<
 *         # Total inventory value in quote asset prices
 *         inventory_value_quote = base_value + quote_asset_amount
 */
  __pyx_t_1 = PyNumber_Multiply(__pyx_v_base_asset_amount, __pyx_v_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 824, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_base_value = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":826
 *         base_value = base_asset_amount * price
 *         # Total inventory value in quote asset prices
 *         inventory_value_quote = base_value + quote_asset_amount             # <<<<<<<<<<<<<<
 *         # Total inventory value in base asset prices
 *         inventory_value_base = inventory_value_quote / price
 */
  __pyx_t_1 = PyNumber_Add(__pyx_v_base_value, __pyx_v_quote_asset_amount); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_inventory_value_quote = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":828
 *         inventory_value_quote = base_value + quote_asset_amount
 *         # Total inventory value in base asset prices
 *         inventory_value_base = inventory_value_quote / price             # <<<<<<<<<<<<<<
 *         return inventory_value_base
 * 
 */
  __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_v_inventory_value_quote, __pyx_v_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 828, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_inventory_value_base = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":829
 *         # Total inventory value in base asset prices
 *         inventory_value_base = inventory_value_quote / price
 *         return inventory_value_base             # <<<<<<<<<<<<<<
 * 
 *     def calculate_inventory(self) -> Decimal:
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_inventory_value_base);
  __pyx_r = __pyx_v_inventory_value_base;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":809
 *         return self.c_calculate_target_inventory()
 * 
 *     cdef c_calculate_inventory(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_calculate_inventory", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_base_asset);
  __Pyx_XDECREF(__pyx_v_quote_asset);
  __Pyx_XDECREF(__pyx_v_base_value);
  __Pyx_XDECREF(__pyx_v_inventory_value_quote);
  __Pyx_XDECREF(__pyx_v_inventory_value_base);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XDECREF(__pyx_v_base_asset_amount);
  __Pyx_XDECREF(__pyx_v_quote_asset_amount);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":831
 *         return inventory_value_base
 * 
 *     def calculate_inventory(self) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_calculate_inventory()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_49calculate_inventory(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_49calculate_inventory = {"calculate_inventory", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_49calculate_inventory, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_49calculate_inventory(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("calculate_inventory (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("calculate_inventory", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "calculate_inventory", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_48calculate_inventory(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_48calculate_inventory(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("calculate_inventory", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":832
 * 
 *     def calculate_inventory(self) -> Decimal:
 *         return self.c_calculate_inventory()             # <<<<<<<<<<<<<<
 * 
 *     cdef bint c_is_algorithm_ready(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_calculate_inventory(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 832, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":831
 *         return inventory_value_base
 * 
 *     def calculate_inventory(self) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_calculate_inventory()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.calculate_inventory", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":834
 *         return self.c_calculate_inventory()
 * 
 *     cdef bint c_is_algorithm_ready(self):             # <<<<<<<<<<<<<<
 *         return self._avg_vol.is_sampling_buffer_full and self._trading_intensity.is_sampling_buffer_full
 * 
 */

static int __pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_is_algorithm_ready(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_is_algorithm_ready", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":835
 * 
 *     cdef bint c_is_algorithm_ready(self):
 *         return self._avg_vol.is_sampling_buffer_full and self._trading_intensity.is_sampling_buffer_full             # <<<<<<<<<<<<<<
 * 
 *     cdef bint c_is_algorithm_changed(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_avg_vol, __pyx_n_s_is_sampling_buffer_full); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_trading_intensity), __pyx_n_s_is_sampling_buffer_full); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":834
 *         return self.c_calculate_inventory()
 * 
 *     cdef bint c_is_algorithm_ready(self):             # <<<<<<<<<<<<<<
 *         return self._avg_vol.is_sampling_buffer_full and self._trading_intensity.is_sampling_buffer_full
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_is_algorithm_ready", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":837
 *         return self._avg_vol.is_sampling_buffer_full and self._trading_intensity.is_sampling_buffer_full
 * 
 *     cdef bint c_is_algorithm_changed(self):             # <<<<<<<<<<<<<<
 *         return self._trading_intensity.is_sampling_buffer_changed or self._avg_vol.is_sampling_buffer_changed
 * 
 */

static int __pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_is_algorithm_changed(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_is_algorithm_changed", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":838
 * 
 *     cdef bint c_is_algorithm_changed(self):
 *         return self._trading_intensity.is_sampling_buffer_changed or self._avg_vol.is_sampling_buffer_changed             # <<<<<<<<<<<<<<
 * 
 *     def is_algorithm_ready(self) -> bool:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->_trading_intensity), __pyx_n_s_is_sampling_buffer_changed); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 838, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_avg_vol, __pyx_n_s_is_sampling_buffer_changed); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 838, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 838, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = __pyx_t_3;
  __pyx_L3_bool_binop_done:;
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":837
 *         return self._avg_vol.is_sampling_buffer_full and self._trading_intensity.is_sampling_buffer_full
 * 
 *     cdef bint c_is_algorithm_changed(self):             # <<<<<<<<<<<<<<
 *         return self._trading_intensity.is_sampling_buffer_changed or self._avg_vol.is_sampling_buffer_changed
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_is_algorithm_changed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":840
 *         return self._trading_intensity.is_sampling_buffer_changed or self._avg_vol.is_sampling_buffer_changed
 * 
 *     def is_algorithm_ready(self) -> bool:             # <<<<<<<<<<<<<<
 *         return self.c_is_algorithm_ready()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_51is_algorithm_ready(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_51is_algorithm_ready = {"is_algorithm_ready", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_51is_algorithm_ready, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_51is_algorithm_ready(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_algorithm_ready (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_algorithm_ready", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_algorithm_ready", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_50is_algorithm_ready(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_50is_algorithm_ready(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_algorithm_ready", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":841
 * 
 *     def is_algorithm_ready(self) -> bool:
 *         return self.c_is_algorithm_ready()             # <<<<<<<<<<<<<<
 * 
 *     def is_algorithm_changed(self) -> bool:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_is_algorithm_ready(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 841, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 841, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":840
 *         return self._trading_intensity.is_sampling_buffer_changed or self._avg_vol.is_sampling_buffer_changed
 * 
 *     def is_algorithm_ready(self) -> bool:             # <<<<<<<<<<<<<<
 *         return self.c_is_algorithm_ready()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.is_algorithm_ready", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":843
 *         return self.c_is_algorithm_ready()
 * 
 *     def is_algorithm_changed(self) -> bool:             # <<<<<<<<<<<<<<
 *         return self.c_is_algorithm_changed()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_53is_algorithm_changed(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_53is_algorithm_changed = {"is_algorithm_changed", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_53is_algorithm_changed, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_53is_algorithm_changed(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_algorithm_changed (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("is_algorithm_changed", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "is_algorithm_changed", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_52is_algorithm_changed(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_52is_algorithm_changed(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_algorithm_changed", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":844
 * 
 *     def is_algorithm_changed(self) -> bool:
 *         return self.c_is_algorithm_changed()             # <<<<<<<<<<<<<<
 * 
 *     def _get_level_spreads(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_is_algorithm_changed(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 844, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":843
 *         return self.c_is_algorithm_ready()
 * 
 *     def is_algorithm_changed(self) -> bool:             # <<<<<<<<<<<<<<
 *         return self.c_is_algorithm_changed()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.is_algorithm_changed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":846
 *         return self.c_is_algorithm_changed()
 * 
 *     def _get_level_spreads(self):             # <<<<<<<<<<<<<<
 *         level_step = ((self._optimal_spread / 2) / 100) * self.level_distances
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_55_get_level_spreads(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_55_get_level_spreads = {"_get_level_spreads", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_55_get_level_spreads, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_55_get_level_spreads(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_get_level_spreads (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("_get_level_spreads", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "_get_level_spreads", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_54_get_level_spreads(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_54_get_level_spreads(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_v_level_step = NULL;
  PyObject *__pyx_v_bid_level_spreads = NULL;
  PyObject *__pyx_v_ask_level_spreads = NULL;
  PyObject *__pyx_9genexpr11__pyx_v_i = NULL;
  PyObject *__pyx_9genexpr12__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_get_level_spreads", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":847
 * 
 *     def _get_level_spreads(self):
 *         level_step = ((self._optimal_spread / 2) / 100) * self.level_distances             # <<<<<<<<<<<<<<
 * 
 *         bid_level_spreads = [i * level_step for i in range(self.order_levels)]
 */
  __pyx_t_1 = __Pyx_PyInt_TrueDivideObjC(__pyx_v_self->_optimal_spread, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_TrueDivideObjC(__pyx_t_1, __pyx_int_100, 0x64, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_level_distances); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_2, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 847, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_level_step = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":849
 *         level_step = ((self._optimal_spread / 2) / 100) * self.level_distances
 * 
 *         bid_level_spreads = [i * level_step for i in range(self.order_levels)]             # <<<<<<<<<<<<<<
 *         ask_level_spreads = [i * level_step for i in range(self.order_levels)]
 * 
 */
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 849, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_levels); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 849, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 849, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 849, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 849, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 849, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 849, __pyx_L5_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 849, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 849, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 849, __pyx_L5_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 849, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_5(__pyx_t_1);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 849, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr11__pyx_v_i, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Multiply(__pyx_9genexpr11__pyx_v_i, __pyx_v_level_step); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 849, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 849, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_i); __pyx_9genexpr11__pyx_v_i = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_i); __pyx_9genexpr11__pyx_v_i = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_v_bid_level_spreads = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":850
 * 
 *         bid_level_spreads = [i * level_step for i in range(self.order_levels)]
 *         ask_level_spreads = [i * level_step for i in range(self.order_levels)]             # <<<<<<<<<<<<<<
 * 
 *         return bid_level_spreads, ask_level_spreads
 */
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 850, __pyx_L12_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_levels); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 850, __pyx_L12_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 850, __pyx_L12_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 850, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 850, __pyx_L12_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 850, __pyx_L12_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 850, __pyx_L12_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 850, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 850, __pyx_L12_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 850, __pyx_L12_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 850, __pyx_L12_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_5(__pyx_t_1);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 850, __pyx_L12_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr12__pyx_v_i, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_Multiply(__pyx_9genexpr12__pyx_v_i, __pyx_v_level_step); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 850, __pyx_L12_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 850, __pyx_L12_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_i); __pyx_9genexpr12__pyx_v_i = 0;
    goto __pyx_L16_exit_scope;
    __pyx_L12_error:;
    __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_i); __pyx_9genexpr12__pyx_v_i = 0;
    goto __pyx_L1_error;
    __pyx_L16_exit_scope:;
  } /* exit inner scope */
  __pyx_v_ask_level_spreads = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":852
 *         ask_level_spreads = [i * level_step for i in range(self.order_levels)]
 * 
 *         return bid_level_spreads, ask_level_spreads             # <<<<<<<<<<<<<<
 * 
 *     cdef _create_proposal_based_on_order_override(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 852, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_bid_level_spreads);
  __Pyx_GIVEREF(__pyx_v_bid_level_spreads);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_bid_level_spreads)) __PYX_ERR(0, 852, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_ask_level_spreads);
  __Pyx_GIVEREF(__pyx_v_ask_level_spreads);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_ask_level_spreads)) __PYX_ERR(0, 852, __pyx_L1_error);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":846
 *         return self.c_is_algorithm_changed()
 * 
 *     def _get_level_spreads(self):             # <<<<<<<<<<<<<<
 *         level_step = ((self._optimal_spread / 2) / 100) * self.level_distances
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy._get_level_spreads", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_level_step);
  __Pyx_XDECREF(__pyx_v_bid_level_spreads);
  __Pyx_XDECREF(__pyx_v_ask_level_spreads);
  __Pyx_XDECREF(__pyx_9genexpr11__pyx_v_i);
  __Pyx_XDECREF(__pyx_9genexpr12__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":854
 *         return bid_level_spreads, ask_level_spreads
 * 
 *     cdef _create_proposal_based_on_order_override(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy__create_proposal_based_on_order_override(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_buys = 0;
  PyObject *__pyx_v_sells = 0;
  PyObject *__pyx_v_reference_price = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_key = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_v_list_to_be_appended = NULL;
  PyObject *__pyx_v_size = NULL;
  PyObject *__pyx_v_price = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  int __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create_proposal_based_on_order_override", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":856
 *     cdef _create_proposal_based_on_order_override(self):
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *             list buys = []
 *             list sells = []
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 856, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 856, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":857
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 *             list buys = []             # <<<<<<<<<<<<<<
 *             list sells = []
 *         reference_price = self.get_price()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 857, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buys = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":858
 *             ExchangeBase market = self._market_info.market
 *             list buys = []
 *             list sells = []             # <<<<<<<<<<<<<<
 *         reference_price = self.get_price()
 *         if self.order_override is not None:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sells = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":859
 *             list buys = []
 *             list sells = []
 *         reference_price = self.get_price()             # <<<<<<<<<<<<<<
 *         if self.order_override is not None:
 *             for key, value in self.order_override.items():
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 859, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 859, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_reference_price = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":860
 *             list sells = []
 *         reference_price = self.get_price()
 *         if self.order_override is not None:             # <<<<<<<<<<<<<<
 *             for key, value in self.order_override.items():
 *                 if str(value[0]) in ["buy", "sell"]:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_override); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 860, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_5) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":861
 *         reference_price = self.get_price()
 *         if self.order_override is not None:
 *             for key, value in self.order_override.items():             # <<<<<<<<<<<<<<
 *                 if str(value[0]) in ["buy", "sell"]:
 *                     list_to_be_appended = buys if str(value[0]) == "buy" else sells
 */
    __pyx_t_6 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_override); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 861, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__pyx_t_2 == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 861, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_dict_iterator(__pyx_t_2, 0, __pyx_n_s_items, (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 861, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
    while (1) {
      __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_7, &__pyx_t_6, &__pyx_t_3, &__pyx_t_2, NULL, __pyx_t_8);
      if (unlikely(__pyx_t_9 == 0)) break;
      if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 861, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_3);
      __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":862
 *         if self.order_override is not None:
 *             for key, value in self.order_override.items():
 *                 if str(value[0]) in ["buy", "sell"]:             # <<<<<<<<<<<<<<
 *                     list_to_be_appended = buys if str(value[0]) == "buy" else sells
 *                     size = Decimal(str(value[2]))
 */
      __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_value, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 862, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = __Pyx_PyObject_Unicode(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 862, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_buy, Py_EQ)); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 862, __pyx_L1_error)
      if (!__pyx_t_10) {
      } else {
        __pyx_t_5 = __pyx_t_10;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_sell, Py_EQ)); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 862, __pyx_L1_error)
      __pyx_t_5 = __pyx_t_10;
      __pyx_L7_bool_binop_done:;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_10 = __pyx_t_5;
      if (__pyx_t_10) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":863
 *             for key, value in self.order_override.items():
 *                 if str(value[0]) in ["buy", "sell"]:
 *                     list_to_be_appended = buys if str(value[0]) == "buy" else sells             # <<<<<<<<<<<<<<
 *                     size = Decimal(str(value[2]))
 *                     size = market.c_quantize_order_amount(self.trading_pair, size)
 */
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_value, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 863, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_11 = __Pyx_PyObject_Unicode(__pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 863, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_t_11, __pyx_n_u_buy, Py_EQ)); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 863, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_10) {
          __Pyx_INCREF(__pyx_v_buys);
          __pyx_t_3 = __pyx_v_buys;
        } else {
          __Pyx_INCREF(__pyx_v_sells);
          __pyx_t_3 = __pyx_v_sells;
        }
        __Pyx_XDECREF_SET(__pyx_v_list_to_be_appended, ((PyObject*)__pyx_t_3));
        __pyx_t_3 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":864
 *                 if str(value[0]) in ["buy", "sell"]:
 *                     list_to_be_appended = buys if str(value[0]) == "buy" else sells
 *                     size = Decimal(str(value[2]))             # <<<<<<<<<<<<<<
 *                     size = market.c_quantize_order_amount(self.trading_pair, size)
 *                     if str(value[0]) == "buy":
 */
        __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 864, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_value, 2, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 864, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_12 = __Pyx_PyObject_Unicode(__pyx_t_2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 864, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = NULL;
        __pyx_t_4 = 0;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_11);
          if (likely(__pyx_t_2)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
            __Pyx_INCREF(__pyx_t_2);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_11, function);
            __pyx_t_4 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_12};
          __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 864, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        }
        __Pyx_XDECREF_SET(__pyx_v_size, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":865
 *                     list_to_be_appended = buys if str(value[0]) == "buy" else sells
 *                     size = Decimal(str(value[2]))
 *                     size = market.c_quantize_order_amount(self.trading_pair, size)             # <<<<<<<<<<<<<<
 *                     if str(value[0]) == "buy":
 *                         price = reference_price * (Decimal("1") - Decimal(str(value[1])) / Decimal("100"))
 */
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 865, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_3))) __PYX_ERR(0, 865, __pyx_L1_error)
        __pyx_t_11 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_3), __pyx_v_size, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 865, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __Pyx_DECREF_SET(__pyx_v_size, __pyx_t_11);
        __pyx_t_11 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":866
 *                     size = Decimal(str(value[2]))
 *                     size = market.c_quantize_order_amount(self.trading_pair, size)
 *                     if str(value[0]) == "buy":             # <<<<<<<<<<<<<<
 *                         price = reference_price * (Decimal("1") - Decimal(str(value[1])) / Decimal("100"))
 *                     elif str(value[0]) == "sell":
 */
        __pyx_t_11 = __Pyx_GetItemInt(__pyx_v_value, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 866, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_3 = __Pyx_PyObject_Unicode(__pyx_t_11); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 866, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_t_3, __pyx_n_u_buy, Py_EQ)); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 866, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (__pyx_t_10) {

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":867
 *                     size = market.c_quantize_order_amount(self.trading_pair, size)
 *                     if str(value[0]) == "buy":
 *                         price = reference_price * (Decimal("1") - Decimal(str(value[1])) / Decimal("100"))             # <<<<<<<<<<<<<<
 *                     elif str(value[0]) == "sell":
 *                         price = reference_price * (Decimal("1") + Decimal(str(value[1])) / Decimal("100"))
 */
          __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 867, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_12 = NULL;
          __pyx_t_4 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_4 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_kp_u_1_2};
            __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 867, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 867, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __pyx_t_2 = __Pyx_GetItemInt(__pyx_v_value, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 867, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_13 = __Pyx_PyObject_Unicode(__pyx_t_2); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 867, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = NULL;
          __pyx_t_4 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_12))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_12);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_12, function);
              __pyx_t_4 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_13};
            __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 867, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          }
          __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 867, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_2 = NULL;
          __pyx_t_4 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_2)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_2);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_4 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_kp_u_100};
            __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
            __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 867, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __pyx_t_13 = __Pyx_PyNumber_Divide(__pyx_t_11, __pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 867, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_12 = PyNumber_Subtract(__pyx_t_3, __pyx_t_13); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 867, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __pyx_t_13 = PyNumber_Multiply(__pyx_v_reference_price, __pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 867, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_XDECREF_SET(__pyx_v_price, __pyx_t_13);
          __pyx_t_13 = 0;

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":866
 *                     size = Decimal(str(value[2]))
 *                     size = market.c_quantize_order_amount(self.trading_pair, size)
 *                     if str(value[0]) == "buy":             # <<<<<<<<<<<<<<
 *                         price = reference_price * (Decimal("1") - Decimal(str(value[1])) / Decimal("100"))
 *                     elif str(value[0]) == "sell":
 */
          goto __pyx_L9;
        }

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":868
 *                     if str(value[0]) == "buy":
 *                         price = reference_price * (Decimal("1") - Decimal(str(value[1])) / Decimal("100"))
 *                     elif str(value[0]) == "sell":             # <<<<<<<<<<<<<<
 *                         price = reference_price * (Decimal("1") + Decimal(str(value[1])) / Decimal("100"))
 *                     price = market.c_quantize_order_price(self.trading_pair, price)
 */
        __pyx_t_13 = __Pyx_GetItemInt(__pyx_v_value, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 868, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_13);
        __pyx_t_12 = __Pyx_PyObject_Unicode(__pyx_t_13); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 868, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        __pyx_t_10 = (__Pyx_PyUnicode_Equals(__pyx_t_12, __pyx_n_u_sell, Py_EQ)); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 868, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        if (__pyx_t_10) {

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":869
 *                         price = reference_price * (Decimal("1") - Decimal(str(value[1])) / Decimal("100"))
 *                     elif str(value[0]) == "sell":
 *                         price = reference_price * (Decimal("1") + Decimal(str(value[1])) / Decimal("100"))             # <<<<<<<<<<<<<<
 *                     price = market.c_quantize_order_price(self.trading_pair, price)
 *                     if size > 0 and price > 0:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_13, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 869, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_3 = NULL;
          __pyx_t_4 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_3)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_3);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_4 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_kp_u_1_2};
            __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
            __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
            if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 869, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 869, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __pyx_t_11 = __Pyx_GetItemInt(__pyx_v_value, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 869, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_2 = __Pyx_PyObject_Unicode(__pyx_t_11); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 869, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = NULL;
          __pyx_t_4 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_3))) {
            __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
            if (likely(__pyx_t_11)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
              __Pyx_INCREF(__pyx_t_11);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_3, function);
              __pyx_t_4 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_t_2};
            __pyx_t_13 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 869, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_13);
            __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          }
          __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 869, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_11 = NULL;
          __pyx_t_4 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_11)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_11);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_4 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_kp_u_100};
            __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 869, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __pyx_t_2 = __Pyx_PyNumber_Divide(__pyx_t_13, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 869, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __pyx_t_3 = PyNumber_Add(__pyx_t_12, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 869, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          __pyx_t_2 = PyNumber_Multiply(__pyx_v_reference_price, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 869, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_XDECREF_SET(__pyx_v_price, __pyx_t_2);
          __pyx_t_2 = 0;

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":868
 *                     if str(value[0]) == "buy":
 *                         price = reference_price * (Decimal("1") - Decimal(str(value[1])) / Decimal("100"))
 *                     elif str(value[0]) == "sell":             # <<<<<<<<<<<<<<
 *                         price = reference_price * (Decimal("1") + Decimal(str(value[1])) / Decimal("100"))
 *                     price = market.c_quantize_order_price(self.trading_pair, price)
 */
        }
        __pyx_L9:;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":870
 *                     elif str(value[0]) == "sell":
 *                         price = reference_price * (Decimal("1") + Decimal(str(value[1])) / Decimal("100"))
 *                     price = market.c_quantize_order_price(self.trading_pair, price)             # <<<<<<<<<<<<<<
 *                     if size > 0 and price > 0:
 *                         list_to_be_appended.append(PriceSize(price, size))
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 870, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_2))) __PYX_ERR(0, 870, __pyx_L1_error)
        if (unlikely(!__pyx_v_price)) { __Pyx_RaiseUnboundLocalError("price"); __PYX_ERR(0, 870, __pyx_L1_error) }
        __pyx_t_3 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_2), __pyx_v_price); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 870, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_price, __pyx_t_3);
        __pyx_t_3 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":871
 *                         price = reference_price * (Decimal("1") + Decimal(str(value[1])) / Decimal("100"))
 *                     price = market.c_quantize_order_price(self.trading_pair, price)
 *                     if size > 0 and price > 0:             # <<<<<<<<<<<<<<
 *                         list_to_be_appended.append(PriceSize(price, size))
 *         return buys, sells
 */
        __pyx_t_3 = PyObject_RichCompare(__pyx_v_size, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 871, __pyx_L1_error)
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 871, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (__pyx_t_5) {
        } else {
          __pyx_t_10 = __pyx_t_5;
          goto __pyx_L11_bool_binop_done;
        }
        __pyx_t_3 = PyObject_RichCompare(__pyx_v_price, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 871, __pyx_L1_error)
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 871, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_10 = __pyx_t_5;
        __pyx_L11_bool_binop_done:;
        if (__pyx_t_10) {

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":872
 *                     price = market.c_quantize_order_price(self.trading_pair, price)
 *                     if size > 0 and price > 0:
 *                         list_to_be_appended.append(PriceSize(price, size))             # <<<<<<<<<<<<<<
 *         return buys, sells
 * 
 */
          if (unlikely(__pyx_v_list_to_be_appended == Py_None)) {
            PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
            __PYX_ERR(0, 872, __pyx_L1_error)
          }
          __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_PriceSize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 872, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          __pyx_t_12 = NULL;
          __pyx_t_4 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_2))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_2);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_2, function);
              __pyx_t_4 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_12, __pyx_v_price, __pyx_v_size};
            __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 872, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_3);
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
          }
          __pyx_t_14 = __Pyx_PyList_Append(__pyx_v_list_to_be_appended, __pyx_t_3); if (unlikely(__pyx_t_14 == ((int)-1))) __PYX_ERR(0, 872, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":871
 *                         price = reference_price * (Decimal("1") + Decimal(str(value[1])) / Decimal("100"))
 *                     price = market.c_quantize_order_price(self.trading_pair, price)
 *                     if size > 0 and price > 0:             # <<<<<<<<<<<<<<
 *                         list_to_be_appended.append(PriceSize(price, size))
 *         return buys, sells
 */
        }

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":862
 *         if self.order_override is not None:
 *             for key, value in self.order_override.items():
 *                 if str(value[0]) in ["buy", "sell"]:             # <<<<<<<<<<<<<<
 *                     list_to_be_appended = buys if str(value[0]) == "buy" else sells
 *                     size = Decimal(str(value[2]))
 */
      }
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":860
 *             list sells = []
 *         reference_price = self.get_price()
 *         if self.order_override is not None:             # <<<<<<<<<<<<<<
 *             for key, value in self.order_override.items():
 *                 if str(value[0]) in ["buy", "sell"]:
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":873
 *                     if size > 0 and price > 0:
 *                         list_to_be_appended.append(PriceSize(price, size))
 *         return buys, sells             # <<<<<<<<<<<<<<
 * 
 *     def create_proposal_based_on_order_override(self) -> Tuple[List[Proposal], List[Proposal]]:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 873, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_buys);
  __Pyx_GIVEREF(__pyx_v_buys);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_buys)) __PYX_ERR(0, 873, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_sells);
  __Pyx_GIVEREF(__pyx_v_sells);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_sells)) __PYX_ERR(0, 873, __pyx_L1_error);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":854
 *         return bid_level_spreads, ask_level_spreads
 * 
 *     cdef _create_proposal_based_on_order_override(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy._create_proposal_based_on_order_override", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_buys);
  __Pyx_XDECREF(__pyx_v_sells);
  __Pyx_XDECREF(__pyx_v_reference_price);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_list_to_be_appended);
  __Pyx_XDECREF(__pyx_v_size);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":875
 *         return buys, sells
 * 
 *     def create_proposal_based_on_order_override(self) -> Tuple[List[Proposal], List[Proposal]]:             # <<<<<<<<<<<<<<
 *         return self._create_proposal_based_on_order_override()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_57create_proposal_based_on_order_override(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_57create_proposal_based_on_order_override = {"create_proposal_based_on_order_override", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_57create_proposal_based_on_order_override, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_57create_proposal_based_on_order_override(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_proposal_based_on_order_override (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("create_proposal_based_on_order_override", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "create_proposal_based_on_order_override", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_56create_proposal_based_on_order_override(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_56create_proposal_based_on_order_override(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_proposal_based_on_order_override", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":876
 * 
 *     def create_proposal_based_on_order_override(self) -> Tuple[List[Proposal], List[Proposal]]:
 *         return self._create_proposal_based_on_order_override()             # <<<<<<<<<<<<<<
 * 
 *     cdef _create_proposal_based_on_order_levels(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_create_proposal_based_on_order_override(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":875
 *         return buys, sells
 * 
 *     def create_proposal_based_on_order_override(self) -> Tuple[List[Proposal], List[Proposal]]:             # <<<<<<<<<<<<<<
 *         return self._create_proposal_based_on_order_override()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.create_proposal_based_on_order_override", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":878
 *         return self._create_proposal_based_on_order_override()
 * 
 *     cdef _create_proposal_based_on_order_levels(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy__create_proposal_based_on_order_levels(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_buys = 0;
  PyObject *__pyx_v_sells = 0;
  PyObject *__pyx_v_bid_level_spreads = NULL;
  PyObject *__pyx_v_ask_level_spreads = NULL;
  PyObject *__pyx_v_size = NULL;
  PyObject *__pyx_v_level = NULL;
  PyObject *__pyx_v_bid_price = NULL;
  PyObject *__pyx_v_ask_price = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create_proposal_based_on_order_levels", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":880
 *     cdef _create_proposal_based_on_order_levels(self):
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *             list buys = []
 *             list sells = []
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 880, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":881
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 *             list buys = []             # <<<<<<<<<<<<<<
 *             list sells = []
 *         bid_level_spreads, ask_level_spreads = self._get_level_spreads()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 881, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buys = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":882
 *             ExchangeBase market = self._market_info.market
 *             list buys = []
 *             list sells = []             # <<<<<<<<<<<<<<
 *         bid_level_spreads, ask_level_spreads = self._get_level_spreads()
 *         size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 882, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sells = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":883
 *             list buys = []
 *             list sells = []
 *         bid_level_spreads, ask_level_spreads = self._get_level_spreads()             # <<<<<<<<<<<<<<
 *         size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)
 *         if size > 0:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_level_spreads); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 883, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 883, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 883, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_5);
    index = 0; __pyx_t_2 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 2) < 0) __PYX_ERR(0, 883, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 883, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_bid_level_spreads = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_ask_level_spreads = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":884
 *             list sells = []
 *         bid_level_spreads, ask_level_spreads = self._get_level_spreads()
 *         size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)             # <<<<<<<<<<<<<<
 *         if size > 0:
 *             for level in range(self.order_levels):
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 884, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_order_amount); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_1), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 884, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_size = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":885
 *         bid_level_spreads, ask_level_spreads = self._get_level_spreads()
 *         size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)
 *         if size > 0:             # <<<<<<<<<<<<<<
 *             for level in range(self.order_levels):
 *                 bid_price = market.c_quantize_order_price(self.trading_pair,
 */
  __pyx_t_2 = PyObject_RichCompare(__pyx_v_size, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 885, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_7) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":886
 *         size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)
 *         if size > 0:
 *             for level in range(self.order_levels):             # <<<<<<<<<<<<<<
 *                 bid_price = market.c_quantize_order_price(self.trading_pair,
 *                                                           self._optimal_bid - Decimal(str(bid_level_spreads[level])))
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_levels); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 886, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 886, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_2 = __pyx_t_3; __Pyx_INCREF(__pyx_t_2);
      __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_8 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 886, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 886, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_2))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 886, __pyx_L1_error)
            #endif
            if (__pyx_t_8 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 886, __pyx_L1_error)
          #else
          __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 886, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 886, __pyx_L1_error)
            #endif
            if (__pyx_t_8 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_8); __Pyx_INCREF(__pyx_t_3); __pyx_t_8++; if (unlikely((0 < 0))) __PYX_ERR(0, 886, __pyx_L1_error)
          #else
          __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 886, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_9(__pyx_t_2);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 886, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_level, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":887
 *         if size > 0:
 *             for level in range(self.order_levels):
 *                 bid_price = market.c_quantize_order_price(self.trading_pair,             # <<<<<<<<<<<<<<
 *                                                           self._optimal_bid - Decimal(str(bid_level_spreads[level])))
 *                 ask_price = market.c_quantize_order_price(self.trading_pair,
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 887, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_3))) __PYX_ERR(0, 887, __pyx_L1_error)

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":888
 *             for level in range(self.order_levels):
 *                 bid_price = market.c_quantize_order_price(self.trading_pair,
 *                                                           self._optimal_bid - Decimal(str(bid_level_spreads[level])))             # <<<<<<<<<<<<<<
 *                 ask_price = market.c_quantize_order_price(self.trading_pair,
 *                                                           self._optimal_ask + Decimal(str(ask_level_spreads[level])))
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_v_bid_level_spreads, __pyx_v_level); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = __Pyx_PyObject_Unicode(__pyx_t_10); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_10)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_10);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_11};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 888, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_t_5 = PyNumber_Subtract(__pyx_v_self->_optimal_bid, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 888, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":887
 *         if size > 0:
 *             for level in range(self.order_levels):
 *                 bid_price = market.c_quantize_order_price(self.trading_pair,             # <<<<<<<<<<<<<<
 *                                                           self._optimal_bid - Decimal(str(bid_level_spreads[level])))
 *                 ask_price = market.c_quantize_order_price(self.trading_pair,
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_3), __pyx_t_5); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 887, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_bid_price, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":889
 *                 bid_price = market.c_quantize_order_price(self.trading_pair,
 *                                                           self._optimal_bid - Decimal(str(bid_level_spreads[level])))
 *                 ask_price = market.c_quantize_order_price(self.trading_pair,             # <<<<<<<<<<<<<<
 *                                                           self._optimal_ask + Decimal(str(ask_level_spreads[level])))
 * 
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 889, __pyx_L1_error)

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":890
 *                                                           self._optimal_bid - Decimal(str(bid_level_spreads[level])))
 *                 ask_price = market.c_quantize_order_price(self.trading_pair,
 *                                                           self._optimal_ask + Decimal(str(ask_level_spreads[level])))             # <<<<<<<<<<<<<<
 * 
 *                 buys.append(PriceSize(bid_price, size))
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_11 = __Pyx_PyObject_GetItem(__pyx_v_ask_level_spreads, __pyx_v_level); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_10 = __Pyx_PyObject_Unicode(__pyx_t_11); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_11 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_t_10};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 890, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_t_3 = PyNumber_Add(__pyx_v_self->_optimal_ask, __pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 890, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":889
 *                 bid_price = market.c_quantize_order_price(self.trading_pair,
 *                                                           self._optimal_bid - Decimal(str(bid_level_spreads[level])))
 *                 ask_price = market.c_quantize_order_price(self.trading_pair,             # <<<<<<<<<<<<<<
 *                                                           self._optimal_ask + Decimal(str(ask_level_spreads[level])))
 * 
 */
      __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_1), __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 889, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_ask_price, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":892
 *                                                           self._optimal_ask + Decimal(str(ask_level_spreads[level])))
 * 
 *                 buys.append(PriceSize(bid_price, size))             # <<<<<<<<<<<<<<
 *                 sells.append(PriceSize(ask_price, size))
 *         return buys, sells
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_PriceSize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 892, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_v_bid_price, __pyx_v_size};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 892, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_buys, __pyx_t_5); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 892, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":893
 * 
 *                 buys.append(PriceSize(bid_price, size))
 *                 sells.append(PriceSize(ask_price, size))             # <<<<<<<<<<<<<<
 *         return buys, sells
 * 
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_PriceSize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 893, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_v_ask_price, __pyx_v_size};
        __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 2+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 893, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
      __pyx_t_12 = __Pyx_PyList_Append(__pyx_v_sells, __pyx_t_5); if (unlikely(__pyx_t_12 == ((int)-1))) __PYX_ERR(0, 893, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":886
 *         size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)
 *         if size > 0:
 *             for level in range(self.order_levels):             # <<<<<<<<<<<<<<
 *                 bid_price = market.c_quantize_order_price(self.trading_pair,
 *                                                           self._optimal_bid - Decimal(str(bid_level_spreads[level])))
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":885
 *         bid_level_spreads, ask_level_spreads = self._get_level_spreads()
 *         size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)
 *         if size > 0:             # <<<<<<<<<<<<<<
 *             for level in range(self.order_levels):
 *                 bid_price = market.c_quantize_order_price(self.trading_pair,
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":894
 *                 buys.append(PriceSize(bid_price, size))
 *                 sells.append(PriceSize(ask_price, size))
 *         return buys, sells             # <<<<<<<<<<<<<<
 * 
 *     def create_proposal_based_on_order_levels(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_buys);
  __Pyx_GIVEREF(__pyx_v_buys);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_buys)) __PYX_ERR(0, 894, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_sells);
  __Pyx_GIVEREF(__pyx_v_sells);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_v_sells)) __PYX_ERR(0, 894, __pyx_L1_error);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":878
 *         return self._create_proposal_based_on_order_override()
 * 
 *     cdef _create_proposal_based_on_order_levels(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy._create_proposal_based_on_order_levels", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_buys);
  __Pyx_XDECREF(__pyx_v_sells);
  __Pyx_XDECREF(__pyx_v_bid_level_spreads);
  __Pyx_XDECREF(__pyx_v_ask_level_spreads);
  __Pyx_XDECREF(__pyx_v_size);
  __Pyx_XDECREF(__pyx_v_level);
  __Pyx_XDECREF(__pyx_v_bid_price);
  __Pyx_XDECREF(__pyx_v_ask_price);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":896
 *         return buys, sells
 * 
 *     def create_proposal_based_on_order_levels(self):             # <<<<<<<<<<<<<<
 *         return self._create_proposal_based_on_order_levels()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_59create_proposal_based_on_order_levels(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_59create_proposal_based_on_order_levels = {"create_proposal_based_on_order_levels", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_59create_proposal_based_on_order_levels, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_59create_proposal_based_on_order_levels(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_proposal_based_on_order_levels (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("create_proposal_based_on_order_levels", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "create_proposal_based_on_order_levels", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_58create_proposal_based_on_order_levels(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_58create_proposal_based_on_order_levels(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_proposal_based_on_order_levels", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":897
 * 
 *     def create_proposal_based_on_order_levels(self):
 *         return self._create_proposal_based_on_order_levels()             # <<<<<<<<<<<<<<
 * 
 *     cdef _create_basic_proposal(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_create_proposal_based_on_order_levels(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 897, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":896
 *         return buys, sells
 * 
 *     def create_proposal_based_on_order_levels(self):             # <<<<<<<<<<<<<<
 *         return self._create_proposal_based_on_order_levels()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.create_proposal_based_on_order_levels", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":899
 *         return self._create_proposal_based_on_order_levels()
 * 
 *     cdef _create_basic_proposal(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy__create_basic_proposal(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_buys = 0;
  PyObject *__pyx_v_sells = 0;
  PyObject *__pyx_v_price = NULL;
  PyObject *__pyx_v_size = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned int __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_create_basic_proposal", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":901
 *     cdef _create_basic_proposal(self):
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *             list buys = []
 *             list sells = []
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 901, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 901, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":902
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 *             list buys = []             # <<<<<<<<<<<<<<
 *             list sells = []
 *         price = market.c_quantize_order_price(self.trading_pair, Decimal(str(self._optimal_bid)))
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 902, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buys = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":903
 *             ExchangeBase market = self._market_info.market
 *             list buys = []
 *             list sells = []             # <<<<<<<<<<<<<<
 *         price = market.c_quantize_order_price(self.trading_pair, Decimal(str(self._optimal_bid)))
 *         size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sells = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":904
 *             list buys = []
 *             list sells = []
 *         price = market.c_quantize_order_price(self.trading_pair, Decimal(str(self._optimal_bid)))             # <<<<<<<<<<<<<<
 *         size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)
 *         if size > 0:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_Unicode(__pyx_v_self->_optimal_bid); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 904, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 904, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_price = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":905
 *             list sells = []
 *         price = market.c_quantize_order_price(self.trading_pair, Decimal(str(self._optimal_bid)))
 *         size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)             # <<<<<<<<<<<<<<
 *         if size > 0:
 *             buys.append(PriceSize(price, size))
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_3))) __PYX_ERR(0, 905, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_order_amount); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_3), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 905, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_size = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":906
 *         price = market.c_quantize_order_price(self.trading_pair, Decimal(str(self._optimal_bid)))
 *         size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)
 *         if size > 0:             # <<<<<<<<<<<<<<
 *             buys.append(PriceSize(price, size))
 * 
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_size, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 906, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 906, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":907
 *         size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)
 *         if size > 0:
 *             buys.append(PriceSize(price, size))             # <<<<<<<<<<<<<<
 * 
 *         price = market.c_quantize_order_price(self.trading_pair, Decimal(str(self._optimal_ask)))
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_PriceSize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 907, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_price, __pyx_v_size};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 907, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_buys, __pyx_t_1); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 907, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":906
 *         price = market.c_quantize_order_price(self.trading_pair, Decimal(str(self._optimal_bid)))
 *         size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)
 *         if size > 0:             # <<<<<<<<<<<<<<
 *             buys.append(PriceSize(price, size))
 * 
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":909
 *             buys.append(PriceSize(price, size))
 * 
 *         price = market.c_quantize_order_price(self.trading_pair, Decimal(str(self._optimal_ask)))             # <<<<<<<<<<<<<<
 *         size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)
 *         if size > 0:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 909, __pyx_L1_error)
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyObject_Unicode(__pyx_v_self->_optimal_ask); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = NULL;
  __pyx_t_6 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_6 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, __pyx_t_4};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_6, 1+__pyx_t_6);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 909, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_t_3 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_1), __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 909, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_price, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":910
 * 
 *         price = market.c_quantize_order_price(self.trading_pair, Decimal(str(self._optimal_ask)))
 *         size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)             # <<<<<<<<<<<<<<
 *         if size > 0:
 *             sells.append(PriceSize(price, size))
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_3))) __PYX_ERR(0, 910, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_order_amount); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_3), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 910, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_size, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":911
 *         price = market.c_quantize_order_price(self.trading_pair, Decimal(str(self._optimal_ask)))
 *         size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)
 *         if size > 0:             # <<<<<<<<<<<<<<
 *             sells.append(PriceSize(price, size))
 *         return buys, sells
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_size, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 911, __pyx_L1_error)
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 911, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":912
 *         size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)
 *         if size > 0:
 *             sells.append(PriceSize(price, size))             # <<<<<<<<<<<<<<
 *         return buys, sells
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_PriceSize); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 912, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    __pyx_t_6 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_6 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_price, __pyx_v_size};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_6, 2+__pyx_t_6);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 912, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_sells, __pyx_t_1); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 912, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":911
 *         price = market.c_quantize_order_price(self.trading_pair, Decimal(str(self._optimal_ask)))
 *         size = market.c_quantize_order_amount(self.trading_pair, self._config_map.order_amount)
 *         if size > 0:             # <<<<<<<<<<<<<<
 *             sells.append(PriceSize(price, size))
 *         return buys, sells
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":913
 *         if size > 0:
 *             sells.append(PriceSize(price, size))
 *         return buys, sells             # <<<<<<<<<<<<<<
 * 
 *     def create_basic_proposal(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 913, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_buys);
  __Pyx_GIVEREF(__pyx_v_buys);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_buys)) __PYX_ERR(0, 913, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_sells);
  __Pyx_GIVEREF(__pyx_v_sells);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_sells)) __PYX_ERR(0, 913, __pyx_L1_error);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":899
 *         return self._create_proposal_based_on_order_levels()
 * 
 *     cdef _create_basic_proposal(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy._create_basic_proposal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_buys);
  __Pyx_XDECREF(__pyx_v_sells);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XDECREF(__pyx_v_size);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":915
 *         return buys, sells
 * 
 *     def create_basic_proposal(self):             # <<<<<<<<<<<<<<
 *         return self._create_basic_proposal()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_61create_basic_proposal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_61create_basic_proposal = {"create_basic_proposal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_61create_basic_proposal, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_61create_basic_proposal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_basic_proposal (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("create_basic_proposal", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "create_basic_proposal", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_60create_basic_proposal(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_60create_basic_proposal(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_basic_proposal", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":916
 * 
 *     def create_basic_proposal(self):
 *         return self._create_basic_proposal()             # <<<<<<<<<<<<<<
 * 
 *     cdef object c_create_base_proposal(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_create_basic_proposal(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":915
 *         return buys, sells
 * 
 *     def create_basic_proposal(self):             # <<<<<<<<<<<<<<
 *         return self._create_basic_proposal()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.create_basic_proposal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":918
 *         return self._create_basic_proposal()
 * 
 *     cdef object c_create_base_proposal(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_create_base_proposal(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  CYTHON_UNUSED struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_buys = 0;
  PyObject *__pyx_v_sells = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  unsigned int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_create_base_proposal", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":920
 *     cdef object c_create_base_proposal(self):
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *             list buys = []
 *             list sells = []
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 920, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 920, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":921
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 *             list buys = []             # <<<<<<<<<<<<<<
 *             list sells = []
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buys = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":922
 *             ExchangeBase market = self._market_info.market
 *             list buys = []
 *             list sells = []             # <<<<<<<<<<<<<<
 * 
 *         if self.order_override is not None and len(self.order_override) > 0:
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 922, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_sells = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":924
 *             list sells = []
 * 
 *         if self.order_override is not None and len(self.order_override) > 0:             # <<<<<<<<<<<<<<
 *             # If order_override is set, it will override order_levels
 *             buys, sells = self._create_proposal_based_on_order_override()
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_override); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_t_1 != Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_override); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 924, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_4 > 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":926
 *         if self.order_override is not None and len(self.order_override) > 0:
 *             # If order_override is set, it will override order_levels
 *             buys, sells = self._create_proposal_based_on_order_override()             # <<<<<<<<<<<<<<
 *         elif self.order_levels > 0:
 *             # Simple order levels
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_create_proposal_based_on_order_override(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 926, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 926, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 926, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 926, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7);
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 926, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 926, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    if (!(likely(PyList_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_5))) __PYX_ERR(0, 926, __pyx_L1_error)
    if (!(likely(PyList_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_6))) __PYX_ERR(0, 926, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_buys, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_sells, ((PyObject*)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":924
 *             list sells = []
 * 
 *         if self.order_override is not None and len(self.order_override) > 0:             # <<<<<<<<<<<<<<
 *             # If order_override is set, it will override order_levels
 *             buys, sells = self._create_proposal_based_on_order_override()
 */
    goto __pyx_L3;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":927
 *             # If order_override is set, it will override order_levels
 *             buys, sells = self._create_proposal_based_on_order_override()
 *         elif self.order_levels > 0:             # <<<<<<<<<<<<<<
 *             # Simple order levels
 *             buys, sells = self._create_proposal_based_on_order_levels()
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_levels); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 927, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyObject_RichCompare(__pyx_t_1, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 927, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 927, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_2) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":929
 *         elif self.order_levels > 0:
 *             # Simple order levels
 *             buys, sells = self._create_proposal_based_on_order_levels()             # <<<<<<<<<<<<<<
 *         else:
 *             # No order levels nor order_overrides. Just 1 bid and 1 ask order
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_create_proposal_based_on_order_levels(__pyx_v_self); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 929, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
      PyObject* sequence = __pyx_t_6;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 929, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_1 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      #else
      __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 929, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 929, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 929, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7);
      index = 0; __pyx_t_1 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_1)) goto __pyx_L8_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      index = 1; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L8_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 929, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L9_unpacking_done;
      __pyx_L8_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 929, __pyx_L1_error)
      __pyx_L9_unpacking_done:;
    }
    if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(0, 929, __pyx_L1_error)
    if (!(likely(PyList_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_5))) __PYX_ERR(0, 929, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_buys, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_sells, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":927
 *             # If order_override is set, it will override order_levels
 *             buys, sells = self._create_proposal_based_on_order_override()
 *         elif self.order_levels > 0:             # <<<<<<<<<<<<<<
 *             # Simple order levels
 *             buys, sells = self._create_proposal_based_on_order_levels()
 */
    goto __pyx_L3;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":932
 *         else:
 *             # No order levels nor order_overrides. Just 1 bid and 1 ask order
 *             buys, sells = self._create_basic_proposal()             # <<<<<<<<<<<<<<
 * 
 *         return Proposal(buys, sells)
 */
  /*else*/ {
    __pyx_t_6 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->_create_basic_proposal(__pyx_v_self); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 932, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if ((likely(PyTuple_CheckExact(__pyx_t_6))) || (PyList_CheckExact(__pyx_t_6))) {
      PyObject* sequence = __pyx_t_6;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 932, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_1 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_1 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_1);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 932, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_1 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 932, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      #endif
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 932, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7);
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L10_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_1 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_1)) goto __pyx_L10_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_1);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 932, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L11_unpacking_done;
      __pyx_L10_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 932, __pyx_L1_error)
      __pyx_L11_unpacking_done:;
    }
    if (!(likely(PyList_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_5))) __PYX_ERR(0, 932, __pyx_L1_error)
    if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(0, 932, __pyx_L1_error)
    __Pyx_DECREF_SET(__pyx_v_buys, ((PyObject*)__pyx_t_5));
    __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_sells, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;
  }
  __pyx_L3:;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":934
 *             buys, sells = self._create_basic_proposal()
 * 
 *         return Proposal(buys, sells)             # <<<<<<<<<<<<<<
 * 
 *     def create_base_proposal(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_Proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = NULL;
  __pyx_t_9 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_9 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_5, __pyx_v_buys, __pyx_v_sells};
    __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_9, 2+__pyx_t_9);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 934, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":918
 *         return self._create_basic_proposal()
 * 
 *     cdef object c_create_base_proposal(self):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_create_base_proposal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_buys);
  __Pyx_XDECREF(__pyx_v_sells);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":936
 *         return Proposal(buys, sells)
 * 
 *     def create_base_proposal(self):             # <<<<<<<<<<<<<<
 *         return self.c_create_base_proposal()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_63create_base_proposal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_63create_base_proposal = {"create_base_proposal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_63create_base_proposal, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_63create_base_proposal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_base_proposal (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("create_base_proposal", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "create_base_proposal", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_62create_base_proposal(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_62create_base_proposal(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("create_base_proposal", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":937
 * 
 *     def create_base_proposal(self):
 *         return self.c_create_base_proposal()             # <<<<<<<<<<<<<<
 * 
 *     cdef tuple c_get_adjusted_available_balance(self, list orders):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_create_base_proposal(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 937, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":936
 *         return Proposal(buys, sells)
 * 
 *     def create_base_proposal(self):             # <<<<<<<<<<<<<<
 *         return self.c_create_base_proposal()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.create_base_proposal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":939
 *         return self.c_create_base_proposal()
 * 
 *     cdef tuple c_get_adjusted_available_balance(self, list orders):             # <<<<<<<<<<<<<<
 *         """
 *         Calculates the available balance, plus the amount attributed to orders.
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_get_adjusted_available_balance(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_orders) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_base_balance = 0;
  PyObject *__pyx_v_quote_balance = 0;
  PyObject *__pyx_v_order = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_get_adjusted_available_balance", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":945
 *         """
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *             object base_balance = market.c_get_available_balance(self.base_asset)
 *             object quote_balance = market.c_get_available_balance(self.quote_asset)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 945, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 945, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":946
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 *             object base_balance = market.c_get_available_balance(self.base_asset)             # <<<<<<<<<<<<<<
 *             object quote_balance = market.c_get_available_balance(self.quote_asset)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 946, __pyx_L1_error)
  __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_get_available_balance(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 946, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_base_balance = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":947
 *             ExchangeBase market = self._market_info.market
 *             object base_balance = market.c_get_available_balance(self.base_asset)
 *             object quote_balance = market.c_get_available_balance(self.quote_asset)             # <<<<<<<<<<<<<<
 * 
 *         for order in orders:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_quote_asset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_2))) __PYX_ERR(0, 947, __pyx_L1_error)
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_get_available_balance(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 947, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_quote_balance = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":949
 *             object quote_balance = market.c_get_available_balance(self.quote_asset)
 * 
 *         for order in orders:             # <<<<<<<<<<<<<<
 *             if order.is_buy:
 *                 quote_balance += order.quantity * order.price
 */
  if (unlikely(__pyx_v_orders == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 949, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_orders; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_3 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 949, __pyx_L1_error)
      #endif
      if (__pyx_t_3 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 949, __pyx_L1_error)
    #else
    __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 949, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_order, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":950
 * 
 *         for order in orders:
 *             if order.is_buy:             # <<<<<<<<<<<<<<
 *                 quote_balance += order.quantity * order.price
 *             else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_n_s_is_buy); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 950, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_4 < 0))) __PYX_ERR(0, 950, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_4) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":951
 *         for order in orders:
 *             if order.is_buy:
 *                 quote_balance += order.quantity * order.price             # <<<<<<<<<<<<<<
 *             else:
 *                 base_balance += order.quantity
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_n_s_quantity); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 951, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_n_s_price); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 951, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyNumber_Multiply(__pyx_t_2, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 951, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_v_quote_balance, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 951, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF_SET(__pyx_v_quote_balance, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":950
 * 
 *         for order in orders:
 *             if order.is_buy:             # <<<<<<<<<<<<<<
 *                 quote_balance += order.quantity * order.price
 *             else:
 */
      goto __pyx_L5;
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":953
 *                 quote_balance += order.quantity * order.price
 *             else:
 *                 base_balance += order.quantity             # <<<<<<<<<<<<<<
 * 
 *         return base_balance, quote_balance
 */
    /*else*/ {
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_n_s_quantity); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 953, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyNumber_InPlaceAdd(__pyx_v_base_balance, __pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 953, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF_SET(__pyx_v_base_balance, __pyx_t_6);
      __pyx_t_6 = 0;
    }
    __pyx_L5:;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":949
 *             object quote_balance = market.c_get_available_balance(self.quote_asset)
 * 
 *         for order in orders:             # <<<<<<<<<<<<<<
 *             if order.is_buy:
 *                 quote_balance += order.quantity * order.price
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":955
 *                 base_balance += order.quantity
 * 
 *         return base_balance, quote_balance             # <<<<<<<<<<<<<<
 * 
 *     def get_adjusted_available_balance(self, orders: List[LimitOrder]):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 955, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_base_balance);
  __Pyx_GIVEREF(__pyx_v_base_balance);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_base_balance)) __PYX_ERR(0, 955, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_quote_balance);
  __Pyx_GIVEREF(__pyx_v_quote_balance);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_quote_balance)) __PYX_ERR(0, 955, __pyx_L1_error);
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":939
 *         return self.c_create_base_proposal()
 * 
 *     cdef tuple c_get_adjusted_available_balance(self, list orders):             # <<<<<<<<<<<<<<
 *         """
 *         Calculates the available balance, plus the amount attributed to orders.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_get_adjusted_available_balance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_base_balance);
  __Pyx_XDECREF(__pyx_v_quote_balance);
  __Pyx_XDECREF(__pyx_v_order);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":957
 *         return base_balance, quote_balance
 * 
 *     def get_adjusted_available_balance(self, orders: List[LimitOrder]):             # <<<<<<<<<<<<<<
 *         return self.c_get_adjusted_available_balance(orders)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_65get_adjusted_available_balance(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_65get_adjusted_available_balance = {"get_adjusted_available_balance", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_65get_adjusted_available_balance, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_65get_adjusted_available_balance(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_orders = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_adjusted_available_balance (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_orders,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_orders)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 957, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "get_adjusted_available_balance") < 0)) __PYX_ERR(0, 957, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_orders = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_adjusted_available_balance", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 957, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.get_adjusted_available_balance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_64get_adjusted_available_balance(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_orders);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_64get_adjusted_available_balance(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_orders) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_adjusted_available_balance", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":958
 * 
 *     def get_adjusted_available_balance(self, orders: List[LimitOrder]):
 *         return self.c_get_adjusted_available_balance(orders)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_apply_order_price_modifiers(self, object proposal):
 */
  __Pyx_XDECREF(__pyx_r);
  if (!(likely(PyList_CheckExact(__pyx_v_orders))||((__pyx_v_orders) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_v_orders))) __PYX_ERR(0, 958, __pyx_L1_error)
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_get_adjusted_available_balance(__pyx_v_self, ((PyObject*)__pyx_v_orders)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 958, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":957
 *         return base_balance, quote_balance
 * 
 *     def get_adjusted_available_balance(self, orders: List[LimitOrder]):             # <<<<<<<<<<<<<<
 *         return self.c_get_adjusted_available_balance(orders)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.get_adjusted_available_balance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":960
 *         return self.c_get_adjusted_available_balance(orders)
 * 
 *     cdef c_apply_order_price_modifiers(self, object proposal):             # <<<<<<<<<<<<<<
 *         if self._config_map.order_optimization_enabled:
 *             self.c_apply_order_optimization(proposal)
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_apply_order_price_modifiers(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_apply_order_price_modifiers", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":961
 * 
 *     cdef c_apply_order_price_modifiers(self, object proposal):
 *         if self._config_map.order_optimization_enabled:             # <<<<<<<<<<<<<<
 *             self.c_apply_order_optimization(proposal)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_order_optimization_enabled); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":962
 *     cdef c_apply_order_price_modifiers(self, object proposal):
 *         if self._config_map.order_optimization_enabled:
 *             self.c_apply_order_optimization(proposal)             # <<<<<<<<<<<<<<
 * 
 *         if self._config_map.add_transaction_costs:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_order_optimization(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 962, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":961
 * 
 *     cdef c_apply_order_price_modifiers(self, object proposal):
 *         if self._config_map.order_optimization_enabled:             # <<<<<<<<<<<<<<
 *             self.c_apply_order_optimization(proposal)
 * 
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":964
 *             self.c_apply_order_optimization(proposal)
 * 
 *         if self._config_map.add_transaction_costs:             # <<<<<<<<<<<<<<
 *             self.c_apply_add_transaction_costs(proposal)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_add_transaction_costs); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 964, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":965
 * 
 *         if self._config_map.add_transaction_costs:
 *             self.c_apply_add_transaction_costs(proposal)             # <<<<<<<<<<<<<<
 * 
 *     def apply_order_price_modifiers(self, proposal: Proposal):
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_add_transaction_costs(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 965, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":964
 *             self.c_apply_order_optimization(proposal)
 * 
 *         if self._config_map.add_transaction_costs:             # <<<<<<<<<<<<<<
 *             self.c_apply_add_transaction_costs(proposal)
 * 
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":960
 *         return self.c_get_adjusted_available_balance(orders)
 * 
 *     cdef c_apply_order_price_modifiers(self, object proposal):             # <<<<<<<<<<<<<<
 *         if self._config_map.order_optimization_enabled:
 *             self.c_apply_order_optimization(proposal)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_apply_order_price_modifiers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":967
 *             self.c_apply_add_transaction_costs(proposal)
 * 
 *     def apply_order_price_modifiers(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         self.c_apply_order_price_modifiers(proposal)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_67apply_order_price_modifiers(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_67apply_order_price_modifiers = {"apply_order_price_modifiers", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_67apply_order_price_modifiers, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_67apply_order_price_modifiers(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_proposal = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("apply_order_price_modifiers (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_proposal,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_proposal)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 967, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "apply_order_price_modifiers") < 0)) __PYX_ERR(0, 967, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_proposal = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("apply_order_price_modifiers", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 967, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.apply_order_price_modifiers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_66apply_order_price_modifiers(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_proposal);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_66apply_order_price_modifiers(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("apply_order_price_modifiers", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":968
 * 
 *     def apply_order_price_modifiers(self, proposal: Proposal):
 *         self.c_apply_order_price_modifiers(proposal)             # <<<<<<<<<<<<<<
 * 
 *     def apply_budget_constraint(self, proposal: Proposal):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_order_price_modifiers(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 968, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":967
 *             self.c_apply_add_transaction_costs(proposal)
 * 
 *     def apply_order_price_modifiers(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         self.c_apply_order_price_modifiers(proposal)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.apply_order_price_modifiers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":970
 *         self.c_apply_order_price_modifiers(proposal)
 * 
 *     def apply_budget_constraint(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_apply_budget_constraint(proposal)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_69apply_budget_constraint(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_69apply_budget_constraint = {"apply_budget_constraint", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_69apply_budget_constraint, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_69apply_budget_constraint(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_proposal = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("apply_budget_constraint (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_proposal,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_proposal)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 970, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "apply_budget_constraint") < 0)) __PYX_ERR(0, 970, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_proposal = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("apply_budget_constraint", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 970, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.apply_budget_constraint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_68apply_budget_constraint(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_proposal);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_68apply_budget_constraint(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("apply_budget_constraint", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":971
 * 
 *     def apply_budget_constraint(self, proposal: Proposal):
 *         return self.c_apply_budget_constraint(proposal)             # <<<<<<<<<<<<<<
 * 
 *     def adjusted_available_balance_for_orders_budget_constrain(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_budget_constraint(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 971, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":970
 *         self.c_apply_order_price_modifiers(proposal)
 * 
 *     def apply_budget_constraint(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_apply_budget_constraint(proposal)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.apply_budget_constraint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":973
 *         return self.c_apply_budget_constraint(proposal)
 * 
 *     def adjusted_available_balance_for_orders_budget_constrain(self):             # <<<<<<<<<<<<<<
 *         candidate_hanging_orders = self.hanging_orders_tracker.candidate_hanging_orders_from_pairs()
 *         non_hanging = []
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_71adjusted_available_balance_for_orders_budget_constrain(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_71adjusted_available_balance_for_orders_budget_constrain = {"adjusted_available_balance_for_orders_budget_constrain", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_71adjusted_available_balance_for_orders_budget_constrain, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_71adjusted_available_balance_for_orders_budget_constrain(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("adjusted_available_balance_for_orders_budget_constrain (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("adjusted_available_balance_for_orders_budget_constrain", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "adjusted_available_balance_for_orders_budget_constrain", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_70adjusted_available_balance_for_orders_budget_constrain(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_70adjusted_available_balance_for_orders_budget_constrain(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_v_candidate_hanging_orders = NULL;
  PyObject *__pyx_v_non_hanging = NULL;
  PyObject *__pyx_v_all_orders = NULL;
  PyObject *__pyx_v_all_non_hanging_orders = NULL;
  PyObject *__pyx_9genexpr13__pyx_v_order = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("adjusted_available_balance_for_orders_budget_constrain", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":974
 * 
 *     def adjusted_available_balance_for_orders_budget_constrain(self):
 *         candidate_hanging_orders = self.hanging_orders_tracker.candidate_hanging_orders_from_pairs()             # <<<<<<<<<<<<<<
 *         non_hanging = []
 *         if self.market_info in self._sb_order_tracker.get_limit_orders():
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_hanging_orders_tracker); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_candidate_hanging_orders_from_pa); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 974, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 974, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_candidate_hanging_orders = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":975
 *     def adjusted_available_balance_for_orders_budget_constrain(self):
 *         candidate_hanging_orders = self.hanging_orders_tracker.candidate_hanging_orders_from_pairs()
 *         non_hanging = []             # <<<<<<<<<<<<<<
 *         if self.market_info in self._sb_order_tracker.get_limit_orders():
 *             all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_non_hanging = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":976
 *         candidate_hanging_orders = self.hanging_orders_tracker.candidate_hanging_orders_from_pairs()
 *         non_hanging = []
 *         if self.market_info in self._sb_order_tracker.get_limit_orders():             # <<<<<<<<<<<<<<
 *             all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()
 *             non_hanging = [order for order in all_orders
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_market_info); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base._sb_order_tracker), __pyx_n_s_get_limit_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 976, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 976, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_6 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_t_3, Py_EQ)); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 976, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_6) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":977
 *         non_hanging = []
 *         if self.market_info in self._sb_order_tracker.get_limit_orders():
 *             all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()             # <<<<<<<<<<<<<<
 *             non_hanging = [order for order in all_orders
 *                            if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)]
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base._sb_order_tracker), __pyx_n_s_get_limit_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 977, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_market_info); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_values); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 977, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 977, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_v_all_orders = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":978
 *         if self.market_info in self._sb_order_tracker.get_limit_orders():
 *             all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()
 *             non_hanging = [order for order in all_orders             # <<<<<<<<<<<<<<
 *                            if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)]
 *         all_non_hanging_orders = list(set(non_hanging) - set(candidate_hanging_orders))
 */
    { /* enter inner scope */
      __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 978, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (likely(PyList_CheckExact(__pyx_v_all_orders)) || PyTuple_CheckExact(__pyx_v_all_orders)) {
        __pyx_t_2 = __pyx_v_all_orders; __Pyx_INCREF(__pyx_t_2);
        __pyx_t_7 = 0;
        __pyx_t_8 = NULL;
      } else {
        __pyx_t_7 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_all_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 978, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 978, __pyx_L6_error)
      }
      for (;;) {
        if (likely(!__pyx_t_8)) {
          if (likely(PyList_CheckExact(__pyx_t_2))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 978, __pyx_L6_error)
              #endif
              if (__pyx_t_7 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 978, __pyx_L6_error)
            #else
            __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 978, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            #endif
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 978, __pyx_L6_error)
              #endif
              if (__pyx_t_7 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_5); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 978, __pyx_L6_error)
            #else
            __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 978, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_5);
            #endif
          }
        } else {
          __pyx_t_5 = __pyx_t_8(__pyx_t_2);
          if (unlikely(!__pyx_t_5)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 978, __pyx_L6_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_5);
        }
        __Pyx_XDECREF_SET(__pyx_9genexpr13__pyx_v_order, __pyx_t_5);
        __pyx_t_5 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":979
 *             all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()
 *             non_hanging = [order for order in all_orders
 *                            if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)]             # <<<<<<<<<<<<<<
 *         all_non_hanging_orders = list(set(non_hanging) - set(candidate_hanging_orders))
 *         return self.c_get_adjusted_available_balance(all_non_hanging_orders)
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_n_s_is_order_id_in_hanging_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 979, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr13__pyx_v_order, __pyx_n_s_client_order_id); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 979, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_10 = NULL;
        __pyx_t_4 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_10)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_10);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
            __pyx_t_4 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_t_9};
          __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 979, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 979, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_11 = (!__pyx_t_6);
        if (__pyx_t_11) {

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":978
 *         if self.market_info in self._sb_order_tracker.get_limit_orders():
 *             all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()
 *             non_hanging = [order for order in all_orders             # <<<<<<<<<<<<<<
 *                            if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)]
 *         all_non_hanging_orders = list(set(non_hanging) - set(candidate_hanging_orders))
 */
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_9genexpr13__pyx_v_order))) __PYX_ERR(0, 978, __pyx_L6_error)

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":979
 *             all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()
 *             non_hanging = [order for order in all_orders
 *                            if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)]             # <<<<<<<<<<<<<<
 *         all_non_hanging_orders = list(set(non_hanging) - set(candidate_hanging_orders))
 *         return self.c_get_adjusted_available_balance(all_non_hanging_orders)
 */
        }

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":978
 *         if self.market_info in self._sb_order_tracker.get_limit_orders():
 *             all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()
 *             non_hanging = [order for order in all_orders             # <<<<<<<<<<<<<<
 *                            if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)]
 *         all_non_hanging_orders = list(set(non_hanging) - set(candidate_hanging_orders))
 */
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_order); __pyx_9genexpr13__pyx_v_order = 0;
      goto __pyx_L11_exit_scope;
      __pyx_L6_error:;
      __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_order); __pyx_9genexpr13__pyx_v_order = 0;
      goto __pyx_L1_error;
      __pyx_L11_exit_scope:;
    } /* exit inner scope */
    __Pyx_DECREF_SET(__pyx_v_non_hanging, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":976
 *         candidate_hanging_orders = self.hanging_orders_tracker.candidate_hanging_orders_from_pairs()
 *         non_hanging = []
 *         if self.market_info in self._sb_order_tracker.get_limit_orders():             # <<<<<<<<<<<<<<
 *             all_orders = self._sb_order_tracker.get_limit_orders()[self.market_info].values()
 *             non_hanging = [order for order in all_orders
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":980
 *             non_hanging = [order for order in all_orders
 *                            if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)]
 *         all_non_hanging_orders = list(set(non_hanging) - set(candidate_hanging_orders))             # <<<<<<<<<<<<<<
 *         return self.c_get_adjusted_available_balance(all_non_hanging_orders)
 * 
 */
  __pyx_t_3 = PySet_New(__pyx_v_non_hanging); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = PySet_New(__pyx_v_candidate_hanging_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = PyNumber_Subtract(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PySequence_ListKeepNew(__pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 980, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_v_all_non_hanging_orders = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":981
 *                            if not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order.client_order_id)]
 *         all_non_hanging_orders = list(set(non_hanging) - set(candidate_hanging_orders))
 *         return self.c_get_adjusted_available_balance(all_non_hanging_orders)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_apply_budget_constraint(self, object proposal):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_get_adjusted_available_balance(__pyx_v_self, __pyx_v_all_non_hanging_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 981, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":973
 *         return self.c_apply_budget_constraint(proposal)
 * 
 *     def adjusted_available_balance_for_orders_budget_constrain(self):             # <<<<<<<<<<<<<<
 *         candidate_hanging_orders = self.hanging_orders_tracker.candidate_hanging_orders_from_pairs()
 *         non_hanging = []
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.adjusted_available_balance_for_orders_budget_constrain", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_candidate_hanging_orders);
  __Pyx_XDECREF(__pyx_v_non_hanging);
  __Pyx_XDECREF(__pyx_v_all_orders);
  __Pyx_XDECREF(__pyx_v_all_non_hanging_orders);
  __Pyx_XDECREF(__pyx_9genexpr13__pyx_v_order);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":983
 *         return self.c_get_adjusted_available_balance(all_non_hanging_orders)
 * 
 *     cdef c_apply_budget_constraint(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_apply_budget_constraint(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_quote_size = 0;
  PyObject *__pyx_v_base_size = 0;
  PyObject *__pyx_v_adjusted_amount = 0;
  PyObject *__pyx_v_base_balance = NULL;
  PyObject *__pyx_v_quote_balance = NULL;
  PyObject *__pyx_v_buy = NULL;
  PyObject *__pyx_v_buy_fee = NULL;
  PyObject *__pyx_v_sell = NULL;
  PyObject *__pyx_9genexpr14__pyx_v_o = NULL;
  PyObject *__pyx_9genexpr15__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_apply_budget_constraint", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":985
 *     cdef c_apply_budget_constraint(self, object proposal):
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *             object quote_size
 *             object base_size
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 985, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 985, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":990
 *             object adjusted_amount
 * 
 *         base_balance, quote_balance = self.adjusted_available_balance_for_orders_budget_constrain()             # <<<<<<<<<<<<<<
 * 
 *         for buy in proposal.buys:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_adjusted_available_balance_for_o); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 990, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 990, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_2 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 1); 
    } else {
      __pyx_t_2 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_3 = PyList_GET_ITEM(sequence, 1); 
    }
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_3);
    #else
    __pyx_t_2 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 990, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 990, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 990, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_5);
    index = 0; __pyx_t_2 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_2)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_2);
    index = 1; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L3_unpacking_failed;
    __Pyx_GOTREF(__pyx_t_3);
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 2) < 0) __PYX_ERR(0, 990, __pyx_L1_error)
    __pyx_t_6 = NULL;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L4_unpacking_done;
    __pyx_L3_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 990, __pyx_L1_error)
    __pyx_L4_unpacking_done:;
  }
  __pyx_v_base_balance = __pyx_t_2;
  __pyx_t_2 = 0;
  __pyx_v_quote_balance = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":992
 *         base_balance, quote_balance = self.adjusted_available_balance_for_orders_budget_constrain()
 * 
 *         for buy in proposal.buys:             # <<<<<<<<<<<<<<
 *             buy_fee = market.c_get_fee(self.base_asset, self.quote_asset, OrderType.LIMIT, TradeType.BUY,
 *                                        buy.size, buy.price)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 992, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_3 = __pyx_t_1; __Pyx_INCREF(__pyx_t_3);
    __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 992, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 992, __pyx_L1_error)
          #endif
          if (__pyx_t_7 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 992, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 992, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 992, __pyx_L1_error)
          #endif
          if (__pyx_t_7 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_1); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 992, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 992, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_8(__pyx_t_3);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 992, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_buy, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":993
 * 
 *         for buy in proposal.buys:
 *             buy_fee = market.c_get_fee(self.base_asset, self.quote_asset, OrderType.LIMIT, TradeType.BUY,             # <<<<<<<<<<<<<<
 *                                        buy.size, buy.price)
 *             quote_size = buy.size * buy.price * (Decimal(1) + buy_fee.percent)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 993, __pyx_L1_error)
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_quote_asset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_2))) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_LIMIT); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_BUY); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":994
 *         for buy in proposal.buys:
 *             buy_fee = market.c_get_fee(self.base_asset, self.quote_asset, OrderType.LIMIT, TradeType.BUY,
 *                                        buy.size, buy.price)             # <<<<<<<<<<<<<<
 *             quote_size = buy.size * buy.price * (Decimal(1) + buy_fee.percent)
 * 
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_n_s_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 994, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_n_s_price); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 994, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":993
 * 
 *         for buy in proposal.buys:
 *             buy_fee = market.c_get_fee(self.base_asset, self.quote_asset, OrderType.LIMIT, TradeType.BUY,             # <<<<<<<<<<<<<<
 *                                        buy.size, buy.price)
 *             quote_size = buy.size * buy.price * (Decimal(1) + buy_fee.percent)
 */
    __pyx_t_12 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_get_fee(__pyx_v_market, ((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_2), __pyx_t_9, __pyx_t_10, __pyx_t_5, __pyx_t_11, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 993, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_XDECREF_SET(__pyx_v_buy_fee, __pyx_t_12);
    __pyx_t_12 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":995
 *             buy_fee = market.c_get_fee(self.base_asset, self.quote_asset, OrderType.LIMIT, TradeType.BUY,
 *                                        buy.size, buy.price)
 *             quote_size = buy.size * buy.price * (Decimal(1) + buy_fee.percent)             # <<<<<<<<<<<<<<
 * 
 *             # Adjust buy order size to use remaining balance if less than the order amount
 */
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_n_s_size); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_n_s_price); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_5 = PyNumber_Multiply(__pyx_t_12, __pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_10 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_12))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_12);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_12, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_10, __pyx_int_1};
      __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 995, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    }
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy_fee, __pyx_n_s_percent); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_10 = PyNumber_Add(__pyx_t_11, __pyx_t_12); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_12 = PyNumber_Multiply(__pyx_t_5, __pyx_t_10); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 995, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_XDECREF_SET(__pyx_v_quote_size, __pyx_t_12);
    __pyx_t_12 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":998
 * 
 *             # Adjust buy order size to use remaining balance if less than the order amount
 *             if quote_balance < quote_size:             # <<<<<<<<<<<<<<
 *                 adjusted_amount = quote_balance / (buy.price * (Decimal("1") + buy_fee.percent))
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, adjusted_amount)
 */
    __pyx_t_12 = PyObject_RichCompare(__pyx_v_quote_balance, __pyx_v_quote_size, Py_LT); __Pyx_XGOTREF(__pyx_t_12); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 998, __pyx_L1_error)
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 998, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    if (__pyx_t_13) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":999
 *             # Adjust buy order size to use remaining balance if less than the order amount
 *             if quote_balance < quote_size:
 *                 adjusted_amount = quote_balance / (buy.price * (Decimal("1") + buy_fee.percent))             # <<<<<<<<<<<<<<
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, adjusted_amount)
 *                 buy.size = adjusted_amount
 */
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_n_s_price); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 999, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 999, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_11 = NULL;
      __pyx_t_4 = 0;
      #if CYTHON_UNPACK_METHODS
      if (unlikely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_4 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_kp_u_1_2};
        __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 999, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy_fee, __pyx_n_s_percent); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 999, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_11 = PyNumber_Add(__pyx_t_10, __pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 999, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyNumber_Multiply(__pyx_t_12, __pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 999, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_11 = __Pyx_PyNumber_Divide(__pyx_v_quote_balance, __pyx_t_5); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 999, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF_SET(__pyx_v_adjusted_amount, __pyx_t_11);
      __pyx_t_11 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1000
 *             if quote_balance < quote_size:
 *                 adjusted_amount = quote_balance / (buy.price * (Decimal("1") + buy_fee.percent))
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, adjusted_amount)             # <<<<<<<<<<<<<<
 *                 buy.size = adjusted_amount
 *                 quote_balance = s_decimal_zero
 */
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1000, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_11))||((__pyx_t_11) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_11))) __PYX_ERR(0, 1000, __pyx_L1_error)
      __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_11), __pyx_v_adjusted_amount, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1000, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF_SET(__pyx_v_adjusted_amount, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1001
 *                 adjusted_amount = quote_balance / (buy.price * (Decimal("1") + buy_fee.percent))
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, adjusted_amount)
 *                 buy.size = adjusted_amount             # <<<<<<<<<<<<<<
 *                 quote_balance = s_decimal_zero
 *             elif quote_balance == s_decimal_zero:
 */
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_buy, __pyx_n_s_size, __pyx_v_adjusted_amount) < 0) __PYX_ERR(0, 1001, __pyx_L1_error)

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1002
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, adjusted_amount)
 *                 buy.size = adjusted_amount
 *                 quote_balance = s_decimal_zero             # <<<<<<<<<<<<<<
 *             elif quote_balance == s_decimal_zero:
 *                 buy.size = s_decimal_zero
 */
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_s_decimal_zero); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1002, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF_SET(__pyx_v_quote_balance, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":998
 * 
 *             # Adjust buy order size to use remaining balance if less than the order amount
 *             if quote_balance < quote_size:             # <<<<<<<<<<<<<<
 *                 adjusted_amount = quote_balance / (buy.price * (Decimal("1") + buy_fee.percent))
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, adjusted_amount)
 */
      goto __pyx_L7;
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1003
 *                 buy.size = adjusted_amount
 *                 quote_balance = s_decimal_zero
 *             elif quote_balance == s_decimal_zero:             # <<<<<<<<<<<<<<
 *                 buy.size = s_decimal_zero
 *             else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_s_decimal_zero); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1003, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_11 = PyObject_RichCompare(__pyx_v_quote_balance, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1003, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 1003, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (__pyx_t_13) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1004
 *                 quote_balance = s_decimal_zero
 *             elif quote_balance == s_decimal_zero:
 *                 buy.size = s_decimal_zero             # <<<<<<<<<<<<<<
 *             else:
 *                 quote_balance -= quote_size
 */
      __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_s_decimal_zero); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1004, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_buy, __pyx_n_s_size, __pyx_t_11) < 0) __PYX_ERR(0, 1004, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1003
 *                 buy.size = adjusted_amount
 *                 quote_balance = s_decimal_zero
 *             elif quote_balance == s_decimal_zero:             # <<<<<<<<<<<<<<
 *                 buy.size = s_decimal_zero
 *             else:
 */
      goto __pyx_L7;
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1006
 *                 buy.size = s_decimal_zero
 *             else:
 *                 quote_balance -= quote_size             # <<<<<<<<<<<<<<
 * 
 *         proposal.buys = [o for o in proposal.buys if o.size > 0]
 */
    /*else*/ {
      __pyx_t_11 = PyNumber_InPlaceSubtract(__pyx_v_quote_balance, __pyx_v_quote_size); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1006, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_DECREF_SET(__pyx_v_quote_balance, __pyx_t_11);
      __pyx_t_11 = 0;
    }
    __pyx_L7:;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":992
 *         base_balance, quote_balance = self.adjusted_available_balance_for_orders_budget_constrain()
 * 
 *         for buy in proposal.buys:             # <<<<<<<<<<<<<<
 *             buy_fee = market.c_get_fee(self.base_asset, self.quote_asset, OrderType.LIMIT, TradeType.BUY,
 *                                        buy.size, buy.price)
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1008
 *                 quote_balance -= quote_size
 * 
 *         proposal.buys = [o for o in proposal.buys if o.size > 0]             # <<<<<<<<<<<<<<
 * 
 *         for sell in proposal.sells:
 */
  { /* enter inner scope */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1008, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1008, __pyx_L11_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (likely(PyList_CheckExact(__pyx_t_11)) || PyTuple_CheckExact(__pyx_t_11)) {
      __pyx_t_5 = __pyx_t_11; __Pyx_INCREF(__pyx_t_5);
      __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1008, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1008, __pyx_L11_error)
    }
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_5))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1008, __pyx_L11_error)
            #endif
            if (__pyx_t_7 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_11 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_11); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 1008, __pyx_L11_error)
          #else
          __pyx_t_11 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1008, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_11);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1008, __pyx_L11_error)
            #endif
            if (__pyx_t_7 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_11 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_11); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 1008, __pyx_L11_error)
          #else
          __pyx_t_11 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1008, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_11);
          #endif
        }
      } else {
        __pyx_t_11 = __pyx_t_8(__pyx_t_5);
        if (unlikely(!__pyx_t_11)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1008, __pyx_L11_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_11);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr14__pyx_v_o, __pyx_t_11);
      __pyx_t_11 = 0;
      __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr14__pyx_v_o, __pyx_n_s_size); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1008, __pyx_L11_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_12 = PyObject_RichCompare(__pyx_t_11, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_12); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1008, __pyx_L11_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 1008, __pyx_L11_error)
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (__pyx_t_13) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_9genexpr14__pyx_v_o))) __PYX_ERR(0, 1008, __pyx_L11_error)
      }
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_o); __pyx_9genexpr14__pyx_v_o = 0;
    goto __pyx_L16_exit_scope;
    __pyx_L11_error:;
    __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_o); __pyx_9genexpr14__pyx_v_o = 0;
    goto __pyx_L1_error;
    __pyx_L16_exit_scope:;
  } /* exit inner scope */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_proposal, __pyx_n_s_buys, __pyx_t_3) < 0) __PYX_ERR(0, 1008, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1010
 *         proposal.buys = [o for o in proposal.buys if o.size > 0]
 * 
 *         for sell in proposal.sells:             # <<<<<<<<<<<<<<
 *             base_size = sell.size
 * 
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1010, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
    __pyx_t_5 = __pyx_t_3; __Pyx_INCREF(__pyx_t_5);
    __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1010, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1010, __pyx_L1_error)
          #endif
          if (__pyx_t_7 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 1010, __pyx_L1_error)
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1010, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1010, __pyx_L1_error)
          #endif
          if (__pyx_t_7 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 1010, __pyx_L1_error)
        #else
        __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1010, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        #endif
      }
    } else {
      __pyx_t_3 = __pyx_t_8(__pyx_t_5);
      if (unlikely(!__pyx_t_3)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1010, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_XDECREF_SET(__pyx_v_sell, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1011
 * 
 *         for sell in proposal.sells:
 *             base_size = sell.size             # <<<<<<<<<<<<<<
 * 
 *             # Adjust sell order size to use remaining balance if less than the order amount
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_sell, __pyx_n_s_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1011, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_base_size, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1014
 * 
 *             # Adjust sell order size to use remaining balance if less than the order amount
 *             if base_balance < base_size:             # <<<<<<<<<<<<<<
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, base_balance)
 *                 sell.size = adjusted_amount
 */
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_base_balance, __pyx_v_base_size, Py_LT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1014, __pyx_L1_error)
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 1014, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_13) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1015
 *             # Adjust sell order size to use remaining balance if less than the order amount
 *             if base_balance < base_size:
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, base_balance)             # <<<<<<<<<<<<<<
 *                 sell.size = adjusted_amount
 *                 base_balance = s_decimal_zero
 */
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1015, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_3))) __PYX_ERR(0, 1015, __pyx_L1_error)
      __pyx_t_12 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_3), __pyx_v_base_balance, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1015, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF_SET(__pyx_v_adjusted_amount, __pyx_t_12);
      __pyx_t_12 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1016
 *             if base_balance < base_size:
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, base_balance)
 *                 sell.size = adjusted_amount             # <<<<<<<<<<<<<<
 *                 base_balance = s_decimal_zero
 *             elif base_balance == s_decimal_zero:
 */
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_sell, __pyx_n_s_size, __pyx_v_adjusted_amount) < 0) __PYX_ERR(0, 1016, __pyx_L1_error)

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1017
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, base_balance)
 *                 sell.size = adjusted_amount
 *                 base_balance = s_decimal_zero             # <<<<<<<<<<<<<<
 *             elif base_balance == s_decimal_zero:
 *                 sell.size = s_decimal_zero
 */
      __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_s_decimal_zero); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1017, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF_SET(__pyx_v_base_balance, __pyx_t_12);
      __pyx_t_12 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1014
 * 
 *             # Adjust sell order size to use remaining balance if less than the order amount
 *             if base_balance < base_size:             # <<<<<<<<<<<<<<
 *                 adjusted_amount = market.c_quantize_order_amount(self.trading_pair, base_balance)
 *                 sell.size = adjusted_amount
 */
      goto __pyx_L19;
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1018
 *                 sell.size = adjusted_amount
 *                 base_balance = s_decimal_zero
 *             elif base_balance == s_decimal_zero:             # <<<<<<<<<<<<<<
 *                 sell.size = s_decimal_zero
 *             else:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_12, __pyx_n_s_s_decimal_zero); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1018, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    __pyx_t_3 = PyObject_RichCompare(__pyx_v_base_balance, __pyx_t_12, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1018, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 1018, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_13) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1019
 *                 base_balance = s_decimal_zero
 *             elif base_balance == s_decimal_zero:
 *                 sell.size = s_decimal_zero             # <<<<<<<<<<<<<<
 *             else:
 *                 base_balance -= base_size
 */
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_s_decimal_zero); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1019, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__Pyx_PyObject_SetAttrStr(__pyx_v_sell, __pyx_n_s_size, __pyx_t_3) < 0) __PYX_ERR(0, 1019, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1018
 *                 sell.size = adjusted_amount
 *                 base_balance = s_decimal_zero
 *             elif base_balance == s_decimal_zero:             # <<<<<<<<<<<<<<
 *                 sell.size = s_decimal_zero
 *             else:
 */
      goto __pyx_L19;
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1021
 *                 sell.size = s_decimal_zero
 *             else:
 *                 base_balance -= base_size             # <<<<<<<<<<<<<<
 * 
 *         proposal.sells = [o for o in proposal.sells if o.size > 0]
 */
    /*else*/ {
      __pyx_t_3 = PyNumber_InPlaceSubtract(__pyx_v_base_balance, __pyx_v_base_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1021, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF_SET(__pyx_v_base_balance, __pyx_t_3);
      __pyx_t_3 = 0;
    }
    __pyx_L19:;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1010
 *         proposal.buys = [o for o in proposal.buys if o.size > 0]
 * 
 *         for sell in proposal.sells:             # <<<<<<<<<<<<<<
 *             base_size = sell.size
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1023
 *                 base_balance -= base_size
 * 
 *         proposal.sells = [o for o in proposal.sells if o.size > 0]             # <<<<<<<<<<<<<<
 * 
 *     def apply_budget_constraint(self, proposal: Proposal):
 */
  { /* enter inner scope */
    __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1023, __pyx_L23_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1023, __pyx_L23_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_12 = __pyx_t_3; __Pyx_INCREF(__pyx_t_12);
      __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
    } else {
      __pyx_t_7 = -1; __pyx_t_12 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1023, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_8 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_12); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1023, __pyx_L23_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_8)) {
        if (likely(PyList_CheckExact(__pyx_t_12))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_12);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1023, __pyx_L23_error)
            #endif
            if (__pyx_t_7 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_12, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 1023, __pyx_L23_error)
          #else
          __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_12, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1023, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_12);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1023, __pyx_L23_error)
            #endif
            if (__pyx_t_7 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_12, __pyx_t_7); __Pyx_INCREF(__pyx_t_3); __pyx_t_7++; if (unlikely((0 < 0))) __PYX_ERR(0, 1023, __pyx_L23_error)
          #else
          __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_12, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1023, __pyx_L23_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_8(__pyx_t_12);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1023, __pyx_L23_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr15__pyx_v_o, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr15__pyx_v_o, __pyx_n_s_size); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1023, __pyx_L23_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_11 = PyObject_RichCompare(__pyx_t_3, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1023, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely((__pyx_t_13 < 0))) __PYX_ERR(0, 1023, __pyx_L23_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (__pyx_t_13) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_5, (PyObject*)__pyx_9genexpr15__pyx_v_o))) __PYX_ERR(0, 1023, __pyx_L23_error)
      }
    }
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_o); __pyx_9genexpr15__pyx_v_o = 0;
    goto __pyx_L28_exit_scope;
    __pyx_L23_error:;
    __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_o); __pyx_9genexpr15__pyx_v_o = 0;
    goto __pyx_L1_error;
    __pyx_L28_exit_scope:;
  } /* exit inner scope */
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_proposal, __pyx_n_s_sells, __pyx_t_5) < 0) __PYX_ERR(0, 1023, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":983
 *         return self.c_get_adjusted_available_balance(all_non_hanging_orders)
 * 
 *     cdef c_apply_budget_constraint(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_apply_budget_constraint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_quote_size);
  __Pyx_XDECREF(__pyx_v_base_size);
  __Pyx_XDECREF(__pyx_v_adjusted_amount);
  __Pyx_XDECREF(__pyx_v_base_balance);
  __Pyx_XDECREF(__pyx_v_quote_balance);
  __Pyx_XDECREF(__pyx_v_buy);
  __Pyx_XDECREF(__pyx_v_buy_fee);
  __Pyx_XDECREF(__pyx_v_sell);
  __Pyx_XDECREF(__pyx_9genexpr14__pyx_v_o);
  __Pyx_XDECREF(__pyx_9genexpr15__pyx_v_o);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1025
 *         proposal.sells = [o for o in proposal.sells if o.size > 0]
 * 
 *     def apply_budget_constraint(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_apply_budget_constraint(proposal)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_73apply_budget_constraint(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_73apply_budget_constraint = {"apply_budget_constraint", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_73apply_budget_constraint, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_73apply_budget_constraint(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_proposal = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("apply_budget_constraint (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_proposal,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_proposal)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1025, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "apply_budget_constraint") < 0)) __PYX_ERR(0, 1025, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_proposal = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("apply_budget_constraint", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1025, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.apply_budget_constraint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_72apply_budget_constraint(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_proposal);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_72apply_budget_constraint(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("apply_budget_constraint", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1026
 * 
 *     def apply_budget_constraint(self, proposal: Proposal):
 *         return self.c_apply_budget_constraint(proposal)             # <<<<<<<<<<<<<<
 * 
 *     # Compare the market price with the top bid and top ask price
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_budget_constraint(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1026, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1025
 *         proposal.sells = [o for o in proposal.sells if o.size > 0]
 * 
 *     def apply_budget_constraint(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_apply_budget_constraint(proposal)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.apply_budget_constraint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1055
 *             # If the price_above_bid is lower than the price suggested by the top pricing proposal,
 *             # lower the price and from there apply the best_order_spread to each order in the next levels
 *             proposal.buys = sorted(proposal.buys, key = lambda p: p.price, reverse = True)             # <<<<<<<<<<<<<<
 *             for i, proposed in enumerate(proposal.buys):
 *                 if proposal.buys[i].price > price_above_bid:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26c_apply_order_optimization_lambda2(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26c_apply_order_optimization_lambda2 = {"lambda2", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26c_apply_order_optimization_lambda2, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26c_apply_order_optimization_lambda2(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_p = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda2 (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_p,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_p)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1055, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "lambda2") < 0)) __PYX_ERR(0, 1055, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_p = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lambda2", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1055, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_apply_order_optimization.lambda2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_lambda_funcdef_lambda2(__pyx_self, __pyx_v_p);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda2(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda2", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_p, __pyx_n_s_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1055, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_apply_order_optimization.lambda2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1073
 *             # If the price_below_ask is higher than the price suggested by the pricing proposal,
 *             # increase your price and from there apply the best_order_spread to each order in the next levels
 *             proposal.sells = sorted(proposal.sells, key = lambda p: p.price)             # <<<<<<<<<<<<<<
 *             for i, proposed in enumerate(proposal.sells):
 *                 if proposal.sells[i].price < price_below_ask:
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26c_apply_order_optimization_1lambda3(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26c_apply_order_optimization_1lambda3 = {"lambda3", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26c_apply_order_optimization_1lambda3, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26c_apply_order_optimization_1lambda3(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_p = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda3 (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_p,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_p)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1073, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "lambda3") < 0)) __PYX_ERR(0, 1073, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_p = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("lambda3", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1073, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_apply_order_optimization.lambda3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_lambda_funcdef_lambda3(__pyx_self, __pyx_v_p);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda3(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_p) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("lambda3", 1);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_p, __pyx_n_s_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_apply_order_optimization.lambda3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1029
 * 
 *     # Compare the market price with the top bid and top ask price
 *     cdef c_apply_order_optimization(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_apply_order_optimization(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_own_buy_size = 0;
  PyObject *__pyx_v_own_sell_size = 0;
  PyObject *__pyx_v_order = NULL;
  PyObject *__pyx_v_top_bid_price = NULL;
  PyObject *__pyx_v_price_quantum = NULL;
  PyObject *__pyx_v_price_above_bid = NULL;
  PyObject *__pyx_v_i = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_proposed = NULL;
  PyObject *__pyx_v_top_ask_price = NULL;
  PyObject *__pyx_v_price_below_ask = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_apply_order_optimization", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1031
 *     cdef c_apply_order_optimization(self, object proposal):
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *             object own_buy_size = s_decimal_zero
 *             object own_sell_size = s_decimal_zero
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1031, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 1031, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1032
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 *             object own_buy_size = s_decimal_zero             # <<<<<<<<<<<<<<
 *             object own_sell_size = s_decimal_zero
 *             object best_order_spread
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_s_decimal_zero); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1032, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_own_buy_size = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1033
 *             ExchangeBase market = self._market_info.market
 *             object own_buy_size = s_decimal_zero
 *             object own_sell_size = s_decimal_zero             # <<<<<<<<<<<<<<
 *             object best_order_spread
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_s_decimal_zero); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1033, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_own_sell_size = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1036
 *             object best_order_spread
 * 
 *         for order in self.active_orders:             # <<<<<<<<<<<<<<
 *             if order.is_buy:
 *                 own_buy_size = order.quantity
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_active_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1036, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1036, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1036, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1036, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1036, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1036, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1036, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1036, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1036, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1036, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_order, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1037
 * 
 *         for order in self.active_orders:
 *             if order.is_buy:             # <<<<<<<<<<<<<<
 *                 own_buy_size = order.quantity
 *             else:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_n_s_is_buy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1037, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_5) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1038
 *         for order in self.active_orders:
 *             if order.is_buy:
 *                 own_buy_size = order.quantity             # <<<<<<<<<<<<<<
 *             else:
 *                 own_sell_size = order.quantity
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_n_s_quantity); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1038, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_own_buy_size, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1037
 * 
 *         for order in self.active_orders:
 *             if order.is_buy:             # <<<<<<<<<<<<<<
 *                 own_buy_size = order.quantity
 *             else:
 */
      goto __pyx_L5;
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1040
 *                 own_buy_size = order.quantity
 *             else:
 *                 own_sell_size = order.quantity             # <<<<<<<<<<<<<<
 * 
 *         if len(proposal.buys) > 0:
 */
    /*else*/ {
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_n_s_quantity); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1040, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF_SET(__pyx_v_own_sell_size, __pyx_t_1);
      __pyx_t_1 = 0;
    }
    __pyx_L5:;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1036
 *             object best_order_spread
 * 
 *         for order in self.active_orders:             # <<<<<<<<<<<<<<
 *             if order.is_buy:
 *                 own_buy_size = order.quantity
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1042
 *                 own_sell_size = order.quantity
 * 
 *         if len(proposal.buys) > 0:             # <<<<<<<<<<<<<<
 *             # Get the top bid price in the market using order_optimization_depth and your buy order volume
 *             top_bid_price = self._market_info.get_price_for_volume(
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1042, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1042, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (__pyx_t_3 > 0);
  if (__pyx_t_5) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1044
 *         if len(proposal.buys) > 0:
 *             # Get the top bid price in the market using order_optimization_depth and your buy order volume
 *             top_bid_price = self._market_info.get_price_for_volume(             # <<<<<<<<<<<<<<
 *                 False, own_buy_size).result_price
 *             price_quantum = market.c_get_order_price_quantum(
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_get_price_for_volume); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1045
 *             # Get the top bid price in the market using order_optimization_depth and your buy order volume
 *             top_bid_price = self._market_info.get_price_for_volume(
 *                 False, own_buy_size).result_price             # <<<<<<<<<<<<<<
 *             price_quantum = market.c_get_order_price_quantum(
 *                 self.trading_pair,
 */
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, Py_False, __pyx_v_own_buy_size};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1044, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_result_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_top_bid_price = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1047
 *                 False, own_buy_size).result_price
 *             price_quantum = market.c_get_order_price_quantum(
 *                 self.trading_pair,             # <<<<<<<<<<<<<<
 *                 top_bid_price
 *             )
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1047, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 1047, __pyx_L1_error)

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1046
 *             top_bid_price = self._market_info.get_price_for_volume(
 *                 False, own_buy_size).result_price
 *             price_quantum = market.c_get_order_price_quantum(             # <<<<<<<<<<<<<<
 *                 self.trading_pair,
 *                 top_bid_price
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_get_order_price_quantum(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_1), __pyx_v_top_bid_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1046, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_price_quantum = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1051
 *             )
 *             # Get the price above the top bid
 *             price_above_bid = (ceil(top_bid_price / price_quantum) + 1) * price_quantum             # <<<<<<<<<<<<<<
 * 
 *             # If the price_above_bid is lower than the price suggested by the top pricing proposal,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_ceil); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyNumber_Divide(__pyx_v_top_bid_price, __pyx_v_price_quantum); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_6};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1051, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_t_2, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyNumber_Multiply(__pyx_t_1, __pyx_v_price_quantum); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1051, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_price_above_bid = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1055
 *             # If the price_above_bid is lower than the price suggested by the top pricing proposal,
 *             # lower the price and from there apply the best_order_spread to each order in the next levels
 *             proposal.buys = sorted(proposal.buys, key = lambda p: p.price, reverse = True)             # <<<<<<<<<<<<<<
 *             for i, proposed in enumerate(proposal.buys):
 *                 if proposal.buys[i].price > price_above_bid:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1055, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1055, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2)) __PYX_ERR(0, 1055, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1055, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26c_apply_order_optimization_lambda2, 0, __pyx_n_s_AvellanedaMarketMakingStrategy_c_2, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1055, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_key, __pyx_t_6) < 0) __PYX_ERR(0, 1055, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_reverse, Py_True) < 0) __PYX_ERR(0, 1055, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_sorted, __pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1055, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_proposal, __pyx_n_s_buys, __pyx_t_6) < 0) __PYX_ERR(0, 1055, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1056
 *             # lower the price and from there apply the best_order_spread to each order in the next levels
 *             proposal.buys = sorted(proposal.buys, key = lambda p: p.price, reverse = True)
 *             for i, proposed in enumerate(proposal.buys):             # <<<<<<<<<<<<<<
 *                 if proposal.buys[i].price > price_above_bid:
 *                     proposal.buys[i].price = market.c_quantize_order_price(self.trading_pair, price_above_bid)
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_6 = __pyx_int_0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1056, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_1 = __pyx_t_2; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1056, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1056, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1056, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1056, __pyx_L1_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1056, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1056, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1056, __pyx_L1_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1056, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_4(__pyx_t_1);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1056, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_proposed, __pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_6);
      __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_6, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1056, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6);
      __pyx_t_6 = __pyx_t_2;
      __pyx_t_2 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1057
 *             proposal.buys = sorted(proposal.buys, key = lambda p: p.price, reverse = True)
 *             for i, proposed in enumerate(proposal.buys):
 *                 if proposal.buys[i].price > price_above_bid:             # <<<<<<<<<<<<<<
 *                     proposal.buys[i].price = market.c_quantize_order_price(self.trading_pair, price_above_bid)
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1057, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_v_i); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1057, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1057, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = PyObject_RichCompare(__pyx_t_2, __pyx_v_price_above_bid, Py_GT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1057, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1057, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (__pyx_t_5) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1058
 *             for i, proposed in enumerate(proposal.buys):
 *                 if proposal.buys[i].price > price_above_bid:
 *                     proposal.buys[i].price = market.c_quantize_order_price(self.trading_pair, price_above_bid)             # <<<<<<<<<<<<<<
 * 
 *         if len(proposal.sells) > 0:
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1058, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (!(likely(PyUnicode_CheckExact(__pyx_t_8))||((__pyx_t_8) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_8))) __PYX_ERR(0, 1058, __pyx_L1_error)
        __pyx_t_2 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_8), __pyx_v_price_above_bid); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1058, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1058, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = __Pyx_PyObject_GetItem(__pyx_t_8, __pyx_v_i); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1058, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (__Pyx_PyObject_SetAttrStr(__pyx_t_9, __pyx_n_s_price, __pyx_t_2) < 0) __PYX_ERR(0, 1058, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1057
 *             proposal.buys = sorted(proposal.buys, key = lambda p: p.price, reverse = True)
 *             for i, proposed in enumerate(proposal.buys):
 *                 if proposal.buys[i].price > price_above_bid:             # <<<<<<<<<<<<<<
 *                     proposal.buys[i].price = market.c_quantize_order_price(self.trading_pair, price_above_bid)
 * 
 */
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1056
 *             # lower the price and from there apply the best_order_spread to each order in the next levels
 *             proposal.buys = sorted(proposal.buys, key = lambda p: p.price, reverse = True)
 *             for i, proposed in enumerate(proposal.buys):             # <<<<<<<<<<<<<<
 *                 if proposal.buys[i].price > price_above_bid:
 *                     proposal.buys[i].price = market.c_quantize_order_price(self.trading_pair, price_above_bid)
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1042
 *                 own_sell_size = order.quantity
 * 
 *         if len(proposal.buys) > 0:             # <<<<<<<<<<<<<<
 *             # Get the top bid price in the market using order_optimization_depth and your buy order volume
 *             top_bid_price = self._market_info.get_price_for_volume(
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1060
 *                     proposal.buys[i].price = market.c_quantize_order_price(self.trading_pair, price_above_bid)
 * 
 *         if len(proposal.sells) > 0:             # <<<<<<<<<<<<<<
 *             # Get the top ask price in the market using order_optimization_depth and your sell order volume
 *             top_ask_price = self._market_info.get_price_for_volume(
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1060, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_5 = (__pyx_t_3 > 0);
  if (__pyx_t_5) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1062
 *         if len(proposal.sells) > 0:
 *             # Get the top ask price in the market using order_optimization_depth and your sell order volume
 *             top_ask_price = self._market_info.get_price_for_volume(             # <<<<<<<<<<<<<<
 *                 True, own_sell_size).result_price
 *             price_quantum = market.c_get_order_price_quantum(
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_get_price_for_volume); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1062, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1063
 *             # Get the top ask price in the market using order_optimization_depth and your sell order volume
 *             top_ask_price = self._market_info.get_price_for_volume(
 *                 True, own_sell_size).result_price             # <<<<<<<<<<<<<<
 *             price_quantum = market.c_get_order_price_quantum(
 *                 self.trading_pair,
 */
    __pyx_t_9 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_9, Py_True, __pyx_v_own_sell_size};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1062, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_result_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1063, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_v_top_ask_price = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1065
 *                 True, own_sell_size).result_price
 *             price_quantum = market.c_get_order_price_quantum(
 *                 self.trading_pair,             # <<<<<<<<<<<<<<
 *                 top_ask_price
 *             )
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1065, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 1065, __pyx_L1_error)

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1064
 *             top_ask_price = self._market_info.get_price_for_volume(
 *                 True, own_sell_size).result_price
 *             price_quantum = market.c_get_order_price_quantum(             # <<<<<<<<<<<<<<
 *                 self.trading_pair,
 *                 top_ask_price
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_get_order_price_quantum(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_1), __pyx_v_top_ask_price); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1064, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF_SET(__pyx_v_price_quantum, __pyx_t_6);
    __pyx_t_6 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1069
 *             )
 *             # Get the price below the top ask
 *             price_below_ask = (floor(top_ask_price / price_quantum) - 1) * price_quantum             # <<<<<<<<<<<<<<
 * 
 *             # If the price_below_ask is higher than the price suggested by the pricing proposal,
 */
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_floor); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyNumber_Divide(__pyx_v_top_ask_price, __pyx_v_price_quantum); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_2 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_9};
      __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_7, 1+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1069, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }
    __pyx_t_1 = __Pyx_PyInt_SubtractObjC(__pyx_t_6, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = PyNumber_Multiply(__pyx_t_1, __pyx_v_price_quantum); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_price_below_ask = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1073
 *             # If the price_below_ask is higher than the price suggested by the pricing proposal,
 *             # increase your price and from there apply the best_order_spread to each order in the next levels
 *             proposal.sells = sorted(proposal.sells, key = lambda p: p.price)             # <<<<<<<<<<<<<<
 *             for i, proposed in enumerate(proposal.sells):
 *                 if proposal.sells[i].price < price_below_ask:
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_6);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6)) __PYX_ERR(0, 1073, __pyx_L1_error);
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_9 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26c_apply_order_optimization_1lambda3, 0, __pyx_n_s_AvellanedaMarketMakingStrategy_c_2, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (PyDict_SetItem(__pyx_t_6, __pyx_n_s_key, __pyx_t_9) < 0) __PYX_ERR(0, 1073, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_sorted, __pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_proposal, __pyx_n_s_sells, __pyx_t_9) < 0) __PYX_ERR(0, 1073, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1074
 *             # increase your price and from there apply the best_order_spread to each order in the next levels
 *             proposal.sells = sorted(proposal.sells, key = lambda p: p.price)
 *             for i, proposed in enumerate(proposal.sells):             # <<<<<<<<<<<<<<
 *                 if proposal.sells[i].price < price_below_ask:
 *                     proposal.sells[i].price = market.c_quantize_order_price(self.trading_pair, price_below_ask)
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_9 = __pyx_int_0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1074, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
      __pyx_t_1 = __pyx_t_6; __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = 0;
      __pyx_t_4 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1074, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1074, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    for (;;) {
      if (likely(!__pyx_t_4)) {
        if (likely(PyList_CheckExact(__pyx_t_1))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1074, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1074, __pyx_L1_error)
          #else
          __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1074, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1074, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_6); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1074, __pyx_L1_error)
          #else
          __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1074, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          #endif
        }
      } else {
        __pyx_t_6 = __pyx_t_4(__pyx_t_1);
        if (unlikely(!__pyx_t_6)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1074, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_6);
      }
      __Pyx_XDECREF_SET(__pyx_v_proposed, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_9);
      __pyx_t_6 = __Pyx_PyInt_AddObjC(__pyx_t_9, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1074, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_9);
      __pyx_t_9 = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1075
 *             proposal.sells = sorted(proposal.sells, key = lambda p: p.price)
 *             for i, proposed in enumerate(proposal.sells):
 *                 if proposal.sells[i].price < price_below_ask:             # <<<<<<<<<<<<<<
 *                     proposal.sells[i].price = market.c_quantize_order_price(self.trading_pair, price_below_ask)
 * 
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_6, __pyx_v_i); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_price); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyObject_RichCompare(__pyx_t_6, __pyx_v_price_below_ask, Py_LT); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1075, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (__pyx_t_5) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1076
 *             for i, proposed in enumerate(proposal.sells):
 *                 if proposal.sells[i].price < price_below_ask:
 *                     proposal.sells[i].price = market.c_quantize_order_price(self.trading_pair, price_below_ask)             # <<<<<<<<<<<<<<
 * 
 *     def apply_order_optimization(self, proposal: Proposal):
 */
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1076, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_2))) __PYX_ERR(0, 1076, __pyx_L1_error)
        __pyx_t_6 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_2), __pyx_v_price_below_ask); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1076, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1076, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_8 = __Pyx_PyObject_GetItem(__pyx_t_2, __pyx_v_i); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1076, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        if (__Pyx_PyObject_SetAttrStr(__pyx_t_8, __pyx_n_s_price, __pyx_t_6) < 0) __PYX_ERR(0, 1076, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1075
 *             proposal.sells = sorted(proposal.sells, key = lambda p: p.price)
 *             for i, proposed in enumerate(proposal.sells):
 *                 if proposal.sells[i].price < price_below_ask:             # <<<<<<<<<<<<<<
 *                     proposal.sells[i].price = market.c_quantize_order_price(self.trading_pair, price_below_ask)
 * 
 */
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1074
 *             # increase your price and from there apply the best_order_spread to each order in the next levels
 *             proposal.sells = sorted(proposal.sells, key = lambda p: p.price)
 *             for i, proposed in enumerate(proposal.sells):             # <<<<<<<<<<<<<<
 *                 if proposal.sells[i].price < price_below_ask:
 *                     proposal.sells[i].price = market.c_quantize_order_price(self.trading_pair, price_below_ask)
 */
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1060
 *                     proposal.buys[i].price = market.c_quantize_order_price(self.trading_pair, price_above_bid)
 * 
 *         if len(proposal.sells) > 0:             # <<<<<<<<<<<<<<
 *             # Get the top ask price in the market using order_optimization_depth and your sell order volume
 *             top_ask_price = self._market_info.get_price_for_volume(
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1029
 * 
 *     # Compare the market price with the top bid and top ask price
 *     cdef c_apply_order_optimization(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_apply_order_optimization", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_own_buy_size);
  __Pyx_XDECREF(__pyx_v_own_sell_size);
  __Pyx_XDECREF(__pyx_v_order);
  __Pyx_XDECREF(__pyx_v_top_bid_price);
  __Pyx_XDECREF(__pyx_v_price_quantum);
  __Pyx_XDECREF(__pyx_v_price_above_bid);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_proposed);
  __Pyx_XDECREF(__pyx_v_top_ask_price);
  __Pyx_XDECREF(__pyx_v_price_below_ask);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1078
 *                     proposal.sells[i].price = market.c_quantize_order_price(self.trading_pair, price_below_ask)
 * 
 *     def apply_order_optimization(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_apply_order_optimization(proposal)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_75apply_order_optimization(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_75apply_order_optimization = {"apply_order_optimization", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_75apply_order_optimization, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_75apply_order_optimization(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_proposal = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("apply_order_optimization (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_proposal,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_proposal)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1078, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "apply_order_optimization") < 0)) __PYX_ERR(0, 1078, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_proposal = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("apply_order_optimization", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1078, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.apply_order_optimization", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_74apply_order_optimization(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_proposal);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_74apply_order_optimization(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("apply_order_optimization", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1079
 * 
 *     def apply_order_optimization(self, proposal: Proposal):
 *         return self.c_apply_order_optimization(proposal)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_apply_order_amount_eta_transformation(self, object proposal):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_order_optimization(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1079, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1078
 *                     proposal.sells[i].price = market.c_quantize_order_price(self.trading_pair, price_below_ask)
 * 
 *     def apply_order_optimization(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_apply_order_optimization(proposal)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.apply_order_optimization", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1081
 *         return self.c_apply_order_optimization(proposal)
 * 
 *     cdef c_apply_order_amount_eta_transformation(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_apply_order_amount_eta_transformation(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_trading_pair = 0;
  PyObject *__pyx_v_inventory = NULL;
  PyObject *__pyx_v_q_target = NULL;
  PyObject *__pyx_v_q = NULL;
  PyObject *__pyx_v_i = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_proposed = NULL;
  PyObject *__pyx_9genexpr16__pyx_v_o = NULL;
  PyObject *__pyx_9genexpr17__pyx_v_o = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_apply_order_amount_eta_transformation", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1083
 *     cdef c_apply_order_amount_eta_transformation(self, object proposal):
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *             str trading_pair = self._market_info.trading_pair
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1083, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 1083, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1084
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 *             str trading_pair = self._market_info.trading_pair             # <<<<<<<<<<<<<<
 * 
 *         # Order amounts should be changed only if order_override is not active
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 1084, __pyx_L1_error)
  __pyx_v_trading_pair = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1087
 * 
 *         # Order amounts should be changed only if order_override is not active
 *         if (self.order_override is None) or (len(self.order_override) == 0):             # <<<<<<<<<<<<<<
 *             # eta parameter is described in the paper as the shape parameter for having exponentially decreasing order amount
 *             # for orders that go against inventory target (i.e. Want to buy when excess inventory or sell when deficit inventory)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_override); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_t_1 == Py_None);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (!__pyx_t_3) {
  } else {
    __pyx_t_2 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_override); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1087, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__pyx_t_4 == 0);
  __pyx_t_2 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1093
 *             # q cannot be in absolute values - cannot be dependent on the size of the inventory or amount of base asset
 *             # because it's a scaling factor
 *             inventory = Decimal(str(self.c_calculate_inventory()))             # <<<<<<<<<<<<<<
 * 
 *             if inventory == 0:
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1093, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_calculate_inventory(__pyx_v_self); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1093, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_Unicode(__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1093, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_t_7};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1093, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_v_inventory = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1095
 *             inventory = Decimal(str(self.c_calculate_inventory()))
 * 
 *             if inventory == 0:             # <<<<<<<<<<<<<<
 *                 return
 * 
 */
    __pyx_t_2 = (__Pyx_PyInt_BoolEqObjC(__pyx_v_inventory, __pyx_int_0, 0, 0)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1095, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1096
 * 
 *             if inventory == 0:
 *                 return             # <<<<<<<<<<<<<<
 * 
 *             q_target = Decimal(str(self.c_calculate_target_inventory()))
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1095
 *             inventory = Decimal(str(self.c_calculate_inventory()))
 * 
 *             if inventory == 0:             # <<<<<<<<<<<<<<
 *                 return
 * 
 */
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1098
 *                 return
 * 
 *             q_target = Decimal(str(self.c_calculate_target_inventory()))             # <<<<<<<<<<<<<<
 *             q = (market.get_balance(self.base_asset) - q_target) / (inventory)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1098, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_calculate_target_inventory(__pyx_v_self); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1098, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = __Pyx_PyObject_Unicode(__pyx_t_7); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1098, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1098, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_v_q_target = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1099
 * 
 *             q_target = Decimal(str(self.c_calculate_target_inventory()))
 *             q = (market.get_balance(self.base_asset) - q_target) / (inventory)             # <<<<<<<<<<<<<<
 * 
 *             if len(proposal.buys) > 0:
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_market), __pyx_n_s_get_balance); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1099, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base_asset); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1099, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    __pyx_t_8 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_8 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_t_6};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1099, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
    __pyx_t_5 = PyNumber_Subtract(__pyx_t_1, __pyx_v_q_target); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1099, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_5, __pyx_v_inventory); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1099, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_q = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1101
 *             q = (market.get_balance(self.base_asset) - q_target) / (inventory)
 * 
 *             if len(proposal.buys) > 0:             # <<<<<<<<<<<<<<
 *                 if q > 0:
 *                     for i, proposed in enumerate(proposal.buys):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1101, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1101, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = (__pyx_t_4 > 0);
    if (__pyx_t_2) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1102
 * 
 *             if len(proposal.buys) > 0:
 *                 if q > 0:             # <<<<<<<<<<<<<<
 *                     for i, proposed in enumerate(proposal.buys):
 * 
 */
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_q, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1102, __pyx_L1_error)
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1102, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_2) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1103
 *             if len(proposal.buys) > 0:
 *                 if q > 0:
 *                     for i, proposed in enumerate(proposal.buys):             # <<<<<<<<<<<<<<
 * 
 *                         proposal.buys[i].size = market.c_quantize_order_amount(trading_pair, proposal.buys[i].size * Decimal.exp(-self.eta * q))
 */
        __Pyx_INCREF(__pyx_int_0);
        __pyx_t_1 = __pyx_int_0;
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1103, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
          __pyx_t_6 = __pyx_t_5; __Pyx_INCREF(__pyx_t_6);
          __pyx_t_4 = 0;
          __pyx_t_9 = NULL;
        } else {
          __pyx_t_4 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1103, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1103, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        for (;;) {
          if (likely(!__pyx_t_9)) {
            if (likely(PyList_CheckExact(__pyx_t_6))) {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_6);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1103, __pyx_L1_error)
                #endif
                if (__pyx_t_4 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_5 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 1103, __pyx_L1_error)
              #else
              __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1103, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
              #endif
            } else {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_6);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1103, __pyx_L1_error)
                #endif
                if (__pyx_t_4 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 1103, __pyx_L1_error)
              #else
              __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_6, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1103, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
              #endif
            }
          } else {
            __pyx_t_5 = __pyx_t_9(__pyx_t_6);
            if (unlikely(!__pyx_t_5)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 1103, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_5);
          }
          __Pyx_XDECREF_SET(__pyx_v_proposed, __pyx_t_5);
          __pyx_t_5 = 0;
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
          __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1103, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1);
          __pyx_t_1 = __pyx_t_5;
          __pyx_t_5 = 0;

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1105
 *                     for i, proposed in enumerate(proposal.buys):
 * 
 *                         proposal.buys[i].size = market.c_quantize_order_amount(trading_pair, proposal.buys[i].size * Decimal.exp(-self.eta * q))             # <<<<<<<<<<<<<<
 *                     proposal.buys = [o for o in proposal.buys if o.size > 0]
 * 
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1105, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_7 = __Pyx_PyObject_GetItem(__pyx_t_5, __pyx_v_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1105, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1105, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1105, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_exp); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1105, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_eta); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1105, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_12 = PyNumber_Negative(__pyx_t_10); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1105, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_10 = PyNumber_Multiply(__pyx_t_12, __pyx_v_q); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1105, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_12 = NULL;
          __pyx_t_8 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_11))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_11);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_11, function);
              __pyx_t_8 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_t_10};
            __pyx_t_7 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
            if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1105, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          }
          __pyx_t_11 = PyNumber_Multiply(__pyx_t_5, __pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1105, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __pyx_t_7 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), __pyx_v_trading_pair, __pyx_t_11, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1105, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1105, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_t_11, __pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1105, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_size, __pyx_t_7) < 0) __PYX_ERR(0, 1105, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1103
 *             if len(proposal.buys) > 0:
 *                 if q > 0:
 *                     for i, proposed in enumerate(proposal.buys):             # <<<<<<<<<<<<<<
 * 
 *                         proposal.buys[i].size = market.c_quantize_order_amount(trading_pair, proposal.buys[i].size * Decimal.exp(-self.eta * q))
 */
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1106
 * 
 *                         proposal.buys[i].size = market.c_quantize_order_amount(trading_pair, proposal.buys[i].size * Decimal.exp(-self.eta * q))
 *                     proposal.buys = [o for o in proposal.buys if o.size > 0]             # <<<<<<<<<<<<<<
 * 
 *             if len(proposal.sells) > 0:
 */
        { /* enter inner scope */
          __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1106, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1106, __pyx_L14_error)
          __Pyx_GOTREF(__pyx_t_6);
          if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
            __pyx_t_5 = __pyx_t_6; __Pyx_INCREF(__pyx_t_5);
            __pyx_t_4 = 0;
            __pyx_t_9 = NULL;
          } else {
            __pyx_t_4 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1106, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1106, __pyx_L14_error)
          }
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          for (;;) {
            if (likely(!__pyx_t_9)) {
              if (likely(PyList_CheckExact(__pyx_t_5))) {
                {
                  Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
                  #if !CYTHON_ASSUME_SAFE_MACROS
                  if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1106, __pyx_L14_error)
                  #endif
                  if (__pyx_t_4 >= __pyx_temp) break;
                }
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_6 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 1106, __pyx_L14_error)
                #else
                __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1106, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_6);
                #endif
              } else {
                {
                  Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
                  #if !CYTHON_ASSUME_SAFE_MACROS
                  if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1106, __pyx_L14_error)
                  #endif
                  if (__pyx_t_4 >= __pyx_temp) break;
                }
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_4); __Pyx_INCREF(__pyx_t_6); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 1106, __pyx_L14_error)
                #else
                __pyx_t_6 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1106, __pyx_L14_error)
                __Pyx_GOTREF(__pyx_t_6);
                #endif
              }
            } else {
              __pyx_t_6 = __pyx_t_9(__pyx_t_5);
              if (unlikely(!__pyx_t_6)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 1106, __pyx_L14_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_6);
            }
            __Pyx_XDECREF_SET(__pyx_9genexpr16__pyx_v_o, __pyx_t_6);
            __pyx_t_6 = 0;
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr16__pyx_v_o, __pyx_n_s_size); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1106, __pyx_L14_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_7 = PyObject_RichCompare(__pyx_t_6, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1106, __pyx_L14_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1106, __pyx_L14_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            if (__pyx_t_2) {
              if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_9genexpr16__pyx_v_o))) __PYX_ERR(0, 1106, __pyx_L14_error)
            }
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_o); __pyx_9genexpr16__pyx_v_o = 0;
          goto __pyx_L19_exit_scope;
          __pyx_L14_error:;
          __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_o); __pyx_9genexpr16__pyx_v_o = 0;
          goto __pyx_L1_error;
          __pyx_L19_exit_scope:;
        } /* exit inner scope */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_proposal, __pyx_n_s_buys, __pyx_t_1) < 0) __PYX_ERR(0, 1106, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1102
 * 
 *             if len(proposal.buys) > 0:
 *                 if q > 0:             # <<<<<<<<<<<<<<
 *                     for i, proposed in enumerate(proposal.buys):
 * 
 */
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1101
 *             q = (market.get_balance(self.base_asset) - q_target) / (inventory)
 * 
 *             if len(proposal.buys) > 0:             # <<<<<<<<<<<<<<
 *                 if q > 0:
 *                     for i, proposed in enumerate(proposal.buys):
 */
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1108
 *                     proposal.buys = [o for o in proposal.buys if o.size > 0]
 * 
 *             if len(proposal.sells) > 0:             # <<<<<<<<<<<<<<
 *                 if q < 0:
 *                     for i, proposed in enumerate(proposal.sells):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1108, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = (__pyx_t_4 > 0);
    if (__pyx_t_2) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1109
 * 
 *             if len(proposal.sells) > 0:
 *                 if q < 0:             # <<<<<<<<<<<<<<
 *                     for i, proposed in enumerate(proposal.sells):
 *                         proposal.sells[i].size = market.c_quantize_order_amount(trading_pair, proposal.sells[i].size * Decimal.exp(self.eta * q))
 */
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_q, __pyx_int_0, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1109, __pyx_L1_error)
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1109, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_2) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1110
 *             if len(proposal.sells) > 0:
 *                 if q < 0:
 *                     for i, proposed in enumerate(proposal.sells):             # <<<<<<<<<<<<<<
 *                         proposal.sells[i].size = market.c_quantize_order_amount(trading_pair, proposal.sells[i].size * Decimal.exp(self.eta * q))
 *                     proposal.sells = [o for o in proposal.sells if o.size > 0]
 */
        __Pyx_INCREF(__pyx_int_0);
        __pyx_t_1 = __pyx_int_0;
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1110, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
          __pyx_t_7 = __pyx_t_5; __Pyx_INCREF(__pyx_t_7);
          __pyx_t_4 = 0;
          __pyx_t_9 = NULL;
        } else {
          __pyx_t_4 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1110, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1110, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        for (;;) {
          if (likely(!__pyx_t_9)) {
            if (likely(PyList_CheckExact(__pyx_t_7))) {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_7);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1110, __pyx_L1_error)
                #endif
                if (__pyx_t_4 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_5 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 1110, __pyx_L1_error)
              #else
              __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_7, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1110, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
              #endif
            } else {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_7);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1110, __pyx_L1_error)
                #endif
                if (__pyx_t_4 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_4); __Pyx_INCREF(__pyx_t_5); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 1110, __pyx_L1_error)
              #else
              __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_7, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1110, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_5);
              #endif
            }
          } else {
            __pyx_t_5 = __pyx_t_9(__pyx_t_7);
            if (unlikely(!__pyx_t_5)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 1110, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_5);
          }
          __Pyx_XDECREF_SET(__pyx_v_proposed, __pyx_t_5);
          __pyx_t_5 = 0;
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
          __pyx_t_5 = __Pyx_PyInt_AddObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1110, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_1);
          __pyx_t_1 = __pyx_t_5;
          __pyx_t_5 = 0;

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1111
 *                 if q < 0:
 *                     for i, proposed in enumerate(proposal.sells):
 *                         proposal.sells[i].size = market.c_quantize_order_amount(trading_pair, proposal.sells[i].size * Decimal.exp(self.eta * q))             # <<<<<<<<<<<<<<
 *                     proposal.sells = [o for o in proposal.sells if o.size > 0]
 * 
 */
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_t_5, __pyx_v_i); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_size); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_t_11, __pyx_n_s_exp); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_eta); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_12 = PyNumber_Multiply(__pyx_t_11, __pyx_v_q); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          __pyx_t_11 = NULL;
          __pyx_t_8 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_10))) {
            __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
            if (likely(__pyx_t_11)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
              __Pyx_INCREF(__pyx_t_11);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_10, function);
              __pyx_t_8 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_11, __pyx_t_12};
            __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
            __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1111, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          }
          __pyx_t_10 = PyNumber_Multiply(__pyx_t_5, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __pyx_t_6 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_amount(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), __pyx_v_trading_pair, __pyx_t_10, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_5 = __Pyx_PyObject_GetItem(__pyx_t_10, __pyx_v_i); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (__Pyx_PyObject_SetAttrStr(__pyx_t_5, __pyx_n_s_size, __pyx_t_6) < 0) __PYX_ERR(0, 1111, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1110
 *             if len(proposal.sells) > 0:
 *                 if q < 0:
 *                     for i, proposed in enumerate(proposal.sells):             # <<<<<<<<<<<<<<
 *                         proposal.sells[i].size = market.c_quantize_order_amount(trading_pair, proposal.sells[i].size * Decimal.exp(self.eta * q))
 *                     proposal.sells = [o for o in proposal.sells if o.size > 0]
 */
        }
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1112
 *                     for i, proposed in enumerate(proposal.sells):
 *                         proposal.sells[i].size = market.c_quantize_order_amount(trading_pair, proposal.sells[i].size * Decimal.exp(self.eta * q))
 *                     proposal.sells = [o for o in proposal.sells if o.size > 0]             # <<<<<<<<<<<<<<
 * 
 *     def apply_order_amount_eta_transformation(self, proposal: Proposal):
 */
        { /* enter inner scope */
          __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1112, __pyx_L27_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1112, __pyx_L27_error)
          __Pyx_GOTREF(__pyx_t_7);
          if (likely(PyList_CheckExact(__pyx_t_7)) || PyTuple_CheckExact(__pyx_t_7)) {
            __pyx_t_5 = __pyx_t_7; __Pyx_INCREF(__pyx_t_5);
            __pyx_t_4 = 0;
            __pyx_t_9 = NULL;
          } else {
            __pyx_t_4 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_7); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1112, __pyx_L27_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_9 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1112, __pyx_L27_error)
          }
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          for (;;) {
            if (likely(!__pyx_t_9)) {
              if (likely(PyList_CheckExact(__pyx_t_5))) {
                {
                  Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
                  #if !CYTHON_ASSUME_SAFE_MACROS
                  if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1112, __pyx_L27_error)
                  #endif
                  if (__pyx_t_4 >= __pyx_temp) break;
                }
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_7 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_4); __Pyx_INCREF(__pyx_t_7); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 1112, __pyx_L27_error)
                #else
                __pyx_t_7 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1112, __pyx_L27_error)
                __Pyx_GOTREF(__pyx_t_7);
                #endif
              } else {
                {
                  Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
                  #if !CYTHON_ASSUME_SAFE_MACROS
                  if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1112, __pyx_L27_error)
                  #endif
                  if (__pyx_t_4 >= __pyx_temp) break;
                }
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_4); __Pyx_INCREF(__pyx_t_7); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 1112, __pyx_L27_error)
                #else
                __pyx_t_7 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1112, __pyx_L27_error)
                __Pyx_GOTREF(__pyx_t_7);
                #endif
              }
            } else {
              __pyx_t_7 = __pyx_t_9(__pyx_t_5);
              if (unlikely(!__pyx_t_7)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 1112, __pyx_L27_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_7);
            }
            __Pyx_XDECREF_SET(__pyx_9genexpr17__pyx_v_o, __pyx_t_7);
            __pyx_t_7 = 0;
            __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr17__pyx_v_o, __pyx_n_s_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1112, __pyx_L27_error)
            __Pyx_GOTREF(__pyx_t_7);
            __pyx_t_6 = PyObject_RichCompare(__pyx_t_7, __pyx_int_0, Py_GT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1112, __pyx_L27_error)
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1112, __pyx_L27_error)
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            if (__pyx_t_2) {
              if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_9genexpr17__pyx_v_o))) __PYX_ERR(0, 1112, __pyx_L27_error)
            }
          }
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_o); __pyx_9genexpr17__pyx_v_o = 0;
          goto __pyx_L32_exit_scope;
          __pyx_L27_error:;
          __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_o); __pyx_9genexpr17__pyx_v_o = 0;
          goto __pyx_L1_error;
          __pyx_L32_exit_scope:;
        } /* exit inner scope */
        if (__Pyx_PyObject_SetAttrStr(__pyx_v_proposal, __pyx_n_s_sells, __pyx_t_1) < 0) __PYX_ERR(0, 1112, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1109
 * 
 *             if len(proposal.sells) > 0:
 *                 if q < 0:             # <<<<<<<<<<<<<<
 *                     for i, proposed in enumerate(proposal.sells):
 *                         proposal.sells[i].size = market.c_quantize_order_amount(trading_pair, proposal.sells[i].size * Decimal.exp(self.eta * q))
 */
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1108
 *                     proposal.buys = [o for o in proposal.buys if o.size > 0]
 * 
 *             if len(proposal.sells) > 0:             # <<<<<<<<<<<<<<
 *                 if q < 0:
 *                     for i, proposed in enumerate(proposal.sells):
 */
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1087
 * 
 *         # Order amounts should be changed only if order_override is not active
 *         if (self.order_override is None) or (len(self.order_override) == 0):             # <<<<<<<<<<<<<<
 *             # eta parameter is described in the paper as the shape parameter for having exponentially decreasing order amount
 *             # for orders that go against inventory target (i.e. Want to buy when excess inventory or sell when deficit inventory)
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1081
 *         return self.c_apply_order_optimization(proposal)
 * 
 *     cdef c_apply_order_amount_eta_transformation(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_apply_order_amount_eta_transformation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_trading_pair);
  __Pyx_XDECREF(__pyx_v_inventory);
  __Pyx_XDECREF(__pyx_v_q_target);
  __Pyx_XDECREF(__pyx_v_q);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XDECREF(__pyx_v_proposed);
  __Pyx_XDECREF(__pyx_9genexpr16__pyx_v_o);
  __Pyx_XDECREF(__pyx_9genexpr17__pyx_v_o);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1114
 *                     proposal.sells = [o for o in proposal.sells if o.size > 0]
 * 
 *     def apply_order_amount_eta_transformation(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         self.c_apply_order_amount_eta_transformation(proposal)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_77apply_order_amount_eta_transformation(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_77apply_order_amount_eta_transformation = {"apply_order_amount_eta_transformation", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_77apply_order_amount_eta_transformation, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_77apply_order_amount_eta_transformation(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_proposal = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("apply_order_amount_eta_transformation (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_proposal,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_proposal)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1114, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "apply_order_amount_eta_transformation") < 0)) __PYX_ERR(0, 1114, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_proposal = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("apply_order_amount_eta_transformation", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1114, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.apply_order_amount_eta_transformation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_76apply_order_amount_eta_transformation(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_proposal);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_76apply_order_amount_eta_transformation(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("apply_order_amount_eta_transformation", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1115
 * 
 *     def apply_order_amount_eta_transformation(self, proposal: Proposal):
 *         self.c_apply_order_amount_eta_transformation(proposal)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_apply_add_transaction_costs(self, object proposal):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_order_amount_eta_transformation(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1114
 *                     proposal.sells = [o for o in proposal.sells if o.size > 0]
 * 
 *     def apply_order_amount_eta_transformation(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         self.c_apply_order_amount_eta_transformation(proposal)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.apply_order_amount_eta_transformation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1117
 *         self.c_apply_order_amount_eta_transformation(proposal)
 * 
 *     cdef c_apply_add_transaction_costs(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_apply_add_transaction_costs(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *__pyx_v_market = 0;
  PyObject *__pyx_v_buy = NULL;
  PyObject *__pyx_v_fee = NULL;
  PyObject *__pyx_v_price = NULL;
  PyObject *__pyx_v_sell = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  unsigned int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_apply_add_transaction_costs", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1119
 *     cdef c_apply_add_transaction_costs(self, object proposal):
 *         cdef:
 *             ExchangeBase market = self._market_info.market             # <<<<<<<<<<<<<<
 *         for buy in proposal.buys:
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase))))) __PYX_ERR(0, 1119, __pyx_L1_error)
  __pyx_v_market = ((struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1120
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 *         for buy in proposal.buys:             # <<<<<<<<<<<<<<
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
 *                                    self._limit_order_type, TradeType.BUY, buy.size, buy.price)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1120, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1120, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1120, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1120, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_2);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1120, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_1); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1120, __pyx_L1_error)
        #else
        __pyx_t_1 = __Pyx_PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1120, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1120, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_buy, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1121
 *             ExchangeBase market = self._market_info.market
 *         for buy in proposal.buys:
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,             # <<<<<<<<<<<<<<
 *                                    self._limit_order_type, TradeType.BUY, buy.size, buy.price)
 *             price = buy.price * (Decimal(1) - fee.percent)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 1121, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_quote_asset); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_5))) __PYX_ERR(0, 1121, __pyx_L1_error)

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1122
 *         for buy in proposal.buys:
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
 *                                    self._limit_order_type, TradeType.BUY, buy.size, buy.price)             # <<<<<<<<<<<<<<
 *             price = buy.price * (Decimal(1) - fee.percent)
 *             buy.price = market.c_quantize_order_price(self.trading_pair, price)
 */
    __pyx_t_6 = __pyx_v_self->_limit_order_type;
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_BUY); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_n_s_size); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_n_s_price); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1122, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1121
 *             ExchangeBase market = self._market_info.market
 *         for buy in proposal.buys:
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,             # <<<<<<<<<<<<<<
 *                                    self._limit_order_type, TradeType.BUY, buy.size, buy.price)
 *             price = buy.price * (Decimal(1) - fee.percent)
 */
    __pyx_t_10 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_get_fee(__pyx_v_market, ((PyObject*)__pyx_t_1), ((PyObject*)__pyx_t_5), __pyx_t_6, __pyx_t_8, __pyx_t_7, __pyx_t_9, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1121, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_XDECREF_SET(__pyx_v_fee, __pyx_t_10);
    __pyx_t_10 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1123
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
 *                                    self._limit_order_type, TradeType.BUY, buy.size, buy.price)
 *             price = buy.price * (Decimal(1) - fee.percent)             # <<<<<<<<<<<<<<
 *             buy.price = market.c_quantize_order_price(self.trading_pair, price)
 *         for sell in proposal.sells:
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_n_s_price); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_int_1};
      __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1123, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_fee, __pyx_n_s_percent); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = PyNumber_Subtract(__pyx_t_9, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyNumber_Multiply(__pyx_t_10, __pyx_t_8); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_XDECREF_SET(__pyx_v_price, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1124
 *                                    self._limit_order_type, TradeType.BUY, buy.size, buy.price)
 *             price = buy.price * (Decimal(1) - fee.percent)
 *             buy.price = market.c_quantize_order_price(self.trading_pair, price)             # <<<<<<<<<<<<<<
 *         for sell in proposal.sells:
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_7))) __PYX_ERR(0, 1124, __pyx_L1_error)
    __pyx_t_8 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_7), __pyx_v_price); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1124, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_buy, __pyx_n_s_price, __pyx_t_8) < 0) __PYX_ERR(0, 1124, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1120
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 *         for buy in proposal.buys:             # <<<<<<<<<<<<<<
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
 *                                    self._limit_order_type, TradeType.BUY, buy.size, buy.price)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1125
 *             price = buy.price * (Decimal(1) - fee.percent)
 *             buy.price = market.c_quantize_order_price(self.trading_pair, price)
 *         for sell in proposal.sells:             # <<<<<<<<<<<<<<
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
 *                                    self._limit_order_type, TradeType.SELL, sell.size, sell.price)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_8 = __pyx_t_2; __Pyx_INCREF(__pyx_t_8);
    __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1125, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_4 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1125, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_8))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_8);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1125, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1125, __pyx_L1_error)
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1125, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_8);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1125, __pyx_L1_error)
          #endif
          if (__pyx_t_3 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1125, __pyx_L1_error)
        #else
        __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1125, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_4(__pyx_t_8);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1125, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_sell, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1126
 *             buy.price = market.c_quantize_order_price(self.trading_pair, price)
 *         for sell in proposal.sells:
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,             # <<<<<<<<<<<<<<
 *                                    self._limit_order_type, TradeType.SELL, sell.size, sell.price)
 *             price = sell.price * (Decimal(1) + fee.percent)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base_asset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_2))) __PYX_ERR(0, 1126, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_quote_asset); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_7))||((__pyx_t_7) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_7))) __PYX_ERR(0, 1126, __pyx_L1_error)

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1127
 *         for sell in proposal.sells:
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
 *                                    self._limit_order_type, TradeType.SELL, sell.size, sell.price)             # <<<<<<<<<<<<<<
 *             price = sell.price * (Decimal(1) + fee.percent)
 *             sell.price = market.c_quantize_order_price(self.trading_pair, price)
 */
    __pyx_t_10 = __pyx_v_self->_limit_order_type;
    __Pyx_INCREF(__pyx_t_10);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_SELL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_sell, __pyx_n_s_size); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_sell, __pyx_n_s_price); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1126
 *             buy.price = market.c_quantize_order_price(self.trading_pair, price)
 *         for sell in proposal.sells:
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,             # <<<<<<<<<<<<<<
 *                                    self._limit_order_type, TradeType.SELL, sell.size, sell.price)
 *             price = sell.price * (Decimal(1) + fee.percent)
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_get_fee(__pyx_v_market, ((PyObject*)__pyx_t_2), ((PyObject*)__pyx_t_7), __pyx_t_10, __pyx_t_6, __pyx_t_9, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1126, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_fee, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1128
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
 *                                    self._limit_order_type, TradeType.SELL, sell.size, sell.price)
 *             price = sell.price * (Decimal(1) + fee.percent)             # <<<<<<<<<<<<<<
 *             sell.price = market.c_quantize_order_price(self.trading_pair, price)
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sell, __pyx_n_s_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_6 = NULL;
    __pyx_t_11 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
        __pyx_t_11 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_int_1};
      __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_9, __pyx_callargs+1-__pyx_t_11, 1+__pyx_t_11);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1128, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_fee, __pyx_n_s_percent); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_6 = PyNumber_Add(__pyx_t_5, __pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyNumber_Multiply(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_price, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1129
 *                                    self._limit_order_type, TradeType.SELL, sell.size, sell.price)
 *             price = sell.price * (Decimal(1) + fee.percent)
 *             sell.price = market.c_quantize_order_price(self.trading_pair, price)             # <<<<<<<<<<<<<<
 * 
 *     def apply_add_transaction_costs(self, proposal: Proposal):
 */
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (!(likely(PyUnicode_CheckExact(__pyx_t_9))||((__pyx_t_9) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_9))) __PYX_ERR(0, 1129, __pyx_L1_error)
    __pyx_t_6 = ((struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase *)__pyx_v_market->__pyx_base.__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_quantize_order_price(((struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase *)__pyx_v_market), ((PyObject*)__pyx_t_9), __pyx_v_price); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_sell, __pyx_n_s_price, __pyx_t_6) < 0) __PYX_ERR(0, 1129, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1125
 *             price = buy.price * (Decimal(1) - fee.percent)
 *             buy.price = market.c_quantize_order_price(self.trading_pair, price)
 *         for sell in proposal.sells:             # <<<<<<<<<<<<<<
 *             fee = market.c_get_fee(self.base_asset, self.quote_asset,
 *                                    self._limit_order_type, TradeType.SELL, sell.size, sell.price)
 */
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1117
 *         self.c_apply_order_amount_eta_transformation(proposal)
 * 
 *     cdef c_apply_add_transaction_costs(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             ExchangeBase market = self._market_info.market
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_apply_add_transaction_costs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_buy);
  __Pyx_XDECREF(__pyx_v_fee);
  __Pyx_XDECREF(__pyx_v_price);
  __Pyx_XDECREF(__pyx_v_sell);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1131
 *             sell.price = market.c_quantize_order_price(self.trading_pair, price)
 * 
 *     def apply_add_transaction_costs(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         self.c_apply_add_transaction_costs(proposal)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_79apply_add_transaction_costs(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_79apply_add_transaction_costs = {"apply_add_transaction_costs", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_79apply_add_transaction_costs, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_79apply_add_transaction_costs(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_proposal = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("apply_add_transaction_costs (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_proposal,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_proposal)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1131, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "apply_add_transaction_costs") < 0)) __PYX_ERR(0, 1131, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_proposal = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("apply_add_transaction_costs", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1131, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.apply_add_transaction_costs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_78apply_add_transaction_costs(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_proposal);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_78apply_add_transaction_costs(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("apply_add_transaction_costs", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1132
 * 
 *     def apply_add_transaction_costs(self, proposal: Proposal):
 *         self.c_apply_add_transaction_costs(proposal)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_did_fill_order(self, object order_filled_event):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_apply_add_transaction_costs(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1131
 *             sell.price = market.c_quantize_order_price(self.trading_pair, price)
 * 
 *     def apply_add_transaction_costs(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         self.c_apply_add_transaction_costs(proposal)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.apply_add_transaction_costs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1134
 *         self.c_apply_add_transaction_costs(proposal)
 * 
 *     cdef c_did_fill_order(self, object order_filled_event):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str order_id = order_filled_event.order_id
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_did_fill_order(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_order_filled_event) {
  PyObject *__pyx_v_order_id = 0;
  PyObject *__pyx_v_market_info = 0;
  CYTHON_UNUSED PyObject *__pyx_v_order_fill_record = 0;
  struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *__pyx_v_limit_order_record = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  Py_UCS4 __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_did_fill_order", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1136
 *     cdef c_did_fill_order(self, object order_filled_event):
 *         cdef:
 *             str order_id = order_filled_event.order_id             # <<<<<<<<<<<<<<
 *             object market_info = self._sb_order_tracker.c_get_shadow_market_pair_from_order_id(order_id)
 *             tuple order_fill_record
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_filled_event, __pyx_n_s_order_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 1136, __pyx_L1_error)
  __pyx_v_order_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1137
 *         cdef:
 *             str order_id = order_filled_event.order_id
 *             object market_info = self._sb_order_tracker.c_get_shadow_market_pair_from_order_id(order_id)             # <<<<<<<<<<<<<<
 *             tuple order_fill_record
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_13order_tracker_OrderTracker *)__pyx_v_self->__pyx_base._sb_order_tracker->__pyx_base.__pyx_base.__pyx_vtab)->c_get_shadow_market_pair_from_order_id(__pyx_v_self->__pyx_base._sb_order_tracker, __pyx_v_order_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_market_info = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1140
 *             tuple order_fill_record
 * 
 *         if market_info is not None:             # <<<<<<<<<<<<<<
 *             limit_order_record = self._sb_order_tracker.c_get_shadow_limit_order(order_id)
 *             order_fill_record = (limit_order_record, order_filled_event)
 */
  __pyx_t_2 = (__pyx_v_market_info != Py_None);
  if (__pyx_t_2) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1141
 * 
 *         if market_info is not None:
 *             limit_order_record = self._sb_order_tracker.c_get_shadow_limit_order(order_id)             # <<<<<<<<<<<<<<
 *             order_fill_record = (limit_order_record, order_filled_event)
 * 
 */
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_10hummingbot_8strategy_13order_tracker_OrderTracker *)__pyx_v_self->__pyx_base._sb_order_tracker->__pyx_base.__pyx_base.__pyx_vtab)->c_get_shadow_limit_order(__pyx_v_self->__pyx_base._sb_order_tracker, __pyx_v_order_id)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1141, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_limit_order_record = ((struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1142
 *         if market_info is not None:
 *             limit_order_record = self._sb_order_tracker.c_get_shadow_limit_order(order_id)
 *             order_fill_record = (limit_order_record, order_filled_event)             # <<<<<<<<<<<<<<
 * 
 *             if order_filled_event.trade_type is TradeType.BUY:
 */
    __pyx_t_1 = PyTuple_New(2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1142, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF((PyObject *)__pyx_v_limit_order_record);
    __Pyx_GIVEREF((PyObject *)__pyx_v_limit_order_record);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)__pyx_v_limit_order_record))) __PYX_ERR(0, 1142, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_order_filled_event);
    __Pyx_GIVEREF(__pyx_v_order_filled_event);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_v_order_filled_event)) __PYX_ERR(0, 1142, __pyx_L1_error);
    __pyx_v_order_fill_record = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1144
 *             order_fill_record = (limit_order_record, order_filled_event)
 * 
 *             if order_filled_event.trade_type is TradeType.BUY:             # <<<<<<<<<<<<<<
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:
 *                     self.log_with_clock(
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_filled_event, __pyx_n_s_trade_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_BUY); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_2 = (__pyx_t_1 == __pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_2) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1145
 * 
 *             if order_filled_event.trade_type is TradeType.BUY:
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:             # <<<<<<<<<<<<<<
 *                     self.log_with_clock(
 *                         logging.INFO,
 */
      __pyx_t_4 = __Pyx_PyInt_From_int64_t(__pyx_v_self->_logging_options); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1145, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_OPTION_LOG_MAKER_ORDER_FILLED); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1145, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = PyNumber_And(__pyx_t_4, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1145, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1145, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_2) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1146
 *             if order_filled_event.trade_type is TradeType.BUY:
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:
 *                     self.log_with_clock(             # <<<<<<<<<<<<<<
 *                         logging.INFO,
 *                         f"({market_info.trading_pair}) Maker buy order of "
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_log_with_clock); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1146, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1147
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:
 *                     self.log_with_clock(
 *                         logging.INFO,             # <<<<<<<<<<<<<<
 *                         f"({market_info.trading_pair}) Maker buy order of "
 *                         f"{order_filled_event.amount} {market_info.base_asset} filled."
 */
        __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_logging); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1147, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_INFO); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1147, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1148
 *                     self.log_with_clock(
 *                         logging.INFO,
 *                         f"({market_info.trading_pair}) Maker buy order of "             # <<<<<<<<<<<<<<
 *                         f"{order_filled_event.amount} {market_info.base_asset} filled."
 *                     )
 */
        __pyx_t_4 = PyTuple_New(7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1148, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = 0;
        __pyx_t_7 = 127;
        __Pyx_INCREF(__pyx_kp_u__11);
        __pyx_t_6 += 1;
        __Pyx_GIVEREF(__pyx_kp_u__11);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_kp_u__11);
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_market_info, __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1148, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1148, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) : __pyx_t_7;
        __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_9);
        __pyx_t_9 = 0;
        __Pyx_INCREF(__pyx_kp_u_Maker_buy_order_of);
        __pyx_t_6 += 21;
        __Pyx_GIVEREF(__pyx_kp_u_Maker_buy_order_of);
        PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_kp_u_Maker_buy_order_of);

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1149
 *                         logging.INFO,
 *                         f"({market_info.trading_pair}) Maker buy order of "
 *                         f"{order_filled_event.amount} {market_info.base_asset} filled."             # <<<<<<<<<<<<<<
 *                     )
 *             else:
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_filled_event, __pyx_n_s_amount); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1149, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_9, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1149, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_7;
        __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_8);
        __pyx_t_8 = 0;
        __Pyx_INCREF(__pyx_kp_u__12);
        __pyx_t_6 += 1;
        __Pyx_GIVEREF(__pyx_kp_u__12);
        PyTuple_SET_ITEM(__pyx_t_4, 4, __pyx_kp_u__12);
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_market_info, __pyx_n_s_base_asset); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1149, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_9 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_empty_unicode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1149, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_9) : __pyx_t_7;
        __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_9);
        __Pyx_GIVEREF(__pyx_t_9);
        PyTuple_SET_ITEM(__pyx_t_4, 5, __pyx_t_9);
        __pyx_t_9 = 0;
        __Pyx_INCREF(__pyx_kp_u_filled);
        __pyx_t_6 += 8;
        __Pyx_GIVEREF(__pyx_kp_u_filled);
        PyTuple_SET_ITEM(__pyx_t_4, 6, __pyx_kp_u_filled);

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1148
 *                     self.log_with_clock(
 *                         logging.INFO,
 *                         f"({market_info.trading_pair}) Maker buy order of "             # <<<<<<<<<<<<<<
 *                         f"{order_filled_event.amount} {market_info.base_asset} filled."
 *                     )
 */
        __pyx_t_9 = __Pyx_PyUnicode_Join(__pyx_t_4, 7, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1148, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_4 = NULL;
        __pyx_t_10 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
            __pyx_t_10 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_5, __pyx_t_9};
          __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_10, 2+__pyx_t_10);
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1146, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1145
 * 
 *             if order_filled_event.trade_type is TradeType.BUY:
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:             # <<<<<<<<<<<<<<
 *                     self.log_with_clock(
 *                         logging.INFO,
 */
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1144
 *             order_fill_record = (limit_order_record, order_filled_event)
 * 
 *             if order_filled_event.trade_type is TradeType.BUY:             # <<<<<<<<<<<<<<
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:
 *                     self.log_with_clock(
 */
      goto __pyx_L4;
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1152
 *                     )
 *             else:
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:             # <<<<<<<<<<<<<<
 *                     self.log_with_clock(
 *                         logging.INFO,
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyInt_From_int64_t(__pyx_v_self->_logging_options); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_OPTION_LOG_MAKER_ORDER_FILLED); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_9 = PyNumber_And(__pyx_t_3, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(0, 1152, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (__pyx_t_2) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1153
 *             else:
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:
 *                     self.log_with_clock(             # <<<<<<<<<<<<<<
 *                         logging.INFO,
 *                         f"({market_info.trading_pair}) Maker sell order of "
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_log_with_clock); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1153, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1154
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:
 *                     self.log_with_clock(
 *                         logging.INFO,             # <<<<<<<<<<<<<<
 *                         f"({market_info.trading_pair}) Maker sell order of "
 *                         f"{order_filled_event.amount} {market_info.base_asset} filled."
 */
        __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_logging); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1154, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_INFO); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1154, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1155
 *                     self.log_with_clock(
 *                         logging.INFO,
 *                         f"({market_info.trading_pair}) Maker sell order of "             # <<<<<<<<<<<<<<
 *                         f"{order_filled_event.amount} {market_info.base_asset} filled."
 *                     )
 */
        __pyx_t_3 = PyTuple_New(7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1155, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = 0;
        __pyx_t_7 = 127;
        __Pyx_INCREF(__pyx_kp_u__11);
        __pyx_t_6 += 1;
        __Pyx_GIVEREF(__pyx_kp_u__11);
        PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_kp_u__11);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_market_info, __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1155, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1155, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_7;
        __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_8);
        __pyx_t_8 = 0;
        __Pyx_INCREF(__pyx_kp_u_Maker_sell_order_of);
        __pyx_t_6 += 22;
        __Pyx_GIVEREF(__pyx_kp_u_Maker_sell_order_of);
        PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_kp_u_Maker_sell_order_of);

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1156
 *                         logging.INFO,
 *                         f"({market_info.trading_pair}) Maker sell order of "
 *                         f"{order_filled_event.amount} {market_info.base_asset} filled."             # <<<<<<<<<<<<<<
 *                     )
 * 
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_filled_event, __pyx_n_s_amount); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1156, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_4 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_empty_unicode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1156, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_4) : __pyx_t_7;
        __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_4);
        __pyx_t_4 = 0;
        __Pyx_INCREF(__pyx_kp_u__12);
        __pyx_t_6 += 1;
        __Pyx_GIVEREF(__pyx_kp_u__12);
        PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_kp_u__12);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_market_info, __pyx_n_s_base_asset); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1156, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_4, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1156, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __pyx_t_7 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_7) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_7;
        __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_t_8);
        __pyx_t_8 = 0;
        __Pyx_INCREF(__pyx_kp_u_filled);
        __pyx_t_6 += 8;
        __Pyx_GIVEREF(__pyx_kp_u_filled);
        PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_kp_u_filled);

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1155
 *                     self.log_with_clock(
 *                         logging.INFO,
 *                         f"({market_info.trading_pair}) Maker sell order of "             # <<<<<<<<<<<<<<
 *                         f"{order_filled_event.amount} {market_info.base_asset} filled."
 *                     )
 */
        __pyx_t_8 = __Pyx_PyUnicode_Join(__pyx_t_3, 7, __pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1155, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_3 = NULL;
        __pyx_t_10 = 0;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_1))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
          if (likely(__pyx_t_3)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
            __Pyx_INCREF(__pyx_t_3);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_1, function);
            __pyx_t_10 = 1;
          }
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_t_5, __pyx_t_8};
          __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_10, 2+__pyx_t_10);
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1153, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        }
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1152
 *                     )
 *             else:
 *                 if self._logging_options & self.OPTION_LOG_MAKER_ORDER_FILLED:             # <<<<<<<<<<<<<<
 *                     self.log_with_clock(
 *                         logging.INFO,
 */
      }
    }
    __pyx_L4:;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1140
 *             tuple order_fill_record
 * 
 *         if market_info is not None:             # <<<<<<<<<<<<<<
 *             limit_order_record = self._sb_order_tracker.c_get_shadow_limit_order(order_id)
 *             order_fill_record = (limit_order_record, order_filled_event)
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1134
 *         self.c_apply_add_transaction_costs(proposal)
 * 
 *     cdef c_did_fill_order(self, object order_filled_event):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str order_id = order_filled_event.order_id
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_did_fill_order", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_order_id);
  __Pyx_XDECREF(__pyx_v_market_info);
  __Pyx_XDECREF(__pyx_v_order_fill_record);
  __Pyx_XDECREF((PyObject *)__pyx_v_limit_order_record);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1159
 *                     )
 * 
 *     cdef c_did_complete_buy_order(self, object order_completed_event):             # <<<<<<<<<<<<<<
 *         self.c_did_complete_order(order_completed_event)
 * 
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_did_complete_buy_order(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_order_completed_event) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_did_complete_buy_order", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1160
 * 
 *     cdef c_did_complete_buy_order(self, object order_completed_event):
 *         self.c_did_complete_order(order_completed_event)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_did_complete_sell_order(self, object order_completed_event):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_did_complete_order(__pyx_v_self, __pyx_v_order_completed_event); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1159
 *                     )
 * 
 *     cdef c_did_complete_buy_order(self, object order_completed_event):             # <<<<<<<<<<<<<<
 *         self.c_did_complete_order(order_completed_event)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_did_complete_buy_order", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1162
 *         self.c_did_complete_order(order_completed_event)
 * 
 *     cdef c_did_complete_sell_order(self, object order_completed_event):             # <<<<<<<<<<<<<<
 *         self.c_did_complete_order(order_completed_event)
 * 
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_did_complete_sell_order(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_order_completed_event) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_did_complete_sell_order", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1163
 * 
 *     cdef c_did_complete_sell_order(self, object order_completed_event):
 *         self.c_did_complete_order(order_completed_event)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_did_complete_order(self, object order_completed_event):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_did_complete_order(__pyx_v_self, __pyx_v_order_completed_event); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1162
 *         self.c_did_complete_order(order_completed_event)
 * 
 *     cdef c_did_complete_sell_order(self, object order_completed_event):             # <<<<<<<<<<<<<<
 *         self.c_did_complete_order(order_completed_event)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_did_complete_sell_order", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1165
 *         self.c_did_complete_order(order_completed_event)
 * 
 *     cdef c_did_complete_order(self, object order_completed_event):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str order_id = order_completed_event.order_id
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_did_complete_order(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_order_completed_event) {
  PyObject *__pyx_v_order_id = 0;
  struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *__pyx_v_limit_order_record = 0;
  PyObject *__pyx_v_order_action_string = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  unsigned int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  double __pyx_t_8;
  double __pyx_t_9;
  double __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  Py_ssize_t __pyx_t_12;
  Py_UCS4 __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_did_complete_order", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1167
 *     cdef c_did_complete_order(self, object order_completed_event):
 *         cdef:
 *             str order_id = order_completed_event.order_id             # <<<<<<<<<<<<<<
 *             LimitOrder limit_order_record = self._sb_order_tracker.c_get_limit_order(self._market_info, order_id)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_order_completed_event, __pyx_n_s_order_id); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_1))) __PYX_ERR(0, 1167, __pyx_L1_error)
  __pyx_v_order_id = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1168
 *         cdef:
 *             str order_id = order_completed_event.order_id
 *             LimitOrder limit_order_record = self._sb_order_tracker.c_get_limit_order(self._market_info, order_id)             # <<<<<<<<<<<<<<
 * 
 *         if limit_order_record is None:
 */
  __pyx_t_1 = __pyx_v_self->_market_info;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = ((PyObject *)((struct __pyx_vtabstruct_10hummingbot_8strategy_13order_tracker_OrderTracker *)__pyx_v_self->__pyx_base._sb_order_tracker->__pyx_base.__pyx_base.__pyx_vtab)->c_get_limit_order(__pyx_v_self->__pyx_base._sb_order_tracker, __pyx_t_1, __pyx_v_order_id)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_limit_order_record = ((struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1170
 *             LimitOrder limit_order_record = self._sb_order_tracker.c_get_limit_order(self._market_info, order_id)
 * 
 *         if limit_order_record is None:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_3 = (((PyObject *)__pyx_v_limit_order_record) == Py_None);
  if (__pyx_t_3) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1171
 * 
 *         if limit_order_record is None:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         # Continue only if the order is not a hanging order
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1170
 *             LimitOrder limit_order_record = self._sb_order_tracker.c_get_limit_order(self._market_info, order_id)
 * 
 *         if limit_order_record is None:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1174
 * 
 *         # Continue only if the order is not a hanging order
 *         if (not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order_id)             # <<<<<<<<<<<<<<
 *                 and not self.hanging_orders_tracker.is_order_id_in_completed_hanging_orders(order_id)):
 *             # delay order creation by filled_order_delay (in seconds)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_n_s_is_order_id_in_hanging_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_order_id};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1174, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 1174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_7 = (!__pyx_t_6);
  if (__pyx_t_7) {
  } else {
    __pyx_t_3 = __pyx_t_7;
    goto __pyx_L5_bool_binop_done;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1175
 *         # Continue only if the order is not a hanging order
 *         if (not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order_id)
 *                 and not self.hanging_orders_tracker.is_order_id_in_completed_hanging_orders(order_id)):             # <<<<<<<<<<<<<<
 *             # delay order creation by filled_order_delay (in seconds)
 *             self._create_timestamp = self._current_timestamp + self.filled_order_delay
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_hanging_orders_tracker); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_is_order_id_in_completed_hanging); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  __pyx_t_5 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_4, function);
      __pyx_t_5 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_order_id};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_6 = (!__pyx_t_7);
  __pyx_t_3 = __pyx_t_6;
  __pyx_L5_bool_binop_done:;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1174
 * 
 *         # Continue only if the order is not a hanging order
 *         if (not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order_id)             # <<<<<<<<<<<<<<
 *                 and not self.hanging_orders_tracker.is_order_id_in_completed_hanging_orders(order_id)):
 *             # delay order creation by filled_order_delay (in seconds)
 */
  if (__pyx_t_3) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1177
 *                 and not self.hanging_orders_tracker.is_order_id_in_completed_hanging_orders(order_id)):
 *             # delay order creation by filled_order_delay (in seconds)
 *             self._create_timestamp = self._current_timestamp + self.filled_order_delay             # <<<<<<<<<<<<<<
 *             self._cancel_timestamp = min(self._cancel_timestamp, self._create_timestamp)
 * 
 */
    __pyx_t_2 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_filled_order_delay); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyNumber_Add(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1177, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_self->_create_timestamp = __pyx_t_8;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1178
 *             # delay order creation by filled_order_delay (in seconds)
 *             self._create_timestamp = self._current_timestamp + self.filled_order_delay
 *             self._cancel_timestamp = min(self._cancel_timestamp, self._create_timestamp)             # <<<<<<<<<<<<<<
 * 
 *             if limit_order_record.is_buy:
 */
    __pyx_t_8 = __pyx_v_self->_create_timestamp;
    __pyx_t_9 = __pyx_v_self->_cancel_timestamp;
    __pyx_t_3 = (__pyx_t_8 < __pyx_t_9);
    if (__pyx_t_3) {
      __pyx_t_10 = __pyx_t_8;
    } else {
      __pyx_t_10 = __pyx_t_9;
    }
    __pyx_v_self->_cancel_timestamp = __pyx_t_10;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1180
 *             self._cancel_timestamp = min(self._cancel_timestamp, self._create_timestamp)
 * 
 *             if limit_order_record.is_buy:             # <<<<<<<<<<<<<<
 *                 self._filled_buys_balance += 1
 *                 order_action_string = "buy"
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_n_s_is_buy); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1180, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 1180, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_3) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1181
 * 
 *             if limit_order_record.is_buy:
 *                 self._filled_buys_balance += 1             # <<<<<<<<<<<<<<
 *                 order_action_string = "buy"
 *             else:
 */
      __pyx_v_self->_filled_buys_balance = (__pyx_v_self->_filled_buys_balance + 1);

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1182
 *             if limit_order_record.is_buy:
 *                 self._filled_buys_balance += 1
 *                 order_action_string = "buy"             # <<<<<<<<<<<<<<
 *             else:
 *                 self._filled_sells_balance += 1
 */
      __Pyx_INCREF(__pyx_n_u_buy);
      __pyx_v_order_action_string = __pyx_n_u_buy;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1180
 *             self._cancel_timestamp = min(self._cancel_timestamp, self._create_timestamp)
 * 
 *             if limit_order_record.is_buy:             # <<<<<<<<<<<<<<
 *                 self._filled_buys_balance += 1
 *                 order_action_string = "buy"
 */
      goto __pyx_L7;
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1184
 *                 order_action_string = "buy"
 *             else:
 *                 self._filled_sells_balance += 1             # <<<<<<<<<<<<<<
 *                 order_action_string = "sell"
 * 
 */
    /*else*/ {
      __pyx_v_self->_filled_sells_balance = (__pyx_v_self->_filled_sells_balance + 1);

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1185
 *             else:
 *                 self._filled_sells_balance += 1
 *                 order_action_string = "sell"             # <<<<<<<<<<<<<<
 * 
 *             self._last_own_trade_price = limit_order_record.price
 */
      __Pyx_INCREF(__pyx_n_u_sell);
      __pyx_v_order_action_string = __pyx_n_u_sell;
    }
    __pyx_L7:;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1187
 *                 order_action_string = "sell"
 * 
 *             self._last_own_trade_price = limit_order_record.price             # <<<<<<<<<<<<<<
 * 
 *             self.log_with_clock(
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_n_s_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->_last_own_trade_price);
    __Pyx_DECREF(__pyx_v_self->_last_own_trade_price);
    __pyx_v_self->_last_own_trade_price = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1189
 *             self._last_own_trade_price = limit_order_record.price
 * 
 *             self.log_with_clock(             # <<<<<<<<<<<<<<
 *                 logging.INFO,
 *                 f"({self.trading_pair}) Maker {order_action_string} order {order_id} "
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_log_with_clock); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1190
 * 
 *             self.log_with_clock(
 *                 logging.INFO,             # <<<<<<<<<<<<<<
 *                 f"({self.trading_pair}) Maker {order_action_string} order {order_id} "
 *                 f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_logging); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_INFO); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1191
 *             self.log_with_clock(
 *                 logging.INFO,
 *                 f"({self.trading_pair}) Maker {order_action_string} order {order_id} "             # <<<<<<<<<<<<<<
 *                 f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *                 f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
 */
    __pyx_t_2 = PyTuple_New(15); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_12 = 0;
    __pyx_t_13 = 127;
    __Pyx_INCREF(__pyx_kp_u__11);
    __pyx_t_12 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__11);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_kp_u__11);
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = __Pyx_PyObject_FormatSimple(__pyx_t_14, __pyx_empty_unicode); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_13 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) > __pyx_t_13) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) : __pyx_t_13;
    __pyx_t_12 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_15);
    __pyx_t_15 = 0;
    __Pyx_INCREF(__pyx_kp_u_Maker);
    __pyx_t_12 += 8;
    __Pyx_GIVEREF(__pyx_kp_u_Maker);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_kp_u_Maker);
    __Pyx_INCREF(__pyx_v_order_action_string);
    __pyx_t_13 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_v_order_action_string) > __pyx_t_13) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_v_order_action_string) : __pyx_t_13;
    __pyx_t_12 += __Pyx_PyUnicode_GET_LENGTH(__pyx_v_order_action_string);
    __Pyx_GIVEREF(__pyx_v_order_action_string);
    PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_v_order_action_string);
    __Pyx_INCREF(__pyx_kp_u_order);
    __pyx_t_12 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_order);
    PyTuple_SET_ITEM(__pyx_t_2, 4, __pyx_kp_u_order);
    __pyx_t_15 = __Pyx_PyUnicode_Unicode(__pyx_v_order_id); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_13 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) > __pyx_t_13) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) : __pyx_t_13;
    __pyx_t_12 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_2, 5, __pyx_t_15);
    __pyx_t_15 = 0;
    __Pyx_INCREF(__pyx_kp_u__13);
    __pyx_t_12 += 2;
    __Pyx_GIVEREF(__pyx_kp_u__13);
    PyTuple_SET_ITEM(__pyx_t_2, 6, __pyx_kp_u__13);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1192
 *                 logging.INFO,
 *                 f"({self.trading_pair}) Maker {order_action_string} order {order_id} "
 *                 f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "             # <<<<<<<<<<<<<<
 *                 f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
 *             )
 */
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_n_s_quantity); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_14 = __Pyx_PyObject_FormatSimple(__pyx_t_15, __pyx_empty_unicode); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_13 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_14) > __pyx_t_13) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_14) : __pyx_t_13;
    __pyx_t_12 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_2, 7, __pyx_t_14);
    __pyx_t_14 = 0;
    __Pyx_INCREF(__pyx_kp_u__12);
    __pyx_t_12 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__12);
    PyTuple_SET_ITEM(__pyx_t_2, 8, __pyx_kp_u__12);
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_n_s_base_currency); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = __Pyx_PyObject_FormatSimple(__pyx_t_14, __pyx_empty_unicode); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_13 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) > __pyx_t_13) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) : __pyx_t_13;
    __pyx_t_12 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_2, 9, __pyx_t_15);
    __pyx_t_15 = 0;
    __Pyx_INCREF(__pyx_kp_u__14);
    __pyx_t_12 += 3;
    __Pyx_GIVEREF(__pyx_kp_u__14);
    PyTuple_SET_ITEM(__pyx_t_2, 10, __pyx_kp_u__14);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1193
 *                 f"({self.trading_pair}) Maker {order_action_string} order {order_id} "
 *                 f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *                 f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."             # <<<<<<<<<<<<<<
 *             )
 *             self.notify_hb_app_with_timestamp(
 */
    __pyx_t_15 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_n_s_price); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_14 = __Pyx_PyObject_FormatSimple(__pyx_t_15, __pyx_empty_unicode); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_13 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_14) > __pyx_t_13) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_14) : __pyx_t_13;
    __pyx_t_12 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_14);
    __Pyx_GIVEREF(__pyx_t_14);
    PyTuple_SET_ITEM(__pyx_t_2, 11, __pyx_t_14);
    __pyx_t_14 = 0;
    __Pyx_INCREF(__pyx_kp_u__12);
    __pyx_t_12 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__12);
    PyTuple_SET_ITEM(__pyx_t_2, 12, __pyx_kp_u__12);
    __pyx_t_14 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_n_s_quote_currency); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __pyx_t_15 = __Pyx_PyObject_FormatSimple(__pyx_t_14, __pyx_empty_unicode); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __pyx_t_13 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) > __pyx_t_13) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_15) : __pyx_t_13;
    __pyx_t_12 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_2, 13, __pyx_t_15);
    __pyx_t_15 = 0;
    __Pyx_INCREF(__pyx_kp_u_has_been_completely_filled);
    __pyx_t_12 += 29;
    __Pyx_GIVEREF(__pyx_kp_u_has_been_completely_filled);
    PyTuple_SET_ITEM(__pyx_t_2, 14, __pyx_kp_u_has_been_completely_filled);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1191
 *             self.log_with_clock(
 *                 logging.INFO,
 *                 f"({self.trading_pair}) Maker {order_action_string} order {order_id} "             # <<<<<<<<<<<<<<
 *                 f"({limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *                 f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
 */
    __pyx_t_15 = __Pyx_PyUnicode_Join(__pyx_t_2, 15, __pyx_t_12, __pyx_t_13); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_2)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_2);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_t_11, __pyx_t_15};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 2+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1195
 *                 f"{limit_order_record.price} {limit_order_record.quote_currency}) has been completely filled."
 *             )
 *             self.notify_hb_app_with_timestamp(             # <<<<<<<<<<<<<<
 *                 f"Maker {order_action_string.upper()} order "
 *                 f"{limit_order_record.quantity} {limit_order_record.base_currency} @ "
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_notify_hb_app_with_timestamp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1196
 *             )
 *             self.notify_hb_app_with_timestamp(
 *                 f"Maker {order_action_string.upper()} order "             # <<<<<<<<<<<<<<
 *                 f"{limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *                 f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."
 */
    __pyx_t_15 = PyTuple_New(11); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_12 = 0;
    __pyx_t_13 = 127;
    __Pyx_INCREF(__pyx_kp_u_Maker_2);
    __pyx_t_12 += 6;
    __Pyx_GIVEREF(__pyx_kp_u_Maker_2);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_kp_u_Maker_2);
    __pyx_t_11 = __Pyx_CallUnboundCMethod0(&__pyx_umethod_PyUnicode_Type_upper, __pyx_v_order_action_string); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_11, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_13 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_13) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_13;
    __pyx_t_12 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_order);
    __pyx_t_12 += 7;
    __Pyx_GIVEREF(__pyx_kp_u_order);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_kp_u_order);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1197
 *             self.notify_hb_app_with_timestamp(
 *                 f"Maker {order_action_string.upper()} order "
 *                 f"{limit_order_record.quantity} {limit_order_record.base_currency} @ "             # <<<<<<<<<<<<<<
 *                 f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."
 *             )
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_n_s_quantity); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = __Pyx_PyObject_FormatSimple(__pyx_t_2, __pyx_empty_unicode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_13 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) > __pyx_t_13) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) : __pyx_t_13;
    __pyx_t_12 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_11);
    PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_t_11);
    __pyx_t_11 = 0;
    __Pyx_INCREF(__pyx_kp_u__12);
    __pyx_t_12 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__12);
    PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_kp_u__12);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_n_s_base_currency); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_11, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_13 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_13) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_13;
    __pyx_t_12 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u__14);
    __pyx_t_12 += 3;
    __Pyx_GIVEREF(__pyx_kp_u__14);
    PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_kp_u__14);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1198
 *                 f"Maker {order_action_string.upper()} order "
 *                 f"{limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *                 f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."             # <<<<<<<<<<<<<<
 *             )
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_n_s_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_11 = __Pyx_PyObject_FormatSimple(__pyx_t_2, __pyx_empty_unicode); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_13 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) > __pyx_t_13) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_11) : __pyx_t_13;
    __pyx_t_12 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_11);
    __Pyx_GIVEREF(__pyx_t_11);
    PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_11);
    __pyx_t_11 = 0;
    __Pyx_INCREF(__pyx_kp_u__12);
    __pyx_t_12 += 1;
    __Pyx_GIVEREF(__pyx_kp_u__12);
    PyTuple_SET_ITEM(__pyx_t_15, 8, __pyx_kp_u__12);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_limit_order_record), __pyx_n_s_quote_currency); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_2 = __Pyx_PyObject_FormatSimple(__pyx_t_11, __pyx_empty_unicode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_13 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) > __pyx_t_13) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_2) : __pyx_t_13;
    __pyx_t_12 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_2);
    __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_kp_u_is_filled);
    __pyx_t_12 += 11;
    __Pyx_GIVEREF(__pyx_kp_u_is_filled);
    PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_kp_u_is_filled);

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1196
 *             )
 *             self.notify_hb_app_with_timestamp(
 *                 f"Maker {order_action_string.upper()} order "             # <<<<<<<<<<<<<<
 *                 f"{limit_order_record.quantity} {limit_order_record.base_currency} @ "
 *                 f"{limit_order_record.price} {limit_order_record.quote_currency} is filled."
 */
    __pyx_t_2 = __Pyx_PyUnicode_Join(__pyx_t_15, 11, __pyx_t_12, __pyx_t_13); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1196, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __pyx_t_15 = NULL;
    __pyx_t_5 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_15)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_15);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_5 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_15, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_5, 1+__pyx_t_5);
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1195, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1174
 * 
 *         # Continue only if the order is not a hanging order
 *         if (not self._hanging_orders_tracker.is_order_id_in_hanging_orders(order_id)             # <<<<<<<<<<<<<<
 *                 and not self.hanging_orders_tracker.is_order_id_in_completed_hanging_orders(order_id)):
 *             # delay order creation by filled_order_delay (in seconds)
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1165
 *         self.c_did_complete_order(order_completed_event)
 * 
 *     cdef c_did_complete_order(self, object order_completed_event):             # <<<<<<<<<<<<<<
 *         cdef:
 *             str order_id = order_completed_event.order_id
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_did_complete_order", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_order_id);
  __Pyx_XDECREF((PyObject *)__pyx_v_limit_order_record);
  __Pyx_XDECREF(__pyx_v_order_action_string);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1201
 *             )
 * 
 *     cdef bint c_is_within_tolerance(self, list current_prices, list proposal_prices):             # <<<<<<<<<<<<<<
 *         if len(current_prices) != len(proposal_prices):
 *             return False
 */

static int __pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_is_within_tolerance(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_current_prices, PyObject *__pyx_v_proposal_prices) {
  PyObject *__pyx_v_current = NULL;
  PyObject *__pyx_v_proposal = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *(*__pyx_t_7)(PyObject *);
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *(*__pyx_t_11)(PyObject *);
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_is_within_tolerance", 0);
  __Pyx_INCREF(__pyx_v_current_prices);
  __Pyx_INCREF(__pyx_v_proposal_prices);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1202
 * 
 *     cdef bint c_is_within_tolerance(self, list current_prices, list proposal_prices):
 *         if len(current_prices) != len(proposal_prices):             # <<<<<<<<<<<<<<
 *             return False
 *         current_prices = sorted(current_prices)
 */
  if (unlikely(__pyx_v_current_prices == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1202, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyList_GET_SIZE(__pyx_v_current_prices); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1202, __pyx_L1_error)
  if (unlikely(__pyx_v_proposal_prices == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 1202, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyList_GET_SIZE(__pyx_v_proposal_prices); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1202, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_1 != __pyx_t_2);
  if (__pyx_t_3) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1203
 *     cdef bint c_is_within_tolerance(self, list current_prices, list proposal_prices):
 *         if len(current_prices) != len(proposal_prices):
 *             return False             # <<<<<<<<<<<<<<
 *         current_prices = sorted(current_prices)
 *         proposal_prices = sorted(proposal_prices)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1202
 * 
 *     cdef bint c_is_within_tolerance(self, list current_prices, list proposal_prices):
 *         if len(current_prices) != len(proposal_prices):             # <<<<<<<<<<<<<<
 *             return False
 *         current_prices = sorted(current_prices)
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1204
 *         if len(current_prices) != len(proposal_prices):
 *             return False
 *         current_prices = sorted(current_prices)             # <<<<<<<<<<<<<<
 *         proposal_prices = sorted(proposal_prices)
 *         for current, proposal in zip(current_prices, proposal_prices):
 */
  __pyx_t_5 = PySequence_List(__pyx_v_current_prices); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1204, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_6 = PyList_Sort(__pyx_t_4); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1204, __pyx_L1_error)
  __Pyx_DECREF_SET(__pyx_v_current_prices, ((PyObject*)__pyx_t_4));
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1205
 *             return False
 *         current_prices = sorted(current_prices)
 *         proposal_prices = sorted(proposal_prices)             # <<<<<<<<<<<<<<
 *         for current, proposal in zip(current_prices, proposal_prices):
 *             # if spread diff is more than the tolerance or order quantities are different, return false.
 */
  __pyx_t_5 = PySequence_List(__pyx_v_proposal_prices); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_4 = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;
  __pyx_t_6 = PyList_Sort(__pyx_t_4); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 1205, __pyx_L1_error)
  __Pyx_DECREF_SET(__pyx_v_proposal_prices, ((PyObject*)__pyx_t_4));
  __pyx_t_4 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1206
 *         current_prices = sorted(current_prices)
 *         proposal_prices = sorted(proposal_prices)
 *         for current, proposal in zip(current_prices, proposal_prices):             # <<<<<<<<<<<<<<
 *             # if spread diff is more than the tolerance or order quantities are different, return false.
 *             if abs(proposal - current) / current > self.order_refresh_tolerance:
 */
  __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_current_prices);
  __Pyx_GIVEREF(__pyx_v_current_prices);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_current_prices)) __PYX_ERR(0, 1206, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_proposal_prices);
  __Pyx_GIVEREF(__pyx_v_proposal_prices);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_proposal_prices)) __PYX_ERR(0, 1206, __pyx_L1_error);
  __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_4, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
    __pyx_t_4 = __pyx_t_5; __Pyx_INCREF(__pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_7 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1206, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  for (;;) {
    if (likely(!__pyx_t_7)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1206, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_2); __Pyx_INCREF(__pyx_t_5); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 1206, __pyx_L1_error)
        #else
        __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1206, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1206, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_2); __Pyx_INCREF(__pyx_t_5); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 1206, __pyx_L1_error)
        #else
        __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1206, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_7(__pyx_t_4);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1206, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
      PyObject* sequence = __pyx_t_5;
      Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 1206, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1206, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1206, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_10 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1206, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_11 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_10);
      index = 0; __pyx_t_8 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_8)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      index = 1; __pyx_t_9 = __pyx_t_11(__pyx_t_10); if (unlikely(!__pyx_t_9)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_11(__pyx_t_10), 2) < 0) __PYX_ERR(0, 1206, __pyx_L1_error)
      __pyx_t_11 = NULL;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_11 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 1206, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XDECREF_SET(__pyx_v_current, __pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_XDECREF_SET(__pyx_v_proposal, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1208
 *         for current, proposal in zip(current_prices, proposal_prices):
 *             # if spread diff is more than the tolerance or order quantities are different, return false.
 *             if abs(proposal - current) / current > self.order_refresh_tolerance:             # <<<<<<<<<<<<<<
 *                 return False
 *         return True
 */
    __pyx_t_5 = PyNumber_Subtract(__pyx_v_proposal, __pyx_v_current); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_9 = __Pyx_PyNumber_Absolute(__pyx_t_5); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyNumber_Divide(__pyx_t_9, __pyx_v_current); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_refresh_tolerance); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_8 = PyObject_RichCompare(__pyx_t_5, __pyx_t_9, Py_GT); __Pyx_XGOTREF(__pyx_t_8); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_8); if (unlikely((__pyx_t_3 < 0))) __PYX_ERR(0, 1208, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    if (__pyx_t_3) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1209
 *             # if spread diff is more than the tolerance or order quantities are different, return false.
 *             if abs(proposal - current) / current > self.order_refresh_tolerance:
 *                 return False             # <<<<<<<<<<<<<<
 *         return True
 * 
 */
      __pyx_r = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      goto __pyx_L0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1208
 *         for current, proposal in zip(current_prices, proposal_prices):
 *             # if spread diff is more than the tolerance or order quantities are different, return false.
 *             if abs(proposal - current) / current > self.order_refresh_tolerance:             # <<<<<<<<<<<<<<
 *                 return False
 *         return True
 */
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1206
 *         current_prices = sorted(current_prices)
 *         proposal_prices = sorted(proposal_prices)
 *         for current, proposal in zip(current_prices, proposal_prices):             # <<<<<<<<<<<<<<
 *             # if spread diff is more than the tolerance or order quantities are different, return false.
 *             if abs(proposal - current) / current > self.order_refresh_tolerance:
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1210
 *             if abs(proposal - current) / current > self.order_refresh_tolerance:
 *                 return False
 *         return True             # <<<<<<<<<<<<<<
 * 
 *     def is_within_tolerance(self, current_prices: List[Decimal], proposal_prices: List[Decimal]) -> bool:
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1201
 *             )
 * 
 *     cdef bint c_is_within_tolerance(self, list current_prices, list proposal_prices):             # <<<<<<<<<<<<<<
 *         if len(current_prices) != len(proposal_prices):
 *             return False
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_is_within_tolerance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_current);
  __Pyx_XDECREF(__pyx_v_proposal);
  __Pyx_XDECREF(__pyx_v_current_prices);
  __Pyx_XDECREF(__pyx_v_proposal_prices);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1212
 *         return True
 * 
 *     def is_within_tolerance(self, current_prices: List[Decimal], proposal_prices: List[Decimal]) -> bool:             # <<<<<<<<<<<<<<
 *         return self.c_is_within_tolerance(current_prices, proposal_prices)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_81is_within_tolerance(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_81is_within_tolerance = {"is_within_tolerance", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_81is_within_tolerance, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_81is_within_tolerance(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_current_prices = 0;
  PyObject *__pyx_v_proposal_prices = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_within_tolerance (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_current_prices,&__pyx_n_s_proposal_prices,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  2: values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_current_prices)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1212, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_proposal_prices)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[1]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1212, __pyx_L3_error)
        else {
          __Pyx_RaiseArgtupleInvalid("is_within_tolerance", 1, 2, 2, 1); __PYX_ERR(0, 1212, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "is_within_tolerance") < 0)) __PYX_ERR(0, 1212, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
      values[1] = __Pyx_Arg_FASTCALL(__pyx_args, 1);
    }
    __pyx_v_current_prices = values[0];
    __pyx_v_proposal_prices = values[1];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("is_within_tolerance", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 1212, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.is_within_tolerance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_80is_within_tolerance(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_current_prices, __pyx_v_proposal_prices);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_80is_within_tolerance(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_current_prices, PyObject *__pyx_v_proposal_prices) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_within_tolerance", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1213
 * 
 *     def is_within_tolerance(self, current_prices: List[Decimal], proposal_prices: List[Decimal]) -> bool:
 *         return self.c_is_within_tolerance(current_prices, proposal_prices)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_cancel_active_orders_on_max_age_limit(self):
 */
  __Pyx_XDECREF(__pyx_r);
  if (!(likely(PyList_CheckExact(__pyx_v_current_prices))||((__pyx_v_current_prices) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_v_current_prices))) __PYX_ERR(0, 1213, __pyx_L1_error)
  if (!(likely(PyList_CheckExact(__pyx_v_proposal_prices))||((__pyx_v_proposal_prices) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_v_proposal_prices))) __PYX_ERR(0, 1213, __pyx_L1_error)
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_is_within_tolerance(__pyx_v_self, ((PyObject*)__pyx_v_current_prices), ((PyObject*)__pyx_v_proposal_prices)); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1213, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1213, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1212
 *         return True
 * 
 *     def is_within_tolerance(self, current_prices: List[Decimal], proposal_prices: List[Decimal]) -> bool:             # <<<<<<<<<<<<<<
 *         return self.c_is_within_tolerance(current_prices, proposal_prices)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.is_within_tolerance", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1215
 *         return self.c_is_within_tolerance(current_prices, proposal_prices)
 * 
 *     cdef c_cancel_active_orders_on_max_age_limit(self):             # <<<<<<<<<<<<<<
 *         """
 *         Cancels active non hanging orders if they are older than max age limit
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_cancel_active_orders_on_max_age_limit(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_v_active_orders = 0;
  PyObject *__pyx_v_order = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  unsigned int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_cancel_active_orders_on_max_age_limit", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1220
 *         """
 *         cdef:
 *             list active_orders = self.active_non_hanging_orders             # <<<<<<<<<<<<<<
 *         for order in active_orders:
 *             if order_age(order, self._current_timestamp) > self._config_map.max_order_age:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_active_non_hanging_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(0, 1220, __pyx_L1_error)
  __pyx_v_active_orders = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1221
 *         cdef:
 *             list active_orders = self.active_non_hanging_orders
 *         for order in active_orders:             # <<<<<<<<<<<<<<
 *             if order_age(order, self._current_timestamp) > self._config_map.max_order_age:
 *                 self.c_cancel_order(self._market_info, order.client_order_id)
 */
  if (unlikely(__pyx_v_active_orders == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 1221, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_active_orders; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_MACROS
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1221, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 1221, __pyx_L1_error)
    #else
    __pyx_t_3 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1221, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_order, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1222
 *             list active_orders = self.active_non_hanging_orders
 *         for order in active_orders:
 *             if order_age(order, self._current_timestamp) > self._config_map.max_order_age:             # <<<<<<<<<<<<<<
 *                 self.c_cancel_order(self._market_info, order.client_order_id)
 * 
 */
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_order_age); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    __pyx_t_7 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_7 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_6, __pyx_v_order, __pyx_t_5};
      __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_7, 2+__pyx_t_7);
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1222, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_max_order_age); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyObject_RichCompare(__pyx_t_3, __pyx_t_4, Py_GT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1222, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_8 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 1222, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (__pyx_t_8) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1223
 *         for order in active_orders:
 *             if order_age(order, self._current_timestamp) > self._config_map.max_order_age:
 *                 self.c_cancel_order(self._market_info, order.client_order_id)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_cancel_active_orders(self, object proposal):
 */
      __pyx_t_5 = __pyx_v_self->_market_info;
      __Pyx_INCREF(__pyx_t_5);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_n_s_client_order_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1223, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(PyUnicode_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_4))) __PYX_ERR(0, 1223, __pyx_L1_error)
      __pyx_t_3 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_cancel_order(((struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *)__pyx_v_self), __pyx_t_5, ((PyObject*)__pyx_t_4)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1223, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1222
 *             list active_orders = self.active_non_hanging_orders
 *         for order in active_orders:
 *             if order_age(order, self._current_timestamp) > self._config_map.max_order_age:             # <<<<<<<<<<<<<<
 *                 self.c_cancel_order(self._market_info, order.client_order_id)
 * 
 */
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1221
 *         cdef:
 *             list active_orders = self.active_non_hanging_orders
 *         for order in active_orders:             # <<<<<<<<<<<<<<
 *             if order_age(order, self._current_timestamp) > self._config_map.max_order_age:
 *                 self.c_cancel_order(self._market_info, order.client_order_id)
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1215
 *         return self.c_is_within_tolerance(current_prices, proposal_prices)
 * 
 *     cdef c_cancel_active_orders_on_max_age_limit(self):             # <<<<<<<<<<<<<<
 *         """
 *         Cancels active non hanging orders if they are older than max age limit
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_cancel_active_orders_on_max_age_limit", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_active_orders);
  __Pyx_XDECREF(__pyx_v_order);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1225
 *                 self.c_cancel_order(self._market_info, order.client_order_id)
 * 
 *     cdef c_cancel_active_orders(self, object proposal):             # <<<<<<<<<<<<<<
 *         if self._cancel_timestamp > self._current_timestamp:
 *             return
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_cancel_active_orders(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_v_active_buy_prices = 0;
  CYTHON_UNUSED PyObject *__pyx_v_active_sells = 0;
  int __pyx_v_to_defer_canceling;
  PyObject *__pyx_v_active_sell_prices = NULL;
  PyObject *__pyx_v_proposal_buys = NULL;
  PyObject *__pyx_v_proposal_sells = NULL;
  PyObject *__pyx_v_order = NULL;
  PyObject *__pyx_9genexpr18__pyx_v_o = NULL;
  PyObject *__pyx_9genexpr19__pyx_v_o = NULL;
  PyObject *__pyx_9genexpr20__pyx_v_buy = NULL;
  PyObject *__pyx_9genexpr21__pyx_v_sell = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  unsigned int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_cancel_active_orders", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1226
 * 
 *     cdef c_cancel_active_orders(self, object proposal):
 *         if self._cancel_timestamp > self._current_timestamp:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_cancel_timestamp > __pyx_v_self->__pyx_base.__pyx_base._current_timestamp);
  if (__pyx_t_1) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1227
 *     cdef c_cancel_active_orders(self, object proposal):
 *         if self._cancel_timestamp > self._current_timestamp:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         cdef:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1226
 * 
 *     cdef c_cancel_active_orders(self, object proposal):
 *         if self._cancel_timestamp > self._current_timestamp:             # <<<<<<<<<<<<<<
 *             return
 * 
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1230
 * 
 *         cdef:
 *             list active_buy_prices = []             # <<<<<<<<<<<<<<
 *             list active_sells = []
 *             bint to_defer_canceling = False
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_active_buy_prices = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1231
 *         cdef:
 *             list active_buy_prices = []
 *             list active_sells = []             # <<<<<<<<<<<<<<
 *             bint to_defer_canceling = False
 * 
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_active_sells = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1232
 *             list active_buy_prices = []
 *             list active_sells = []
 *             bint to_defer_canceling = False             # <<<<<<<<<<<<<<
 * 
 *         if len(self.active_non_hanging_orders) == 0:
 */
  __pyx_v_to_defer_canceling = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1234
 *             bint to_defer_canceling = False
 * 
 *         if len(self.active_non_hanging_orders) == 0:             # <<<<<<<<<<<<<<
 *             return
 *         if proposal is not None:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_active_non_hanging_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyObject_Length(__pyx_t_2); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1234, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_1 = (__pyx_t_3 == 0);
  if (__pyx_t_1) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1235
 * 
 *         if len(self.active_non_hanging_orders) == 0:
 *             return             # <<<<<<<<<<<<<<
 *         if proposal is not None:
 *             active_buy_prices = [Decimal(str(o.price)) for o in self.active_non_hanging_orders if o.is_buy]
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1234
 *             bint to_defer_canceling = False
 * 
 *         if len(self.active_non_hanging_orders) == 0:             # <<<<<<<<<<<<<<
 *             return
 *         if proposal is not None:
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1236
 *         if len(self.active_non_hanging_orders) == 0:
 *             return
 *         if proposal is not None:             # <<<<<<<<<<<<<<
 *             active_buy_prices = [Decimal(str(o.price)) for o in self.active_non_hanging_orders if o.is_buy]
 *             active_sell_prices = [Decimal(str(o.price)) for o in self.active_non_hanging_orders if not o.is_buy]
 */
  __pyx_t_1 = (__pyx_v_proposal != Py_None);
  if (__pyx_t_1) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1237
 *             return
 *         if proposal is not None:
 *             active_buy_prices = [Decimal(str(o.price)) for o in self.active_non_hanging_orders if o.is_buy]             # <<<<<<<<<<<<<<
 *             active_sell_prices = [Decimal(str(o.price)) for o in self.active_non_hanging_orders if not o.is_buy]
 *             proposal_buys = [buy.price for buy in proposal.buys]
 */
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1237, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_active_non_hanging_orders); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1237, __pyx_L8_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
        __pyx_t_5 = __pyx_t_4; __Pyx_INCREF(__pyx_t_5);
        __pyx_t_3 = 0;
        __pyx_t_6 = NULL;
      } else {
        __pyx_t_3 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1237, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1237, __pyx_L8_error)
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      for (;;) {
        if (likely(!__pyx_t_6)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1237, __pyx_L8_error)
              #endif
              if (__pyx_t_3 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1237, __pyx_L8_error)
            #else
            __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1237, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1237, __pyx_L8_error)
              #endif
              if (__pyx_t_3 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1237, __pyx_L8_error)
            #else
            __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1237, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          }
        } else {
          __pyx_t_4 = __pyx_t_6(__pyx_t_5);
          if (unlikely(!__pyx_t_4)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 1237, __pyx_L8_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_XDECREF_SET(__pyx_9genexpr18__pyx_v_o, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr18__pyx_v_o, __pyx_n_s_is_buy); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1237, __pyx_L8_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1237, __pyx_L8_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (__pyx_t_1) {
          __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1237, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr18__pyx_v_o, __pyx_n_s_price); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1237, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_9 = __Pyx_PyObject_Unicode(__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1237, __pyx_L8_error)
          __Pyx_GOTREF(__pyx_t_9);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_8 = NULL;
          __pyx_t_10 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_8)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_10 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_8, __pyx_t_9};
            __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
            if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1237, __pyx_L8_error)
            __Pyx_GOTREF(__pyx_t_4);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 1237, __pyx_L8_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_o); __pyx_9genexpr18__pyx_v_o = 0;
      goto __pyx_L13_exit_scope;
      __pyx_L8_error:;
      __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_o); __pyx_9genexpr18__pyx_v_o = 0;
      goto __pyx_L1_error;
      __pyx_L13_exit_scope:;
    } /* exit inner scope */
    __Pyx_DECREF_SET(__pyx_v_active_buy_prices, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1238
 *         if proposal is not None:
 *             active_buy_prices = [Decimal(str(o.price)) for o in self.active_non_hanging_orders if o.is_buy]
 *             active_sell_prices = [Decimal(str(o.price)) for o in self.active_non_hanging_orders if not o.is_buy]             # <<<<<<<<<<<<<<
 *             proposal_buys = [buy.price for buy in proposal.buys]
 *             proposal_sells = [sell.price for sell in proposal.sells]
 */
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1238, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_active_non_hanging_orders); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1238, __pyx_L16_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
        __pyx_t_4 = __pyx_t_5; __Pyx_INCREF(__pyx_t_4);
        __pyx_t_3 = 0;
        __pyx_t_6 = NULL;
      } else {
        __pyx_t_3 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1238, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1238, __pyx_L16_error)
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      for (;;) {
        if (likely(!__pyx_t_6)) {
          if (likely(PyList_CheckExact(__pyx_t_4))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1238, __pyx_L16_error)
              #endif
              if (__pyx_t_3 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_5 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1238, __pyx_L16_error)
            #else
            __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1238, __pyx_L16_error)
            __Pyx_GOTREF(__pyx_t_5);
            #endif
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1238, __pyx_L16_error)
              #endif
              if (__pyx_t_3 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1238, __pyx_L16_error)
            #else
            __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1238, __pyx_L16_error)
            __Pyx_GOTREF(__pyx_t_5);
            #endif
          }
        } else {
          __pyx_t_5 = __pyx_t_6(__pyx_t_4);
          if (unlikely(!__pyx_t_5)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 1238, __pyx_L16_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_5);
        }
        __Pyx_XDECREF_SET(__pyx_9genexpr19__pyx_v_o, __pyx_t_5);
        __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr19__pyx_v_o, __pyx_n_s_is_buy); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1238, __pyx_L16_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely((__pyx_t_1 < 0))) __PYX_ERR(0, 1238, __pyx_L16_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_11 = (!__pyx_t_1);
        if (__pyx_t_11) {
          __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1238, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_7);
          __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr19__pyx_v_o, __pyx_n_s_price); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1238, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_8 = __Pyx_PyObject_Unicode(__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1238, __pyx_L16_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
          __pyx_t_9 = NULL;
          __pyx_t_10 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_7))) {
            __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
            if (likely(__pyx_t_9)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
              __Pyx_INCREF(__pyx_t_9);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_7, function);
              __pyx_t_10 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_8};
            __pyx_t_5 = __Pyx_PyObject_FastCall(__pyx_t_7, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
            __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1238, __pyx_L16_error)
            __Pyx_GOTREF(__pyx_t_5);
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          }
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 1238, __pyx_L16_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_9genexpr19__pyx_v_o); __pyx_9genexpr19__pyx_v_o = 0;
      goto __pyx_L21_exit_scope;
      __pyx_L16_error:;
      __Pyx_XDECREF(__pyx_9genexpr19__pyx_v_o); __pyx_9genexpr19__pyx_v_o = 0;
      goto __pyx_L1_error;
      __pyx_L21_exit_scope:;
    } /* exit inner scope */
    __pyx_v_active_sell_prices = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1239
 *             active_buy_prices = [Decimal(str(o.price)) for o in self.active_non_hanging_orders if o.is_buy]
 *             active_sell_prices = [Decimal(str(o.price)) for o in self.active_non_hanging_orders if not o.is_buy]
 *             proposal_buys = [buy.price for buy in proposal.buys]             # <<<<<<<<<<<<<<
 *             proposal_sells = [sell.price for sell in proposal.sells]
 * 
 */
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1239, __pyx_L24_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1239, __pyx_L24_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
        __pyx_t_5 = __pyx_t_4; __Pyx_INCREF(__pyx_t_5);
        __pyx_t_3 = 0;
        __pyx_t_6 = NULL;
      } else {
        __pyx_t_3 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1239, __pyx_L24_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1239, __pyx_L24_error)
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      for (;;) {
        if (likely(!__pyx_t_6)) {
          if (likely(PyList_CheckExact(__pyx_t_5))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_5);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1239, __pyx_L24_error)
              #endif
              if (__pyx_t_3 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1239, __pyx_L24_error)
            #else
            __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1239, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_5);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1239, __pyx_L24_error)
              #endif
              if (__pyx_t_3 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_3); __Pyx_INCREF(__pyx_t_4); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1239, __pyx_L24_error)
            #else
            __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_5, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1239, __pyx_L24_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          }
        } else {
          __pyx_t_4 = __pyx_t_6(__pyx_t_5);
          if (unlikely(!__pyx_t_4)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 1239, __pyx_L24_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_XDECREF_SET(__pyx_9genexpr20__pyx_v_buy, __pyx_t_4);
        __pyx_t_4 = 0;
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr20__pyx_v_buy, __pyx_n_s_price); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1239, __pyx_L24_error)
        __Pyx_GOTREF(__pyx_t_4);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 1239, __pyx_L24_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_9genexpr20__pyx_v_buy); __pyx_9genexpr20__pyx_v_buy = 0;
      goto __pyx_L28_exit_scope;
      __pyx_L24_error:;
      __Pyx_XDECREF(__pyx_9genexpr20__pyx_v_buy); __pyx_9genexpr20__pyx_v_buy = 0;
      goto __pyx_L1_error;
      __pyx_L28_exit_scope:;
    } /* exit inner scope */
    __pyx_v_proposal_buys = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1240
 *             active_sell_prices = [Decimal(str(o.price)) for o in self.active_non_hanging_orders if not o.is_buy]
 *             proposal_buys = [buy.price for buy in proposal.buys]
 *             proposal_sells = [sell.price for sell in proposal.sells]             # <<<<<<<<<<<<<<
 * 
 *             if self.c_is_within_tolerance(active_buy_prices, proposal_buys) and \
 */
    { /* enter inner scope */
      __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1240, __pyx_L31_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1240, __pyx_L31_error)
      __Pyx_GOTREF(__pyx_t_5);
      if (likely(PyList_CheckExact(__pyx_t_5)) || PyTuple_CheckExact(__pyx_t_5)) {
        __pyx_t_4 = __pyx_t_5; __Pyx_INCREF(__pyx_t_4);
        __pyx_t_3 = 0;
        __pyx_t_6 = NULL;
      } else {
        __pyx_t_3 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1240, __pyx_L31_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1240, __pyx_L31_error)
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      for (;;) {
        if (likely(!__pyx_t_6)) {
          if (likely(PyList_CheckExact(__pyx_t_4))) {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1240, __pyx_L31_error)
              #endif
              if (__pyx_t_3 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_5 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1240, __pyx_L31_error)
            #else
            __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1240, __pyx_L31_error)
            __Pyx_GOTREF(__pyx_t_5);
            #endif
          } else {
            {
              Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
              #if !CYTHON_ASSUME_SAFE_MACROS
              if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1240, __pyx_L31_error)
              #endif
              if (__pyx_t_3 >= __pyx_temp) break;
            }
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1240, __pyx_L31_error)
            #else
            __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1240, __pyx_L31_error)
            __Pyx_GOTREF(__pyx_t_5);
            #endif
          }
        } else {
          __pyx_t_5 = __pyx_t_6(__pyx_t_4);
          if (unlikely(!__pyx_t_5)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 1240, __pyx_L31_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_5);
        }
        __Pyx_XDECREF_SET(__pyx_9genexpr21__pyx_v_sell, __pyx_t_5);
        __pyx_t_5 = 0;
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr21__pyx_v_sell, __pyx_n_s_price); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1240, __pyx_L31_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_5))) __PYX_ERR(0, 1240, __pyx_L31_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_9genexpr21__pyx_v_sell); __pyx_9genexpr21__pyx_v_sell = 0;
      goto __pyx_L35_exit_scope;
      __pyx_L31_error:;
      __Pyx_XDECREF(__pyx_9genexpr21__pyx_v_sell); __pyx_9genexpr21__pyx_v_sell = 0;
      goto __pyx_L1_error;
      __pyx_L35_exit_scope:;
    } /* exit inner scope */
    __pyx_v_proposal_sells = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1242
 *             proposal_sells = [sell.price for sell in proposal.sells]
 * 
 *             if self.c_is_within_tolerance(active_buy_prices, proposal_buys) and \             # <<<<<<<<<<<<<<
 *                     self.c_is_within_tolerance(active_sell_prices, proposal_sells):
 *                 to_defer_canceling = True
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_is_within_tolerance(__pyx_v_self, __pyx_v_active_buy_prices, __pyx_v_proposal_buys); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1242, __pyx_L1_error)
    if (__pyx_t_1) {
    } else {
      __pyx_t_11 = __pyx_t_1;
      goto __pyx_L37_bool_binop_done;
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1243
 * 
 *             if self.c_is_within_tolerance(active_buy_prices, proposal_buys) and \
 *                     self.c_is_within_tolerance(active_sell_prices, proposal_sells):             # <<<<<<<<<<<<<<
 *                 to_defer_canceling = True
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_is_within_tolerance(__pyx_v_self, __pyx_v_active_sell_prices, __pyx_v_proposal_sells); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1243, __pyx_L1_error)
    __pyx_t_11 = __pyx_t_1;
    __pyx_L37_bool_binop_done:;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1242
 *             proposal_sells = [sell.price for sell in proposal.sells]
 * 
 *             if self.c_is_within_tolerance(active_buy_prices, proposal_buys) and \             # <<<<<<<<<<<<<<
 *                     self.c_is_within_tolerance(active_sell_prices, proposal_sells):
 *                 to_defer_canceling = True
 */
    if (__pyx_t_11) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1244
 *             if self.c_is_within_tolerance(active_buy_prices, proposal_buys) and \
 *                     self.c_is_within_tolerance(active_sell_prices, proposal_sells):
 *                 to_defer_canceling = True             # <<<<<<<<<<<<<<
 * 
 *         if not to_defer_canceling:
 */
      __pyx_v_to_defer_canceling = 1;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1242
 *             proposal_sells = [sell.price for sell in proposal.sells]
 * 
 *             if self.c_is_within_tolerance(active_buy_prices, proposal_buys) and \             # <<<<<<<<<<<<<<
 *                     self.c_is_within_tolerance(active_sell_prices, proposal_sells):
 *                 to_defer_canceling = True
 */
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1236
 *         if len(self.active_non_hanging_orders) == 0:
 *             return
 *         if proposal is not None:             # <<<<<<<<<<<<<<
 *             active_buy_prices = [Decimal(str(o.price)) for o in self.active_non_hanging_orders if o.is_buy]
 *             active_sell_prices = [Decimal(str(o.price)) for o in self.active_non_hanging_orders if not o.is_buy]
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1246
 *                 to_defer_canceling = True
 * 
 *         if not to_defer_canceling:             # <<<<<<<<<<<<<<
 *             self._hanging_orders_tracker.update_strategy_orders_with_equivalent_orders()
 *             for order in self.active_non_hanging_orders:
 */
  __pyx_t_11 = (!__pyx_v_to_defer_canceling);
  if (__pyx_t_11) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1247
 * 
 *         if not to_defer_canceling:
 *             self._hanging_orders_tracker.update_strategy_orders_with_equivalent_orders()             # <<<<<<<<<<<<<<
 *             for order in self.active_non_hanging_orders:
 *                 # If is about to be added to hanging_orders then don't cancel
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_n_s_update_strategy_orders_with_equi); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_10 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_10 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+1-__pyx_t_10, 0+__pyx_t_10);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1248
 *         if not to_defer_canceling:
 *             self._hanging_orders_tracker.update_strategy_orders_with_equivalent_orders()
 *             for order in self.active_non_hanging_orders:             # <<<<<<<<<<<<<<
 *                 # If is about to be added to hanging_orders then don't cancel
 *                 if not self._hanging_orders_tracker.is_potential_hanging_order(order):
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_active_non_hanging_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_4 = __pyx_t_2; __Pyx_INCREF(__pyx_t_4);
      __pyx_t_3 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1248, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1248, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1248, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1248, __pyx_L1_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1248, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_4);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1248, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_3); __Pyx_INCREF(__pyx_t_2); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1248, __pyx_L1_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_4, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1248, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_6(__pyx_t_4);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1248, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_v_order, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1250
 *             for order in self.active_non_hanging_orders:
 *                 # If is about to be added to hanging_orders then don't cancel
 *                 if not self._hanging_orders_tracker.is_potential_hanging_order(order):             # <<<<<<<<<<<<<<
 *                     self.c_cancel_order(self._market_info, order.client_order_id)
 *         else:
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_n_s_is_potential_hanging_order); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1250, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_7 = NULL;
      __pyx_t_10 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_10 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_v_order};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+1-__pyx_t_10, 1+__pyx_t_10);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1250, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
      __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_11 < 0))) __PYX_ERR(0, 1250, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_1 = (!__pyx_t_11);
      if (__pyx_t_1) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1251
 *                 # If is about to be added to hanging_orders then don't cancel
 *                 if not self._hanging_orders_tracker.is_potential_hanging_order(order):
 *                     self.c_cancel_order(self._market_info, order.client_order_id)             # <<<<<<<<<<<<<<
 *         else:
 *             self.c_set_timers()
 */
        __pyx_t_2 = __pyx_v_self->_market_info;
        __Pyx_INCREF(__pyx_t_2);
        __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_order, __pyx_n_s_client_order_id); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1251, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (!(likely(PyUnicode_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None) || __Pyx_RaiseUnexpectedTypeError("unicode", __pyx_t_5))) __PYX_ERR(0, 1251, __pyx_L1_error)
        __pyx_t_7 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_cancel_order(((struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *)__pyx_v_self), __pyx_t_2, ((PyObject*)__pyx_t_5)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1251, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1250
 *             for order in self.active_non_hanging_orders:
 *                 # If is about to be added to hanging_orders then don't cancel
 *                 if not self._hanging_orders_tracker.is_potential_hanging_order(order):             # <<<<<<<<<<<<<<
 *                     self.c_cancel_order(self._market_info, order.client_order_id)
 *         else:
 */
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1248
 *         if not to_defer_canceling:
 *             self._hanging_orders_tracker.update_strategy_orders_with_equivalent_orders()
 *             for order in self.active_non_hanging_orders:             # <<<<<<<<<<<<<<
 *                 # If is about to be added to hanging_orders then don't cancel
 *                 if not self._hanging_orders_tracker.is_potential_hanging_order(order):
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1246
 *                 to_defer_canceling = True
 * 
 *         if not to_defer_canceling:             # <<<<<<<<<<<<<<
 *             self._hanging_orders_tracker.update_strategy_orders_with_equivalent_orders()
 *             for order in self.active_non_hanging_orders:
 */
    goto __pyx_L39;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1253
 *                     self.c_cancel_order(self._market_info, order.client_order_id)
 *         else:
 *             self.c_set_timers()             # <<<<<<<<<<<<<<
 * 
 *     def cancel_active_orders(self, proposal: Proposal = None):
 */
  /*else*/ {
    __pyx_t_4 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_set_timers(__pyx_v_self); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_L39:;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1225
 *                 self.c_cancel_order(self._market_info, order.client_order_id)
 * 
 *     cdef c_cancel_active_orders(self, object proposal):             # <<<<<<<<<<<<<<
 *         if self._cancel_timestamp > self._current_timestamp:
 *             return
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_cancel_active_orders", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_active_buy_prices);
  __Pyx_XDECREF(__pyx_v_active_sells);
  __Pyx_XDECREF(__pyx_v_active_sell_prices);
  __Pyx_XDECREF(__pyx_v_proposal_buys);
  __Pyx_XDECREF(__pyx_v_proposal_sells);
  __Pyx_XDECREF(__pyx_v_order);
  __Pyx_XDECREF(__pyx_9genexpr18__pyx_v_o);
  __Pyx_XDECREF(__pyx_9genexpr19__pyx_v_o);
  __Pyx_XDECREF(__pyx_9genexpr20__pyx_v_buy);
  __Pyx_XDECREF(__pyx_9genexpr21__pyx_v_sell);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1255
 *             self.c_set_timers()
 * 
 *     def cancel_active_orders(self, proposal: Proposal = None):             # <<<<<<<<<<<<<<
 *         return self.c_cancel_active_orders(proposal)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_83cancel_active_orders(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_83cancel_active_orders = {"cancel_active_orders", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_83cancel_active_orders, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_83cancel_active_orders(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_proposal = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cancel_active_orders (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_proposal,0};
    values[0] = __Pyx_Arg_NewRef_FASTCALL(((PyObject *)Py_None));
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_proposal);
          if (value) { values[0] = __Pyx_Arg_NewRef_FASTCALL(value); kw_args--; }
          else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1255, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "cancel_active_orders") < 0)) __PYX_ERR(0, 1255, __pyx_L3_error)
      }
    } else {
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_proposal = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cancel_active_orders", 0, 0, 1, __pyx_nargs); __PYX_ERR(0, 1255, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.cancel_active_orders", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_82cancel_active_orders(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_proposal);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_82cancel_active_orders(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("cancel_active_orders", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1256
 * 
 *     def cancel_active_orders(self, proposal: Proposal = None):
 *         return self.c_cancel_active_orders(proposal)             # <<<<<<<<<<<<<<
 * 
 *     cdef bint c_to_create_orders(self, object proposal):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_cancel_active_orders(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1256, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1255
 *             self.c_set_timers()
 * 
 *     def cancel_active_orders(self, proposal: Proposal = None):             # <<<<<<<<<<<<<<
 *         return self.c_cancel_active_orders(proposal)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.cancel_active_orders", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1258
 *         return self.c_cancel_active_orders(proposal)
 * 
 *     cdef bint c_to_create_orders(self, object proposal):             # <<<<<<<<<<<<<<
 *         non_hanging_orders_non_cancelled = [o for o in self.active_non_hanging_orders if not
 *                                             self._hanging_orders_tracker.is_potential_hanging_order(o)]
 */

static int __pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_to_create_orders(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_v_non_hanging_orders_non_cancelled = NULL;
  PyObject *__pyx_9genexpr22__pyx_v_o = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  unsigned int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_to_create_orders", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1259
 * 
 *     cdef bint c_to_create_orders(self, object proposal):
 *         non_hanging_orders_non_cancelled = [o for o in self.active_non_hanging_orders if not             # <<<<<<<<<<<<<<
 *                                             self._hanging_orders_tracker.is_potential_hanging_order(o)]
 * 
 */
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1259, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_active_non_hanging_orders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1259, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
      __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = 0;
      __pyx_t_5 = NULL;
    } else {
      __pyx_t_4 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1259, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1259, __pyx_L5_error)
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    for (;;) {
      if (likely(!__pyx_t_5)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1259, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 1259, __pyx_L5_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1259, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1259, __pyx_L5_error)
            #endif
            if (__pyx_t_4 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_4); __Pyx_INCREF(__pyx_t_2); __pyx_t_4++; if (unlikely((0 < 0))) __PYX_ERR(0, 1259, __pyx_L5_error)
          #else
          __pyx_t_2 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_4); __pyx_t_4++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1259, __pyx_L5_error)
          __Pyx_GOTREF(__pyx_t_2);
          #endif
        }
      } else {
        __pyx_t_2 = __pyx_t_5(__pyx_t_3);
        if (unlikely(!__pyx_t_2)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1259, __pyx_L5_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_XDECREF_SET(__pyx_9genexpr22__pyx_v_o, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1260
 *     cdef bint c_to_create_orders(self, object proposal):
 *         non_hanging_orders_non_cancelled = [o for o in self.active_non_hanging_orders if not
 *                                             self._hanging_orders_tracker.is_potential_hanging_order(o)]             # <<<<<<<<<<<<<<
 * 
 *         return (self._create_timestamp < self._current_timestamp
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_n_s_is_potential_hanging_order); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1260, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_8 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_7, __pyx_9genexpr22__pyx_v_o};
        __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_8, 1+__pyx_t_8);
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1260, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
      __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 1260, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1259
 * 
 *     cdef bint c_to_create_orders(self, object proposal):
 *         non_hanging_orders_non_cancelled = [o for o in self.active_non_hanging_orders if not             # <<<<<<<<<<<<<<
 *                                             self._hanging_orders_tracker.is_potential_hanging_order(o)]
 * 
 */
      __pyx_t_10 = (!__pyx_t_9);
      if (__pyx_t_10) {
        if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_9genexpr22__pyx_v_o))) __PYX_ERR(0, 1259, __pyx_L5_error)
      }
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_9genexpr22__pyx_v_o); __pyx_9genexpr22__pyx_v_o = 0;
    goto __pyx_L10_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_9genexpr22__pyx_v_o); __pyx_9genexpr22__pyx_v_o = 0;
    goto __pyx_L1_error;
    __pyx_L10_exit_scope:;
  } /* exit inner scope */
  __pyx_v_non_hanging_orders_non_cancelled = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1262
 *                                             self._hanging_orders_tracker.is_potential_hanging_order(o)]
 * 
 *         return (self._create_timestamp < self._current_timestamp             # <<<<<<<<<<<<<<
 *                 and (not self._config_map.should_wait_order_cancel_confirmation or
 *                      len(self._sb_order_tracker.in_flight_cancels) == 0)
 */
  __pyx_t_9 = (__pyx_v_self->_create_timestamp < __pyx_v_self->__pyx_base.__pyx_base._current_timestamp);
  if (__pyx_t_9) {
  } else {
    __pyx_t_10 = __pyx_t_9;
    goto __pyx_L11_bool_binop_done;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1263
 * 
 *         return (self._create_timestamp < self._current_timestamp
 *                 and (not self._config_map.should_wait_order_cancel_confirmation or             # <<<<<<<<<<<<<<
 *                      len(self._sb_order_tracker.in_flight_cancels) == 0)
 *                 and proposal is not None
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_config_map, __pyx_n_s_should_wait_order_cancel_confirm); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_9 < 0))) __PYX_ERR(0, 1263, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = (!__pyx_t_9);
  if (!__pyx_t_11) {
  } else {
    goto __pyx_L13_next_and;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1264
 *         return (self._create_timestamp < self._current_timestamp
 *                 and (not self._config_map.should_wait_order_cancel_confirmation or
 *                      len(self._sb_order_tracker.in_flight_cancels) == 0)             # <<<<<<<<<<<<<<
 *                 and proposal is not None
 *                 and len(non_hanging_orders_non_cancelled) == 0)
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self->__pyx_base._sb_order_tracker), __pyx_n_s_in_flight_cancels); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1264, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = (__pyx_t_4 == 0);
  if (__pyx_t_11) {
  } else {
    __pyx_t_10 = __pyx_t_11;
    goto __pyx_L11_bool_binop_done;
  }
  __pyx_L13_next_and:;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1265
 *                 and (not self._config_map.should_wait_order_cancel_confirmation or
 *                      len(self._sb_order_tracker.in_flight_cancels) == 0)
 *                 and proposal is not None             # <<<<<<<<<<<<<<
 *                 and len(non_hanging_orders_non_cancelled) == 0)
 * 
 */
  __pyx_t_11 = (__pyx_v_proposal != Py_None);
  if (__pyx_t_11) {
  } else {
    __pyx_t_10 = __pyx_t_11;
    goto __pyx_L11_bool_binop_done;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1266
 *                      len(self._sb_order_tracker.in_flight_cancels) == 0)
 *                 and proposal is not None
 *                 and len(non_hanging_orders_non_cancelled) == 0)             # <<<<<<<<<<<<<<
 * 
 *     def to_create_orders(self, proposal: Proposal) -> bool:
 */
  __pyx_t_4 = __Pyx_PyList_GET_SIZE(__pyx_v_non_hanging_orders_non_cancelled); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1266, __pyx_L1_error)
  __pyx_t_11 = (__pyx_t_4 == 0);
  __pyx_t_10 = __pyx_t_11;
  __pyx_L11_bool_binop_done:;
  __pyx_r = __pyx_t_10;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1258
 *         return self.c_cancel_active_orders(proposal)
 * 
 *     cdef bint c_to_create_orders(self, object proposal):             # <<<<<<<<<<<<<<
 *         non_hanging_orders_non_cancelled = [o for o in self.active_non_hanging_orders if not
 *                                             self._hanging_orders_tracker.is_potential_hanging_order(o)]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_to_create_orders", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_non_hanging_orders_non_cancelled);
  __Pyx_XDECREF(__pyx_9genexpr22__pyx_v_o);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1268
 *                 and len(non_hanging_orders_non_cancelled) == 0)
 * 
 *     def to_create_orders(self, proposal: Proposal) -> bool:             # <<<<<<<<<<<<<<
 *         return self.c_to_create_orders(proposal)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_85to_create_orders(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_85to_create_orders = {"to_create_orders", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_85to_create_orders, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_85to_create_orders(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_proposal = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_create_orders (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_proposal,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_proposal)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1268, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "to_create_orders") < 0)) __PYX_ERR(0, 1268, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_proposal = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("to_create_orders", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1268, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.to_create_orders", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_84to_create_orders(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_proposal);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_84to_create_orders(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("to_create_orders", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1269
 * 
 *     def to_create_orders(self, proposal: Proposal) -> bool:
 *         return self.c_to_create_orders(proposal)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_execute_orders_proposal(self, object proposal):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_to_create_orders(__pyx_v_self, __pyx_v_proposal); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1269, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1269, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1268
 *                 and len(non_hanging_orders_non_cancelled) == 0)
 * 
 *     def to_create_orders(self, proposal: Proposal) -> bool:             # <<<<<<<<<<<<<<
 *         return self.c_to_create_orders(proposal)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.to_create_orders", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25c_execute_orders_proposal_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1298
 *                 orders_created = True
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))             # <<<<<<<<<<<<<<
 *                     if order:
 *                         self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(
 */

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25c_execute_orders_proposal_genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr *)__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1298, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal *) __pyx_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25c_execute_orders_proposal_2generator1, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_AvellanedaMarketMakingStrategy_c_3, __pyx_n_s_hummingbot_strategy_avellaneda_m); if (unlikely(!gen)) __PYX_ERR(0, 1298, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_execute_orders_proposal.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25c_execute_orders_proposal_2generator1(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr *__pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1298, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 1298, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_genexpr_arg_0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1298, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1298, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 1298, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1298, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1298, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 1298, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1298, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1298, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_o);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_o, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_o, __pyx_n_s_client_order_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1298, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_bid_order_id)) { __Pyx_RaiseClosureNameError("bid_order_id"); __PYX_ERR(0, 1298, __pyx_L1_error) }
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_4, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_bid_order_id, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1298, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_5) {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_o);
      __pyx_r = __pyx_cur_scope->__pyx_v_o;
      __Pyx_XGIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L7_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1298, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25c_execute_orders_proposal_5generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value); /* proto */

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1321
 *                 orders_created = True
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))             # <<<<<<<<<<<<<<
 *                     if order:
 *                         self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order
 */

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25c_execute_orders_proposal_3genexpr(PyObject *__pyx_self, PyObject *__pyx_genexpr_arg_0) {
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr *)__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1321, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal *) __pyx_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_outer_scope);
  __pyx_cur_scope->__pyx_genexpr_arg_0 = __pyx_genexpr_arg_0;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_genexpr_arg_0);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25c_execute_orders_proposal_5generator2, NULL, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_AvellanedaMarketMakingStrategy_c_3, __pyx_n_s_hummingbot_strategy_avellaneda_m); if (unlikely(!gen)) __PYX_ERR(0, 1321, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_execute_orders_proposal.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25c_execute_orders_proposal_5generator2(__pyx_CoroutineObject *__pyx_generator, CYTHON_UNUSED PyThreadState *__pyx_tstate, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr *__pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1321, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_genexpr_arg_0)) { __Pyx_RaiseUnboundLocalError(".0"); __PYX_ERR(0, 1321, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_genexpr_arg_0)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_genexpr_arg_0; __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_genexpr_arg_0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1321, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1321, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 1321, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1321, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        {
          Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
          #if !CYTHON_ASSUME_SAFE_MACROS
          if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1321, __pyx_L1_error)
          #endif
          if (__pyx_t_2 >= __pyx_temp) break;
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely((0 < 0))) __PYX_ERR(0, 1321, __pyx_L1_error)
        #else
        __pyx_t_4 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1321, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 1321, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_o);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_o, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_o, __pyx_n_s_client_order_id); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1321, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_ask_order_id)) { __Pyx_RaiseClosureNameError("ask_order_id"); __PYX_ERR(0, 1321, __pyx_L1_error) }
    __pyx_t_5 = (__Pyx_PyUnicode_Equals(__pyx_t_4, __pyx_cur_scope->__pyx_outer_scope->__pyx_v_ask_order_id, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 1321, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_5) {
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_o);
      __pyx_r = __pyx_cur_scope->__pyx_v_o;
      __Pyx_XGIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L7_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 1321, __pyx_L1_error)
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_Generator_Replace_StopIteration(0);
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  #if !CYTHON_USE_EXC_INFO_STACK
  __Pyx_Coroutine_ResetAndClearException(__pyx_generator);
  #endif
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1271
 *         return self.c_to_create_orders(proposal)
 * 
 *     cdef c_execute_orders_proposal(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double expiration_seconds = NaN
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_execute_orders_proposal(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal *__pyx_cur_scope;
  double __pyx_v_expiration_seconds;
  int __pyx_v_orders_created;
  Py_ssize_t __pyx_v_number_of_pairs;
  PyObject *__pyx_v_price_quote_str = NULL;
  PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_buy = NULL;
  PyObject *__pyx_v_order = NULL;
  PyObject *__pyx_v_sell = NULL;
  PyObject *__pyx_9genexpr23__pyx_v_buy = NULL;
  PyObject *__pyx_gb_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25c_execute_orders_proposal_2generator1 = 0;
  PyObject *__pyx_9genexpr25__pyx_v_sell = NULL;
  PyObject *__pyx_gb_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25c_execute_orders_proposal_5generator2 = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  double __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  Py_UCS4 __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  unsigned int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  struct __pyx_opt_args_10hummingbot_8strategy_13strategy_base_12StrategyBase_c_buy_with_specific_market __pyx_t_17;
  struct __pyx_opt_args_10hummingbot_8strategy_13strategy_base_12StrategyBase_c_sell_with_specific_market __pyx_t_18;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_execute_orders_proposal", 0);
  __pyx_cur_scope = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal *)__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 1271, __pyx_L1_error)
  } else {
    __Pyx_GOTREF((PyObject *)__pyx_cur_scope);
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1273
 *     cdef c_execute_orders_proposal(self, object proposal):
 *         cdef:
 *             double expiration_seconds = NaN             # <<<<<<<<<<<<<<
 *             str bid_order_id, ask_order_id
 *             bint orders_created = False
 */
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_NaN); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __pyx_PyFloat_AsDouble(__pyx_t_1); if (unlikely((__pyx_t_2 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_expiration_seconds = __pyx_t_2;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1275
 *             double expiration_seconds = NaN
 *             str bid_order_id, ask_order_id
 *             bint orders_created = False             # <<<<<<<<<<<<<<
 *         # Number of pair of orders to track for hanging orders
 *         number_of_pairs = min((len(proposal.buys), len(proposal.sells))) if self._hanging_orders_enabled else 0
 */
  __pyx_v_orders_created = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1277
 *             bint orders_created = False
 *         # Number of pair of orders to track for hanging orders
 *         number_of_pairs = min((len(proposal.buys), len(proposal.sells))) if self._hanging_orders_enabled else 0             # <<<<<<<<<<<<<<
 * 
 *         if len(proposal.buys) > 0:
 */
  if (__pyx_v_self->_hanging_orders_enabled) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1277, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1277, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1277, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = (__pyx_t_4 < __pyx_t_5);
    if (__pyx_t_7) {
      __pyx_t_6 = __pyx_t_4;
    } else {
      __pyx_t_6 = __pyx_t_5;
    }
    __pyx_t_3 = __pyx_t_6;
  } else {
    __pyx_t_3 = 0;
  }
  __pyx_v_number_of_pairs = __pyx_t_3;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1279
 *         number_of_pairs = min((len(proposal.buys), len(proposal.sells))) if self._hanging_orders_enabled else 0
 * 
 *         if len(proposal.buys) > 0:             # <<<<<<<<<<<<<<
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1279, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = (__pyx_t_3 > 0);
  if (__pyx_t_7) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1280
 * 
 *         if len(proposal.buys) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:             # <<<<<<<<<<<<<<
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "
 *                                    f"{buy.price.normalize()} {self.quote_asset}"
 */
    __pyx_t_1 = __Pyx_PyInt_From_int64_t(__pyx_v_self->_logging_options); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_OPTION_LOG_CREATE_ORDER); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyNumber_And(__pyx_t_1, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1280, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1280, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    if (__pyx_t_7) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1281
 *         if len(proposal.buys) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "             # <<<<<<<<<<<<<<
 *                                    f"{buy.price.normalize()} {self.quote_asset}"
 *                                    for buy in proposal.buys]
 */
      { /* enter inner scope */
        __pyx_t_9 = PyList_New(0); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1281, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_9);

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1283
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "
 *                                    f"{buy.price.normalize()} {self.quote_asset}"
 *                                    for buy in proposal.buys]             # <<<<<<<<<<<<<<
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.buys)} bid orders "
 */
        __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1283, __pyx_L7_error)
        __Pyx_GOTREF(__pyx_t_8);
        if (likely(PyList_CheckExact(__pyx_t_8)) || PyTuple_CheckExact(__pyx_t_8)) {
          __pyx_t_1 = __pyx_t_8; __Pyx_INCREF(__pyx_t_1);
          __pyx_t_3 = 0;
          __pyx_t_10 = NULL;
        } else {
          __pyx_t_3 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1283, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1283, __pyx_L7_error)
        }
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        for (;;) {
          if (likely(!__pyx_t_10)) {
            if (likely(PyList_CheckExact(__pyx_t_1))) {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1283, __pyx_L7_error)
                #endif
                if (__pyx_t_3 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_8 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_8); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1283, __pyx_L7_error)
              #else
              __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1283, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_8);
              #endif
            } else {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_1);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1283, __pyx_L7_error)
                #endif
                if (__pyx_t_3 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_3); __Pyx_INCREF(__pyx_t_8); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1283, __pyx_L7_error)
              #else
              __pyx_t_8 = __Pyx_PySequence_ITEM(__pyx_t_1, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1283, __pyx_L7_error)
              __Pyx_GOTREF(__pyx_t_8);
              #endif
            }
          } else {
            __pyx_t_8 = __pyx_t_10(__pyx_t_1);
            if (unlikely(!__pyx_t_8)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 1283, __pyx_L7_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_8);
          }
          __Pyx_XDECREF_SET(__pyx_9genexpr23__pyx_v_buy, __pyx_t_8);
          __pyx_t_8 = 0;

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1281
 *         if len(proposal.buys) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "             # <<<<<<<<<<<<<<
 *                                    f"{buy.price.normalize()} {self.quote_asset}"
 *                                    for buy in proposal.buys]
 */
          __pyx_t_8 = PyTuple_New(7); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1281, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_6 = 0;
          __pyx_t_11 = 127;
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr23__pyx_v_buy, __pyx_n_s_size); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1281, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_t_13, __pyx_n_s_normalize); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1281, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __pyx_t_13 = NULL;
          __pyx_t_15 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_14))) {
            __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_14);
            if (likely(__pyx_t_13)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
              __Pyx_INCREF(__pyx_t_13);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_14, function);
              __pyx_t_15 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_13, NULL};
            __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_14, __pyx_callargs+1-__pyx_t_15, 0+__pyx_t_15);
            __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
            if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1281, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          }
          __pyx_t_14 = __Pyx_PyObject_FormatSimple(__pyx_t_12, __pyx_empty_unicode); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1281, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_14);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_14) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_14) : __pyx_t_11;
          __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_14);
          __Pyx_GIVEREF(__pyx_t_14);
          PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_14);
          __pyx_t_14 = 0;
          __Pyx_INCREF(__pyx_kp_u__12);
          __pyx_t_6 += 1;
          __Pyx_GIVEREF(__pyx_kp_u__12);
          PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_kp_u__12);
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base_asset); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1281, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_12 = __Pyx_PyObject_FormatSimple(__pyx_t_14, __pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1281, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) : __pyx_t_11;
          __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12);
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_12);
          __pyx_t_12 = 0;
          __Pyx_INCREF(__pyx_kp_u__15);
          __pyx_t_6 += 2;
          __Pyx_GIVEREF(__pyx_kp_u__15);
          PyTuple_SET_ITEM(__pyx_t_8, 3, __pyx_kp_u__15);

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1282
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "
 *                                    f"{buy.price.normalize()} {self.quote_asset}"             # <<<<<<<<<<<<<<
 *                                    for buy in proposal.buys]
 *                 self.logger().info(
 */
          __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr23__pyx_v_buy, __pyx_n_s_price); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1282, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_14);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_t_14, __pyx_n_s_normalize); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1282, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
          __pyx_t_14 = NULL;
          __pyx_t_15 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_13))) {
            __pyx_t_14 = PyMethod_GET_SELF(__pyx_t_13);
            if (likely(__pyx_t_14)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_13);
              __Pyx_INCREF(__pyx_t_14);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_13, function);
              __pyx_t_15 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_14, NULL};
            __pyx_t_12 = __Pyx_PyObject_FastCall(__pyx_t_13, __pyx_callargs+1-__pyx_t_15, 0+__pyx_t_15);
            __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1282, __pyx_L7_error)
            __Pyx_GOTREF(__pyx_t_12);
            __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          }
          __pyx_t_13 = __Pyx_PyObject_FormatSimple(__pyx_t_12, __pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1282, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_13);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_11;
          __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
          __Pyx_GIVEREF(__pyx_t_13);
          PyTuple_SET_ITEM(__pyx_t_8, 4, __pyx_t_13);
          __pyx_t_13 = 0;
          __Pyx_INCREF(__pyx_kp_u__12);
          __pyx_t_6 += 1;
          __Pyx_GIVEREF(__pyx_kp_u__12);
          PyTuple_SET_ITEM(__pyx_t_8, 5, __pyx_kp_u__12);
          __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_quote_asset); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1282, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_12 = __Pyx_PyObject_FormatSimple(__pyx_t_13, __pyx_empty_unicode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1282, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_12) : __pyx_t_11;
          __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_12);
          __Pyx_GIVEREF(__pyx_t_12);
          PyTuple_SET_ITEM(__pyx_t_8, 6, __pyx_t_12);
          __pyx_t_12 = 0;

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1281
 *         if len(proposal.buys) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "             # <<<<<<<<<<<<<<
 *                                    f"{buy.price.normalize()} {self.quote_asset}"
 *                                    for buy in proposal.buys]
 */
          __pyx_t_12 = __Pyx_PyUnicode_Join(__pyx_t_8, 7, __pyx_t_6, __pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1281, __pyx_L7_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_9, (PyObject*)__pyx_t_12))) __PYX_ERR(0, 1281, __pyx_L7_error)
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1283
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "
 *                                    f"{buy.price.normalize()} {self.quote_asset}"
 *                                    for buy in proposal.buys]             # <<<<<<<<<<<<<<
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.buys)} bid orders "
 */
        }
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_XDECREF(__pyx_9genexpr23__pyx_v_buy); __pyx_9genexpr23__pyx_v_buy = 0;
        goto __pyx_L11_exit_scope;
        __pyx_L7_error:;
        __Pyx_XDECREF(__pyx_9genexpr23__pyx_v_buy); __pyx_9genexpr23__pyx_v_buy = 0;
        goto __pyx_L1_error;
        __pyx_L11_exit_scope:;
      } /* exit inner scope */
      __pyx_v_price_quote_str = ((PyObject*)__pyx_t_9);
      __pyx_t_9 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1284
 *                                    f"{buy.price.normalize()} {self.quote_asset}"
 *                                    for buy in proposal.buys]
 *                 self.logger().info(             # <<<<<<<<<<<<<<
 *                     f"({self.trading_pair}) Creating {len(proposal.buys)} bid orders "
 *                     f"at (Size, Price): {price_quote_str}"
 */
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_8 = NULL;
      __pyx_t_15 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_8)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_8);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
          __pyx_t_15 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_8, NULL};
        __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_15, 0+__pyx_t_15);
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1284, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_info); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1284, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1285
 *                                    for buy in proposal.buys]
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.buys)} bid orders "             # <<<<<<<<<<<<<<
 *                     f"at (Size, Price): {price_quote_str}"
 *                 )
 */
      __pyx_t_1 = PyTuple_New(6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = 0;
      __pyx_t_11 = 127;
      __Pyx_INCREF(__pyx_kp_u__11);
      __pyx_t_3 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__11);
      PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_kp_u__11);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_13 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_11;
      __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
      __Pyx_GIVEREF(__pyx_t_13);
      PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_13);
      __pyx_t_13 = 0;
      __Pyx_INCREF(__pyx_kp_u_Creating);
      __pyx_t_3 += 11;
      __Pyx_GIVEREF(__pyx_kp_u_Creating);
      PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_kp_u_Creating);
      __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_6 = PyObject_Length(__pyx_t_13); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1285, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_13 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_6, 0, ' ', 'd'); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
      __Pyx_GIVEREF(__pyx_t_13);
      PyTuple_SET_ITEM(__pyx_t_1, 3, __pyx_t_13);
      __pyx_t_13 = 0;
      __Pyx_INCREF(__pyx_kp_u_bid_orders_at_Size_Price);
      __pyx_t_3 += 30;
      __Pyx_GIVEREF(__pyx_kp_u_bid_orders_at_Size_Price);
      PyTuple_SET_ITEM(__pyx_t_1, 4, __pyx_kp_u_bid_orders_at_Size_Price);

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1286
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.buys)} bid orders "
 *                     f"at (Size, Price): {price_quote_str}"             # <<<<<<<<<<<<<<
 *                 )
 *             for idx, buy in enumerate(proposal.buys):
 */
      __pyx_t_13 = __Pyx_PyObject_FormatSimple(__pyx_v_price_quote_str, __pyx_empty_unicode); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1286, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_13) : __pyx_t_11;
      __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_13);
      __Pyx_GIVEREF(__pyx_t_13);
      PyTuple_SET_ITEM(__pyx_t_1, 5, __pyx_t_13);
      __pyx_t_13 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1285
 *                                    for buy in proposal.buys]
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.buys)} bid orders "             # <<<<<<<<<<<<<<
 *                     f"at (Size, Price): {price_quote_str}"
 *                 )
 */
      __pyx_t_13 = __Pyx_PyUnicode_Join(__pyx_t_1, 6, __pyx_t_3, __pyx_t_11); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = NULL;
      __pyx_t_15 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_12))) {
        __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_12);
        if (likely(__pyx_t_1)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
          __Pyx_INCREF(__pyx_t_1);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_12, function);
          __pyx_t_15 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_t_13};
        __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_12, __pyx_callargs+1-__pyx_t_15, 1+__pyx_t_15);
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1284, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1280
 * 
 *         if len(proposal.buys) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:             # <<<<<<<<<<<<<<
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "
 *                                    f"{buy.price.normalize()} {self.quote_asset}"
 */
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1288
 *                     f"at (Size, Price): {price_quote_str}"
 *                 )
 *             for idx, buy in enumerate(proposal.buys):             # <<<<<<<<<<<<<<
 *                 bid_order_id = self.c_buy_with_specific_market(
 *                     self._market_info,
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_9 = __pyx_int_0;
    __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_buys); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1288, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);
    if (likely(PyList_CheckExact(__pyx_t_12)) || PyTuple_CheckExact(__pyx_t_12)) {
      __pyx_t_13 = __pyx_t_12; __Pyx_INCREF(__pyx_t_13);
      __pyx_t_3 = 0;
      __pyx_t_10 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_13 = PyObject_GetIter(__pyx_t_12); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_13); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1288, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
    for (;;) {
      if (likely(!__pyx_t_10)) {
        if (likely(PyList_CheckExact(__pyx_t_13))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_13);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1288, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_12 = PyList_GET_ITEM(__pyx_t_13, __pyx_t_3); __Pyx_INCREF(__pyx_t_12); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1288, __pyx_L1_error)
          #else
          __pyx_t_12 = __Pyx_PySequence_ITEM(__pyx_t_13, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1288, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_13);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1288, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_12 = PyTuple_GET_ITEM(__pyx_t_13, __pyx_t_3); __Pyx_INCREF(__pyx_t_12); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1288, __pyx_L1_error)
          #else
          __pyx_t_12 = __Pyx_PySequence_ITEM(__pyx_t_13, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1288, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          #endif
        }
      } else {
        __pyx_t_12 = __pyx_t_10(__pyx_t_13);
        if (unlikely(!__pyx_t_12)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1288, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_12);
      }
      __Pyx_XDECREF_SET(__pyx_v_buy, __pyx_t_12);
      __pyx_t_12 = 0;
      __Pyx_INCREF(__pyx_t_9);
      __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_9);
      __pyx_t_12 = __Pyx_PyInt_AddObjC(__pyx_t_9, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1288, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_9);
      __pyx_t_9 = __pyx_t_12;
      __pyx_t_12 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1290
 *             for idx, buy in enumerate(proposal.buys):
 *                 bid_order_id = self.c_buy_with_specific_market(
 *                     self._market_info,             # <<<<<<<<<<<<<<
 *                     buy.size,
 *                     order_type=self._limit_order_type,
 */
      __pyx_t_12 = __pyx_v_self->_market_info;
      __Pyx_INCREF(__pyx_t_12);

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1291
 *                 bid_order_id = self.c_buy_with_specific_market(
 *                     self._market_info,
 *                     buy.size,             # <<<<<<<<<<<<<<
 *                     order_type=self._limit_order_type,
 *                     price=buy.price,
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1291, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1292
 *                     self._market_info,
 *                     buy.size,
 *                     order_type=self._limit_order_type,             # <<<<<<<<<<<<<<
 *                     price=buy.price,
 *                     expiration_seconds=expiration_seconds
 */
      __pyx_t_8 = __pyx_v_self->_limit_order_type;
      __Pyx_INCREF(__pyx_t_8);

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1293
 *                     buy.size,
 *                     order_type=self._limit_order_type,
 *                     price=buy.price,             # <<<<<<<<<<<<<<
 *                     expiration_seconds=expiration_seconds
 *                 )
 */
      __pyx_t_14 = __Pyx_PyObject_GetAttrStr(__pyx_v_buy, __pyx_n_s_price); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_14);

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1289
 *                 )
 *             for idx, buy in enumerate(proposal.buys):
 *                 bid_order_id = self.c_buy_with_specific_market(             # <<<<<<<<<<<<<<
 *                     self._market_info,
 *                     buy.size,
 */
      __pyx_t_17.__pyx_n = 3;
      __pyx_t_17.order_type = __pyx_t_8;
      __pyx_t_17.price = __pyx_t_14;
      __pyx_t_17.expiration_seconds = __pyx_v_expiration_seconds;
      __pyx_t_16 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_buy_with_specific_market(((struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *)__pyx_v_self), __pyx_t_12, __pyx_t_1, &__pyx_t_17); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1289, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_bid_order_id);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_bid_order_id, ((PyObject*)__pyx_t_16));
      __Pyx_GIVEREF(__pyx_t_16);
      __pyx_t_16 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1296
 *                     expiration_seconds=expiration_seconds
 *                 )
 *                 orders_created = True             # <<<<<<<<<<<<<<
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))
 */
      __pyx_v_orders_created = 1;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1297
 *                 )
 *                 orders_created = True
 *                 if idx < number_of_pairs:             # <<<<<<<<<<<<<<
 *                     order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))
 *                     if order:
 */
      __pyx_t_16 = PyInt_FromSsize_t(__pyx_v_number_of_pairs); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1297, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __pyx_t_14 = PyObject_RichCompare(__pyx_v_idx, __pyx_t_16, Py_LT); __Pyx_XGOTREF(__pyx_t_14); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1297, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1297, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (__pyx_t_7) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1298
 *                 orders_created = True
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))             # <<<<<<<<<<<<<<
 *                     if order:
 *                         self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(
 */
        __pyx_t_14 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_active_orders); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1298, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_16 = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25c_execute_orders_proposal_genexpr(((PyObject*)__pyx_cur_scope), __pyx_t_14); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1298, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_14 = __Pyx_PyIter_Next(__pyx_t_16); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1298, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_XDECREF_SET(__pyx_v_order, __pyx_t_14);
        __pyx_t_14 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1299
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))
 *                     if order:             # <<<<<<<<<<<<<<
 *                         self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(
 *                             CreatedPairOfOrders(order, None))
 */
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_order); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1299, __pyx_L1_error)
        if (__pyx_t_7) {

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1300
 *                     order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))
 *                     if order:
 *                         self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(             # <<<<<<<<<<<<<<
 *                             CreatedPairOfOrders(order, None))
 *         if len(proposal.sells) > 0:
 */
          __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_n_s_add_current_pairs_of_proposal_or); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1300, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1301
 *                     if order:
 *                         self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(
 *                             CreatedPairOfOrders(order, None))             # <<<<<<<<<<<<<<
 *         if len(proposal.sells) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 */
          __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_CreatedPairOfOrders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1301, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_12 = NULL;
          __pyx_t_15 = 0;
          #if CYTHON_UNPACK_METHODS
          if (unlikely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_12)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_12);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_15 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[3] = {__pyx_t_12, __pyx_v_order, Py_None};
            __pyx_t_8 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_15, 2+__pyx_t_15);
            __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1301, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __pyx_t_1 = NULL;
          __pyx_t_15 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_16))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_16);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_16, function);
              __pyx_t_15 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_t_8};
            __pyx_t_14 = __Pyx_PyObject_FastCall(__pyx_t_16, __pyx_callargs+1-__pyx_t_15, 1+__pyx_t_15);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
            if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1300, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_14);
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          }
          __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1299
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))
 *                     if order:             # <<<<<<<<<<<<<<
 *                         self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(
 *                             CreatedPairOfOrders(order, None))
 */
        }

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1297
 *                 )
 *                 orders_created = True
 *                 if idx < number_of_pairs:             # <<<<<<<<<<<<<<
 *                     order = next((o for o in self.active_orders if o.client_order_id == bid_order_id))
 *                     if order:
 */
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1288
 *                     f"at (Size, Price): {price_quote_str}"
 *                 )
 *             for idx, buy in enumerate(proposal.buys):             # <<<<<<<<<<<<<<
 *                 bid_order_id = self.c_buy_with_specific_market(
 *                     self._market_info,
 */
    }
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1279
 *         number_of_pairs = min((len(proposal.buys), len(proposal.sells))) if self._hanging_orders_enabled else 0
 * 
 *         if len(proposal.buys) > 0:             # <<<<<<<<<<<<<<
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{buy.size.normalize()} {self.base_asset}, "
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1302
 *                         self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(
 *                             CreatedPairOfOrders(order, None))
 *         if len(proposal.sells) > 0:             # <<<<<<<<<<<<<<
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_3 = PyObject_Length(__pyx_t_9); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1302, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_7 = (__pyx_t_3 > 0);
  if (__pyx_t_7) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1303
 *                             CreatedPairOfOrders(order, None))
 *         if len(proposal.sells) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:             # <<<<<<<<<<<<<<
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "
 *                                    f"{sell.price.normalize()} {self.quote_asset}"
 */
    __pyx_t_9 = __Pyx_PyInt_From_int64_t(__pyx_v_self->_logging_options); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_OPTION_LOG_CREATE_ORDER); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_13);
    __pyx_t_14 = PyNumber_And(__pyx_t_9, __pyx_t_13); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1303, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1303, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (__pyx_t_7) {

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1304
 *         if len(proposal.sells) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "             # <<<<<<<<<<<<<<
 *                                    f"{sell.price.normalize()} {self.quote_asset}"
 *                                    for sell in proposal.sells]
 */
      { /* enter inner scope */
        __pyx_t_14 = PyList_New(0); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1304, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_14);

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1306
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "
 *                                    f"{sell.price.normalize()} {self.quote_asset}"
 *                                    for sell in proposal.sells]             # <<<<<<<<<<<<<<
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.sells)} ask "
 */
        __pyx_t_13 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1306, __pyx_L21_error)
        __Pyx_GOTREF(__pyx_t_13);
        if (likely(PyList_CheckExact(__pyx_t_13)) || PyTuple_CheckExact(__pyx_t_13)) {
          __pyx_t_9 = __pyx_t_13; __Pyx_INCREF(__pyx_t_9);
          __pyx_t_3 = 0;
          __pyx_t_10 = NULL;
        } else {
          __pyx_t_3 = -1; __pyx_t_9 = PyObject_GetIter(__pyx_t_13); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1306, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_9);
          __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1306, __pyx_L21_error)
        }
        __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        for (;;) {
          if (likely(!__pyx_t_10)) {
            if (likely(PyList_CheckExact(__pyx_t_9))) {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_9);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1306, __pyx_L21_error)
                #endif
                if (__pyx_t_3 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_13 = PyList_GET_ITEM(__pyx_t_9, __pyx_t_3); __Pyx_INCREF(__pyx_t_13); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1306, __pyx_L21_error)
              #else
              __pyx_t_13 = __Pyx_PySequence_ITEM(__pyx_t_9, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1306, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_13);
              #endif
            } else {
              {
                Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_9);
                #if !CYTHON_ASSUME_SAFE_MACROS
                if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1306, __pyx_L21_error)
                #endif
                if (__pyx_t_3 >= __pyx_temp) break;
              }
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_13 = PyTuple_GET_ITEM(__pyx_t_9, __pyx_t_3); __Pyx_INCREF(__pyx_t_13); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1306, __pyx_L21_error)
              #else
              __pyx_t_13 = __Pyx_PySequence_ITEM(__pyx_t_9, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1306, __pyx_L21_error)
              __Pyx_GOTREF(__pyx_t_13);
              #endif
            }
          } else {
            __pyx_t_13 = __pyx_t_10(__pyx_t_9);
            if (unlikely(!__pyx_t_13)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 1306, __pyx_L21_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_13);
          }
          __Pyx_XDECREF_SET(__pyx_9genexpr25__pyx_v_sell, __pyx_t_13);
          __pyx_t_13 = 0;

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1304
 *         if len(proposal.sells) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "             # <<<<<<<<<<<<<<
 *                                    f"{sell.price.normalize()} {self.quote_asset}"
 *                                    for sell in proposal.sells]
 */
          __pyx_t_13 = PyTuple_New(7); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1304, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_13);
          __pyx_t_6 = 0;
          __pyx_t_11 = 127;
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr25__pyx_v_sell, __pyx_n_s_size); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1304, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_normalize); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1304, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_8 = NULL;
          __pyx_t_15 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_1))) {
            __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
            if (likely(__pyx_t_8)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
              __Pyx_INCREF(__pyx_t_8);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_1, function);
              __pyx_t_15 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_8, NULL};
            __pyx_t_16 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+1-__pyx_t_15, 0+__pyx_t_15);
            __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
            if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1304, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_16);
            __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          }
          __pyx_t_1 = __Pyx_PyObject_FormatSimple(__pyx_t_16, __pyx_empty_unicode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1304, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_1);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_1) : __pyx_t_11;
          __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_1);
          __Pyx_GIVEREF(__pyx_t_1);
          PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_1);
          __pyx_t_1 = 0;
          __Pyx_INCREF(__pyx_kp_u__12);
          __pyx_t_6 += 1;
          __Pyx_GIVEREF(__pyx_kp_u__12);
          PyTuple_SET_ITEM(__pyx_t_13, 1, __pyx_kp_u__12);
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base_asset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1304, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_16 = __Pyx_PyObject_FormatSimple(__pyx_t_1, __pyx_empty_unicode); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1304, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_16) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_16) : __pyx_t_11;
          __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_16);
          __Pyx_GIVEREF(__pyx_t_16);
          PyTuple_SET_ITEM(__pyx_t_13, 2, __pyx_t_16);
          __pyx_t_16 = 0;
          __Pyx_INCREF(__pyx_kp_u__15);
          __pyx_t_6 += 2;
          __Pyx_GIVEREF(__pyx_kp_u__15);
          PyTuple_SET_ITEM(__pyx_t_13, 3, __pyx_kp_u__15);

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1305
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "
 *                                    f"{sell.price.normalize()} {self.quote_asset}"             # <<<<<<<<<<<<<<
 *                                    for sell in proposal.sells]
 *                 self.logger().info(
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_9genexpr25__pyx_v_sell, __pyx_n_s_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1305, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_normalize); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1305, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          __pyx_t_1 = NULL;
          __pyx_t_15 = 0;
          #if CYTHON_UNPACK_METHODS
          if (likely(PyMethod_Check(__pyx_t_8))) {
            __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_8);
            if (likely(__pyx_t_1)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
              __Pyx_INCREF(__pyx_t_1);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_8, function);
              __pyx_t_15 = 1;
            }
          }
          #endif
          {
            PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
            __pyx_t_16 = __Pyx_PyObject_FastCall(__pyx_t_8, __pyx_callargs+1-__pyx_t_15, 0+__pyx_t_15);
            __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
            if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1305, __pyx_L21_error)
            __Pyx_GOTREF(__pyx_t_16);
            __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          }
          __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_16, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1305, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_8);
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_11;
          __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_13, 4, __pyx_t_8);
          __pyx_t_8 = 0;
          __Pyx_INCREF(__pyx_kp_u__12);
          __pyx_t_6 += 1;
          __Pyx_GIVEREF(__pyx_kp_u__12);
          PyTuple_SET_ITEM(__pyx_t_13, 5, __pyx_kp_u__12);
          __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_quote_asset); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1305, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_8);
          __pyx_t_16 = __Pyx_PyObject_FormatSimple(__pyx_t_8, __pyx_empty_unicode); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1305, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_16) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_16) : __pyx_t_11;
          __pyx_t_6 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_16);
          __Pyx_GIVEREF(__pyx_t_16);
          PyTuple_SET_ITEM(__pyx_t_13, 6, __pyx_t_16);
          __pyx_t_16 = 0;

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1304
 *         if len(proposal.sells) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "             # <<<<<<<<<<<<<<
 *                                    f"{sell.price.normalize()} {self.quote_asset}"
 *                                    for sell in proposal.sells]
 */
          __pyx_t_16 = __Pyx_PyUnicode_Join(__pyx_t_13, 7, __pyx_t_6, __pyx_t_11); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1304, __pyx_L21_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
          if (unlikely(__Pyx_ListComp_Append(__pyx_t_14, (PyObject*)__pyx_t_16))) __PYX_ERR(0, 1304, __pyx_L21_error)
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1306
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "
 *                                    f"{sell.price.normalize()} {self.quote_asset}"
 *                                    for sell in proposal.sells]             # <<<<<<<<<<<<<<
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.sells)} ask "
 */
        }
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_XDECREF(__pyx_9genexpr25__pyx_v_sell); __pyx_9genexpr25__pyx_v_sell = 0;
        goto __pyx_L25_exit_scope;
        __pyx_L21_error:;
        __Pyx_XDECREF(__pyx_9genexpr25__pyx_v_sell); __pyx_9genexpr25__pyx_v_sell = 0;
        goto __pyx_L1_error;
        __pyx_L25_exit_scope:;
      } /* exit inner scope */
      __Pyx_XDECREF_SET(__pyx_v_price_quote_str, ((PyObject*)__pyx_t_14));
      __pyx_t_14 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1307
 *                                    f"{sell.price.normalize()} {self.quote_asset}"
 *                                    for sell in proposal.sells]
 *                 self.logger().info(             # <<<<<<<<<<<<<<
 *                     f"({self.trading_pair}) Creating {len(proposal.sells)} ask "
 *                     f"orders at (Size, Price): {price_quote_str}"
 */
      __pyx_t_16 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_logger); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __pyx_t_13 = NULL;
      __pyx_t_15 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_16))) {
        __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_16);
        if (likely(__pyx_t_13)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
          __Pyx_INCREF(__pyx_t_13);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_16, function);
          __pyx_t_15 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_13, NULL};
        __pyx_t_9 = __Pyx_PyObject_FastCall(__pyx_t_16, __pyx_callargs+1-__pyx_t_15, 0+__pyx_t_15);
        __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
        if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1307, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      }
      __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_info); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1308
 *                                    for sell in proposal.sells]
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.sells)} ask "             # <<<<<<<<<<<<<<
 *                     f"orders at (Size, Price): {price_quote_str}"
 *                 )
 */
      __pyx_t_9 = PyTuple_New(6); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_3 = 0;
      __pyx_t_11 = 127;
      __Pyx_INCREF(__pyx_kp_u__11);
      __pyx_t_3 += 1;
      __Pyx_GIVEREF(__pyx_kp_u__11);
      PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_kp_u__11);
      __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_trading_pair); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_13);
      __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_t_13, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_11;
      __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_INCREF(__pyx_kp_u_Creating);
      __pyx_t_3 += 11;
      __Pyx_GIVEREF(__pyx_kp_u_Creating);
      PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_kp_u_Creating);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_6 = PyObject_Length(__pyx_t_8); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 1308, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyUnicode_From_Py_ssize_t(__pyx_t_6, 0, ' ', 'd'); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_INCREF(__pyx_kp_u_ask_orders_at_Size_Price);
      __pyx_t_3 += 30;
      __Pyx_GIVEREF(__pyx_kp_u_ask_orders_at_Size_Price);
      PyTuple_SET_ITEM(__pyx_t_9, 4, __pyx_kp_u_ask_orders_at_Size_Price);

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1309
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.sells)} ask "
 *                     f"orders at (Size, Price): {price_quote_str}"             # <<<<<<<<<<<<<<
 *                 )
 *             for idx, sell in enumerate(proposal.sells):
 */
      __pyx_t_8 = __Pyx_PyObject_FormatSimple(__pyx_v_price_quote_str, __pyx_empty_unicode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1309, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_11 = (__Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) > __pyx_t_11) ? __Pyx_PyUnicode_MAX_CHAR_VALUE(__pyx_t_8) : __pyx_t_11;
      __pyx_t_3 += __Pyx_PyUnicode_GET_LENGTH(__pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_9, 5, __pyx_t_8);
      __pyx_t_8 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1308
 *                                    for sell in proposal.sells]
 *                 self.logger().info(
 *                     f"({self.trading_pair}) Creating {len(proposal.sells)} ask "             # <<<<<<<<<<<<<<
 *                     f"orders at (Size, Price): {price_quote_str}"
 *                 )
 */
      __pyx_t_8 = __Pyx_PyUnicode_Join(__pyx_t_9, 6, __pyx_t_3, __pyx_t_11); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1308, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      __pyx_t_15 = 0;
      #if CYTHON_UNPACK_METHODS
      if (likely(PyMethod_Check(__pyx_t_16))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_16);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_16, function);
          __pyx_t_15 = 1;
        }
      }
      #endif
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_9, __pyx_t_8};
        __pyx_t_14 = __Pyx_PyObject_FastCall(__pyx_t_16, __pyx_callargs+1-__pyx_t_15, 1+__pyx_t_15);
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1307, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_14);
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      }
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1303
 *                             CreatedPairOfOrders(order, None))
 *         if len(proposal.sells) > 0:
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:             # <<<<<<<<<<<<<<
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "
 *                                    f"{sell.price.normalize()} {self.quote_asset}"
 */
    }

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1311
 *                     f"orders at (Size, Price): {price_quote_str}"
 *                 )
 *             for idx, sell in enumerate(proposal.sells):             # <<<<<<<<<<<<<<
 *                 ask_order_id = self.c_sell_with_specific_market(
 *                     self._market_info,
 */
    __Pyx_INCREF(__pyx_int_0);
    __pyx_t_14 = __pyx_int_0;
    __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_proposal, __pyx_n_s_sells); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1311, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    if (likely(PyList_CheckExact(__pyx_t_16)) || PyTuple_CheckExact(__pyx_t_16)) {
      __pyx_t_8 = __pyx_t_16; __Pyx_INCREF(__pyx_t_8);
      __pyx_t_3 = 0;
      __pyx_t_10 = NULL;
    } else {
      __pyx_t_3 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_t_16); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_10 = __Pyx_PyObject_GetIterNextFunc(__pyx_t_8); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1311, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
    for (;;) {
      if (likely(!__pyx_t_10)) {
        if (likely(PyList_CheckExact(__pyx_t_8))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_8);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1311, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_16 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_3); __Pyx_INCREF(__pyx_t_16); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1311, __pyx_L1_error)
          #else
          __pyx_t_16 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1311, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          #endif
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_8);
            #if !CYTHON_ASSUME_SAFE_MACROS
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 1311, __pyx_L1_error)
            #endif
            if (__pyx_t_3 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_16 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_3); __Pyx_INCREF(__pyx_t_16); __pyx_t_3++; if (unlikely((0 < 0))) __PYX_ERR(0, 1311, __pyx_L1_error)
          #else
          __pyx_t_16 = __Pyx_PySequence_ITEM(__pyx_t_8, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1311, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          #endif
        }
      } else {
        __pyx_t_16 = __pyx_t_10(__pyx_t_8);
        if (unlikely(!__pyx_t_16)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 1311, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_16);
      }
      __Pyx_XDECREF_SET(__pyx_v_sell, __pyx_t_16);
      __pyx_t_16 = 0;
      __Pyx_INCREF(__pyx_t_14);
      __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_14);
      __pyx_t_16 = __Pyx_PyInt_AddObjC(__pyx_t_14, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 1311, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_16);
      __Pyx_DECREF(__pyx_t_14);
      __pyx_t_14 = __pyx_t_16;
      __pyx_t_16 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1313
 *             for idx, sell in enumerate(proposal.sells):
 *                 ask_order_id = self.c_sell_with_specific_market(
 *                     self._market_info,             # <<<<<<<<<<<<<<
 *                     sell.size,
 *                     order_type=self._limit_order_type,
 */
      __pyx_t_16 = __pyx_v_self->_market_info;
      __Pyx_INCREF(__pyx_t_16);

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1314
 *                 ask_order_id = self.c_sell_with_specific_market(
 *                     self._market_info,
 *                     sell.size,             # <<<<<<<<<<<<<<
 *                     order_type=self._limit_order_type,
 *                     price=sell.price,
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_sell, __pyx_n_s_size); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1314, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1315
 *                     self._market_info,
 *                     sell.size,
 *                     order_type=self._limit_order_type,             # <<<<<<<<<<<<<<
 *                     price=sell.price,
 *                     expiration_seconds=expiration_seconds
 */
      __pyx_t_13 = __pyx_v_self->_limit_order_type;
      __Pyx_INCREF(__pyx_t_13);

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1316
 *                     sell.size,
 *                     order_type=self._limit_order_type,
 *                     price=sell.price,             # <<<<<<<<<<<<<<
 *                     expiration_seconds=expiration_seconds
 *                 )
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_sell, __pyx_n_s_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1316, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1312
 *                 )
 *             for idx, sell in enumerate(proposal.sells):
 *                 ask_order_id = self.c_sell_with_specific_market(             # <<<<<<<<<<<<<<
 *                     self._market_info,
 *                     sell.size,
 */
      __pyx_t_18.__pyx_n = 3;
      __pyx_t_18.order_type = __pyx_t_13;
      __pyx_t_18.price = __pyx_t_1;
      __pyx_t_18.expiration_seconds = __pyx_v_expiration_seconds;
      __pyx_t_12 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->__pyx_base.c_sell_with_specific_market(((struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *)__pyx_v_self), __pyx_t_16, __pyx_t_9, &__pyx_t_18); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1312, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_ask_order_id);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_ask_order_id, ((PyObject*)__pyx_t_12));
      __Pyx_GIVEREF(__pyx_t_12);
      __pyx_t_12 = 0;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1319
 *                     expiration_seconds=expiration_seconds
 *                 )
 *                 orders_created = True             # <<<<<<<<<<<<<<
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))
 */
      __pyx_v_orders_created = 1;

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1320
 *                 )
 *                 orders_created = True
 *                 if idx < number_of_pairs:             # <<<<<<<<<<<<<<
 *                     order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))
 *                     if order:
 */
      __pyx_t_12 = PyInt_FromSsize_t(__pyx_v_number_of_pairs); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1320, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_1 = PyObject_RichCompare(__pyx_v_idx, __pyx_t_12, Py_LT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1320, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1320, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (__pyx_t_7) {

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1321
 *                 orders_created = True
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))             # <<<<<<<<<<<<<<
 *                     if order:
 *                         self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order
 */
        __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_active_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1321, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_12 = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25c_execute_orders_proposal_3genexpr(((PyObject*)__pyx_cur_scope), __pyx_t_1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1321, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_1 = __Pyx_PyIter_Next(__pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1321, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_XDECREF_SET(__pyx_v_order, __pyx_t_1);
        __pyx_t_1 = 0;

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1322
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))
 *                     if order:             # <<<<<<<<<<<<<<
 *                         self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order
 *         if orders_created:
 */
        __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_order); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1322, __pyx_L1_error)
        if (__pyx_t_7) {

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1323
 *                     order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))
 *                     if order:
 *                         self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order             # <<<<<<<<<<<<<<
 *         if orders_created:
 *             self.c_set_timers()
 */
          __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_hanging_orders_tracker, __pyx_n_s_current_created_pairs_of_orders); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1323, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_1);
          __pyx_t_12 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_idx); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1323, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
          if (__Pyx_PyObject_SetAttrStr(__pyx_t_12, __pyx_n_s_sell_order, __pyx_v_order) < 0) __PYX_ERR(0, 1323, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

          /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1322
 *                 if idx < number_of_pairs:
 *                     order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))
 *                     if order:             # <<<<<<<<<<<<<<
 *                         self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order
 *         if orders_created:
 */
        }

        /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1320
 *                 )
 *                 orders_created = True
 *                 if idx < number_of_pairs:             # <<<<<<<<<<<<<<
 *                     order = next((o for o in self.active_orders if o.client_order_id == ask_order_id))
 *                     if order:
 */
      }

      /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1311
 *                     f"orders at (Size, Price): {price_quote_str}"
 *                 )
 *             for idx, sell in enumerate(proposal.sells):             # <<<<<<<<<<<<<<
 *                 ask_order_id = self.c_sell_with_specific_market(
 *                     self._market_info,
 */
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1302
 *                         self._hanging_orders_tracker.add_current_pairs_of_proposal_orders_executed_by_strategy(
 *                             CreatedPairOfOrders(order, None))
 *         if len(proposal.sells) > 0:             # <<<<<<<<<<<<<<
 *             if self._logging_options & self.OPTION_LOG_CREATE_ORDER:
 *                 price_quote_str = [f"{sell.size.normalize()} {self.base_asset}, "
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1324
 *                     if order:
 *                         self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order
 *         if orders_created:             # <<<<<<<<<<<<<<
 *             self.c_set_timers()
 * 
 */
  if (__pyx_v_orders_created) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1325
 *                         self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order
 *         if orders_created:
 *             self.c_set_timers()             # <<<<<<<<<<<<<<
 * 
 *     def execute_orders_proposal(self, proposal: Proposal):
 */
    __pyx_t_14 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_set_timers(__pyx_v_self); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1324
 *                     if order:
 *                         self._hanging_orders_tracker.current_created_pairs_of_orders[idx].sell_order = order
 *         if orders_created:             # <<<<<<<<<<<<<<
 *             self.c_set_timers()
 * 
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1271
 *         return self.c_to_create_orders(proposal)
 * 
 *     cdef c_execute_orders_proposal(self, object proposal):             # <<<<<<<<<<<<<<
 *         cdef:
 *             double expiration_seconds = NaN
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_execute_orders_proposal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_price_quote_str);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_buy);
  __Pyx_XDECREF(__pyx_v_order);
  __Pyx_XDECREF(__pyx_v_sell);
  __Pyx_XDECREF(__pyx_9genexpr23__pyx_v_buy);
  __Pyx_XDECREF(__pyx_gb_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25c_execute_orders_proposal_2generator1);
  __Pyx_XDECREF(__pyx_9genexpr25__pyx_v_sell);
  __Pyx_XDECREF(__pyx_gb_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25c_execute_orders_proposal_5generator2);
  __Pyx_DECREF((PyObject *)__pyx_cur_scope);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1327
 *             self.c_set_timers()
 * 
 *     def execute_orders_proposal(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         self.c_execute_orders_proposal(proposal)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_87execute_orders_proposal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_87execute_orders_proposal = {"execute_orders_proposal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_87execute_orders_proposal, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_87execute_orders_proposal(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_proposal = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("execute_orders_proposal (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_proposal,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_proposal)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1327, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "execute_orders_proposal") < 0)) __PYX_ERR(0, 1327, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_proposal = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("execute_orders_proposal", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1327, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.execute_orders_proposal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_86execute_orders_proposal(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_proposal);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_86execute_orders_proposal(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_proposal) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("execute_orders_proposal", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1328
 * 
 *     def execute_orders_proposal(self, proposal: Proposal):
 *         self.c_execute_orders_proposal(proposal)             # <<<<<<<<<<<<<<
 * 
 *     cdef c_set_timers(self):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_execute_orders_proposal(__pyx_v_self, __pyx_v_proposal); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1328, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1327
 *             self.c_set_timers()
 * 
 *     def execute_orders_proposal(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         self.c_execute_orders_proposal(proposal)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.execute_orders_proposal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1330
 *         self.c_execute_orders_proposal(proposal)
 * 
 *     cdef c_set_timers(self):             # <<<<<<<<<<<<<<
 *         cdef double next_cycle = self._current_timestamp + self.order_refresh_time
 *         if self._create_timestamp <= self._current_timestamp:
 */

static PyObject *__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_set_timers(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  double __pyx_v_next_cycle;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  int __pyx_t_5;
  double __pyx_t_6;
  double __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("c_set_timers", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1331
 * 
 *     cdef c_set_timers(self):
 *         cdef double next_cycle = self._current_timestamp + self.order_refresh_time             # <<<<<<<<<<<<<<
 *         if self._create_timestamp <= self._current_timestamp:
 *             self._create_timestamp = next_cycle
 */
  __pyx_t_1 = PyFloat_FromDouble(__pyx_v_self->__pyx_base.__pyx_base._current_timestamp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_order_refresh_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = __pyx_PyFloat_AsDouble(__pyx_t_3); if (unlikely((__pyx_t_4 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 1331, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_next_cycle = __pyx_t_4;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1332
 *     cdef c_set_timers(self):
 *         cdef double next_cycle = self._current_timestamp + self.order_refresh_time
 *         if self._create_timestamp <= self._current_timestamp:             # <<<<<<<<<<<<<<
 *             self._create_timestamp = next_cycle
 *         if self._cancel_timestamp <= self._current_timestamp:
 */
  __pyx_t_5 = (__pyx_v_self->_create_timestamp <= __pyx_v_self->__pyx_base.__pyx_base._current_timestamp);
  if (__pyx_t_5) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1333
 *         cdef double next_cycle = self._current_timestamp + self.order_refresh_time
 *         if self._create_timestamp <= self._current_timestamp:
 *             self._create_timestamp = next_cycle             # <<<<<<<<<<<<<<
 *         if self._cancel_timestamp <= self._current_timestamp:
 *             self._cancel_timestamp = min(self._create_timestamp, next_cycle)
 */
    __pyx_v_self->_create_timestamp = __pyx_v_next_cycle;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1332
 *     cdef c_set_timers(self):
 *         cdef double next_cycle = self._current_timestamp + self.order_refresh_time
 *         if self._create_timestamp <= self._current_timestamp:             # <<<<<<<<<<<<<<
 *             self._create_timestamp = next_cycle
 *         if self._cancel_timestamp <= self._current_timestamp:
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1334
 *         if self._create_timestamp <= self._current_timestamp:
 *             self._create_timestamp = next_cycle
 *         if self._cancel_timestamp <= self._current_timestamp:             # <<<<<<<<<<<<<<
 *             self._cancel_timestamp = min(self._create_timestamp, next_cycle)
 * 
 */
  __pyx_t_5 = (__pyx_v_self->_cancel_timestamp <= __pyx_v_self->__pyx_base.__pyx_base._current_timestamp);
  if (__pyx_t_5) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1335
 *             self._create_timestamp = next_cycle
 *         if self._cancel_timestamp <= self._current_timestamp:
 *             self._cancel_timestamp = min(self._create_timestamp, next_cycle)             # <<<<<<<<<<<<<<
 * 
 *     def set_timers(self):
 */
    __pyx_t_4 = __pyx_v_next_cycle;
    __pyx_t_6 = __pyx_v_self->_create_timestamp;
    __pyx_t_5 = (__pyx_t_4 < __pyx_t_6);
    if (__pyx_t_5) {
      __pyx_t_7 = __pyx_t_4;
    } else {
      __pyx_t_7 = __pyx_t_6;
    }
    __pyx_v_self->_cancel_timestamp = __pyx_t_7;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1334
 *         if self._create_timestamp <= self._current_timestamp:
 *             self._create_timestamp = next_cycle
 *         if self._cancel_timestamp <= self._current_timestamp:             # <<<<<<<<<<<<<<
 *             self._cancel_timestamp = min(self._create_timestamp, next_cycle)
 * 
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1330
 *         self.c_execute_orders_proposal(proposal)
 * 
 *     cdef c_set_timers(self):             # <<<<<<<<<<<<<<
 *         cdef double next_cycle = self._current_timestamp + self.order_refresh_time
 *         if self._create_timestamp <= self._current_timestamp:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.c_set_timers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1337
 *             self._cancel_timestamp = min(self._create_timestamp, next_cycle)
 * 
 *     def set_timers(self):             # <<<<<<<<<<<<<<
 *         self.c_set_timers()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_89set_timers(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_89set_timers = {"set_timers", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_89set_timers, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_89set_timers(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_timers (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("set_timers", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "set_timers", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_88set_timers(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_88set_timers(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_timers", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1338
 * 
 *     def set_timers(self):
 *         self.c_set_timers()             # <<<<<<<<<<<<<<
 * 
 *     def notify_hb_app(self, msg: str):
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_set_timers(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1338, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1337
 *             self._cancel_timestamp = min(self._create_timestamp, next_cycle)
 * 
 *     def set_timers(self):             # <<<<<<<<<<<<<<
 *         self.c_set_timers()
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.set_timers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1340
 *         self.c_set_timers()
 * 
 *     def notify_hb_app(self, msg: str):             # <<<<<<<<<<<<<<
 *         if self._hb_app_notification:
 *             super().notify_hb_app(msg)
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_91notify_hb_app(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_91notify_hb_app = {"notify_hb_app", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_91notify_hb_app, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_91notify_hb_app(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_msg = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("notify_hb_app (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_msg,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_msg)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1340, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "notify_hb_app") < 0)) __PYX_ERR(0, 1340, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v_msg = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("notify_hb_app", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 1340, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.notify_hb_app", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_90notify_hb_app(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v_msg);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_90notify_hb_app(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, PyObject *__pyx_v_msg) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("notify_hb_app", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1341
 * 
 *     def notify_hb_app(self, msg: str):
 *         if self._hb_app_notification:             # <<<<<<<<<<<<<<
 *             super().notify_hb_app(msg)
 * 
 */
  if (__pyx_v_self->_hb_app_notification) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1342
 *     def notify_hb_app(self, msg: str):
 *         if self._hb_app_notification:
 *             super().notify_hb_app(msg)             # <<<<<<<<<<<<<<
 * 
 *     def dump_debug_variables(self):
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);
    __Pyx_GIVEREF((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, ((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy))) __PYX_ERR(0, 1342, __pyx_L1_error);
    __Pyx_INCREF((PyObject *)__pyx_v_self);
    __Pyx_GIVEREF((PyObject *)__pyx_v_self);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_self))) __PYX_ERR(0, 1342, __pyx_L1_error);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_super, __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_notify_hb_app); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_msg};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1342, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1341
 * 
 *     def notify_hb_app(self, msg: str):
 *         if self._hb_app_notification:             # <<<<<<<<<<<<<<
 *             super().notify_hb_app(msg)
 * 
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1340
 *         self.c_set_timers()
 * 
 *     def notify_hb_app(self, msg: str):             # <<<<<<<<<<<<<<
 *         if self._hb_app_notification:
 *             super().notify_hb_app(msg)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.notify_hb_app", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1344
 *             super().notify_hb_app(msg)
 * 
 *     def dump_debug_variables(self):             # <<<<<<<<<<<<<<
 *         market = self._market_info.market
 *         mid_price = self.get_price()
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_93dump_debug_variables(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_93dump_debug_variables = {"dump_debug_variables", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_93dump_debug_variables, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_93dump_debug_variables(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dump_debug_variables (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("dump_debug_variables", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "dump_debug_variables", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_92dump_debug_variables(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_92dump_debug_variables(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_v_market = NULL;
  PyObject *__pyx_v_mid_price = NULL;
  PyObject *__pyx_v_spread = NULL;
  PyObject *__pyx_v_best_ask = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_new_ask = NULL;
  PyObject *__pyx_v_best_bid = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_new_bid = NULL;
  PyObject *__pyx_v_vol = NULL;
  PyObject *__pyx_v_mid_price_variance = NULL;
  PyObject *__pyx_v_df_header = NULL;
  PyObject *__pyx_v_time_left_fraction = NULL;
  PyObject *__pyx_v_df = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned int __pyx_t_4;
  double __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("dump_debug_variables", 1);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1345
 * 
 *     def dump_debug_variables(self):
 *         market = self._market_info.market             # <<<<<<<<<<<<<<
 *         mid_price = self.get_price()
 *         spread = Decimal(str(self.c_get_spread()))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_market_info, __pyx_n_s_market); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_market = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1346
 *     def dump_debug_variables(self):
 *         market = self._market_info.market
 *         mid_price = self.get_price()             # <<<<<<<<<<<<<<
 *         spread = Decimal(str(self.c_get_spread()))
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1346, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1346, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_mid_price = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1347
 *         market = self._market_info.market
 *         mid_price = self.get_price()
 *         spread = Decimal(str(self.c_get_spread()))             # <<<<<<<<<<<<<<
 * 
 *         best_ask = mid_price + spread / 2
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_get_spread(__pyx_v_self); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1347, __pyx_L1_error)
  __pyx_t_3 = PyFloat_FromDouble(__pyx_t_5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = __Pyx_PyObject_Unicode(__pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1347, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_6};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1347, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_spread = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1349
 *         spread = Decimal(str(self.c_get_spread()))
 * 
 *         best_ask = mid_price + spread / 2             # <<<<<<<<<<<<<<
 *         new_ask = self._reservation_price + self._optimal_spread / 2
 *         best_bid = mid_price - spread / 2
 */
  __pyx_t_1 = __Pyx_PyInt_TrueDivideObjC(__pyx_v_spread, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Add(__pyx_v_mid_price, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_best_ask = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1350
 * 
 *         best_ask = mid_price + spread / 2
 *         new_ask = self._reservation_price + self._optimal_spread / 2             # <<<<<<<<<<<<<<
 *         best_bid = mid_price - spread / 2
 *         new_bid = self._reservation_price - self._optimal_spread / 2
 */
  __pyx_t_2 = __Pyx_PyInt_TrueDivideObjC(__pyx_v_self->_optimal_spread, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Add(__pyx_v_self->_reservation_price, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1350, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_new_ask = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1351
 *         best_ask = mid_price + spread / 2
 *         new_ask = self._reservation_price + self._optimal_spread / 2
 *         best_bid = mid_price - spread / 2             # <<<<<<<<<<<<<<
 *         new_bid = self._reservation_price - self._optimal_spread / 2
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_TrueDivideObjC(__pyx_v_spread, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyNumber_Subtract(__pyx_v_mid_price, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_best_bid = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1352
 *         new_ask = self._reservation_price + self._optimal_spread / 2
 *         best_bid = mid_price - spread / 2
 *         new_bid = self._reservation_price - self._optimal_spread / 2             # <<<<<<<<<<<<<<
 * 
 *         vol = self.get_volatility()
 */
  __pyx_t_2 = __Pyx_PyInt_TrueDivideObjC(__pyx_v_self->_optimal_spread, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Subtract(__pyx_v_self->_reservation_price, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_new_bid = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1354
 *         new_bid = self._reservation_price - self._optimal_spread / 2
 * 
 *         vol = self.get_volatility()             # <<<<<<<<<<<<<<
 *         mid_price_variance = vol ** 2
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_get_volatility); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 0+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_v_vol = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1355
 * 
 *         vol = self.get_volatility()
 *         mid_price_variance = vol ** 2             # <<<<<<<<<<<<<<
 * 
 *         if not os.path.exists(self._debug_csv_path):
 */
  __pyx_t_1 = PyNumber_Power(__pyx_v_vol, __pyx_int_2, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1355, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_mid_price_variance = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1357
 *         mid_price_variance = vol ** 2
 * 
 *         if not os.path.exists(self._debug_csv_path):             # <<<<<<<<<<<<<<
 *             df_header = pd.DataFrame([('mid_price',
 *                                        'best_bid',
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_os); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_path); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_exists); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_v_self->_debug_csv_path};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 1357, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_8 = (!__pyx_t_7);
  if (__pyx_t_8) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1358
 * 
 *         if not os.path.exists(self._debug_csv_path):
 *             df_header = pd.DataFrame([('mid_price',             # <<<<<<<<<<<<<<
 *                                        'best_bid',
 *                                        'best_ask',
 */
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pd); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_DataFrame); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_tuple__16);
    __Pyx_GIVEREF(__pyx_tuple__16);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_tuple__16)) __PYX_ERR(0, 1358, __pyx_L1_error);
    __pyx_t_3 = NULL;
    __pyx_t_4 = 0;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_4 = 1;
      }
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1358, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __pyx_v_df_header = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1379
 *                                        'mid_price_variance',
 *                                        'inventory_target_pct')])
 *             df_header.to_csv(self._debug_csv_path, mode='a', header=False, index=False)             # <<<<<<<<<<<<<<
 * 
 *         if self._execution_state.time_left is not None and self._execution_state.closing_time is not None:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_df_header, __pyx_n_s_to_csv); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_v_self->_debug_csv_path);
    __Pyx_GIVEREF(__pyx_v_self->_debug_csv_path);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_self->_debug_csv_path)) __PYX_ERR(0, 1379, __pyx_L1_error);
    __pyx_t_2 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_mode, __pyx_n_u_a) < 0) __PYX_ERR(0, 1379, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_header, Py_False) < 0) __PYX_ERR(0, 1379, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_index, Py_False) < 0) __PYX_ERR(0, 1379, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1357
 *         mid_price_variance = vol ** 2
 * 
 *         if not os.path.exists(self._debug_csv_path):             # <<<<<<<<<<<<<<
 *             df_header = pd.DataFrame([('mid_price',
 *                                        'best_bid',
 */
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1381
 *             df_header.to_csv(self._debug_csv_path, mode='a', header=False, index=False)
 * 
 *         if self._execution_state.time_left is not None and self._execution_state.closing_time is not None:             # <<<<<<<<<<<<<<
 *             time_left_fraction = self._execution_state.time_left / self._execution_state.closing_time
 *         else:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_execution_state, __pyx_n_s_time_left); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = (__pyx_t_3 != Py_None);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__pyx_t_7) {
  } else {
    __pyx_t_8 = __pyx_t_7;
    goto __pyx_L5_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_execution_state, __pyx_n_s_closing_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_7 = (__pyx_t_3 != Py_None);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_8 = __pyx_t_7;
  __pyx_L5_bool_binop_done:;
  if (__pyx_t_8) {

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1382
 * 
 *         if self._execution_state.time_left is not None and self._execution_state.closing_time is not None:
 *             time_left_fraction = self._execution_state.time_left / self._execution_state.closing_time             # <<<<<<<<<<<<<<
 *         else:
 *             time_left_fraction = None
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_execution_state, __pyx_n_s_time_left); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_execution_state, __pyx_n_s_closing_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = __Pyx_PyNumber_Divide(__pyx_t_3, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_time_left_fraction = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1381
 *             df_header.to_csv(self._debug_csv_path, mode='a', header=False, index=False)
 * 
 *         if self._execution_state.time_left is not None and self._execution_state.closing_time is not None:             # <<<<<<<<<<<<<<
 *             time_left_fraction = self._execution_state.time_left / self._execution_state.closing_time
 *         else:
 */
    goto __pyx_L4;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1384
 *             time_left_fraction = self._execution_state.time_left / self._execution_state.closing_time
 *         else:
 *             time_left_fraction = None             # <<<<<<<<<<<<<<
 * 
 *         df = pd.DataFrame([(mid_price,
 */
  /*else*/ {
    __Pyx_INCREF(Py_None);
    __pyx_v_time_left_fraction = Py_None;
  }
  __pyx_L4:;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1386
 *             time_left_fraction = None
 * 
 *         df = pd.DataFrame([(mid_price,             # <<<<<<<<<<<<<<
 *                             best_bid,
 *                             best_ask,
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_pd); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_DataFrame); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1393
 *                             self._optimal_bid,
 *                             self._optimal_ask,
 *                             (mid_price - (self._reservation_price - self._optimal_spread / 2)) / mid_price,             # <<<<<<<<<<<<<<
 *                             ((self._reservation_price + self._optimal_spread / 2) - mid_price) / mid_price,
 *                             market.get_balance(self.base_asset),
 */
  __pyx_t_2 = __Pyx_PyInt_TrueDivideObjC(__pyx_v_self->_optimal_spread, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyNumber_Subtract(__pyx_v_self->_reservation_price, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Subtract(__pyx_v_mid_price, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyNumber_Divide(__pyx_t_2, __pyx_v_mid_price); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1393, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1394
 *                             self._optimal_ask,
 *                             (mid_price - (self._reservation_price - self._optimal_spread / 2)) / mid_price,
 *                             ((self._reservation_price + self._optimal_spread / 2) - mid_price) / mid_price,             # <<<<<<<<<<<<<<
 *                             market.get_balance(self.base_asset),
 *                             self.c_calculate_target_inventory(),
 */
  __pyx_t_2 = __Pyx_PyInt_TrueDivideObjC(__pyx_v_self->_optimal_spread, __pyx_int_2, 2, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_9 = PyNumber_Add(__pyx_v_self->_reservation_price, __pyx_t_2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_Subtract(__pyx_t_9, __pyx_v_mid_price); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_PyNumber_Divide(__pyx_t_2, __pyx_v_mid_price); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1395
 *                             (mid_price - (self._reservation_price - self._optimal_spread / 2)) / mid_price,
 *                             ((self._reservation_price + self._optimal_spread / 2) - mid_price) / mid_price,
 *                             market.get_balance(self.base_asset),             # <<<<<<<<<<<<<<
 *                             self.c_calculate_target_inventory(),
 *                             time_left_fraction,
 */
  __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_market, __pyx_n_s_get_balance); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base_asset); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1395, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (likely(PyMethod_Check(__pyx_t_10))) {
    __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_10);
    if (likely(__pyx_t_12)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
      __Pyx_INCREF(__pyx_t_12);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_10, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_12, __pyx_t_11};
    __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  }

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1396
 *                             ((self._reservation_price + self._optimal_spread / 2) - mid_price) / mid_price,
 *                             market.get_balance(self.base_asset),
 *                             self.c_calculate_target_inventory(),             # <<<<<<<<<<<<<<
 *                             time_left_fraction,
 *                             self._avg_vol.current_value,
 */
  __pyx_t_10 = ((struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab)->c_calculate_target_inventory(__pyx_v_self); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1396, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1398
 *                             self.c_calculate_target_inventory(),
 *                             time_left_fraction,
 *                             self._avg_vol.current_value,             # <<<<<<<<<<<<<<
 *                             self.gamma,
 *                             self._alpha,
 */
  __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_avg_vol, __pyx_n_s_current_value); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 1398, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1399
 *                             time_left_fraction,
 *                             self._avg_vol.current_value,
 *                             self.gamma,             # <<<<<<<<<<<<<<
 *                             self._alpha,
 *                             self._kappa,
 */
  __pyx_t_12 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_gamma); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 1399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1402
 *                             self._alpha,
 *                             self._kappa,
 *                             self.eta,             # <<<<<<<<<<<<<<
 *                             vol,
 *                             mid_price_variance,
 */
  __pyx_t_13 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_eta); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 1402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1405
 *                             vol,
 *                             mid_price_variance,
 *                             self.inventory_target_base_pct)])             # <<<<<<<<<<<<<<
 *         df.to_csv(self._debug_csv_path, mode='a', header=False, index=False)
 */
  __pyx_t_14 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_inventory_target_base_pct); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1386
 *             time_left_fraction = None
 * 
 *         df = pd.DataFrame([(mid_price,             # <<<<<<<<<<<<<<
 *                             best_bid,
 *                             best_ask,
 */
  __pyx_t_15 = PyTuple_New(20); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_INCREF(__pyx_v_mid_price);
  __Pyx_GIVEREF(__pyx_v_mid_price);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_v_mid_price)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_best_bid);
  __Pyx_GIVEREF(__pyx_v_best_bid);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_v_best_bid)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_best_ask);
  __Pyx_GIVEREF(__pyx_v_best_ask);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_best_ask)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_reservation_price);
  __Pyx_GIVEREF(__pyx_v_self->_reservation_price);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 3, __pyx_v_self->_reservation_price)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_optimal_spread);
  __Pyx_GIVEREF(__pyx_v_self->_optimal_spread);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 4, __pyx_v_self->_optimal_spread)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_optimal_bid);
  __Pyx_GIVEREF(__pyx_v_self->_optimal_bid);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 5, __pyx_v_self->_optimal_bid)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_optimal_ask);
  __Pyx_GIVEREF(__pyx_v_self->_optimal_ask);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 6, __pyx_v_self->_optimal_ask)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 7, __pyx_t_1)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_9);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 8, __pyx_t_9)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 9, __pyx_t_2)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_10);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 10, __pyx_t_10)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_time_left_fraction);
  __Pyx_GIVEREF(__pyx_v_time_left_fraction);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 11, __pyx_v_time_left_fraction)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_11);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 12, __pyx_t_11)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_12);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 13, __pyx_t_12)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_alpha);
  __Pyx_GIVEREF(__pyx_v_self->_alpha);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 14, __pyx_v_self->_alpha)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_kappa);
  __Pyx_GIVEREF(__pyx_v_self->_kappa);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 15, __pyx_v_self->_kappa)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_13);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 16, __pyx_t_13)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_vol);
  __Pyx_GIVEREF(__pyx_v_vol);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 17, __pyx_v_vol)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_mid_price_variance);
  __Pyx_GIVEREF(__pyx_v_mid_price_variance);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 18, __pyx_v_mid_price_variance)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_14);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_15, 19, __pyx_t_14)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_9 = 0;
  __pyx_t_2 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_t_14 = PyList_New(1); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1386, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __Pyx_GIVEREF(__pyx_t_15);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_14, 0, __pyx_t_15)) __PYX_ERR(0, 1386, __pyx_L1_error);
  __pyx_t_15 = 0;
  __pyx_t_15 = NULL;
  __pyx_t_4 = 0;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_15)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_15);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_4 = 1;
    }
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_15, __pyx_t_14};
    __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+1-__pyx_t_4, 1+__pyx_t_4);
    __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1386, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_v_df = __pyx_t_6;
  __pyx_t_6 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1406
 *                             mid_price_variance,
 *                             self.inventory_target_base_pct)])
 *         df.to_csv(self._debug_csv_path, mode='a', header=False, index=False)             # <<<<<<<<<<<<<<
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_df, __pyx_n_s_to_csv); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_self->_debug_csv_path);
  __Pyx_GIVEREF(__pyx_v_self->_debug_csv_path);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_self->_debug_csv_path)) __PYX_ERR(0, 1406, __pyx_L1_error);
  __pyx_t_14 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_mode, __pyx_n_u_a) < 0) __PYX_ERR(0, 1406, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_header, Py_False) < 0) __PYX_ERR(0, 1406, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_14, __pyx_n_s_index, Py_False) < 0) __PYX_ERR(0, 1406, __pyx_L1_error)
  __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_3, __pyx_t_14); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 1406, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1344
 *             super().notify_hb_app(msg)
 * 
 *     def dump_debug_variables(self):             # <<<<<<<<<<<<<<
 *         market = self._market_info.market
 *         mid_price = self.get_price()
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.dump_debug_variables", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_market);
  __Pyx_XDECREF(__pyx_v_mid_price);
  __Pyx_XDECREF(__pyx_v_spread);
  __Pyx_XDECREF(__pyx_v_best_ask);
  __Pyx_XDECREF(__pyx_v_new_ask);
  __Pyx_XDECREF(__pyx_v_best_bid);
  __Pyx_XDECREF(__pyx_v_new_bid);
  __Pyx_XDECREF(__pyx_v_vol);
  __Pyx_XDECREF(__pyx_v_mid_price_variance);
  __Pyx_XDECREF(__pyx_v_df_header);
  __Pyx_XDECREF(__pyx_v_time_left_fraction);
  __Pyx_XDECREF(__pyx_v_df);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_95__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_95__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_95__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_95__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) {
    __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL;}
  if (unlikely(__pyx_kwds) && __Pyx_NumKwargs_FASTCALL(__pyx_kwds) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "__reduce_cython__", 0))) return NULL;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_94__reduce_cython__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_94__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 1);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self__events_cannot_be_converted, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 */

/* Python wrapper */
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_97__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_97__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_97__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_97__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_MACROS
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_state,0};
    if (__pyx_kwds) {
      Py_ssize_t kw_args;
      switch (__pyx_nargs) {
        case  1: values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = __Pyx_NumKwargs_FASTCALL(__pyx_kwds);
      switch (__pyx_nargs) {
        case  0:
        if (likely((values[0] = __Pyx_GetKwValue_FASTCALL(__pyx_kwds, __pyx_kwvalues, __pyx_n_s_pyx_state)) != 0)) {
          (void)__Pyx_Arg_NewRef_FASTCALL(values[0]);
          kw_args--;
        }
        else if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 3, __pyx_L3_error)
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        const Py_ssize_t kwd_pos_args = __pyx_nargs;
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values + 0, kwd_pos_args, "__setstate_cython__") < 0)) __PYX_ERR(1, 3, __pyx_L3_error)
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_Arg_FASTCALL(__pyx_args, 0);
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_96__setstate_cython__(((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  {
    Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
      __Pyx_Arg_XDECREF_FASTCALL(values[__pyx_temp]);
    }
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_96__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 1);

  /* "(tree fragment)":4
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 */
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_kp_s_self__events_cannot_be_converted, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy;

static PyObject *__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *p;
  PyObject *o = __Pyx_PyType_GetSlot(__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase, tp_new, newfunc)(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)o);
  p->__pyx_base.__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub*)__pyx_vtabptr_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy;
  p->_config_map = Py_None; Py_INCREF(Py_None);
  p->_market_info = Py_None; Py_INCREF(Py_None);
  p->_price_delegate = Py_None; Py_INCREF(Py_None);
  p->_minimum_spread = Py_None; Py_INCREF(Py_None);
  p->_hanging_orders_cancel_pct = Py_None; Py_INCREF(Py_None);
  p->_hanging_orders_tracker = Py_None; Py_INCREF(Py_None);
  p->_limit_order_type = Py_None; Py_INCREF(Py_None);
  p->_last_own_trade_price = Py_None; Py_INCREF(Py_None);
  p->_alpha = Py_None; Py_INCREF(Py_None);
  p->_kappa = Py_None; Py_INCREF(Py_None);
  p->_gamma = Py_None; Py_INCREF(Py_None);
  p->_eta = Py_None; Py_INCREF(Py_None);
  p->_execution_mode = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_execution_timeframe = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_execution_state = Py_None; Py_INCREF(Py_None);
  p->_start_time = Py_None; Py_INCREF(Py_None);
  p->_end_time = Py_None; Py_INCREF(Py_None);
  p->_q_adjustment_factor = Py_None; Py_INCREF(Py_None);
  p->_reservation_price = Py_None; Py_INCREF(Py_None);
  p->_optimal_spread = Py_None; Py_INCREF(Py_None);
  p->_optimal_bid = Py_None; Py_INCREF(Py_None);
  p->_optimal_ask = Py_None; Py_INCREF(Py_None);
  p->_debug_csv_path = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_avg_vol = Py_None; Py_INCREF(Py_None);
  p->_trading_intensity = ((struct __pyx_obj_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator *)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy(PyObject *o) {
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *p = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_config_map);
  Py_CLEAR(p->_market_info);
  Py_CLEAR(p->_price_delegate);
  Py_CLEAR(p->_minimum_spread);
  Py_CLEAR(p->_hanging_orders_cancel_pct);
  Py_CLEAR(p->_hanging_orders_tracker);
  Py_CLEAR(p->_limit_order_type);
  Py_CLEAR(p->_last_own_trade_price);
  Py_CLEAR(p->_alpha);
  Py_CLEAR(p->_kappa);
  Py_CLEAR(p->_gamma);
  Py_CLEAR(p->_eta);
  Py_CLEAR(p->_execution_mode);
  Py_CLEAR(p->_execution_timeframe);
  Py_CLEAR(p->_execution_state);
  Py_CLEAR(p->_start_time);
  Py_CLEAR(p->_end_time);
  Py_CLEAR(p->_q_adjustment_factor);
  Py_CLEAR(p->_reservation_price);
  Py_CLEAR(p->_optimal_spread);
  Py_CLEAR(p->_optimal_bid);
  Py_CLEAR(p->_optimal_ask);
  Py_CLEAR(p->_debug_csv_path);
  Py_CLEAR(p->_avg_vol);
  Py_CLEAR(p->_trading_intensity);
  #if PY_MAJOR_VERSION < 3
  if (!(__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase) || PyType_IS_GC(__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase)) PyObject_GC_Track(o);
  #else
  if (PyType_IS_GC(__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase)) PyObject_GC_Track(o);
  #endif
  if (likely(__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase)) __Pyx_PyType_GetSlot(__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase, tp_dealloc, destructor)(o); else __Pyx_call_next_tp_dealloc(o, __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);
}

static int __pyx_tp_traverse_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *p = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)o;
  e = ((likely(__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase)) ? ((__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase->tp_traverse) ? __pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase->tp_traverse(o, v, a) : 0) : __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy)); if (e) return e;
  if (p->_config_map) {
    e = (*v)(p->_config_map, a); if (e) return e;
  }
  if (p->_market_info) {
    e = (*v)(p->_market_info, a); if (e) return e;
  }
  if (p->_price_delegate) {
    e = (*v)(p->_price_delegate, a); if (e) return e;
  }
  if (p->_minimum_spread) {
    e = (*v)(p->_minimum_spread, a); if (e) return e;
  }
  if (p->_hanging_orders_cancel_pct) {
    e = (*v)(p->_hanging_orders_cancel_pct, a); if (e) return e;
  }
  if (p->_hanging_orders_tracker) {
    e = (*v)(p->_hanging_orders_tracker, a); if (e) return e;
  }
  if (p->_limit_order_type) {
    e = (*v)(p->_limit_order_type, a); if (e) return e;
  }
  if (p->_last_own_trade_price) {
    e = (*v)(p->_last_own_trade_price, a); if (e) return e;
  }
  if (p->_alpha) {
    e = (*v)(p->_alpha, a); if (e) return e;
  }
  if (p->_kappa) {
    e = (*v)(p->_kappa, a); if (e) return e;
  }
  if (p->_gamma) {
    e = (*v)(p->_gamma, a); if (e) return e;
  }
  if (p->_eta) {
    e = (*v)(p->_eta, a); if (e) return e;
  }
  if (p->_execution_state) {
    e = (*v)(p->_execution_state, a); if (e) return e;
  }
  if (p->_start_time) {
    e = (*v)(p->_start_time, a); if (e) return e;
  }
  if (p->_end_time) {
    e = (*v)(p->_end_time, a); if (e) return e;
  }
  if (p->_q_adjustment_factor) {
    e = (*v)(p->_q_adjustment_factor, a); if (e) return e;
  }
  if (p->_reservation_price) {
    e = (*v)(p->_reservation_price, a); if (e) return e;
  }
  if (p->_optimal_spread) {
    e = (*v)(p->_optimal_spread, a); if (e) return e;
  }
  if (p->_optimal_bid) {
    e = (*v)(p->_optimal_bid, a); if (e) return e;
  }
  if (p->_optimal_ask) {
    e = (*v)(p->_optimal_ask, a); if (e) return e;
  }
  if (p->_avg_vol) {
    e = (*v)(p->_avg_vol, a); if (e) return e;
  }
  if (p->_trading_intensity) {
    e = (*v)(((PyObject *)p->_trading_intensity), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *p = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *)o;
  if (likely(__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase)) { if (__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase->tp_clear) __pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase->tp_clear(o); } else __Pyx_call_next_tp_clear(o, __pyx_tp_clear_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);
  tmp = ((PyObject*)p->_config_map);
  p->_config_map = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_market_info);
  p->_market_info = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_price_delegate);
  p->_price_delegate = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_minimum_spread);
  p->_minimum_spread = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_hanging_orders_cancel_pct);
  p->_hanging_orders_cancel_pct = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_hanging_orders_tracker);
  p->_hanging_orders_tracker = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_limit_order_type);
  p->_limit_order_type = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_last_own_trade_price);
  p->_last_own_trade_price = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_alpha);
  p->_alpha = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_kappa);
  p->_kappa = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_gamma);
  p->_gamma = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_eta);
  p->_eta = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_execution_state);
  p->_execution_state = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_start_time);
  p->_start_time = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_end_time);
  p->_end_time = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_q_adjustment_factor);
  p->_q_adjustment_factor = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_reservation_price);
  p->_reservation_price = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_optimal_spread);
  p->_optimal_spread = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_optimal_bid);
  p->_optimal_bid = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_optimal_ask);
  p->_optimal_ask = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_avg_vol);
  p->_avg_vol = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_trading_intensity);
  p->_trading_intensity = ((struct __pyx_obj_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_min_spread(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10min_spread_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_avg_vol(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7avg_vol_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_avg_vol(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7avg_vol_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_trading_intensity(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17trading_intensity_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_trading_intensity(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17trading_intensity_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_market_info(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11market_info_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_order_refresh_tolerance_pct(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_27order_refresh_tolerance_pct_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_order_refresh_tolerance(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_23order_refresh_tolerance_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_order_amount(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12order_amount_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_inventory_target_base_pct(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25inventory_target_base_pct_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_inventory_target_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_21inventory_target_base_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_inventory_target_base(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_21inventory_target_base_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_order_optimization_enabled(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_26order_optimization_enabled_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_order_refresh_time(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_18order_refresh_time_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_filled_order_delay(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_18filled_order_delay_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_order_override(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_14order_override_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_order_levels(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12order_levels_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_level_distances(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15level_distances_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_max_order_age(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_13max_order_age_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_add_transaction_costs_to_orders(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_31add_transaction_costs_to_orders_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_base_asset(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10base_asset_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_quote_asset(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11quote_asset_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_trading_pair(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12trading_pair_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_gamma(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5gamma_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_alpha(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5alpha_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_alpha(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5alpha_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_kappa(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5kappa_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_kappa(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5kappa_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_eta(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_3eta_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_reservation_price(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17reservation_price_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_reservation_price(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17reservation_price_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_optimal_spread(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_14optimal_spread_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_optimal_ask(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_ask_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_optimal_ask(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_ask_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_optimal_bid(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_bid_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_optimal_bid(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11optimal_bid_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_execution_timeframe(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_19execution_timeframe_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_start_time(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10start_time_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_start_time(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_10start_time_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_end_time(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_8end_time_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_end_time(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_8end_time_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_market_info_to_active_orders(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_28market_info_to_active_orders_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_active_orders(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_13active_orders_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_active_non_hanging_orders(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25active_non_hanging_orders_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_active_buys(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11active_buys_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_active_sells(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_12active_sells_1__get__(o);
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_logging_options(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15logging_options_1__get__(o);
}

static int __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_logging_options(PyObject *o, PyObject *v, CYTHON_UNUSED void *x) {
  if (v) {
    return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15logging_options_3__set__(o, v);
  }
  else {
    PyErr_SetString(PyExc_NotImplementedError, "__del__");
    return -1;
  }
}

static PyObject *__pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_hanging_orders_tracker(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_22hanging_orders_tracker_1__get__(o);
}

static PyMethodDef __pyx_methods_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy[] = {
  {"logger", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_1logger, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"init_params", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_3init_params, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"all_markets_ready", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5all_markets_ready, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_price", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7get_price, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_last_price", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_9get_last_price, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_mid_price", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11get_mid_price, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"update_from_config_map", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_13update_from_config_map, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_config_map_execution_mode", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15get_config_map_execution_mode, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_config_map_hanging_orders", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17get_config_map_hanging_orders, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_config_map_indicators", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_19get_config_map_indicators, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"pure_mm_assets_df", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_21pure_mm_assets_df, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"active_orders_df", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_23active_orders_df, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"market_status_data_frame", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25market_status_data_frame, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"format_status", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_27format_status, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"execute_orders_proposal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_29execute_orders_proposal, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"cancel_order", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_31cancel_order, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"start", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_33start, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"process_tick", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_35process_tick, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"collect_market_variables", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_37collect_market_variables, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_spread", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_39get_spread, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_volatility", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_41get_volatility, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"measure_order_book_liquidity", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_43measure_order_book_liquidity, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"calculate_reservation_price_and_optimal_spread", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_45calculate_reservation_price_and_optimal_spread, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"calculate_target_inventory", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_47calculate_target_inventory, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"calculate_inventory", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_49calculate_inventory, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"is_algorithm_ready", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_51is_algorithm_ready, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"is_algorithm_changed", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_53is_algorithm_changed, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"_get_level_spreads", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_55_get_level_spreads, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"create_proposal_based_on_order_override", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_57create_proposal_based_on_order_override, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"create_proposal_based_on_order_levels", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_59create_proposal_based_on_order_levels, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"create_basic_proposal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_61create_basic_proposal, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"create_base_proposal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_63create_base_proposal, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"get_adjusted_available_balance", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_65get_adjusted_available_balance, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"apply_order_price_modifiers", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_67apply_order_price_modifiers, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"apply_budget_constraint", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_69apply_budget_constraint, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"adjusted_available_balance_for_orders_budget_constrain", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_71adjusted_available_balance_for_orders_budget_constrain, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"apply_budget_constraint", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_73apply_budget_constraint, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"apply_order_optimization", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_75apply_order_optimization, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"apply_order_amount_eta_transformation", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_77apply_order_amount_eta_transformation, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"apply_add_transaction_costs", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_79apply_add_transaction_costs, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"is_within_tolerance", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_81is_within_tolerance, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"cancel_active_orders", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_83cancel_active_orders, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"to_create_orders", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_85to_create_orders, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"execute_orders_proposal", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_87execute_orders_proposal, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"set_timers", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_89set_timers, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"notify_hb_app", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_91notify_hb_app, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"dump_debug_variables", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_93dump_debug_variables, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__reduce_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_95__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void*)(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_97__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy[] = {
  {(char *)"min_spread", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_min_spread, 0, (char *)0, 0},
  {(char *)"avg_vol", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_avg_vol, __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_avg_vol, (char *)0, 0},
  {(char *)"trading_intensity", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_trading_intensity, __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_trading_intensity, (char *)0, 0},
  {(char *)"market_info", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_market_info, 0, (char *)0, 0},
  {(char *)"order_refresh_tolerance_pct", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_order_refresh_tolerance_pct, 0, (char *)0, 0},
  {(char *)"order_refresh_tolerance", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_order_refresh_tolerance, 0, (char *)0, 0},
  {(char *)"order_amount", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_order_amount, 0, (char *)0, 0},
  {(char *)"inventory_target_base_pct", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_inventory_target_base_pct, 0, (char *)0, 0},
  {(char *)"inventory_target_base", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_inventory_target_base, __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_inventory_target_base, (char *)0, 0},
  {(char *)"order_optimization_enabled", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_order_optimization_enabled, 0, (char *)0, 0},
  {(char *)"order_refresh_time", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_order_refresh_time, 0, (char *)0, 0},
  {(char *)"filled_order_delay", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_filled_order_delay, 0, (char *)0, 0},
  {(char *)"order_override", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_order_override, 0, (char *)0, 0},
  {(char *)"order_levels", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_order_levels, 0, (char *)0, 0},
  {(char *)"level_distances", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_level_distances, 0, (char *)0, 0},
  {(char *)"max_order_age", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_max_order_age, 0, (char *)0, 0},
  {(char *)"add_transaction_costs_to_orders", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_add_transaction_costs_to_orders, 0, (char *)0, 0},
  {(char *)"base_asset", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_base_asset, 0, (char *)0, 0},
  {(char *)"quote_asset", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_quote_asset, 0, (char *)0, 0},
  {(char *)"trading_pair", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_trading_pair, 0, (char *)0, 0},
  {(char *)"gamma", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_gamma, 0, (char *)0, 0},
  {(char *)"alpha", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_alpha, __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_alpha, (char *)0, 0},
  {(char *)"kappa", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_kappa, __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_kappa, (char *)0, 0},
  {(char *)"eta", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_eta, 0, (char *)0, 0},
  {(char *)"reservation_price", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_reservation_price, __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_reservation_price, (char *)0, 0},
  {(char *)"optimal_spread", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_optimal_spread, 0, (char *)0, 0},
  {(char *)"optimal_ask", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_optimal_ask, __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_optimal_ask, (char *)0, 0},
  {(char *)"optimal_bid", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_optimal_bid, __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_optimal_bid, (char *)0, 0},
  {(char *)"execution_timeframe", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_execution_timeframe, 0, (char *)0, 0},
  {(char *)"start_time", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_start_time, __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_start_time, (char *)0, 0},
  {(char *)"end_time", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_end_time, __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_end_time, (char *)0, 0},
  {(char *)"market_info_to_active_orders", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_market_info_to_active_orders, 0, (char *)0, 0},
  {(char *)"active_orders", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_active_orders, 0, (char *)0, 0},
  {(char *)"active_non_hanging_orders", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_active_non_hanging_orders, 0, (char *)0, 0},
  {(char *)"active_buys", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_active_buys, 0, (char *)0, 0},
  {(char *)"active_sells", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_active_sells, 0, (char *)0, 0},
  {(char *)"logging_options", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_logging_options, __pyx_setprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_logging_options, (char *)0, 0},
  {(char *)"hanging_orders_tracker", __pyx_getprop_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_hanging_orders_tracker, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy},
  {Py_tp_clear, (void *)__pyx_tp_clear_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy},
  {Py_tp_methods, (void *)__pyx_methods_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy},
  {Py_tp_getset, (void *)__pyx_getsets_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy_spec = {
  "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.AvellanedaMarketMakingStrategy",
  sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.""AvellanedaMarketMakingStrategy", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, /*tp_traverse*/
  __pyx_tp_clear_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *__pyx_freelist_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start[8];
static int __pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start = 0;
#endif

static PyObject *__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start)))) {
    o = (PyObject*)__pyx_freelist_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start[--__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start];
    memset(o, 0, sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start(PyObject *o) {
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *p = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_order_id);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start)))) {
    __pyx_freelist_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start[__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start++] = ((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *p = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *)o;
  if (p->__pyx_v_order_id) {
    e = (*v)(p->__pyx_v_order_id, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *p = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start *)o;
  tmp = ((PyObject*)p->__pyx_v_order_id);
  p->__pyx_v_order_id = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start},
  {Py_tp_clear, (void *)__pyx_tp_clear_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start_spec = {
  "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.__pyx_scope_struct____pyx_base____pyx_base__c_start",
  sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.""__pyx_scope_struct____pyx_base____pyx_base__c_start", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start, /*tp_traverse*/
  __pyx_tp_clear_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr *__pyx_freelist_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr = 0;
#endif

static PyObject *__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr[--__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_o);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr[__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_o) {
    e = (*v)(p->__pyx_v_o, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr_spec = {
  "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.__pyx_scope_struct_1_genexpr",
  sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.""__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal *__pyx_freelist_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal[8];
static int __pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal = 0;
#endif

static PyObject *__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal)))) {
    o = (PyObject*)__pyx_freelist_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal[--__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal];
    memset(o, 0, sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal));
    (void) PyObject_INIT(o, t);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal(PyObject *o) {
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal *p = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  Py_CLEAR(p->__pyx_v_ask_order_id);
  Py_CLEAR(p->__pyx_v_bid_order_id);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal)))) {
    __pyx_freelist_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal[__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal++] = ((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal_spec = {
  "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.__pyx_scope_struct_2_c_execute_orders_proposal",
  sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.""__pyx_scope_struct_2_c_execute_orders_proposal", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr *__pyx_freelist_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr[8];
static int __pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr = 0;
#endif

static PyObject *__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr)))) {
    o = (PyObject*)__pyx_freelist_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr[--__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr(PyObject *o) {
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_o);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr)))) {
    __pyx_freelist_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr[__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr++] = ((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr *p = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_o) {
    e = (*v)(p->__pyx_v_o, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr_spec = {
  "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.__pyx_scope_struct_3_genexpr",
  sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.""__pyx_scope_struct_3_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

#if CYTHON_USE_FREELISTS
static struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr *__pyx_freelist_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr[8];
static int __pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr = 0;
#endif

static PyObject *__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  #if CYTHON_USE_FREELISTS
  if (likely((int)(__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr > 0) & (int)(t->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr)))) {
    o = (PyObject*)__pyx_freelist_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr[--__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else
  #endif
  {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  #endif
  return o;
}

static void __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr(PyObject *o) {
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_genexpr_arg_0);
  Py_CLEAR(p->__pyx_v_o);
  Py_CLEAR(p->__pyx_t_0);
  #if CYTHON_USE_FREELISTS
  if (((int)(__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr < 8) & (int)(Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr)))) {
    __pyx_freelist_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr[__pyx_freecount_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr++] = ((struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr *)o);
  } else
  #endif
  {
    #if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
    (*Py_TYPE(o)->tp_free)(o);
    #else
    {
      freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
      if (tp_free) tp_free(o);
    }
    #endif
  }
}

static int __pyx_tp_traverse_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject *)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_genexpr_arg_0) {
    e = (*v)(p->__pyx_genexpr_arg_0, a); if (e) return e;
  }
  if (p->__pyx_v_o) {
    e = (*v)(p->__pyx_v_o, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr},
  {Py_tp_new, (void *)__pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr},
  {0, 0},
};
static PyType_Spec __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr_spec = {
  "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.__pyx_scope_struct_4_genexpr",
  sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE,
  __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr_slots,
};
#else

static PyTypeObject __pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making.""__pyx_scope_struct_4_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC|Py_TPFLAGS_HAVE_FINALIZE, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif
/* #### Code section: pystring_table ### */

static int __Pyx_CreateStringTabAndInitStrings(void) {
  __Pyx_StringTabEntry __pyx_string_tab[] = {
    {&__pyx_kp_u_0, __pyx_k_0, sizeof(__pyx_k_0), 0, 1, 0, 0},
    {&__pyx_kp_u_1, __pyx_k_1, sizeof(__pyx_k_1), 0, 1, 0, 0},
    {&__pyx_kp_u_10, __pyx_k_10, sizeof(__pyx_k_10), 0, 1, 0, 0},
    {&__pyx_kp_u_100, __pyx_k_100, sizeof(__pyx_k_100), 0, 1, 0, 0},
    {&__pyx_kp_u_10f, __pyx_k_10f, sizeof(__pyx_k_10f), 0, 1, 0, 0},
    {&__pyx_kp_u_1_2, __pyx_k_1_2, sizeof(__pyx_k_1_2), 0, 1, 0, 0},
    {&__pyx_kp_u_2, __pyx_k_2, sizeof(__pyx_k_2), 0, 1, 0, 0},
    {&__pyx_kp_u_3f, __pyx_k_3f, sizeof(__pyx_k_3f), 0, 1, 0, 0},
    {&__pyx_kp_u_4f, __pyx_k_4f, sizeof(__pyx_k_4f), 0, 1, 0, 0},
    {&__pyx_kp_u_5E, __pyx_k_5E, sizeof(__pyx_k_5E), 0, 1, 0, 0},
    {&__pyx_n_u_Age, __pyx_k_Age, sizeof(__pyx_k_Age), 0, 1, 0, 1},
    {&__pyx_kp_u_Amount_Adj, __pyx_k_Amount_Adj, sizeof(__pyx_k_Amount_Adj), 0, 1, 0, 0},
    {&__pyx_kp_u_Amount_Orig, __pyx_k_Amount_Orig, sizeof(__pyx_k_Amount_Orig), 0, 1, 0, 0},
    {&__pyx_kp_u_Assets, __pyx_k_Assets, sizeof(__pyx_k_Assets), 0, 1, 0, 0},
    {&__pyx_n_s_AttributeError, __pyx_k_AttributeError, sizeof(__pyx_k_AttributeError), 0, 0, 1, 1},
    {&__pyx_kp_u_Available_Balance, __pyx_k_Available_Balance, sizeof(__pyx_k_Available_Balance), 0, 1, 0, 0},
    {&__pyx_n_s_AvellanedaMarketMakingConfigMap, __pyx_k_AvellanedaMarketMakingConfigMap, sizeof(__pyx_k_AvellanedaMarketMakingConfigMap), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy, __pyx_k_AvellanedaMarketMakingStrategy, sizeof(__pyx_k_AvellanedaMarketMakingStrategy), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_2, __pyx_k_AvellanedaMarketMakingStrategy_2, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_2), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_3, __pyx_k_AvellanedaMarketMakingStrategy_3, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_3), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_4, __pyx_k_AvellanedaMarketMakingStrategy_4, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_4), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_a, __pyx_k_AvellanedaMarketMakingStrategy_a, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_a), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_a_2, __pyx_k_AvellanedaMarketMakingStrategy_a_2, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_a_2), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_a_3, __pyx_k_AvellanedaMarketMakingStrategy_a_3, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_a_3), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_a_4, __pyx_k_AvellanedaMarketMakingStrategy_a_4, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_a_4), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_a_5, __pyx_k_AvellanedaMarketMakingStrategy_a_5, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_a_5), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_a_6, __pyx_k_AvellanedaMarketMakingStrategy_a_6, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_a_6), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_a_7, __pyx_k_AvellanedaMarketMakingStrategy_a_7, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_a_7), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_a_8, __pyx_k_AvellanedaMarketMakingStrategy_a_8, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_a_8), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_c, __pyx_k_AvellanedaMarketMakingStrategy_c, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_c), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_c_10, __pyx_k_AvellanedaMarketMakingStrategy_c_10, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_c_10), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_c_11, __pyx_k_AvellanedaMarketMakingStrategy_c_11, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_c_11), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_c_12, __pyx_k_AvellanedaMarketMakingStrategy_c_12, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_c_12), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_c_13, __pyx_k_AvellanedaMarketMakingStrategy_c_13, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_c_13), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_c_2, __pyx_k_AvellanedaMarketMakingStrategy_c_2, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_c_2), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_c_3, __pyx_k_AvellanedaMarketMakingStrategy_c_3, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_c_3), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_c_4, __pyx_k_AvellanedaMarketMakingStrategy_c_4, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_c_4), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_c_5, __pyx_k_AvellanedaMarketMakingStrategy_c_5, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_c_5), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_c_6, __pyx_k_AvellanedaMarketMakingStrategy_c_6, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_c_6), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_c_7, __pyx_k_AvellanedaMarketMakingStrategy_c_7, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_c_7), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_c_8, __pyx_k_AvellanedaMarketMakingStrategy_c_8, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_c_8), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_c_9, __pyx_k_AvellanedaMarketMakingStrategy_c_9, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_c_9), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_d, __pyx_k_AvellanedaMarketMakingStrategy_d, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_d), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_e, __pyx_k_AvellanedaMarketMakingStrategy_e, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_e), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_f, __pyx_k_AvellanedaMarketMakingStrategy_f, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_f), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_g, __pyx_k_AvellanedaMarketMakingStrategy_g, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_g), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_g_2, __pyx_k_AvellanedaMarketMakingStrategy_g_2, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_g_2), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_g_3, __pyx_k_AvellanedaMarketMakingStrategy_g_3, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_g_3), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_g_4, __pyx_k_AvellanedaMarketMakingStrategy_g_4, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_g_4), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_g_5, __pyx_k_AvellanedaMarketMakingStrategy_g_5, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_g_5), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_g_6, __pyx_k_AvellanedaMarketMakingStrategy_g_6, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_g_6), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_g_7, __pyx_k_AvellanedaMarketMakingStrategy_g_7, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_g_7), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_g_8, __pyx_k_AvellanedaMarketMakingStrategy_g_8, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_g_8), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_g_9, __pyx_k_AvellanedaMarketMakingStrategy_g_9, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_g_9), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_i, __pyx_k_AvellanedaMarketMakingStrategy_i, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_i), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_i_2, __pyx_k_AvellanedaMarketMakingStrategy_i_2, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_i_2), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_i_3, __pyx_k_AvellanedaMarketMakingStrategy_i_3, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_i_3), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_i_4, __pyx_k_AvellanedaMarketMakingStrategy_i_4, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_i_4), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_l, __pyx_k_AvellanedaMarketMakingStrategy_l, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_l), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_m, __pyx_k_AvellanedaMarketMakingStrategy_m, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_m), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_m_2, __pyx_k_AvellanedaMarketMakingStrategy_m_2, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_m_2), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_n, __pyx_k_AvellanedaMarketMakingStrategy_n, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_n), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_p, __pyx_k_AvellanedaMarketMakingStrategy_p, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_p), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_p_2, __pyx_k_AvellanedaMarketMakingStrategy_p_2, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_p_2), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_s, __pyx_k_AvellanedaMarketMakingStrategy_s, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_s), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_s_2, __pyx_k_AvellanedaMarketMakingStrategy_s_2, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_s_2), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_t, __pyx_k_AvellanedaMarketMakingStrategy_t, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_t), 0, 0, 1, 1},
    {&__pyx_n_s_AvellanedaMarketMakingStrategy_u, __pyx_k_AvellanedaMarketMakingStrategy_u, sizeof(__pyx_k_AvellanedaMarketMakingStrategy_u), 0, 0, 1, 1},
    {&__pyx_n_s_BUY, __pyx_k_BUY, sizeof(__pyx_k_BUY), 0, 0, 1, 1},
    {&__pyx_kp_u_Best_Ask, __pyx_k_Best_Ask, sizeof(__pyx_k_Best_Ask), 0, 1, 0, 0},
    {&__pyx_kp_u_Best_Bid, __pyx_k_Best_Bid, sizeof(__pyx_k_Best_Bid), 0, 1, 0, 0},
    {&__pyx_n_s_CONNECTED, __pyx_k_CONNECTED, sizeof(__pyx_k_CONNECTED), 0, 0, 1, 1},
    {&__pyx_kp_u_Calculating_volatility_estimatin, __pyx_k_Calculating_volatility_estimatin, sizeof(__pyx_k_Calculating_volatility_estimatin), 0, 1, 0, 0},
    {&__pyx_kp_u_Calculating_volatility_estimatin_2, __pyx_k_Calculating_volatility_estimatin_2, sizeof(__pyx_k_Calculating_volatility_estimatin_2), 0, 1, 0, 0},
    {&__pyx_n_s_ClientConfigAdapter, __pyx_k_ClientConfigAdapter, sizeof(__pyx_k_ClientConfigAdapter), 0, 0, 1, 1},
    {&__pyx_n_s_Clock, __pyx_k_Clock, sizeof(__pyx_k_Clock), 0, 0, 1, 1},
    {&__pyx_n_s_CreatedPairOfOrders, __pyx_k_CreatedPairOfOrders, sizeof(__pyx_k_CreatedPairOfOrders), 0, 0, 1, 1},
    {&__pyx_kp_u_Creating, __pyx_k_Creating, sizeof(__pyx_k_Creating), 0, 1, 0, 0},
    {&__pyx_kp_u_Current, __pyx_k_Current, sizeof(__pyx_k_Current), 0, 1, 0, 0},
    {&__pyx_kp_u_Current_Value, __pyx_k_Current_Value, sizeof(__pyx_k_Current_Value), 0, 1, 0, 0},
    {&__pyx_n_s_DailyBetweenTimesModel, __pyx_k_DailyBetweenTimesModel, sizeof(__pyx_k_DailyBetweenTimesModel), 0, 0, 1, 1},
    {&__pyx_n_s_DataFrame, __pyx_k_DataFrame, sizeof(__pyx_k_DataFrame), 0, 0, 1, 1},
    {&__pyx_n_s_Decimal, __pyx_k_Decimal, sizeof(__pyx_k_Decimal), 0, 0, 1, 1},
    {&__pyx_n_s_Dict, __pyx_k_Dict, sizeof(__pyx_k_Dict), 0, 0, 1, 1},
    {&__pyx_n_u_Exchange, __pyx_k_Exchange, sizeof(__pyx_k_Exchange), 0, 1, 0, 1},
    {&__pyx_n_s_ExchangeBase, __pyx_k_ExchangeBase, sizeof(__pyx_k_ExchangeBase), 0, 0, 1, 1},
    {&__pyx_n_s_FileNotFoundError, __pyx_k_FileNotFoundError, sizeof(__pyx_k_FileNotFoundError), 0, 0, 1, 1},
    {&__pyx_n_s_FromDateToDateModel, __pyx_k_FromDateToDateModel, sizeof(__pyx_k_FromDateToDateModel), 0, 0, 1, 1},
    {&__pyx_kp_u_H_M_S, __pyx_k_H_M_S, sizeof(__pyx_k_H_M_S), 0, 1, 0, 0},
    {&__pyx_n_s_HangingOrdersTracker, __pyx_k_HangingOrdersTracker, sizeof(__pyx_k_HangingOrdersTracker), 0, 0, 1, 1},
    {&__pyx_n_s_INFO, __pyx_k_INFO, sizeof(__pyx_k_INFO), 0, 0, 1, 1},
    {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
    {&__pyx_n_s_InstantVolatilityIndicator, __pyx_k_InstantVolatilityIndicator, sizeof(__pyx_k_InstantVolatilityIndicator), 0, 0, 1, 1},
    {&__pyx_n_s_LIMIT, __pyx_k_LIMIT, sizeof(__pyx_k_LIMIT), 0, 0, 1, 1},
    {&__pyx_n_u_Level, __pyx_k_Level, sizeof(__pyx_k_Level), 0, 1, 0, 1},
    {&__pyx_n_s_LimitOrder, __pyx_k_LimitOrder, sizeof(__pyx_k_LimitOrder), 0, 0, 1, 1},
    {&__pyx_n_s_List, __pyx_k_List, sizeof(__pyx_k_List), 0, 0, 1, 1},
    {&__pyx_kp_s_List_Decimal, __pyx_k_List_Decimal, sizeof(__pyx_k_List_Decimal), 0, 0, 1, 0},
    {&__pyx_kp_s_List_LimitOrder, __pyx_k_List_LimitOrder, sizeof(__pyx_k_List_LimitOrder), 0, 0, 1, 0},
    {&__pyx_kp_s_List_MarketTradingPairTuple, __pyx_k_List_MarketTradingPairTuple, sizeof(__pyx_k_List_MarketTradingPairTuple), 0, 0, 1, 0},
    {&__pyx_kp_u_Maker, __pyx_k_Maker, sizeof(__pyx_k_Maker), 0, 1, 0, 0},
    {&__pyx_kp_u_Maker_2, __pyx_k_Maker_2, sizeof(__pyx_k_Maker_2), 0, 1, 0, 0},
    {&__pyx_kp_u_Maker_buy_order_of, __pyx_k_Maker_buy_order_of, sizeof(__pyx_k_Maker_buy_order_of), 0, 1, 0, 0},
    {&__pyx_kp_u_Maker_sell_order_of, __pyx_k_Maker_sell_order_of, sizeof(__pyx_k_Maker_sell_order_of), 0, 1, 0, 0},
    {&__pyx_n_u_Market, __pyx_k_Market, sizeof(__pyx_k_Market), 0, 1, 0, 1},
    {&__pyx_n_s_MarketTradingPairTuple, __pyx_k_MarketTradingPairTuple, sizeof(__pyx_k_MarketTradingPairTuple), 0, 0, 1, 1},
    {&__pyx_kp_u_Market_connectors_are_not_ready, __pyx_k_Market_connectors_are_not_ready, sizeof(__pyx_k_Market_connectors_are_not_ready), 0, 1, 0, 0},
    {&__pyx_kp_u_Markets, __pyx_k_Markets, sizeof(__pyx_k_Markets), 0, 1, 0, 0},
    {&__pyx_kp_u_Markets_are_not_ready_No_market, __pyx_k_Markets_are_not_ready_No_market, sizeof(__pyx_k_Markets_are_not_ready_No_market), 0, 1, 0, 0},
    {&__pyx_n_s_MidPrice, __pyx_k_MidPrice, sizeof(__pyx_k_MidPrice), 0, 0, 1, 1},
    {&__pyx_n_u_MidPrice, __pyx_k_MidPrice, sizeof(__pyx_k_MidPrice), 0, 1, 0, 1},
    {&__pyx_n_s_MultiOrderLevelModel, __pyx_k_MultiOrderLevelModel, sizeof(__pyx_k_MultiOrderLevelModel), 0, 0, 1, 1},
    {&__pyx_n_s_NaN, __pyx_k_NaN, sizeof(__pyx_k_NaN), 0, 0, 1, 1},
    {&__pyx_n_s_NetworkStatus, __pyx_k_NetworkStatus, sizeof(__pyx_k_NetworkStatus), 0, 0, 1, 1},
    {&__pyx_kp_u_No_active_maker_orders, __pyx_k_No_active_maker_orders, sizeof(__pyx_k_No_active_maker_orders), 0, 1, 0, 0},
    {&__pyx_kp_u_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 1, 0, 0},
    {&__pyx_n_s_OPTION_LOG_ALL, __pyx_k_OPTION_LOG_ALL, sizeof(__pyx_k_OPTION_LOG_ALL), 0, 0, 1, 1},
    {&__pyx_n_s_OPTION_LOG_CREATE_ORDER, __pyx_k_OPTION_LOG_CREATE_ORDER, sizeof(__pyx_k_OPTION_LOG_CREATE_ORDER), 0, 0, 1, 1},
    {&__pyx_n_s_OPTION_LOG_MAKER_ORDER_FILLED, __pyx_k_OPTION_LOG_MAKER_ORDER_FILLED, sizeof(__pyx_k_OPTION_LOG_MAKER_ORDER_FILLED), 0, 0, 1, 1},
    {&__pyx_n_s_OPTION_LOG_STATUS_REPORT, __pyx_k_OPTION_LOG_STATUS_REPORT, sizeof(__pyx_k_OPTION_LOG_STATUS_REPORT), 0, 0, 1, 1},
    {&__pyx_kp_u_Optimal_Spread, __pyx_k_Optimal_Spread, sizeof(__pyx_k_Optimal_Spread), 0, 1, 0, 0},
    {&__pyx_n_s_OrderBookAssetPriceDelegate, __pyx_k_OrderBookAssetPriceDelegate, sizeof(__pyx_k_OrderBookAssetPriceDelegate), 0, 0, 1, 1},
    {&__pyx_n_s_OrderType, __pyx_k_OrderType, sizeof(__pyx_k_OrderType), 0, 0, 1, 1},
    {&__pyx_kp_u_Orders, __pyx_k_Orders, sizeof(__pyx_k_Orders), 0, 1, 0, 0},
    {&__pyx_n_u_Price, __pyx_k_Price, sizeof(__pyx_k_Price), 0, 1, 0, 1},
    {&__pyx_n_s_PriceSize, __pyx_k_PriceSize, sizeof(__pyx_k_PriceSize), 0, 0, 1, 1},
    {&__pyx_n_s_PriceType, __pyx_k_PriceType, sizeof(__pyx_k_PriceType), 0, 0, 1, 1},
    {&__pyx_n_s_Proposal, __pyx_k_Proposal, sizeof(__pyx_k_Proposal), 0, 0, 1, 1},
    {&__pyx_kp_u_Reservation_Price, __pyx_k_Reservation_Price, sizeof(__pyx_k_Reservation_Price), 0, 1, 0, 0},
    {&__pyx_n_s_RunAlwaysExecutionState, __pyx_k_RunAlwaysExecutionState, sizeof(__pyx_k_RunAlwaysExecutionState), 0, 0, 1, 1},
    {&__pyx_n_s_RunInTimeConditionalExecutionSta, __pyx_k_RunInTimeConditionalExecutionSta, sizeof(__pyx_k_RunInTimeConditionalExecutionSta), 0, 0, 1, 1},
    {&__pyx_n_s_SELL, __pyx_k_SELL, sizeof(__pyx_k_SELL), 0, 0, 1, 1},
    {&__pyx_n_u_Spread, __pyx_k_Spread, sizeof(__pyx_k_Spread), 0, 1, 0, 1},
    {&__pyx_n_s_StrategyBase, __pyx_k_StrategyBase, sizeof(__pyx_k_StrategyBase), 0, 0, 1, 1},
    {&__pyx_kp_u_Strategy_parameters, __pyx_k_Strategy_parameters, sizeof(__pyx_k_Strategy_parameters), 0, 1, 0, 0},
    {&__pyx_n_s_Timestamp, __pyx_k_Timestamp, sizeof(__pyx_k_Timestamp), 0, 0, 1, 1},
    {&__pyx_kp_u_Total_Balance, __pyx_k_Total_Balance, sizeof(__pyx_k_Total_Balance), 0, 1, 0, 0},
    {&__pyx_n_s_TrackHangingOrdersModel, __pyx_k_TrackHangingOrdersModel, sizeof(__pyx_k_TrackHangingOrdersModel), 0, 0, 1, 1},
    {&__pyx_n_s_TradeType, __pyx_k_TradeType, sizeof(__pyx_k_TradeType), 0, 0, 1, 1},
    {&__pyx_n_s_TradingIntensityIndicator, __pyx_k_TradingIntensityIndicator, sizeof(__pyx_k_TradingIntensityIndicator), 0, 0, 1, 1},
    {&__pyx_n_s_Tuple, __pyx_k_Tuple, sizeof(__pyx_k_Tuple), 0, 0, 1, 1},
    {&__pyx_kp_s_Tuple_List_Proposal_List_Proposa, __pyx_k_Tuple_List_Proposal_List_Proposa, sizeof(__pyx_k_Tuple_List_Proposal_List_Proposa), 0, 0, 1, 0},
    {&__pyx_n_u_Type, __pyx_k_Type, sizeof(__pyx_k_Type), 0, 1, 0, 1},
    {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
    {&__pyx_n_s_Union, __pyx_k_Union, sizeof(__pyx_k_Union), 0, 0, 1, 1},
    {&__pyx_kp_s_Union_AvellanedaMarketMakingConf, __pyx_k_Union_AvellanedaMarketMakingConf, sizeof(__pyx_k_Union_AvellanedaMarketMakingConf), 0, 0, 1, 0},
    {&__pyx_kp_u_WARNINGS, __pyx_k_WARNINGS, sizeof(__pyx_k_WARNINGS), 0, 1, 0, 0},
    {&__pyx_kp_u_WARNING_Some_markets_are_not_con, __pyx_k_WARNING_Some_markets_are_not_con, sizeof(__pyx_k_WARNING_Some_markets_are_not_con), 0, 1, 0, 0},
    {&__pyx_kp_u__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 1, 0, 0},
    {&__pyx_kp_u__11, __pyx_k__11, sizeof(__pyx_k__11), 0, 1, 0, 0},
    {&__pyx_kp_u__12, __pyx_k__12, sizeof(__pyx_k__12), 0, 1, 0, 0},
    {&__pyx_kp_u__13, __pyx_k__13, sizeof(__pyx_k__13), 0, 1, 0, 0},
    {&__pyx_kp_u__14, __pyx_k__14, sizeof(__pyx_k__14), 0, 1, 0, 0},
    {&__pyx_kp_u__15, __pyx_k__15, sizeof(__pyx_k__15), 0, 1, 0, 0},
    {&__pyx_n_s__17, __pyx_k__17, sizeof(__pyx_k__17), 0, 0, 1, 1},
    {&__pyx_kp_u__18, __pyx_k__18, sizeof(__pyx_k__18), 0, 1, 0, 0},
    {&__pyx_kp_u__4, __pyx_k__4, sizeof(__pyx_k__4), 0, 1, 0, 0},
    {&__pyx_kp_u__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 1, 0, 0},
    {&__pyx_kp_u__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 1, 0, 0},
    {&__pyx_kp_u__7, __pyx_k__7, sizeof(__pyx_k__7), 0, 1, 0, 0},
    {&__pyx_kp_u__8, __pyx_k__8, sizeof(__pyx_k__8), 0, 1, 0, 0},
    {&__pyx_kp_u__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 1, 0, 0},
    {&__pyx_n_s__96, __pyx_k__96, sizeof(__pyx_k__96), 0, 0, 1, 1},
    {&__pyx_n_u_a, __pyx_k_a, sizeof(__pyx_k_a), 0, 1, 0, 1},
    {&__pyx_n_s_active_markets, __pyx_k_active_markets, sizeof(__pyx_k_active_markets), 0, 0, 1, 1},
    {&__pyx_n_s_active_non_hanging_orders, __pyx_k_active_non_hanging_orders, sizeof(__pyx_k_active_non_hanging_orders), 0, 0, 1, 1},
    {&__pyx_n_s_active_orders, __pyx_k_active_orders, sizeof(__pyx_k_active_orders), 0, 0, 1, 1},
    {&__pyx_n_s_active_orders_df, __pyx_k_active_orders_df, sizeof(__pyx_k_active_orders_df), 0, 0, 1, 1},
    {&__pyx_n_s_active_orders_df_locals_lambda, __pyx_k_active_orders_df_locals_lambda, sizeof(__pyx_k_active_orders_df_locals_lambda), 0, 0, 1, 1},
    {&__pyx_n_s_add_as_hanging_order, __pyx_k_add_as_hanging_order, sizeof(__pyx_k_add_as_hanging_order), 0, 0, 1, 1},
    {&__pyx_n_s_add_current_pairs_of_proposal_or, __pyx_k_add_current_pairs_of_proposal_or, sizeof(__pyx_k_add_current_pairs_of_proposal_or), 0, 0, 1, 1},
    {&__pyx_n_s_add_sample, __pyx_k_add_sample, sizeof(__pyx_k_add_sample), 0, 0, 1, 1},
    {&__pyx_n_s_add_transaction_costs, __pyx_k_add_transaction_costs, sizeof(__pyx_k_add_transaction_costs), 0, 0, 1, 1},
    {&__pyx_n_s_adjusted_available_balance_for_o, __pyx_k_adjusted_available_balance_for_o, sizeof(__pyx_k_adjusted_available_balance_for_o), 0, 0, 1, 1},
    {&__pyx_n_s_age, __pyx_k_age, sizeof(__pyx_k_age), 0, 0, 1, 1},
    {&__pyx_n_s_all, __pyx_k_all, sizeof(__pyx_k_all), 0, 0, 1, 1},
    {&__pyx_n_s_all_markets_ready, __pyx_k_all_markets_ready, sizeof(__pyx_k_all_markets_ready), 0, 0, 1, 1},
    {&__pyx_n_s_all_non_hanging_orders, __pyx_k_all_non_hanging_orders, sizeof(__pyx_k_all_non_hanging_orders), 0, 0, 1, 1},
    {&__pyx_n_s_all_orders, __pyx_k_all_orders, sizeof(__pyx_k_all_orders), 0, 0, 1, 1},
    {&__pyx_kp_u_alpha, __pyx_k_alpha, sizeof(__pyx_k_alpha), 0, 1, 0, 0},
    {&__pyx_n_u_alpha_2, __pyx_k_alpha_2, sizeof(__pyx_k_alpha_2), 0, 1, 0, 1},
    {&__pyx_n_s_amount, __pyx_k_amount, sizeof(__pyx_k_amount), 0, 0, 1, 1},
    {&__pyx_n_s_amount_orig, __pyx_k_amount_orig, sizeof(__pyx_k_amount_orig), 0, 0, 1, 1},
    {&__pyx_n_s_apply, __pyx_k_apply, sizeof(__pyx_k_apply), 0, 0, 1, 1},
    {&__pyx_n_s_apply_add_transaction_costs, __pyx_k_apply_add_transaction_costs, sizeof(__pyx_k_apply_add_transaction_costs), 0, 0, 1, 1},
    {&__pyx_n_s_apply_budget_constraint, __pyx_k_apply_budget_constraint, sizeof(__pyx_k_apply_budget_constraint), 0, 0, 1, 1},
    {&__pyx_n_s_apply_order_amount_eta_transform, __pyx_k_apply_order_amount_eta_transform, sizeof(__pyx_k_apply_order_amount_eta_transform), 0, 0, 1, 1},
    {&__pyx_n_s_apply_order_optimization, __pyx_k_apply_order_optimization, sizeof(__pyx_k_apply_order_optimization), 0, 0, 1, 1},
    {&__pyx_n_s_apply_order_price_modifiers, __pyx_k_apply_order_price_modifiers, sizeof(__pyx_k_apply_order_price_modifiers), 0, 0, 1, 1},
    {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
    {&__pyx_n_s_ask_level_spreads, __pyx_k_ask_level_spreads, sizeof(__pyx_k_ask_level_spreads), 0, 0, 1, 1},
    {&__pyx_kp_u_ask_orders_at_Size_Price, __pyx_k_ask_orders_at_Size_Price, sizeof(__pyx_k_ask_orders_at_Size_Price), 0, 1, 0, 0},
    {&__pyx_n_s_ask_price, __pyx_k_ask_price, sizeof(__pyx_k_ask_price), 0, 0, 1, 1},
    {&__pyx_n_s_assets_df, __pyx_k_assets_df, sizeof(__pyx_k_assets_df), 0, 0, 1, 1},
    {&__pyx_n_s_asyncio_coroutines, __pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 0, 1, 1},
    {&__pyx_n_s_available_base_balance, __pyx_k_available_base_balance, sizeof(__pyx_k_available_base_balance), 0, 0, 1, 1},
    {&__pyx_n_s_available_quote_balance, __pyx_k_available_quote_balance, sizeof(__pyx_k_available_quote_balance), 0, 0, 1, 1},
    {&__pyx_n_s_balance_warning, __pyx_k_balance_warning, sizeof(__pyx_k_balance_warning), 0, 0, 1, 1},
    {&__pyx_n_s_base_asset, __pyx_k_base_asset, sizeof(__pyx_k_base_asset), 0, 0, 1, 1},
    {&__pyx_n_s_base_balance, __pyx_k_base_balance, sizeof(__pyx_k_base_balance), 0, 0, 1, 1},
    {&__pyx_n_s_base_currency, __pyx_k_base_currency, sizeof(__pyx_k_base_currency), 0, 0, 1, 1},
    {&__pyx_n_s_base_ratio, __pyx_k_base_ratio, sizeof(__pyx_k_base_ratio), 0, 0, 1, 1},
    {&__pyx_n_s_base_value, __pyx_k_base_value, sizeof(__pyx_k_base_value), 0, 0, 1, 1},
    {&__pyx_n_s_best_ask, __pyx_k_best_ask, sizeof(__pyx_k_best_ask), 0, 0, 1, 1},
    {&__pyx_n_u_best_ask, __pyx_k_best_ask, sizeof(__pyx_k_best_ask), 0, 1, 0, 1},
    {&__pyx_n_s_best_bid, __pyx_k_best_bid, sizeof(__pyx_k_best_bid), 0, 0, 1, 1},
    {&__pyx_n_u_best_bid, __pyx_k_best_bid, sizeof(__pyx_k_best_bid), 0, 1, 0, 1},
    {&__pyx_n_s_bid_level_spreads, __pyx_k_bid_level_spreads, sizeof(__pyx_k_bid_level_spreads), 0, 0, 1, 1},
    {&__pyx_kp_u_bid_orders_at_Size_Price, __pyx_k_bid_orders_at_Size_Price, sizeof(__pyx_k_bid_orders_at_Size_Price), 0, 1, 0, 0},
    {&__pyx_n_s_bid_price, __pyx_k_bid_price, sizeof(__pyx_k_bid_price), 0, 0, 1, 1},
    {&__pyx_n_s_bool, __pyx_k_bool, sizeof(__pyx_k_bool), 0, 0, 1, 1},
    {&__pyx_n_u_buy, __pyx_k_buy, sizeof(__pyx_k_buy), 0, 1, 0, 1},
    {&__pyx_n_s_buys, __pyx_k_buys, sizeof(__pyx_k_buys), 0, 0, 1, 1},
    {&__pyx_n_s_calculate, __pyx_k_calculate, sizeof(__pyx_k_calculate), 0, 0, 1, 1},
    {&__pyx_n_s_calculate_inventory, __pyx_k_calculate_inventory, sizeof(__pyx_k_calculate_inventory), 0, 0, 1, 1},
    {&__pyx_n_s_calculate_reservation_price_and, __pyx_k_calculate_reservation_price_and, sizeof(__pyx_k_calculate_reservation_price_and), 0, 0, 1, 1},
    {&__pyx_n_s_calculate_target_inventory, __pyx_k_calculate_target_inventory, sizeof(__pyx_k_calculate_target_inventory), 0, 0, 1, 1},
    {&__pyx_n_s_cancel_active_orders, __pyx_k_cancel_active_orders, sizeof(__pyx_k_cancel_active_orders), 0, 0, 1, 1},
    {&__pyx_n_s_cancel_order, __pyx_k_cancel_order, sizeof(__pyx_k_cancel_order), 0, 0, 1, 1},
    {&__pyx_n_s_candidate_hanging_orders, __pyx_k_candidate_hanging_orders, sizeof(__pyx_k_candidate_hanging_orders), 0, 0, 1, 1},
    {&__pyx_n_s_candidate_hanging_orders_from_pa, __pyx_k_candidate_hanging_orders_from_pa, sizeof(__pyx_k_candidate_hanging_orders_from_pa), 0, 0, 1, 1},
    {&__pyx_n_s_ceil, __pyx_k_ceil, sizeof(__pyx_k_ceil), 0, 0, 1, 1},
    {&__pyx_n_s_class_getitem, __pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 0, 1, 1},
    {&__pyx_n_s_client_order_id, __pyx_k_client_order_id, sizeof(__pyx_k_client_order_id), 0, 0, 1, 1},
    {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
    {&__pyx_n_s_clock, __pyx_k_clock, sizeof(__pyx_k_clock), 0, 0, 1, 1},
    {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
    {&__pyx_n_s_closing_time, __pyx_k_closing_time, sizeof(__pyx_k_closing_time), 0, 0, 1, 1},
    {&__pyx_n_s_cls, __pyx_k_cls, sizeof(__pyx_k_cls), 0, 0, 1, 1},
    {&__pyx_n_s_collect_market_variables, __pyx_k_collect_market_variables, sizeof(__pyx_k_collect_market_variables), 0, 0, 1, 1},
    {&__pyx_n_s_columns, __pyx_k_columns, sizeof(__pyx_k_columns), 0, 0, 1, 1},
    {&__pyx_n_s_config_map, __pyx_k_config_map, sizeof(__pyx_k_config_map), 0, 0, 1, 1},
    {&__pyx_n_s_create_base_proposal, __pyx_k_create_base_proposal, sizeof(__pyx_k_create_base_proposal), 0, 0, 1, 1},
    {&__pyx_n_s_create_basic_proposal, __pyx_k_create_basic_proposal, sizeof(__pyx_k_create_basic_proposal), 0, 0, 1, 1},
    {&__pyx_n_s_create_proposal_based_on_order_l, __pyx_k_create_proposal_based_on_order_l, sizeof(__pyx_k_create_proposal_based_on_order_l), 0, 0, 1, 1},
    {&__pyx_n_s_create_proposal_based_on_order_o, __pyx_k_create_proposal_based_on_order_o, sizeof(__pyx_k_create_proposal_based_on_order_o), 0, 0, 1, 1},
    {&__pyx_n_s_current_created_pairs_of_orders, __pyx_k_current_created_pairs_of_orders, sizeof(__pyx_k_current_created_pairs_of_orders), 0, 0, 1, 1},
    {&__pyx_n_u_current_inv, __pyx_k_current_inv, sizeof(__pyx_k_current_inv), 0, 1, 0, 1},
    {&__pyx_n_s_current_prices, __pyx_k_current_prices, sizeof(__pyx_k_current_prices), 0, 0, 1, 1},
    {&__pyx_n_s_current_value, __pyx_k_current_value, sizeof(__pyx_k_current_value), 0, 0, 1, 1},
    {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
    {&__pyx_n_s_datetime, __pyx_k_datetime, sizeof(__pyx_k_datetime), 0, 0, 1, 1},
    {&__pyx_n_s_debug_csv_path, __pyx_k_debug_csv_path, sizeof(__pyx_k_debug_csv_path), 0, 0, 1, 1},
    {&__pyx_n_s_decimal, __pyx_k_decimal, sizeof(__pyx_k_decimal), 0, 0, 1, 1},
    {&__pyx_n_s_df, __pyx_k_df, sizeof(__pyx_k_df), 0, 0, 1, 1},
    {&__pyx_n_s_df_header, __pyx_k_df_header, sizeof(__pyx_k_df_header), 0, 0, 1, 1},
    {&__pyx_n_s_df_lines, __pyx_k_df_lines, sizeof(__pyx_k_df_lines), 0, 0, 1, 1},
    {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
    {&__pyx_kp_u_disable, __pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0, 0},
    {&__pyx_n_s_display_name, __pyx_k_display_name, sizeof(__pyx_k_display_name), 0, 0, 1, 1},
    {&__pyx_n_s_dump_debug_variables, __pyx_k_dump_debug_variables, sizeof(__pyx_k_dump_debug_variables), 0, 0, 1, 1},
    {&__pyx_kp_u_enable, __pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0, 0},
    {&__pyx_n_s_end_datetime, __pyx_k_end_datetime, sizeof(__pyx_k_end_datetime), 0, 0, 1, 1},
    {&__pyx_n_s_end_time, __pyx_k_end_time, sizeof(__pyx_k_end_time), 0, 0, 1, 1},
    {&__pyx_n_s_end_timestamp, __pyx_k_end_timestamp, sizeof(__pyx_k_end_timestamp), 0, 0, 1, 1},
    {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
    {&__pyx_n_s_eta, __pyx_k_eta, sizeof(__pyx_k_eta), 0, 0, 1, 1},
    {&__pyx_n_u_eta, __pyx_k_eta, sizeof(__pyx_k_eta), 0, 1, 0, 1},
    {&__pyx_n_s_execute_orders_proposal, __pyx_k_execute_orders_proposal, sizeof(__pyx_k_execute_orders_proposal), 0, 0, 1, 1},
    {&__pyx_n_s_execution_mode, __pyx_k_execution_mode, sizeof(__pyx_k_execution_mode), 0, 0, 1, 1},
    {&__pyx_n_s_execution_state, __pyx_k_execution_state, sizeof(__pyx_k_execution_state), 0, 0, 1, 1},
    {&__pyx_n_s_execution_timeframe, __pyx_k_execution_timeframe, sizeof(__pyx_k_execution_timeframe), 0, 0, 1, 1},
    {&__pyx_n_s_execution_timeframe_mode, __pyx_k_execution_timeframe_mode, sizeof(__pyx_k_execution_timeframe_mode), 0, 0, 1, 1},
    {&__pyx_n_s_exists, __pyx_k_exists, sizeof(__pyx_k_exists), 0, 0, 1, 1},
    {&__pyx_n_s_exp, __pyx_k_exp, sizeof(__pyx_k_exp), 0, 0, 1, 1},
    {&__pyx_kp_u_filled, __pyx_k_filled, sizeof(__pyx_k_filled), 0, 1, 0, 0},
    {&__pyx_n_s_filled_order_delay, __pyx_k_filled_order_delay, sizeof(__pyx_k_filled_order_delay), 0, 0, 1, 1},
    {&__pyx_n_s_first_col_length, __pyx_k_first_col_length, sizeof(__pyx_k_first_col_length), 0, 0, 1, 1},
    {&__pyx_n_s_float, __pyx_k_float, sizeof(__pyx_k_float), 0, 0, 1, 1},
    {&__pyx_n_s_floor, __pyx_k_floor, sizeof(__pyx_k_floor), 0, 0, 1, 1},
    {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
    {&__pyx_n_s_format_status, __pyx_k_format_status, sizeof(__pyx_k_format_status), 0, 0, 1, 1},
    {&__pyx_n_s_formatters, __pyx_k_formatters, sizeof(__pyx_k_formatters), 0, 0, 1, 1},
    {&__pyx_n_s_gamma, __pyx_k_gamma, sizeof(__pyx_k_gamma), 0, 0, 1, 1},
    {&__pyx_n_u_gamma, __pyx_k_gamma, sizeof(__pyx_k_gamma), 0, 1, 0, 1},
    {&__pyx_kp_u_gc, __pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0, 0},
    {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
    {&__pyx_n_s_getLogger, __pyx_k_getLogger, sizeof(__pyx_k_getLogger), 0, 0, 1, 1},
    {&__pyx_n_s_get_adjusted_available_balance, __pyx_k_get_adjusted_available_balance, sizeof(__pyx_k_get_adjusted_available_balance), 0, 0, 1, 1},
    {&__pyx_n_s_get_available_balance, __pyx_k_get_available_balance, sizeof(__pyx_k_get_available_balance), 0, 0, 1, 1},
    {&__pyx_n_s_get_balance, __pyx_k_get_balance, sizeof(__pyx_k_get_balance), 0, 0, 1, 1},
    {&__pyx_n_s_get_config_map_execution_mode, __pyx_k_get_config_map_execution_mode, sizeof(__pyx_k_get_config_map_execution_mode), 0, 0, 1, 1},
    {&__pyx_n_s_get_config_map_hanging_orders, __pyx_k_get_config_map_hanging_orders, sizeof(__pyx_k_get_config_map_hanging_orders), 0, 0, 1, 1},
    {&__pyx_n_s_get_config_map_indicators, __pyx_k_get_config_map_indicators, sizeof(__pyx_k_get_config_map_indicators), 0, 0, 1, 1},
    {&__pyx_n_s_get_last_price, __pyx_k_get_last_price, sizeof(__pyx_k_get_last_price), 0, 0, 1, 1},
    {&__pyx_n_s_get_level_spreads, __pyx_k_get_level_spreads, sizeof(__pyx_k_get_level_spreads), 0, 0, 1, 1},
    {&__pyx_n_s_get_limit_orders, __pyx_k_get_limit_orders, sizeof(__pyx_k_get_limit_orders), 0, 0, 1, 1},
    {&__pyx_n_s_get_maker_order_type, __pyx_k_get_maker_order_type, sizeof(__pyx_k_get_maker_order_type), 0, 0, 1, 1},
    {&__pyx_n_s_get_mid_price, __pyx_k_get_mid_price, sizeof(__pyx_k_get_mid_price), 0, 0, 1, 1},
    {&__pyx_n_s_get_price, __pyx_k_get_price, sizeof(__pyx_k_get_price), 0, 0, 1, 1},
    {&__pyx_n_s_get_price_by_type, __pyx_k_get_price_by_type, sizeof(__pyx_k_get_price_by_type), 0, 0, 1, 1},
    {&__pyx_n_s_get_price_for_volume, __pyx_k_get_price_for_volume, sizeof(__pyx_k_get_price_for_volume), 0, 0, 1, 1},
    {&__pyx_n_s_get_spread, __pyx_k_get_spread, sizeof(__pyx_k_get_spread), 0, 0, 1, 1},
    {&__pyx_n_s_get_volatility, __pyx_k_get_volatility, sizeof(__pyx_k_get_volatility), 0, 0, 1, 1},
    {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
    {&__pyx_n_u_hang, __pyx_k_hang, sizeof(__pyx_k_hang), 0, 1, 0, 1},
    {&__pyx_n_s_hanging_orders_cancel_pct, __pyx_k_hanging_orders_cancel_pct, sizeof(__pyx_k_hanging_orders_cancel_pct), 0, 0, 1, 1},
    {&__pyx_n_s_hanging_orders_enabled, __pyx_k_hanging_orders_enabled, sizeof(__pyx_k_hanging_orders_enabled), 0, 0, 1, 1},
    {&__pyx_n_s_hanging_orders_mode, __pyx_k_hanging_orders_mode, sizeof(__pyx_k_hanging_orders_mode), 0, 0, 1, 1},
    {&__pyx_n_s_hanging_orders_tracker, __pyx_k_hanging_orders_tracker, sizeof(__pyx_k_hanging_orders_tracker), 0, 0, 1, 1},
    {&__pyx_kp_u_has_been_completely_filled, __pyx_k_has_been_completely_filled, sizeof(__pyx_k_has_been_completely_filled), 0, 1, 0, 0},
    {&__pyx_n_s_hb_app_notification, __pyx_k_hb_app_notification, sizeof(__pyx_k_hb_app_notification), 0, 0, 1, 1},
    {&__pyx_n_s_header, __pyx_k_header, sizeof(__pyx_k_header), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_client_config_config, __pyx_k_hummingbot_client_config_config, sizeof(__pyx_k_hummingbot_client_config_config), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_connector_exchange_ba, __pyx_k_hummingbot_connector_exchange_ba, sizeof(__pyx_k_hummingbot_connector_exchange_ba), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_core_data_type_common, __pyx_k_hummingbot_core_data_type_common, sizeof(__pyx_k_hummingbot_core_data_type_common), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_core_data_type_limit, __pyx_k_hummingbot_core_data_type_limit, sizeof(__pyx_k_hummingbot_core_data_type_limit), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_core_network_iterator, __pyx_k_hummingbot_core_network_iterator, sizeof(__pyx_k_hummingbot_core_network_iterator), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_core_utils, __pyx_k_hummingbot_core_utils, sizeof(__pyx_k_hummingbot_core_utils), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_strategy___utils___tr, __pyx_k_hummingbot_strategy___utils___tr, sizeof(__pyx_k_hummingbot_strategy___utils___tr), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_strategy___utils___tr_2, __pyx_k_hummingbot_strategy___utils___tr_2, sizeof(__pyx_k_hummingbot_strategy___utils___tr_2), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_k_hummingbot_strategy_avellaneda_m, sizeof(__pyx_k_hummingbot_strategy_avellaneda_m), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_strategy_avellaneda_m_2, __pyx_k_hummingbot_strategy_avellaneda_m_2, sizeof(__pyx_k_hummingbot_strategy_avellaneda_m_2), 0, 0, 1, 1},
    {&__pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_k_hummingbot_strategy_avellaneda_m_3, sizeof(__pyx_k_hummingbot_strategy_avellaneda_m_3), 0, 0, 1, 0},
    {&__pyx_n_s_hummingbot_strategy_conditional, __pyx_k_hummingbot_strategy_conditional, sizeof(__pyx_k_hummingbot_strategy_conditional), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_strategy_data_types, __pyx_k_hummingbot_strategy_data_types, sizeof(__pyx_k_hummingbot_strategy_data_types), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_strategy_hanging_orde, __pyx_k_hummingbot_strategy_hanging_orde, sizeof(__pyx_k_hummingbot_strategy_hanging_orde), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_strategy_market_tradi, __pyx_k_hummingbot_strategy_market_tradi, sizeof(__pyx_k_hummingbot_strategy_market_tradi), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_strategy_order_book_a, __pyx_k_hummingbot_strategy_order_book_a, sizeof(__pyx_k_hummingbot_strategy_order_book_a), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_strategy_strategy_bas, __pyx_k_hummingbot_strategy_strategy_bas, sizeof(__pyx_k_hummingbot_strategy_strategy_bas), 0, 0, 1, 1},
    {&__pyx_n_s_hummingbot_strategy_utils, __pyx_k_hummingbot_strategy_utils, sizeof(__pyx_k_hummingbot_strategy_utils), 0, 0, 1, 1},
    {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
    {&__pyx_n_s_idx, __pyx_k_idx, sizeof(__pyx_k_idx), 0, 0, 1, 1},
    {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
    {&__pyx_n_s_in_flight_cancels, __pyx_k_in_flight_cancels, sizeof(__pyx_k_in_flight_cancels), 0, 0, 1, 1},
    {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
    {&__pyx_n_s_info, __pyx_k_info, sizeof(__pyx_k_info), 0, 0, 1, 1},
    {&__pyx_n_s_init_params, __pyx_k_init_params, sizeof(__pyx_k_init_params), 0, 0, 1, 1},
    {&__pyx_n_s_initializing, __pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 0, 1, 1},
    {&__pyx_n_s_int, __pyx_k_int, sizeof(__pyx_k_int), 0, 0, 1, 1},
    {&__pyx_n_s_inventory_target_base, __pyx_k_inventory_target_base, sizeof(__pyx_k_inventory_target_base), 0, 0, 1, 1},
    {&__pyx_n_s_inventory_target_base_pct, __pyx_k_inventory_target_base_pct, sizeof(__pyx_k_inventory_target_base_pct), 0, 0, 1, 1},
    {&__pyx_n_u_inventory_target_pct, __pyx_k_inventory_target_pct, sizeof(__pyx_k_inventory_target_pct), 0, 1, 0, 1},
    {&__pyx_n_s_is_algorithm_changed, __pyx_k_is_algorithm_changed, sizeof(__pyx_k_is_algorithm_changed), 0, 0, 1, 1},
    {&__pyx_n_s_is_algorithm_ready, __pyx_k_is_algorithm_ready, sizeof(__pyx_k_is_algorithm_ready), 0, 0, 1, 1},
    {&__pyx_n_s_is_buy, __pyx_k_is_buy, sizeof(__pyx_k_is_buy), 0, 0, 1, 1},
    {&__pyx_n_s_is_coroutine, __pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 0, 1, 1},
    {&__pyx_n_s_is_debug, __pyx_k_is_debug, sizeof(__pyx_k_is_debug), 0, 0, 1, 1},
    {&__pyx_kp_u_is_filled, __pyx_k_is_filled, sizeof(__pyx_k_is_filled), 0, 1, 0, 0},
    {&__pyx_n_s_is_hanging_order, __pyx_k_is_hanging_order, sizeof(__pyx_k_is_hanging_order), 0, 0, 1, 1},
    {&__pyx_n_s_is_order_id_in_completed_hanging, __pyx_k_is_order_id_in_completed_hanging, sizeof(__pyx_k_is_order_id_in_completed_hanging), 0, 0, 1, 1},
    {&__pyx_n_s_is_order_id_in_hanging_orders, __pyx_k_is_order_id_in_hanging_orders, sizeof(__pyx_k_is_order_id_in_hanging_orders), 0, 0, 1, 1},
    {&__pyx_n_s_is_potential_hanging_order, __pyx_k_is_potential_hanging_order, sizeof(__pyx_k_is_potential_hanging_order), 0, 0, 1, 1},
    {&__pyx_n_s_is_sampling_buffer_changed, __pyx_k_is_sampling_buffer_changed, sizeof(__pyx_k_is_sampling_buffer_changed), 0, 0, 1, 1},
    {&__pyx_n_s_is_sampling_buffer_full, __pyx_k_is_sampling_buffer_full, sizeof(__pyx_k_is_sampling_buffer_full), 0, 0, 1, 1},
    {&__pyx_n_s_is_within_tolerance, __pyx_k_is_within_tolerance, sizeof(__pyx_k_is_within_tolerance), 0, 0, 1, 1},
    {&__pyx_kp_u_isenabled, __pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0, 0},
    {&__pyx_n_s_isnan, __pyx_k_isnan, sizeof(__pyx_k_isnan), 0, 0, 1, 1},
    {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
    {&__pyx_kp_u_kappa, __pyx_k_kappa, sizeof(__pyx_k_kappa), 0, 1, 0, 0},
    {&__pyx_n_u_kappa_2, __pyx_k_kappa_2, sizeof(__pyx_k_kappa_2), 0, 1, 0, 1},
    {&__pyx_n_s_key, __pyx_k_key, sizeof(__pyx_k_key), 0, 0, 1, 1},
    {&__pyx_n_s_len, __pyx_k_len, sizeof(__pyx_k_len), 0, 0, 1, 1},
    {&__pyx_n_s_level, __pyx_k_level, sizeof(__pyx_k_level), 0, 0, 1, 1},
    {&__pyx_n_s_level_distances, __pyx_k_level_distances, sizeof(__pyx_k_level_distances), 0, 0, 1, 1},
    {&__pyx_n_s_level_step, __pyx_k_level_step, sizeof(__pyx_k_level_step), 0, 0, 1, 1},
    {&__pyx_n_s_limit_orders, __pyx_k_limit_orders, sizeof(__pyx_k_limit_orders), 0, 0, 1, 1},
    {&__pyx_n_s_line, __pyx_k_line, sizeof(__pyx_k_line), 0, 0, 1, 1},
    {&__pyx_n_s_lines, __pyx_k_lines, sizeof(__pyx_k_lines), 0, 0, 1, 1},
    {&__pyx_n_s_ln, __pyx_k_ln, sizeof(__pyx_k_ln), 0, 0, 1, 1},
    {&__pyx_n_s_log_with_clock, __pyx_k_log_with_clock, sizeof(__pyx_k_log_with_clock), 0, 0, 1, 1},
    {&__pyx_n_s_logger, __pyx_k_logger, sizeof(__pyx_k_logger), 0, 0, 1, 1},
    {&__pyx_n_s_logging, __pyx_k_logging, sizeof(__pyx_k_logging), 0, 0, 1, 1},
    {&__pyx_n_s_logging_options, __pyx_k_logging_options, sizeof(__pyx_k_logging_options), 0, 0, 1, 1},
    {&__pyx_n_s_lvl_buy, __pyx_k_lvl_buy, sizeof(__pyx_k_lvl_buy), 0, 0, 1, 1},
    {&__pyx_n_s_lvl_sell, __pyx_k_lvl_sell, sizeof(__pyx_k_lvl_sell), 0, 0, 1, 1},
    {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
    {&__pyx_n_s_map_df_to_str, __pyx_k_map_df_to_str, sizeof(__pyx_k_map_df_to_str), 0, 0, 1, 1},
    {&__pyx_n_s_market, __pyx_k_market, sizeof(__pyx_k_market), 0, 0, 1, 1},
    {&__pyx_n_s_market_books, __pyx_k_market_books, sizeof(__pyx_k_market_books), 0, 0, 1, 1},
    {&__pyx_n_s_market_info, __pyx_k_market_info, sizeof(__pyx_k_market_info), 0, 0, 1, 1},
    {&__pyx_n_s_market_info_to_active_orders, __pyx_k_market_info_to_active_orders, sizeof(__pyx_k_market_info_to_active_orders), 0, 0, 1, 1},
    {&__pyx_n_s_market_pair_to_active_orders, __pyx_k_market_pair_to_active_orders, sizeof(__pyx_k_market_pair_to_active_orders), 0, 0, 1, 1},
    {&__pyx_n_s_market_status_data_frame, __pyx_k_market_status_data_frame, sizeof(__pyx_k_market_status_data_frame), 0, 0, 1, 1},
    {&__pyx_n_s_market_trading_pair_tuples, __pyx_k_market_trading_pair_tuples, sizeof(__pyx_k_market_trading_pair_tuples), 0, 0, 1, 1},
    {&__pyx_n_s_markets_columns, __pyx_k_markets_columns, sizeof(__pyx_k_markets_columns), 0, 0, 1, 1},
    {&__pyx_n_s_markets_data, __pyx_k_markets_data, sizeof(__pyx_k_markets_data), 0, 0, 1, 1},
    {&__pyx_n_s_markets_df, __pyx_k_markets_df, sizeof(__pyx_k_markets_df), 0, 0, 1, 1},
    {&__pyx_n_s_math, __pyx_k_math, sizeof(__pyx_k_math), 0, 0, 1, 1},
    {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
    {&__pyx_n_s_max_order_age, __pyx_k_max_order_age, sizeof(__pyx_k_max_order_age), 0, 0, 1, 1},
    {&__pyx_n_s_measure_order_book_liquidity, __pyx_k_measure_order_book_liquidity, sizeof(__pyx_k_measure_order_book_liquidity), 0, 0, 1, 1},
    {&__pyx_kp_u_mid_price, __pyx_k_mid_price, sizeof(__pyx_k_mid_price), 0, 1, 0, 0},
    {&__pyx_n_s_mid_price_2, __pyx_k_mid_price_2, sizeof(__pyx_k_mid_price_2), 0, 0, 1, 1},
    {&__pyx_n_u_mid_price_2, __pyx_k_mid_price_2, sizeof(__pyx_k_mid_price_2), 0, 1, 0, 1},
    {&__pyx_kp_u_mid_price_std_dev, __pyx_k_mid_price_std_dev, sizeof(__pyx_k_mid_price_std_dev), 0, 1, 0, 0},
    {&__pyx_n_s_mid_price_variance, __pyx_k_mid_price_variance, sizeof(__pyx_k_mid_price_variance), 0, 0, 1, 1},
    {&__pyx_n_u_mid_price_variance, __pyx_k_mid_price_variance, sizeof(__pyx_k_mid_price_variance), 0, 1, 0, 1},
    {&__pyx_n_s_min_spread, __pyx_k_min_spread, sizeof(__pyx_k_min_spread), 0, 0, 1, 1},
    {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
    {&__pyx_n_s_model_config, __pyx_k_model_config, sizeof(__pyx_k_model_config), 0, 0, 1, 1},
    {&__pyx_n_s_msg, __pyx_k_msg, sizeof(__pyx_k_msg), 0, 0, 1, 1},
    {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
    {&__pyx_n_s_nan, __pyx_k_nan, sizeof(__pyx_k_nan), 0, 0, 1, 1},
    {&__pyx_n_u_nan, __pyx_k_nan, sizeof(__pyx_k_nan), 0, 1, 0, 1},
    {&__pyx_n_s_network_status, __pyx_k_network_status, sizeof(__pyx_k_network_status), 0, 0, 1, 1},
    {&__pyx_n_s_network_warning, __pyx_k_network_warning, sizeof(__pyx_k_network_warning), 0, 0, 1, 1},
    {&__pyx_n_s_new_ask, __pyx_k_new_ask, sizeof(__pyx_k_new_ask), 0, 0, 1, 1},
    {&__pyx_n_s_new_bid, __pyx_k_new_bid, sizeof(__pyx_k_new_bid), 0, 0, 1, 1},
    {&__pyx_n_s_no_sells, __pyx_k_no_sells, sizeof(__pyx_k_no_sells), 0, 0, 1, 1},
    {&__pyx_n_s_non_hanging, __pyx_k_non_hanging, sizeof(__pyx_k_non_hanging), 0, 0, 1, 1},
    {&__pyx_n_s_normalize, __pyx_k_normalize, sizeof(__pyx_k_normalize), 0, 0, 1, 1},
    {&__pyx_n_s_notify_hb_app, __pyx_k_notify_hb_app, sizeof(__pyx_k_notify_hb_app), 0, 0, 1, 1},
    {&__pyx_n_s_notify_hb_app_with_timestamp, __pyx_k_notify_hb_app_with_timestamp, sizeof(__pyx_k_notify_hb_app_with_timestamp), 0, 0, 1, 1},
    {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
    {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
    {&__pyx_kp_u_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 1, 0, 0},
    {&__pyx_kp_u_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 1, 0, 0},
    {&__pyx_n_s_o, __pyx_k_o, sizeof(__pyx_k_o), 0, 0, 1, 1},
    {&__pyx_kp_u_optimal_ask, __pyx_k_optimal_ask, sizeof(__pyx_k_optimal_ask), 0, 1, 0, 0},
    {&__pyx_n_u_optimal_ask_2, __pyx_k_optimal_ask_2, sizeof(__pyx_k_optimal_ask_2), 0, 1, 0, 1},
    {&__pyx_kp_u_optimal_ask_to_mid, __pyx_k_optimal_ask_to_mid, sizeof(__pyx_k_optimal_ask_to_mid), 0, 1, 0, 0},
    {&__pyx_kp_u_optimal_bid, __pyx_k_optimal_bid, sizeof(__pyx_k_optimal_bid), 0, 1, 0, 0},
    {&__pyx_n_u_optimal_bid_2, __pyx_k_optimal_bid_2, sizeof(__pyx_k_optimal_bid_2), 0, 1, 0, 1},
    {&__pyx_kp_u_optimal_bid_to_mid, __pyx_k_optimal_bid_to_mid, sizeof(__pyx_k_optimal_bid_to_mid), 0, 1, 0, 0},
    {&__pyx_kp_u_optimal_spread, __pyx_k_optimal_spread, sizeof(__pyx_k_optimal_spread), 0, 1, 0, 0},
    {&__pyx_n_u_optimal_spread_2, __pyx_k_optimal_spread_2, sizeof(__pyx_k_optimal_spread_2), 0, 1, 0, 1},
    {&__pyx_kp_u_order, __pyx_k_order, sizeof(__pyx_k_order), 0, 1, 0, 0},
    {&__pyx_n_s_order_2, __pyx_k_order_2, sizeof(__pyx_k_order_2), 0, 0, 1, 1},
    {&__pyx_n_s_order_age, __pyx_k_order_age, sizeof(__pyx_k_order_age), 0, 0, 1, 1},
    {&__pyx_n_s_order_amount, __pyx_k_order_amount, sizeof(__pyx_k_order_amount), 0, 0, 1, 1},
    {&__pyx_n_s_order_amount_shape_factor, __pyx_k_order_amount_shape_factor, sizeof(__pyx_k_order_amount_shape_factor), 0, 0, 1, 1},
    {&__pyx_n_s_order_book, __pyx_k_order_book, sizeof(__pyx_k_order_book), 0, 0, 1, 1},
    {&__pyx_kp_u_order_book_depth_factor, __pyx_k_order_book_depth_factor, sizeof(__pyx_k_order_book_depth_factor), 0, 1, 0, 0},
    {&__pyx_kp_u_order_book_intensity_factor, __pyx_k_order_book_intensity_factor, sizeof(__pyx_k_order_book_intensity_factor), 0, 1, 0, 0},
    {&__pyx_n_s_order_id, __pyx_k_order_id, sizeof(__pyx_k_order_id), 0, 0, 1, 1},
    {&__pyx_n_s_order_levels, __pyx_k_order_levels, sizeof(__pyx_k_order_levels), 0, 0, 1, 1},
    {&__pyx_n_s_order_levels_mode, __pyx_k_order_levels_mode, sizeof(__pyx_k_order_levels_mode), 0, 0, 1, 1},
    {&__pyx_n_s_order_optimization_enabled, __pyx_k_order_optimization_enabled, sizeof(__pyx_k_order_optimization_enabled), 0, 0, 1, 1},
    {&__pyx_n_s_order_override, __pyx_k_order_override, sizeof(__pyx_k_order_override), 0, 0, 1, 1},
    {&__pyx_n_s_order_refresh_time, __pyx_k_order_refresh_time, sizeof(__pyx_k_order_refresh_time), 0, 0, 1, 1},
    {&__pyx_n_s_order_refresh_tolerance, __pyx_k_order_refresh_tolerance, sizeof(__pyx_k_order_refresh_tolerance), 0, 0, 1, 1},
    {&__pyx_n_s_order_refresh_tolerance_pct, __pyx_k_order_refresh_tolerance_pct, sizeof(__pyx_k_order_refresh_tolerance_pct), 0, 0, 1, 1},
    {&__pyx_n_s_orders, __pyx_k_orders, sizeof(__pyx_k_orders), 0, 0, 1, 1},
    {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
    {&__pyx_n_s_p, __pyx_k_p, sizeof(__pyx_k_p), 0, 0, 1, 1},
    {&__pyx_n_s_pandas, __pyx_k_pandas, sizeof(__pyx_k_pandas), 0, 0, 1, 1},
    {&__pyx_n_s_path, __pyx_k_path, sizeof(__pyx_k_path), 0, 0, 1, 1},
    {&__pyx_n_s_pd, __pyx_k_pd, sizeof(__pyx_k_pd), 0, 0, 1, 1},
    {&__pyx_kp_s_pd_DataFrame, __pyx_k_pd_DataFrame, sizeof(__pyx_k_pd_DataFrame), 0, 0, 1, 0},
    {&__pyx_n_s_percent, __pyx_k_percent, sizeof(__pyx_k_percent), 0, 0, 1, 1},
    {&__pyx_n_s_pmm_logger, __pyx_k_pmm_logger, sizeof(__pyx_k_pmm_logger), 0, 0, 1, 1},
    {&__pyx_n_s_price, __pyx_k_price, sizeof(__pyx_k_price), 0, 0, 1, 1},
    {&__pyx_n_s_price_delegate, __pyx_k_price_delegate, sizeof(__pyx_k_price_delegate), 0, 0, 1, 1},
    {&__pyx_n_s_process_tick, __pyx_k_process_tick, sizeof(__pyx_k_process_tick), 0, 0, 1, 1},
    {&__pyx_n_s_proposal, __pyx_k_proposal, sizeof(__pyx_k_proposal), 0, 0, 1, 1},
    {&__pyx_n_s_proposal_prices, __pyx_k_proposal_prices, sizeof(__pyx_k_proposal_prices), 0, 0, 1, 1},
    {&__pyx_n_s_pure_mm_assets_df, __pyx_k_pure_mm_assets_df, sizeof(__pyx_k_pure_mm_assets_df), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
    {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
    {&__pyx_kp_u_q, __pyx_k_q, sizeof(__pyx_k_q), 0, 1, 0, 0},
    {&__pyx_n_s_quantity, __pyx_k_quantity, sizeof(__pyx_k_quantity), 0, 0, 1, 1},
    {&__pyx_n_s_quote_asset, __pyx_k_quote_asset, sizeof(__pyx_k_quote_asset), 0, 0, 1, 1},
    {&__pyx_n_s_quote_balance, __pyx_k_quote_balance, sizeof(__pyx_k_quote_balance), 0, 0, 1, 1},
    {&__pyx_n_s_quote_currency, __pyx_k_quote_currency, sizeof(__pyx_k_quote_currency), 0, 0, 1, 1},
    {&__pyx_n_s_quote_ratio, __pyx_k_quote_ratio, sizeof(__pyx_k_quote_ratio), 0, 0, 1, 1},
    {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
    {&__pyx_n_s_ready, __pyx_k_ready, sizeof(__pyx_k_ready), 0, 0, 1, 1},
    {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
    {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
    {&__pyx_n_s_ref_price, __pyx_k_ref_price, sizeof(__pyx_k_ref_price), 0, 0, 1, 1},
    {&__pyx_n_s_regex, __pyx_k_regex, sizeof(__pyx_k_regex), 0, 0, 1, 1},
    {&__pyx_n_s_register_events, __pyx_k_register_events, sizeof(__pyx_k_register_events), 0, 0, 1, 1},
    {&__pyx_n_s_replace, __pyx_k_replace, sizeof(__pyx_k_replace), 0, 0, 1, 1},
    {&__pyx_kp_u_reservation_price, __pyx_k_reservation_price, sizeof(__pyx_k_reservation_price), 0, 1, 0, 0},
    {&__pyx_n_u_reservation_price_2, __pyx_k_reservation_price_2, sizeof(__pyx_k_reservation_price_2), 0, 1, 0, 1},
    {&__pyx_n_s_result_price, __pyx_k_result_price, sizeof(__pyx_k_result_price), 0, 0, 1, 1},
    {&__pyx_n_s_return, __pyx_k_return, sizeof(__pyx_k_return), 0, 0, 1, 1},
    {&__pyx_n_s_reverse, __pyx_k_reverse, sizeof(__pyx_k_reverse), 0, 0, 1, 1},
    {&__pyx_n_s_risk_factor, __pyx_k_risk_factor, sizeof(__pyx_k_risk_factor), 0, 0, 1, 1},
    {&__pyx_n_u_risk_factor, __pyx_k_risk_factor, sizeof(__pyx_k_risk_factor), 0, 1, 0, 1},
    {&__pyx_kp_u_risk_factor_2, __pyx_k_risk_factor_2, sizeof(__pyx_k_risk_factor_2), 0, 1, 0, 0},
    {&__pyx_n_s_round, __pyx_k_round, sizeof(__pyx_k_round), 0, 0, 1, 1},
    {&__pyx_n_u_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 1, 0, 1},
    {&__pyx_n_s_s_decimal_neg_one, __pyx_k_s_decimal_neg_one, sizeof(__pyx_k_s_decimal_neg_one), 0, 0, 1, 1},
    {&__pyx_n_s_s_decimal_one, __pyx_k_s_decimal_one, sizeof(__pyx_k_s_decimal_one), 0, 0, 1, 1},
    {&__pyx_n_s_s_decimal_zero, __pyx_k_s_decimal_zero, sizeof(__pyx_k_s_decimal_zero), 0, 0, 1, 1},
    {&__pyx_n_s_sampling_length, __pyx_k_sampling_length, sizeof(__pyx_k_sampling_length), 0, 0, 1, 1},
    {&__pyx_n_s_seconds, __pyx_k_seconds, sizeof(__pyx_k_seconds), 0, 0, 1, 1},
    {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
    {&__pyx_kp_s_self__events_cannot_be_converted, __pyx_k_self__events_cannot_be_converted, sizeof(__pyx_k_self__events_cannot_be_converted), 0, 0, 1, 0},
    {&__pyx_n_u_sell, __pyx_k_sell, sizeof(__pyx_k_sell), 0, 1, 0, 1},
    {&__pyx_n_s_sell_order, __pyx_k_sell_order, sizeof(__pyx_k_sell_order), 0, 0, 1, 1},
    {&__pyx_n_s_sells, __pyx_k_sells, sizeof(__pyx_k_sells), 0, 0, 1, 1},
    {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
    {&__pyx_n_s_set_timers, __pyx_k_set_timers, sizeof(__pyx_k_set_timers), 0, 0, 1, 1},
    {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
    {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
    {&__pyx_n_s_should_wait_order_cancel_confirm, __pyx_k_should_wait_order_cancel_confirm, sizeof(__pyx_k_should_wait_order_cancel_confirm), 0, 0, 1, 1},
    {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
    {&__pyx_n_s_sort, __pyx_k_sort, sizeof(__pyx_k_sort), 0, 0, 1, 1},
    {&__pyx_n_s_sorted, __pyx_k_sorted, sizeof(__pyx_k_sorted), 0, 0, 1, 1},
    {&__pyx_n_s_spec, __pyx_k_spec, sizeof(__pyx_k_spec), 0, 0, 1, 1},
    {&__pyx_n_s_split, __pyx_k_split, sizeof(__pyx_k_split), 0, 0, 1, 1},
    {&__pyx_n_s_spread, __pyx_k_spread, sizeof(__pyx_k_spread), 0, 0, 1, 1},
    {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
    {&__pyx_n_s_start_datetime, __pyx_k_start_datetime, sizeof(__pyx_k_start_datetime), 0, 0, 1, 1},
    {&__pyx_n_s_start_time, __pyx_k_start_time, sizeof(__pyx_k_start_time), 0, 0, 1, 1},
    {&__pyx_n_s_start_timestamp, __pyx_k_start_timestamp, sizeof(__pyx_k_start_timestamp), 0, 0, 1, 1},
    {&__pyx_n_s_status_report_interval, __pyx_k_status_report_interval, sizeof(__pyx_k_status_report_interval), 0, 0, 1, 1},
    {&__pyx_n_s_str, __pyx_k_str, sizeof(__pyx_k_str), 0, 0, 1, 1},
    {&__pyx_n_s_strftime, __pyx_k_strftime, sizeof(__pyx_k_strftime), 0, 0, 1, 1},
    {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
    {&__pyx_n_s_super, __pyx_k_super, sizeof(__pyx_k_super), 0, 0, 1, 1},
    {&__pyx_n_u_target_inv, __pyx_k_target_inv, sizeof(__pyx_k_target_inv), 0, 1, 0, 1},
    {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
    {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
    {&__pyx_n_s_ticks_to_be_ready_after, __pyx_k_ticks_to_be_ready_after, sizeof(__pyx_k_ticks_to_be_ready_after), 0, 0, 1, 1},
    {&__pyx_n_s_ticks_to_be_ready_before, __pyx_k_ticks_to_be_ready_before, sizeof(__pyx_k_ticks_to_be_ready_before), 0, 0, 1, 1},
    {&__pyx_kp_u_ticks_to_fill_buffers, __pyx_k_ticks_to_fill_buffers, sizeof(__pyx_k_ticks_to_fill_buffers), 0, 1, 0, 0},
    {&__pyx_n_s_time, __pyx_k_time, sizeof(__pyx_k_time), 0, 0, 1, 1},
    {&__pyx_n_s_time_left, __pyx_k_time_left, sizeof(__pyx_k_time_left), 0, 0, 1, 1},
    {&__pyx_n_s_time_left_fraction, __pyx_k_time_left_fraction, sizeof(__pyx_k_time_left_fraction), 0, 0, 1, 1},
    {&__pyx_n_u_time_left_fraction, __pyx_k_time_left_fraction, sizeof(__pyx_k_time_left_fraction), 0, 1, 0, 1},
    {&__pyx_kp_u_time_until_end_of_trading_cycle, __pyx_k_time_until_end_of_trading_cycle, sizeof(__pyx_k_time_until_end_of_trading_cycle), 0, 1, 0, 0},
    {&__pyx_kp_u_time_until_end_of_trading_cycle_2, __pyx_k_time_until_end_of_trading_cycle_2, sizeof(__pyx_k_time_until_end_of_trading_cycle_2), 0, 1, 0, 0},
    {&__pyx_n_s_timedelta, __pyx_k_timedelta, sizeof(__pyx_k_timedelta), 0, 0, 1, 1},
    {&__pyx_n_s_timestamp, __pyx_k_timestamp, sizeof(__pyx_k_timestamp), 0, 0, 1, 1},
    {&__pyx_n_s_title, __pyx_k_title, sizeof(__pyx_k_title), 0, 0, 1, 1},
    {&__pyx_n_u_title, __pyx_k_title, sizeof(__pyx_k_title), 0, 1, 0, 1},
    {&__pyx_n_s_to_create_orders, __pyx_k_to_create_orders, sizeof(__pyx_k_to_create_orders), 0, 0, 1, 1},
    {&__pyx_n_s_to_csv, __pyx_k_to_csv, sizeof(__pyx_k_to_csv), 0, 0, 1, 1},
    {&__pyx_n_s_to_show_current_pct, __pyx_k_to_show_current_pct, sizeof(__pyx_k_to_show_current_pct), 0, 0, 1, 1},
    {&__pyx_n_s_to_string, __pyx_k_to_string, sizeof(__pyx_k_to_string), 0, 0, 1, 1},
    {&__pyx_n_s_total_in_quote, __pyx_k_total_in_quote, sizeof(__pyx_k_total_in_quote), 0, 0, 1, 1},
    {&__pyx_n_s_trade_type, __pyx_k_trade_type, sizeof(__pyx_k_trade_type), 0, 0, 1, 1},
    {&__pyx_n_s_trading_intensity_buffer_size, __pyx_k_trading_intensity_buffer_size, sizeof(__pyx_k_trading_intensity_buffer_size), 0, 0, 1, 1},
    {&__pyx_n_s_trading_pair, __pyx_k_trading_pair, sizeof(__pyx_k_trading_pair), 0, 0, 1, 1},
    {&__pyx_n_s_typing, __pyx_k_typing, sizeof(__pyx_k_typing), 0, 0, 1, 1},
    {&__pyx_n_s_unit, __pyx_k_unit, sizeof(__pyx_k_unit), 0, 0, 1, 1},
    {&__pyx_n_s_unlink, __pyx_k_unlink, sizeof(__pyx_k_unlink), 0, 0, 1, 1},
    {&__pyx_n_s_unregister_events, __pyx_k_unregister_events, sizeof(__pyx_k_unregister_events), 0, 0, 1, 1},
    {&__pyx_n_s_update_from_config_map, __pyx_k_update_from_config_map, sizeof(__pyx_k_update_from_config_map), 0, 0, 1, 1},
    {&__pyx_n_s_update_strategy_orders_with_equi, __pyx_k_update_strategy_orders_with_equi, sizeof(__pyx_k_update_strategy_orders_with_equi), 0, 0, 1, 1},
    {&__pyx_n_s_upper, __pyx_k_upper, sizeof(__pyx_k_upper), 0, 0, 1, 1},
    {&__pyx_n_s_values, __pyx_k_values, sizeof(__pyx_k_values), 0, 0, 1, 1},
    {&__pyx_kp_u_vol, __pyx_k_vol, sizeof(__pyx_k_vol), 0, 1, 0, 0},
    {&__pyx_n_s_vol_2, __pyx_k_vol_2, sizeof(__pyx_k_vol_2), 0, 0, 1, 1},
    {&__pyx_kp_u_volatility, __pyx_k_volatility, sizeof(__pyx_k_volatility), 0, 1, 0, 0},
    {&__pyx_n_u_volatility_2, __pyx_k_volatility_2, sizeof(__pyx_k_volatility_2), 0, 1, 0, 1},
    {&__pyx_n_s_volatility_buffer_size, __pyx_k_volatility_buffer_size, sizeof(__pyx_k_volatility_buffer_size), 0, 0, 1, 1},
    {&__pyx_n_s_volatility_pct, __pyx_k_volatility_pct, sizeof(__pyx_k_volatility_pct), 0, 0, 1, 1},
    {&__pyx_n_s_warning, __pyx_k_warning, sizeof(__pyx_k_warning), 0, 0, 1, 1},
    {&__pyx_n_s_warning_lines, __pyx_k_warning_lines, sizeof(__pyx_k_warning_lines), 0, 0, 1, 1},
    {&__pyx_n_s_x, __pyx_k_x, sizeof(__pyx_k_x), 0, 0, 1, 1},
    {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
    {0, 0, 0, 0, 0, 0, 0}
  };
  return __Pyx_InitStrings(__pyx_string_tab);
}
/* #### Code section: cached_builtins ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_all = __Pyx_GetBuiltinName(__pyx_n_s_all); if (!__pyx_builtin_all) __PYX_ERR(0, 133, __pyx_L1_error)
  __pyx_builtin_AttributeError = __Pyx_GetBuiltinName(__pyx_n_s_AttributeError); if (!__pyx_builtin_AttributeError) __PYX_ERR(0, 381, __pyx_L1_error)
  __pyx_builtin_round = __Pyx_GetBuiltinName(__pyx_n_s_round); if (!__pyx_builtin_round) __PYX_ERR(0, 451, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 470, __pyx_L1_error)
  __pyx_builtin_max = __Pyx_GetBuiltinName(__pyx_n_s_max); if (!__pyx_builtin_max) __PYX_ERR(0, 532, __pyx_L1_error)
  __pyx_builtin_sorted = __Pyx_GetBuiltinName(__pyx_n_s_sorted); if (!__pyx_builtin_sorted) __PYX_ERR(0, 1055, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 1056, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 1206, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_n_s_super); if (!__pyx_builtin_super) __PYX_ERR(0, 1342, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(2, 984, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":984
 *         __pyx_import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_u_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple_)) __PYX_ERR(2, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "../../miniconda3/envs/hummingbot/lib/python3.12/site-packages/numpy/__init__.cython-30.pxd":990
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_u_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(2, 990, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1358
 * 
 *         if not os.path.exists(self._debug_csv_path):
 *             df_header = pd.DataFrame([('mid_price',             # <<<<<<<<<<<<<<
 *                                        'best_bid',
 *                                        'best_ask',
 */
  __pyx_tuple__16 = PyTuple_Pack(21, __pyx_n_u_mid_price_2, __pyx_n_u_best_bid, __pyx_n_u_best_ask, __pyx_n_u_reservation_price_2, __pyx_n_u_optimal_spread_2, __pyx_n_u_optimal_bid_2, __pyx_n_u_optimal_ask_2, __pyx_kp_u_optimal_bid_to_mid, __pyx_kp_u_optimal_ask_to_mid, __pyx_n_u_current_inv, __pyx_n_u_target_inv, __pyx_n_u_time_left_fraction, __pyx_kp_u_mid_price_std_dev, __pyx_n_u_risk_factor, __pyx_n_u_gamma, __pyx_n_u_alpha_2, __pyx_n_u_kappa_2, __pyx_n_u_eta, __pyx_n_u_volatility_2, __pyx_n_u_mid_price_variance, __pyx_n_u_inventory_target_pct); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 1358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":54
 * 
 * NaN = float("nan")
 * s_decimal_zero = Decimal(0)             # <<<<<<<<<<<<<<
 * s_decimal_neg_one = Decimal(-1)
 * s_decimal_one = Decimal(1)
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_int_0); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":55
 * NaN = float("nan")
 * s_decimal_zero = Decimal(0)
 * s_decimal_neg_one = Decimal(-1)             # <<<<<<<<<<<<<<
 * s_decimal_one = Decimal(1)
 * pmm_logger = None
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_int_neg_1); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":56
 * s_decimal_zero = Decimal(0)
 * s_decimal_neg_one = Decimal(-1)
 * s_decimal_one = Decimal(1)             # <<<<<<<<<<<<<<
 * pmm_logger = None
 * 
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_int_1); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":66
 *     OPTION_LOG_ALL = 0x7fffffffffffffff
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def logger(cls):
 *         global pmm_logger
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_n_s_cls); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);
  __pyx_codeobj__23 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__22, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_logger, 66, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__23)) __PYX_ERR(0, 66, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":73
 *         return pmm_logger
 * 
 *     def init_params(self,             # <<<<<<<<<<<<<<
 *                     config_map: Union[AvellanedaMarketMakingConfigMap, ClientConfigAdapter],
 *                     market_info: MarketTradingPairTuple,
 */
  __pyx_tuple__24 = PyTuple_Pack(8, __pyx_n_s_self, __pyx_n_s_config_map, __pyx_n_s_market_info, __pyx_n_s_logging_options, __pyx_n_s_status_report_interval, __pyx_n_s_hb_app_notification, __pyx_n_s_debug_csv_path, __pyx_n_s_is_debug); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);
  __pyx_codeobj__25 = (PyObject*)__Pyx_PyCode_New(8, 0, 0, 8, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__24, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_init_params, 73, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__25)) __PYX_ERR(0, 73, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":132
 *         self.get_config_map_hanging_orders()
 * 
 *     def all_markets_ready(self):             # <<<<<<<<<<<<<<
 *         return all([market.ready for market in self._sb_markets])
 * 
 */
  __pyx_tuple__26 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_market); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);
  __pyx_codeobj__27 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__26, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_all_markets_ready, 132, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__27)) __PYX_ERR(0, 132, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":305
 *         self._end_time = value
 * 
 *     def get_price(self) -> float:             # <<<<<<<<<<<<<<
 *         return self.get_mid_price()
 * 
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_n_s_self); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);
  __pyx_codeobj__29 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_get_price, 305, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__29)) __PYX_ERR(0, 305, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":308
 *         return self.get_mid_price()
 * 
 *     def get_last_price(self) -> float:             # <<<<<<<<<<<<<<
 *         return self._market_info.get_last_price()
 * 
 */
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_get_last_price, 308, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(0, 308, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":311
 *         return self._market_info.get_last_price()
 * 
 *     def get_mid_price(self) -> float:             # <<<<<<<<<<<<<<
 *         return self.c_get_mid_price()
 * 
 */
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_get_mid_price, 311, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(0, 311, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":352
 *         return self._hanging_orders_tracker
 * 
 *     def update_from_config_map(self):             # <<<<<<<<<<<<<<
 *         self.get_config_map_execution_mode()
 *         self.get_config_map_hanging_orders()
 */
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_update_from_config_map, 352, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 352, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":357
 *         self.get_config_map_indicators()
 * 
 *     def get_config_map_execution_mode(self):             # <<<<<<<<<<<<<<
 *         try:
 *             execution_mode = self._config_map.execution_timeframe_mode.title
 */
  __pyx_tuple__33 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_execution_mode, __pyx_n_s_execution_timeframe, __pyx_n_s_start_time, __pyx_n_s_end_time, __pyx_n_s_execution_state); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_get_config_map_execution_mode, 357, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 357, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":385
 *             pass
 * 
 *     def get_config_map_hanging_orders(self):             # <<<<<<<<<<<<<<
 *         if self._config_map.hanging_orders_mode.title == TrackHangingOrdersModel.model_config['title']:
 *             hanging_orders_enabled = True
 */
  __pyx_tuple__35 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_hanging_orders_enabled, __pyx_n_s_hanging_orders_cancel_pct); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__35, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_get_config_map_hanging_orders, 385, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 385, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":405
 *             self._hanging_orders_tracker.hanging_orders_cancel_pct = hanging_orders_cancel_pct / Decimal('100')
 * 
 *     def get_config_map_indicators(self):             # <<<<<<<<<<<<<<
 *         volatility_buffer_size = self._config_map.volatility_buffer_size
 *         trading_intensity_buffer_size = self._config_map.trading_intensity_buffer_size
 */
  __pyx_tuple__37 = PyTuple_Pack(5, __pyx_n_s_self, __pyx_n_s_volatility_buffer_size, __pyx_n_s_trading_intensity_buffer_size, __pyx_n_s_ticks_to_be_ready_after, __pyx_n_s_ticks_to_be_ready_before); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__37, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_get_config_map_indicators, 405, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 405, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":438
 *             self._ticks_to_be_ready = 0
 * 
 *     def pure_mm_assets_df(self, to_show_current_pct: bool) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self._price_delegate.get_price_by_type(PriceType.MidPrice)
 */
  __pyx_tuple__39 = PyTuple_Pack(17, __pyx_n_s_self, __pyx_n_s_to_show_current_pct, __pyx_n_s_market, __pyx_n_s_trading_pair, __pyx_n_s_base_asset, __pyx_n_s_quote_asset, __pyx_n_s_price, __pyx_n_s_base_balance, __pyx_n_s_quote_balance, __pyx_n_s_available_base_balance, __pyx_n_s_available_quote_balance, __pyx_n_s_base_value, __pyx_n_s_total_in_quote, __pyx_n_s_base_ratio, __pyx_n_s_quote_ratio, __pyx_n_s_data, __pyx_n_s_df); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 17, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_pure_mm_assets_df, 438, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(0, 438, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":460
 *         return df
 * 
 *     def active_orders_df(self) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self.get_price()
 */
  __pyx_tuple__41 = PyTuple_Pack(20, __pyx_n_s_self, __pyx_n_s_market, __pyx_n_s_trading_pair, __pyx_n_s_base_asset, __pyx_n_s_quote_asset, __pyx_n_s_price, __pyx_n_s_active_orders, __pyx_n_s_no_sells, __pyx_n_s_columns, __pyx_n_s_data, __pyx_n_s_lvl_buy, __pyx_n_s_lvl_sell, __pyx_n_s_idx, __pyx_n_s_order_2, __pyx_n_s_is_hanging_order, __pyx_n_s_level, __pyx_n_s_spread, __pyx_n_s_age, __pyx_n_s_amount_orig, __pyx_n_s_o); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 20, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_active_orders_df, 460, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 460, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":499
 *         return pd.DataFrame(data=data, columns=columns)
 * 
 *     def market_status_data_frame(self, market_trading_pair_tuples: List[MarketTradingPairTuple]) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         markets_data = []
 *         markets_columns = ["Exchange", "Market", "Best Bid", "Best Ask", f"MidPrice"]
 */
  __pyx_tuple__43 = PyTuple_Pack(10, __pyx_n_s_self, __pyx_n_s_market_trading_pair_tuples, __pyx_n_s_markets_data, __pyx_n_s_markets_columns, __pyx_n_s_market_books, __pyx_n_s_market, __pyx_n_s_trading_pair, __pyx_n_s_bid_price, __pyx_n_s_ask_price, __pyx_n_s_ref_price); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 10, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_market_status_data_frame, 499, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 499, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":520
 *         return pd.DataFrame(data=markets_data, columns=markets_columns).replace(np.nan, '', regex=True)
 * 
 *     def format_status(self) -> str:             # <<<<<<<<<<<<<<
 *         if not self._all_markets_ready:
 *             return "Market connectors are not ready."
 */
  __pyx_tuple__45 = PyTuple_Pack(12, __pyx_n_s_self, __pyx_n_s_lines, __pyx_n_s_warning_lines, __pyx_n_s_markets_df, __pyx_n_s_assets_df, __pyx_n_s_first_col_length, __pyx_n_s_df_lines, __pyx_n_s_df, __pyx_n_s_volatility_pct, __pyx_n_s_line, __pyx_n_s_line, __pyx_n_s_line); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 12, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__45, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_format_status, 520, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(0, 520, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":563
 *         return "\n".join(lines)
 * 
 *     def execute_orders_proposal(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_execute_orders_proposal(proposal)
 * 
 */
  __pyx_tuple__47 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_proposal); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_execute_orders_proposal, 563, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(0, 563, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":566
 *         return self.c_execute_orders_proposal(proposal)
 * 
 *     def cancel_order(self, order_id: str):             # <<<<<<<<<<<<<<
 *         return self.c_cancel_order(self._market_info, order_id)
 * 
 */
  __pyx_tuple__49 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_order_id); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__49, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_cancel_order, 566, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(0, 566, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":586
 *         self._execution_state.time_left = self._execution_state.closing_time
 * 
 *     def start(self, clock: Clock, timestamp: float):             # <<<<<<<<<<<<<<
 *         self.c_start(clock, timestamp)
 * 
 */
  __pyx_tuple__51 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_clock, __pyx_n_s_timestamp); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);
  __pyx_codeobj__52 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__51, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_start, 586, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__52)) __PYX_ERR(0, 586, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":645
 *             self._last_timestamp = timestamp
 * 
 *     def process_tick(self, timestamp: float):             # <<<<<<<<<<<<<<
 *         proposal = None
 *         # Trading is allowed
 */
  __pyx_tuple__53 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_timestamp, __pyx_n_s_proposal); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);
  __pyx_codeobj__54 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__53, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_process_tick, 645, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__54)) __PYX_ERR(0, 645, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":682
 *         inventory_in_base = quote_balance / price + base_balance
 * 
 *     def collect_market_variables(self, timestamp: float):             # <<<<<<<<<<<<<<
 *         self.c_collect_market_variables(timestamp)
 * 
 */
  __pyx_tuple__55 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_timestamp); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);
  __pyx_codeobj__56 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__55, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_collect_market_variables, 682, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__56)) __PYX_ERR(0, 682, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":692
 *         return market.c_get_price(trading_pair, True) - market.c_get_price(trading_pair, False)
 * 
 *     def get_spread(self):             # <<<<<<<<<<<<<<
 *         return self.c_get_spread()
 * 
 */
  __pyx_codeobj__57 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_get_spread, 692, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__57)) __PYX_ERR(0, 692, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":695
 *         return self.c_get_spread()
 * 
 *     def get_volatility(self):             # <<<<<<<<<<<<<<
 *         vol = Decimal(str(self._avg_vol.current_value))
 *         return vol
 */
  __pyx_tuple__58 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_vol_2); if (unlikely(!__pyx_tuple__58)) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__58);
  __Pyx_GIVEREF(__pyx_tuple__58);
  __pyx_codeobj__59 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__58, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_get_volatility, 695, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__59)) __PYX_ERR(0, 695, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":710
 *                                f"kappa={self._kappa:.4f}")
 * 
 *     def measure_order_book_liquidity(self):             # <<<<<<<<<<<<<<
 *         return self.c_measure_order_book_liquidity()
 * 
 */
  __pyx_codeobj__60 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_measure_order_book_liquidity, 710, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__60)) __PYX_ERR(0, 710, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":779
 *                                    f"optimal_ask={((self._reservation_price + self._optimal_spread / 2) - price) / price * 100:.4f}%")
 * 
 *     def calculate_reservation_price_and_optimal_spread(self):             # <<<<<<<<<<<<<<
 *         return self.c_calculate_reservation_price_and_optimal_spread()
 * 
 */
  __pyx_codeobj__61 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_calculate_reservation_price_and, 779, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__61)) __PYX_ERR(0, 779, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":806
 *         return market.c_quantize_order_amount(trading_pair, Decimal(str(target_inventory_amount)))
 * 
 *     def calculate_target_inventory(self) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_calculate_target_inventory()
 * 
 */
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_calculate_target_inventory, 806, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(0, 806, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":831
 *         return inventory_value_base
 * 
 *     def calculate_inventory(self) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_calculate_inventory()
 * 
 */
  __pyx_codeobj__63 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_calculate_inventory, 831, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__63)) __PYX_ERR(0, 831, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":840
 *         return self._trading_intensity.is_sampling_buffer_changed or self._avg_vol.is_sampling_buffer_changed
 * 
 *     def is_algorithm_ready(self) -> bool:             # <<<<<<<<<<<<<<
 *         return self.c_is_algorithm_ready()
 * 
 */
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_is_algorithm_ready, 840, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) __PYX_ERR(0, 840, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":843
 *         return self.c_is_algorithm_ready()
 * 
 *     def is_algorithm_changed(self) -> bool:             # <<<<<<<<<<<<<<
 *         return self.c_is_algorithm_changed()
 * 
 */
  __pyx_codeobj__65 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_is_algorithm_changed, 843, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__65)) __PYX_ERR(0, 843, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":846
 *         return self.c_is_algorithm_changed()
 * 
 *     def _get_level_spreads(self):             # <<<<<<<<<<<<<<
 *         level_step = ((self._optimal_spread / 2) / 100) * self.level_distances
 * 
 */
  __pyx_tuple__66 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_level_step, __pyx_n_s_bid_level_spreads, __pyx_n_s_ask_level_spreads, __pyx_n_s_i, __pyx_n_s_i); if (unlikely(!__pyx_tuple__66)) __PYX_ERR(0, 846, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__66);
  __Pyx_GIVEREF(__pyx_tuple__66);
  __pyx_codeobj__67 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__66, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_get_level_spreads, 846, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__67)) __PYX_ERR(0, 846, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":875
 *         return buys, sells
 * 
 *     def create_proposal_based_on_order_override(self) -> Tuple[List[Proposal], List[Proposal]]:             # <<<<<<<<<<<<<<
 *         return self._create_proposal_based_on_order_override()
 * 
 */
  __pyx_codeobj__68 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_create_proposal_based_on_order_o, 875, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__68)) __PYX_ERR(0, 875, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":896
 *         return buys, sells
 * 
 *     def create_proposal_based_on_order_levels(self):             # <<<<<<<<<<<<<<
 *         return self._create_proposal_based_on_order_levels()
 * 
 */
  __pyx_codeobj__69 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_create_proposal_based_on_order_l, 896, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__69)) __PYX_ERR(0, 896, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":915
 *         return buys, sells
 * 
 *     def create_basic_proposal(self):             # <<<<<<<<<<<<<<
 *         return self._create_basic_proposal()
 * 
 */
  __pyx_codeobj__70 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_create_basic_proposal, 915, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__70)) __PYX_ERR(0, 915, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":936
 *         return Proposal(buys, sells)
 * 
 *     def create_base_proposal(self):             # <<<<<<<<<<<<<<
 *         return self.c_create_base_proposal()
 * 
 */
  __pyx_codeobj__71 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_create_base_proposal, 936, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__71)) __PYX_ERR(0, 936, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":957
 *         return base_balance, quote_balance
 * 
 *     def get_adjusted_available_balance(self, orders: List[LimitOrder]):             # <<<<<<<<<<<<<<
 *         return self.c_get_adjusted_available_balance(orders)
 * 
 */
  __pyx_tuple__72 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_orders); if (unlikely(!__pyx_tuple__72)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__72);
  __Pyx_GIVEREF(__pyx_tuple__72);
  __pyx_codeobj__73 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__72, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_get_adjusted_available_balance, 957, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__73)) __PYX_ERR(0, 957, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":967
 *             self.c_apply_add_transaction_costs(proposal)
 * 
 *     def apply_order_price_modifiers(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         self.c_apply_order_price_modifiers(proposal)
 * 
 */
  __pyx_codeobj__74 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_apply_order_price_modifiers, 967, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__74)) __PYX_ERR(0, 967, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":970
 *         self.c_apply_order_price_modifiers(proposal)
 * 
 *     def apply_budget_constraint(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_apply_budget_constraint(proposal)
 * 
 */
  __pyx_codeobj__75 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_apply_budget_constraint, 970, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__75)) __PYX_ERR(0, 970, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":973
 *         return self.c_apply_budget_constraint(proposal)
 * 
 *     def adjusted_available_balance_for_orders_budget_constrain(self):             # <<<<<<<<<<<<<<
 *         candidate_hanging_orders = self.hanging_orders_tracker.candidate_hanging_orders_from_pairs()
 *         non_hanging = []
 */
  __pyx_tuple__76 = PyTuple_Pack(6, __pyx_n_s_self, __pyx_n_s_candidate_hanging_orders, __pyx_n_s_non_hanging, __pyx_n_s_all_orders, __pyx_n_s_all_non_hanging_orders, __pyx_n_s_order_2); if (unlikely(!__pyx_tuple__76)) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__76);
  __Pyx_GIVEREF(__pyx_tuple__76);
  __pyx_codeobj__77 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 6, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__76, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_adjusted_available_balance_for_o, 973, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__77)) __PYX_ERR(0, 973, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1025
 *         proposal.sells = [o for o in proposal.sells if o.size > 0]
 * 
 *     def apply_budget_constraint(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_apply_budget_constraint(proposal)
 * 
 */
  __pyx_codeobj__78 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_apply_budget_constraint, 1025, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__78)) __PYX_ERR(0, 1025, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1078
 *                     proposal.sells[i].price = market.c_quantize_order_price(self.trading_pair, price_below_ask)
 * 
 *     def apply_order_optimization(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_apply_order_optimization(proposal)
 * 
 */
  __pyx_codeobj__79 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_apply_order_optimization, 1078, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__79)) __PYX_ERR(0, 1078, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1114
 *                     proposal.sells = [o for o in proposal.sells if o.size > 0]
 * 
 *     def apply_order_amount_eta_transformation(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         self.c_apply_order_amount_eta_transformation(proposal)
 * 
 */
  __pyx_codeobj__80 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_apply_order_amount_eta_transform, 1114, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__80)) __PYX_ERR(0, 1114, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1131
 *             sell.price = market.c_quantize_order_price(self.trading_pair, price)
 * 
 *     def apply_add_transaction_costs(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         self.c_apply_add_transaction_costs(proposal)
 * 
 */
  __pyx_codeobj__81 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_apply_add_transaction_costs, 1131, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__81)) __PYX_ERR(0, 1131, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1212
 *         return True
 * 
 *     def is_within_tolerance(self, current_prices: List[Decimal], proposal_prices: List[Decimal]) -> bool:             # <<<<<<<<<<<<<<
 *         return self.c_is_within_tolerance(current_prices, proposal_prices)
 * 
 */
  __pyx_tuple__82 = PyTuple_Pack(3, __pyx_n_s_self, __pyx_n_s_current_prices, __pyx_n_s_proposal_prices); if (unlikely(!__pyx_tuple__82)) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__82);
  __Pyx_GIVEREF(__pyx_tuple__82);
  __pyx_codeobj__83 = (PyObject*)__Pyx_PyCode_New(3, 0, 0, 3, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__82, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_is_within_tolerance, 1212, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__83)) __PYX_ERR(0, 1212, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1255
 *             self.c_set_timers()
 * 
 *     def cancel_active_orders(self, proposal: Proposal = None):             # <<<<<<<<<<<<<<
 *         return self.c_cancel_active_orders(proposal)
 * 
 */
  __pyx_codeobj__84 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_cancel_active_orders, 1255, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__84)) __PYX_ERR(0, 1255, __pyx_L1_error)
  __pyx_tuple__85 = PyTuple_Pack(1, Py_None); if (unlikely(!__pyx_tuple__85)) __PYX_ERR(0, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__85);
  __Pyx_GIVEREF(__pyx_tuple__85);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1268
 *                 and len(non_hanging_orders_non_cancelled) == 0)
 * 
 *     def to_create_orders(self, proposal: Proposal) -> bool:             # <<<<<<<<<<<<<<
 *         return self.c_to_create_orders(proposal)
 * 
 */
  __pyx_codeobj__86 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_to_create_orders, 1268, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__86)) __PYX_ERR(0, 1268, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1327
 *             self.c_set_timers()
 * 
 *     def execute_orders_proposal(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         self.c_execute_orders_proposal(proposal)
 * 
 */
  __pyx_codeobj__87 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_execute_orders_proposal, 1327, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__87)) __PYX_ERR(0, 1327, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1337
 *             self._cancel_timestamp = min(self._create_timestamp, next_cycle)
 * 
 *     def set_timers(self):             # <<<<<<<<<<<<<<
 *         self.c_set_timers()
 * 
 */
  __pyx_codeobj__88 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_set_timers, 1337, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__88)) __PYX_ERR(0, 1337, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1340
 *         self.c_set_timers()
 * 
 *     def notify_hb_app(self, msg: str):             # <<<<<<<<<<<<<<
 *         if self._hb_app_notification:
 *             super().notify_hb_app(msg)
 */
  __pyx_tuple__89 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_msg); if (unlikely(!__pyx_tuple__89)) __PYX_ERR(0, 1340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__89);
  __Pyx_GIVEREF(__pyx_tuple__89);
  __pyx_codeobj__90 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__89, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_notify_hb_app, 1340, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__90)) __PYX_ERR(0, 1340, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1344
 *             super().notify_hb_app(msg)
 * 
 *     def dump_debug_variables(self):             # <<<<<<<<<<<<<<
 *         market = self._market_info.market
 *         mid_price = self.get_price()
 */
  __pyx_tuple__91 = PyTuple_Pack(13, __pyx_n_s_self, __pyx_n_s_market, __pyx_n_s_mid_price_2, __pyx_n_s_spread, __pyx_n_s_best_ask, __pyx_n_s_new_ask, __pyx_n_s_best_bid, __pyx_n_s_new_bid, __pyx_n_s_vol_2, __pyx_n_s_mid_price_variance, __pyx_n_s_df_header, __pyx_n_s_time_left_fraction, __pyx_n_s_df); if (unlikely(!__pyx_tuple__91)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__91);
  __Pyx_GIVEREF(__pyx_tuple__91);
  __pyx_codeobj__92 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 13, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__91, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_hummingbot_strategy_avellaneda_m_3, __pyx_n_s_dump_debug_variables, 1344, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__92)) __PYX_ERR(0, 1344, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_codeobj__93 = (PyObject*)__Pyx_PyCode_New(1, 0, 0, 1, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__28, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_reduce_cython, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__93)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 */
  __pyx_tuple__94 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_pyx_state); if (unlikely(!__pyx_tuple__94)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__94);
  __Pyx_GIVEREF(__pyx_tuple__94);
  __pyx_codeobj__95 = (PyObject*)__Pyx_PyCode_New(2, 0, 0, 2, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__94, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_setstate_cython, 3, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__95)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static CYTHON_SMALL_CODE int __Pyx_InitConstants(void) {
  __pyx_umethod_PyUnicode_Type_upper.type = (PyObject*)&PyUnicode_Type;
  __pyx_umethod_PyUnicode_Type_upper.method_name = &__pyx_n_s_upper;
  if (__Pyx_CreateStringTabAndInitStrings() < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_float_100_0 = PyFloat_FromDouble(100.0); if (unlikely(!__pyx_float_100_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_16 = PyInt_FromLong(16); if (unlikely(!__pyx_int_16)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_32 = PyInt_FromLong(32); if (unlikely(!__pyx_int_32)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_100 = PyInt_FromLong(100); if (unlikely(!__pyx_int_100)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_900 = PyInt_FromLong(900); if (unlikely(!__pyx_int_900)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_0x7fffffffffffffff = PyInt_FromString((char *)"0x7fffffffffffffff", 0, 0); if (unlikely(!__pyx_int_0x7fffffffffffffff)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_globals ### */

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  return 0;
}
/* #### Code section: init_module ### */

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_t_1 = PyImport_ImportModule("hummingbot.strategy.strategy_base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.strategy.strategy_base", "StrategyBase", sizeof(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_8strategy_13strategy_base_StrategyBase = (struct __pyx_vtabstruct_10hummingbot_8strategy_13strategy_base_StrategyBase*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase); if (unlikely(!__pyx_vtabptr_10hummingbot_8strategy_13strategy_base_StrategyBase)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy = &__pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.__pyx_base = *__pyx_vtabptr_10hummingbot_8strategy_13strategy_base_StrategyBase;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.__pyx_base.__pyx_base.c_start = (PyObject *(*)(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *, struct __pyx_obj_10hummingbot_4core_5clock_Clock *, double))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_start;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.__pyx_base.__pyx_base.c_stop = (PyObject *(*)(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *, struct __pyx_obj_10hummingbot_4core_5clock_Clock *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_stop;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.__pyx_base.__pyx_base.c_tick = (PyObject *(*)(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator *, double))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_tick;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.__pyx_base.c_did_fill_order = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_did_fill_order;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.__pyx_base.c_did_complete_buy_order = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_did_complete_buy_order;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.__pyx_base.c_did_complete_sell_order = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_13strategy_base_StrategyBase *, PyObject *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_did_complete_sell_order;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_get_mid_price = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_get_mid_price;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy._create_proposal_based_on_order_levels = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy__create_proposal_based_on_order_levels;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy._create_proposal_based_on_order_override = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy__create_proposal_based_on_order_override;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy._create_basic_proposal = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy__create_basic_proposal;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_create_base_proposal = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_create_base_proposal;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_get_adjusted_available_balance = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_get_adjusted_available_balance;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_apply_order_price_modifiers = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_apply_order_price_modifiers;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_apply_order_amount_eta_transformation = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_apply_order_amount_eta_transformation;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_apply_budget_constraint = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_apply_budget_constraint;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_apply_order_optimization = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_apply_order_optimization;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_apply_add_transaction_costs = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_apply_add_transaction_costs;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_is_within_tolerance = (int (*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *, PyObject *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_is_within_tolerance;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_cancel_active_orders = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_cancel_active_orders;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_cancel_active_orders_on_max_age_limit = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_cancel_active_orders_on_max_age_limit;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_to_create_orders = (int (*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_to_create_orders;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_execute_orders_proposal = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_execute_orders_proposal;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_set_timers = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_set_timers;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_get_spread = (double (*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_get_spread;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_collect_market_variables = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, double))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_collect_market_variables;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_is_algorithm_ready = (int (*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_is_algorithm_ready;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_is_algorithm_changed = (int (*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_is_algorithm_changed;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_measure_order_book_liquidity = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_measure_order_book_liquidity;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_calculate_reservation_price_and_optimal_spread = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_calculate_reservation_price_and_optimal_spread;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_calculate_target_inventory = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_calculate_target_inventory;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_calculate_inventory = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_calculate_inventory;
  __pyx_vtable_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy.c_did_complete_order = (PyObject *(*)(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy *, PyObject *))__pyx_f_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_c_did_complete_order;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_2 = PyTuple_Pack(1, (PyObject *)__pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy_spec, __pyx_t_2);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (unlikely(!__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy)) __PYX_ERR(0, 60, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy_spec, __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy) < 0) __PYX_ERR(0, 60, __pyx_L1_error)
  #else
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy = &__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy->tp_base = __pyx_ptype_10hummingbot_8strategy_13strategy_base_StrategyBase;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy) < 0) __PYX_ERR(0, 60, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy->tp_dictoffset && __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy->tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_vtabptr_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy) < 0) __PYX_ERR(0, 60, __pyx_L1_error)
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_MergeVtables(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy) < 0) __PYX_ERR(0, 60, __pyx_L1_error)
  #endif
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_AvellanedaMarketMakingStrategy, (PyObject *) __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy) < 0) __PYX_ERR(0, 60, __pyx_L1_error)
  if (__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy->tp_weaklistoffset == 0) __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy->tp_weaklistoffset = offsetof(struct __pyx_obj_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_base.__pyx_base.__pyx_base.__weakref__);
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if (__Pyx_setup_reduce((PyObject *) __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy) < 0) __PYX_ERR(0, 60, __pyx_L1_error)
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start_spec, NULL); if (unlikely(!__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start)) __PYX_ERR(0, 569, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start_spec, __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start) < 0) __PYX_ERR(0, 569, __pyx_L1_error)
  #else
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start = &__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start) < 0) __PYX_ERR(0, 569, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start->tp_dictoffset && __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct____pyx_base____pyx_base__c_start->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr)) __PYX_ERR(0, 580, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr_spec, __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 580, __pyx_L1_error)
  #else
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr = &__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 580, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr->tp_dictoffset && __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_1_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal_spec, NULL); if (unlikely(!__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal)) __PYX_ERR(0, 1271, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal_spec, __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal) < 0) __PYX_ERR(0, 1271, __pyx_L1_error)
  #else
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal = &__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal) < 0) __PYX_ERR(0, 1271, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal->tp_dictoffset && __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_2_c_execute_orders_proposal->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr)) __PYX_ERR(0, 1298, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr_spec, __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr) < 0) __PYX_ERR(0, 1298, __pyx_L1_error)
  #else
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr = &__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr) < 0) __PYX_ERR(0, 1298, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr->tp_dictoffset && __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_3_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  #if CYTHON_USE_TYPE_SPECS
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr_spec, NULL); if (unlikely(!__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr)) __PYX_ERR(0, 1321, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr_spec, __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr) < 0) __PYX_ERR(0, 1321, __pyx_L1_error)
  #else
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr = &__pyx_type_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr) < 0) __PYX_ERR(0, 1321, __pyx_L1_error)
  #endif
  #if PY_MAJOR_VERSION < 3
  __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr->tp_print = 0;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr->tp_dictoffset && __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making___pyx_scope_struct_4_genexpr->tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
  }
  #endif
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.event.event_listener"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.event.event_listener", "EventListener", sizeof(struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_5event_14event_listener_EventListener),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener) __PYX_ERR(3, 6, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_5event_14event_listener_EventListener = (struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_listener_EventListener*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_4core_5event_14event_listener_EventListener); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_5event_14event_listener_EventListener)) __PYX_ERR(3, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.pubsub"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_6pubsub_PubSub = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.pubsub", "PubSub", sizeof(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_6pubsub_PubSub),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_6pubsub_PubSub) __PYX_ERR(4, 17, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_6pubsub_PubSub = (struct __pyx_vtabstruct_10hummingbot_4core_6pubsub_PubSub*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_4core_6pubsub_PubSub); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_6pubsub_PubSub)) __PYX_ERR(4, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.data_type.order_book_query_result"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.data_type.order_book_query_result", "OrderBookQueryResult", sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_OrderBookQueryResult) __PYX_ERR(5, 3, __pyx_L1_error)
  __pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.data_type.order_book_query_result", "ClientOrderBookQueryResult", sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_9data_type_23order_book_query_result_ClientOrderBookQueryResult) __PYX_ERR(5, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType_3_0_12(__pyx_t_1, __Pyx_BUILTIN_MODULE_NAME, "type", 
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyTypeObject),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(PyTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyHeapTypeObject),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_7cpython_4type_type) __PYX_ERR(6, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("numpy"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "dtype", sizeof(PyArray_Descr), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyArray_Descr),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_dtype) __PYX_ERR(2, 202, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "flatiter", sizeof(PyArrayIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyArrayIterObject),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_flatiter) __PYX_ERR(2, 225, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "broadcast", sizeof(PyArrayMultiIterObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyArrayMultiIterObject),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_broadcast) __PYX_ERR(2, 229, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "ndarray", sizeof(PyArrayObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyArrayObject),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_ndarray) __PYX_ERR(2, 238, __pyx_L1_error)
  __pyx_ptype_5numpy_generic = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "generic", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_generic) __PYX_ERR(2, 809, __pyx_L1_error)
  __pyx_ptype_5numpy_number = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "number", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_number) __PYX_ERR(2, 811, __pyx_L1_error)
  __pyx_ptype_5numpy_integer = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "integer", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_integer) __PYX_ERR(2, 813, __pyx_L1_error)
  __pyx_ptype_5numpy_signedinteger = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "signedinteger", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_signedinteger) __PYX_ERR(2, 815, __pyx_L1_error)
  __pyx_ptype_5numpy_unsignedinteger = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "unsignedinteger", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_unsignedinteger) __PYX_ERR(2, 817, __pyx_L1_error)
  __pyx_ptype_5numpy_inexact = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "inexact", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_inexact) __PYX_ERR(2, 819, __pyx_L1_error)
  __pyx_ptype_5numpy_floating = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "floating", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_floating) __PYX_ERR(2, 821, __pyx_L1_error)
  __pyx_ptype_5numpy_complexfloating = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "complexfloating", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_complexfloating) __PYX_ERR(2, 823, __pyx_L1_error)
  __pyx_ptype_5numpy_flexible = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "flexible", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_flexible) __PYX_ERR(2, 825, __pyx_L1_error)
  __pyx_ptype_5numpy_character = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "character", sizeof(PyObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyObject),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_5numpy_character) __PYX_ERR(2, 827, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType_3_0_12(__pyx_t_1, "numpy", "ufunc", sizeof(PyUFuncObject), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(PyUFuncObject),__Pyx_ImportType_CheckSize_Ignore_3_0_12); if (!__pyx_ptype_5numpy_ufunc) __PYX_ERR(2, 866, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.data_type.order_book"); if (unlikely(!__pyx_t_1)) __PYX_ERR(7, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.data_type.order_book", "OrderBook", sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_9data_type_10order_book_OrderBook),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook) __PYX_ERR(7, 12, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_9data_type_10order_book_OrderBook = (struct __pyx_vtabstruct_10hummingbot_4core_9data_type_10order_book_OrderBook*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_4core_9data_type_10order_book_OrderBook); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_9data_type_10order_book_OrderBook)) __PYX_ERR(7, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.strategy.__utils__.trailing_indicators.trading_intensity"); if (unlikely(!__pyx_t_1)) __PYX_ERR(8, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.strategy.__utils__.trailing_indicators.trading_intensity", "TradingIntensityIndicator", sizeof(struct __pyx_obj_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator) __PYX_ERR(8, 11, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator = (struct __pyx_vtabstruct_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator); if (unlikely(!__pyx_vtabptr_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradingIntensityIndicator)) __PYX_ERR(8, 11, __pyx_L1_error)
  __pyx_ptype_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradesForwarder = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.strategy.__utils__.trailing_indicators.trading_intensity", "TradesForwarder", sizeof(struct __pyx_obj_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradesForwarder), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradesForwarder),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradesForwarder) __PYX_ERR(8, 28, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradesForwarder = (struct __pyx_vtabstruct_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradesForwarder*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradesForwarder); if (unlikely(!__pyx_vtabptr_10hummingbot_8strategy_9__utils___19trailing_indicators_17trading_intensity_TradesForwarder)) __PYX_ERR(8, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.clock"); if (unlikely(!__pyx_t_1)) __PYX_ERR(9, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_5clock_Clock = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.clock", "Clock", sizeof(struct __pyx_obj_10hummingbot_4core_5clock_Clock), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_5clock_Clock),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_5clock_Clock) __PYX_ERR(9, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.time_iterator"); if (unlikely(!__pyx_t_1)) __PYX_ERR(10, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.time_iterator", "TimeIterator", sizeof(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_13time_iterator_TimeIterator),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator) __PYX_ERR(10, 7, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_13time_iterator_TimeIterator = (struct __pyx_vtabstruct_10hummingbot_4core_13time_iterator_TimeIterator*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_4core_13time_iterator_TimeIterator); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_13time_iterator_TimeIterator)) __PYX_ERR(10, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.data_type.limit_order"); if (unlikely(!__pyx_t_1)) __PYX_ERR(11, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.data_type.limit_order", "LimitOrder", sizeof(struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_9data_type_11limit_order_LimitOrder),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder) __PYX_ERR(11, 6, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_9data_type_11limit_order_LimitOrder = (struct __pyx_vtabstruct_10hummingbot_4core_9data_type_11limit_order_LimitOrder*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_4core_9data_type_11limit_order_LimitOrder); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_9data_type_11limit_order_LimitOrder)) __PYX_ERR(11, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.strategy.order_tracker"); if (unlikely(!__pyx_t_1)) __PYX_ERR(12, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_8strategy_13order_tracker_OrderTracker = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.strategy.order_tracker", "OrderTracker", sizeof(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_8strategy_13order_tracker_OrderTracker),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_8strategy_13order_tracker_OrderTracker) __PYX_ERR(12, 7, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_8strategy_13order_tracker_OrderTracker = (struct __pyx_vtabstruct_10hummingbot_8strategy_13order_tracker_OrderTracker*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_8strategy_13order_tracker_OrderTracker); if (unlikely(!__pyx_vtabptr_10hummingbot_8strategy_13order_tracker_OrderTracker)) __PYX_ERR(12, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.event.event_reporter"); if (unlikely(!__pyx_t_1)) __PYX_ERR(13, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.event.event_reporter", "EventReporter", sizeof(struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_5event_14event_reporter_EventReporter),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter) __PYX_ERR(13, 4, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_5event_14event_reporter_EventReporter = (struct __pyx_vtabstruct_10hummingbot_4core_5event_14event_reporter_EventReporter*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_4core_5event_14event_reporter_EventReporter); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_5event_14event_reporter_EventReporter)) __PYX_ERR(13, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.event.event_logger"); if (unlikely(!__pyx_t_1)) __PYX_ERR(14, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.event.event_logger", "EventLogger", sizeof(struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_5event_12event_logger_EventLogger),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger) __PYX_ERR(14, 4, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_5event_12event_logger_EventLogger = (struct __pyx_vtabstruct_10hummingbot_4core_5event_12event_logger_EventLogger*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_4core_5event_12event_logger_EventLogger); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_5event_12event_logger_EventLogger)) __PYX_ERR(14, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.core.network_iterator"); if (unlikely(!__pyx_t_1)) __PYX_ERR(15, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.core.network_iterator", "NetworkIterator", sizeof(struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_4core_16network_iterator_NetworkIterator),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator) __PYX_ERR(15, 6, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_4core_16network_iterator_NetworkIterator = (struct __pyx_vtabstruct_10hummingbot_4core_16network_iterator_NetworkIterator*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_4core_16network_iterator_NetworkIterator); if (unlikely(!__pyx_vtabptr_10hummingbot_4core_16network_iterator_NetworkIterator)) __PYX_ERR(15, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.connector.connector_base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(16, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.connector.connector_base", "ConnectorBase", sizeof(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_9connector_14connector_base_ConnectorBase),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase) __PYX_ERR(16, 5, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_9connector_14connector_base_ConnectorBase = (struct __pyx_vtabstruct_10hummingbot_9connector_14connector_base_ConnectorBase*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_9connector_14connector_base_ConnectorBase); if (unlikely(!__pyx_vtabptr_10hummingbot_9connector_14connector_base_ConnectorBase)) __PYX_ERR(16, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("hummingbot.connector.exchange_base"); if (unlikely(!__pyx_t_1)) __PYX_ERR(17, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase = __Pyx_ImportType_3_0_12(__pyx_t_1, "hummingbot.connector.exchange_base", "ExchangeBase", sizeof(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase), __PYX_GET_STRUCT_ALIGNMENT_3_0_12(struct __pyx_obj_10hummingbot_9connector_13exchange_base_ExchangeBase),__Pyx_ImportType_CheckSize_Warn_3_0_12); if (!__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase) __PYX_ERR(17, 10, __pyx_L1_error)
  __pyx_vtabptr_10hummingbot_9connector_13exchange_base_ExchangeBase = (struct __pyx_vtabstruct_10hummingbot_9connector_13exchange_base_ExchangeBase*)__Pyx_GetVtable(__pyx_ptype_10hummingbot_9connector_13exchange_base_ExchangeBase); if (unlikely(!__pyx_vtabptr_10hummingbot_9connector_13exchange_base_ExchangeBase)) __PYX_ERR(17, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_avellaneda_market_making(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_avellaneda_market_making},
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "avellaneda_market_making",
      0, /* m_doc */
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      0, /* m_size */
    #elif CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstate), /* m_size */
    #else
      -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif
#endif

#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initavellaneda_market_making(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initavellaneda_market_making(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_avellaneda_market_making(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_avellaneda_market_making(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *module, const char* from_name, const char* to_name, int allow_none)
#else
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
#endif
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
#if CYTHON_COMPILING_IN_LIMITED_API
            result = PyModule_AddObject(module, to_name, value);
#else
            result = PyDict_SetItemString(moddict, to_name, value);
#endif
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    moddict = module;
#else
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
#endif
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_avellaneda_market_making(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  double __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'avellaneda_market_making' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("avellaneda_market_making", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #elif CYTHON_USE_MODULE_STATE
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  {
    int add_module_result = PyState_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "avellaneda_market_making" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #endif
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = __Pyx_PyImport_AddModuleRef((const char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_avellaneda_market_making(void)", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(WITH_THREAD) && PY_VERSION_HEX < 0x030700F0 && defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  PyEval_InitThreads();
  #endif
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_hummingbot__strategy__avellaneda_market_making__avellaneda_market_making) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making")) {
      if (unlikely((PyDict_SetItemString(modules, "hummingbot.strategy.avellaneda_market_making.avellaneda_market_making", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely((__Pyx_modinit_type_init_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code() < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1
 * import datetime             # <<<<<<<<<<<<<<
 * import logging
 * import os
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_datetime, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_datetime, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":2
 * import datetime
 * import logging             # <<<<<<<<<<<<<<
 * import os
 * import time
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_logging, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_logging, __pyx_t_2) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":3
 * import datetime
 * import logging
 * import os             # <<<<<<<<<<<<<<
 * import time
 * from decimal import Decimal
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_os, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_2) < 0) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":4
 * import logging
 * import os
 * import time             # <<<<<<<<<<<<<<
 * from decimal import Decimal
 * from math import ceil, floor, isnan
 */
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_n_s_time, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_time, __pyx_t_2) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":5
 * import os
 * import time
 * from decimal import Decimal             # <<<<<<<<<<<<<<
 * from math import ceil, floor, isnan
 * from typing import Dict, List, Tuple, Union
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_Decimal);
  __Pyx_GIVEREF(__pyx_n_s_Decimal);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Decimal)) __PYX_ERR(0, 5, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_decimal, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Decimal, __pyx_t_2) < 0) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":6
 * import time
 * from decimal import Decimal
 * from math import ceil, floor, isnan             # <<<<<<<<<<<<<<
 * from typing import Dict, List, Tuple, Union
 * 
 */
  __pyx_t_3 = PyList_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_ceil);
  __Pyx_GIVEREF(__pyx_n_s_ceil);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_ceil)) __PYX_ERR(0, 6, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_floor);
  __Pyx_GIVEREF(__pyx_n_s_floor);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_floor)) __PYX_ERR(0, 6, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_isnan);
  __Pyx_GIVEREF(__pyx_n_s_isnan);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 2, __pyx_n_s_isnan)) __PYX_ERR(0, 6, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_math, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_ceil); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ceil, __pyx_t_3) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_floor); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_floor, __pyx_t_3) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_isnan); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_isnan, __pyx_t_3) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":7
 * from decimal import Decimal
 * from math import ceil, floor, isnan
 * from typing import Dict, List, Tuple, Union             # <<<<<<<<<<<<<<
 * 
 * import numpy as np
 */
  __pyx_t_2 = PyList_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_Dict);
  __Pyx_GIVEREF(__pyx_n_s_Dict);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Dict)) __PYX_ERR(0, 7, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_List);
  __Pyx_GIVEREF(__pyx_n_s_List);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_List)) __PYX_ERR(0, 7, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_Tuple);
  __Pyx_GIVEREF(__pyx_n_s_Tuple);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 2, __pyx_n_s_Tuple)) __PYX_ERR(0, 7, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_Union);
  __Pyx_GIVEREF(__pyx_n_s_Union);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 3, __pyx_n_s_Union)) __PYX_ERR(0, 7, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_typing, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Dict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Dict, __pyx_t_2) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_List); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_List, __pyx_t_2) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Tuple); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Tuple, __pyx_t_2) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_Union); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Union, __pyx_t_2) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":9
 * from typing import Dict, List, Tuple, Union
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * import pandas as pd
 * 
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_numpy, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_3) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":10
 * 
 * import numpy as np
 * import pandas as pd             # <<<<<<<<<<<<<<
 * 
 * from hummingbot.connector.exchange_base import ExchangeBase
 */
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_n_s_pandas, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pd, __pyx_t_3) < 0) __PYX_ERR(0, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":12
 * import pandas as pd
 * 
 * from hummingbot.connector.exchange_base import ExchangeBase             # <<<<<<<<<<<<<<
 * from hummingbot.connector.exchange_base cimport ExchangeBase
 * from hummingbot.core.clock cimport Clock
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_ExchangeBase);
  __Pyx_GIVEREF(__pyx_n_s_ExchangeBase);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_ExchangeBase)) __PYX_ERR(0, 12, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_hummingbot_connector_exchange_ba, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":16
 * from hummingbot.core.clock cimport Clock
 * 
 * from hummingbot.client.config.config_helpers import ClientConfigAdapter             # <<<<<<<<<<<<<<
 * from hummingbot.core.data_type.common import (
 *     OrderType,
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_ClientConfigAdapter);
  __Pyx_GIVEREF(__pyx_n_s_ClientConfigAdapter);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_ClientConfigAdapter)) __PYX_ERR(0, 16, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_hummingbot_client_config_config, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_ClientConfigAdapter); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ClientConfigAdapter, __pyx_t_2) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":18
 * from hummingbot.client.config.config_helpers import ClientConfigAdapter
 * from hummingbot.core.data_type.common import (
 *     OrderType,             # <<<<<<<<<<<<<<
 *     PriceType,
 *     TradeType
 */
  __pyx_t_3 = PyList_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_OrderType);
  __Pyx_GIVEREF(__pyx_n_s_OrderType);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_OrderType)) __PYX_ERR(0, 18, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_PriceType);
  __Pyx_GIVEREF(__pyx_n_s_PriceType);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_PriceType)) __PYX_ERR(0, 18, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_TradeType);
  __Pyx_GIVEREF(__pyx_n_s_TradeType);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 2, __pyx_n_s_TradeType)) __PYX_ERR(0, 18, __pyx_L1_error);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":17
 * 
 * from hummingbot.client.config.config_helpers import ClientConfigAdapter
 * from hummingbot.core.data_type.common import (             # <<<<<<<<<<<<<<
 *     OrderType,
 *     PriceType,
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_hummingbot_core_data_type_common, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_OrderType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OrderType, __pyx_t_3) < 0) __PYX_ERR(0, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_PriceType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PriceType, __pyx_t_3) < 0) __PYX_ERR(0, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_TradeType); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TradeType, __pyx_t_3) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":23
 * )
 * from hummingbot.core.data_type.limit_order cimport LimitOrder
 * from hummingbot.core.data_type.limit_order import LimitOrder             # <<<<<<<<<<<<<<
 * from hummingbot.core.network_iterator import NetworkStatus
 * from hummingbot.core.utils import map_df_to_str
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_LimitOrder);
  __Pyx_GIVEREF(__pyx_n_s_LimitOrder);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_LimitOrder)) __PYX_ERR(0, 23, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_hummingbot_core_data_type_limit, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":24
 * from hummingbot.core.data_type.limit_order cimport LimitOrder
 * from hummingbot.core.data_type.limit_order import LimitOrder
 * from hummingbot.core.network_iterator import NetworkStatus             # <<<<<<<<<<<<<<
 * from hummingbot.core.utils import map_df_to_str
 * from hummingbot.strategy.__utils__.trailing_indicators.instant_volatility import InstantVolatilityIndicator
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_NetworkStatus);
  __Pyx_GIVEREF(__pyx_n_s_NetworkStatus);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_NetworkStatus)) __PYX_ERR(0, 24, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_hummingbot_core_network_iterator, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_NetworkStatus); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NetworkStatus, __pyx_t_3) < 0) __PYX_ERR(0, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":25
 * from hummingbot.core.data_type.limit_order import LimitOrder
 * from hummingbot.core.network_iterator import NetworkStatus
 * from hummingbot.core.utils import map_df_to_str             # <<<<<<<<<<<<<<
 * from hummingbot.strategy.__utils__.trailing_indicators.instant_volatility import InstantVolatilityIndicator
 * from hummingbot.strategy.__utils__.trailing_indicators.trading_intensity import TradingIntensityIndicator
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_map_df_to_str);
  __Pyx_GIVEREF(__pyx_n_s_map_df_to_str);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_map_df_to_str)) __PYX_ERR(0, 25, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_hummingbot_core_utils, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_map_df_to_str); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_map_df_to_str, __pyx_t_2) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":26
 * from hummingbot.core.network_iterator import NetworkStatus
 * from hummingbot.core.utils import map_df_to_str
 * from hummingbot.strategy.__utils__.trailing_indicators.instant_volatility import InstantVolatilityIndicator             # <<<<<<<<<<<<<<
 * from hummingbot.strategy.__utils__.trailing_indicators.trading_intensity import TradingIntensityIndicator
 * from hummingbot.strategy.avellaneda_market_making.avellaneda_market_making_config_map_pydantic import (
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_InstantVolatilityIndicator);
  __Pyx_GIVEREF(__pyx_n_s_InstantVolatilityIndicator);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_InstantVolatilityIndicator)) __PYX_ERR(0, 26, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_hummingbot_strategy___utils___tr, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_InstantVolatilityIndicator); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_InstantVolatilityIndicator, __pyx_t_3) < 0) __PYX_ERR(0, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":27
 * from hummingbot.core.utils import map_df_to_str
 * from hummingbot.strategy.__utils__.trailing_indicators.instant_volatility import InstantVolatilityIndicator
 * from hummingbot.strategy.__utils__.trailing_indicators.trading_intensity import TradingIntensityIndicator             # <<<<<<<<<<<<<<
 * from hummingbot.strategy.avellaneda_market_making.avellaneda_market_making_config_map_pydantic import (
 *     AvellanedaMarketMakingConfigMap,
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_TradingIntensityIndicator);
  __Pyx_GIVEREF(__pyx_n_s_TradingIntensityIndicator);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_TradingIntensityIndicator)) __PYX_ERR(0, 27, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_hummingbot_strategy___utils___tr_2, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":29
 * from hummingbot.strategy.__utils__.trailing_indicators.trading_intensity import TradingIntensityIndicator
 * from hummingbot.strategy.avellaneda_market_making.avellaneda_market_making_config_map_pydantic import (
 *     AvellanedaMarketMakingConfigMap,             # <<<<<<<<<<<<<<
 *     DailyBetweenTimesModel,
 *     FromDateToDateModel,
 */
  __pyx_t_3 = PyList_New(5); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_AvellanedaMarketMakingConfigMap);
  __Pyx_GIVEREF(__pyx_n_s_AvellanedaMarketMakingConfigMap);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_AvellanedaMarketMakingConfigMap)) __PYX_ERR(0, 29, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_DailyBetweenTimesModel);
  __Pyx_GIVEREF(__pyx_n_s_DailyBetweenTimesModel);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_DailyBetweenTimesModel)) __PYX_ERR(0, 29, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_FromDateToDateModel);
  __Pyx_GIVEREF(__pyx_n_s_FromDateToDateModel);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 2, __pyx_n_s_FromDateToDateModel)) __PYX_ERR(0, 29, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_MultiOrderLevelModel);
  __Pyx_GIVEREF(__pyx_n_s_MultiOrderLevelModel);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 3, __pyx_n_s_MultiOrderLevelModel)) __PYX_ERR(0, 29, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_TrackHangingOrdersModel);
  __Pyx_GIVEREF(__pyx_n_s_TrackHangingOrdersModel);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 4, __pyx_n_s_TrackHangingOrdersModel)) __PYX_ERR(0, 29, __pyx_L1_error);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":28
 * from hummingbot.strategy.__utils__.trailing_indicators.instant_volatility import InstantVolatilityIndicator
 * from hummingbot.strategy.__utils__.trailing_indicators.trading_intensity import TradingIntensityIndicator
 * from hummingbot.strategy.avellaneda_market_making.avellaneda_market_making_config_map_pydantic import (             # <<<<<<<<<<<<<<
 *     AvellanedaMarketMakingConfigMap,
 *     DailyBetweenTimesModel,
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_hummingbot_strategy_avellaneda_m_2, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_AvellanedaMarketMakingConfigMap); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_AvellanedaMarketMakingConfigMap, __pyx_t_3) < 0) __PYX_ERR(0, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_DailyBetweenTimesModel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DailyBetweenTimesModel, __pyx_t_3) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_FromDateToDateModel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FromDateToDateModel, __pyx_t_3) < 0) __PYX_ERR(0, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_MultiOrderLevelModel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MultiOrderLevelModel, __pyx_t_3) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_TrackHangingOrdersModel); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TrackHangingOrdersModel, __pyx_t_3) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":36
 * )
 * from hummingbot.strategy.conditional_execution_state import (
 *     RunAlwaysExecutionState,             # <<<<<<<<<<<<<<
 *     RunInTimeConditionalExecutionState
 * )
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_RunAlwaysExecutionState);
  __Pyx_GIVEREF(__pyx_n_s_RunAlwaysExecutionState);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_RunAlwaysExecutionState)) __PYX_ERR(0, 36, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_RunInTimeConditionalExecutionSta);
  __Pyx_GIVEREF(__pyx_n_s_RunInTimeConditionalExecutionSta);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_RunInTimeConditionalExecutionSta)) __PYX_ERR(0, 36, __pyx_L1_error);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":35
 *     TrackHangingOrdersModel,
 * )
 * from hummingbot.strategy.conditional_execution_state import (             # <<<<<<<<<<<<<<
 *     RunAlwaysExecutionState,
 *     RunInTimeConditionalExecutionState
 */
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_hummingbot_strategy_conditional, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_RunAlwaysExecutionState); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RunAlwaysExecutionState, __pyx_t_2) < 0) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_RunInTimeConditionalExecutionSta); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RunInTimeConditionalExecutionSta, __pyx_t_2) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":40
 * )
 * from hummingbot.strategy.data_types import (
 *     PriceSize,             # <<<<<<<<<<<<<<
 *     Proposal,
 * )
 */
  __pyx_t_3 = PyList_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_PriceSize);
  __Pyx_GIVEREF(__pyx_n_s_PriceSize);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_PriceSize)) __PYX_ERR(0, 40, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_Proposal);
  __Pyx_GIVEREF(__pyx_n_s_Proposal);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 1, __pyx_n_s_Proposal)) __PYX_ERR(0, 40, __pyx_L1_error);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":39
 *     RunInTimeConditionalExecutionState
 * )
 * from hummingbot.strategy.data_types import (             # <<<<<<<<<<<<<<
 *     PriceSize,
 *     Proposal,
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_hummingbot_strategy_data_types, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_PriceSize); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PriceSize, __pyx_t_3) < 0) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Proposal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Proposal, __pyx_t_3) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":44
 * )
 * from hummingbot.strategy.hanging_orders_tracker import (
 *     CreatedPairOfOrders,             # <<<<<<<<<<<<<<
 *     HangingOrdersTracker,
 * )
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_CreatedPairOfOrders);
  __Pyx_GIVEREF(__pyx_n_s_CreatedPairOfOrders);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_CreatedPairOfOrders)) __PYX_ERR(0, 44, __pyx_L1_error);
  __Pyx_INCREF(__pyx_n_s_HangingOrdersTracker);
  __Pyx_GIVEREF(__pyx_n_s_HangingOrdersTracker);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_HangingOrdersTracker)) __PYX_ERR(0, 44, __pyx_L1_error);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":43
 *     Proposal,
 * )
 * from hummingbot.strategy.hanging_orders_tracker import (             # <<<<<<<<<<<<<<
 *     CreatedPairOfOrders,
 *     HangingOrdersTracker,
 */
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_hummingbot_strategy_hanging_orde, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_CreatedPairOfOrders); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CreatedPairOfOrders, __pyx_t_2) < 0) __PYX_ERR(0, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_HangingOrdersTracker); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HangingOrdersTracker, __pyx_t_2) < 0) __PYX_ERR(0, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":47
 *     HangingOrdersTracker,
 * )
 * from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple             # <<<<<<<<<<<<<<
 * from hummingbot.strategy.order_book_asset_price_delegate import OrderBookAssetPriceDelegate
 * from hummingbot.strategy.order_tracker cimport OrderTracker
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_MarketTradingPairTuple);
  __Pyx_GIVEREF(__pyx_n_s_MarketTradingPairTuple);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_MarketTradingPairTuple)) __PYX_ERR(0, 47, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_hummingbot_strategy_market_tradi, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_MarketTradingPairTuple); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MarketTradingPairTuple, __pyx_t_3) < 0) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":48
 * )
 * from hummingbot.strategy.market_trading_pair_tuple import MarketTradingPairTuple
 * from hummingbot.strategy.order_book_asset_price_delegate import OrderBookAssetPriceDelegate             # <<<<<<<<<<<<<<
 * from hummingbot.strategy.order_tracker cimport OrderTracker
 * from hummingbot.strategy.strategy_base import StrategyBase
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_OrderBookAssetPriceDelegate);
  __Pyx_GIVEREF(__pyx_n_s_OrderBookAssetPriceDelegate);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_OrderBookAssetPriceDelegate)) __PYX_ERR(0, 48, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_hummingbot_strategy_order_book_a, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_OrderBookAssetPriceDelegate); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_OrderBookAssetPriceDelegate, __pyx_t_2) < 0) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":50
 * from hummingbot.strategy.order_book_asset_price_delegate import OrderBookAssetPriceDelegate
 * from hummingbot.strategy.order_tracker cimport OrderTracker
 * from hummingbot.strategy.strategy_base import StrategyBase             # <<<<<<<<<<<<<<
 * from hummingbot.strategy.utils import order_age
 * 
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_StrategyBase);
  __Pyx_GIVEREF(__pyx_n_s_StrategyBase);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_StrategyBase)) __PYX_ERR(0, 50, __pyx_L1_error);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_hummingbot_strategy_strategy_bas, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":51
 * from hummingbot.strategy.order_tracker cimport OrderTracker
 * from hummingbot.strategy.strategy_base import StrategyBase
 * from hummingbot.strategy.utils import order_age             # <<<<<<<<<<<<<<
 * 
 * NaN = float("nan")
 */
  __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_order_age);
  __Pyx_GIVEREF(__pyx_n_s_order_age);
  if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_order_age)) __PYX_ERR(0, 51, __pyx_L1_error);
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_hummingbot_strategy_utils, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_order_age); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_order_age, __pyx_t_2) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":53
 * from hummingbot.strategy.utils import order_age
 * 
 * NaN = float("nan")             # <<<<<<<<<<<<<<
 * s_decimal_zero = Decimal(0)
 * s_decimal_neg_one = Decimal(-1)
 */
  __pyx_t_4 = __Pyx_PyUnicode_AsDouble(__pyx_n_u_nan); if (unlikely(__pyx_t_4 == ((double)((double)-1)) && PyErr_Occurred())) __PYX_ERR(0, 53, __pyx_L1_error)
  __pyx_t_3 = PyFloat_FromDouble(__pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NaN, __pyx_t_3) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":54
 * 
 * NaN = float("nan")
 * s_decimal_zero = Decimal(0)             # <<<<<<<<<<<<<<
 * s_decimal_neg_one = Decimal(-1)
 * s_decimal_one = Decimal(1)
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_s_decimal_zero, __pyx_t_2) < 0) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":55
 * NaN = float("nan")
 * s_decimal_zero = Decimal(0)
 * s_decimal_neg_one = Decimal(-1)             # <<<<<<<<<<<<<<
 * s_decimal_one = Decimal(1)
 * pmm_logger = None
 */
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_s_decimal_neg_one, __pyx_t_3) < 0) __PYX_ERR(0, 55, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":56
 * s_decimal_zero = Decimal(0)
 * s_decimal_neg_one = Decimal(-1)
 * s_decimal_one = Decimal(1)             # <<<<<<<<<<<<<<
 * pmm_logger = None
 * 
 */
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_n_s_Decimal); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_s_decimal_one, __pyx_t_2) < 0) __PYX_ERR(0, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":57
 * s_decimal_neg_one = Decimal(-1)
 * s_decimal_one = Decimal(1)
 * pmm_logger = None             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pmm_logger, Py_None) < 0) __PYX_ERR(0, 57, __pyx_L1_error)

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":61
 * 
 * cdef class AvellanedaMarketMakingStrategy(StrategyBase):
 *     OPTION_LOG_CREATE_ORDER = 1 << 3             # <<<<<<<<<<<<<<
 *     OPTION_LOG_MAKER_ORDER_FILLED = 1 << 4
 *     OPTION_LOG_STATUS_REPORT = 1 << 5
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_OPTION_LOG_CREATE_ORDER, __pyx_int_8) < 0) __PYX_ERR(0, 61, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":62
 * cdef class AvellanedaMarketMakingStrategy(StrategyBase):
 *     OPTION_LOG_CREATE_ORDER = 1 << 3
 *     OPTION_LOG_MAKER_ORDER_FILLED = 1 << 4             # <<<<<<<<<<<<<<
 *     OPTION_LOG_STATUS_REPORT = 1 << 5
 *     OPTION_LOG_ALL = 0x7fffffffffffffff
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_OPTION_LOG_MAKER_ORDER_FILLED, __pyx_int_16) < 0) __PYX_ERR(0, 62, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":63
 *     OPTION_LOG_CREATE_ORDER = 1 << 3
 *     OPTION_LOG_MAKER_ORDER_FILLED = 1 << 4
 *     OPTION_LOG_STATUS_REPORT = 1 << 5             # <<<<<<<<<<<<<<
 *     OPTION_LOG_ALL = 0x7fffffffffffffff
 * 
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_OPTION_LOG_STATUS_REPORT, __pyx_int_32) < 0) __PYX_ERR(0, 63, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":64
 *     OPTION_LOG_MAKER_ORDER_FILLED = 1 << 4
 *     OPTION_LOG_STATUS_REPORT = 1 << 5
 *     OPTION_LOG_ALL = 0x7fffffffffffffff             # <<<<<<<<<<<<<<
 * 
 *     @classmethod
 */
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_OPTION_LOG_ALL, __pyx_int_0x7fffffffffffffff) < 0) __PYX_ERR(0, 64, __pyx_L1_error)
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":66
 *     OPTION_LOG_ALL = 0x7fffffffffffffff
 * 
 *     @classmethod             # <<<<<<<<<<<<<<
 *     def logger(cls):
 *         global pmm_logger
 */
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_1logger, __Pyx_CYFUNCTION_CLASSMETHOD | __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_l, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__23)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_logger, __pyx_t_2) < 0) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);
  __Pyx_GetNameInClass(__pyx_t_2, (PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_logger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Method_ClassMethod(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_logger, __pyx_t_3) < 0) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":76
 *                     config_map: Union[AvellanedaMarketMakingConfigMap, ClientConfigAdapter],
 *                     market_info: MarketTradingPairTuple,
 *                     logging_options: int = OPTION_LOG_ALL,             # <<<<<<<<<<<<<<
 *                     status_report_interval: float = 900,
 *                     hb_app_notification: bool = False,
 */
  __Pyx_GetNameInClass(__pyx_t_3, (PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_OPTION_LOG_ALL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_k__3 = __pyx_t_3;
  __Pyx_GIVEREF(__pyx_t_3);
  __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":73
 *         return pmm_logger
 * 
 *     def init_params(self,             # <<<<<<<<<<<<<<
 *                     config_map: Union[AvellanedaMarketMakingConfigMap, ClientConfigAdapter],
 *                     market_info: MarketTradingPairTuple,
 */
  __Pyx_GetNameInClass(__pyx_t_3, (PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_OPTION_LOG_ALL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":80
 *                     hb_app_notification: bool = False,
 *                     debug_csv_path: str = '',
 *                     is_debug: bool = False,             # <<<<<<<<<<<<<<
 *                     ):
 *         self._sb_order_tracker = OrderTracker()
 */
  __pyx_t_2 = PyTuple_New(5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3)) __PYX_ERR(0, 73, __pyx_L1_error);
  __Pyx_INCREF(__pyx_int_900);
  __Pyx_GIVEREF(__pyx_int_900);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_int_900)) __PYX_ERR(0, 73, __pyx_L1_error);
  __Pyx_INCREF(Py_False);
  __Pyx_GIVEREF(Py_False);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 2, Py_False)) __PYX_ERR(0, 73, __pyx_L1_error);
  __Pyx_INCREF(__pyx_kp_u__4);
  __Pyx_GIVEREF(__pyx_kp_u__4);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_kp_u__4)) __PYX_ERR(0, 73, __pyx_L1_error);
  __Pyx_INCREF(Py_False);
  __Pyx_GIVEREF(Py_False);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 4, Py_False)) __PYX_ERR(0, 73, __pyx_L1_error);
  __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":73
 *         return pmm_logger
 * 
 *     def init_params(self,             # <<<<<<<<<<<<<<
 *                     config_map: Union[AvellanedaMarketMakingConfigMap, ClientConfigAdapter],
 *                     market_info: MarketTradingPairTuple,
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_config_map, __pyx_kp_s_Union_AvellanedaMarketMakingConf) < 0) __PYX_ERR(0, 73, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_market_info, __pyx_n_s_MarketTradingPairTuple) < 0) __PYX_ERR(0, 73, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_logging_options, __pyx_n_s_int) < 0) __PYX_ERR(0, 73, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_status_report_interval, __pyx_n_s_float) < 0) __PYX_ERR(0, 73, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_hb_app_notification, __pyx_n_s_bool) < 0) __PYX_ERR(0, 73, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_debug_csv_path, __pyx_n_s_str) < 0) __PYX_ERR(0, 73, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_is_debug, __pyx_n_s_bool) < 0) __PYX_ERR(0, 73, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_3init_params, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_i, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__25)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_t_2);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_init_params, __pyx_t_5) < 0) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":132
 *         self.get_config_map_hanging_orders()
 * 
 *     def all_markets_ready(self):             # <<<<<<<<<<<<<<
 *         return all([market.ready for market in self._sb_markets])
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_5all_markets_ready, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_a, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__27)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_all_markets_ready, __pyx_t_5) < 0) __PYX_ERR(0, 132, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":305
 *         self._end_time = value
 * 
 *     def get_price(self) -> float:             # <<<<<<<<<<<<<<
 *         return self.get_mid_price()
 * 
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_float) < 0) __PYX_ERR(0, 305, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_7get_price, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_g, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__29)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_get_price, __pyx_t_3) < 0) __PYX_ERR(0, 305, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":308
 *         return self.get_mid_price()
 * 
 *     def get_last_price(self) -> float:             # <<<<<<<<<<<<<<
 *         return self._market_info.get_last_price()
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_float) < 0) __PYX_ERR(0, 308, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_9get_last_price, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_g_2, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_get_last_price, __pyx_t_5) < 0) __PYX_ERR(0, 308, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":311
 *         return self._market_info.get_last_price()
 * 
 *     def get_mid_price(self) -> float:             # <<<<<<<<<<<<<<
 *         return self.c_get_mid_price()
 * 
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_float) < 0) __PYX_ERR(0, 311, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_11get_mid_price, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_g_3, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__31)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_get_mid_price, __pyx_t_3) < 0) __PYX_ERR(0, 311, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":352
 *         return self._hanging_orders_tracker
 * 
 *     def update_from_config_map(self):             # <<<<<<<<<<<<<<
 *         self.get_config_map_execution_mode()
 *         self.get_config_map_hanging_orders()
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_13update_from_config_map, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_u, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__32)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_update_from_config_map, __pyx_t_3) < 0) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":357
 *         self.get_config_map_indicators()
 * 
 *     def get_config_map_execution_mode(self):             # <<<<<<<<<<<<<<
 *         try:
 *             execution_mode = self._config_map.execution_timeframe_mode.title
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_15get_config_map_execution_mode, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_g_4, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__34)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_get_config_map_execution_mode, __pyx_t_3) < 0) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":385
 *             pass
 * 
 *     def get_config_map_hanging_orders(self):             # <<<<<<<<<<<<<<
 *         if self._config_map.hanging_orders_mode.title == TrackHangingOrdersModel.model_config['title']:
 *             hanging_orders_enabled = True
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_17get_config_map_hanging_orders, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_g_5, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__36)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_get_config_map_hanging_orders, __pyx_t_3) < 0) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":405
 *             self._hanging_orders_tracker.hanging_orders_cancel_pct = hanging_orders_cancel_pct / Decimal('100')
 * 
 *     def get_config_map_indicators(self):             # <<<<<<<<<<<<<<
 *         volatility_buffer_size = self._config_map.volatility_buffer_size
 *         trading_intensity_buffer_size = self._config_map.trading_intensity_buffer_size
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_19get_config_map_indicators, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_g_6, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__38)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_get_config_map_indicators, __pyx_t_3) < 0) __PYX_ERR(0, 405, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":438
 *             self._ticks_to_be_ready = 0
 * 
 *     def pure_mm_assets_df(self, to_show_current_pct: bool) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self._price_delegate.get_price_by_type(PriceType.MidPrice)
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_to_show_current_pct, __pyx_n_s_bool) < 0) __PYX_ERR(0, 438, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_kp_s_pd_DataFrame) < 0) __PYX_ERR(0, 438, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_21pure_mm_assets_df, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_p, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__40)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_pure_mm_assets_df, __pyx_t_5) < 0) __PYX_ERR(0, 438, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":460
 *         return df
 * 
 *     def active_orders_df(self) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         market, trading_pair, base_asset, quote_asset = self._market_info
 *         price = self.get_price()
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_kp_s_pd_DataFrame) < 0) __PYX_ERR(0, 460, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_23active_orders_df, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_a_2, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_active_orders_df, __pyx_t_3) < 0) __PYX_ERR(0, 460, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":499
 *         return pd.DataFrame(data=data, columns=columns)
 * 
 *     def market_status_data_frame(self, market_trading_pair_tuples: List[MarketTradingPairTuple]) -> pd.DataFrame:             # <<<<<<<<<<<<<<
 *         markets_data = []
 *         markets_columns = ["Exchange", "Market", "Best Bid", "Best Ask", f"MidPrice"]
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_market_trading_pair_tuples, __pyx_kp_s_List_MarketTradingPairTuple) < 0) __PYX_ERR(0, 499, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_kp_s_pd_DataFrame) < 0) __PYX_ERR(0, 499, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_25market_status_data_frame, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_m, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_market_status_data_frame, __pyx_t_5) < 0) __PYX_ERR(0, 499, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":520
 *         return pd.DataFrame(data=markets_data, columns=markets_columns).replace(np.nan, '', regex=True)
 * 
 *     def format_status(self) -> str:             # <<<<<<<<<<<<<<
 *         if not self._all_markets_ready:
 *             return "Market connectors are not ready."
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_str) < 0) __PYX_ERR(0, 520, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_27format_status, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_f, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__46)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_format_status, __pyx_t_3) < 0) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":563
 *         return "\n".join(lines)
 * 
 *     def execute_orders_proposal(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_execute_orders_proposal(proposal)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_proposal, __pyx_n_s_Proposal) < 0) __PYX_ERR(0, 563, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_29execute_orders_proposal, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_e, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__48)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_execute_orders_proposal, __pyx_t_5) < 0) __PYX_ERR(0, 563, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":566
 *         return self.c_execute_orders_proposal(proposal)
 * 
 *     def cancel_order(self, order_id: str):             # <<<<<<<<<<<<<<
 *         return self.c_cancel_order(self._market_info, order_id)
 * 
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_order_id, __pyx_n_s_str) < 0) __PYX_ERR(0, 566, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_31cancel_order, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_c_4, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__50)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_cancel_order, __pyx_t_3) < 0) __PYX_ERR(0, 566, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":586
 *         self._execution_state.time_left = self._execution_state.closing_time
 * 
 *     def start(self, clock: Clock, timestamp: float):             # <<<<<<<<<<<<<<
 *         self.c_start(clock, timestamp)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_clock, __pyx_n_s_Clock) < 0) __PYX_ERR(0, 586, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_timestamp, __pyx_n_s_float) < 0) __PYX_ERR(0, 586, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_33start, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_s, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__52)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_start, __pyx_t_5) < 0) __PYX_ERR(0, 586, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":645
 *             self._last_timestamp = timestamp
 * 
 *     def process_tick(self, timestamp: float):             # <<<<<<<<<<<<<<
 *         proposal = None
 *         # Trading is allowed
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_timestamp, __pyx_n_s_float) < 0) __PYX_ERR(0, 645, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_35process_tick, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_p_2, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__54)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_process_tick, __pyx_t_3) < 0) __PYX_ERR(0, 645, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":682
 *         inventory_in_base = quote_balance / price + base_balance
 * 
 *     def collect_market_variables(self, timestamp: float):             # <<<<<<<<<<<<<<
 *         self.c_collect_market_variables(timestamp)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_timestamp, __pyx_n_s_float) < 0) __PYX_ERR(0, 682, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_37collect_market_variables, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_c_5, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__56)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_collect_market_variables, __pyx_t_5) < 0) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":692
 *         return market.c_get_price(trading_pair, True) - market.c_get_price(trading_pair, False)
 * 
 *     def get_spread(self):             # <<<<<<<<<<<<<<
 *         return self.c_get_spread()
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_39get_spread, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_g_7, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__57)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 692, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_get_spread, __pyx_t_5) < 0) __PYX_ERR(0, 692, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":695
 *         return self.c_get_spread()
 * 
 *     def get_volatility(self):             # <<<<<<<<<<<<<<
 *         vol = Decimal(str(self._avg_vol.current_value))
 *         return vol
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_41get_volatility, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_g_8, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__59)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_get_volatility, __pyx_t_5) < 0) __PYX_ERR(0, 695, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":710
 *                                f"kappa={self._kappa:.4f}")
 * 
 *     def measure_order_book_liquidity(self):             # <<<<<<<<<<<<<<
 *         return self.c_measure_order_book_liquidity()
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_43measure_order_book_liquidity, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_m_2, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__60)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 710, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_measure_order_book_liquidity, __pyx_t_5) < 0) __PYX_ERR(0, 710, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":779
 *                                    f"optimal_ask={((self._reservation_price + self._optimal_spread / 2) - price) / price * 100:.4f}%")
 * 
 *     def calculate_reservation_price_and_optimal_spread(self):             # <<<<<<<<<<<<<<
 *         return self.c_calculate_reservation_price_and_optimal_spread()
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_45calculate_reservation_price_and_optimal_spread, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_c_6, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__61)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_calculate_reservation_price_and, __pyx_t_5) < 0) __PYX_ERR(0, 779, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":806
 *         return market.c_quantize_order_amount(trading_pair, Decimal(str(target_inventory_amount)))
 * 
 *     def calculate_target_inventory(self) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_calculate_target_inventory()
 * 
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_Decimal) < 0) __PYX_ERR(0, 806, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_47calculate_target_inventory, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_c_7, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__62)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_calculate_target_inventory, __pyx_t_3) < 0) __PYX_ERR(0, 806, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":831
 *         return inventory_value_base
 * 
 *     def calculate_inventory(self) -> Decimal:             # <<<<<<<<<<<<<<
 *         return self.c_calculate_inventory()
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_Decimal) < 0) __PYX_ERR(0, 831, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_49calculate_inventory, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_c_8, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__63)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_calculate_inventory, __pyx_t_5) < 0) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":840
 *         return self._trading_intensity.is_sampling_buffer_changed or self._avg_vol.is_sampling_buffer_changed
 * 
 *     def is_algorithm_ready(self) -> bool:             # <<<<<<<<<<<<<<
 *         return self.c_is_algorithm_ready()
 * 
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 840, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_51is_algorithm_ready, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_i_2, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_is_algorithm_ready, __pyx_t_3) < 0) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":843
 *         return self.c_is_algorithm_ready()
 * 
 *     def is_algorithm_changed(self) -> bool:             # <<<<<<<<<<<<<<
 *         return self.c_is_algorithm_changed()
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 843, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_53is_algorithm_changed, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_i_3, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__65)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_is_algorithm_changed, __pyx_t_5) < 0) __PYX_ERR(0, 843, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":846
 *         return self.c_is_algorithm_changed()
 * 
 *     def _get_level_spreads(self):             # <<<<<<<<<<<<<<
 *         level_step = ((self._optimal_spread / 2) / 100) * self.level_distances
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_55_get_level_spreads, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_2, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__67)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 846, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_get_level_spreads, __pyx_t_5) < 0) __PYX_ERR(0, 846, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":875
 *         return buys, sells
 * 
 *     def create_proposal_based_on_order_override(self) -> Tuple[List[Proposal], List[Proposal]]:             # <<<<<<<<<<<<<<
 *         return self._create_proposal_based_on_order_override()
 * 
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_kp_s_Tuple_List_Proposal_List_Proposa) < 0) __PYX_ERR(0, 875, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_57create_proposal_based_on_order_override, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_c_9, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__68)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_create_proposal_based_on_order_o, __pyx_t_3) < 0) __PYX_ERR(0, 875, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":896
 *         return buys, sells
 * 
 *     def create_proposal_based_on_order_levels(self):             # <<<<<<<<<<<<<<
 *         return self._create_proposal_based_on_order_levels()
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_59create_proposal_based_on_order_levels, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_c_10, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__69)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_create_proposal_based_on_order_l, __pyx_t_3) < 0) __PYX_ERR(0, 896, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":915
 *         return buys, sells
 * 
 *     def create_basic_proposal(self):             # <<<<<<<<<<<<<<
 *         return self._create_basic_proposal()
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_61create_basic_proposal, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_c_11, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__70)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_create_basic_proposal, __pyx_t_3) < 0) __PYX_ERR(0, 915, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":936
 *         return Proposal(buys, sells)
 * 
 *     def create_base_proposal(self):             # <<<<<<<<<<<<<<
 *         return self.c_create_base_proposal()
 * 
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_63create_base_proposal, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_c_12, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__71)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_create_base_proposal, __pyx_t_3) < 0) __PYX_ERR(0, 936, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":957
 *         return base_balance, quote_balance
 * 
 *     def get_adjusted_available_balance(self, orders: List[LimitOrder]):             # <<<<<<<<<<<<<<
 *         return self.c_get_adjusted_available_balance(orders)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_orders, __pyx_kp_s_List_LimitOrder) < 0) __PYX_ERR(0, 957, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_65get_adjusted_available_balance, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_g_9, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__73)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_get_adjusted_available_balance, __pyx_t_5) < 0) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":967
 *             self.c_apply_add_transaction_costs(proposal)
 * 
 *     def apply_order_price_modifiers(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         self.c_apply_order_price_modifiers(proposal)
 * 
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_proposal, __pyx_n_s_Proposal) < 0) __PYX_ERR(0, 967, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_67apply_order_price_modifiers, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_a_3, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__74)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 967, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_apply_order_price_modifiers, __pyx_t_3) < 0) __PYX_ERR(0, 967, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":970
 *         self.c_apply_order_price_modifiers(proposal)
 * 
 *     def apply_budget_constraint(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_apply_budget_constraint(proposal)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_proposal, __pyx_n_s_Proposal) < 0) __PYX_ERR(0, 970, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_69apply_budget_constraint, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_a_4, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__75)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_apply_budget_constraint, __pyx_t_5) < 0) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":973
 *         return self.c_apply_budget_constraint(proposal)
 * 
 *     def adjusted_available_balance_for_orders_budget_constrain(self):             # <<<<<<<<<<<<<<
 *         candidate_hanging_orders = self.hanging_orders_tracker.candidate_hanging_orders_from_pairs()
 *         non_hanging = []
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_71adjusted_available_balance_for_orders_budget_constrain, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_a_5, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__77)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_adjusted_available_balance_for_o, __pyx_t_5) < 0) __PYX_ERR(0, 973, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1025
 *         proposal.sells = [o for o in proposal.sells if o.size > 0]
 * 
 *     def apply_budget_constraint(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_apply_budget_constraint(proposal)
 * 
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_proposal, __pyx_n_s_Proposal) < 0) __PYX_ERR(0, 1025, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_73apply_budget_constraint, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_a_4, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__78)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1025, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_apply_budget_constraint, __pyx_t_3) < 0) __PYX_ERR(0, 1025, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1078
 *                     proposal.sells[i].price = market.c_quantize_order_price(self.trading_pair, price_below_ask)
 * 
 *     def apply_order_optimization(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         return self.c_apply_order_optimization(proposal)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_proposal, __pyx_n_s_Proposal) < 0) __PYX_ERR(0, 1078, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_75apply_order_optimization, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_a_6, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__79)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_apply_order_optimization, __pyx_t_5) < 0) __PYX_ERR(0, 1078, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1114
 *                     proposal.sells = [o for o in proposal.sells if o.size > 0]
 * 
 *     def apply_order_amount_eta_transformation(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         self.c_apply_order_amount_eta_transformation(proposal)
 * 
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_proposal, __pyx_n_s_Proposal) < 0) __PYX_ERR(0, 1114, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_77apply_order_amount_eta_transformation, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_a_7, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__80)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_apply_order_amount_eta_transform, __pyx_t_3) < 0) __PYX_ERR(0, 1114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1131
 *             sell.price = market.c_quantize_order_price(self.trading_pair, price)
 * 
 *     def apply_add_transaction_costs(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         self.c_apply_add_transaction_costs(proposal)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_proposal, __pyx_n_s_Proposal) < 0) __PYX_ERR(0, 1131, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_79apply_add_transaction_costs, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_a_8, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__81)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_apply_add_transaction_costs, __pyx_t_5) < 0) __PYX_ERR(0, 1131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1212
 *         return True
 * 
 *     def is_within_tolerance(self, current_prices: List[Decimal], proposal_prices: List[Decimal]) -> bool:             # <<<<<<<<<<<<<<
 *         return self.c_is_within_tolerance(current_prices, proposal_prices)
 * 
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_current_prices, __pyx_kp_s_List_Decimal) < 0) __PYX_ERR(0, 1212, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_proposal_prices, __pyx_kp_s_List_Decimal) < 0) __PYX_ERR(0, 1212, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 1212, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_81is_within_tolerance, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_i_4, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__83)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_is_within_tolerance, __pyx_t_3) < 0) __PYX_ERR(0, 1212, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1255
 *             self.c_set_timers()
 * 
 *     def cancel_active_orders(self, proposal: Proposal = None):             # <<<<<<<<<<<<<<
 *         return self.c_cancel_active_orders(proposal)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_proposal, __pyx_n_s_Proposal) < 0) __PYX_ERR(0, 1255, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_83cancel_active_orders, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_c_13, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__84)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetDefaultsTuple(__pyx_t_5, __pyx_tuple__85);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_cancel_active_orders, __pyx_t_5) < 0) __PYX_ERR(0, 1255, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1268
 *                 and len(non_hanging_orders_non_cancelled) == 0)
 * 
 *     def to_create_orders(self, proposal: Proposal) -> bool:             # <<<<<<<<<<<<<<
 *         return self.c_to_create_orders(proposal)
 * 
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_proposal, __pyx_n_s_Proposal) < 0) __PYX_ERR(0, 1268, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_return, __pyx_n_s_bool) < 0) __PYX_ERR(0, 1268, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_85to_create_orders, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_t, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__86)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_to_create_orders, __pyx_t_3) < 0) __PYX_ERR(0, 1268, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1327
 *             self.c_set_timers()
 * 
 *     def execute_orders_proposal(self, proposal: Proposal):             # <<<<<<<<<<<<<<
 *         self.c_execute_orders_proposal(proposal)
 * 
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_proposal, __pyx_n_s_Proposal) < 0) __PYX_ERR(0, 1327, __pyx_L1_error)
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_87execute_orders_proposal, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_e, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__87)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_5, __pyx_t_3);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_execute_orders_proposal, __pyx_t_5) < 0) __PYX_ERR(0, 1327, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1337
 *             self._cancel_timestamp = min(self._create_timestamp, next_cycle)
 * 
 *     def set_timers(self):             # <<<<<<<<<<<<<<
 *         self.c_set_timers()
 * 
 */
  __pyx_t_5 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_89set_timers, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_s_2, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__88)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_set_timers, __pyx_t_5) < 0) __PYX_ERR(0, 1337, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1340
 *         self.c_set_timers()
 * 
 *     def notify_hb_app(self, msg: str):             # <<<<<<<<<<<<<<
 *         if self._hb_app_notification:
 *             super().notify_hb_app(msg)
 */
  __pyx_t_5 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_t_5, __pyx_n_s_msg, __pyx_n_s_str) < 0) __PYX_ERR(0, 1340, __pyx_L1_error)
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_91notify_hb_app, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_n, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__90)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_3, __pyx_t_5);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_notify_hb_app, __pyx_t_3) < 0) __PYX_ERR(0, 1340, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1344
 *             super().notify_hb_app(msg)
 * 
 *     def dump_debug_variables(self):             # <<<<<<<<<<<<<<
 *         market = self._market_info.market
 *         mid_price = self.get_price()
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_93dump_debug_variables, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_d, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__92)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (__Pyx_SetItemOnTypeDict((PyObject *)__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy, __pyx_n_s_dump_debug_variables, __pyx_t_3) < 0) __PYX_ERR(0, 1344, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  PyType_Modified(__pyx_ptype_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_AvellanedaMarketMakingStrategy);

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_95__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_3, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__93)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_reduce_cython, __pyx_t_3) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._events cannot be converted to a Python object for pickling"
 */
  __pyx_t_3 = __Pyx_CyFunction_New(&__pyx_mdef_10hummingbot_8strategy_24avellaneda_market_making_24avellaneda_market_making_30AvellanedaMarketMakingStrategy_97__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_n_s_AvellanedaMarketMakingStrategy_4, NULL, __pyx_n_s_hummingbot_strategy_avellaneda_m, __pyx_d, ((PyObject *)__pyx_codeobj__95)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_setstate_cython, __pyx_t_3) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "hummingbot/strategy/avellaneda_market_making/avellaneda_market_making.pyx":1
 * import datetime             # <<<<<<<<<<<<<<
 * import logging
 * import os
 */
  __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_3) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  if (__pyx_m) {
    if (__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init hummingbot.strategy.avellaneda_market_making.avellaneda_market_making", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init hummingbot.strategy.avellaneda_market_making.avellaneda_market_making");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
#if __PYX_LIMITED_VERSION_HEX < 0x030d00A1
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d00A1
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK && CYTHON_FAST_THREAD_STATE
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_value == NULL || exc_info->exc_value == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    PyObject *exc_value = exc_info->exc_value;
    if (exc_value == NULL || exc_value == Py_None) {
        *value = NULL;
        *type = NULL;
        *tb = NULL;
    } else {
        *value = exc_value;
        Py_INCREF(*value);
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        *tb = PyException_GetTraceback(exc_value);
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
  #endif
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    PyObject *tmp_value = exc_info->exc_value;
    exc_info->exc_value = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
  #else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
  #endif
}
#endif

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type = NULL, *local_value, *local_tb = NULL;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if PY_VERSION_HEX >= 0x030C00A6
    local_value = tstate->current_exception;
    tstate->current_exception = 0;
    if (likely(local_value)) {
        local_type = (PyObject*) Py_TYPE(local_value);
        Py_INCREF(local_type);
        local_tb = PyException_GetTraceback(local_value);
    }
  #else
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
  #endif
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE && PY_VERSION_HEX >= 0x030C00A6
    if (unlikely(tstate->current_exception))
#elif CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
      #if PY_VERSION_HEX >= 0x030B00a4
        tmp_value = exc_info->exc_value;
        exc_info->exc_value = local_value;
        tmp_type = NULL;
        tmp_tb = NULL;
        Py_XDECREF(local_type);
        Py_XDECREF(local_tb);
      #else
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
      #endif
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    __Pyx_PyThreadState_declare
    CYTHON_UNUSED_VAR(cause);
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
      #if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
      #elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* TupleAndListFromArray */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        Py_INCREF(__pyx_empty_tuple);
        return __pyx_empty_tuple;
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = PyTuple_GET_SIZE(kwnames);
    for (i = 0; i < n; i++)
    {
        if (s == PyTuple_GET_ITEM(kwnames, i)) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        int eq = __Pyx_PyUnicode_Equals(s, PyTuple_GET_ITEM(kwnames, i), Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs = PyTuple_GET_SIZE(kwnames);
    PyObject *dict;
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* KeywordStringCheck */
static int __Pyx_CheckKeywordStrings(
    PyObject *kw,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kw, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kw))) {
        Py_ssize_t kwsize;
#if CYTHON_ASSUME_SAFE_MACROS
        kwsize = PyTuple_GET_SIZE(kw);
#else
        kwsize = PyTuple_Size(kw);
        if (kwsize < 0) return 0;
#endif
        if (unlikely(kwsize == 0))
            return 1;
        if (!kw_allowed) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, 0);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            goto invalid_keyword;
        }
#if PY_VERSION_HEX < 0x03090000
        for (pos = 0; pos < kwsize; pos++) {
#if CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kw, pos);
#else
            key = PyTuple_GetItem(kw, pos);
            if (!key) return 0;
#endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
        }
#endif
        return 1;
    }
    while (PyDict_Next(kw, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if (!kw_allowed && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && PY_VERSION_HEX < 0x030d0000
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#elif CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object"))) {
        return NULL;
    }
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object"))) {
        return NULL;
    }
    #endif
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    #if PY_MAJOR_VERSION < 3
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    #else
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    #endif
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject **args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) < 0) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject **args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    #if PY_VERSION_HEX >= 0x030700A1
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL
        #if PY_VERSION_HEX < 0x03090000
        vectorcallfunc f = _PyVectorcall_Function(func);
        #else
        vectorcallfunc f = PyVectorcall_Function(func);
        #endif
        if (f) {
            return f(func, args, (size_t)nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, (size_t)nargs, NULL);
        }
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject *const *kwvalues,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    int kwds_is_tuple = CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds));
    while (1) {
        Py_XDECREF(key); key = NULL;
        Py_XDECREF(value); value = NULL;
        if (kwds_is_tuple) {
            Py_ssize_t size;
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(kwds);
#else
            size = PyTuple_Size(kwds);
            if (size < 0) goto bad;
#endif
            if (pos >= size) break;
#if CYTHON_AVOID_BORROWED_REFS
            key = __Pyx_PySequence_ITEM(kwds, pos);
            if (!key) goto bad;
#elif CYTHON_ASSUME_SAFE_MACROS
            key = PyTuple_GET_ITEM(kwds, pos);
#else
            key = PyTuple_GetItem(kwds, pos);
            if (!key) goto bad;
#endif
            value = kwvalues[pos];
            pos++;
        }
        else
        {
            if (!PyDict_Next(kwds, &pos, &key, &value)) break;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
#endif
        }
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(value);
            Py_DECREF(key);
#endif
            key = NULL;
            value = NULL;
            continue;
        }
#if !CYTHON_AVOID_BORROWED_REFS
        Py_INCREF(key);
#endif
        Py_INCREF(value);
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key)
                );
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
#if CYTHON_AVOID_BORROWED_REFS
                    value = NULL;
#endif
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    Py_XDECREF(key);
    Py_XDECREF(value);
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    #if PY_MAJOR_VERSION < 3
    PyErr_Format(PyExc_TypeError,
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    Py_XDECREF(key);
    Py_XDECREF(value);
    return -1;
}

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* set_iter */
static CYTHON_INLINE PyObject* __Pyx_set_iterator(PyObject* iterable, int is_set,
                                                  Py_ssize_t* p_orig_length, int* p_source_is_set) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    is_set = is_set || likely(PySet_CheckExact(iterable) || PyFrozenSet_CheckExact(iterable));
    *p_source_is_set = is_set;
    if (likely(is_set)) {
        *p_orig_length = PySet_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
    }
#else
    CYTHON_UNUSED_VAR(is_set);
    *p_source_is_set = 0;
#endif
    *p_orig_length = 0;
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_set_iter_next(
        PyObject* iter_obj, Py_ssize_t orig_length,
        Py_ssize_t* ppos, PyObject **value,
        int source_is_set) {
    if (!CYTHON_COMPILING_IN_CPYTHON || PY_VERSION_HEX >= 0x030d0000 || unlikely(!source_is_set)) {
        *value = PyIter_Next(iter_obj);
        if (unlikely(!*value)) {
            return __Pyx_IterFinish();
        }
        CYTHON_UNUSED_VAR(orig_length);
        CYTHON_UNUSED_VAR(ppos);
        return 1;
    }
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    if (unlikely(PySet_GET_SIZE(iter_obj) != orig_length)) {
        PyErr_SetString(
            PyExc_RuntimeError,
            "set changed size during iteration");
        return -1;
    }
    {
        Py_hash_t hash;
        int ret = _PySet_NextEntry(iter_obj, ppos, value, &hash);
        assert (ret != -1);
        if (likely(ret)) {
            Py_INCREF(*value);
            return 1;
        }
    }
#endif
    return 0;
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* DictGetItem */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            if (unlikely(PyTuple_Check(key))) {
                PyObject* args = PyTuple_Pack(1, key);
                if (likely(args)) {
                    PyErr_SetObject(PyExc_KeyError, args);
                    Py_DECREF(args);
                }
            } else {
                PyErr_SetObject(PyExc_KeyError, key);
            }
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#endif

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyInt_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_n_s_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* pybytes_as_double */
static double __Pyx_SlowPyString_AsDouble(PyObject *obj) {
    PyObject *float_value;
#if PY_MAJOR_VERSION >= 3
    float_value = PyFloat_FromString(obj);
#else
    float_value = PyFloat_FromString(obj, 0);
#endif
    if (likely(float_value)) {
#if CYTHON_ASSUME_SAFE_MACROS
        double value = PyFloat_AS_DOUBLE(float_value);
#else
        double value = PyFloat_AsDouble(float_value);
#endif
        Py_DECREF(float_value);
        return value;
    }
    return (double)-1;
}
static const char* __Pyx__PyBytes_AsDouble_Copy(const char* start, char* buffer, Py_ssize_t length) {
    int last_was_punctuation = 1;
    Py_ssize_t i;
    for (i=0; i < length; i++) {
        char chr = start[i];
        int is_punctuation = (chr == '_') | (chr == '.') | (chr == 'e') | (chr == 'E');
        *buffer = chr;
        buffer += (chr != '_');
        if (unlikely(last_was_punctuation & is_punctuation)) goto parse_failure;
        last_was_punctuation = is_punctuation;
    }
    if (unlikely(last_was_punctuation)) goto parse_failure;
    *buffer = '\0';
    return buffer;
parse_failure:
    return NULL;
}
static double __Pyx__PyBytes_AsDouble_inf_nan(const char* start, Py_ssize_t length) {
    int matches = 1;
    char sign = start[0];
    int is_signed = (sign == '+') | (sign == '-');
    start += is_signed;
    length -= is_signed;
    switch (start[0]) {
        #ifdef Py_NAN
        case 'n':
        case 'N':
            if (unlikely(length != 3)) goto parse_failure;
            matches &= (start[1] == 'a' || start[1] == 'A');
            matches &= (start[2] == 'n' || start[2] == 'N');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_NAN : Py_NAN;
        #endif
        case 'i':
        case 'I':
            if (unlikely(length < 3)) goto parse_failure;
            matches &= (start[1] == 'n' || start[1] == 'N');
            matches &= (start[2] == 'f' || start[2] == 'F');
            if (likely(length == 3 && matches))
                return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
            if (unlikely(length != 8)) goto parse_failure;
            matches &= (start[3] == 'i' || start[3] == 'I');
            matches &= (start[4] == 'n' || start[4] == 'N');
            matches &= (start[5] == 'i' || start[5] == 'I');
            matches &= (start[6] == 't' || start[6] == 'T');
            matches &= (start[7] == 'y' || start[7] == 'Y');
            if (unlikely(!matches)) goto parse_failure;
            return (sign == '-') ? -Py_HUGE_VAL : Py_HUGE_VAL;
        case '.': case '0': case '1': case '2': case '3': case '4': case '5': case '6': case '7': case '8': case '9':
            break;
        default:
            goto parse_failure;
    }
    return 0.0;
parse_failure:
    return -1.0;
}
static CYTHON_INLINE int __Pyx__PyBytes_AsDouble_IsSpace(char ch) {
    return (ch == 0x20) | !((ch < 0x9) | (ch > 0xd));
}
CYTHON_UNUSED static double __Pyx__PyBytes_AsDouble(PyObject *obj, const char* start, Py_ssize_t length) {
    double value;
    Py_ssize_t i, digits;
    const char *last = start + length;
    char *end;
    while (__Pyx__PyBytes_AsDouble_IsSpace(*start))
        start++;
    while (start < last - 1 && __Pyx__PyBytes_AsDouble_IsSpace(last[-1]))
        last--;
    length = last - start;
    if (unlikely(length <= 0)) goto fallback;
    value = __Pyx__PyBytes_AsDouble_inf_nan(start, length);
    if (unlikely(value == -1.0)) goto fallback;
    if (value != 0.0) return value;
    digits = 0;
    for (i=0; i < length; digits += start[i++] != '_');
    if (likely(digits == length)) {
        value = PyOS_string_to_double(start, &end, NULL);
    } else if (digits < 40) {
        char number[40];
        last = __Pyx__PyBytes_AsDouble_Copy(start, number, length);
        if (unlikely(!last)) goto fallback;
        value = PyOS_string_to_double(number, &end, NULL);
    } else {
        char *number = (char*) PyMem_Malloc((digits + 1) * sizeof(char));
        if (unlikely(!number)) goto fallback;
        last = __Pyx__PyBytes_AsDouble_Copy(start, number, length);
        if (unlikely(!last)) {
            PyMem_Free(number);
            goto fallback;
        }
        value = PyOS_string_to_double(number, &end, NULL);
        PyMem_Free(number);
    }
    if (likely(end == last) || (value == (double)-1 && PyErr_Occurred())) {
        return value;
    }
fallback:
    return __Pyx_SlowPyString_AsDouble(obj);
}

/* pyobject_as_double */
static double __Pyx__PyObject_AsDouble(PyObject* obj) {
    if (PyUnicode_CheckExact(obj)) {
        return __Pyx_PyUnicode_AsDouble(obj);
    } else if (PyBytes_CheckExact(obj)) {
        return __Pyx_PyBytes_AsDouble(obj);
    } else if (PyByteArray_CheckExact(obj)) {
        return __Pyx_PyByteArray_AsDouble(obj);
    } else {
        PyObject* float_value;
#if !CYTHON_USE_TYPE_SLOTS
        float_value = PyNumber_Float(obj);  if ((0)) goto bad;
        (void)__Pyx_PyObject_CallOneArg;
#else
        PyNumberMethods *nb = Py_TYPE(obj)->tp_as_number;
        if (likely(nb) && likely(nb->nb_float)) {
            float_value = nb->nb_float(obj);
            if (likely(float_value) && unlikely(!PyFloat_Check(float_value))) {
                __Pyx_TypeName float_value_type_name = __Pyx_PyType_GetName(Py_TYPE(float_value));
                PyErr_Format(PyExc_TypeError,
                    "__float__ returned non-float (type " __Pyx_FMT_TYPENAME ")",
                    float_value_type_name);
                __Pyx_DECREF_TypeName(float_value_type_name);
                Py_DECREF(float_value);
                goto bad;
            }
        } else {
            float_value = __Pyx_PyObject_CallOneArg((PyObject*)&PyFloat_Type, obj);
        }
#endif
        if (likely(float_value)) {
            double value = PyFloat_AS_DOUBLE(float_value);
            Py_DECREF(float_value);
            return value;
        }
    }
bad:
    return (double)-1;
}

/* JoinPyUnicode */
static PyObject* __Pyx_PyUnicode_Join(PyObject* value_tuple, Py_ssize_t value_count, Py_ssize_t result_ulength,
                                      Py_UCS4 max_char) {
#if CYTHON_USE_UNICODE_INTERNALS && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    PyObject *result_uval;
    int result_ukind, kind_shift;
    Py_ssize_t i, char_pos;
    void *result_udata;
    CYTHON_MAYBE_UNUSED_VAR(max_char);
#if CYTHON_PEP393_ENABLED
    result_uval = PyUnicode_New(result_ulength, max_char);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = (max_char <= 255) ? PyUnicode_1BYTE_KIND : (max_char <= 65535) ? PyUnicode_2BYTE_KIND : PyUnicode_4BYTE_KIND;
    kind_shift = (result_ukind == PyUnicode_4BYTE_KIND) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_DATA(result_uval);
#else
    result_uval = PyUnicode_FromUnicode(NULL, result_ulength);
    if (unlikely(!result_uval)) return NULL;
    result_ukind = sizeof(Py_UNICODE);
    kind_shift = (result_ukind == 4) ? 2 : result_ukind - 1;
    result_udata = PyUnicode_AS_UNICODE(result_uval);
#endif
    assert(kind_shift == 2 || kind_shift == 1 || kind_shift == 0);
    char_pos = 0;
    for (i=0; i < value_count; i++) {
        int ukind;
        Py_ssize_t ulength;
        void *udata;
        PyObject *uval = PyTuple_GET_ITEM(value_tuple, i);
        if (unlikely(__Pyx_PyUnicode_READY(uval)))
            goto bad;
        ulength = __Pyx_PyUnicode_GET_LENGTH(uval);
        if (unlikely(!ulength))
            continue;
        if (unlikely((PY_SSIZE_T_MAX >> kind_shift) - ulength < char_pos))
            goto overflow;
        ukind = __Pyx_PyUnicode_KIND(uval);
        udata = __Pyx_PyUnicode_DATA(uval);
        if (!CYTHON_PEP393_ENABLED || ukind == result_ukind) {
            memcpy((char *)result_udata + (char_pos << kind_shift), udata, (size_t) (ulength << kind_shift));
        } else {
            #if PY_VERSION_HEX >= 0x030d0000
            if (unlikely(PyUnicode_CopyCharacters(result_uval, char_pos, uval, 0, ulength) < 0)) goto bad;
            #elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030300F0 || defined(_PyUnicode_FastCopyCharacters)
            _PyUnicode_FastCopyCharacters(result_uval, char_pos, uval, 0, ulength);
            #else
            Py_ssize_t j;
            for (j=0; j < ulength; j++) {
                Py_UCS4 uchar = __Pyx_PyUnicode_READ(ukind, udata, j);
                __Pyx_PyUnicode_WRITE(result_ukind, result_udata, char_pos+j, uchar);
            }
            #endif
        }
        char_pos += ulength;
    }
    return result_uval;
overflow:
    PyErr_SetString(PyExc_OverflowError, "join() result is too long for a Python string");
bad:
    Py_DECREF(result_uval);
    return NULL;
#else
    CYTHON_UNUSED_VAR(max_char);
    CYTHON_UNUSED_VAR(result_ulength);
    CYTHON_UNUSED_VAR(value_count);
    return PyUnicode_Join(__pyx_empty_unicode, value_tuple);
#endif
}

/* PyObjectFormat */
#if CYTHON_USE_UNICODE_WRITER
static PyObject* __Pyx_PyObject_Format(PyObject* obj, PyObject* format_spec) {
    int ret;
    _PyUnicodeWriter writer;
    if (likely(PyFloat_CheckExact(obj))) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x03040000
        _PyUnicodeWriter_Init(&writer, 0);
#else
        _PyUnicodeWriter_Init(&writer);
#endif
        ret = _PyFloat_FormatAdvancedWriter(
            &writer,
            obj,
            format_spec, 0, PyUnicode_GET_LENGTH(format_spec));
    } else if (likely(PyLong_CheckExact(obj))) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x03040000
        _PyUnicodeWriter_Init(&writer, 0);
#else
        _PyUnicodeWriter_Init(&writer);
#endif
        ret = _PyLong_FormatAdvancedWriter(
            &writer,
            obj,
            format_spec, 0, PyUnicode_GET_LENGTH(format_spec));
    } else {
        return PyObject_Format(obj, format_spec);
    }
    if (unlikely(ret == -1)) {
        _PyUnicodeWriter_Dealloc(&writer);
        return NULL;
    }
    return _PyUnicodeWriter_Finish(&writer);
}
#endif

/* FixUpExtensionType */
#if CYTHON_USE_TYPE_SPECS
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if PY_VERSION_HEX > 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
#else
    const PyType_Slot *slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
        int changed = 0;
#if !(PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON)
        const
#endif
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
#if PY_VERSION_HEX < 0x030900b1
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif
#else
                if ((0));
#endif
#if PY_VERSION_HEX <= 0x030900b1 && CYTHON_COMPILING_IN_CPYTHON
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    if (unlikely(PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr) < 0)) {
                        Py_DECREF(descr);
                        return -1;
                    }
                    Py_DECREF(descr);
                    changed = 1;
                }
#endif
            }
            memb++;
        }
        if (changed)
            PyType_Modified(type);
    }
#endif
    return 0;
}
#endif

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef((char*) __PYX_ABI_MODULE_NAME);
}

/* FetchCommonType */
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t basicsize,
                               Py_ssize_t expected_basicsize) {
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
#if !CYTHON_USE_TYPE_SPECS
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* abi_module;
    const char* object_name;
    PyTypeObject *cached_type = NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    object_name = strrchr(type->tp_name, '.');
    object_name = object_name ? object_name+1 : type->tp_name;
    cached_type = (PyTypeObject*) PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        if (__Pyx_VerifyCachedType(
              (PyObject *)cached_type,
              object_name,
              cached_type->tp_basicsize,
              type->tp_basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    if (PyType_Ready(type) < 0) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, (PyObject *)type) < 0)
        goto bad;
    Py_INCREF(type);
    cached_type = type;
done:
    Py_DECREF(abi_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#else
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module, *cached_type = NULL;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) return NULL;
    cached_type = PyObject_GetAttrString(abi_module, object_name);
    if (cached_type) {
        Py_ssize_t basicsize;
#if CYTHON_COMPILING_IN_LIMITED_API
        PyObject *py_basicsize;
        py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
        if (unlikely(!py_basicsize)) goto bad;
        basicsize = PyLong_AsSsize_t(py_basicsize);
        Py_DECREF(py_basicsize);
        py_basicsize = 0;
        if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
#else
        basicsize = likely(PyType_Check(cached_type)) ? ((PyTypeObject*) cached_type)->tp_basicsize : -1;
#endif
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              basicsize,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    }
    if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
    PyErr_Clear();
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromModuleAndSpec(abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    if (PyObject_SetAttrString(abi_module, object_name, cached_type) < 0) goto bad;
done:
    Py_DECREF(abi_module);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &= Py_TYPE(key)->tp_flags;
        Py_INCREF(key);
        Py_INCREF(value);
        PyTuple_SET_ITEM(kwnames, i, key);
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    if (likely(kw == NULL) || PyDict_GET_SIZE(kw) == 0) {
        return vc(func, args, nargs, NULL);
    }
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void *cfunc) {
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure)
{
    CYTHON_UNUSED_VAR(closure);
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#elif PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_tuple;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->defaults_kwdict;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = op->func_annotations;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    int is_coroutine;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
#if PY_VERSION_HEX >= 0x03050000
    if (is_coroutine) {
        PyObject *module, *fromlist, *marker = __pyx_n_s_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_n_s_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        op->func_is_coroutine = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(op->func_is_coroutine)) {
            return __Pyx_NewRef(op->func_is_coroutine);
        }
ignore:
        PyErr_Clear();
    }
#endif
    op->func_is_coroutine = __Pyx_PyBool_FromLong(is_coroutine);
    return __Pyx_NewRef(op->func_is_coroutine);
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {(char *) "_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
#if !CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#endif
#endif
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    CYTHON_UNUSED_VAR(args);
#if PY_MAJOR_VERSION >= 3
    Py_INCREF(m->func_qualname);
    return m->func_qualname;
#else
    return PyString_FromString(((PyCFunctionObject*)m)->m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults_size = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyObject *py_name = NULL;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void*)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_MACROS
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
#if CYTHON_COMPILING_IN_LIMITED_API
            py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
            if (!py_name) return NULL;
            PyErr_Format(PyExc_TypeError,
                "%.200S() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                py_name, size);
            Py_DECREF(py_name);
#else
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
#endif
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    py_name = __Pyx_CyFunction_get_name((__pyx_CyFunctionObject*)func, NULL);
    if (!py_name) return NULL;
    PyErr_Format(PyExc_TypeError, "%.200S() takes no keyword arguments",
                 py_name);
    Py_DECREF(py_name);
#else
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
#endif
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_MACROS
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(!argc) < 0) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
#if PY_MAJOR_VERSION > 2
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
#else
            PyErr_SetString(PyExc_TypeError,
                            "unbound method needs an argument");
#endif
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            PyErr_Format(PyExc_TypeError, "%.200s() needs an argument",
                         ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(PyTuple_GET_SIZE(kwnames))) {
        PyErr_Format(PyExc_TypeError,
                     "%.200s() takes no keyword arguments", ((PyCFunctionObject*)cyfunc)->m_ml->ml_name);
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
            def->ml_name, nargs);
        return NULL;
    }
    return def->ml_meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))def->ml_meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyMethodDef* def = ((PyCFunctionObject*)cyfunc)->m_ml;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
        self = ((PyCFunctionObject*)cyfunc)->m_self;
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))def->ml_meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if (defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
#else
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
#if !CYTHON_METH_FASTCALL
    0,
#elif CYTHON_BACKPORT_VECTORCALL
    (printfunc)offsetof(__pyx_CyFunctionObject, func_vectorcall),
#else
    offsetof(PyCFunctionObject, vectorcall),
#endif
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if defined(_Py_TPFLAGS_HAVE_VECTORCALL) && CYTHON_METH_FASTCALL
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_PyMethod_New,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_CyFunction_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_CyFunctionType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
#endif
    if (unlikely(__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (unlikely(!m->defaults))
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    m->defaults_size = size;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
        
            x = (long)((unsigned long)a + (unsigned long)b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return __Pyx_NewRef(op2);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
            double result;
            
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyIntCompare */
static CYTHON_INLINE int __Pyx_PyInt_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 1;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        return (a == b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) == 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 0;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 0;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal == 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
        return ((double)a == (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* py_abs */
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject *__Pyx_PyLong_AbsNeg(PyObject *n) {
#if PY_VERSION_HEX >= 0x030C00A7
    if (likely(__Pyx_PyLong_IsCompact(n))) {
        return PyLong_FromSize_t(__Pyx_PyLong_CompactValueUnsigned(n));
    }
#else
    if (likely(Py_SIZE(n) == -1)) {
        return PyLong_FromUnsignedLong(__Pyx_PyLong_Digits(n)[0]);
    }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    {
        PyObject *copy = _PyLong_Copy((PyLongObject*)n);
        if (likely(copy)) {
            #if PY_VERSION_HEX >= 0x030C00A7
            ((PyLongObject*)copy)->long_value.lv_tag = ((PyLongObject*)copy)->long_value.lv_tag & ~_PyLong_SIGN_MASK;
            #else
            __Pyx_SET_SIZE(copy, -Py_SIZE(copy));
            #endif
        }
        return copy;
    }
#else
    return PyNumber_Negative(n);
#endif
}
#endif

/* RaiseUnboundLocalError */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* pynumber_float */
static CYTHON_INLINE PyObject* __Pyx__PyNumber_Float(PyObject* obj) {
    double val;
    if (PyLong_CheckExact(obj)) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (likely(__Pyx_PyLong_IsCompact(obj))) {
            val = (double) __Pyx_PyLong_CompactValue(obj);
            goto no_error;
        }
#endif
        val = PyLong_AsDouble(obj);
    } else if (PyUnicode_CheckExact(obj)) {
        val = __Pyx_PyUnicode_AsDouble(obj);
    } else if (PyBytes_CheckExact(obj)) {
        val = __Pyx_PyBytes_AsDouble(obj);
    } else if (PyByteArray_CheckExact(obj)) {
        val = __Pyx_PyByteArray_AsDouble(obj);
    } else {
        return PyNumber_Float(obj);
    }
    if (unlikely(val == -1 && PyErr_Occurred())) {
        return NULL;
    }
#if CYTHON_USE_PYLONG_INTERNALS
no_error:
#endif
    return PyFloat_FromDouble(val);
}

/* UnicodeConcatInPlace */
# if CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
static int
__Pyx_unicode_modifiable(PyObject *unicode)
{
    if (Py_REFCNT(unicode) != 1)
        return 0;
    if (!PyUnicode_CheckExact(unicode))
        return 0;
    if (PyUnicode_CHECK_INTERNED(unicode))
        return 0;
    return 1;
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_ConcatInPlaceImpl(PyObject **p_left, PyObject *right
        #if CYTHON_REFNANNY
        , void* __pyx_refnanny
        #endif
    ) {
    PyObject *left = *p_left;
    Py_ssize_t left_len, right_len, new_len;
    if (unlikely(__Pyx_PyUnicode_READY(left) == -1))
        return NULL;
    if (unlikely(__Pyx_PyUnicode_READY(right) == -1))
        return NULL;
    left_len = PyUnicode_GET_LENGTH(left);
    if (left_len == 0) {
        Py_INCREF(right);
        return right;
    }
    right_len = PyUnicode_GET_LENGTH(right);
    if (right_len == 0) {
        Py_INCREF(left);
        return left;
    }
    if (unlikely(left_len > PY_SSIZE_T_MAX - right_len)) {
        PyErr_SetString(PyExc_OverflowError,
                        "strings are too large to concat");
        return NULL;
    }
    new_len = left_len + right_len;
    if (__Pyx_unicode_modifiable(left)
            && PyUnicode_CheckExact(right)
            && PyUnicode_KIND(right) <= PyUnicode_KIND(left)
            && !(PyUnicode_IS_ASCII(left) && !PyUnicode_IS_ASCII(right))) {
        int ret;
        __Pyx_GIVEREF(*p_left);
        ret = PyUnicode_Resize(p_left, new_len);
        __Pyx_GOTREF(*p_left);
        if (unlikely(ret != 0))
            return NULL;
        #if PY_VERSION_HEX >= 0x030d0000
        if (unlikely(PyUnicode_CopyCharacters(*p_left, left_len, right, 0, right_len) < 0)) return NULL;
        #else
        _PyUnicode_FastCopyCharacters(*p_left, left_len, right, 0, right_len);
        #endif
        __Pyx_INCREF(*p_left);
        __Pyx_GIVEREF(*p_left);
        return *p_left;
    } else {
        return __Pyx_PyUnicode_Concat(left, right);
    }
  }
#endif

/* RaiseClosureNameError */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* pep479 */
static void __Pyx_Generator_Replace_StopIteration(int in_async_gen) {
    PyObject *exc, *val, *tb, *cur_exc;
    __Pyx_PyThreadState_declare
    #ifdef __Pyx_StopAsyncIteration_USED
    int is_async_stopiteration = 0;
    #endif
    CYTHON_MAYBE_UNUSED_VAR(in_async_gen);
    cur_exc = PyErr_Occurred();
    if (likely(!__Pyx_PyErr_GivenExceptionMatches(cur_exc, PyExc_StopIteration))) {
        #ifdef __Pyx_StopAsyncIteration_USED
        if (in_async_gen && unlikely(__Pyx_PyErr_GivenExceptionMatches(cur_exc, __Pyx_PyExc_StopAsyncIteration))) {
            is_async_stopiteration = 1;
        } else
        #endif
            return;
    }
    __Pyx_PyThreadState_assign
    __Pyx_GetException(&exc, &val, &tb);
    Py_XDECREF(exc);
    Py_XDECREF(val);
    Py_XDECREF(tb);
    PyErr_SetString(PyExc_RuntimeError,
        #ifdef __Pyx_StopAsyncIteration_USED
        is_async_stopiteration ? "async generator raised StopAsyncIteration" :
        in_async_gen ? "async generator raised StopIteration" :
        #endif
        "generator raised StopIteration");
}

/* IterNext */
static PyObject *__Pyx_PyIter_Next2Default(PyObject* defval) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (!defval || unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return NULL;
        __Pyx_PyErr_Clear();
        Py_INCREF(defval);
        return defval;
    }
    if (defval) {
        Py_INCREF(defval);
        return defval;
    }
    __Pyx_PyErr_SetNone(PyExc_StopIteration);
    return NULL;
}
static void __Pyx_PyIter_Next_ErrorNoIterator(PyObject *iterator) {
    __Pyx_TypeName iterator_type_name = __Pyx_PyType_GetName(Py_TYPE(iterator));
    PyErr_Format(PyExc_TypeError,
        __Pyx_FMT_TYPENAME " object is not an iterator", iterator_type_name);
    __Pyx_DECREF_TypeName(iterator_type_name);
}
static CYTHON_INLINE PyObject *__Pyx_PyIter_Next2(PyObject* iterator, PyObject* defval) {
    PyObject* next;
    iternextfunc iternext = Py_TYPE(iterator)->tp_iternext;
    if (likely(iternext)) {
#if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
        next = iternext(iterator);
        if (likely(next))
            return next;
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
        if (unlikely(iternext == &_PyObject_NextNotImplemented))
            return NULL;
#endif
#else
        next = PyIter_Next(iterator);
        if (likely(next))
            return next;
#endif
    } else if (CYTHON_USE_TYPE_SLOTS || unlikely(!PyIter_Check(iterator))) {
        __Pyx_PyIter_Next_ErrorNoIterator(iterator);
        return NULL;
    }
#if !CYTHON_USE_TYPE_SLOTS
    else {
        next = PyIter_Next(iterator);
        if (likely(next))
            return next;
    }
#endif
    return __Pyx_PyIter_Next2Default(defval);
}

/* ModInt[long] */
static CYTHON_INLINE long __Pyx_mod_long(long a, long b) {
    long r = a % b;
    r += ((r != 0) & ((r ^ b) < 0)) * b;
    return r;
}

/* CIntToDigits */
static const char DIGIT_PAIRS_10[2*10*10+1] = {
    "00010203040506070809"
    "10111213141516171819"
    "20212223242526272829"
    "30313233343536373839"
    "40414243444546474849"
    "50515253545556575859"
    "60616263646566676869"
    "70717273747576777879"
    "80818283848586878889"
    "90919293949596979899"
};
static const char DIGIT_PAIRS_8[2*8*8+1] = {
    "0001020304050607"
    "1011121314151617"
    "2021222324252627"
    "3031323334353637"
    "4041424344454647"
    "5051525354555657"
    "6061626364656667"
    "7071727374757677"
};
static const char DIGITS_HEX[2*16+1] = {
    "0123456789abcdef"
    "0123456789ABCDEF"
};

/* BuildPyUnicode */
static PyObject* __Pyx_PyUnicode_BuildFromAscii(Py_ssize_t ulength, char* chars, int clength,
                                                int prepend_sign, char padding_char) {
    PyObject *uval;
    Py_ssize_t uoffset = ulength - clength;
#if CYTHON_USE_UNICODE_INTERNALS
    Py_ssize_t i;
#if CYTHON_PEP393_ENABLED
    void *udata;
    uval = PyUnicode_New(ulength, 127);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_DATA(uval);
#else
    Py_UNICODE *udata;
    uval = PyUnicode_FromUnicode(NULL, ulength);
    if (unlikely(!uval)) return NULL;
    udata = PyUnicode_AS_UNICODE(uval);
#endif
    if (uoffset > 0) {
        i = 0;
        if (prepend_sign) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, 0, '-');
            i++;
        }
        for (; i < uoffset; i++) {
            __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, i, padding_char);
        }
    }
    for (i=0; i < clength; i++) {
        __Pyx_PyUnicode_WRITE(PyUnicode_1BYTE_KIND, udata, uoffset+i, chars[i]);
    }
#else
    {
        PyObject *sign = NULL, *padding = NULL;
        uval = NULL;
        if (uoffset > 0) {
            prepend_sign = !!prepend_sign;
            if (uoffset > prepend_sign) {
                padding = PyUnicode_FromOrdinal(padding_char);
                if (likely(padding) && uoffset > prepend_sign + 1) {
                    PyObject *tmp;
                    PyObject *repeat = PyInt_FromSsize_t(uoffset - prepend_sign);
                    if (unlikely(!repeat)) goto done_or_error;
                    tmp = PyNumber_Multiply(padding, repeat);
                    Py_DECREF(repeat);
                    Py_DECREF(padding);
                    padding = tmp;
                }
                if (unlikely(!padding)) goto done_or_error;
            }
            if (prepend_sign) {
                sign = PyUnicode_FromOrdinal('-');
                if (unlikely(!sign)) goto done_or_error;
            }
        }
        uval = PyUnicode_DecodeASCII(chars, clength, NULL);
        if (likely(uval) && padding) {
            PyObject *tmp = PyNumber_Add(padding, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
        if (likely(uval) && sign) {
            PyObject *tmp = PyNumber_Add(sign, uval);
            Py_DECREF(uval);
            uval = tmp;
        }
done_or_error:
        Py_XDECREF(padding);
        Py_XDECREF(sign);
    }
#endif
    return uval;
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_int(int value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(int)*3+2];
    char *dpos, *end = digits + sizeof(int)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    int remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (int) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (int) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (int) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
  #if CYTHON_USE_EXC_INFO_STACK && PY_VERSION_HEX >= 0x030B00a4
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_value = exc_info->exc_value;
    exc_info->exc_value = *value;
    if (tmp_value == NULL || tmp_value == Py_None) {
        Py_XDECREF(tmp_value);
        tmp_value = NULL;
        tmp_type = NULL;
        tmp_tb = NULL;
    } else {
        tmp_type = (PyObject*) Py_TYPE(tmp_value);
        Py_INCREF(tmp_type);
        #if CYTHON_COMPILING_IN_CPYTHON
        tmp_tb = ((PyBaseExceptionObject*) tmp_value)->traceback;
        Py_XINCREF(tmp_tb);
        #else
        tmp_tb = PyException_GetTraceback(tmp_value);
        #endif
    }
  #elif CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
  #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
  #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* PyIntCompare */
static CYTHON_INLINE int __Pyx_PyInt_BoolNeObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 0;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        return (a != b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) != 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 1;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 1;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal != 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
        return ((double)a != (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_NE));
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_MultiplyCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op2))) {
        const long a = intval;
        long b = PyInt_AS_LONG(op2);
        
#ifdef HAVE_LONG_LONG
            if (sizeof(PY_LONG_LONG) > sizeof(long)) {
                PY_LONG_LONG result = (PY_LONG_LONG)a * (PY_LONG_LONG)b;
                return (result >= LONG_MIN && result <= LONG_MAX) ?
                    PyInt_FromLong((long)result) : PyLong_FromLongLong(result);
            }
#endif
#if CYTHON_USE_TYPE_SLOTS
            return PyInt_Type.tp_as_number->nb_multiply(op1, op2);
#else
            return PyNumber_Multiply(op1, op2);
#endif
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op2))) {
        const long a = intval;
        long b, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG lla = intval;
        PY_LONG_LONG llb, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op2))) {
            return __Pyx_NewRef(op2);
        }
        if (likely(__Pyx_PyLong_IsCompact(op2))) {
            b = __Pyx_PyLong_CompactValue(op2);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op2);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op2);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT+30) {
                        b = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT+30) {
                        llb = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT+30) {
                        b = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT+30) {
                        llb = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT+30) {
                        b = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT+30) {
                        llb = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT+30) {
                        b = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT+30) {
                        llb = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT+30) {
                        b = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT+30) {
                        llb = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT+30) {
                        b = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT+30) {
                        llb = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_multiply(op1, op2);
            }
        }
                CYTHON_UNUSED_VAR(a);
                CYTHON_UNUSED_VAR(b);
                #ifdef HAVE_LONG_LONG
                llb = b;
                goto long_long;
                #else
                return PyLong_Type.tp_as_number->nb_multiply(op1, op2);
                #endif
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla * llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op2)) {
        const long a = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double b = __pyx_PyFloat_AsDouble(op2);
#else
        double b = PyFloat_AS_DOUBLE(op2);
#endif
            double result;
            
            PyFPE_START_PROTECT("multiply", return NULL)
            result = ((double)a) * (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceMultiply : PyNumber_Multiply)(op1, op2);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddCObj(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op2))) {
        const long a = intval;
        long x;
        long b = PyInt_AS_LONG(op2);
        
            x = (long)((unsigned long)a + (unsigned long)b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op2))) {
        const long a = intval;
        long b, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG lla = intval;
        PY_LONG_LONG llb, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op2))) {
            return __Pyx_NewRef(op1);
        }
        if (likely(__Pyx_PyLong_IsCompact(op2))) {
            b = __Pyx_PyLong_CompactValue(op2);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op2);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op2);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        b = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        llb = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        b = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        llb = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        b = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        llb = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        b = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        llb = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        b = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        llb = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        b = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        llb = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op2)) {
        const long a = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double b = __pyx_PyFloat_AsDouble(op2);
#else
        double b = PyFloat_AS_DOUBLE(op2);
#endif
            double result;
            
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_TrueDivideObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        
            if (8 * sizeof(long) <= 53 || likely(labs(a) <= ((PY_LONG_LONG)1 << 53))) {
                return PyFloat_FromDouble((double)a / (double)b);
            }
            return PyInt_Type.tp_as_number->nb_true_divide(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT && 1 * PyLong_SHIFT < 53) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT && 1 * PyLong_SHIFT < 53) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT && 2 * PyLong_SHIFT < 53) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT && 2 * PyLong_SHIFT < 53) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT && 3 * PyLong_SHIFT < 53) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT && 3 * PyLong_SHIFT < 53) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_true_divide(op1, op2);
            }
        }
                if ((8 * sizeof(long) <= 53 || likely(labs(a) <= ((PY_LONG_LONG)1 << 53)))
                        || __Pyx_PyLong_DigitCount(op1) <= 52 / PyLong_SHIFT) {
                    return PyFloat_FromDouble((double)a / (double)b);
                }
                return PyLong_Type.tp_as_number->nb_true_divide(op1, op2);
            return PyLong_FromLong(x);
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
            double result;
            
            PyFPE_START_PROTECT("divide", return NULL)
            result = ((double)a) / (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceTrueDivide : PyNumber_TrueDivide)(op1, op2);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_MultiplyObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        
#ifdef HAVE_LONG_LONG
            if (sizeof(PY_LONG_LONG) > sizeof(long)) {
                PY_LONG_LONG result = (PY_LONG_LONG)a * (PY_LONG_LONG)b;
                return (result >= LONG_MIN && result <= LONG_MAX) ?
                    PyInt_FromLong((long)result) : PyLong_FromLongLong(result);
            }
#endif
#if CYTHON_USE_TYPE_SLOTS
            return PyInt_Type.tp_as_number->nb_multiply(op1, op2);
#else
            return PyNumber_Multiply(op1, op2);
#endif
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return __Pyx_NewRef(op1);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT+30) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT+30) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT+30) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT+30) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT+30) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT+30) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT+30) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT+30) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT+30) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT+30) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT+30) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT+30) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_multiply(op1, op2);
            }
        }
                CYTHON_UNUSED_VAR(a);
                CYTHON_UNUSED_VAR(b);
                #ifdef HAVE_LONG_LONG
                lla = a;
                goto long_long;
                #else
                return PyLong_Type.tp_as_number->nb_multiply(op1, op2);
                #endif
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla * llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
            double result;
            
            PyFPE_START_PROTECT("multiply", return NULL)
            result = ((double)a) * (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceMultiply : PyNumber_Multiply)(op1, op2);
}
#endif

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#elif PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
#if CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
#include <string.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
        
            x = (long)((unsigned long)a - (unsigned long)b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        if (unlikely(__Pyx_PyLong_IsZero(op1))) {
            return PyLong_FromLong(-intval);
        }
        if (likely(__Pyx_PyLong_IsCompact(op1))) {
            a = __Pyx_PyLong_CompactValue(op1);
        } else {
            const digit* digits = __Pyx_PyLong_Digits(op1);
            const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    #ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    #endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
#if CYTHON_COMPILING_IN_LIMITED_API
        double a = __pyx_PyFloat_AsDouble(op1);
#else
        double a = PyFloat_AS_DOUBLE(op1);
#endif
            double result;
            
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* PyUnicode_Unicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_Unicode(PyObject *obj) {
    if (unlikely(obj == Py_None))
        obj = __pyx_kp_u_None;
    return __Pyx_NewRef(obj);
}

/* UnpackUnboundCMethod */
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
     METH_VARARGS | METH_KEYWORDS,
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    target->method = method;
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION >= 3
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    #else
    if (likely(!__Pyx_CyOrPyCFunction_Check(method)))
    #endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            target->method = unbound_method;
        }
    }
    return 0;
}

/* CallUnboundCMethod0 */
static PyObject* __Pyx__CallUnboundCMethod0(__Pyx_CachedCFunction* cfunc, PyObject* self) {
    PyObject *args, *result = NULL;
    if (unlikely(!cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_ASSUME_SAFE_MACROS
    args = PyTuple_New(1);
    if (unlikely(!args)) goto bad;
    Py_INCREF(self);
    PyTuple_SET_ITEM(args, 0, self);
#else
    args = PyTuple_Pack(1, self);
    if (unlikely(!args)) goto bad;
#endif
    result = __Pyx_PyObject_Call(cfunc->method, args, NULL);
    Py_DECREF(args);
bad:
    return result;
}

/* CIntToPyUnicode */
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_From_Py_ssize_t(Py_ssize_t value, Py_ssize_t width, char padding_char, char format_char) {
    char digits[sizeof(Py_ssize_t)*3+2];
    char *dpos, *end = digits + sizeof(Py_ssize_t)*3+2;
    const char *hex_digits = DIGITS_HEX;
    Py_ssize_t length, ulength;
    int prepend_sign, last_one_off;
    Py_ssize_t remaining;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const Py_ssize_t neg_one = (Py_ssize_t) -1, const_zero = (Py_ssize_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (format_char == 'X') {
        hex_digits += 16;
        format_char = 'x';
    }
    remaining = value;
    last_one_off = 0;
    dpos = end;
    do {
        int digit_pos;
        switch (format_char) {
        case 'o':
            digit_pos = abs((int)(remaining % (8*8)));
            remaining = (Py_ssize_t) (remaining / (8*8));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_8 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 8);
            break;
        case 'd':
            digit_pos = abs((int)(remaining % (10*10)));
            remaining = (Py_ssize_t) (remaining / (10*10));
            dpos -= 2;
            memcpy(dpos, DIGIT_PAIRS_10 + digit_pos * 2, 2);
            last_one_off = (digit_pos < 10);
            break;
        case 'x':
            *(--dpos) = hex_digits[abs((int)(remaining % 16))];
            remaining = (Py_ssize_t) (remaining / 16);
            break;
        default:
            assert(0);
            break;
        }
    } while (unlikely(remaining != 0));
    assert(!last_one_off || *dpos == '0');
    dpos += last_one_off;
    length = end - dpos;
    ulength = length;
    prepend_sign = 0;
    if (!is_unsigned && value <= neg_one) {
        if (padding_char == ' ' || width <= length + 1) {
            *(--dpos) = '-';
            ++length;
        } else {
            prepend_sign = 1;
        }
        ++ulength;
    }
    if (width > ulength) {
        ulength = width;
    }
    if (ulength == 1) {
        return PyUnicode_FromOrdinal(*dpos);
    }
    return __Pyx_PyUnicode_BuildFromAscii(ulength, dpos, (int) length, prepend_sign, padding_char);
}

/* CallNextTpDealloc */
static void __Pyx_call_next_tp_dealloc(PyObject* obj, destructor current_tp_dealloc) {
    PyTypeObject* type = Py_TYPE(obj);
    destructor tp_dealloc = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_dealloc, destructor) != current_tp_dealloc)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_dealloc = __Pyx_PyType_GetSlot(type, tp_dealloc, destructor)) == current_tp_dealloc)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type)
        tp_dealloc(obj);
}

/* CallNextTpTraverse */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    traverseproc tp_traverse = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc) != current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_traverse = __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc)) == current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_traverse)
        return tp_traverse(obj, v, a);
    return 0;
}

/* CallNextTpClear */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    inquiry tp_clear = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_clear, inquiry) != current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_clear = __Pyx_PyType_GetSlot(type, tp_clear, inquiry)) == current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_clear)
        tp_clear(obj);
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_0_12
#define __PYX_HAVE_RT_ImportType_3_0_12
static PyTypeObject *__Pyx_ImportType_3_0_12(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_0_12 check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_0_12 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_0_12 && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_n_s_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_n_s_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* ValidateBasesTuple */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_MACROS
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (n < 0) return -1;
#endif
    for (i = 1; i < n; i++)
    {
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        PyTypeObject *b;
#if PY_MAJOR_VERSION < 3
        if (PyClass_Check(b0))
        {
            PyErr_Format(PyExc_TypeError, "base class '%.200s' is an old-style class",
                         PyString_AS_STRING(((PyClassObject*)b0)->cl_name));
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !(CYTHON_USE_TYPE_SLOTS || CYTHON_COMPILING_IN_PYPY)
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !(CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API) || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if PY_VERSION_HEX >= 0x030700a1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400)
        gc = PyImport_GetModule(__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if PY_VERSION_HEX >= 0x03050000 && !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    __Pyx_TypeName type_name = __Pyx_PyType_GetName(tp);
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, attr_name);
#else
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%.400s'",
                 type_name, PyString_AS_STRING(attr_name));
#endif
    __Pyx_DECREF_TypeName(type_name);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_n_s_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_n_s_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* MergeVTables */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i;
    void** base_vtables;
    __Pyx_TypeName tp_base_name;
    __Pyx_TypeName base_name;
    void* unknown = (void*)-1;
    PyObject* bases = type->tp_bases;
    int base_depth = 0;
    {
        PyTypeObject* base = type->tp_base;
        while (base) {
            base_depth += 1;
            base = base->tp_base;
        }
    }
    base_vtables = (void**) malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
    for (i = 1; i < PyTuple_GET_SIZE(bases); i++) {
        void* base_vtable = __Pyx_GetVtable(((PyTypeObject*)PyTuple_GET_ITEM(bases, i)));
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = type->tp_base;
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = base->tp_base;
            }
        }
    }
    PyErr_Clear();
    free(base_vtables);
    return 0;
bad:
    tp_base_name = __Pyx_PyType_GetName(type->tp_base);
    base_name = __Pyx_PyType_GetName((PyTypeObject*)PyTuple_GET_ITEM(bases, i));
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
    free(base_vtables);
    return -1;
}
#endif

/* SetupReduce */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_n_s_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_n_s_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (unlikely(!py_import))
        goto bad;
    if (!from_list) {
        empty_list = PyList_New(0);
        if (unlikely(!empty_list))
            goto bad;
        from_list = empty_list;
    }
    #endif
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.') != NULL) {
                module = PyImport_ImportModuleLevelObject(
                    name, __pyx_d, empty_dict, from_list, 1);
                if (unlikely(!module)) {
                    if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (unlikely(!py_level))
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, __pyx_d, empty_dict, from_list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_d, empty_dict, from_list, level);
            #endif
        }
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    return module;
}

/* ImportDottedModule */
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
    if (likely(PyTuple_GET_SIZE(parts_tuple) == count)) {
        partial_name = name;
    } else {
        slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
#if PY_MAJOR_VERSION < 3
        PyExc_ImportError,
        "No module named '%s'", PyString_AS_STRING(partial_name));
#else
#if PY_VERSION_HEX >= 0x030600B1
        PyExc_ModuleNotFoundError,
#else
        PyExc_ImportError,
#endif
        "No module named '%U'", partial_name);
#endif
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
#endif
#if PY_MAJOR_VERSION >= 3
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple) {
    Py_ssize_t i, nparts;
    nparts = PyTuple_GET_SIZE(parts_tuple);
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = PySequence_ITEM(parts_tuple, i);
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (unlikely(!module)) {
        return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
    }
    return module;
}
#endif
static PyObject *__Pyx__ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if PY_MAJOR_VERSION < 3
    PyObject *module, *from_list, *star = __pyx_n_s__17;
    CYTHON_UNUSED_VAR(parts_tuple);
    from_list = PyList_New(1);
    if (unlikely(!from_list))
        return NULL;
    Py_INCREF(star);
    PyList_SET_ITEM(from_list, 0, star);
    module = __Pyx_Import(name, from_list, 0);
    Py_DECREF(from_list);
    return module;
#else
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    return __Pyx_ImportDottedModule_WalkParts(module, name, parts_tuple);
#endif
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030400B1
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_n_s_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_n_s_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_kp_u__18);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if PY_VERSION_HEX < 0x030700A1 || (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400)
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* ClassMethod */
static PyObject* __Pyx_Method_ClassMethod(PyObject *method) {
#if CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM <= 0x05080000
    if (PyObject_TypeCheck(method, &PyWrapperDescr_Type)) {
        return PyClassMethod_New(method);
    }
#else
#if CYTHON_COMPILING_IN_PYPY
    if (PyMethodDescr_Check(method))
#else
    #if PY_MAJOR_VERSION == 2
    static PyTypeObject *methoddescr_type = NULL;
    if (unlikely(methoddescr_type == NULL)) {
       PyObject *meth = PyObject_GetAttrString((PyObject*)&PyList_Type, "append");
       if (unlikely(!meth)) return NULL;
       methoddescr_type = Py_TYPE(meth);
       Py_DECREF(meth);
    }
    #else
    PyTypeObject *methoddescr_type = &PyMethodDescr_Type;
    #endif
    if (__Pyx_TypeCheck(method, methoddescr_type))
#endif
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject *)method;
        #if PY_VERSION_HEX < 0x03020000
        PyTypeObject *d_type = descr->d_type;
        #else
        PyTypeObject *d_type = descr->d_common.d_type;
        #endif
        return PyDescr_NewClassMethod(d_type, descr->d_method);
    }
#endif
    else if (PyMethod_Check(method)) {
        return PyClassMethod_New(PyMethod_GET_FUNCTION(method));
    }
    else {
        return PyClassMethod_New(method);
    }
}

/* GetNameInClass */
static PyObject *__Pyx__GetNameInClass(PyObject *nmspace, PyObject *name) {
    PyObject *result;
    PyObject *dict;
    assert(PyType_Check(nmspace));
#if CYTHON_USE_TYPE_SLOTS
    dict = ((PyTypeObject*)nmspace)->tp_dict;
    Py_XINCREF(dict);
#else
    dict = PyObject_GetAttr(nmspace, __pyx_n_s_dict);
#endif
    if (likely(dict)) {
        result = PyObject_GetItem(dict, name);
        Py_DECREF(dict);
        if (result) {
            return result;
        }
    }
    PyErr_Clear();
    __Pyx_GetModuleGlobalNameUncached(result, name);
    return result;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
#if !CYTHON_COMPILING_IN_LIMITED_API
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}
#endif

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result;
        result = PyObject_Call(replace, __pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    #if __PYX_LIMITED_VERSION_HEX < 0x030780000
    {
        PyObject *compiled = NULL, *result = NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "code", code))) return NULL;
        if (unlikely(PyDict_SetItemString(scratch_dict, "type", (PyObject*)(&PyType_Type)))) return NULL;
        compiled = Py_CompileString(
            "out = type(code)(\n"
            "  code.co_argcount, code.co_kwonlyargcount, code.co_nlocals, code.co_stacksize,\n"
            "  code.co_flags, code.co_code, code.co_consts, code.co_names,\n"
            "  code.co_varnames, code.co_filename, co_name, co_firstlineno,\n"
            "  code.co_lnotab)\n", "<dummy>", Py_file_input);
        if (!compiled) return NULL;
        result = PyEval_EvalCode(compiled, scratch_dict, scratch_dict);
        Py_DECREF(compiled);
        if (!result) PyErr_Print();
        Py_DECREF(result);
        result = PyDict_GetItemString(scratch_dict, "out");
        if (result) Py_INCREF(result);
        return result;
    }
    #else
    return NULL;
    #endif
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
    if (unlikely(!code_object)) goto bad;
    py_py_line = PyLong_FromLong(py_line);
    if (unlikely(!py_py_line)) goto bad;
    py_funcname = PyUnicode_FromString(funcname);
    if (unlikely(!py_funcname)) goto bad;
    dict = PyDict_New();
    if (unlikely(!dict)) goto bad;
    {
        PyObject *old_code_object = code_object;
        code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
        Py_DECREF(old_code_object);
    }
    if (unlikely(!code_object)) goto bad;
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_srcfile = NULL;
    py_srcfile = PyString_FromString(filename);
    if (!py_srcfile) goto bad;
    #endif
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        if (!py_funcname) goto bad;
        #endif
    }
    #if PY_MAJOR_VERSION < 3
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    #else
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    #endif
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_srcfile);
    #endif
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* Declarations */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = (float)(1.0) / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = (float)(1.0) / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
#if CYTHON_CCOMPLEX && (1) && (!0 || __cplusplus)
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = (double)(1.0) / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = (double)(1.0) / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if ((b.imag == 0) && (a.real >= 0)) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0.0, -1.0);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntFromPy */
static CYTHON_INLINE int64_t __Pyx_PyInt_As_int64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int64_t) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int64_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int64_t) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        int64_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int64_t) -1;
        val = __Pyx_PyInt_As_int64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 2 * PyLong_SHIFT)) {
                            return (int64_t) (((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 3 * PyLong_SHIFT)) {
                            return (int64_t) (((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 4 * PyLong_SHIFT)) {
                            return (int64_t) (((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int64_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int64_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int64_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) ((((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int64_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int64_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int64_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int64_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int64_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int64_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int64_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int64_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int64_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int64_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int64_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int64_t) 1) << (sizeof(int64_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int64_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int64_t");
    return (int64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int64_t");
    return (int64_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int64_t(int64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(int) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL;
        PyObject *py_bytes = NULL, *arg_tuple = NULL, *kwds = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        arg_tuple = PyTuple_Pack(2, py_bytes, order_str);
        if (!arg_tuple) goto limited_bad;
        if (!is_unsigned) {
            kwds = PyDict_New();
            if (!kwds) goto limited_bad;
            if (PyDict_SetItemString(kwds, "signed", __Pyx_NewRef(Py_True))) goto limited_bad;
        }
        result = PyObject_Call(from_bytes, arg_tuple, kwds);
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(arg_tuple);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API
static __Pyx_TypeName
__Pyx_PyType_GetName(PyTypeObject* tp)
{
    PyObject *name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_n_s_name);
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) {
        PyErr_Clear();
        Py_XDECREF(name);
        name = __Pyx_NewRef(__pyx_n_s__96);
    }
    return name;
}
#endif

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if ((sizeof(long) < sizeof(long))) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    }
#endif
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallMethod1 */
#if !(CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2)
static PyObject* __Pyx__PyObject_CallMethod1(PyObject* method, PyObject* arg) {
    PyObject *result = __Pyx_PyObject_CallOneArg(method, arg);
    Py_DECREF(method);
    return result;
}
#endif
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
#if CYTHON_VECTORCALL && __PYX_LIMITED_VERSION_HEX >= 0x030C00A2
    PyObject *args[2] = {obj, arg};
    (void) __Pyx_PyObject_GetMethod;
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_Call2Args;
    return PyObject_VectorcallMethod(method_name, args, 2 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_Call2Args(method, obj, arg);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) return NULL;
    return __Pyx__PyObject_CallMethod1(method, arg);
#endif
}

/* CoroutineBase */
#include <frameobject.h>
#if PY_VERSION_HEX >= 0x030b00a6 && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
static int __Pyx_PyGen__FetchStopIterationValue(PyThreadState *__pyx_tstate, PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    CYTHON_UNUSED_VAR(__pyx_tstate);
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (likely(__Pyx_IS_TYPE(ev, (PyTypeObject*)PyExc_StopIteration))) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!__Pyx_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!__Pyx_PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__Pyx_ExcInfoStruct *exc_state) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_CLEAR(exc_state->exc_value);
#else
    PyObject *t, *v, *tb;
    t = exc_state->exc_type;
    v = exc_state->exc_value;
    tb = exc_state->exc_traceback;
    exc_state->exc_type = NULL;
    exc_state->exc_value = NULL;
    exc_state->exc_traceback = NULL;
    Py_XDECREF(t);
    Py_XDECREF(v);
    Py_XDECREF(tb);
#endif
}
#define __Pyx_Coroutine_AlreadyRunningError(gen)  (__Pyx__Coroutine_AlreadyRunningError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyRunningError(__pyx_CoroutineObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check((PyObject*)gen)) {
        msg = "coroutine already executing";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact((PyObject*)gen)) {
        msg = "async generator already executing";
    #endif
    } else {
        msg = "generator already executing";
    }
    PyErr_SetString(PyExc_ValueError, msg);
}
#define __Pyx_Coroutine_NotStartedError(gen)  (__Pyx__Coroutine_NotStartedError(gen), (PyObject*)NULL)
static void __Pyx__Coroutine_NotStartedError(PyObject *gen) {
    const char *msg;
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if ((0)) {
    #ifdef __Pyx_Coroutine_USED
    } else if (__Pyx_Coroutine_Check(gen)) {
        msg = "can't send non-None value to a just-started coroutine";
    #endif
    #ifdef __Pyx_AsyncGen_USED
    } else if (__Pyx_AsyncGen_CheckExact(gen)) {
        msg = "can't send non-None value to a just-started async generator";
    #endif
    } else {
        msg = "can't send non-None value to a just-started generator";
    }
    PyErr_SetString(PyExc_TypeError, msg);
}
#define __Pyx_Coroutine_AlreadyTerminatedError(gen, value, closing)  (__Pyx__Coroutine_AlreadyTerminatedError(gen, value, closing), (PyObject*)NULL)
static void __Pyx__Coroutine_AlreadyTerminatedError(PyObject *gen, PyObject *value, int closing) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    CYTHON_MAYBE_UNUSED_VAR(closing);
    #ifdef __Pyx_Coroutine_USED
    if (!closing && __Pyx_Coroutine_Check(gen)) {
        PyErr_SetString(PyExc_RuntimeError, "cannot reuse already awaited coroutine");
    } else
    #endif
    if (value) {
        #ifdef __Pyx_AsyncGen_USED
        if (__Pyx_AsyncGen_CheckExact(gen))
            PyErr_SetNone(__Pyx_PyExc_StopAsyncIteration);
        else
        #endif
        PyErr_SetNone(PyExc_StopIteration);
    }
}
static
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value, int closing) {
    __Pyx_PyThreadState_declare
    PyThreadState *tstate;
    __Pyx_ExcInfoStruct *exc_state;
    PyObject *retval;
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            return __Pyx_Coroutine_NotStartedError((PyObject*)self);
        }
    }
    if (unlikely(self->resume_label == -1)) {
        return __Pyx_Coroutine_AlreadyTerminatedError((PyObject*)self, value, closing);
    }
#if CYTHON_FAST_THREAD_STATE
    __Pyx_PyThreadState_assign
    tstate = __pyx_tstate;
#else
    tstate = __Pyx_PyThreadState_Current;
#endif
    exc_state = &self->gi_exc_state;
    if (exc_state->exc_value) {
        #if CYTHON_COMPILING_IN_PYPY
        #else
        PyObject *exc_tb;
        #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
        exc_tb = PyException_GetTraceback(exc_state->exc_value);
        #elif PY_VERSION_HEX >= 0x030B00a4
        exc_tb = ((PyBaseExceptionObject*) exc_state->exc_value)->traceback;
        #else
        exc_tb = exc_state->exc_traceback;
        #endif
        if (exc_tb) {
            PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
            PyFrameObject *f = tb->tb_frame;
            assert(f->f_back == NULL);
            #if PY_VERSION_HEX >= 0x030B00A1
            f->f_back = PyThreadState_GetFrame(tstate);
            #else
            Py_XINCREF(tstate->frame);
            f->f_back = tstate->frame;
            #endif
            #if PY_VERSION_HEX >= 0x030B00a4 && !CYTHON_COMPILING_IN_CPYTHON
            Py_DECREF(exc_tb);
            #endif
        }
        #endif
    }
#if CYTHON_USE_EXC_INFO_STACK
    exc_state->previous_item = tstate->exc_info;
    tstate->exc_info = exc_state;
#else
    if (exc_state->exc_type) {
        __Pyx_ExceptionSwap(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(exc_state);
        __Pyx_ExceptionSave(&exc_state->exc_type, &exc_state->exc_value, &exc_state->exc_traceback);
    }
#endif
    self->is_running = 1;
    retval = self->body(self, tstate, value);
    self->is_running = 0;
#if CYTHON_USE_EXC_INFO_STACK
    exc_state = &self->gi_exc_state;
    tstate->exc_info = exc_state->previous_item;
    exc_state->previous_item = NULL;
    __Pyx_Coroutine_ResetFrameBackpointer(exc_state);
#endif
    return retval;
}
static CYTHON_INLINE void __Pyx_Coroutine_ResetFrameBackpointer(__Pyx_ExcInfoStruct *exc_state) {
#if CYTHON_COMPILING_IN_PYPY
    CYTHON_UNUSED_VAR(exc_state);
#else
    PyObject *exc_tb;
    #if PY_VERSION_HEX >= 0x030B00a4
    if (!exc_state->exc_value) return;
    exc_tb = PyException_GetTraceback(exc_state->exc_value);
    #else
    exc_tb = exc_state->exc_traceback;
    #endif
    if (likely(exc_tb)) {
        PyTracebackObject *tb = (PyTracebackObject *) exc_tb;
        PyFrameObject *f = tb->tb_frame;
        Py_CLEAR(f->f_back);
        #if PY_VERSION_HEX >= 0x030B00a4
        Py_DECREF(exc_tb);
        #endif
    }
#endif
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(PyObject* gen, PyObject *retval) {
    CYTHON_MAYBE_UNUSED_VAR(gen);
    if (unlikely(!retval)) {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        if (!__Pyx_PyErr_Occurred()) {
            PyObject *exc = PyExc_StopIteration;
            #ifdef __Pyx_AsyncGen_USED
            if (__Pyx_AsyncGen_CheckExact(gen))
                exc = __Pyx_PyExc_StopAsyncIteration;
            #endif
            __Pyx_PyErr_SetNone(exc);
        }
    }
    return retval;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
static CYTHON_INLINE
PyObject *__Pyx_PyGen_Send(PyGenObject *gen, PyObject *arg) {
#if PY_VERSION_HEX <= 0x030A00A1
    return _PyGen_Send(gen, arg);
#else
    PyObject *result;
    if (PyIter_Send((PyObject*)gen, arg ? arg : Py_None, &result) == PYGEN_RETURN) {
        if (PyAsyncGen_CheckExact(gen)) {
            assert(result == Py_None);
            PyErr_SetNone(PyExc_StopAsyncIteration);
        }
        else if (result == Py_None) {
            PyErr_SetNone(PyExc_StopIteration);
        }
        else {
#if PY_VERSION_HEX < 0x030d00A1
            _PyGen_SetStopIterationValue(result);
#else
            if (!PyTuple_Check(result) && !PyExceptionInstance_Check(result)) {
                PyErr_SetObject(PyExc_StopIteration, result);
            } else {
                PyObject *exc = __Pyx_PyObject_CallOneArg(PyExc_StopIteration, result);
                if (likely(exc != NULL)) {
                    PyErr_SetObject(PyExc_StopIteration, exc);
                    Py_DECREF(exc);
                }
            }
#endif
        }
        Py_DECREF(result);
        result = NULL;
    }
    return result;
#endif
}
#endif
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen__FetchStopIterationValue(__Pyx_PyThreadState_Current, &val);
    ret = __Pyx_Coroutine_SendEx(gen, val, 0);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_AsyncGen_USED
        if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
            ret = __Pyx_async_gen_asend_send(yf, value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03050000 && defined(PyCoro_CheckExact) && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyCoro_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, value == Py_None ? NULL : value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value, 0);
    }
    return __Pyx_Coroutine_MethodReturn(self, retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_Check(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    if (__Pyx_CoroutineAwait_CheckExact(yf)) {
        retval = __Pyx_CoroutineAwait_Close((__pyx_CoroutineAwaitObject*)yf, NULL);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_AsyncGen_USED
    if (__pyx_PyAsyncGenASend_CheckExact(yf)) {
        retval = __Pyx_async_gen_asend_close(yf, NULL);
    } else
    if (__pyx_PyAsyncGenAThrow_CheckExact(yf)) {
        retval = __Pyx_async_gen_athrow_close(yf, NULL);
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (unlikely(PyErr_Occurred())) {
                PyErr_WriteUnraisable(yf);
            }
        } else {
            retval = __Pyx_PyObject_CallNoArg(meth);
            Py_DECREF(meth);
            if (unlikely(!retval))
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03030000 && (defined(__linux__) || PY_VERSION_HEX >= 0x030600B3)
        if (PyGen_CheckExact(yf)) {
            ret = __Pyx_PyGen_Send((PyGenObject*)yf, NULL);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_Check(yf)) {
            ret = __Pyx_Coroutine_Send(yf, Py_None);
        } else
        #endif
            ret = __Pyx_PyObject_GetIterNextFunc(yf)(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None, 0);
}
static PyObject *__Pyx_Coroutine_Close_Method(PyObject *self, PyObject *arg) {
    CYTHON_UNUSED_VAR(arg);
    return __Pyx_Coroutine_Close(self);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL, 1);
    if (unlikely(retval)) {
        const char *msg;
        Py_DECREF(retval);
        if ((0)) {
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_Coroutine_Check(self)) {
            msg = "coroutine ignored GeneratorExit";
        #endif
        #ifdef __Pyx_AsyncGen_USED
        } else if (__Pyx_AsyncGen_CheckExact(self)) {
#if PY_VERSION_HEX < 0x03060000
            msg = "async generator ignored GeneratorExit - might require Python 3.6+ finalisation (PEP 525)";
#else
            msg = "async generator ignored GeneratorExit";
#endif
        #endif
        } else {
            msg = "generator ignored GeneratorExit";
        }
        PyErr_SetString(PyExc_RuntimeError, msg);
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (likely(!raised_exception || __Pyx_PyErr_GivenExceptionMatches2(raised_exception, PyExc_GeneratorExit, PyExc_StopIteration))) {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx__Coroutine_Throw(PyObject *self, PyObject *typ, PyObject *val, PyObject *tb,
                                        PyObject *args, int close_on_genexit) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(gen->is_running))
        return __Pyx_Coroutine_AlreadyRunningError(gen);
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (__Pyx_PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit) && close_on_genexit) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
            goto throw_here;
        }
        gen->is_running = 1;
        if (0
        #ifdef __Pyx_Generator_USED
            || __Pyx_Generator_CheckExact(yf)
        #endif
        #ifdef __Pyx_Coroutine_USED
            || __Pyx_Coroutine_Check(yf)
        #endif
            ) {
            ret = __Pyx__Coroutine_Throw(yf, typ, val, tb, args, close_on_genexit);
        #ifdef __Pyx_Coroutine_USED
        } else if (__Pyx_CoroutineAwait_CheckExact(yf)) {
            ret = __Pyx__Coroutine_Throw(((__pyx_CoroutineAwaitObject*)yf)->coroutine, typ, val, tb, args, close_on_genexit);
        #endif
        } else {
            PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (unlikely(PyErr_Occurred())) {
                    gen->is_running = 0;
                    return NULL;
                }
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            if (likely(args)) {
                ret = __Pyx_PyObject_Call(meth, args, NULL);
            } else {
                PyObject *cargs[4] = {NULL, typ, val, tb};
                ret = __Pyx_PyObject_FastCall(meth, cargs+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
            }
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(self, ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(self, __Pyx_Coroutine_SendEx(gen, NULL, 0));
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    PyObject *typ;
    PyObject *val = NULL;
    PyObject *tb = NULL;
    if (unlikely(!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb)))
        return NULL;
    return __Pyx__Coroutine_Throw(self, typ, val, tb, args, 1);
}
static CYTHON_INLINE int __Pyx_Coroutine_traverse_excstate(__Pyx_ExcInfoStruct *exc_state, visitproc visit, void *arg) {
#if PY_VERSION_HEX >= 0x030B00a4
    Py_VISIT(exc_state->exc_value);
#else
    Py_VISIT(exc_state->exc_type);
    Py_VISIT(exc_state->exc_value);
    Py_VISIT(exc_state->exc_traceback);
#endif
    return 0;
}
static int __Pyx_Coroutine_traverse(__pyx_CoroutineObject *gen, visitproc visit, void *arg) {
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    return __Pyx_Coroutine_traverse_excstate(&gen->gi_exc_state, visit, arg);
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    __Pyx_Coroutine_ExceptionClear(&gen->gi_exc_state);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        Py_CLEAR(((__pyx_PyAsyncGenObject*)gen)->ag_finalizer);
    }
#endif
    Py_CLEAR(gen->gi_code);
    Py_CLEAR(gen->gi_frame);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    Py_CLEAR(gen->gi_modulename);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label >= 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1 && CYTHON_USE_TP_FINALIZE
        if (unlikely(PyObject_CallFinalizerFromDealloc(self)))
#else
        Py_TYPE(gen)->tp_del(self);
        if (unlikely(Py_REFCNT(self) > 0))
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        /* We have to handle this case for asynchronous generators
           right here, because this code has to be between UNTRACK
           and GC_Del. */
        Py_CLEAR(((__pyx_PyAsyncGenObject*)self)->ag_finalizer);
    }
#endif
    __Pyx_Coroutine_clear(self);
    __Pyx_PyHeapTypeObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label < 0) {
        return;
    }
#if !CYTHON_USE_TP_FINALIZE
    assert(self->ob_refcnt == 0);
    __Pyx_SET_REFCNT(self, 1);
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
#ifdef __Pyx_AsyncGen_USED
    if (__Pyx_AsyncGen_CheckExact(self)) {
        __pyx_PyAsyncGenObject *agen = (__pyx_PyAsyncGenObject*)self;
        PyObject *finalizer = agen->ag_finalizer;
        if (finalizer && !agen->ag_closed) {
            PyObject *res = __Pyx_PyObject_CallOneArg(finalizer, self);
            if (unlikely(!res)) {
                PyErr_WriteUnraisable(self);
            } else {
                Py_DECREF(res);
            }
            __Pyx_ErrRestore(error_type, error_value, error_traceback);
            return;
        }
    }
#endif
    if (unlikely(gen->resume_label == 0 && !error_value)) {
#ifdef __Pyx_Coroutine_USED
#ifdef __Pyx_Generator_USED
    if (!__Pyx_Generator_CheckExact(self))
#endif
        {
        PyObject_GC_UnTrack(self);
#if PY_MAJOR_VERSION >= 3  || defined(PyErr_WarnFormat)
        if (unlikely(PyErr_WarnFormat(PyExc_RuntimeWarning, 1, "coroutine '%.50S' was never awaited", gen->gi_qualname) < 0))
            PyErr_WriteUnraisable(self);
#else
        {PyObject *msg;
        char *cmsg;
        #if CYTHON_COMPILING_IN_PYPY
        msg = NULL;
        cmsg = (char*) "coroutine was never awaited";
        #else
        char *cname;
        PyObject *qualname;
        qualname = gen->gi_qualname;
        cname = PyString_AS_STRING(qualname);
        msg = PyString_FromFormat("coroutine '%.50s' was never awaited", cname);
        if (unlikely(!msg)) {
            PyErr_Clear();
            cmsg = (char*) "coroutine was never awaited";
        } else {
            cmsg = PyString_AS_STRING(msg);
        }
        #endif
        if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, cmsg, 1) < 0))
            PyErr_WriteUnraisable(self);
        Py_XDECREF(msg);}
#endif
        PyObject_GC_Track(self);
        }
#endif
    } else {
        PyObject *res = __Pyx_Coroutine_Close(self);
        if (unlikely(!res)) {
            if (PyErr_Occurred())
                PyErr_WriteUnraisable(self);
        } else {
            Py_DECREF(res);
        }
    }
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if !CYTHON_USE_TP_FINALIZE
    assert(Py_REFCNT(self) > 0);
    if (likely(--self->ob_refcnt == 0)) {
        return;
    }
    {
        Py_ssize_t refcnt = Py_REFCNT(self);
        _Py_NewReference(self);
        __Pyx_SET_REFCNT(self, refcnt);
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(Py_TYPE(self)) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_name;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_name, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self, void *context)
{
    PyObject *name = self->gi_qualname;
    CYTHON_UNUSED_VAR(context);
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value)))
#else
    if (unlikely(value == NULL || !PyString_Check(value)))
#endif
    {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_Py_XDECREF_SET(self->gi_qualname, value);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_frame(__pyx_CoroutineObject *self, void *context)
{
    PyObject *frame = self->gi_frame;
    CYTHON_UNUSED_VAR(context);
    if (!frame) {
        if (unlikely(!self->gi_code)) {
            Py_RETURN_NONE;
        }
        frame = (PyObject *) PyFrame_New(
            PyThreadState_Get(),            /*PyThreadState *tstate,*/
            (PyCodeObject*) self->gi_code,  /*PyCodeObject *code,*/
            __pyx_d,                 /*PyObject *globals,*/
            0                               /*PyObject *locals*/
        );
        if (unlikely(!frame))
            return NULL;
        self->gi_frame = frame;
    }
    Py_INCREF(frame);
    return frame;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (unlikely(!gen))
        return NULL;
    return __Pyx__Coroutine_NewInit(gen, body, code, closure, name, qualname, module_name);
}
static __pyx_CoroutineObject *__Pyx__Coroutine_NewInit(
            __pyx_CoroutineObject *gen, __pyx_coroutine_body_t body, PyObject *code, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    #if PY_VERSION_HEX >= 0x030B00a4
    gen->gi_exc_state.exc_value = NULL;
    #else
    gen->gi_exc_state.exc_type = NULL;
    gen->gi_exc_state.exc_value = NULL;
    gen->gi_exc_state.exc_traceback = NULL;
    #endif
#if CYTHON_USE_EXC_INFO_STACK
    gen->gi_exc_state.previous_item = NULL;
#endif
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    Py_XINCREF(code);
    gen->gi_code = code;
    gen->gi_frame = NULL;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
#ifndef CYTHON_REGISTER_ABCS
#define CYTHON_REGISTER_ABCS 1
#endif
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (CYTHON_REGISTER_ABCS && !abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_MAJOR_VERSION >= 3) ? "collections.abc" : "collections");
        if (unlikely(!module)) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_MAJOR_VERSION >= 3) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if ((0)) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close_Method, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {(char*) "gi_code", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_code), READONLY, NULL},
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CoroutineObject, gi_modulename), 0, 0},
#if CYTHON_USE_TYPE_SPECS
    {(char *) "__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CoroutineObject, gi_weakreflist), READONLY, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {(char *) "gi_frame", (getter)__Pyx_Coroutine_get_frame, NULL,
     (char*) PyDoc_STR("Frame of the generator"), 0},
    {0, 0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_GeneratorType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_Coroutine_dealloc},
    {Py_tp_traverse, (void *)__Pyx_Coroutine_traverse},
    {Py_tp_iter, (void *)PyObject_SelfIter},
    {Py_tp_iternext, (void *)__Pyx_Generator_Next},
    {Py_tp_methods, (void *)__pyx_Generator_methods},
    {Py_tp_members, (void *)__pyx_Generator_memberlist},
    {Py_tp_getset, (void *)__pyx_Generator_getsets},
    {Py_tp_getattro, (void *) __Pyx_PyObject_GenericGetAttrNoDict},
#if CYTHON_USE_TP_FINALIZE
    {Py_tp_finalize, (void *)__Pyx_Coroutine_del},
#endif
    {0, 0},
};
static PyType_Spec __pyx_GeneratorType_spec = {
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    __pyx_GeneratorType_slots
};
#else
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    __PYX_TYPE_MODULE_PREFIX "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if CYTHON_USE_TP_FINALIZE
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if CYTHON_USE_TP_FINALIZE
    __Pyx_Coroutine_del,
#elif PY_VERSION_HEX >= 0x030400a1
    0,
#endif
#if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
    0,
#endif
#if __PYX_NEED_TP_PRINT_SLOT
    0,
#endif
#if PY_VERSION_HEX >= 0x030C0000
    0,
#endif
#if PY_VERSION_HEX >= 0x030d00A4
    0,
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
    0,
#endif
};
#endif
static int __pyx_Generator_init(PyObject *module) {
#if CYTHON_USE_TYPE_SPECS
    __pyx_GeneratorType = __Pyx_FetchCommonTypeFromSpec(module, &__pyx_GeneratorType_spec, NULL);
#else
    CYTHON_UNUSED_VAR(module);
    __pyx_GeneratorType_type.tp_getattro = __Pyx_PyObject_GenericGetAttrNoDict;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
#endif
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030B00A4
    return Py_Version & ~0xFFUL;
#else
    const char* rt_version = Py_GetVersion();
    unsigned long version = 0;
    unsigned long factor = 0x01000000UL;
    unsigned int digit = 0;
    int i = 0;
    while (factor) {
        while ('0' <= rt_version[i] && rt_version[i] <= '9') {
            digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
            ++i;
        }
        version += factor * digit;
        if (rt_version[i] != '.')
            break;
        digit = 0;
        factor >>= 8;
        ++i;
    }
    return version;
#endif
}
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* InitStrings */
#if PY_MAJOR_VERSION >= 3
static int __Pyx_InitString(__Pyx_StringTabEntry t, PyObject **str) {
    if (t.is_unicode | t.is_str) {
        if (t.intern) {
            *str = PyUnicode_InternFromString(t.s);
        } else if (t.encoding) {
            *str = PyUnicode_Decode(t.s, t.n - 1, t.encoding, NULL);
        } else {
            *str = PyUnicode_FromStringAndSize(t.s, t.n - 1);
        }
    } else {
        *str = PyBytes_FromStringAndSize(t.s, t.n - 1);
    }
    if (!*str)
        return -1;
    if (PyObject_Hash(*str) == -1)
        return -1;
    return 0;
}
#endif
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION >= 3
        __Pyx_InitString(*t, t->p);
        #else
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        #endif
        ++t;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY && !CYTHON_COMPILING_IN_LIMITED_API) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetName(Py_TYPE(result));
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type " __Pyx_FMT_TYPENAME ")",
                 type_name, type_name, result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
#if PY_MAJOR_VERSION < 3
  } else if (likely(PyInt_CheckExact(o))) {
    return PyInt_AS_LONG(o);
#endif
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyInt_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
